 /* Seamless3d

Copyright (C) 2001-2010 Graham John Perrett (AKA thyme) thyme@seamless3d.com www.seamless3d.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this
software and associated documentation files (the "Software"), to deal in the Software
without restriction, including without limitation the rights to use, copy, modify,
merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to the following
conditions:

The above copyright notice and this permission notice shall be included in all copies or
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT
OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.
*/


#include <windows.h>
#include <direct.h>

#include "seamless3d.h"
#include "techuelife_triangle_editor.h"
#include "gui.h"
#include "scene_tree.h"
#include "resource.h"
#include "input.h"
#include "robot_aid.h"
#include "seamless3d_gui_render.h"

bool isDoubleLeftClick();

void paintBlock2(const unsigned fromColor, const unsigned toColor, int tx, int ty, int bx, int by, int dstWidth, int dstHeight, int* dstPix);
ProtoDeclare* getSceneProtoDeclare(int selLine);
void commonDoDrag(const Vec2i ptCursor, Vec2i& pos, Vec2i& dragPos);
void exitFileWindow();

inline bool FieldButton::checkClickTextCell(Vec2i ptCursor, int& newCaretPos)const{
	if(dynamic_cast<const SFBoolButton*>(this)==0){
		const int lessThanValPosAllowed = 5;
		if(valPos!=Vec2i()){
			if(ptCursor.x >= valPos.x-lessThanValPosAllowed && ptCursor.x < valPos.x + valSize.x && ptCursor.y >= valPos.y && ptCursor.y < valPos.y+title_bar_height){
				if(ptCursor.x < valPos.x){
					ptCursor.x=valPos.x;
				}
				newCaretPos=(ptCursor.x + charWidth/2 - valPos.x)/charWidth;
				return true;
			}
		}
	}
	return false;
}
void ControlPanel::buttonHover(Vec2i ptCursor){
/*
	for(int k=g_propWinLine.length; --k>=0;){
		Button* button=g_propWinLine[k].button;
		if(button && button->hover(ptCursor)){
			return;
		}
	}
*/
	for(int k=buttonList.length; --k>=0;){
		Button* button=buttonList[k];
		if(button && button->initHover(ptCursor)){
			return;
		}
	}
	if(exitWindow.initHover(ptCursor)){
		return;
	}
	if(OpenNetFile* openNetFile = dynamic_cast<OpenNetFile*>(node.p)){
		openNetFile->hoverUrls(ptCursor);
	}
}
inline void ControlPanel::setUpIndexControl(Field& field, Node& node, const int fieldIndex, const Properties& properties, const int& index, const int& length, int& yPos){
	int groupLength=1;
	for(int i=fieldIndex; i < node.getFLen() && properties.groupIndex==node.getProperties()[i].groupIndex; i++){
		groupLength++;
	}
	const int valPosIdxX=colum-24+18*1+4;
	buttonList+=new IndexButton(Vec2i(18, 24), Vec2i(colum-24, yPos), Vec2i(0, 0), "select element", *this, node, field, properties, Vec2i(valPosIdxX, yPos), Vec2i(size.x-valPosIdxX, 16));
	const int valPosLenX=240+18+14;
	buttonList+=new LengthButton(Vec2i(18, 24), Vec2i(240+10, yPos), Vec2i(0, 0), "number of elements", *this, node, field, properties, fieldIndex, groupLength, Vec2i(valPosLenX, yPos), Vec2i(size.x-valPosLenX, 16));
}
void ControlPanel::setUpControlPanel(Node* node, const M_Properties &properties){
	this->node=node;

	g_controlPanelOpen=true;
	int currentGroup=-1;
	g_propWinLine.length=0;
	buttonList.length=0;
	int yPos=startFieldLineYpos;
	int lastYPos=yPos;
	int i=0;

	if(Art* part=dynamic_cast<Art*>(node)){
		int index = 0;
		int xPos=colum;
		//scale
		FieldButton* button=part->scale.setUpControl(*node, part->scale, properties[0], 0, xPos, yPos, *this);
		buttonList+=button;
		g_propWinLine+=PropWinLine(yPos, button);
		yPos+=25;

		button=new ArtRotationButton(this, part, part->rotation, properties[1]);
		buttonList+=button;
		g_propWinLine+=PropWinLine(yPos, button);
		yPos+=25;

		button = new ArtTranslationButton(this, part, part->translation, properties[2]);
		buttonList += button;
		g_propWinLine += PropWinLine(yPos, button);
		yPos += 25;

		//track
		//button=part->track.setUpControl(*node, part->track, properties[3], 0, 60, yPos, *this);
		//buttonList+=button;
		//g_propWinLine+=PropWinLine(yPos, button);

		//mirror
		button=part->mirror.setUpControl(*node, part->mirror, properties[4], 0, 80, yPos, *this);
		buttonList+=button;
		g_propWinLine+=PropWinLine(yPos, button);

		//ik
		button = new ArtIKButton(this, part, part->ik, properties[5]);
		//button=part->ik.setUpControl(*node, part->ik, properties[4], 0, 150, yPos, *this);
		buttonList+=button;
		g_propWinLine+=PropWinLine(yPos, button);

		//pin
		button=part->pin.setUpControl(*node, part->pin, properties[6], 0, 240, yPos, *this);
		buttonList+=button;
		g_propWinLine+=PropWinLine(yPos, button);
		yPos+=25;

		if(part->ik){
			button = new ArtIKRotButton(this, part, part->ikRot, properties[7]);
			buttonList += button;
			g_propWinLine += PropWinLine(yPos, button);
			yPos += 25;

			button = new ArtIKPosButton(this, part, part->ikPos, properties[8]);
			buttonList += button;
			g_propWinLine += PropWinLine(yPos, button);
			yPos += 25;


			//fka
			//button=part->fka.setUpControl(*node, part->fka, properties[8], 0, 60, yPos, *this);
			button = new ArtFkaButton(this, part, part->fka, properties[9]);
			buttonList+=button;
			g_propWinLine+=PropWinLine(yPos, button);
			//fkb
			button = new ArtFkbButton(this, part, part->fkb, properties[10]);
			buttonList += button;
			g_propWinLine += PropWinLine(yPos, button);
			yPos += 25;
		}

		const Properties& sprop=properties[11];
		setUpIndexControl(part->coord, *part, 11, sprop, index, part->coord.size(), yPos);
		yPos+=25;
		currentGroup = sprop.groupIndex;

		button = new ArtCoordPointButton(this, part, part->coord, yPos, sprop);
		buttonList += button;
		g_propWinLine += PropWinLine(yPos, button);
		yPos += 25;

		i = 12;
		goto here;
	}
	for(; i<node->getFLen();i++){
here:
		const Properties& sprop=properties[i];
		if(sprop.disable()){
			continue;
		}
		Field* field=node->getField(i);
		if((sprop.isWritable()^sprop.isReadable())==false&&field->getType()!=MFNODE&&field->getType()!=SFNODE){
			int index=0;
			if(sprop.groupIndex!=-1){
				index=node->getFieldGroupIndex(sprop.groupIndex);
				if(field->size()&&index>=(field->size()+sprop.startAtIndex)){
					index=field->size()-1+sprop.startAtIndex;
					node->setFieldGroupIndex(sprop.groupIndex,index);
				}
				if(currentGroup!=sprop.groupIndex){
					setUpIndexControl(*field, *node, i, sprop, index, field->size(), yPos);
					yPos+=25;
					currentGroup=sprop.groupIndex;
				}
			}
			lastYPos=yPos;

			int xPos=colum;
			if(sprop.name.length>14){
				xPos=sprop.name.length*8+12;
			}
			FieldButton* button=field->setUpControl(*node, *field, sprop, index-sprop.startAtIndex, xPos, yPos, *this);
			buttonList+=button;
			g_propWinLine+=PropWinLine(yPos, button);
			yPos+=25;
		}
		else{
			g_propWinLine+=PropWinLine(lastYPos);
		}
	}
	if(node->getType()!=PROTOINTERFACE){
		for(int i=0; i<node->getFLen();i++){
			const M_Route& route=node->getField(i)->getRoute();
			for (int r=0; r<route.length; r++){
				g_propWinLine+=PropWinLine(yPos);
				yPos+=pixSymbolCY;
			}
		}
	}
	int height=yPos+10;
	if(dynamic_cast<TextureMapper*>(node)){
		height += 35;
	}
	else if(dynamic_cast<Demo*>(node)){
		//operation;
		int xPos=82;
		int valXPos=xPos+18*1+4;
		FieldButton* button=buttonList[0];
		button->dstPos=Vec2i(xPos,26);
		button->valPos=Vec2i(valXPos,26);
		//play;
		xPos=160;
		valXPos=xPos+18*1+4;
		button=buttonList[1];
		button->dstPos=Vec2i(xPos,26);
		button->valPos=Vec2i(valXPos,26);
		//stop;
		xPos=250;
		valXPos=xPos+18*1+4;
		button=buttonList[2];
		button->dstPos=Vec2i(xPos,26);
		button->valPos=Vec2i(valXPos,26);
		//express;
		xPos=400;
		valXPos=xPos+18*1+4;
		button=buttonList[3];
		button->dstPos=Vec2i(xPos,26+8);
		button->valPos=Vec2i(valXPos,26);
		//status;
		xPos=60;
		valXPos=xPos+4;
		button=buttonList[4];
		button->dstPos=Vec2i(xPos,26+25);
		button->valPos=Vec2i(valXPos,26+25);
		height=startFieldLineYpos+25*2+10;
	}
	else if(dynamic_cast<FuseSurface*>(node)){
		FieldButton* button=buttonList[7]; //numOf
		int yPos = button->valPos.y;
		button->dstPos=Vec2i(60,yPos);
		button->valPos=Vec2i(60+22,yPos);

		button=buttonList[8]; //Qty
		button->dstPos=Vec2i(200,yPos);
		button->valPos=Vec2i(200+22,yPos);
/*
		yPos+=25;
		button=buttonList[9]; //orientation
		button->dstPos.y=yPos;
		button->valPos.y=yPos;
*/
		yPos+=25;
		button=buttonList[9]; //bond
		button->dstPos=Vec2i(60,yPos+8);
		button->valPos=Vec2i(60+22,yPos);

		button=buttonList[10]; //coordChange
		button->dstPos=Vec2i(200,yPos);
		button->valPos=Vec2i(200+22,yPos);

		yPos+=25;

		button=buttonList[11]; //spread
		button->dstPos=Vec2i(60,yPos);
		button->valPos=Vec2i(60+22,yPos);

		button=buttonList[12]; //spreadRes
		button->dstPos=Vec2i(200,yPos);
		button->valPos=Vec2i(200+22,yPos);

		yPos+=25;
		button=buttonList[13]; //polyMapAid
		button->dstPos.y=yPos+8;
		button->valPos.y=yPos;

		yPos+=25;
		button=buttonList[14]; //srcMatch Index
		button->dstPos.y=yPos;
		button->valPos.y=yPos;

		button=buttonList[15]; //dstMatch Length
		button->dstPos.y=yPos;
		button->valPos.y=yPos;

		yPos+=25;
		button=buttonList[16]; //srcMatch
		button->dstPos=Vec2i(80,yPos);
		button->valPos=Vec2i(80+22,yPos);

		button=buttonList[17]; //dstMatch
		button->dstPos=Vec2i(230,yPos);
		button->valPos=Vec2i(230+22,yPos);

		yPos+=25;
		button=buttonList[18]; //in
		button->dstPos=Vec2i(60,yPos+8);
		button->valPos=Vec2i(60+22,yPos);

		button=buttonList[19]; //range
		button->dstPos=Vec2i(200,yPos);
		button->valPos=Vec2i(200+22,yPos);

		yPos+=25;
		button=buttonList[20]; //snap
		button->dstPos.y=yPos+8;
		button->valPos.y=yPos;

		yPos+=25;
		button=buttonList[21]; //evolve
		button->dstPos=Vec2i(60,yPos+8);
		button->valPos=Vec2i(60+22,yPos);

		button=buttonList[22]; //pointChanhe
		button->dstPos=Vec2i(200,yPos);
		button->valPos=Vec2i(200+22,yPos);

		height=startFieldLineYpos+25*15+10;
	}
	/*
	else if(dynamic_cast<FileAlreadyExists*>(node)){
		FieldButton* button=buttonList[1]; //yes
		int yPos = button->valPos.y;
		button->dstPos=Vec2i(100,yPos);
		button->valPos=Vec2i(100+22,yPos);

		button=buttonList[2]; //no
		button->dstPos=Vec2i(150,yPos);
		button->valPos=Vec2i(150+22,yPos);

		height=startFieldLineYpos+25*2+10;
	}
	*/
	else if(dynamic_cast<QuestionMessage3Button*>(node)){
		FieldButton* button=buttonList[1]; //yes
		int yPos = button->valPos.y;
		button->dstPos=Vec2i(100,yPos);
		button->valPos=Vec2i(100+22,yPos);

		button=buttonList[2]; //no
		button->dstPos=Vec2i(150,yPos);
		button->valPos=Vec2i(150+22,yPos);

		button=buttonList[3]; //cancel
		button->dstPos=Vec2i(200,yPos);
		button->valPos=Vec2i(200+22,yPos);

		height=startFieldLineYpos+25*2+10;
	}
	else if(dynamic_cast<Stem*>(node)){
		int first = 90;
		int second = first + 100;
		int third = first + 110 * 2;

		FieldButton* button=buttonList[0]; //add
		int valPosY = button->valPos.y;
		int dstPosY = button->dstPos.y;
		button->dstPos=Vec2i(first, dstPosY);
		button->valPos=Vec2i(first + 22, valPosY);

		button=buttonList[1]; //mirror
		button->dstPos=Vec2i(second, dstPosY);
		button->valPos=Vec2i(second + 22, valPosY);

		button=buttonList[2]; //store
		button->dstPos=Vec2i(third, dstPosY);
		button->valPos=Vec2i(third + 22, valPosY);

		button=buttonList[3]; //slide
		button->dstPos.y -= 25 * 2;
		button->valPos.y -= 25 * 2;

		//button=buttonList[4]; //orientation
		//button->dstPos.y -= 25 * 2;
		//button->valPos.y -= 25 * 2;

		button=buttonList[4]; //staves
		button->dstPos=Vec2i(60, 25 * 3);
		button->valPos=Vec2i(60 + 22, 25 * 3);

		button=buttonList[5]; //xDistance
		button->dstPos=Vec2i(220, 25 * 3);
		button->valPos=Vec2i(220 + 22, 25 * 3);

		button=buttonList[6]; //index
		button->dstPos.y -= 25 * 3;
		button->valPos.y -= 25 * 3;

		button=buttonList[7]; //length
		button->dstPos.y -= 25 * 3;
		button->valPos.y -= 25 * 3;

		button=buttonList[8]; //height
		valPosY = button->valPos.y - 25 * 3;
		dstPosY = button->dstPos.y - 25 * 3;
		button->dstPos=Vec2i(60, dstPosY);
		button->valPos=Vec2i(60 + 22, valPosY);

		button=buttonList[9]; //yDistance
		button->dstPos=Vec2i(250, dstPosY);
		button->valPos=Vec2i(250 + 22, valPosY);

		valPosY += 25;
		dstPosY += 25;

		button=buttonList[10]; //bands
		button->dstPos=Vec2i(60, dstPosY);
		button->valPos=Vec2i(60 + 22, valPosY);



		button=buttonList[11]; //sync
		valPosY = button->valPos.y - 25 * 4;
		dstPosY = button->dstPos.y - 25 * 4;


		button->dstPos=Vec2i(first, dstPosY);
		button->valPos=Vec2i(first + 22, valPosY);

		button=buttonList[12]; //mold
		button->dstPos=Vec2i(second, dstPosY);
		button->valPos=Vec2i(second + 22, valPosY);

		button=buttonList[15]; //insideOut
		button->dstPos=Vec2i(third, dstPosY);
		button->valPos=Vec2i(third + 22, valPosY);


		valPosY += 25;
		dstPosY += 25;
		button=buttonList[16]; //xCentred
		button->dstPos=Vec2i(first, dstPosY);
		button->valPos=Vec2i(first + 22, valPosY);


		button=buttonList[18]; //beginSP
		button->dstPos=Vec2i(second, dstPosY);
		button->valPos=Vec2i(second + 22, valPosY);

		button=buttonList[13]; //southPole
		button->dstPos=Vec2i(third, dstPosY);
		button->valPos=Vec2i(third + 22, valPosY);

		valPosY += 25;
		dstPosY += 25;
		button=buttonList[17]; //xJoinEnds
		button->dstPos=Vec2i(first, dstPosY);
		button->valPos=Vec2i(first + 22, valPosY);

		button=buttonList[14]; //northPole
		button->dstPos=Vec2i(third, dstPosY);
		button->valPos=Vec2i(third + 22, valPosY);

		button=buttonList[19]; //endSP
		button->dstPos=Vec2i(second, dstPosY);
		button->valPos=Vec2i(second + 22, valPosY);

		height=startFieldLineYpos+25*9+10;
	}
	else if(dynamic_cast<NurbsPatch*>(node)){
		int first = 70;
		int second = first + 100;
		int third = first + 100 * 2;

		FieldButton* button=buttonList[4]; //showAll
		int valPosY = button->valPos.y;
		int dstPosY = button->dstPos.y;
		button->dstPos=Vec2i(first, dstPosY);
		button->valPos=Vec2i(first + 22, valPosY);

		button=buttonList[5]; //mirror
		button->dstPos=Vec2i(second, dstPosY);
		button->valPos=Vec2i(second + 22, valPosY);

		button=buttonList[6]; //dim
		valPosY = button->valPos.y - 25 * 2;
		dstPosY = button->dstPos.y - 25 * 2;
		button->dstPos=Vec2i(third - 30, dstPosY);
		button->valPos=Vec2i(third - 30 + 22 + 18, valPosY);


		button=buttonList[7]; //xTug
		valPosY = button->valPos.y - 25 * 2;
		dstPosY = button->dstPos.y - 25 * 2;
		button->dstPos=Vec2i(first - 23, dstPosY);
		button->valPos=Vec2i(first - 23 + 22, valPosY);

		button=buttonList[8]; //yTug
		button->dstPos=Vec2i(second - 60, dstPosY);
		button->valPos=Vec2i(second - 60 + 22, valPosY);

		button=buttonList[9]; //zTug
		button->dstPos=Vec2i(third - 100, dstPosY);
		button->valPos=Vec2i(third - 100 + 22, valPosY);

		button=buttonList[10]; //endSP
		button->dstPos=Vec2i(third + 20, dstPosY);
		button->valPos=Vec2i(third + 20 + 22, valPosY);

		button=buttonList[11]; //uJoinEnds
		valPosY += 25;
		dstPosY += 25;
		button->dstPos=Vec2i(second, dstPosY);
		button->valPos=Vec2i(second + 22, valPosY);

		button=buttonList[12]; //begiSP 
		button->dstPos=Vec2i(third + 20, dstPosY);
		button->valPos=Vec2i(third + 20 + 22, valPosY);


		height=startFieldLineYpos+25*6+10;
	}
	else if(dynamic_cast<Viewpoint*>(node)){
		FieldButton* button=buttonList[5]; //capture
		int valPosY = button->valPos.y;
		int dstPosY = button->dstPos.y;
		int first = 30;
		int second = first + 80;
		int third = first + 95 * 2;
		button->dstPos=Vec2i(first, dstPosY);
		button->valPos=Vec2i(first + 22, valPosY);

		button=buttonList[6]; //animCapture
		button->dstPos=Vec2i(second, dstPosY);
		button->valPos=Vec2i(second + 22, valPosY);
		button=buttonList[7]; //animDelete
		button->dstPos=Vec2i(third, dstPosY);
		button->valPos=Vec2i(third + 22, valPosY);
		height=startFieldLineYpos+25*6+10;
	}

	size=Vec2i(size.x, height);
}

void ControlPanel::setupNodeWindow(Node* node){
	int yReduction=0;

	setUpControlPanel(node,node->getProperties());
	paint();
	houseKeepVisualClue();
}

void ControlPanel::onLoseFocus(){
	textCell.cellActive=false;
	textCell.doNoThinkWinkOff();
}

void SceneTree::onLoseFocus(){
	if(textCell.cellActive){
		textCell.cellActive=false;
		textCell.doNoThinkWinkOff();
	}
}

void ControlPanel::CPTextCell::onChange(){
	const Properties& sprop=win.node->getProperties(button->bField);
	int index=0;
	if(sprop.groupIndex!=-1){
		index=win.node->getFieldGroupIndex(sprop.groupIndex);
	}
	int startAtIndex=sprop.startAtIndex;
	if(startAtIndex<=index){
		if(button->updateValFromTextCell(cellStr, index-startAtIndex, sprop)){
			if(win.node->getType() != FILE_NODE){
				updateBuild3dWindow();
				houseKeepVisualClue();
			}
		}
	}
}

void ControlPanel::colorBlockColorRgn(int*& dstPix){
	String& str=Button::str;
	int bottom=size.y-10;
	int top=title_bar_height;
	int lastGroup=-1;
	bool togleColor=false;
	for(int i=0; i<buttonList.length; i++){
		bottom = top;
		lastGroup = buttonList[i]->sprop.groupIndex;

		do{
			FieldButton* button = buttonList[i++];
			bottom = button->valPos.y+25;
		}while(i < buttonList.length && buttonList[i]->sprop.groupIndex == lastGroup);
		--i;

		if(dynamic_cast<AnimControlPanel*>(this)){
			bottom+=17;
		}
		if(togleColor=!togleColor){
			if(dynamic_cast<TextureMapper*>(node.p)){
				bottom+=35;
			}
//			if(dynamic_cast<File*>(node.p)){
//				bottom = size.y - 8;
//			}
			paintBlock2(0xffffff, 8, top, size.x-9, bottom, size.x, size.y, dstPix);
			paintSideEdge(&iconSrcBuf[(iconSrcBufCY-1-10)*480+444], 8, 0, top, bottom, size.x, size.y, dstPix);
			paintSideEdge(&iconSrcBuf[(iconSrcBufCY-1-10)*480+452], 9, size.x-9, top, bottom, size.x, size.y, dstPix);
		}
		else{
			paintBlock2(iconSrcBuf[(iconSrcBufCY-1)*480+451], iconSrcBuf[(iconSrcBufCY-1)*480+452], 8, top, size.x-9, bottom, size.x, size.y, dstPix);
			paintSideEdge(&iconSrcBuf[(iconSrcBufCY-1)*480+444], 8, 0, top, bottom, size.x, size.y, dstPix);
			paintSideEdge(&iconSrcBuf[(iconSrcBufCY-1)*480+452], 9, size.x-9, top, bottom, size.x, size.y, dstPix);
		}
		top=bottom;
	}
	togleColor=!togleColor;
	if(togleColor){
		paintBlock2(0xffffff, 8, top, size.x-9, size.y-10, size.x, size.y, dstPix);
		paintSideEdge(&iconSrcBuf[(iconSrcBufCY-1-10)*480+444], 8, 0, top, size.y-10, size.x, size.y, dstPix);
		paintSideEdge(&iconSrcBuf[(iconSrcBufCY-1-10)*480+452], 9, size.x-9, top, size.y-10, size.x, size.y, dstPix);

		blitHorizontalColorSweep(Vec2i(8, size.y-10), Vec2i(451, 10), Vec2i(452, 10), Vec2i(size.x-8-9, 10));
		pasteSprite2( 0, size.y-10, 444, 10, size.x, size.y, iconSrcBufCX, iconSrcBufCY, 8, 10, dstPix);
		pasteSprite2( size.x-9, size.y-10, 452, 10, size.x, size.y, iconSrcBufCX, iconSrcBufCY, 9, 10, dstPix);

	}
	else{

		paintBlock2(iconSrcBuf[(iconSrcBufCY-1)*480+451], iconSrcBuf[(iconSrcBufCY-1)*480+452], 8, top, size.x-9, size.y-10, size.x, size.y, dstPix);
		paintSideEdge(&iconSrcBuf[(iconSrcBufCY-1)*480+444], 8, 0, top, size.y-10, size.x, size.y, dstPix);
		paintSideEdge(&iconSrcBuf[(iconSrcBufCY-1)*480+452], 9, size.x-9, top, size.y-10, size.x, size.y, dstPix);

		blitHorizontalColorSweep(Vec2i(8, size.y-10), Vec2i(451, 0), Vec2i(452, 0), Vec2i(size.x-8-9, 10));
		pasteSprite2( 0, size.y-10, 444, 0, size.x, size.y, iconSrcBufCX, iconSrcBufCY, 8, 10, dstPix);
		pasteSprite2( size.x-9, size.y-10, 452, 0, size.x, size.y, iconSrcBufCX, iconSrcBufCY, 9, 10, dstPix);
	}
	{
		if(node->getType()!=PROTOINTERFACE){
			for(int j=0; j<node->getFLen(); j++){
				M_Route& m_route=node->getField(j)->refRoute();
				int len=m_route.length;
				if(len){
					str = node->getProperties()[j].name;

					int xPos=colum-12;
					if(str.length>14){
						xPos=str.length*8;
					}

					drawTextR(str, PROPERTY_COLOR, xPos, top+12, size.x, size.y, dstPix);
					if(len > 1){
						drawVerticleLine2(0, xPos+8+4, top+9, top+9+(m_route.length-1)*pixSymbolCY, size.x, size.y, dstPix);
					}
					for(int i = 0; i<len; i++){
						S_Route& sroute=m_route[i];


						drawHorizontalLine2(0, top+9, xPos+8+4, xPos+8+4+6, size.x, size.y, dstPix);
						pasteSprite2(xPos+8+4+10-4, top+6, 403, 13, size.x, size.y, iconSrcBufCX, iconSrcBufCY, 4, 7, dstPix);  
						Node* toNode=sroute.toNode;
						int type=toNode->getType();
						pasteSprite2(xPos+8+4+10, top+2, nodeIconLookUp[type].srcPos.x, nodeIconLookUp[type].srcPos.y, size.x, size.y, iconSrcBufCX, iconSrcBufCY, nodeIconLookUp[type].size.x, nodeIconLookUp[type].size.y, dstPix);  

						const M_Properties& mprop=toNode->getProperties();
						const Properties &sprop=mprop[sroute.toFieldIdx];
						str=sroute.toNode->name;
						str+='.';
						drawText2(str, NODE_COLOR, Vec2i(xPos+41, top+12), size.x, size.y, dstPix);
						int xp=str.length*8;
						str=sprop.name;
						drawText2(str, PROPERTY_COLOR, Vec2i(xPos+41+xp, top+12), size.x, size.y, dstPix);

						top+=pixSymbolCY;
					}
				}
			}
		}
	}
}
String findMatchingIsName(Field* field){

	ProtoInterface* protoInterface=getSceneProtoDeclare(g_stw->selLine)->protoInterface;

	for(int i=0; i<protoInterface->getFLen(); i++){
		if(protoInterface->getField(i)->testRouteTo(field)){
			return protoInterface->properties[i].name;
		}
	}
	throw protoInterface;
}
inline void TextureMapper::drawTextureGrid(int& dstWidth, int& dstHeight, int* pixBuf){

	for(int y=0; y<8; y++){
		for(int x=y&1; x<8; x+=2){
			paintBlock2(0x808000, pos.x+x*tile, pos.y+y*tile, pos.x+x*tile+tile, pos.y+y*tile+tile, dstWidth, dstHeight, pixBuf);
		}
	}
	for(int y=0; y<8; y++){
		for(int x=(y+1)&1; x<8; x+=2){
			paintBlock2(0x000080, pos.x+x*tile, pos.y+y*tile, pos.x+x*tile+tile, pos.y+y*tile+tile, dstWidth, dstHeight, pixBuf);
		}
	}
//	drawVerticleLine(	0x808080, pos.x-1, pos.y-1, pos.y+tile*8+1, dstWidth, dstHeight, pixBuf);
//	drawVerticleLine(	0x808080, pos.x+tile*8, pos.y-1, pos.y+tile*8+1, dstWidth, dstHeight, pixBuf);
//	drawHorizontalLine(	0x808080, pos.y-1, pos.x, pos.x+tile*8, dstWidth, dstHeight, pixBuf);
//	drawHorizontalLine(	0x808080, pos.y+tile*8, pos.x, pos.x+tile*8, dstWidth, dstHeight, pixBuf);

	float y=pos.y + tile * 8 - tile * 8 * position2d.y;
	float x=pos.x + tile * 8 * position2d.x;
	Vec2i fillPosLT=Vec2i(x, y - tile * 8 * size2d.y);
	Vec2i fillPosRB=Vec2i(x + tile * 8 * size2d.x + .5, y + .5);
	changePixelColorInBlock(0x808000, 0xFFFF00, fillPosLT.x, fillPosLT.y, fillPosRB.x, fillPosRB.y, dstWidth, dstHeight, pixBuf);
	changePixelColorInBlock(0x000080, 0x0000FF, fillPosLT.x, fillPosLT.y, fillPosRB.x, fillPosRB.y, dstWidth, dstHeight, pixBuf);
}

void ArtCoordPointButton::paintFieldName(){
	if(part.smls->workAnimCP && (Window::animMode() || g_colorOwnership)){
		drawTextR(str = "controlPoint", PROPERTY_COLOR, dstPos.x - 12, valPos.y + 16, win.size.x, win.size.y, win.pixBuf.getArray());
	}
	else{
		drawTextR(str = sprop.name, PROPERTY_COLOR, dstPos.x - 12, valPos.y + 16, win.size.x, win.size.y, win.pixBuf.getArray());
	}
}

void FieldButton::paintFieldName(){
	drawTextR(str=sprop.name, PROPERTY_COLOR, dstPos.x-12, valPos.y+16, win.size.x, win.size.y, win.pixBuf.getArray());
}

void LengthButton::paintFieldName(){
	drawTextR(str="length", PROPERTY_COLOR, dstPos.x-12, valPos.y+16, win.size.x, win.size.y, win.pixBuf.getArray());
}

void IndexButton::paintFieldName(){
	drawTextR(str="index", PROPERTY_COLOR, dstPos.x-12, valPos.y+16, win.size.x, win.size.y, win.pixBuf.getArray());
}

int getWorldList(String& str);

String g_serverWorldListHTML;
void OpenNetFile::printWorldsOpen(const int dstWidth, const int dstHeight, int* dstPix){
	if(worldListRefreshing){
		return;
	}
	static int i = 0;
	if(worldListRefreshed==false){
		worldListRefreshing=true;
		i = getWorldList(g_serverWorldListHTML="");
		worldListRefreshed=true;
		worldListRefreshing=false;
	}

	int y = 16*4;
	if(i==-2){
		drawText2(g_serverWorldListHTML, 0, Vec2i(10, y), dstWidth, dstHeight, dstPix);	
	}
	else if(i > 0){
		for(int k = 0; k<worldListUrls.length; k++){
			drawText2(worldListTitles[k], 0x0000ff, Vec2i(10, y), dstWidth, dstHeight, dstPix);
			int x = 10+worldListTitles[k].length*8;
			drawHorizontalLine2(0x0000ff, y+3, 10, x, dstWidth, dstHeight, dstPix);
			drawText2(worldListOccupants[k], 0, Vec2i(x+8, y), dstWidth, dstHeight, dstPix);
			y+=16;
		}
	}
}
void ControlPanel::paint(){
	if(node==0)return;

	int dstWidth=size.x;
	int dstHeight=size.y;
	int* dstPix=pixBuf.getArray();
	const M_Properties& properties=node->getProperties();

//	paintBlock2( 0xf0f0ff, 1, 1, size.x-1, 25, size.x, size.y, dstPix);

	colorBlockColorRgn(dstPix);
//	for(int i=0; i<24; i++){
//		drawHorizontalLine2(iconSrcBuf[(24-i+24*2)*480+24*7], 1+i, 1, dstWidth-1, dstWidth, dstHeight, dstPix);
//	}
	//paintBlock2(0xf0f0ff, 1, 25, dstWidth, dstHeight, dstPix);
	blitHorizontalColorSweep(Vec2i(10, 0), Vec2i(470, 0), Vec2i(471, 0), Vec2i(dstWidth-9-10, title_bar_height));
	pasteSprite2( 0, 0, 461, 0, dstWidth, dstHeight, iconSrcBufCX, iconSrcBufCY, 10, 25, dstPix);
	pasteSprite2( dstWidth-9, 0, 471, 0, dstWidth, dstHeight, iconSrcBufCX, iconSrcBufCY, 9, 25, dstPix);

//	drawHorizontalLine2( 0x808080, 0, 0, size.x, size.x, size.y, dstPix);
//	drawHorizontalLine2( 0xb5b5c9, size.y-1, 0, size.x, size.x, size.y, dstPix);
//	drawHorizontalLine2( 0xffffff, size.y-2, 0, size.x, size.x, size.y, dstPix);
//	drawHorizontalLine2( 0, size.y-3, 0, size.x, size.x, size.y, dstPix);

//	drawVerticleLine2( 0xb5b5c9, 0, 0, size.y, size.x, size.y, dstPix);
//	drawVerticleLine2( 0xffffff, 1, 0, size.y-1, size.x, size.y, dstPix);
//	drawVerticleLine2( 0, 2, 0, size.y-2, size.x, size.y, dstPix);

//	drawVerticleLine2( 0x0, size.x-3, 0, size.y-2, size.x, size.y, dstPix);
//	drawVerticleLine2( 0xffffff, size.x-2, 0, size.y-1, size.x, size.y, dstPix);
//	drawVerticleLine2( 0xb5b5c9, size.x-1, 0, size.y, size.x, size.y, dstPix);

	static String str;
	int i = 0;
	int j = 0;
	int currentGroup=-1;

	int g=str.length*8+15;
	int type=node->getType();
	int iconNodeType = type;
	if(type>=g_userProtoBegin){
		iconNodeType=PROTOINSTANCE;
	}
	int iconExtraX = (nodeIconLookUp[iconNodeType].size.x-16)>>1;
	pasteSprite2( 5-iconExtraX, 5-((nodeIconLookUp[iconNodeType].size.y-16)>>1), nodeIconLookUp[iconNodeType].srcPos.x, nodeIconLookUp[iconNodeType].srcPos.y, dstWidth, dstHeight, iconSrcBufCX, iconSrcBufCY, nodeIconLookUp[iconNodeType].size.x, nodeIconLookUp[iconNodeType].size.y, dstPix);

	str=node->getTypeStr();

	//drawText2(str, 0xffff00, 10+0, 15, dstWidth, dstHeight, dstPix);
	//drawText2(str, 0xff0000, 10+2, 17, dstWidth, dstHeight, dstPix);


	if(node->name!=""){
		str+=' ';
		str+=node->name;
	}
//	drawText2(str, 0xc0c0c0, 21+1, 16, dstWidth, dstHeight, dstPix);
//	drawText2(str, 0x804000, 21+2, 17, dstWidth, dstHeight, dstPix);
//	drawText2(str, 0xff4000, 21+2, 17, dstWidth, dstHeight, dstPix);
//	drawText2(str, 0x900080, 21+2, 17, dstWidth, dstHeight, dstPix);
//	drawText2(str, 0xff0000, 21+2, 17, dstWidth, dstHeight, dstPix);
	drawText2(str, 0x0, Vec2i(21+2+iconExtraX, 16), dstWidth, dstHeight, dstPix);

	exitWindow.dstPos=Vec2i(size.x-22, 4);
	exitWindow.paint();

	if(OpenNetFile* openNetFile=dynamic_cast<OpenNetFile*>(node.p)){
		if(openNetFile->worldList){
			FieldButton* button = buttonList[i];
			button->paintFieldName();
			button->paint();
			openNetFile->printWorldsOpen(dstWidth, dstHeight, dstPix);
			goto here;
		}
	}
	for(int i=buttonList.length; --i>=0;){
		FieldButton* button = buttonList[i];

		if(button->sprop.doNotDisplayName() == false && button->sprop.push() == false && button->sprop.link() == false){
			button->paintFieldName();
		}
		
		if(button->bField.isRef() && !dynamic_cast<IndexButton*>(button) && !dynamic_cast<LengthButton*>(button)){
			str="IS";
			drawText2(str, 0x0ff, Vec2i(button->dstPos.x, button->valPos.y+16), size.x, size.y, dstPix);
			str=findMatchingIsName(&button->bField);
			drawText2(str, PROPERTY_COLOR, Vec2i(button->dstPos.x+3*8, button->valPos.y+16), size.x, size.y, dstPix);
		}
		else{
			button->paint();
			Vec2i valPos=button->valPos;
			valPos.y+=16;
			if(button->sprop.doNotDisplayValue()==false){
				if(button->editMode){
					drawText2(textCell.cellStr, textCell.firstVisChar, textCell.cellStr.length, 0, valPos, size.x, size.y, dstPix);
					textCell.highLightBlock();
				}
				else{
					button->toString(Button::str="");
					drawText2(Button::str, 0, valPos, size.x, size.y, dstPix);
				}
			}
		}
	}
	if(TextureMapper* textureMapper=dynamic_cast<TextureMapper*>(node.p)){
		textureMapper->drawTextureGrid(dstWidth, dstHeight, dstPix);
	}
here:
	if(g_settings->invertWindowColor){
		invert();
	}
	loadSurfaceFromMemory();
	if(textCell.cellActive==true){
		textCell.doNoThinkWink();
	}
}
inline void ControlPanel::CPTextCell::selectField(FieldButton& bf){
	g_callBackList += this;
	bf.editMode = true;
	cellActive = true;//not sure how much this can be relied appoun to tell the state of the caret but seems ok so far
	button = &bf;
	bf.toString(cellStr="");
	cellPos = bf.valPos;
	cellPos.y += 5;
	cellSize = bf.valSize;
	firstVisChar = 0;
}
inline bool TextureMapper::clickGrid(const Vec2i ptCursor){
	float x =ptCursor.x-pos.x;
	float y =ptCursor.y-pos.y;
	if(x < 0 || y < 0 || x > tile * 8 || y > tile * 8){
		return false;
	}

	y=tile*8-y;

	Vec2f positionRT;
	if(position2d==Vec2f() && size2d==Vec2f()){
		position2d.refValue() = Vec2f(floor(x / tile) / 8, floor(y / tile) / 8);
		positionRT = position2d + Vec2f(1.0f/8, 1.0f/8);
	}
	else{
		positionRT = position2d + size2d;

		if(x < position2d.refValue().x * tile * 8){
			position2d.refValue().x=floor(x / tile) / 8;
		}
		else if(x > positionRT.x * tile * 8){
			positionRT.x=ceil(x / tile) / 8;
		}
		if(y < position2d.refValue().y * tile * 8){
			position2d.refValue().y=floor(y / tile) / 8;
		}
		else if(y > positionRT.y * tile * 8){
			positionRT.y=ceil(y / tile) / 8;
		}
		position2d.refValue().x=floor(position2d.refValue().x * 8) / 8;
		position2d.refValue().y=floor(position2d.refValue().y * 8) / 8;

		positionRT.x=ceil(positionRT.x * 8) / 8;
		positionRT.y=ceil(positionRT.y * 8) / 8;

	}
	position2d.refValue()+=Vec2f(.002, .002);
	positionRT-=Vec2f(.002, .002);

	size2d.refValue()=positionRT-position2d.refValue();

	updateBuild3dWindow();
	return true;
}

bool ControlPanel::click(const Vec2i ptCursor){
	if(ptCursor >= pos && ptCursor < pos+size){
		activeWindow=this;
		if(tugBorder){
			tugBorder->window=this;
			g_drag=tugBorder;
			return true;
		}

		const Vec2i localCursor=ptCursor-pos;

		if(OpenNetFile* openNetFile=dynamic_cast<OpenNetFile*>(node.p)){
			int i = openNetFile->clickUrl(localCursor);
			if(i > 0){
				return true;
			}
			else if(i==-1){
				return false;
			}
		}
		else if(TextureMapper* textureMapper=dynamic_cast<TextureMapper*>(node.p)){
			if(textureMapper->clickGrid(localCursor)){
				return true;
			}
		}
		{
			int newCaretPos = 0;
			for(int k=buttonList.length; --k >= 0;){
				FieldButton* button=buttonList[k];
				if(button->checkClickTextCell(localCursor, newCaretPos)){
					if(button->editMode == false){
						textCell.cellActive = false;
						textCell.selectField(*button);
					}
					textCell.highlightPos = textCell.caretPos = newCaretPos + textCell.firstVisChar;
					textCell.updateCaret();

					textCell.checkForDoubleLeftClick();

					g_drag = &textCell;
					paint();
					return true;
				}
			}
		}
		textCell.cellActive=false;
		if(localCursor.y < title_bar_height){
			g_drag = this;
			dragPos = ptCursor;
			return true;
		}
		return true;
	}
	else{
		return false;
	}
}
void ControlPanel::doDrag(Vec2i ptCursor){
	commonDoDrag(ptCursor, pos, dragPos);
}
void CaretWindow::doDrag(const Vec2i ptCursor){
	caretPos = firstVisChar + (ptCursor.x - win.pos.x - cellPos.x + charWidth / 2) / charWidth;
	updateCaret();
	win.paint();
}
void ControlPanel::CPTextCell::setCellActiveFalse(){
	M_<Ptr<FieldButton> >& buttonList=win.buttonList;
	for(int k=buttonList.length; --k>=0;){
		buttonList[k]->editMode=false;
	}
	removeFromList<CallBack>(g_callBackList, this);
	win.paint();
}

void CaretWindow::checkForDoubleLeftClick(){
	if(cellStr.length){

		if(isDoubleLeftClick()){

			if(caretPos==cellStr.length){
				caretPos--;
			}
			highlightPos=caretPos;	
			while(caretPos < cellStr.length && cellStr[caretPos] > ' ' && cellStr[caretPos] != ':'){
				caretPos++;
			}
			while(highlightPos > 0 && cellStr[highlightPos-1] > ' ' && cellStr[highlightPos-1] != ':'){
				highlightPos--;
			}
		}
	}
}
void ControlPanel::onGainFocus(){
	textCell.onWindowGainFocus();
}
void ControlPanel::setWinSize(Vec2i& raw){
	pixBuf.length = raw.x * raw.y;
	for(int k=buttonList.length; --k>=0;){
		FieldButton* button=buttonList[k];
		if(button->sprop.doNotResize()==false){
			button->valSize.x = size.x - button->valPos.x;
		}
	}
	if(E_FAIL==initWindow()){
		throw;
	}
}
