/* Seamless3d

Copyright (C) 2001-2010 Graham John Perrett (AKA thyme) thyme@seamless3d.com www.seamless3d.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this
software and associated documentation files (the "Software"), to deal in the Software
without restriction, including without limitation the rights to , copy, modify,
merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to the following
conditions:

The above copyright notice and this permission notice shall be included in all copies or
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT
OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.
*/

#include <windows.h>
#include <direct.h>
#include <process.h>

#include "seamless3d.h"
#include "resource.h"
#include "techuelife_triangle_editor.h"
#include "gui.h"
#include "scene_tree.h"

#include "robot_aid.h"
extern const char* Create_a_New_Seamless_File;
extern const char* Create_Skeleton_of_Parts;
extern const char* Add_a_Group;
extern const char* Build_the_Pelvis;
extern const char* Add_a_pink_group;
extern const char* Build_the_Left_Eye;
extern const char* Copy_the_Left_Eye_to_the_Right;
extern const char* Build_the_Neck;
extern const char* Build_the_Head;
extern const char* Add_a_Cream_Group;
extern const char* Build_the_Left_Wrist;
extern const char* Build_the_Left_Elbow;
extern const char* Build_the_Left_Shoulder;
extern const char* Copy_the_Left_Arm_to_the_Right;
extern const char* Add_a_Blue_Group;
extern const char* Build_the_Left_Ankle;
extern const char* Build_the_Left_Knee;
extern const char* Build_the_Left_Hip;
extern const char* Copy_the_Left_Leg_to_the_Right;
extern const char* Position_the_Limbs_and_Neck;
extern const char* Add_a_Green_Group;
extern const char* Place_the_Avatar_to_the_Floor;
extern const char* Move_the_Pivot_Points;
extern const char* Move_the_Eye_Pivot_Points;
extern const char* Move_the_Rest_of_the_Pivot_Points;
extern const char* Add_a_Gesture_Group;
extern const char* Setup_Animation;
extern const char* Record_the_poses;
extern const char* Complete;

extern const char* name_pelvis;
extern const char* name_neck;
extern const char* name_head;
extern const char* name_l_eyeball;
extern const char* name_r_eyeball;

extern const char* name_l_shoulder;
extern const char* name_l_elbow;
extern const char* name_l_wrist;

extern const char* name_r_shoulder;
extern const char* name_r_elbow;
extern const char* name_r_wrist;

extern const char* name_l_hip;
extern const char* name_l_knee;
extern const char* name_l_ankle;
	
extern const char* name_r_hip;
extern const char* name_r_knee;
extern const char* name_r_ankle;

extern const char* name_front_view;
extern const char* name_left_view;
extern const char* name_right_view;
extern const char* name_back_view;
extern const char* name_birds_eye_view;
extern const char* name_worms_eye_view;
extern const char* name_showcase_view;

extern const char* str_hello;

void InputEvent::preNurbsLatheAvDemo(){
	g_robotDemo->fileName="nurbs_lathe_av";
	g_robotDemo->function=InputEvent::nurbsLatheAvDemo;
	g_robotDemo->name=Create_a_New_Seamless_File;
	setupSequence();
}
void InputEvent::nurbsLatheAvDemo(){

	const int seamless=1;
	const int pelvis=seamless+2;

	const int neck=	pelvis+1;
	const int head=		neck+1;
	const int l_eyeball=	head+1;
	const int r_eyeball=	l_eyeball+1;

	const int l_shoulder=	r_eyeball+1;
	const int l_elbow=		l_shoulder+1;
	const int l_wrist=		l_elbow+1;

	const int r_shoulder=	l_wrist+1;
	const int r_elbow=		r_shoulder+1;
	const int r_wrist=		r_elbow+1;

	const int l_hip=		r_wrist+1;
	const int l_knee=		l_hip+1;
	const int l_ankle=		l_knee+1;
	
	const int r_hip=		l_ankle+1;
	const int r_knee=		r_hip+1;
	const int r_ankle=		r_knee+1;

	const int pinkGroup=r_ankle+1;
	const int pelvisLathe=pinkGroup+1;
	const int neckLathe=pelvisLathe+2;
	const int headLathe=neckLathe+2;
	const int eyeSphere=headLathe+2;
	const int eyeCopyPart=eyeSphere+2;
	
	const int creamGroup=eyeCopyPart+3;
	const int wristLathe=creamGroup+1;
	const int elbowLathe=wristLathe+2;
	const int shoulderLathe=elbowLathe+2;
	const int armCopyPart=shoulderLathe+2;

	const int blueGroup=armCopyPart+7;
	const int ankleLathe=blueGroup+1;
	const int kneeLathe=ankleLathe+2;
	const int hipLathe=kneeLathe+2;
	const int legCopyPart=hipLathe+2;


	const int gestureGroup=legCopyPart+7;
	const int anim = gestureGroup+1;

	list+=new Section(Create_a_New_Seamless_File);

	clickButton(newSmls);
	list+=new TypeString(deleteOldFileAndCreateFileName());
	list+=new Move_CreateNewFile_Save();
	list+=new Click;

	InputEvent* create_Skeleton_of_PartsArray[]={
		new Section(Create_Skeleton_of_Parts),
		new ClickButton(openSceneTree),

		new NameNode(pelvis, name_pelvis),
		new NewNodeAsChild(PART, pelvis),
		new NameNode(neck, name_neck),
		new NewNodeAsChild(PART, neck),
		new NameNode(head, name_head),
		new NewNodeAsChild(PART, head),
		new NameNode(l_eyeball, name_l_eyeball),
		new NewNodeAsSiblingBelow(PART, l_eyeball),
		new NameNode(r_eyeball, name_r_eyeball),

		new NewNodeAsChild(PART, pelvis),
		new NameNode(l_shoulder, name_l_shoulder),
		new NewNodeAsChild(PART, l_shoulder),
		new NameNode(l_elbow, name_l_elbow),
		new NewNodeAsChild(PART, l_elbow),
		new NameNode(l_wrist, name_l_wrist),

		new NewNodeAsChild(PART, pelvis),
		new NameNode(r_shoulder, name_r_shoulder),
		new NewNodeAsChild(PART, r_shoulder),
		new NameNode(r_elbow, name_r_elbow),
		new NewNodeAsChild(PART, r_elbow),
		new NameNode(r_wrist, name_r_wrist),

		new NewNodeAsChild(PART, pelvis),
		new NameNode(l_hip, name_l_hip),
		new NewNodeAsChild(PART, l_hip),
		new NameNode(l_knee, name_l_knee),
		new NewNodeAsChild(PART, l_knee),
		new NameNode(l_ankle, name_l_ankle),

		new NewNodeAsChild(PART, pelvis),
		new NameNode(r_hip, name_r_hip),
		new NewNodeAsChild(PART, r_hip),
		new NameNode(r_knee, name_r_knee),
		new NewNodeAsChild(PART, r_knee),
		new NameNode(r_ankle, name_r_ankle),

		new ClickButton(saveSmls),
		new Section("Add floor & height guide"),
		new ClickButton(floorBuild),
//		new MouseZoom(1.4),
	};
	list+=EVENT_LIST(create_Skeleton_of_PartsArray);

	InputEvent* add_a_Pink_GroupArray[]={
		new Section(Add_a_pink_group),
		new NewNodeAsChild(GROUPBUILD, seamless),
		new SelectNode(pinkGroup),
		new DecFieldY("marker", 10),
		new DecFieldZ("marker", 10),

		new Section(Build_the_Pelvis),

		new SelectNode(pelvis),
		new ClickButton(buildAvSegmentForSelPart),
		new DecFieldY("translation", 93),
		new MoveToBezierPoint(4),
		new Move3d(Vec3f(0, .61, 0)),
		new MoveToBezierPoint(3),
		new Move3d(Vec3f(.15, .61, 0)),
		new MoveToBezierPoint(2),
		new Move3d(Vec3f(.114768, .298663, 0)),
		new MoveToBezierPoint(1),
		new Move3d(Vec3f(.157916, 0, 0)),
		new IncFieldGroup("point"),
		new MoveToBezierPoint(2),
		new Move3d(Vec3f(.145376, .125742, 0)),
		new SelectVertex(3),
		new IncFieldGroup("point"),
		new MoveToBezierPoint(4),
		new Move3d(Vec3f(.150782, .477376, 0)),
		new ClickButton(setLeftViewOrientation),
		new DecFieldZ("scale", 100-70),
		new DecFieldX("color", 100),
		new DecFieldY("color", 100-80),

		new Section(Build_the_Neck),
		new SelectNode(neck),
		new ClickButton(buildAvSegmentForSelPart),
		new DecFieldY("translation", 35),
		new DecFieldX("rotation", 3),
		new DecFieldY("scale", 100-35),
		new DecFieldX("color", 100-90),
		new DecFieldY("color", 100-67),
		new DecFieldZ("color", 100-55),

		new Section(Build_the_Head),
		new ClickButton(setFrontViewOrientation),
		new SelectNode(head),
		new ClickButton(buildAvSegmentForSelPart),
		new DecFieldY("translation", 28),
		new MoveToBezierPoint(1),
		new Move3d(Vec3f(.0685, 0, 0)),
		new MoveToBezierPoint(2),
		new Move3d(Vec3f(.092602, .082352, 0)),
		new MoveToBezierPoint(3),
		new Move3d(Vec3f(.074807, .212197, 0)),
		new MoveToBezierPoint(4),
		new Move3d(Vec3f(0, .213292, 0)),
		new ClickButton(setLeftViewOrientation),
		new IncFieldZ("translation", 3),
		new DecFieldX("color", 100-90),
		new DecFieldY("color", 100-67),
		new DecFieldZ("color", 100-55),
		//////////
	//build eyes
		new Section(Build_the_Left_Eye),
		new SelectNode(l_eyeball),
		new ClickButton(buildSphereForSelPart),
		new DecFieldCtrlX("rotation", 1),
		new DecFieldX("radius", 25-3),
		new DecFieldY("translation", 17),
		new IncFieldZ("translation", 7),
		new ClickButton(setFrontViewOrientation),
		new ClickButton(toggleWireFrame),
		new IncFieldX("translation", 3),
		new DecFieldX("color", 100),
		new DecFieldY("color", 100),
		new DecFieldZ("color", 100),

	//Copy the Left Eye to the Right
		new Section(Copy_the_Left_Eye_to_the_Right),
		new SelectNode(l_eyeball),
		new VirtKeyDown(VK_SHIFT),
		new VirtKeyDown(VK_DOWN),
		new VirtKeyUp(VK_DOWN),
		new VirtKeyUp(VK_SHIFT),
		new ClickButton(copyPartForSelParts),
		new ToggleField("mirrorX"),
	};
	list+=EVENT_LIST(add_a_Pink_GroupArray);

	InputEvent* add_a_Cream_GroupArray[]={
		new Section(Add_a_Cream_Group),
		new Collapse(pinkGroup),
		new NewNodeAsChild(GROUPBUILD, seamless),
		new SelectNode(creamGroup),
		new DecFieldZ("marker", 20),
		new Section(Build_the_Left_Shoulder),
		new SelectNode(l_shoulder),
		new ClickButton(buildAvSegmentForSelPart),
		new IncFieldX("translation", 16),
		new DecFieldY("translation", 65),
		new DecFieldY("scale", 3),
		new ClickButton(toggleWireFrame),
		new MoveToBezierPoint(1),
		new Move3d(Vec3f(.034578, 0, 0)),
		new MoveToBezierPoint(2),
		new Move3d(Vec3f(.041518, .148037, 0)),
		new DecFieldZ("rotation", 3),

		new ClickButton(setLeftViewOrientation),
		new IncFieldZ("translation", 2),
		new DecFieldX("color", 100),
		new DecFieldY("color", 100-80),

		new Section(Build_the_Left_Elbow),
		new SelectNode(l_elbow),
		new ClickButton(buildAvSegmentForSelPart),
		new DecFieldY("translation", 88),
		new IncFieldZ("translation", 2),
		new DecFieldY("scale", 5),
		new ClickButton(setFrontViewOrientation),
		new IncFieldX("translation", 16),
		new MoveToBezierPoint(1),
		new Move3d(Vec3f(.024554, 0, 0)),
		new MoveToBezierPoint(2),
		new Move3d(Vec3f(.032073, .187374, 0)),
		new MoveToBezierPoint(3),
		new Move3d(Vec3f(.035735, .293102, 0)),
		new DecFieldX("color", 100),
		new DecFieldY("color", 100-80),

		new Section(Build_the_Left_Wrist),
		new SelectNode(l_wrist),
		new ClickButton(buildAvSegmentForSelPart),
		new IncFieldX("translation", 16),
		new DecFieldY("translation", 101),
		new DecFieldY("scale", 100-54),
		new DecFieldX("scale", 100-57),
		new MoveToBezierPoint(3),
		new Move3d(Vec3f(.048195, .252597, 0)),
		new ClickButton(setLeftViewOrientation),
		new IncFieldZ("translation", 2),
		new DecFieldZ("scale", 100-90),
		new DecFieldX("color", 100-90),
		new DecFieldY("color", 100-67),
		new DecFieldZ("color", 100-55),
		new ClickButton(setFrontViewOrientation),

		//Copy the Left Arm to the Right
		new Section(Copy_the_Left_Arm_to_the_Right),
		new SelectNode(l_shoulder),
		new VirtKeyDown(VK_SHIFT),
		new VirtKeyDown(VK_DOWN),
		new VirtKeyUp(VK_DOWN),
		new VirtKeyUp(VK_SHIFT),
		new ClickButton(copyPartForSelParts),
		new ToggleField("mirrorX"),

		new ClickButton(saveSmls),
	};
	list+=EVENT_LIST(add_a_Cream_GroupArray);

	InputEvent* add_a_Blue_GroupArray[]={
		new Section(Add_a_Blue_Group),
		new Collapse(creamGroup),
		new NewNodeAsChild(GROUPBUILD, seamless),
		new SelectNode(blueGroup),
		new DecFieldX("marker", 10),
		new DecFieldY("marker", 10),
		new Section(Build_the_Left_Hip),
		new SelectNode(l_hip),
		new ClickButton(buildAvSegmentForSelPart),
		new IncFieldX("translation", 9),
		new DecFieldY("translation", 133),
		new IncFieldY("scale", 53),
		new MoveToBezierPoint(2),
		new Move3d(Vec3f(.065422, .147206, 0)),
		new MoveToBezierPoint(3),
		new Move3d(Vec3f(.067349, .3, 0)),
		new DecFieldX("color", 100),
		new DecFieldY("color", 100-80),

		new Section(Build_the_Left_Knee),
		new SelectNode(l_knee),
		new ClickButton(buildAvSegmentForSelPart),
		new DecFieldY("translation", 172),
		new IncFieldX("translation", 9),
		new IncFieldY("scale", 53),
		new MoveToBezierPoint(1),
		new Move3d(Vec3f(.037226, -.002623, 0)),
		new MoveToBezierPoint(2),
		new Move3d(Vec3f(.048573, .137332, 0)),
		new MoveToBezierPoint(3),
		new Move3d(Vec3f(.056204, .300033, 0)),
		new DecFieldX("color", 100),
		new DecFieldY("color", 100-80),

		new Section(Build_the_Left_Ankle),
		new SelectNode(l_ankle),
		new ClickButton(buildAvSegmentForSelPart),
		new DecFieldY("translation", 175),
		new IncFieldX("translation", 9),
		new DecFieldY("scale", 100-28),
		new MoveToBezierPoint(1),
		new Move3d(Vec3f(.059609, -.008864, 0)),
		new MoveToBezierPoint(2),
		new Move3d(Vec3f(.062153, .158957, 0)),
		new MoveToBezierPoint(3),
		new Move3d(Vec3f(.026871, .282803, 0)),
		new ClickButton(setLeftViewOrientation),
		new IncFieldZ("scale", 50),
		new IncFieldZ("translation", 4),
		new ClickButton(setFrontViewOrientation),
		new DecFieldY("color", 100),
		new DecFieldZ("color", 100-70),

		//Copy the Left Leg to the Right
		new Section(Copy_the_Left_Leg_to_the_Right),
		new SelectNode(l_hip),
		new VirtKeyDown(VK_SHIFT),
		new VirtKeyDown(VK_DOWN),
		new VirtKeyUp(VK_DOWN),
		new VirtKeyUp(VK_SHIFT),
		new ClickButton(copyPartForSelParts),
		new ToggleField("mirrorX"),

		new ClickButton(saveSmls),
	};
	list+=EVENT_LIST(add_a_Blue_GroupArray);

	InputEvent* move_the_Pivot_PointsArray[]={
		new Section(Move_the_Pivot_Points),
		new SelectNode(pelvis),
		new DecFieldY("pivotPoint", 88),
		new SelectNode(neck),
		new IncFieldY("pivotPoint", 55),
		new SelectNode(head),
		new IncFieldY("pivotPoint", 7),

		new SelectNode(l_shoulder),
		new IncFieldY("pivotPoint", 49),
		new IncFieldX("pivotPoint", 16),

		new ClickButton(setLeftViewOrientation),

		new IncFieldZ("pivotPoint", 2),
		new SelectNode(l_elbow),
		new DecFieldY("pivotPoint", 24),
		new SelectNode(l_wrist),
		new DecFieldY("pivotPoint", 24),

		new ClickButton(setFrontViewOrientation),

		new SelectNode(r_shoulder),
		new IncFieldY("pivotPoint", 49),
		new DecFieldX("pivotPoint", 16),

		new ClickButton(setRightViewOrientation),

		new IncFieldZ("pivotPoint", 2),
		new SelectNode(r_elbow),
		new DecFieldY("pivotPoint", 24),
		new SelectNode(r_wrist),
		new DecFieldY("pivotPoint", 24),

		new ClickButton(setFrontViewOrientation),
		new SelectNode(l_hip),
		new IncFieldX("pivotPoint", 9),
		new SelectNode(l_knee),
		new DecFieldY("pivotPoint", 41),
		new SelectNode(l_ankle),
		new DecFieldY("pivotPoint", 41),

		new SelectNode(r_hip),
		new DecFieldX("pivotPoint", 9),
		new SelectNode(r_knee),
		new DecFieldY("pivotPoint", 41),
		new SelectNode(r_ankle),
		new DecFieldY("pivotPoint", 41),
	};
	list+=EVENT_LIST(move_the_Pivot_PointsArray);

	InputEvent* Add_a_Gesture_GroupArray[]={
		new Section(Add_a_Gesture_Group),
		new Collapse(blueGroup),
		new SelectNode(gestureGroup), //at this stage this is the no output group
		new NewNodeAsSiblingAbove(GESTUREGROUP, gestureGroup),
		new Section(Setup_Animation),
		new NewNodeAsChild(ANIM, gestureGroup),
		new NameNode(anim, str_hello),
		new IncFieldX("period", 1),

		new Section(Record_the_poses),
		//new MouseZoom(1),
		new DragAnimSlider(.25),
		new SelectNode(r_shoulder),

		new MoveToAnimRotPoint(),
		new Move3d(Vec3f(.265,.3,.020002)),
		new SelectNode(r_elbow),

		new MoveToAnimRotPoint(),
		new Move3d(Vec3f(.311474,.42654,.020002)),
		new SelectNode(l_shoulder),

		new MoveToAnimRotPoint(),
		new Move3d(Vec3f(.635,.53,.020002)),
		new SelectNode(l_elbow),

		new MoveToAnimRotPoint(),
		new Move3d(Vec3f(.705608,.114373,.020002)),
		new SelectNode(l_wrist),

		new MoveToAnimRotPoint(),
		new Move3d(Vec3f(.732901,.064384,.020002)),
		new ClickButton(setRightViewOrientation),

		new SelectNode(r_shoulder),

		new MoveToAnimRotPoint(),
		new Move3d(Vec3f(.015002,1.03,.17)),
		new SelectNode(r_elbow),

		new MoveToAnimRotPoint(),
		new Move3d(Vec3f(.523454,.916936,.243526)),
		new SelectNode(r_wrist),

		new MoveToAnimRotPoint(),
		new Move3d(Vec3f(.766737,.895295,.229557)),
		new ClickButton(setBirdsEyeViewOrientation),

		new SelectNode(r_shoulder),

		new MoveToAnimRotPoint(),
		new Move3d(Vec3f(.27,-.255002,.44)),
		new SelectNode(r_wrist),

		new MoveToAnimRotPoint(),
		new Move3d(Vec3f(-.085883,-.687791,.650296)),
		new ClickButton(setRightViewOrientation),

		new SelectNode(pelvis),

		new MoveToAnimRotPoint(),
		new Move3d(Vec3f(.500002,-.11,0)),
		new SelectNode(r_hip),

		new MoveToAnimRotPoint(),
		new Move3d(Vec3f(.525002,.08,.09)),
		new SelectNode(r_knee),

		new MoveToAnimRotPoint(),
		new Move3d(Vec3f(.530427,-.566887,.09)),
		new SelectNode(r_ankle),

		new MoveToAnimRotPoint(),
		new Move3d(Vec3f(.513688,-.819757,.09)),
		new SelectNode(l_hip),

		new MoveToAnimRotPoint(),
		new Move3d(Vec3f(.530002,.03,-.09)),
		new SelectNode(l_knee),

		new MoveToAnimRotPoint(),
		new Move3d(Vec3f(.432457,-.654811,-.09)),
		new SelectNode(l_ankle),

		new MoveToAnimRotPoint(),
		new Move3d(Vec3f(.361178,-.726758,-.09)),
		new SelectNode(neck),

		new MoveToAnimRotPoint(),
		new Move3d(Vec3f(.58054,.556081,0)),
		new SelectNode(head),

		new MoveToAnimRotPoint(),
		new Move3d(Vec3f(.570782,.626081,0)),
		new ClickButton(setFrontViewOrientation),

		new SelectNode(pelvis),

		new MoveToAnimRotPoint(),
		new Move3d(Vec3f(.5,.005,.000002)),
		new SelectNode(r_hip),

		new MoveToAnimRotPoint(),
		new Move3d(Vec3f(.41054,-.119841,.000002)),
		new SelectNode(l_hip),

		new MoveToAnimRotPoint(),
		new Move3d(Vec3f(.55946,-.065159,.000002)),
		new SelectNode(neck),

		new MoveToAnimRotPoint(),
		new Move3d(Vec3f(.430292,.457807,.06554)),
		new SelectNode(head),

		new MoveToAnimRotPoint(),
		new Move3d(Vec3f(.432615,.532756,.065782)),

		new ClickButton(setRightViewOrientation),
		new ClickButton(invertBackgroundColor),
		new DragViewRot(.002363,.999995,.002291,1.587749),
		new SelectNode(pelvis),
		new MoveToAnimTranPoint(),
		new Move3d(Vec3f(-.025,-.045,0)),

		new ClickButton(setFrontViewOrientation),
		new ToggleField("copy", 1),
		new DragAnimSlider(.75),
		new ToggleField("paste"),
		new DragAnimSlider(.5),
		new ToggleField("paste"),

		new SelectNode(r_shoulder),

		new MoveToAnimRotPoint(),
		new Move3d(Vec3f(.269089,.207605,.053353)),
		new SelectNode(r_elbow),

		new MoveToAnimRotPoint(),
		new Move3d(Vec3f(.148984,.20125,.219281)),
		new SelectNode(r_wrist),

		new MoveToAnimRotPoint(),
		new Move3d(Vec3f(.201875,.843881,.354247)),
		//////////////
		new SelectNode(pelvis),

		new MoveToAnimRotPoint(),
		new Move3d(Vec3f(.491004,-.06503,-.024897)),
		new SelectNode(r_hip),

		new MoveToAnimRotPoint(),
		new Move3d(Vec3f(.426214,-.026176,-.024897)),
		new SelectNode(pelvis),

		new MoveToAnimTranPoint(),
		//new Move3d(Vec3f(.011004,-.04503,-.024897)),
		new Move3d(Vec3f(0,-.05,-.024897)),
		new SelectNode(l_shoulder),

		new MoveToAnimRotPoint(),
		new Move3d(Vec3f(.637545,.424586,.053281)),
		new SelectNode(l_elbow),

		new MoveToAnimRotPoint(),
		new Move3d(Vec3f(.742213,.261463,.025237)),
		new SelectNode(l_wrist),

		new MoveToAnimRotPoint(),
		new Move3d(Vec3f(.688207,-.105409,-.003327)),
		new ClickButton(toggleWireFrame),
		new DragViewRot(.552273,-.814241,-.178901,.430118),

		new ToggleField("play", 1),
		new Section(Complete),

	};
	list+=EVENT_LIST(Add_a_Gesture_GroupArray);
}
