/* Seamless3d

Copyright (C) 2001-2010 Graham John Perrett (AKA thyme) thyme@seamless3d.com www.seamless3d.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this
software and associated documentation files (the "Software"), to deal in the Software
without restriction, including without limitation the rights to use, copy, modify,
merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to the following
conditions:

The above copyright notice and this permission notice shall be included in all copies or
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT
OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.
*/

#define STRICT
#include <windows.h>
#include <windowsx.h>
#include <dinput.h>

#include "seamless3d.h"
#include "techuelife_triangle_editor.h"
#include "seamless3d_gui_render.h"
#include "gui.h"


#define BUFFER_SIZE           16
#define RECORD_3D_POINTER 1

LPDIRECTINPUT8  g_lpDI; 
IDirectInput8*           g_pDI           = NULL;
IDirectInputDevice8*     g_pMouse        = NULL;
HANDLE                  g_hMouseEvent   = NULL;

extern HINSTANCE               g_hInst;
extern SettingsPtr g_settings;

bool g_zDragMode = false;
bool g_viewPosDragMode = false;
bool g_viewRotDragMode = false;
bool g_insertControlPoint = false;
Vec3f g_viewDragMousePos;
bool MouseFlag = false;
bool tugFlag = false;
bool g_tugRotFlag = false;
bool orientationTugFlag = false;
bool dragPartFlag = false;
bool g_dragMorphFlag = false;
bool g_dragActuallyHappned = false;

bool	g_mirrorModeVertexFound=false;
bool	mouse3dMode=false;
static int cursorCounter=0;
Rotation g_beginViewRot;
Vec2i g_rightDragCounter;
void get2dMousePosFrom3d();
void restoreRobotFrom3d();
Vec2i get2dMousePosFrom3d(Vec3f par);

void updateBuild3dWindow(bool ifPartBuild=false);
void backUpSmlsCoord(const M_Art& partList);
extern Ptr<PartClue> g_vcArtTriBase;
extern Ptr<AnimClue> g_vcArtAnimBase;
extern Ptr<PartCageClue> g_partCageClue;
extern TransformPtr g_vcBezierBase;
extern Ptr<Palette> g_palette;
void commonIncDecColorChange(Seamless* smls, Art* part, M_Color3f& color, const int colorIdx);


BOOL   g_MousebActive      = TRUE;
BOOL   g_bSwapMouseButtons;

//-----------------------------------------------------------------------------
// Function: SetAcquire
//
// Description: 
//      Acquire or unacquire the mouse, depending on if the app is active
//       Input device must be acquired before the GetDeviceState is called
//
//-----------------------------------------------------------------------------
HRESULT SetAcquire()
{
    // nothing to do if g_pMouse is NULL
    if (NULL == g_pMouse)
        return S_FALSE;

    if (g_MousebActive) 
    {
        g_pMouse->Acquire();
    } 
    else 
    {
        g_pMouse->Unacquire();
    }

    return S_OK;
}




//-----------------------------------------------------------------------------
// Function: FreeDirectInput
//
// Description: 
//      Initialize the DirectInput variables.
//
//-----------------------------------------------------------------------------
HRESULT FreeDirectInput()
{
    // Unacquire and release any DirectInputDevice objects.
    if ( NULL != g_pMouse ) 
    {
        // Unacquire the device one last time just in case 
        // the app tried to exit while the device is still acquired.
        g_pMouse->Unacquire();

        g_pMouse->Release();
        g_pMouse = NULL;
    }

    // Release any DirectInput objects.
    if ( NULL != g_pDI )  
    {
        g_pDI->Release();
        g_pDI = NULL;
    }

    // Close event handle
    if ( NULL != g_hMouseEvent )
    {
        CloseHandle( g_hMouseEvent );
        g_hMouseEvent = NULL;
    }

    return S_OK;
}




//-----------------------------------------------------------------------------
// Function: OnMouseInput
//
// Description: 
//      Handles responding to any mouse input that is generated from
//      the mouse event being triggered.
//
//-----------------------------------------------------------------------------

HRESULT InitDirectInput( HWND hWnd )
{
	HRESULT         hr; 
    // Create a DInput object
    if( FAILED( hr = DirectInput8Create( GetModuleHandle(NULL), DIRECTINPUT_VERSION, 
                                         IID_IDirectInput8, (VOID**)&g_pDI, NULL ) ) )
        return hr;
    
    // Obtain an interface to the system mouse device.
    if( FAILED( hr = g_pDI->CreateDevice( GUID_SysMouse, &g_pMouse, NULL ) ) )
        return hr;
    
    // Set the data format to "mouse format" - a predefined data format 
    //
    // A data format specifies which controls on a device we
    // are interested in, and how they should be reported.
    //
    // This tells DirectInput that we will be passing a
    // DIMOUSESTATE2 structure to IDirectInputDevice::GetDeviceState.
    if( FAILED( hr = g_pMouse->SetDataFormat( &c_dfDIMouse2 ) ) )
        return hr;
    
    // Set the cooperativity level to let DirectInput know how
    // this device should interact with the system and with other
    // DirectInput applications.
//    hr = g_pMouse->SetCooperativeLevel( hWnd, DISCL_EXCLUSIVE|DISCL_FOREGROUND );
//    hr = g_pMouse->SetCooperativeLevel( hWnd, DISCL_FOREGROUND | DISCL_NONEXCLUSIVE);
//    if ( FAILED(hr) ) 
//        return hr;


    // Acquire the newly created device



    hr = g_pDI->CreateDevice( GUID_SysMouse , &g_pMouse, NULL );
    if ( FAILED(hr) ) 
        return hr;
	

    // Set the data format to "mouse format" - a predefined data format 
    //
    // A data format specifies which controls on a device we
    // are interested in, and how they should be reported.
    //
    // This tells DirectInput that we will be passing a
    // DIMOUSESTATE structure to IDirectInputDevice::GetDeviceState.
    hr = g_pMouse->SetDataFormat( &c_dfDIMouse );
    if ( FAILED(hr) ) 
        return hr;

    // Set the cooperativity level to let DirectInput know how
    // this device should interact with the system and with other
    // DirectInput applications.
//    hr = g_pMouse->SetCooperativeLevel( hWnd, 
//                                        DISCL_EXCLUSIVE | DISCL_FOREGROUND);
    hr = g_pMouse->SetCooperativeLevel( hWnd,DISCL_NONEXCLUSIVE | DISCL_FOREGROUND);

    if ( FAILED(hr) ) 
        return hr;

    g_pMouse->Acquire();


    // create a win32 event which is signaled when mouse data is availible
   // g_hMouseEvent = CreateEvent( NULL, FALSE, FALSE, NULL );
   // if ( NULL == g_hMouseEvent )
    //    return E_FAIL;

    // give the event to the mouse device
    //hr = g_pMouse->SetEventNotification( g_hMouseEvent );
    //if ( FAILED(hr) ) 
     //   return hr;

    // setup the buffer size for the mouse data
    DIPROPDWORD dipdw;

    dipdw.diph.dwSize = sizeof(DIPROPDWORD);
    dipdw.diph.dwHeaderSize = sizeof(DIPROPHEADER);
    dipdw.diph.dwObj = 0;
    dipdw.diph.dwHow = DIPH_DEVICE;
    dipdw.dwData = BUFFER_SIZE;

    //hr = g_pMouse->SetProperty( DIPROP_BUFFERSIZE, &dipdw.diph );
    //if ( FAILED(hr) ) 
    //    return hr;

    // not necessary, but nice for left handed users that have
    // their swapped mouse buttons
    g_bSwapMouseButtons = GetSystemMetrics( SM_SWAPBUTTON );

    return S_OK;
}

void UnAcquire(){
	g_pMouse->Unacquire();
	if(NurbsPatch* patch = dynamic_cast<NurbsPatch*>(g_selNode.getRaw())){
		if(g_transformCP->dragRotPointMode){
			g_transformCP->quantizeRotationLever();
		}
		else if(g_transformCP->dragPivotPointMode){
			g_transformCP->pivotPoint.refValue() += g_transformCP->translation.refValue();
			g_transformCP->refreshRemPoints();
		}
		if(patch->smls->animCP == false && g_dragActuallyHappned){
			if(Anim* anim = Window::getAnimBar()){
				if(Composition* composition = dynamic_cast<Composition*>(anim->soundTrack.p)){
					composition->updateAndPlay(anim, patch);
				}
			}
		}
	}
	else if(Oscillator* osc = dynamic_cast<Oscillator*>(g_selNode.getRaw())){
		if(MouseFlag){
			if(Anim* anim = Window::getAnimBar()){
				if(Composition* composition = dynamic_cast<Composition*>(anim->soundTrack.p)){
					composition->updateAndPlay(anim, osc);
				}
			}
		}
	}
	else if(Chords* chords = dynamic_cast<Chords*>(g_selNode.getRaw())){
		if(MouseFlag){
			if(Anim* anim = Window::getAnimBar()){
				if(Composition* composition = dynamic_cast<Composition*>(anim->soundTrack.p)){
					composition->updateAndPlay(anim, chords);
				}
			}
		}
	}
	if(MultiSelect::selectRectMode){
		MultiSelect::selectRectMode=false;
		g_visualClue=g_vcBezierBase;
		if(BezierPatch* bezierPatch = dynamic_cast<BezierPatch*>(g_selNode.getRaw())){
			bezierPatch->joinEdge();
		}
		else if(NurbsPatch* nurbsPatch = dynamic_cast<NurbsPatch*>(g_selNode.getRaw())){
			nurbsPatch->smls->selectControlPoints();
			nurbsPatch->houseKeep();
		}
		else if(Art* part = dynamic_cast<Art*>(g_selNode.getRaw())){
			MultiSelect::selectTriangles(part);
			if(g_palette->rectangle){
				commonEndVertexChange();
			}
		}
	}
	else if(g_viewRotDragMode==false && g_viewPosDragMode==false && g_insertControlPoint == false){
		get2dMousePosFrom3d();
		//SetCursorPos(screenPos.x,screenPos.y);
	}
	else {
		g_viewRotDragMode = false;
		g_viewPosDragMode = false;
		g_insertControlPoint = false;
	}
	restoreRobotFrom3d();

	showCursor(TRUE);

	g_zDragMode = false;
	g_viewPosDragMode = false;
	g_viewRotDragMode = false;
	g_insertControlPoint = false;
	MouseFlag = false;
	tugFlag = false;
	g_tugRotFlag = false;
	orientationTugFlag = false;
	dragPartFlag = false;
	g_dragMorphFlag = false;
	g_vcArtAnimBase->dragTranPointMode = false;

	g_mirrorModeVertexFound = false;
	mouse3dMode = false;
}

Vec3f localDrag;
unsigned g_state=0;

bool checkRobotActive();

void commonPostTug(){
	g_scene->updateBuild();
	BackUp::monostable.beginCycle();
}
void EC2::drag(Vec3f pt){
	M_Vec2f& point=this->point;
	Vec3f svec3f= g_scene->getIViewMatrix() * translationMatrix(-location) * rotationMatrix(-orientation) * pt;
	//Vec3f svec3f=getIViewMatrix() * pt;
	int idx=index[0];
	if(orientationTugFlag==true){
		float diffX=point[idx].x-svec3f.x;
		float diffY=point[idx].y-svec3f.y;

		tangent[idx]=atan2(diffY,diffX);
	}
	else{
		point[idx]=Vec2f(svec3f.x,svec3f.y);
	}
	g_selNode->processFields();
	ecHouseKeep();
	commonPostTug();
}
void EC3::drag(Vec3f pt){
	M_Vec2f& point=this->point;
	Vec3f svec3f= g_scene->getIViewMatrix() * translationMatrix(-location) * rotationMatrix(-orientation) * pt;
	//Vec3f svec3f=getIViewMatrix() * pt;
	int idx=index[0];
	if(orientationTugFlag==true){
		float diffX=point[idx].x-svec3f.x;
		float diffY=point[idx].y-svec3f.y;

		tangent[idx]=atan2(diffY,diffX);
	}
	else{
		point[idx].x=svec3f.x;
		point[idx].y=svec3f.y;
	}
	g_selNode->processFields();
	ecHouseKeep();
	commonPostTug();
}
void Bezier::drag(Vec3f pt){
	M_Vec2f& point=this->point;
	Vec3f vec= g_scene->getIViewMatrix() * imatrix * pt;
	matchBorderTangents(Vec2f(vec.x,vec.y));
	g_selNode->processFields();
	bezHouseKeep();
	commonPostTug();
}
void Bezier3f::drag(Vec3f pt){
	M_Vec3f& point=this->point;
	Vec3f vec= g_scene->getIViewMatrix() * imatrix * pt;
	matchBorderTangents(Vec3f(vec.x,vec.y,vec.z));
	g_selNode->processFields();
	bezHouseKeep();
	commonPostTug();
}
void BezierTape::drag(Vec3f pt){
	M_Vec2f& point=this->point;
	Vec3f vec= g_scene->getIViewMatrix() * rotationMatrix(Rotation(0,1,0,PI*.5)) * pt;
	matchBorderTangents(Vec2f(vec.x,vec.y));
	g_selNode->processFields();
	bezHouseKeep();
	commonPostTug();
}
int nextLink(int bp){
	bp+=4;
	if(bp==16){
		bp=0;
	}
	return bp;
}
Vec3f setOppositePoint(Vec3f hinge, Vec3f swingA, Vec3f swingB){
	float disA = (swingA - hinge).length();
	float disB = (swingB - hinge).length();
	float ratio = disB / disA;
//	return hinge + (swingA - hinge) * Vec3f(-ratio, -ratio, -ratio);

	return hinge + (hinge - swingA) * Vec3f(ratio, ratio, ratio);;
}
void setCorner(M_Vec3f& point, const M_Int& cornerLink, const int bp, const Vec3f vecA){
	Vec3f vecB = point[cornerLink[bp+1]];
	Vec3f vecC = point[cornerLink[bp+3]];

//	point[cornerLink[bp + 1]] = vecB;
//	point[cornerLink[bp + 3]] = vecC;

	point[cornerLink[bp + 2]] = vecB + (vecC - vecA);

//	point[cornerLink[nextLink(bp) + 1]] = point[cornerLink[bp + 3]];
}
void setCorner0(M_Vec3f& point, const M_Int& cornerLink, const int bp, const Vec3f vecA){
	Vec3f vecB = point[cornerLink[bp+2]];
	Vec3f vecC = point[cornerLink[bp+3]];

//	point[cornerLink[bp + 1]] = vecB;
//	point[cornerLink[bp + 3]] = vecC;

	point[cornerLink[bp + 1]] = vecB + (vecA - vecC);

//	point[cornerLink[nextLink(bp) + 1]] = point[cornerLink[bp + 3]];
}

void setCorners(M_Vec3f& point, const M_Int& cornerLink, const Vec3f pivot){
	setCorner(point, cornerLink, 0, pivot);
	setCorner(point, cornerLink, 4, pivot);
	setCorner(point, cornerLink, 8, pivot);
	setCorner(point, cornerLink, 12, pivot);
}
void ensureCoplaner(M_Vec3f& point, const M_Int& cornerLink){
	const int b0=0;
	const int b1=4;
	const int b2=8;
	const int b3=12;

	Vec3f normal = calcNormal(point[cornerLink[b0+3]], point[cornerLink[b0+1]], point[cornerLink[b0+0]]);
	Vec3f rightAngle = calcNormal(point[cornerLink[b0+3]], point[cornerLink[b0+0]], point[cornerLink[b0+0]]+normal);

	const Vec3f centerPoint = point[cornerLink[0]];

	{
		float dis = (point[cornerLink[b0+1]] - centerPoint).length();
		point[cornerLink[b0+1]]= rightAngle * dis + centerPoint;
	}
	{
		float dis = (point[cornerLink[b1+3]] - centerPoint).length();
		point[cornerLink[b1+3]] = -rightAngle * dis + centerPoint;
	}
	{
		point[cornerLink[b2+3]] = setOppositePoint(centerPoint, point[cornerLink[b0+3]], point[cornerLink[b2+3]]);
		point[cornerLink[b2+1]] = point[cornerLink[b1+3]];
	}
	{
		point[cornerLink[b3+1]] = point[cornerLink[b2+3]];
		point[cornerLink[b3+3]] = point[cornerLink[b0+1]];
	}

	setCorners(point, cornerLink, centerPoint);
}

void rotatePoint(M_Vec3f& point, const M_Int& cornerLink, int bp, Rotation rot, Vec3f pivot){
	point[cornerLink[bp+3]] = rot * (point[cornerLink[bp+3]] - pivot) + pivot;
}
void dragJoinedCornerEdge(int idx, Vec3f mousePos, M_Vec3f& point, const M_Int& cornerLink, int b0){
	int b1 = nextLink(b0);
	int b2 = nextLink(b1);
	int b3 = nextLink(b2);

	Vec3f pivot = point[cornerLink[b0+0]];
	Rotation rot = Rotation(point[idx]-pivot, mousePos-pivot);
	point[cornerLink[b3+3]] = mousePos;
	point[cornerLink[b0+1]] = mousePos;

	rotatePoint(point, cornerLink, b0, rot, pivot);
	rotatePoint(point, cornerLink, b1, rot, pivot);
	rotatePoint(point, cornerLink, b2, rot, pivot);

	point[cornerLink[b1+1]] = point[cornerLink[b0+3]];
	point[cornerLink[b2+1]] = point[cornerLink[b1+3]];
	point[cornerLink[b3+1]] = point[cornerLink[b2+3]];
	point[cornerLink[b3+3]] = point[cornerLink[b0+1]];

	ensureCoplaner(point, cornerLink);
}
void setPoint(M_Vec3f& point, const M_Int& cornerLink, const int bp, const Vec3f& mousePos, const Vec3f diff){
	point[cornerLink[bp+0]] = mousePos;
	//point[cornerLink[bp+2]] += diff;
	point[cornerLink[bp+3]] += diff;
	point[cornerLink[nextLink(bp) + 1]] = point[cornerLink[bp + 3]];
}
bool BezierPatch::doCornerLink(M_Vec3f& point, const int idx, const Vec3f mousePos){
	static M_Int cornerLink;
	const M_Int& link = this->link;
	
	cornerLink.length=4*4;

	for(int l = 0; l<link.length; ){
		for(int k = 0; k < 4; k++){
			int i = linkToPatchIdx[l];
			int noU=dim[i].x;
			int noV=dim[i].y;
			int pointLen=noU*noV;

			int j = link[l++];
			int h = j - m_bp[i];
			cornerLink[k*4+0] = j;
			if(h == 0){
				cornerLink[k*4+1] = j+1;
				cornerLink[k*4+2] = j+1+noU;
				cornerLink[k*4+3] = j+noU;
			}
			else if(h == noU - 1){
				cornerLink[k*4+1] = j+noU;
				cornerLink[k*4+2] = j+noU-1;
				cornerLink[k*4+3] = j-1;
			}
			else if(h == pointLen - noU){
				cornerLink[k*4+1] = j-noU;
				cornerLink[k*4+2] = j-noU+1;
				cornerLink[k*4+3] = j+1;
			}
			else if(h == pointLen - 1){
				cornerLink[k*4+1] = j-1;
				cornerLink[k*4+2] = j-1-noU;
				cornerLink[k*4+3] = j-noU;
			}
		}


		const int b0=0;
		const int b1=4;
		const int b2=8;
		const int b3=12;
		Vec3f diff;

		if(	cornerLink[b0] == idx){
			diff = mousePos-point[cornerLink[b0]];
			goto here;
		}
		else if(cornerLink[b1] == idx){
			diff = mousePos-point[cornerLink[b1]];
			goto here;
		}
		else if(cornerLink[b2] == idx){
			diff = mousePos-point[cornerLink[b2]];
			goto here;
		}
		else if(cornerLink[b3] == idx){
			diff = mousePos-point[cornerLink[b3]];
here:		
			point[cornerLink[b0+0]] = mousePos;
			point[cornerLink[b1+0]] = mousePos;
			point[cornerLink[b2+0]] = mousePos;
			point[cornerLink[b3+0]] = mousePos;


			point[cornerLink[b0+1]] += diff;
			point[cornerLink[b0+3]] += diff;
			point[cornerLink[b1+3]] += diff;
			point[cornerLink[b2+3]] += diff;

			point[cornerLink[nextLink(b0) + 1]] = point[cornerLink[b0 + 3]];

			ensureCoplaner(point, cornerLink);

			//point[cornerLink[4*2+3]] = setOppositePoint(point[cornerLink[0]], point[cornerLink[3]], point[cornerLink[4*2+3]]);
			return true;
		}
		else if(cornerLink[b0+1] == idx || cornerLink[b3+3] == idx){
			dragJoinedCornerEdge(idx, mousePos, point, cornerLink, b0);
			return true;
		}
		else if(cornerLink[b0+3] == idx || cornerLink[b1+1] == idx){
			dragJoinedCornerEdge(idx, mousePos, point, cornerLink, b1);
			return true;
		}
		else if(cornerLink[b1+3] == idx || cornerLink[b2+1] == idx){
			dragJoinedCornerEdge(idx, mousePos, point, cornerLink, b2);
			return true;
		}
		else if(cornerLink[b2+3] == idx || cornerLink[b3+1] == idx){
			dragJoinedCornerEdge(idx, mousePos, point, cornerLink, b3);
			return true;
		}
	/*
		else if(cornerLink[2] == idx){
			Vec3f pivot = point[cornerLink[b0+0]];
			Rotation rot = Rotation(point[idx]-pivot, mousePos-pivot);
			point[idx] = mousePos;

			rotatePoint(point, cornerLink, b0, rot, pivot);
			rotatePoint(point, cornerLink, b1, rot, pivot);
			rotatePoint(point, cornerLink, b2, rot, pivot);
			rotatePoint(point, cornerLink, b3, rot, pivot);

			point[cornerLink[b0+1]] = point[cornerLink[b3+3]];
			point[cornerLink[b1+1]] = point[cornerLink[b0+3]];
			point[cornerLink[b2+1]] = point[cornerLink[b1+3]];
			point[cornerLink[b3+1]] = point[cornerLink[b2+3]];

			Vec3f normal = calcNormal(point[cornerLink[b0+3]], point[cornerLink[b0+2]], pivot);
			Rotation rotB = Rotation(normal, Vec3f(0,0,1));

	//		point[cornerLink[1]] = rot * (point[cornerLink[1]] - pivot) + pivot;
			point[cornerLink[2]] = rotB * (point[cornerLink[2]] - pivot) + pivot;
			point[cornerLink[3]] = rotB * (point[cornerLink[3]] - pivot) + pivot;

			Rotation rotC = Rotation(point[cornerLink[b0+3]]-pivot, Vec3f(0,-1,0));
			point[cornerLink[2]] = rotC * (point[cornerLink[2]] - pivot) + pivot;
			point[cornerLink[3]] = rotC * (point[cornerLink[3]] - pivot) + pivot;

			point[cornerLink[b0+3]].y=point[cornerLink[2]].y;

			point[cornerLink[2]] = -rotC * (point[cornerLink[2]] - pivot) + pivot;
			point[cornerLink[3]] = -rotC * (point[cornerLink[3]] - pivot) + pivot;

			point[cornerLink[2]] = -rotB * (point[cornerLink[2]] - pivot) + pivot;
			point[cornerLink[3]] = -rotB * (point[cornerLink[3]] - pivot) + pivot;
			
			setCorner0(point, cornerLink, 0, pivot);

			//		rotatePoint(point, cornerLink, b0, rot, pivot);
	//		rotatePoint(point, cornerLink, b1, rot, pivot);
	//		rotatePoint(point, cornerLink, b2, rot, pivot);
	//		rotatePoint(point, cornerLink, b3, rot, pivot);

	//		point[cornerLink[b0+1]] = point[cornerLink[b3+3]];
	//		point[cornerLink[b1+1]] = point[cornerLink[b0+3]];
	//		point[cornerLink[b2+1]] = point[cornerLink[b1+3]];
	//		point[cornerLink[b3+1]] = point[cornerLink[b2+3]];

			//Vec3f rightAngle = calcNormal(point[cornerLink[b0+2]], pivot+normal, pivot);
			//point[cornerLink[b1+2]]
			//Vec3f v = (Vec3f(1,1,1)-rightAngle)*((point[cornerLink[b0+2]]-pivot).length())+pivot;
			//Vec3f v = rightAngle*(point[cornerLink[b0+2]]-pivot)+pivot;
			//point[cornerLink[b1+2]]=v;
			//Vec3f diff = (v - point[cornerLink[b0+2]]) * .5;
			//point[cornerLink[b0+3]] = point[cornerLink[b1+1]] = point[cornerLink[b0+2]] + diff;
			//Vec3f vecA = pivot;
			//Vec3f vecB = point[cornerLink[b0+2]];
			//Vec3f vecC = point[cornerLink[b0+3]];
			//point[cornerLink[b0 + 1]] = vecB + (vecA - vecC);
	//		setCorners(point, cornerLink, pivot);
			return true;
		}
	*/
	}
	return false;
}
void BezierPatch::doLink(M_Vec3f& point, const M_Int& link, int idx, Vec3f diff, int exclude){

	for(int i=link.length; (i-=4)>=0;){
		if(i==exclude)continue;

		if(link[i+hingeB]==idx){
			int sA=link[i+swingA];
			int hA=link[i+hingeA];
			int hB=link[i+hingeB];
			int sB=link[i+swingB];

			point[hA]=point[hB];
			point[sA]+=diff;
			point[sB]=point[hA]+(point[hA]-point[sA]);

			//doLink(point, link, sA, Vec3f(), hA);
			//doLink(point, link, sB, Vec3f(), hA);
		}
		else if(link[i+hingeA]==idx){
			int sA=link[i+swingA];
			int hA=link[i+hingeA];
			int hB=link[i+hingeB];
			int sB=link[i+swingB];

			point[hB]=point[hA];
			point[sB]+=diff;
			point[sA]=point[hB]+(point[hB]-point[sB]);

			//doLink(point, link, sA, Vec3f(), hB);
			//doLink(point, link, sB, Vec3f(), hB);
		}
		else if(link[i+swingA]==idx){
			int sB=link[i+swingB];
			int sA=link[i+swingA];
			int h=link[i+hingeB];

			point[sB]=point[h]+(point[h]-point[sA]);
//			doLink(point, link, sA, Vec3f(), sB);
		}
		else if(link[i+swingB]==idx){
			int sB=link[i+swingB];
			int sA=link[i+swingA];
			int h=link[i+hingeB];

			point[sA]=point[h]+(point[h]-point[sB]);
//			doLink(point, link, sB, Vec3f(), sA);
		}
	}
}
void BezierPatch::refreshLink(M_Int& link){
	return;
	for(int i=link.length; (i-=4)>=0;){
		for(int j=i; (j-=4)>=0;){
			if(link[j+hingeA]==link[i+hingeA] && link[j+hingeB] != link[i+hingeA]){
				link[j+hingeA]=link[i+hingeB];
			}
			if(link[j+hingeB]==link[i+hingeA] && link[j+hingeA] != link[i+hingeA]){
				link[j+hingeB]=link[i+hingeB];
			}
		}
	}
}
/*
void SineTug::drag(Vec3f pt){
	int i=index[0];
	const M_Vec3f& location = this->location;
	M_Rotation& orientation = this->orientation.refValue();
	M_Float& distance = this->distance.refValue();

	Vec3f displacement = g_scene->getIViewMatrix() * pt - location[i];
	distance[i] = displacement.length() /8;
	orientation[i] = Rotation(Vec3f(0,1,0), displacement);

	commonPostTug();
	houseKeep();
}
*/
void SineTug::drag(Vec3f pt){
	int i=index[0];
	const M_Vec3f& location = this->location;
	M_Rotation& orientation = this->orientation.refValue();
	M_Float& distance = this->distance.refValue();

	Matrix matrix = rotationMatrix(orientation[i]) * translationMatrix(location[i]) * g_scene->getViewMatrix();

	Vec3f pivotPoint = matrix * Vec3f();

	Vec3f controlPoint;
	{
		Vec3f vec;
		float* ptr=0;
		ptr=&distance[i];
		controlPoint = matrix * Vec3f(0, *ptr*8, 0);

		Vec3f loc = location[i];
		Rotation orien = orientation[i];

		Matrix imatrix = translationMatrix(Vec3f(-loc.x, -loc.y, -loc.z)) * rotationMatrix(-orien);

		Vec3f newPt = g_scene->getIViewMatrix() * imatrix * pt;

		if(distance[i]>=0){
			*ptr = sqrt(newPt.x*newPt.x + newPt.y*newPt.y + newPt.z*newPt.z)/8;
		}
		else{
			*ptr = -sqrt(newPt.x*newPt.x + newPt.y*newPt.y + newPt.z*newPt.z)/8;
		}

		{

			//would there not be an easier way to calculate the orientation with the information we have at this point in time?

			Vec3f toDiff = Vec3f(pt.x-pivotPoint.x, pt.y-pivotPoint.y, pt.z-pivotPoint.z);
			Vec2f fromDiff = Vec2f(controlPoint.x-pivotPoint.x, controlPoint.y-pivotPoint.y);
			float toAngle = Vec2f(toDiff.x, toDiff.y).getAngle();
			float fromAngle = fromDiff.getAngle();
			float angle = toAngle - fromAngle;

			Vec3f axis = (-g_scene->rotation * Vec3f(0, 0, 1)).normalize();
			orientation[i] = orientation[i] * Rotation(axis, angle);
		}
		Vec3f sr;
		{
			sr= Vec3f(0, distance[i]*8, 0);
		}
		Matrix postMatrix = rotationMatrix(orientation[i]) * translationMatrix(location[i]) * g_scene->getViewMatrix();
		g_pointer.translation= postMatrix * sr;
	}
	houseKeep();
	commonPostTug();
}


void BezierPatch::drag(Vec3f pt){
	M_Vec3f& point=this->point;
	Vec3f vec= g_scene->getIViewMatrix() * pt;
	int idx=index[0];

	if(doCornerLink(point, idx, vec) == false){
		point[idx] = vec;
	}

	processFields();
	commonPostTug();
	houseKeep();
}
void Subdivision::drag(Vec3f pt){
	M_Vec3f& point=this->point;
	Vec3f vec= g_scene->getIViewMatrix() * pt;
	int idx=index[0];
	point[idx]=vec;
	commonPostTug();
	houseKeep();
}
int findMirrorVertex(const M_Vec3f& cPtr, Vec3f svec);

inline bool NurbsPatch::doSameViewLocation(const Vec3f& pt, const Vec3f& svec, const int& c2, const int& r, const int& xDim, const int& yDim, M_Vec3f& point, const M_Vec3f& dpoint, const M_Float& weight, const Matrix& imatrix, const Matrix& mtrx){
	int idx2 = r * xDim + c2;
	Vec3f svec2 = mtrx * dpoint[idx2];
	if(0.001 > abs(svec.x - svec2.x) && 0.001 > abs(svec.y - svec2.y)){
		Vec3f pt2 = pt;
		pt2.z = svec2.z;
		Vec3f vec2 = g_scene->getIViewMatrix() * imatrix * pt2;

		if(mirror){
			Vec3f svec = dpoint[idx2];
			if(svec.x==0){
				vec2.x=0;
			}
			else{
				svec.x=-svec.x;
				int mirrorIdx = findMirrorVertex(dpoint, svec);
				if(mirrorIdx>=0){
					svec=vec2;
					svec.x=-svec.x;
					//point[mirrorIdx]=svec*weight[mirrorIdx];
					setFilteredDimension(point, mirrorIdx, svec*weight[mirrorIdx]);
				}
			}
		}


		//vec2 = Vec3f();
		//point[idx2] = vec2 * weight[idx2];
		setFilteredDimension(point, idx2, vec2 * weight[idx2]);
//		String str;
//		vec2.toString(str);
//		dLog(str);
		return true;
	}
	return false;
}


void processAllPatches(M_<NurbsPatch*>& cpNodeList);
void updateCage(Seamless* smls);

Vec2f getScreenCoordinate(const bool perspective, Vec3f sr){
	Vec2f v;

	if(perspective){
		v.x = sr.x / (-sr.z + g_viewDist);
		v.y = sr.y / (-sr.z + g_viewDist);
	}
	else{
		v.x = sr.x;
		v.y = sr.y;
	}
	return v;
}

extern M_Vec3f g_backUpCoord;


//	Vec3f vec= g_scene->getIViewMatrix() * imatrix * pt;
//
//	if(g_visualClue == g_vcRotateEditBase){
//		g_vcRotateEditBase->drag(pt);
//		return;
//	}
//	else{
//		if(sameViewPositionTug){
//			int incAmount = 1;
//			if(showAll == false){
//				incAmount++;
//			}
//			Matrix mtrx = g_viewMatrix * matrix;
//			Vec3f svec = mtrx * dpoint[idx];
//			int xDim = dim.x;
//			int yDim = dim.y;
//			int c = idx%xDim;
//			int r = idx/xDim;
//			int c2 = c + incAmount;
//			if(c2 > xDim){
//				c2 = xDim - c2; 
//			}
//			if(c2 < 0){
//				c2 += xDim - 1;
//			}
//			if(!doSameViewLocation(pt, svec, c2, r, xDim, yDim, point, dpoint, weight, imatrix, mtrx)){
//				int c2 = c - incAmount;
//				if(c2 < 0){
//					c2 += xDim - 1;
//				}
//				doSameViewLocation(pt, svec, c2, r, xDim, yDim, point, dpoint, weight, imatrix, mtrx);
//			}
//		}
//		//if(mirrorTug){
//		//	Vec3f svec = dpoint[idx];
//		//	if(mirrorPatch){
//		//	}
//		//	else{
//		//		if(svec.x==0){
//		//			vec.x=0;
//		//		}
//		//		else{
//		//			svec.x=-svec.x;
//		//			int mirrorIdx = findMirrorVertex(dpoint, svec);
//		//			if(mirrorIdx>=0){
//		//				svec=vec;
//		//				svec.x=-svec.x;
//		//				//point[mirrorIdx]=svec*weight[mirrorIdx];
//		//				setFilteredDimension(point, mirrorIdx, svec*weight[mirrorIdx]);
//		//			}
//		//		}
//		//	}
//		//}
//		//point[idx]=vec*weight[idx];
////		setFilteredDimension(point, idx, vec*weight[idx]);
//
//
//		if(MultiSelect::pointList.length){
//			Vec3f newPt = g_scene->getIViewMatrix() * g_catchIMatrix * imatrix * pt;
//			Vec3f diff = newPt - MultiSelect::beginDragPos[MultiSelect::clickedIndex];
//			for(int i = MultiSelect::pointList.length; --i>=0;){
//				int idx = MultiSelect::pointList[i].i;
//				NurbsPatch* patch = MultiSelect::pointList[i].p;
//				Vec3f svec = MultiSelect::beginDragPos[i] + diff;
//				patch->setFilteredDimension(patch->point.refValue(), idx, svec * patch->weight[idx]);
//				patch->dpoint.refValue()[idx] = patch->point[idx] / patch->weight[idx];
//				if(patch->mirror){
//					if(NurbsPatch* mirrorPatch = patch->mirrorPatch){
//						Vec3f vec = patch->point[idx];
//						vec.x = -vec.x;
//						mirrorPatch->point.refValue()[idx] = vec;
//						vec = patch->dpoint[idx];
//						vec.x = -vec.x;
//						mirrorPatch->dpoint.refValue()[idx] = vec;
//					}
//					else{
//						int uDim = patch->dim.x;
//						int vDim = patch->dim.y;
//						int u = idx % uDim;
//						int v = idx / uDim;
//						int middleIdx = (uDim >> 1);
//
//						if(u == middleIdx){
//							patch->point.refValue()[idx].x = 0;
//							patch->dpoint.refValue()[idx].x = 0;
//						}
//						else{
//							int midx = v * uDim + uDim - 1 - u;
//							Vec3f vec = patch->point[idx];
//							vec.x = -vec.x;
//							patch->point.refValue()[midx] = vec;
//							vec = patch->dpoint[idx];
//							vec.x = -vec.x;
//							patch->dpoint.refValue()[midx] = vec;
//						}
//					}
//				}
//			}
//			smls->calcAllPolyMidPoints();
//			commonPostTug();
//			updateCage(smls);
//			houseKeep();
//			return;
//		}
//	}
//
//	processFields();
//	commonPostTug();
//	houseKeep();
//}

void Oscillator::drag(Vec3f pt){
	M_Vec2f& fPoint = this->note;
	Vec3f vec= g_scene->getIViewMatrix() * pt;
	int idx = index[0];
	if(idx == 0){
		vec.x = 0;
	}
	float w = 1;
	if(fWeight.length){
		w = fWeight[idx];
	}
	fPoint[idx] = Vec2f(vec.x, vec.y*10) * w;
	processFields();
	commonPostTug();
	houseKeep();
	g_dragActuallyHappned = true;
}

void Envelope::drag(Vec3f pt){
	M_Vec2f& point = this->point;
	Vec3f vec= g_scene->getIViewMatrix() * pt;
	int idx = index[0];
	if(idx == 0){
		vec.x = 0;
	}
	float w = 1;
	if(weight.length){
		w = weight[idx];
	}
	point[idx] = Vec2f(vec.x, vec.y*10) * w;
	processFields();
	commonPostTug();
	houseKeep();
	g_dragActuallyHappned = true;
}

void AmplitudeModulator::drag(Vec3f pt){
	M_Vec2f& point = this->wet;
	Vec3f vec= g_scene->getIViewMatrix() * pt;
	int idx = index[0];
	if(idx == 0){
		vec.x = 0;
	}
	float w = 1;
	if(weight.length){
		w = weight[idx];
	}
	point[idx] = Vec2f(vec.x, vec.y*10) * w;
	processFields();
	commonPostTug();
	houseKeep();
	g_dragActuallyHappned = true;
}

void FrequencyModulator::drag(Vec3f pt){
	M_Vec2f& point = this->wet;
	Vec3f vec= g_scene->getIViewMatrix() * pt;
	int idx = index[0];
	if(idx == 0){
		vec.x = 0;
	}
	float w = 1;
	if(weight.length){
		w = weight[idx];
	}
	point[idx] = Vec2f(vec.x, vec.y*10) * w;
	processFields();
	commonPostTug();
	houseKeep();
	g_dragActuallyHappned = true;
}

void Chords::drag(Vec3f pt){
	M_Vec2f& point = this->spectrum;
	Vec3f vec= g_scene->getIViewMatrix() * pt;
	int idx = index[0];
	if(idx == 0){
		vec.x = 0;
	}
	float w = 1;
	if(weight.length){
		w = weight[idx];
	}
	point[idx] = Vec2f(vec.x, vec.y*10) * w;
	processFields();
	commonPostTug();
	houseKeep();
	g_dragActuallyHappned = true;
}


NurbsLathe* getMirrorLatheLoop(M_Build& build, const String& mirrorName){
	for(int k = 0; k < build.length; k++){
		Build* bld = build[k];
		if(NurbsLathe* lathe = dynamic_cast<NurbsLathe*>(bld)){
			if(Art* part = lathe->part.refValue().p){
				if(part->name == mirrorName){
					return lathe;
				}
			}
		}
		else if(GroupBuild* groupBuild = dynamic_cast<GroupBuild*>(bld)){
			if(NurbsLathe* lathe = getMirrorLatheLoop(groupBuild->build.refValue(), mirrorName)){
				return lathe;
			}
		}
	}
	return 0;
}

NurbsLathe* NurbsLathe::getMirrorLathe()const{
	if(const Art* part = this->part.refValue().p){
		if(part->mirrorPart){
			return getMirrorLatheLoop(part->smls->build.refValue(), part->mirrorPart->name);
		}
	}
	return 0;
}

SphereBuild* getMirrorSphereLoop(M_Build& build, const String& mirrorName){
	for(int k = 0; k < build.length; k++){
		Build* bld = build[k];
		if(SphereBuild* sphere = dynamic_cast<SphereBuild*>(bld)){
			if(Art* part = sphere->part.refValue().p){
				if(part->name == mirrorName){
					return sphere;
				}
			}
		}
		else if(GroupBuild* groupBuild = dynamic_cast<GroupBuild*>(bld)){
			if(SphereBuild* sphere = getMirrorSphereLoop(groupBuild->build.refValue(), mirrorName)){
				return sphere;
			}
		}
	}
	return 0;
}

SphereBuild* SphereBuild::getMirrorSphere()const{
	if(const Art* part = this->part.refValue().p){
		if(part->mirrorPart){
			return getMirrorSphereLoop(part->smls->build.refValue(), part->mirrorPart->name);
		}
	}
	return 0;
}


void NurbsLathe::drag(Vec3f pt){
	M_Vec2f& point = this->point.refValue();
	int idx=index[0];
	Vec3f vec= g_scene->getIViewMatrix() * imatrix * pt;
	if(beginSP && idx == 0){
		vec.x = 0;
	}
	if(endSP && idx == point.length - 1){
		vec.x = 0;
	}

	float w = 1;
	if(weight.length){
		w = weight[idx];
	}
	point[idx] = Vec2f(vec.x, vec.y) * w;

	if(NurbsLathe* mirrorLathe = getMirrorLathe()){
		mirrorLathe->point.refValue()[idx] = Vec2f(vec.x, vec.y).mirrorX() * w;
		mirrorLathe->processFields();
	}
	processFields();
	commonPostTug();
	houseKeep();
}

void BendyLathe::drag(Vec3f pt){
	Vec3f vec = g_scene->getIViewMatrix() * imatrix * pt;
	M_Vec3f& point=this->point;

	if(showRingControlPoints){
		int idx=index[2];
		if(radIdx<0){
			dragRingBase(pt, idx, point, weight, knot, minT(), maxT(), ringBase[idx]);
		}
		else{
			dragRadius(pt, idx);
		}
//		M_Float& radius = this->radius;
//		M_Float& at = this->at;
	}
	else{
		int idx=index[0];
		float w = 1;
		if(weight.length){
			w = weight[idx];
		}
		point[idx]=vec*w;
	}
	processFields();
	commonPostTug();
	houseKeep();
}

void NurbsTrack::drag(Vec3f pt){
	Vec3f vec = g_scene->getIViewMatrix() * pt;
	M_Vec3f& point = this->keyValue.refValue();

	int idx = index[1];
	float w = 1;
	if(weight.length){
		w = weight[idx];
	}
	point[idx] = vec * w;

	processFields();
	commonPostTug();
	houseKeep();
	if(Anim* anim = Window::animMode()){
		anim->update = true;
		anim->renderNodeFirst();
		anim->update = false;
	}
}

void Harmonic::drag(Vec3f pt){
}

void HarmonicPatch::drag(Vec3f pt){
	M_Vec3f& point=this->point;
	Vec3f vec= g_scene->getIViewMatrix() * pt;
	int idx=index[0];
	point[idx]=vec*weight[idx];
	processFields();
	commonPostTug();
	houseKeep();
}
void ColorSweep::drag(Vec3f pt){
	int idx=index[0];
	M_Float& distance=this->distance;
	if(idx==0){
		Vec3f vec= g_scene->getIViewMatrix() * pt;
		location.refValue()=vec;
	}
	else{
		Vec3f vec= g_scene->getIViewMatrix() * imatrix * pt;
		for(int i=0; i<idx-1; i++){
			vec.y-=distance[i];
		}
		distance[idx-1]=vec.y;
	}
	processFields();
	commonPostTug();
	houseKeep();
}
void CCLathe::drag(Vec3f pt) 
{
	int idx = index[0];
	Vec3f vec = g_scene->getIViewMatrix() * pt;
	Vec2f pos = Vec2f(vec.x, vec.y);

	// adjust angle  
	if ( angleDrag )  {
		if ( idx == 0 ) {
			pos -= seg[idx].Pf;
			double a = pos.angle() / (PI / 4);
			if ( 0x8000 & GetKeyState(VK_LSHIFT) ) {
				double ia = floor(a + 0.5);
				if ( fabs(a - ia) * controlLen < 0.05 ){
					a = ia;
				}
			}
			curve.refValue()[idx] = a * (PI / 4);
		}
		else {
			bool notches = (0x8000 & GetKeyState(VK_LSHIFT)) != 0;
			double ro = seg[idx].Ro.length();
			double ao = seg[idx].Ro.angle();
			Vec2f Rf = pos - seg[idx].C;
			Rf *= (ro / Rf.length());
			double a = fmod(Rf.angle() - ao + 2*PI, 2*PI);

			if ( notches ) {
				double zoom = g_scene->getZoom();
				double s = sin(Rf.angle());
				double ae = 0.05 / (ro * zoom + controlLen);

				if ( a - ae < 0 || a + ae > 2 * PI ) {
					a = 2 * PI;
				}
				else if ( fabs(Rf.x + seg[idx].C.x) * zoom < 0.05 * fabs(s) ) {
					Rf.x = -seg[idx].C.x;
					Rf.y = sqrt(ro * ro - Rf.x * Rf.x) * (s < 0 ? -1 : 1);
					a = fmod(Rf.angle() - ao + 2*PI, 2*PI);
				}
				else {
					double aa = Rf.angle() / (PI / 4);
					double ia = floor(aa + 0.5);
					if ( fabs(aa - ia) < ae ){
						a = fmod(ia * (PI / 4) - ao + 2*PI, 2*PI);						
					}
				}
			}

			if ( curve[idx] < 0 )
				a -= 2 * PI;

			if ( fabs(a) < 0.0001 )
				a = (curve[idx] < 0) ? -2 * PI : 2 * PI;

			curve.refValue()[idx] = a;
			distance.refValue()[idx - 1] = abs(a * ro);
		}
	}
	else
	{
		if ( 0x8000 & GetKeyState(VK_LSHIFT) ) {
			float zoom = g_scene->getZoom();
			if ( abs(pos.x) * zoom < 0.05 ) {
				pos.x = 1e-4;			
			}
		}
		setPoint(idx, pos);
	}

	g_selNode->processFields();
	commonPostTug();
	houseKeep();
}

void CCLathe::addControlPoint( Vec3f pt ) {

	Vec3f vec = g_scene->getIViewMatrix() * pt;
   Vec2f pos = Vec2f(vec.x, vec.y);

	// don't add a control point if it is very close to an existing one
	for(int i = 0; i < seg.length; i++ ){
		if((seg[i].Pf - pos).length() < 0.0001 ){
			return;
		}
	}


   Stem* stem = getStemOwner();
	if (stem!=nullPtr){
		if (stem->bands.length == 0){
			stem->sync.refValue() = true;
		}
	}

	int idx = max(0, index[0]);
	int oldLen = curve.length;
	int newLen = max(2, oldLen + 1);

   curve.resize(newLen);
   curve.set(&curve);
   distance.resize(newLen - 1);
   distance.set(&distance);

   if ( (idx < oldLen - 1) && (0x8000 & GetKeyState(VK_LSHIFT)) ) {
      idx++;
      for ( int i = newLen; --i > idx; )
         curve.refValue()[i] = curve[i - 1];
      for ( int i = newLen - 1; --i > idx - 1; )
         distance.refValue()[i] = distance[i - 1];
   }
   else
   {
   	idx = newLen - 1;

      autoClose = false;
      if ( oldLen == 0 )
      {
         curve.refValue()[0] = - PI / 2;
      }
   }
	index[0] = idx;
   setPoint(idx, pos);

	if ( stem && stem->sync == true ){
		int stemIndex = max(0, idx - 1);
		stem->index[0] = stemIndex;

		M_Int& bands = stem->bands;
		getBandAndYDist(bands, Stem::bandDist, stem->yDistance);
		bands.length = stem->yDistance.length;

      for ( int i = bands.length; --i > stemIndex; ) {
         bands[i] = bands[i - 1];
      }
		int lastVal = 5;
		for (int i = 0; i < bands.length; i++){
			if (i == stemIndex || bands[i] == 0) {
				bands[i] = lastVal;
			}
			lastVal = bands[i];
		}
	}

	stem->convertYDistanceToHeight();
	processFields();
	commonPostTug();
	updateBuild3dWindow();
}

void CCLathe::deleteControlPoint()
{
	int oldLen = curve.length;
	if ( oldLen == 0 )
	{
		return;
	}

	int newLen = oldLen > 2 ? oldLen - 1 : 0;
	int idx = min(index[0], oldLen - 1);

	if ( (idx > 0) && (idx < oldLen - 1) ) {
		for ( int i = idx; i < oldLen - 1; i++ )
			curve.refValue()[i] = curve[i + 1];
		for ( int i = idx - 1; i < oldLen - 2; i++ )
			distance.refValue()[i] = distance[i + 1];
	}
	else {
		idx = max(0, newLen - 1);
	}

	curve.resize(newLen);
	curve.set(&curve);
	distance.resize(max(0, newLen - 1));
	distance.set(&distance);

	Stem* stem = getStemOwner();
	if ( stem && stem->sync == true ){
		int stemIndex = max(0, index[0] - 1);
		stem->index[0] = max(0, stemIndex - 1);
		M_Int& bands = stem->bands;

		if ( stemIndex < bands.length - 1 ) {
			for ( int i = stemIndex; i < bands.length - 1; i++ ) {
				bands[i] = bands[i + 1];
			}
		}

		getBandAndYDist(bands, Stem::bandDist, stem->yDistance);
		bands.length = stem->yDistance.length;
	}

	index[0] = idx;
	stem->convertYDistanceToHeight();
	processFields();
	commonPostTug();
	updateBuild3dWindow();
}


Art* RingBuild::getSelPart(){

	Art* prt = part[0];
	int idx = index[0];

	for(int i = 0, r = 0; i < ownership.length; i++){
		if(i < ownership.length){
			r+=ownership[i];
			if(i<part.length){
				prt = part[i];
			}
		}
		if(r >= idx){
			break;
		}
	}
	return prt;
}

void RingBuild::senseMirrorRing(){

	Art* part = RingBuild::getSelPart();
	Seamless* smls = part->smls;
	M_Vec3f& coord = part->coord.refValue();

	if(part->mirror && part->mirrorPart == 0){
		int idx = index[0];
		if(quadrant == 1){
			scaleX.refValue()[idx].y = scaleX.refValue()[idx].x;
		}
		else if(quadrant == 3){
			scaleX.refValue()[idx].x = scaleX.refValue()[idx].y;
		}
	}
}

void RingBuild::drag(Vec3f pt){
	M_Vec2f& scaleX=this->scaleX.refValue();
	M_Vec2f& scaleZ=this->scaleZ.refValue();
	M_Rotation& orientation=this->orientation.refValue();
	M_Vec3f& location=this->location.refValue();
	const Rotation& rotation=this->rotation.refValue();
	const Vec3f& translation=this->translation.refValue();
	const Vec3f& scale = this->scale.refValue();

	int i=index[0];

	Matrix matrix = rotationMatrix(orientation[i]) * translationMatrix(location[i]) * 
	scaleMatrix(scale) * rotationMatrix(rotation) * translationMatrix(translation) * g_scene->getViewMatrix();


	Vec3f pivotPoint = matrix * Vec3f();

	Vec3f controlPoint;
	if(quadrant==4){
		Matrix pivotPointIMatrix = translationMatrix(Vec3f(-translation.x, -translation.y, -translation.z)) * rotationMatrix( -rotation) * scaleMatrix(Vec3f(1./scale.x, 1./scale.y, 1./scale.z));
		Vec3f newPt=g_scene->getIViewMatrix() * pivotPointIMatrix * pt;

		if(dragLocationY){
			location[i].y = newPt.y;
		}
		if(dragLocationXZ){
			location[i].x = newPt.x;
			location[i].z = newPt.z;
		}
	}
	else{
		Vec3f vec;
		float* ptr=0;
		switch(quadrant){
			case 0:
				ptr=&scaleZ[i].x;
				controlPoint = matrix * Vec3f(0, 0, -*ptr);
				vec = Vec3f(0, 0, -*ptr);
				break;
			case 1:
				ptr=&scaleX[i].x;
				controlPoint = matrix * Vec3f(-*ptr, 0, 0);
				vec = Vec3f(-*ptr, 0, 0);
				break;
			case 2:
				ptr=&scaleZ[i].y;
				controlPoint = matrix * Vec3f(0, 0, *ptr);
				vec = Vec3f(0, 0, *ptr);
				break;
			case 3:
				ptr=&scaleX[i].y;
				controlPoint = matrix * Vec3f(*ptr, 0, 0);
				vec = Vec3f(*ptr, 0, 0);
				break;
		}
		Vec3f loc = this->location.refValue()[i];
		Rotation orien = this->orientation.refValue()[i];

		Matrix imatrix=translationMatrix(Vec3f(-translation.x, -translation.y, -translation.z)) * rotationMatrix( -rotation) * scaleMatrix(Vec3f(1./scale.x, 1./scale.y, 1./scale.z)) *
			translationMatrix(Vec3f(-loc.x, -loc.y, -loc.z)) * rotationMatrix(-orien);

		Vec3f newPt=g_scene->getIViewMatrix() * imatrix * pt;

		*ptr = sqrt(newPt.x*newPt.x + newPt.y*newPt.y + newPt.z*newPt.z);

//		{
//			Matrix baseMatrix = scaleMatrix(scale) * rotationMatrix(rotation) * translationMatrix(translation) * g_viewMatrix;
//			Matrix matrix = rotationMatrix(orientation[i]) * translationMatrix(location[i]) * baseMatrix;

		//}

		senseMirrorRing();


		if(dragOrientation){

			//would there not be an easier way to calculate the orientation with the information we have at this point in time?

			Vec3f toDiff = Vec3f(pt.x-pivotPoint.x, pt.y-pivotPoint.y, pt.z-pivotPoint.z);
			Vec2f fromDiff = Vec2f(controlPoint.x-pivotPoint.x, controlPoint.y-pivotPoint.y);
			float toAngle = Vec2f(toDiff.x, toDiff.y).getAngle();
			float fromAngle = fromDiff.getAngle();
			float angle = toAngle - fromAngle;

			Rotation unwind = -rotation;
			Vec3f axis = (-g_scene->rotation * unwind * Vec3f(0, 0, 1)).normalize();
			orientation[i] = orientation[i] * Rotation(axis, angle);
		}
		Vec3f sr;
		switch(quadrant){
			case 0:
				sr= Vec3f(0, 0, -scaleZ[i].x);
				break;
			case 1:
				sr= Vec3f(-scaleX[i].x, 0, 0);
				break;
			case 2:
				sr= Vec3f(0, 0, scaleZ[i].y);
				break;
			case 3:
				sr= Vec3f(scaleX[i].y, 0, 0);
				break;
		}
		Matrix postMatrix = rotationMatrix(orientation[i]) * translationMatrix(location[i]) * 
		scaleMatrix(scale) * rotationMatrix(rotation) * translationMatrix(translation) * g_scene->getViewMatrix();
		g_pointer.translation= postMatrix * sr;
	}
	houseKeep();
	commonPostTug();
}
void Art::drag(Vec3f pt){
	dynamic_cast<PartAnimClue*>(g_visualClue)->dragPart(pt);
}
extern bool tugFlag2;
void openTransformCP();

void getInput(){
	if(checkRobotActive()){
		return;
	}
	Vec3f pt;
	Vec3f* translationPtr=0;
	Rotation* rotationPtr=0;

	Seamless* smls = 0;
	Art* rootPart;
	if(Art* part=dynamic_cast<Art*>(g_selNode.getRaw())){
		smls=part->smls;
		rootPart=smls->skeleton.refValue()[0];
		translationPtr=&part->translation.refValue();
		rotationPtr=&part->rotation.refValue();
	}
	Vec3f& partTranslation=*translationPtr;
	Rotation& partRotation=*rotationPtr;

	DIMOUSESTATE dims;      // DirectInput mouse state structure
	HRESULT hr=1;

	#define KEYDOWN(name, key) (name[key] & 0x80);

	if(g_pMouse==NULL)return;
	g_state = g_pMouse->GetDeviceState( sizeof(DIMOUSESTATE), &dims );
	if(g_state==DI_OK){
		/*
		if(!(dims.rgbButtons[0] & 0x80) && mouse3dMode==true){
			UnAcquire();
			mouse3dMode=false;
		}
		*/
/*
		if(g_selNode->getType()==PART && !(0x8000&GetKeyState(VK_LCONTROL))&&g_localControl&&dims.rgbButtons[0] & 0x80){//left mouse down

			//pj=g_selNode;
			if(dims.lY||dims.lX){
				partTranslation+=rootPart->unWindRot(pj).multVec(Vec3f(.001*(float)dims.lX,0,0));
				
				if(ZDragMode){
					partTranslation+=rootPart->unWindRot(pj).multVec(Vec3f(0,0,.001*(float)dims.lY));
				}
				else{
					partTranslation+=rootPart->unWindRot(pj).multVec(Vec3f(0,-.001*(float)dims.lY,0));
				}
				if(pj->smls->skeleton.size()){
					buildSceneMatrices( pj->smls->skeleton[0], Matrix());
					pj->smls->processRenderIfs();
					pj->smls->convertToRenderFormat();
				}
				resetSmlsSaved();
			}
		}
		else if(g_selNode->getType()==PART && g_localControl&&dims.rgbButtons[1] & 0x80){//right mouse down
			pj=g_selNode;
			partRotation=partRotation.multiply(Rotation(rootPart->unWindRot(pj).multVec(Vec3f(0,1,0)), PI*.005*(float)dims.lX));
			partRotation=partRotation.multiply(Rotation(rootPart->unWindRot(pj).multVec(Vec3f(1,0,0)), PI*.005*(float)dims.lY));
			if(pj->smls->skeleton.size()){
				buildSceneMatrices( pj->smls->skeleton[0], Matrix());
				pj->smls->processRenderIfs();
				pj->smls->convertToRenderFormat();
			}
			resetSmlsSaved();
		}
		else*/
		{
			if(dims.lY||dims.lX){

				Art* part=dynamic_cast<Art*>(g_selNode.getRaw());

				if(g_viewRotDragMode == true){
					//if(part!=0 && 0x8000&GetKeyState(VK_LSHIFT) && !(0x8000&GetKeyState(VK_LCONTROL))){
					//	part->rotation.refValue()*=Rotation(0,1,0, PI*.01*(float)dims.lX)*Rotation(1,0,0, PI*.01*(float)dims.lY);
					//	part->rotation=part->rotation;
					//}
					//else {
						g_scene->rotateScene(dims.lX,dims.lY);
					//}
				}
				else{ 
					MouseFlag = true;

					if(MultiSelect::selectRectMode){
						float a = MouseSense*.25/g_scene->scale.x;
						float b = MouseSense*.25;

						g_pointer.translation.y-=dims.lY*MouseSense*.25;
						g_pointer.translation.x+=dims.lX*MouseSense*.25;


//						Vec3f v = Vec3f(dims.lX*b, -dims.lY*b, 0);
						Vec3f v = g_pointer.translation;
						MultiSelect::bottomRight=v;
//						Patch::bottomRight+=Vec3f(dims.lX*b, -dims.lY*b, 0);

						g_viewDragMousePos+=Vec3f(dims.lX*b, -dims.lY*b, 0);
						Vec2i vec=get2dMousePosFrom3d(g_viewDragMousePos);
						WINDOWINFO wi;
						GetWindowInfo(hWndMain, &wi);
						SetCursorPos(vec.x+wi.rcClient.left, vec.y+wi.rcClient.top); 
						houseKeepVisualClue();
					}
					else if(g_viewPosDragMode && !(0x8000&GetKeyState(VK_LCONTROL))){
						float a = MouseSense*.25/g_scene->scale.x;
						float b = MouseSense*.25;

						Vec3f v=-g_scene->rotation*Vec3f(dims.lX*a, -dims.lY*a, 0);
						g_scene->translation+=v;
						
						g_viewDragMousePos+=Vec3f(dims.lX*b, -dims.lY*b, 0);
						Vec2i vec=get2dMousePosFrom3d(g_viewDragMousePos);
						WINDOWINFO wi;
						GetWindowInfo(hWndMain, &wi);
						g_mousePos.x = vec.x + wi.rcClient.left;
						g_mousePos.y = vec.y + wi.rcClient.top;
						SetCursorPos(g_mousePos.x, g_mousePos.y); 
					}
					else{
						if(g_zDragMode){
							g_pointer.translation.z+=dims.lY*MouseSense*.5;
						}
						else{
							g_pointer.translation.y-=dims.lY*MouseSense*.5;
						}
						g_pointer.translation.x+=dims.lX*MouseSense*.5;
						if(tugFlag){
							if(!MultiSelect::selectRectMode){
								if(VisualAid* visualAid = dynamic_cast<VisualAid*>(g_selNode.getRaw())){
									if((!(0x8000&GetKeyState(VK_LCONTROL))) || dynamic_cast<NurbsPatch*>(visualAid)){
										visualAid->drag(g_pointer.translation);
									}
								}
							}
						}
					}
				}
			}
			if(dims.rgbButtons[1] & 0x80){
				if(g_viewRotDragMode){
					g_viewRotDragMode = true;
					g_rightDragCounter = Vec2i();
					g_beginViewRot = g_scene->rotation;
				}
				else if(tugFlag==false){
//					SenseTugGlobals::initTug();
//					if(tugFlag==false){
					g_viewRotDragMode = true;
					g_rightDragCounter = Vec2i();
					g_beginViewRot = g_scene->rotation;
//					}
//					else{
//						g_tugRotFlag=true;
//					}
				}
			}
			else{
				//RotateMode=FALSE;
				if(dims.rgbButtons[0] & 0x80){
					if(tugFlag==false && g_viewPosDragMode == false && g_insertControlPoint == false && MultiSelect::selectRectMode == false){
						//g_viewMatrix=scaleMatrix(g_scene->scale) * translationMatrix(g_scene->translation) * rotationMatrix(g_scene->rotation);
						SenseTugGlobals::initTug();
						if(tugFlag==false && g_insertControlPoint == false && MultiSelect::selectRectMode == false){
							g_viewPosDragMode = true;
						}
						else if(ImmediatePart* part=dynamic_cast<ImmediatePart*>(g_selNode.getRaw())){
							if(Window::animMode() == 0){
								if(g_palette->rectangle){
									Seamless* smls = part->smls;
									M_Vec3f& coord = part->coord;
									ColorEffect* colorEffect = dynamic_cast<ColorEffect*>(part->workEffect);
									if(coord.length > 0 && colorEffect){
										M_Color3f& color = colorEffect->color;
										int colorIdx = smls->actTriPart->triangle.refTexIndex(smls->ActTri * 3 + smls->ActTriCorner);
										color[colorIdx] = g_palette->color;
										commonIncDecColorChange(smls, part, color, colorIdx);
									}
								}
							}
						}

						g_viewDragMousePos=g_pointer.translation;
					}
				}
				else {
	if(g_viewRotDragMode){
		houseKeepVisualClueViewChange();
	}

#ifdef PROG_ROB
	if(MultiSelect::selectRectMode){
		if(0x8000&GetKeyState(VK_LSHIFT)){
			if(0x8000&GetKeyState(VK_LCONTROL)){
				dLog("new UnRectVert("+String(MultiSelect::topLeft.x) + "," + String(MultiSelect::topLeft.y) + "," + String(MultiSelect::bottomRight.x) + "," + String(MultiSelect::bottomRight.y) + "),");
			}
			else{
				dLog("new RectVert("+String(MultiSelect::topLeft.x) + "," + String(MultiSelect::topLeft.y) + "," + String(MultiSelect::bottomRight.x) + "," + String(MultiSelect::bottomRight.y) + "),");
			}
		}
		if(0x8000&GetKeyState(VK_RSHIFT)){
			dLog("new RectPoly("+String(MultiSelect::topLeft.x) + "," + String(MultiSelect::topLeft.y) + "," + String(MultiSelect::bottomRight.x) + "," + String(MultiSelect::bottomRight.y) + "),");
		}
	}
	else if(g_viewRotDragMode){
		dLog("new DragViewRot("+String(g_scene->rotation.x) + "," + String(g_scene->rotation.y) + "," + String(g_scene->rotation.z) + "," + String(g_scene->rotation.angle)+"),");
	}
	else if(g_viewPosDragMode){
		dLog("new DragViewPos("+String(g_scene->translation.x) + "," + String(g_scene->translation.y) + "," + String(g_scene->translation.z) + "),");
	}
	else if(NurbsPatch* patch = dynamic_cast<NurbsPatch*>(g_selNode.getRaw())){
		//dLog("new TugVertex(" + String(nurbsPatch->index[0]) + "," +String(g_pointer.translation.x)+','+String(g_pointer.translation.y) + "),");
		if(g_transformCP->dragRotPointMode){
			if((0x8000&GetKeyState(VK_LSHIFT))){
				if((0x8000&GetKeyState(VK_LCONTROL))){
					if((0x8000&GetKeyState(VK_LMENU))){
					}
					else{
						dLog("\nnew Stretch("+String(g_pointer.translation.x)+','+String(g_pointer.translation.y)+"),");
					}
				}
				else{
					
				}
			}
			else if((0x8000&GetKeyState(VK_LMENU))){
				if((0x8000&GetKeyState(VK_LCONTROL))){
					dLog("\nnew Scale2d("+String(g_pointer.translation.x)+','+String(g_pointer.translation.y)+"),");
				}
				else{
					dLog("\nnew Scale1d("+String(g_pointer.translation.x)+','+String(g_pointer.translation.y)+"),");
				}
			}
			else if((0x8000&GetKeyState(VK_LCONTROL))){
				dLog("\nnew Neutral("+String(g_pointer.translation.x)+','+String(g_pointer.translation.y)+"),");
			}
			else{
				dLog("\nnew Rotate("+String(g_pointer.translation.x)+','+String(g_pointer.translation.y)+"),");
			}
		}
		else if(g_transformCP->dragPivotPointMode){
			dLog("\nnew MoveToMultiSelectPivotPoint(),\nnew Move3d(Vec3f("+String(g_pointer.translation.x)+','+String(g_pointer.translation.y)+','+String(g_pointer.translation.z)+")),");
		}
		else if(0x8000&GetKeyState(VK_LMENU)){
			dLog("new Scale3d(" + String(MultiSelect::pointList[MultiSelect::clickedIndex].i + MultiSelect::pointList[MultiSelect::clickedIndex].p->fromPoint) + "," +String(g_pointer.translation.x)+','+String(g_pointer.translation.y) + "),");
		}
		else{
			dLog("new TugCP(" + String(MultiSelect::pointList[MultiSelect::clickedIndex].i + MultiSelect::pointList[MultiSelect::clickedIndex].p->fromPoint) + "," +String(g_pointer.translation.x)+','+String(g_pointer.translation.y) + "),");
		}
	}
	else if(Art* art = dynamic_cast<Art*>(g_selNode.getRaw())){
		if(Window::animMode()){
			dLog("\nnew MoveToAnimRotPoint(),\nnew Move3d(Vec3f("+String(g_pointer.translation.x)+','+String(g_pointer.translation.y)+','+String(g_pointer.translation.z)+")),");
		}
		else if(0x8000&GetKeyState(VK_LMENU)){
			int idx = dynamic_cast<Art*>(g_selNode.getRaw())->index[0];
			dLog("new SelectTriangle(" + String(art->smls->ActTri*3+smls->ActTriCorner) + "," + art->smls->actTriPart->name + "),");
		}
		else if(0x8000&GetKeyState(VK_LCONTROL)){
			dLog("new SelectVertex(" + String(art->smls->ActCtrlPnt) + "," + art->name + "),");
//			dLog("new TugVertex(" + String(art->smls->ActCtrlPnt) + "," +String(g_pointer.translation.x)+','+String(g_pointer.translation.y)+','+art->name + "),");
		}
		else{
			dLog("new TugVertex(" + String(art->smls->ActCtrlPnt) + "," +String(g_pointer.translation.x)+','+String(g_pointer.translation.y)+','+art->name + "),");
		}
	}
	else if(NurbsLathe* lathe = dynamic_cast<NurbsLathe*>(g_selNode.getRaw())){
		if(0x8000&GetKeyState(VK_RCONTROL)){
			dLog("new AddCtrlPnt(" + String(g_pointer.translation.x) + ',' + String(g_pointer.translation.y) + "),");
		}
		else{
			dLog("new TugVertex(" + String(lathe->index[0]) + "," +String(g_pointer.translation.x) + ',' + String(g_pointer.translation.y) + "),");
		}
	}
	else if(RingBuild* ringBuild = dynamic_cast<RingBuild*>(g_selNode.getRaw())){
		int i = ringBuild->index[0];
		dLog("new TugRing(" + String(i*5+ringBuild->quadrant) + ","+String(g_pointer.translation.x)+','+String(g_pointer.translation.y)+"),");
	}
	//dLog("\t{" + String(idx) + ", Vec2f("+String(g_pointer.translation.x)+','+String(g_pointer.translation.y)+")},");
#endif
#ifdef RECORD_3D_POINTER
//					if(Art* art = dynamic_cast<Art*>(g_selNode.getRaw())){
//					dLog("\tselNode("+g_selNode->name+");");
//					dLog("\tlist+=new MoveToAnimPoint();\n\tlist+=new Move3d(Vec3f("+String(g_pointer.translation.x)+','+String(g_pointer.translation.y)+','+String(g_pointer.translation.z)+"));");
//					dLog("\tselNode("+g_selNode->name+");");
//					if(g_pointer.translation.z!=0){
//						int idx = dynamic_cast<Art*>(g_selNode.getRaw())->index[0];
//						dLog("\t{" + g_selNode->name + ", " + String(idx) + ", Vec2f("+String(g_pointer.translation.x)+','+String(g_pointer.translation.y)+")},");
//						dLog("\t{" + String(idx) + ", Vec2f("+String(g_pointer.translation.x)+','+String(g_pointer.translation.y)+")},");

						//dLog("\t" + String(idx) + ", " + g_selNode->name + ", " );
//						dLog(String(art->smls->ActTri) + "*3, ");
//					}
//					dLog("\t{" + String(idx) + ", Vec3f("+String(g_pointer.translation.x)+','+String(g_pointer.translation.y)+','+String(g_pointer.translation.z)+")},");
//					}
#endif
					if(AnimClue* animClue = dynamic_cast<AnimClue*>(SenseTugGlobals::node)){
						Art* part = dynamic_cast<Art*>(g_selNode.getRaw());
						if(g_dragActuallyHappned){
							if(animClue->dragTranPointMode){
								if(Anim::autoCapture){
									if(part->ik){
										animClue->anim->captureSinglePartPose(part, &part->ikPos);
									}
									else{
										animClue->anim->captureSinglePartPose(part, &part->translation);
									}
								}
								g_vcArtAnimBase->dragTranPointMode = false;
								part->houseKeep();
							}
							else{
								if(Anim::autoCapture){
									if(part->ik){
										animClue->anim->captureSinglePartPose(part, &part->ikRot);
									}
									else{
										animClue->anim->captureSinglePartPose(part, &part->rotation);
									}
								}
							}
							Window::paintAllControlPanels();
						}
					}
					if(tugFlag == true && !(0x8000&GetKeyState(VK_LCONTROL)) && !(0x8000&GetKeyState(VK_LMENU))){
						if(AnimClue* animClue = dynamic_cast<AnimClue*>(g_visualClue)){
							if(smls->animCP == false && smls->workAnimCP == false && g_dragMorphFlag && g_dragActuallyHappned && !(0x8000&GetKeyState(VK_LCONTROL)) && !(0x8000&GetKeyState(VK_LMENU))){
//								smls->genDynamicMatricesAndPopulate();
								animClue->anim->capturePolyMorph(smls);
								tugFlag2 = tugFlag = false;
							}
							else if(Part* part = dynamic_cast<Part*>(g_selNode.getRaw())){
								Seamless* smls = part->smls;
								if(smls->workAnimCP && g_partCageClue->cpTug){
									animClue->anim->captureMorph(smls);
								}
							}
						}
						else if(Art* part = dynamic_cast<Art*>(g_selNode.getRaw())){
							tugFlag2=tugFlag = false;
							if(g_dragActuallyHappned && part->smls->build.length){
								g_dragActuallyHappned = false;
								part->smls->doBuild();
							}
							g_dragActuallyHappned = false;
							part->smls->ctrf = true;
						}
						BackUp::monostable.beginCycle();
					}
					tugFlag=false;
					orientationTugFlag=false;
					dragPartFlag=false;
					g_mirrorModeVertexFound=false;
					mouse3dMode=false;
					UnAcquire();
					g_dragActuallyHappned = false;
					showCursor(TRUE);
					g_stw->cp->paint();
					//dLog("new DragViewRot("+String(g_scene->rotation.x) + "," + String(g_scene->rotation.y) + "," + String(g_scene->rotation.z) + "," + String(g_scene->rotation.angle)+"),");
					//dLog("new DragViewPos("+String(g_scene->translation.x) + "," + String(g_scene->translation.y) + "," + String(g_scene->translation.z) + "),");
				}
			}
		}
	}
	else {
		if(DIERR_INPUTLOST==g_state){
			if(mouse3dMode){
				SetAcquire();
			}
		}
		/*
		else if(DIERR_INVALIDPARAM ==g_state){
			dLog("DIERR_INVALIDPARAM ");
		}
		else if(DIERR_NOTACQUIRED==g_state){
			dLog("DIERR_NOTACQUIRED");
		}
		else if(DIERR_NOTINITIALIZED==g_state){
			dLog("DIERR_NOTINITIALIZED");
		}
		else if(E_PENDING ==g_state){
			dLog("E_PENDING ");
		}
		else {
			dLog("what ever");
		}
	*/
	}
}
