/* Seamless3d

Copyright (C) 2001-2010 Graham John Perrett (AKA thyme) thyme@seamless3d.com www.seamless3d.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this
software and associated documentation files (the "Software"), to deal in the Software
without restriction, including without limitation the rights to use, copy, modify,
merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to the following
conditions:

The above copyright notice and this permission notice shall be included in all copies or
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT
OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.
*/

#include <windows.h>
#include <direct.h>

#include "seamless3d.h"
#include "techuelife_triangle_editor.h"
#include "seamless3d_gui_render.h"
#include "gui.h"
#include "scene_tree.h"
#include "robot_aid.h"
#include "hipng.h"

void insertShort(M_<char>& buf, short data);
void insertInt32(M_<char>& buf, int data);
void convertBMPToPng(const char* filename, int width, int height, M_<char>& header, Pix24* pixels);

void createBMP24(const char* filename, int width, int height, Pix24* pixels){
	M_<char> header;
	const int header_size = 0x36;
	const int file_size = height * width * sizeof(Pix24) + header_size;
	header += 'B';
	header += 'M';
	insertInt32(header, file_size);
	header.length += 4; //reserved
	insertInt32(header, header_size); //pixels offset
	insertInt32(header, 40); //the size of this header 
	insertInt32(header, width);
	insertInt32(header, height);
	insertShort(header, 1); //number of color planes being used.
	insertShort(header, 24); //number of bits per pixel
	insertInt32(header, 0); //Compression method none;
	insertInt32(header, height * width * sizeof(Pix24)); //The image size field;
	insertInt32(header, width); //horizontal resolution of the image. (pixel per meter, signed integer)
	insertInt32(header, height); //vertical resolution of the image. (pixel per meter, signed integer)
	header.length += 8;

	FILE* stream = fopen(filename, "w+b");
	fwrite(header.getArray(), header_size, 1, stream);
	fwrite(pixels, width * height, sizeof(Pix24), stream);
	fclose(stream);
}

void createPNG24(const char* filename, int width, int height, Pix24* pixels){
	HIPNG hipng;
	// TODO: Place code here.
	hipng.file_name = filename;
	hipng.width = width;
	hipng.height = height;
	hipng.image = reinterpret_cast<png_byte*>(pixels);

	hipng.bit_depth = 8;
	hipng.color_type = PNG_COLOR_TYPE_RGB;
	hipng.interlaced = FALSE;
	hipng.num_colors = 0;
	hipng.palettep = NULL;
	hipng.gamma = 0.0;
	hipng.num_text = 0;
	hipng.pComment = 0;
	hipng.current_time = FALSE;
	hipng.num_transparent = 0;
	hipng.transparent_palette = NULL;
	hipng.RGBtransparent_value = NULL;
	hipng.row_callback_fn = NULL;
	hipng.rows = NULL;
	
	BOOL bl = write_standard_png(&hipng);   
}

void plotPixelRGB(Pix24 color, int x, int y, int width, int height, M_<Pix24>& pixels){
	pixels[(height - 1 - y) * width + x] = Pix24(color.r, color.g, color.b); //swap color to BGR
}

inline void scaleTri(Vec2f* tri2d){
	const float scaleAmmount = 1.0001;
	Vec2f d = (tri2d[0] + tri2d[1] + tri2d[2]) / 3.0;
	Vec2f e = d * scaleAmmount;
	Vec2f diff = e - d;
	tri2d[0] = (tri2d[0] * scaleAmmount) - diff;
	tri2d[1] = (tri2d[1] * scaleAmmount) - diff;
	tri2d[2] = (tri2d[2] * scaleAmmount) - diff;
}
void convertTo2dTriangle(Vec2f* tri2d, Color3f* tri2dC, const S_Triangle& blueTri, const S_Triangle& whiteTri, const M_Vec3f& blueCoord, const M_Color3f& m_color){
	tri2d[0] = Vec2f(blueCoord[blueTri.ia].x, blueCoord[blueTri.ia].y);
	tri2d[1] = Vec2f(blueCoord[blueTri.ib].x, blueCoord[blueTri.ib].y);
	tri2d[2] = Vec2f(blueCoord[blueTri.ic].x, blueCoord[blueTri.ic].y);

	tri2dC[0] = m_color[whiteTri.ta];
	tri2dC[1] = m_color[whiteTri.tb];
	tri2dC[2] = m_color[whiteTri.tc];

	scaleTri(tri2d);
}

bool pointInTriangle(const Vec2f& p, const Vec2f& a, const Vec2f& b, const Vec2f& c);

inline bool findLineIntersectInTriangle2(int& c, const Vec2f& pt, const Vec2f* v, const Color3f* colorArray, Vec2f& edgeHit, Color3f& edgeHitC){
	for(int i = 0; i < 2; i++, c++){
		c = c % 3;
		int n = (c + 1) % 3;
		if(pt.y >= v[n].y && pt.y <= v[c].y){
			Vec2f diff = v[c] - v[n];
			edgeHit.y = pt.y - v[n].y;
			float yRatio = edgeHit.y / diff.y;
			edgeHit.x = diff.x * yRatio;
			edgeHit += v[n];
			edgeHitC = (colorArray[c] * yRatio) + (colorArray[n] * (1.0 - yRatio));
			return true;
		}
	}
	return false;
}
inline bool findLineIntersectInTriangle(int& c, const Vec2f& pt, const Vec2f* v, const Color3f* colorArray, Vec2f& edgeHit, Color3f& edgeHitC){
	for(int i = 0; i < 3; i++, c++){
		c = c % 3;
		int n = (c + 1) % 3;
		if(pt.y >= v[c].y && pt.y <= v[n].y){
			Vec2f diff = v[n] - v[c];
			edgeHit.y = pt.y - v[c].y;
			float yRatio = edgeHit.y / diff.y;
			edgeHit.x = diff.x * yRatio;
			edgeHit += v[c];
			edgeHitC = (colorArray[n] * yRatio) + (colorArray[c] * (1.0 - yRatio));
			return true;
		}
	}
	return false;
}
bool getTriangleHitColor(const Vec2f& pt, Color3f& clickColor, const Vec2f* v, const Color3f* colorArray){
	Vec2f edgeHitA;
	Color3f edgeHitCA;
	Vec2f edgeHitB;
	Color3f edgeHitCB;
	int c = 0;
	if(	findLineIntersectInTriangle(c,	  pt, v, colorArray, edgeHitA, edgeHitCA) &&
		findLineIntersectInTriangle2(++c, pt, v, colorArray, edgeHitB, edgeHitCB)){

		if(pt.x <= edgeHitA.x && pt.x >= edgeHitB.x){

			float xDiff = edgeHitA.x - edgeHitB.x;
			float xRatio = (pt.x - edgeHitB.x) / xDiff;

			clickColor = edgeHitCA * xRatio + edgeHitCB * (1.0 - xRatio);

			return true;
		}
	}
	return false;
}
inline int getHighPix(const Vec2f* tri2d, const int height){
	float most = tri2d[0].y;
	int a = 0;
	for(int i = 1; i < 3; i++){
		if(tri2d[i].y > most){
			most = tri2d[i].y;
			a = i;
		}
	}
	return tri2d[a].y * (height - 1) + 1.0;
}
inline int getRightPix(const Vec2f* tri2d, const int width){
	float most = tri2d[0].x;
	int a = 0;
	for(int i = 1; i < 3; i++){
		if(tri2d[i].x > most){
			most = tri2d[i].x;
			a = i;
		}
	}
	return tri2d[a].x * (width - 1) + 1.0;
}

inline int getLowPix(const Vec2f* tri2d, const int height){
	float most = tri2d[0].y;
	int a = 0;
	for(int i = 1; i < 3; i++){
		if(tri2d[i].y < most){
			most = tri2d[i].y;
			a = i;
		}
	}
	return tri2d[a].y * (height - 1);
}

inline int getLeftPix(const Vec2f* tri2d, const int width){
	float most = tri2d[0].x;
	int a = 0;
	for(int i = 1; i < 3; i++){
		if(tri2d[i].x < most){
			most = tri2d[i].x;
			a = i;
		}
	}
	return tri2d[a].x * (width - 1);
}

void writeColorToTexture(const char* filename){
	M_<Pix24> pixels;
	int width = 2048;
	int height = 2048;
	pixels.length = width * height;
	
	if(Stem* stem = dynamic_cast<Stem*>(g_selNode.getRaw())){
		if(stem->source.length){
			if(Art* bluePart = dynamic_cast<Art*>(stem->source.refValue()[0].p)){
				if(stem->part.length){
					if(Art* whitePart = dynamic_cast<Art*>(stem->part.refValue()[0].p)){
						if(ColorEffect* colorEffect = dynamic_cast<ColorEffect*>(whitePart->workEffect)){

							const M_Color3f& m_color = colorEffect->color;
							M_Triangle& blueTri = bluePart->triangle;
							M_Triangle& whiteTri = whitePart->triangle;
							const M_Vec3f& blueCoord = bluePart->coord;
							const M_Vec3f& whiteCoord = whitePart->coord;

							if(blueTri.length <= whiteTri.length && blueCoord.length <= whiteCoord.length){
								Vec2f tri2d[3];
								Color3f tri2dC[3];
								const float halfPix = 1.0 / width;

								for(int t = 0; t < blueTri.length; t++){
									convertTo2dTriangle(tri2d, tri2dC, blueTri[t], whiteTri[t], blueCoord, m_color);

									int high = getHighPix(tri2d, height);
									int low = getLowPix(tri2d, height);
									int left = getLeftPix(tri2d, width);
									int right = getRightPix(tri2d, width);
									for(int y = low; y < high; y++){
										for(int x = left; x < right; x++){
											float fx = static_cast<float>(x) / width + halfPix;
											float fy = static_cast<float>(y) / height + halfPix;

											if(pointInTriangle(Vec2f(fx, fy), tri2d[0], tri2d[1], tri2d[2])){
												Color3f clickColor;
												if(getTriangleHitColor(Vec2f(fx, fy), clickColor, tri2d, tri2dC)){
													plotPixelRGB(Pix24(clickColor.r * 255, clickColor.g * 255, clickColor.b * 255), x, y, width, height, pixels);
												}
											}
										}
									}
								}
								createPNG24(filename, width, height, pixels.getArray());
								return;
							}
						}
					}
				}
			}
		}
	}
	MessageWin::msg("Stem must have source and destination parts with equal triangle and coord length");
}

extern String g_mainFilename;
extern unsigned int g_counter;

void saveBackBuffer(unsigned int* pBits, M_<Pix24>& pixels, unsigned int screenWidth, unsigned int width, unsigned int height){
	Pix24* lpDest = pixels.getArray();
	for(unsigned int r = 0; r < height; r++){
		for(unsigned int c = 0; c < width; c++){
			int idx = r * width + c;
			unsigned int rawPix = pBits[idx];
			Pix24 pixel;
			pixel.b = rawPix & 0x00ff;
			rawPix >>= 8;
			pixel.g = rawPix & 0x00ff;
			rawPix >>= 8;
			pixel.r = rawPix & 0x00ff;
			//pixel = Pix24(0xff,0xff,0);
			lpDest[idx] = pixel;
		}
	}
	static String filename;
	filename = g_mainFilename;
	filename += g_counter;
	filename += ".png";
	createBMP24(filename.c_str(), width, height, lpDest);
}
