/* Seamless3d

Copyright (C) 2001-2010 Graham John Perrett (AKA thyme) thyme@seamless3d.com www.seamless3d.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this
software and associated documentation files (the "Software"), to deal in the Software
without restriction, including without limitation the rights to use, copy, modify,
merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to the following
conditions:

The above copyright notice and this permission notice shall be included in all copies or
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT
OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.
*/

#include <windows.h>
#include <shlobj.h>
#include <direct.h>

#include "resource.h"
#include "seamless3d.h"
#include "script.h"
#include "techuelife_triangle_editor.h"
#include "gui.h"
#include "scene_tree.h"
#include "robot_aid.h"

extern const char* menu_short_copy;
Seamless* getSeamless2();
void renderScene();

class File : public Node, public MField4{
	public:
enum Type{
	FOLDER,
	SMLS,
	VRML,
	BVH,
	MID,
	PNG,
	WAV
};
struct FileTime{

	unsigned int lo;
	unsigned int hi;
	FileTime():
		lo(0),
		hi(0)
	{}
	bool operator>(const FileTime& par)const{
		if(hi > par.hi) return true;
		if(hi < par.hi) return false;
		return lo > par.lo; 
	}
	bool operator<(const FileTime& par)const{
		if(hi < par.hi) return true;
		if(hi > par.hi) return false;
		return lo < par.lo;
	}
};
class SortByDate : public SFBool {
public:
	File& home;

	SortByDate(const bool& par, File& home):
		SFBool(par),
		home(home)
	{};
	void operator = (const bool& par);
};

class Open : public SFBool {
public:
	File& home;

	Open(const bool& par, File& home):
		SFBool(par),
		home(home)
	{};
	void operator = (const bool& par);
};

////////////////
	int getType()const{return FILE_NODE;}
	int getFLen()const{return size();};
	Field* getField(const int i)const{return (*this)[i];}
	int getFieldGroupIndex(const int &i)const{throw this;}
	void setFieldGroupIndex(const int &i, const int &val){throw this;}
	M_Int* getIndexPtr(){return nullPtr;}
	
	SortByDate sortByDate;
	SFString filename;
	Open open;
	MFNode folder;

	File():
		MField4(
			&sortByDate,
			&filename,
			&open,
			&folder
		),
		sortByDate(false, *this),
		filename(String()),
		open(false, *this),
		folder(M_Node()),

		isFile(false),
		type(FOLDER)
	{
	}
	bool isFile;
	char type;
	FileTime time;

};

class FileCP : public ControlPanel{
public:
	M_<SceneTreeListItem> fileTreeList;
	enum{
		//beginYOffset = 50
		beginYOffset = 25,
		endYOffset = 25 + 8,
		driveWidth = 70,
	};

class VerticalScrollBarSlider : public Button, public Drag{
public:
	FileCP& win;

	VerticalScrollBarSlider(Vec2i i_size, Vec2i i_srcPos, const char* i_tip, int i_command, FileCP& i_win, int refCountValue);

	void doDrag(const Vec2i ptCursor);
	void endDrag(const Vec2i ptCursor);

	void paint();
	void action();
	Vec2i getSimClickPos(int subButtonIdx){
		return Vec2i(bSize.x/2, bSize.y/2);
	};
	bool VerticalScrollBarSlider::updateDrag(Vec2i ptCursor);
	void VerticalScrollBarSlider::setPixelPos();
};

class VerticalScrollBarTopButton : public Button{
public:
	VerticalScrollBarTopButton(Vec2i i_size, Vec2i i_srcPos, const char* i_tip, int i_command, FileCP& i_win, int refCountValue);
	void paint();
	void action();
	Vec2i getSimClickPos(int subButtonIdx){
		return Vec2i(bSize.x/2, bSize.y/2);
	};
	FileCP& win;
};
class VerticalScrollBarBottomButton : public Button{
public:
	VerticalScrollBarBottomButton(Vec2i i_size, Vec2i i_srcPos, const char* i_tip, int i_command, FileCP& i_win, int refCountValue);
	void paint();
	void action();
	Vec2i getSimClickPos(int subButtonIdx){
		return Vec2i(bSize.x/2, bSize.y/2);
	};
	FileCP& win;
};
class PageUp : public Button{
public:
	PageUp(Vec2i i_size, Vec2i i_srcPos, const char* i_tip, int i_command, FileCP& i_win, int refCountValue);
	void paint();
	void action();
	Vec2i getSimClickPos(int subButtonIdx){
		return Vec2i(bSize.x/2, bSize.y/2);
	};
	FileCP& win;
};
class PageDown : public Button{
public:
	PageDown(Vec2i i_size, Vec2i i_srcPos, const char* i_tip, int i_command, FileCP& i_win, int refCountValue);
	void paint(){};
	void action();
	Vec2i getSimClickPos(int subButtonIdx){
		return Vec2i(bSize.x/2, bSize.y/2);
	};
	FileCP& win;
};
	FileCP():
		setupSelectedLine(true),
		numOfVisLines(0),
		scroll(0),
		selLine(0),
		numOfPathFolders(0),
		filter(0),
		root(new File),

		topScroll(Vec2i(16,15), Vec2i(268,176), tip_scroll_up_a_line, 0, *this, 1),
		bottomScroll( Vec2i(16,15), Vec2i(284,176), tip_scroll_down_a_line, 0, *this, 1),
		pageUp( Vec2i(), Vec2i(407,13), tip_scroll_up_a_page, 0, *this, 1),
		pageDown( Vec2i(), Vec2i(), tip_scroll_down_a_page, 0, *this, 1),
		slider( Vec2i(16,15), Vec2i(300,176), tip_drag_to_scroll, 0, *this, 1)
	{
		pixBuf.length=size.x*size.y;
		hog = true;
		if(E_FAIL==initWindow()){
			throw;
		}
	}
	M_<Ptr<SingleButton> > driveList;

	VerticalScrollBarTopButton topScroll;
	VerticalScrollBarBottomButton bottomScroll;
	PageUp pageUp;
	PageDown pageDown;
	VerticalScrollBarSlider slider;

	bool setupSelectedLine;
	int numOfVisLines;
	int scroll;
	int selLine;
	int numOfPathFolders;
	const char* filter;
	void setScrolPos(int offset);
	void resizePageUpPageDown();
	void keepLeagal();
	void buttonHover(Vec2i ptCursor);

	void setWinSize(Vec2i& raw);
	void genNodeList(Node* node);
	void setupNodeWindow(Node* node);
	void paint();
	bool click(const Vec2i ptCursor);
	bool hover(Vec2i ptCursor);

	void openFolder(M_<SceneTreeListItem>& fileTreeList, const int selLine);
	void commonSetupFileTree(String fullFilename, bool filterFilename = false);
	void postCommonSetupFileTree();
	void commonSaveFileAsAction();
	void continueExportPartAction();
	void continueExportPatchAction();
	void procedeWithSavingNewFile();
	void renameFile();
	//void openSmlsLastPhase(String fullFilePath, const bool recover = false);

	void createFullFilePath(const String& filename);
	
	void genNodeListRecursion(const int parent, int indent, File* file);
	Ptr<File> root;
	String workingFolder;
	String fullFilePath;
	
	void highLightTextCell();
	void findAutoCompleteLine(String& cellStr, const int caretPos);
	void onMouseWheel(const int par);
	void onGainFocus();
	void onClickButton(){};
	void onClose(){
		textCell.cellActive = false;
	}
	bool rightClick(const Vec2i ptCursor);
	void fixKeyScroll();
};

static void (* openFunc)(String filename) = 0;
static char (* checkFileExtension)(const char* s) = 0;
static void (* queFunction)() = 0;
void doQueFunction(){
	queFunction();
}
static int fieldIndex = 0;
static FileCP* fileCP = 0;
static String recoverFilename;
static String recoverFilter;
static bool selectLastFile = false;
bool smlsIsSaved();
String droppedFilename;

void (* QuestionMessage3Button::noFunc)() = 0;
void (* QuestionMessage3Button::yesFunc)() = 0;
void (* QuestionMessage3Button::cancelFunc)() = 0;

Ptr<QuestionMessage3Button> g_questionMessage3Button = new QuestionMessage3Button;

extern bool g_browseMode;
extern float g_remVersionLoaded;
extern bool g_forceSaveAs;
extern const char* msg_needPartOrMorpher;

void insertRecentFile(const char* fullFileName);
void changeExtensionToSmls(String& str);
void openMidi(const String& filename);
void openBVH(const String& filename);
void importNodesFromFileField(const String& filename, M_Node& children, int task, Seamless* destSmls = 0);
void importNodesFromFileAndInsert(const String& filename, int importWhere, int task, Seamless* destSmls = 0);
void writeColorToTexture(const char* filename);
bool saveSmlsCommon();
void closeRecentFileWindow();
void openWithHtmlBrowser(String url, const String htmlBrowser);

bool g_dontContinueButtonAction = false;

String g_currentFilename;

static bool createFullFilePathAndCheckFileExists(const String& filename);

static void newSmlsAction(String filename);
static void openSmlsAction(String filename);
static void openVideoAction(String filename);
static void openAudioAction(String filename);
static void saveSmlsAsAction(String filename);
static void importAllNodesAction(String filename);
static void importAllNodesReplaceAction(String filename);
static void importPartAction(String filename);
static void exportPartAction(String filename);
static void importPatchAction(String filename);
static void exportPatchAction(String filename);

static void exportColorToTextureAction(String filename);

static void saveSmlsFirstYes();
static void saveSmlsFirstNo();
static void saveSmlsFirstCancel();

static void openRecentSaveSmlsFirstYes();
static void openRecentSaveSmlsFirstNo();
static void openRecentSaveSmlsFirstCancel();

static void newSmlsOverwriteYes();
static void newSmlsOverwriteNo();
static void newSmlsOverwriteCancel();

static void saveSmlsAsOverwriteYes();
static void saveSmlsAsOverwriteNo();
static void saveSmlsAsOverwriteCancel();

static void exportColorToTextureOverwriteYes();
static void exportColorToTextureOverwriteNo();
static void exportColorToTextureOverwriteCancel();

static void exportPartOverwriteYes();
static void exportPartOverwriteNo();
static void exportPartOverwriteCancel();

static void exportPatchOverwriteYes();
static void exportPatchOverwriteNo();
static void exportPatchOverwriteCancel();

static void recoverYes();
static void recoverNo();
static void recoverCancel();

static void recoverSmlsAction(String filename);

static void deleteRecoverFiles();
bool isDoubleLeftClick();
void createSmlsFromBxxAvStudio();
extern bool g_processComnandLine;
bool openFileFromCommandLine();
void openDroppedFile(void (* i_queFunction)());
void saveSmlsAs(void (* i_queFunction)());
void exitPaletteWindow();

FileCP* fileCPAllreadyOpen(){
	for(int i=Window::visWinList.length; --i>=0;){
		if(FileCP* fileCP = dynamic_cast<FileCP*>(Window::visWinList[i].p)){
			return fileCP;
		}
	}
	return 0;
}

void insertRecentFile(const char* fullFileName){
	M_String& file = g_settings->file.refValue();
	file.length++;
	for(int i = file.length - 1; --i >= 0;){
		file[i + 1] = file[i];
	}
	file[0] = fullFileName;
	g_currentFilename = fullFileName;
	bool skip = false;
	for(int from = 1, to = 1; from < file.length; from++, to++){
		if(file[from] == fullFileName){
			from++;
			skip = true;
		}
		if(from < file.length){
			file[to] = file[from];
		}
	}
	if(skip){
		file.length--;
	}
	if(file.length > 6){
		file.length = 6;
	}
	closeRecentFileWindow();
}


void setTypeStr(const int type, const char* str){
	nodeLookUpStr[type] = str;
}
String getFilename(const String& filename){
	for(int i = filename.length; --i >= 0;){
		if(filename[i] == '\\'){
			return filename.substring(i + 1);
		}
	}
	return "";
}

bool extractFolderName(const String& str, int& from, int& i){
	from = i;
	while(i < str.length){
		if(str[i] == '\\'){
			break;
		}
		i++;
	}
	return i > from;
}
void extractFolderPath(String& path, const String& str){
	for(int i = str.length; --i>=0;){
		if(str[i] == '\\'){
			path = str.substring(0, i - 1);
			break;
		}
	}
}
void buildPathNodes(File*& file, int& selLine, const String& str){
	int i = 0;
	String folderName;
	int from = i;
	while(extractFolderName(str, from, i)){
		File* newFile = new File;
		newFile->name = str.substring(from, i - 1);
		i++;
		file->folder.refValue() += newFile;
		file = newFile;
		selLine++;
	}
}
void addWrlExtension(String& par){
	const String filename = par.toUpperCase();

	int end = filename.length - 4;

	const char* p = filename.c_str() + end;

	if(end >= 0){
		if(	p[0] == '.' &&
			p[1] == 'W' &&
			p[2] == 'R' &&
			p[3] == 'L'){
			return;
		}
		if(	p[0] == '.' &&
			p[1] == 'W' &&
			p[2] == 'R' &&
			p[3] == 'Z'){
			return;
		}
	}

	end -= 3;
	p -= 3;

	if(end >= 0 &&
		p[0] == '.' &&
		p[1] == 'W' &&
		p[2] == 'R' &&
		p[3] == 'L' &&
		p[4] == '.' &&
		p[5] == 'G' &&
		p[6] == 'Z'){
		return;
	}

	if(filename.length && filename[filename.length-1] != '.'){
		par+='.';
	}
	par += "wrl";
}

void addSmlsExtension(String& par){
	const String filename = par.toUpperCase();

	int end = filename.length - 5;

	if(	end >= 0 &&
		filename[end + 0] == '.' &&
		filename[end + 1] == 'S' &&
		filename[end + 2] == 'M' &&
		filename[end + 3] == 'L' &&
		filename[end + 4] == 'S'){
			return;
	}

	if(filename.length && filename[filename.length-1] != '.'){
		par += '.';
	}
	par += "smls";
}

void addBvhExtension(String& par){
	const String filename = par.toUpperCase();

	int end = filename.length - 4;

	if(	end >= 0 &&
		filename[end + 0] == '.' &&
		filename[end + 1] == 'B' &&
		filename[end + 2] == 'V' &&
		filename[end + 3] == 'H'){
			return;
	}

	if(filename.length && filename[filename.length-1] != '.'){
		par += '.';
	}
	par += "bvh";
}

void addPngExtension(String& par){
	const String filename = par.toUpperCase();

	int end = filename.length - 4;

	const char* p = filename.c_str() + end;

	if(end >= 0){
		if(	p[0] == '.' &&
			p[1] == 'P' &&
			p[2] == 'N' &&
			p[3] == 'G'){
			return;
		}
	}
	if(filename.length && filename[filename.length-1] != '.'){
		par += '.';
	}
	par += "png";
}

void addWavExtension(String& par){
	const String filename = par.toUpperCase();

	int end = filename.length - 4;

	const char* p = filename.c_str() + end;

	if(end >= 0){
		if(	p[0] == '.' &&
			p[1] == 'W' &&
			p[2] == 'A' &&
			p[3] == 'V'){
			return;
		}
	}
	if(filename.length && filename[filename.length-1] != '.'){
		par += '.';
	}
	par += "wav";
}

void addOpenExtension(String& par){
	const String filename = par.toUpperCase();

	int end = filename.length - 4;

	if(	end >= 0 &&
		filename[end + 0] == '.' &&
		filename[end + 1] == 'W' &&
		filename[end + 2] == 'R' &&
		filename[end + 3] == 'L'){
			return;
	}
	if(	end >= 0 &&
		filename[end + 0] == '.' &&
		filename[end + 1] == 'W' &&
		filename[end + 2] == 'R' &&
		filename[end + 3] == 'Z'){
			return;
	}
	if(	end >= 0 &&
		filename[end + 0] == '.' &&
		filename[end + 1] == 'O' &&
		filename[end + 2] == 'B' &&
		filename[end + 3] == 'J'){
			return;
	}

	if(	end >= 0 &&
		filename[end + 0] == '.' &&
		filename[end + 1] == 'B' &&
		filename[end + 2] == 'V' &&
		filename[end + 3] == 'H'){
			return;
	}

	if(	end >= 0 &&
		filename[end + 0] == '.' &&
		filename[end + 1] == 'M' &&
		filename[end + 2] == 'I' &&
		filename[end + 3] == 'D'){
			return;
	}

	end--;

	if(	end >= 0 &&
		filename[end + 0] == '.' &&
		filename[end + 1] == 'S' &&
		filename[end + 2] == 'M' &&
		filename[end + 3] == 'L' &&
		filename[end + 4] == 'S'){
			return;
	}
	if(	end >= 0 &&
		filename[end + 0] == '.' &&
		filename[end + 1] == 'X' &&
		filename[end + 2] == '3' &&
		filename[end + 3] == 'D' &&
		filename[end + 4] == 'V'){
			return;
	}

	end--;

	if(	end >= 0 &&
		filename[end + 0] == '.' &&
		filename[end + 1] == 'X' &&
		filename[end + 2] == '3' &&
		filename[end + 3] == 'D' &&
		filename[end + 4] == 'V' &&
		filename[end + 5] == 'Z'){
			return;
	}

	end--;
	if(	end >= 0 &&
		filename[end + 0] == '.' &&
		filename[end + 1] == 'W' &&
		filename[end + 2] == 'R' &&
		filename[end + 3] == 'L' &&
		filename[end + 4] == '.' &&
		filename[end + 5] == 'G' &&
		filename[end + 6] == 'Z'){
			return;
	}

	if(filename.length && filename[filename.length-1] != '.'){
		par+='.';
	}
	par += "smls";
}

void openSmlsLastPhase(String fullFilePath, const bool recover = false){

	String dir;
	String ext;
	String drive;
	MultiSelect::reset();

	splitPath(fullFilePath, drive, dir, String(), ext);
	if(recover == false){
		g_seamlessDir = drive;
		g_seamlessDir += dir;
	}
	_chdir(g_seamlessDir.c_str());

	String EXT = ext.toUpperCase();
	if(EXT.length){
		EXT = EXT.substring(1);
	}

	g_lookupProtoInterface.length=g_userProtoBegin;
	nodeLookUpStr.length=g_userProtoBegin;
	containerFieldLookUpStr.length=g_userProtoBegin;

	if(EXT == "MID"){
		openMidi(fullFilePath);
		return;
	}
	if(EXT == "BVH"){
		openBVH(fullFilePath);
		return;
	}

	initCompile();
	Browser::createSceneFromFile(fullFilePath);
	g_wireFrameMode=g_scene->wireframe;
	g_scene->genNodeList2();

	decodeFunctions(g_loadStr);
	g_scene->genNodeList2();
	postLoadProcessNodes(g_sceneTreeList);

	g_scene->postLoadProcessCollapse();
	resetLooseEnds();
	Window::postLoadProcessWin(*g_scene);
	if(g_remVersionLoaded < 2.093-.00001){
		g_stw->selectFirstPart();
	}
	Viewpoint::gotoFirstViewpoint();
	setTitleName();
	initializeScript();

	if(recover){
		insertRecentFile(recoverFilename.c_str());
	}
	else{
		insertRecentFile(fullFilePath.c_str());
		g_settings->writeINIFile();
	}
	BackUp::deleteBackupFiles();
	BackUp::createUniqueBackupName(true);
	if(recover){
		BackUp::smlsSaved.unSet();
		g_forceSaveAs = true;
	}
	else{
		recoverFilename = "";
		g_forceSaveAs = false;
	}
	if(EXT!="SMLS"){
//		changeExtensionToSmls(fullFilePath);
		g_forceSaveAs = true;
	}
	g_tb->updateToolBarContext();
	setTitleName();
	g_stw->paint();
	Window::paintAllControlPanels();

	renderScene();

	if(Anim* anim = Window::animMode()){
		houseKeepVisualClue();
		if(anim->play == false){
			anim->update = true;
			anim->updateDisplayFrom();
		}
	}
}

void openSmlsAction(String filename){
	addOpenExtension(filename);

	if(createFullFilePathAndCheckFileExists(filename)){
		openSmlsLastPhase(fileCP->fullFilePath);
	}
}

void openVideoAction(String filename){
	addPngExtension(filename);

	if(false == createFullFilePathAndCheckFileExists(filename)){
		return;
	}

	filename = fileCP->fullFilePath;
	fileCP->postCommonSetupFileTree();

	g_vcp->openMovieFile(filename.c_str());
}

void openAudioAction(String filename){
	addWavExtension(filename);

	if(false == createFullFilePathAndCheckFileExists(filename)){
		return;
	}

	filename = fileCP->fullFilePath;
	fileCP->postCommonSetupFileTree();

	g_wtpcp->openWaveFile(filename.c_str());
}

void recoverSmlsAction(String filename){
	if(createFullFilePathAndCheckFileExists(filename)){
		openSmlsLastPhase(fileCP->fullFilePath, true);
	}
}

void importAllNodesAction(String filename){
	addOpenExtension(filename);

	if(false == createFullFilePathAndCheckFileExists(filename)){
		return;
	}

	filename = fileCP->fullFilePath;
	fileCP->postCommonSetupFileTree();

	importNodesFromFileField(filename, dynamic_cast<Progenitor*>(g_selNode.getRaw())->children, 0);
}

Seamless* getSeamless(int& selLine);

void importBuildNodesAction(String filename){
	addSmlsExtension(filename);

	if(false == createFullFilePathAndCheckFileExists(filename)){
		return;
	}

	filename = fileCP->fullFilePath;
	fileCP->postCommonSetupFileTree();

	MFNode* mfnode = 0;
	if(Seamless* smls = dynamic_cast<Seamless*>(g_selNode.getRaw())){
		mfnode = &smls->build;
	}
	else if(GroupBuild* group = dynamic_cast<GroupBuild*>(g_selNode.getRaw())){
		mfnode = &group->build;
	}

	importNodesFromFileField(filename, *mfnode, ONLY_BUILD, getSeamless2());
}

void importAllNodesReplaceAction(String filename){
	addOpenExtension(filename);

	if(false == createFullFilePathAndCheckFileExists(filename)){
		return;
	}

	filename = fileCP->fullFilePath;
	fileCP->postCommonSetupFileTree();

	importNodesFromFileAndInsert(filename, IMPORT_REPLACE, 0);
}


void addImportNodeAboveAction(String filename){
	addOpenExtension(filename);

	if(false == createFullFilePathAndCheckFileExists(filename)){
		return;
	}

	filename = fileCP->fullFilePath;
	fileCP->postCommonSetupFileTree();

	importNodesFromFileAndInsert(filename, IMPORT_ABOVE, 0);
}

void addImportNodeBelowAction(String filename){
	addOpenExtension(filename);

	if(false == createFullFilePathAndCheckFileExists(filename)){
		return;
	}

	filename = fileCP->fullFilePath;
	fileCP->postCommonSetupFileTree();

	importNodesFromFileAndInsert(filename, IMPORT_BELOW, 0);
}


void importAnimationReplaceAction(String filename){
	addOpenExtension(filename);

	if(false == createFullFilePathAndCheckFileExists(filename)){
		return;
	}

	filename = fileCP->fullFilePath;
	fileCP->postCommonSetupFileTree();

	importNodesFromFileAndInsert(filename, IMPORT_REPLACE, ONLY_ANIMATION);
}

void addImportAnimAboveAction(String filename){
	addOpenExtension(filename);

	if(false == createFullFilePathAndCheckFileExists(filename)){
		return;
	}

	filename = fileCP->fullFilePath;
	fileCP->postCommonSetupFileTree();

	importNodesFromFileAndInsert(filename, IMPORT_ABOVE, ONLY_ANIMATION);
}

void addImportAnimBelowAction(String filename){
	addOpenExtension(filename);

	if(false == createFullFilePathAndCheckFileExists(filename)){
		return;
	}

	filename = fileCP->fullFilePath;
	fileCP->postCommonSetupFileTree();

	importNodesFromFileAndInsert(filename, IMPORT_BELOW, ONLY_ANIMATION);
}


void importBuildReplaceAction(String filename){
	addSmlsExtension(filename);

	if(false == createFullFilePathAndCheckFileExists(filename)){
		return;
	}

	filename = fileCP->fullFilePath;
	fileCP->postCommonSetupFileTree();

	importNodesFromFileAndInsert(filename, IMPORT_REPLACE, ONLY_BUILD, getSeamless2());
}

void addImportBuildAboveAction(String filename){
	addSmlsExtension(filename);

	if(false == createFullFilePathAndCheckFileExists(filename)){
		return;
	}

	filename = fileCP->fullFilePath;
	fileCP->postCommonSetupFileTree();

	importNodesFromFileAndInsert(filename, IMPORT_ABOVE, ONLY_BUILD, getSeamless2());
}

void addImportBuildBelowAction(String filename){
	addSmlsExtension(filename);

	if(false == createFullFilePathAndCheckFileExists(filename)){
		return;
	}

	filename = fileCP->fullFilePath;
	fileCP->postCommonSetupFileTree();

	importNodesFromFileAndInsert(filename, IMPORT_BELOW, ONLY_BUILD, getSeamless2());
}


void importAnimationAction(String filename){
	addOpenExtension(filename);

	if(false == createFullFilePathAndCheckFileExists(filename)){
		return;
	}

	filename = fileCP->fullFilePath;
	fileCP->postCommonSetupFileTree();

	importNodesFromFileField(filename, dynamic_cast<Progenitor*>(g_selNode.getRaw())->children, ONLY_ANIMATION);
}
void importAnimRecalcAction(String filename){
	addOpenExtension(filename);

	if(false == createFullFilePathAndCheckFileExists(filename)){
		return;
	}

	filename = fileCP->fullFilePath;
	fileCP->postCommonSetupFileTree();

	importNodesFromFileField(filename, dynamic_cast<Progenitor*>(g_selNode.getRaw())->children, RECALC);
}

void importPivotPointsAction(String filename){
	addSmlsExtension(filename);

	if(false == createFullFilePathAndCheckFileExists(filename)){
		return;
	}

	filename = fileCP->fullFilePath;
	fileCP->postCommonSetupFileTree();

	importNodesFromFileField(filename, g_scene->children, IMPORT_PIVOT_POINTS);
}

void importPivotPointDistancesAction(String filename){
	addSmlsExtension(filename);

	if(false == createFullFilePathAndCheckFileExists(filename)){
		return;
	}

	filename = fileCP->fullFilePath;
	fileCP->postCommonSetupFileTree();

	importNodesFromFileField(filename, g_scene->children, IMPORT_PIVOT_POINTS_DISTANCES);
}

/*
bool compareUpperOrLower(const char* a, const char* b, int len){
	for(int i = len; --i >= 0;){
		char ca = upperCaseChar(a[i]);
		char cb = upperCaseChar(b[i]);
		if(ca != cb){
			return false;
		}
	}
	return true;
}
*/

char checkSmlsFileExtensions(const char* s){
	int len = strlen(s);
	if(len<5){
		return false;
	}
	const char* d = &s[len-5];
	if(!stricmp(d,".smls")){
		return File::SMLS;
	}
	return 0;
}

char checkAllFileExtensions(const char* s){
	int len = strlen(s);
	if(len<5){
		return false;
	}
	const char* d = &s[len-4];
	if(!stricmp(d,".wrl")){
		return File::VRML;
	}
	if(!stricmp(d,".wrz")){
		return File::VRML;
	}
	if(!stricmp(d,".bvh")){
		return File::BVH;
	}
	if(!stricmp(d,".mid")){
		return File::MID;
	}
	d--;
	if(!stricmp(d,".smls")){
		return File::SMLS;
	}
	if(!stricmp(d,".x3dv")){
		return File::VRML;
	}
	d--;
	if(!stricmp(d,".x3dvz")){
		return File::VRML;
	}

	return 0;
}

char checkWrlFileExtension(const char* s){
	int len = strlen(s);
	if(len<5){
		return false;
	}
	const char* d = &s[len-4];
	if(!stricmp(d,".wrl")){
		return File::VRML;
	}
	if(!stricmp(d,".wrz")){
		return File::VRML;
	}
	d--;
	if(!stricmp(d,".x3dv")){
		return File::VRML;
	}
	d--;
	if(!stricmp(d,".x3dvz")){
		return File::VRML;
	}
	return 0;
}

char checkWavFileExtension(const char* s){
	int len = strlen(s);
	if(len<5){
		return false;
	}
	const char* d = &s[len-4];
	if(!stricmp(d,".wav")){
		return File::WAV;
	}
	return false;
}

char checkPngFileExtension(const char* s){
	int len = strlen(s);
	if(len<5){
		return false;
	}
	const char* d = &s[len-4];
	if(!stricmp(d,".png")){
		return File::PNG;
	}
	return false;
}

char checkBvhFileExtension(const char* s){
	int len = strlen(s);
	if(len<5){
		return false;
	}
	const char* d = &s[len-4];
	if(!stricmp(d,".bvh")){
		return File::BVH;
	}
	return false;
}

void bubbleSortByName(M_Node& m_node){
	bool flag = false;
	do{
		flag = false;
		for(int i = 0; i < m_node.length - 1; i++){

			Ptr<File> fa = dynamic_cast<File*>(m_node[i].p);
			Ptr<File> fb = dynamic_cast<File*>(m_node[i + 1].p);

			if((fa->isFile > fb->isFile) || (fa->isFile == fb->isFile && fa->name.greaterThanCaseInsensitive(fb->name))){
				m_node[i + 1] = fa;
				m_node[i] = fb;

				flag = true;
			}
		}
	}while(flag);
}

void bubbleSortByDate(M_Node& m_node){
	bool flag = false;
	do{
		flag = false;
		for(int i = 0; i < m_node.length - 1; i++){

			Ptr<File> fa = dynamic_cast<File*>(m_node[i].p);
			Ptr<File> fb = dynamic_cast<File*>(m_node[i + 1].p);

			if((fa->isFile > fb->isFile) || (fa->isFile == fb->isFile && fa->time > fb->time)){
			//if(fa->time > fb->time){
				m_node[i + 1] = fa;
				m_node[i] = fb;

				flag = true;
			}
		}
	}while(flag);
}


void FileCP::commonSetupFileTree(String fullFilename, bool filterFilename){
	if(g_browseMode){
		return;
	}
	WIN32_FIND_DATA w32fd;
	HANDLE SearchHandle;
	extractFolderPath(workingFolder, fullFilename);

	File* file = root;
	file->folder.refValue().length = 0;
	buildPathNodes(file, selLine = -1, workingFolder);
	numOfPathFolders = selLine;

	String bPath = workingFolder;
	bPath += '\\';

	if(filterFilename){
		bPath += root->filename;
	}
	else{
		bPath += filter;
	}

	if((SearchHandle = FindFirstFile(bPath.c_str(), &w32fd)) != INVALID_HANDLE_VALUE){
		do{
			if((w32fd.dwFileAttributes&FILE_ATTRIBUTE_DIRECTORY)==FILE_ATTRIBUTE_DIRECTORY){
				if(!stricmp(w32fd.cFileName,"."))continue;
				if(!stricmp(w32fd.cFileName,".."))continue;
				File* folder = new File;
				folder->name = w32fd.cFileName;
				folder->time.hi = w32fd.ftLastWriteTime.dwHighDateTime;
				folder->time.lo = w32fd.ftLastWriteTime.dwLowDateTime;
				file->folder += folder;
			}
			else{
				char type = checkFileExtension(w32fd.cFileName);
				if(type == 0)continue;
				File* newFile = new File;
				newFile->isFile = true;
				newFile->type = type;
				newFile->name = w32fd.cFileName;
				newFile->time.hi = w32fd.ftLastWriteTime.dwHighDateTime;
				newFile->time.lo = w32fd.ftLastWriteTime.dwLowDateTime;
				file->folder += newFile;
			}
		}while(FindNextFile(SearchHandle, &w32fd));
	}
	if(root->sortByDate){
		bubbleSortByDate(file->folder.refValue());
	}
	else{
		bubbleSortByName(file->folder.refValue());
	}

	if(root->folder.length){
		genNodeList(root->folder.refValue()[0].p);
	}
	if(setupSelectedLine){
		scroll = 0;
		bool matchingNameFound = false;
		if(selectLastFile){
			selLine = fileTreeList.length - 1;
		}
		else{
			for(int k = fileTreeList.length; --k >= 0;){
				if(fileTreeList[k].node->name == root->filename){
					selLine = k;
					matchingNameFound = true;
					break;
				}
			}
		}
		if(selLine >= fileTreeList.length){
			selLine = fileTreeList.length - 1;
		}
		if(selLine < 0){
			selLine = 0;
		}
		if(matchingNameFound || selectLastFile){
			root->filename = textCell.cellStr = fileTreeList[selLine].node->name;
		}
		if(selLine >= numOfVisLines){
			if(selectLastFile){
				setScrolPos(selLine - (numOfVisLines - 1));
			}
			else {
				int offset = selLine - (numOfVisLines / 2);
				if(fileTreeList.length - offset < numOfVisLines){
					offset = fileTreeList.length - numOfVisLines;
				}
				setScrolPos(offset);
			}
		}
		setupSelectedLine = false;
		selectLastFile = false;
	}

	paint();
}
void FileCP::openFolder(M_<SceneTreeListItem>& fileTreeList, const int selLine){
	int line = selLine;

	M_Node m_folder;

	while(true){
		m_folder += fileTreeList[line].node;
		if(line == 0) break;
		line = fileTreeList[line].parent;
	}

	workingFolder = "";
	for(int k = m_folder.length; --k>=0;){
		workingFolder += m_folder[k]->name;
		workingFolder += '\\';
	}

	commonSetupFileTree(workingFolder);
	
	scroll = 0;
	if(this->selLine >= numOfVisLines){
		int offset = this->selLine - (numOfVisLines / 2);
		if(fileTreeList.length - offset < numOfVisLines){
			offset = fileTreeList.length - numOfVisLines;
		}
		setScrolPos(offset);
	}
	else{
		setScrolPos(0);
	}
}
static void setupFileTree(const char* title, const char* openName, void (*i_openFunc)(String filename), char (* i_checkFileExtension)(const char* s), String fullFilename, bool sortByDate, const char* filter = "*.*" ){
	Window::activeWindow->onSeamlessWindowsLoseFocus();
	setTypeStr(FILE_NODE, title);
	g_lookupProtoInterface[FILE_NODE]->properties[2].name = openName;

	openFunc = i_openFunc;
	checkFileExtension = i_checkFileExtension;

	fileCP = new FileCP;
	fileCP->filter = filter;
	fileCP->root->sortByDate.refValue() = sortByDate;

	if(fullFilename == ""){
		fileCP->fullFilePath = g_seamlessDir;
	}
	else{
		fileCP->fullFilePath = fullFilename;
		String ext;
		splitPath(fullFilename, String(), String(), fileCP->root->filename.refValue(), ext);
		fileCP->root->filename.refValue() += ext;
	}
	
	fileCP->setupNodeWindow(fileCP->root);
	Window::activeWindow = fileCP;

	fileCP->commonSetupFileTree(fileCP->fullFilePath);
	fileCP->highLightTextCell();
	fileCP->paint();
}

void FileCP::postCommonSetupFileTree(){
	Window::closeWindow(this);
}


void FileCP::commonSaveFileAsAction(){
	String dir;
	splitPath(fullFilePath, g_seamlessDir, dir, String(), String() );
	g_seamlessDir += dir;

	g_settings->writeINIFile();
	
	writeSmlsFile(fullFilePath);

	BackUp::deleteBackupFiles();
	BackUp::createUniqueBackupName();

	insertRecentFile(fullFilePath.c_str());
	g_forceSaveAs = false;
	resetLooseEnds();
	if(recoverFilename != ""){
		deleteRecoverFiles();
	}
	setSmlsSaved();
}
void FileCP::continueExportPartAction(){
	dynamic_cast<Art*>(g_selNode.getRaw())->saveToWrl(fullFilePath);
}

NurbsPatch* g_importExportPatch = 0;

void FileCP::continueExportPatchAction(){
	g_importExportPatch->saveToWrl(fullFilePath);
}

void addNodesForNewFile(){
	Seamless* seamless=new Seamless;
	seamless=new Seamless;
	seamless->effect=new ColorEffect;
	seamless->skeleton+=new Part;
	seamless->process();
	g_scene->children.refValue().length = 0;
	g_scene->children.refValue() += seamless;
	g_scene->genNodeList2();
	g_scene->selLine = 3;
	g_scene->scroll = 0;
	g_currentFilename = g_seamlessDir;
	g_currentFilename += "untitled.smls";
}
void FileCP::procedeWithSavingNewFile(){
	addNodesForNewFile();

	g_stw->paint();

	commonSaveFileAsAction();
	postCommonSetupFileTree();
}
void newSmlsAction(String filename){
	addSmlsExtension(filename);
	fileCP->createFullFilePath(filename);

	FILE* stream = fopen(fileCP->fullFilePath.c_str(), "r+b");
	if(stream){
		fclose(stream);

		String msg = "overwrite ";
		msg += filename;
		msg += " ?";

		QuestionMessage3Button::setupMessageWindow(msg, newSmlsOverwriteYes, newSmlsOverwriteNo, newSmlsOverwriteCancel);
	}
	else{
		fileCP->procedeWithSavingNewFile();
	}
}
void FileCP::createFullFilePath(const String& filename){
	fullFilePath = workingFolder;
	fullFilePath += '\\';
	fullFilePath += filename;
}

void saveSmlsAsAction(String filename){
	addSmlsExtension(filename);
	fileCP->createFullFilePath(filename);

	FILE* stream = fopen(fileCP->fullFilePath.c_str(), "r+b");
	if(stream){
		fclose(stream);

		String msg = "overwrite ";
		msg += filename;
		msg += " ?";

		QuestionMessage3Button::setupMessageWindow(msg, saveSmlsAsOverwriteYes, saveSmlsAsOverwriteNo, saveSmlsAsOverwriteCancel);
	}
	else{
		fileCP->commonSaveFileAsAction();
		fileCP->postCommonSetupFileTree();
		doQueFunction();
	}
}
void exportPartAction(String filename){
	addWrlExtension(filename);
	fileCP->createFullFilePath(filename);

	FILE* stream = fopen(fileCP->fullFilePath.c_str(), "r+b");
	if(stream){
		fclose(stream);

		String msg = "overwrite ";
		msg += filename;
		msg += " ?";

		QuestionMessage3Button::setupMessageWindow(msg, exportPartOverwriteYes, exportPartOverwriteNo, exportPartOverwriteCancel);
	}
	else{
		fileCP->continueExportPartAction();
		fileCP->postCommonSetupFileTree();
	}
}

void exportPatchAction(String filename){
	addWrlExtension(filename);
	fileCP->createFullFilePath(filename);

	FILE* stream = fopen(fileCP->fullFilePath.c_str(), "r+b");
	if(stream){
		fclose(stream);

		String msg = "overwrite ";
		msg += filename;
		msg += " ?";

		QuestionMessage3Button::setupMessageWindow(msg, exportPatchOverwriteYes, exportPatchOverwriteNo, exportPatchOverwriteCancel);
	}
	else{
		fileCP->continueExportPatchAction();
		fileCP->postCommonSetupFileTree();
	}
}

void exportColorToTextureAction(String filename){
	addPngExtension(filename);

	fileCP->createFullFilePath(filename);
	FILE* stream = fopen(fileCP->fullFilePath.c_str(), "rb");
	if(stream){
		fclose(stream);

		String msg = "overwrite ";
		msg += filename;
		msg += " ?";
		
		QuestionMessage3Button::setupMessageWindow(msg, exportColorToTextureOverwriteYes, exportColorToTextureOverwriteNo, exportColorToTextureOverwriteCancel);
	}
	else{
		writeColorToTexture(fileCP->fullFilePath.c_str());
		fileCP->postCommonSetupFileTree();
	}
}
void FileCP::renameFile(){

	String dir;
	String drive;
	String filename;
	String ext;
	splitPath(fullFilePath, drive, dir, filename, ext);
	String newName = drive;
	newName += dir;
	newName += filename;

	String newName2;
	int i = 0;
	do{
		newName2 = newName;
		if(i < 10){
			newName2 += '0';
		}
		newName2 += i;
		i++;
		newName2 += ext;
	}while(rename(fullFilePath.c_str(), newName2.c_str()));
}
void newSmlsOverwriteYes(){
	fileCP->renameFile();
	fileCP->procedeWithSavingNewFile();
	Window::closeWindow(Window::activeWindow.getActiveWindow());
}
void newSmlsOverwriteNo(){
	Window::closeWindow(Window::activeWindow.getActiveWindow());
}
void newSmlsOverwriteCancel(){
	Window::closeWindow(Window::activeWindow.getActiveWindow());
}

void saveSmlsAsOverwriteYes(){
	fileCP->renameFile();
	fileCP->commonSaveFileAsAction();
	Window::closeWindow(Window::activeWindow.getActiveWindow());
	fileCP->postCommonSetupFileTree();
	doQueFunction();
}
void saveSmlsAsOverwriteNo(){
	Window::closeWindow(Window::activeWindow.getActiveWindow());
	doQueFunction();
}
void saveSmlsAsOverwriteCancel(){
	Window::closeWindow(Window::activeWindow.getActiveWindow());
}

void exportPartOverwriteYes(){
	fileCP->renameFile();
	Window::closeWindow(Window::activeWindow.getActiveWindow());
	fileCP->continueExportPartAction();
	fileCP->postCommonSetupFileTree();
}

void exportPatchOverwriteYes(){
	fileCP->renameFile();
	Window::closeWindow(Window::activeWindow.getActiveWindow());
	fileCP->continueExportPatchAction();
	fileCP->postCommonSetupFileTree();
}

void exportPartOverwriteNo(){
	Window::closeWindow(Window::activeWindow.getActiveWindow());
}

void exportPatchOverwriteNo(){
	Window::closeWindow(Window::activeWindow.getActiveWindow());
}

void exportPartOverwriteCancel(){
	Window::closeWindow(Window::activeWindow.getActiveWindow());
}

void exportPatchOverwriteCancel(){
	Window::closeWindow(Window::activeWindow.getActiveWindow());
}

void exportColorToTextureOverwriteYes(){
	fileCP->renameFile();
	writeColorToTexture(fileCP->fullFilePath.c_str());
	Window::closeWindow(Window::activeWindow.getActiveWindow());
	fileCP->postCommonSetupFileTree();
}
void exportColorToTextureOverwriteNo(){
	Window::closeWindow(Window::activeWindow.getActiveWindow());
}
void exportColorToTextureOverwriteCancel(){
	Window::closeWindow(Window::activeWindow.getActiveWindow());
}

void exportAnimToBvhYes(){
	fileCP->renameFile();
	if(Anim* anim = dynamic_cast<Anim*>(g_selNode.getRaw())){
		anim->saveBvh(fileCP->fullFilePath.c_str());
	}
	Window::closeWindow(Window::activeWindow.getActiveWindow());
	fileCP->postCommonSetupFileTree();
}
void exportAnimToBvhNo(){
	Window::closeWindow(Window::activeWindow.getActiveWindow());
}
void exportAnimToBvhCancel(){
	Window::closeWindow(Window::activeWindow.getActiveWindow());
}

void exportAnimToBvhAction(String filename){
	addBvhExtension(filename);
	fileCP->createFullFilePath(filename);
	FILE* stream = fopen(fileCP->fullFilePath.c_str(), "rb");
	if(stream){
		fclose(stream);

		String msg = "overwrite ";
		msg += filename;
		msg += " ?";
		
		QuestionMessage3Button::setupMessageWindow(msg, exportAnimToBvhYes, exportAnimToBvhNo, exportAnimToBvhCancel);
	}
	else{
		if(Anim* anim = Window::animMode()){
			anim->saveBvh(filename.c_str());
		}
		fileCP->postCommonSetupFileTree();
	}
}

bool createFullFilePathAndCheckFileExists(const String& filename){
	fileCP->createFullFilePath(filename);
	FILE* stream = fopen(fileCP->fullFilePath.c_str(), "rb");
	if(stream == 0){
		MessageWin::msg(filename, String(" does not exist!"), String(), MB_OK | MB_ICONEXCLAMATION);
		return false;
	}
	fclose(stream);
	return true;
}
void importPartAction(String filename){
	addWrlExtension(filename);

	if(false == createFullFilePathAndCheckFileExists(filename)){
		return;
	}

	g_loadStr = readFile(fileCP->fullFilePath);

	fileCP->postCommonSetupFileTree();

	String& str = g_loadStr;

	if(str==""){
		g_stw->paint();
		Window::paintAllControlPanels();
		return;
	}

	if(Art* part=dynamic_cast<Art*>(g_selNode.getRaw())){
		fastFullBuildForAutoIn(true);
		part->createPartFromString(str);
		part->smls->doBuild();
		updateBuild3dWindow();
	}
	else if(Morpher* morpher=dynamic_cast<Morpher*>(g_selNode.getRaw())){
		CopyToMorph copyToMorph;
		RelateMorpher relateMorpher;
		CopyPart copyPart;
		Seamless* smls=morpher->smls;
		Art* parent=morpher->parent;
		Art* part=new Part;
		parent->part+=part;
		part->morpher+=morpher;
		part->createPartFromString(str);
		copyToMorph.morpher=morpher;
		copyToMorph.part=part;
		copyToMorph.doBuild();
		copyPart.source.refValue()+=parent;
		copyPart.dest.refValue()+=part;
		copyPart.doBuild();
		relateMorpher.part=part;
		relateMorpher.doBuild();
		parent->part.refValue().length--;
		smls->process();
		g_stw->cp->paint();
		part->smls->ctrf=true;
		updateBuild3dWindow();
	}
	else if(NurbsPatch* patch = dynamic_cast<NurbsPatch*>(g_selNode.getRaw())){
		patch->loadDataFromString(str);
		updateBuild3dWindow();
	}
	else{
		msG(msg_needPartOrMorpher);
		return;
	}
	BackUp::monostable.beginCycle();
	g_stw->cp->setupNodeWindow(g_selNode.getRaw());
	g_stw->paint();
}

void importPatchAction(String filename){
	addWrlExtension(filename);

	if(false == createFullFilePathAndCheckFileExists(filename)){
		return;
	}

	g_loadStr = readFile(fileCP->fullFilePath);

	fileCP->postCommonSetupFileTree();

	String& str = g_loadStr;

	if(str==""){
		g_stw->paint();
		Window::paintAllControlPanels();
		return;
	}

	g_importExportPatch->loadDataFromString(str);
	updateBuild3dWindow();

	BackUp::monostable.beginCycle();
	g_stw->cp->setupNodeWindow(g_selNode.getRaw());
	g_stw->paint();
}

void importBxxAvStudioAction(String filename){

	if(false == createFullFilePathAndCheckFileExists(filename)){
		return;
	}
	filename = fileCP->fullFilePath;
	fileCP->postCommonSetupFileTree();

	addWrlExtension(filename);

	String dir;
	String ext;
	splitPath(filename, g_seamlessDir, dir, String(), ext);
	g_seamlessDir += dir;
	_chdir(g_seamlessDir.c_str());


	String& str = g_loadStr;
///
	g_online = false;
	g_viewer = false;

	g_lookupProtoInterface.length=g_userProtoBegin;
	nodeLookUpStr.length=g_userProtoBegin;
	containerFieldLookUpStr.length=g_userProtoBegin;

	initCompile();
	Browser::createSceneFromFile(filename);

	g_wireFrameMode=g_scene->wireframe;
	g_scene->genNodeList2();

	postLoadProcessNodes(g_sceneTreeList);

	g_scene->postLoadProcessCollapse();
	resetLooseEnds();
	Window::postLoadProcessWin(*g_scene);
	
	createSmlsFromBxxAvStudio();

	String EXT = ext.toUpperCase();
	if(EXT.length){
		EXT = EXT.substring(1);
	}
	if(EXT!="SMLS"){
		changeExtensionToSmls(filename);
		g_forceSaveAs = true;
	}
	g_currentFilename = filename;

	setTitleName();
}

bool smlsIsSaved(){
	if(g_browseMode){
		return true;
	}
	return BackUp::smlsSaved;
}

void File::SortByDate::operator = (const bool& par){
	g_dontContinueButtonAction = true;
	refValue() = par;
	g_settings->sortByDate = par;
	fileCP->setupSelectedLine = true;
	fileCP->commonSetupFileTree(String());
}

bool scanFilenameForDrive(const String& filename){
	if(filename.length > 1){
		if(filename[1] == ':'){
			return true;
		}
	}
	return false;
}

bool scanFilenameForAsterisk(const String& filename){
	for(int i = filename.length; --i >= 0;){
		if(filename[i] == '*'){
			return true;
		}
	}
	return false;
}

void File::Open::operator = (const bool& par){
	g_dontContinueButtonAction = true;
	String filename = home.filename;
	if(scanFilenameForDrive(filename)){
		fileCP->setScrolPos(fileCP->scroll = 0);
		if(filename.length == 2){
			filename += '\\';
		}
		fileCP->commonSetupFileTree(filename);
	}
	else if(scanFilenameForAsterisk(filename)){
		fileCP->setScrolPos(fileCP->scroll = 0);
		fileCP->commonSetupFileTree(fileCP->fullFilePath, true);
	}
	else{
		removeTypeFromList <DownButton>(g_callBackList);
		
		openFunc(home.filename);
	}
}
void exitFileWindow(){
	removeTypeFromList <DownButton>(g_callBackList);
	fileCP->postCommonSetupFileTree();
}
bool saveSmlsCommon(){
	g_stw->textCell.cellActive=false;
	if(writeSmlsFile(g_currentFilename)){
		setSmlsSaved();
		BackUp::backupToCache(true);
		return true;
	}
	return false;
}
void saveSmls(){
	if(g_forceSaveAs){
		saveSmlsAs();
	}
	else if(false == saveSmlsCommon()){
		saveSmlsAs();
	}
}

void openRecentSaveSmlsFirstYes(){

	Window::closeWindow(Window::activeWindow.getActiveWindow());

	if(g_forceSaveAs){
		saveSmlsAs();
	}
	else if(saveSmlsCommon()){
		openSmlsLastPhase(g_settings->file[fieldIndex]);
	}
	else{
		saveSmlsAs();
	}
}
void openRecentSaveSmlsFirstNo(){
	openSmlsLastPhase(g_settings->file[fieldIndex]);
}
void openRecentSaveSmlsFirstCancel(){
	Window::closeWindow(Window::activeWindow.getActiveWindow());
}


void returnToDropFile(){
	openSmlsLastPhase(droppedFilename);
}


void saveSmlsFirstYes(){
	if(g_forceSaveAs){
		Window::closeWindow(Window::activeWindow.getActiveWindow());
		saveSmlsAs(queFunction);
	}
	else{
		g_stw->textCell.cellActive = false;
		
		if(writeSmlsFile(g_currentFilename)){
			setSmlsSaved();
			recoverFilename = "";
			g_settings->writeINIFile();
			Window::closeWindow(Window::activeWindow.getActiveWindow());
			queFunction();
		}
		else{
			Window::closeWindow(Window::activeWindow.getActiveWindow());
			saveSmlsAs(queFunction);
		}
	}
}

void saveSmlsFirstNo(){
	Window::closeWindow(Window::activeWindow.getActiveWindow());
	queFunction();
}

void saveSmlsFirstCancel(){
	Window::closeWindow(Window::activeWindow.getActiveWindow());
}

void dropFileSaveSmlsFirstYes(){
	if(g_forceSaveAs){
		Window::closeWindow(Window::activeWindow.getActiveWindow());
		saveSmlsAs(returnToDropFile);
	}
	else{
		g_stw->textCell.cellActive = false;
		
		if(writeSmlsFile(g_currentFilename)){
			setSmlsSaved();
			recoverFilename = "";
			g_settings->writeINIFile();
			Window::closeWindow(Window::activeWindow.getActiveWindow());
			openSmlsLastPhase(droppedFilename);
		}
		else{
			Window::closeWindow(Window::activeWindow.getActiveWindow());
			saveSmlsAs(returnToDropFile);
		}
	}
}
void dropFileSaveSmlsFirstNo(){
	Window::closeWindow(Window::activeWindow.getActiveWindow());
	openSmlsLastPhase(droppedFilename);
}
void dropFileSaveSmlsFirstCancel(){
	Window::closeWindow(Window::activeWindow.getActiveWindow());
}

void QuestionMessage3Button::Yes::operator = (const bool& par){
	g_dontContinueButtonAction = true;
	removeTypeFromList <DownButton>(g_callBackList);
	yesFunc();
}
void QuestionMessage3Button::No::operator = (const bool& par){
	g_dontContinueButtonAction = true;
	removeTypeFromList <DownButton>(g_callBackList);
	noFunc();
}
void QuestionMessage3Button::Cancel::operator = (const bool& par){
	g_dontContinueButtonAction = true;
	removeTypeFromList <DownButton>(g_callBackList);
	cancelFunc();
}

void openSmls(){//			OPEN_SMLS
	if(fileCPAllreadyOpen()){
		return;
	}
	String filename = g_currentFilename;
	if(g_settings->file.length){
		filename = g_settings->file[0];
	}
	setupFileTree("Open Seamless3d File", "open", openSmlsAction, checkAllFileExtensions, filename, g_settings->sortByDate);
}

void Video::Open::operator=(const bool& par){
	setupFileTree("Open png file sequence", "open", openVideoAction, checkPngFileExtension, String(), g_settings->sortByDate);
}

void WaveToPoly::Open::operator=(const bool& par){
	setupFileTree("Open wave file", "open", openAudioAction, checkWavFileExtension, String(), g_settings->sortByDate);
}

void importAllNodesField(){//IMPORT_NODES_FROM_FILE
	setupFileTree("Import All Nodes", "open", importAllNodesAction, checkAllFileExtensions, String(), g_settings->sortByDate);
}

void importBuildNodesField(){
	setupFileTree("Import Build Nodes", "open", importBuildNodesAction, checkAllFileExtensions, String(), g_settings->sortByDate);
}

void importPart(){
	String filename = g_seamlessDir;
	if(g_selNode->name != ""){
		filename += g_selNode->name;
		filename += ".wrl";
	}
	setupFileTree("Import Shape From File", "open", importPartAction, checkWrlFileExtension, filename, g_settings->sortByDate);
}
void importBxxAvStudio(){
	setupFileTree("Import Blaxxun/Canal Avatar Studio Avatar", "open", importBxxAvStudioAction, checkWrlFileExtension, g_seamlessDir, g_settings->sortByDate);
}

void importAllNodesReplace(int i_fieldIndex){
	fieldIndex = i_fieldIndex;
	setupFileTree("Import All Nodes", "open", importAllNodesReplaceAction, checkAllFileExtensions, String(), g_settings->sortByDate);
}
void addImportNodeAbove(int fieldIndex){
	setupFileTree("Import All Nodes", "open", addImportNodeAboveAction, checkAllFileExtensions, String(), g_settings->sortByDate);
}
void addImportNodeBelow(int fieldIndex){
	setupFileTree("Import All Nodes", "open", addImportNodeBelowAction, checkAllFileExtensions, String(), g_settings->sortByDate);
}

void importAnimationReplace(int fieldIndex){
	setupFileTree("Import Animation Nodes & Script", "open", importAnimationReplaceAction, checkAllFileExtensions, g_seamlessDir, g_settings->sortByDate);
}
void addImportAnimAbove(int fieldIndex){
	setupFileTree("Import Animation Nodes & Script", "open", addImportAnimAboveAction, checkAllFileExtensions, g_seamlessDir, g_settings->sortByDate);
}
void addImportAnimBelow(int fieldIndex){
	setupFileTree("Import Animation Nodes & Script", "open", addImportAnimBelowAction, checkAllFileExtensions, g_seamlessDir, g_settings->sortByDate);
}

void importBuildReplace(int fieldIndex){
	setupFileTree("Import Build Nodes", "open", importBuildReplaceAction, checkAllFileExtensions, g_seamlessDir, g_settings->sortByDate);
}
void addImportBuildAbove(int fieldIndex){
	setupFileTree("Import Build Nodes", "open", addImportBuildAboveAction, checkAllFileExtensions, g_seamlessDir, g_settings->sortByDate);
}
void addImportBuildBelow(int fieldIndex){
	setupFileTree("Import Build Nodes", "open", addImportBuildBelowAction, checkAllFileExtensions, g_seamlessDir, g_settings->sortByDate);
}

void importAnimation(){//IMPORT_ANIMATION
	setupFileTree("Import Animation Nodes & Script", "open", importAnimationAction, checkAllFileExtensions, g_seamlessDir, g_settings->sortByDate);
}

void importAnimRecalc(){
	setupFileTree("Import Animation Nodes & Script", "open", importAnimRecalcAction, checkAllFileExtensions, g_seamlessDir, g_settings->sortByDate);
}

void importPivotPoints(){
	setupFileTree("Import Pivot Points", "open", importPivotPointsAction, checkSmlsFileExtensions, g_seamlessDir, g_settings->sortByDate);
}

void importPivotPointDistances(){
	setupFileTree("Import Pivot Point Distances", "open", importPivotPointDistancesAction, checkSmlsFileExtensions, g_seamlessDir, g_settings->sortByDate);
}

void saveBeforeClosing(void (*i_yesFunc)(), void (*i_noFunc)(), void (*i_cancelFunc)()){
	String filename;
	String ext;
	splitPath(g_currentFilename, String(), String(), filename, ext);

	String msg = "save ";
	msg += filename;
	msg += ext;
	msg += " before closing ?";

	QuestionMessage3Button::setupMessageWindow(msg, i_yesFunc, i_noFunc, i_cancelFunc);
}


void saveSmlsFirst(void (* i_queFunction)()){
	if(smlsIsSaved()){
		i_queFunction();
	}
	else{
		queFunction = i_queFunction;
		saveBeforeClosing(saveSmlsFirstYes, saveSmlsFirstNo, saveSmlsFirstCancel);
	}
}

void openDroppedFileLastPhase(){
	openSmlsLastPhase(droppedFilename);
}

void newSmlsLastPhase(){
	setupFileTree("Create New Seamless3d File As", "save", newSmlsAction, checkAllFileExtensions, String(), g_settings->sortByDate);
}

void newSmls(){//			NEW_SMLS
	if(fileCPAllreadyOpen()){
		return;
	}
	if(g_browseMode){
		return;
	}
	saveSmlsFirst(newSmlsLastPhase);
}

void openDroppedFile(){
	if(FileCP* fileCP = fileCPAllreadyOpen()){
		Window::closeWindow(fileCP);
	}
	saveSmlsFirst(openDroppedFileLastPhase);
}

void robotDemoStart(void (* i_queFunction)()){
	if(InputEvent::robotActive)return;
	saveSmlsFirst(i_queFunction);
}

void dummy(){
}

void saveSmlsAs(void (* i_queFunction)()){
	queFunction = i_queFunction;
	if(fileCPAllreadyOpen()){
		return;
	}
	if(g_browseMode){
		return;
	}
	setupFileTree("Save Seamless3d File As", "save", saveSmlsAsAction, checkAllFileExtensions, g_currentFilename, g_settings->sortByDate);
}
void saveSmlsAs(){
	saveSmlsAs(dummy);
}
void exportPart(){//SAVE_PART_AS:
	if(g_browseMode){
		return;
	}
	if(dynamic_cast<Art*>(g_selNode.getRaw())){
		String filename = g_seamlessDir;
		if(g_selNode->name != ""){
			filename += g_selNode->name;
			filename += ".wrl";
		}
		setupFileTree("Save Part As", "save", exportPartAction, checkWrlFileExtension, filename, g_settings->sortByDate);
	}
	else{
		needPartMsg();
	}
}

void exportPatch(){
	if(g_browseMode){
		return;
	}
	if(g_importExportPatch = dynamic_cast<NurbsPatch*>(g_selNode.getRaw())){
		String filename = g_seamlessDir;
		if(g_selNode->name != ""){
			filename += g_selNode->name;
			filename += ".wrl";
		}
		setupFileTree("Save Patch As", "save", exportPatchAction, checkWrlFileExtension, filename, g_settings->sortByDate);
	}
	else{
		needPartMsg();
	}
}


void exportMorphedPatch(){
	if(g_browseMode){
		return;
	}
	if(Art* part = dynamic_cast<Art*>(g_selNode.getRaw())){
		int idx = part->index[0];
		if(idx < part->controlPoint.length){
			g_importExportPatch = part->controlPoint[idx].p;
			String filename = g_seamlessDir;
			if(g_selNode->name != ""){
				filename += g_selNode->name;
				filename += ".wrl";
			}
			setupFileTree("Save Patch As", "save", exportPatchAction, checkWrlFileExtension, filename, g_settings->sortByDate);
		}
	}
	else{
		needPartMsg();
	}
}

void importMorphedPatch(){
	if(g_browseMode){
		return;
	}
	if(Art* part = dynamic_cast<Art*>(g_selNode.getRaw())){
		int idx = part->index[0];
		if(idx < part->controlPoint.length){
			g_importExportPatch = part->controlPoint[idx].p;
			String filename = g_seamlessDir;
			if(g_selNode->name != ""){
				filename += g_selNode->name;
				filename += ".wrl";
			}
			setupFileTree("Open Patch", "open", importPatchAction, checkWrlFileExtension, filename, g_settings->sortByDate);
		}
	}
	else{
		needPartMsg();
	}
}

void exportColorToTexture(){
	if(g_browseMode){
		return;
	}
	setupFileTree("Export Color To Texture", "save", exportColorToTextureAction, checkPngFileExtension, g_seamlessDir, g_settings->sortByDate);
}

void exportAnimToBvh(){
	if(g_browseMode){
		return;
	}
	setupFileTree("Export Anim To BVH Motion Capture", "save", exportAnimToBvhAction, checkBvhFileExtension, g_seamlessDir, g_settings->sortByDate);
}

void QuestionMessage3Button::setupMessageWindow(const String& i_msg, void (*i_yesFunc)(), void (*i_noFunc)(), void (*i_cancelFunc)()){
	yesFunc = i_yesFunc;
	noFunc = i_noFunc;
	cancelFunc = i_cancelFunc;
	
	ControlPanel* controlPanel = new ControlPanel;
	controlPanel->hog = true;

	g_lookupProtoInterface[QUESTION_MESSAGE_3_BUTTON]->properties[0].name = i_msg;

	int width = i_msg.length * 8 + 50;
	if(width < 300){
		width = 330;
	}
	controlPanel->size = Vec2i(width, controlPanel->size.y);
	controlPanel->pos = Vec2i(g_3d->size.x / 2 - controlPanel->size.x / 2, g_3d->size.y / 2 - 50);
	
	controlPanel->setupNodeWindow(g_questionMessage3Button);
	Window::activeWindow = controlPanel;
	PlaySound("SystemExclamation", 0, SND_ASYNC);
}

void exportColorToTexture(char* filename);

void openDrive(){
	String& fullFilePath = fileCP->fullFilePath;
	if(Button* button = DownButton::getDownButton()){
		fileCP->setScrolPos(fileCP->scroll = 0);
		char chr = 'A' + button->command;
		fullFilePath = chr;
		fullFilePath += ":\\";
		fileCP->commonSetupFileTree(fullFilePath);
	}
}

void upOneLevel(){
	fileCP->setScrolPos(fileCP->scroll = 0);

	String& fullFilePath = fileCP->fullFilePath;
	bool b = false;
	for(int i = fullFilePath.length; --i >= 0;){
		if(fullFilePath[i] == '\\'){
			if(b){
				fullFilePath = fullFilePath.substring(0, i);
				fileCP->commonSetupFileTree(fullFilePath);
				break;
			}
			b = true;
		}
	}
}

void openSeamlessFolder(){
	fileCP->setScrolPos(fileCP->scroll = 0);
	String& fullFilePath = fileCP->fullFilePath;
	getFolderPath(fullFilePath="", CSIDL_MYDOCUMENTS);
	fullFilePath += "\\Seamless3d\\";
	fileCP->commonSetupFileTree(fullFilePath);
}
void openMyDocuments(){
	fileCP->setScrolPos(fileCP->scroll = 0);
	String& fullFilePath = fileCP->fullFilePath;
	getFolderPath(fullFilePath="", CSIDL_MYDOCUMENTS);
	fullFilePath += "\\";
	fileCP->commonSetupFileTree(fullFilePath);
}
void openDesktop(){
	fileCP->setScrolPos(fileCP->scroll = 0);
	String& fullFilePath = fileCP->fullFilePath;
	getFolderPath(fullFilePath="", CSIDL_DESKTOP);
	fullFilePath += "\\";
	fileCP->commonSetupFileTree(fullFilePath);
}
void FileCP::setupNodeWindow(Node* node){
	this->node = node;
	File* file = dynamic_cast<File*>(node);

	g_controlPanelOpen = true;
	g_propWinLine.length = 0;
	buttonList.length = 0;

	const M_Properties& properties = node->getProperties();

	FieldButton* button = file->sortByDate.setUpControl(*node, file->sortByDate, properties[0], 0, 0, 0, *this);
	int yPos = 8;
	button->dstPos = Vec2i(500, yPos);
	button->valPos = Vec2i(500 + 22, yPos - 8);

	buttonList += button;
	g_propWinLine += PropWinLine(yPos, button);

	button = file->filename.setUpControl(*node, file->filename, properties[1], 0, 74, 0, *this);
//	button->bSize=Vec2i(24, 24);
//	button->srcPos=Vec2i(24*14, 24);
//	button->iconSize=Vec2i(24, 24);
//	button->valSize.x=8*7;
//	button->valPos.x = 70;
	buttonList += button;
	g_propWinLine += PropWinLine(yPos, button);


	button = file->open.setUpControl(*node, file->open, properties[2], 0, 0, 0, *this);
//	button = file->open.setUpControl(*node, file->open, properties[1], 0, 80, title_bar_height + 1 + 25, *this);
//	button->srcPos=Vec2i(16,192);
//	button->bSize=Vec2i(24,24);
//	button->iconSize=Vec2i(16,16);
	button->dstPos.x = 4;
	buttonList += button;

	SingleButton* driveButton = new SingleButton(Vec2i(24, 24), Vec2i(24, 24), Vec2i(0, 72), "up one level", 0, upOneLevel, *this);
	driveButton->dstPos = Vec2i(370, 1);
	driveList += driveButton;

	int y = 30;
	driveButton = new SingleButton(Vec2i(24, 24), Vec2i(16, 16), Vec2i(432, 160), "open Seamless3d folder", 0, openSeamlessFolder, *this);
	driveButton->dstPos = Vec2i(10, y);
	driveList += driveButton;
	y += 22;
	driveButton = new SingleButton(Vec2i(24, 24), Vec2i(16, 16), Vec2i(432, 112), "open My Documents", 0, openMyDocuments, *this);
	driveButton->dstPos = Vec2i(10, y);
	driveList += driveButton;
	y += 22;
	driveButton = new SingleButton(Vec2i(24, 24), Vec2i(16, 16), Vec2i(416, 112), "open Desktop", 0, openDesktop, *this);
	driveButton->dstPos = Vec2i(10, y);
	driveList += driveButton;
	y += 22;
	DWORD drives = GetLogicalDrives();
	for(int k = 0; k < 32; k++, drives >>= 1){
		if(drives & 1){
			SingleButton* driveButton = new SingleButton(Vec2i(46, 24), Vec2i(24, 24), Vec2i(216, 48), "open drive", 0, openDrive, *this);
			driveButton->command = k;
			driveButton->dstPos = Vec2i(10, y);
			driveButton->offSetIcon = -9;
			driveList += driveButton;
			y += 22;
		}
	}

	g_propWinLine+=PropWinLine(yPos, button);

	size = Vec2i(560, 500);
	pos = Vec2i(g_3d->size.x / 2 - size.x / 2, 50);

	hover(g_mousePos);
	paint();
}

const Vec2i g_fileIconLookUp[]={
	Vec2i(240,96),	//FOLDER
	Vec2i(368,160), //SMLS
	Vec2i(368,112), //VRML
	Vec2i(368,144), //BVH
	Vec2i(400,112), //MID
	Vec2i(368,128), //PNG
	Vec2i(400,112), //WAV
};

void paintBlock2(const unsigned fromColor, const unsigned toColor, int tx, int ty, int bx, int by, int dstWidth, int dstHeight, int* dstPix);

/*
void deleteFile(){
	String recycleBin;

	String targetFile;
	String dir;
	splitPath(fileCP->fullFilePath, targetFile, dir, String(), String());
	recycleBin = targetFile;
	recycleBin += "RECYCLED\\";

	targetFile += dir;
	targetFile += fileCP->textCell.cellStr.c_str();

	MoveFile(targetFile.c_str(), recycleBin.c_str());
	fileCP->commonSetupFileTree(fileCP->fullFilePath);
}
*/

void windowsExplorerOpenFolder(){
	String drive;
	String dir;
	splitPath(fileCP->fullFilePath, drive, dir, String(), String());
	drive+=dir;
	openWithHtmlBrowser(drive, "");
}

void refresh(){
	fileCP->commonSetupFileTree(fileCP->fullFilePath);
}

bool FileCP::click(const Vec2i ptCursor){
	if(ptCursor >= pos && ptCursor < pos+size){
		activeWindow=this;
		if(tugBorder){
			tugBorder->window = this;
			g_drag = tugBorder;
			return true;
		}

		const Vec2i localCursor=ptCursor-pos;

		{
			int newCaretPos = 0;
			for(int k = buttonList.length; --k >= 0;){
				FieldButton* button=buttonList[k];
				if(button->checkClickTextCell(localCursor, newCaretPos)){
					if(button->editMode == false){
						textCell.cellActive = false;
						textCell.selectField(*button);
					}
					textCell.highlightPos = textCell.caretPos = newCaretPos + textCell.firstVisChar;
					textCell.updateCaret();

					textCell.checkForDoubleLeftClick();

					g_drag = &textCell;
					paint();
					return true;
				}
			}
		}//above is same as ControlPanel::click except for the OpenNetFile & TextureMapper conditions
		if(localCursor.y < title_bar_height){
			g_drag = this;
			dragPos = ptCursor;
			return true;
		}
		if(localCursor.y >= beginYOffset && localCursor.y < beginYOffset + numOfVisLines * pixSymbolCY){
			
			int i = (localCursor.y - beginYOffset) / pixSymbolCY + scroll;
			
			if(i >= 0 && i < fileTreeList.length){

				selLine = i;

				if(File* file = dynamic_cast<File*>(fileTreeList[selLine].node.p)){
					if(file->isFile == false){
						if(isDoubleLeftClick()){
							openFolder(fileTreeList, selLine);
						}
					}
					else{
						dynamic_cast<File*>(this->node.p)->filename = file->name;
						if(isDoubleLeftClick()){
							openFunc(file->name);
							return true;
						}
						textCell.cellActive = false;
						highLightTextCell();
					}
				}
				paint();
			}
		}
		return true;
	}
	else{
		return false;
	}
}

bool FileCP::rightClick(const Vec2i ptCursor){
	if(ptCursor >= pos && ptCursor < pos+size){

		//textCell.cellActive = false;
		const Vec2i localCursor = ptCursor-pos;

		Window::activeWindow=this;

		g_menuGroupIndex=-1;

		g_remMenu=new Menu;
		M_<MenuItemPtr>& list=g_remMenu->list;
		g_remMenu->list.length=0;

		int selFrom=0;
		int selLen=0;
		int newCaretPos=0;
		if(textCell.isInCell(localCursor, selFrom, selLen, newCaretPos) && selLen > 0){
			list+=new MenuItem_("cut  (Ctrl X)", cut, *g_remMenu);
			list+=new MenuItem_(Vec2i(16,16), Vec2i(remember_x, remember_y), menu_short_copy, remember, *g_remMenu);
			list+=new MenuItem_(Vec2i(16,16), Vec2i(16, 192), "paste (Ctrl V)", pasteFromClipboard, *g_remMenu);
		}
		else{
			int newCaretPos=0;
			for(int k=buttonList.length; --k>=0;){
				FieldButton* button=buttonList[k];
				if(button->checkClickTextCell(localCursor, newCaretPos)){
					if(button->editMode==false){
						textCell.cellActive=false;
						textCell.selectField(*button);
					}
					textCell.highlightPos=textCell.caretPos=newCaretPos+textCell.firstVisChar;
					textCell.updateCaret();

					paint();

					list+=new MenuItem_(Vec2i(16,16), Vec2i(16, 192), "paste (Ctrl V)", pasteFromClipboard, *g_remMenu);
				}
			}
		}
		//if(localCursor.y >= beginYOffset && localCursor.y < beginYOffset + numOfVisLines * pixSymbolCY){
		{
			
			int i = (localCursor.y - beginYOffset) / pixSymbolCY + scroll;
			
			if(i >= 0 && i < fileTreeList.length){

				if(File* file = dynamic_cast<File*>(fileTreeList[i].node.p)){
					if(list.length){
						list+=new Seperator(*g_remMenu);
					}

//					if(file->isFile){
//						selLine = i;
//						dynamic_cast<File*>(this->node.p)->filename = file->name;
//						paint();
//					}
					//highLightTextCell();
					//list += new MenuItem_("delete file", deleteFile, *g_remMenu);
					list += new MenuItem_("open windows explorer", windowsExplorerOpenFolder, *g_remMenu);
					list += new MenuItem_("refresh", refresh, *g_remMenu);
				}
			}
		}
		if(list.length){
			g_remMenu->autoSizeAndStuff();
			g_remMenu->pos=ptCursor;
			return true;
		}
	}
	return false;
}

void FileCP::paint(){
	if(node==0)return;

	int dstWidth=size.x;
	int dstHeight=size.y;
	int* dstPix=pixBuf.getArray();
	const M_Properties& properties=node->getProperties();


	blitHorizontalColorSweep(Vec2i(10, 0), Vec2i(470, 0), Vec2i(471, 0), Vec2i(dstWidth-9-10, title_bar_height));
	pasteSprite2( 0, 0, 461, 0, dstWidth, dstHeight, iconSrcBufCX, iconSrcBufCY, 10, 25, dstPix);
	pasteSprite2( dstWidth-9, 0, 471, 0, dstWidth, dstHeight, iconSrcBufCX, iconSrcBufCY, 9, 25, dstPix);
	
	paintBlock2(0xffffff, 8, title_bar_height, dstWidth-9, size.y-10, dstWidth, dstHeight, dstPix);
	paintSideEdge(&iconSrcBuf[(iconSrcBufCY-1-10)*480+444], 8, 0, title_bar_height, dstHeight-10, dstWidth, dstHeight, dstPix);
	paintSideEdge(&iconSrcBuf[(iconSrcBufCY-1-10)*480+452], 9, dstWidth-9, title_bar_height, dstHeight-10, dstWidth, dstHeight, dstPix);

	blitHorizontalColorSweep(Vec2i(8, dstHeight-10), Vec2i(451, 10), Vec2i(452, 10), Vec2i(dstWidth-8-9, 10));
	pasteSprite2( 0, dstHeight-10, 444, 10, dstWidth, dstHeight, iconSrcBufCX, iconSrcBufCY, 8, 10, dstPix);
	pasteSprite2( dstWidth-9, dstHeight-10, 452, 10, dstWidth, dstHeight, iconSrcBufCX, iconSrcBufCY, 9, 10, dstPix);



//	paintBlock2(iconSrcBuf[(iconSrcBufCY-1)*480+451], iconSrcBuf[(iconSrcBufCY-1)*480+452], textCell.cellPos.x, dstHeight-27, textCell.cellPos.x + textCell.cellSize.x, dstHeight-8, size.x, size.y, dstPix);
	paintBlock2(iconSrcBuf[(iconSrcBufCY-1)*480+451], iconSrcBuf[(iconSrcBufCY-1)*480+452], 8, dstHeight-27, size.x-9, dstHeight-8, size.x, size.y, dstPix);
	paintSideEdge(&iconSrcBuf[(iconSrcBufCY-1)*480+444], 8, 0, dstHeight-27, dstHeight-8, size.x, size.y, dstPix);
	paintSideEdge(&iconSrcBuf[(iconSrcBufCY-1)*480+452], 9, size.x-9, dstHeight-27, dstHeight-8, size.x, size.y, dstPix);

	String& str = Button::str;

	str = node->getTypeStr();

	if(node->name!=""){
		str += ' ';
		str += node->name;
	}
	drawText2(str, 0x0, Vec2i(10, 16), dstWidth, dstHeight, dstPix);

	buttonList[0]->paintFieldName(); //sortByDate


	for(int i = buttonList.length; --i>=0;){
		FieldButton* button = buttonList[i];
		if(button->sprop.doNotDisplayName() == false && button->sprop.push() == false && button->sprop.link() == false){
			button->paintFieldName();
		}
		{
			button->paint();
			Vec2i valPos = button->valPos;
			valPos.y+=16;
			if(button->sprop.doNotDisplayValue()==false){
				if(button->editMode){
					textCell.paintCellStr();
					//const int visCharLen = textCell.cellSize.x / charWidth;
					//int lastChar = textCell.cellStr.length;
					//if(lastChar - textCell.firstVisChar > visCharLen){
					//	lastChar = textCell.firstVisChar + visCharLen;
					//}
					//drawText2(textCell.cellStr, textCell.firstVisChar, lastChar, 0, valPos, size.x, size.y, dstPix);
					//textCell.highLightBlock(valPos);
				}
				else{
					button->toString(str="");
					const int visCharLen = button->valSize.x / charWidth;
					int lastChar = str.length;
					if(lastChar > visCharLen){
						lastChar = visCharLen;
					}
					drawText2(str, 0, lastChar, 0, valPos, size.x, size.y, dstPix);
				}
			}
		}
	}
	
	int len = fileTreeList.length - scroll;
	if(len > numOfVisLines){
		len = numOfVisLines;
	}
	if(selLine >= scroll && selLine < scroll + len){
		int y = beginYOffset+(selLine-scroll)*pixSymbolCY;

		paintBlock2(iconSrcBuf[(iconSrcBufCY-1)*480+451], iconSrcBuf[(iconSrcBufCY-1)*480+452], driveWidth + 8, y, size.x - 9, y + 16, size.x, size.y, dstPix);
		//paintSideEdge(&iconSrcBuf[(iconSrcBufCY-1)*480+444], 8, 0, y, y + 16, size.x, size.y, dstPix);
		paintSideEdge(&iconSrcBuf[(iconSrcBufCY-1)*480+452], 9, size.x-9, y, y + 16, size.x, size.y, dstPix);
	}

	String strTime;
	for(int line = 0; line < len; line++){
		File* file = dynamic_cast<File*>(fileTreeList[scroll + line].node.p);
		int indent = fileTreeList[scroll + line].indent;
		int x = driveWidth + tLnOffst + indent * pixSymbolCX;
		int y = beginYOffset+line * pixSymbolCY;
		if(line <= numOfPathFolders - scroll){
			pasteSprite2(x, y, g_fileIconLookUp[file->type].x, g_fileIconLookUp[file->type].y, fileCP->size.x, fileCP->size.y, iconSrcBufCX, iconSrcBufCY, pixSymbolCX, pixSymbolCY, fileCP->pixBuf.getArray());
			drawText2(file->name, 0, file->name.length, NODE_COLOR, Vec2i(x + pixSymbolCX + 3, y + 12), dstWidth, dstHeight, dstPix);
		}
		else{
			FILETIME localFileTime;
			SYSTEMTIME systemTime;
			FileTimeToLocalFileTime(reinterpret_cast<FILETIME*>(&file->time), &localFileTime);
			FileTimeToSystemTime(&localFileTime, &systemTime);

			//sprintf(strTime,"%i/%.2i/%.2i  %.2i:%.2i", systemTime.wYear, systemTime.wMonth, systemTime.wDay, systemTime.wHour, systemTime.wMinute);

			strTime = systemTime.wYear;
			strTime += '/';
			if(systemTime.wMonth < 10){
				strTime += '0';
			}
			strTime += systemTime.wMonth;
			strTime += '/';
			if(systemTime.wDay < 10){
				strTime += '0';
			}
			strTime += systemTime.wDay;
			strTime += ' ';
			if(systemTime.wHour < 10){
				strTime += '0';
			}
			strTime += systemTime.wHour;
			strTime += ':';
			if(systemTime.wMinute < 10){
				strTime += '0';
			}
			strTime += systemTime.wMinute;

			x = driveWidth + 10;

			drawText2(strTime, 0, strTime.length, NODE_COLOR, Vec2i(x, y + 12), dstWidth, dstHeight, dstPix);
			x += strTime.length * 8 + 8;
			pasteSprite2(x, y, g_fileIconLookUp[file->type].x, g_fileIconLookUp[file->type].y, fileCP->size.x, fileCP->size.y, iconSrcBufCX, iconSrcBufCY, pixSymbolCX, pixSymbolCY, fileCP->pixBuf.getArray());
			x += pixSymbolCX + 4;
			drawText2(file->name, 0, file->name.length, NODE_COLOR, Vec2i(x, y + 12), dstWidth, dstHeight, dstPix);
		}
	}

	exitWindow.dstPos = Vec2i(size.x-22, 4);
	exitWindow.paint();
	topScroll.paint();
	bottomScroll.paint();
	pageUp.paint();
	slider.paint();
	exitWindow.dstPos=Vec2i(size.x-22, 4);

	for(int k = 0; k < driveList.length; k++){
		SingleButton* button = driveList[k];
		button->paint();
		if(k >= 4){
			Vec2i dstPos = button->dstPos;
			char chr = 'A' + button->command;
			str = chr;
			str += ':';
			drawText2(str, 0, button->dstPos + Vec2i(30, 16), dstWidth, dstHeight, dstPix);
		}
	}

	if(g_settings->invertWindowColor){
		invert();
	}
	loadSurfaceFromMemory();
	if(textCell.cellActive==true){
		textCell.doNoThinkWink();
	}
}
void FileCP::genNodeListRecursion(const int parent, int indent, File* file){
	int currentLine = fileTreeList.length;
	fileTreeList += SceneTreeListItem(false, file, parent, indent, Color3f());

	indent++;
	M_Node& folder = file->folder.refValue();
	for(int k = 0; k < folder.length; k++){
		File* file = dynamic_cast<File*>(folder[k].p);
		genNodeListRecursion(currentLine, indent, file);
	}
}

void FileCP::genNodeList(Node* node){
	fileTreeList.length = 0;
	int indent = 0;
	genNodeListRecursion(-1, indent, dynamic_cast<File*>(node));
}
FileCP::VerticalScrollBarSlider::VerticalScrollBarSlider(Vec2i i_size, Vec2i i_srcPos, const char* i_tip, int i_command, FileCP& i_win, int refCountValue):
	Button(i_size, i_srcPos, i_tip, i_command, i_win, refCountValue),
	win(i_win)
{}
FileCP::VerticalScrollBarTopButton::VerticalScrollBarTopButton(Vec2i i_size, Vec2i i_srcPos, const char* i_tip, int i_command, FileCP& i_win, int refCountValue):
	Button(i_size, i_srcPos, i_tip, i_command, i_win, refCountValue),
	win(i_win)
{}
FileCP::VerticalScrollBarBottomButton::VerticalScrollBarBottomButton(Vec2i i_size, Vec2i i_srcPos, const char* i_tip, int i_command, FileCP& i_win, int refCountValue):
	Button(i_size, i_srcPos, i_tip, i_command, i_win, refCountValue),
	win(i_win)
{}
FileCP::PageUp::PageUp(Vec2i i_size, Vec2i i_srcPos, const char* i_tip, int i_command, FileCP& i_win, int refCountValue):
	Button(i_size, i_srcPos, i_tip, i_command, i_win, refCountValue),
	win(i_win)
{}
FileCP::PageDown::PageDown(Vec2i i_size, Vec2i i_srcPos, const char* i_tip, int i_command, FileCP& i_win, int refCountValue):
	Button(i_size, i_srcPos, i_tip, i_command, i_win, refCountValue),
	win(i_win)
{}



void FileCP::VerticalScrollBarSlider::paint(){
	pasteSprite2( dstPos.x, dstPos.y, srcPos.x, srcPos.y, win.size.x, win.size.y, iconSrcBufCX, iconSrcBufCY, bSize.x, bSize.y, win.pixBuf.getArray());
}
void FileCP::VerticalScrollBarTopButton::paint(){
	pasteSprite2( dstPos.x, dstPos.y, srcPos.x, srcPos.y, win.size.x, win.size.y, iconSrcBufCX, iconSrcBufCY, bSize.x, bSize.y, win.pixBuf.getArray());
}
void FileCP::VerticalScrollBarBottomButton::paint(){
	pasteSprite2( dstPos.x, dstPos.y, srcPos.x, srcPos.y, win.size.x, win.size.y, iconSrcBufCX, iconSrcBufCY, bSize.x, bSize.y, win.pixBuf.getArray());
}
void FileCP::PageUp::paint(){
	//win.blitBlockFromHorizontalLine(dstPos, srcPos, Vec2i(9, win.size.y-dstPos.y-22));
	win.blitVerticalColorSweep(dstPos, srcPos - Vec2i(0,1), srcPos, Vec2i(6, win.bottomScroll.dstPos.y - win.topScroll.dstPos.y - win.bottomScroll.bSize.y));
}
void FileCP::VerticalScrollBarSlider::action(){
	removeTypeFromList < Button>(g_callBackList);
	removeTypeFromList < DownButton>(g_callBackList);
	dragPos=g_mousePos;
	g_drag=this;
}
void FileCP::VerticalScrollBarTopButton::action(){
	win.setScrolPos(-1);
}
void FileCP::VerticalScrollBarBottomButton::action(){
	win.setScrolPos(1);
}
void FileCP::PageUp::action(){
	win.setScrolPos(-win.numOfVisLines);
}
void FileCP::PageDown::action(){
	win.setScrolPos(win.numOfVisLines);
}
void FileCP::VerticalScrollBarSlider::setPixelPos(){
	const int top = win.topScroll.dstPos.y + win.topScroll.bSize.y;
	const int bottom = win.bottomScroll.dstPos.y - win.slider.bSize.y;
	const int relBottom = bottom - top;

	int relY = static_cast<float>(relBottom) / win.fileTreeList.length * win.scroll;
	dstPos = Vec2i(win.size.x - 20, top + relY);

}
void FileCP::VerticalScrollBarSlider::doDrag(const Vec2i ptCursor){
	int diff=ptCursor.y-dragPos.y;

	const int top = win.topScroll.dstPos.y + win.topScroll.bSize.y;
	const int bottom = win.bottomScroll.dstPos.y - win.slider.bSize.y;
	const int relBottom = bottom - top;
	const int relDstPos = dstPos.y - top;


	if( diff < 0 && relDstPos + diff < 0){
		diff = -relDstPos;
	}
	if( diff > 0 && relDstPos + diff >= relBottom){
		diff = relBottom - relDstPos;
	}

	dstPos.y+=diff;
	dragPos.y+=diff;

	win.scroll = static_cast<float>(relDstPos + diff) / relBottom * win.fileTreeList.length;
	win.resizePageUpPageDown();
	win.keepLeagal();//why is this needed?
	win.paint();
}
void FileCP::VerticalScrollBarSlider::endDrag(const Vec2i ptCursor){
	g_drag=0;
	const Vec2i localCursor=ptCursor-win.pos;
	if( localCursor >= dstPos && localCursor < dstPos + bSize){
		open=true;//prevent tool tip opening
		g_callBackList+=this;
	}
}
void FileCP::keepLeagal(){
	if(scroll >= fileTreeList.length){
		scroll = fileTreeList.length - 1;
	}
	if(scroll < 0){//why is this ever needed if there is always at least one node the scene node?
		scroll = 0;
	}
	if(selLine >= fileTreeList.length){
		selLine = fileTreeList.length - 1;
	}
}
void FileCP::setScrolPos(int offset){
	scroll+=offset;
	keepLeagal();
	slider.setPixelPos();
	resizePageUpPageDown();
	textCell.updateCaret();
	paint();
}
void FileCP::resizePageUpPageDown(){
	pageUp.dstPos = topScroll.dstPos + Vec2i(5, slider.bSize.y);
	pageUp.bSize = Vec2i(6, slider.dstPos.y - topScroll.dstPos.y - topScroll.bSize.y);
	pageDown.dstPos = slider.dstPos + Vec2i(5, slider.bSize.y);
	pageDown.bSize = Vec2i(6, bottomScroll.dstPos.y - pageDown.dstPos.y);
}
void FileCP::setWinSize(Vec2i& raw){
	pixBuf.length = raw.x * raw.y;

	FieldButton* filenameButton = buttonList[1]; //filename
	filenameButton->valPos = Vec2i(74, size.y - endYOffset + 3);
	textCell.cellPos = filenameButton->valPos;
	textCell.cellPos.y += 5;
	

	FieldButton* openButton = buttonList[2];//open
	openButton->dstPos = Vec2i(size.x - 50, size.y - endYOffset + 3);
	textCell.cellSize = Vec2i(openButton->dstPos.x - textCell.cellPos.x, 16);
	filenameButton->valSize.x = textCell.cellSize.x;

	topScroll.dstPos = Vec2i(size.x - 20, beginYOffset);
	bottomScroll.dstPos = Vec2i(size.x - 20, size.y - endYOffset - 10);
	slider.setPixelPos();
	resizePageUpPageDown();
	numOfVisLines = (size.y - beginYOffset - endYOffset) / pixSymbolCY;

	if(E_FAIL==initWindow()){
		throw;
	}
}
void FileCP::buttonHover(Vec2i ptCursor){
	for(int k=buttonList.length; --k>=0;){
		Button* button=buttonList[k];
		if(button && button->initHover(ptCursor)){
			return;
		}
	}
	for(int k = driveList.length; --k>=0;){
		Button* button = driveList[k];
		if(button && button->initHover(ptCursor)){
			return;
		}
	}

	if(topScroll.initHover(ptCursor)){
		return;
	}
	if(bottomScroll.initHover(ptCursor)){
		return;
	}
	if(pageUp.initHover(ptCursor)){
		return;
	}
	if(pageDown.initHover(ptCursor)){
		return;
	}
	if(slider.initHover(ptCursor)){
		return;
	}
	if(exitWindow.initHover(ptCursor)){
		return;
	}
	else{
		int a=0;
	}
}
bool FileCP::hover(Vec2i ptCursor){
	return ResizeableWindow::hover(ptCursor);
}
Ptr<RecentFiles> g_recentFiles = new RecentFiles;
void setupRecentFiles(){
	int len = g_settings->file.refValue().length;
	if(len == 0){
		return;
	}
	if(len > 6){
		len = 6;
	}
	M_Properties& properties = g_lookupProtoInterface[RECENT_FILES]->properties;

	for(int i = 0; i < len; i++){
		String filename;
		String ext;
		splitPath(g_settings->file[i], String(), String(), filename, ext);
		filename += ext;
		Properties& sprop = properties[i];
		sprop.name = filename;
		sprop.flags &= ~DISABLE;
	}

	ControlPanel* recentFilesCP = new ControlPanel;
	recentFilesCP->setupNodeWindow(g_recentFiles);
	Window::activeWindow = recentFilesCP;
}
void RecentFiles::Select::operator=(bool const& par){
	g_dontContinueButtonAction = true;
	removeTypeFromList <DownButton>(g_callBackList);

	fieldIndex = home.getFieldIndex(this);

	if(smlsIsSaved()){
		openSmlsLastPhase(g_settings->file[fieldIndex]);
	}
	else{
		saveBeforeClosing(openRecentSaveSmlsFirstYes, openRecentSaveSmlsFirstNo, openRecentSaveSmlsFirstCancel);
	}

}

void openRecentMenu(int recentFileIndex){

	fieldIndex = recentFileIndex;

	if(smlsIsSaved()){
		openSmlsLastPhase(g_settings->file[fieldIndex]);
	}
	else{
		saveBeforeClosing(openRecentSaveSmlsFirstYes, openRecentSaveSmlsFirstNo, openRecentSaveSmlsFirstCancel);
	}
}
int countBackups(const String& recoverFilter){
	String recoverBackUpName = g_undoDir;
	recoverBackUpName += recoverFilter.substr(0, recoverFilter.length - 6);
	String filename;

	int i = 0;
	while(true){
		filename = recoverBackUpName;
		filename += i;
		filename += ".smls";
		if(FILE* stream = fopen(filename.c_str(), "r+b")){
			fclose(stream);
		}
		else break;
		i++;
	};
	return i;
}
bool checkForLockFiles(){
	WIN32_FIND_DATA w32fd;
	HANDLE SearchHandle;


	String bPath = g_undoDir;
	bPath += "*.lock";

	if((SearchHandle = FindFirstFile(bPath.c_str(), &w32fd)) != INVALID_HANDLE_VALUE){
		do{
			if((w32fd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) == FILE_ATTRIBUTE_DIRECTORY){
				continue;
			}
			else{
				recoverFilter = w32fd.cFileName;
				recoverFilter = recoverFilter.substr(0, recoverFilter.length - 5);
				if(recoverFilter != BackUp::ourBackUpName){
					recoverFilter += "*.smls";

					String lockFilename = g_undoDir;
					lockFilename += w32fd.cFileName;
					if(FILE* stream = fopen(lockFilename.c_str(), "rb")){
						fclose(stream);
						int numOf = countBackups(recoverFilter);
						
						recoverFilename = readFile(lockFilename);
						String msg = numOf;
						msg += " backups detected for ";
						msg += getFilename(recoverFilename);
						msg += ", recover ?";
						dLog(msg);

						QuestionMessage3Button::setupMessageWindow(msg, recoverYes, recoverNo, recoverCancel);
						return true;
					}
				}
			}
		}while(FindNextFile(SearchHandle, &w32fd));
	}
	return false;
}
void deleteRecoverFiles(){
	String recoverBackUpName = recoverFilter.substr(0, recoverFilter.length - 6);

	String filename = g_undoDir;
	filename += recoverBackUpName;
	filename += ".lock";
	DeleteFile(filename.c_str());

	BOOL noError = FALSE;
	int i = 0;
	do{
		filename = g_undoDir;
		filename += recoverBackUpName;
		filename += i++;
		filename += ".smls";
		noError = DeleteFile(filename.c_str());
	}while (noError);
	recoverFilename = "";
}

void recoverYes(){
	Window::closeWindow(Window::activeWindow.getActiveWindow());
	selectLastFile = true;
	g_processComnandLine = false;
	setupFileTree("Recover File from Cache", "open", recoverSmlsAction, checkAllFileExtensions, g_undoDir, true, recoverFilter.c_str());
}
void recoverNo(){
	deleteRecoverFiles();
	recoverFilename = "";
	Window::closeWindow(Window::activeWindow.getActiveWindow());

	if(checkForLockFiles() == false){
		if(g_processComnandLine && openFileFromCommandLine() == false){
			setupRecentFiles();
		}
	}
}
void recoverCancel(){
	recoverFilename = "";
	Window::closeWindow(Window::activeWindow.getActiveWindow());

	if(g_processComnandLine && openFileFromCommandLine() == false){
		setupRecentFiles();
	}
}
void closeRecentFileWindow(){
	bool alreadyClosed = false;
	if(alreadyClosed){
		return;
	}
	for(int k = Window::visWinList.length; --k >= 0;){
		if(ControlPanel* controlPanel = dynamic_cast<ControlPanel*>(Window::visWinList[k].p)){
			if(dynamic_cast<RecentFiles*>(controlPanel->node.p)){
				Window::closeWindow(controlPanel);
				return;
			}
		}
	}
}

void ExitButton::action(){
	ControlPanel* controlPanel = dynamic_cast<ControlPanel*>(&win);
	if(File* file = dynamic_cast<File*>(controlPanel->node.p)){
		exitFileWindow();
	}
	else if(Palette* palette = dynamic_cast<Palette*>(controlPanel->node.p)){
		exitPaletteWindow();
	}
	else if(Anim* anim = dynamic_cast<Anim*>(controlPanel->node.p)){
		if(anim->play == false && 0 == (0x8000&GetKeyState(VK_LCONTROL))){
			resetRotations();
		}
		Window::closeWindow(&win);
		g_tb->updateToolBarContext();
	}
	else{
		Window::closeWindow(&win);
	}
}

bool match(const String& name, String& cellStr, const int caretPos){
	if(name.length < caretPos || cellStr.length < caretPos){
		return false;
	}
	for(int i = 0; i < caretPos; i++){
		if(upperCaseChar(name[i]) != upperCaseChar(cellStr[i])){
			return false;
		}
	}
	cellStr = cellStr.substr(0, caretPos);
	cellStr += name.substr(caretPos);
	return true;
}

void FileCP::findAutoCompleteLine(String& cellStr, const int caretPos){
	for(int k = 0; k < fileTreeList.length; k++){
		if(match(fileTreeList[k].node->name, cellStr, caretPos)){
			selLine = k;
			scroll = selLine - (numOfVisLines / 2);
			if(scroll < 0){
				scroll = 0;
			}
			setScrolPos(0);
			textCell.highlightPos = cellStr.length;
			break;
		}
	}
}

bool ControlPanel::CPTextCell::charInput(unsigned wParam){
	if(FileCP* fileCP = dynamic_cast<FileCP*>(&win)){
		if(wParam==VK_RETURN){
			cellActive = false;
			fileCP->root->open = true;
			return false;
		}
	}
	if(wParam==VK_BACK){
		if(deleteHighlight()){
			onChange();
			win.paint();
		}
		else if(caretPos>0){
			decCaretPosNH();
			cellStr.deleteChar(caretPos);
			win.paint();
		}
	}
	else if(wParam<' ')return true;
	else{
		static double time = 0;
		deleteHighlight();
		if(cellStr.length == 0){
			time = g_time;
		}
		cellStr.insert(wParam, caretPos);
		incCaretPosNH();
		if(FileCP* fileCP = dynamic_cast<FileCP*>(&win)){
			if(g_time - time < 2){
				time = g_time;
				fileCP->findAutoCompleteLine(cellStr, caretPos);
			}
		}
		win.paint();
	}
	onChange();
	return true;
}

void FileCP::highLightTextCell(){
	FieldButton* button = buttonList[1];
	if(button->editMode == false){
		textCell.cellActive = false;
		textCell.selectField(*button);
	}
	textCell.caretPos = textCell.cellStr.length;
	textCell.highlightPos = 0;
	textCell.updateCaret();
}

void FileCP::onMouseWheel(const int par){
	setScrolPos(par);
}

void FileCP::fixKeyScroll(){
	if(selLine < 0){
		selLine = 0;
	}
	if(selLine >= fileTreeList.length){
		selLine = fileTreeList.length - 1;
	}
	if(selLine < scroll){
		scroll = selLine;
	}
	if(selLine >= scroll + numOfVisLines){
		scroll = selLine - (numOfVisLines - 1);
	}
	if(scroll < 0){
		scroll = 0;
	}
	if(scroll >= fileTreeList.length){
		scroll = fileTreeList.length-1;
	}
	textCell.cellStr = fileTreeList[selLine].node->name;
	textCell.highlightPos = textCell.cellStr.length;
	textCell.caretPos = 0;
	
	setScrolPos(0);
}

void scrollUpCP(){
	if(FileCP* fileCP = dynamic_cast<FileCP*>(Window::activeWindow.getActiveWindow())){
		fileCP->selLine--;
		fileCP->fixKeyScroll();
	}
}

void scrollDownCP(){
	if(FileCP* fileCP = dynamic_cast<FileCP*>(Window::activeWindow.getActiveWindow())){
		if(fileCP->fileTreeList.length){
			fileCP->selLine++;
			fileCP->fixKeyScroll();
		}
	}
}

void scrollPageUpCP(){
	if(FileCP* fileCP = dynamic_cast<FileCP*>(Window::activeWindow.getActiveWindow())){
		fileCP->scroll -= fileCP->numOfVisLines;
		fileCP->selLine = fileCP->scroll;
		fileCP->fixKeyScroll();
	}
}

void scrollPageDownCP(){
	if(FileCP* fileCP = dynamic_cast<FileCP*>(Window::activeWindow.getActiveWindow())){
		if(fileCP->fileTreeList.length){
			fileCP->scroll += fileCP->numOfVisLines;
			fileCP->selLine = fileCP->scroll + fileCP->numOfVisLines - 1;
			fileCP->fixKeyScroll();
		}
	}
}

void FileCP::onGainFocus(){
	textCell.onWindowGainFocus();
	FieldButton* button = buttonList[1];
	if(button->editMode == false){
		textCell.cellActive = false;
		textCell.selectField(*button);
	}
}

