/* Seamless3d

Copyright (C) 2001-2010 Graham John Perrett (AKA thyme) thyme@seamless3d.com www.seamless3d.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this
software and associated documentation files (the "Software"), to deal in the Software
without restriction, including without limitation the rights to use, copy, modify,
merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to the following
conditions:

The above copyright notice and this permission notice shall be included in all copies or
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT
OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.
*/

#include <windows.h>
#include <shlobj.h>
#include <direct.h>

#include "seamless3d.h"
#include "techuelife_triangle_editor.h"
#include "gui.h"
#include "scene_tree.h"
#include "robot_aid.h"

bool backFace(const Vec3f& v0, const Vec3f& v1, const Vec3f& v2);
Vec2f convertPointToScreen2d(Vec3f point);
bool triIntersectsRect(Vec2f a, Vec2f b, Vec2f c);
bool isRectCornerInTriangle(const Vec2f* tri);

CommonSpec* getCommonSpec(const Art* part){
	const String& name = part->name;
	const M_Node& m_commonSpec = g_settings->commonSpec;

	for(int k = m_commonSpec.length; --k >= 0;){
		if(CommonSpec* commonSpec = dynamic_cast<CommonSpec*>(m_commonSpec[k].p)){
			if(commonSpec->name == name){
				return commonSpec;
			}
			else{
				const M_String& alternative = commonSpec->alternative;
				for(int i = alternative.length; --i >= 0;){
					if(alternative[i] == name){
						return commonSpec;
					}
				}
			}
		}
	}
	return 0;
}

extern Ptr<PartClue> g_vcArtTriBase;


Ptr<Palette> g_palette;
const float Palette::initCustomColorData[3 * PaletteCP::tabColums * PaletteCP::tabRows] = {
1, 0, 0, 0, .5, .25, 0, 0, 1, 1, 1, 0, 0, 1, 1, 1, .5, 0, 1, 0, 1, 0, 0, 0, .5, .8, .8, .85, .45, .39, .68, .08, .15, .81, .05, .05, .878, .416, .463, .9, .6, .42, .98, .74, .63, 1, .14, 0, .5, 1, 0, 0, 0, .5, .7, .7, 0, .5, 1, 1, .784, .251, .251, .9, 0, .6, .25, .25, .25, .25, .25, .5, .68, .38, .298, .88, .36, .34, .82, .322, .302, .89, .43, .31, .82, .45, .35, .9, .67, .55, .86, .08, .24, 0, .5, .5, 0, .5, 1, 1, 1, .5, 0, .8, .8, .5, 0, 1, 1, .5, 1, .5, .5, .5, .41, .41, 1, .51, .31, .21, .753, .184, .133, .878, .416, .463, 1, .604, .714, .965, .745, .647, .953, .753, .639, 1, .9, .9, .8, 1, .8, .9, .9, 1, 1, 1, .8, .8, 1, 1, 1, .9, .8, 1, .9, 1, .75, .75, .75, .008, .608, .984, .48, .22, .13, .5, 0, .25, .68, .08, .15, .941, .553, .471, .996, .886, .843, .98, .64, .54, .5, 0, .25, .25, .5, .5, .5, .5, 1, 1, 1, 0, .008, .608, .984, 1, 1, .8, .91, .702, .58, .25, 0, .5, .41, .41, 1, .29, .17, .06, .51, .31, .21, .98, .74, .63, .878, .416, .463, .996, .886, .843, .98, .64, .54, .192, .078, .031, 0, 1, 0, 0, 0, 1, .86, .08, .24, 0, 1, 1, 1, .5, 0, 1, 1, 0, 0, 0, 0, 1, .5, 1, .68, .38, .298, .85, .45, .39, .5, .25, 0, .81, .05, .05, .878, .416, .463, .98, .74, .63
};

Art* getPart(Node* node){
	if(Art* part = dynamic_cast<Art*>(node)){
		return part;
	}
	else if(TransferOwnership* transferOwnership = dynamic_cast<TransferOwnership*>(node)){
		return transferOwnership->getSelectedPart();
	}
	else if(TransferBCyl* transferBCyl = dynamic_cast<TransferBCyl*>(node)){
		return transferBCyl->dest;
	}
	return 0;
}

Art* getPart(){
	return getPart(g_selNode.getRaw());
}

PaletteCP* paletteCPAllreadyOpen(){
	for(int i=Window::visWinList.length; --i>=0;){
		if(PaletteCP* paletteCP = dynamic_cast<PaletteCP*>(Window::visWinList[i].p)){
			return paletteCP;
		}
	}
	return 0;
}

void exitPaletteWindow(){
	if(g_colorOwnership){
		for(int k = g_sceneTreeList.length; --k>=0;){
			if(Seamless* smls=dynamic_cast<Seamless*>(g_sceneTreeList[k].node.p)){
				smls->ctrf = true;
			}
		}
	}
	g_paletteMode = false;
	g_colorOwnership = false;
	Window::closeWindow(paletteCPAllreadyOpen());
	houseKeepVisualClue();
}

void openOwnership(){
	if(paletteCPAllreadyOpen()){
		if(g_colorOwnership){
			exitPaletteWindow();
			return;
		}
		else{
			exitPaletteWindow();
		}
	}
	if(getPart() == 0){
		MessageWin::msg("you need a Part (or a TransferOwnership or TransferBCyl referencing a Part) selected to open color ownership mode");
		return;
	}

	g_paletteMode = false;
	g_colorOwnership = true;
	for(int k = g_sceneTreeList.length; --k>=0;){
		if(Seamless* smls = dynamic_cast<Seamless*>(g_sceneTreeList[k].node.p)){
			smls->ctrf = true;
		}
	}
	if(Art* part = getPart()){
		if(part->smls->workAnimCP){
			g_lookupProtoInterface[PALETTE]->properties[2].name = "points";
		}
		else{
			g_lookupProtoInterface[PALETTE]->properties[2].name = "coords";
		}
	}
	PaletteCP* paletteCP = new PaletteCP;
	paletteCP->setupNodeWindow(g_palette);
	Window::activeWindow = paletteCP;
	houseKeepVisualClue();
}


void openPalette(){
	if(paletteCPAllreadyOpen()){
		if(g_colorOwnership){
			exitPaletteWindow();
		}
		else{
			return;
		}
	}

	g_colorOwnership = false;
	g_paletteMode = true;

	PaletteCP* paletteCP = new PaletteCP;
	paletteCP->setupColorSelectorNodeWindow(g_palette);
	Window::activeWindow = paletteCP;
}

//////////////////////

class OwnerGroup{
public:
	Art* part;
	int offset;
	int fromCoord;
	int len;
	float weight;
	OwnerGroup():
		part(0),
		offset(0),
		fromCoord(0),
		len(0),
		weight(1)
	{}
};

class ContigCoord{
public:
	Vec3f coord;
	int gi;
	int ci;
	ContigCoord():
		gi(0),
		ci(0)
	{}
};

int Art::numOfGroups()const{
	const M_Int& coordGroup = this->coordGroup;
	int h = 0;
	int numOfGroups = 0;
	if(coordGroup.length == 0){
		return 1;
	}
	for(int i = 0; i < coordGroup.length; i++){
		h += coordGroup[i];
		numOfGroups++;
	}
	numOfGroups++;
	return numOfGroups;
}

int countGroups(Art* selPart, M_Art& partList, int& selGroup){
	int numOfgroups = 0;
	for(int k = 0; k < partList.length; k++){
		Art* part = partList[k];
		if(part == selPart){
			selGroup = numOfgroups;
		}
		numOfgroups += part->numOfGroups();
	}
	return numOfgroups;
}

int countGroups(M_Art& partList, M_Int& partGroupIndex){
	partGroupIndex.length = partList.length;
	int numOfgroups = 0;
	for(int k = 0; k < partList.length; k++){
		Art* part = partList[k];
		partGroupIndex[k] = numOfgroups;
		numOfgroups += part->numOfGroups();
	}
	return numOfgroups;
}

void concatCoord(M_Art& partList, M_<OwnerGroup>& ownerGroup, M_<ContigCoord>& contigCoord, M_Triangle& contigTriangle){
	int len = 0;
	int triLen = 0;
	for(int k = 0; k < partList.length; k++){
		Art* part = partList[k];
		len += part->coord.length;
		triLen += part->triangle.length;
	}
	contigCoord.length = len;
	contigTriangle.length = triLen;

	int gi = 0;
	int contigIdx = 0;
	int contigTriIdx = 0;

	for(int k = 0; k < partList.length; k++){
		Art* part = partList[k];
		//part->fromCoord = contigIdx;
		M_Int& coordGroup = part->coordGroup;
		M_Float& weight = part->weight;
		M_Vec3f& coord = part->coord;
		int coordLen = part->coord.length;
		int coordGroupLen = coordGroup.length;


		int cg =0, h = 0;
		if(coordGroupLen == 0){
			ownerGroup[gi].part = part;
			ownerGroup[gi].offset = cg;
			while(h < coordLen){
				contigCoord[contigIdx].coord = coord[h++];
				contigCoord[contigIdx].gi = gi;
				contigIdx++;
			}
			gi++;
		}
		else{
			do{
				ownerGroup[gi].part = part;
				ownerGroup[gi].offset = cg;
				ownerGroup[gi].weight = weight[cg];
				int portion = coordGroup[cg];
				if(h + portion > coordLen){
					portion = coordLen - h;
				}
				for(int j = portion; --j >= 0;){
					contigCoord[contigIdx].coord = coord[h++];
					contigCoord[contigIdx].gi = gi;
					contigIdx++;
				}
				gi++;
			}while(++cg < coordGroupLen);
			if(h <= coordLen){
				ownerGroup[gi].part = part;
				ownerGroup[gi].offset = cg;
				while(h < coordLen){
					contigCoord[contigIdx].coord = coord[h++];
					contigCoord[contigIdx].gi = gi;
					contigIdx++;
				}
				gi++;
			}
		}



		part->fromPoly = contigTriIdx;
		M_Triangle& triangle = part->triangle;
		int tLen = part->triangle.length;
		for(int t = 0; t < tLen;){
			contigTriangle[contigTriIdx++] = triangle[t++];
		}
		triangle.length = 0;
	}
}

int getDestGroup(){
	return 0;
}

void applySelection(M_<S_VertexIndex>& triList, int destGroup, M_<ContigCoord>& contigCoord){
	for(int k = 0; k < triList.length; k++){
		int crdIdx = triList[k].p->fromCoord + triList[k].i;
		contigCoord[crdIdx].gi = destGroup;
	}
}

void setCoordIndex(M_<OwnerGroup>& ownerGroup, M_<ContigCoord>& contigCoord){
	int contigCoordLen = contigCoord.length;

	for(int i = 0; i < contigCoordLen; i++){
		ContigCoord& cc = contigCoord[i];
		OwnerGroup& og = ownerGroup[cc.gi];

		cc.ci = og.len++;
	}
}

void setPartCoordsLengthsAndSetGroupOffsets(M_<OwnerGroup>& ownerGroup){
	int last = ownerGroup.length - 1;
	int contigIdx = 0;
	int fromCoord = 0;
	for(int i = 0, h = 0; i < ownerGroup.length; i++){
		Art* part = ownerGroup[i].part;
		int len = ownerGroup[i].len;
		if(ownerGroup[i].offset == 0){
			fromCoord = contigIdx;
			h = 0;
			part->coordGroup.refValue().length = 0;
			part->weight.refValue().length = 0;
		}
		ownerGroup[i].fromCoord = fromCoord;
		
		bool lastSubGroup = (i == last || ownerGroup[i + 1].offset == 0);
		
		if(false == (lastSubGroup && ownerGroup[i].weight == 1)){
			part->coordGroup.refValue() += len;
			part->weight.refValue() += ownerGroup[i].weight;
		}

		if(lastSubGroup){
			part->coord.refValue().length = h + len;
		}

		ownerGroup[i].offset = contigIdx;
		h += len;
		contigIdx += len;
	}
}

void addOffsetsToCoordIndex(M_<OwnerGroup>& ownerGroup, M_<ContigCoord>& contigCoord){
	int contigCoordLen = contigCoord.length;

	for(int i = 0; i < contigCoordLen; i++){
		int gi = contigCoord[i].gi;
		Art* part = ownerGroup[gi].part;
		int offset = ownerGroup[gi].offset;
		int fromCoord = ownerGroup[gi].fromCoord;
		int ci = contigCoord[i].ci + offset;
		contigCoord[i].ci = ci;
		part->coord.refValue()[ci - fromCoord] = contigCoord[i].coord;
	}
}

void redirectTriangles(M_Triangle& triangle, M_<OwnerGroup>& ownerGroup, M_<ContigCoord>& contigCoord){
	for(int i = triangle.getTriangleLength(); --i >= 0;){
		Art* part = triangle.refPart(i);
		int idx = part->fromCoord + triangle.refIndex(i);
		int gi = contigCoord[idx].gi;
		int ci = contigCoord[idx].ci - ownerGroup[gi].fromCoord;
		triangle.refPart(i) = ownerGroup[gi].part;
		triangle.refIndex(i) = ci;
	}
}

void copyTrianglesToParts(M_Triangle& triangle){
	for(int i = 0; i < triangle.length; i++){
		S_Triangle& tri = triangle[i];
		tri.getOwner()->triangle += tri;;
	}
}

void setToRefContigCoords(M_Triangle& triangle){
	for(int i = triangle.getTriangleLength(); --i >= 0;){
		triangle.refIndex(i) += triangle.refPart(i)->fromCoord;
	}
}

void setToRefPartCoords(M_Triangle& triangle){
	for(int i = triangle.getTriangleLength(); --i >= 0;){
		triangle.refIndex(i) -= triangle.refPart(i)->fromCoord;
	}
}

void markSelectedAndNonSelectedTriangles(M_<S_VertexIndex>& triList, M_Triangle& triangle, M_Int& triFlag){
	triFlag.length = triangle.length;
	for(int i = triFlag.length; --i >= 0;){ //non selected triangles wiill have second bit set
		triFlag[i] = 2;
	}
	for(int i = triList.length; --i >= 0;){//selected triangles will have first bit set
		triFlag[triList[i].p->fromPoly + triList[i].i] = 1;
	}
}

void markSelectedAndNonSelectedCoords(M_Triangle& triangle, M_Int& triFlag, M_<ContigCoord>& contigCoord, M_Int& coordFlag){
	coordFlag.length = contigCoord.length;

	for(int t = triangle.length; --t >= 0;){

		S_Triangle& tri = triangle[t];

		for(int i = 3; --i >= 0; ){
			int idx = tri.refPart(i)->fromCoord + tri.refIndex(i);
			coordFlag[idx] |= triFlag[t];
		}
	}
}

void addNewCoords(M_<ContigCoord>& contigCoord, M_Int& coordFlag){
	for(int i = contigCoord.length; --i >= 0;){
		if(coordFlag[i] == 3){
			int gi = contigCoord[i].gi;
			int last = contigCoord.length;;
			contigCoord.length++;
			coordFlag[i] = last;
			contigCoord[last].coord = contigCoord[i].coord;
			contigCoord[last].gi = gi;
		}
		else{
			coordFlag[i] = -1;
		}
	}
}

void redirectToNewCoords(M_<OwnerGroup>& ownerGroup, M_Triangle& triangle, M_Int& triFlag, M_<ContigCoord>& contigCoord, M_Int& coordFlag){
	for(int t = triangle.length; --t >=0;){

		S_Triangle& tri = triangle[t];

		for(int i = 3; --i >= 0; ){
			int idx = tri.refPart(i)->fromCoord + tri.refIndex(i);
			if(triFlag[t] == 1 && coordFlag[idx] > 0){
				Art* owner = tri.getOwner();
				if(ColorEffect* colorEffect = dynamic_cast<ColorEffect*>(owner->workEffect)){
					M_Color3f& m_color = colorEffect->color;
					int last = m_color.length;
					m_color += m_color[tri.refTexIndex(i)];
					int cf = coordFlag[idx];
					int ci = contigCoord[cf].ci;
					int gi = contigCoord[cf].gi;
					tri.refIndex(i) = ci - ownerGroup[gi].fromCoord;
					tri.refTexIndex(i) = last;
				}
				else if(TextureEffect* textureEffect = dynamic_cast<TextureEffect*>(owner->workEffect)){
					M_Vec2f& texCoord = textureEffect->texCoord;
					int last = texCoord.length;
					texCoord += texCoord[tri.refTexIndex(i)];
					int cf = coordFlag[idx];
					int ci = contigCoord[cf].ci;
					int gi = contigCoord[cf].gi;
					tri.refIndex(i) = ci - ownerGroup[gi].fromCoord;
					tri.refTexIndex(i) = last;
				}
			}
			else{
				int ci = contigCoord[idx].ci;
				int gi = contigCoord[idx].gi;
				tri.refIndex(i) = ci - ownerGroup[gi].fromCoord;
			}
		}
	}
}

int Seamless::setCoordFroms(){
	int contigLen = 0;
	for(int k = 0; k < partList.length; k++){
		Art* part = partList[k];
		part->fromCoord = contigLen;
		contigLen += part->coord.length;
	}
	return contigLen;
}

void selectVerticesFromSelectedTriangles(Seamless* smls, M_Int& coordFlag, M_<S_VertexIndex>& triList, M_<S_VertexIndex>& coordList){
	coordFlag.length = 0;
	coordFlag.length = smls->setCoordFroms();

	for(int k = triList.length; --k >= 0;){

		const S_Triangle& tri = triList[k].p->triangle[triList[k].i];

		for(int i = 3; --i >= 0; ){
			Art* part = tri.refPart(i);
			int idx = tri.refIndex(i);
			int contigIdx = part->fromCoord + idx;
			if(coordFlag[contigIdx] == 0){
				coordFlag[contigIdx] = 1;
				coordList += S_VertexIndex(part, idx);
			}
		}
	}
}

void multiBreak(Seamless* smls){
	M_Art& partList = smls->partList;
	M_<OwnerGroup> ownerGroup;
	M_<ContigCoord> contigCoord;
	M_Int coordFlag;
	M_Triangle triangle;
	M_Int triFlag;

	int selGroup = 0;

	ownerGroup.length = countGroups(0, partList, selGroup);
	concatCoord(partList, ownerGroup, contigCoord, triangle);

	markSelectedAndNonSelectedTriangles(MultiSelect::triList, triangle, triFlag);
	markSelectedAndNonSelectedCoords(triangle, triFlag, contigCoord, coordFlag);
	addNewCoords(contigCoord, coordFlag);

	setCoordIndex(ownerGroup, contigCoord);
	setPartCoordsLengthsAndSetGroupOffsets(ownerGroup);
	addOffsetsToCoordIndex(ownerGroup, contigCoord);

	//dLog("ci");
	//for(int i = 0; i < contigCoord.length; i++){
	//	dLog(contigCoord[i].ci);
	//}
	//dLog("gi");
	//for(int i = 0; i < contigCoord.length; i++){
	//	dLog(contigCoord[i].gi);
	//}
	
	redirectToNewCoords(ownerGroup, triangle, triFlag, contigCoord, coordFlag);
	copyTrianglesToParts(triangle);

	smls->resetRelevantParts();
	smls->ctrf = true;
	smls->doBuild();
	makePointerEqualIndexLocation();
	BackUp::monostable.beginCycle();
	smls->verifyIntegrity();

	MultiSelect::coordList.length = 0;
	selectVerticesFromSelectedTriangles(smls, coordFlag, MultiSelect::triList, MultiSelect::coordList);
	MultiSelect::triList.length = 0;
}

void MultiSelect::selectFrontAndBackFaceVertices(Seamless* smls){
	M_Art& partList = smls->partList;
	for(int k = 0; k < partList.length; k++){
		Art* part = partList[k];
		if(part->show == false){
			continue;
		}
		M_Vec3f& coord = part->coord;
		int coordLen = coord.length;
		for(int i = 0; i <coordLen ; i++){
			if(isInBox(coord[i])){
				Art* p = part;
				coordList += S_VertexIndex(p, i);
			}
		}
	}
}

void MultiSelect::selectFrontFaceVertices(Seamless* smls){
	M_Art& partList = smls->partList;
	M_Bool selectCoord;
	int selectCoordLen = 0;

	for(int k = 0; k < partList.length; k++){
		Art* part = partList[k];
		selectCoordLen += part->coord.length;
	}
	selectCoord.length = selectCoordLen;

	for(int k = coordList.length; --k >= 0;){
		selectCoord[coordList[k].p->fromCoord + coordList[k].i] = true;
	}

	for(int k = 0; k < partList.length; k++){
		Art* part = partList[k];
		if(part->show == false){
			continue;
		}
		M_Triangle& triangle = part->triangle;
		for(int i = triangle.length; --i>=0;){
			S_Triangle& tri = triangle[i];

			Art* partA = tri.pa;
			Art* partB = tri.pb;
			Art* partC = tri.pc;

			if(backfaceIgnore){
				Vec3f coordA = partA->matrix * g_viewMatrix * tri.pa->coord[tri.ia];
				Vec3f coordB = partB->matrix * g_viewMatrix * tri.pb->coord[tri.ib];
				Vec3f coordC = partC->matrix * g_viewMatrix * tri.pc->coord[tri.ic];

				if(!backFace(coordA, coordB, coordC)){
					continue;
				}
			}
			Vec2f va = convertPointToScreen2d(partA->matrix * partA->coord[tri.ia]);
			Vec2f vb = convertPointToScreen2d(partB->matrix * partB->coord[tri.ib]);
			Vec2f vc = convertPointToScreen2d(partC->matrix * partC->coord[tri.ic]);

			Vec2f tri2d[3] = {va, vb, vc};

			int contigIdx = partA->fromCoord + tri.ia;
			if(selectCoord[contigIdx] == false && isInBox(va)){
				selectCoord[contigIdx] = true;
				coordList += S_VertexIndex(partA, tri.ia);
			}

			contigIdx = partB->fromCoord + tri.ib;
			if(selectCoord[contigIdx] == false && isInBox(vb)){
				selectCoord[contigIdx] = true;
				coordList += S_VertexIndex(partB, tri.ib);
			}

			contigIdx = partC->fromCoord + tri.ic;
			if(selectCoord[contigIdx] == false && isInBox(vc)){
				selectCoord[contigIdx] = true;
				coordList += S_VertexIndex(partC, tri.ic);
			}
		}
	}
}

void applyOwnership(Art* part){
	PaletteCP* paletteCP = paletteCPAllreadyOpen();
	if(paletteCP == 0){
		return;
	}

	Seamless* smls = part->smls;
	M_Art& partList = smls->partList;
	M_<OwnerGroup> ownerGroup;
	M_<ContigCoord> contigCoord;
	M_Triangle triangle;

	int selGroup = 0;
	smls->process();

	int len = ownerGroup.length = countGroups(part, partList, selGroup);
	concatCoord(partList, ownerGroup, contigCoord, triangle);

	applySelection(MultiSelect::coordList, selGroup + paletteCP->selTab.x, contigCoord);

	setCoordIndex(ownerGroup, contigCoord);
	setPartCoordsLengthsAndSetGroupOffsets(ownerGroup);
	addOffsetsToCoordIndex(ownerGroup, contigCoord);
	//copyCoordsToParts(ownerGroup, contigCoord);
	//clearTriangles(partList);
	
	redirectTriangles(triangle, ownerGroup, contigCoord);
	copyTrianglesToParts(triangle);

	MultiSelect::reset();

	smls->process();
	smls->resetRelevantParts();

	smls->ctrf = true;
	smls->doBuild();
	makePointerEqualIndexLocation();
	paletteCP->updateOwnershipFields(part);
	paletteCP->paint();
	BackUp::monostable.beginCycle();
	smls->verifyIntegrity();
}

void createContigLookUpList(M_<CPContigElement>& contigLookUpList, M_<NurbsPatch*>& cpNodeList){
	for(int k = 0; k < cpNodeList.length; k++){
		NurbsPatch* patch = cpNodeList[k];
		for(int i = patch->point.length; --i >= 0;){
			contigLookUpList += CPContigElement(patch);
		}
	}
}

class CPOwnerGroup{
public:
	Art* part;
	M_Int contigIdx;
	CPOwnerGroup():
		part(0)
	{}
};

void applyOriginalOwnership(M_<CPContigElement>& contigLookUpList, M_Art& partList, M_<CPOwnerGroup>& ownerGroup){
	int groupIdx = 0;
	for(int k = 0; k < partList.length; k++){
		Art* part = partList[k];
		M_ControlPoint& controlPoint = part->controlPoint;
		M_Int& coordGroup = part->coordGroup;

		int q = 0;
		for(int cg = 0; cg < coordGroup.length; cg++){
			for(int c = coordGroup[cg]; --c >= 0 && q < controlPoint.length; q++){
				ControlPoint& cp = controlPoint[q];
				int idx = cp.i + cp.p->fromPoint;
				contigLookUpList[idx].group = groupIdx;
			}
			ownerGroup[groupIdx].part = part;
			groupIdx++;
		}
		if(q < controlPoint.length){
			for(; q < controlPoint.length; q++){
				ControlPoint& cp = controlPoint[q];
				int idx = cp.i + cp.p->fromPoint;
				contigLookUpList[idx].group = groupIdx;
			}
		}
		ownerGroup[groupIdx].part = part;
		groupIdx++;
	}
}

void NurbsPatch::mirrorSingleCPOwnership(M_<CPContigElement>& contigLookUpList, int idx, const int& selGroupIdx){
	if(mirror){
		if(mirrorPatch){
			contigLookUpList[mirrorPatch->fromPoint + idx].group = selGroupIdx;
		}
		else{
			int uDim = dim.x;
			int vDim = dim.y;
			int u = idx % uDim;
			int v = idx / uDim;
			int middleIdx = (uDim >> 1);
			int midx = v * uDim + uDim - 1 - u;
			contigLookUpList[fromPoint + midx].group = selGroupIdx;
		}
	}
}

inline void NurbsPatch::applyRectOwnership(const Art* part, M_<CPContigElement>& contigLookUpList, int i, const int& selGroupIdx, const int& selGroupIdxMirror){
	if(MultiSelect::isInBox(dpoint[i])){
		contigLookUpList[fromPoint + i].group = selGroupIdx;
		if(part->mirror){
			if(const Art* mirrorPart = part->mirrorPart){
				if(mirrorPart->coordGroup.length == part->coordGroup.length){
					mirrorSingleCPOwnership(contigLookUpList, i, selGroupIdxMirror);
				}
			}
			else{
				mirrorSingleCPOwnership(contigLookUpList, i, selGroupIdx);
			}
		}
	}
}


inline void NurbsPatch::applyRectOwnership(const Art* part, M_<CPContigElement>& contigLookUpList, const int& selGroupIdx, const int& selGroupIdxMirror){
	int uDim = this->dim.x;
	int vDim = this->dim.y;

	if(cage == false){
		return;
	}

	if(showAll){
		M_Vec3f& dpoint = this->dpoint.refValue();
		int len = dpoint.length;
		for(int i = 0; i < len; i++){
			if(MultiSelect::isInBox(dpoint[i])){
				applyRectOwnership(part, contigLookUpList, i, selGroupIdx, selGroupIdxMirror);
			}
		}
	}
	else{
		if(uJoinEnds){
			if(isCtrlPolyRef){
				for(int u = 1; u < uDim - 1; u += 2){
					for(int v = 1; v < vDim - 1; v++){
						int i = v * uDim + u;
						if(MultiSelect::isInBox(dpoint[i])){
							applyRectOwnership(part, contigLookUpList, i, selGroupIdx, selGroupIdxMirror);
						}
					}
				}
			}
			else{
				for(int u = 1; u < uDim; u += 2){
					for(int v=0; v < vDim; v++){
						int i = v * uDim + u;
						if(MultiSelect::isInBox(dpoint[i])){
							applyRectOwnership(part, contigLookUpList, i, selGroupIdx, selGroupIdxMirror);
						}
					}
				}
			}
		}
		else{
			if(NurbsPatch::MID_1 == checkNurbsType(vDim, vKnot)){
				if(isCtrlPolyRef){
					for(int v = 1; v < vDim - 1; v += 2){
						for(int u = 1; u < uDim - 1; u += 2){
							int i = v * uDim + u;
							if(MultiSelect::isInBox(dpoint[i])){
								applyRectOwnership(part, contigLookUpList, i, selGroupIdx, selGroupIdxMirror);
							}
						}
					}
				}
				else{
					for(int v = 0; v < vDim; v += 2){
						for(int u = 0; u < uDim; u += 2){
							int i = v * uDim + u;
							if(MultiSelect::isInBox(dpoint[i])){
								applyRectOwnership(part, contigLookUpList, i, selGroupIdx, selGroupIdxMirror);
							}
							if(u == 0 || u == uDim - 2){
								u--;
							}
						}
						if(v == 0 || v == vDim - 2){
							v--;
						}
					}
				}
			}
			else{
				if(NurbsPatch::MID_1 == checkNurbsType(uDim, uKnot)){
					if(isCtrlPolyRef){
						for(int v = 1; v < vDim - 1; v++){
							for(int u = 1; u < uDim - 1; u += 2){
								int i = v * uDim+u;
								if(MultiSelect::isInBox(dpoint[i])){
									applyRectOwnership(part, contigLookUpList, i, selGroupIdx, selGroupIdxMirror);
								}
							}
						}
					}
					else{
						for(int v = 0; v < vDim; v++){
							for(int u = 1; u < uDim; u +=2){
								int i = v * uDim + u;
								if(MultiSelect::isInBox(dpoint[i])){
									applyRectOwnership(part, contigLookUpList, i, selGroupIdx, selGroupIdxMirror);
								}
							}
						}
					}
				}
				else{
					if(isCtrlPolyRef){
						for(int v = 1; v < vDim - 1; v++){
							for(int u = 1; u < uDim - 1; u++){
								int i = v * uDim + u;
								if(MultiSelect::isInBox(dpoint[i])){
									applyRectOwnership(part, contigLookUpList, i, selGroupIdx, selGroupIdxMirror);
								}
							}
						}
					}
					else {
						for(int v = 0; v < vDim; v++){
							for(int u = 0; u < uDim; u++){
								int i = v * uDim + u;
								if(MultiSelect::isInBox(dpoint[i])){
									applyRectOwnership(part, contigLookUpList, i, selGroupIdx, selGroupIdxMirror);
								}
							}
						}
					}
				}
			}
		}
	}
}

/*
inline void NurbsPatch::applyRectOwnership(const Art* part, M_<CPContigElement>& contigLookUpList, const int& selGroupIdx, const int& selGroupIdxMirror){
	const int uDim = dim.x;
	const int vDim = dim.y;

	if(showAll){
		for(int v=0; v < vDim; v++){
			for(int u=0; u < uDim; u++){
				int i = v * uDim + u;
				applyRectOwnership(part, contigLookUpList, i, selGroupIdx, selGroupIdxMirror);
			}
		}
	}
	else{
		if(uJoinEnds){
			for(int v = 0; v < vDim; v++){
				for(int u = 1; u < uDim; u+=2){
					int i = v * uDim + u;
					applyRectOwnership(part, contigLookUpList, i, selGroupIdx, selGroupIdxMirror);
				}
			}
		}
		else{
			if(NurbsPatch::MID_1 == checkNurbsType(vDim, vKnot)){
				for(int v=0; v < vDim; v += 2){
					for(int u=0; u<uDim; u += 2){
						int i = v * uDim + u;
						applyRectOwnership(part, contigLookUpList, i, selGroupIdx, selGroupIdxMirror);
						if(u == 0 || u == uDim - 2){
							u--;
						}
					}
					if(v == 0 || v == vDim - 2){
						v--;
					}
				}
			}
			else{
				for(int v = 0; v < vDim; v++){
					for(int u = 0; u < uDim; u += 2){
						int i = v * uDim + u;
						applyRectOwnership(part, contigLookUpList, i, selGroupIdx, selGroupIdxMirror);
						if(u == 0 || u == uDim - 2){
							u--;
						}
					}
				}
			}
		}
	}
}
*/

void Seamless::applyRectOwnership(const Art* part, M_<CPContigElement>& contigLookUpList, M_<NurbsPatch*>& cpNodeList, const int& selGroupIdx, const int& selGroupIdxMirror){
	for(int k = 0; k < cpNodeList.length; k++){
		NurbsPatch* patch = cpNodeList[k];
		if(patch->cage){
			patch->applyRectOwnership(part, contigLookUpList, selGroupIdx, selGroupIdxMirror);
		}
	}
	for(int i = 0; i < ctrlPolyPoint.length;){
		int ctrlPoly = i;
		int len = ctrlPolyPoint[ctrlPoly].len;
		for(int k = len; --k >= 0;){
			NurbsPatch* patch = ctrlPolyPoint[i].p;
			int idx = ctrlPolyPoint[i].i;
			if(patch->cage && MultiSelect::isInBox(patch->dpoint[idx])){
				//contigList[patch->fromPoint + idx] = patch;
				patch->applyRectOwnership(part, contigLookUpList, idx, selGroupIdx, selGroupIdxMirror);
			}
			i++;
		}
		i = getNextCtrlPoly(ctrlPoly);
	}

}

void clearCoordGroups(M_Art& partList){
	for(int k = partList.length; --k >= 0;){
		Art* part = partList[k];
		M_Int& coordGroup = part->coordGroup;
		for(int i = coordGroup.length; --i >= 0;){
			coordGroup[i] = 0;
		}
	}
}
void addBackOwnerPoints(M_<CPOwnerGroup>& ownerGroup, M_<CPContigElement>& contigLookUpList){
	int groupLen = ownerGroup.length;
	Art* prevPart = 0;
	int gi = 0;
	for(int k = 0; k < groupLen; k++){
		M_Int& contigIdx = ownerGroup[k].contigIdx;
		Art* part = ownerGroup[k].part;
		M_Int& coordGroup = part->coordGroup.refValue();
		if(prevPart != part){
			part->controlPoint.length = 0;
			prevPart = part;
			gi = 0;
		}
		M_ControlPoint& controlPoint = part->controlPoint;

		if(gi < coordGroup.length){
			coordGroup[gi] = contigIdx.length;
		}

		for(int i = 0; i < contigIdx.length; i++){
			int idx = contigIdx[i];
			NurbsPatch* patch = dynamic_cast<NurbsPatch*>(contigLookUpList[idx].cpNode);
			controlPoint += ControlPoint(patch, idx - patch->fromPoint);
		}
		gi++;
	}
}
void fillGroups(M_<CPOwnerGroup>& ownerGroup, M_<CPContigElement> contigLookUpList){
	for(int k = 0; k < contigLookUpList.length; k++){
		int gi = contigLookUpList[k].group;
		if(gi >= 0){
			ownerGroup[gi].contigIdx += k;
		}
	}
}


void redirectCPMorpherIndicesToContigList(Seamless* smls){
	for(int k = g_animBarList.length; --k>=0;){
		g_animBarList[k]->redirectCPMorpherIndicesToContigList(smls);
	}
}

void redirectContigListBackToCPMorpherIndices(Seamless* smls, M_<CPContigElement>& contigLookUpList){
	M_Art& partList = smls->partList;

	for(int k = contigLookUpList.length; --k>=0;){
		contigLookUpList[k] = CPContigElement();
	}
	for(int k = partList.length; --k>=0;){
		Art* part = partList[k];
		M_ControlPoint& m_controlPoint = part->controlPoint;
		for(int i = m_controlPoint.length; --i>=0;){
			ControlPoint& controlPoint = part->controlPoint[i];
			contigLookUpList[controlPoint.p->fromPoint + controlPoint.i] = CPContigElement(part, i);
		}
	}
	for(int k = g_animBarList.length; --k>=0;){
		g_animBarList[k]->redirectContigListBackToCPMorpherIndices(smls, contigLookUpList);
	}
}

void applyCPOwnership(Art* part){
	PaletteCP* paletteCP = paletteCPAllreadyOpen();
	if(paletteCP == 0){
		return;
	}
	int selGroup = 0;
	int selGroupMirror = 0;
	Seamless* smls = part->smls;
	M_Art& partList = smls->partList;
	smls->process();

	M_<CPOwnerGroup> ownerGroup;
	int len = ownerGroup.length = countGroups(part, partList, selGroup);
	if(part->mirror){
		if(part->mirrorPart){
			countGroups(part->mirrorPart, partList, selGroupMirror);
		}
	}

	M_<NurbsPatch*>& cpNodeList = smls->cpNodeList;

	M_<CPContigElement> contigLookUpList;
	createContigLookUpList(contigLookUpList, cpNodeList);

	redirectCPMorpherIndicesToContigList(smls);

	applyOriginalOwnership(contigLookUpList, partList, ownerGroup);
	smls->applyRectOwnership(part, contigLookUpList, cpNodeList, selGroup + paletteCP->selTab.x, selGroupMirror + paletteCP->selTab.x);
	fillGroups(ownerGroup, contigLookUpList);
	clearCoordGroups(partList);
	addBackOwnerPoints(ownerGroup, contigLookUpList);

	redirectContigListBackToCPMorpherIndices(smls, contigLookUpList);

	paletteCP->updateOwnershipFields(part);
	paletteCP->paint();
	BackUp::monostable.beginCycle();
}


void addExtraGroups(M_Int& srcCoordGroup, M_Float& srcWeight, M_Int& dstCoordGroup, M_Float& dstWeight){
	for(int g = 0; g < srcCoordGroup.length; g++){
		float slider = srcWeight[g];

		for(int i = 0; i < dstWeight.length; i++){
			if(slider == dstWeight[i]){
				break;
			}
			else if(dstWeight[i] > slider){
				dstWeight.insertElement(i, slider);
				dstCoordGroup.insertElement(i, 0);
				break;
			}
		}
	}
}

void addExtraGroups(Seamless* srcSmls, Seamless* dstSmls){
	M_Art& srcPartList = srcSmls->partList;
	M_Art& dstPartList = dstSmls->partList;
	for(int i = srcPartList.length; --i>=0;){
		if(Art* importPart = srcPartList[i]){
			for(int j = dstPartList.length; --j>=0;){
				if(Art* existingPart = dstPartList[j]){
					if(importPart->name == existingPart->name){
						addExtraGroups(importPart->coordGroup.refValue(), importPart->weight.refValue(), existingPart->coordGroup.refValue(), existingPart->weight.refValue());
						break;
					}
				}
			}
		}
	}
}


int getGroupOffset(M_Int& dstCoordGroup, M_Float& dstWeight, float srcWeight){
	int groupOffset = 0;
	for(int i = 0; i < dstCoordGroup.length; i++){
		if(dstWeight[i] == srcWeight){
			break;
		}
		groupOffset++;
	}
	return groupOffset;
}

void addPartOwnership2(M_ControlPoint& controlPoint, M_Int& dstCoordGroup, M_Float& dstWeight, M_Int& srcCoordGroup, M_Float& srcWeight, M_<CPContigElement>& contigLookUpList, M_<CPOwnerGroup>& ownerGroup, const int groupIdx){
	int q = 0;
	for(int cg = 0; cg < srcCoordGroup.length; cg++){
		for(int c = srcCoordGroup[cg]; --c >= 0 && q < controlPoint.length; q++){
			int groupOffset = getGroupOffset(dstCoordGroup, dstWeight, srcWeight[cg]);

			ControlPoint& cp = controlPoint[q];
			int idx = cp.i + cp.p->fromPoint;
			contigLookUpList[idx].group = groupIdx + groupOffset;
		}
		//ownerGroup[groupIdx].part = part;
	}
	if(q < controlPoint.length){
		for(; q < controlPoint.length; q++){
			ControlPoint& cp = controlPoint[q];
			int idx = cp.i + cp.p->fromPoint;
			contigLookUpList[idx].group = groupIdx + dstCoordGroup.length;
		}
	}
}


void copyOwnership(Seamless* srcSmls, Seamless* dstSmls, M_<CPContigElement>& contigLookUpList, M_<CPOwnerGroup>& ownerGroup, M_Int& partGroupIndex){
	M_Art& srcPartList = srcSmls->partList;
	M_Art& dstPartList = dstSmls->partList;
	for(int i = srcPartList.length; --i>=0;){
		if(Art* importPart = srcPartList[i]){
			for(int j = dstPartList.length; --j>=0;){
				if(Art* existingPart = dstPartList[j]){
					if(importPart->name == existingPart->name){
						int groupIdx = partGroupIndex[j];
						addPartOwnership2(importPart->controlPoint, existingPart->coordGroup.refValue(), existingPart->weight.refValue(), importPart->coordGroup.refValue(), importPart->weight.refValue(), contigLookUpList, ownerGroup, groupIdx);
						break;
					}
				}
			}
		}
	}
}

void importControlPointOwnership(Seamless* srcSmls, Seamless* dstSmls){
	M_<NurbsPatch*>& cpNodeList = dstSmls->cpNodeList;

	M_Art& srcPartList = srcSmls->partList;
	M_Art& dstPartList = dstSmls->partList;

	dstSmls->process();
	dstSmls->populateMirrorNodes();
	addExtraGroups(srcSmls, dstSmls);
	dstSmls->cpNodeList.length = 0;
	dstSmls->cpNodeListAdd(dstSmls->build.refValue());


	M_<CPOwnerGroup> ownerGroup;
	M_Int partGroupIndex;

	ownerGroup.length = countGroups(dstPartList, partGroupIndex);

	M_<CPContigElement> contigLookUpList;
	createContigLookUpList(contigLookUpList, cpNodeList);
	applyOriginalOwnership(contigLookUpList, dstPartList, ownerGroup);
	//smls->applyRectOwnership(srcPart, contigLookUpList, cpNodeList, selGroup + paletteCP->selTab.x, selGroupMirror + paletteCP->selTab.x);
	copyOwnership(srcSmls, dstSmls, contigLookUpList, ownerGroup, partGroupIndex);


	fillGroups(ownerGroup, contigLookUpList);
	clearCoordGroups(dstPartList);
	addBackOwnerPoints(ownerGroup, contigLookUpList);
}

bool colorBuildNodes(M_Build& m_build, Art* part){
	for(int k = m_build.length; --k >= 0;){
		Build* bld = m_build[k];

		if(SphereBuild* sphereBuild = dynamic_cast<SphereBuild*>(bld)){
			if(sphereBuild->part.p == part){
				sphereBuild->color = g_palette->color;
				return true;
			}
		}
		else if(CylinderBuild* cylinderBuild = dynamic_cast<CylinderBuild*>(bld)){
			if(cylinderBuild->part.p == part){
				cylinderBuild->color = g_palette->color;
				return true;
			}
		}
		else if(ConeBuild* coneBuild = dynamic_cast<ConeBuild*>(bld)){
			if(coneBuild->part.p == part){
				coneBuild->color = g_palette->color;
				return true;
			}
		}
		else if(BoxBuild* boxBuild = dynamic_cast<BoxBuild*>(bld)){
			if(boxBuild->part.p == part){
				boxBuild->color = g_palette->color;
				return true;
			}
		}
		else if(TorusBuild* torusBuild = dynamic_cast<TorusBuild*>(bld)){
			if(torusBuild->part.p == part){
				torusBuild->color = g_palette->color;
				return true;
			}
		}
		else if(BezierLatheBuild* bezierLatheBuild = dynamic_cast<BezierLatheBuild*>(bld)){
			if(bezierLatheBuild->part.p == part){
				bezierLatheBuild->color = g_palette->color;
				return true;
			}
		}
		else if(RectangleBuild* rectangleBuild = dynamic_cast<RectangleBuild*>(bld)){
			if(rectangleBuild->part.p == part){
				rectangleBuild->color = g_palette->color;
				return true;
			}
		}
		else if(NurbsLathe* nurbsLathe = dynamic_cast<NurbsLathe*>(bld)){
			if(nurbsLathe->part.p == part){
				nurbsLathe->color = g_palette->color;
				return true;
			}
		}
		else if(BendyLathe* bendyLathe = dynamic_cast<BendyLathe*>(bld)){
			if(bendyLathe->part.p == part){
				bendyLathe->color = g_palette->color;
				return true;
			}
		}
		else if(GroupBuild* groupBuild = dynamic_cast<GroupBuild*>(bld)){
			if(colorBuildNodes(groupBuild->build.refValue(), part)){
				return true;
			}
		}
	}
	return false;
}

void checkBuildNodesForColorRect(Art* part){
	if(part->show == false){
		return;
	}
	M_Vec3f& coord = part->coord;
	int coordLen = coord.length;
	for(int i = 0; i < coordLen ; i++){
		if(MultiSelect::isInBox(coord[i])){
			if(colorBuildNodes(part->smls->build.refValue(), part)){
				return;
			}
		}
	}
}

void applyColor(Seamless* smls){
	PaletteCP* paletteCP = paletteCPAllreadyOpen();
	if(paletteCP == 0){
		return;
	}
	M_Art& partList = smls->partList;


	for(int k = 0; k < partList.length; k++){
		Art* part = partList[k];
		if(dynamic_cast<Part*>(part)){
			if(dynamic_cast<Part*>(part)){
				checkBuildNodesForColorRect(part);
				continue;
			}
		}

		M_Triangle& triangle = part->triangle;


		for(int t = triangle.length; --t >= 0;){
			S_Triangle& tri = triangle[t];
			Art* owner = tri.getOwner();

			if(ColorEffect* colorEffect = dynamic_cast<ColorEffect*>(owner->workEffect)){
				M_Color3f& m_color = colorEffect->color.refValue();
				for(int i = 3; --i >= 0; ){
					Art* part = tri.refPart(i);
					int idx = tri.refIndex(i);
					if(MultiSelect::isInBox(part->coord[idx])){
						m_color[tri.refTexIndex(i)] = g_palette->color;
					}
				}
			}
		}
	}
	MultiSelect::reset();

	smls->resetRelevantParts();

	smls->ctrf = true;
	smls->doBuild();
	makePointerEqualIndexLocation();
	BackUp::monostable.beginCycle();
	smls->verifyIntegrity();
}

void addEffectToList(M_<Effect*>& list, Effect* effect){
	for(int k = list.length; --k >= 0;){
		if(list[k] == effect)return;
	}
	list += effect;
}
void optimizeAllEffects(Seamless* smls){
	M_<Effect*> list;

	M_Art& partList = smls->partList;
	for(int k = 0; k < partList.length; k++){
		addEffectToList(list, partList[k]->workEffect);
	}

	for(int k = 0; k < list.length; k++){
		list[k]->optimize(partList);
	}
}


void applyTriangleColor(Seamless* smls){
	PaletteCP* paletteCP = paletteCPAllreadyOpen();
	if(paletteCP == 0){
		return;
	}
	M_Art& partList = smls->partList;
	bool bluePartProcessed = false;


	for(int k = 0; k < partList.length; k++){
		Art* part = partList[k];
		if(dynamic_cast<Part*>(part)){
			if(dynamic_cast<Part*>(part)){
				checkBuildNodesForColorRect(part);
				continue;
			}
		}
		M_Triangle& triangle = part->triangle;
		for(int i = triangle.length; --i>=0;){
			S_Triangle& tri = triangle[i];

			Art* partA = tri.pa;
			Art* partB = tri.pb;
			Art* partC = tri.pc;

			if(MultiSelect::backfaceIgnore){
				Vec3f coordA = partA->matrix * g_viewMatrix * tri.pa->coord[tri.ia];
				Vec3f coordB = partB->matrix * g_viewMatrix * tri.pb->coord[tri.ib];
				Vec3f coordC = partC->matrix * g_viewMatrix * tri.pc->coord[tri.ic];

				if(!backFace(coordA, coordB, coordC)){
					continue;
				}
			}
			Vec2f va = convertPointToScreen2d(partA->matrix * partA->coord[tri.ia]);
			Vec2f vb = convertPointToScreen2d(partB->matrix * partB->coord[tri.ib]);
			Vec2f vc = convertPointToScreen2d(partC->matrix * partC->coord[tri.ic]);

			Vec2f tri2d[3] = {va, vb, vc};

			if(MultiSelect::isInBox(va) || MultiSelect::isInBox(vb) || MultiSelect::isInBox(vc) || triIntersectsRect(va, vb, vc) || isRectCornerInTriangle(tri2d)){
				Art* owner = tri.getOwner();
				if(ColorEffect* colorEffect = dynamic_cast<ColorEffect*>(owner->workEffect)){
					M_Color3f& m_color = colorEffect->color.refValue();
					for(int i = 3; --i >= 0; ){
						bluePartProcessed = true;
						tri.refTexIndex(i) = m_color.length;
						m_color += g_palette->color;
					}
				}
			}
		}
	}
	if(bluePartProcessed){
		optimizeAllEffects(smls);
	}

	MultiSelect::reset();
	smls->resetRelevantParts();

	smls->ctrf = true;
	smls->doBuild();
	makePointerEqualIndexLocation();
	BackUp::monostable.beginCycle();
	smls->verifyIntegrity();
}

void genTriangles(Art* part, M_<S_VertexIndex>& triList);
extern Ptr<PartClue> g_vcArtTriBase;

bool notInList(Art* part, int idx, M_<S_VertexIndex>& triList, int oldLen){
	for(int k = 0; k < oldLen; k++){
		if(triList[k].p == part && triList[k].i == idx){
			return false;
		}
	}
	return true;
}


void selectMultipleTriangles(Seamless* smls){

	M_Art& partList = smls->partList;
	int oldLen = MultiSelect::triList.length;

	for(int k = 0; k < partList.length; k++){
		Art* part = partList[k];
		if(part->show == false){
			continue;
		}
		M_Triangle& triangle = part->triangle;
		for(int i = triangle.length; --i>=0;){
			S_Triangle& tri = triangle[i];

			Art* partA = tri.pa;
			Art* partB = tri.pb;
			Art* partC = tri.pc;

			if(MultiSelect::backfaceIgnore){
				Vec3f coordA = partA->matrix * g_viewMatrix * tri.pa->coord[tri.ia];
				Vec3f coordB = partB->matrix * g_viewMatrix * tri.pb->coord[tri.ib];
				Vec3f coordC = partC->matrix * g_viewMatrix * tri.pc->coord[tri.ic];

				if(!backFace(coordA, coordB, coordC)){
					continue;
				}
			}
			Vec2f va = convertPointToScreen2d(partA->matrix * partA->coord[tri.ia]);
			Vec2f vb = convertPointToScreen2d(partB->matrix * partB->coord[tri.ib]);
			Vec2f vc = convertPointToScreen2d(partC->matrix * partC->coord[tri.ic]);

			Vec2f tri2d[3] = {va, vb, vc};

			if(MultiSelect::isInBox(va) || MultiSelect::isInBox(vb) || MultiSelect::isInBox(vc) || triIntersectsRect(va, vb, vc) || isRectCornerInTriangle(tri2d)){
				Art* p = tri.getOwner();
				if(notInList(p, i, MultiSelect::triList, oldLen)){
					MultiSelect::triList+=S_VertexIndex(p, i);
				}
			}
		}
		genTriangles(g_vcArtTriBase->transpSmls->skeleton.refValue()[0], MultiSelect::triList);
	}
}

void applyCleanEdgeColor(Seamless* smls){
	PaletteCP* paletteCP = paletteCPAllreadyOpen();
	if(paletteCP == 0){
		return;
	}

	M_<S_VertexIndex>& triList = MultiSelect::triList;

	for(int k = 0; k < triList.length; k++){
		Art* owner = triList[k].p;
		S_Triangle& tri = owner->triangle[triList[k].i];
		if(ColorEffect* colorEffect = dynamic_cast<ColorEffect*>(owner->workEffect)){
			M_Color3f& m_color = colorEffect->color.refValue();
			for(int i = 3; --i >= 0; ){
				Art* part = tri.refPart(i);
				int idx = tri.refIndex(i);
				if(MultiSelect::isInBox(part->coord[idx])){
					tri.refTexIndex(i) = m_color.length;
					m_color += g_palette->color;
				}
			}
		}
	}
	optimizeAllEffects(smls);

	MultiSelect::reset();
	MultiSelect::triList.length = 0;

	smls->resetRelevantParts();

	smls->ctrf = true;
	smls->doBuild();
	makePointerEqualIndexLocation();
	BackUp::monostable.beginCycle();
	smls->verifyIntegrity();
}

void PaletteCP::updateOwnershipFields(Art* part){
	float weight = 1;

	if(selTab.x < part->weight.length){
		weight = part->weight[selTab.x];
	}
	g_palette->weight.refValue() = weight;

	CommonSpec* commonSpec = getCommonSpec(part);

	if(commonSpec && selTab.x < commonSpec->color.length){
		if(part->coordGroup.length == 0){
			g_palette->color.refValue() = commonSpec->color[commonSpec->color.length-1];
		}
		else{
			g_palette->color.refValue() = commonSpec->color.refValue()[selTab.x];
		}
	}

	int coords = 0;
	if(part->smls->workAnimCP){
		if(part->coordGroup.length == 0){
			coords = part->controlPoint.length;
		}
		else if(selTab.x < part->coordGroup.length){
			coords = part->coordGroup[selTab.x];
		}
		else{
			for(int i = part->coordGroup.length; --i >= 0;){
				coords += part->coordGroup[i];
			}
			coords = part->controlPoint.length - coords;
		}
		g_palette->coords.refValue() = coords;
	}
	else{
		if(part->coordGroup.length == 0){
			coords = part->coord.length;
		}
		else if(selTab.x < part->coordGroup.length){
			coords = part->coordGroup[selTab.x];
		}
		else{
			for(int i = part->coordGroup.length; --i >= 0;){
				coords += part->coordGroup[i];
			}
			coords = part->coord.length - coords;
		}
		g_palette->coords.refValue() = coords;
	}
}

/*
void setNodeColor(const Color3f& color){
	Node* node = g_selNode.getRaw();
	if(GroupBuild* groupBuild = dynamic_cast<GroupBuild*>(node)){
		groupBuild->marker = color;
		g_stw->paint();
		g_stw->cp->paint();
	}
	else if(SphereBuild* sphereBuild = dynamic_cast<SphereBuild*>(node)){
		sphereBuild->color = color;
		updateBuild3dWindow();
		return;
	}
	else if(CylinderBuild* cylinderBuild = dynamic_cast<CylinderBuild*>(node)){
		cylinderBuild->color = color;
	}
	else if(ConeBuild* coneBuild = dynamic_cast<ConeBuild*>(node)){
		coneBuild->color = color;
	}
	else if(BoxBuild* boxBuild = dynamic_cast<BoxBuild*>(node)){
		boxBuild->color = color;
	}
	else if(TorusBuild* torusBuild = dynamic_cast<TorusBuild*>(node)){
		torusBuild->color = color;
	}
	else if(BezierLatheBuild* bezierLatheBuild = dynamic_cast<BezierLatheBuild*>(node)){
		bezierLatheBuild->color = color;
	}
	else if(RectangleBuild* rectangleBuild = dynamic_cast<RectangleBuild*>(node)){
		rectangleBuild->color = color;
	}
	else if(NurbsLathe* nurbsLathe = dynamic_cast<NurbsLathe*>(node)){
		nurbsLathe->color = color;
	}
	else if(BendyLathe* bendyLathe = dynamic_cast<BendyLathe*>(node)){
		bendyLathe->color = color;
	}
	else if(ColorSweep* colorSweep = dynamic_cast<ColorSweep*>(node)){
		int index = colorSweep->index[0];
		if(index < colorSweep->color.length){
			colorSweep->color.refValue()[index] = color;
		}
	}
	else{
		return;
	}
	updateBuild3dWindow();
}
*/

void syncCommonSpecLength(Art* part){
	if(CommonSpec* commonSpec = getCommonSpec(part)){
		M_Color3f& m_color = commonSpec->color.refValue();
		int len = part->coordGroup.length + 1;
		if(m_color.length < len){
			for(int i = m_color.length; i < len; i++){
				m_color.insertElement(0, Color3f(.5, .5, .5));
			}
		}
	}
}


void clickNodeForOwnershipWindow(){
	if(PaletteCP* paletteCP = paletteCPAllreadyOpen()){
		if(g_colorOwnership){
			if(Art* part = getPart()){
				syncCommonSpecLength(part);
				paletteCP->selTab = Vec2i();
				paletteCP->updateOwnershipFields(part);
				paletteCP->paint();
				part->smls->ctrf = true;
			}
			else{
				exitPaletteWindow();
			}
		}
	}
}

void PaletteCP::setupNodeWindow(Node* node){
	this->node = node;
	Palette* palette = dynamic_cast<Palette*>(node);

	g_controlPanelOpen = true;
	g_propWinLine.length = 0;
	buttonList.length = 0;

	const M_Properties& properties = node->getProperties();

	int yPos = startFieldLineYpos;

	FieldButton* button = palette->insert.setUpControl(*node, palette->insert, properties[0], 0, 50, yPos, *this);
	button->dstPos = Vec2i(4, yPos);
	//button->valPos = Vec2i(, yPos - 8);
	buttonList += button;
	g_propWinLine += PropWinLine(yPos, button);

	button = palette->_delete.setUpControl(*node, palette->_delete, properties[1], 0, 50, yPos, *this);
	button->dstPos = Vec2i(68, yPos);
	buttonList += button;
	g_propWinLine += PropWinLine(yPos, button);

	button = palette->coords.setUpControl(*node, palette->coords, properties[2], 0, 190, yPos, *this);
	buttonList += button;
	g_propWinLine += PropWinLine(yPos, button);

	button = palette->weight.setUpControl(*node, palette->weight, properties[3], 0, 300, yPos, *this);
	buttonList += button;
	g_propWinLine += PropWinLine(yPos, button);

	button = palette->autoWeightDivide.setUpControl(*node, palette->autoWeightDivide, properties[4], 0, 140, yPos + 25 * 2, *this);
	buttonList += button;
	g_propWinLine += PropWinLine(yPos, button);

	button = palette->color.setUpControl(*node, palette->color, properties[5], 0, 210, yPos + 25 * 2, *this);
	//button->dstPos = Vec2i(500, yPos);
	//button->valPos = Vec2i(500 + 22, yPos - 8);
	buttonList += button;
	g_propWinLine += PropWinLine(yPos, button);


	size = Vec2i(384, 120);
	pos = Vec2i(0, 500);

	selTab = Vec2i();

	updateOwnershipFields(getPart());

	hover(g_mousePos);
}

void PaletteCP::setupColorSelectorNodeWindow(Node* node){
	this->node = node;
	Palette* palette = dynamic_cast<Palette*>(node);

	g_controlPanelOpen = true;
	g_propWinLine.length = 0;
	buttonList.length = 0;

	const M_Properties& properties = node->getProperties();

	int yPos = startFieldLineYpos;

	FieldButton* button = palette->color.setUpControl(*node, palette->color, properties[5], 0, 54, yPos, *this);
	//button->dstPos = Vec2i(500, yPos);
	//button->valPos = Vec2i(500 + 22, yPos - 8);
	buttonList += button;
	g_propWinLine += PropWinLine(yPos, button);

	button = palette->define.setUpControl(*node, palette->define, properties[6], 0, 190, 1, *this);
	buttonList += button;
	g_propWinLine += PropWinLine(yPos, button);


	size = Vec2i(390, 240);
	pos = Vec2i(0, 500);

	selTab = Vec2i();

	//updateOwnershipFields(getPart());

	hover(g_mousePos);
}

void sweepArray(const Color3f& fromColor, const Color3f& toColor, Color3f* verticalPixel){
	
	Color3f step = (toColor - fromColor) / 49.0f;

	for(int i = 0; i < 50; i++){
		verticalPixel[i] = fromColor + step * i;
	}
}

inline void plotSweepRow(const Color3f& fromColor, const Color3f& toColor, const int& x, const int& y, int dstWidth, int dstHeight, int* dstPix){
	
	Color3f step = (toColor - fromColor) / 24.0f;

	Color3f color;
	for(int i = 0; i < 25; i++){
		color = fromColor + step * i;
		plotPixel2(color.colorToHex(), x + i, y, dstWidth, dstHeight, dstPix);
	}
}

inline void PaletteCP::paintColor(String& str, int dstWidth, int dstHeight, int* dstPix, const M_Properties& properties){
	str = "Palette";

	drawText2(str, 0x0, Vec2i(8, 16), dstWidth, dstHeight, dstPix);
	int colorY = colorTabsYpos;

	int intColor = g_palette->color.refValue().colorToHex();

	paintBlock2(intColor, colorTabsBeginX, selColorTabYpos, colorTabsBeginX + selColorTabCX, selColorTabYpos + selColorTabCY, dstWidth, dstHeight, dstPix);

	for(int r = 0; r < tabRows - 1; r++){
		int colorX = colorTabsBeginX;
		for(int c = 0; c < tabColums; c++){
			int intColor = g_palette->cell[r * tabColums + c].colorToHex();
			paintBlock2(intColor, colorX, colorY, colorX + colorTabsCX, colorY + colorTabsCY, dstWidth, dstHeight, dstPix);
			colorX += colorTabsCX + colorTabsXSpace;
		}
		colorY += colorTabsCY + colorTabsXSpace;
	}

	//str = "custom colors";
	//drawText2(str, 0x0, Vec2i(8, colorY + 15), dstWidth, dstHeight, dstPix);

	colorY = customColorTabsYpos;

	int colorX = colorTabsBeginX;
	for(int c = 0; c < tabColums; c++){
		int intColor = g_palette->cell[bottomRow + c].colorToHex();
		paintBlock2(intColor, colorX, colorY, colorX + colorTabsCX, colorY + colorTabsCY, dstWidth, dstHeight, dstPix);
		colorX += colorTabsCX + colorTabsXSpace;
	}
	colorY += colorTabsCY + colorTabsXSpace;


	Color3f verticalPixelA[50];
	Color3f verticalPixelB[50];
	
	Color3f fromColor = g_palette->cell[upperLastRow];
	Color3f toColor = g_palette->cell[bottomRow];
	sweepArray(fromColor, toColor, verticalPixelA);

	for(int i = 0; i < 50; i++){
		int intColor = verticalPixelA[i].colorToHex();
		plotPixel2(intColor, selBlendPosX - 1, selBlendPosY + i, dstWidth, dstHeight, dstPix);
		plotPixel2(intColor, selBlendPosX - 2, selBlendPosY + i, dstWidth, dstHeight, dstPix);
	}

	for(int k = 0, g = 1; ; k++, g++){
		fromColor = g_palette->cell[upperLastRow + g];
		toColor = g_palette->cell[bottomRow + g];
		sweepArray(fromColor, toColor, verticalPixelB);
		int x = selBlendPosX + k * 25;
		for(int r = 0; r < 50; r++){
			plotSweepRow(verticalPixelA[r], verticalPixelB[r], x, selBlendPosY + r, dstWidth, dstHeight, dstPix);
		}
		if(k == 13){
			break;
		}
		for(int i = 50; --i >= 0;){
			verticalPixelA[i] = verticalPixelB[i];
		}
	}
	for(int i = 0; i < 50; i++){
		int intColor = verticalPixelB[i].colorToHex();
		plotPixel2(intColor, selBlendPosX + 7 * 50, selBlendPosY + i, dstWidth, dstHeight, dstPix);
		plotPixel2(intColor, selBlendPosX + 7 * 50 + 1, selBlendPosY + i, dstWidth, dstHeight, dstPix);
	}
	//for(int k = 0; k < 8; k++){
	//	int x = selBlendPosX + k * 50;
	//	paintBlock2(g_palette->cell[upperLastRow + k * 2].colorToHex(), x - 2, selBlendPosY - 5, x + 2, selBlendPosY, dstWidth, dstHeight, dstPix);
	//	paintBlock2(g_palette->cell[bottomRow + k * 2].colorToHex(), x - 2, selBlendPosY + 50, x + 2, selBlendPosY + 50 + 5, dstWidth, dstHeight, dstPix);
	//}

	drawHorizontalLine2(0, selColorTabYpos - 2, colorTabsBeginX - 2, colorTabsBeginX + selColorTabCX + 2, dstWidth, dstHeight, dstPix);
	drawHorizontalLine2(0, selColorTabYpos + selColorTabCY + 1, colorTabsBeginX - 2, colorTabsBeginX + selColorTabCX + 2, dstWidth, dstHeight, dstPix);
	drawVerticleLine2(0, colorTabsBeginX - 2, selColorTabYpos - 1, selColorTabYpos + selColorTabCY + 1, dstWidth, dstHeight, dstPix);
	drawVerticleLine2(0, colorTabsBeginX + selColorTabCX + 1, selColorTabYpos - 1, selColorTabYpos + selColorTabCY + 1, dstWidth, dstHeight, dstPix);
}

inline void PaletteCP::paintOwnership(Art* part, String& str, int dstWidth, int dstHeight, int* dstPix, const M_Properties& properties){

	int i = 0;
	int j = 0;
	int currentGroup=-1;

	str = "Ownership ";
	str += part->name;
	str += " group ";
	str += selTab.x;

	drawText2(str, 0x0, Vec2i(8, 16), dstWidth, dstHeight, dstPix);

	const CommonSpec* commonSpec = getCommonSpec(part);

	const M_Int& coordGroup = part->coordGroup;
	int colorX = colorTabsBeginX;
	for(int cg = 0; cg < coordGroup.length; cg++){
		int intColor = 0x808080;
		if(commonSpec && cg < commonSpec->color.length){
			intColor = commonSpec->color[cg].colorToHex();
		}
		paintBlock2(intColor, colorX, colorTabsYpos, colorX + colorTabsCX, colorTabsYpos + colorTabsCY, dstWidth, dstHeight, dstPix);
		colorX += colorTabsCX + colorTabsXSpace;
	}
	int intColor = 0x808080;
	if(commonSpec && commonSpec->color.length){
		intColor = commonSpec->color[commonSpec->color.length-1].colorToHex();
	}
	paintBlock2(intColor, colorX, colorTabsYpos, colorX + colorTabsCX, colorTabsYpos + colorTabsCY, dstWidth, dstHeight, dstPix);

}

void PaletteCP::paint(){

	int dstWidth=size.x;
	int dstHeight=size.y;
	int* dstPix=pixBuf.getArray();
	const M_Properties& properties = node->getProperties();

	blitHorizontalColorSweep(Vec2i(10, 0), Vec2i(470, 0), Vec2i(471, 0), Vec2i(dstWidth-9-10, title_bar_height));
	pasteSprite2( 0, 0, 461, 0, dstWidth, dstHeight, iconSrcBufCX, iconSrcBufCY, 10, 25, dstPix);
	pasteSprite2( dstWidth-9, 0, 471, 0, dstWidth, dstHeight, iconSrcBufCX, iconSrcBufCY, 9, 25, dstPix);
	
	paintBlock2(0xffffff, 8, title_bar_height, dstWidth-9, size.y-10, dstWidth, dstHeight, dstPix);
	paintSideEdge(&iconSrcBuf[(iconSrcBufCY-1-10)*480+444], 8, 0, title_bar_height, dstHeight-10, dstWidth, dstHeight, dstPix);
	paintSideEdge(&iconSrcBuf[(iconSrcBufCY-1-10)*480+452], 9, dstWidth-9, title_bar_height, dstHeight-10, dstWidth, dstHeight, dstPix);

	blitHorizontalColorSweep(Vec2i(8, dstHeight-10), Vec2i(451, 10), Vec2i(452, 10), Vec2i(dstWidth-8-9, 10));
	pasteSprite2( 0, dstHeight-10, 444, 10, dstWidth, dstHeight, iconSrcBufCX, iconSrcBufCY, 8, 10, dstPix);
	pasteSprite2( dstWidth-9, dstHeight-10, 452, 10, dstWidth, dstHeight, iconSrcBufCX, iconSrcBufCY, 9, 10, dstPix);

	static String str;

	if(g_colorOwnership){
		if(Art* part = getPart()){
			paintOwnership(part, str, dstWidth, dstHeight, dstPix, properties);
		}
	}
	else{
		paintColor(str, dstWidth, dstHeight, dstPix, properties);
	}

	int colorX = colorTabsBeginX + (colorTabsCX + colorTabsXSpace) * selTab.x;
	int colorY = (colorTabsCY + colorTabsXSpace) * selTab.y;
	if(selCustom){
		colorY = customColorTabsYpos;
	}
	else{
		colorY += colorTabsYpos;
	}

	drawHorizontalLine2(0, colorY - 2, colorX - 2, colorX + colorTabsCX + 2, dstWidth, dstHeight, dstPix);
	drawHorizontalLine2(0, colorY + colorTabsCY + 1, colorX - 2, colorX + colorTabsCX + 2, dstWidth, dstHeight, dstPix);
	drawVerticleLine2(0, colorX - 2, colorY - 1, colorY + colorTabsCY + 1, dstWidth, dstHeight, dstPix);
	drawVerticleLine2(0, colorX + colorTabsCX + 1, colorY - 1, colorY + colorTabsCY + 1, dstWidth, dstHeight, dstPix);

	exitWindow.dstPos=Vec2i(size.x-22, 4);
	exitWindow.paint();

	for(int i=buttonList.length; --i>=0;){
		FieldButton* button = buttonList[i];

		if(button->sprop.doNotDisplayName() == false && button->sprop.push() == false && button->sprop.link() == false){
			button->paintFieldName();
		}
		
		button->paint();
		Vec2i valPos=button->valPos;
		valPos.y+=16;
		if(button->sprop.doNotDisplayValue()==false){
			if(button->editMode){
				drawText2(textCell.cellStr, textCell.firstVisChar, textCell.cellStr.length, 0, valPos, size.x, size.y, dstPix);
				textCell.highLightBlock();
			}
			else{
				button->toString(Button::str="");
				drawText2(Button::str, 0, valPos, size.x, size.y, dstPix);
			}
		}
	}

	//if(g_settings->invertWindowColor){
	//	invert();
	//}
	loadSurfaceFromMemory();
	if(textCell.cellActive==true){
		textCell.doNoThinkWink();
	}
}

void PaletteCP::clickColorTabs(const Vec2i ptCursor){
	int x = ptCursor.x - colorTabsBeginX + 2;
	int y = ptCursor.y - colorTabsYpos + 2;
	if(x > 0 && y > 0 && x < (colorTabsCX + colorTabsXSpace) * tabColums && y < (colorTabsCY + colorTabsXSpace) * (tabRows - 1)){
		selCustom = false;
		selTab = Vec2i(
			x / (colorTabsCX + colorTabsXSpace),
			y / (colorTabsCY + colorTabsXSpace)
		);
		if(g_palette->define){
			g_palette->cell.refValue()[selTab.y * tabColums + selTab.x] = g_palette->color;
			g_palette->define.refValue() = false;
		}
		else{
			g_palette->color.refValue() = g_palette->cell[selTab.y * tabColums + selTab.x];
		}

		textCell.cellActive = false;
//		setNodeColor(g_palette->color);
		paint();
		return;
	}
	
	y = ptCursor.y - customColorTabsYpos + 2;
	if(x > 0 && y > 0 && x < (colorTabsCX + colorTabsXSpace) * tabColums && y < (colorTabsCY + colorTabsXSpace)){
		selCustom = true;
		selTab.x = x / (colorTabsCX + colorTabsXSpace);

		if(g_palette->define){
			g_palette->cell.refValue()[bottomRow + selTab.x] = g_palette->color;
			g_palette->define.refValue() = false;
		}
		else{
			g_palette->color.refValue() = g_palette->cell[bottomRow + selTab.x];
		}
		textCell.cellActive = false;
//		setNodeColor(g_palette->color);
		return;
	}

	x = ptCursor.x - selBlendPosX - 2;
	y = ptCursor.y - selBlendPosY;
	if(x > 0 && y > 0 && x < 50 * 7 + 2 && y < 50){
		int intColor = pixBuf[ptCursor.y * size.x + ptCursor.x];
		Color3f color;
		color.b = (intColor & 0xff) / 255.0f;
		color.g = ((intColor & 0xff00) >> 8) / 255.0f;
		color.r = ((intColor & 0xff0000) >> 16) / 255.0f;
		g_palette->color.refValue() = color;
		textCell.cellActive = false;
//		setNodeColor(g_palette->color);
		paint();
	}

}

void PaletteCP::clickOwnershipColorTabs(Art* part, const Vec2i ptCursor, int numOfTabs){
	float x = ptCursor.x - colorTabsBeginX + 2;
	float y = ptCursor.y - colorTabsYpos;
	if(x < 0 || y < 0 || x > (colorTabsCX + colorTabsXSpace) * numOfTabs || y > colorTabsCY){
		return;
	}
	selTab.x = x / (colorTabsCX + colorTabsXSpace); 
	updateOwnershipFields(part);

	paint();
}

bool PaletteCP::click(const Vec2i ptCursor){
	if(ptCursor >= pos && ptCursor < pos+size){
		activeWindow=this;
		if(tugBorder){
			tugBorder->window = this;
			g_drag = tugBorder;
			return true;
		}

		const Vec2i localCursor = ptCursor-pos;

		{
			int newCaretPos = 0;
			for(int k = buttonList.length; --k >= 0;){
				FieldButton* button=buttonList[k];
				if(button->checkClickTextCell(localCursor, newCaretPos)){
					if(button->editMode == false){
						textCell.cellActive = false;
						textCell.selectField(*button);
					}
					textCell.highlightPos = textCell.caretPos = newCaretPos + textCell.firstVisChar;
					textCell.updateCaret();

					textCell.checkForDoubleLeftClick();

					g_drag = &textCell;
					paint();
					return true;
				}
			}
		}//above is same as ControlPanel::click except for the OpenNetFile & TextureMapper conditions
		if(localCursor.y < title_bar_height){
			g_drag = this;
			dragPos = ptCursor;
			return true;
		}
		if(g_colorOwnership){
			if(Art* part = getPart()){
				clickOwnershipColorTabs(part, localCursor, part->numOfGroups());
			}
		}
		else{
			clickColorTabs(localCursor);
		}
		return true;
	}
	else{
		return false;
	}
}

void doAutoWeightDivide(Art* part){
	if(int numOfGroups = part->numOfGroups()){
		float w = 1.0f / numOfGroups;
		M_Float& weight = part->weight;
		for(int i = 0; i < weight.length; i++){
			weight[i] = w * (i + 1);
		}
	}
}

void Palette::Insert::operator=(const bool& par){
	PaletteCP* paletteCP = paletteCPAllreadyOpen();
	if(Art* part = getPart()){
		float weight = 1 / part->numOfGroups() * paletteCP->selTab.x;
		part->coordGroup.refValue().insertElement(paletteCP->selTab.x, 0);
		part->weight.refValue().insertElement(paletteCP->selTab.x, weight);
		syncCommonSpecLength(part);

		if(g_palette->autoWeightDivide){
			doAutoWeightDivide(part);
		}
		paletteCP->updateOwnershipFields(part);
		part->smls->ctrf = true;
	}
}

void Palette::Delete::operator=(const bool& par){
	PaletteCP* paletteCP = paletteCPAllreadyOpen();
	if(Art* part = getPart()){
		if(part->coordGroup.length){
			float weight = 1 / part->numOfGroups() * paletteCP->selTab.x;
			part->coordGroup.refValue().deleteElement(paletteCP->selTab.x);
			part->weight.refValue().deleteElement(paletteCP->selTab.x);
			if(paletteCP->selTab.x > 0){
				paletteCP->selTab.x--;
			}
			if(g_palette->autoWeightDivide){
				doAutoWeightDivide(part);
			}
			paletteCP->updateOwnershipFields(part);
			part->smls->ctrf = true;
		}
	}
}

void Palette::Coords::operator=(const int& par){
	PaletteCP* paletteCP = paletteCPAllreadyOpen();
	if(Art* part = getPart()){
		if(paletteCP->selTab.x < part->weight.length){
			part->coordGroup.refValue()[paletteCP->selTab.x] = par;
			refValue() = par;
			part->smls->ctrf = true;
		}
	}
}

void Palette::Weight::operator=(const float& par){
	PaletteCP* paletteCP = paletteCPAllreadyOpen();
	if(Art* part = getPart()){
		if(paletteCP->selTab.x < part->weight.length){
			part->weight.refValue()[paletteCP->selTab.x] = par;
			refValue() = par;
			part->smls->ctrf = true;
		}
	}
}

void Palette::Color::operator=(const Color3f& par){
	PaletteCP* paletteCP = paletteCPAllreadyOpen();
	if(g_colorOwnership){
		if(Art* part = getPart()){
			if(part->name == ""){
				return;
			}
			CommonSpec* commonSpec = getCommonSpec(part);

			if(commonSpec == 0){
				commonSpec = new CommonSpec;
				commonSpec->name = part->name;
				commonSpec->color.refValue() += Color3f();
				g_settings->commonSpec.refValue() += commonSpec;
			}

			if(commonSpec){
				if(paletteCP->selTab.x < commonSpec->color.length){
					commonSpec->color.refValue()[paletteCP->selTab.x] = par;
					refValue() = par;
					part->smls->ctrf = true;
				}
				else{
					commonSpec->color.refValue().length = paletteCP->selTab.x + 1;
					commonSpec->color.refValue()[paletteCP->selTab.x] = par;
				}
			}
		}
	}
	else{
		refValue() = par;
//		if(paletteCP){
//			setNodeColor(par);
//		}
	}
	if(paletteCP){
		paletteCP->paint();
	}
}

bool ClickColorCell::doEvent(){
	static int phase2=0;
	if(phase==0){
		if(PaletteCP* paletteCP = paletteCPAllreadyOpen()){
			Vec2i pos = Vec2i(PaletteCP::colorTabsBeginX + cellIndex.x * (PaletteCP::colorTabsCX + PaletteCP::colorTabsXSpace), PaletteCP::colorTabsYpos + cellIndex.y * (PaletteCP::colorTabsCY + PaletteCP::colorTabsXSpace));
			pos += Vec2i(PaletteCP::colorTabsCX / 2, PaletteCP::colorTabsCY / 2) + paletteCP->pos;
			if(localMove(pos)>=0){
				phase++;
				beginRobotTime = g_time;
				phase2=0;
			}
		}
	}
	else{
		return doSingleClick(phase2);
	}
	return false;
}

extern const char* commonSpecStr;

void Settings::LoadColorCodePresets::operator=(const bool& par){
	PaletteCP* paletteCP = paletteCPAllreadyOpen();

	g_palette = dynamic_cast<Palette*>(g_settings->palette.p);
	g_settings->commonSpec.refValue().length = 0;
	createSmlsFromString(g_settings->commonSpec.refValue(), g_loadStr = commonSpecStr);
	if(paletteCP){
		if(Art* part = dynamic_cast<Art*>(g_selNode.getRaw())){
			clickNodeForOwnershipWindow();

			//syncCommonSpecLength(part);
			//paletteCP->updateOwnershipFields(part);
			//part->smls->ctrf = true;
		}
	}
}

void clearOwnershipPatch(){
	if(NurbsPatch* patch = dynamic_cast<NurbsPatch*>(g_selNode.getRaw())){
		if(Seamless* smls = patch->smls){
			smls->removeOwnershipPatch(patch);

			BackUp::monostable.beginCycle();
		}
	}
}

void deleteControlPointFromAllMorphers(Art* part, int i);

void clearOwnership(){
	if(Art* part=dynamic_cast<Art*>(g_selNode.getRaw())){
		if(Anim* anim = Window::animMode()){
		M_ControlPoint& controlPoint = part->controlPoint;
			for(int i = controlPoint.length; --i>=0;){
				deleteControlPointFromAllMorphers(part, i);
				controlPoint.deleteElement(i);
			}
//			part->controlPoint.length = 0;

			if(Art* mirrorPart = part->mirrorPart){
				M_ControlPoint& controlPoint = mirrorPart->controlPoint;
				for(int i = controlPoint.length; --i>=0;){
					deleteControlPointFromAllMorphers(mirrorPart, i);
					controlPoint.deleteElement(i);
				}
//				part->mirrorPart->controlPoint.length = 0;
			}
			if(PaletteCP* paletteCP = paletteCPAllreadyOpen()){
				paletteCP->updateOwnershipFields(part);
				paletteCP->paint();
			}
			part->houseKeep();
			if(Anim* anim = Window::animMode()){
				if(anim->play == false){
					anim->update = true;
				}
			}
			BackUp::monostable.beginCycle();

		}
	}
	else{
		needPartMsg();
	}
}

void removeAllPolyMidPointsOwnership(){
	if(Art* part=dynamic_cast<Art*>(g_selNode.getRaw())){
		part->smls->removeAllPolyMidPointsOwnership();
		if(PaletteCP* paletteCP = paletteCPAllreadyOpen()){
			paletteCP->updateOwnershipFields(part);
			paletteCP->paint();
		}
		part->houseKeep();
		if(Anim* anim = Window::animMode()){
			if(anim->play == false){
				anim->update = true;
			}
		}
		BackUp::monostable.beginCycle();
	}
	else{
		needPartMsg();
	}
}
