/* Seamless3d

Copyright (C) 2001-2010 Graham John Perrett (AKA thyme) thyme@seamless3d.com www.seamless3d.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this
software and associated documentation files (the "Software"), to deal in the Software
without restriction, including without limitation the rights to use, copy, modify,
merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to the following
conditions:

The above copyright notice and this permission notice shall be included in all copies or
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT
OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.
*/


#include <windows.h>

#include "seamless3d.h"
#include "techuelife_triangle_editor.h"
#include "gui.h"
#include "scene_tree.h"
#include "robot_aid.h"

extern const char* title_scene_tree;
extern const char* menu_short_copy;
extern Ptr<AnimClue> g_vcArtAnimBase;

SelNode g_selNode;
CopyNode g_copyNode;
int g_rememberIndex = 0;
bool checkPoint(const bool perspective, Art*& part, int i, Matrix &partMatrix);
int getChildrenLen(int index);
void countUpViewpoints();
void commonDoDrag(const Vec2i ptCursor, Vec2i& pos, Vec2i& dragPos);

extern int g_viewIndex;
extern M_<Ptr<Viewpoint> > g_viewpoints;
bool isDoubleLeftClick();
void updateAutoInForClick();
void paintBlock2(const unsigned fromColor, const unsigned toColor, int tx, int ty, int bx, int by, int dstWidth, int dstHeight, int* dstPix);

void SceneTree::setSelNodeLine(Node* node){
	if(g_sceneTreeList[selLine].node==node){
		return;
	}
	for(selLine=0; selLine<g_sceneTreeList.length;selLine++){
		if(g_sceneTreeList[selLine].node==node){

			int collapsedLine;
			bool visible = getCollapsedLine(selLine, collapsedLine);
			createActiveLabel(collapsedLine, visible);
			break;
		}
	}
}
void SceneTree::incEditLine(HWND hWnd, int incStep){
	textCell.cellActive=false;

	int collapsedLine;
	if(selLine+incStep<g_sceneTreeList.length && selLine+incStep>0){
		bool visible = getCollapsedLine(selLine, collapsedLine);
		collapsedLine+=incStep;
		selLine=getLineForNonCollapsed(collapsedLine);
		lastSelNodeLine=selLine;
		paint();

		makePointerEqualIndexLocation();
		setTitleName();
		createActiveLabel(collapsedLine, visible);
	}
}
void SceneTree::createActiveLabel(int collapsedLine, bool visible){
	removeTypeFromList<CaretWindow>(g_callBackList);
	g_callBackList+=&textCell;

	g_selNode->getSceneDisplayString(textCell.cellStr="");//****** why can't we use local member "selNode" instead?
	lastSelNodeLine=selLine;
	if(visible){
		textCell.cellActive = true;
		textCell.updateCaret();
	}
	else{
		textCell.cellActive = false;
	}
}
void SceneTree::newNodePopupMenu (Vec2i ptCursor){
	const Vec2i localCursor = ptCursor-pos;

	g_remMenu=new Menu;
	M_<MenuItemPtr>& list=g_remMenu->list;
	g_remMenu->list.length=0;

	//closeNodeNameEditBox();// needed?

	int tmpInt=(localCursor.y-size.y-beginYOffset)/pixSymbolCY+g_scene->scroll;

	if(tmpInt>g_sceneTreeCollapsed.length)return;
	int insertOrder=1;

	if(selLine!=0){
		list+=new MenuItem_(Vec2i(16,16), Vec2i(remember_x, remember_y), menu_short_copy, remember, *g_remMenu);
	}


	buildRoutePopupMenu(*g_remMenu);

	g_remMenu->autoSizeAndStuff();
	//menu.buildMenu();

	//find the width

	g_remMenu->pos=ptCursor;
}
bool SceneTree::STTextCell::charInput(unsigned wParam){
	if(	dynamic_cast<ProtoInterface*>(g_selNode.getRaw())||
		dynamic_cast<ProtoBody*>(g_selNode.getRaw())||
		dynamic_cast<Scene*>(g_selNode.getRaw())){
	}
	else if(wParam==VK_ESCAPE){
		InputEvent::resumeSequence();
		cellStr=g_selNode->name;
		if(caretPos>cellStr.length){
			highlightPos=caretPos=cellStr.length;
		}
		updateCaret();
		win.paint();
	}
	else if(wParam==VK_RETURN){
		//if(Window::activeWindow != g_msg){
		if(cellActive){
			if(dynamic_cast<CodeLine*>(g_selNode.getRaw())){
				cellReadyToEnter = true;
			}
			cellActive = false;
			doNoThinkWinkOff();
		}
	}
	else if(wParam==VK_BACK){
		if(deleteHighlight()){
			win.paint();
		}
		else if(caretPos>0){
			decCaretPosNH();
			cellStr.deleteChar(caretPos);
			win.paint();
		}
	}
	else if(wParam<' ')return true;
	else {
		deleteHighlight();
		cellStr.insert(wParam, caretPos);
		incCaretPosNH();
		win.paint();
	}
	return true;
}
extern bool tugFlag;

/*
void SceneTree::senseAllPartsVertex( SeamlessPtr smls){
	const bool perspective=g_settings->perspective;
	NodePtr before = g_selNode.getRaw();
	int k;
	M_Art &partList=smls->partList;
	for(k=0;k<partList.length;k++){

		Art* part=partList[k];
		M_Triangle& triangle = part->triangle;

		if(part->show){
			for(int i=triangle.getTriangleLength(); --i>=0; ){
				Art* part = triangle.refPart(i);
				if (part->checkPointInVertex(perspective, triangle.refIndex(i), part->matrix * g_viewMatrix)){
					if(MultiSelect::coordList.length == 0 && Window::animMode()){
						part->index[0] = triangle.refIndex(i);
						MultiSelect::coordList += S_VertexIndex(part, part->index[0]);
						smls->ActCtrlPnt = part->index[0];
						MultiSelect::checkToReset(part);
						MultiSelect::backUpCoords(part);
						MultiSelect::calcPivotPoint();
					//	g_visualClue = g_vcArtAnimBase;
					//	SenseTugGlobals::node = g_vcArtAnimBase;
					//	tugFlag = true;
					}
					return;
				}
			}
		}
	}
/*


	const bool perspective=g_settings->perspective;
	NodePtr before = g_selNode.getRaw();
	int k;
	M_Art &partList=smls->partList;
	for(k=0;k<partList.length;k++){
		Art* part=partList[k];
		if(part->show){
			Matrix partMatrix = part->matrix * g_viewMatrix;
			for(int i=part->coord.refValue().length; --i>=0; ){
				if (part->checkPointInVertex(perspective, i, partMatrix)){
					return true;
				}
			}
		}
	}
	return false;
	*/
//}

void SceneTree::senseAllPartsTriangle(SeamlessPtr smls){
	const bool perspective=g_settings->perspective;
	NodePtr before = g_selNode.getRaw();
	int k;
	M_Art &partList=smls->partList;
	for(k=0;k<partList.length;k++){

		Art* part=partList[k];
		M_Triangle& triangle = part->triangle;

		if(part->show){
			for(int i=triangle.getTriangleLength(); (i-=3)>=0; ){
				part->checkPointInTriangle(perspective, i);
			}
		}
	}
}
void SceneTree::senseAllPartsTriangle(SeamlessPtr smls, TouchSensor* sensor){
	const bool perspective=g_settings->perspective;
	NodePtr before = g_selNode.getRaw();
	int k;
	M_Art &partList=smls->partList;
	for(k=0;k<partList.length;k++){

		Art* part=partList[k];
		M_Triangle& triangle = part->triangle;

		if(part->show){
			for(int i=triangle.getTriangleLength(); (i-=3)>=0; ){
				part->checkPointInTriangle(perspective, i, sensor);
			}
		}
	}
}
void SceneTree::paintBlueDot(int indent, int selLine){
	int paintLine=selLine-g_scene->scroll;
	pasteSprite2( tLnOffst+indent*pixSymbolCX-4, beginYOffset+6+paintLine*pixSymbolCY, 192, 103, size.x, size.y, iconSrcBufCX, iconSrcBufCY, 4, 4, pixBuf.getArray());
}
int RingBuild::getPartOffset(){
	int i = index[1];
	if(part.length == 0 || i >= part.length){
		return -1;
	}
	if(g_colorOwnership){
		part[0]->smls->ctrf=true;
	}
	return i;
}
int TransferOwnership::getPartOffset(){
	int i = index[0];
	if(dest.length == 0 || i >= dest.length){
		return -1;
	}
	if(source.p){
		i++;
	}
	if(g_colorOwnership){
		dest[0]->smls->ctrf=true;
	}
	return i;
}

void SceneTree::paint(){

	int indent=0;//,selLine=0;

	int dstWidth=size.x;
	int dstHeight=size.y;
	int* dstPix=pixBuf.getArray();

	int* pEnd=&pixBuf[0]+dstWidth*dstHeight;

	int nodeMax=(dstHeight-beginYOffset)/pixSymbolCY+1;
	int nodeCnt=nodeMax-1;
	int lastNode=g_scene->scroll+nodeCnt;
	if(g_sceneTreeCollapsed.length<=lastNode){
		nodeCnt=g_sceneTreeCollapsed.length-g_scene->scroll;
		lastNode=g_scene->scroll+nodeCnt-1;
	}
	if(nodeCnt<1)return;
	int collapsedLine;
	bool visible=getCollapsedLine(selLine, collapsedLine);

	paintNodeBackground(size.x, size.y, nodeCnt, lastNode, pEnd, collapsedLine);

	paintSideEdge(&iconSrcBuf[(iconSrcBufCY-1-10)*480+444], 8, 0, title_bar_height, dstHeight-10, dstWidth, dstHeight, dstPix);
	paintSideEdge(&iconSrcBuf[(iconSrcBufCY-1-10)*480+452], 9, dstWidth-9, title_bar_height, dstHeight-10, dstWidth, dstHeight, dstPix);

	int y = 0;
	int x = 0;
	if(visible){
		//y = textCell.cellPos.y;
		//x = tLnOffst + g_sceneTreeCollapsed[collapsedLine].indent*pixSymbolCX + pixSymbolCX;
		//paintBlock2(iconSrcBuf[(iconSrcBufCY-1)*480+451], iconSrcBuf[(iconSrcBufCY-1)*480+452], x, y, size.x-9, y + 16, size.x, size.y, dstPix);
		//paintSideEdge(&iconSrcBuf[(iconSrcBufCY-1)*480+452], 9, size.x-9, y, y + 16, size.x, size.y, dstPix);

		y = beginYOffset+(collapsedLine-g_scene->scroll)*pixSymbolCY;
		x = tLnOffst + g_sceneTreeCollapsed[collapsedLine].indent*pixSymbolCX + pixSymbolCX;

		paintColorSweepForCurvedBar(&iconSrcBuf[(iconSrcBufCY-1-25-24)*480+470], &iconSrcBuf[(iconSrcBufCY-1-25-24)*480+471], x+10, y, Vec2i(dstWidth - 9, 16), dstWidth, dstHeight, dstPix);
		pasteSprite2( x, y, 461, 25+24, dstWidth, dstHeight, iconSrcBufCX, iconSrcBufCY, 10, 16, dstPix);
		pasteSprite2( dstWidth-9, y, 471, 25+24, dstWidth, dstHeight, iconSrcBufCX, iconSrcBufCY, 9, 16, dstPix);
	}
	{
		int sl=0;
		g_scene->paintSceneTree(indent, sl, collapsedLine, visible, Window::animMode() != 0, 0);
	}

	blitHorizontalColorSweep(Vec2i(10, 0), Vec2i(470, 0), Vec2i(471, 0), Vec2i(dstWidth-9-10, title_bar_height));
	pasteSprite2( 0, 0, 461, 0, dstWidth, dstHeight, iconSrcBufCX, iconSrcBufCY, 10, title_bar_height, dstPix);
	pasteSprite2( dstWidth-9, 0, 471, 0, dstWidth, dstHeight, iconSrcBufCX, iconSrcBufCY, 9, title_bar_height, dstPix);

	pasteSprite2( 5, 1, openSceneTree_x, openSceneTree_y, dstWidth, dstHeight, iconSrcBufCX, iconSrcBufCY, 24, 24, dstPix);
	drawText2(title_scene_tree, 0x0, Vec2i(21+2+24-16, 16), dstWidth, dstHeight, dstPix);

	blitHorizontalColorSweep(Vec2i(8, dstHeight-10), Vec2i(451, 10), Vec2i(452, 10), Vec2i(dstWidth-8-9, 10));
	pasteSprite2( 0, dstHeight-10, 444, 10, dstWidth, dstHeight, iconSrcBufCX, iconSrcBufCY, 8, 10, dstPix);
	pasteSprite2( dstWidth-9, dstHeight-10, 452, 10, dstWidth, dstHeight, iconSrcBufCX, iconSrcBufCY, 9, 10, dstPix);

	
	if(MarkPart* markPart = dynamic_cast<MarkPart*>(g_selNode.getRaw())){
		if(visible && g_selNode->collapsed == false){
			int po = markPart->getPartOffset();
			if(po>=0){
				pasteSprite2(x-19, y + po * 16 + 16, 240, 176, dstWidth, dstHeight, iconSrcBufCX, iconSrcBufCY, 16, 16, dstPix);
			}
		}
	}

	newNodeBar.paint(dstWidth, dstHeight, iconSrcBufCX, iconSrcBufCY, dstPix);

	if(g_drag==&newNodeBar.item){
		newNodeBar.paintDragClues(dstWidth, dstHeight, iconSrcBufCX, iconSrcBufCY, dstPix);
	}
	topScroll.paint();
	bottomScroll.paint();
	pageUp.paint();
	slider.paint();
	exitWindow.dstPos=Vec2i(size.x-22, 4);
	exitWindow.paint();

	if(g_settings->invertWindowColor){
		invert();
	}
	loadSurfaceFromMemory();
	if(textCell.cellActive==true){
		textCell.doNoThinkWink();
	}
}
void SceneTree::buttonHover(Vec2i ptCursor){
	if(topScroll.initHover(ptCursor)){
		return;
	}
	if(bottomScroll.initHover(ptCursor)){
		return;
	}
	if(pageUp.initHover(ptCursor)){
		return;
	}
	if(pageDown.initHover(ptCursor)){
		return;
	}
	if(slider.initHover(ptCursor)){
		return;
	}
	if(exitWindow.initHover(ptCursor)){
		return;
	}
	if(newNodeBar.initHover(ptCursor)){
		return;
	}
	else{
		int a=0;
	}
}
void SceneTree::STTextCell::doDrag(const Vec2i ptCursor){
	const Vec2i localCursor=ptCursor-win.pos;
	if(localCursor.y < cellPos.y || localCursor.y > cellPos.y+16){
		win.newNodeBar.dragNodeFromSceneTree(*(g_selNode.getRaw()));
	}
	else{
		CaretWindow::doDrag(ptCursor);
	}
}
bool SceneTree::rightClick(const Vec2i ptCursor){
	if(ptCursor >= pos && ptCursor < pos+size){
		Window::activeWindow=this;

		Vec2i localCursor=ptCursor-pos;


		int tmpInt=(localCursor.y-SceneTree::beginYOffset)/pixSymbolCY+g_scene->scroll;

		if(tmpInt>=g_sceneTreeCollapsed.length)return true;
		Node* node=g_sceneTreeCollapsed[tmpInt].node;

		{
			int textIndentPix = tLnOffst+g_sceneTreeCollapsed[tmpInt].indent*pixSymbolCX+pixSymbolCX;
			if(localCursor.x<=textIndentPix){
				localCursor.x=textIndentPix;
			}
		}
		int newCaretPos=(localCursor.x+charWidth/2-(tLnOffst+g_sceneTreeCollapsed[tmpInt].indent*pixSymbolCX+16+3))/charWidth;
		int newSelNodeLine=getLineForNonCollapsed(tmpInt);

		if(newSelNodeLine != selLine){
			int collapsedLine;
			getCollapsedLine(selLine, collapsedLine);
			if(g_selCount==1 || tmpInt<collapsedLine || tmpInt>=collapsedLine+g_selCount.all){
				g_selCount.reset();
				selectNode(newSelNodeLine, node, newCaretPos, tmpInt);
				g_drag=0;
			}
			g_tb->updateToolBarContext();
		}

		newNodePopupMenu(ptCursor);

		return true;
	}
	else{
		return false;
	}
}
void recurseShow(Art* part, bool show){
	part->show=show;
	M_SArt m_art=part->part.refValue();
	for(int k=m_art.length; --k>=0;){
		recurseShow(m_art[k], show);
	}
}
void SceneTree::selectNode(int newSelNodeLine, Node* node, int newCaretPos, int tmpInt){
	String tmpStr; //do we really need this as in why dont we just write to the node here or work out the caretPos later?
	if(newSelNodeLine!=selLine){
		if(dynamic_cast<CodeLine*>(g_selNode.getRaw())){
			textCell.cellReadyToEnter = true;
		}
		textCell.cellActive=false;
		lastSelNodeLine=selLine=newSelNodeLine;
		if(Art* part = dynamic_cast<Art*>(g_selNode.getRaw())){
			part->smls->actTriPart=part;
			makePointerEqualIndexLocation();
		}
		textCell.caretPos=newCaretPos;
		g_selNode->getSceneDisplayString(tmpStr="");
		if(textCell.caretPos>tmpStr.length){
			textCell.caretPos=tmpStr.length;
		}
		createActiveLabel(tmpInt, true);
		setTitleName();

		updateAutoInForClick();

		if(dynamic_cast<Video*>(node)){
			if(g_vcp == 0){
				g_vcp = new VideoControlPanel;
			}
			g_vcp->setupNodeWindow(node);
		}
		else if(dynamic_cast<WaveToPoly*>(node)){
			if(g_wtpcp == 0){
				g_wtpcp = new WaveToPolyControlPanel;
			}
			g_wtpcp->setupNodeWindow(node);
		}
		else if(dynamic_cast<Anim*>(node)){
			g_acp->setupNodeWindow(node);
		}
		else{
			cp->setupNodeWindow(node);
		}
	}
	else{
		if(textCell.cellActive==false){
			createActiveLabel(tmpInt, true);
		}
		textCell.caretPos = textCell.firstVisChar + newCaretPos;

		if(textCell.caretPos>textCell.cellStr.length){
			textCell.caretPos=textCell.cellStr.length;
		}
//		if(dynamic_cast<Anim*>(node)){
//		}
//		else{
//			if(cp->node!=node){
//				cp->setupNodeWindow(node);
//			}
//		}
	}
	if(!g_drag){
		g_drag=&textCell;
	}
	textCell.highlightPos=textCell.caretPos;
	textCell.updateCaret();
	paint();
}

void clickNodeForOwnershipWindow();

bool SceneTree::click(const Vec2i ptCursor){

	if(ptCursor >= pos && ptCursor < pos+size){
		Window::activeWindow=this;

		if(tugBorder){
			tugBorder->window=this;
			g_drag=tugBorder;
			return true;
		}

		if(Button* button=Button::getHoverButton()){
			button->action();
			return true;
		}

		Vec2i localCursor=ptCursor-pos;
		if(localCursor.y < title_bar_height){
			textCell.cellActive=false;
			g_drag=this;
			dragPos=ptCursor;
			return true;
		}

		g_selCount.reset();
		int tmpInt=(localCursor.y-beginYOffset)/pixSymbolCY+g_scene->scroll;

		if(tmpInt>=g_sceneTreeCollapsed.length)return true;
		Node* node=g_sceneTreeCollapsed[tmpInt].node;
		int newCaretPos=(localCursor.x+charWidth/2-(tLnOffst+g_sceneTreeCollapsed[tmpInt].indent*pixSymbolCX+16+3))/charWidth;
		int indentPix = tLnOffst+g_sceneTreeCollapsed[tmpInt].indent*pixSymbolCX;


		if(node->getType()!=SCENE && localCursor.x<=indentPix){

			if(node->collapsed || getChildrenLen(tmpInt)){
				node->collapsed=!node->collapsed;
				g_stw->genNodeList();
				paint();
#ifdef PROG_ROB
				int newSelNodeLine=getLineForNonCollapsed(tmpInt);
				g_upStr="new Collapse(";
				nameOrLineNumber(newSelNodeLine, g_upStr);
				g_upStr+="),";
#endif

			}
			return true;
		}
		else if(Art* part=dynamic_cast<Art*>(node)){
			if(0x8000&GetKeyState(VK_LCONTROL) || 0x8000&GetKeyState(VK_LMENU)){
			}
			else if(localCursor.x > indentPix && localCursor.x < pixSymbolCX + indentPix-2 && isDoubleLeftClick()){
				part->show = !part->show;
				if(part->collapsed && g_sceneTreeCollapsed[tmpInt].ref == false){
					recurseShow(part, part->show);
				}
				int newSelNodeLine=getLineForNonCollapsed(tmpInt);
				selectNode(newSelNodeLine, node, newCaretPos, tmpInt);

				if(cp->isVisible()==false){
					Window::activeWindow=cp.p;
				}
#ifdef PROG_ROB
				g_upStr="new Hide(";
				nameOrLineNumber(newSelNodeLine, g_upStr);
				g_upStr+="),";
#endif

				part->smls->convertToRenderFormat();
				BackUp::monostable.beginCycle();
				paint();
				g_tb->updateToolBarContext();
				clickNodeForOwnershipWindow();
				return true;
			}
		}
		else if(NurbsPatch* patch=dynamic_cast<NurbsPatch*>(node)){
			if(localCursor.x > indentPix && localCursor.x < pixSymbolCX + indentPix-2 && isDoubleLeftClick()){
				patch->cage = !patch->cage;
				int newSelNodeLine=getLineForNonCollapsed(tmpInt);
				selectNode(newSelNodeLine, node, newCaretPos, tmpInt);

				if(cp->isVisible()==false){
					Window::activeWindow=cp.p;
				}
#ifdef PROG_ROB
				g_upStr="new Hide(";
				nameOrLineNumber(newSelNodeLine, g_upStr);
				g_upStr+="),";
#endif

				patch->houseKeep();
				BackUp::monostable.beginCycle();
				paint();
				g_tb->updateToolBarContext();
				clickNodeForOwnershipWindow();
				return true;
			}
		}

		int newSelNodeLine=getLineForNonCollapsed(tmpInt);
		selectNode(newSelNodeLine, node, newCaretPos, tmpInt);

		textCell.checkForDoubleLeftClick();
#ifdef PROG_ROB
		g_upStr="new SelectNode(";
		nameOrLineNumber(newSelNodeLine, g_upStr);
		g_upStr+="),";
#endif


		if(Viewpoint* viewpoint = dynamic_cast<Viewpoint*>(node)){
			countUpViewpoints();
			for(int k = g_viewpoints.length; -- k >= 0;){
				Viewpoint* v = g_viewpoints[k];
				if(v == viewpoint){
					g_viewIndex = k;
					break;
				}
			}
			g_scene->setSceneFromViewpoint(*viewpoint);
		}


		if(dynamic_cast<Video*>(node)){
			Window::insertUnderActive(g_vcp.p);
		}
		else if(dynamic_cast<WaveToPoly*>(node)){
			Window::insertUnderActive(g_wtpcp.p);
		}
		if(Anim* anim = dynamic_cast<Anim*>(node)){
			Window::insertUnderActive(g_acp.p);
			if(anim->play == false){
				anim->update = true;
			}
		}
		else{
			Window::insertUnderActive(cp.p);
		}

		paint();

		clickNodeForOwnershipWindow();
		g_tb->updateToolBarContext();
		return true;
	}
	else{
		return false;
	}
}
void SceneTree::release(Vec2i ptCursor){

//	if(inside==false)return;

//	if(dragMode){
//		ReleaseCapture();
//		dragMode=false;
//	}

}

void SceneTree::doDrag(Vec2i ptCursor){
	commonDoDrag(ptCursor, pos, dragPos);
}
void SceneTree::exitSceneTreeWindow(){
#ifdef PROG_ROB
	dLog("new MoveToExitSceneTree(),\nnew Click(),");
#endif
	closeWindow(g_stw.p);
}
void SceneTree::reset(){
	remPJ=0;
	copyLine=0;

	keepLeagal();
}
void SceneTree::keepLeagal(){
	if(g_scene->scroll>=g_sceneTreeCollapsed.length){
		g_scene->scroll=g_sceneTreeCollapsed.length-1;
	}
	if(g_scene->scroll < 0){//why is this ever needed if there is always at least one node the scene node?
		g_scene->scroll = 0;
	}
	if(selLine >= g_sceneTreeList.length){
		selLine=g_sceneTreeList.length-1;
	}
}
void SceneTree::setScrolPos(int offset){
	g_scene->scroll.refValue()+=offset;
	keepLeagal();
	slider.setPixelPos();
	resizePageUpPageDown();
	textCell.updateCaret();
	paint();
}
void SceneTree::selectFirstPart(){
	cp->node=g_scene;
	reset();
	for(int k=0; k<g_sceneTreeList.length; k++){
		if(Art* part=dynamic_cast<Art*>(g_sceneTreeList[k].node.p)){
			cp->node=part;
			selLine=k;
			break;
		}
	}
	setupSelectedNodeWindow();
	makePointerEqualIndexLocation();
	textCell.cellStr=g_selNode->name;
	textCell.caretPos=textCell.highlightPos=0;
	g_selCount.reset();
	g_visualClue=nullPtr;
}
void SceneTree::selectNextLine(){
	int collapsedLine; if(getCollapsedLine(selLine, collapsedLine)){
		if(collapsedLine < g_sceneTreeCollapsed.length-1){
			int x = tLnOffst+g_sceneTreeCollapsed[collapsedLine].indent*pixSymbolCX+16+3 + textCell.caretPos*charWidth;
			collapsedLine++;
			if(collapsedLine + 1 > g_scene->scroll+numOfVisLines){
				g_scene->scroll=collapsedLine-numOfVisLines + 1;
			}

			int _x = tLnOffst+g_sceneTreeCollapsed[collapsedLine].indent*pixSymbolCX+16+3;
			int y = beginYOffset+(collapsedLine-g_scene->scroll)*pixSymbolCY+12;

			if(_x > x){
				x = _x;
			}
			click(Vec2i(x, y));
//					SetScrollPos( hSceneTreeWnd, SB_VERT, g_scrollPosCollapsed, TRUE );
		}
	}
	else{
//				HideCaret(hSceneTreeWnd);
	}
}
void SceneTree::paintNodeBackground(unsigned dstWidth, unsigned dstHeight, int nodeCnt, int lastNode, int* pEnd, int collapsedLine){

	int* dstPix=pixBuf.getArray();

	int ty = beginYOffset;
	for(int i = 0; i < nodeCnt; i++){
		unsigned color=g_sceneTreeCollapsed[g_scene->scroll+i].marker.colorToHex();
		if(g_selCount > 1 && g_scene->scroll + i >= collapsedLine && g_scene->scroll + i < collapsedLine + g_selCount.all){
			color ^= 0xffffffff;
		}
		paintBlock2(color, 8, ty, dstWidth-11, ty+pixSymbolCY, dstWidth, dstHeight, dstPix);
		ty += pixSymbolCY;
	}
	paintBlock2(0xffffff, 8, ty, dstWidth-11, size.y-10, dstWidth, dstHeight, dstPix);
}
void SceneTree::onMouseWheel(const int par){
	setScrolPos(par);
}

void SceneTree::onGainFocus(){
	textCell.onWindowGainFocus();
}

Node* SelNode::operator ->()const{
	return g_sceneTreeList[g_stw->selLine].node;
}
Node* SelNode::getRaw()const{
	return g_sceneTreeList[g_stw->selLine].node;
}

Node* CopyNode::operator ->()const{
	return g_sceneTreeList[g_stw->copyLine].node;
}
Node* CopyNode::getRaw()const{
	return g_sceneTreeList[g_stw->copyLine].node;
}
void SceneTree::escape(){
	if(g_drag){
		g_drag->endDrag(Vec2i(-1000,-1000));
	}
}
void SceneTree::STTextCell::updateText(){
	cellStr=g_selNode->name;
}
