/* Seamless3d

Copyright (C) 2001-2010 Graham John Perrett (AKA thyme) thyme@seamless3d.com www.seamless3d.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this
software and associated documentation files (the "Software"), to deal in the Software
without restriction, including without limitation the rights to use, copy, modify,
merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to the following
conditions:

The above copyright notice and this permission notice shall be included in all copies or
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT
OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.
*/

#include "seamless3d.h"


void Stem::convertCoordIndexToTriangle(Art* part, int h, int c, int s){
	M_Triangle& triangle = part->triangle;
	const M_Int& rc = this->rc;
	int ia = 0;
	int ib = 0;
	int ic = 0;

	int ta = 0;
	int tb = 0;
	int tc = 0;
	for(int ci=0; ci < coordIndex.length; ){
		
		ia = rcCoord(ci++, ta);
		ib = rcCoord(ci++, tb);
		ic = rcCoord(ci++, tc);

		triangle[s++] = S_Triangle(part,ia+h, ta+c, part,ib+h,tb+c, part,ic+h,tc+c);
	}
}

void Polymap::convertCoordIndexToTriangle(Art* part, int h, int c, int s){
	M_Triangle& triangle = part->triangle;
	const M_Int& rc = this->rc;
	int ia = 0;
	int ib = 0;
	int ic = 0;

	int ta = 0;
	int tb = 0;
	int tc = 0;
	for(int ci=0; ci < coordIndex.length; ){
		
		ia = rcCoord(ci++, ta);
		ib = rcCoord(ci++, tb);
		ic = rcCoord(ci++, tc);

		triangle[s++] = S_Triangle(part,ia+h, ta+c, part,ib+h,tb+c, part,ic+h,tc+c);
	}
}

/*
void Stem::postLoadProcess(){
	const M_Int& rc = this->rc;
	for(int i, ci=0; ci<coordIndex.length; ){
		int ia;
		int ib;
		int ic;

		int ta;
		int tb;
		int tc;
		{
		
			i=coordIndex[ci++];
			ta=i;
			ia=i;
			int g=ia-(coord.length-rc.length);
			if(g>=0){
				ia=rc[g];
			}
		}
		{
			i=coordIndex[ci++];
			tb=i;
			ib=i;
			int g=ib-(coord.length-rc.length);
			if(g>=0){
				ib=rc[g];
			}
		}
		{
			i=coordIndex[ci++];
			tc=i;
			ic=i;
			int g=ic-(coord.length-rc.length);
			if(g>=0){
				ic=rc[g];
			}
		}
		LocalTriangle tri=LocalTriangle(ia,ta, ib,tb, ic,tc);

		triangle+=tri;
	}
}
*/

void inline sweepColorY(const int s, const M_Triangle& triangle, M_Vec3f& coord, M_Color3f& color, const Color3f beginColor,const Color3f endColor, float beginOpacity, float endOpacity, float beginHeight, float endHeight){
	float opacityDiff=endOpacity-beginOpacity;
	float rColorDiff=endColor.r-beginColor.r;
	float gColorDiff=endColor.g-beginColor.g;
	float bColorDiff=endColor.b-beginColor.b;
	float heightDiff=endHeight-beginHeight;
	for(int i = triangle.getTriangleLength(); --i >= s;){
		//if(triangle.refPart(i)==part){
		int ci=triangle.refTexIndex(i);
		Vec3f vec = coord[triangle.refIndex(i)];
		float length = vec.y;
		if(length >= beginHeight && length <= endHeight){
			
			float frac = (length-beginHeight)/heightDiff;
			float opacity = (beginOpacity+frac*opacityDiff);
			float clarity = 1.0 - opacity;

			color[ci]=Color3f(
				color[ci].r*clarity+opacity*(beginColor.r+frac*rColorDiff),
				color[ci].g*clarity+opacity*(beginColor.g+frac*gColorDiff),
				color[ci].b*clarity+opacity*(beginColor.b+frac*bColorDiff)
				);
		}
	}
}

void inline sweepColorSpherical(const int s, const M_Triangle& triangle, M_Vec3f& coord, M_Color3f& color, const Color3f beginColor,const Color3f endColor, float beginOpacity, float endOpacity, float beginHeight, float endHeight){

	float opacityDiff=endOpacity-beginOpacity;
	float rColorDiff=endColor.r-beginColor.r;
	float gColorDiff=endColor.g-beginColor.g;
	float bColorDiff=endColor.b-beginColor.b;
	float heightDiff=endHeight-beginHeight;
	for(int i=triangle.length*3;--i>=0;){
		int ci=triangle.refTexIndex(i);
		Art* prt = triangle.refPart(i);
		Vec3f vec=prt->matrix * prt->coord[triangle.refIndex(i)];

		float length = vec.length();
		if(length >= beginHeight && length <= endHeight){
			float frac = (length-beginHeight)/heightDiff;
			float opacity = (beginOpacity+frac*opacityDiff);
			float clarity = 1.0-opacity;
			color[ci] = Color3f(
				color[ci].r*clarity+opacity*(beginColor.r+frac*rColorDiff),
				color[ci].g*clarity+opacity*(beginColor.g+frac*gColorDiff),
				color[ci].b*clarity+opacity*(beginColor.b+frac*bColorDiff)
			);
		}
	}
}

void ColorSweep::doColorForStem(M_Triangle& m_triangle, M_Vec3f& coord, M_Color3f& eColor, const int h, const int c, const int s){
	if(color.length == 0){
		return;
	}
	float posFrom=0;
	float posTo=distance.refValue()[0];
	int s3 = s * 3;

	for(int i=0;;){
		if(straight){
			sweepColorY(s3, m_triangle, coord, eColor, color[i], color[i+1], opacity[i], opacity[i+1], posFrom, posTo);
		}
		else if(spherical){
			sweepColorSpherical(s3, m_triangle, coord, eColor, color[i], color[i+1], opacity[i], opacity[i+1], posFrom, posTo);
		}

		if(i>=distance.length-1)break;


		posFrom = posTo;
		i++;
		posTo += distance[i];

	}
}
