/* Seamless3d

Copyright (C) 2001-2010 Graham John Perrett (AKA thyme) thyme@seamless3d.com www.seamless3d.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this
software and associated documentation files (the "Software"), to deal in the Software
without restriction, including without limitation the rights to use, copy, modify,
merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to the following
conditions:

The above copyright notice and this permission notice shall be included in all copies or
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT
OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.
*/

#include <windows.h>
#include <direct.h>

#include <sys/stat.h>
#include <share.h>
#include <fcntl.h>
#include <io.h>

#include "seamless3d.h"
#include "techuelife_triangle_editor.h"
#include "seamless3d_gui_render.h"
#include "gui.h"
#include "scene_tree.h"
#include "robot_aid.h"

void processAllPatches(M_<NurbsPatch*>& cpNodeList);
void commonPostTug();
void updateCage(Seamless* smls);
int getEdgeInnerPoint(NurbsPatch* patch, int idx);
void getOppositeEdgePoint(NurbsPatch* patch, int idx, CornerEdge& ce);
void calcMidPoints(const int i, const M_CtrlPolyPoint& ctrlPolyPoint);
static M_<CornerEdge> m_cornerEdge;

void makeStitchPointsPlanar(Seamless* smls, const M_CtrlPolyPoint& ctrlPolyPoint){
	
	for(int i = 0; i < ctrlPolyPoint.length; ){
		int h = i;
		const int len = ctrlPolyPoint[i].len;
		if(len < 5){
			calcMidPoints(i, ctrlPolyPoint);
			i += len;
			for(int k = len + 1; --k >= 0;){
				int midLen = ctrlPolyPoint[i].len;
				if(midLen == 0){
					midLen++;
				}
				i += midLen;
			}
			continue;
		}
		int centerIdx = i + len;

		for(int k = len; --k >= 0;){
			centerIdx += ctrlPolyPoint[centerIdx].len;
		}

		Vec3f a = ctrlPolyPoint[centerIdx].p->dpoint[ctrlPolyPoint[centerIdx].i];

		Vec3f centerNormal;

		for(int q = len; --q >= 0;){
			Vec3f b = ctrlPolyPoint[h].p->dpoint[ctrlPolyPoint[h].i];

			if(++h >= i + len){
				h = i;
			}
			Vec3f c = ctrlPolyPoint[h].p->dpoint[ctrlPolyPoint[h].i];

			centerNormal += calcNormal(a, b, c);
		}
/*
		int i = stitchIdxIdx[sii].index;
		int numOf = stitchIdxIdx[sii].numOf;
		int cornerLen = numOf - 1;
		m_cornerEdge.length = 0;

		for(int k = 1; k < numOf; k++){
			int w = i + k;
			int innerPointIdx = getEdgeInnerPoint(patch, ctrlPolyPoint[w].i);
			if(idx == innerPointIdx){
				Vec3f centerNormal;

				for(int j = 0; j < numOf; j++){
					CornerEdge ce;
					int w = i + j;
					getOppositeEdgePoint(ctrlPolyPoint[w].p, ctrlPolyPoint[w].i, ce);
					m_cornerEdge += ce;
				}

				Vec3f cornerPoint = m_cornerEdge[0].patch->dpoint.refValue()[m_cornerEdge[0].cornerIdx];

				for(int j = 0; j < cornerLen; j++){
					int w = i + j;
					Vec3f normal;
					Vec3f v1 = m_cornerEdge[j].patch->dpoint.refValue()[m_cornerEdge[j].edgeIdx];
					Vec3f v2 = m_cornerEdge[j].patch->dpoint.refValue()[m_cornerEdge[j].oppositeEdgeIdx];
					calcNormal(cornerPoint, v1, v2, normal);
					centerNormal += normal;
				}
*/

		centerNormal = centerNormal.normalize();
		Rotation rot = Rotation(centerNormal, Vec3f(0, 1, 0));
		a = rot * a;

		for(int q = len, h = i; --q >= 0; h++){
			NurbsPatch* p = ctrlPolyPoint[h].p;
			int idx = ctrlPolyPoint[h].i;

			const Vec3f v0 = p->dpoint[idx];
			Vec3f v1 = rot * v0;
			v1.y = a.y;
			v1 = -rot * v1;
			if(	abs(v1.x - v0.x) > .00001 ||
				abs(v1.y - v0.y) > .00001 ||
				abs(v1.z - v0.z) > .00001){

				for(int q = len, h = i; --q >= 0; h++){
					NurbsPatch* p = ctrlPolyPoint[h].p;
					int idx = ctrlPolyPoint[h].i;

					Vec3f v = rot * p->dpoint[idx];
					v.y = a.y;
					v = -rot * v;
					p->point.refValue()[idx] = v * p->weight[idx];
				}
					
				break;
			}
		}

		calcMidPoints(i, ctrlPolyPoint);
		i = centerIdx + ctrlPolyPoint[centerIdx].len;
	}
}

void makeStitchPointsPlanar(){
	if(NurbsPatch* nurbsPatch = dynamic_cast<NurbsPatch*>(g_selNode.getRaw())){
		Seamless* smls = nurbsPatch->smls;
		makeStitchPointsPlanar(smls, smls->ctrlPolyPoint);
		processAllPatches(smls->cpNodeList);
//		nurbsPatch->houseKeep();
		smls->doBuild();
	}
	else if(Art* part=dynamic_cast<Art*>(g_selNode.getRaw())){

		if(MultiSelect::pointList.length){
			for(int i = MultiSelect::pointList.length; --i>=0;){
				int idx = MultiSelect::pointList[i].i;
				NurbsPatch* patch = MultiSelect::pointList[i].p;
				Seamless* smls = part->smls;
				makeStitchPointsPlanar(smls, smls->ctrlPolyPoint);
				processAllPatches(smls->cpNodeList);
				commonPostTug();
				updateCage(smls);
			}
		}
	}
	updateBuild3dWindow();
	BackUp::monostable.beginCycle();
}

/*
int getEdgeInnerPoint(NurbsPatch* patch, int idx){
	int uDim = patch->dim.x;
	int vDim = patch->dim.y;
	int u = idx % uDim;
	int v = idx / uDim;
	//which edge?
	if(u == 0){//left edge
		return idx + 1;
	}
	else if(u == uDim - 1){//right edge
		return idx - 1;
	}
	else if(v == 0){//south edge
		return idx + uDim;
	}
	else if(v == vDim - 1){//north edge
		return idx - uDim;
	}
	return 0;
}

void getOppositeEdgePoint(NurbsPatch* patch, int idx, CornerEdge& ce){
	ce.patch=patch;
	ce.edgeIdx=idx;

	int uDim = patch->dim.x;
	int vDim = patch->dim.y;
	int u = idx % uDim;
	int v = idx / uDim;

	//which edge?
	if(u == 0){//left edge
		ce.innerPointIdx = idx + 1;
		if(v == 1){
			ce.oppositeEdgeIdx = ce.innerPointIdx - uDim;
			ce.cornerIdx = 0;
		}
		else{
			ce.oppositeEdgeIdx = ce.innerPointIdx + uDim;
			ce.cornerIdx = uDim * (vDim - 1);
		}
	}
	else if(u == uDim - 1){//right edge
		ce.innerPointIdx = idx - 1;
		if(v == 1){
			ce.oppositeEdgeIdx = ce.innerPointIdx - uDim;
			ce.cornerIdx = uDim - 1;
		}
		else{
			ce.oppositeEdgeIdx = ce.innerPointIdx + uDim;
			ce.cornerIdx = uDim * vDim - 1;
		}
	}
	else if(v == 0){//south edge
		ce.innerPointIdx = idx + uDim;
		if(u == 1){
			ce.oppositeEdgeIdx = ce.innerPointIdx - 1;
			ce.cornerIdx = 0;
		}
		else{
			ce.oppositeEdgeIdx = ce.innerPointIdx + 1;
			ce.cornerIdx = uDim - 1;
		}
	}
	else if(v == vDim - 1){//north edge
		ce.innerPointIdx = idx - uDim;
		if(u == 1){
			ce.oppositeEdgeIdx = ce.innerPointIdx - 1;
			ce.cornerIdx = uDim * (vDim - 1);
		}
		else{
			ce.oppositeEdgeIdx = ce.innerPointIdx + 1;
			ce.cornerIdx = uDim * vDim - 1;
		}
	}
}
*/

void deleteControlPointFromAllMorphers(Art* part, int i);

void Seamless::removeOwnershipPatch(const NurbsPatch* patch){
	for(int k = partList.length; --k>=0;){
		Art* part = partList[k];
		M_ControlPoint& controlPoint = part->controlPoint;
		for(int i = controlPoint.length; --i>=0;){
			if(controlPoint[i].p == patch){
				deleteControlPointFromAllMorphers(part, i);
				controlPoint.deleteElement(i);
			}
		}
	}
}
void Seamless::removeMidPointsOwnership(const NurbsPatch* patch, const int idx){
	for(int k = partList.length; --k>=0;){
		M_ControlPoint& controlPoint = partList[k]->controlPoint;
		for(int i = controlPoint.length; --i>=0;){
			if(controlPoint[i].p == patch && controlPoint[i].i == idx){
				controlPoint.deleteElement(i);
			}
		}
	}
}

void Seamless::removeMidPointsOwnership(const int i, const M_CtrlPolyPoint& ctrlPolyPoint){
	const int len = ctrlPolyPoint[i].len;
	const int midIdx = i + len;
	int g = midIdx;
	int h = i;
	for(int q = 0; q < len; q++){

		if(++h >= i + len){
			h = i;
		}
		int midLen = ctrlPolyPoint[g].len;
		if(midLen){
			for(int m = midLen; --m >= 0;){
				NurbsPatch* p = ctrlPolyPoint[g].p;
				int idx = ctrlPolyPoint[g++].i;
				removeMidPointsOwnership(p, idx);
			}
		}
		else{
			g++;
		}
	}

	int centerLen = ctrlPolyPoint[g].len;
	for(int q = centerLen; --q >= 0;){
		NurbsPatch* p = ctrlPolyPoint[g].p;
		int idx = ctrlPolyPoint[g++].i;
		removeMidPointsOwnership(p, idx);
	}
}

void Seamless::removeAllPolyMidPointsOwnership(){
	for(int i = 0; i < ctrlPolyPoint.length;){
		removeMidPointsOwnership(i, ctrlPolyPoint);
		i = getNextCtrlPoly(i);
	}
	processAllPatches(cpNodeList);
}


void calcMidPoints(const int i, const M_CtrlPolyPoint& ctrlPolyPoint){
	const int len = ctrlPolyPoint[i].len;
	const int midIdx = i + len;
	int g = midIdx;
	int h = i;
	Vec3f centerPoint;
	//mid points
	for(int q = 0; q < len; q++){
		//NurbsPatch* patch = ctrlPolyPoint[h].p;
		//int idx = ctrlPolyPoint[h].i;
		//Vec3f v = patch->dpoint.refValue()[idx] = patch->point[idx] / patch->weight[idx];

		Vec3f v = ctrlPolyPoint[h].p->dpoint[ctrlPolyPoint[h].i];
		centerPoint += v;

		if(++h >= i + len){
			h = i;
		}
		int midLen = ctrlPolyPoint[g].len;
		if(midLen){
			Vec3f midPoint = (v + ctrlPolyPoint[h].p->dpoint[ctrlPolyPoint[h].i]) / 2.0f;

			for(int m = midLen; --m >= 0;){
				NurbsPatch* p = ctrlPolyPoint[g].p;
				int idx = ctrlPolyPoint[g++].i;
				p->point.refValue()[idx] = midPoint * p->weight[idx];
			}
		}
		else{
			g++;
		}
	}
	centerPoint /= len;

	int centerLen = ctrlPolyPoint[g].len;
	for(int q = centerLen; --q >= 0;){
		NurbsPatch* p = ctrlPolyPoint[g].p;
		int idx = ctrlPolyPoint[g++].i;
		p->point.refValue()[idx] = centerPoint * p->weight[idx];
	}
}


void Seamless::dragCtrlPoly(NurbsPatch* patch, int idx){

	for(int i = 0; i < ctrlPolyPoint.length;){
		int selCtrlPoly = i;
		int len = ctrlPolyPoint[selCtrlPoly].len;
		for(int k = len; --k >= 0;){
			if(patch == ctrlPolyPoint[i].p && idx == ctrlPolyPoint[i].i){
				calcMidPoints(selCtrlPoly, ctrlPolyPoint);
				break;
			}
			i++;
		}
		i = selCtrlPoly + len;
		for(int k = len + 1; --k >= 0;){
			int midLen = ctrlPolyPoint[i].len;
			if(midLen == 0){
				midLen++;
			}
			i += midLen;
		}
	}
	processAllPatches(cpNodeList);
}

void Seamless::decodeStitchIndex(M_<NurbsPatch*>& contigLookUpList){
	M_Int& stitchIndex = this->stitchIndex.refValue();
	for(int i = 0; i < stitchIndex.length;){

		CtrlPolyPoint cpp;
		cpp.len = stitchIndex[i++];

		for(int q = cpp.len; --q >= 0;){
			int b = stitchIndex[i++];
			cpp.p = contigLookUpList[b];
			cpp.i = b - cpp.p->fromPoint;
			ctrlPolyPoint += cpp;
		}
		//mid points
		for(int q = cpp.len; --q >= 0;){
			if(cpp.len = stitchIndex[i++]){
				for(int m = cpp.len; --m >= 0;){
					int b = stitchIndex[i++];
					cpp.p = contigLookUpList[b];
					cpp.i = b - cpp.p->fromPoint;
					ctrlPolyPoint += cpp;
				}
			}
			else{
				ctrlPolyPoint += cpp;
			}
		}
		//center points
		if(cpp.len = stitchIndex[i++]){
			for(int m = cpp.len; --m >= 0;){
				int b = stitchIndex[i++];
				cpp.p = contigLookUpList[b];
				cpp.i = b - cpp.p->fromPoint;
				ctrlPolyPoint += cpp;
			}
		}
		else{
			ctrlPolyPoint += cpp;
		}
	}
}

bool isFixed(NurbsPatch* patch, int idx){
	int uDim = patch->dim.x;
	int vDim = patch->dim.y;
	int u = idx % uDim;
	int v = idx / uDim;
	if(u % 3 == 0){
		return true;
	}
	if(v % 3 == 0){
		return true;
	}
	return false;
}

void Seamless::createControlPolygon(M_ControlPoint& pointList){
	MultiSelect::ctrlPolySelList.length = 0;
	MultiSelect::ctrlPolySelList += ctrlPolyPoint.length;

	M_Vec2f screenPoint;
	screenPoint.length = pointList.length;
	const bool perspective = g_settings->perspective;

	for(int k = pointList.length; --k >=0;){
		screenPoint[k] = procPerspective(perspective, g_viewMatrix * pointList[k].p->dpoint[pointList[k].i]);
	}


	Vec2f center;

	for(int k = pointList.length; --k >=0;){
		center += screenPoint[k];
	}
	center /= pointList.length;

	M_Float angle;
	angle.length = pointList.length;
	for(int k = pointList.length; --k >=0;){
		angle[k] = (screenPoint[k] - center).getAngle3();
	}

	M_ControlPoint orderedPoint;

	while(pointList.length){
		float mostVal = -1;
		int mostIdx = 0;
		for(int k = pointList.length; --k >=0;){
			if(angle[k] > mostVal){
				mostVal = angle[k];
				mostIdx = k;
			}
		}
		orderedPoint += pointList[mostIdx];
		pointList.deleteElement(mostIdx);
		angle.deleteElement(mostIdx);
	}

	int len = orderedPoint.length;
	for(int k = 0; k < len; k++){
		ctrlPolyPoint += CtrlPolyPoint(orderedPoint[k].p, orderedPoint[k].i, len);
	}
	ctrlPolyPoint.length += len + 1;


	processAllPatches(cpNodeList);
	houseKeepVisualClue();
	BackUp::monostable.beginCycle();
}

void createMidPointList(const int i, const M_CtrlPolyPoint& ctrlPolyPoint, M_Vec3f& midPoint){
	const int len = ctrlPolyPoint[i].len;
	midPoint.length = 0;
	int h = i;
	Vec3f centerPoint;
	//mid points
	for(int q = 0; q < len; q++){
		Vec3f v = ctrlPolyPoint[h].p->dpoint[ctrlPolyPoint[h].i];
		centerPoint += v;

		if(++h >= i + len){
			h = i;
		}
		midPoint += (v + ctrlPolyPoint[h].p->dpoint[ctrlPolyPoint[h].i]) / 2.0f;
	}
	midPoint += centerPoint / len;
}

void insertMidPoint(M_CtrlPolyPoint& ctrlPolyPoint, int ctrlPolyIdx, NurbsPatch* patch, int idx, int shortestK){
	int midPointIdx = ctrlPolyIdx + ctrlPolyPoint[ctrlPolyIdx].len;
	for(int m = 0; ; m++){
		if(m == shortestK){
			break;
		}
		int midLen = ctrlPolyPoint[midPointIdx].len;
		if(midLen == 0){
			midLen++;
		}
		midPointIdx += midLen;
	}

	int newLen = ctrlPolyPoint[midPointIdx].len + 1;

	if(newLen == 1){
		ctrlPolyPoint[midPointIdx] = CtrlPolyPoint(patch, idx, newLen);
	}
	else{
		ctrlPolyPoint.insertElement(midPointIdx, CtrlPolyPoint(patch, idx, newLen));
	}
}

void Seamless::addMidPoint(M_ControlPoint& pointList){
	M_Vec3f midPoint;
	int selCtrlPoly = MultiSelect::ctrlPolySelList[0];

	createMidPointList(selCtrlPoly, ctrlPolyPoint, midPoint);
	int midPointIdx = selCtrlPoly + ctrlPolyPoint[selCtrlPoly].len;

	for(int q = 0; q < pointList.length; q++){
		float shortest = 10000000;
		int shortestK = 0;
		int shortestQ = 0;
	
		for(int k = 0; k < midPoint.length; k++){
			NurbsPatch* patch = pointList[q].p;
			int idx = pointList[q].i;
			float distance = (midPoint[k] - patch->dpoint[idx]).length();
			if(distance < shortest){
				shortest = distance;
				shortestK = k;
				shortestQ = q;
			}
		}
		insertMidPoint(ctrlPolyPoint, selCtrlPoly, pointList[shortestQ].p, pointList[shortestQ].i, shortestK);
	}

	pointList.length = 0;
	calcMidPoints(selCtrlPoly, ctrlPolyPoint);
	//removeAllPolyMidPointsOwnership();
	processAllPatches(cpNodeList);
	houseKeepVisualClue();
	BackUp::monostable.beginCycle();
}

void encodeStitchIndex(M_Int& stitchIndex, const M_CtrlPolyPoint& ctrlPolyPoint){
	stitchIndex.length = 0;
	for(int k = 0; k < ctrlPolyPoint.length; ){
		int len = ctrlPolyPoint[k].len;
		stitchIndex += len;
		if(len){
			for(int g = 0; g < len; g++){
				stitchIndex += ctrlPolyPoint[k].i + ctrlPolyPoint[k].p->fromPoint;
				k++;
			}
		}
		else{
			k++;
		}
	}
}

int Seamless::getNextCtrlPoly(int ctrlPoly){
	int len = ctrlPolyPoint[ctrlPoly].len;
	ctrlPoly += len;
	for(int k = len + 1; --k >= 0;){
		int midLen = ctrlPolyPoint[ctrlPoly].len;
		if(midLen == 0){
			midLen++;
		}
		ctrlPoly += midLen;
	}
	return ctrlPoly;
}

void Seamless::getNextCtrlPoly(){
	if(ctrlPolyPoint.length == 0){
		return;
	}
	MultiSelect::ctrlPolySelList.length = 1;
	int selCtrlPoly = MultiSelect::ctrlPolySelList[0];

	if(selCtrlPoly >= ctrlPolyPoint.length){
		selCtrlPoly = 0;
	}
	selCtrlPoly = getNextCtrlPoly(selCtrlPoly);
	if(selCtrlPoly >= ctrlPolyPoint.length){
		selCtrlPoly = 0;
	}
	MultiSelect::ctrlPolySelList[0] = selCtrlPoly;
}

void Seamless::getPrevCtrlPoly(){
	if(ctrlPolyPoint.length == 0){
		return;
	}
	MultiSelect::ctrlPolySelList.length = 1;

	int targetPoly = MultiSelect::ctrlPolySelList[0];
	int prevPoly = MultiSelect::ctrlPolySelList[0];
	while(1){
		getNextCtrlPoly();
		if(targetPoly == MultiSelect::ctrlPolySelList[0]){
			MultiSelect::ctrlPolySelList[0] = prevPoly;
			return;
		}
		prevPoly = MultiSelect::ctrlPolySelList[0];
	}
}

void Seamless::deleteCtrlPoly(int ctrlPoly){
	if(ctrlPolyPoint.length){
		int nextCtrlPoly = getNextCtrlPoly(ctrlPoly);
		ctrlPolyPoint.deleteElement(ctrlPoly, nextCtrlPoly - ctrlPoly);
	}
}

void Seamless::deleteCtrlPoly(M_Int& deletePolyList){
	
	bubbleSort(deletePolyList);
	
	deleteLinksToPolys(deletePolyList);

	for(int q = deletePolyList.length; --q >= 0;){
		deleteCtrlPoly(deletePolyList[q]);
	}

}

void Seamless::calcAllPolyMidPoints(){
	for(int k = 0; k < cpNodeList.length; k++){
		cpNodeList[k]->pointsToDpoints();
	}

	for(int i = 0; i < ctrlPolyPoint.length;){
		calcMidPoints(i, ctrlPolyPoint);
		i = getNextCtrlPoly(i);
	}

	calcAllLinks();

	processAllPatches(cpNodeList);
}


void Seamless::redirect(NurbsPatch* patch, M_Int& lookup){
	for(int i = ctrlPolyPoint.length; --i >= 0;){
		if(ctrlPolyPoint[i].p == patch){
			ctrlPolyPoint[i].i = lookup[ctrlPolyPoint[i].i];
		}
	}
}

int getCorner(int corner, int offset, int len){
	return (corner + offset) % len;
}

int getMidPointAbsoluteIdx(int midIndex, int corner, const M_CtrlPolyPoint& ctrlPolyPoint){
	for(int c = 0; c < corner; c++){
		midIndex += ctrlPolyPoint[midIndex + c].len;
	}
	return midIndex;
}

void Seamless::mergeControlPolygon(int polyA, int a0, int a1, int polyB, int b0, int b1, M_Int& deleteCtrlPolySelList){

	int midB = getMidOffset(polyB, b1);
	int midA = getMidOffset(polyA, a0);

//	MultiSelect::pointList += ControlPoint(ctrlPolyPoint[polyB + b1].p, ctrlPolyPoint[polyB + b1].i);
	//MultiSelect::pointList += ControlPoint(ctrlPolyPoint[polyB + b1].p, ctrlPolyPoint[polyB + b1].i);

	int lenA = ctrlPolyPoint[midA].len;
	int lenB = ctrlPolyPoint[midB].len;
	//return;

	//MultiSelect::pointList += ControlPoint(ctrlPolyPoint[polyA + 4 + a0].p, ctrlPolyPoint[polyA + 4 + a0].i);
	//MultiSelect::pointList += ControlPoint(ctrlPolyPoint[polyA + a0].p, ctrlPolyPoint[polyA + a0].i);
	//MultiSelect::pointList += ControlPoint(ctrlPolyPoint[polyA + a1].p, ctrlPolyPoint[polyA + a1].i);

//	MultiSelect::pointList += ControlPoint(ctrlPolyPoint[polyB + b0].p, ctrlPolyPoint[polyB + b0].i);
//	MultiSelect::pointList += ControlPoint(ctrlPolyPoint[polyB + b1].p, ctrlPolyPoint[polyB + b1].i);

	CtrlPolyPoint cpp;

	int len = 4;
	cpp.len = len;
	int idx = getCorner(a0, 3, len);
	cpp.p = ctrlPolyPoint[polyA + idx].p;
	cpp.i = ctrlPolyPoint[polyA + idx].i;
	ctrlPolyPoint += cpp;
	//MultiSelect::pointList += ControlPoint(cpp.p, cpp.i);

	idx = getCorner(b0, 1, len);
	cpp.p = ctrlPolyPoint[polyB + idx].p;
	cpp.i = ctrlPolyPoint[polyB + idx].i;
	ctrlPolyPoint += cpp;
	//MultiSelect::pointList += ControlPoint(cpp.p, cpp.i);

	idx = getCorner(b0, 2, len);
	cpp.p = ctrlPolyPoint[polyB + idx].p;
	cpp.i = ctrlPolyPoint[polyB + idx].i;
	ctrlPolyPoint += cpp;
	//MultiSelect::pointList += ControlPoint(cpp.p, cpp.i);

	idx = getCorner(a0, 2, len);
	cpp.p = ctrlPolyPoint[polyA + idx].p;
	cpp.i = ctrlPolyPoint[polyA + idx].i;
	ctrlPolyPoint += cpp;
	//MultiSelect::pointList += ControlPoint(cpp.p, cpp.i);

	ctrlPolyPoint += CtrlPolyPoint(ctrlPolyPoint[polyA + a0].p, ctrlPolyPoint[polyA + a0].i, 2);
	ctrlPolyPoint += CtrlPolyPoint(ctrlPolyPoint[polyB + b0].p, ctrlPolyPoint[polyB + b0].i, 2);
	//------------
	ctrlPolyPoint.length++;
	ctrlPolyPoint += CtrlPolyPoint(ctrlPolyPoint[polyA + a1].p, ctrlPolyPoint[polyA + a1].i, 2);
	ctrlPolyPoint += CtrlPolyPoint(ctrlPolyPoint[polyB + b1].p, ctrlPolyPoint[polyB + b1].i, 2);
	ctrlPolyPoint.length++;

	int centerLen = lenA + lenB;
	if(centerLen){
		for(int i = 0; i < lenA; i++){
			ctrlPolyPoint += CtrlPolyPoint(ctrlPolyPoint[midA + i].p, ctrlPolyPoint[midA + i].i, centerLen);
		}
		for(int i = 0; i < lenB; i++){
			ctrlPolyPoint += CtrlPolyPoint(ctrlPolyPoint[midB + i].p, ctrlPolyPoint[midB + i].i, centerLen);
		}
	}
	else{
		ctrlPolyPoint.length++;
	}
	deleteCtrlPolySelList += polyA;
	deleteCtrlPolySelList += polyB;
	ctrlPolyPoint[polyB].p = 0;
	ctrlPolyPoint[polyA].p = 0;
}

bool Seamless::findCtrlPolyWith2SelectedClockwise(M_Int& ctrlPolySelList, NurbsPatch* patchA0, int polyA, int a0, int a1, const Vec3f vec0, const Vec3f vec1, M_Int& deleteCtrlPolySelList){
	for(int q = 0; q < ctrlPolySelList.length; q++){
		int polyB = ctrlPolySelList[q];
		int len = ctrlPolyPoint[polyB].len;
		for(int b0 = 0; b0 < len; b0++){
			if(ctrlPolyPoint[polyB].p){
				NurbsPatch* patchB0 = ctrlPolyPoint[polyB + b0].p;
				int iB0 = ctrlPolyPoint[polyB + b0].i;
	////
				int b1 = getCorner(b0, len - 1, len);
				NurbsPatch* patchB1 = ctrlPolyPoint[polyB + b1].p;
				int iB1 = ctrlPolyPoint[polyB + b1].i;

				if((vec0 - patchB0->dpoint[iB0]).length() < .001){
					if((vec1 - patchB1->dpoint[iB1]).length() < .001){
						mergeControlPolygon(polyA, a0, a1, polyB, b0, b1, deleteCtrlPolySelList);
						return true;
					}
				}
			}
		}
	}
	return false;
}

void Seamless::findCtrlPolyWith2SelectedAntiClockwise(M_Int& ctrlPolySelList, M_Int& deleteCtrlPolySelList){
	for(int q = 0; q < ctrlPolySelList.length; q++){
		int polyA = ctrlPolySelList[q];
		int len = ctrlPolyPoint[polyA].len;
		for(int a0 = 0; a0 < len; a0++){
			if(ctrlPolyPoint[polyA].p){
				NurbsPatch* patchA0 = ctrlPolyPoint[polyA + a0].p;
				int iA0 = ctrlPolyPoint[polyA + a0].i;

				int a1 = getCorner(a0, 1, len);
				NurbsPatch* patchA1 = ctrlPolyPoint[polyA + a1].p;
				//if(patchA0 == patchA1){
					int iA1 = ctrlPolyPoint[polyA + a1].i;

					Vec3f vecA0 = patchA0->dpoint[iA0];
					Vec3f vecA1 = patchA1->dpoint[iA1];

					if(findCtrlPolyWith2SelectedClockwise(MultiSelect::ctrlPolySelList, patchA0, polyA, a0, a1, vecA0, vecA1, deleteCtrlPolySelList)){
						//MultiSelect::pointList += ControlPoint(ctrlPolyPoint[polyA + a0].p, ctrlPolyPoint[polyA + a0].i);
						//MultiSelect::pointList += ControlPoint(ctrlPolyPoint[polyA + a1].p, ctrlPolyPoint[polyA + a1].i);
						return;
					}
				//}
			}
		}
	}
}

inline bool polyInList(int poly, M_Int& deleteCtrlPolySelList, int checkToPoly){
	for(int i = deleteCtrlPolySelList.length; --i >= checkToPoly;){
		if(deleteCtrlPolySelList[i] == poly){
			return true;
		}
	}
	return false;
}

bool Seamless::findCloseCornerLoop(int polyA, int corner, M_Int& ctrlPolySelList, M_Int& deleteCtrlPolySelList, M_Int& m_corner, int checkToPoly){
	int len = ctrlPolyPoint[polyA].len;
	int a = corner;
	for(int c = 1; c < len; c++){
		a = getCorner(a, 1, len);
		NurbsPatch* patchA = ctrlPolyPoint[polyA + a].p;
		int iA = ctrlPolyPoint[polyA + a].i;
		Vec3f vecA = patchA->dpoint[iA];
	
		for(int q = 0; q < ctrlPolySelList.length; q++){
			int polyB = ctrlPolySelList[q];
			int len = ctrlPolyPoint[polyA].len;
			if(false == polyInList(polyB, deleteCtrlPolySelList, checkToPoly)){

				for(int b = 0; b < len; b++){
					NurbsPatch* patchB = ctrlPolyPoint[polyB + b].p;
					if(patchA != patchB){
						int iB = ctrlPolyPoint[polyB + b].i;
						Vec3f vecB = patchB->dpoint[iB];
						if(	abs(vecA.x - vecB.x) < .00001 &&
							abs(vecA.y - vecB.y) < .00001 &&
							abs(vecA.z - vecB.z) < .00001){

							//MultiSelect::pointList += ControlPoint(ctrlPolyPoint[polyB + b].p, ctrlPolyPoint[polyB + b].i);
							if(m_corner.length == 3){
								checkToPoly++;
							}
							m_corner += a;
							m_corner += b;
							if(polyB == deleteCtrlPolySelList[checkToPoly-1]){
								m_corner[checkToPoly-1] = b;
								return true;
							}
							deleteCtrlPolySelList += polyB;
							return findCloseCornerLoop(polyB, b, ctrlPolySelList, deleteCtrlPolySelList, m_corner, checkToPoly);
						}
					}
				}
			}
		}
	}
	return false;
}

bool is5Poly(M_Int& m_corner, M_Int& m_ab, int& c){
	bool flag = false;
	for(int k = 0; k < (m_corner.length - 1) / 2; k++){
		if(m_corner[k * 2] - m_corner[k * 2 + 1] == 2){
			flag = true;
			c = k;
		}
		else if(m_corner[k * 2] - m_corner[k * 2 + 1] == -2){
			flag = true;
			c = k;
		}
		else{
			m_ab += k;
		}
	}
	return flag;
}

void Seamless::create5SidedPoly(int polyA, int a1, int a0, int polyB, int b1, int b0, int polyC, int c1, int c3){
//	int polyC = getNurbsCorner(c0, MultiSelect::ctrlPolySelList);
	//MultiSelect::pointList += ControlPoint(ctrlPolyPoint[polyA + a0].p, ctrlPolyPoint[polyA + a0].i);
	int c0 = getCorner(c1, 3, 4);
	//MultiSelect::pointList += ControlPoint(ctrlPolyPoint[polyC + c0].p, ctrlPolyPoint[polyC + c0].i);
	//MultiSelect::pointList += ControlPoint(ctrlPolyPoint[polyC + c3].p, ctrlPolyPoint[polyC + c3].i);

//	MultiSelect::pointList += ControlPoint(ctrlPolyPoint[polyA + a0].p, ctrlPolyPoint[polyA + a0].i);
//	MultiSelect::pointList += ControlPoint(ctrlPolyPoint[polyA + a1].p, ctrlPolyPoint[polyA + a1].i);
	//MultiSelect::pointList += ControlPoint(ctrlPolyPoint[polyB + b0].p, ctrlPolyPoint[polyB + b0].i);

	NurbsPatch* blockPatch = 0;
	int blockIdxA = -1;

	
	Vec3f selVec = ctrlPolyPoint[polyC + c0].p->dpoint[ctrlPolyPoint[polyC + c0].i];

//	int polyA = findClosestMidPoint(selVec, a0, blockIdxA);
//	int blockIdxB = blockIdxA;
	//MultiSelect::pointList += ControlPoint(ctrlPolyPoint[blockIdxA].p, ctrlPolyPoint[blockIdxA].i);
//	int polyB = findClosestMidPoint(selVec, b0, blockIdxB);
	//MultiSelect::pointList += ControlPoint(ctrlPolyPoint[blockIdxB].p, ctrlPolyPoint[blockIdxB].i);

	//int a1 = getCorner(a0, 1, 4);
	//int b1 = getCorner(b0, 1, 4);

	Vec3f nearA = ctrlPolyPoint[polyA + a1].p->dpoint[ctrlPolyPoint[polyA + a1].i];
	Vec3f nearB = ctrlPolyPoint[polyB + b1].p->dpoint[ctrlPolyPoint[polyB + b1].i];
	//MultiSelect::pointList += ControlPoint(ctrlPolyPoint[polyA + a1].p, ctrlPolyPoint[polyA + a1].i);
	//MultiSelect::pointList += ControlPoint(ctrlPolyPoint[polyB + b1].p, ctrlPolyPoint[polyB + b1].i);

	//int c3 = getCorner(c0, 3, 4);
	Vec3f selVec3 = ctrlPolyPoint[polyC + c3].p->dpoint[ctrlPolyPoint[polyC + c3].i];
	//MultiSelect::pointList += ControlPoint(ctrlPolyPoint[polyC + c3].p, ctrlPolyPoint[polyC + c3].i);

	if((selVec3 - nearA).length() > (selVec3 - nearB).length()){
		mergeTo5CtrlPoly2(polyC, c0, polyA, a0, a1, polyB, b0, b1);
	}
	else{
		mergeTo5CtrlPoly2(polyC, c0, polyB, b0, b1, polyA, a0, a1);
	}
}

void Seamless::findCloseCorner(M_Int& ctrlPolySelList, M_Int& deleteCtrlPolySelList, M_Int& m_corner){
	M_Int m_poly;
	for(int q = 0; q < ctrlPolySelList.length; q++){
		int polyA = ctrlPolySelList[q];
		if(false == polyInList(polyA, deleteCtrlPolySelList, 0)){
			int aLen = ctrlPolyPoint[polyA].len;
			for(int a = 0; a < aLen; a++){
				NurbsPatch* patchA = ctrlPolyPoint[polyA + a].p;
				int iA = ctrlPolyPoint[polyA + a].i;
				Vec3f vecA = patchA->dpoint[iA];
			
				for(int q = 0; q < ctrlPolySelList.length; q++){
					int polyB = ctrlPolySelList[q];
					int len = ctrlPolyPoint[polyA].len;
					if(polyA != polyB){

						for(int b = 0; b < len; b++){
							NurbsPatch* patchB = ctrlPolyPoint[polyB + b].p;
							if(patchA != patchB){
								int iB = ctrlPolyPoint[polyB + b].i;
								Vec3f vecB = patchB->dpoint[iB];
								if(	abs(vecA.x - vecB.x) < .00001 &&
									abs(vecA.y - vecB.y) < .00001 &&
									abs(vecA.z - vecB.z) < .00001){

									//deleteCtrlPolySelList += polyA;
									//deleteCtrlPolySelList += polyB;

									m_poly.length = 0;
									m_poly += polyA;
									m_poly += polyB;

									m_corner.length = 0;
									m_corner.length++;
									m_corner += a;
									m_corner += b;
									//int end = ctrlPolyPoint[polyA + a].len;
									//end = ctrlPolyPoint[polyB + b].len;
									//MultiSelect::pointList += ControlPoint(ctrlPolyPoint[polyA + a].p, ctrlPolyPoint[polyA + a].i);
									//MultiSelect::pointList += ControlPoint(ctrlPolyPoint[polyB + b].p, ctrlPolyPoint[polyB + b].i);
									//int checkToPoly = deleteCtrlPolySelList.length - 2;
									int checkToPoly = 0;
									if(findCloseCornerLoop(polyB, b, ctrlPolySelList, m_poly, m_corner, checkToPoly)){
										M_Int m_ab;
										int c = 0;
										if(is5Poly(m_corner, m_ab, c)){
											int polyA = m_poly[m_ab[0]];
											int a0 = m_corner[m_ab[0] * 2];
											int a1 = m_corner[m_ab[0] * 2 + 1];
											int polyB = m_poly[m_ab[1]];
											int b0 = m_corner[m_ab[1] * 2];
											int b1 = m_corner[m_ab[1] * 2 + 1];
											int polyC = m_poly[c];
											int c0 = m_corner[c * 2];
											int c1 = m_corner[c * 2 + 1];
											create5SidedPoly(polyA, a0, a1, polyB, b0, b1, polyC, c0, c1);
										}
										else{
											for(int k = checkToPoly; k < m_poly.length; k++){
												int poly = m_poly[k];
												int c1 = getCorner(m_corner[(k - checkToPoly) * 2 + 1], 3, 4);
												int c0 = getCorner(m_corner[(k - checkToPoly) * 2], 1, 4);
												//MultiSelect::pointList += ControlPoint(ctrlPolyPoint[poly + c0].p, ctrlPolyPoint[poly + c0].i);
												//MultiSelect::pointList += ControlPoint(ctrlPolyPoint[poly + c1].p, ctrlPolyPoint[poly + c1].i);
												ctrlPolyPoint += CtrlPolyPoint(ctrlPolyPoint[poly + c0].p, ctrlPolyPoint[poly + c0].i, 6);
												ctrlPolyPoint += CtrlPolyPoint(ctrlPolyPoint[poly + c1].p, ctrlPolyPoint[poly + c1].i, 6);
											}
											for(int k0 = checkToPoly; k0 < m_poly.length; k0++){
												int poly0 = m_poly[k0];
												int k1 = k0 + 1;
												if(k1 >= m_poly.length){
													k1 = checkToPoly;
												}
												int poly1 = m_poly[k1];
												int c0 = m_corner[(k0 - checkToPoly) * 2 + 1];
												int c1 = m_corner[(k1 - checkToPoly) * 2];
												ctrlPolyPoint.length++;
												ctrlPolyPoint += CtrlPolyPoint(ctrlPolyPoint[poly0 + c0].p, ctrlPolyPoint[poly0 + c0].i, 2);
												ctrlPolyPoint += CtrlPolyPoint(ctrlPolyPoint[poly1 + c1].p, ctrlPolyPoint[poly1 + c1].i, 2);
											}
											int end = 0;
											for(int k = checkToPoly; k < m_poly.length; k++){
												int poly = m_poly[k];
												int m = getMidOffset(poly, m_corner[(k - checkToPoly) * 2 + 1]);
												end += ctrlPolyPoint[m].len;
											}
											for(int k = checkToPoly; k < m_poly.length; k++){
												int poly = m_poly[k];
												int m = getMidOffset(poly, m_corner[(k - checkToPoly) * 2 + 1]);

												MultiSelect::pointList += ControlPoint(ctrlPolyPoint[m].p, ctrlPolyPoint[m].i);
												ctrlPolyPoint += CtrlPolyPoint(ctrlPolyPoint[m].p, ctrlPolyPoint[m].i, end);
											}
										}
										deleteCtrlPolySelList += m_poly;
										q = ctrlPolySelList.length;
										a = aLen;
										break;
									}
								}
							}
						}
					}
				}
			}
		}
	}
}
void Seamless::commonEndSitchPolys(){
	MultiSelect::ctrlPolySelList.length = 0;
	BackUp::monostable.beginCycle();
	calcAllPolyMidPoints();
	//removeAllPolyMidPointsOwnership();
	updateBuild3dWindow();
	BackUp::monostable.beginCycle();
}

void Seamless::stitchControlPolygons(){

	//if(MultiSelect::ctrlPolySelList.length == 3){
	//	mergeTo5CtrlPoly();
	//}
	if(MultiSelect::ctrlPolySelList.length == 2){
		M_Int deleteCtrlPolySelList;

		findCtrlPolyWith2SelectedAntiClockwise(MultiSelect::ctrlPolySelList, deleteCtrlPolySelList);

		bubbleSort(deleteCtrlPolySelList);
		deleteCtrlPoly(deleteCtrlPolySelList);

		commonEndSitchPolys();
	}
}

bool Seamless::isCtrlPolyTaken(int a, int b, NurbsPatch* patch){
	
	for(int i = 0; i < ctrlPolyPoint.length;){
		int len = ctrlPolyPoint[i].len;
		for(int k = len; --k >= 0;){
			if(ctrlPolyPoint[i + k].p == patch){
				if(ctrlPolyPoint[i + k].i == a){
					while(--k >= 0){
						if(ctrlPolyPoint[i + k].p == patch){
							if(ctrlPolyPoint[i + k].i == b){
								return true;
							}
						}
					}
					break;
				}
				if(ctrlPolyPoint[i + k].i == b){
					while(--k >= 0){
						if(ctrlPolyPoint[i + k].p == patch){
							if(ctrlPolyPoint[i + k].i == a){
								return true;
							}
						}
					}
					break;
				}
			}
		}
		i = getNextCtrlPoly(i);
	}
	return false;
}

bool checkCtrlPoly(M_CtrlPolyPoint& ctrlPolyPoint, int k, int a, int b, NurbsPatch* patch, int ctrlPoly, int len){
	while(--k >= 0){
		if(ctrlPolyPoint[ctrlPoly + k].p == patch && ctrlPolyPoint[ctrlPoly + k].i == b){
			return true;
		}
	}
	ctrlPoly += len;
	for(int k = len + 1; --k >= 0;){
		int midLen = ctrlPolyPoint[ctrlPoly].len;
		if(midLen){
			for(int i = 0; i < midLen; i++){
				if(ctrlPolyPoint[ctrlPoly + i].p == patch && ctrlPolyPoint[ctrlPoly + i].i == b){
					return true;
				}
			}			
		}
		else{
			midLen++;
		}
		ctrlPoly += midLen;
	}

	return false;
}

bool Seamless::isCtrlPolyCornerTaken(int a, int b, NurbsPatch* patch){
	
	for(int i = 0; i < ctrlPolyPoint.length;){
		int len = ctrlPolyPoint[i].len;
		for(int k = len; --k >= 0;){
			if(ctrlPolyPoint[i + k].p == patch){
				if(ctrlPolyPoint[i + k].i == a){
					if(checkCtrlPoly(ctrlPolyPoint, k, a, b, patch, i, len)){
						return true;
					}
					break;
				}
				if(ctrlPolyPoint[i + k].i == b){
					if(checkCtrlPoly(ctrlPolyPoint, k, b, a, patch, i, len)){
						return true;
					}
					break;
				}
			}
		}
		i = getNextCtrlPoly(i);
	}
	return false;
}

void NurbsPatch::addOneByOneCtrlPolyPnt(const int u, const int uDim, CtrlPolyPoint& cpp, M_CtrlPolyPoint& ctrlPolyPoint, M_CtrlPolyPoint& insideOutCtrlPoly){
	int poly = ctrlPolyPoint.length;
	int a = u;
	int b = u + 1;
	int c = u + 1 + uDim;
	int d = u + uDim;
	if(smls->isCtrlPolyTaken(c, d, cpp.p)){
		return;
	}

	cpp.i = a;
	ctrlPolyPoint += cpp;

	cpp.i = b;
	ctrlPolyPoint += cpp;

	cpp.i = c;
	ctrlPolyPoint += cpp;

	cpp.i = d;
	ctrlPolyPoint += cpp;

	ctrlPolyPoint.length += 4 + 1;
	if(isInsideOut()){
		makeCtrlPolyInsideOut(poly, ctrlPolyPoint);
	}
}

void NurbsPatch::addSouthEastCornerOneByOneCtrlPolyPnt(const int u, const int uDim, CtrlPolyPoint& cpp, M_CtrlPolyPoint& ctrlPolyPoint, M_CtrlPolyPoint& insideOutCtrlPoly){
	int poly = ctrlPolyPoint.length;
	int a = u;
	int b = u + 1;
	int c = u + 1 + uDim;
	int d = u + uDim;
	if(smls->isCtrlPolyCornerTaken(d, b, cpp.p)){
		return;
	}

	cpp.i = a;
	ctrlPolyPoint += cpp;

	cpp.i = b;
	ctrlPolyPoint += cpp;

	cpp.i = c;
	ctrlPolyPoint += cpp;

	cpp.i = d;
	ctrlPolyPoint += cpp;

	ctrlPolyPoint.length += 4 + 1;
	if(isInsideOut()){
		makeCtrlPolyInsideOut(poly, ctrlPolyPoint);
	}
}

void NurbsPatch::addNorthEastCornerOneByOneCtrlPolyPnt(const int u, const int uDim, CtrlPolyPoint& cpp, M_CtrlPolyPoint& ctrlPolyPoint, M_CtrlPolyPoint& insideOutCtrlPoly){
	int poly = ctrlPolyPoint.length;
	int a = u;
	int b = u + 1;
	int c = u + 1 + uDim;
	int d = u + uDim;
	if(smls->isCtrlPolyCornerTaken(a, c, cpp.p)){
		return;
	}

	cpp.i = a;
	ctrlPolyPoint += cpp;

	cpp.i = b;
	ctrlPolyPoint += cpp;

	cpp.i = c;
	ctrlPolyPoint += cpp;

	cpp.i = d;
	ctrlPolyPoint += cpp;

	ctrlPolyPoint.length += 4 + 1;
	if(isInsideOut()){
		makeCtrlPolyInsideOut(poly, ctrlPolyPoint);
	}
}

void NurbsPatch::addNorthWestCornerOneByOneCtrlPolyPnt(const int u, const int uDim, CtrlPolyPoint& cpp, M_CtrlPolyPoint& ctrlPolyPoint, M_CtrlPolyPoint& insideOutCtrlPoly){
	int poly = ctrlPolyPoint.length;
	int a = u;
	int b = u + 1;
	int c = u + 1 + uDim;
	int d = u + uDim;
	if(smls->isCtrlPolyCornerTaken(b, d, cpp.p)){
		return;
	}

	cpp.i = a;
	ctrlPolyPoint += cpp;

	cpp.i = b;
	ctrlPolyPoint += cpp;

	cpp.i = c;
	ctrlPolyPoint += cpp;

	cpp.i = d;
	ctrlPolyPoint += cpp;

	ctrlPolyPoint.length += 4 + 1;
	if(isInsideOut()){
		makeCtrlPolyInsideOut(poly, ctrlPolyPoint);
	}
}

void NurbsPatch::addSouthWestCornerOneByOneCtrlPolyPnt(const int u, const int uDim, CtrlPolyPoint& cpp, M_CtrlPolyPoint& ctrlPolyPoint, M_CtrlPolyPoint& insideOutCtrlPoly){
	int poly = ctrlPolyPoint.length;
	int a = u;
	int b = u + 1;
	int c = u + 1 + uDim;
	int d = u + uDim;
	if(smls->isCtrlPolyCornerTaken(c, a, cpp.p)){
		return;
	}

	cpp.i = a;
	ctrlPolyPoint += cpp;

	cpp.i = b;
	ctrlPolyPoint += cpp;

	cpp.i = c;
	ctrlPolyPoint += cpp;

	cpp.i = d;
	ctrlPolyPoint += cpp;

	ctrlPolyPoint.length += 4 + 1;
	if(isInsideOut()){
		makeCtrlPolyInsideOut(poly, ctrlPolyPoint);
	}
}


void NurbsPatch::addTwoByOneCtrlPolyPntSouthMid(const int u, const int uDim, CtrlPolyPoint& cpp, M_CtrlPolyPoint& ctrlPolyPoint, M_CtrlPolyPoint& insideOutCtrlPoly){
	int poly = ctrlPolyPoint.length;
	int a = u;
	int b = u + 2;
	int c = u + 2 + uDim;
	int d = u + uDim;
	if(smls->isCtrlPolyTaken(c, d, cpp.p)){
		return;
	}

	cpp.i = a;
	ctrlPolyPoint += cpp;

	cpp.i = b;
	ctrlPolyPoint += cpp;

	cpp.i = c;
	ctrlPolyPoint += cpp;

	cpp.i = d;
	ctrlPolyPoint += cpp;

	// mid point
	cpp.len = 1;
	cpp.i = u + 1;
	ctrlPolyPoint += cpp;
	cpp.len = 4;

	ctrlPolyPoint.length += 4;
	if(isInsideOut()){
		makeCtrlPolyInsideOut(poly, ctrlPolyPoint);
	}
}

void NurbsPatch::addTwoByOneCtrlPolyPntNorthMid(const int u, const int uDim, CtrlPolyPoint& cpp, M_CtrlPolyPoint& ctrlPolyPoint, M_CtrlPolyPoint& insideOutCtrlPoly){
	int poly = ctrlPolyPoint.length;
	int a = u;
	int b = u - 2;
	int c = u - 2 - uDim;
	int d = u - uDim;
	if(smls->isCtrlPolyTaken(c, d, cpp.p)){
		return;
	}

	cpp.i = a;
	ctrlPolyPoint += cpp;

	cpp.i = b;
	ctrlPolyPoint += cpp;

	cpp.i = c;
	ctrlPolyPoint += cpp;

	cpp.i = d;
	ctrlPolyPoint += cpp;

	// mid point
	cpp.len = 1;
	cpp.i = u - 1;
	ctrlPolyPoint += cpp;
	cpp.len = 4;

	ctrlPolyPoint.length += 4;
	if(isInsideOut()){
		makeCtrlPolyInsideOut(poly, ctrlPolyPoint);
	}
}

void NurbsPatch::addUJoinCtrlPolyPntSouth(const int u, const int uDim, CtrlPolyPoint& cpp, M_CtrlPolyPoint& ctrlPolyPoint, M_CtrlPolyPoint& insideOutCtrlPoly){
	int poly = ctrlPolyPoint.length;
	int a = u + uDim - 2;
	int b = u + 1;
	int c = u + 1 + uDim;
	int d = u + uDim - 2 + uDim;
	if(smls->isCtrlPolyTaken(c, d, cpp.p)){
		return;
	}

	cpp.i = a;
	ctrlPolyPoint += cpp;

	cpp.i = b;
	ctrlPolyPoint += cpp;

	cpp.i = c;
	ctrlPolyPoint += cpp;

	cpp.i = d;
	ctrlPolyPoint += cpp;

	//2 mid point1
	cpp.len = 2;
	cpp.i = u;
	ctrlPolyPoint += cpp;

	cpp.i = uDim - 1;
	ctrlPolyPoint += cpp;

	cpp.len = 4;

	ctrlPolyPoint.length += 4;
	if(isInsideOut()){
		makeCtrlPolyInsideOut(poly, ctrlPolyPoint);
	}
}

void NurbsPatch::addUJoinCtrlPolyPntNorth(const int u, const int uDim, CtrlPolyPoint& cpp, M_CtrlPolyPoint& ctrlPolyPoint, M_CtrlPolyPoint& insideOutCtrlPoly){
	int poly = ctrlPolyPoint.length;
	int a = u - uDim + 2;
	int b = u - 1;
	int c = u - 1 - uDim;
	int d = u - uDim + 2 - uDim;

	if(smls->isCtrlPolyTaken(c, d, cpp.p)){
		return;
	}

	cpp.i = a;
	ctrlPolyPoint += cpp;

	cpp.i = b;
	ctrlPolyPoint += cpp;

	cpp.i = c;
	ctrlPolyPoint += cpp;

	cpp.i = d;
	ctrlPolyPoint += cpp;

	//2 mid point1
	cpp.len = 2;
	cpp.i = u;
	ctrlPolyPoint += cpp;

	cpp.i = u - uDim + 1;
	ctrlPolyPoint += cpp;

	cpp.len = 4;

	ctrlPolyPoint.length += 4;
	if(isInsideOut()){
		makeCtrlPolyInsideOut(poly, ctrlPolyPoint);
	}
}

void NurbsPatch::addOneByTwoCtrlPolyPntEast1Mid(const int u, const int uDim, CtrlPolyPoint& cpp, M_CtrlPolyPoint& ctrlPolyPoint, M_CtrlPolyPoint& insideOutCtrlPoly){
	int poly = ctrlPolyPoint.length;
	int a = u;
	int b = u - uDim * 2;
	int c = u - uDim * 2 + 1;
	int d = u + 1;
	if(smls->isCtrlPolyTaken(c, d, cpp.p)){
		return;
	}

	cpp.i = a;
	ctrlPolyPoint += cpp;

	cpp.i = b;
	ctrlPolyPoint += cpp;

	cpp.i = c;
	ctrlPolyPoint += cpp;

	cpp.i = d;
	ctrlPolyPoint += cpp;

	// mid point

	cpp.len = 1;
	cpp.i = u - uDim;
	ctrlPolyPoint += cpp;
	cpp.len = 4;

	ctrlPolyPoint.length += 4;
	if(isInsideOut()){
		makeCtrlPolyInsideOut(poly, ctrlPolyPoint);
	}
}

void NurbsPatch::addOneByTwoCtrlPolyPntEast0Mid(const int u, const int uDim, CtrlPolyPoint& cpp, M_CtrlPolyPoint& ctrlPolyPoint, M_CtrlPolyPoint& insideOutCtrlPoly){
	int poly = ctrlPolyPoint.length;
	int a = u;
	int b = u - uDim;
	int c = u - uDim + 1;
	int d = u + 1;
	if(smls->isCtrlPolyTaken(c, d, cpp.p)){
		return;
	}

	cpp.i = a;
	ctrlPolyPoint += cpp;

	cpp.i = b;
	ctrlPolyPoint += cpp;

	cpp.i = c;
	ctrlPolyPoint += cpp;

	cpp.i = d;
	ctrlPolyPoint += cpp;

	// mid point

	ctrlPolyPoint.length += 5;

	if(isInsideOut()){
		makeCtrlPolyInsideOut(poly, ctrlPolyPoint);
	}
}

void NurbsPatch::addOneByTwoCtrlPolyPntWest1Mid(const int u, const int uDim, CtrlPolyPoint& cpp, M_CtrlPolyPoint& ctrlPolyPoint, M_CtrlPolyPoint& insideOutCtrlPoly){
	int poly = ctrlPolyPoint.length;
	int a = u;
	int b = u + uDim * 2;
	int c = u + uDim * 2 - 1;
	int d = u - 1;
	if(smls->isCtrlPolyTaken(c, d, cpp.p)){
		return;
	}

	cpp.i = a;
	ctrlPolyPoint += cpp;

	cpp.i = b;
	ctrlPolyPoint += cpp;

	cpp.i = c;
	ctrlPolyPoint += cpp;

	cpp.i = d;
	ctrlPolyPoint += cpp;

	// mid point

	cpp.len = 1;
	cpp.i = u + uDim;
	ctrlPolyPoint += cpp;
	cpp.len = 4;

	ctrlPolyPoint.length += 4;

	if(isInsideOut()){
		makeCtrlPolyInsideOut(poly, ctrlPolyPoint);
	}
}

void NurbsPatch::addOneByTwoCtrlPolyPntWest0Mid(const int u, const int uDim, CtrlPolyPoint& cpp, M_CtrlPolyPoint& ctrlPolyPoint, M_CtrlPolyPoint& insideOutCtrlPoly){
	int poly = ctrlPolyPoint.length;
	int a = u;
	int b = u + uDim;
	int c = u + uDim - 1;
	int d = u - 1;
	if(smls->isCtrlPolyTaken(c, d, cpp.p)){
		return;
	}

	cpp.i = a;
	ctrlPolyPoint += cpp;

	cpp.i = b;
	ctrlPolyPoint += cpp;

	cpp.i = c;
	ctrlPolyPoint += cpp;

	cpp.i = d;
	ctrlPolyPoint += cpp;

	// mid point

	ctrlPolyPoint.length += 5;

	if(isInsideOut()){
		makeCtrlPolyInsideOut(poly, ctrlPolyPoint);
	}
}

void NurbsPatch::makeCtrlPolyInsideOut(const int poly, M_CtrlPolyPoint& ctrlPolyPoint){
	const int uDim = this->dim.x;
	const int vDim = this->dim.y;

	int len = ctrlPolyPoint[poly].len;
	for(int c = 0; c < len; c++){
		int idx = ctrlPolyPoint[poly + c].i;
		int u = idx % uDim;
		int v = idx / uDim;
		u = uDim - 1 - u;//mirror
		ctrlPolyPoint[poly + c].i = v * uDim + u;
	}
	int midIdx = poly + len;
	for(int c = 0; c < len; c++){
		int len = ctrlPolyPoint[midIdx].len;
		if(len == 0){
			midIdx++;
			continue;
		}
		for(int c = 0; c < len; c++){
			int idx = ctrlPolyPoint[midIdx].i;
			int u = idx % uDim;
			int v = idx / uDim;
			u = uDim - 1 - u;//mirror
			ctrlPolyPoint[midIdx++].i = v * uDim + u;
		}
	}
}


void makeCtrlPolyInsideOut(const int poly, M_CtrlPolyPoint& ctrlPolyPoint, M_CtrlPolyPoint& insideOutCtrlPoly){
	insideOutCtrlPoly.length = 0;
	const int len = ctrlPolyPoint[poly].len;
	for(int c = 0; c < len; c++){
		insideOutCtrlPoly += ctrlPolyPoint[poly + c];
	}
	int midIdx = poly + len;
	for(int c = 0; c < len; c++){
		int ln = ctrlPolyPoint[midIdx].len;
		int rLen = ln;
		if(ln == 0){
			ln = 1;
		}
		for(int c = 0; c < ln; c++){
			ctrlPolyPoint[midIdx].len = rLen;
			insideOutCtrlPoly += ctrlPolyPoint[midIdx++];
		}
	}
	// copy back inside out poly
	int i = poly;
	for(int r = len; --r >= 0;){
		int corner = r + 1;
		if(corner > len - 1){
			corner = 0;
		}
		ctrlPolyPoint[i++] = insideOutCtrlPoly[corner];
	}
	for(int r = insideOutCtrlPoly.length; --r >= len;){
		ctrlPolyPoint[i++] = insideOutCtrlPoly[r];
	}

}

void invertSelCtrlPoly(){
	if(NurbsPatch* patch = dynamic_cast<NurbsPatch*>(g_selNode.getRaw())){
		M_CtrlPolyPoint insideOutCtrlPoly;
		Seamless* smls = patch->smls;
		M_CtrlPolyPoint& ctrlPolyPoint = smls->ctrlPolyPoint;
		M_Link& link = smls->link;

		for(int i = MultiSelect::ctrlPolySelList.length; --i >= 0;){
			int poly = MultiSelect::ctrlPolySelList[i];
			makeCtrlPolyInsideOut(poly, smls->ctrlPolyPoint, insideOutCtrlPoly);
			for(int j = link.length; --j >= 0;){
				if(poly == link[j].poly){
					int len = ctrlPolyPoint[poly].len;
					int corner = len - link[j].corner - 1;
					//if(corner < 0){
					//	corner = len - 1;
					//}
					link[j].corner = corner;
					link[j].between = 1.0 - link[j].between;
				}
			}
		}

		patch->houseKeep();
		BackUp::monostable.beginCycle();
	}
	else if(Oscillator* osc = dynamic_cast<Oscillator*>(g_selNode.getRaw())){
		osc->alignWithPatch();
	}
}


void NurbsPatch::fillEdgesWithCtrlPolys(){
	const int uDim = this->dim.x;
	const int vDim = this->dim.y;
	M_CtrlPolyPoint ctrlPolyPoint;
	M_CtrlPolyPoint insideOutCtrlPoly;

	CtrlPolyPoint cpp;
	cpp.len = 4;
	cpp.p = this;

	int u = 0;
	int q = uDim * (vDim - 2);

	int vMid = checkNurbsType(vDim, vKnot);

	if(uJoinEnds){
		if(beginSP == false){
			addUJoinCtrlPolyPntSouth(u, uDim, cpp, ctrlPolyPoint, insideOutCtrlPoly);
		}
		if(endSP == false){
			addUJoinCtrlPolyPntNorth(uDim * vDim - 1, uDim, cpp, ctrlPolyPoint, insideOutCtrlPoly);
		}
	}
	else{
		if(beginSP == false){
			addSouthWestCornerOneByOneCtrlPolyPnt(u, uDim, cpp, ctrlPolyPoint, insideOutCtrlPoly);
		}
		if(endSP == false){
			addNorthWestCornerOneByOneCtrlPolyPnt(q + u, uDim, cpp, ctrlPolyPoint, insideOutCtrlPoly);
		}
	}

	for(u = 1; u < uDim - 2; u += 2){
		if(beginSP == false){
			addTwoByOneCtrlPolyPntSouthMid(u, uDim, cpp, ctrlPolyPoint, insideOutCtrlPoly);
		}
		if(endSP == false){
			addTwoByOneCtrlPolyPntNorthMid(q + uDim + u + 2, uDim, cpp, ctrlPolyPoint, insideOutCtrlPoly);
		}
	}

	if(uJoinEnds){
	}
	else{
		if(beginSP == false){
			addSouthEastCornerOneByOneCtrlPolyPnt(u, uDim, cpp, ctrlPolyPoint, insideOutCtrlPoly);
		}
		if(endSP == false){
			addNorthEastCornerOneByOneCtrlPolyPnt(q + u, uDim, cpp, ctrlPolyPoint, insideOutCtrlPoly);
		}
		if(NurbsPatch::MID_0 == vMid){
			for(int v = 1; v < vDim - 2; v++){
				addOneByTwoCtrlPolyPntEast0Mid((v + 1) * uDim, uDim, cpp, ctrlPolyPoint, insideOutCtrlPoly);
				addOneByTwoCtrlPolyPntWest0Mid(v * uDim + uDim - 1, uDim, cpp, ctrlPolyPoint, insideOutCtrlPoly);
			}
		}
		else{
			for(int v = 1; v < vDim - 2; v += 2){
				addOneByTwoCtrlPolyPntEast1Mid((v + 2) * uDim, uDim, cpp, ctrlPolyPoint, insideOutCtrlPoly);
				addOneByTwoCtrlPolyPntWest1Mid(v * uDim + uDim - 1, uDim, cpp, ctrlPolyPoint, insideOutCtrlPoly);
			}
		}
	}
	smls->ctrlPolyPoint += ctrlPolyPoint;
	smls->calcAllPolyMidPoints();
	houseKeepVisualClue();
	BackUp::monostable.beginCycle();
}

/*
int Seamless::findClosestMidPoint(const M_Int& ctrlPolySelList, const Vec3f selCorner, int& corner, int& blockIdx){//returns poly
	int poly = -1;
	int midIdx = 0;
	float shortest = 1000000.0f;
	for(int i = 0; i < ctrlPolyPoint.length;){
		int len = ctrlPolyPoint[i].len;
		int m = i + len;
		for(int c = 0; c <len; c++){
			int midLen = ctrlPolyPoint[m].len;
			for(int q = 0; q < midLen; q++){
				NurbsPatch* patch = ctrlPolyPoint[m].p;
				int idx = ctrlPolyPoint[m].i;
				if(m != blockIdx){

					Vec3f vec = patch->dpoint[idx];
					
					float distance = (selCorner - vec).length();
					if(distance < shortest){
						//MultiSelect::pointList += ControlPoint(ctrlPolyPoint[m].p, ctrlPolyPoint[m].i);
						shortest = distance;
						midIdx = m;
						poly = i;
						corner = c;
					}
					m++;
				}
			}
		}
		i = getNextCtrlPoly(i);
	}
	blockIdx = midIdx;
	return poly;
}
*/
int Seamless::findClosestMidPoint(const M_Int& ctrlPolySelList, const Vec3f selCorner, int& corner, int& blockIdx){//returns poly
	int poly = -1;
	int midIdx = 0;
	float shortest = 1000000.0f;

	for(int q = ctrlPolySelList.length; --q>=0;){
		int i = ctrlPolySelList[q];
		int len = ctrlPolyPoint[i].len;
		int m = i + len;
		for(int c = 0; c <len; c++){
			int midLen = ctrlPolyPoint[m].len;
			if(midLen){
				for(int q = 0; q < midLen; q++){
					NurbsPatch* patch = ctrlPolyPoint[m].p;
					int idx = ctrlPolyPoint[m].i;
					if(m != blockIdx){
						Vec3f vec = patch->dpoint[idx];
						float distance = (selCorner - vec).length();
						if(distance < shortest){
							//MultiSelect::pointList += ControlPoint(ctrlPolyPoint[m].p, ctrlPolyPoint[m].i);
							shortest = distance;
							midIdx = m;
							poly = i;
							corner = c;
						}
					}
				}
			}
			else{
				midLen++;
			}
			m += midLen;
		}
	}
	blockIdx = midIdx;
	return poly;
}

int Seamless::getNurbsCorner(int& c0, M_Int& ctrlPolySelList){//return poly

	for(int q = ctrlPolySelList.length; --q>=0;){
		int i = ctrlPolySelList[q];
		int len = ctrlPolyPoint[i].len;

		for(int c = 0; c < len; c++){
			NurbsPatch* patch = ctrlPolyPoint[i + c].p;
			int uDim = patch->dim.x;
			int vDim = patch->dim.y;
			int idx = ctrlPolyPoint[i + c].i;
			int u = idx % uDim;
			int v = idx / uDim;
			if(u == 0 || u == uDim - 1){
				if(v == 0 || v == vDim - 1){
					c0 = c;
					return i;
				}
			}
			else if(v == 0 || v == vDim - 1){
				if(u == 0 || u == uDim - 1){
					c0 = c;
					return i;
				}
			}
		}
	}
	return -1;
}

void Seamless::mergeTo5CtrlPoly2(int polyC, int c0, int polyA, int a0, int a1, int polyB, int b0, int b1){
	int c2 = getCorner(c0, 2, 4);
	//MultiSelect::pointList += ControlPoint(ctrlPolyPoint[polyC + c2].p, ctrlPolyPoint[polyC + c2].i);
	ctrlPolyPoint += CtrlPolyPoint(ctrlPolyPoint[polyC + c2].p, ctrlPolyPoint[polyC + c2].i, 5);
	int b2 = getCorner(b1, 1, 4);
	//MultiSelect::pointList += ControlPoint(ctrlPolyPoint[polyB + b2].p, ctrlPolyPoint[polyB + b2].i);
	ctrlPolyPoint += CtrlPolyPoint(ctrlPolyPoint[polyB + b2].p, ctrlPolyPoint[polyB + b2].i, 5);
	int b3 = getCorner(b2, 1, 4);
	//MultiSelect::pointList += ControlPoint(ctrlPolyPoint[polyB + b3].p, ctrlPolyPoint[polyB + b3].i);
	ctrlPolyPoint += CtrlPolyPoint(ctrlPolyPoint[polyB + b3].p, ctrlPolyPoint[polyB + b3].i, 5);
	int a2 = getCorner(a1, 1, 4);
	//MultiSelect::pointList += ControlPoint(ctrlPolyPoint[polyA + a2].p, ctrlPolyPoint[polyA + a2].i);
	ctrlPolyPoint += CtrlPolyPoint(ctrlPolyPoint[polyA + a2].p, ctrlPolyPoint[polyA + a2].i, 5);
	int a3 = getCorner(a2, 1, 4);
	//MultiSelect::pointList += ControlPoint(ctrlPolyPoint[polyA + a3].p, ctrlPolyPoint[polyA + a3].i);
	ctrlPolyPoint += CtrlPolyPoint(ctrlPolyPoint[polyA + a3].p, ctrlPolyPoint[polyA + a3].i, 5);

	int c3 = getCorner(c2, 1, 4);
	ctrlPolyPoint += CtrlPolyPoint(ctrlPolyPoint[polyC + c3].p, ctrlPolyPoint[polyC + c3].i, 2);
	ctrlPolyPoint += CtrlPolyPoint(ctrlPolyPoint[polyB + b1].p, ctrlPolyPoint[polyB + b1].i, 2);
	ctrlPolyPoint.length++;
	ctrlPolyPoint += CtrlPolyPoint(ctrlPolyPoint[polyB + b0].p, ctrlPolyPoint[polyB + b0].i, 2);
	ctrlPolyPoint += CtrlPolyPoint(ctrlPolyPoint[polyA + a1].p, ctrlPolyPoint[polyA + a1].i, 2);
	ctrlPolyPoint.length++;
	ctrlPolyPoint += CtrlPolyPoint(ctrlPolyPoint[polyA + a0].p, ctrlPolyPoint[polyA + a0].i, 2);
	int c1 = getCorner(c0, 1, 4);
	ctrlPolyPoint += CtrlPolyPoint(ctrlPolyPoint[polyC + c1].p, ctrlPolyPoint[polyC + c1].i, 2);

	int midB = getMidOffset(polyB, b0);
	int midA = getMidOffset(polyA, a0);

	int centerLen = ctrlPolyPoint[midB].len + ctrlPolyPoint[midA].len + 1;
	ctrlPolyPoint += CtrlPolyPoint(ctrlPolyPoint[polyC + c0].p, ctrlPolyPoint[polyC + c0].i, centerLen);
	for(int i = 0; i < ctrlPolyPoint[midB].len; i++){
		ctrlPolyPoint += CtrlPolyPoint(ctrlPolyPoint[midB + i].p, ctrlPolyPoint[midB + i].i, centerLen);
	}
	for(int i = 0; i < ctrlPolyPoint[midA].len; i++){
		ctrlPolyPoint += CtrlPolyPoint(ctrlPolyPoint[midA + i].p, ctrlPolyPoint[midA + i].i, centerLen);
	}
}

int Seamless::getMidOffset(int poly, int corner){

	int m = poly + ctrlPolyPoint[poly].len;

	for(int c = 0; c < corner; c++){
		int midLen = ctrlPolyPoint[m].len;
		if(midLen == 0){
			m++;
			continue;
		}
		for(int k = 0; k < midLen; k++){
			m++;
		}
	}
	return m;
}

void Seamless::mergeTo6CtrlPoly(){
	M_Int deleteCtrlPolySelList;
	M_Int m_corner;

	findCloseCorner(MultiSelect::ctrlPolySelList, deleteCtrlPolySelList, m_corner);

	//ctrlPolyPoint.length += 1;

	bubbleSort(deleteCtrlPolySelList);
	deleteCtrlPoly(deleteCtrlPolySelList);

	commonEndSitchPolys();
}

void Seamless::mergeTo5CtrlPoly(){
	int c0 = 0;
	int polyC = getNurbsCorner(c0, MultiSelect::ctrlPolySelList);
	//MultiSelect::pointList += ControlPoint(ctrlPolyPoint[polyC + c0].p, ctrlPolyPoint[polyC + c0].i);

	int a0 = 0;
	int b0 = 0;

	NurbsPatch* blockPatch = 0;
	int blockIdxA = -1;

	
	Vec3f selVec = ctrlPolyPoint[polyC + c0].p->dpoint[ctrlPolyPoint[polyC + c0].i];

	int polyA = findClosestMidPoint(MultiSelect::ctrlPolySelList, selVec, a0, blockIdxA);
	int blockIdxB = blockIdxA;
	//MultiSelect::pointList += ControlPoint(ctrlPolyPoint[blockIdxA].p, ctrlPolyPoint[blockIdxA].i);
	int polyB = findClosestMidPoint(MultiSelect::ctrlPolySelList, selVec, b0, blockIdxB);
	//MultiSelect::pointList += ControlPoint(ctrlPolyPoint[blockIdxB].p, ctrlPolyPoint[blockIdxB].i);

	int a1 = getCorner(a0, 1, 4);
	int b1 = getCorner(b0, 1, 4);

	Vec3f nearA = ctrlPolyPoint[polyA + a1].p->dpoint[ctrlPolyPoint[polyA + a1].i];
	Vec3f nearB = ctrlPolyPoint[polyB + b1].p->dpoint[ctrlPolyPoint[polyB + b1].i];

	//MultiSelect::pointList += ControlPoint(ctrlPolyPoint[polyA + a0].p, ctrlPolyPoint[polyA + a0].i);
	//MultiSelect::pointList += ControlPoint(ctrlPolyPoint[polyA + a1].p, ctrlPolyPoint[polyA + a1].i);
//	MultiSelect::pointList += ControlPoint(ctrlPolyPoint[polyB + b0].p, ctrlPolyPoint[polyB + b0].i);

	//MultiSelect::pointList += ControlPoint(ctrlPolyPoint[polyA + a1].p, ctrlPolyPoint[polyA + a1].i);
	//MultiSelect::pointList += ControlPoint(ctrlPolyPoint[polyB + b1].p, ctrlPolyPoint[polyB + b1].i);

	int c3 = getCorner(c0, 3, 4);
	Vec3f selVec3 = ctrlPolyPoint[polyC + c3].p->dpoint[ctrlPolyPoint[polyC + c3].i];
	//MultiSelect::pointList += ControlPoint(ctrlPolyPoint[polyC + c3].p, ctrlPolyPoint[polyC + c3].i);

	if((selVec3 - nearA).length() > (selVec3 - nearB).length()){
		mergeTo5CtrlPoly2(polyC, c0, polyA, a0, a1, polyB, b0, b1);
	}
	else{
		mergeTo5CtrlPoly2(polyC, c0, polyB, b0, b1, polyA, a0, a1);
	}
	M_Int m_poly = M_Int(polyA, polyB, polyC);

	bubbleSort(m_poly);
	deleteCtrlPoly(m_poly);

	commonEndSitchPolys();
}

Vec2f convertPointToScreen2d(Vec3f point);
bool triIntersectsRect(Vec2f a, Vec2f b, Vec2f c);
bool isRectCornerInTriangle(const Vec2f* tri);
bool backFace(const Vec2f& v0, const Vec2f& v1, const Vec2f& v2);

void Seamless::selectCtrlPolys(){
	g_viewMatrix = g_scene->getViewMatrix();
	for(int i = 0; i < ctrlPolyPoint.length;){
		Vec3f va3f = ctrlPolyPoint[i].p->dpoint[ctrlPolyPoint[i].i];
		int len = ctrlPolyPoint[i].len;
		for(int j = 1; j < len - 1; j++){
			Vec3f vb3f = ctrlPolyPoint[i + j].p->dpoint[ctrlPolyPoint[i + j].i];
			Vec3f vc3f = ctrlPolyPoint[i + j + 1].p->dpoint[ctrlPolyPoint[i + j + 1].i];
		
			Vec2f va = convertPointToScreen2d(va3f);
			Vec2f vb = convertPointToScreen2d(vb3f);
			Vec2f vc = convertPointToScreen2d(vc3f);

			if(MultiSelect::backfaceIgnore){
				if(!backFace(va, vb, vc)){
					continue;
				}
			}

			Vec2f tri2d[3] = {va, vb, vc};

			if(MultiSelect::isInBox(va) || MultiSelect::isInBox(vb) || MultiSelect::isInBox(vc) || triIntersectsRect(va, vb, vc) || isRectCornerInTriangle(tri2d)){
				MultiSelect::ctrlPolySelList += i;
				break;
			}
		}
		i = getNextCtrlPoly(i);
	}

	//order list incremntaly and remove any double selections

	M_Bool sortList;
	sortList.length = ctrlPolyPoint.length;
	for(int k = MultiSelect::ctrlPolySelList.length; --k>=0;){
		sortList[MultiSelect::ctrlPolySelList[k]] = true;
	}
	MultiSelect::ctrlPolySelList.length = 0;
	for(int k = 0; k < sortList.length; k++){
		if(sortList[k]){
			MultiSelect::ctrlPolySelList+=k;
		}
	}
}

void Seamless::calcAllLinks(){
	for(int k = link.length; --k>=0;){
		Link lnk = link[k];
		int idxA = lnk.poly + lnk.corner;
		Vec3f vecA = ctrlPolyPoint[idxA].p->dpoint[ctrlPolyPoint[idxA].i];
		int idxB = lnk.poly + getCorner(lnk.corner, 1, ctrlPolyPoint[lnk.poly].len);
		Vec3f vecB = ctrlPolyPoint[idxB].p->dpoint[ctrlPolyPoint[idxB].i];
		Vec3f linkVec = vecA + (vecB - vecA) * lnk.between;
		lnk.patch->point.refValue()[lnk.point] = linkVec * lnk.patch->weight[lnk.point];
	}
}

void Seamless::linkPointToEdge(int poly){
	NurbsPatch* patch = MultiSelect::pointList[0].p;
	int idx = MultiSelect::pointList[0].i;
	Vec3f selVec = patch->dpoint[idx];
	int len = ctrlPolyPoint[poly].len;
	float shortest = 10000000;
	Vec3f shortestEdgePoint;
	int shortestCorner = 0;
	float shortestRatio = 0;
	for(int a = 0; a < len; a++){
		Vec3f vecA = ctrlPolyPoint[poly + a].p->dpoint[ctrlPolyPoint[poly + a].i];
		int b = getCorner(a, 1, len);
		Vec3f vecB = ctrlPolyPoint[poly + b].p->dpoint[ctrlPolyPoint[poly + b].i];
		float edgeDistance = (vecB - vecA).length();
		float selDistance = (selVec - vecA).length();
		float ratio = selDistance / edgeDistance;

		Vec3f edgePoint = vecA + (vecB - vecA) * ratio;
		float diffDistance = (edgePoint - selVec).length();
		if(diffDistance < shortest){
			shortest = diffDistance;
			shortestEdgePoint = edgePoint;
			shortestCorner = a;
			shortestRatio = ratio;
		}
	}
	patch->point.refValue()[idx] = shortestEdgePoint * patch->weight[idx];
	link += Link(patch, idx, poly, shortestCorner, shortestRatio);

	MultiSelect::pointList.length = 0;
	BackUp::monostable.beginCycle();
	calcAllPolyMidPoints();
	houseKeepVisualClue();
}

void Seamless::deleteLinksToPolys(M_Int& deletePolyList){
	if(deletePolyList.length == 0){
		return;
	}
	M_Int redirectCtrlPolyPoint;
	redirectCtrlPolyPoint.length = ctrlPolyPoint.length;

	for(int k = 0; k < deletePolyList.length; k++){
		int from = deletePolyList[k];
		int to = getNextCtrlPoly(from);
		for(int i = from; i < to; i++){
			redirectCtrlPolyPoint[i] = -1;
		}
	}

	for(int i = 0, cntr = -1; i < redirectCtrlPolyPoint.length; i++){
		if(redirectCtrlPolyPoint[i] >= 0){
			cntr++;
			redirectCtrlPolyPoint[i] = cntr;
		}
	}

	int d = 0;
	for(int s = 0; s < link.length; s++){
		int poly = link[s].poly;
		int i = redirectCtrlPolyPoint[poly];
		if(i == -1){
			continue;
		}
		link[d] = link[s];
		link[d++].poly = i;
	}
	link.length = d;
}

void Seamless::deleteLinksToNonExistentPatches(){
	int d = 0;
	for(int s = 0; s < link.length; s++){
		NurbsPatch* patch = link[s].patch;
		if(patch->plidx >= cpNodeList.length || cpNodeList[patch->plidx] != patch){
			continue;
		}
		link[d++] = link[s];
	}
	link.length = d;
}

bool Seamless::pointsExist(M_Int& deletePolyList, int poly, int& m, int len){
	for(int k = len; --k >= 0;){
		NurbsPatch* patch = ctrlPolyPoint[m++].p;
		if(patch->plidx >= cpNodeList.length || cpNodeList[patch->plidx] != patch){
			deletePolyList += poly;
			return false;
		}
	}
	return true;
}

void Seamless::deletePolysWithNonExistingPoints(){
	M_Int deletePolyList;
	int ctrlPolyPointLen = ctrlPolyPoint.length;

	for(int i = 0; i < ctrlPolyPointLen;){
		int poly = i;
		int m = poly;
		int len = ctrlPolyPoint[poly].len;
		if(false == pointsExist(deletePolyList, poly, m, len)){
			i = getNextCtrlPoly(poly);
			continue;
		}

		for(int o = len + 1; --o>=0;){
			int len = ctrlPolyPoint[m].len;
			if(len == 0){
				m++;
				continue;
			}
			if(false == pointsExist(deletePolyList, poly, m, len)){
				i = m = getNextCtrlPoly(poly);
				break;
			}
		}
		i = m;
	}

	deleteCtrlPoly(deletePolyList);
}
