/* Seamless3d

Copyright (C) 2001-2010 Graham John Perrett (AKA thyme) thyme@seamless3d.com www.seamless3d.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this
software and associated documentation files (the "Software"), to deal in the Software
without restriction, including without limitation the rights to use, copy, modify,
merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to the following
conditions:

The above copyright notice and this permission notice shall be included in all copies or
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT
OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.
*/

#include <windows.h>
#include <direct.h>

#include "seamless3d.h"
#include "techuelife_triangle_editor.h"
#include "seamless3d_gui_render.h"
#include "gui.h"
#include "scene_tree.h"
#include "robot_aid.h"


bool backFace(const Vec2f& v0, const Vec2f& v1, const Vec2f& v2);

void Subdivision::selNextPoly(){
	M_Int& coordIndex = this->coordIndex;
	actPoly++;
	if(actPoly*4>=coordIndex.length){
		actPoly=coordIndex.length/4-1;
	}
	actCorner=0;
	index[0]=coordIndex[actPoly*4+actCorner];
	houseKeep();
}
void Subdivision::selPrevPoly(){
	M_Int& coordIndex = this->coordIndex;
	actPoly--;
	if(--actPoly<0){
		actPoly=0;
	}
	actCorner=0;
	index[0]=0;
	houseKeep();
}
void Subdivision::delPoly(){
	M_Int&	coordIndex = this->coordIndex;
	if(coordIndex.length==0) return;

	int selPolyLen = 4;

	int len = coordIndex.length-selPolyLen;

	coordIndex.deleteElement(actPoly*4, selPolyLen);
	deleteNonIndexedPoints();

	processFields();
	if(actPoly>=polyIndex.length){
		actPoly--;
	}

	BackUp::monostable.beginCycle();
	if(part.p){
		Seamless* smls = part->smls;
		smls->doBuild();
		smls->ctrf=true;
	}
	processFields();
	makePointerEqualIndexLocation();
	updateBuild3dWindow();
	houseKeepVisualClue();
	BackUp::monostable.beginCycle();
}
void Subdivision::incToNextPolyThatSharesActivePoint(){
	M_Int&	coordIndex = this->coordIndex;
	if(coordIndex.length==0) return;

	int actCtrlPnt = index[0];

	for(int j = polyIndex.length - 1, i = actPoly + 1; --j >=0; i++){
		if(i >= polyIndex.length){
			i = 0;
		}
		for(int q = polyIndex[i].len; --q >= 0;){
			if(coordIndex[polyIndex[i].idx + q] == actCtrlPnt){
				actPoly = i;
				break;
			}
		}
	}
	makePointerEqualIndexLocation();
	houseKeep();
}
void Subdivision::decToNextPolyThatSharesActivePoint(){
	M_Int&	coordIndex = this->coordIndex;
	if(coordIndex.length==0) return;

	int actCtrlPnt = index[0];

	for(int j = polyIndex.length - 1, i = actPoly - 1; --j >=0; i--){
		if(i < 0){
			i = polyIndex.length - 1;
		}
		for(int q = polyIndex[i].len; --q >= 0;){
			if(coordIndex[polyIndex[i].idx + q] == actCtrlPnt){
				actPoly = i;
				break;
			}
		}
	}
	makePointerEqualIndexLocation();
	houseKeep();
}
void Subdivision::mirrorPoly(){
	M_Int&	coordIndex = this->coordIndex;
	if(coordIndex.length==0) return;

	M_Vec3f& point = this->point.refValue();

	point.length+=2;
	Vec3f& newCoordA = point[point.length-2];
	Vec3f& newCoordB = point[point.length-1];


	int a = actCorner;  //abc=triangle we r mirroring;
	int b = (actCorner+1)%4;
	int c = (actCorner+2)%4;
	int d = (actCorner+3)%4;

	int idx = polyIndex[actPoly].idx;

	int ia = coordIndex[idx + a];
	int ib = coordIndex[idx + b];
	int ic = coordIndex[idx + c];
	int id = coordIndex[idx + d];

	Vec3f vecA = point[ia];
	Vec3f vecB = point[ib];
	Vec3f vecC = point[ic];
	Vec3f vecD = point[id];
	
//	newCoordA = vecB + (vecA - vecC);
	newCoordB = vecA + (vecB - vecD);
	newCoordA = newCoordB + (vecA - vecB);


//	newCoordB = vecA + (newCoordA - vecC);
//	newCoordB = vecA + (vecC - newCoordA);
//	newCoordA = newCoordA;
//	newCoordA.x-=.1;

	coordIndex += point.length-2;
	coordIndex += point.length-1;
	coordIndex += ib;
	coordIndex += ia;

	processFields();

	processFields();
	makePointerEqualIndexLocation();
	updateBuild3dWindow();
	houseKeepVisualClue();
	BackUp::monostable.beginCycle();
}
void Subdivision::extrudePoly(){
	M_Int& coordIndex = this->coordIndex;
	M_Vec3f& point = this->point; 

	coordIndex+=coordIndex[actPoly*4+0];
	coordIndex+=coordIndex[actPoly*4+1];
	coordIndex+=point.length+1;
	coordIndex+=point.length+0;

	coordIndex+=coordIndex[actPoly*4+1];
	coordIndex+=coordIndex[actPoly*4+2];
	coordIndex+=point.length+2;
	coordIndex+=point.length+1;

	coordIndex+=coordIndex[actPoly*4+2];
	coordIndex+=coordIndex[actPoly*4+3];
	coordIndex+=point.length+3;
	coordIndex+=point.length+2;

	coordIndex+=coordIndex[actPoly*4+3];
	coordIndex+=coordIndex[actPoly*4+0];
	coordIndex+=point.length+0;
	coordIndex+=point.length+3;

	for(int i=0; i<4;  i++){
		point+=point[coordIndex[actPoly*4+i]];
		coordIndex[actPoly*4+i]=point.length-1;
	}

/*
	for(int i=(actPoly+1)*4; i<coordIndex.length;  i++){
		coordIndex[i-4] = coordIndex[i];
	}
	coordIndex.length-=4;
*/
	processFields();
	makePointerEqualIndexLocation();
	updateBuild3dWindow();
	houseKeepVisualClue();
	BackUp::monostable.beginCycle();
}
bool Subdivision::pointIsInPoly(){
	const int actCtrlPnt = index[0];
	int idx = polyIndex[actPoly].idx;
	int len = polyIndex[actPoly].len;

	for(int q = len; --q >= 0;){
		if(coordIndex[idx + q] == actCtrlPnt){
			actCorner = q;
			return true;
		}
	};
	return false;
}

void Subdivision::makePointerEqualIndexLocation(){
	Art* part=this->part.refValue();
	if(part==0)return;
	const M_Vec3f& coord=this->point;

	int& actCtrlPnt=this->index[0];

	if((unsigned)actCtrlPnt>=(unsigned)coord.length){
		actCtrlPnt=coord.length-1;
	}
	if(pointIsInPoly()){
		return;
	}
	for(int i=coordIndex.length; --i>=0;){
		if(coordIndex[i]==actCtrlPnt){
			actPoly=i/4;
			actCorner=i%4;
		}
	}
}
inline Vec2f get2dCoord(const Subdivision* subdiv, const int& i, const int& perspective){
	Vec3f crd3f = g_scene->getViewMatrix() * subdiv->point[i];
	Vec2f crd;
	if(perspective){
		crd=Vec2f(
			crd3f.x/(-crd3f.z + g_viewDist),
			crd3f.y/(-crd3f.z + g_viewDist)
		);
	}
	else{
		crd=Vec2f(crd3f.x, crd3f.y);
	}
	return crd;
}
bool Subdivision::getClosestVertex(int& dstIndex, int& dstTriangle, int& dstCorner){
	bool found = false;
	M_Vec2f v;
	Art* part=this->part.refValue();
	if(part==0)return -1;
	const int actCtrlPnt = index[0];

	const bool perspective = g_settings->perspective;
	Vec2f selCoord = get2dCoord(this, actCtrlPnt, perspective);

	float minDist = 10000000;

	for(int i = polyIndex.length; --i>=0;){
		int idx = polyIndex[i].idx;
		int len = polyIndex[i].len;
		v.length = len;

		for(int q = len; --q >= 0;){
			v[q] = get2dCoord(this, coordIndex[idx + q], perspective);
		};
		for(int q = polyIndex[i].len; --q >= 0;){
			if(false == backFace(v[q], v[(q + 1) % len], v[(q + len - 1) % len])){
				continue;
			}
			float dist = (v[q] - selCoord).length();


			if(dist < minDist && coordIndex[idx + q] != actCtrlPnt){
				dstCorner = q;
				minDist = dist;
				dstIndex = coordIndex[idx + q];
				dstTriangle = i;
				found = true;
			}
		}
	}
	return found;
}
void Subdivision::deleteNonIndexedPoints(){
	M_Vec3f& point = this->point.refValue();
	M_Int& coordIndex = this->coordIndex.refValue();
	M_Int lookUp;
	lookUp.length = point.length;

	for(int i = coordIndex.length; --i>=0;){
		lookUp[coordIndex[i]] = 1;
	}
	int g = -1;
	for(int i = 0; i < lookUp.length; i++){
		if(lookUp[i] == 1){
			g++;
			point[g] = point[i];
		}
		lookUp[i] = g;
	}
	if(g==-1){
		point.length = 0;
		return;
	}

	point.length = g + 1;

	for(int i = coordIndex.length; --i>=0;){
		coordIndex[i] = lookUp[coordIndex[i]];
	}
}
void Subdivision::joinVertex(){
	M_Vec3f& point = this->point.refValue();
	M_Int& coordIndex = this->coordIndex.refValue();
	const int actCtrlPnt = index[0];
	int dstIndex = 0;
	int dstTriangle = 0;
	int dstCorner = 0;

	if(getClosestVertex(dstIndex, dstTriangle, dstCorner)){
		for(int i = coordIndex.length; --i>=0;){
			if(coordIndex[i] == actCtrlPnt){
				coordIndex[i] = dstIndex;
			}
		}
		processFields();
		deleteNonIndexedPoints();
		processFields();
		makePointerEqualIndexLocation();
		updateBuild3dWindow();
		houseKeepVisualClue();
		BackUp::monostable.beginCycle();
	}
}
void Subdivision::joinCloseVertices(){
	M_Vec3f& point = this->point.refValue();
	M_Int& coordIndex = this->coordIndex.refValue();
	M_Int redirect;
	M_Bool mask;
	redirect.length = point.length;
	mask.length = point.length;


	float range = .001;

	for(int d=point.length;--d>=0;){
		if(mask[d] || redirect[d] == -1){
			continue;
		}
		for(int s=point.length;--s>=0;){
			if(mask[s] || redirect[s] == -1 || s == d){
				continue;
			}
			Vec3f dstVec=point[d];
			Vec3f srcVec=point[s];
			if( fabs(dstVec.x-srcVec.x)<range&&
				fabs(dstVec.y-srcVec.y)<range&&
				fabs(dstVec.z-srcVec.z)<range
				){
				redirect[s]=d;
				redirect[d]=-1;
				mask[s]=true;
			}
		}
	}
	for(int i = coordIndex.length; --i>=0;){
		int ci = coordIndex[i];
		if(mask[ci]){
			coordIndex[i] = redirect[ci];
		}
	}
	deleteNonIndexedPoints();
	processFields();
	makePointerEqualIndexLocation();
	updateBuild3dWindow();
	houseKeepVisualClue();
	BackUp::monostable.beginCycle();
}
inline bool isAllLeftOfSelection(const Vec3f& selCoord, const int& i,  const M_Vec3f& point, const M_<PolyIndex>& polyIndex, const M_Int& coordIndex){
	for(int q = polyIndex[i].len; --q >= 0;){
		if(point[coordIndex[polyIndex[i].idx + q]].x > selCoord.x){
			return false;
		}
	}
	return true;
}
void Subdivision::cutLeft(){
	M_Vec3f& point = this->point.refValue();
	M_Int& coordIndex = this->coordIndex.refValue();

	int actCtrlPnt = index[0];

	if(actCtrlPnt >= point.length){
		return;
	}

	Vec3f selCoord = point[actCtrlPnt];

	int g = 0;
	for(int i = 0; i < polyIndex.length; i++){
		polyIndex[g]=polyIndex[i];
		if(false == isAllLeftOfSelection(selCoord, i,  point, polyIndex, coordIndex)){
			g++;
		}
	}
	polyIndex.length = g;

	int j = 0;
	for(int i = 0; i < polyIndex.length; i++){
		for(int q = 0; q < polyIndex[i].len; q++){
			coordIndex[j++] = coordIndex[polyIndex[i].idx + q];
		}
	}
	coordIndex.length = j;

	deleteNonIndexedPoints();
	updateBuild3dWindow();
	houseKeepVisualClue();
	BackUp::monostable.beginCycle();
}
void Subdivision::mirrorControlPoints(){
	M_Vec3f& point = this->point.refValue();
	M_Int& coordIndex = this->coordIndex.refValue();

	int halfPointLen = point.length;
	int halfPolyIndexLen = polyIndex.length;
	point += point;
	for(int i = point.length; --i >= halfPointLen;){
		point[i].x = -point[i].x;
	}
	for(int i = 0; i < halfPolyIndexLen; i++){
		polyIndex+=PolyIndex(coordIndex.length, polyIndex[i].len);
		for(int q = polyIndex[i].len; --q >= 0;){
			coordIndex+=coordIndex[polyIndex[i].idx + q] + halfPointLen;
		}
	}
	updateBuild3dWindow();
	houseKeepVisualClue();
	BackUp::monostable.beginCycle();
}
