/* Seamless3d

Copyright (C) 2001-2010 Graham John Perrett (AKA thyme) thyme@seamless3d.com www.seamless3d.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this
software and associated documentation files (the "Software"), to deal in the Software
without restriction, including without limitation the rights to use, copy, modify,
merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to the following
conditions:

The above copyright notice and this permission notice shall be included in all copies or
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT
OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.
*/

#include <windows.h>
#include <direct.h>

#include "seamless3d.h"
#include "techuelife_triangle_editor.h"
#include "gui.h"
#include "scene_tree.h"
#include "resource.h"
#include "input.h"
#include "robot_aid.h"
#include "seamless3d_gui_render.h"

Window::TopWindow Window::topWindow;
Window::ActiveWindow Window::activeWindow;
M_<Ptr<Window> > Window::visWinList;
/*
void Window::removeAll2dWindows(){
	visWinList.length=0;
	visWinList+=g_3d;
}
*/
inline bool cmpMenuStr(const String& label, const String& str){
	int strLen = str.length;
	int labelLen = label.length;
	if(labelLen != strLen){
		if(labelLen > strLen){
			labelLen = label.indexOf("(");
			if(labelLen != -1){
				while(label[--labelLen]<=' '){
				}
				labelLen++;
			}
		}
	}
	if(labelLen == strLen){
		for(int i = 0; i<labelLen; i++){
			if(label[i]!=str[i]){
				return false;
			}
		}
		return true;
	}
	return false;
}
MenuItem* Window::getMenuItem(const String str){
	for(int k = visWinList.length; --k>=0;){
		if(Menu* menu=dynamic_cast<Menu*>(visWinList[k].p)){
			M_<Ptr<MenuItem> >& list = menu->list;
			for(int i = list.length; --i>=0;){
				MenuItem* menuItem = list[i];
				//int dummy=0;
				//if(menuItem->label.cmpCppWord(str, dummy)){
				if(cmpMenuStr(menuItem->label, str)){
					return menuItem;
				}
			}
		}
	}
	InputEvent::_killTimer();
	return 0;
}
void Window::blitBlockFromHorizontalLine(Vec2i dstPos, Vec2i srcPos, Vec2i blockSize){
	const unsigned* srcPix = &iconSrcBuf[(iconSrcBufCY-1-srcPos.y)*480+srcPos.x];
	int* dstPix=pixBuf.getArray();
	for(int i=0; i<blockSize.x; i++){
		drawVerticleLine2( srcPix[i], dstPos.x + i,	dstPos.y, dstPos.y + blockSize.y, size.x, size.y, dstPix);
	}
}
void paintBlock2(const unsigned fromColor, const unsigned toColor, int tx, int ty, int bx, int by, int dstWidth, int dstHeight, int* dstPix);
void Window::blitHorizontalColorSweep(Vec2i dstPos, Vec2i srcPosFrom, Vec2i srcPosTo, Vec2i blockSize){
	const unsigned* srcPixFrom = &iconSrcBuf[(iconSrcBufCY-1-srcPosFrom.y)*480+srcPosFrom.x];
	const unsigned* srcPixTo = &iconSrcBuf[(iconSrcBufCY-1-srcPosTo.y)*480+srcPosTo.x];
	int* dstPix=pixBuf.getArray();

	for(int i=0; i<blockSize.y; i++, srcPixFrom-=480, srcPixTo-=480){
		paintBlock2(*srcPixFrom, *srcPixTo, dstPos.x, dstPos.y+i, dstPos.x+blockSize.x, dstPos.y+i+1, size.x, size.y, dstPix);
	}
}
void paintVBlock2(const unsigned fromColor, const unsigned toColor, int tx, int ty, int by, int dstWidth, int dstHeight, int* dstPix){
	int changeInc=(by-ty)>>5;

	int fr=fromColor;
	int fg=fr;
	int fb=fr;
	fr=(fr&0xff0000)>>16;
	fg=(fg&0x00ff00)>>8;
	fb&=0x0000ff;

	int tr=toColor;
	int tg=tr;
	int tb=tr;
	tr=(tr&0xff0000)>>16;
	tg=(tg&0x00ff00)>>8;
	tb&=0x0000ff;

	int i=0;
	unsigned color=0;
	for(; fr!=tr || fg!=tg || fb!=tb; i+=changeInc){
		if(ty+i+changeInc>=by)break;
		color=fr;
		color<<=8;
		color|=fg;
		color<<=8;
		color|=fb;
		if(fr>tr){
			fr--;
		}
		else if(fr<tr){
			fr++;
		}

		if(fg>tg){
			fg--;
		}
		else if(fg<tg){
			fg++;
		}

		if(fb>tb){
			fb--;
		}
		else if(fb<tb){
			fb++;
		}
		drawVerticleLine2(color, tx, ty+i, ty+i+changeInc, dstWidth, dstHeight, dstPix);
	}
	drawVerticleLine2(color, tx, ty+i, by, dstWidth, dstHeight, dstPix);
}
void Window::blitVerticalColorSweep(Vec2i dstPos, Vec2i srcPosFrom, Vec2i srcPosTo, Vec2i blockSize){
	const unsigned* srcPixFrom = &iconSrcBuf[(iconSrcBufCY-1-srcPosFrom.y)*480+srcPosFrom.x];
	const unsigned* srcPixTo = &iconSrcBuf[(iconSrcBufCY-1-srcPosTo.y)*480+srcPosTo.x];
	int* dstPix=pixBuf.getArray();

	for(int i=0; i<blockSize.x; i++, srcPixFrom++, srcPixTo++){
		paintVBlock2(*srcPixFrom, *srcPixTo, dstPos.x+i, dstPos.y, dstPos.y+blockSize.y, size.x, size.y, dstPix);
	}
}

bool Window::hover(Vec2i ptCursor){
	if(dynamic_cast<_3DWin*>(this)){
		return false;
	}
	if(
		ptCursor.x>=pos.x &&
		ptCursor.x< pos.x+size.x &&
		ptCursor.y>=pos.y &&
		ptCursor.y< pos.y+size.y
		)
	{
		changeCursor(IDC_ARROW);

		buttonHover(ptCursor-pos);
		return true;
	}
	return false;
}
bool Window::click(const Vec2i ptCursor){
	if(ptCursor >= pos && ptCursor < pos+size){
		activeWindow=this;
		if(Button::getHoverButton()){
			Button* button=Button::getHoverButton();
			DownButton::setDownButton(button);
			button->action();

			paint();
		}
		else if(Drag* drag=dynamic_cast<Drag*>(this)){
			g_drag=drag;
			drag->dragPos=ptCursor;
		}
		return true;
	}
	return false;
}
void Window::release(Vec2i ptCursor){
//	if(inside==false)return;

//	if(dragMode){
//		ReleaseCapture();
//		dragMode=false;
//	}
//	g_bt.active=false;
	paint();
}
void Window::clear(){
	int*p = pixBuf.getArray();
	for(int i=pixBuf.length; --i>=0; p++){
//		*p=0xf8f3f9;
		*p=0xffffff;
	}
}
void Window::invert(){
	int*p = pixBuf.getArray();
	for(int i=pixBuf.length; --i>=0; p++){
		*p^=0xffffff;
	}
}
void Window::removeToolTip(){
	removeTypeFromList<ToolTipWin, Window>(visWinList);
}
void Window::removeAllAbove(Ptr<Window> par){
	for(int k=visWinList.length; visWinList[--k]!=par;){
		visWinList[k]=0;
		visWinList.length--;
	}
}
/*
void removeTypeFromList(M_<CallBack*>& list){
	int d=0;
	int s=0;
	while(s<list.length){
		if(dynamic_cast<MenuItem*>(list[s])){
			s++;
			continue;
		}
		list[d++]=list[s++];
	}
	list.length=d;
};
*/
void Window::removeMenu(){
	removeTypeFromList<Menu, Window>(visWinList);
	removeTypeFromList<MenuItem>(g_callBackList);
	g_remMenu=0;
}
void Window::initializeWindows(){
	visWinList+=g_3d.p;
}
Window* Window::TopWindow::operator ->()const{
	return visWinList[visWinList.length-1];
}
Window::TopWindow::operator Window*()const{
	return visWinList[visWinList.length-1];
}
void M_<Ptr<Window> >::insertElement(int targetIndex, Ptr<Window> value){iinsertElement(*this, targetIndex, value);}

void Window::insertUnderActive(Ptr<Window> par){
	Ptr<Window> remActiveWindow = Window::activeWindow.getActiveWindow();

	if(remActiveWindow==par){
		return;
	}
	removeFromList(visWinList, par.p);
	for(int i=0; i<visWinList.length; i++){
		if(visWinList[i]==remActiveWindow){
			visWinList.insertElement(i, par);
			par->visible=true;
			return;
		}
	}
}
void Window::TopWindow::operator = (Ptr<Window> par){
	
	Ptr<Window> remActiveWindow = Window::activeWindow.getActiveWindow();

	if(topWindow==par){
		return;
	}
	removeFromList(visWinList, par.p);
	if(par){
		visWinList+=par;
	}
	Ptr<Window> activeWindow = Window::activeWindow.getActiveWindow();
	if(activeWindow != remActiveWindow){
		remActiveWindow->onLoseFocus();
		activeWindow->onGainFocus();
	}
	if(par->isVisible()){
		par->visible=true;
		par->paint();
	}
}
Window* Window::ActiveWindow::getActiveWindow()const{
	int i=visWinList.length-1;
//	for(; dynamic_cast<Menu*>(visWinList[i].p) || dynamic_cast<ToolTipWin*>(visWinList[i].p) || dynamic_cast<MessageWin*>(visWinList[i].p); --i){
	for(; dynamic_cast<Menu*>(visWinList[i].p) || dynamic_cast<ToolTipWin*>(visWinList[i].p); --i){
	}
	return visWinList[i];
}
Window* Window::ActiveWindow::operator ->()const{
	return getActiveWindow();
}
void Window::ActiveWindow::operator = (Ptr<Window> par){
	if(activeWindow != par && dynamic_cast<Menu*>(par.p)==0){
		topWindow=par;
		par->hover(g_mousePos);
	}
}
void Window::closeWindow(Ptr<Window> par){
	par->onClose();
	Button* button=Button::getHoverButton();
	if(button && &button->win==par){
		removeTypeFromList<Button>(g_callBackList);
	}
	button=DownButton::getDownButton();
	if(button && &button->win==par){
		removeTypeFromList < DownButton>(g_callBackList);
	}
	par->visible=false;
	par->releaseWindow();
	removeFromList(visWinList, par.p);

	if(par==g_acp){
		houseKeepVisualClue();
	}
}
ResizeableWindow::TB_Top ResizeableWindow::tB_Top;
ResizeableWindow::TB_Bottom ResizeableWindow::tB_Bottom;
ResizeableWindow::TB_Left ResizeableWindow::tB_Left;
ResizeableWindow::TB_Right ResizeableWindow::tB_Right;
ResizeableWindow::TB_TopLeft ResizeableWindow::tB_TopLeft;
ResizeableWindow::TB_TopRight ResizeableWindow::tB_TopRight;
ResizeableWindow::TB_BottomLeft ResizeableWindow::tB_BottomLeft;
ResizeableWindow::TB_BottomRight ResizeableWindow::tB_BottomRight;

void ResizeableWindow::TB_Top::doDrag(Vec2i ptCursor){
	window->size=Vec2i(window->size.x, window->pos.y + window->size.y - ptCursor.y);
	window->pos=Vec2i(window->pos.x, ptCursor.y);
	window->paint();
}
void ResizeableWindow::TB_Bottom::doDrag(Vec2i ptCursor){
	window->size=Vec2i(window->size.x, ptCursor.y - window->pos.y);
	window->paint();
}
void ResizeableWindow::TB_Left::doDrag(Vec2i ptCursor){
	window->size=Vec2i(window->pos.x + window->size.x - ptCursor.x, window->size.y);
	window->pos=Vec2i(ptCursor.x, window->pos.y);
	window->paint();
}
void ResizeableWindow::TB_Right::doDrag(Vec2i ptCursor){
	window->size=Vec2i(ptCursor.x - window->pos.x, window->size.y);
	window->paint();
}
void ResizeableWindow::TB_TopLeft::doDrag(Vec2i ptCursor){
	window->size=Vec2i(window->pos.x + window->size.x - ptCursor.x, window->pos.y + window->size.y - ptCursor.y);
	window->pos=Vec2i(ptCursor.x, ptCursor.y);
	window->paint();
}
void ResizeableWindow::TB_TopRight::doDrag(Vec2i ptCursor){
	window->size=Vec2i(ptCursor.x - window->pos.x, window->pos.y + window->size.y - ptCursor.y);
	window->pos=Vec2i(window->pos.x, ptCursor.y);
	window->paint();
}
void ResizeableWindow::TB_BottomLeft::doDrag(Vec2i ptCursor){
	window->size=Vec2i(window->pos.x + window->size.x - ptCursor.x, ptCursor.y - window->pos.y);
	window->pos=Vec2i(ptCursor.x, window->pos.y);
	window->paint();
}
void ResizeableWindow::TB_BottomRight::doDrag(Vec2i ptCursor){
	window->size=Vec2i(ptCursor.x - window->pos.x, ptCursor.y - window->pos.y);
	window->paint();
}
void Window::nullAllControlPanels(){
	for(int i=visWinList.length; --i>=0;){
		if(dynamic_cast<ControlPanel*>(visWinList[i].p)){
			visWinList[i]=0;
		}
	}
}
void Window::paintAllControlPanels(){
	for(int i=visWinList.length; --i>=0;){
		if(ControlPanel* controlPanel=dynamic_cast<ControlPanel*>(visWinList[i].p)){
			controlPanel->paint();
		}
	}
}
void Window::paintAllControlPanelsExceptActive(){
	Window* activeWindow=Window::activeWindow.getActiveWindow();
	for(int i=visWinList.length; --i>=0;){
		if(ControlPanel* controlPanel=dynamic_cast<ControlPanel*>(visWinList[i].p)){
			if(controlPanel!=activeWindow){
				controlPanel->paint();
			}
		}
	}
}
void Window::reSetupAllControlPanels(){
	for(int i=visWinList.length; --i>=0;){
		if(ControlPanel* controlPanel=dynamic_cast<ControlPanel*>(visWinList[i].p)){
			controlPanel->setupNodeWindow(controlPanel->node);
		}
	}
}
ControlPanel* Window::getControlPanel(const int layer){
	int counter=-1;
	for(int i=visWinList.length; --i>=0;){
		if(ControlPanel* controlPanel=dynamic_cast<ControlPanel*>(visWinList[i].p)){
			if(dynamic_cast<Demo*>(controlPanel->node.p)==0){
				if(++counter==layer){
					return controlPanel;
				}
			}
		}
	}
	return 0;
}
Window* Window::getWindow(const int layer){
	int idx=visWinList.length-layer-1;
	if(static_cast<unsigned>(idx) < static_cast<unsigned>(visWinList.length)){
		return visWinList[idx].p;
	}
	return 0;
}
void Window::closeAllControlPanelsExceptDemo(){
	for(int i=visWinList.length; --i>=0;){
		if(ControlPanel* controlPanel=dynamic_cast<ControlPanel*>(visWinList[i].p)){
			if(dynamic_cast<Demo*>(controlPanel->node.p)==0){
				closeWindow(controlPanel);
			}
		}
	}
}
Anim* Window::getAnimBar(){
	for(int i=visWinList.length; --i>=0;){
		if(ControlPanel* controlPanel=dynamic_cast<ControlPanel*>(visWinList[i].p)){
			if(Anim* anim=dynamic_cast<Anim*>(controlPanel->node.p)){
				return anim;
			}
		}
	}
	return 0;
}
Anim* Window::animMode(){
	if(dynamic_cast<NurbsPatch*>(g_selNode.getRaw()) == 0){
		for(int i=visWinList.length; --i>=0;){
			if(ControlPanel* controlPanel=dynamic_cast<ControlPanel*>(visWinList[i].p)){
				if(Anim* anim=dynamic_cast<Anim*>(controlPanel->node.p)){
					return anim;
				}
			}
		}
	}
	return 0;
}
ChatWin* Window::chatMode(){
	for(int i=visWinList.length; --i>=0;){
		if(ChatWin* chatWin=dynamic_cast<ChatWin*>(visWinList[i].p)){
			return chatWin;
		}
	}
	return 0;
}
Art* Window::returnOpenArt(){
	for(int i=visWinList.length; --i>=0;){
		if(ControlPanel* controlPanel=dynamic_cast<ControlPanel*>(visWinList[i].p)){
			if(Art* art=dynamic_cast<Art*>(controlPanel->node.p)){
				return art;
			}
		}
	}
	return 0;
}
void Window::preSaveProcessWin(Scene& scene){
	M_Int& wOpen = scene.wOpen;
	M_Vec2f& wPos = scene.wPos;
	M_Vec2f& wSize = scene.wSize;

	wOpen.length=0;
	wPos.length=0;
	wSize.length=0;
	for(int i=0; i<visWinList.length; i++){
		Window* win=visWinList[i];
		if(PaletteCP* palette = dynamic_cast<PaletteCP*>(win)){
			if(g_paletteMode){
				wOpen += -4;
			}
			else{
				wOpen += -5; //colorOwnership
			}
			wPos += Vec2f(palette->pos.x, palette->pos.y);
			wSize += Vec2f(palette->size.x, palette->size.y);
		}
		else if(ControlPanel* controlPanel=dynamic_cast<ControlPanel*>(win)){
//			if(controlPanel==g_stw->cp){
//				wOpen+=-2;
//				wPos+=Vec2f(controlPanel->pos.x, controlPanel->pos.y);
//				wSize+=Vec2f(controlPanel->size.x, controlPanel->size.y);
//			}
//			else{
				int line = getNodeLine(controlPanel->node);
				if(line >=0){
					wOpen += line;
					wPos += Vec2f(controlPanel->pos.x, controlPanel->pos.y);
					wSize += Vec2f(controlPanel->size.x, controlPanel->size.y);
				}
//			}
		}
		else if(SceneTree* sceneTree=dynamic_cast<SceneTree*>(win)){
			wOpen+=-1;
			wPos+=Vec2f(sceneTree->pos.x, sceneTree->pos.y);
			wSize+=Vec2f(sceneTree->size.x, sceneTree->size.y);

			wOpen+=-3;//new node bar
			wPos+=Vec2f(sceneTree->newNodeBar.dstPos.x, sceneTree->newNodeBar.dstPos.y);
			wSize.length++;
		}
	}
}

PaletteCP* paletteCPAllreadyOpen();
void openOwnership();
void openPalette();

void Window::postLoadProcessWin(Scene& scene){
	M_Int& wOpen = scene.wOpen;
	M_Vec2f& wPos = scene.wPos;
	M_Vec2f& wSize = scene.wSize;

	closeAllControlPanelsExceptDemo();
	closeWindow(g_stw.p);

	for(int k=0; k < wOpen.length; k++){
		if(wOpen[k]==-1){
			g_stw->pos=Vec2i(wPos[k].x, wPos[k].y);
			g_stw->size=Vec2i(wSize[k].x, wSize[k].y);
			//insertUnderActive(g_stw.p);
			activeWindow=g_stw.p;
		}
		else if(wOpen[k] == -3){
			g_stw->newNodeBar.dstPos=Vec2i(wPos[k].x, wPos[k].y);
		}
		else if(wOpen[k] == -4){//palette
			openPalette();
			PaletteCP* paletteCP = paletteCPAllreadyOpen();
			paletteCP->pos = Vec2i(wPos[k].x, wPos[k].y);
			paletteCP->size = Vec2i(wSize[k].x, wSize[k].y);
		}
		else if(wOpen[k] == -5){//colorOwnership
			openOwnership();
			PaletteCP* paletteCP = paletteCPAllreadyOpen();
			paletteCP->pos = Vec2i(wPos[k].x, wPos[k].y);
			paletteCP->size = Vec2i(wSize[k].x, wSize[k].y);
		}
		else{
			Node* node = 0;
			if(wOpen[k]==-2){
//				g_stw->cp->pos=Vec2i(wPos[k].x, wPos[k].y);
//				g_stw->cp->size=Vec2i(wSize[k].x, wSize[k].y);
//				activeWindow=g_stw->cp.p;
				node = g_selNode.getRaw();
			}
			else{
				node = g_sceneTreeList[wOpen[k]].node;
			}
			ControlPanel* controlPanel=0;

			if(dynamic_cast<Video*>(node)){
				if(g_vcp == 0){
					g_vcp = new VideoControlPanel;
				}
				controlPanel=g_vcp;
			}
			else if(dynamic_cast<WaveToPoly*>(node)){
				if(g_wtpcp == 0){
					g_wtpcp = new WaveToPolyControlPanel;
				}
				controlPanel=g_wtpcp;
			}
			else if(dynamic_cast<Anim*>(node)){
				//g_acp=new AnimControlPanel;
				controlPanel=g_acp;
			}
			else{
				controlPanel=new ControlPanel;
				g_stw->cp=controlPanel;
			}
			controlPanel->pos=Vec2i(wPos[k].x, wPos[k].y);
			controlPanel->size=Vec2i(wSize[k].x, wSize[k].y);
			controlPanel->setupNodeWindow(node);
			//insertUnderActive(controlPanel);
			activeWindow=controlPanel;
		}
	}
	g_stw->checkWinVis();
	g_stw->cp->checkWinVis();
}

void Window::activeOnMouseWheel(const int par){//originaly coded by bumpy, later simplified by thyme
	for(int i = visWinList.length; i-- > 0; ){
		Window *win = visWinList[i];
		if (g_mousePos >= win->pos && g_mousePos < win->pos + win->size){
			win->onMouseWheel(par);
			break;
		}
	}
}

void Window::checkWinVis(){
	if(pos.x > g_3d->size.x-1){
		pos.x = g_3d->size.x - 100;
		if(pos.x < 0){
			pos.x = 0;
		}
	}
	if(pos.y > g_3d->size.y-1){
		pos.y = g_3d->size.y - 100;
		if(pos.y < 0){
			pos.y = 0;
		}
	}
}
