/* Seamless3d

Copyright (C) 2001-2010 Graham John Perrett (AKA thyme) thyme@seamless3d.com www.seamless3d.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this
software and associated documentation files (the "Software"), to deal in the Software
without restriction, including without limitation the rights to use, copy, modify,
merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to the following
conditions:

The above copyright notice and this permission notice shall be included in all copies or
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT
OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.
*/

#define INITGUID
//#define OPENGL 1 //uncomment this line if using opengl if not defined else where

#include <windows.h>
#include <direct.h>
#include <process.h>
#include <shlobj.h>

#include "seamless3d.h"
#include "techuelife_triangle_editor.h"
#include "gui.h"
#include "scene_tree.h"
#include "resource.h"
#include "input.h"

#include "seamless3d_gui_render.h"
#include "robot_aid.h"
#include "chat_window.h"

#define ENGLISH 1
//#define FRENCH 1

#ifdef ENGLISH
#include "global_strings.h"
#elif FRENCH
#include "global_strings.fr.h"
#endif

#ifdef RECORD_SOUNDMOVIE
extern HANDLE g_hNotificationEvent;
#endif

Vec2i g_extraWinPix;
void outputScreenShotForDemo();
extern const char* menu_to_open_F10;
bool smlsIsSaved();
void saveBeforeClosing(void (*i_yesFunc)(FileCP& fileCP), void (*i_noFunc)(FileCP& fileCP), void (*i_cancelFunc)(FileCP& fileCP));

HACCEL hAccel;
extern bool g_browseMode;
void renderScene();
void initializeDeveloperScript();
void insertRecentFile(const char* fullFileName);
void openDroppedFile();
extern String droppedFilename;
extern bool g_recordSoundMovie;
extern bool g_playSoundMovie;
extern bool g_playing;

GeomRenderPtr g_redPoint=new GeomRender;
GeomRenderPtr g_lastPoint=new GeomRender;
GeomRenderPtr g_nextPoint=new GeomRender;
GeomRenderPtr g_yellowPoint=new GeomRender;
GeomRenderPtr g_cyanPoint=new GeomRender;
bool g_online = false;
bool g_viewer = false;
bool g_minimized = false;
bool g_processComnandLine = true;

double g_previousClickTime=0;
Vec2i g_previousLeftClickPos;
bool g_leftDoubleClicked=false;
double g_currentClickTime=0;

extern bool g_windowSizeWindowBool;
extern double g_windowSizeWindowTime;

//void mouseLeaveCleanUp();

bool g_bActive  = true;//false; // Whether the app is active (not minimized)
bool g_bReady   = false; // Whether the app is ready to render frames

HINSTANCE  g_hinst;

unsigned int colorToHex( Color3f color);
bool setDirectInput( HWND hWnd );
void openSmlsLastPhase(String fullFilePath, const bool recover = false);
void setupRecentFiles();
bool checkForLockFiles();
void updatePlayBuffer();
extern bool g_playing;

HRESULT InitDirectInput( HWND hWnd );
HRESULT FreeDirectInput();
BackUp::Monostable BackUp::monostable;
RECT clientOffest;

extern bool g_dontCloseMenu;
bool g_createdINI=false;
bool outPutIndices=false;
extern int	g_mirrorVertex=0;
//BOOL g_localControl=false;
BackUp::SmlsSaved BackUp::smlsSaved;

double	g_timeBegan;
double g_time;
int g_viewIndex=0;
extern float g_version;
extern bool initWinINet();
extern int g_numOfVisLines;

//void convertJoinPartToTEPart(IndexedFaceSet* part);
NodePtr g_before;
ArtPtr remPJ;

SettingsPtr g_settings;


Drag* g_drag=0;
//-----------------------------------------------------------------------------
// Global variables
//-----------------------------------------------------------------------------
//Vec2i g_3dwinPos;
//Vec2i g_3dwinSize=Vec2i(800,572-24*2-2);
//Vec2i g_sceneTreePos;
//Vec2i g_sceneTreeSize=Vec2i(400,0);
//Vec2i g_nodeWinPos=Vec2i(400, 570);


HWND hWndMain=NULL;

//#define GET_WM_COMMAND_ID(wp, lp)               LOWORD(wp)
//#define GET_WM_COMMAND_HWND(wp, lp)             (HWND)(lp)
//#define GET_WM_COMMAND_CMD(wp, lp)              HIWORD(wp)
//BOOL CALLBACK selectedVertexProc(HWND hDlg,UINT iMessage, WPARAM wParam, LPARAM lParam);

void buildF10Menu();
bool windowSizeHasChanged();
void convertVisualCluesToRenderFormat();
void resetLooseEnds();
void addNodesForNewFile();
void initVisualClue();
void initRadialTug();

void mouse3dHandler(UINT msg,WPARAM wParam);
//void pointToPartByDefault();

char defaultMsg[]={"message from thyme's code"};

String g_seamlessDir;
const String& createSeamlessConstDirString(String& str){
	getFolderPath(str, CSIDL_MYDOCUMENTS);
	//getFolderPath(str, CSIDL_PROGRAM_FILES);
	int strPathLen=str.length;
	if(strPathLen>0&&str[strPathLen-1]=='\\'||strPathLen>0&&str[strPathLen-1]=='/'){
		str+="Seamless3d\\";
	}
	else{
		str+="\\Seamless3d\\";
	}
	return str;
}
const String g_appDataDir = createSeamlessConstDirString(g_seamlessDir) + "Application Data For Seamless3d\\";
//const String g_undoDir = g_appDataDir + "undoCache\\";
const String g_undoDir = g_appDataDir + "Recover\\";
const String g_demoDir = g_appDataDir + "For Robot Demos Only\\";
const String g_demoCacheDir = g_demoDir + "Cache\\";
const String g_setupPath = g_appDataDir + "seamless3dSetup.wrl";
const String g_debugLogPath = g_appDataDir + "debugLog.txt";
const String g_screenshotDir = g_seamlessDir + "Screenshot\\";


void showCursor(BOOL show){
	static BOOL remVis=TRUE;
	if(remVis!=show){
		remVis=show;
		ShowCursor(show);
	}
	
}
void getFolderPath(String& str, unsigned id){
	TCHAR buf[MAX_PATH];
	SHGetSpecialFolderPath(
		NULL,
		buf,
		id,
		NULL
	);
//	str="";
	for(int i=0; buf[i]!=0; i++){
		str+=buf[i];
	}
}
void makeTitleStr(char* str){
	char tempStr[0x200];
	char drive[0x200];
	char dir[0x200];
	char ext[0x200];

	_splitpath(g_currentFilename.c_str(), drive, dir, str, ext );
	strcat(str, ext);

#ifdef OPENGL
	sprintf(tempStr," - Seamless3d %.3f OpenGL",g_version);
#elif TTS
	sprintf(tempStr," - Seamless3d TTS %.3f",g_version);
#else
	sprintf(tempStr," - Seamless3d %.3f",g_version);
//	sprintf(tempStr," - Seamless3d %.5f",g_version);
#endif


	if(!BackUp::smlsSaved){
		strcat(str," *");
	}
	strcat(str, tempStr);
	strcat(str," - ");
	strcat(str, g_selNode->name.c_str());
}
void setTitleName(){
	//selectPartFromVertex(g_seamless);
	char TempStr[0x200];
	makeTitleStr(TempStr);
	SetWindowText(hWndMain, TempStr);
	//SendMessage(g_hDlg, WM_USER, 0, 0);
}
void setSmlsSaved(){
	BackUp::smlsSaved.set();
	setTitleName();
}
void BackUp::Monostable::beginCycle(){
	if(beginBool==false){
		backedUpFileCntr++;
	}
	beginBool=true;
	setTitleName(); //this might as well be done here in case its been forgotten and cant do any significant harm slowing things down
	beginTime=g_time;
	g_selCount.reset();//this might as well be done here in case its been forgotten and cant do any significant harm slowing things down
}
bool BackUp::Monostable::state()const{
	return beginBool && (g_time>beginTime+g_settings->backupDelay);
}
inline void BackUp::checkForBackUpOperation(){
	if(monostable.state()){
		backupToCache();
		monostable.clear();
	}
}

int msG(String str1,String str2,String str3,unsigned dlgTypeAndIcon){

	String tempStr=String(str1);
	tempStr+=str2;
	tempStr+=str3;
	
	return MessageBox(hWndMain,tempStr.c_str(),defaultMsg,dlgTypeAndIcon);
}

int msG(String str,unsigned dlgTypeAndIcon){
	return MessageBox(hWndMain,str.c_str(),defaultMsg,dlgTypeAndIcon);
}

extern float g_remVersionLoaded;

void changeExtensionToSmls(String& str){
	for(int i=str.length; --i>=0;){
		if(str[i]=='.'){
			str = str.substr(0, i);
			break;
		}
		else if(str[i]=='\\' || str[i]=='/'){
			break;
		}
	}
	for(int i=2; ;i++){
		String newName;
		newName+=str;
		newName+=i;
		newName+=".smls";
		FILE* file;
		if((file=fopen(newName.c_str(), "rb"))==0){
			str=newName;
			return;
		}
		else{
			fclose(file);
		}
	}
}


void searchTillText(char*& p);

void openSceneTreeInterface(){

	g_stw->newNodeBar.dstPos=Vec2i(220, 25+16+8);
	g_numOfVisLines=(g_stw->size.y-SceneTree::beginYOffset)/pixSymbolCY;

	g_stw->size = Vec2i(g_settings->sceneTreeSize.x, g_settings->sceneTreeSize.y);
	if(g_stw->size.x < 10){
		g_stw->size = Vec2i(350, g_stw->size.y);
	}
	if(g_stw->size.y < 10){
		g_stw->size = Vec2i(g_stw->size.x, 400);
	}
	g_stw->pos=Vec2i(g_settings->sceneTreePos.x, g_settings->sceneTreePos.y);
	g_stw->pixBuf.length=g_stw->size.x*g_stw->size.y;


//	g_stw->paint();
}
/*
void roundToNearest(){
	ArtPtr part=g_selNode;
	SeamlessPtr smls=part->smls;
	M_Vec3f& point=smls->refPoint();
	Vec3f vec;
	vec=point[smls->ActCtrlPnt].multiply(100);
	vec.x=floor(vec.x+.5);
	vec.y=floor(vec.y+.5);
	vec.z=floor(vec.z+.5);
	point[smls->ActCtrlPnt]=vec.multiply(.01);
	
	smls->updateAllVertexInBuffers();
	makePointerEqualIndexLocation();
}
*/
void autoAddColor(ArtPtr part){
	if(part->color.length==0){
		M_Art& partList=part->smls->partList;
		for(int k=partList.length; --k>=0; ){
			Art* part=partList[k];
			part->color.length=part->coord.length;
		}
	}
}

void workOutVertexSameParts(int& newFrom, int& newTo, M_Int fromVertex, M_Int toVertex){
	bubbleSort(fromVertex, toVertex);

	for(int i=0; i<fromVertex.length; i++){
		if(fromVertex[i]<=newFrom){
			newFrom++;
		}
		if(fromVertex[i]<=newTo){
			newTo++;
		}
	}
}
void workOutVertexDiffParts(int& newFrom, M_Int fromVertex, M_Int toVertex){
	bubbleSort(fromVertex, toVertex);

	for(int i=0; i<fromVertex.length; i++){
		if(fromVertex[i]<=newFrom){
			newFrom++;
		}
	}
}

/*
void loopBackIFS(Seamless* smls){
	M_Build& build=smls->build.refValue();
	for(int k=build.length; --k>=0;){
		if(build[k]->getType()==STEM){
			Stem* stem=dynamic_cast<Stem*>(build[k].getRawPtr());
			M_Node& source=stem->source.refValue();
			M_SArt& part=stem->part.refValue();
			for(int i=source.length; --i>=0; ){
				if(i<part.length){
					Art* prt=dynamic_cast<Art*>(part[i].getRawPtr());
					IndexedFaceSet* ifs=dynamic_cast<IndexedFaceSet*>(source[i].getRawPtr());
					if(ifs!=0){
						int remPointSize=ifs->refPoint().length;
						int remColorSize=ifs->refColor().length;
						int remTriSize=ifs->refCoordIndex().length/4;
						//source[i]=dynamic_cast<Art*>(part[i].getRawPtr())->extractGeometry();
						ifs->copyNotClose(prt->extractGeometry());
						if(remPointSize!=ifs->refPoint().length){
							String str=prt->getName();
							str+=" coord diff ";
							str+=ifs->refPoint().length-remPointSize;
							dLog(str);
						}
						if(remColorSize!=ifs->refColor().length){
							String str=prt->getName();
							str+=" color diff ";
							str+=ifs->refColor().length-remColorSize;
							dLog(str);
						}
						if(remTriSize!=ifs->refCoordIndex().length/4){
							String str=prt->getName();
							str+=" triangle diff ";
							str+=ifs->refCoordIndex().length/4-remTriSize;
							dLog(str);
						}
					}
				}
			}
		}
	}
	smls->doBuild();
	smls->convertToRenderFormat();
	genNodeList();
	SendMessage(hSceneTreeWnd, WM_PAINT, 0, 0);
	makePointerEqualIndexLocation();
}
*/
bool checkVertexBelongsToTriangle(Art* vertexPart, Art* triPart, int vertexIndex, int &corner, int triangleIndex);
/*
void transferSingleVertex(){
	if(remPJ==NULL)return;
	Color3f scolor;
	ArtPtr part=g_selNode;
	SeamlessPtr &smls=part->smls;
	BackUp::monostable.beginCycle();
	int _actCtrlPnt=smls->ActCtrlPnt;
	M_Int &coordIndex=smls->geometry->coordIndex;

	Vec3f svec = smls->geometry->coord->point[_actCtrlPnt];
	svec=part->matrix*svec-remPJ->matrix*Vec3f();

	if(smls->geometry->color!=NULL){
		scolor=smls->geometry->color->color[_actCtrlPnt];
	}

	smls->genVertex(remPJ, svec, scolor);

	if(part->plidx>remPJ->plidx){
		_actCtrlPnt++;
	}

	//replace old vertex reference with the new vertex reference
	for(int i = -1; i<coordIndex.length-1 ; ){
		if(coordIndex[++i]==_actCtrlPnt){
			coordIndex[i]=remPJ->toCoord-1;
		}
		if(coordIndex[++i]==_actCtrlPnt){
			coordIndex[i]=remPJ->toCoord-1;
		}
		if(coordIndex[++i]==_actCtrlPnt){
			coordIndex[i]=remPJ->toCoord-1;
		}
		i++;
	}

	smls->deleteVertex(_actCtrlPnt); //delete old vertex

	smls->orderPolygonsFromVertices();

	smls->convertToRenderFormat();

	smls->ActCtrlPnt=remPJ->toCoord-1;
	makePointerEqualIndexLocation();
}
*/
/*
void joinCloseVertexImmediate(){
	ArtPtr part=g_selNode;
	SeamlessPtr &smls=part->smls;
	BackUp::monostable.beginCycle();

	JoinCloseVertex joinCloseVertex;
	joinCloseVertex.part=part;
	joinCloseVertex.joinRange=g_joinRange;
	joinCloseVertex.build();

	smls->convertToRenderFormat();

	makePointerEqualIndexLocation();
}
*/
/*
void reset3dCursor(){
	g_pointer->translation=Vec3f(0,0,0);
	g_pointer->rotation=Rotation();
	g_pointer->scale=Vec3f(1,1,1);
}
*/
void mouse2dSync(HWND &hWnd);
bool _3DWin::click(const Vec2i ptCursor){
	return false;
}
inline void Window::checkIfAWindowHasBeenClickedAndUpdate(const Vec2i& mousePos){
	if(Window::activeWindow->hog){
		if(false == Window::activeWindow->click(mousePos)){
			PlaySound("SystemDefault", 0, SND_ASYNC);
		}
	}
	else{
		for(int i=visWinList.length; --i>=0;){
			if(visWinList[i]->click(mousePos)){
				return;
			}
		}
		activeWindow=g_3d.p;
		mouse3dMode=true;
		mouse2dSync(hWndMain);
		SetAcquire();
	}
}
inline void Window::checkIfAWindowHasBeenRightClickedAndUpdate(const Vec2i& mousePos){
	if(Window::activeWindow->hog){
		if(false == Window::activeWindow->rightClick(mousePos)){
			PlaySound("SystemDefault", 0, SND_ASYNC);
		}
	}
	else{
		for(int i=visWinList.length; --i>=0;){
			if(visWinList[i]->rightClick(mousePos)){
				return;
			}
		}
		activeWindow=g_3d.p;
		mouse3dMode=true;
		mouse2dSync(hWndMain);
		SetAcquire();
	}
}

inline void Window::checkIfMouseIsHoveringAWindowAndUpdate(const Vec2i& mousePos){
	if(g_remMenu == 0 && Window::activeWindow->hog){
		if(Window::activeWindow->hover(mousePos) == false){
			changeCursor(IDC_ARROW);
		}
	}
	else{
		for(int i = visWinList.length; --i>=0;){
			if(visWinList[i]->hover(mousePos)){
				return;
			}
			changeCursor(IDC_CROSS);
		}
	}
}

bool isDoubleLeftClick(){
	if(g_leftDoubleClicked==false && g_previousLeftClickPos==g_mousePos){

		double difference = g_currentClickTime - g_previousClickTime;
		double doubleClickTime = GetDoubleClickTime() * MILISECONDS_TO_SECONDS;

		if(difference <= doubleClickTime){
			return true;
		}
	}
	return false;
}

BOOL mouse2dHandler(UINT msg,WPARAM wParam, LPARAM lParam){
	static POINT ptDownMousePos2d;
    switch( msg )
    {
	case WM_RBUTTONDOWN:{
		Button::repeatCounter=0;
		removeTypeFromList<Button>(g_callBackList);
		Window::removeToolTip();
		Window::removeMenu();

		g_mousePos=Vec2i(LOWORD(lParam), HIWORD(lParam));
		Window::checkIfAWindowHasBeenRightClickedAndUpdate(g_mousePos);
		return TRUE;
	}
	case WM_LBUTTONDOWN:{
		Button::repeatCounter=0;
		g_previousClickTime=g_currentClickTime;
		g_currentClickTime=g_time;

		Window::removeToolTip();

		g_mousePos=Vec2i(LOWORD(lParam), HIWORD(lParam));
		if(Button* button = Button::getHoverButton()){

			DownButton::setDownButton(button);

			button->win.onClickButton();
			Window::activeWindow=&button->win;
			button->action();
			//button->win.paint();//do we need this?
			if(g_dontCloseMenu==false){
				Window::removeMenu();
			}
		}
		else{
			Window::removeMenu();
			Window::checkIfAWindowHasBeenClickedAndUpdate(g_mousePos);
		}
		g_dontCloseMenu=false;
		g_leftDoubleClicked=isDoubleLeftClick();
		g_previousLeftClickPos=g_mousePos;
		return TRUE;
	}
	case WM_LBUTTONUP:{
		g_mousePos=Vec2i(LOWORD(lParam), HIWORD(lParam));

		if(g_drag){
			g_drag->endDrag(g_mousePos);
			g_drag=0;
		}
		else if(Button* button=DownButton::getDownButton()){
			button->onButtonUp();
		}
		else{
			Window::activeWindow->release(g_mousePos);
		}
#ifdef PROG_ROB
		InputEvent::sceneTreeScrolPos=0;
		if(g_upStr!=""){
			dLog(g_upStr);
		}
		g_upStr="";

		InputEvent::sceneTreeScrolPos=g_stw->getScrollPosCollapsed();
#endif

		return TRUE;
	}
	case WM_MOUSEMOVE:{
		g_mousePos = Vec2i(LOWORD(lParam), HIWORD(lParam));
		Button* hoverButton = Button::getHoverButton();
		if(g_remMenu == 0 && (unsigned)g_mousePos.x < 100 && (unsigned)g_mousePos.y < 50 && Window::activeWindow->hog == false){
			buildF10Menu();
		}
		else if(g_drag){
			g_drag->doDrag(g_mousePos);
		}
		else if(hoverButton && hoverButton->onHoverMove(g_mousePos - hoverButton->win.pos)){
		}
		else{
			Window::checkIfMouseIsHoveringAWindowAndUpdate(g_mousePos);
		}
		break;
	}
		/*
	case WM_MOUSEMOVE:
		if(MK_LBUTTON&wParam){
			
			
			GetCursorPos( &ptCursor );
			g_scene->translation=g_scene->rotation.multVec(g_scene->translation);
			g_scene->translation.x+=.0025*(float)(ptCursor.x-ptDownMousePos2d.x);
			g_scene->translation=-g_scene->rotation*g_scene->translation;
			
			if(0x8000&GetKeyState(VK_LCONTROL)){
				g_scene->translation=g_scene->rotation.multVec(g_scene->translation);
				g_scene->translation.z-=.0025*(float)(ptCursor.y-ptDownMousePos2d.y);
				g_scene->translation=-g_scene->rotation*g_scene->translation;
			}
			else{
				g_scene->translation=g_scene->rotation*g_scene->translation;
				g_scene->translation.y-=.0025*(float)(ptCursor.y-ptDownMousePos2d.y);
				g_scene->translation=-g_scene->rotation*g_scene->translation;
			}
			
			GetCursorPos( &ptDownMousePos2d );
			return TRUE;
		}
		if(MK_RBUTTON&wParam){
			GetCursorPos( &ptCursor );
			rotateScene(ptCursor.x-ptDownMousePos2d.x, ptCursor.y-ptDownMousePos2d.y);
			ptDownMousePos2d=ptCursor;
			return TRUE;
		}
		*/
	}
	return FALSE;
}
void updateNavKey();

bool keyDown(HWND &hWnd, UINT &message, WPARAM &wParam, LPARAM &lParam){
	if(Window::activeWindow==g_3d){
		static String str;
		float tran=(1/g_scene->scale.x)*.01;
		switch (wParam){
			case VK_LEFT:{
				if(g_leftKeyDown==false){
					g_leftKeyTime=g_time;
					g_leftKeyDown=true;
					updateNavKey();
				}
				else{
					g_leftKeyDown=true;
				}
				break;
			}
			case VK_RIGHT:{
				if(g_rightKeyDown==false){
					g_rightKeyTime=g_time;
					g_rightKeyDown=true;
					updateNavKey();
				}
				else{
					g_rightKeyDown=true;
				}
				break;
			}
			case VK_DOWN:{
				if(g_backwardKeyDown==false){
					g_backwardKeyTime=g_time;
					g_backwardKeyDown=true;
					updateNavKey();
				}
				else{
					g_backwardKeyDown=true;
				}
				break;
			}
			case VK_UP:{
				if(g_forwardKeyDown==false){
					g_forwardKeyTime=g_time;
					g_forwardKeyDown=true;
					updateNavKey();
				}
				else{
					g_forwardKeyDown=true;
				}
				break;
			}
			default:{
				return false;
			}
		}
		return true;
	}
	return false;
}

	/*
	int shiftKeyDown(HWND &hWnd, UINT &message, WPARAM &wParam, LPARAM &lParam){
		switch (wParam){
		case '0':
			setupNodeWindow(g_settings);
			break;
	//	case '9':
	//		setupNodeWindow(g_freezeTime);
	//		break;
		default:
			if(dynamic_cast<Art*>(g_selNode.getRawPtr())==0)return false;
			switch (wParam){
	//			case VK_ESCAPE:	//,
	//				reset3dCursor();
	//				break;
				case '1':
					buildSphereForSelPart();
					break;
				case '2':
					createNewCylinderForPart();
					break;
				case '3':
					genTriangle();
					break;
				case '4':
					createNewRectangleForPart(dynamic_cast<Art*>(g_selNode.getRawPtr()));
					setupNodeWindow(g_immediateRectangle);
					//DialogBoxParam( (HINSTANCE)GetModuleHandle(NULL),MAKEINTRESOURCE(FLAT_DIALOGUE), hWnd,flatDlgProc, lParam  );
					break;
				default:
					return false;
			}
		}
		return true;
}
*/
bool getOldColor(Color3f& oldColor){
	Art* part=dynamic_cast<Art*>(g_selNode.getRaw());
	if(part==0)return false;
	Seamless* smls=part->smls;
	Art* triPart=smls->actTriPart;
	if(triPart==0)return false;
	ColorEffect* colorEffect=dynamic_cast<ColorEffect*>(part->workEffect);
	if(colorEffect==0)return false;
	if(triPart->triangle.length <= smls->ActTri){
		return false;
	}
	M_Triangle& triangle=triPart->triangle;
	const M_Color3f& color = colorEffect->color;
	oldColor=color[triangle.refTexIndex(smls->ActTri*3+smls->ActTriCorner)];
	return true;
}
//-----------------------------------------------------------------------------
// Name: MsgProc()
// Desc: The window's message handler
//-----------------------------------------------------------------------------

void keyboardRadialTug(Vec3f coord, Vec3f pt){
	g_radialTug->radius.refValue()[0]=g_settings->tugRadius.refValue();
	Art* part=dynamic_cast<Art*>(g_selNode.getRaw());
	Seamless* smls=part->smls;
	M_Art& partList=smls->partList;

	g_radialTug->displacement.refValue()[0]=pt;

	g_radialTug->slope.refValue()=g_settings->tugSlope.refValue();
	if(g_settings->multiPartTug.refValue()){
		for(int k=partList.length; --k>=0;){
			Art* prt=partList[k];
			g_radialTugStem->part.refValue()[0]=prt;
			g_radialTug->location.refValue()[0]=part->matrix * (prt->imatrix * coord);
			g_radialTugStem->doBuild();
		}
	}
	else{
		g_radialTugStem->part.refValue()[0]=part;
		g_radialTug->location.refValue()[0]=coord;
		g_radialTugStem->doBuild();
	}
}

void openChatLog(){
	String url = g_appDataDir;
	url+="chat.txt";

	PROCESS_INFORMATION pi;
	STARTUPINFO si;
	ZeroMemory( &si, sizeof(si) );
	ZeroMemory( &pi, sizeof(pi) );
	si.cb = sizeof(si);
	// Start the child process.

	String commandLine='\"';
	getFolderPath(commandLine, CSIDL_WINDOWS);
	if(commandLine[commandLine.length-1]!='\\'){
		commandLine+='\\';
	}
	commandLine+="notepad.EXE\" ";
	commandLine+=url;

	if(CreateProcess(
		NULL, // No module name (use command line).
		const_cast<char*>(commandLine.c_str()), // Command line. 
		NULL,             // Process handle not inheritable. 
		NULL,             // Thread handle not inheritable. 
		FALSE,            // Set handle inheritance to FALSE. 
		0,                // No creation flags. 
		NULL,             // Use parent's environment block. 
		NULL,             // Use parent's starting directory. 
		&si,              // Pointer to STARTUPINFO structure.
		&pi )){             // Pointer to PROCESS_INFORMATION structure.    ) 
		
		//WaitForSingleObject( pi.hProcess, INFINITE );
		// Close process and thread handles.
		CloseHandle( pi.hProcess );
		CloseHandle( pi.hThread );
	}
}


void openWithHtmlBrowser(String url, const String htmlBrowser){
	PROCESS_INFORMATION pi;
	STARTUPINFO si;
	ZeroMemory( &si, sizeof(si) );
	ZeroMemory( &pi, sizeof(pi) );
	si.cb = sizeof(si);
	// Start the child process.

	String commandLine='\"';
	if(htmlBrowser == ""){
		getFolderPath(commandLine, CSIDL_PROGRAM_FILES);
		if(commandLine[commandLine.length-1]!='\\'){
			commandLine+='\\';
		}
		commandLine+="Internet Explorer\\IEXPLORE.EXE\"";
	}
	else{
		commandLine = htmlBrowser;
	}
	commandLine += ' ';
	commandLine += url;

	if( CreateProcess(
		NULL, // No module name (use command line).
		const_cast<char*>(commandLine.c_str()), // Command line. 
		NULL,             // Process handle not inheritable. 
		NULL,             // Thread handle not inheritable. 
		FALSE,            // Set handle inheritance to FALSE. 
		0,                // No creation flags. 
		NULL,             // Use parent's environment block. 
		NULL,             // Use parent's starting directory. 
		&si,              // Pointer to STARTUPINFO structure.
		&pi )){             // Pointer to PROCESS_INFORMATION structure.    ) 
		
		//WaitForSingleObject( pi.hProcess, INFINITE );
		// Close process and thread handles.
		CloseHandle( pi.hProcess );
		CloseHandle( pi.hThread );
	}
}
void getFFmpegPath(String& path){
    char buffer[MAX_PATH];
	GetModuleFileName(NULL, buffer, MAX_PATH);
	path = buffer;

	String drive;
	String dir;
	String fname;
	String ext;
	splitPath(path, drive, dir, fname, ext);
	path = '\"';
	path += drive;
	path += dir;
	path += "ffmpeg\\ffmpeg.exe\"";
}
void openWithFFmpeg(String url){
	PROCESS_INFORMATION pi;
	STARTUPINFO si;
	ZeroMemory( &si, sizeof(si));
	ZeroMemory( &pi, sizeof(pi));
	si.cb = sizeof(si);
	// Start the child process.
	String commandLine;
	getFFmpegPath(commandLine);


//	String commandLine='\"';
//	getFolderPath(commandLine, CSIDL_PROGRAM_FILES);
//	if(commandLine[commandLine.length-1]!='\\'){
//		commandLine+='\\';
//	}
//	commandLine+="Seamless3d\\ffmpeg.exe\"";
	commandLine+=url;
	//commandLine+='\"';

	if( CreateProcess(
		NULL, // No module name (use command line).
		const_cast<char*>(commandLine.c_str()), // Command line. 
		NULL,             // Process handle not inheritable. 
		NULL,             // Thread handle not inheritable. 
		FALSE,            // Set handle inheritance to FALSE. 
		0,                // No creation flags. 
		NULL,             // Use parent's environment block. 
		NULL,             // Use parent's starting directory. 
		&si,              // Pointer to STARTUPINFO structure.
		&pi )){             // Pointer to PROCESS_INFORMATION structure.    ) 
		
		//WaitForSingleObject( pi.hProcess, INFINITE );
		// Close process and thread handles.
		CloseHandle( pi.hProcess );
		CloseHandle( pi.hThread );
	}
}

void openWithWindowsMediaPlayer(String url){
	PROCESS_INFORMATION pi;
	STARTUPINFO si;
	ZeroMemory( &si, sizeof(si) );
	ZeroMemory( &pi, sizeof(pi) );
	si.cb = sizeof(si);
	// Start the child process.

	String commandLine='\"';
	getFolderPath(commandLine, CSIDL_PROGRAM_FILES);
	if(commandLine[commandLine.length-1]!='\\'){
		commandLine+='\\';
	}
	commandLine+="Windows Media Player\\wmplayer.exe\" \"";
	commandLine+=url;
	commandLine+='\"';

	if( CreateProcess(
		NULL, // No module name (use command line).
		const_cast<char*>(commandLine.c_str()), // Command line. 
		NULL,             // Process handle not inheritable. 
		NULL,             // Thread handle not inheritable. 
		FALSE,            // Set handle inheritance to FALSE. 
		0,                // No creation flags. 
		NULL,             // Use parent's environment block. 
		NULL,             // Use parent's starting directory. 
		&si,              // Pointer to STARTUPINFO structure.
		&pi )){             // Pointer to PROCESS_INFORMATION structure.    ) 
		
		//WaitForSingleObject( pi.hProcess, INFINITE );
		// Close process and thread handles.
		CloseHandle( pi.hProcess );
		CloseHandle( pi.hThread );
	}
}
void openWithPovRay(String url){
	PROCESS_INFORMATION pi;
	STARTUPINFO si;
	ZeroMemory( &si, sizeof(si) );
	ZeroMemory( &pi, sizeof(pi) );
	si.cb = sizeof(si);
	// Start the child process.

	String commandLine='\"';
	getFolderPath(commandLine, CSIDL_PROGRAM_FILES);
	if(commandLine[commandLine.length-1]!='\\'){
		commandLine+='\\';
	}
	commandLine+="POV-Ray for Windows v3.62\\bin\\pvengine.exe\" /EDIT \"";
	commandLine+=url;
	commandLine+='\"';

	if( CreateProcess(
		NULL, // No module name (use command line).
		const_cast<char*>(commandLine.c_str()), // Command line. 
		NULL,             // Process handle not inheritable. 
		NULL,             // Thread handle not inheritable. 
		FALSE,            // Set handle inheritance to FALSE. 
		0,                // No creation flags. 
		NULL,             // Use parent's environment block. 
		NULL,             // Use parent's starting directory. 
		&si,              // Pointer to STARTUPINFO structure.
		&pi )){             // Pointer to PROCESS_INFORMATION structure.    ) 
		
		CloseHandle( pi.hProcess );
		CloseHandle( pi.hThread );
	}
}

void returnToExitSeamless3d(){
	g_settings->writeINIFile();
	BackUp::deleteBackupFiles();
	DestroyWindow(hWndMain);
}

LRESULT CALLBACK MsgProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{

	g_before=g_selNode.getRaw();
	Art* part=dynamic_cast<Art*>(g_selNode.getRaw());
	Seamless* smls=0;
	float tran=(1/g_scene->getZoom())*.01F;


	Vec3f vec;
	Rotation tmpRot;
	
	if(!mouse3dMode){
		if(TRUE==mouse2dHandler(message, wParam, lParam)){
			return TRUE;
		}
	}
//	if(g_activeWindow->input(message, wParam, lParam)){
//		return 0;
//	}
	
#ifdef	OPENGL
	if(mouse3dMode){
		mouse3dHandler(message, wParam);
	}
#endif
	
    switch( message ){

		case WM_CREATE:{
			DragAcceptFiles(hWnd, TRUE);
			InitDirectInput(hWnd);
			g_bReady = true;
			//InputEvent::setupSequence(hWnd);
			break;
		}
		case WM_MOUSEWHEEL:{
			Window::activeOnMouseWheel(-(short)HIWORD(wParam)/WHEEL_DELTA);
			break;
		}
		case WM_SIZE:
				if(wParam==SIZE_MINIMIZED){
					g_minimized=true;
					break;
				}
				else if(wParam==SIZE_MAXIMIZED || wParam==SIZE_RESTORED){
					g_minimized=false;
				}
		case WM_MOVE:{
				RECT rect;
				GetWindowRect(hWndMain, &rect);
				g_settings->win3dPos.refValue().x = rect.left;
				g_settings->win3dPos.refValue().y = rect.top;
				g_settings->win3dSize.refValue().x = rect.right-rect.left - g_extraWinPix.x;
				g_settings->win3dSize.refValue().y = rect.bottom-rect.top - g_extraWinPix.y;
				g_windowSizeWindowBool=true;
				g_windowSizeWindowTime=g_time;
			}
			break;
		case WM_PAINT:
			PAINTSTRUCT ps;
			
			BeginPaint(hWnd, &ps);
			renderScene();
			//outputScreenShotForDemo();
			EndPaint(hWnd, &ps);
			break;

		case WM_KEYUP:
			switch (wParam){
			case 0x5a:	//"Z"
				g_zDragMode=false;
			case VK_LEFT:{
				g_leftKeyDown=false;
				break;
			}
			case VK_RIGHT:{
				g_rightKeyDown=false;
				break;
			}
			case VK_DOWN:{
				g_backwardKeyDown=false;
				break;
			}
			case VK_UP:{
				g_forwardKeyDown=false;
				break;
			}
			default:
				return DefWindowProc(hWnd, message, wParam, lParam);
			}
			break;
		case WM_SYSKEYDOWN:{
	//		if(0x8000&GetKeyState(VK_LMENU)){
				if(keyDownAlt(wParam)){
					break;
				}
	//		}
			return DefWindowProc(hWnd, message, wParam, lParam);
		}
		case WM_KEYDOWN:
			if(0x8000&GetKeyState(VK_LCONTROL)){
				if(keyDownCtrl(wParam)){
					break;
				}
			}
			else if(0x8000&GetKeyState(VK_LSHIFT)){
				if(keyDownShift(wParam)){
					break;
				}
			}
			else {
				if(keyDown(hWnd, message, wParam, lParam)){
					break;
				}
				else if(keyDown(wParam)){
					break;
				}
			}
			return DefWindowProc(hWnd, message, wParam, lParam);
		case WM_CHAR:{
			for(int i=g_callBackList.length; --i>=0;){
				if(CaretWindow* textCell=dynamic_cast<CaretWindow*>(g_callBackList[i])){
					textCell->charInput(wParam);
					return 0;
				}
			}
			return DefWindowProc(hWnd, message, wParam, lParam);
		}
		case WM_GETMINMAXINFO:
			// Prevent the window from going smaller than some minimum size
			((MINMAXINFO*)lParam)->ptMinTrackSize.x = 100;
			((MINMAXINFO*)lParam)->ptMinTrackSize.y = 100;
			break;
		case WM_MOUSEACTIVATE:   // sent when window changes active state
			return DefWindowProc( hWnd, MA_NOACTIVATE, wParam, lParam );
#ifndef FRAME_TIME_SYNC
		case WM_ACTIVATE:   // sent when window changes active state
			if(mouse3dMode){
				UnAcquire();
			}
			if ( WA_INACTIVE == wParam ){
				if(g_drag){
					g_drag->endDrag(g_mousePos);
				}
				Window::onSeamlessWindowsLoseFocus();
				g_bActive = false;
				return 0;
			}
			else{
				Window::activeWindow->onGainFocus();
				g_bActive = true;
			}
			// Set exclusive mode access to the mouse based on active state
			break;
		case WM_ACTIVATEAPP:
			InputEvent::resumeSequence();
			break;
#endif
		case WM_DROPFILES:{
			HDROP hDrop = (HDROP) wParam;
			char szGile[0x200];
			DragQueryFile(hDrop, 0, szGile, sizeof(szGile));
			DragFinish(hDrop);
			droppedFilename = szGile;
			openDroppedFile();
			break;
		}
		case WM_COMMAND:
			switch LOWORD(wParam){

	/*
			case MOUSE_3D:
				showCursor(FALSE);
				break;
			case MOUSE_2D:
				mouse3dMode=false;
				UnAcquire();
				showCursor(TRUE);
				break;
	*/
	//		case STRING_FROM_CHAT_WIN_COMMAND:{
	//			doStringFromChat(lParam);
	//			break;
	//		}

			case RESET_3D_CURSOR:
				//reset3dCursor();
				break;
			default:
				return DefWindowProc(hWnd, message, wParam, lParam);
			}
			break;
		case WM_CLOSE:
			if(Window::activeWindow->hog){
				PlaySound("SystemDefault", 0, SND_ASYNC);
				break;
			}
			saveSmlsFirst(returnToExitSeamless3d);
			break;
		case WM_DESTROY:
			destroyRender();
			
			FreeDirectInput();
			PostQuitMessage(0);
			
			ReleaseDC(hWnd, GetDC(hWnd));
			
			break;
			
		default:
			return DefWindowProc(hWnd, message, wParam, lParam);
	}
	return 0;
}




void initializeShortCuts();
void getMemberAddresses();
void convert_raw_bitmap_to_cpp();

void genButtons();

#include <crtdbg.h>
//static _CrtMemState s1, s2, s3;

//void processWords(int);
//void printOutLines();

const char* g_lpszCmdLine = 0;
bool openFileFromCommandLine(){
	g_processComnandLine = false;	
	if(strcmp(g_lpszCmdLine,"") != 0){
		String file = g_lpszCmdLine;
		
		if(file[0]=='\"'){
			char longName[0x200];
			file = file.substring(1,file.length-2);

			GetLongPathName( file.c_str(), longName, sizeof(longName));
			openSmlsLastPhase(longName);
		}
		else{
			openSmlsLastPhase(g_lpszCmdLine);
		}
		
		return true;
	}
	return false;
}

void saveMemToFiles(unsigned int width, unsigned int height);


INT WINAPI WinMain(HINSTANCE hInst, HINSTANCE, LPSTR lpszCmdLine, INT ){//
	g_lpszCmdLine = lpszCmdLine;

//	double a = 2.000001;
//	double b = 1;
//	bool result = roughGE(a, b);
//	result = result;

	// o  = old r = new
	//Vec3f v = Vec3f(0, 1, 0);
	//Rotation o = Rotation(0, 0, 1, PI*.5);
	//Vec3f v2 = o * v;
	//Rotation r = Rotation(0, 1, 0, PI);
	//r = o * r;
	//v = r * v;
	//v = v;

//	int* p = new int;
//	_CrtMemCheckpoint( &s1 );
	//delete p;
//	p = new int;
//	_CrtMemCheckpoint( &s2 );
//	int flag = _CrtMemDifference(&s3, &s1, &s2);

//	dLog(s1.lTotalCount);
//	dLog(s2.lTotalCount);

//	Rotation r = -Rotation(.92497, .268729, .268729, 1.648711);
//	genButtons();

//	convert_raw_bitmap_to_cpp();

	//	double freq=220;
//	double cent = 1200;
//	double answer = freq * pow((pow(2.0,1.0/1200.0)),cent);

//	drawText2("", 0, Vec2i(0,0), 0, 0, 0);

//	return 0;

//	unsigned char ch = '';
//	Rotation r = Rotation(1, 0, 0, -PI);
//	r*= Rotation(1, 0, 0, PI*.25);
	//return 0;

//	M_Vec3f ma=M_Vec3f(Vec3f(0,1,2));
//	M_Vec3f mb;
//	ma=mb;
//	__asm{
//		push 0xffffff
//		mov	 ecx,0xffffffff 
//	}

//	Vec3f vec = Vec3f(1,0,0) * translationMatrix(Vec3f(1,0,0)) * rotationMatrix(Rotation(0,1,0, PI*.5));


	setMathHiPresision();
	getMemberAddresses();

	//g_chatLines+=String("NSPE allows the user to hand edit the polygons on NURBS surfaces. This includes being able to drag the vertices anywhere along the NURBS surface as well as join the vertices together, break the vertices apart and color them. NSPE has a significant advantage over simply converting a NURBS surface to a polygon mesh for editing because NSPE lets the user be able continue to modify the NURBS surface for the hand edited polygon structure.");
	//g_chatLines+=String("http://www.seamless3d.com/forum/viewtopic.php?f=1&t=411");
	//g_chatLines+=String("Because NSPE ensures that when a polygon's vertex is dragged it will always be on the NURBS surface, NSPE greatly helps the user to avoid unintentionally changing the shape of the model when optimizing for real time animation.");
	//processWords(8);
	//printOutLines();

	g_extraWinPix.x = GetSystemMetrics(SM_CXSIZEFRAME) * 2;
	g_extraWinPix.y = GetSystemMetrics(SM_CYCAPTION) + GetSystemMetrics(SM_CYSIZEFRAME) * 2;
	


	try{
		mkdir(g_seamlessDir.c_str());
		mkdir(g_appDataDir.c_str());
		mkdir(g_undoDir.c_str());
		mkdir(g_screenshotDir.c_str());
		mkdir(g_demoDir.c_str());
		mkdir(g_demoCacheDir.c_str());

		SYSTEMTIME ToDaySystemTime;

		String debugStr="\n\n\nSeamless3d ";
		debugStr+=g_version;
#ifdef OPENGL
		debugStr+=" OpenGL";
#endif
		dLog(debugStr );
		GetSystemTime(&ToDaySystemTime);
		char tmpChar[0x200];
		sprintf(tmpChar,"GMT: %.2i/%.2i/%i  %.2i:%.2i",ToDaySystemTime.wDay,ToDaySystemTime.wMonth,ToDaySystemTime.wYear,ToDaySystemTime.wHour,ToDaySystemTime.wMinute);
		dLog(tmpChar);
		
		//sprintf(tmpChar,"OS: - Version: %i.%i - build: %i ",_winmajor,_winminor,_osver);
		dLog(tmpChar);
		
#ifndef OPENGL		
		if (!checkForRenderCaperbilities()){
			return TRUE;
		}
#endif
		
		g_hinst = hInst;
		initializeShortCuts();

		initLookup();

		Settings::loadINI();

		addNodesForNewFile();
		BackUp::createUniqueBackupName();

		Window::initializeWindows();
		initWinINet();

		// bumpy hack - recover auto-save file was inserted here

		initVisualClue();
		initRadialTug();


//		createSting();

		//////////////////////////////////
		// Register the window class
		WNDCLASSEX wcMainWin = { sizeof(WNDCLASSEX), CS_VREDRAW|CS_HREDRAW, MsgProc, 0L, 0L,
			GetModuleHandle(NULL), NULL, NULL, NULL, NULL,
			"wcMainWin", NULL };
		wcMainWin.hIcon= LoadIcon( hInst, MAKEINTRESOURCE(IDI_MAIN_ICON));
		wcMainWin.hCursor		= LoadCursor(NULL, IDC_ARROW);
		
		RegisterClassEx( &wcMainWin );
		
		char TempStr[0x200];
		makeTitleStr(TempStr);
 		
		/* Create a window of the previously defined class */
		hWndMain = CreateWindow( "wcMainWin", TempStr,
			//WS_SIZEBOX | WS_SYSMENU | WS_CAPTION | WS_MAXIMIZEBOX,
			WS_OVERLAPPEDWINDOW,
			g_settings->win3dPos.x, g_settings->win3dPos.y, 
			g_settings->win3dSize.x + g_extraWinPix.x, g_settings->win3dSize.y + g_extraWinPix.y,
			GetDesktopWindow(), NULL, wcMainWin.hInstance, NULL );

		
		GetClientRect(hWndMain, &clientOffest);
		ToolBarWin::initToolbar();
		g_stw=new SceneTree;
		g_rcp=new ControlPanel;
		g_acp=new AnimControlPanel;
		openSceneTreeInterface();

		// Initialize Direct3D
		if( SUCCEEDED(InitRender(hWndMain))){
			dLog("InitRender SUCCEEDED");
#ifdef BIND_DEVELOPER_SCRIPT
			initializeDeveloperScript();
#endif
			convertVisualCluesToRenderFormat();
			
			
			// Show the window
			ShowWindow( hWndMain, SW_SHOWDEFAULT );
			UpdateWindow( hWndMain );

			resetLooseEnds();
			Viewpoint::gotoFirstViewpoint();

			//Window::topWindow=g_stw.p;
			if(g_settings->showToolBar){
				Window::topWindow=g_tb.p;
			}
			g_stw->size=Vec2i(330,660); 
			g_stw->pos=Vec2i(g_3d->size.x-330,0);
			g_stw->newNodeBar.dstPos=Vec2i(g_stw->size.x-80, 30);
			g_stw->cp->size=Vec2i(350, g_stw->cp->size.y);
			g_stw->cp->pos=Vec2i(0,200);
			Window::postLoadProcessWin(*g_scene);
			if(g_tb->pos.y > g_3d->size.y - g_tb->size.y){
				g_tb->pos=Vec2i(0, g_3d->size.y - g_tb->size.y);
			}




			//Window::topWindow=g_pw.p;

			
			// Load keyboard accelerators
			hAccel = LoadAccelerators( hInst, MAKEINTRESOURCE(IDR_MAIN_ACCEL) );


#ifdef FRAME_TIME_SYNC
			g_timeBegan = g_time = 0;
#else
			g_timeBegan = g_time = timeGetTime() * MILISECONDS_TO_SECONDS;
#endif
		

			// Enter the message loop
			MSG msg;
			ZeroMemory( &msg, sizeof(msg) );

			if(g_createdINI){
				msG(menu_to_open_F10);
			}
			BOOL bGotMsg=FALSE;
			g_stw->size = Vec2i(g_stw->size.x, g_stw->size.y); //do we really need this?
			g_stw->paint();

			if(checkForLockFiles() == false){
				if(openFileFromCommandLine() == false){
					setupRecentFiles();
				}
			}


			while(WM_QUIT != msg.message){
				// Use PeekMessage() if the app is active, so we can use idle time to
				// render the scene. Else, use GetMessage() to avoid eating CPU time.

#ifdef RECORD_SOUNDMOVIE
				//if(g_playSoundMovie && g_playing){
				//	DWORD dwResult = MsgWaitForMultipleObjects( 1, &g_hNotificationEvent, FALSE, INFINITE, QS_ALLEVENTS);
				//	switch(dwResult){
				//		case WAIT_OBJECT_0 + 0:{
				//			if(g_playing){
				//				updatePlayBuffer();
				//				renderScene();
				//			}
				//			break;
				//		}
				//		case WAIT_OBJECT_0 + 1:{
				//			// Windows messages are available
				//			while(PeekMessage( &msg, NULL, 0, 0, PM_REMOVE )){ 
				//				TranslateMessage(&msg);
				//				DispatchMessage(&msg);
				//			}
				//		}
				//	}
				//}else
				if(g_recordSoundMovie){
					DWORD dwResult = MsgWaitForMultipleObjects( 1, &g_hNotificationEvent, FALSE, INFINITE, QS_ALLEVENTS);
					switch(dwResult){
						case WAIT_OBJECT_0 + 0:{
							renderScene();
							//outputScreenShotForDemo();
							break;
						}
						case WAIT_OBJECT_0 + 1:{
							// Windows messages are available
							while(PeekMessage( &msg, NULL, 0, 0, PM_REMOVE )){ 
								TranslateMessage(&msg);
								DispatchMessage(&msg);
							}
						}
					}
				}
				else{
					bGotMsg = PeekMessage( &msg, NULL, 0U, 0U, PM_REMOVE );
					if(!g_bActive){
	//					bGotMsg = GetMessage( &msg, NULL, 0U, 0U );
						//bGotMsg = PeekMessage( &msg, NULL, 0U, 0U, PM_REMOVE );
						Sleep(100);
					}
					
					if( bGotMsg ){
						// Translate and dispatch the message
						if( 0 == TranslateAccelerator( hWndMain, hAccel, &msg ) ){
							TranslateMessage( &msg );
							DispatchMessage( &msg );
						}
					}
					else{
						// Render a frame during idle time (no messages are waiting)
						if( g_bReady ){
							renderScene();
						}
						
						BackUp::checkForBackUpOperation();
					}
				}
#else
				bGotMsg = PeekMessage( &msg, NULL, 0U, 0U, PM_REMOVE );
				if(!g_bActive){
//					bGotMsg = GetMessage( &msg, NULL, 0U, 0U );
					//bGotMsg = PeekMessage( &msg, NULL, 0U, 0U, PM_REMOVE );
					Sleep(100);
				}
				
				if( bGotMsg ){
					// Translate and dispatch the message
					if( 0 == TranslateAccelerator( hWndMain, hAccel, &msg ) ){
						TranslateMessage( &msg );
						DispatchMessage( &msg );
					}
				}
				else{
					// Render a frame during idle time (no messages are waiting)
					if( g_bReady ){
						renderScene();
					}
					
					BackUp::checkForBackUpOperation();
				}
#endif
			}			
		}

		// Clean up everything and exit the app
		Cleanup();

		UnregisterClass( "chatWc", hInst);
		UnregisterClass( "wcMainWin", hInst);
		
		GetSystemTime(&ToDaySystemTime);
		sprintf(tmpChar,"GMT: %.2i/%.2i/%i  %.2i:%.2i",ToDaySystemTime.wDay,ToDaySystemTime.wMonth,ToDaySystemTime.wYear,ToDaySystemTime.wHour,ToDaySystemTime.wMinute);
		dLog(tmpChar);
		dLog("end");
		
	}
	catch(Field*e){
		dLog("error in "+e->getTypeStr());
	}
	catch(M_Float* e){
		dLog("error in M_Float");
		dLog((int)e);
		dLog((char *)e->showReason());
	}
	catch(M_Vec3f* e){
		dLog("error in M_Vec3f");
		dLog((int)e);
		dLog((char *)e->showReason());
	}
	catch(M_Vec2f* e){
		dLog("error in M_Vec2f");
		dLog((char *)e->showReason());
	}
	catch(String* e){
		dLog("error in String");
		dLog((char *)e->showReason());
	}
	catch(M_Color3f* e){
		dLog("error in M_Color3f");
		dLog((char *)e->showReason());
	}
	catch(M_Int* e){
		dLog("error in M_Int");
		dLog((char *)e->showReason());
	}
	catch(M_Node* e){
		dLog("error in M_Node");
		dLog((char *)e->showReason());
	}

	catch(M_Triangle* e){
		dLog("error in M_Triangle");
		dLog((int)e);
		dLog((char *)e->showReason());
	}

	catch(Node* e){
		dLog(String("Exception in ")+e->getTypeStr());
	}
	catch(...){
		dLog(String("Exception ..."));
	}

	InputEvent::resumeSequence();
	return 0;
}
void Scene::OpenOutputFolder::operator=(const bool& par){
	openWithHtmlBrowser(g_seamlessDir, "");
}
/*
inline void convertConcaveIFSIndexToTriIndex(const M_Vec3f& src_point, M_Int& index){
	M_Vec3f point;
	M_Int& tmpIndex=g_tmpIndex;
	tmpIndex.length=0;
	M_Int poly;
	for(int i = 0; i < index.length && index[i] >= 0; i++){
		poly+=index[i];
	}

	Rotation rot;

	for(int s = 0 ;;){
		int idx0 = poly[s++];
		if(s>=poly.length){
			s=0;
		}
		int idx1 = poly[s++];
		if(s>=poly.length){
			s=0;
		}
		int idx2 = poly[s++];
		if(s>=poly.length){
			s=0;
		}

		Vec3f v0 = src_point[idx0];
		Vec3f v1 = src_point[idx1];
		Vec3f v2 = src_point[idx2];

		if(thinTriangle(v0, v1, v2)){
			s -= 2;
			continue;
		}
		Vec3f normal = calcNormal(src_point[index[idx0]], src_point[index[idx1]], src_point[index[idx2]]);
		rot = Rotation(Vec3f(0, 0, 1), normal);
		break;
	}

	point.length = src_point.length;
	for(int i=0; i<point.length; i++){
		point[i] = rot * src_point[i];
	}

	int s=0;
	int c=0;
	while(poly.length>3){
		int idx0 = poly[s++];
		if(s>=poly.length){
			s=0;
			c++;
		}
		int idx1 = poly[s++];
		if(s>=poly.length){
			s=0;
			c++;
		}
		int idx2 = poly[s++];
		if(s>=poly.length){
			s=0;
			c++;
		}

		Vec3f v0 = point[idx0];
		Vec3f v1 = point[idx1];
		Vec3f v2 = point[idx2];

		if(!backFace(v0, v1, v2) || thinTriangle(v0, v1, v2) || vertexInsideTri(point, idx0, idx1, idx2)){
			if(c==3){
				s=0;
				c=0;
				rot=-rot;
				for(int i=0; i<point.length; i++){
					point[i] = rot * src_point[i];
				}
				poly.length=0;
				for(int i = 0; i < index.length && index[i] >= 0; i++){
					poly+=index[i];
				}
				tmpIndex.length=0;
				continue;
			}
			s -= 2;
			if(s < 0){
				s=poly.length + s;
			}
			continue;
		}
		tmpIndex+=idx0;
		tmpIndex+=idx1;
		tmpIndex+=idx2;
		tmpIndex+=-1;

		int g=s-2;
		if(g < 0){
			g=poly.length + g;
		}
		poly.deleteElement(g);
		s=0;
		c=0;
	}
	tmpIndex+=poly;
	index=tmpIndex;
}
*/
