/* Seamless3d

Copyright (C) 2001-2011 Graham John Perrett (AKA thyme) thyme@seamless3d.com www.seamless3d.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this
software and associated documentation files (the "Software"), to deal in the Software
without restriction, including without limitation the rights to use, copy, modify,
merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to the following
conditions:

The above copyright notice and this permission notice shall be included in all copies or
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT
OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.
*/

#include <windows.h>
#include <direct.h>

#include "seamless3d.h"
#include "techuelife_triangle_editor.h"
#include "seamless3d_gui_render.h"
#include "gui.h"
#include "scene_tree.h"
#include "robot_aid.h"


extern const char* msg_needPartOrCenterJoint;
extern const char* msg_needPartOrBezier;
extern const char* msg_centerJointHasNoRef;
extern const char* msg_needPart;
extern const char* msg_needWhitePart;
extern const char* msg_needBluePart;
extern const char* msg_needPartOrSmls;
extern const char* msg_needPartCopied;
extern const char* msg_needPartOrMorpher;
extern const char* msg_reverse_operation;
extern const char* msg_vertex_belongs_to_only_one;
extern const char* msg_divideMessage;
extern const char* msg_one_tri_must_exist;
extern const char* msg_both_tri_must_not_be_same;
extern const char* msg_both_tri_can_not_shar_3_points;
extern const char* msg_both_tri_must_share_2_points;
extern const char* msg_num_of_sel_must_be_more_than_one;
extern const char* msg_num_of_sel_must_be_even;

extern const char* name_front_view;
extern const char* name_left_view;
extern const char* name_right_view;
extern const char* name_back_view;
extern const char* name_birds_eye_view;
extern const char* name_worms_eye_view;
extern const char* name_showcase_view;
extern const char* name_floor;
extern const char* name_ruller;

extern ShapePtr g_vcSelectRectShape;
extern M_Matrix m_matrix;
extern M_Int m_smi;
extern M_Vec3f g_m_coord;

M_<S_VertexIndex> MultiSelect::triList;
M_<S_VertexIndex> MultiSelect::coordList;
M_ControlPoint MultiSelect::pointList;
M_Int MultiSelect::ctrlPolySelList;
int MultiSelect::selectedPoint = 0;
bool Viewpoint::neutral = false;
bool Anim::autoCapture = true;

ImmediatePart* convertPartToImmediate(Art* part);
Part* convertImmediateToPart(Art* immediate);
int getSeamlessParent(int line);
Seamless* getSeamless(int& selLine);
void buildModelMorpherForSelSmls(Seamless* smls);
StretchCurvePtr g_stretchCurve = new StretchCurve;
Ptr<TransformCP> g_transformCP = new TransformCP;
Ptr<PasteOptions> g_pasteOptions = new PasteOptions;
Ptr<DeleteOptions> g_deleteOptions = new DeleteOptions;
ConvertToPatchPtr g_convertToPatch=new ConvertToPatch;
Ptr<ReduceCapturedPoses> g_reduceCapturedPoses = new ReduceCapturedPoses;
void updateNurbsCurveKnotMid0(int dim, M_Float& knot);
void updateNurbsCurveKnotMid1(int dim, M_Float& knot);
void openWithHtmlBrowser(String url, const String htmlBrowser);

void replaceNodeRefInSceneTree(NodePtr from, Node* to);
void commonInitInsertNodeStuff(Node* node, Field*& field, Node*& newNode, unsigned fieldIdx, unsigned nodeType);
void checkForNameClash(M_Node& importDefList, M_Node& existingDefList);
void resetDefList();

bool g_forceSaveAs = true;
extern bool g_browseMode;

float g_joinRange=.001;
CenterJointPtr g_immediateCenterJoint=new CenterJoint;
extern Ptr<PartClue> g_vcArtTriBase;
extern Ptr<AnimClue> g_vcArtAnimBase;
extern Shape* g_vcEllipticalPolyline2D;
extern Shape* g_transformLeverClue;
extern bool g_optimized;

void searchDoBuild(M_Node& children);
void keyboardRadialTug(Vec3f coord, Vec3f pt);
void removeRouteFromProto(Node* windowNode);
ProtoDeclare* getSceneProtoDeclare(int selLine);
void insertElement(bool move, bool replace, int after);
void createUniqueName(Node* node);
void quantizeInc(float& v);
void quantizeDec(float& v);
void recurseShow(Art* part, bool show);
bool backFace(const Vec3f& v0, const Vec3f& v1, const Vec3f& v2);
bool backFace(const Vec2f& v0, const Vec2f& v1, const Vec2f& v2);
bool makeNameLegalForCpp(String& str);
void extractFileName(String& filename, const String& path);
void getFFmpegPath(String& commandLine);
float getHighestPoint(M_Vec2f& point);
float getLowestPoint(M_Vec2f& point);
bool getPolyMap(NurbsPatch*& patch, Stem*& stem, Polymap*& polymap, Art* part, int selTriIndex, ImmediatePart*& sourcePart);
bool isTriangleHit(const Vec2f& pt, const Vec2f* v);
void saveRelWinMousePos();
bool MultiSelect::backfaceIgnore = false;
bool MultiSelect::accumulate = false;
void closeRecentFileWindow();
void scrollUpCP();
void scrollDownCP();
void scrollPageUpCP();
void scrollPageDownCP();
extern Ptr<Palette> g_palette;
void applyOwnership(Art* smls);
void applyColor(Seamless* smls);
PaletteCP* paletteCPAllreadyOpen();
void captureFrame();
void showAllCages();
void hideSelectedCages();
void showSingleCage();
void incorporate();
void toggleRecordSoundMovie();
void calcAllPolyMidPoints();
void fillEdgesWithCtrlPolys();
void mergeCtrlPolys();
void mergeTo5CtrlPoly();
void mergeTo6CtrlPoly();
extern bool g_showControlPolys;
void toggleShowControlPolys();
void linkPointToEdge();
void openTransformCP();
void snapPivotPointToControlPoint();
void invertSelCtrlPoly();
void pasteColumn();
void deleteLink();
void compileMorph();
void randomizeHarmonicFreq();
void loadSDIF();
void makeTrianglesTiny();

extern bool g_allowPolyMorphCapture;
SingleButton* g_toggleBackfaceSelectButton = 0;
SingleButton* g_toggleAccumulateSelectButton = 0;
SingleButton* g_togglePolyMorphCaptureButton = 0;

void togglePolyMorphCapture(){
	g_allowPolyMorphCapture = !g_allowPolyMorphCapture;
	if(g_allowPolyMorphCapture){
		g_togglePolyMorphCaptureButton->srcPos = Vec2i(24 * 2, 208);
	}
	else{
		g_togglePolyMorphCaptureButton->srcPos = Vec2i(24 * 1, 208);
	}
}

Ptr<FloorHeight> g_floorHeight = new FloorHeight;

#ifdef PROG_ROB
	String g_upStr;
#endif

//void matchEndsForPatch();
void alignMidPoint();
void alignAllMidPoints();

void makeStitchPointsPlanar();
void updateNurbsCurveKnotBezier(int dim, M_Float& knot);

M_<ShortCutKey> globalShortCut = M_<ShortCutKey>(
	ShortCutKey(VK_F1, help),
	ShortCutKey(VK_F3, selectPreviousVertex),
	ShortCutKey(VK_F4, selectNextVertex),
	ShortCutKey(VK_F5, prevTriangle),
	ShortCutKey(VK_F6, nextTriangle),
	ShortCutKey(VK_F7, selectPreviousCornerInTri),
	ShortCutKey(VK_F8, selectNextCornerInTri),
	ShortCutKey(VK_F9, doBuild),
	//ShortCutKey(VK_F11, toggleRecordSoundMovie),
	ShortCutKey(VK_F12, outputScreenShotSingle)
//    ShortCutKey(VK_F11, spaghetti)
);

M_<ShortCutKey> globalShortCutAlt = M_<ShortCutKey>(
	ShortCutKey('R',	resetRotations),
	ShortCutKey('X',	zeroVetexX),
//	ShortCutKey('M',	matchEndsForPatch),
	ShortCutKey('J',	joinVertexKeepSourceColor),
	ShortCutKey('F',	captureFrame),
	ShortCutKey('I',	incorporate),


	ShortCutKey(VK_F10, buildF10Menu),
	ShortCutKey(VK_NUMPAD1, centerViewToControlPoint),
	ShortCutKey(VK_NUMPAD0, centerViewToPivotPoint),
//	ShortCutKey(VK_NUMPAD4, decPivotX),
//	ShortCutKey(VK_NUMPAD6, incPivotX),
//	ShortCutKey(VK_NUMPAD2, decPivotY),
//	ShortCutKey(VK_NUMPAD8, incPivotY),
//	ShortCutKey(VK_NUMPAD9, decPivotZ),
//	ShortCutKey(VK_NUMPAD3, incPivotZ)
	ShortCutKey(VK_NUMPAD4, decCoordX),
	ShortCutKey(VK_NUMPAD6, incCoordX),
	ShortCutKey(VK_NUMPAD2, decCoordY),
	ShortCutKey(VK_NUMPAD8, incCoordY),
	ShortCutKey(VK_NUMPAD9, decCoordZ),
	ShortCutKey(VK_NUMPAD3, incCoordZ)
);
M_<ShortCutKey> globalShortCutCtrl = M_<ShortCutKey>(
	ShortCutKey(VK_F2, openNodeWindow),
	ShortCutKey('O', openSmls),
	ShortCutKey('S', saveSmls),
	ShortCutKey('A', saveSmlsAs),
	ShortCutKey('N', newSmls),
	ShortCutKey('I', reverseIndices),
	ShortCutKey('B', invertBackgroundColor),
	ShortCutKey('C', remember),
	ShortCutKey('X', cut),
	ShortCutKey(VK_F10, buildF10Menu),
    //ShortCutKey(VK_F11, openNodeWindow),
    //ShortCutKey(VK_F11, openNodeWindow),

	ShortCutKey(219, decRed),
	ShortCutKey(221, incRed),
	ShortCutKey(186, decGreen)

);
M_<ShortCutKey> globalShortCutShift = M_<ShortCutKey>();
void initializeShortCuts(){
	globalShortCutCtrl+=ShortCutKey('Z', unDo);
	globalShortCutCtrl+=ShortCutKey('Y', reDo);
	globalShortCutCtrl+=ShortCutKey('J',	joinVertexKeepBothColors);
	globalShortCutCtrl+=ShortCutKey('P',	makeStitchPointsPlanar);
	globalShortCutCtrl+=ShortCutKey('N',	calcAllPolyMidPoints);
	globalShortCutCtrl+=ShortCutKey('T',	openTransformCP);
	globalShortCutCtrl+=ShortCutKey('E',	fillEdgesWithCtrlPolys);
	globalShortCutCtrl+=ShortCutKey('M', mergeCtrlPolys);
	globalShortCutCtrl+=ShortCutKey('5', mergeTo5CtrlPoly);
	globalShortCutCtrl+=ShortCutKey('6', mergeTo6CtrlPoly);
	globalShortCutCtrl+=ShortCutKey('G', toggleShowControlPolys);
	globalShortCutCtrl+=ShortCutKey('I', invertSelCtrlPoly);
	globalShortCutCtrl+=ShortCutKey('L', linkPointToEdge);
	globalShortCutCtrl+=ShortCutKey('F', flipPolarity);
	globalShortCutCtrl+=ShortCutKey('H', replaceColorPart);
	globalShortCutCtrl+=ShortCutKey('W', snapPivotPointToControlPoint);
	globalShortCutCtrl+=ShortCutKey('Q', pasteColumn);
	//globalShortCutCtrl+=ShortCutKey('I', alignOscToPatch);
	globalShortCutCtrl+=ShortCutKey('D', deleteLink);
	globalShortCutCtrl+=ShortCutKey('K', togglePolyMorphCapture);
	globalShortCutCtrl+=ShortCutKey('R', compileMorph);

	globalShortCutCtrl+=ShortCutKey(VK_F7, saveRelWinMousePos);

	globalShortCutCtrl+=ShortCutKey(222, incGreen);
	globalShortCutCtrl+=ShortCutKey(190, decBlue);
	globalShortCutCtrl+=ShortCutKey(191, incBlue);
	globalShortCutCtrl+=ShortCutKey(187, lightenVertex);
	globalShortCutCtrl+=ShortCutKey(189, darkenVertex);

	globalShortCutCtrl+=ShortCutKey(VK_NUMPAD4, decCoordX);
	globalShortCutCtrl+=ShortCutKey(VK_NUMPAD6, incCoordX);
	globalShortCutCtrl+=ShortCutKey(VK_NUMPAD2, decCoordY);
	globalShortCutCtrl+=ShortCutKey(VK_NUMPAD8, incCoordY);
	globalShortCutCtrl+=ShortCutKey(VK_NUMPAD9, decCoordZ);
	globalShortCutCtrl+=ShortCutKey(VK_NUMPAD3, incCoordZ);

	globalShortCutCtrl+=ShortCutKey(VK_SUBTRACT, decTugRadius);
	globalShortCutCtrl+=ShortCutKey(VK_ADD, incTugRadius);



	_3DWin::shortCut+=ShortCutKey(VK_ESCAPE, reset3dScene);
	_3DWin::shortCut+=ShortCutKey(0xdb,	decViewOrientationX);//[
	_3DWin::shortCut+=ShortCutKey(0xdd,	incViewOrientationX);//]
	_3DWin::shortCut+=ShortCutKey(0xba,	decViewOrientationY);//;
	_3DWin::shortCut+=ShortCutKey(0xde,	incViewOrientationY);//'
	_3DWin::shortCut+=ShortCutKey(0xbe,	decViewOrientationZ);//.
	_3DWin::shortCut+=ShortCutKey(0xbf,	incViewOrientationZ);///
	_3DWin::shortCut+=ShortCutKey(0xbc, resetViewOrientation);

	_3DWin::shortCut+=ShortCutKey(0xbb, zoomIn);
	_3DWin::shortCut+=ShortCutKey(0xbd, zoom);

	_3DWin::shortCut+=ShortCutKey(VK_PRIOR, Viewpoint::gotoPrevViewpoint);
	_3DWin::shortCut+=ShortCutKey(VK_NEXT, Viewpoint::gotoNextViewpoint);
	_3DWin::shortCut+=ShortCutKey(VK_NEXT, Viewpoint::gotoNextViewpoint);
	_3DWin::shortCut+=ShortCutKey(VK_NUMPAD2, decViewPositionY);
	_3DWin::shortCut+=ShortCutKey(VK_NUMPAD3, incViewPositionZ);
	_3DWin::shortCut+=ShortCutKey(VK_NUMPAD4, decViewPositionX);
	_3DWin::shortCut+=ShortCutKey(VK_NUMPAD5, resetViewPosition);
	_3DWin::shortCut+=ShortCutKey(VK_NUMPAD6, incViewPositionX);
	_3DWin::shortCut+=ShortCutKey(VK_NUMPAD8, incViewPositionY);
	_3DWin::shortCut+=ShortCutKey(VK_NUMPAD9, decViewPositionZ);
	_3DWin::shortCut+=ShortCutKey('1', decViewOrientation90X);
	_3DWin::shortCut+=ShortCutKey('2', incViewOrientation90X);
	_3DWin::shortCut+=ShortCutKey('3', decViewOrientation90Y);
	_3DWin::shortCut+=ShortCutKey('4', incViewOrientation90Y);
	_3DWin::shortCut+=ShortCutKey('5', decViewOrientation90Z);
	_3DWin::shortCut+=ShortCutKey('6', incViewOrientation90Z);

	_3DWin::shortCut+=ShortCutKey(' ',	toggleWireFrame);
	_3DWin::shortCut+=ShortCutKey('B',	breakVertexJoin);
	_3DWin::shortCut+=ShortCutKey('C',	colorVertex);
	_3DWin::shortCut+=ShortCutKey('D',	swapDivide);
	_3DWin::shortCut+=ShortCutKey('E',	incToNextTriThatSharesActivePoint);
	_3DWin::shortCut+=ShortCutKey('F',	forceToMirrorX);
	_3DWin::shortCut+=ShortCutKey('I',	swapVertexOrderInc);
	_3DWin::shortCut+=ShortCutKey('J',	joinVertexKeepDestColor);//			joinVertexImmediate(part,false)??????????????
	_3DWin::shortCut+=ShortCutKey('K',	pasteColorToTriangle);
	_3DWin::shortCut+=ShortCutKey('M',	mirrorTriangleImmediate);
	_3DWin::shortCut+=ShortCutKey('P',	projectSingleVertexOntoPoly);
	_3DWin::shortCut+=ShortCutKey('Q',	quickTriangle);
	_3DWin::shortCut+=ShortCutKey('R',	remember);
	_3DWin::shortCut+=ShortCutKey('S',	splitTriangleImmediate);
	_3DWin::shortCut+=ShortCutKey('T',	openSceneTree);
	_3DWin::shortCut+=ShortCutKey('U',	swapVertexOrderDec);
	_3DWin::shortCut+=ShortCutKey('W',	decToNextTriThatSharesActivePoint);
	_3DWin::shortCut+=ShortCutKey('Z',	setZDragNode);
	_3DWin::shortCut+=ShortCutKey('A',	alignAllMidPoints);
	_3DWin::shortCut+=ShortCutKey('Y',	unMultiSelTri);

	globalShortCutAlt += ShortCutKey('E',	randomizeHarmonicFreq);
	globalShortCutAlt += ShortCutKey('W',	loadSDIF);
	globalShortCutAlt += ShortCutKey('Y',	makeTrianglesTiny);

}
M_<ShortCutKey> _3DWin::shortCut = M_<ShortCutKey>(
	ShortCutKey(VK_SUBTRACT,	zoom),
	ShortCutKey(VK_ADD,			zoomIn),
	ShortCutKey(VK_DELETE, 		deleteTriangle)
);
M_<ShortCutKey> _3DWin::shortCutAlt = M_<ShortCutKey>(
	ShortCutKey('B',	breakColorJoin),
	ShortCutKey('C',	colorPart),
	ShortCutKey('F',	findThinTriangle),
	ShortCutKey('L',	findLayeredTriangle),
	ShortCutKey('M',	moveVertex),
	//ShortCutKey('O',	transferSingleVertex),
	ShortCutKey('A',	showAllCages),
	ShortCutKey('S',	showSingleCage),
	ShortCutKey('H',	hideSelectedCages)
);
M_<ShortCutKey> _3DWin::shortCutCtrl = M_<ShortCutKey>(
);
M_<ShortCutKey> _3DWin::shortCutShift = M_<ShortCutKey>(
	ShortCutKey('R',	cutRightFromVertex),
	ShortCutKey('J',	transferCloseVerticesJoin),
	ShortCutKey('L',	cutLeftFromVertex),
	ShortCutKey('U',	cutUpFromVertex),
	ShortCutKey('D',	cutDownFromVertex),

	ShortCutKey('X',	pasteAddCopyOfMirrorX),
	ShortCutKey('Y',	pasteAddCopyOfMirrorY),
	ShortCutKey('Z',	pasteAddCopyOfMirrorZ),
	ShortCutKey('0',	openSettings),
	ShortCutKey('A',	alignMidPoint),

	ShortCutKey(0xbb,	zoomIn)
);
M_<ShortCutKey> SceneTree::shortCut = M_<ShortCutKey>(
	ShortCutKey(VK_RIGHT,	incCaret),
	ShortCutKey(VK_LEFT,	decCaret),
	ShortCutKey(VK_HOME,	homeCaret),
	ShortCutKey(VK_END,		endCaret),
	ShortCutKey(VK_DELETE,	deleteChar),
	ShortCutKey(VK_ESCAPE,	SceneTree::escape)
);
M_<ShortCutKey> ChatWin::shortCut = M_<ShortCutKey>(
	ShortCutKey(VK_RIGHT,	incCaret),
	ShortCutKey(VK_LEFT,	decCaret),
	ShortCutKey(VK_HOME,	homeCaret),
	ShortCutKey(VK_END,		endCaret),
	ShortCutKey(VK_DELETE,	deleteChar),
	ShortCutKey(VK_ESCAPE,	SceneTree::escape)
);
M_<ShortCutKey> ChatWin::shortCutCtrl = M_<ShortCutKey>(
	ShortCutKey('V',	pasteFromClipboard)
);
M_<ShortCutKey> ChatWin::shortCutShift = M_<ShortCutKey>(
	ShortCutKey(VK_RIGHT,	incCaretShift),
	ShortCutKey(VK_LEFT,	decCaretShift),
	ShortCutKey(VK_HOME,	homeCaretShift),
	ShortCutKey(VK_END,		endCaretShift),
	ShortCutKey(VK_DELETE,	deleteCharShift)
);
M_<ShortCutKey> SceneTree::shortCutAlt = M_<ShortCutKey>();
M_<ShortCutKey> SceneTree::shortCutCtrl = M_<ShortCutKey>(
	ShortCutKey('V',	pasteFromClipboard)
);
M_<ShortCutKey> SceneTree::shortCutShift = M_<ShortCutKey>(
	ShortCutKey(VK_RIGHT,	incCaretShift),
	ShortCutKey(VK_LEFT,	decCaretShift),
	ShortCutKey(VK_HOME,	homeCaretShift),
	ShortCutKey(VK_END,		endCaretShift),
	ShortCutKey(VK_DELETE,	deleteCharShift),
	ShortCutKey(VK_UP,		SceneTree::decMultipleNodeSelection),
	ShortCutKey(VK_DOWN,	SceneTree::incMultipleNodeSelection)
);
M_<ShortCutKey> ControlPanel::shortCut = M_<ShortCutKey>(
	ShortCutKey(VK_RIGHT,	incCaret),
	ShortCutKey(VK_LEFT,	decCaret),
	ShortCutKey(VK_HOME,	homeCaret),
	ShortCutKey(VK_END,		endCaret),
	ShortCutKey(VK_DELETE,	deleteChar),
	ShortCutKey(VK_UP,		scrollUpCP),
	ShortCutKey(VK_DOWN,	scrollDownCP),
	ShortCutKey(VK_PRIOR,	scrollPageUpCP),
	ShortCutKey(VK_NEXT,	scrollPageDownCP)
);

M_<ShortCutKey> ControlPanel::shortCutAlt = M_<ShortCutKey>();
M_<ShortCutKey> ControlPanel::shortCutCtrl = M_<ShortCutKey>(
	ShortCutKey('V',	pasteFromClipboard)
);
M_<ShortCutKey> ControlPanel::shortCutShift = M_<ShortCutKey>(
	ShortCutKey(VK_RIGHT,	incCaretShift),
	ShortCutKey(VK_LEFT,	decCaretShift),
	ShortCutKey(VK_HOME,	homeCaretShift),
	ShortCutKey(VK_END,		endCaretShift),
	ShortCutKey(VK_DELETE,	deleteCharShift)
);
void houseKeepForNurbsViewChange(){
	if(NurbsPatch* patch = dynamic_cast<NurbsPatch*>(g_selNode.getRaw())){
		if(g_settings->tugRadius > 0){
			patch->houseKeep();
		}
	}
}
void needPartOrCenterJointMsg(){
	msG(msg_needPartOrCenterJoint);
}
void needPartOrBezierMsg(){
	msG(msg_needPartOrBezier);
}

void centerJointHasNoRefMsg(){
	msG(msg_centerJointHasNoRef);
}

void needWhitePartMsg(){
	msG(msg_needWhitePart);
}
void needBluePartMsg(){
	msG(msg_needBluePart);
}
void needPartMsg(){
	msG(msg_needPart);
}
void needPartOrSmlsMsg(){
	msG(msg_needPartOrSmls);
}
void needPartCopiedMsg(){
	msG(msg_needPartCopied);
}
void commonEndPivotPointChange(){
	g_stw->cp->paint();
	BackUp::monostable.beginCycle();
}
void commonEndVertexChange(){
	Seamless* smls = dynamic_cast<Art*>(g_selNode.getRaw())->smls;
	smls->doBuild();
	smls->updateAllVertexInBuffers();
	g_stw->cp->paint();
	BackUp::monostable.beginCycle();
}

void commonEndVertexBuild(){
	Seamless* smls = dynamic_cast<Art*>(g_selNode.getRaw())->smls;
	smls->doBuild();
	makePointerEqualIndexLocation();
	smls->ctrf=true;
	Window::paintAllControlPanels();
}

void replaceColorInPart(ArtPtr part, Color3f oldColor, Color3f newColor){
	ColorEffect* colorEffect=dynamic_cast<ColorEffect*>(part->workEffect);
	if(colorEffect==0)return;
	M_Color3f& color=colorEffect->color.refValue();
	M_Art& relevantPart=part->relevantPart;
	for(int k=relevantPart.length; --k>=0; ){
		M_Triangle& triangle=relevantPart[k]->triangle;
		for(int i=triangle.length*3; --i>=0; ){
			if(triangle.refPart(i)==part){
				if(color[triangle.refTexIndex(i)]==oldColor){
					color[triangle.refTexIndex(i)]=newColor;
				}
			}
		}
	}
	part->smls->updateAllVertexInBuffers();
	BackUp::monostable.beginCycle();
}
void transferCloseVertices(){
	if(Art* part=dynamic_cast<Art*>(g_selNode.getRaw())){
		if(Art* copied=dynamic_cast<Art*>(g_copyNode.getRaw())){
			copied->transferCloseVertices(part, g_joinRange);
			part->smls->convertToRenderFormat();
			makePointerEqualIndexLocation();
			BackUp::monostable.beginCycle();
		}
		else{
			needPartCopiedMsg();
		}
	}
	else{
		needPartMsg();
	}
}
void pasteCloseVerticesJoin(Art* source, Art* copied, bool dontShareColor){

	SeamlessPtr smls=copied->smls;

	JoinCloseVertex jcv;
	jcv.part+=copied;
	jcv.source+=source;

	if(copied->plidx<source->plidx){
		int question=msG(msg_reverse_operation ,MB_OKCANCEL|MB_ICONQUESTION);
		if (question==IDOK){
			jcv.part+=source;
			jcv.source+=copied;
		}
		else{
			return;
		}

	}


	jcv.srcColor=dontShareColor;
	jcv.doBuild();

}
inline bool checkForMatchingVertex(M_Triangle& triangle, int i,int j){
	for(int b=3; --b>=0;j++){
		if(triangle.refPart(i)==triangle.refPart(j)&&triangle.refIndex(i)==triangle.refIndex(j)){
			return true;;
		}
	}
	return false;
}
inline bool layeredTri(M_Triangle& triangle, int i,int j){
	for(int a=3; --a>=0;i++){
		if(checkForMatchingVertex(triangle, i,j)==false){
			return false;
		}
	}
	return true;
}
void findLayeredTriangle(){
	Seamless* smls=dynamic_cast<Seamless*>(g_selNode.getRaw());
	if(Art* part=dynamic_cast<Art*>(g_selNode.getRaw())){
		smls=part->smls;
	}
	if(smls){
		DelPoly dp;
		dp.poly.resize(1);
		int delCount=0;
		M_Art& partList = smls->partList;
		for(int k=partList.length; --k>=0;){
			ArtPtr part=partList[k];
			M_Triangle& triangle=part->triangle;
			for(int i=triangle.length*3; --i>=0;){
				for(int j=triangle.length*3; --j>=0;){
					if(triangle.refPart(i)==triangle.refPart(j)&&triangle.refIndex(i)==triangle.refIndex(j)){
						int ib=i/3;
						int jb=j/3;
						if(ib==jb)continue; // if comapiring same triangle forget it
						if(layeredTri(triangle ,ib*3,jb*3)){
							delCount++;
							dp.part=part;
							dp.poly.refValue()[0]=jb;
							dp.doBuild();
							i=triangle.length*3;
							j=0;// get us back to i loop start
						}
					}
				}
			}
		}
		dLog(String("deleted ")+String(delCount)+" layered triangles");
		smls->convertToRenderFormat();
		makePointerEqualIndexLocation();
		g_stw->genNodeList();
		g_stw->paint();
	}
	else{
		needPartOrSmlsMsg();
	}
}
void help(){
	openWithHtmlBrowser("http://www.seamless3d.com/tut", g_settings->htmlBrowser.refValue());
}
void about(){
			msG("Seamless3d Copyright  2001-2010 Graham John Perrett\n\n\
Seamless3d is FREE Open Source 3d modelling software under the MIT license\n\
The Seamless3d modelling language is an open standard designed for the real world, verifiable by Seamless3d Software\n\
Seamless3d is typically very stable when used with experience but I must stress it is still a program being developed so use with care. Backup any files you are working on regularly!\n*Important*:\nWhen the output button is pressed Seamless3d will generate a file with the same filename as the .smls file that is open with the appropriate filename extension. (.x3d;.x3dz;.x3dv;.x3dvz;.wrl;.wrz) Any file that has this name or extension will get overwritten so take care where you leave files with matching names if you don't want them overwritten.\n\
Please send bug reports or questions or files that won't open any more to the Seamless3d forum: www.seamless3d.com or my email thyme@seamless3d.com\n\
\nSeamless3d developed and written by:\n\n\tGraham John Perrett (thyme) thyme@seamless3d.com\n\twww.seamless3d.com\n\ngzip C code (zlib 1.2.3 general purpose data compression library) written by:\n\n\tJean-loup Gailly gzip@prep.ai.mit.edu\n\tMark Adler madler@alumni.caltech.edu\n\nSome quaternion/rotation maths C code that has been adapted was originally written by:\n\n\tGavin Bell");
}
void openSceneTree(){//IDD_SCENE_TREE_WINDOW
#ifdef PROG_ROB
	dLog("new ClickButton(openSceneTree),");
#endif
	closeRecentFileWindow();
	Window::activeWindow = g_stw.p;
	Window::activeWindow = g_stw->cp.p;
}
/*
void spaghetti(){//SPAGHETTI
	g_spaghetti=(bool)(g_spaghetti^true);
	houseKeepVisualClue();
}
*/
void resetViewPosition(){//RESET_POSITION
	g_scene->translation=Vec3f(0,0,0);
}
void resetViewOrientation(){//RESET_ORIENTATION
#ifdef PROG_ROB
			dLog("new ClickButton(resetViewOrientation),");
#endif
	g_scene->rotation=Rotation();
	houseKeepForNurbsViewChange();
}
void reset3dScene(){//RESET_ALL	VK_ESCAPE:
	InputEvent::resumeSequence();
	g_pointer.translation=Vec3f(0,0,0);
//	g_pointer.rotation=Rotation();
//	g_pointer.scale=Vec3f(1,1,1);

	g_scene->translation=Vec3f();
	g_scene->rotation=Rotation();
	g_scene->scale=Vec3f(1,1,1);

	Viewpoint::gotoFirstViewpoint();
//	BackUp::monostable.beginCycle();

	houseKeepVisualClue();
}

void commonCenterViewToPivotPoint(Art* part){
#ifdef PROG_ROB
	dLog("new ClickButton(centerViewToPivotPoint),");
#endif
//	g_scene->rotation=Rotation();
	part->smls->catIMatrix();
	part->smls->catMatrix();

	g_scene->translation=-(part->matrix * part->absolutePivotPoint);

	g_pointer.translation=Vec3f();
//	g_pointer.rotation=Rotation();
}
void Art::centerViewToControlPoint(){
	Vec3f crd;
	if(Window::animMode()){
		if(smls->animCP == false && MultiSelect::coordList.length == 1){
			smls->genDynamicMatricesAndPopulate();
			int idx = index[0];
			if(idx < coord.length){
				crd = m_matrix[m_smi[fromCoord + idx]] * (g_m_coord[fromCoord + idx] - absolutePivotPoint);
			}
			g_scene->translation = -crd;
			g_pointer.translation = Vec3f();
			return;
		}
		if(controlPoint.length == 0){
			return;
		}
		if(MultiSelect::selectedPoint >= controlPoint.length){
			MultiSelect::selectedPoint = 0;
		}
		ControlPoint& cp = controlPoint[MultiSelect::selectedPoint];
		NurbsPatch* patch = cp.p;
		patch->index[0] = cp.i;

		crd = patch->dpoint[cp.i];
	}
	else{
		int idx = index[0];
		if(idx<coord.length){
			crd = coord[idx];
		}
	}
	g_scene->translation = -crd;
	g_pointer.translation = Vec3f();
}
void BezierLathe::centerViewToControlPoint(){
	int idx=index[0];
	Vec3f vec=Vec3f(point[idx].x, point[idx].y, 0);
	g_scene->translation=-vec;

	g_pointer.translation=Vec3f();
}
void BezierLatheBuild::centerViewToControlPoint(){
	int idx=index[0];
	Vec3f vec = scaleMatrix(scale) * rotationMatrix(rotation) * translationMatrix(translation) * Vec3f(point[idx].x, point[idx].y, 0);
	g_scene->translation=-vec;

	g_pointer.translation=Vec3f();
}
void NurbsLathe::centerViewToControlPoint(){
	int idx = index[0];
	Vec3f vec = scaleMatrix(scale) * rotationMatrix(rotation) * translationMatrix(translation) * Vec3f(dpoint[idx].x, dpoint[idx].y, 0);

	g_scene->translation = -vec;
}
void NurbsTrack::centerViewToControlPoint(){
	int idx = index[1];
	Vec3f vec = keyValue[idx];
	g_scene->translation = -vec;
	g_pointer.translation = Vec3f();
}
void BendyLathe::centerViewToControlPoint(){
	Vec3f vec;
	if(showRingControlPoints){
		vec = ringBase[index[2]];
	}
	else{
		vec = dpoint[index[0]];
	}
	vec = scaleMatrix(scale) * rotationMatrix(rotation) * translationMatrix(translation) * vec;
	g_scene->translation=-vec;
}

void Oscillator::centerViewToControlPoint(){
	int idx = index[0];
	Vec3f vec = Vec3f(fdpoint[idx].x, fdpoint[idx].y / 10, 0);;
	g_scene->translation = -vec;

	g_pointer.translation = Vec3f();
}

void Envelope::centerViewToControlPoint(){
	int idx = index[0];
	Vec3f vec = Vec3f(point[idx].x, point[idx].y / 10, 0);;
	g_scene->translation = -vec;

	g_pointer.translation = Vec3f();
}

void AmplitudeModulator::centerViewToControlPoint(){
	int idx = index[0];
	Vec3f vec = Vec3f(wet[idx].x, wet[idx].y / 10, 0);;
	g_scene->translation = -vec;

	g_pointer.translation = Vec3f();
}

void FrequencyModulator::centerViewToControlPoint(){
	int idx = index[0];
	Vec3f vec = Vec3f(wet[idx].x, wet[idx].y / 10, 0);;
	g_scene->translation = -vec;

	g_pointer.translation = Vec3f();
}

void Chords::centerViewToControlPoint(){
	int idx = index[0];
	Vec3f vec = Vec3f(spectrum[idx].x, spectrum[idx].y / 10, 0);;
	g_scene->translation = -vec;

	g_pointer.translation = Vec3f();
}

void NurbsPatch::centerViewToControlPoint(){
	int idx = index[0];

	Vec3f vec = scaleMatrix(scale) * rotationMatrix(rotation) * translationMatrix(translation) * dpoint[idx];

	g_scene->translation = -vec;

	g_pointer.translation = Vec3f();
}

void Harmonic::centerViewToControlPoint(){
}

void HarmonicPatch::centerViewToControlPoint(){
	int idx=index[0];
	Vec3f vec=dpoint[idx];
	g_scene->translation=-vec;

	g_pointer.translation=Vec3f();
}
void RingBuild::centerViewToControlPoint(){
	Vec3f vec = translationMatrix(location.refValue()[index[0]]) * 
		scaleMatrix(scale) * rotationMatrix(rotation) * translationMatrix(translation) * Vec3f();

	g_scene->translation=-vec;

	g_pointer.translation=Vec3f();
}
void CCLathe::centerViewToControlPoint(){
	int idx=index[0];
	Vec2f point = seg[idx].Pf;
	Vec3f vec=Vec3f(point.x, point.y, 0);

	g_scene->translation=-vec;
	g_pointer.translation=Vec3f();
}
void VisualAid::centerViewToControlPoint(){
	needPartOrBezierMsg();
}
void centerViewToControlPoint(){
#ifdef PROG_ROB
	dLog("new ClickButton(centerViewToControlPoint),");
#endif
	if(Subdivision* subdivision=dynamic_cast<Subdivision*>(g_selNode.getRaw())){
		Vec3f crd;
		int idx=subdivision->index[0];
		if(idx<subdivision->point.length){
			crd=subdivision->point[idx];
		}
		g_scene->translation=-crd;
		g_pointer.translation=Vec3f();
		
	}
	else if(VisualAid* visualAid = dynamic_cast<VisualAid*>(g_selNode.getRaw())){
		visualAid->centerViewToControlPoint();
	}
	else{
		needPartOrBezierMsg();
	}
}
void centerViewToPivotPoint(){//CENTRE_PART_VIEW
	if(Art* part=dynamic_cast<Art*>(g_selNode.getRaw())){
		commonCenterViewToPivotPoint(part);
	}
	else if(CenterJoint* cj=dynamic_cast<CenterJoint*>(g_selNode.getRaw())){
		if(Art* part=dynamic_cast<Art*>(cj->part.p)){
			commonCenterViewToPivotPoint(part);
		}
		else{
			centerJointHasNoRefMsg();
		}
	}
	else if(RingBuild* ringBuild=dynamic_cast<RingBuild*>(g_selNode.getRaw())){
		Vec3f vec = translationMatrix(ringBuild->location.refValue()[ringBuild->index[0]]) * 
			scaleMatrix(ringBuild->scale.refValue()) * rotationMatrix(ringBuild->rotation.refValue()) * translationMatrix(ringBuild->translation.refValue()) * Vec3f();

		g_scene->translation=-vec;

		g_pointer.translation=Vec3f();
	}
	else{
		needPartOrCenterJointMsg();
	}
}
void findThinTriangle(){//FIND_THIN_TRIANGLE
	if(Seamless* smls=dynamic_cast<Seamless*>(g_selNode.getRaw())){
		findThinTriangle(smls);
	}
	else if(Art* part=dynamic_cast<Art*>(g_selNode.getRaw())){
		findThinTriangle(part->smls);
	}
}
void selectPreviousVertex(){//SELECT_PREV_VERTEX
	if(Art* part=dynamic_cast<Art*>(g_selNode.getRaw())){
		Seamless* smls=part->smls;
		if(smls->ActCtrlPnt>0){
			smls->ActCtrlPnt--;
		}
		else if(part->plidx>1){
			part=smls->partList[part->plidx-1];;
			g_stw->setSelNodeLine(part);
			smls->ActCtrlPnt=part->coord.length-1;
			if(smls->ActCtrlPnt<0){
				smls->ActCtrlPnt=0;
			}
		}
		makePointerEqualIndexLocation();
	}
	else{
		needPartMsg();
	}
}
void selectNextVertex(){//SELECT_NEXT_VERTEX
#ifdef PROG_ROB
	dLog("new IncButtonY(selectPrevNextVertex),");
#endif
	if(Art* part=dynamic_cast<Art*>(g_selNode.getRaw())){
		Seamless* smls=part->smls;
		if(smls->ActCtrlPnt<part->coord.length-1){
			smls->ActCtrlPnt++;
		}
		else if(part->plidx<smls->partList.length-1){
			g_stw->setSelNodeLine(smls->partList[part->plidx+1]);
			smls->ActCtrlPnt=0;
		}
		makePointerEqualIndexLocation();
	}
	else{
		needPartMsg();
	}
}
void nextTriangle(){
	if(Art* part=dynamic_cast<Art*>(g_selNode.getRaw())){
		Seamless* smls=part->smls;
		if(smls->animCP){
			g_transformCP->rotAxisX = true;
			return;
		}
		part=smls->actTriPart;
		if(part==0)return;
		int k=part->plidx;
		M_Art& partList=smls->partList;
		int actTri=smls->ActTri+1;
		while(1){
			if(actTri>=part->triangle.length){
				if(++k>=partList.length){
					return;
				}
				part=partList[k];
				actTri=0;
				continue;
			}
			break;
		}
		smls->ActTri=actTri;
		smls->actTriPart=part;
		g_stw->setSelNodeLine(part);
		S_Triangle tri=part->triangle[smls->ActTri];
		while(tri.refPart(smls->ActTriCorner)!=part){
			smls->ActTriCorner=(smls->ActTriCorner+1)%3;
		}
		smls->ActCtrlPnt=tri.refIndex(smls->ActTriCorner);;
		makePointerEqualIndexLocation();
	}
	else if(NurbsPatch* nurbsPatch = dynamic_cast<NurbsPatch*>(g_selNode.getRaw())){
		g_transformCP->rotAxisX = true;

		//nurbsPatch->smls->getNextCtrlPoly();
		//nurbsPatch->houseKeep();
	}
	else if(Subdivision* subdivision=dynamic_cast<Subdivision*>(g_selNode.getRaw())){
		subdivision->selNextPoly();
	}
	else{
		needPartMsg();
	}
}
void prevTriangle(){
	if(Art* part=dynamic_cast<Art*>(g_selNode.getRaw())){
		Seamless* smls=part->smls;
		part=smls->actTriPart;
		if(part==0)return;
		int k=part->plidx;
		M_Art& partList=smls->partList;
		int actTri=smls->ActTri-1;
		while(1){
			if(actTri<0){
				if(--k<0){
					return;
				}
				part=partList[k];
				actTri=part->triangle.length-1;
				continue;
			}
			break;
		}
		if(actTri<0){
			actTri=0;
		}
		smls->ActTri=actTri;
		smls->actTriPart=part;
		g_stw->setSelNodeLine(part);
		S_Triangle tri=part->triangle[smls->ActTri];
		while(tri.refPart(smls->ActTriCorner)!=part){
			smls->ActTriCorner=(smls->ActTriCorner+5)%3;
		}
		smls->ActCtrlPnt=tri.refIndex(smls->ActTriCorner);;
		makePointerEqualIndexLocation();
	}
	else if(NurbsPatch* nurbsPatch = dynamic_cast<NurbsPatch*>(g_selNode.getRaw())){
		nurbsPatch->smls->getPrevCtrlPoly();
		nurbsPatch->houseKeep();
	}
	else if(Subdivision* subdivision=dynamic_cast<Subdivision*>(g_selNode.getRaw())){
		subdivision->selPrevPoly();
	}
	else{
		needPartMsg();
	}
}

void selectCornerInTri(Seamless* smls){
	g_stw->setSelNodeLine(smls->actTriPart->triangle[smls->ActTri].refPart(smls->ActTriCorner));
	smls->ActCtrlPnt=smls->actTriPart->triangle[smls->ActTri].refIndex(smls->ActTriCorner);
	makePointerEqualIndexLocation();
}
void selectPreviousCornerInTri(){//SELECT_PREV_CORNER_IN_TRI
	if(Art* part=dynamic_cast<Art*>(g_selNode.getRaw())){
		Seamless* smls=part->smls;
		if(smls->animCP){
			g_transformCP->rotAxisY = true;
			return;
		}
		if(smls->actTriPart!=0){
			smls->ActTriCorner=(smls->ActTriCorner+5)%3;
			selectCornerInTri(smls);
		}
	}
	if(dynamic_cast<NurbsPatch*>(g_selNode.getRaw())){
		g_transformCP->rotAxisY = true;
	}
	else{
		needPartMsg();
	}
}
void selectNextCornerInTri(){//SELECT_NEXT_CORNER_IN_TRI
	if(Art* part=dynamic_cast<Art*>(g_selNode.getRaw())){
		Seamless* smls=part->smls;
		if(smls->animCP){
			g_transformCP->rotAxisZ = true;
			return;
		}
		if(smls->actTriPart!=0){
			smls->ActTriCorner=( smls->ActTriCorner+1)%3;
			selectCornerInTri(smls);
		}
	}
	if(dynamic_cast<NurbsPatch*>(g_selNode.getRaw())){
		g_transformCP->rotAxisZ = true;
	}
	else{
		needPartMsg();
	}
}

void joinVertexImmediate(Art* copyPart, ImmediatePart* part, Art* remActTriPart, Art* actTriPart, const int remActCtrlPnt, const int ActCtrlPnt, const int remActTri, const int ActTri, const int remActTriCorner, const int ActTriCorner, bool srcColor, bool dstColor, bool move, bool endWithBuild = true){
#ifdef PROG_ROB
	if(move){
		dLog("new MoveToF10(),");
		dLog("new Move_MenuItem(menu_join_move),");
		dLog("new Move_MenuItem(menu_join_vertex_keep_both_colors),");
		dLog("new Move_MenuItem(menu_move_vertex),");
		dLog("new Click,");
	}
	else if(srcColor && dstColor){
		dLog("new ClickButton(joinVertexKeepBothColors),");
	}
	else if(dstColor){
		dLog("new ClickButton(joinVertexKeepDestColor),");
	}
	else if(srcColor){
		dLog("new MoveToF10(),");
		dLog("new Move_MenuItem(menu_join_move),");
		dLog("new Move_MenuItem(menu_join_vertex_keep_source_color),");
		dLog("new Click,");
	}
#endif

	Seamless* smls=part->smls;
	BackUp::monostable.beginCycle();

	fastFullBuildForAutoIn(true);

	if(actTriPart==remActTriPart && actTriPart->triangle.refIndex(ActTri*3+ActTriCorner)==remActTriPart->triangle.refIndex(remActTri*3+remActTriCorner) && ActTri == remActTri){
		actTriPart->triangle.refTexIndex(ActTri*3+ActTriCorner)=remActTriPart->triangle.refTexIndex(remActTri*3+remActTriCorner);
	}
	else {
		if(move){
			M_Vec3f& coord=part->coord.refValue();
			Vec3f pt = copyPart->matrix*copyPart->coord.refValue()[remActCtrlPnt];

			Vec3f newPt=pt-part->matrix*Vec3f();
			part->smls->senseMirrorVertex(part, ActCtrlPnt, newPt);
			coord[ActCtrlPnt]=newPt;

		}
		else{
			JoinVertex joinVertex;
			joinVertex.srcCoord=getIndexRelativeToPart( part, actTriPart, ActTri*3+ActTriCorner);
			joinVertex.dstCoord=getIndexRelativeToPart( copyPart, remActTriPart, remActTri*3+remActTriCorner);
			joinVertex.part=copyPart;
			joinVertex.source=part;
			joinVertex.srcColor=srcColor;
			joinVertex.dstColor=dstColor;
			joinVertex.doBuild();
		}
	}
	if(endWithBuild){
		smls->doBuild();
		smls->ctrf=true;
		if(move==false){
			//g_stw->selLine=getNodeLine(copyPart);
			smls->ActTri=0;
			smls->ActTriCorner=0;
			smls->actTriPart=0;
			MultiSelect::triList.length = 0;
			MultiSelect::coordList.length = 0;
			copyPart->index[0] = 0;
			g_stw->setResultsForSelectingVertex(copyPart);

	//		if(copyPart == part && smls->ActCtrlPnt < remActCtrlPnt){
	//			smls->ActCtrlPnt=remActCtrlPnt-1;
	//		}
	//		else{
	//			smls->ActCtrlPnt=remActCtrlPnt;
	//		}
	//		makePointerEqualIndexLocation();
			g_stw->paint();
		}
		Window::paintAllControlPanels();
	}
}
inline Vec2f get2dCoord(const Art* part, const int& i, const int& perspective){
	Vec3f crd3f = part->matrix * g_scene->getViewMatrix() * part->coord.refValue()[i];
	Vec2f crd;
	if(perspective){
		crd=Vec2f(
			crd3f.x / (-crd3f.z + g_viewDist),
			crd3f.y / (-crd3f.z + g_viewDist)
		);
	}
	else{
		crd=Vec2f(crd3f.x, crd3f.y);
	}
	return crd;
}
int getClosestCorner(const float* dist){
	float minDist = dist[0];
	int minCorner = 0;
	for(int i = 3; --i>=1;){
		if(dist[i] < minDist){
			minDist = dist[i];
			minCorner = i;
		}
	}
	return minCorner;
}
Art* getClosestVertex(Art* selPart, const int selIdx, Art*& dstTriPart, int& dstIndex, int& dstTriangle, int& dstCorner, const int perspective){
	if(selIdx >= selPart->coord.length){
		return 0;
	}
	Vec2f selCoord = get2dCoord(selPart, selIdx, perspective);

	Seamless* smls = selPart->smls;
	M_Art& partList = smls->partList;

	float minDist = 10000000;
	Art* dstPart = 0;

	for(int k = partList.length; --k>=0;){
		Art* testTriPart = partList[k];
		const M_Triangle& triangle = testTriPart->triangle;
		for(int i = triangle.length; --i>=0;){
			S_Triangle tri = triangle[i];
			const Vec2f v[3] = {
				get2dCoord(tri.pa, tri.ia, perspective),
				get2dCoord(tri.pb, tri.ib, perspective),
				get2dCoord(tri.pc, tri.ic, perspective)
			};
			if(false == backFace(v[0], v[1], v[2])){
				continue;
			}
			const float dist[3] = {
				(v[0] - selCoord).length(),
				(v[1] - selCoord).length(),
				(v[2] - selCoord).length()
			};
			if(isTriangleHit(selCoord, v)){
				if((
					(tri.pa == selPart && selIdx == tri.ia) || 
					(tri.pb == selPart && selIdx == tri.ib) ||
					(tri.pc == selPart && selIdx == tri.ic)) == false){

					dstCorner = getClosestCorner(dist);
					if(tri.refPart(dstCorner)->show){
						switch(dstCorner){
							case 0:
								dstPart = tri.pa;
								dstIndex = tri.ia;
								break;
							case 1:
								dstPart = tri.pb;
								dstIndex = tri.ib;
								break;
							case 2:
								dstPart = tri.pc;
								dstIndex = tri.ic;
								break;
						}
						dstTriPart = testTriPart;
						dstTriangle = i;
						return dstPart;
					}
				}
			}
			if(dist[0] < minDist && (tri.pa == selPart && selIdx == tri.ia) == false && tri.pa->show){
				dstCorner = 0;
				minDist = dist[0];
				dstPart = tri.pa;
				dstIndex = tri.ia;
				dstTriPart = testTriPart;
				dstTriangle = i;
			}
			if(dist[1] < minDist && (tri.pb == selPart && selIdx == tri.ib) == false && tri.pb->show){
				dstCorner = 1;
				minDist = dist[1];
				dstPart = tri.pb;
				dstIndex = tri.ib;
				dstTriPart = testTriPart;
				dstTriangle = i;
			}
			if(dist[2] < minDist && (tri.pc == selPart && selIdx == tri.ic) == false && tri.pc->show){
				dstCorner = 2;
				minDist = dist[2];
				dstPart = tri.pc;
				dstIndex = tri.ic;
				dstTriPart = testTriPart;
				dstTriangle = i;
			}
		}
	}
	return dstPart;
}

void Seamless::moveNurbsCP(){
	if(MultiSelect::pointList.length != 1){
		return;
	}

	for(int i = MultiSelect::pointList.length; --i >= 0;){
		NurbsPatch* selPatch = MultiSelect::pointList[i].p;
		int selIdx = MultiSelect::pointList[i].i;

		Vec3f selPoint = selPatch->dpoint[selIdx];

		float shortest = 100000000;
		int shortestIdx = 0;
		int g = 0;
		NurbsPatch* shortestPatch = 0;

		for(int k = cpNodeList.length; --k >= 0;){
			NurbsPatch* patch = cpNodeList[k];
			M_Vec3f& dpoint = patch->dpoint.refValue();
			for(g = dpoint.length; --g >= 0;){
				if(patch != selPatch || selIdx != g){
					Vec3f vec = patch->dpoint[g];
					float distance = (selPoint - vec).length();
					if(distance < shortest){
						shortest = distance;
						shortestPatch = patch;
						shortestIdx = g;
					}
				}
			}
		}
		selPatch->point.refValue()[selIdx] = shortestPatch->dpoint[shortestIdx] * selPatch->weight[selIdx];
		//selPatch->weight.refValue()[selIdx] = shortestPatch->weight[shortestIdx];
		selPatch->processFields();
	}
	BackUp::monostable.beginCycle();
	doBuild();
	houseKeepVisualClue();
}

extern ImmediatePart* g_workPolyMapColor;
extern ImmediatePart* g_workPolyMapTexture;

ImmediatePart* selectPolyMapPart(Art* part){
	if(dynamic_cast<ColorEffect*>(part->workEffect)){
		return g_workPolyMapColor;
	}
	else{
		return g_workPolyMapTexture;
	}
}

void TextureEffect::clear(){
	texCoord.refValue().length = 0;
}

void ColorEffect::clear(){
	color.refValue().length = 0;
}
void MonoColorEffect::clear(){
	color.refValue().length = 0;
}

void Stem::copyPolyMapToPart(Art* part){
	part->coord.refValue().length = 0;
	part->normal.refValue().length = 0;
	part->workEffect->clear();
	part->triangle.length = 0;
	part->workEffect->stemCopyLocalTriNonMirror(this, part, part->coord.refValue(), 0, 0);
	part->smls->resetRelevantParts();
}

void moveVertex(){//MOVE_VERTEX
	int dstIndex=0;
	int dstTri=0;
	int dstCorner=0;
	Art* dstTriPart=0;
	if(Part* whitePart=dynamic_cast<Part*>(g_selNode.getRaw())){
		Seamless* smls = whitePart->smls;
		if(Art* dstPart=getClosestVertex(whitePart, smls->ActCtrlPnt, dstTriPart, dstIndex, dstTri, dstCorner, g_settings->perspective)){
			if(dstPart == whitePart){
				NurbsPatch* patch = 0;
				Stem* srcStem = 0;
				Stem* dstStem = 0;
				Polymap* polymap = 0;
				ImmediatePart* bluePart = 0;
				if(getPolyMap(patch, srcStem, polymap, whitePart, smls->ActTri, bluePart)){
					if(getPolyMap(patch, dstStem, polymap, whitePart, dstTri, bluePart)){
						if(bluePart){
							joinVertexImmediate(bluePart, bluePart, bluePart, bluePart, dstIndex-dstStem->beginCoordOffset, smls->ActCtrlPnt-srcStem->beginCoordOffset, dstTri-dstStem->beginTriOffset, smls->ActTri-srcStem->beginTriOffset, dstCorner, smls->ActTriCorner, true, true, true);
						}
						else{
							bluePart = selectPolyMapPart(whitePart);
							dstStem->copyPolyMapToPart(bluePart);
							joinVertexImmediate(bluePart, bluePart, bluePart, bluePart, dstIndex-dstStem->beginCoordOffset, smls->ActCtrlPnt-srcStem->beginCoordOffset, dstTri-dstStem->beginTriOffset, smls->ActTri-srcStem->beginTriOffset, dstCorner, smls->ActTriCorner, true, true, true, false);
							dstStem->copyPartToPolyMap(bluePart);
							commonEndVertexBuild();
						}
					}
				}
			}
		}
	}
	else if(ImmediatePart* bluePart=dynamic_cast<ImmediatePart*>(g_selNode.getRaw())){
		Seamless* smls = bluePart->smls;
		Art* dstTriPart=0;
		if(Art* dstPart=getClosestVertex(bluePart, smls->ActCtrlPnt, dstTriPart, dstIndex, dstTri, dstCorner, g_settings->perspective)){
			joinVertexImmediate(dstPart, bluePart, dstTriPart, smls->actTriPart, dstIndex, smls->ActCtrlPnt, dstTri, smls->ActTri, dstCorner, smls->ActTriCorner, true, true, true);
		}
	}
	else if(NurbsPatch* patch = dynamic_cast<NurbsPatch*>(g_selNode.getRaw())){
		patch->smls->moveNurbsCP();
	}
	else{
		needPartMsg();
	}
}

int getRelIndex(Stem*& stem, int& ActTri, int& ActTriCorner){//returns coord ActCtrlPnt
	ActTri = (ActTri - stem->beginTriOffset) * 3;
	if(stem->mirror){
		if(stem->mirrorStem == 0 && ActTri >= stem->coordIndex.length){
			//ActCtrlPnt -= stem->coord.length;
			ActTri -= stem->coordIndex.length;
			ActTriCorner = abs(3 - ActTriCorner) % 3;
		}
		if(stem->mirrorStem && stem->coordIndex.length == 0 && stem->mirrorStem->coordIndex.length){
			stem = stem->mirrorStem;
			ActTriCorner = abs(3 - ActTriCorner) % 3;
		}
	}
	if(ActTri + ActTriCorner >= stem->coordIndex.length){
		return -1;
	}
	return stem->coordIndex[ActTri + ActTriCorner];
}

void joinVertexKeepDestColor(){//JOIN_VERTEX //JOIN_COLOR
	int dstIndex=0;
	int dstTri=0;
	int dstCorner=0;
	Art* dstTriPart=0;
	if(Part* whitePart=dynamic_cast<Part*>(g_selNode.getRaw())){
		Seamless* smls = whitePart->smls;
		if(Art* dstPart=getClosestVertex(whitePart, smls->ActCtrlPnt, dstTriPart, dstIndex, dstTri, dstCorner, g_settings->perspective)){
			if(dstPart == whitePart){
				NurbsPatch* patch = 0;
				Stem* srcStem = 0;
				Stem* dstStem = 0;
				Polymap* polymap = 0;
				ImmediatePart* bluePart = 0;
				if(getPolyMap(patch, srcStem, polymap, whitePart, smls->ActTri, bluePart)){
					if(getPolyMap(patch, dstStem, polymap, whitePart, dstTri, bluePart)){
						if(bluePart){
							joinVertexImmediate(bluePart, bluePart, bluePart, bluePart, dstIndex-dstStem->beginCoordOffset, smls->ActCtrlPnt-srcStem->beginCoordOffset, dstTri-dstStem->beginTriOffset, smls->ActTri-srcStem->beginTriOffset, dstCorner, smls->ActTriCorner, false, true, false);
						}
						else{
							int ActTri = smls->ActTri;
							int ActTriCorner = smls->ActTriCorner;
							if(0 > getRelIndex(srcStem, ActTri, ActTriCorner)){
								return;
							}
							int ActCtrlPnt = srcStem->rcCoord(ActTri + ActTriCorner);
				

							if(0 > getRelIndex(dstStem, dstTri, dstCorner)){
								return;
							}
							int dstIndex = dstStem->rcCoord(dstTri + dstCorner);

							bluePart = selectPolyMapPart(whitePart);
							dstStem->copyPolyMapToPart(bluePart);
							joinVertexImmediate(bluePart, bluePart, bluePart, bluePart, dstIndex, ActCtrlPnt, dstTri / 3, ActTri / 3, dstCorner, ActTriCorner, false, true, false, false);
							dstStem->copyPartToPolyMap(bluePart);
							commonEndVertexBuild();
						}
					}
				}
			}
		}
	}
	else if(ImmediatePart* bluePart=dynamic_cast<ImmediatePart*>(g_selNode.getRaw())){
		Seamless* smls = bluePart->smls;
		if(Art* dstPart=getClosestVertex(bluePart, smls->ActCtrlPnt, dstTriPart, dstIndex, dstTri, dstCorner, g_settings->perspective)){
			joinVertexImmediate(dstPart, bluePart, dstTriPart, smls->actTriPart, dstIndex, smls->ActCtrlPnt, dstTri, smls->ActTri, dstCorner, smls->ActTriCorner, false, true, false);
		}
	}
	else if(Subdivision* subdivision=dynamic_cast<Subdivision*>(g_selNode.getRaw())){
		subdivision->joinVertex();
	}
	else if(BezierPatch* bezierPatch=dynamic_cast<BezierPatch*>(g_selNode.getRaw())){
		MultiSelect::selectRectMode=true;
	}
	else{
		needPartMsg();
	}
}

void addMidPoint(M_CtrlPolyPoint& ctrlPolyPoint, int ctrlPolyIdx, M_ControlPoint& pointList);

void joinVertexKeepSourceColor(){
	int dstIndex=0;
	int dstTri=0;
	int dstCorner=0;
	Art* dstTriPart=0;
	if(Part* whitePart=dynamic_cast<Part*>(g_selNode.getRaw())){
		Seamless* smls = whitePart->smls;
		if(smls->workAnimCP && MultiSelect::pointList.length){
			smls->addMidPoint(MultiSelect::pointList);
			return;
		}
		if(Art* dstPart=getClosestVertex(whitePart, smls->ActCtrlPnt, dstTriPart, dstIndex, dstTri, dstCorner, g_settings->perspective)){
			if(dstPart == whitePart){
				NurbsPatch* patch = 0;
				Stem* srcStem = 0;
				Stem* dstStem = 0;
				Polymap* polymap = 0;
				ImmediatePart* bluePart = 0;
				if(getPolyMap(patch, srcStem, polymap, whitePart, smls->ActTri, bluePart)){
					if(getPolyMap(patch, dstStem, polymap, whitePart, dstTri, bluePart)){
						if(bluePart){
							joinVertexImmediate(bluePart, bluePart, bluePart, bluePart, dstIndex-dstStem->beginCoordOffset, smls->ActCtrlPnt-srcStem->beginCoordOffset, dstTri-dstStem->beginTriOffset, smls->ActTri-srcStem->beginTriOffset, dstCorner, smls->ActTriCorner, true, false, false);
						}
						else{
							bluePart = selectPolyMapPart(whitePart);
							dstStem->copyPolyMapToPart(bluePart);
							joinVertexImmediate(bluePart, bluePart, bluePart, bluePart, dstIndex-dstStem->beginCoordOffset, smls->ActCtrlPnt-srcStem->beginCoordOffset, dstTri-dstStem->beginTriOffset, smls->ActTri-srcStem->beginTriOffset, dstCorner, smls->ActTriCorner, true, false, false, false);
							dstStem->copyPartToPolyMap(bluePart);
							commonEndVertexBuild();
						}
					}
				}
			}
		}
	}
	else if(ImmediatePart* bluePart=dynamic_cast<ImmediatePart*>(g_selNode.getRaw())){
		Seamless* smls = bluePart->smls;
		Art* dstTriPart=0;
		if(Art* dstPart=getClosestVertex(bluePart, smls->ActCtrlPnt, dstTriPart, dstIndex, dstTri, dstCorner, g_settings->perspective)){
			joinVertexImmediate(dstPart, bluePart, dstTriPart, smls->actTriPart, dstIndex, smls->ActCtrlPnt, dstTri, smls->ActTri, dstCorner, smls->ActTriCorner, true, false, false);
		}
	}
	else if(NurbsPatch* nurbsPatch = dynamic_cast<NurbsPatch*>(g_selNode.getRaw())){
		Seamless* smls = nurbsPatch->smls;
		if(MultiSelect::pointList.length){
			smls->addMidPoint(MultiSelect::pointList);
			return;
		}
	}
	else if(BezierPatch* bezierPatch=dynamic_cast<BezierPatch*>(g_selNode.getRaw())){
		MultiSelect::selectRectMode=true;
	}
	else{
		needPartMsg();
	}
}
void joinVertexKeepBothColors(){
	int dstIndex=0;
	int dstTri=0;
	int dstCorner=0;
	Art* dstTriPart=0;
	if(Part* whitePart=dynamic_cast<Part*>(g_selNode.getRaw())){
		Seamless* smls = whitePart->smls;
		if(MultiSelect::pointList.length){
			smls->createControlPolygon(MultiSelect::pointList);
			return;
		}
		if(Art* dstPart=getClosestVertex(whitePart, smls->ActCtrlPnt, dstTriPart, dstIndex, dstTri, dstCorner, g_settings->perspective)){
			if(dstPart == whitePart){
				NurbsPatch* patch = 0;
				Stem* srcStem = 0;
				Stem* dstStem = 0;
				Polymap* polymap = 0;
				ImmediatePart* bluePart = 0;
				if(getPolyMap(patch, srcStem, polymap, whitePart, smls->ActTri, bluePart)){
					if(getPolyMap(patch, dstStem, polymap, whitePart, dstTri, bluePart)){
						if(bluePart){
							joinVertexImmediate(bluePart, bluePart, bluePart, bluePart, dstIndex-dstStem->beginCoordOffset, smls->ActCtrlPnt-srcStem->beginCoordOffset, dstTri-dstStem->beginTriOffset, smls->ActTri-srcStem->beginTriOffset, dstCorner, smls->ActTriCorner, true, true, false);
						}
						else{
							bluePart = selectPolyMapPart(whitePart);
							dstStem->copyPolyMapToPart(bluePart);
							joinVertexImmediate(bluePart, bluePart, bluePart, bluePart, dstIndex-dstStem->beginCoordOffset, smls->ActCtrlPnt-srcStem->beginCoordOffset, dstTri-dstStem->beginTriOffset, smls->ActTri-srcStem->beginTriOffset, dstCorner, smls->ActTriCorner, true, true, false, false);
							dstStem->copyPartToPolyMap(bluePart);
							commonEndVertexBuild();
						}
					}
				}
			}
		}
	}
	else if(ImmediatePart* bluePart=dynamic_cast<ImmediatePart*>(g_selNode.getRaw())){
		Seamless* smls = bluePart->smls;
		Art* dstTriPart=0;
		if(Art* dstPart=getClosestVertex(bluePart, smls->ActCtrlPnt, dstTriPart, dstIndex, dstTri, dstCorner, g_settings->perspective)){
			Seamless* smls = bluePart->smls;
			joinVertexImmediate(dstPart, bluePart, dstTriPart, smls->actTriPart, dstIndex, smls->ActCtrlPnt, dstTri, smls->ActTri, dstCorner, smls->ActTriCorner, true, true, false);
		}
		else{
			needPartCopiedMsg();
		}
	}
	else if(NurbsPatch* nurbsPatch = dynamic_cast<NurbsPatch*>(g_selNode.getRaw())){
		Seamless* smls = nurbsPatch->smls;
		if(MultiSelect::pointList.length){
			smls->createControlPolygon(MultiSelect::pointList);
			return;
		}
	}
	else if(BezierPatch* bezierPatch=dynamic_cast<BezierPatch*>(g_selNode.getRaw())){
		MultiSelect::selectRectMode=true;
	}
	else{
		needPartMsg();
	}
}
void joinMatchingColors(){//JOIN_MATCHING_COLORS
	Seamless* smls=dynamic_cast<Seamless*>(g_selNode.getRaw());
	if(Art* part=dynamic_cast<Art*>(g_selNode.getRaw())){
		smls=part->smls;
	}
	if(smls){
		dynamic_cast<Effect*>(smls->effect.refValue().p)->optimize(smls->partList);
		smls->process();
		BackUp::monostable.beginCycle();
	}
	else{
		needPartOrSmlsMsg();
	}
}
void colorPart(){//COLOR_PART
#ifdef PROG_ROB
			dLog("new ClickButton(colorPart),");
#endif
	if(ImmediatePart* part = dynamic_cast<ImmediatePart*>(g_selNode.getRaw())){
		if(ColorEffect* colorEffect = dynamic_cast<ColorEffect*>(part->workEffect)){
			part->colorize(g_palette->color);
			commonEndVertexChange();
		}
	}
	else{
		needBluePartMsg();
	}
}


void replaceColorPart(){//REPLACE_COLOR_PART
	if(Part* whitePart=dynamic_cast<Part*>(g_selNode.getRaw())){
		if(dynamic_cast<ColorEffect*>(whitePart->workEffect)==0)return;

		Color3f oldColor;
		if(getOldColor(oldColor)){
			Seamless* smls = whitePart->smls;
			NurbsPatch* patch = 0;
			Stem* stem = 0;
			Polymap* polymap = 0;
			ImmediatePart* bluePart = 0;
			if(getPolyMap(patch, stem, polymap, whitePart, smls->ActTri, bluePart)){
				if(bluePart){
				}
				else{
					int ActTri = smls->ActTri;
					int ActTriCorner = smls->ActTriCorner;
					if(0 > getRelIndex(stem, ActTri, ActTriCorner)){
						return;
					}

					M_Color3f& color = stem->color.refValue();
					M_Int& coordIndex = stem->coordIndex.refValue();
					for(int i = coordIndex.length; --i >= 0;){
						int ci = coordIndex[i];
						if(color[ci] == oldColor){
							color[ci] = g_palette->color;
						}
					}
					commonEndVertexBuild();
				}
			}
		}
	}
	else if(ImmediatePart* part=dynamic_cast<ImmediatePart*>(g_selNode.getRaw())){
		Color3f oldColor;
		if(getOldColor(oldColor)){
			replaceColorInPart(part, oldColor, g_palette->color);
		}
	}
	else{
		needBluePartMsg();
	}
	BackUp::monostable.beginCycle();
}
void replaceColorSkeleton(){//REPLACE_COLOR_SKELETON
	if(Art* part=dynamic_cast<Art*>(g_selNode.getRaw())){
		if(Art* copied=dynamic_cast<Art*>(g_copyNode.getRaw())){
			Color3f oldColor;
			if(getOldColor(oldColor)){
				M_Art& partList=part->smls->partList;
				for(int k=partList.length; --k>=0;){
					replaceColorInPart(partList[k], oldColor, g_palette->color);
				}
			}
		}
		else{
			needPartCopiedMsg();
		}
	}
	else{
		needBluePartMsg();
	}
}
void transferCloseVerticesJoin(){//	 TRANSFERCLOSEVERTICES_JOIN
	if(Art* source=dynamic_cast<Art*>(g_selNode.getRaw())){
		if(Art* copied=dynamic_cast<Art*>(g_copyNode.getRaw())){
			fastFullBuildForAutoIn(true);
			pasteCloseVerticesJoin(source, copied, true);
			Seamless* smls=source->smls;
			smls->verifyIntegrity();

			smls->convertToRenderFormat();
			makePointerEqualIndexLocation();
		//			g_stw->setSelNodeLine();
			g_stw->paint();
			g_stw->cp->paint();
			BackUp::monostable.beginCycle();
		}
		else{
			needPartCopiedMsg();
		}
	}
	else if(Subdivision* subdivision=dynamic_cast<Subdivision*>(g_selNode.getRaw())){
		subdivision->joinCloseVertices();
	}
	else{
		needPartMsg();
	}
}
inline bool checkVertexIsShared(M_Art& partList, Art* targetVertexPart, int& targetVertexIndex, Art* targetTrianglePart, int targetTriangleIndex){
	for(int k=partList.length; --k>=0; ){
		Art* part=partList[k];
		M_Triangle& triangle=part->triangle;
		for(int i=triangle.length*3, t=triangle.length; --t>=0; ){
			if(targetTriangleIndex==t && targetTrianglePart==part){ //first check its not the same triangle before asking any more questions
				i-=3;
			}
			else {
				--i;
				if(triangle.refPart(i)==targetVertexPart&&triangle.refIndex(i)==targetVertexIndex){
					return true;
				}
				--i;
				if(triangle.refPart(i)==targetVertexPart&&triangle.refIndex(i)==targetVertexIndex){
					return true;
				}
				--i;
				if(triangle.refPart(i)==targetVertexPart&&triangle.refIndex(i)==targetVertexIndex){
					return true;
				}
			}
		}
	}
	return false;
}
void joinCloseVerticesMacro(){//	 TRANSFERCLOSEVERTICES_JOIN
#ifdef PROG_ROB
	dLog("new ClickButton(joinCloseVerticesMacro),");
#endif
	if(Art* part=dynamic_cast<Art*>(g_selNode.getRaw())){
		M_Art& partList=part->smls->partList;
		Seamless* smls=part->smls;
		for(int k=partList.length; --k>=0;){
			Art* partK=partList[k];
			if(partK->show==false){
				continue;
			}
			pasteCloseVerticesJoin(partK, partK, true);
		}
		for(int k=partList.length; --k>=0;){
			Art* partK=partList[k];
			for(int j=partList.length; --j>=0;){
				Art* partJ=partList[j];
				if(partJ->show==false){
					continue;
				}
				if(j==k){
					continue;
				}
				if(j>k){
					pasteCloseVerticesJoin(partK, partJ, true);	
				}
				else{
					pasteCloseVerticesJoin(partJ, partK, true);	
				}
			}
		}
		smls->verifyIntegrity();

		smls->convertToRenderFormat();
		makePointerEqualIndexLocation();
	//			g_stw->setSelNodeLine();
		g_stw->paint();
		g_stw->cp->paint();
		BackUp::monostable.beginCycle();
	}
	else{
		needPartMsg();
	}
}

void multiBreak(Seamless* smls);

void breakVertexJoin(){
#ifdef PROG_ROB
	dLog("new MoveToF10(),");
	dLog("new Move_MenuItem(menu_break_join),");
	dLog("new Move_MenuItem(menu_break_coord),");
	dLog("new Click,");
#endif

	if(Art* part=dynamic_cast<Art*>(g_selNode.getRaw())){
		fastFullBuildForAutoIn(true);

		Seamless* smls = part->smls;

		if(MultiSelect::triList.length){
			multiBreak(smls);
			return;
		}

		M_Vec3f& coord=part->coord.refValue();
		M_Vec3f& normal=part->normal.refValue();

		if(coord.length==0){
			return;
		}
		if(checkVertexIsShared(smls->partList, part, smls->ActCtrlPnt, smls->actTriPart, smls->ActTri)==false){
			msG(msg_vertex_belongs_to_only_one);
			return;
		}
		BackUp::monostable.beginCycle();

		int i=smls->ActTri*3+smls->ActTriCorner;

		M_Triangle& triangle=smls->actTriPart->triangle;

		int ia = triangle.refIndex(i);
		int cli = triangle.refTexIndex(i);

		triangle.refIndex(i)=coord.length;

		smls->ActCtrlPnt=coord.length;

		coord+=coord[ia];


		if(normal.length == coord.length){
			normal+=normal[smls->ActCtrlPnt];
		}

		if(part->workEffect!=0){
			if(part->workEffect->getType()==COLOREFFECT){
				M_Color3f& color=dynamic_cast<ColorEffect*>(part->workEffect)->color.refValue();
				triangle.refTexIndex(i)=color.length;
				color+=color[cli];
			}
			else if(part->workEffect->getType()==TEXTUREEFFECT){
				M_Vec2f& texCoord=dynamic_cast<TextureEffect*>(part->workEffect)->texCoord.refValue();
				triangle.refTexIndex(i)=texCoord.length;
				texCoord+=texCoord[cli];
			}
		}
		MultiSelect::coordList.length = 0;
		MultiSelect::beginDragPos.length = 0;

		smls->doBuild();
		smls->convertToRenderFormat();
		makePointerEqualIndexLocation();
	}
	else{
		needPartMsg();
	}

}

void breakColorJoin(){//BREAK_COLOR_JOIN
	if(Art* part=dynamic_cast<Art*>(g_selNode.getRaw())){
#ifdef PROG_ROB
		dLog("new MoveToF10(),");
		dLog("new Move_MenuItem(menu_break_join),");
		dLog("new Move_MenuItem(menu_break_color),");
		dLog("new Click,");
#endif
		part->workEffect->breakColorJoin(*(part->smls));
	}
	else{
		needPartMsg();
	}
}
void cutWholePart(){//CUT_WHOLE_PART
	if(Art* part=dynamic_cast<Art*>(g_selNode.getRaw())){
		fastFullBuildForAutoIn(true);
		part->deleteGeometry();
		part->smls->convertToRenderFormat();
		makePointerEqualIndexLocation();
		BackUp::monostable.beginCycle();
	}
	else{
		needPartMsg();
	}
}
Vec2f g_storeClickPos=Vec2f(.195313, .728516);
void Art::tripleSplitTriangle(int triIdx){
	int newCoordIndex=coord.length;
	Vec3f newCoord;

	int a = triIdx;
	int b = triIdx+1;
	int c = triIdx+2;

	int ia=triangle.refIndex(a);
	int ib=triangle.refIndex(b);
	int ic=triangle.refIndex(c);

	Art* pa=triangle.refPart(a);
	Art* pb=triangle.refPart(b);
	Art* pc=triangle.refPart(c);

	Vec3f vecA = pa->matrix * g_viewMatrix * pa->coord.refValue()[ia];
	Vec3f vecB = pb->matrix * g_viewMatrix * pb->coord.refValue()[ib];
	Vec3f vecC = pc->matrix * g_viewMatrix * pc->coord.refValue()[ic];

//	newCoord = (vecA + vecB + vecC) / 3.0f;
//	newCoord = g_scene->getIViewMatrix() * imatrix * newCoord;

	Vec3f v[3]={vecA, vecB, vecC};

	Vec3f clickPos;
	if(getTriangleHitPos(clickPos, g_storeClickPos, v)){

		newCoord = g_scene->getIViewMatrix() * imatrix * clickPos;

		int newColIndex=0;

		if(workEffect->getType()==COLOREFFECT){
			ColorEffect* effect=dynamic_cast<ColorEffect*>(workEffect);
			M_Color3f& color=effect->color.refValue();
			newColIndex=color.length;

			Color3f colA=color[triangle.refTexIndex(a)];
			Color3f colB=color[triangle.refTexIndex(b)];
			Color3f colC=color[triangle.refTexIndex(b)];

			Color3f newColor;
			newColor.r = colA.r - (colA.r - colB.r)/2.;
			newColor.g = colA.g - (colA.g - colB.g)/2.;
			newColor.b = colA.b - (colA.b - colB.b)/2.;
			color+=newColor;
		}
		else if(workEffect->getType()==TEXTUREEFFECT){
			TextureEffect* effect=dynamic_cast<TextureEffect*>(workEffect);
			M_Vec2f& texCoord=effect->texCoord.refValue();
			newColIndex=texCoord.length;
			Vec2f texA=texCoord[triangle.refTexIndex(a)];;
			Vec2f texB=texCoord[triangle.refTexIndex(b)];;

			Vec2f newTexCoord;
			newTexCoord.x = texA.x - (texA.x - texB.x)/2.;
			newTexCoord.y = texA.y - (texA.y - texB.y)/2.;
			texCoord+=newTexCoord;
		}
		else throw this;

		coord.refValue()+=newCoord;

		if(normal.length){
			normal.refValue()+=Vec3f();
		}

		pc=triangle.refPart(c);
		ic=triangle.refIndex(c);
		int tc=triangle.refTexIndex(c);
		
		triangle.refPart(c)=this;
		triangle.refIndex(c)=newCoordIndex;
		triangle.refTexIndex(c)=newColIndex;

		triangle+=S_Triangle(
			pb, ib, triangle.refTexIndex(b),
			pc, ic, tc,
			this, newCoordIndex,newColIndex
		);
		triangle+=S_Triangle(
			pc, ic, tc,
			pa, ia, triangle.refTexIndex(a),
			this, newCoordIndex,newColIndex
		);
	}
}
void  projectSingleVertexOntoPoly(){
#ifdef PROG_ROB
	dLog("new ClickButton(projectSingleVertexOntoPoly),");
#endif

	if(Art* selPart=dynamic_cast<Art*>(g_selNode.getRaw())){

		const bool perspective=g_settings->perspective;
		Seamless* smls=selPart->smls;
		smls->catMatrix();
		smls->catIMatrix();
		const int selTri=smls->ActTri*3;
		const int ActCtrlPnt=smls->ActCtrlPnt;

		g_viewMatrix=g_scene->getViewMatrix();
		g_pointer.translation=selPart->matrix * g_scene->getViewMatrix() * selPart->coord[smls->ActCtrlPnt];

		M_Art &partList=smls->partList;
		for(int k=0;k<partList.length;k++){

			Art* part=partList[k];
			M_Triangle& triangle = part->triangle;

			if(part->show){
				for(int i=triangle.getTriangleLength(); (i-=3)>=0; ){
					if(part->checkPointInTriangleForProjectVertex(perspective, i)){
						
						int a = i;
						int b = i+1;
						int c = i+2;

						int ia=triangle.refIndex(a);
						int ib=triangle.refIndex(b);
						int ic=triangle.refIndex(c);

						Art* pa=triangle.refPart(a);
						Art* pb=triangle.refPart(b);
						Art* pc=triangle.refPart(c);
						if(selPart == part){
							if(selTri == i ||
								pa==selPart && ia == ActCtrlPnt ||
								pb==selPart && ib == ActCtrlPnt ||
								pc==selPart && ic == ActCtrlPnt
								){
								continue;
							}
						}

						Vec3f vecA = pa->matrix * g_scene->getViewMatrix() * pa->coord.refValue()[ia];
						Vec3f vecB = pb->matrix * g_scene->getViewMatrix() * pb->coord.refValue()[ib];
						Vec3f vecC = pc->matrix * g_scene->getViewMatrix() * pc->coord.refValue()[ic];


						Vec3f v[3]={vecA, vecB, vecC};

						Vec3f clickPos;
						if(getTriangleHitPos(clickPos, g_storeClickPos, v)){
							ImmediatePart* part=dynamic_cast<ImmediatePart*>(selPart);

							Vec3f newPt = g_scene->getIViewMatrix() * part->imatrix * clickPos;

							M_Vec3f& coord=part->coord.refValue();

							smls->senseMirrorVertex(part, smls->ActCtrlPnt, newPt);
							coord[smls->ActCtrlPnt]=newPt;

							makePointerEqualIndexLocation();
							smls->convertToRenderFormat();

							BackUp::monostable.beginCycle();
							return;
						}

					}
				}
			}
		}

	}
	else{
		needPartMsg();
	}
};
/*
void  tripleSplitTriangleImmediate(){
	if(Art* part=dynamic_cast<Art*>(g_selNode.getRaw())){
		Seamless* smls=part->smls;
		part=smls->actTriPart;

     	part->tripleSplitTriangle(smls->ActTri*3);

		makePointerEqualIndexLocation();
		smls->convertToRenderFormat();

		BackUp::monostable.beginCycle();
	}
	else{
		needPartMsg();
	}
}
*/
Art* findTriangleJoinedBy2Vertices(const Art* triPart, int triIndex, int& otherTriIndex, const Art* selPartA, const int selIdxA, const Art* selPartB, const int selIdxB, int& notConected, M_<S_VertexIndex>& triList){
	for(int k = triList.length; --k>=0; ){
		Art* part = triList[k].p;
		int tri = triList[k].i;
		if(triPart == part && triIndex == tri){
			continue;
		}
		S_Triangle& triangle = part->triangle[tri];
		if(	triangle.refPart(0) == selPartA && triangle.refIndex(0) == selIdxA){
			if(	triangle.refPart(1) == selPartB && triangle.refIndex(1) == selIdxB){
				otherTriIndex = tri;
				notConected = 2;
				return part;
			}
			if(	triangle.refPart(2) == selPartB && triangle.refIndex(2) == selIdxB){
				otherTriIndex = tri;
				notConected = 1;
				return part;
			}
		}
		if(	triangle.refPart(1) == selPartA && triangle.refIndex(1) == selIdxA){
			if(	triangle.refPart(0) == selPartB && triangle.refIndex(0) == selIdxB){
				otherTriIndex = tri;
				notConected = 2;
				return part;
			}
			if( triangle.refPart(2) == selPartB && triangle.refIndex(2) == selIdxB){
				otherTriIndex = tri;
				notConected = 0;
				return part;
			}
		}
		if(	triangle.refPart(2) == selPartA && triangle.refIndex(2) == selIdxA){
			if(	triangle.refPart(0) == selPartB && triangle.refIndex(0) == selIdxB){
				otherTriIndex = tri;
				notConected = 1;
				return part;
			}
			if(	triangle.refPart(1) == selPartB && triangle.refIndex(1) == selIdxB){
				otherTriIndex = tri;
				notConected = 0;
				return part;
			}
		}
	}
	return 0;
}

Art* findTriangleJoinedBy2Vertices(const Art* triPart, int triIndex, int& otherTriIndex, const Art* selPartA, const int selIdxA, const Art* selPartB, const int selIdxB, int& notConected){
	M_Art& partList = selPartA->smls->partList;
	for(int k = partList.length; --k>=0; ){
		Art* part = partList[k];
		M_Triangle& triangle = part->triangle;
		for(int i = triangle.getTriangleLength(); (i-=3)>=0;){
			if(triPart == part && triIndex == i){
				continue;
			}
			if(	triangle.refPart(i+0) == selPartA && triangle.refIndex(i+0) == selIdxA){
				if(	triangle.refPart(i+1) == selPartB && triangle.refIndex(i+1) == selIdxB){
					otherTriIndex = i / 3;
					notConected = 2;
					return part;
				}
				if(	triangle.refPart(i+2) == selPartB && triangle.refIndex(i+2) == selIdxB){
					otherTriIndex = i / 3;
					notConected = 1;
					return part;
				}
			}
			if(	triangle.refPart(i+1) == selPartA && triangle.refIndex(i+1) == selIdxA){
				if(	triangle.refPart(i+0) == selPartB && triangle.refIndex(i+0) == selIdxB){
					otherTriIndex = i / 3;
					notConected = 2;
					return part;
				}
				if( triangle.refPart(i+2) == selPartB && triangle.refIndex(i+2) == selIdxB){
					otherTriIndex = i / 3;
					notConected = 0;
					return part;
				}
			}
			if(	triangle.refPart(i+2) == selPartA && triangle.refIndex(i+2) == selIdxA){
				if(	triangle.refPart(i+0) == selPartB && triangle.refIndex(i+0) == selIdxB){
					otherTriIndex = i / 3;
					notConected = 1;
					return part;
				}
				if(	triangle.refPart(i+1) == selPartB && triangle.refIndex(i+1) == selIdxB){
					otherTriIndex = i / 3;
					notConected = 0;
					return part;
				}
			}
		}
	}
	return 0;
}
Art* findTriangleJoinedBy2Vertices(const Art* selPartA, const int selIdxA, const Art* selPartB, const int selIdxB, int& triIndex, int& triCorner){
	M_Art& partList = selPartA->smls->partList;
	for(int k = partList.length; --k>=0; ){
		Art* part = partList[k];
		for(int tri = part->triangle.length; --tri>=0;){
			S_Triangle& triangle = part->triangle[tri];
			if(	triangle.refPart(0) == selPartA && triangle.refIndex(0) == selIdxA){
				if(	triangle.refPart(1) == selPartB && triangle.refIndex(1) == selIdxB){
					triIndex = tri;
					triCorner = 0;
					return part;
				}
			}
			if(	triangle.refPart(1) == selPartA && triangle.refIndex(1) == selIdxA){
				if(	triangle.refPart(2) == selPartB && triangle.refIndex(2) == selIdxB){
					triIndex = tri;
					triCorner = 1;
					return part;
				}
			}
			if(	triangle.refPart(2) == selPartA && triangle.refIndex(2) == selIdxA){
				if(	triangle.refPart(0) == selPartB && triangle.refIndex(0) == selIdxB){
					triIndex = tri;
					triCorner = 2;
					return part;
				}
			}
		}
	}
	return 0;
}

void genTriangles(Art* part, M_<S_VertexIndex>& triList){
	if(part->workEffect==0)return;
	part->deleteGeometry();
	Seamless* smls = part->smls;
	Effect* effect=part->workEffect;
	int effectLength=effect->getEffectLength();
	int triLen = triList.length;
	M_Color3f& color = dynamic_cast<ColorEffect*>(effect)->color.refValue();
	color.length=triLen*3;
	for(int i = color.length; --i>=0;){
		color[i] = Color3f(0,1,1);
	}

	M_Vec3f& coord = part->coord.refValue();
	coord.length = triLen * 3;
	M_Triangle& m_triangle = part->triangle;
	m_triangle.length = triLen;

	for(int i = 0; i < triLen; i++){
		int t = triList[i].i;
		S_Triangle tri = triList[i].p->triangle[t];
		coord[i*3+0] = tri.pa->coord[tri.ia];
		coord[i*3+1] = tri.pb->coord[tri.ib];
		coord[i*3+2] = tri.pc->coord[tri.ic];

		m_triangle[i]=S_Triangle(part, i*3+0, i*3+0, part, i*3+1, i*3+1, part, i*3+2, i*3+2);
	}

	part->relevantPart=M_Art(part);
	smls->process();
}
bool triIntersectsRect(Vec2f a, Vec2f b, Vec2f c);
Vec2f convertPointToScreen2d(Vec3f point);
bool isRectCornerInTriangle(const Vec2f* tri);

void colorVertex(Art* part, int crdIdx){
	Seamless* smls = part->smls;
	M_Art& partList = smls->partList;
	for(int k = partList.length; --k>=0;){
		Art* p = partList[k];
		M_Triangle& triangle = p->triangle;

		for(int i = triangle.getTriangleLength(); --i >= 0;){
			int crdIdxB = triangle.refIndex(i);
			Art* partB = triangle.refPart(i);
			if(partB == part && crdIdxB == crdIdx){
				if(ColorEffect* colorEffect = dynamic_cast<ColorEffect*>(partB->workEffect)){
					M_Color3f& color = colorEffect->color;
					color[triangle.refTexIndex(i)] = g_palette->color;
				}
			}
		}
	}
}

void applyTriangleColor(Seamless* smls);
void applyCleanEdgeColor(Seamless* smls);
void selectMultipleTriangles(Seamless* smls);
void applyCPOwnership(Art* part);

void NurbsPatch::selectControlPoints(M_<NurbsPatch*>& contigList){
	int uDim = this->dim.x;
	int vDim = this->dim.y;

	if(cage == false){
		return;
	}

	if(showAll){
		M_Vec3f& dpoint = this->dpoint.refValue();
		int len = dpoint.length;
		for(int i = 0; i < len; i++){
			if(MultiSelect::isInBox(dpoint[i])){
				contigList[fromPoint + i] = this;
			}
		}
	}
	else{
		if(uJoinEnds){
			if(isCtrlPolyRef){
				for(int u = 1; u < uDim - 1; u += 2){
					for(int v = 1; v < vDim - 1; v++){
						int i = v * uDim + u;
						if(MultiSelect::isInBox(dpoint[i])){
							contigList[fromPoint + i] = this;
						}
					}
				}
			}
			else{
				for(int u = 1; u < uDim; u += 2){
					for(int v=0; v < vDim; v++){
						int i = v * uDim + u;
						if(MultiSelect::isInBox(dpoint[i])){
							contigList[fromPoint + i] = this;
						}
					}
				}
			}
		}
		else{
			if(NurbsPatch::MID_1 == checkNurbsType(vDim, vKnot)){
				if(isCtrlPolyRef){
					for(int v = 1; v < vDim - 1; v += 2){
						for(int u = 1; u < uDim - 1; u += 2){
							int i = v * uDim + u;
							if(MultiSelect::isInBox(dpoint[i])){
								contigList[fromPoint + i] = this;
							}
						}
					}
				}
				else{
					for(int v = 0; v < vDim; v += 2){
						for(int u = 0; u < uDim; u += 2){
							int i = v * uDim + u;
							if(MultiSelect::isInBox(dpoint[i])){
								contigList[fromPoint + i] = this;
							}
							if(u == 0 || u == uDim - 2){
								u--;
							}
						}
						if(v == 0 || v == vDim - 2){
							v--;
						}
					}
				}
			}
			else{
				if(NurbsPatch::MID_1 == checkNurbsType(uDim, uKnot)){
					if(isCtrlPolyRef){
						for(int v = 1; v < vDim - 1; v++){
							for(int u = 1; u < uDim - 1; u += 2){
								int i = v * uDim+u;
								if(MultiSelect::isInBox(dpoint[i])){
									contigList[fromPoint + i] = this;
								}
							}
						}
					}
					else{
						for(int v = 0; v < vDim; v++){
							for(int u = 1; u < uDim; u +=2){
								int i = v * uDim + u;
								if(MultiSelect::isInBox(dpoint[i])){
									contigList[fromPoint + i] = this;
								}
							}
						}
					}
				}
				else{
					if(isCtrlPolyRef){
						for(int v = 1; v < vDim - 1; v++){
							for(int u = 1; u < uDim - 1; u++){
								int i = v * uDim + u;
								if(MultiSelect::isInBox(dpoint[i])){
									contigList[fromPoint + i] = this;
								}
							}
						}
					}
					else {
						for(int v = 0; v < vDim; v++){
							for(int u = 0; u < uDim; u++){
								int i = v * uDim + u;
								if(MultiSelect::isInBox(dpoint[i])){
									contigList[fromPoint + i] = this;
								}
							}
						}
					}
				}
			}
		}
	}
}

void Seamless::createSelContigList(M_<NurbsPatch*>& contigList){
	contigList.length = pointTotal;

	for(int k = MultiSelect::pointList.length; --k >= 0;){
		NurbsPatch* patch = MultiSelect::pointList[k].p;
		int i = MultiSelect::pointList[k].i;
		contigList[patch->fromPoint + i] = patch;
	}
}

void Seamless::unselectCtrlPolys(){
	g_viewMatrix = g_scene->getViewMatrix();
	int d = 0;
	for(int s = 0; s < MultiSelect::ctrlPolySelList.length; s++){
		const int i = MultiSelect::ctrlPolySelList[s];
		Vec3f va3f = ctrlPolyPoint[i].p->dpoint[ctrlPolyPoint[i].i];
		int len = ctrlPolyPoint[i].len;
		bool selPoly = false;
		for(int j = 1; j < len - 1; j++){
			Vec3f vb3f = ctrlPolyPoint[i + j].p->dpoint[ctrlPolyPoint[i + j].i];
			Vec3f vc3f = ctrlPolyPoint[i + j + 1].p->dpoint[ctrlPolyPoint[i + j + 1].i];
		
			Vec2f va = convertPointToScreen2d(va3f);
			Vec2f vb = convertPointToScreen2d(vb3f);
			Vec2f vc = convertPointToScreen2d(vc3f);

			if(MultiSelect::backfaceIgnore){
				if(!backFace(va, vb, vc)){
					break;
				}
			}

			Vec2f tri2d[3] = {va, vb, vc};

			if(MultiSelect::isInBox(va) || MultiSelect::isInBox(vb) || MultiSelect::isInBox(vc) || triIntersectsRect(va, vb, vc) || isRectCornerInTriangle(tri2d)){
				selPoly = true;
				break;
			}
		}
		if(selPoly == false){
			MultiSelect::ctrlPolySelList[d++] = i;
		}
	}
	MultiSelect::ctrlPolySelList.length = d;
}

void Seamless::selectControlPoints(){
	MultiSelect::processRectangle();

	if(0x8000&GetKeyState(VK_RSHIFT) || InputEvent::r_shiftDown){
		if((0x8000&GetKeyState(VK_RCONTROL)) || InputEvent::r_ctrlDown){
			unselectCtrlPolys();
		}
		else{
			selectCtrlPolys();
		}
		return;
	}
	if((0x8000&GetKeyState(VK_LCONTROL)) || InputEvent::l_ctrlDown){
		int d = 0;
		for(int s = 0; s < MultiSelect::pointList.length; s++){
			NurbsPatch* patch = MultiSelect::pointList[s].p;
			int idx = MultiSelect::pointList[s].i;
			if(MultiSelect::isInBox(patch->dpoint[idx])){
				continue;
			}
			MultiSelect::pointList[d++] = ControlPoint(patch, idx);
		}
		MultiSelect::pointList.length = d;
	}
	else{
		M_<NurbsPatch*> contigList;  //make sure we dont add points twice by making contig list
		createSelContigList(contigList);

		for(int k = 0; k < cpNodeList.length; k++){
			NurbsPatch* patch = cpNodeList[k];
			patch->selectControlPoints(contigList);
		}

		for(int i = 0; i < ctrlPolyPoint.length;){
			int ctrlPoly = i;
			int len = ctrlPolyPoint[ctrlPoly].len;
			for(int k = len; --k >= 0;){
				NurbsPatch* patch = ctrlPolyPoint[i].p;
				int idx = ctrlPolyPoint[i].i;
				if(patch->cage && MultiSelect::isInBox(patch->dpoint[idx])){
					contigList[patch->fromPoint + idx] = patch;
				}
				i++;
			}
			i = getNextCtrlPoly(ctrlPoly);
		}

		MultiSelect::pointList.length = 0;
		for(int k = 0; k < pointTotal; k++){
			if(NurbsPatch* patch = contigList[k]){
				MultiSelect::pointList += ControlPoint(patch, k - patch->fromPoint);
			}
		}
		g_transformCP->calcPivotPoint();
	}
}

void MultiSelect::selectTriangles(Art* part){
	Seamless* smls = part->smls;
	M_Art& partList = smls->partList;

	//if(MultiSelect::accumulate == false){
	//	if(g_paletteMode == false){
	//		Art::triList.length = 0;
	//	}
	//	Art::coordList.length = 0;
	//}

	g_viewMatrix = g_scene->getViewMatrix();

	selectRectMode=false;
	IndexedLineSet* polyLine=dynamic_cast<IndexedLineSet*>(g_vcSelectRectShape->geometry.p);
	dynamic_cast<Coordinate*>(polyLine->coord.p)->point.refValue().length=0;
	polyLine->coordIndex.refValue().length=0;

	processRectangle();

	if((0x8000&GetKeyState(VK_RSHIFT)) || InputEvent::r_shiftDown){
		if(g_paletteMode){
			applyTriangleColor(smls);
			return;
		}
		selectMultipleTriangles(smls);
	}
	else if((0x8000&GetKeyState(VK_LSHIFT)) || InputEvent::l_shiftDown){
		if(smls->workAnimCP){
			if(g_colorOwnership){
				applyCPOwnership(part);
				part->houseKeep();
				return;
			}
			else{
				smls->selectControlPoints();
				part->houseKeep();
				return;
			}
		}
		if(g_paletteMode){
			if(triList.length){
				applyCleanEdgeColor(smls);
				g_visualClue = g_vcArtTriBase;
			}
			else{
				applyColor(smls);
				g_visualClue = g_vcArtTriBase;
			}
		}
		else{
			if(backfaceIgnore){
				//selectFrontFaceVertices(smls);
				selectFrontFaceRenderedVertices(smls);
			}
			else{
				//selectFrontAndBackFaceVertices(smls);
				selectFrontAndBackFaceTransformedVertices(smls);
			}
			if(g_colorOwnership){
				applyOwnership(part);
			}
		}
	}
	else{
		return;
	}
	if(smls->animCP == false && Window::animMode()){
		g_visualClue = g_vcArtAnimBase;
	}
	else{
		g_visualClue = g_vcArtTriBase;
	}
}

void unMultiSelTri(){
	if(ImmediatePart* part=dynamic_cast<ImmediatePart*>(g_selNode.getRaw())){
		if(MultiSelect::triList.length){
			Seamless* smls = part->smls;
			for(int k = MultiSelect::triList.length; --k>=0;){
				if(MultiSelect::triList[k].i == smls->ActTri && smls->actTriPart == MultiSelect::triList[k].p){
					MultiSelect::triList.deleteElement(k);
					genTriangles(g_vcArtTriBase->transpSmls->skeleton.refValue()[0], MultiSelect::triList);
					return;
				}
			}
		}
		else{
			msG("You must have multiple triangles selected for this command");
		}
	}
	else{
		needBluePartMsg();
	}
}


void splitTriangleImmediate(ImmediatePart* part, Art* actTriPart, const int& ActTri, bool complete=false){
	fastFullBuildForAutoIn(true);

	Seamless* smls=part->smls;

	const int ActTriCorner = smls->ActTriCorner;
	
	int idxRed = ActTri * 3 + ActTriCorner;
	int idxGreen = ActTri * 3 + (ActTriCorner + 1) % 3;
	int notConectedA = (ActTriCorner + 2) % 3;
	const int cRed = actTriPart->triangle.refIndex(idxRed);
	const int cGreen = actTriPart->triangle.refIndex(idxGreen);
	Art* red = actTriPart->triangle.refPart(idxRed);
	Art* green = actTriPart->triangle.refPart(idxGreen);
	int otherTriIndex=0;
	int notConectedB = 0;
	
	Art* otherTriPart = findTriangleJoinedBy2Vertices(actTriPart, ActTri*3, otherTriIndex, red, cRed, green, cGreen, notConectedB);

	smls->catMatrix();
	smls->catIMatrix();
	if(otherTriPart){
		actTriPart->split2Triangles(ActTri, ActTriCorner, otherTriPart, otherTriIndex, (notConectedB+2)%3);
	}
	else{
		actTriPart->splitTriangle(ActTri, smls->ActTriCorner);
	}
	if(complete){

	//	if(smls->build.length && red == green){
		if(red == green){
			int to = cGreen;
			if(cRed > cGreen){
				to = cRed;
				smls->ActCtrlPnt++;
			}
			to++;
			for(int i = red->coord.length; --i>=to;){
				red->swapVertex(i, i - 1);
			}
		}

		smls->doBuild();
		makePointerEqualIndexLocation();
		smls->convertToRenderFormat();
		BackUp::monostable.beginCycle();
	}
}
void swapDivide(ImmediatePart* part, Art* actTriPart, const int ActTri, const int ActCorner);
class Edge{
public:
	Edge():
		part(0),
		x(0),
		y(0)
	{}
	Edge(Art* part, int x, int y):
		part(part),
		x(x),
		y(y)
	{
	}
	Art* part;
	int x;
	int y;
};

class TwoPoint{
public:
	TwoPoint():
		partA(0),
		idxA(0),
		partB(0),
		idxB(0)
	{}
	TwoPoint(Art* partA, int idxA, Art* partB, int idxB):
		partA(partA),
		idxA(idxA),
		partB(partB),
		idxB(idxB)
	{
	}
	Art* partA;
	int idxA;
	Art* partB;
	int idxB;
};

inline bool splitFan(ImmediatePart* part, Art* actTriPart, M_<Edge>& edges, M_<S_VertexIndex>& triList){
	Seamless* smls = part->smls;

	for(int i = triList.length; --i>=0;){
		Art* triPart = triList[i].p;

		Art* vertPartA = triPart->triangle[triList[i].i].pa;
		int crdIdxA = triPart->triangle[triList[i].i].ia;

		Art* vertPartB = triPart->triangle[triList[i].i].pb;
		int crdIdxB = triPart->triangle[triList[i].i].ib;

		Art* vertPartC = triPart->triangle[triList[i].i].pc;
		int crdIdxC = triPart->triangle[triList[i].i].ic;

		if( vertPartA == part && crdIdxA == smls->ActCtrlPnt){
			continue;
		}
		if( vertPartB == part && crdIdxB == smls->ActCtrlPnt){
			continue;
		}
		if( vertPartC == part && crdIdxC == smls->ActCtrlPnt){
			continue;
		}
		return false;
	}
	int lastSplitCoordA = smls->actTriPart->triangle[smls->ActTri].refIndex(smls->ActTriCorner);
	int lastSplitCoordB = smls->actTriPart->triangle[smls->ActTri].refIndex((smls->ActTriCorner + 1) % 3);

	int ActTri = smls->ActTri;
	int ActTriCorner = smls->ActTriCorner;
	Art* remActTriPart=actTriPart;


	while(1){
		edges += Edge(smls->actTriPart, smls->ActTri, smls->ActTriCorner);

		int ActTri = smls->ActTri; //triArray[nextTriIdx++];
		const int ActTriCorner = (smls->ActTriCorner + 2) % 3;
		
		int idxRed = ActTri * 3 + ActTriCorner;
		int idxGreen = ActTri * 3 + (ActTriCorner + 1) % 3;
		int notConectedA = (ActTriCorner + 2) % 3;
		const int cRed = actTriPart->triangle.refIndex(idxRed);
		const int cGreen = actTriPart->triangle.refIndex(idxGreen);
		Art* red = actTriPart->triangle.refPart(idxRed);
		Art* green = actTriPart->triangle.refPart(idxGreen);
		int otherTriIndex=0;
		int notConectedB = 0;

		Art* otherTriPart = findTriangleJoinedBy2Vertices(actTriPart, ActTri, otherTriIndex, red, cRed, green, cGreen, notConectedB, MultiSelect::triList);
		if(edges[0].x==otherTriIndex){
			break;
		}
		if(otherTriPart){
			smls->actTriPart = otherTriPart;
			smls->ActTri = otherTriIndex;
			smls->ActTriCorner = (notConectedB + 1) % 3;
			smls->ActCtrlPnt = smls->actTriPart->triangle[smls->ActTri].refIndex(smls->ActTriCorner);

			lastSplitCoordA = smls->ActCtrlPnt;
			lastSplitCoordB = smls->actTriPart->triangle[smls->ActTri].refIndex((smls->ActTriCorner + 1) % 3);
		}
		else{
			//take care of the end edge
			for(int i = 3; --i>=0;){
				int cornerA = (smls->ActTriCorner + i) % 3;
				int cornerB = (cornerA + 1) % 3;
				int a = smls->actTriPart->triangle[smls->ActTri].refIndex(cornerA);
				int b = smls->actTriPart->triangle[smls->ActTri].refIndex(cornerB);

				if(a != lastSplitCoordA && b != lastSplitCoordA && b != lastSplitCoordB){
					edges += Edge(smls->actTriPart, smls->ActTri, cornerA);
					break;
				}
				else if(b != lastSplitCoordA && a != lastSplitCoordA && a != lastSplitCoordB){
					edges += Edge(smls->actTriPart, smls->ActTri, cornerB);
					break;
				}
			}
			break;
		}
	}
	int divTri = 0;
	for(int i = 0; i < edges.length; i++){
		smls->ActTri = edges[i].x;
		smls->ActTriCorner = edges[i].y;
		smls->ActCtrlPnt = smls->actTriPart->triangle[smls->ActTri].refIndex(smls->ActTriCorner);
		splitTriangleImmediate(part, actTriPart, smls->ActTri);

		if(i == 0){
			divTri = part->triangle.length-1;
		}
	}
	smls->actTriPart = remActTriPart;
	smls->ActTri = ActTri;
	smls->ActTriCorner = ActTriCorner;
	smls->ActCtrlPnt = smls->actTriPart->triangle[smls->ActTri].refIndex(smls->ActTriCorner);


	smls->actTriPart = remActTriPart;
	smls->ActTri = divTri;
	smls->ActTriCorner = 0;
	smls->ActCtrlPnt = smls->actTriPart->triangle[smls->ActTri].refIndex(smls->ActTriCorner);

	swapDivide(part, smls->actTriPart, smls->ActTri, smls->ActTriCorner);
	return true;


	int idxRed = smls->ActTri * 3 + ActTriCorner;
	int idxGreen = smls->ActTri * 3 + (ActTriCorner + 1) % 3;
	int notConectedA = (ActTriCorner + 2) % 3;
	int cRed = remActTriPart->triangle.refIndex(idxRed);
	int cGreen = remActTriPart->triangle.refIndex(idxGreen);
	Art* red = remActTriPart->triangle.refPart(idxRed);
	Art* green = remActTriPart->triangle.refPart(idxGreen);
	int otherTriIndex=0;
	int notConectedB = 0;

	Art* otherTriPart = findTriangleJoinedBy2Vertices(remActTriPart, smls->ActTri, otherTriIndex, red, cRed, green, cGreen, notConectedB);

	smls->actTriPart = otherTriPart;
	smls->ActTri = otherTriIndex;
	smls->ActTriCorner = (notConectedB + 1 + 2) % 3;
	smls->ActCtrlPnt = smls->actTriPart->triangle[smls->ActTri].refIndex(smls->ActTriCorner);
////////////////
	ActTriCorner = smls->ActTriCorner;

	idxRed = smls->ActTri * 3 + ActTriCorner;
	idxGreen = smls->ActTri * 3 + (ActTriCorner + 1) % 3;
	notConectedA = (ActTriCorner + 2) % 3;
	cRed = remActTriPart->triangle.refIndex(idxRed);
	cGreen = remActTriPart->triangle.refIndex(idxGreen);
	red = remActTriPart->triangle.refPart(idxRed);
	green = remActTriPart->triangle.refPart(idxGreen);
	otherTriIndex=0;
	notConectedB = 0;

	otherTriPart = findTriangleJoinedBy2Vertices(remActTriPart, smls->ActTri, otherTriIndex, red, cRed, green, cGreen, notConectedB);

	smls->actTriPart = otherTriPart;
	smls->ActTri = otherTriIndex;
	smls->ActTriCorner = (notConectedB + 1 + 2) % 3;
	smls->ActCtrlPnt = smls->actTriPart->triangle[smls->ActTri].refIndex(smls->ActTriCorner);


	swapDivide(part, smls->actTriPart, smls->ActTri, smls->ActTriCorner);

	return true;	
}
void splitBand(ImmediatePart* part, Art* _actTriPart){
	//M_Int triArray = M_Int(6, 15, 14, 23, 7, 16, 22, 17, 25, 24, 26, 27);
	//M_Art partArray = M_Art(actTriPart, actTriPart, actTriPart, actTriPart, actTriPart, actTriPart, actTriPart, actTriPart, actTriPart, actTriPart, actTriPart, actTriPart);

	M_<Edge> edges;
	M_<TwoPoint> divIdx;
	M_Art divPartA;
	M_Art divPartB;


	fastFullBuildForAutoIn(true);

	Seamless* smls = part->smls;
	smls->actTriPart = _actTriPart;

	if(splitFan(part, smls->actTriPart, edges, MultiSelect::triList)){
		goto here;
	}


	int lastSplitCoordA = smls->actTriPart->triangle[smls->ActTri].refIndex(smls->ActTriCorner);
	int lastSplitCoordB = smls->actTriPart->triangle[smls->ActTri].refIndex((smls->ActTriCorner + 1) % 3);

	//if(nextTriIdx>=triArray.length){
	//	nextTriIdx=0;
	//}
	//S_Triangle& tri=part->triangle[smls->ActTri];
	//int nsel = 0;
	////for(int i = 0; i < 3; i++){
	////	if(tri.refPart(i) == part && tri.refIndex(i) == cRed){
	////		nsel = (i + 1) % 3;;
	////	}
	////}
	//smls->ActTriCorner=nsel;
	//smls->ActCtrlPnt=tri.refIndex(nsel);

	bool splitCycle = true;
	while(1){
		if(splitCycle){
			edges += Edge(smls->actTriPart, smls->ActTri, smls->ActTriCorner);
			//dLog(String("Tri ") + smls->ActTri + String(" ") + String(smls->ActTriCorner));
		}
		else{
			int next = (smls->ActTriCorner + 1) % 3;
			divIdx += TwoPoint(
				smls->actTriPart->triangle[smls->ActTri].refPart(smls->ActTriCorner),
				smls->actTriPart->triangle[smls->ActTri].refIndex(smls->ActTriCorner),
				smls->actTriPart->triangle[smls->ActTri].refPart(next),
				smls->actTriPart->triangle[smls->ActTri].refIndex(next)
			);
		}
		splitCycle = !splitCycle;

		int ActTri = smls->ActTri; //triArray[nextTriIdx++];
		const int ActTriCorner = (smls->ActTriCorner + 1) % 3;
		
		int idxRed = ActTri * 3 + ActTriCorner;
		int idxGreen = ActTri * 3 + (ActTriCorner + 1) % 3;
		int notConectedA = (ActTriCorner + 2) % 3;
		const int cRed = smls->actTriPart->triangle.refIndex(idxRed);
		const int cGreen = smls->actTriPart->triangle.refIndex(idxGreen);
		Art* red = smls->actTriPart->triangle.refPart(idxRed);
		Art* green = smls->actTriPart->triangle.refPart(idxGreen);
		int otherTriIndex=0;
		int notConectedB = 0;

		Art* otherTriPart = findTriangleJoinedBy2Vertices(smls->actTriPart, ActTri, otherTriIndex, red, cRed, green, cGreen, notConectedB, MultiSelect::triList);

		//smls->catMatrix();
		//smls->catIMatrix();
		if(edges[0].x==otherTriIndex){
			break;
		}
		if(otherTriPart){
	//		actTriPart->split2Triangles(ActTri, ActTriCorner, otherTriPart, otherTriIndex, (notConectedB+2)%3);

			smls->actTriPart = otherTriPart;
			smls->ActTri = otherTriIndex;
			smls->ActTriCorner = (notConectedB + 1) % 3;
			smls->ActCtrlPnt = smls->actTriPart->triangle[smls->ActTri].refIndex(smls->ActTriCorner);

		}
		else{
			const int ActTriCorner = (smls->ActTriCorner + 2) % 3;
			
			int idxRed = ActTri * 3 + ActTriCorner;
			int idxGreen = ActTri * 3 + (ActTriCorner + 1) % 3;
			int notConectedA = (ActTriCorner + 2) % 3;
			const int cRed = smls->actTriPart->triangle.refIndex(idxRed);
			const int cGreen = smls->actTriPart->triangle.refIndex(idxGreen);
			Art* red = smls->actTriPart->triangle.refPart(idxRed);
			Art* green = smls->actTriPart->triangle.refPart(idxGreen);
			int otherTriIndex=0;
			int notConectedB = 0;

			Art* otherTriPart = findTriangleJoinedBy2Vertices(smls->actTriPart, ActTri, otherTriIndex, red, cRed, green, cGreen, notConectedB, MultiSelect::triList);
			if(edges[0].x==otherTriIndex){
				break;
			}
			if(otherTriPart){
				smls->actTriPart = otherTriPart;
				smls->ActTri = otherTriIndex;
				smls->ActTriCorner = (notConectedB + 1) % 3;
				smls->ActCtrlPnt = smls->actTriPart->triangle[smls->ActTri].refIndex(smls->ActTriCorner);

				lastSplitCoordA = smls->ActCtrlPnt;
				lastSplitCoordB = smls->actTriPart->triangle[smls->ActTri].refIndex((smls->ActTriCorner + 1) % 3);
			}
			else{
				//take care of the end edge
				for(int i = 3; --i>=0;){
					int cornerA = (smls->ActTriCorner + i) % 3;
					int cornerB = (cornerA + 1) % 3;
					int a = smls->actTriPart->triangle[smls->ActTri].refIndex(cornerA);
					int b = smls->actTriPart->triangle[smls->ActTri].refIndex(cornerB);

					if(a != lastSplitCoordA && b != lastSplitCoordA && b != lastSplitCoordB){
						edges += Edge(smls->actTriPart, smls->ActTri, cornerA);
						break;
					}
					else if(b != lastSplitCoordA && a != lastSplitCoordA && a != lastSplitCoordB){
						edges += Edge(smls->actTriPart, smls->ActTri, cornerB);
						break;
					}
				}
				break;
			}
		}
	}
	for(int i = 0; i < edges.length; i++){
		smls->actTriPart = edges[i].part;
		smls->ActTri = edges[i].x;
		smls->ActTriCorner = edges[i].y;
		smls->ActCtrlPnt = smls->actTriPart->triangle[smls->ActTri].refIndex(smls->ActTriCorner);
		splitTriangleImmediate(dynamic_cast<ImmediatePart*>(smls->actTriPart), smls->actTriPart, smls->ActTri);

		//smls->ActTri = part->triangle.length-1;
		//smls->ActTriCorner = 2;
		//smls->ActCtrlPnt = smls->actTriPart->triangle[smls->ActTri].refIndex(smls->ActTriCorner);
		//swapDivide(part, actTriPart, smls->ActTri);
	}
	for(int i = 0; i < divIdx.length; i++){
	//	smls->ActTri = edges[i].x;
	//	smls->ActTriCorner = edges[i].y;
	//	smls->ActCtrlPnt = smls->actTriPart->triangle[smls->ActTri].refIndex(smls->ActTriCorner);
		int triIndex = 0, triCorner = 0;
		Art* otherPart = findTriangleJoinedBy2Vertices(divIdx[i].partA, divIdx[i].idxA, divIdx[i].partB, divIdx[i].idxB, triIndex, triCorner);
		swapDivide(dynamic_cast<ImmediatePart*>(otherPart), otherPart, triIndex, triCorner);

		//dLog(String("div ") + triIndex + String(" ") + triCorner);
	////	//swapDivide(part, actTriPart, smls->ActTri);
	////	//splitTriangleImmediate(part, actTriPart, smls->ActTri);
	}
here:
	MultiSelect::reset();
	smls->doBuild();
	makePointerEqualIndexLocation();
	smls->convertToRenderFormat();
	BackUp::monostable.beginCycle();
}
void splitBand(){
#ifdef PROG_ROB
			dLog("new ClickButton(splitBand),");
#endif
	if(Part* whitePart=dynamic_cast<Part*>(g_selNode.getRaw())){
		Seamless* smls = whitePart->smls;
		NurbsPatch* patch = 0;
		Stem* stem = 0;
		Polymap* polymap = 0;
		ImmediatePart* bluePart = 0;
		if(getPolyMap(patch, stem, polymap, whitePart, smls->ActTri, bluePart)){
			if(bluePart){
				splitBand(bluePart, bluePart);
			}
		}
	}
	else if(ImmediatePart* bluePart=dynamic_cast<ImmediatePart*>(g_selNode.getRaw())){
		Seamless* smls = bluePart->smls;
		splitBand(bluePart, smls->actTriPart);
	}
	else{
		needPartMsg();
	}
}
void splitTriangleImmediate(){
#ifdef PROG_ROB
			dLog("new ClickButton(splitTriangleImmediate),");
#endif
	if(Part* whitePart=dynamic_cast<Part*>(g_selNode.getRaw())){
		Seamless* smls = whitePart->smls;
		NurbsPatch* patch = 0;
		Stem* stem = 0;
		Polymap* polymap = 0;
		ImmediatePart* bluePart = 0;
		if(getPolyMap(patch, stem, polymap, whitePart, smls->ActTri, bluePart)){
			if(bluePart){
				if(MultiSelect::triList.length){
					splitBand(bluePart, smls->actTriPart);
				}
				else{
					splitTriangleImmediate(bluePart, bluePart, smls->ActTri-stem->beginTriOffset, true);
				}
			}
			else{
				int ActTri = smls->ActTri;
				int ActTriCorner = smls->ActTriCorner;
				if(stem->mirror){
					if(	(stem->mirrorStem && stem->insideOut) ||
						(stem->mirrorStem == 0 && ActTri >= stem->coordIndex.length / 3)
						){
						ActTriCorner = abs(ActTriCorner + 1) % 3;
					}
				}
				if(0 > getRelIndex(stem, ActTri, ActTriCorner)){
					return;
				}

				bluePart = selectPolyMapPart(whitePart);
				stem->copyPolyMapToPart(bluePart);
				bluePart->smls->ActTriCorner = ActTriCorner;
				splitTriangleImmediate(bluePart, bluePart, ActTri / 3);
				stem->copyPartToPolyMap(bluePart);
				BackUp::monostable.beginCycle();
				commonEndVertexBuild();
			}
		}
	}
	else if(ImmediatePart* bluePart=dynamic_cast<ImmediatePart*>(g_selNode.getRaw())){
		Seamless* smls = bluePart->smls;

		if(MultiSelect::triList.length){
			splitBand(bluePart, smls->actTriPart);
		}
		else{
			splitTriangleImmediate(bluePart, smls->actTriPart, smls->ActTri, true);
		}
	}
	else if(Subdivision* subdivision=dynamic_cast<Subdivision*>(g_selNode.getRaw())){
		subdivision->extrudePoly();
	}
	else{
		needPartMsg();
	}
}
void lightenVertex(){
	if(Art* part=dynamic_cast<Art*>(g_selNode.getRaw())){
		Seamless* smls=part->smls;
		M_Vec3f& coord=part->coord;
		if(coord.length>0 &&part->workEffect!=0&&part->workEffect->getType()==COLOREFFECT){
			M_Color3f& color=dynamic_cast<ColorEffect*>(part->workEffect)->color;
			Vec3f vec=Vec3f(color[smls->ActCtrlPnt].r,color[smls->ActCtrlPnt].g,color[smls->ActCtrlPnt].b);
			vec=vec+vec*.005;
			if(vec.x>1)vec.x=1;
			if(vec.y>1)vec.y=1;
			if(vec.z>1)vec.z=1;
			color[smls->ActCtrlPnt]=Color3f(vec.x,vec.y,vec.z);
			commonEndVertexChange();
		}
	}
	else{
		needPartMsg();
	}
}
void darkenVertex(){
	if(Art* part=dynamic_cast<Art*>(g_selNode.getRaw())){
		Seamless* smls=part->smls;
		M_Vec3f& coord=part->coord;
		if(coord.length>0 &&part->workEffect!=0&&part->workEffect->getType()==COLOREFFECT){
			M_Color3f& color=dynamic_cast<ColorEffect*>(part->workEffect)->color;
			Vec3f vec=Vec3f(color[smls->ActCtrlPnt].r,color[smls->ActCtrlPnt].g,color[smls->ActCtrlPnt].b);
			vec=vec-vec*.005;
			if(vec.x<0)vec.x=0;
			if(vec.y<0)vec.y=0;
			if(vec.z<0)vec.z=0;
			color[smls->ActCtrlPnt]=Color3f(vec.x,vec.y,vec.z);
			commonEndVertexChange();
		}
	}
	else{
		needPartMsg();
	}
}
void reverseIndices(){
	if(Art* part=dynamic_cast<Art*>(g_selNode.getRaw())){
		part->insideOut();
		makePointerEqualIndexLocation();
		part->smls->convertToRenderFormat();
		BackUp::monostable.beginCycle();
	}
	else{
		needPartMsg();
	}
}

void doBuild(){
	if(Oscillator* osc = dynamic_cast<Oscillator*>(g_selNode.getRaw())){
		for(int k = 0; k < osc->timbre.length; k++){
			if(NurbsPatch* patch = dynamic_cast<NurbsPatch*>(osc->timbre[k].p)){
				patch->updateVCoord(*osc);
				patch->processFields();
				//patch->doBuild();
				//patch->smls->ctrf = true;
			}
		}
	}
	if(HarmonicPatch* patch = dynamic_cast<HarmonicPatch*>(g_selNode.getRaw())){
		int selLine = g_stw->selLine;
		int parentLine = getParent(selLine);
		if(Oscillator* osc = dynamic_cast<Oscillator*>(g_sceneTreeList[parentLine].node.p)){
			patch->updateVCoord(*osc);
			patch->processFields();
			patch->houseKeep();
		}
	}
	else{
		searchDoBuild(g_scene->children.refValue());
		makePointerEqualIndexLocation();
		Window::paintAllControlPanels();
		if(NurbsPatch* nurbsPatch=dynamic_cast<NurbsPatch*>(g_selNode.getRaw())){
			nurbsPatch->houseKeep();
		}
	}
}

void resetRotations(){
	if(Window::chatMode())return;
	g_scene->resetRotations();
	Window::paintAllControlPanels();
}


void zeroVetexX(){
	if(Window::chatMode())return;
	if(Part* whitePart=dynamic_cast<Part*>(g_selNode.getRaw())){
		Seamless* smls = whitePart->smls;
		NurbsPatch* patch = 0;
		Stem* stem = 0;
		Polymap* polymap = 0;
		ImmediatePart* bluePart = 0;
		if(getPolyMap(patch, stem, polymap, whitePart, smls->ActTri, bluePart)){
			if(bluePart){
			}
			else{
				stem->coord.refValue()[smls->ActCtrlPnt-stem->beginCoordOffset].x=.5;
				commonEndVertexBuild();
			}
		}
	}
	else if(ImmediatePart* part=dynamic_cast<ImmediatePart*>(g_selNode.getRaw())){
		Seamless* smls=part->smls;

		part->coord.refValue()[smls->ActCtrlPnt].x=0;
		smls->convertToRenderFormat();
		makePointerEqualIndexLocation();
		BackUp::monostable.beginCycle();
	}
	else{
		needPartMsg();
	}
}
void BackUp::unDo(){
	if(Window::chatMode())return;
	if(monostable.state())return;
	bool accumulate = MultiSelect::accumulate;
	MultiSelect::accumulate = false;
	MultiSelect::reset();
	MultiSelect::accumulate = accumulate;

	static String filename;

	float storeZoom; Vec3f storeTranslation; Rotation storeRotation; g_scene->storeZoomRotTran(storeZoom, storeTranslation, storeRotation);

	if(backedUpFileCntr<1)return;
	filename=g_undoDir;
	filename+=ourBackUpName;
	filename+=--backedUpFileCntr;
	filename+=".smls";

	initCompile();
	Browser::createSceneFromFile(filename);
	g_scene->restoreZoomRotTran(storeZoom, storeTranslation, storeRotation);

	commonUndoRedo();
}
void BackUp::reDo(){
	if(Window::chatMode())return;
	if(monostable.state())return;
	static String filename;

	float storeZoom; Vec3f storeTranslation; Rotation storeRotation; g_scene->storeZoomRotTran(storeZoom, storeTranslation, storeRotation);

	if(backedUpFileCntr>=backedUpFileCntrEnd)return;
	filename=g_undoDir;
	filename+=ourBackUpName;
	filename+=++backedUpFileCntr;
	filename+=".smls";

	initCompile();
	Browser::createSceneFromFile(filename);

	g_scene->restoreZoomRotTran(storeZoom, storeTranslation, storeRotation);

	commonUndoRedo();
}
void unDo(){
#ifdef PROG_ROB
	dLog("new ClickButtonY(BackUp::unDo),");
#endif
	BackUp::unDo();
}
void reDo(){
#ifdef PROG_ROB
	dLog("new ClickButtonY(BackUp::reDo),");
#endif
	BackUp::reDo();
}
void pivotPointToClipBoard(){
	if(Art* part=dynamic_cast<Art*>(g_selNode.getRaw())){
		if (OpenClipboard(hWndMain)){
			EmptyClipboard(); 

			// Allocate a global memory object for the text. 
		
			String str;
			part->smls->catIMatrix();
			(part->matrix * part->absolutePivotPoint).toString(str);

			HGLOBAL hglbCopy = GlobalAlloc(GMEM_MOVEABLE, str.length+1);
			if (hglbCopy == NULL) 
			{ 
				CloseClipboard(); 
			}
			else{
		
				// Lock the handle and copy the text to the buffer. 
			
				LPVOID lptstrCopy = GlobalLock(hglbCopy); 
				memcpy(lptstrCopy, str.c_str(), str.length+1); 

				GlobalUnlock(hglbCopy); 
			
				// Place the handle on the clipboard. 
			
				SetClipboardData(CF_TEXT, hglbCopy);
				CloseClipboard();
			}
		}
	}
}
void cut(){
	Window::activeWindow->copyTextCellToClipboard(true);
}
void remember(){
	Window::activeWindow->copyTextCellToClipboard();

	g_stw->copyLine = g_stw->selLine;
	g_copySibCount=g_selCount;
	g_copyAllCount=g_selCount.all;
	M_Int* indexPtr=g_selNode->getIndexPtr();
	if(indexPtr!=nullPtr){
		g_copyIndex=*indexPtr;
	}
	if(Art* part = dynamic_cast<Art*>(g_selNode.getRaw())){
		remPJ = part;
		Seamless* smls = part->smls;
		smls->remActCtrlPnt = smls->ActCtrlPnt;
		smls->remActTriPart = smls->actTriPart;
		smls->remActTri = smls->ActTri;
		smls->remActTriCorner = smls->ActTriCorner;
		if(ColorEffect* colorEffect = dynamic_cast<ColorEffect*>(part->workEffect)){
			int triIdx = smls->ActTri * 3 + smls->ActTriCorner;
			if(triIdx < part->triangle.getTriangleLength()){
				int colorIdx = part->triangle.refTexIndex(triIdx);
				if(colorIdx < colorEffect->color.length){
					g_palette->color = colorEffect->color[colorIdx];
					if(PaletteCP* paletteCP = paletteCPAllreadyOpen()){
						paletteCP->paint();
					}
				}
			}
		}
	}
	else if(NurbsPatch* patch = dynamic_cast<NurbsPatch*>(g_selNode.getRaw())){
		 g_rememberIndex = patch->index[0];
	}
//		triVertexRem.refPart(1)=triVertexRem.refPart(2);
///		triVertexRem.refPart(2)=smls->actTriPart;
//		triVertexRem.refIndex(1)=triVertexRem.refIndex(2);
//		triVertexRem.refIndex(2)=smls->ActTri*3+smls->ActTriCorner;
#ifdef PROG_ROB
		dLog("new ClickButton(remember),");
#endif

}
void swapVertexOrderDec(ImmediatePart* part, int ActCtrlPnt){
	Seamless* smls=part->smls;
	if(ActCtrlPnt>0){
		fastFullBuildForAutoIn(true);
		part->swapVertex(ActCtrlPnt,ActCtrlPnt-1);
		smls->ActCtrlPnt--;
		smls->doBuild();
		smls->ctrf=true;
		makePointerEqualIndexLocation();
		BackUp::monostable.beginCycle();
	}
}
void swapVertexOrderInc(ImmediatePart* part, int ActCtrlPnt){
	Seamless* smls=part->smls;
	M_Vec3f& coord = part->coord.refValue();
	if( ActCtrlPnt<coord.length-1){
		fastFullBuildForAutoIn(true);
		part->swapVertex(ActCtrlPnt,ActCtrlPnt+1);
		smls->ActCtrlPnt++;
		smls->doBuild();
		smls->ctrf=true;
		makePointerEqualIndexLocation();
		BackUp::monostable.beginCycle();
	}
}
ImmediatePart* swapVertexOrderSub(Part* whitePart, int& ActCtrlPnt, Stem*& stem, Seamless* smls){
	smls->atThisStage=0;
	smls->doBuild();
	if(TransferOwnership* transferOwnership = dynamic_cast<TransferOwnership*>(smls->atThisStage)){
		if(Part* source = dynamic_cast<Part*>(transferOwnership->source.p)){
			M_Int& qty = transferOwnership->qty.refValue();
			M_Int& numOf = transferOwnership->numOf.refValue();
			M_SArt& m_dest = transferOwnership->dest.refValue();
			for(int k = 0; k < qty.length && k < m_dest.length && whitePart != m_dest[k]; k++){
				ActCtrlPnt += qty[k] * numOf[k];
			}
			if(transferOwnership->beginSP && ActCtrlPnt > 0){
				ActCtrlPnt-=qty[0]-1;
			}
			ActCtrlPnt+=smls->ActCtrlPnt;
			NurbsPatch* patch = 0;
			ImmediatePart* bluePart = 0;
			Polymap* polymap = 0;
			if(getPolyMap(patch, stem, polymap, source, smls->ActTri, bluePart)){
				return bluePart;	
			}
		}
	}
	return 0;
}
void swapVertexOrderSub(const Vec3f& selCoordPos, Seamless* smls){
	M_Art& partList = smls->partList;
	for(int k = partList.length; --k>=0;){
		Art* part = partList[k];
		const M_Vec3f& coord = part->coord;
		for(int i = coord.length; --i>=0;){
			if(coord[i] == selCoordPos){
				smls->ActCtrlPnt=i;
				g_stw->setSelNodeLine(partList[part->plidx]);
			}
		}
	}
	makePointerEqualIndexLocation();
}
void swapVertexOrderDec(){
	if(Part* whitePart=dynamic_cast<Part*>(g_selNode.getRaw())){
		if(g_colorOwnership){
			Seamless* smls = whitePart->smls;
			M_Vec3f& coord = whitePart->coord;

			NurbsPatch* patch = 0;
			Stem* stem = 0;
			Polymap* polymap = 0;
			ImmediatePart* bluePart = 0;
			if(getPolyMap(patch, stem, polymap, whitePart, smls->ActTri, bluePart)){
				if(bluePart){
					swapVertexOrderDec(bluePart, smls->ActCtrlPnt-stem->beginCoordOffset);
				}
			}
			else if(smls->ActCtrlPnt < coord.length){
				smls->selCoordPos = coord[smls->ActCtrlPnt];
				int ActCtrlPnt = 0;
				if(ImmediatePart* bluePart = swapVertexOrderSub(whitePart, ActCtrlPnt, stem, smls)){
					Vec3f selCoordPos = smls->selCoordPos;
					swapVertexOrderDec(bluePart, ActCtrlPnt-stem->beginCoordOffset);
					swapVertexOrderSub(selCoordPos, smls);
				}
				smls->atThisStage=smls;
			}
		}
	}
	else if(ImmediatePart* part=dynamic_cast<ImmediatePart*>(g_selNode.getRaw())){
		swapVertexOrderDec(part, part->smls->ActCtrlPnt);
	}
	else{
		needPartMsg();
	}
}
void swapVertexOrderInc(){
	if(Part* whitePart=dynamic_cast<Part*>(g_selNode.getRaw())){
		if(g_colorOwnership){
			Seamless* smls = whitePart->smls;
			M_Vec3f& coord = whitePart->coord;

			NurbsPatch* patch = 0;
			Stem* stem = 0;
			Polymap* polymap = 0;
			ImmediatePart* bluePart = 0;
			if(getPolyMap(patch, stem, polymap, whitePart, smls->ActTri, bluePart)){
				if(bluePart){
					swapVertexOrderInc(bluePart, smls->ActCtrlPnt-stem->beginCoordOffset);
				}
			}
			else if(smls->ActCtrlPnt < coord.length){
				smls->selCoordPos = coord[smls->ActCtrlPnt];
				int ActCtrlPnt = 0;
				if(ImmediatePart* bluePart = swapVertexOrderSub(whitePart, ActCtrlPnt, stem, smls)){
					Vec3f selCoordPos = smls->selCoordPos;
					swapVertexOrderInc(bluePart, ActCtrlPnt-stem->beginCoordOffset);
					swapVertexOrderSub(selCoordPos, smls);
				}
				smls->atThisStage=smls;
			}
		}
	}
	else if(ImmediatePart* part=dynamic_cast<ImmediatePart*>(g_selNode.getRaw())){
		swapVertexOrderInc(part, part->smls->ActCtrlPnt);
	}
	else{
		needPartMsg();
	}
}
void mirrorTriangleImmediate(ImmediatePart* part, Art* actTriPart, const int ActTri){//DEC_RED:
	fastFullBuildForAutoIn(true);
	Seamless* smls=part->smls;
	BackUp::monostable.beginCycle();
	smls->catMatrix();
	smls->catIMatrix();
	actTriPart->mirrorTriangle( ActTri, smls->ActTriCorner);
}
void mirrorTriangleImmediate(){
	if(Part* whitePart=dynamic_cast<Part*>(g_selNode.getRaw())){
		Seamless* smls = whitePart->smls;
		NurbsPatch* patch = 0;
		Stem* stem = 0;
		Polymap* polymap = 0;
		ImmediatePart* bluePart = 0;
		if(getPolyMap(patch, stem, polymap, whitePart, smls->ActTri, bluePart)){
			if(bluePart){
				mirrorTriangleImmediate(bluePart, bluePart, smls->ActTri-stem->beginTriOffset);
				commonEndVertexBuild();
			}
			else{
				bluePart = selectPolyMapPart(whitePart);
				stem->copyPolyMapToPart(bluePart);
				bluePart->smls->ActTriCorner = smls->ActTriCorner;
				mirrorTriangleImmediate(bluePart, bluePart, smls->ActTri-stem->beginTriOffset);
				stem->copyPartToPolyMap(bluePart);
				commonEndVertexBuild();
			}
		}
	}
	else if(ImmediatePart* bluePart=dynamic_cast<ImmediatePart*>(g_selNode.getRaw())){
		Seamless* smls = bluePart->smls;
		mirrorTriangleImmediate(bluePart, smls->actTriPart, smls->ActTri);
		commonEndVertexBuild();
	}
	else if(BezierPatch* bezierPatch=dynamic_cast<BezierPatch*>(g_selNode.getRaw())){
		bezierPatch->mirrorPatch();
	}
	else if(Subdivision* subdivision=dynamic_cast<Subdivision*>(g_selNode.getRaw())){
		subdivision->mirrorPoly();
	}
	else{
		needPartMsg();
	}
}

void genTriangle(){
	if(ImmediatePart* part=dynamic_cast<ImmediatePart*>(g_selNode.getRaw())){
		
		if(part->workEffect==0)return;
		part->deleteGeometry();
		Seamless* smls = part->smls;
		BackUp::monostable.beginCycle();
		Effect* effect=part->workEffect;
		int effectLength=effect->getEffectLength();
		if(effect->getType()==COLOREFFECT){
			dynamic_cast<ColorEffect*>(effect)->color.refValue().length+=3;
		}
		else if(effect->getType()==TEXTUREEFFECT){
			dynamic_cast<TextureEffect*>(effect)->texCoord.refValue().length+=3;
		}
		else if(effect->getType()==MONOCOLOREFFECT){
		}
		else throw;

		part->coord=M_Vec3f(
			Vec3f(-1, 0, 0),
			Vec3f(1, 0, 0),
			Vec3f(0, 1, 0)
			);
		part->triangle+=S_Triangle(part, 0, effectLength, part, 1, effectLength+1, part, 2, effectLength+2);
		part->relevantPart=M_Art(part);
		smls->process();

		smls->convertToRenderFormat();
		makePointerEqualIndexLocation();
	}
	else{
		needBluePartMsg();
	}
}
void transferSingleVertex(){
#ifdef PROG_ROB
	dLog("new ClickButton(transferSingleVertex),");
#endif
	if(Art* part=dynamic_cast<Art*>(g_selNode.getRaw())){
		if(dynamic_cast<Art*>(g_copyNode.getRaw())){
			fastFullBuildForAutoIn(true);
			BackUp::monostable.beginCycle();
			Seamless* smls=part->smls;
			smls->catMatrix();
			smls->catIMatrix();

			transferOneVertex(part, remPJ, smls->ActCtrlPnt);

			smls->process();
			
			smls->catMatrix();
			smls->catIMatrix();
			smls->convertToRenderFormat();

			g_stw->selLine=g_stw->copyLine;
			smls->ActCtrlPnt=remPJ->coord.length;
			g_selNode->getSceneDisplayString(g_stw->textCell.cellStr="");
			g_stw->textCell.updateCaret();
			makePointerEqualIndexLocation();
			setupSelectedNodeWindow();
		}
		else{
			needPartCopiedMsg();
		}
	}
	else{
		needPartMsg();
	}
}
void transferPartVertex(){
	if(remPJ==NULL)return;
	BackUp::monostable.beginCycle();
	Art* part=dynamic_cast<Art*>(g_selNode.getRaw());
	Seamless* &smls=part->smls;
	smls->ActCtrlPnt=remPJ->coord.length+smls->ActCtrlPnt;
	for (int i=part->coord.length; --i>=0; ){
		transferOneVertex(part, remPJ, 0);
	}


	smls->process();
	
	smls->catMatrix();
	smls->catIMatrix();
	smls->convertToRenderFormat();

	g_stw->selLine=g_stw->copyLine;
	smls->ActCtrlPnt=remPJ->coord.length;
	makePointerEqualIndexLocation();
	g_stw->paint();
	Window::paintAllControlPanels();
}
void openSettings(){
	ControlPanel* settings = new ControlPanel;
	settings->setupNodeWindow(g_settings);
	Window::activeWindow = settings;
}
void floorBuild(float floorHeight){
	removeTypeFromList <DownButton>(g_callBackList);
	g_stw->cp->setupNodeWindow(g_selNode.getRaw());

	float yPos = .92 + floorHeight;
	//front_view->position=Vec3f(0,-.83,3);
	//front_view->position=Vec3f(0,.92,3);

	NoOutputGroup* noOutputGroup = new NoOutputGroup;
	Viewpoint* front_view = new Viewpoint;
	front_view->name=name_front_view;
	front_view->perspective=false;
	front_view->orientation=Rotation(0,1,0,0);
	//front_view->position=Vec3f(0,-.83,3);
	front_view->position=Vec3f(0,yPos,3);
	noOutputGroup->children+=front_view;
	Viewpoint* left_view = new Viewpoint;
	left_view->name=name_left_view;
	left_view->perspective=false;
	left_view->orientation=Rotation(0,1,0,1.570796);
	left_view->position=Vec3f(3,yPos,0);
	noOutputGroup->children+=left_view;
	Viewpoint* right_view = new Viewpoint;
	right_view->name=name_right_view;
	right_view->perspective=false;
	right_view->orientation=Rotation(0,-1,0,1.570796);
	right_view->position=Vec3f(-3,yPos,0);
	noOutputGroup->children+=right_view;
	Viewpoint* back_view = new Viewpoint;
	back_view->name=name_back_view;
	back_view->perspective=false;
	back_view->orientation=Rotation(0,1,0,3.141593);
	back_view->position=Vec3f(0,yPos,-3);
	noOutputGroup->children+=back_view;
	Viewpoint* birds_eye_view = new Viewpoint;

	birds_eye_view->name=name_birds_eye_view;
	birds_eye_view->perspective=false;
	birds_eye_view->orientation=Rotation(-1,0,0,1.570796);
	birds_eye_view->position=Vec3f(0,(2.17 + 1.75) + floorHeight, 0);
	noOutputGroup->children+=birds_eye_view;
	Viewpoint* worms_eye_view = new Viewpoint;
	worms_eye_view->name=name_worms_eye_view;
	worms_eye_view->perspective=false;
	worms_eye_view->orientation=Rotation(1,0,0,1.570796);
	worms_eye_view->position=Vec3f(0,(-3.83 + 1.75) + floorHeight, 0);
	noOutputGroup->children+=worms_eye_view;
	Viewpoint* showcase_view = new Viewpoint;
	showcase_view->name=name_showcase_view;
	showcase_view->perspective=false;
	showcase_view->orientation=Rotation(-.483572, .867491, .116694,.675209);
	showcase_view->position=Vec3f(1.589557, (.143424 + 1.75) + floorHeight, 2.350692);
	noOutputGroup->children+=showcase_view;
	Seamless* smls = new Seamless;
	smls->translation.y = floorHeight;
	smls->collapsed=true;
	ColorEffect* n01136290 = new ColorEffect;
	smls->effect=n01136290;
	Part* floor = new Part;
	floor->name=name_floor;
	Part* ruller = new Part;
	ruller->name=name_ruller;
	floor->part+=ruller;
	smls->skeleton+=floor;
	GroupBuild* groupBuild = new GroupBuild;
	groupBuild->collapsed=true;
	Stem* stem = new Stem;
	stem->staves=1;
	stem->xDistance=.25;
	stem->bands=M_Int(1);
	stem->yDistance=M_Float(.25);
	stem->xJoinEnds=false;
	stem->part+=ruller;
	groupBuild->build+=stem;
	ColorSweep* n01138f40 = new ColorSweep;
	n01138f40->location=Vec3f(0,-.001,0);
	n01138f40->color=M_Color3f(Color3f(1,1,.5),Color3f(0,0,1));
	n01138f40->distance=M_Float(.252);
	n01138f40->opacity=M_Float(1,1);
	n01138f40->straight=true;
	n01138f40->part+=ruller;
	groupBuild->build+=n01138f40;
	CopyPart* n011395f8 = new CopyPart;
	n011395f8->add=true;
	n011395f8->insideOut=true;
	n011395f8->source+=ruller;
	n011395f8->dest+=ruller;
	groupBuild->build+=n011395f8;
	CopyPart* n01139a40 = new CopyPart;
	n01139a40->source+=ruller;
	n01139a40->dest+=floor;
	groupBuild->build+=n01139a40;
	Place* n01139e78 = new Place;
	n01139e78->translation=Vec3f(0,.25,0);
	n01139e78->descendants=false;
	n01139e78->part+=floor;
	groupBuild->build+=n01139e78;
	TransferVertex* n0113a290 = new TransferVertex;
	n0113a290->source=floor;
	n0113a290->dest=ruller;
	groupBuild->build+=n0113a290;
	CopyPart* n0113a620 = new CopyPart;
	n0113a620->source+=ruller;
	n0113a620->dest+=floor;
	groupBuild->build+=n0113a620;
	Place* n0113ab18 = new Place;
	n0113ab18->translation=Vec3f(0,.5,0);
	n0113ab18->descendants=false;
	n0113ab18->part+=floor;
	groupBuild->build+=n0113ab18;
	TransferVertex* n0113aee8 = new TransferVertex;
	n0113aee8->source=floor;
	n0113aee8->dest=ruller;
	groupBuild->build+=n0113aee8;
	CopyPart* n0113b278 = new CopyPart;
	n0113b278->source+=ruller;
	n0113b278->dest+=floor;
	groupBuild->build+=n0113b278;
	ColorSweep* n0113b6f8 = new ColorSweep;
	n0113b6f8->color=M_Color3f(Color3f(1,.5,.5),Color3f(1,.5,.5));
	n0113b6f8->distance=M_Float(1);
	n0113b6f8->opacity=M_Float(.5,.5);
	n0113b6f8->straight=true;
	n0113b6f8->part+=ruller;
	groupBuild->build+=n0113b6f8;
	Place* n0113bdb0 = new Place;
	n0113bdb0->translation=Vec3f(0,1,0);
	n0113bdb0->descendants=false;
	n0113bdb0->part+=floor;
	groupBuild->build+=n0113bdb0;
	TransferVertex* n0113c180 = new TransferVertex;
	n0113c180->source=floor;
	n0113c180->dest=ruller;
	groupBuild->build+=n0113c180;
	Place* n0113c4c8 = new Place;
	n0113c4c8->translation=Vec3f(.5,0,0);
	n0113c4c8->part+=ruller;
	groupBuild->build+=n0113c4c8;
	Stem* stem2 = new Stem;
	stem2->add=false;
	stem2->slide=Vec2f(0,-1);
	stem2->orientation=Rotation(1,0,0,-1.57096);
	stem2->staves=1;
	stem2->xDistance=2;
	stem2->bands=M_Int(1);
	stem2->yDistance=M_Float(2);
	stem2->xJoinEnds=false;
	stem2->part+=floor;
	groupBuild->build+=stem2;
	CopyPart* n0113d0d0 = new CopyPart;
	n0113d0d0->add=true;
	n0113d0d0->insideOut=true;
	n0113d0d0->source+=floor;
	n0113d0d0->dest+=floor;
	groupBuild->build+=n0113d0d0;
	smls->build+=groupBuild;
	stem->processFields();
	stem2->processFields();

	//Place* n0113d4c0 = new Place;
	//n0113d4c0->translation=Vec3f(0,-1.75,0);
	//n0113d4c0->part+=floor;
	//smls->build+=n0113d4c0;
	noOutputGroup->children+=smls;

	M_Node importDefList;
	importDefList=M_Node(
		front_view,
		left_view,
		right_view,
		back_view,
		birds_eye_view,
		worms_eye_view,
		showcase_view,
		floor,
		ruller
	);

	resetDefList();
	checkForNameClash(importDefList, g_defLookUp);

	g_scene->children+=noOutputGroup;

	smls->process();
	smls->doBuild();
	g_stw->genNodeList();
	g_stw->selLine=g_sceneTreeList.length-1;
	g_stw->textCell.cellStr = g_selNode->name;
	g_stw->paint();
	updateBuild3dWindow();
	Viewpoint::gotoFirstViewpoint();

/*
	String str = g_scene->children.refValue()[g_scene->children.refValue().length-1]->toCPPString();
	str.write("seamless_1212121_floor.cpp");
*/
}
void FloorHeight::ZeroHeight::operator = (const bool& par){
	floorBuild(0);
}
void FloorHeight::BlaxxunHeight::operator = (const bool& par){
	floorBuild(-1.75);
}
void floorBuild(){
	g_stw->cp->setupNodeWindow(g_floorHeight);
	Window::activeWindow=g_stw->cp.p;
}
void togglePerspective(){
	g_settings->perspective=!g_settings->perspective;
}
/*
void decCoordX(){
	Art* part=dynamic_cast<Art*>(g_selNode.getRaw());
	if(part){
		Seamless* smls=part->smls;
		M_Vec3f& coord=part->coord;
		if(coord.length>0){
			if(g_settings->tugRadius.refValue()>0){
				keyboardRadialTug(coord[smls->ActCtrlPnt], Vec3f(-g_incDecRes,0,0));
			}
			else{
				coord[smls->ActCtrlPnt].x-=g_incDecRes;
			}
			smls->updateAllVertexInBuffers();
			SendMessage(hWndMain, WM_PAINT,	0,	0	);
		}
	}
}
void incCoordX(){
	Art* part=dynamic_cast<Art*>(g_selNode.getRaw());
	if(part){
		Seamless* smls=part->smls;
		M_Vec3f& coord=part->coord;
		if(coord.length>0){
			if(g_settings->tugRadius.refValue()>0){
				keyboardRadialTug(coord[smls->ActCtrlPnt], Vec3f(g_incDecRes,0,0));
			}
			else{
				coord[smls->ActCtrlPnt].x+=g_incDecRes;
			}
			smls->updateAllVertexInBuffers();
			SendMessage(hWndMain, WM_PAINT,	0,	0	);
		}
	}
}
void decCoordY(){
	Art* part=dynamic_cast<Art*>(g_selNode.getRaw());
	if(part){
		Seamless* smls=part->smls;
		M_Vec3f& coord=part->coord;
		if(coord.length>0){
			if(g_settings->tugRadius>0){
				keyboardRadialTug(coord[smls->ActCtrlPnt], Vec3f(0,-g_incDecRes,0));
			}
			else{
				coord[smls->ActCtrlPnt].y-=g_incDecRes;
			}
			smls->updateAllVertexInBuffers();
			SendMessage(hWndMain, WM_PAINT,	0,	0	);
		}
	}
}
void incCoordY(){
	Art* part=dynamic_cast<Art*>(g_selNode.getRaw());
	if(part){
		Seamless* smls=part->smls;
		M_Vec3f& coord=part->coord;
		if(coord.length>0){
			if(g_settings->tugRadius.refValue()>0){
				keyboardRadialTug(coord[smls->ActCtrlPnt], Vec3f(0,g_incDecRes,0));
			}
			else{
				coord[smls->ActCtrlPnt].y+=g_incDecRes;
			}
			smls->updateAllVertexInBuffers();
			SendMessage(hWndMain, WM_PAINT,	0,	0	);
		}
	}
}
void decCoordZ(){
	Art* part=dynamic_cast<Art*>(g_selNode.getRaw());
	if(part){
		Seamless* smls=part->smls;
		M_Vec3f& coord=part->coord;
		if(coord.length>0){
			if(g_settings->tugRadius.refValue()>0){
				keyboardRadialTug(coord[smls->ActCtrlPnt], Vec3f(0,0,-g_incDecRes));
			}
			else{
				coord[smls->ActCtrlPnt].z-=g_incDecRes;
			}
			smls->updateAllVertexInBuffers();
			SendMessage(hWndMain, WM_PAINT,	0,	0	);
		}
	}
}
void incCoordZ(){
	Art* part=dynamic_cast<Art*>(g_selNode.getRaw());
	if(part){
		Seamless* smls=part->smls;
		M_Vec3f& coord=part->coord;
		if(coord.length>0){
			if(g_settings->tugRadius.refValue()>0){
				keyboardRadialTug(coord[smls->ActCtrlPnt], Vec3f(0,0,g_incDecRes));
			}
			else{
				coord[smls->ActCtrlPnt].z+=g_incDecRes;
			}
			smls->updateAllVertexInBuffers();
			SendMessage(hWndMain, WM_PAINT,	0,	0	);
		}
	}
}
*/
void decCoordX(){
	if(ImmediatePart* part=dynamic_cast<ImmediatePart*>(g_selNode.getRaw())){
		float res=getIncDecRes();
		Seamless* smls=part->smls;
		M_Vec3f& coord=part->coord;
		if(coord.length>0){
			if(g_settings->tugRadius.refValue()>0){
				keyboardRadialTug(coord[smls->ActCtrlPnt], Vec3f(-res,0,0));
			}
			else{
				Vec3f pt = coord[smls->ActCtrlPnt];
				Vec3f newPt = pt;
				quantizeDec(newPt.x);
				smls->senseMirrorVertex(part, smls->ActCtrlPnt, newPt);
				coord[smls->ActCtrlPnt]=newPt;
			}
			commonEndVertexChange();
		}
	}
	else{
		needBluePartMsg();
	}
}
void incCoordX(){
	if(ImmediatePart* part=dynamic_cast<ImmediatePart*>(g_selNode.getRaw())){
		float res=getIncDecRes();
		Seamless* smls=part->smls;
		M_Vec3f& coord=part->coord;
		if(coord.length>0){
			if(g_settings->tugRadius.refValue()>0){
				keyboardRadialTug(coord[smls->ActCtrlPnt], Vec3f(res,0,0));
			}
			else{
				Vec3f pt = coord[smls->ActCtrlPnt];
				Vec3f newPt = pt;
				quantizeInc(newPt.x);
				smls->senseMirrorVertex(part, smls->ActCtrlPnt, newPt);
				coord[smls->ActCtrlPnt] = newPt;
			}
			commonEndVertexChange();
		}
	}
	else{
		needBluePartMsg();
	}
}
void decCoordY(){
	if(ImmediatePart* part=dynamic_cast<ImmediatePart*>(g_selNode.getRaw())){
		float res=getIncDecRes();
		Seamless* smls=part->smls;
		M_Vec3f& coord=part->coord;
		if(coord.length>0){
			if(g_settings->tugRadius.refValue()>0){
				keyboardRadialTug(coord[smls->ActCtrlPnt], Vec3f(0,-res,0));
			}
			else{
				Vec3f pt = coord[smls->ActCtrlPnt];
				Vec3f newPt = pt;
				quantizeDec(newPt.y);
				smls->senseMirrorVertex(part, smls->ActCtrlPnt, newPt);
				coord[smls->ActCtrlPnt] = newPt;
			}
			commonEndVertexChange();
		}
	}
	else{
		needBluePartMsg();
	}
}
void incCoordY(){
	if(ImmediatePart* part=dynamic_cast<ImmediatePart*>(g_selNode.getRaw())){
		float res=getIncDecRes();
		Seamless* smls=part->smls;
		M_Vec3f& coord=part->coord;
		if(coord.length>0){
			if(g_settings->tugRadius.refValue()>0){
				keyboardRadialTug(coord[smls->ActCtrlPnt], Vec3f(0,res,0));
			}
			else{
				Vec3f pt = coord[smls->ActCtrlPnt];
				Vec3f newPt = pt;
				quantizeInc(newPt.y);
				smls->senseMirrorVertex(part, smls->ActCtrlPnt, newPt);
				coord[smls->ActCtrlPnt] = newPt;
			}
			commonEndVertexChange();
		}
	}
	else{
		needBluePartMsg();
	}
}
void decCoordZ(){
	if(ImmediatePart* part=dynamic_cast<ImmediatePart*>(g_selNode.getRaw())){
		float res=getIncDecRes();
		Seamless* smls=part->smls;
		M_Vec3f& coord=part->coord;
		if(coord.length>0){
			if(g_settings->tugRadius.refValue()>0){
				keyboardRadialTug(coord[smls->ActCtrlPnt], Vec3f(0,0,-res));
			}
			else{
				Vec3f pt = coord[smls->ActCtrlPnt];
				Vec3f newPt = pt;
				quantizeDec(newPt.z);
				smls->senseMirrorVertex(part, smls->ActCtrlPnt, newPt);
				coord[smls->ActCtrlPnt] = newPt;
			}
			commonEndVertexChange();
		}
	}
	else{
		needBluePartMsg();
	}
}
void incCoordZ(){
	if(ImmediatePart* part=dynamic_cast<ImmediatePart*>(g_selNode.getRaw())){
		float res=getIncDecRes();
		Seamless* smls=part->smls;
		M_Vec3f& coord=part->coord;
		if(coord.length>0){
			if(g_settings->tugRadius.refValue()>0){
				keyboardRadialTug(coord[smls->ActCtrlPnt], Vec3f(0,0,res));
			}
			else{
				Vec3f pt = coord[smls->ActCtrlPnt];
				Vec3f newPt = pt;
				quantizeInc(newPt.z);
				smls->senseMirrorVertex(part, smls->ActCtrlPnt, newPt);
				coord[smls->ActCtrlPnt] = newPt;
			}
			commonEndVertexChange();
		}
	}
	else{
		needBluePartMsg();
	}
}
void quantizeIncColor(float& v);
void quantizeDecColor(float& v);

void commonIncDecColorChange(Seamless* smls, Art* part, M_Color3f& color, const int colorIdx){
	int returnCoordIdxIdx=0;
	if(Art* mirrorPart=smls->getMirrorTriangle(part, returnCoordIdxIdx)){
		color[mirrorPart->triangle.refTexIndex(returnCoordIdxIdx)]=color[colorIdx];
	}
	commonEndVertexChange();
}
void decRed(ImmediatePart* part, Art* actTriPart, const int ActTri){//DEC_RED:
	Seamless* smls = part->smls;
	M_Vec3f& coord = part->coord;
	if(coord.length>0 &&part->workEffect!=0&&part->workEffect->getType()==COLOREFFECT){
		M_Color3f& color=dynamic_cast<ColorEffect*>(part->workEffect)->color;
		int colorIdx=actTriPart->triangle.refTexIndex(ActTri*3+smls->ActTriCorner);
		quantizeDecColor(color[colorIdx].r);

		commonIncDecColorChange(smls, part, color, colorIdx);
	}
}
void decRed(){//DEC_RED:
	if(Part* whitePart=dynamic_cast<Part*>(g_selNode.getRaw())){
		Seamless* smls = whitePart->smls;
		NurbsPatch* patch = 0;
		Stem* stem = 0;
		Polymap* polymap = 0;
		ImmediatePart* bluePart = 0;
		if(getPolyMap(patch, stem, polymap, whitePart, smls->ActTri, bluePart)){
			if(bluePart){
				decRed(bluePart, bluePart, smls->ActTri-stem->beginTriOffset);
			}
		}
	}
	else if(ImmediatePart* bluePart=dynamic_cast<ImmediatePart*>(g_selNode.getRaw())){
		Seamless* smls = bluePart->smls;
		decRed(bluePart, smls->actTriPart, smls->ActTri);
	}
	else{
		needPartMsg();
	}
}
void incRed(ImmediatePart* part, Art* actTriPart, const int ActTri){//DEC_RED:
	Seamless* smls = part->smls;
	M_Vec3f& coord = part->coord;
	if(coord.length>0 &&part->workEffect!=0&&part->workEffect->getType()==COLOREFFECT){
		M_Color3f& color=dynamic_cast<ColorEffect*>(part->workEffect)->color;
		int colorIdx=actTriPart->triangle.refTexIndex(ActTri*3+smls->ActTriCorner);
		quantizeIncColor(color[colorIdx].r);

		commonIncDecColorChange(smls, part, color, colorIdx);
	}
}
void incRed(){//INC_RED:
	if(Part* whitePart=dynamic_cast<Part*>(g_selNode.getRaw())){
		Seamless* smls = whitePart->smls;
		NurbsPatch* patch = 0;
		Stem* stem = 0;
		Polymap* polymap = 0;
		ImmediatePart* bluePart = 0;
		if(getPolyMap(patch, stem, polymap, whitePart, smls->ActTri, bluePart)){
			if(bluePart){
				incRed(bluePart, bluePart, smls->ActTri-stem->beginTriOffset);
			}
		}
	}
	else if(ImmediatePart* bluePart=dynamic_cast<ImmediatePart*>(g_selNode.getRaw())){
		Seamless* smls = bluePart->smls;
		incRed(bluePart, smls->actTriPart, smls->ActTri);
	}
	else{
		needPartMsg();
	}
}
void decGreen(ImmediatePart* part, Art* actTriPart, const int ActTri){//DEC_RED:
	Seamless* smls = part->smls;
	M_Vec3f& coord = part->coord;
	if(coord.length>0 &&part->workEffect!=0&&part->workEffect->getType()==COLOREFFECT){
		M_Color3f& color=dynamic_cast<ColorEffect*>(part->workEffect)->color;
		int colorIdx=actTriPart->triangle.refTexIndex(ActTri*3+smls->ActTriCorner);
		quantizeDecColor(color[colorIdx].g);

		commonIncDecColorChange(smls, part, color, colorIdx);
	}
}
void decGreen(){//DEC_GREEN:
	if(Part* whitePart=dynamic_cast<Part*>(g_selNode.getRaw())){
		Seamless* smls = whitePart->smls;
		NurbsPatch* patch = 0;
		Stem* stem = 0;
		Polymap* polymap = 0;
		ImmediatePart* bluePart = 0;
		if(getPolyMap(patch, stem, polymap, whitePart, smls->ActTri, bluePart)){
			if(bluePart){
				decGreen(bluePart, bluePart, smls->ActTri-stem->beginTriOffset);
			}
		}
	}
	else if(ImmediatePart* bluePart=dynamic_cast<ImmediatePart*>(g_selNode.getRaw())){
		Seamless* smls = bluePart->smls;
		decGreen(bluePart, smls->actTriPart, smls->ActTri);
	}
	else{
		needPartMsg();
	}
}
void incGreen(ImmediatePart* part, Art* actTriPart, const int ActTri){//DEC_RED:
	Seamless* smls = part->smls;
	M_Vec3f& coord = part->coord;
	if(coord.length>0 &&part->workEffect!=0&&part->workEffect->getType()==COLOREFFECT){
		M_Color3f& color=dynamic_cast<ColorEffect*>(part->workEffect)->color;
		int colorIdx=actTriPart->triangle.refTexIndex(ActTri*3+smls->ActTriCorner);
		quantizeIncColor(color[colorIdx].g);

		commonIncDecColorChange(smls, part, color, colorIdx);
	}
}
void incGreen(){//INC_GREEN:
	if(Part* whitePart=dynamic_cast<Part*>(g_selNode.getRaw())){
		Seamless* smls = whitePart->smls;
		NurbsPatch* patch = 0;
		Stem* stem = 0;
		Polymap* polymap = 0;
		ImmediatePart* bluePart = 0;
		if(getPolyMap(patch, stem, polymap, whitePart, smls->ActTri, bluePart)){
			if(bluePart){
				incGreen(bluePart, bluePart, smls->ActTri-stem->beginTriOffset);
			}
		}
	}
	else if(ImmediatePart* bluePart=dynamic_cast<ImmediatePart*>(g_selNode.getRaw())){
		Seamless* smls = bluePart->smls;
		incGreen(bluePart, smls->actTriPart, smls->ActTri);
	}
	else{
		needPartMsg();
	}
}
void decBlue(ImmediatePart* part, Art* actTriPart, const int ActTri){//DEC_RED:
	Seamless* smls = part->smls;
	M_Vec3f& coord = part->coord;
	if(coord.length>0 &&part->workEffect!=0&&part->workEffect->getType()==COLOREFFECT){
		M_Color3f& color=dynamic_cast<ColorEffect*>(part->workEffect)->color;
		int colorIdx=actTriPart->triangle.refTexIndex(ActTri*3+smls->ActTriCorner);
		quantizeDecColor(color[colorIdx].b);

		commonIncDecColorChange(smls, part, color, colorIdx);
	}
}
void decBlue(){//DEC_BLUE:
	if(Part* whitePart=dynamic_cast<Part*>(g_selNode.getRaw())){
		Seamless* smls = whitePart->smls;
		NurbsPatch* patch = 0;
		Stem* stem = 0;
		Polymap* polymap = 0;
		ImmediatePart* bluePart = 0;
		if(getPolyMap(patch, stem, polymap, whitePart, smls->ActTri, bluePart)){
			if(bluePart){
				decBlue(bluePart, bluePart, smls->ActTri-stem->beginTriOffset);
			}
		}
	}
	else if(ImmediatePart* bluePart=dynamic_cast<ImmediatePart*>(g_selNode.getRaw())){
		Seamless* smls = bluePart->smls;
		decBlue(bluePart, smls->actTriPart, smls->ActTri);
	}
	else{
		needPartMsg();
	}
}
void incBlue(ImmediatePart* part, Art* actTriPart, const int ActTri){//DEC_RED:
	Seamless* smls = part->smls;
	M_Vec3f& coord = part->coord;
	if(coord.length>0 &&part->workEffect!=0&&part->workEffect->getType()==COLOREFFECT){
		M_Color3f& color=dynamic_cast<ColorEffect*>(part->workEffect)->color;
		int colorIdx=actTriPart->triangle.refTexIndex(ActTri*3+smls->ActTriCorner);
		quantizeIncColor(color[colorIdx].b);

		commonIncDecColorChange(smls, part, color, colorIdx);
	}
}
void incBlue(){//INC_BLUE:
	if(Part* whitePart=dynamic_cast<Part*>(g_selNode.getRaw())){
		Seamless* smls = whitePart->smls;
		NurbsPatch* patch = 0;
		Stem* stem = 0;
		Polymap* polymap = 0;
		ImmediatePart* bluePart = 0;
		if(getPolyMap(patch, stem, polymap, whitePart, smls->ActTri, bluePart)){
			if(bluePart){
				incBlue(bluePart, bluePart, smls->ActTri-stem->beginTriOffset);
			}
		}
	}
	else if(ImmediatePart* bluePart=dynamic_cast<ImmediatePart*>(g_selNode.getRaw())){
		Seamless* smls = bluePart->smls;
		incBlue(bluePart, smls->actTriPart, smls->ActTri);
	}
	else{
		needPartMsg();
	}
}
void zoom(){
	g_scene->setZoom(g_scene->getZoom()/1.01F);
	houseKeepVisualClue();
}
void zoomIn(){
	g_scene->setZoom(g_scene->getZoom()*1.01F);
	houseKeepVisualClue();
}
void quantiseRotation(Rotation&);

void setFrontViewOrientation(){
#ifdef PROG_ROB
	dLog("new ClickButton(setFrontViewOrientation),");
#endif
	g_scene->rotation=Rotation(0,0,1,0);
	houseKeepVisualClueViewChange();
}
void setBackViewOrientation(){
#ifdef PROG_ROB
	dLog("new ClickButton(setBackViewOrientation),");
#endif
	g_scene->rotation=Rotation(0,1,0,PI);
	houseKeepVisualClueViewChange();
}
void setLeftViewOrientation(){
#ifdef PROG_ROB
	dLog("new ClickButton(setLeftViewOrientation),");
#endif
	g_scene->rotation=Rotation(0,1,0,-PI*.5);
	houseKeepVisualClueViewChange();
}
void setRightViewOrientation(){
#ifdef PROG_ROB
	dLog("new ClickButton(setRightViewOrientation),");
#endif
	g_scene->rotation=Rotation(0,1,0,PI*.5);
	houseKeepVisualClueViewChange();
}
void setBirdsEyeViewOrientation(){
#ifdef PROG_ROB
	dLog("new ClickButton(setBirdsEyeViewOrientation),");
#endif
	g_scene->rotation=Rotation(1,0,0,PI*.5);
	houseKeepVisualClueViewChange();
}
void setWormsEyeViewOrientation(){
#ifdef PROG_ROB
	dLog("new ClickButton(setWormsEyeViewOrientation),");
#endif
	g_scene->rotation=Rotation(1,0,0,-PI*.5);
	houseKeepVisualClueViewChange();
}
void quantiseViewOrientation(){
#ifdef PROG_ROB
	dLog("new ClickButton(quantiseViewOrientation),");
#endif
	quantiseRotation(g_scene->rotation);
	houseKeepVisualClueViewChange();
}
void commonDecIncViewOrientation90(){
	if(g_settings->absoluteRot90 || 0x8000&GetKeyState(VK_LCONTROL)){
		quantiseRotation(g_scene->rotation);
	}
	houseKeepVisualClueViewChange();
}
void decViewOrientation90X(){//DEC_VIEW_ORIENTATION_90_X
	g_scene->rotation=g_scene->rotation*Rotation(1,0,0,-PI*.25);
	commonDecIncViewOrientation90();
}
void incViewOrientation90X(){//INC_VIEW_ORIENTATION_90_X
	g_scene->rotation=g_scene->rotation*Rotation(1,0,0,PI*.25);
	commonDecIncViewOrientation90();
}
void decViewOrientation90Y(){//DEC_VIEW_ORIENTATION_90_Y
	g_scene->rotation=g_scene->rotation*Rotation(0,1,0,-PI*.25);
	commonDecIncViewOrientation90();
}
void incViewOrientation90Y(){//INC_VIEW_ORIENTATION_90_Y
	g_scene->rotation=g_scene->rotation*Rotation(0,1,0,PI*.25);
	commonDecIncViewOrientation90();
}
void decViewOrientation90Z(){//DEC_VIEW_ORIENTATION_90_Z
	g_scene->rotation=g_scene->rotation*Rotation(0,0,1,-PI*.25);
	commonDecIncViewOrientation90();
}
void incViewOrientation90Z(){//INC_VIEW_ORIENTATION_90_Z
	g_scene->rotation=g_scene->rotation*Rotation(0,0,1,PI*.25);
	commonDecIncViewOrientation90();
}
void decViewOrientationX(){//DEC_VIEW_ORIENTATION_X	//[
	if(0x8000&GetKeyState(VK_LCONTROL)){
		decViewOrientation90X();
	}
	else{
		g_scene->rotation=g_scene->rotation*Rotation(1,0,0,-PI*.01);
	}
	houseKeepVisualClueViewChange();
}
void incViewOrientationX(){//INC_VIEW_ORIENTATION_X	//]
	if(0x8000&GetKeyState(VK_LCONTROL)){
		incViewOrientation90X();
	}
	else{
		g_scene->rotation=g_scene->rotation*Rotation(1,0,0,PI*.01);
	}
	houseKeepVisualClueViewChange();
}
void decViewOrientationY(){//DEC_VIEW_ORIENTATION_Y	//;
	if(0x8000&GetKeyState(VK_LCONTROL)){
		decViewOrientation90Y();
	}
	else{
		g_scene->rotation=g_scene->rotation*Rotation(0,1,0,-PI*.01);
	}
	houseKeepVisualClueViewChange();
}
void incViewOrientationY(){//INC_VIEW_ORIENTATION_Y	//'
	if(0x8000&GetKeyState(VK_LCONTROL)){
		incViewOrientation90Y();
	}
	else{
		g_scene->rotation=g_scene->rotation*Rotation(0,1,0,PI*.01);
	}
	houseKeepVisualClueViewChange();
}
void decViewOrientationZ(){//DEC_VIEW_ORIENTATION_Z	//.
	if(0x8000&GetKeyState(VK_LCONTROL)){
		decViewOrientation90Z();
	}
	else{
		g_scene->rotation=g_scene->rotation*Rotation(0,0,1,-PI*.01);
	}
	houseKeepVisualClueViewChange();
}
void incViewOrientationZ(){//INC_VIEW_ORIENTATION_Z	///
	if(0x8000&GetKeyState(VK_LCONTROL)){
		incViewOrientation90Z();
	}
	else{
		g_scene->rotation=g_scene->rotation*Rotation(0,0,1,PI*.01);
	}
	houseKeepVisualClueViewChange();
}
float tranP(){
	return 1/g_scene->getZoom()*.01F;
}
void decViewPositionX(){//DEC_VIEW_POSITION_X
	g_scene->translation+=-g_scene->rotation*Vec3f(-tranP(),0,0);
}
void incViewPositionX(){//INC_VIEW_POSITION_X
	g_scene->translation+=-g_scene->rotation*Vec3f(tranP(),0,0);
}
void decViewPositionY(){//DEC_VIEW_POSITION_Y
	g_scene->translation+=-g_scene->rotation*Vec3f(0,-tranP(),0);
}
void incViewPositionY(){//INC_VIEW_POSITION_Y
	g_scene->translation+=-g_scene->rotation*Vec3f(0,tranP(),0);
}
void decViewPositionZ(){//DEC_VIEW_POSITION_Z
	g_scene->translation+=-g_scene->rotation*Vec3f(0,0,tranP());
}
void incViewPositionZ(){//INC_VIEW_POSITION_Z
	g_scene->translation+=-g_scene->rotation*Vec3f(0,0,-tranP());
}

//////////////
/*
void decViewOrientationX(){//DEC_VIEW_ORIENTATION_X	//[
	g_scene->orientation=g_scene->orientation.refValue()*Rotation(1,0,0,PI*.01);
}
void incViewOrientationX(){//INC_VIEW_ORIENTATION_X	//]
	g_scene->orientation=g_scene->orientation.refValue()*Rotation(1,0,0,-PI*.01);
}
void decViewOrientationY(){//DEC_VIEW_ORIENTATION_Y	//;
	g_scene->orientation=g_scene->orientation.refValue()*Rotation(0,1,0,PI*.01);
}
void incViewOrientationY(){//INC_VIEW_ORIENTATION_Y	//'
	g_scene->orientation=g_scene->orientation.refValue()*Rotation(0,1,0,-PI*.01);
}
void decViewOrientationZ(){//DEC_VIEW_ORIENTATION_Z	//.
	g_scene->orientation=g_scene->orientation.refValue()*Rotation(0,0,1,PI*.01);
}
void incViewOrientationZ(){//INC_VIEW_ORIENTATION_Z	///
	g_scene->orientation=g_scene->orientation.refValue()*Rotation(0,0,1,-PI*.01);
}
void decViewOrientation90X(){//DEC_VIEW_ORIENTATION_90_X
	g_scene->orientation=g_scene->orientation.refValue()*Rotation(1,0,0,PI*.5);
}
void incViewOrientation90X(){//INC_VIEW_ORIENTATION_90_X
	g_scene->orientation=g_scene->orientation.refValue()*Rotation(1,0,0,-PI*.5);
}
void decViewOrientation90Y(){//DEC_VIEW_ORIENTATION_90_Y
	g_scene->orientation=g_scene->orientation.refValue()*Rotation(0,1,0,PI*.5);
}
void incViewOrientation90Y(){//INC_VIEW_ORIENTATION_90_Y
	g_scene->orientation=g_scene->orientation.refValue()*Rotation(0,1,0,-PI*.5);
}
void decViewOrientation90Z(){//DEC_VIEW_ORIENTATION_90_Z
	g_scene->orientation=g_scene->orientation.refValue()*Rotation(0,0,1,PI*.5);
}
void incViewOrientation90Z(){//INC_VIEW_ORIENTATION_90_Z
	g_scene->orientation=g_scene->orientation.refValue()*Rotation(0,0,1,-PI*.5);
}
float tranP(){
	return 1/g_scene->zoom*.01;
}
void decViewPositionX(){//DEC_VIEW_POSITION_X
	g_scene->position=g_scene->position+g_scene->orientation.refValue()*Vec3f(tranP(),0,0);
}
void incViewPositionX(){//INC_VIEW_POSITION_X
	g_scene->position=g_scene->position+g_scene->orientation.refValue()*Vec3f(-tranP(),0,0);
}
void decViewPositionY(){//DEC_VIEW_POSITION_Y
	g_scene->position=g_scene->position+g_scene->orientation.refValue()*Vec3f(0,tranP(),0);
}
void incViewPositionY(){//INC_VIEW_POSITION_Y
	g_scene->position=g_scene->position+g_scene->orientation.refValue()*Vec3f(0,-tranP(),0);
}
void decViewPositionZ(){//DEC_VIEW_POSITION_Z
	g_scene->position=g_scene->position+g_scene->orientation.refValue()*Vec3f(0,0,-tranP());
}
void incViewPositionZ(){//INC_VIEW_POSITION_Z
	g_scene->position=g_scene->position+g_scene->orientation.refValue()*Vec3f(0,0,tranP());
}
*/
Seamless* getSeamless(){
	for(int k=0; k<g_sceneTreeList.length; k++){
		if(Seamless* smls=dynamic_cast<Seamless*>(g_sceneTreeList[k].node.p)){
			return smls;
		}
	}
	Seamless* smls = new Seamless;
	ColorEffect* colorEffect=new ColorEffect;
	smls->effect=colorEffect;
	g_scene->children+=smls;
	return smls;
}
int getNumOfDescendents(const int line){
	int indent = g_sceneTreeList[line].indent;

	int k = line;
	while(++k < g_sceneTreeList.length && indent < g_sceneTreeList[k].indent);
	
	return k-line-1;
}
int getFirstBuildNodeLine(const int line){//line = seamless
	int indent = g_sceneTreeList[line].indent;

	int k = line+1;
	for(; k < g_sceneTreeList.length && indent < g_sceneTreeList[k].indent; k++){
		
		Node* node = g_sceneTreeList[k].node.p;

//		if(dynamic_cast<Build*>(node)){
//			break;
//		}
		if(
			dynamic_cast<Effect*>(node)==0 && //if not a part or effect its the line we want
			dynamic_cast<Art*>(node)==0
			){
			break;
		}
	}
	return k;
}
int getLastBuildNodeLine(const int line){ //line == first build node line
	int indent=g_sceneTreeList[line].indent;
	int k = line+1;
	while(true){
		if(k >= g_sceneTreeList.length || indent > g_sceneTreeList[k].indent){
			break;
		}
		if(indent==g_sceneTreeList[k].indent){
			if(dynamic_cast<Build*>(g_sceneTreeList[k].node.p)==0){
				break;
			}
		}
		k++;
	}
	
	return k-1;
}
bool searchForFirstExitBuildLine(int& line, const int lastLine){
	int indent = g_sceneTreeList[line].indent;

	for(int k=line;k<=lastLine; k++){
		if(ExitBuild* exitBuild = dynamic_cast<ExitBuild*>(g_sceneTreeList[k].node.p)){
			if(exitBuild->enabled){
				line = k;
				return true;
			}
		}
	}
	return false;
}
int getDeepestOpenGroup(Seamless* smls, const int smlsIndex, const int firstLine, const int lastLine){
	int deepest=firstLine;
	for(int k=firstLine; k<=lastLine; k++){
		if(GroupBuild* groupBuild=dynamic_cast<GroupBuild*>(g_sceneTreeList[k].node.p)){
			if(groupBuild->collapsed==false){
				if(g_sceneTreeList[k].indent>=g_sceneTreeList[deepest].indent){
					if(groupBuild->collapsed==false){
						int dummy=0;
						if(getCollapsedLine(k, dummy)){
							deepest=k;
						}
					}
				}
			}
		}
	}
	return deepest;
}
void insertNodeAtExitOrEnd(MFNode& mfbuild, Build* buildNode, int buildLine){
	MFNode* m=&mfbuild;
	M_Node* m_build=&mfbuild.refValue();
	insertNodeAbove(buildLine, *m_build, buildNode);
	g_stw->selLine=buildLine;
}

void createNewBuildNode(Build* buildNode, Art* part){
	if(part->name==""){
		createUniqueName(part);
	}
	Seamless* smls=part->smls;

	{
		int smlsLine=getNodeLine(smls);
		int firstLine=getFirstBuildNodeLine(smlsLine);
		if(smls->build.refValue().length){
			int lastLine=getLastBuildNodeLine(firstLine);
			int numOf=lastLine-firstLine+1;

			int buildLine=firstLine;

			if(searchForFirstExitBuildLine(buildLine, lastLine)){
				int parentLine=getParent(buildLine);
				Node* parentNode= g_sceneTreeList[parentLine].node.p;
				int fieldIdx=0, element=0;

				getFNodeElementForParent(parentLine, buildLine, g_sceneTreeList, fieldIdx, element);

				Field* field=parentNode->getField(fieldIdx);
				MFNode& mfnode=*dynamic_cast<MFNode*>(field);
				M_Node& m_node=mfnode.refValue();
				insertNodeAbove(element, m_node, buildNode);
				g_stw->selLine=buildLine;
			}
			else{
				buildLine=getDeepestOpenGroup(smls, smlsLine, firstLine, lastLine);
				GroupBuild* group=dynamic_cast<GroupBuild*>(g_sceneTreeList[buildLine].node.p);
				if(group && group->collapsed == false){
					group->build.refValue()+=buildNode;
					int numOfDescendents = getNumOfDescendents(buildLine);
					g_stw->selLine=buildLine+1+numOfDescendents;
				}
				else{
					smls->build.refValue()+=buildNode;
					g_stw->selLine=lastLine+1;
				}
			}
		}
		else{
			smls->build.refValue()+=buildNode;
			g_stw->selLine=firstLine;
		}
	}
	g_stw->genNodeList();
	smls->process();
	buildNode->processFields();
	g_stw->textCell.cellStr="";
	g_stw->textCell.cellActive=false;
	g_stw->cp->setupNodeWindow(buildNode);
	g_stw->paint();
	updateBuild3dWindow();
}
void checkAllAreNamedAndFixIfNot(){
	for(int i=0; i<g_selCount.all; i++){
		Node* part = g_sceneTreeList[g_stw->selLine+i].node.p;
		if(part->name==""){
			createUniqueName(part);
		}
	}
}
void copyPartForSelParts(){
	if(Part* part=dynamic_cast<Part*>(g_selNode.getRaw())){
		if(g_selCount.all==1){
			msG(msg_num_of_sel_must_be_more_than_one);
			return;
		}
		if(g_selCount.all % 2){
			msG(msg_num_of_sel_must_be_even);
			return;
		}
		g_stw->textCell.cellActive=false;

		checkAllAreNamedAndFixIfNot();


		CopyPart* cp=new CopyPart;
		for(int i=0; i<g_selCount.all/2; i++){
			cp->source.refValue()+=dynamic_cast<Art*>(g_sceneTreeList[g_stw->selLine+i].node.p);
			cp->dest.refValue()+=dynamic_cast<Art*>(g_sceneTreeList[g_selCount.all/2+g_stw->selLine+i].node.p);
		}
		g_selCount.reset();

		createNewBuildNode(cp, part);
	}
	else{
		needWhitePartMsg();
	}
}
void colorSweepForSelParts(){
	if(Part* part=dynamic_cast<Part*>(g_selNode.getRaw())){
		g_stw->textCell.cellActive=false;

		checkAllAreNamedAndFixIfNot();


		ColorSweep* cs=new ColorSweep;
		for(int i=0; i<g_selCount.all; i++){
			cs->part.refValue()+=dynamic_cast<Art*>(g_sceneTreeList[g_stw->selLine+i].node.p);
		}
		g_selCount.reset();

		createNewBuildNode(cs, part);
		#ifdef PROG_ROB
			dLog("new ClickButton(colorSweepForSelParts),");
		#endif

	}
	else{
		needWhitePartMsg();
	}
}
void placeForSelParts(){
	if(Part* part=dynamic_cast<Part*>(g_selNode.getRaw())){
		g_stw->textCell.cellActive=false;

		checkAllAreNamedAndFixIfNot();


		Place* place=new Place;
		for(int i=0; i<g_selCount.all; i++){
			place->part.refValue()+=dynamic_cast<Art*>(g_sceneTreeList[g_stw->selLine+i].node.p);
		}
		g_selCount.reset();

		createNewBuildNode(place, part);
	}
	else{
		needWhitePartMsg();
	}
}
void averagePivotPoint(){
	if(Art* part=dynamic_cast<Art*>(g_selNode.getRaw())){
		M_Vec3f& coord = part->coord.refValue();
		if(coord.length){
			Vec3f v;
			if(MultiSelect::coordList.length){
				for(int i = MultiSelect::coordList.length; --i>=0;){
					v+=coord[MultiSelect::coordList[i].i];
				}
				v/=MultiSelect::coordList.length;
			}
			else{
				for(int i=coord.length; --i>=0;){
					v+=coord[i];
				}
				v/=coord.length;
			}
			if(part->mirror && part->mirrorPart == 0){
				v.x = 0;
			}
			Vec3f parentAbsolutePivotPoint;
			int selLine = g_stw->selLine;
			int parentLine=getParent(selLine);
			Node* parentNode= g_sceneTreeList[parentLine].node.p;
			if(Art* parentPart = dynamic_cast<Art*>(parentNode)){
				parentAbsolutePivotPoint = parentPart->absolutePivotPoint;
			}

			part->pivotPoint=v-parentAbsolutePivotPoint;
			commonEndPivotPointChange();
		}
	}
}

void Art::PivotPoint::operator=(const Vec3f& par){
	Vec3f _par = par;
	Vec3f store = refValue();
	int numOfParts = 0;
	if(home.mirror){
		if(Art* mirrorPart = home.mirrorPart){
			Vec3f mirrorPar = par;
			mirrorPar.x = -mirrorPar.x;
			mirrorPart->pivotPoint.refValue() = mirrorPar;
		}
		else if(fabs(home.absolutePivotPoint.x) < .000001){
			if(home.isLeftRightName() == false){
				_par.x = 0;
			}
		}
	}
	refValue() = _par;

	if(0x8000&GetKeyState(VK_LMENU)){
		Vec3f diff = store - refValue();
		M_SArt& m_part = home.part.refValue();
		for(int k = m_part.length; --k>=0;){
			m_part[k]->pivotPoint.refValue()+=diff;
		}
		if(Art* mirrorPart = home.mirrorPart){
			M_SArt& m_part = mirrorPart->part.refValue();
			for(int k = m_part.length; --k>=0;){
				m_part[k]->pivotPoint.refValue()-=diff;
			}
		}
	}

	M_SArt& skeleton=home.smls->skeleton.refValue();

	for(int k=0; k<skeleton.length; k++){
		Art* part=skeleton[k];
		part->calcAbsolutePivotPoint(Vec3f());
	}
	home.smls->ctrf=true;
}
Art* getPartForPivotPoint(){
	Node* node = g_selNode.getRaw();
	if(Art* part = dynamic_cast<Art*>(node)){
		return part;
	}
	else if(TransferBCyl* transferBCyl = dynamic_cast<TransferBCyl*>(node)){
		return transferBCyl->dest;
	}
	else if(TransferOwnership* transferOwnership = dynamic_cast<TransferOwnership*>(node)){
		return transferOwnership->getSelectedPart();
	}
	else if(CenterJoint* centerJoint = dynamic_cast<CenterJoint*>(node)){
		return centerJoint->part;
	}
	needPartOrCenterJointMsg();
	return 0;
}
void decPivotX(){//DEC_PIVOT_X:
	if(Art* part = getPartForPivotPoint()){
		Vec3f pivotPoint = part->pivotPoint;

		quantizeDec(pivotPoint.x);

		part->pivotPoint = pivotPoint;
		commonEndPivotPointChange();
	}
}
void incPivotX(){//INC_PIVOT_X:
	if(Art* part = getPartForPivotPoint()){
		Vec3f pivotPoint = part->pivotPoint;

		quantizeInc(pivotPoint.x);

		part->pivotPoint = pivotPoint;
		commonEndPivotPointChange();
	}
}
void decPivotY(){//DEC_PIVOT_Y:
	if(Art* part = getPartForPivotPoint()){
		Vec3f pivotPoint = part->pivotPoint;

		quantizeDec(pivotPoint.y);

		part->pivotPoint = pivotPoint;
		commonEndPivotPointChange();
	}
}
void incPivotY(){//INC_PIVOT_Y:
	if(Art* part = getPartForPivotPoint()){
		Vec3f pivotPoint = part->pivotPoint;

		quantizeInc(pivotPoint.y);

		part->pivotPoint = pivotPoint;
		commonEndPivotPointChange();
	}
}
void decPivotZ(){//DEC_PIVOT_Z:
	if(Art* part = getPartForPivotPoint()){
		Vec3f pivotPoint = part->pivotPoint;

		quantizeDec(pivotPoint.z);

		part->pivotPoint = pivotPoint;
		commonEndPivotPointChange();
	}
}
void incPivotZ(){//INC_PIVOT_Z:
	if(Art* part = getPartForPivotPoint()){
		Vec3f pivotPoint = part->pivotPoint;

		quantizeInc(pivotPoint.z);

		part->pivotPoint = pivotPoint;
		commonEndPivotPointChange();
	}
}
void swapDivide(ImmediatePart* part, Art* actTriPart, const int ActTri){
		fastFullBuildForAutoIn(true);

#ifdef PROG_ROB
		dLog("new ClickButton(swapDivide),");
#endif

	Seamless* smls=part->smls;
	
	int idxRed = ActTri * 3 + smls->ActTriCorner;
	int idxGreen = ActTri * 3 + (smls->ActTriCorner + 1) % 3;
	int notConectedA = (smls->ActTriCorner + 2) % 3;
	const int cRed = actTriPart->triangle.refIndex(idxRed);
	const int cGreen = actTriPart->triangle.refIndex(idxGreen);
	Art* red = actTriPart->triangle.refPart(idxRed);
	Art* green = actTriPart->triangle.refPart(idxGreen);
	int otherTriIndex=0;
	int notConectedB = 0;
	

	if(Art* otherTriPart = findTriangleJoinedBy2Vertices(actTriPart, ActTri*3, otherTriIndex, red, cRed, green, cGreen, notConectedB)){
		Seamless* smls=part->smls;

		BackUp::monostable.beginCycle();
		SwapTriDivide swapTriDivide;
		swapTriDivide.corner.refValue().length=0;
		swapTriDivide.part=actTriPart;
		M_Int& corner=swapTriDivide.corner.refValue();
		corner+=getIndexRelativeToPart(actTriPart, otherTriPart, otherTriIndex*3+notConectedB);
		corner+=ActTri*3+notConectedA;
		swapTriDivide.doBuild();
	}
}
void swapDivide(ImmediatePart* part, Art* actTriPart, const int ActTri, const int ActTriCorner){
		fastFullBuildForAutoIn(true);

#ifdef PROG_ROB
		dLog("new ClickButton(swapDivide),");
#endif

	Seamless* smls=part->smls;
	
	int idxRed = ActTri * 3 + ActTriCorner;
	int idxGreen = ActTri * 3 + (ActTriCorner + 1) % 3;
	int notConectedA = (ActTriCorner + 2) % 3;
	const int cRed = actTriPart->triangle.refIndex(idxRed);
	const int cGreen = actTriPart->triangle.refIndex(idxGreen);
	Art* red = actTriPart->triangle.refPart(idxRed);
	Art* green = actTriPart->triangle.refPart(idxGreen);
	int otherTriIndex=0;
	int notConectedB = 0;
	

	if(Art* otherTriPart = findTriangleJoinedBy2Vertices(actTriPart, ActTri*3, otherTriIndex, red, cRed, green, cGreen, notConectedB)){
		Seamless* smls=part->smls;

		BackUp::monostable.beginCycle();
		SwapTriDivide swapTriDivide;
		swapTriDivide.corner.refValue().length=0;
		swapTriDivide.part=actTriPart;
		M_Int& corner=swapTriDivide.corner.refValue();
		corner+=getIndexRelativeToPart(actTriPart, otherTriPart, otherTriIndex*3+notConectedB);
		corner+=ActTri*3+notConectedA;
		swapTriDivide.doBuild();
	}
}

void swapDivide(){
	if(Part* whitePart=dynamic_cast<Part*>(g_selNode.getRaw())){
		Seamless* smls = whitePart->smls;
		NurbsPatch* patch = 0;
		Stem* stem = 0;
		Polymap* polymap = 0;
		ImmediatePart* bluePart = 0;
		if(getPolyMap(patch, stem, polymap, whitePart, smls->ActTri, bluePart)){
			if(bluePart){
				swapDivide(bluePart, bluePart, smls->ActTri-stem->beginTriOffset);
				BackUp::monostable.beginCycle();
				commonEndVertexBuild();
			}
			else{
				int ActTri = smls->ActTri;
				int ActTriCorner = smls->ActTriCorner;
				if(stem->mirror){
					if(	(stem->mirrorStem && stem->insideOut) ||
						(stem->mirrorStem == 0 && ActTri >= stem->coordIndex.length / 3)
						){
						ActTriCorner = abs(ActTriCorner + 1) % 3;
					}
				}
				if(0 > getRelIndex(stem, ActTri, ActTriCorner)){
					return;
				}

				bluePart = selectPolyMapPart(whitePart);
				stem->copyPolyMapToPart(bluePart);
				bluePart->smls->ActTriCorner = ActTriCorner;
				swapDivide(bluePart, bluePart, ActTri / 3);
				stem->copyPartToPolyMap(bluePart);

				BackUp::monostable.beginCycle();
				commonEndVertexBuild();
			}
		}
	}
	else if(ImmediatePart* bluePart=dynamic_cast<ImmediatePart*>(g_selNode.getRaw())){
		Seamless* smls = bluePart->smls;
		swapDivide(bluePart, smls->actTriPart, smls->ActTri);
		commonEndVertexChange();
	}
	else{
		needPartMsg();
	}
}
void forceToMirrorX(){
	if(ImmediatePart* part=dynamic_cast<ImmediatePart*>(g_selNode.getRaw())){
		if(dynamic_cast<ImmediatePart*>(g_copyNode.getRaw())){
			Seamless* smls=part->smls;
			BackUp::monostable.beginCycle();
			M_Vec3f& coord=part->coord.refValue();

			Vec3f svec = remPJ->coord.refValue()[smls->remActCtrlPnt];
			svec.x=-svec.x;
			coord[smls->ActCtrlPnt]=svec;

			makePointerEqualIndexLocation();
			smls->convertToRenderFormat();
		}
		else{
			needPartCopiedMsg();
		}
	}
	else{
		needBluePartMsg();
	}
}
void openNodeWindow(){ //OPEN_NODE_WINDOW
	throw;
}
void decTugRadius(){ //DEC_RADIUS
	float& tugRadius=g_settings->tugRadius.refValue();
	tugRadius-=getIncDecRes();
	if(tugRadius<=0){
		tugRadius=0;
		g_vcArtTriBase->circle=false;
	}
	houseKeepVisualClue();
}
void incTugRadius(){ //INC_RADIUS
	float& tugRadius=g_settings->tugRadius.refValue();
	tugRadius+=getIncDecRes();
	houseKeepVisualClue();
}
void setZDragNode(){
	g_zDragMode=true;
};
void quickTriangle(Art* copyPart, ImmediatePart* part, Art* remActTriPart, Art* actTriPart, int remActTri, int ActTri){
#ifdef PROG_ROB
			dLog("new ClickButton(quickTriangle),");
#endif
	SeamlessPtr smls=part->smls;
	const int remActTriCorner = smls->remActTriCorner;
	const int ActTriCorner = smls->ActTriCorner;

	fastFullBuildForAutoIn(true);


//		triVertexRem.refPart(0)=triVertexRem.refPart(1);
//		triVertexRem.refPart(1)=triVertexRem.refPart(2);
	triVertexRem.refPart(0)=remActTriPart;
	triVertexRem.refPart(1)=actTriPart;
	triVertexRem.refPart(2)=actTriPart;
//		triVertexRem.refIndex(0)=triVertexRem.refIndex(1);
//		triVertexRem.refIndex(1)=triVertexRem.refIndex(2);
	triVertexRem.refIndex(0)=remActTri*3+remActTriCorner;
	triVertexRem.refIndex(1)=ActTri*3+ActTriCorner;
	triVertexRem.refIndex(2)=ActTri*3+(ActTriCorner+1)%3;

	//find if anti-clockwise
	{
		int idxA = triVertexRem.refIndex(0);
		int idxB = triVertexRem.refIndex(1);
		int idxC = triVertexRem.refIndex(2);
		Art* partA = triVertexRem.refPart(0)->triangle.refPart(idxA);
		Art* partB = triVertexRem.refPart(1)->triangle.refPart(idxB);
		Art* partC = triVertexRem.refPart(2)->triangle.refPart(idxC);

		Vec3f v[3]={
			partA->matrix * g_viewMatrix * partA->coord[triVertexRem.refPart(0)->triangle.refIndex(idxA)],
			partB->matrix * g_viewMatrix * partB->coord[triVertexRem.refPart(1)->triangle.refIndex(idxB)],
			partC->matrix * g_viewMatrix * partC->coord[triVertexRem.refPart(2)->triangle.refIndex(idxC)]
		};
		if(backFace(v[0], v[1], v[2]) == false){
			Art* storeTriPart = triVertexRem.refPart(2);
			triVertexRem.refPart(2)=triVertexRem.refPart(1);
			triVertexRem.refPart(1)=storeTriPart;

			int storeIndex = triVertexRem.refIndex(2);
			triVertexRem.refIndex(2)=triVertexRem.refIndex(1);
			triVertexRem.refIndex(1)=storeIndex;
		}
	}


	BackUp::monostable.beginCycle();

	if(smls->build.length&&g_settings->record==true){
		M_Node* build;
		int bidx=smls->getIndexForLastBuildNode(build);
		if(bidx==-1)return;

		GenPoly* genPoly = dynamic_cast<GenPoly*>((*build)[bidx].p);

		if(genPoly==0||genPoly->part!=0&&(genPoly->part.refValue()!=triVertexRem.pa||genPoly->part.refValue()!=triVertexRem.pb||genPoly->part.refValue()!=triVertexRem.pc)){
			genPoly=new GenPoly;
			genPoly->part=part;
			insertNodeBelow(bidx, *build, genPoly);
			g_stw->genNodeList();
		}
		else if(genPoly->part==0){
			genPoly->vertex.refValue().length=0;
			genPoly->part=part;
			g_stw->genNodeList();
		}
		Art* part=genPoly->part.refValue();

		genPoly->vertex.refValue()+=getIndexRelativeToPart(part, triVertexRem.pa, triVertexRem.ia);
		genPoly->vertex.refValue()+=getIndexRelativeToPart(part, triVertexRem.pb, triVertexRem.ib);
		genPoly->vertex.refValue()+=getIndexRelativeToPart(part, triVertexRem.pc, triVertexRem.ic);

	}
	else {
		GenPoly gp;
		gp.independentColor = true;
		gp.part=part;

		gp.vertex.refValue()+=getIndexRelativeToPart(part, triVertexRem.pa, triVertexRem.ia);
		gp.vertex.refValue()+=getIndexRelativeToPart(part, triVertexRem.pb, triVertexRem.ib);
		gp.vertex.refValue()+=getIndexRelativeToPart(part, triVertexRem.pc, triVertexRem.ic);

		gp.doBuild();
	}
}
void quickTriangle(){
	if(Part* whitePart=dynamic_cast<Part*>(g_selNode.getRaw())){
		openSceneTree();
		if(g_copyNode.getRaw()==whitePart){
			Seamless* smls = whitePart->smls;
			NurbsPatch* patch = 0;
			Stem* selStem = 0;
			Stem* copyStem = 0;
			Polymap* polymap = 0;
			ImmediatePart* bluePart = 0;
			if(getPolyMap(patch, selStem, polymap, whitePart, smls->ActTri, bluePart)){
				if(getPolyMap(patch, copyStem, polymap, whitePart, smls->remActTri, bluePart)){
					if(bluePart){
						quickTriangle(bluePart, bluePart, bluePart, bluePart, smls->remActTri-copyStem->beginTriOffset, smls->ActTri-selStem->beginTriOffset);
						commonEndVertexBuild();
					}
					else{
						bluePart = selectPolyMapPart(whitePart);
						selStem->copyPolyMapToPart(bluePart);
						bluePart->smls->remActTriCorner = smls->remActTriCorner;
						bluePart->smls->ActTriCorner = smls->ActTriCorner;
						quickTriangle(bluePart, bluePart, bluePart, bluePart, smls->remActTri-copyStem->beginTriOffset, smls->ActTri-selStem->beginTriOffset);
						selStem->copyPartToPolyMap(bluePart);
						commonEndVertexBuild();
					}
				}
			}
		}
	}
	else if(ImmediatePart* bluePart=dynamic_cast<ImmediatePart*>(g_selNode.getRaw())){
		if(Art* copyPart=dynamic_cast<Art*>(g_copyNode.getRaw())){
			Seamless* smls = bluePart->smls;
			quickTriangle(copyPart, bluePart, smls->remActTriPart , smls->actTriPart, smls->remActTri, smls->ActTri);
			commonEndVertexBuild();
		}
		else{
			needPartCopiedMsg();
		}
	}
	else{
		needPartMsg();
	}
}


void buildSphereForSelPart(){
#ifdef PROG_ROB
	dLog("new ClickButton(buildSphereForSelPart),");
#endif
	if(Part* part=dynamic_cast<Part*>(g_selNode.getRaw())){
		openSceneTree();

		Vec3f from = part->absolutePivotPoint;
		Vec3f to;
		if(part->part.length){
			to = part->part.refValue()[0]->pivotPoint;
		}
		if(to.length() < .0001){
			to = Vec3f(0, .25, 0);
		}
		if(SphereBuild* sphereBuild = part->sphereFromTwoPoints(from, to)){
			if(dynamic_cast<TextureEffect*>(part->workEffect)){
				sphereBuild->textureMapper = new TextureMapper;
			}
			createNewBuildNode(sphereBuild, part);
			if(Art* mirrorPart = part->mirrorPart){
				from.x = -from.x;
				to.x = -to.x;
				if(SphereBuild* sphereBuild = mirrorPart->sphereFromTwoPoints(from, to)){
					if(dynamic_cast<TextureEffect*>(mirrorPart->workEffect)){
						sphereBuild->textureMapper = new TextureMapper;
					}
					createNewBuildNode(sphereBuild, mirrorPart);
				}
			}
		}
	}
	else{
		needWhitePartMsg();
	}
}

void newRectangle(){
#ifdef PROG_ROB
	dLog("new ClickButton(newRectangle),");
#endif
//	Part* part=new Part;
	if(Part* part=dynamic_cast<Part*>(g_selNode.getRaw())){
		openSceneTree();
		RectangleBuild* rectangleBuild= new RectangleBuild;
		rectangleBuild->part=part;
		if(dynamic_cast<TextureEffect*>(part->workEffect)){
			rectangleBuild->textureMapper = new TextureMapper;
		}
		createNewBuildNode(rectangleBuild, part);
	}
	else{
		needWhitePartMsg();
	}
}
void buildConeForSelPart(){
	//if(Place* place=dynamic_cast<Place*>(g_selNode.getRaw())){
	//	quantiseRotation(place->rotation);
	//	updateBuild3dWindow();
	//}
	//if(Art* part=dynamic_cast<Art*>(g_selNode.getRaw())){
	//	Vec3f axis = (-g_scene->rotation * -Rotation(1, 0, 0, .698132) * Vec3f(0, 0, 1)).normalize();
	//	part->rotation= part->rotation.refValue() * Rotation(axis, PI*.01);
	//	part->smls->convertToRenderFormat();
	//	return;
	//}
	if(Part* part=dynamic_cast<Part*>(g_selNode.getRaw())){
		openSceneTree();
		ConeBuild* coneBuild= new ConeBuild;
		coneBuild->part=part;
		if(dynamic_cast<TextureEffect*>(part->workEffect)){
			coneBuild->textureMapper = new TextureMapper;
		}
		createNewBuildNode(coneBuild, part);
	}
	else{
		needWhitePartMsg();
	}
}
void buildTorusForSelPart(){
#ifdef PROG_ROB
	dLog("new ClickButton(buildTorusForSelPart),");
#endif
	if(Part* part=dynamic_cast<Part*>(g_selNode.getRaw())){
		openSceneTree();
		TorusBuild* torusBuild= new TorusBuild;
		torusBuild->part=part;
		if(dynamic_cast<TextureEffect*>(part->workEffect)){
			torusBuild->textureMapper = new TextureMapper;
		}
		createNewBuildNode(torusBuild, part);
	}
	else{
		needWhitePartMsg();
	}
}
void buildBezierLatheForSelPart(){
	if(Part* part=dynamic_cast<Part*>(g_selNode.getRaw())){
		openSceneTree();
		BezierLatheBuild* bezierLatheBuild= new BezierLatheBuild;
		bezierLatheBuild->point.refValue()+=Vec2f(.25, 0);
		bezierLatheBuild->point.refValue()+=Vec2f(.5, .1);
		bezierLatheBuild->point.refValue()+=Vec2f(.5, .5);
		bezierLatheBuild->point.refValue()+=Vec2f(0, .3);
		bezierLatheBuild->point.refValue()+=Vec2f(0, .75);
		bezierLatheBuild->point.refValue()+=Vec2f(.15, .7);
		bezierLatheBuild->point.refValue()+=Vec2f(.34, .8);
		bezierLatheBuild->point.refValue()+=Vec2f(.25, 1);
		bezierLatheBuild->fix.refValue()+=true;
		bezierLatheBuild->fix.refValue()+=false;
		bezierLatheBuild->fix.refValue()+=false;
		bezierLatheBuild->fix.refValue()+=false;
		bezierLatheBuild->fix.refValue()+=false;
		bezierLatheBuild->fix.refValue()+=false;
		bezierLatheBuild->fix.refValue()+=false;
		bezierLatheBuild->fix.refValue()+=true;
		bezierLatheBuild->part=part;
		if(dynamic_cast<TextureEffect*>(part->workEffect)){
			bezierLatheBuild->textureMapper = new TextureMapper;
		}
		createNewBuildNode(bezierLatheBuild, part);
	}
	else{
		needWhitePartMsg();
	}
}
void buildNurbsLatheForSelPart(){
	if(Part* part=dynamic_cast<Part*>(g_selNode.getRaw())){
		openSceneTree();
		NurbsLathe* nurbsLathe= new NurbsLathe;
		nurbsLathe->beginSP.refValue() = false;
		nurbsLathe->endSP.refValue() = false;
		nurbsLathe->bands.refValue() = 25;
		if(part->isRightMirror()){
			nurbsLathe->mirrorPoints();
		}
		nurbsLathe->part = part;
		if(dynamic_cast<TextureEffect*>(part->workEffect)){
			nurbsLathe->textureMapper = new TextureMapper;
		}
		createNewBuildNode(nurbsLathe, part);
	}
	else{
		needWhitePartMsg();
	}
}

void buildAvSegmentForSelPart(){
	if(Part* part=dynamic_cast<Part*>(g_selNode.getRaw())){
		openSceneTree();

		Vec3f from = part->absolutePivotPoint;
		Vec3f to;
		if(part->part.length){
			to = part->part.refValue()[0]->pivotPoint;
		}


		NurbsLathe* nurbsLathe = part->latheFromTwoPoints(from, to);
		if(nurbsLathe == 0){
			nurbsLathe = new NurbsLathe;
			nurbsLathe->part.refValue() = part;
			nurbsLathe->point.refValue() = M_Vec2f(Vec2f(0, 0), Vec2f(.05, 0), Vec2f(.05, .15), Vec2f(.05, .3), Vec2f(0, .3));
			nurbsLathe->weight.refValue() = M_Float(1, 1, 1, 1, 1);
			nurbsLathe->knot.refValue() = M_Float(0, 0, 0, 1, 2, 3, 3, 3);
			nurbsLathe->translation.refValue() = from;
			if(part->isRightMirror()){
				nurbsLathe->mirrorPoints();
			}
			nurbsLathe->processFields();
		}

		if(dynamic_cast<TextureEffect*>(part->workEffect)){
			nurbsLathe->textureMapper = new TextureMapper;
		}
		createNewBuildNode(nurbsLathe, part);
	}
	else{
		needWhitePartMsg();
	}
}

void buildModelMorpherForSelSmls(){
	openSceneTree();
	if(Seamless* smls=dynamic_cast<Seamless*>(g_selNode.getRaw())){
		buildModelMorpherForSelSmls(smls);
	}
	else if(dynamic_cast<Art*>(g_selNode.getRaw())||dynamic_cast<Effect*>(g_selNode.getRaw())||dynamic_cast<Build*>(g_selNode.getRaw())||dynamic_cast<Generator*>(g_selNode.getRaw())){
		int selLine = g_stw->selLine;
		if(Seamless* smls=getSeamless(selLine)){
			g_stw->selLine=selLine;
			buildModelMorpherForSelSmls(smls);
		}
	}
	else{
		needPartOrSmlsMsg();
	}
}
void buildBendyLatheForSelPart(){
	if(Part* part=dynamic_cast<Part*>(g_selNode.getRaw())){
		openSceneTree();
		BendyLathe* bendyLathe= new BendyLathe;
		bendyLathe->bands.refValue()=25;
		bendyLathe->part=part;
		if(dynamic_cast<TextureEffect*>(part->workEffect)){
			bendyLathe->textureMapper = new TextureMapper;
		}
		createNewBuildNode(bendyLathe, part);
	}
	else{
		needWhitePartMsg();
	}
}
void buildSubdivisionForSelPart(){
	openSceneTree();
	if(Part* part=dynamic_cast<Part*>(g_selNode.getRaw())){
		Subdivision* subdivision = new Subdivision;
		subdivision->part=part;
		createNewBuildNode(subdivision, part);
	}
	else{
		needWhitePartMsg();
	}
}
void buildRingsForSelPart(){
#ifdef PROG_ROB
	dLog("new ClickButton(buildRingsForSelPart),");
#endif
	
	if(Part* part=dynamic_cast<Part*>(g_selNode.getRaw())){
		openSceneTree();
		RingBuild* ringBuild = new RingBuild;
		ringBuild->scaleX=M_Vec2f(Vec2f(.25, .25), Vec2f(.25, .25));
		ringBuild->scaleZ=M_Vec2f(Vec2f(.25, .25), Vec2f(.25, .25));
		ringBuild->orientation=M_Rotation(Rotation(0,0,1,0), Rotation(0,0,1,0));
		ringBuild->location=M_Vec3f(Vec3f(), Vec3f(0, .5, 0));
		ringBuild->part.refValue()+=part;
		createNewBuildNode(ringBuild, part);
	}
	else{
		needWhitePartMsg();
	}
}

void buildBoxForSelPart(){
#ifdef PROG_ROB
	dLog("new ClickButton(buildBoxForSelPart),");
#endif
	if(Part* part=dynamic_cast<Part*>(g_selNode.getRaw())){
		openSceneTree();
		BoxBuild* boxBuild= new BoxBuild;
		boxBuild->part=part;
		if(dynamic_cast<TextureEffect*>(part->workEffect)){
			for(int k=6; --k>=0;){
				boxBuild->textureMapper.refValue() += new TextureMapper;
			}
		}
		createNewBuildNode(boxBuild, part);
	}
	else{
		needWhitePartMsg();
	}
}

void buildCylinderForSelPart(){
#ifdef PROG_ROB
	dLog("new ClickButton(buildCylinderForSelPart),");
#endif
	//Part* part=new Part;
	if(Part* part=dynamic_cast<Part*>(g_selNode.getRaw())){
		openSceneTree();
		CylinderBuild* cylinderBuild= new CylinderBuild;
		cylinderBuild->part=part;
		if(dynamic_cast<TextureEffect*>(part->workEffect)){
			cylinderBuild->textureMapper = new TextureMapper;
		}
		createNewBuildNode(cylinderBuild, part);
	}
	else{
		needWhitePartMsg();
	}
/*
	if(Art* part=dynamic_cast<Art*>(g_selNode.getRaw())){
		buildCylinderForSelPart(part);
		BackUp::monostable.beginCycle();
		g_pw->setupNodeWindow(g_immediateCylinder);
		Window::activeWindow = g_pw;
	}
	else{
		needPartMsg();
	}
	*/
}
void SceneTree::decMultipleNodeSelection(){
	if(g_selCount>1){
		int collapsedLine;
		if(getCollapsedLine(g_stw->selLine, collapsedLine)){
			g_selCount--;
			if(g_selCount.all>g_stw->numOfVisLines){
				g_scene->scroll=collapsedLine+g_selCount.all-1;
			}
			g_stw->setScrolPos(0);
		}
	}
}
void SceneTree::incMultipleNodeSelection(){
	int collapsedLine;
	if(getCollapsedLine(g_stw->selLine, collapsedLine)){
		g_stw->textCell.cellActive=false;

		g_selCount++;
		int s = (collapsedLine+g_selCount.all-g_scene->scroll)*pixSymbolCY+beginYOffset+10;
		int amount=0;
		if( s > g_stw->size.y){
			amount=(s - g_stw->size.y) / pixSymbolCY+1;
		}
		g_stw->setScrolPos(amount);
	}
}
void hide(){
	Art* part=dynamic_cast<Art*>(g_selNode.getRaw());
	part->show=!part->show;
	part->smls->convertToRenderFormat();
	BackUp::monostable.beginCycle();
	g_stw->paint();
}
void hideAndDescendants(){
	Art* part=dynamic_cast<Art*>(g_selNode.getRaw());
	part->show = !part->show;
	recurseShow(part, part->show);
	part->smls->convertToRenderFormat();
	BackUp::monostable.beginCycle();
	g_stw->paint();
}

void recurseTrack(Art* part, bool track){
	part->track=track;
	M_SArt& m_art = part->part.refValue();
	for(int k=m_art.length; --k>=0;){
		recurseTrack(m_art[k], track);
	}
}

void setTrackAndDescendants(){
	if(Art* part = dynamic_cast<Art*>(g_selNode.getRaw())){
		part->track = !part->track;
		recurseTrack(part, part->track);
		g_stw->cp->paint();
	}
	else if(Seamless* smls = dynamic_cast<Seamless*>(g_selNode.getRaw())){
		smls->track = !smls->track;
		bool track = smls->track;
		M_Art& partList = smls->partList;
		for(int k = partList.length; --k>=0;){
			partList[k]->track = track;
		}
	}
	g_stw->paint();
}

void toggleTrack(){
	if(Art* part = dynamic_cast<Art*>(g_selNode.getRaw())){
		part->track = !part->track;
		g_stw->cp->paint();
	}
	else if(Seamless* smls = dynamic_cast<Seamless*>(g_selNode.getRaw())){
		smls->track = !smls->track;
	}
	else if(Viewpoint* view = dynamic_cast<Viewpoint*>(g_selNode.getRaw())){
		view->track = !view->track;
	}
	g_stw->paint();
}

void unselectAllTracks(){
	for(int k = g_sceneTreeList.length; --k >= 0;){
		if(Seamless* smls = dynamic_cast<Seamless*>(g_sceneTreeList[k].node.p)){
			smls->track = false;
		}
		else if(Viewpoint* view = dynamic_cast<Viewpoint*>(g_sceneTreeList[k].node.p)){
			view->track = false;
		}
	}
}

void selectTrackExclusive(){
	unselectAllTracks();
	if(Art* part = dynamic_cast<Art*>(g_selNode.getRaw())){
		part->smls->track = true;
		M_Art& partList = part->smls->partList;
		for(int i = partList.length; --i >= 0;){
			partList[i]->track = false;
		}
		part->track = true;
	}
	else if(Seamless* smls = dynamic_cast<Seamless*>(g_selNode.getRaw())){
		smls->track = true;
	}
	else if(Viewpoint* view = dynamic_cast<Viewpoint*>(g_selNode.getRaw())){
		view->track = true;
	}
	g_stw->paint();
}
void selectModelTracks(){
	Art* part = dynamic_cast<Art*>(g_selNode.getRaw());
	part->smls->track = true;
	M_Art& partList = part->smls->partList;
	for(int i = partList.length; --i >= 0;){
		partList[i]->track = true;
	}
	g_stw->paint();
}

void selectModelTracksExclusive(){
	unselectAllTracks();
	Art* part = dynamic_cast<Art*>(g_selNode.getRaw());
	part->smls->track = true;
	M_Art& partList = part->smls->partList;
	for(int i = partList.length; --i >= 0;){
		partList[i]->track = true;
	}
	g_stw->paint();
}

void selectSceneTracks(){
	for(int k = g_sceneTreeList.length; --k >= 0;){
		if(Seamless* smls = dynamic_cast<Seamless*>(g_sceneTreeList[k].node.p)){
			smls->track = true;
		}
		else if(Art* part = dynamic_cast<Art*>(g_sceneTreeList[k].node.p)){
			part->track = true;
		}
		else if(Viewpoint* view = dynamic_cast<Viewpoint*>(g_sceneTreeList[k].node.p)){
			view->track = true;
		}
	}
	g_stw->paint();
}

//////
void undoIs(){
	removeRouteFromProto(g_selNode.getRaw());
	updateBuild3dWindow();
}
void deleteRoute(int fieldIndex, int routeIndex){
	g_selNode->getField(fieldIndex)->refRoute().deleteElement(routeIndex);
	Window::reSetupAllControlPanels();
	updateBuild3dWindow();
}

void addIsField(int nodeFieldIndex, int protoFieldIndex){
	removeRouteFromProto(g_selNode.getRaw());
	Field* field=g_selNode->getField(nodeFieldIndex);
	ProtoInterface* protoInterface=getSceneProtoDeclare(g_stw->selLine)->protoInterface;
	field->connect(g_selNode.getRaw(), protoInterface, protoInterface->getField(protoFieldIndex));
	g_stw->cp->paint();
}
void elementBeforeCopyOf(){//ELEMENT_BEFORE_COPYOF
	insertElement(false, false, 0);
}
void elementBeforeMove(){//ELEMENT_BEFORE_MOVE:
	insertElement(true, false, 0);
}
void elementReplaceCopyOf(){//ELEMENT_REPLACE_COPYOF
	insertElement(false, true, 0);
}
void elementReplaceMove(){//ELEMENT_REPLACE_MOVE
	insertElement(true, true, 0);
}
void elementAfterCopyOf(){//ELEMENT_AFTER_COLoadCursorPYOF
	insertElement(false, false, 1);
}
void elementAfterMove(){//ELEMENT_AFTER_MOVE
	insertElement(true, false, 1);
}
void addRouteToField(int dstFieldIndex, int srcFieldIndex){

	S_Route( g_selNode.getRaw(), srcFieldIndex, g_copyNode.getRaw(), dstFieldIndex);
	bool paintSceneTree=false;
	if(g_selNode->name==""){
		createUniqueName(g_selNode.getRaw());
		paintSceneTree=true;
	}
	if(g_copyNode->name==""){
		createUniqueName(g_copyNode.getRaw());
		paintSceneTree=true;
	}
	if(paintSceneTree){
		g_stw->genNodeList();
		g_stw->paint();
	}
	setupSelectedNodeWindow();
}
void deleteElement(){//DELETE_ELEMENT
	const M_Properties& m_properties=g_selNode->getProperties();
	M_Int& mdstIndex=*(g_selNode->getIndexPtr());
	int dstNewSelElement=mdstIndex[g_menuGroupIndex];

	for(int fidx=0; fidx<m_properties.length; fidx++){
		const Properties& s_prop=m_properties[fidx];
		if(s_prop.groupIndex==g_menuGroupIndex){
			
			Field* dst=g_selNode->getField(fidx);
			int dstIndex=mdstIndex[g_menuGroupIndex]-s_prop.startAtIndex;

			switch(dst->getType()){
			case MFBOOL:{	MFBool& _dst=	*dynamic_cast<MFBool*>	(dst); _dst.refValue().deleteElement(dstIndex);} break;
			case MFCOLOR:{	MFColor& _dst=	*dynamic_cast<MFColor*>	(dst); _dst.refValue().deleteElement(dstIndex);} break;
			case MFFLOAT:{	MFFloat& _dst=	*dynamic_cast<MFFloat*>	(dst); _dst.refValue().deleteElement(dstIndex);} break;
			case MFIMAGE:{	MFImage& _dst=	*dynamic_cast<MFImage*>	(dst); _dst.refValue().deleteElement(dstIndex);} break;
			case MFINT32:{	MFInt32& _dst=	*dynamic_cast<MFInt32*>	(dst); _dst.refValue().deleteElement(dstIndex);} break;
			case MFROTATION:{ MFRotation& _dst=*dynamic_cast<MFRotation*>(dst); _dst.refValue().deleteElement(dstIndex);} break;
			case MFSTRING:{	MFString& _dst=	*dynamic_cast<MFString*>(dst); _dst.refValue().deleteElement(dstIndex);} break;
			case MFDOUBLE:{	MFTime& _dst=	*dynamic_cast<MFTime*>	(dst); _dst.refValue().deleteElement(dstIndex);} break;
			case MFVEC2F:{	MFVec2f& _dst=	*dynamic_cast<MFVec2f*>	(dst); _dst.refValue().deleteElement(dstIndex);} break;
			case MFVEC3F:{	MFVec3f& _dst=	*dynamic_cast<MFVec3f*>	(dst); _dst.refValue().deleteElement(dstIndex);} break;

			}
			if(dstNewSelElement>=dst->size()+s_prop.startAtIndex){
				dstNewSelElement=dst->size()+s_prop.startAtIndex-1;
				if(dstNewSelElement<0){
					dstNewSelElement=0;
				}
			}
		}
	}
	mdstIndex[g_menuGroupIndex]=dstNewSelElement;
	Window::paintAllControlPanels();
	updateBuild3dWindow();
}
void beamTo(){
	Transform& transform=*dynamic_cast<Transform*>(g_selNode.getRaw());
	Viewpoint viewpoint;
	viewpoint.position=transform.translation + Vec3f(0, 0, g_viewDist);
	g_scene->setSceneFromViewpoint(viewpoint);
}
void optimize(){
	do{
		g_optimized=false;
		optimize(dynamic_cast<AGroup*>(g_selNode.getRaw()));
		g_stw->genNodeList();
	}while(g_optimized);
	g_stw->paint();
}
void newPanel(){
	ControlPanel* controlPanel = new ControlPanel;
	controlPanel->setupNodeWindow(g_selNode.getRaw());
	controlPanel->pos=g_stw->cp->pos+Vec2i(50,50);
	Window::activeWindow=controlPanel;
}
void convertToWhitePart(Ptr<ImmediatePart> immediate, M_Art& accumOthers){
	Ptr<Part> whitePart = convertImmediateToPart(immediate);
	replaceNodeRefInSceneTree(immediate.p, whitePart.p);
	Seamless* smls=immediate->smls;
	smls->process();
	smls->resetRelevantParts();
//	smls->doBuild();
}
void convertToWhitePart(){
#ifdef PROG_ROB
	dLog("new MoveToNode(" + String(g_stw->selLine) + "),");
	dLog("new RightClick,");
	dLog("new Move_MenuItem(menu_convert),");
	dLog("new Click,");
#endif
	M_Art accumOthers;
	Ptr<ImmediatePart> immediate=dynamic_cast<ImmediatePart*>(g_selNode.getRaw());
	Seamless* smls = immediate->smls;
	Ptr<Part> part = convertImmediateToPart(immediate);
	convertToWhitePart(immediate, accumOthers);
	smls->doBuild();
	g_stw->genNodeList();
	g_stw->paint();
	g_stw->cp->setupNodeWindow(g_selNode.getRaw());

	g_stw->cp->setupNodeWindow(part);
	g_tb->updateToolBarContext();
	part->smls->doBuild();
}

void deleteNode(int nodeLine, int selCount = 1);

void convertToImmediatePart(Ptr<Part> part, M_Art& accumOthers){
#ifdef PROG_ROB
	dLog("new MoveToNode(" + String(g_stw->selLine) + "),");
	dLog("new RightClick,");
	dLog("new Move_MenuItem(menu_convert),");
	dLog("new Click,");
#endif
	MultiSelect::coordList.length = 0;
	MultiSelect::beginDragPos.length = 0;
	Ptr<ImmediatePart> immediate = convertPartToImmediate(part);
	replaceNodeRefInSceneTree(part.p, immediate.p);
	Seamless* smls=part->smls;
	smls->process();
	smls->resetRelevantParts();
	g_stw->genNodeList();
	g_stw->paint();
	g_stw->cp->setupNodeWindow(immediate);

	for(int k=g_sceneTreeList.length; --k>=0;){
		if(Build* build=dynamic_cast<Build*>(g_sceneTreeList[k].node.p)){
			if(build->illegalImmediateRef(immediate, accumOthers)){
				deleteNode(k);
			}
		}
	}
}

void convertToImmediatePart(){
	M_Art accumOthers;
	Ptr<Part> part=dynamic_cast<Part*>(g_selNode.getRaw());
	accumOthers+=part;
	while(accumOthers.length){
		if(Ptr<Part> part = dynamic_cast<Part*>(accumOthers[0])){
			convertToImmediatePart(part, accumOthers);
			accumOthers.deleteElement(0);
		}
	}
	g_stw->commonPostInsertNodeStuff();
	houseKeepVisualClue();
	g_tb->updateToolBarContext();
}
void convertAllToImmediateParts(){
	M_Art accumOthers;

	Ptr<Seamless> smls=dynamic_cast<Seamless*>(g_selNode.getRaw());

	M_Art& partList = smls->partList;

	smls->build.refValue().length=0;

	for(int k = partList.length; --k>=0;){
		if(Ptr<Part> part = dynamic_cast<Part*>(partList[k])){
			convertToImmediatePart(part, accumOthers);
		}
	}
	g_stw->commonPostInsertNodeStuff();
	houseKeepVisualClue();
	g_tb->updateToolBarContext();
}
void convertAllToWhiteParts(){
	M_Art accumOthers;

	Ptr<Seamless> smls=dynamic_cast<Seamless*>(g_selNode.getRaw());

	M_Art& partList = smls->partList;

	for(int k = partList.length; --k>=0;){
		if(Ptr<ImmediatePart> immediate = dynamic_cast<ImmediatePart*>(partList[k])){
			convertToWhitePart(immediate, accumOthers);
		}
	}
	smls->doBuild();
	smls->ctrf=true;
	g_stw->commonPostInsertNodeStuff();
	houseKeepVisualClue();
	g_tb->updateToolBarContext();
}
void fastFullBuild(bool inState){
	int smlsLine=getSeamlessParent(g_stw->selLine);
	int smlsIndent=g_sceneTreeList[smlsLine].indent;

	for(int k=g_stw->selLine; k<g_sceneTreeList.length && g_sceneTreeList[k].indent > smlsIndent; k++){
		Node* node=g_sceneTreeList[k].node.p;
		if(JoinVertex* joinVertex=dynamic_cast<JoinVertex*>(node)){
			joinVertex->in=inState;
		}
		else if(TransferBCyl* transferBCyl=dynamic_cast<TransferBCyl*>(node)){
			transferBCyl->in=inState;
		}
		else if(FuseSurface* fuseSurface=dynamic_cast<FuseSurface*>(node)){
			fuseSurface->in=inState;
		}
	}
	dynamic_cast<Seamless*>(g_sceneTreeList[smlsLine].node.p)->doBuild();

	houseKeepVisualClue();
	g_stw->paint();
	g_stw->cp->paint();
}
void setAllIns(M_Build& m_build, const bool inState);

void fastFullBuildForAutoIn(bool inState, Seamless* smls){
	if(smls->workAnimCP){
		return;
	}
	if(smls->autoIn && smls->in != inState){
		smls->in.refValue() = inState;
		setAllIns(smls->build.refValue(), inState);
	}
}
void fastFullBuildForAutoIn(bool inState){
	int selLine = g_stw->selLine;
	if(Seamless* smls = getSeamless(selLine)){
		if(smls->workAnimCP == false){
			fastFullBuildForAutoIn(inState, smls);
		}
	}
}
void updateAutoInForClick(){
	int selLine = g_stw->selLine;
	if(Seamless* smls = getSeamless(selLine)){
		Node* node = g_selNode.getRaw();
		if(
			dynamic_cast<Generator*>(node)||
			dynamic_cast<ColorSweep*>(node)||
			dynamic_cast<TextureMapper*>(node)||
			dynamic_cast<GroupBuild*>(node)||
			dynamic_cast<Effect*>(node)
			){
			fastFullBuildForAutoIn(false, smls);
		}
		else{
			smls->in.refValue()=false; //force all to get set true
			fastFullBuildForAutoIn(true, smls);
		}
	}
}

void fastBuild(){
	fastFullBuild(false);
}
void fullBuild(){
	fastFullBuild(true);
}
void colorVertex(ImmediatePart* part, Art* actTriPart, const int ActTri,const int ActTriCorner, bool endWithBuild = true){
#ifdef PROG_ROB
	dLog("new ClickButton(colorVertex),");
#endif
	Seamless* smls = part->smls;
	BackUp::monostable.beginCycle();
	ColorEffect* effect=dynamic_cast<ColorEffect*>(part->workEffect);
	if(effect==0)return;
	M_Color3f& color=effect->color.refValue();

	M_Triangle& dstTriangle=actTriPart->triangle;
	int dstIndex = ActTri * 3 + ActTriCorner;

	int dstVertex = dstTriangle.refTexIndex(dstIndex);
	color[dstVertex] = g_palette->color;

	int returnCoordIdxIdx = 0;
	if(Art* mirrorPart = smls->getMirrorTriangle(part, returnCoordIdxIdx)){
		color[mirrorPart->triangle.refTexIndex(returnCoordIdxIdx)] = g_palette->color;
	}
	if(endWithBuild){
		smls->doBuild();
		smls->updateAllVertexInBuffers();
	}
}

void colorVertex(){
	if(Part* whitePart=dynamic_cast<Part*>(g_selNode.getRaw())){
		if(dynamic_cast<ColorEffect*>(whitePart->workEffect)==0)return;

		Seamless* smls = whitePart->smls;
		NurbsPatch* patch = 0;
		Stem* stem = 0;
		Polymap* polymap = 0;
		ImmediatePart* bluePart = 0;
		if(getPolyMap(patch, stem, polymap, whitePart, smls->ActTri, bluePart)){
			if(bluePart){
				colorVertex(bluePart, bluePart, smls->ActTri-stem->beginTriOffset, smls->ActTriCorner);
			}
			else{
				int ActTri = smls->ActTri;
				int ActTriCorner = smls->ActTriCorner;
				if(0 > getRelIndex(stem, ActTri, ActTriCorner)){
					return;
				}
				
				int dstIndex = ActTri + ActTriCorner;

				int dstVertex = stem->coordIndex[dstIndex];
				stem->color.refValue()[dstVertex] = g_palette->color;

				smls->doBuild();
				makePointerEqualIndexLocation();
				smls->ctrf=true;
				Window::paintAllControlPanels();
			}
		}
	}
	else if(ImmediatePart* bluePart=dynamic_cast<ImmediatePart*>(g_selNode.getRaw())){
		Seamless* smls = bluePart->smls;
		colorVertex(bluePart, smls->actTriPart, smls->ActTri, smls->ActTriCorner);
	}
	else{
		needPartMsg();
	}
}
void pasteColorToTriangle(ImmediatePart* part, Art* triPart, const int ActTri){
	fastFullBuildForAutoIn(true);
#ifdef PROG_ROB
	dLog("new ClickButton(pasteColorToTriangle),");
#endif
	Seamless* smls=part->smls;
	if(ColorEffect* colorEffect = dynamic_cast<ColorEffect*>(part->workEffect)){
		M_Color3f& color = colorEffect->color.refValue();

		Color3f newColor = g_palette->color;

		color+=newColor;
		color+=newColor;
		color+=newColor;
		triPart->triangle.refTexIndex(ActTri*3+0)=color.length-3;
		triPart->triangle.refTexIndex(ActTri*3+1)=color.length-2;
		triPart->triangle.refTexIndex(ActTri*3+2)=color.length-1;

		int returnCoordIdxIdx=0;
		if(part->mirror){
			if(Art* mirrorPart=smls->getMirrorTriangle(part, returnCoordIdxIdx)){
				color+=newColor;
				color+=newColor;
				color+=newColor;
				returnCoordIdxIdx = returnCoordIdxIdx / 3 * 3;
				mirrorPart->triangle.refTexIndex(returnCoordIdxIdx+0)=color.length-3;
				mirrorPart->triangle.refTexIndex(returnCoordIdxIdx+1)=color.length-2;
				mirrorPart->triangle.refTexIndex(returnCoordIdxIdx+2)=color.length-1;
			}
		}
	}
}

void pasteColorToTriangle(){
	if(Part* whitePart=dynamic_cast<Part*>(g_selNode.getRaw())){
		Seamless* smls = whitePart->smls;
		NurbsPatch* patch = 0;
		Stem* stem = 0;
		Polymap* polymap = 0;
		ImmediatePart* bluePart = 0;
		if(getPolyMap(patch, stem, polymap, whitePart, smls->ActTri, bluePart)){
			if(bluePart){
				pasteColorToTriangle(bluePart, bluePart, smls->ActTri-stem->beginTriOffset);
				commonEndVertexBuild();
			}
			else{
				int ActTri = smls->ActTri;
				int ActTriCorner = smls->ActTriCorner;
				if(0 > getRelIndex(stem, ActTri, ActTriCorner)){
					return;
				}
				bluePart = selectPolyMapPart(whitePart);
				stem->copyPolyMapToPart(bluePart);
				pasteColorToTriangle(bluePart, bluePart, ActTri/3);
				stem->copyPartToPolyMap(bluePart);
				commonEndVertexBuild();
			}
		}
	}
	else if(ImmediatePart* bluePart=dynamic_cast<ImmediatePart*>(g_selNode.getRaw())){
		Seamless* smls = bluePart->smls;
		pasteColorToTriangle(bluePart, smls->actTriPart, smls->ActTri);
		commonEndVertexBuild();
	}
	else{
		needPartMsg();
	}
}

void commonEnd(){
	Art* part=dynamic_cast<Art*>(g_selNode.getRaw());
	part->smls->ctrf=true;
	part->smls->doBuild();
	makePointerEqualIndexLocation();
	BackUp::monostable.beginCycle();
}
void toggleMultiSelectTriVert(){
#ifdef PROG_ROB
	dLog("new ClickButton(cutLeftFromVertex),");
#endif
	MultiSelect::backfaceIgnore = !MultiSelect::backfaceIgnore;
}
void cutLeftFromVertex(){
#ifdef PROG_ROB
	dLog("new ClickButton(cutLeftFromVertex),");
#endif
	if(Art* part=dynamic_cast<Art*>(g_selNode.getRaw())){
		fastFullBuildForAutoIn(true);
		part->cutLeft((part->matrix*part->coord.refValue()[part->smls->ActCtrlPnt]).x+.0001);
		commonEnd();
	}
	else if(Subdivision* subdivision=dynamic_cast<Subdivision*>(g_selNode.getRaw())){
		subdivision->cutLeft();
	}
}
void cutRightFromVertex(){
	Art* part=dynamic_cast<Art*>(g_selNode.getRaw());
	fastFullBuildForAutoIn(true);
	part->cutRight((part->matrix*part->coord.refValue()[part->smls->ActCtrlPnt]).x-.0001);
	commonEnd();
}
void cutUpFromVertex(){
#ifdef PROG_ROB
	dLog("new ClickButton(cutUpFromVertex),");
#endif
	Art* part=dynamic_cast<Art*>(g_selNode.getRaw());
	fastFullBuildForAutoIn(true);
	part->cutUp((part->matrix*part->coord.refValue()[part->smls->ActCtrlPnt]).y-.0001);
	commonEnd();
}

void cutDownFromVertex(){
#ifdef PROG_ROB
	dLog("new ClickButton(cutDownFromVertex),");
#endif
	Art* part=dynamic_cast<Art*>(g_selNode.getRaw());
	fastFullBuildForAutoIn(true);
	part->cutDown((part->matrix*part->coord.refValue()[part->smls->ActCtrlPnt]).y+.0001);
	commonEnd();
}
void commonPasteAddCopyOfMirror(CopyPart& cp, Art* mirrorPart){
	fastFullBuildForAutoIn(true);
	ArtPtr temp = new Part;
	Art* part=dynamic_cast<Art*>(g_selNode.getRaw());
	part->part+=temp;

	GeomSource geomSource;
	mirrorPart->extractGeometryGS(geomSource, g_tmpIndex);
	temp->workEffect->convertGeomSourceToPart(temp.getRawPtr(), geomSource);

	cp.source=temp;
	cp.dest=temp;
	cp.doBuild();
	TransferVertex tv;
	tv.source=temp;
	tv.dest=part;
	tv.doBuild();
	part->part.resize(part->part.refValue().length-1);
	part->smls->process();
	commonEnd();
}
void pasteAddCopyOf(){
	if(Art* part=dynamic_cast<Art*>(g_selNode.getRaw())){
		if(dynamic_cast<Art*>(g_copyNode.getRaw())){
			CopyPart cp;
			commonPasteAddCopyOfMirror(cp, remPJ);
		}
		else{
			needPartCopiedMsg();
		}
	}
	else{
		needPartMsg();
	}
}
void pasteAddCopyOfMirrorX(){
	if(Art* part=dynamic_cast<Art*>(g_selNode.getRaw())){
		if(dynamic_cast<Art*>(g_copyNode.getRaw())){
			CopyPart cp;
			cp.mirrorX=true;
			commonPasteAddCopyOfMirror(cp, remPJ);
		}
		else{
			needPartCopiedMsg();
		}
	}
	else if(Subdivision* subdivision=dynamic_cast<Subdivision*>(g_selNode.getRaw())){
		subdivision->mirrorControlPoints();
	}
	else{
		needPartMsg();
	}
}
void pasteAddCopyOfMirrorY(){
	if(Art* part=dynamic_cast<Art*>(g_selNode.getRaw())){
		if(dynamic_cast<Art*>(g_copyNode.getRaw())){
			CopyPart cp;
			cp.mirrorY=true;
			commonPasteAddCopyOfMirror(cp, remPJ);
		}
		else{
			needPartCopiedMsg();
		}
	}
	else{
		needPartMsg();
	}
}
void pasteAddCopyOfMirrorZ(){
	if(Art* part=dynamic_cast<Art*>(g_selNode.getRaw())){
		if(dynamic_cast<Art*>(g_copyNode.getRaw())){
			CopyPart cp;
			cp.mirrorZ=true;
			commonPasteAddCopyOfMirror(cp, remPJ);
		}
		else{
			needPartCopiedMsg();
		}
	}
	else{
		needPartMsg();
	}
}

void addRelevantPart(M_Art& relevant, Art*& part);

void replaceWithCopyOfRememberedPart(){
	if(Art* part=dynamic_cast<Art*>(g_selNode.getRaw())){
		if(dynamic_cast<Art*>(g_copyNode.getRaw())){
			fastFullBuildForAutoIn(true);
			part->deleteGeometry();
			GeomSource geomSource;
			remPJ->extractGeometryGS(geomSource, g_tmpIndex);
			part->workEffect->convertGeomSourceToPart(part, geomSource);
			commonEnd();
		}
		else if(Stem* stem = dynamic_cast<Stem*>(g_copyNode.getRaw())){
			part->deleteGeometry();
			part->workEffect->stemCopyLocalTri(stem, part, part->coord.refValue(), 0, 0);
//			part->workEffect->stemCopyLocalTri(stem, part, part->coord.refValue(), part->coord.refValue().length, part->triangle.length);
			//stem->copyPolyMapToPart(part);
			addRelevantPart(part->relevantPart, part);
			commonEnd();
		}
		else{
			needPartCopiedMsg();
		}
	}
	else{
		needPartMsg();
	}
}
void commonReplaceWithCopyOfRememberedPart(CopyPart& cp, Art* mirrorPart){
	Art* part=dynamic_cast<Art*>(g_selNode.getRaw());
	fastFullBuildForAutoIn(true);

	//seems to make this work ok but why did I comment the folowing 5 lines once before? does this get used in a build cycle and will it slow things down if it is?
	GeomSource geomSource;
	mirrorPart->extractGeometryGS(geomSource, g_tmpIndex);
	part->deleteGeometry();
	part->workEffect->convertGeomSourceToPart(part, geomSource);
	part->translation.refValue()=mirrorPart->translation;

	cp.source.refValue().length=1;
	cp.dest.refValue().length=1;

	cp.source.refValue()[0] = mirrorPart;
	cp.dest.refValue()[0] = part;
	cp.doBuild();
	part->coordGroup.refValue() = mirrorPart->coordGroup;
	part->weight.refValue() = mirrorPart->weight;
	commonEnd();
	//setupNodeWindow(g_selNode);
	Window::paintAllControlPanels();
}

void replaceWithCopyOfPolyMapMirrorX(){
	if(remPJ==0) return;
	if(Art* part=dynamic_cast<Art*>(g_selNode.getRaw())){
		CopyPart cp;
		cp.insideOut.refValue() = true;
		commonReplaceWithCopyOfRememberedPart(cp, remPJ);
		M_Vec3f& coord = part->coord.refValue();
		for(int i = coord.length; --i>=0;){
			float x = coord[i].x - .5;
			x = -x;
			coord[i].x = x + .5;
		}
		part->smls->process();
		commonEndVertexChange();
	}
}

void replaceWithCopyOfRememberedPartMirrorX(){
	if(Art* part=dynamic_cast<Art*>(g_selNode.getRaw())){
		if(dynamic_cast<Art*>(g_copyNode.getRaw())){
			CopyPart cp;
			cp.mirrorX=true;
			commonReplaceWithCopyOfRememberedPart(cp, remPJ);
		}
		else{
			needPartCopiedMsg();
		}
	}
	else{
		needPartMsg();
	}
}

void replaceWithCopyOfRememberedPartMirrorY(){
	if(Art* part=dynamic_cast<Art*>(g_selNode.getRaw())){
		if(dynamic_cast<Art*>(g_copyNode.getRaw())){
			CopyPart cp;
			cp.mirrorY=true;
			commonReplaceWithCopyOfRememberedPart(cp, remPJ);
		}
		else{
			needPartCopiedMsg();
		}
	}
	else{
		needPartMsg();
	}
}
void replaceWithCopyOfRememberedPartMirrorZ(){
	if(Art* part=dynamic_cast<Art*>(g_selNode.getRaw())){
		if(dynamic_cast<Art*>(g_copyNode.getRaw())){
			CopyPart cp;
			cp.mirrorZ=true;
			commonReplaceWithCopyOfRememberedPart(cp, remPJ);
		}
		else{
			needPartCopiedMsg();
		}
	}
	else{
		needPartMsg();
	}
}

void copyAndPasteMirroredPart(){

#ifdef PROG_ROB
	dLog("new ClickButton(copyAndPasteMirroredPart),");
#endif

	if(Art* part=dynamic_cast<Art*>(g_selNode.getRaw())){
		if(part->mirror){
			CopyPart cp;
			cp.mirrorX = true;
			cp.copyTransform = true;
			if(Art* mirrorPart = part->mirrorPart){
				commonReplaceWithCopyOfRememberedPart(cp, mirrorPart);
			}
			else{
				part->cutLeft(.0001);
				commonPasteAddCopyOfMirror(cp, part);
			}
		}
	}
	else{
		needPartMsg();
	}
}
void invertBackgroundColor(){
	Color3f& backgroundColor = g_scene->backgroundColor.refValue();
	backgroundColor = hexToColor(backgroundColor.colorToHex() ^ 0xffffff);

	Color3f& scene3DFontColor = g_scene->scene3DFontColor.refValue();
	scene3DFontColor = hexToColor(scene3DFontColor.colorToHex() ^ 0xffffff);
	houseKeepVisualClue();

#ifdef PROG_ROB
	dLog("new ClickButton(invertBackgroundColor),");
#endif

}
inline void deleteVertex(Art* vertexPart, int vertexIdx){
	int storeCoordlen=vertexPart->coord.refValue().length;
	
	M_Art& relevantPart=vertexPart->relevantPart;
	for(int k=relevantPart.length; --k>=0;){
		Art* part=relevantPart[k];
		M_Triangle& triangle=part->triangle;
		for(int i=triangle.length; --i>=0; ){
			if(	(triangle[i].pa==vertexPart && triangle[i].ia==vertexIdx) ||
				(triangle[i].pb==vertexPart && triangle[i].ib==vertexIdx) ||
				(triangle[i].pc==vertexPart && triangle[i].ic==vertexIdx))
			{
				vertexPart->deleteTriangle(i);
				if(storeCoordlen!=vertexPart->coord.refValue().length)return;
			}
		}
	}
}
void immediateDeleteVertex(){
	if(Art* part=dynamic_cast<Art*>(g_selNode.getRaw())){
		Seamless* smls = part->smls;
		fastFullBuildForAutoIn(true);

		deleteVertex(part, smls->ActCtrlPnt);

		makePointerEqualIndexLocation();
		smls->convertToRenderFormat();
		BackUp::monostable.beginCycle();
	}
	else{
		needPartMsg();
	}
}

void bubbleSortPartsInTriList(M_<S_VertexIndex>& triList){
	bool flag;
	do{
		flag = false;
		for(int i=0; i < triList.length - 1; i++){

			if(triList[i].p->plidx > triList [i + 1].p->plidx){

				S_VertexIndex swap = triList[i + 1];
				triList[i + 1] = triList[i];
				triList[i] = swap;

				flag = true;
			}
		}
	}while(flag);
}

void deleteTriangle(ImmediatePart* part, Art* actTriPart, const int ActTri, bool endWithBuild = true){
	if((actTriPart == 0 || part->triangle.length==0) && MultiSelect::triList.length == 0){
		return;
	}
#ifdef PROG_ROB
	dLog("new ClickButton(deleteTriangle),");
#endif
	fastFullBuildForAutoIn(true);
	Seamless* smls=part->smls;

	BackUp::monostable.beginCycle();

	DelPoly delPoly;
	M_Int& poly = delPoly.poly.refValue();

	if(MultiSelect::triList.length){
		bubbleSortPartsInTriList(MultiSelect::triList);

		poly += MultiSelect::triList[0].i;
		delPoly.part = MultiSelect::triList[0].p;
		for(int k = 1; k < MultiSelect::triList.length; k++){
			if(MultiSelect::triList[k].p != delPoly.part){
				bubbleSort(poly);
				delPoly.doBuild();
				poly.length = 0;
			}
			poly += MultiSelect::triList[k].i;
			delPoly.part = MultiSelect::triList[k].p;
		}
		bubbleSort(poly);
		delPoly.doBuild();
		MultiSelect::triList.length = 0;
		MultiSelect::coordList.length = 0;
	}
	else{
		delPoly.part = actTriPart;
		poly += ActTri;
		delPoly.doBuild();
	}
	if(endWithBuild == false){
		return;
	}
	smls->doBuild();
	makePointerEqualIndexLocation();
	smls->convertToRenderFormat();
	g_stw->paint();
}

void resetCPNodeList();

void deleteTriangle(){
	if(Part* whitePart=dynamic_cast<Part*>(g_selNode.getRaw())){
		Seamless* smls = whitePart->smls;
		NurbsPatch* patch = 0;
		Stem* stem = 0;
		Polymap* polymap = 0;
		ImmediatePart* bluePart = 0;
		if(getPolyMap(patch, stem, polymap, whitePart, smls->ActTri, bluePart)){
			if(bluePart){
				deleteTriangle(bluePart, bluePart, smls->ActTri-stem->beginTriOffset);
			}
			else{
				int ActTri = smls->ActTri;
				int ActTriCorner = smls->ActTriCorner;
				if(0 > getRelIndex(stem, ActTri, ActTriCorner)){
					return;
				}

				bluePart = selectPolyMapPart(whitePart);
				stem->copyPolyMapToPart(bluePart);
				deleteTriangle(bluePart, bluePart, ActTri / 3, false);
				stem->copyPartToPolyMap(bluePart);
				commonEndVertexBuild();
			}
		}
	}
	else if(ImmediatePart* bluePart = dynamic_cast<ImmediatePart*>(g_selNode.getRaw())){
		Seamless* smls = bluePart->smls;
		deleteTriangle(bluePart, smls->actTriPart, smls->ActTri);
	}
	else if(NurbsPatch* nurbsPatch = dynamic_cast<NurbsPatch*>(g_selNode.getRaw())){
		Seamless* smls = nurbsPatch->smls;
		
		smls->deleteCtrlPoly(MultiSelect::ctrlPolySelList);

		MultiSelect::ctrlPolySelList.length = 0;
		resetCPNodeList();
		nurbsPatch->houseKeep();
	}
	else if(BezierPatch* bezierPatch = dynamic_cast<BezierPatch*>(g_selNode.getRaw())){
		bezierPatch->deletePatch();
	}
	else if(CCLathe* ccLathe = dynamic_cast<CCLathe*>(g_selNode.getRaw())){
		ccLathe->deleteControlPoint();
	}
	else if(Subdivision* subdivision = dynamic_cast<Subdivision*>(g_selNode.getRaw())){
		subdivision->delPoly();
	}
	else if(NurbsLathe* nurbsLathe = dynamic_cast<NurbsLathe*>(g_selNode.getRaw())){
		int newLen=0;
		nurbsLathe->removeControlPoint(newLen);
		nurbsLathe->processFields();
		updateBuild3dWindow();
		houseKeepVisualClue();
	}
	else if(NurbsTrack* nurbsTrack = dynamic_cast<NurbsTrack*>(g_selNode.getRaw())){
		nurbsTrack->removeControlPoint(nurbsTrack->index[1]);
	}
	else{
		needPartMsg();
	}
}

void convertToCNames(){
	for(int k=0; k<g_sceneTreeList.length; k++){
		Node* node=g_sceneTreeList[k].node.p;
		if(makeNameLegalForCpp(node->name)){
			 makeExistingNameUnique(node);
		}
	}
	BackUp::monostable.beginCycle();
}

void NurbsPatch::mirrorModifyWest(){
	M_Vec3f& point = this->point.refValue();
	M_Float& weight = this->weight.refValue();
	const Vec2f& dim = this->dim;
	int uDim=dim.x;
	int vDim=dim.y;
	int middleIdx = (uDim >> 1);
	for(int v=vDim; --v>=0;){

		for(int u = 0; u < uDim>>1; u++){
			Vec3f p = point[v*uDim + uDim - 1 - u];
			p.x = -p.x;
			point[v*uDim + u] = p;

			weight[v*uDim + u] = weight[v*uDim + uDim - 1 - u];;
		}
		if(uDim && 1){
			point[v*uDim + middleIdx].x = 0;
		}
	}
}
void NurbsPatch::mirrorModifyEast(){
	M_Vec3f& point = this->point.refValue();
	M_Float& weight = this->weight.refValue();
	const Vec2f& dim = this->dim;
	int uDim=dim.x;
	int vDim=dim.y;
	int middleIdx = (uDim >> 1);
	for(int v=vDim; --v>=0;){

		for(int u = 0; u < uDim>>1; u++){
			Vec3f p = point[v*uDim + u];
			p.x = -p.x;
			point[v*uDim + uDim - 1 - u] = p;

			weight[v*uDim + uDim - 1 - u] = weight[v*uDim + u];
		}
		if(uDim && 1){
			point[v*uDim + middleIdx].x = 0;
		}
	}
}
/*
void matchEndsForPatch(){
	if(Window::chatMode())return;
	if(NurbsPatch* nurbsPatch=dynamic_cast<NurbsPatch*>(g_selNode.getRaw())){
		M_Vec3f& point = nurbsPatch->point.refValue();
		M_Float& weight = nurbsPatch->weight.refValue();
		const Vec2f& dim = nurbsPatch->dim;
		int uDim=dim.x;
		int vDim=dim.y;
		for(int v=vDim; --v>=0;){

			point[v * uDim + uDim - 1] = point[v * uDim];

			point[v * uDim + uDim - 2].y = point[v * uDim].y;
			point[v * uDim + uDim - 2].z = point[v * uDim].z;

			point[v * uDim + 1].y = point[v * uDim].y;
			point[v * uDim + 1].z = point[v * uDim].z;

		}
		nurbsPatch->processFields();
	}
	else if(BezierPatch* bezierPatch=dynamic_cast<BezierPatch*>(g_selNode.getRaw())){
		M_Vec3f& point = bezierPatch->point.refValue();
		const M_Vec2f& dim = bezierPatch->dim;
		int uDim=dim[0].x;
		int vDim=dim[0].y;
		for(int v=vDim; --v>=0;){

			point[v * uDim + uDim - 1] = point[v * uDim];

			point[v * uDim + uDim - 2].y = point[v * uDim].y;
			point[v * uDim + uDim - 2].z = point[v * uDim].z;

			point[v * uDim + 1].y = point[v * uDim].y;
			point[v * uDim + 1].z = point[v * uDim].z;

		}
	}
	else{
		return;
	}
	updateBuild3dWindow();
	houseKeepVisualClue();
	BackUp::monostable.beginCycle();
}
*/
void flipPolarity(){
	if(NurbsPatch* nurbsPatch=dynamic_cast<NurbsPatch*>(g_selNode.getRaw())){
		M_Vec3f& point = nurbsPatch->point.refValue();
		M_Float& weight = nurbsPatch->weight.refValue();
		const Vec2f& dim = nurbsPatch->dim;
		int uDim=dim.x;
		int vDim=dim.y;
		for(int v = vDim; --v>=(vDim>>1);){
			for(int u = uDim; --u>= 0;){
				Vec3f h = point[(vDim - v - 1) * uDim + u];
				float w = weight[(vDim - v - 1) * uDim + u];
				point[(vDim - v - 1) * uDim + u] = point[v * uDim + u];
				weight[(vDim - v - 1) * uDim + u] = weight[v * uDim + u];
				point[v * uDim + u] = h;
				weight[v * uDim + u] = w;
			}
		}
		for(int v = vDim; --v>=0;){
			for(int u = uDim; --u>=(uDim>>1);){
				Vec3f h = point[v * uDim + (uDim - u - 1)];
				float w = weight[v * uDim + (uDim - u - 1)];
				point[v * uDim + (uDim - u - 1)] = point[v * uDim + u];
				weight[v * uDim + (uDim - u - 1)] = weight[v * uDim + u];
				point[v * uDim + u] = h;
				weight[v * uDim + u] = w;
			}
		}
		nurbsPatch->processFields();
	}
	else if(BezierPatch* bezierPatch=dynamic_cast<BezierPatch*>(g_selNode.getRaw())){
		M_Vec3f& point = bezierPatch->point.refValue();
		const M_Vec2f& dim = bezierPatch->dim;
		int uDim=dim[0].x;
		int vDim=dim[0].y;
		for(int v = vDim; --v>=(vDim>>1);){
			for(int u = uDim; --u>= 0;){
				Vec3f h = point[(vDim - v - 1) * uDim + u];
				point[(vDim - v - 1) * uDim + u] = point[v * uDim + u];
				point[v * uDim + u] = h;
			}
		}
		for(int v = vDim; --v>=0;){
			for(int u = uDim; --u>=(uDim>>1);){
				Vec3f h = point[v * uDim + (uDim - u - 1)];
				point[v * uDim + (uDim - u - 1)] = point[v * uDim + u];
				point[v * uDim + u] = h;
			}
		}
	}
	else if(WaveToPoly* waveToPoly = dynamic_cast<WaveToPoly*>(g_selNode.getRaw())){
		waveToPoly->findSignal();
	}
	else{
		return;
	}
	updateBuild3dWindow();
	houseKeepVisualClue();
	BackUp::monostable.beginCycle();
}

void insideOutPatch(){
	if(NurbsPatch* nurbsPatch=dynamic_cast<NurbsPatch*>(g_selNode.getRaw())){
		M_Vec3f& point = nurbsPatch->point.refValue();
		M_Float& weight = nurbsPatch->weight.refValue();
		M_Vec3f storePoint = point;
		M_Float storeWeight = weight;
		M_Int lookup;
		lookup.length = point.length;

		const Vec2f& dim = nurbsPatch->dim;
		int uDim = dim.x;
		int vDim = dim.y;
		for(int v = vDim; --v >= 0;){
			for(int u = uDim; --u >= 0;){
				int d = v * uDim + u;
				int s = v * uDim + uDim - 1 - u;
				point[d] = storePoint[s];
				weight[d] = storeWeight[s];
				lookup[d] = s;
			}
		}
		nurbsPatch->smls->redirect(nurbsPatch, lookup);
		nurbsPatch->processFields();
		updateBuild3dWindow();
		houseKeepVisualClue();
		BackUp::monostable.beginCycle();
	}
}

extern TransformPtr g_nurbsIlsBase;
void fillPolyLineIndex(M_Vec3f& point, M_Int& coordIndex);

void alignAllMidPoints(){
	if(NurbsPatch* nurbsPatch=dynamic_cast<NurbsPatch*>(g_selNode.getRaw())){
		M_Vec3f& dpoint = nurbsPatch->dpoint.refValue();
		M_Vec3f& point = nurbsPatch->point.refValue();
		const M_Float& weight = nurbsPatch->weight;
		int uDim = nurbsPatch->dim.x;
		int vDim = nurbsPatch->dim.y;

		for(int v=0; v<vDim; v++){
			for(int u=0; u<uDim; u+=2){
				int i = v*uDim + u;
				if(u == 0){
					point[i+uDim-1]=point[i]=(dpoint[i+1]+(dpoint[i+uDim-2]-dpoint[i+1])*.5)/weight[i];
				}
				else if(u == uDim-1){
					point[i-uDim+1]=point[i]=(dpoint[i-1]+(dpoint[i-uDim+2]-dpoint[i-1])*.5)/weight[i];
				}
				else{
					point[i]=(dpoint[i-1]+(dpoint[i+1]-dpoint[i-1])*.5)/weight[i];
				}
			}
		}

		nurbsPatch->processFields();
	}
	else{
		return;
	}
	updateBuild3dWindow();
	houseKeepVisualClue();
	BackUp::monostable.beginCycle();
}
void alignMidPoint(){
	if(NurbsPatch* nurbsPatch=dynamic_cast<NurbsPatch*>(g_selNode.getRaw())){
		M_Vec3f& dpoint = nurbsPatch->dpoint.refValue();
		M_Vec3f& point = nurbsPatch->point.refValue();
		const M_Float& weight = nurbsPatch->weight;
		int uDim = nurbsPatch->dim.x;

		int i = nurbsPatch->index[0];
		if(i%uDim == 0){
			point[i+uDim-1]=point[i]=(dpoint[i+1]+(dpoint[i+uDim-2]-dpoint[i+1])*.5)/weight[i];
		}
		else if(i%uDim == uDim-1){
			point[i-uDim+1]=point[i]=(dpoint[i-1]+(dpoint[i-uDim+2]-dpoint[i-1])*.5)/weight[i];
		}
		else{
			point[i]=(dpoint[i-1]+(dpoint[i+1]-dpoint[i-1])*.5)/weight[i];
		}

		nurbsPatch->processFields();
	}
	else{
		return;
	}
	updateBuild3dWindow();
	houseKeepVisualClue();
	BackUp::monostable.beginCycle();
}

void outputPovRayMovie(){
	Anim* anim = Window::animMode();
	g_scene->outputScene(anim);
}

const char audio[3][5]={
	".wav",
	".mp3",
	".au",
};

void addAudioFileName(const String& fname, String& ffmpegLine){
	for(int i=0; i<3; i++){
		String audioFile=fname;
		audioFile+=audio[i];
		FILE* file=fopen(audioFile.c_str(), "rb");
		if(file){
			fclose(file);
//			if(g_scene->movieAVI){
//				ffmpegLine+=" -acodec mp3"; //caused crash in ffmpeg auto installer
//			}
			ffmpegLine+=" -i ";
			ffmpegLine+=audioFile;
			break;
		}
	}
}

void addMovieFileExtension(String& fname){
	if(g_scene->movieMPG){
		fname+=".mpg";
	}
	else if(g_scene->movieMP4){
		fname+=".mp4";
	}
	else if(g_scene->movieAVI){
		fname+=".avi";
	}
	else if(g_scene->movieFLV){
		fname+=".flv";
	}
}

void createFFmpegCommand(String& ffmpegLine, String& fname, bool batchFile){
	Anim* anim = Window::animMode();
	int numOfFrames = anim->period*anim->stepsPerSecond+1;

	String filename = numOfFrames;
	int numOfDigits = filename.length;

	filename = g_currentFilename;

	String drive;
	String dir;
	String ext;
	String iniFileName;
	splitPath(filename, drive, dir, fname, ext);




	//getFFmpegPath(ffmpegLine);
	ffmpegLine+=" -r ";
	ffmpegLine+=anim->stepsPerSecond;
	ffmpegLine+=" -f image2 -i ";
	ffmpegLine+=fname;
	if(batchFile){
		ffmpegLine+="%";
	}
	ffmpegLine+="%0";
	ffmpegLine+=numOfDigits;
	ffmpegLine+="d.png";
	if(g_scene->movieAVI){
		ffmpegLine+=" -vcodec msmpeg4v2";
	}

	addAudioFileName(fname, ffmpegLine);

	ffmpegLine+=" -sameq ";
	
	//fname+="-movie";
}

void renderScene();
extern Pix24* g_frameBuffer;
void createPNG24(const char* filename, int width, int height, Pix24* pixels);
void createBatchFile(String& batchFileName);
bool g_screenshotEarly = false;
bool g_screenshotLate = false;
extern bool g_recordSoundMovie;
void openWithFFmpeg(String url);

void outputScreenShotMovie(){
	g_recordSoundMovie = false;
	g_screenshotLate = false;
	g_screenshotEarly = false;
	if(g_settings->screenshotWindows){
		g_screenshotLate = true;
	}
	else{
		g_screenshotEarly = true;
	}


	int width = g_3d->size.x;
	int height = g_3d->size.y;

	g_frameBuffer = new Pix24[width * height];
	Anim* anim = Window::animMode();
	String filename = g_currentFilename;
	String drive;
	String dir;
	String fname;
	String ext;
	String iniFileName;
	splitPath(filename, drive, dir, fname, ext);


	float time = 0;	///g_freezeTime->time;//needs to be fixed
	float multiply = 1;
	float period = 1;
	float step = 0;

	if(anim){
		multiply = anim->speed;
		period = anim->period;
		step = 1.0 / (anim->stepsPerSecond * period);
		anim->slider = 0;
	}
	

	int numOfFrames = 1;


	numOfFrames += period * anim->stepsPerSecond;

	String str;
	String cntrStr;

	str = numOfFrames;
	int numOfDigits = str.length;

	//////////// start inc files
	static M_<Effect*> effectList;
	effectList.length=0;
	float remTime=0; //freezeTime->time;
	for(int i = 0; i < numOfFrames; i++){
		anim->slider = step * i * multiply + time;
		//SendMessage(hWndMain, WM_PAINT, 0, 0);
		renderScene();

		str = drive;
		str += dir;
		str += fname;
		//outPovRay(str, effectList, Matrix(), Matrix());
		cntrStr = i;
		while(cntrStr.length < numOfDigits){
			cntrStr = '0' + cntrStr;
		}
		str += cntrStr;

		str += ".png";
		
		createPNG24(str.c_str(), width, height, g_frameBuffer);
	}


	g_screenshotLate = false;
	g_screenshotEarly = false;

	delete[] g_frameBuffer;
	g_frameBuffer = 0;

	//createBatchFile(str);
	String ffmpegLine;
	createFFmpegCommand(ffmpegLine, fname, false);
	addMovieFileExtension(fname);
	DeleteFile(fname.c_str());
	ffmpegLine += fname;
//	ffmpegLine+=fname;

	
	openWithFFmpeg(ffmpegLine);
}

void outputScreenShotSingle(){
	g_recordSoundMovie = false;
	g_screenshotLate = false;
	g_screenshotEarly = false;
	if(g_settings->screenshotWindows){
		g_screenshotLate = true;
	}
	else{
		g_screenshotEarly = true;
	}

	int width = g_3d->size.x;
	int height = g_3d->size.y;
	g_frameBuffer = new Pix24[width * height];

	renderScene();
	
	SYSTEMTIME ToDaySystemTime;
	GetLocalTime(&ToDaySystemTime);
	char tmpChar[0x200];
	ToDaySystemTime.wYear -= 2000;
	sprintf(tmpChar,"%.2i%.2i%.2i_%.2i%.2i%.2i.png",ToDaySystemTime.wYear, ToDaySystemTime.wMonth, ToDaySystemTime.wDay, ToDaySystemTime.wHour,ToDaySystemTime.wMinute, ToDaySystemTime.wSecond);
	
	String str = g_screenshotDir;
	str += tmpChar;
	createPNG24(str.c_str(), width, height, g_frameBuffer);

	
	g_screenshotLate = false;
	g_screenshotEarly = false;

	delete[] g_frameBuffer;
	g_frameBuffer = 0;

}

bool modifyFilename(String& filename, bool writeX3DXML, bool writeX3DVRMLClassic, bool writeObj, bool outputPovRay, Anim* anim);
void writeOutputHeader(String& str,  const Properties* switchProp, const Properties* lodProp);
void changeFloatFormat(int numOfDigits);
void listRoutes(String&);
extern const char* msg_Can_not_create_output_file_name;

void outputVrmlMorph2(Seamless* smls, Anim* anim){
	g_vrmlSquareOpenBracket[0]='[';
	g_vrmlSquareCloseBracket[0]=']';
	g_vrmlQuote[0]='\"';

	String filename = g_currentFilename;
	if(false == modifyFilename(filename,g_scene->outputX3DXML,g_scene->outputX3DVRMLClassic, false, false, false)){
		MessageWin::msg(msg_Can_not_create_output_file_name,MB_OK | MB_ICONEXCLAMATION);
		return;
	};
	if(g_scene->outputGZipped && (g_scene->outputX3DXML || g_scene->outputX3DVRMLClassic)){
		filename+='z';
	}
	bool remGenNormals = g_genNormals;

	g_formatCoords=g_scene->roundCoordsOutput.refValue();
	g_outputMode=true;
	g_genNormals = false;
	g_genIdxTriSetNodes=false;
	g_writeX3DXML = false;
	g_writeX3DVRMLClassic = false;
	g_genHAnimNodes=false;
	g_genContactSBVT=g_scene->genContactSBVT;
	g_genColorIndex = true;

	M_Node temp;
	float multiply = anim->speed;
	float period = anim->period;
	float step = 1.0 / (anim->stepsPerSecond * period);
	anim->slider = 0;
	int numOfFrames = 1;
	numOfFrames += period * anim->stepsPerSecond;
	ScenePtr newScene = new Scene;

	CoordinateInterpolator* interpolator = new CoordinateInterpolator;
	interpolator->name = "interpolator";

	TimeSensor* timeSensor = new TimeSensor;
	timeSensor->cycleInterval = anim->period;
	timeSensor->loop = true;
	timeSensor->name = "timeSensor";
	String str;

	for(int i = 0; i < numOfFrames; i++){
		anim->slider = step * i * multiply;
		//SendMessage(hWndMain, WM_PAINT, 0, 0);
		renderScene();

		str="";
		temp.length = 0;
		g_defLookUp.length=0;
		smls->outputToString(str, false);
		g_defLookUp.length=0;
		createSmlsFromString(temp, str);
		ShapePtr shape = dynamic_cast<Shape*>(temp[temp.length-1].p);
		IndexedFaceSet* ifs = dynamic_cast<IndexedFaceSet*>(shape->geometry.p);
		Coordinate* coord = dynamic_cast<Coordinate*>(ifs->coord.p);
		interpolator->key.refValue()+=anim->slider;
		interpolator->keyValue.refValue()+=coord->point;
		if(i == 0){
			newScene->children.refValue() += temp;
			coord->name = "coord";
			S_Route(timeSensor, timeSensor->getFieldIndex(&timeSensor->fraction_changed), interpolator, interpolator->getFieldIndex(&interpolator->set_fraction));
			S_Route(interpolator, interpolator->getFieldIndex(&interpolator->value_changed), coord, coord->getFieldIndex(&coord->point));
		}
	}
	newScene->children.refValue() += interpolator;
	newScene->children.refValue() += timeSensor;

	g_writeX3DXML=g_scene->outputX3DXML;
	g_writeX3DVRMLClassic=g_scene->outputX3DVRMLClassic;

	_chdir(g_seamlessDir.c_str());
	writeOutputHeader(str,  0, 0);

	g_defLookUp.length=0;
	changeFloatFormat(g_formatCoords=g_settings->roundCoordsSave.getValue());
	if(g_writeX3DXML){
		g_vrmlSquareOpenBracket[0]=0;
		g_vrmlSquareCloseBracket[0]=0;
		g_vrmlQuote[0]=0;
		str+='\n';
		newScene->children.toXMLString(str, false, "children");
		g_vrmlSquareOpenBracket[0]='[';
		g_vrmlSquareCloseBracket[0]=']';
		g_vrmlQuote[0]='\"';
	}
	else{
		M_Node& children = newScene->children;
		for(int i=0;i<children.length;i++){
			str+='\n';
			children[i]->toString(str, false);
		}
	}
	changeFloatFormat(g_formatCoords=g_settings->roundCoordsSave.getValue());

	listRoutes(str);
	if(g_writeX3DXML){
		str+="\n</Scene>\n</X3D>";
	}


	//if(g_writeX3DXML){
	//	g_vrmlSquareOpenBracket[0]=0;
	//	g_vrmlSquareCloseBracket[0]=0;
	//	g_vrmlQuote[0]=0;
	//	str+='\n';
	//	scene->toXMLString(str, true, "children");
	//	str+="\n<ROUTE fromNode='interpolator' fromField='value_changed' toNode='coord' toField='point'/>\n<ROUTE fromNode='timeSensor' fromField='fraction_changed' toNode='interpolator' toField='set_fraction'/>\n</Scene>\n</X3D>";
	//	g_vrmlSquareOpenBracket[0]='[';
	//	g_vrmlSquareCloseBracket[0]=']';
	//	g_vrmlQuote[0]='\"';
	//}
	//else{
	//	scene->toString(str);
	//	//str+="\nROUTE timeSensor.fraction_changed TO interpolator.set_fraction\nROUTE interpolator.value_changed TO coord.point\n";
	//}
	g_writeGzipped=g_scene->outputGZipped;
	str.write(filename);
	g_writeGzipped=g_settings->saveGZipped;

	g_outputMode=false;
	g_writeX3DXML=false;
	g_writeX3DVRMLClassic=false;
	g_genContactSBVT=true;
	g_genNormals=true;
	g_genColorIndex=false;

	g_genNormals = remGenNormals;
}


#ifdef DEMO_TO_MOVIE

int g_frame = 0;
int g_clipCounter = 0;

void outputScreenShotForDemo1(){
	g_recordSoundMovie = false;
	g_screenshotLate = false;
	g_screenshotEarly = false;
	if(g_settings->screenshotWindows){
		g_screenshotLate = true;
	}
	else{
		g_screenshotEarly = true;
	}

	int width = g_3d->size.x;
	int height = g_3d->size.y;
	g_frameBuffer = new Pix24[width * height];
}

void outputScreenShotForDemo2(){
	int width = g_3d->size.x;
	int height = g_3d->size.y;

	String filename = g_screenshotDir;

	filename+= "\\clip";

	String cntrStr = g_clipCounter;
	while(cntrStr.length < 4){
		cntrStr = '0' + cntrStr;
	}
	filename += cntrStr;
	filename += "\\frame";

	cntrStr = g_frame;
	while(cntrStr.length < 4){
		cntrStr = '0' + cntrStr;
	}
	filename += cntrStr;

	filename += ".png";

	createPNG24(filename.c_str(), width, height, g_frameBuffer);
	
	g_screenshotLate = false;
	g_screenshotEarly = false;

	delete[] g_frameBuffer;
	g_frameBuffer = 0;
	g_frame++;

}
#endif

void createBatchFile(String& batchFileName){
	String ffmpegLine;
	String fname;
	getFFmpegPath(ffmpegLine);
	createFFmpegCommand(ffmpegLine, fname, true);

	batchFileName=fname;
	batchFileName+=".bat";
	addMovieFileExtension(fname);

	ffmpegLine+=fname;
	ffmpegLine+="\r\npause";

	g_writeGzipped=false;
	ffmpegLine.write(batchFileName);
	g_writeGzipped=g_scene->outputGZipped;
}
void outputBatchFile(){
	String batchFileName;
	String outputMsg;
	createBatchFile(batchFileName);
	extractFileName(outputMsg, batchFileName);
	outputMsg+="\n\nhas been created";
	msG(outputMsg);
}

Seamless* getSeamless2();

void NurbsLathe::convertToPatch(){
	g_convertToPatch->ctrlStaves.refValue() = ctrlStaves;
	patch->showAll.refValue()=false;

	int selLine = g_stw->selLine;
	int parentLine = getParent(selLine);
	Node* parentNode = g_sceneTreeList[parentLine].node.p;
	int fieldIdx=0, element=0;

	getFNodeElementForParent(parentLine, selLine, g_sceneTreeList, fieldIdx, element);

	Field* field=parentNode->getField(fieldIdx);

	patch->processFields();

	if(field->getType()==MFNODE){
		MFNode *mfnode=dynamic_cast<MFNode*>(field);
		mfnode->refValue()[element]=patch;
		*mfnode=mfnode->refValue();
	}

	if(getSeamless2()->animCP){
		patch->tpoint = patch->point;
	}
	g_stw->commonPostInsertNodeStuff();
}
void BendyLathe::convertToPatch(){
	g_convertToPatch->ctrlStaves.refValue() = ctrlStaves;

	int selLine = g_stw->selLine;
	int parentLine=getParent(selLine);
	Node* parentNode= g_sceneTreeList[parentLine].node.p;
	int fieldIdx=0, element=0;

	getFNodeElementForParent(parentLine, selLine, g_sceneTreeList, fieldIdx, element);

	Field* field=parentNode->getField(fieldIdx);

	patch->processFields();

	if(field->getType()==MFNODE){
		MFNode *mfnode=dynamic_cast<MFNode*>(field);
		mfnode->refValue()[element]=patch;
		*mfnode=mfnode->refValue();
	}
	if(getSeamless2()->animCP){
		patch->tpoint = patch->point;
	}
	g_stw->commonPostInsertNodeStuff();
}
void NurbsLathe::convertToBendy(){
	Ptr<BendyLathe> bendy = new BendyLathe;
	M_Float& radius = bendy->radius.refValue();
	M_Float& at = bendy->at.refValue();
	M_Vec3f& bendyPoint = bendy->point.refValue();
	M_Float& bendyWeight = bendy->weight.refValue();
	M_Float& bendyKnot = bendy->knot.refValue();
	float minT = bendy->minT();
	float maxT = bendy->maxT();

	int selLine = g_stw->selLine;
	int parentLine=getParent(selLine);
	Node* parentNode= g_sceneTreeList[parentLine].node.p;
	int fieldIdx=0, element=0;

	getFNodeElementForParent(parentLine, selLine, g_sceneTreeList, fieldIdx, element);

	Field* field=parentNode->getField(fieldIdx);

	bendy->part.refValue() = part.refValue();
	bendy->stem = stem;
	bendy->patch = patch;
	bendy->color.refValue() = color;
	bendy->scale.refValue() = scale;
	bendy->rotation.refValue() = rotation;
	bendy->translation.refValue() = translation;
	bendy->staves.refValue() = staves;
	bendy->bands.refValue() = bands;
//	bendy->point.length = point.length;
	float lowestPoint = getLowestPoint(point.refValue());
	float highestPoint = getHighestPoint(point.refValue());
	float diff = highestPoint - lowestPoint;
	int numOfRings = point.length;
	int numOfBendyPonts = bendy->point.length;
	radius.length = numOfRings;
	at.length = numOfRings;

	for(int i = 0; i < numOfBendyPonts; i++){
		bendyPoint[i].y = lowestPoint + diff/(numOfBendyPonts-1) * i;
	}
	for(int i = 0; i < numOfRings; i++){
		//at[i] = 1.0/(numOfRings-1) * i;
		at[i] = bendy->zeroinOnVerticalPoint(bendyPoint, bendyWeight, bendyKnot, minT, maxT, point[i].y);
		radius[i] = point[i].x;
	}

	bendy->processFields();

	if(field->getType()==MFNODE){
		MFNode *mfnode=dynamic_cast<MFNode*>(field);
		mfnode->refValue()[element]=bendy;
		*mfnode=mfnode->refValue();
	}
	g_stw->commonPostInsertNodeStuff();
}
void BezierLatheBuild::convertToPatch(){
	g_convertToPatch->ctrlStaves.refValue()=4;
	g_convertToPatch->ctrlBands.refValue()=3;
	
	int selLine = g_stw->selLine;
	int parentLine=getParent(selLine);
	Node* parentNode= g_sceneTreeList[parentLine].node.p;
	int fieldIdx=0, element=0;

	getFNodeElementForParent(parentLine, selLine, g_sceneTreeList, fieldIdx, element);

	Field* field=parentNode->getField(fieldIdx);

	if(nopfs.length > 1){
		createUniqueName(colorSweep);
		if(textureMapper.length < nopfs.length){
			for(int k = textureMapper.length; --k>=0;){
				createUniqueName(textureMapper[k]);
			}
		}
	}

	if(field->getType()==MFNODE){
		MFNode *mfnode=dynamic_cast<MFNode*>(field);
		mfnode->refValue()[element] = m_patch[0];
		for(int k = 1; k < m_patch.length; k++){
			mfnode->refValue().insertElement(++element, m_patch[k].p);
		}
		if(m_patch.length > 1){
			mfnode->refValue().insertElement(++element, joinVertex.p);
		}
		*mfnode=mfnode->refValue();
	}
	if(getSeamless2()->animCP){
		for(int k = m_patch.length; --k >= 0;){
			m_patch[k]->tpoint = m_patch[k]->point;
		}
	}
	g_stw->commonPostInsertNodeStuff();
}
void SphereBuild::convertToPatch(){
	g_convertToPatch->ctrlStaves.refValue()=4;
	g_convertToPatch->ctrlBands.refValue()=3;
	
	int selLine = g_stw->selLine;
	int parentLine=getParent(selLine);
	Node* parentNode= g_sceneTreeList[parentLine].node.p;
	int fieldIdx=0, element=0;

	getFNodeElementForParent(parentLine, selLine, g_sceneTreeList, fieldIdx, element);

	Field* field=parentNode->getField(fieldIdx);

	MFNode *mfnode=dynamic_cast<MFNode*>(field);
	mfnode->refValue()[element]=patch;
	*mfnode=mfnode->refValue();

	if(getSeamless2()->animCP){
		patch->tpoint = patch->point;
	}
	g_stw->commonPostInsertNodeStuff();
}
void TorusBuild::convertToPatch(){

	g_convertToPatch->ctrlStaves.refValue()=6;
	g_convertToPatch->ctrlBands.refValue()=4;
	
	int selLine = g_stw->selLine;
	int parentLine=getParent(selLine);
	Node* parentNode= g_sceneTreeList[parentLine].node.p;
	int fieldIdx=0, element=0;

	getFNodeElementForParent(parentLine, selLine, g_sceneTreeList, fieldIdx, element);

	Field* field=parentNode->getField(fieldIdx);

	if(field->getType()==MFNODE){
		MFNode *mfnode=dynamic_cast<MFNode*>(field);
		mfnode->refValue()[element]=patch;
		mfnode->refValue().insertElement(++element, joinVertex.p);
		*mfnode=mfnode->refValue();
	}
	if(getSeamless2()->animCP){
		patch->tpoint = patch->point;
	}
	g_stw->commonPostInsertNodeStuff();
}
void RectangleBuild::convertToPatch(){
	int selLine = g_stw->selLine;
	int parentLine=getParent(selLine);
	Node* parentNode= g_sceneTreeList[parentLine].node.p;
	int fieldIdx=0, element=0;

	getFNodeElementForParent(parentLine, selLine, g_sceneTreeList, fieldIdx, element);

	Field* field=parentNode->getField(fieldIdx);

	if(field->getType()==MFNODE){
		MFNode *mfnode=dynamic_cast<MFNode*>(field);
		mfnode->refValue()[element]=patch;
		*mfnode=mfnode->refValue();
	}
	if(getSeamless2()->animCP){
		patch->tpoint = patch->point;
	}
	g_stw->commonPostInsertNodeStuff();
}
void BoxBuild::convertToPatch(){
	
	int selLine = g_stw->selLine;
	int parentLine=getParent(selLine);
	Node* parentNode= g_sceneTreeList[parentLine].node.p;
	int fieldIdx=0, element=0;

	for(int k = 0; k < m_patch.length; k++){
		NurbsPatch* patch = m_patch[k];
		Stem* stem = dynamic_cast<Stem*>(patch->stem.refValue()[0].p);
		stem->bands.refValue()[0] = 16;
		stem->staves.refValue() = 16;
	}

	getFNodeElementForParent(parentLine, selLine, g_sceneTreeList, fieldIdx, element);

	Field* field=parentNode->getField(fieldIdx);

	createUniqueName(colorSweep);
	if(textureMapper.length < 6){
		for(int k = textureMapper.length; --k>=0;){
			createUniqueName(textureMapper[k]);
		}
	}

	if(field->getType()==MFNODE){
		MFNode *mfnode=dynamic_cast<MFNode*>(field);
		mfnode->refValue()[element] = m_patch[0];
		for(int k = 1; k < m_patch.length; k++){
			mfnode->refValue().insertElement(++element, m_patch[k].p);
		}
		if(textureMapper.refValue().length){
			g_stw->selLine++;
		}
		*mfnode=mfnode->refValue();
	}
	if(getSeamless2()->animCP){
		for(int k = m_patch.length; --k >= 0;){
			m_patch[k]->tpoint = m_patch[k]->point;
		}
	}
	g_stw->commonPostInsertNodeStuff();
}
void CylinderBuild::convertToPatch(){
	g_convertToPatch->ctrlStaves.refValue()=4;
	g_convertToPatch->ctrlBands.refValue()=1;

	int selLine = g_stw->selLine;
	int parentLine=getParent(selLine);
	Node* parentNode= g_sceneTreeList[parentLine].node.p;
	int fieldIdx=0, element=0;

	getFNodeElementForParent(parentLine, selLine, g_sceneTreeList, fieldIdx, element);

	Field* field=parentNode->getField(fieldIdx);

	if(baseCap || topCap){
		createUniqueName(patch);
		if(stem->textureMapper.refValue().length){
			createUniqueName(stem->textureMapper.refValue()[0]);
		}
	}


	patch->processFields();

	if(field->getType()==MFNODE){
		MFNode *mfnode=dynamic_cast<MFNode*>(field);
		mfnode->refValue()[element]=patch;
		*mfnode=mfnode->refValue();
	}
	if(getSeamless2()->animCP){
		patch->tpoint = patch->point;
	}
	g_stw->commonPostInsertNodeStuff();
}
void ConeBuild::convertToPatch(){
	g_convertToPatch->ctrlStaves.refValue()=4;
	g_convertToPatch->ctrlBands.refValue()=3;
	
	int selLine = g_stw->selLine;
	int parentLine=getParent(selLine);
	Node* parentNode= g_sceneTreeList[parentLine].node.p;
	int fieldIdx=0, element=0;

	getFNodeElementForParent(parentLine, selLine, g_sceneTreeList, fieldIdx, element);

	Field* field=parentNode->getField(fieldIdx);

	if(baseCap && baseRad > 0 || topCap && topRad > 0){
		createUniqueName(patch);
		if(stem->textureMapper.refValue().length){
			createUniqueName(stem->textureMapper.refValue()[0]);
		}
	}

	patch->processFields();

	if(field->getType()==MFNODE){
		MFNode *mfnode=dynamic_cast<MFNode*>(field);
		mfnode->refValue()[element]=patch;
		*mfnode=mfnode->refValue();
	}
	if(getSeamless2()->animCP){
		patch->tpoint = patch->point;
	}
	g_stw->commonPostInsertNodeStuff();
}
void convertToPatch(){
	M_Properties& properties = g_lookupProtoInterface[CONVERTTOPATCH]->properties;

	properties[0].set_disable(false); //ctrlStaves
	properties[1].set_disable(false); //ctrlBands

	if(Ptr<NurbsLathe> nurbsLathe = dynamic_cast<NurbsLathe*>(g_selNode.getRaw())){
		properties[1].set_disable(true); //ctrlBands
		g_convertToPatch->convertNode=nurbsLathe;
		nurbsLathe->convertToPatch();
		g_stw->cp->setupNodeWindow(g_convertToPatch);
		Window::activeWindow=g_stw->cp.p;
	}
	if(Ptr<BezierLatheBuild> bezierLathe = dynamic_cast<BezierLatheBuild*>(g_selNode.getRaw())){
		properties[1].set_disable(true); //ctrlBands
		bezierLathe->convertToPatch();
		g_convertToPatch->convertNode=bezierLathe;
		g_stw->cp->setupNodeWindow(g_convertToPatch);
		Window::activeWindow=g_stw->cp.p;
	}
	else if(Ptr<BendyLathe> bendyLathe = dynamic_cast<BendyLathe*>(g_selNode.getRaw())){
		properties[1].set_disable(true); //ctrlBands
		g_convertToPatch->convertNode=bendyLathe;
		bendyLathe->convertToPatch();
		g_stw->cp->setupNodeWindow(g_convertToPatch);
		Window::activeWindow=g_stw->cp.p;
	}
	else if(Ptr<SphereBuild> sphereBuild = dynamic_cast<SphereBuild*>(g_selNode.getRaw())){
		g_convertToPatch->convertNode=sphereBuild;
		sphereBuild->convertToPatch();
		g_stw->cp->setupNodeWindow(g_convertToPatch);
		Window::activeWindow=g_stw->cp.p;
	}
	else if(Ptr<CylinderBuild> cylinderBuild = dynamic_cast<CylinderBuild*>(g_selNode.getRaw())){
		cylinderBuild->convertToPatch();
		g_convertToPatch->convertNode=cylinderBuild;
		g_stw->cp->setupNodeWindow(g_convertToPatch);
		Window::activeWindow=g_stw->cp.p;
	}
	else if(Ptr<ConeBuild> coneBuild = dynamic_cast<ConeBuild*>(g_selNode.getRaw())){
		coneBuild->convertToPatch();
		g_convertToPatch->convertNode=coneBuild;
		g_stw->cp->setupNodeWindow(g_convertToPatch);
		Window::activeWindow=g_stw->cp.p;
	}
	else if(Ptr<TorusBuild> torusBuild = dynamic_cast<TorusBuild*>(g_selNode.getRaw())){
		torusBuild->convertToPatch();
		g_convertToPatch->convertNode=torusBuild;
		g_stw->cp->setupNodeWindow(g_convertToPatch);
		Window::activeWindow=g_stw->cp.p;
	}
	else if(Ptr<RectangleBuild> rectangleBuild = dynamic_cast<RectangleBuild*>(g_selNode.getRaw())){
		rectangleBuild->convertToPatch();
		g_convertToPatch->convertNode=rectangleBuild;
		g_stw->cp->setupNodeWindow(g_convertToPatch);
		Window::activeWindow=g_stw->cp.p;
		g_convertToPatch->ctrlStaves=8;
		g_convertToPatch->ctrlBands=8;
	}
	else if(Ptr<BoxBuild> boxBuild = dynamic_cast<BoxBuild*>(g_selNode.getRaw())){
		boxBuild->convertToPatch();
		g_convertToPatch->convertNode=boxBuild;
		g_stw->cp->setupNodeWindow(g_convertToPatch);
		Window::activeWindow=g_stw->cp.p;
		g_convertToPatch->ctrlStaves=4;
		g_convertToPatch->ctrlBands=4;
	}
}
void convertToBendy(){
	if(NurbsLathe* nurbsLathe = dynamic_cast<NurbsLathe*>(g_selNode.getRaw())){
		nurbsLathe->convertToBendy();
	}
}

void commonPostTug();
void updateCage(Seamless* smls);
inline Matrix dxWindMatrix(Matrix& parentMatrix, Vec3f& scale, Rotation& rotation, Vec3f& translation);

void TransformCP::SelectRing::operator=(const bool& par){
	refValue() = par;
	home.selectRings.refValue() = false;
	home.sameScreenPosition.refValue() = false;
	if(par){
		MultiSelect::pointList.length = 0;
	}
	home.refreshRemPoints();
	houseKeepVisualClue();
	g_stw->cp->paint();
}

void TransformCP::SelectRings::operator=(const bool& par){
	refValue() = par;
	home.selectRing.refValue() = false;
	home.sameScreenPosition.refValue() = false;
	if(par){
		MultiSelect::pointList.length = 0;
	}
	home.refreshRemPoints();
	houseKeepVisualClue();
	g_stw->cp->paint();
}

void TransformCP::SameScreenPosition::operator=(const bool& par){
	refValue() = par;
	home.selectRings.refValue() = false;
	g_stw->cp->paint();
}

void TransformCP::RotAxisX::operator=(const bool& par){
	refValue() = par;
	home.rotAxisY.refValue() = false;
	home.rotAxisZ.refValue() = false;
}

void TransformCP::RotAxisY::operator=(const bool& par){
	refValue() = par;
	home.rotAxisX.refValue() = false;
	home.rotAxisZ.refValue() = false;
}

void TransformCP::RotAxisZ::operator=(const bool& par){
	refValue() = par;
	home.rotAxisY.refValue() = false;
	home.rotAxisX.refValue() = false;
}

void TransformCP::processFields(){
	if(MultiSelect::beginDragPos.length != MultiSelect::pointList.length){
		refreshRemPoints();
	}
	if(NurbsPatch* patch = dynamic_cast<NurbsPatch*>(g_selNode.getRaw())){
		Seamless* smls = patch->smls;

		Matrix matrix = translationMatrix(-pivotPoint) * rotationMatrix(scaleOrientation) * scaleMatrix(scale) * rotationMatrix(-scaleOrientation) * rotationMatrix(rotation) * translationMatrix(translation) * translationMatrix(pivotPoint);

		if(MultiSelect::pointList.length){
			for(int i = MultiSelect::pointList.length; --i>=0;){
				int idx = MultiSelect::pointList[i].i;
				NurbsPatch* patch = MultiSelect::pointList[i].p;
				Vec3f svec = matrix * MultiSelect::beginDragPos[i];
				patch->setFilteredDimension(patch->point.refValue(), idx, svec * patch->weight[idx]);
				patch->dpoint.refValue()[idx] = patch->point[idx] / patch->weight[idx];
				patch->mirrorSingleCP(idx);
			}
			IndexedLineSet* polyLine = dynamic_cast<IndexedLineSet*>(g_transformLeverClue->geometry.p);
			M_Vec3f& coord = dynamic_cast<Coordinate*>(polyLine->coord.p)->point;
			coord.length = 0;
			coord += g_transformCP->pivotPoint;
			coord += g_transformCP->rotCtrlPnt;
			fillPolyLineIndex(coord, polyLine->coordIndex.refValue());
			polyLine->convertToRenderFormat(g_transformLeverClue->render);

			smls->calcAllPolyMidPoints();
			commonPostTug();
			updateCage(smls);
			patch->houseKeep();
			return;
		}
	}
}

void TransformCP::refreshRemPoints(){
	scaleOrientation.refValue() = Rotation();
	scale.refValue() = Vec3f(1, 1, 1);
	rotation.refValue() = Rotation();
	pivotPoint.refValue() += translation.refValue();
	translation.refValue() = Vec3f();
	remRotCtrlPnt = rotCtrlPnt;

	MultiSelect::beginDragPos.length = MultiSelect::pointList.length;
	for(int i = MultiSelect::pointList.length; --i >= 0;){
		NurbsPatch* patch = MultiSelect::pointList[i].p;
		int idx = MultiSelect::pointList[i].i;
		MultiSelect::beginDragPos[i] = patch->dpoint[idx];
	}
}

void TransformCP::calcPivotPoint(){
	Vec3f& pivotPoint = this->pivotPoint.refValue();

	pivotPoint = Vec3f(0, 0, 0);
	scale.refValue() = Vec3f(1, 1, 1);
	scaleOrientation.refValue() = Rotation();
	rotation.refValue() = Rotation();
	translation.refValue() = Vec3f();

	if(MultiSelect::pointList.length == 0){
		return;
	}

	pivotPoint = Vec3f();
	MultiSelect::beginDragPos.length = MultiSelect::pointList.length;
	for(int i = MultiSelect::pointList.length; --i >= 0;){
		NurbsPatch* patch = MultiSelect::pointList[i].p;
		int idx = MultiSelect::pointList[i].i;
		Vec3f v = patch->dpoint[idx];
		MultiSelect::beginDragPos[i] = v;
		pivotPoint += v;
	}
	pivotPoint = pivotPoint / MultiSelect::beginDragPos.length;

	IndexedLineSet* polyLine = dynamic_cast<IndexedLineSet*>(g_transformLeverClue->geometry.p);
	M_Vec3f& coord = dynamic_cast<Coordinate*>(polyLine->coord.p)->point;
	rotCtrlPnt = g_scene->getViewMatrix() * pivotPoint + leverAxis * .5;
	remRotCtrlPnt = rotCtrlPnt;
	coord.length = 0;
	coord += pivotPoint;
	coord += rotCtrlPnt;
	fillPolyLineIndex(coord, polyLine->coordIndex.refValue());
	polyLine->convertToRenderFormat(g_transformLeverClue->render);
}

void snapPivotPointToControlPoint(){
	if(MultiSelect::clickedIndex < MultiSelect::beginDragPos.length){
		Vec3f v = g_scene->getViewMatrix() * MultiSelect::beginDragPos[MultiSelect::clickedIndex];
		Vec3f pp = g_scene->getViewMatrix() * g_transformCP->pivotPoint;
		pp.y = v.y;

		g_transformCP->refreshRemPoints();
		g_transformCP->rotCtrlPnt = pp + Vec3f(.5, 0, 0);
		g_transformCP->pivotPoint = g_scene->getIViewMatrix() * pp;
	}
}

void openTransformCP(){
	if(dynamic_cast<NurbsPatch*>(g_selNode.getRaw())){
		MultiSelect::beginDragPos.length = MultiSelect::pointList.length;

		g_transformCP->dragPivotPointMode = false;
		g_transformCP->dragRotPointMode = false;


		g_stw->cp->setupNodeWindow(g_transformCP);
		Window::activeWindow=g_stw->cp.p;
	}
	//else if(Oscillator* osc = dynamic_cast<Oscillator*>(g_selNode.getRaw())){
	//	M_Vec2f& fPoint = osc->note.refValue();
	//	for(int i = fPoint.length; --i>=0;){
	//		fPoint[i].y = fPoint[i].y + osc->transpose;
	//	}
	//	osc->processFields();
	//}
}

void stretchCurve(){//STRETCH_CURVE
	g_stw->cp->setupNodeWindow(g_stretchCurve);
	Window::activeWindow=g_stw->cp.p;
}
void ConvertToPatch::processFields(){
	if(NurbsLathe* nurbsLathe = dynamic_cast<NurbsLathe*>(convertNode.p)){
		nurbsLathe->ctrlStaves=ctrlStaves;
	}
	else if(BezierLatheBuild* bezierLathe = dynamic_cast<BezierLatheBuild*>(convertNode.p)){
		bezierLathe->ctrlStaves=ctrlStaves;
	}
	else if(BendyLathe* bendyLathe = dynamic_cast<BendyLathe*>(convertNode.p)){
		bendyLathe->ctrlStaves=ctrlStaves;
	}
	else if(SphereBuild* sphereBuild = dynamic_cast<SphereBuild*>(convertNode.p)){
		sphereBuild->ctrlStaves=ctrlStaves;
		sphereBuild->ctrlBands=ctrlBands;
	}
	else if(CylinderBuild* cylinderBuild = dynamic_cast<CylinderBuild*>(convertNode.p)){
		cylinderBuild->ctrlStaves=ctrlStaves;
		cylinderBuild->ctrlBands=ctrlBands;
	}
	else if(ConeBuild* coneBuild = dynamic_cast<ConeBuild*>(convertNode.p)){
		coneBuild->ctrlStaves=ctrlStaves;
		coneBuild->ctrlBands=ctrlBands;
	}
	else if(TorusBuild* torusBuild = dynamic_cast<TorusBuild*>(convertNode.p)){
		torusBuild->ctrlStaves=ctrlStaves;
		torusBuild->ctrlBands=ctrlBands;
	}
	else if(RectangleBuild* rectangleBuild = dynamic_cast<RectangleBuild*>(convertNode.p)){
		rectangleBuild->ctrlStaves=ctrlStaves;
		rectangleBuild->ctrlBands=ctrlBands;
	}
	else if(BoxBuild* boxBuild = dynamic_cast<BoxBuild*>(convertNode.p)){
		boxBuild->ctrlStaves=ctrlStaves;
		boxBuild->ctrlBands=ctrlBands;
	}
	convertNode->processFields();
	if(NurbsPatch* patch = dynamic_cast<NurbsPatch*>(g_selNode.getRaw())){
		patch->smls->populateMirrorNodes();
	}
	houseKeepVisualClue();
}

void NurbsPatch::rotateIndexOrder90(){
	M_Vec3f& point = this->point.refValue();
	M_Float& weight = this->weight.refValue();
	M_Vec3f _point = point;
	M_Float _weight = weight;
	int uDim = dim.x;
	int vDim = dim.y;

	int i = 0;

	for(int u = uDim; --u >= 0;){
		for(int v = 0; v < vDim; v++){
			point[i] = _point[v * uDim + u];
			weight[i] = _weight[v * uDim + u];
			i++;
		}
	}
	dim.refValue() = Vec2f(vDim, uDim);
	M_Float knotStore = vKnot;
	vKnot.refValue() = uKnot;
	uKnot.refValue() = knotStore;
}
void NurbsPatch::rotateIndexOrder180(){
	M_Vec3f& point = this->point.refValue();
	M_Float& weight = this->weight.refValue();
	M_Vec3f _point = point;
	M_Float _weight = weight;
	for(int i = _point.length; --i>=0;){
		point[i] = _point[point.length - 1 - i];
		weight[i] = _weight[point.length - 1 - i];
	}
}

void NurbsPatch::rotateIndexOrder270(){
	M_Vec3f& point = this->point.refValue();
	M_Float& weight = this->weight.refValue();
	M_Vec3f _point = point;
	M_Float _weight = weight;
	int uDim = dim.x;
	int vDim = dim.y;

	int i = 0;

	for(int u = 0; u < uDim; u++){
		for(int v = vDim; --v >= 0;){
			point[i] = _point[v * uDim + u];
			weight[i] = _weight[v * uDim + u];
			i++;
		}
	}
	dim.refValue() = Vec2f(vDim, uDim);
	M_Float knotStore = vKnot;
	vKnot.refValue() = uKnot;
	uKnot.refValue() = knotStore;
}


void rotate_uv_90(){
	NurbsPatch* patch = dynamic_cast<NurbsPatch*>(g_selNode.getRaw());
	M_Vec3f& point=patch->point.refValue();
	M_Float& weight=patch->weight.refValue();

	M_Vec3f c_point=point;
	M_Float c_weight=weight;
	int uDim = patch->dim.x;
	int vDim = patch->dim.y;

	//for(int i = c_point.length; --i>=0;){
	//	point[i]=c_point[c_point.length-1-i];
	//	weight[i]=c_weight[c_point.length-1-i];
	//}
	//for(int r = 0; r < vDim; r++){
	//	for(int c = 0; c < uDim; c++){
	//		int rev = vDim-r-1;
	//		point[r*vDim + c] = c_point[rev * vDim + c];
	//		weight[r*vDim + c] = c_weight[rev * vDim + c];
	//	}
	//}

	patch->rotateIndexOrder90();

	patch->processFields();
	houseKeepVisualClue();
	updateBuild3dWindow();
}
void reduceCapturedPoses(){
	g_stw->cp->setupNodeWindow(g_reduceCapturedPoses);
	Window::activeWindow=g_stw->cp.p;
}

void ReduceCapturedPoses::processFields(){
	Anim* anim = Window::animMode();
	M_Node& interpolator = anim->children;
	int keepCaptureForEvery = g_reduceCapturedPoses->keepCaptureForEvery;

	for(int k = interpolator.length; --k>=0;){
		if(OrientationInterpolator* oi = dynamic_cast<OrientationInterpolator*>(interpolator[k].p)){
			M_Float& key = oi->key;
			M_Rotation& keyValue = oi->keyValue;

			int i = 1;
			int e = keepCaptureForEvery;

			for(; e<key.length; i++, e+=keepCaptureForEvery){
				key[i] = key[e];
				keyValue[i] = keyValue[e];
			}
			key[i] = key[key.length-1];
			keyValue[i] = keyValue[keyValue.length-1];
			key.length = i + 1;
			keyValue.length = i + 1;
		}
		else if(PositionInterpolator* pi = dynamic_cast<PositionInterpolator*>(interpolator[k].p)){
			M_Float& key = pi->key;
			M_Vec3f& keyValue = pi->keyValue;

			int i = 1;
			int e = keepCaptureForEvery;

			for(; e<key.length; i++, e+=keepCaptureForEvery){
				key[i] = key[e];
				keyValue[i] = keyValue[e];
			}
			key[i] = key[key.length-1];
			keyValue[i] = keyValue[keyValue.length-1];
			key.length = i + 1;
			keyValue.length = i + 1;
		}
	}
}

void convertPivotPointsToTranslations(Art* part, Vec3f accumTran){
	accumTran+=part->pivotPoint;
	part->translation.refValue() += part->pivotPoint;
	M_Vec3f& coord = part->coord;
	coord.translate(-accumTran);
	part->pivotPoint.refValue()=Vec3f();
	part->absolutePivotPoint=Vec3f();
	M_SArt& m_part = part->part.refValue();
	for(int k = 0; k < m_part.length; k++){
		convertPivotPointsToTranslations(m_part[k], accumTran);
	}
}
void convertPivotPointsToTranslations(Seamless* smls){
	M_SArt& skeleton = smls->skeleton.refValue();
	for(int k = 0; k < skeleton.length; k++){
		convertPivotPointsToTranslations(skeleton[k], Vec3f());
	}
}
void convertPivotPointsToTranslations(){
	Seamless* smls = dynamic_cast<Seamless*>(g_selNode.getRaw());
	convertPivotPointsToTranslations(smls);
	smls->convertToRenderFormat();
}
/////////////
void convertTranslationsToPivotPoints(Art* part, Vec3f accumTran){
	accumTran+=part->translation;
	part->pivotPoint.refValue() += part->translation;
	M_Vec3f& coord = part->coord;
	coord.translate(accumTran);
	part->translation.refValue()=Vec3f();
	part->absolutePivotPoint=accumTran;
	M_SArt& m_part = part->part.refValue();
	for(int k = 0; k < m_part.length; k++){
		convertTranslationsToPivotPoints(m_part[k], accumTran);
	}
}
void convertTranslationsToPivotPoints(Seamless* smls){
	M_SArt& skeleton = smls->skeleton.refValue();
	for(int k = 0; k < skeleton.length; k++){
		convertTranslationsToPivotPoints(skeleton[k], Vec3f());
	}
}
void convertTranslationsToPivotPoints(){
	Seamless* smls = dynamic_cast<Seamless*>(g_selNode.getRaw());
	convertTranslationsToPivotPoints(smls);
	smls->convertToRenderFormat();
}
void convertMercatorToNorthPole(M_Vec3f& coord);
void convertMercatorToSouthPole(M_Vec3f& coord);
void convertNorthPoleToMercator(M_Vec3f& coord);
void convertSouthPoleToMercator(M_Vec3f& coord);
void convertPoleToPole(M_Vec3f& coord);

void mercatorToNorthPole(){
	if(ImmediatePart* part = dynamic_cast<ImmediatePart*>(g_selNode.getRaw())){
		convertMercatorToNorthPole(part->coord.refValue());
		Seamless* smls = part->smls;
		smls->ctrf=true;
		smls->doBuild();
		smls->convertToRenderFormat();
		BackUp::monostable.beginCycle();
	}
}
void mercatorToSouthPole(){
	if(ImmediatePart* part = dynamic_cast<ImmediatePart*>(g_selNode.getRaw())){
		convertMercatorToSouthPole(part->coord.refValue());
		Seamless* smls = part->smls;
		smls->ctrf=true;
		smls->doBuild();
		smls->convertToRenderFormat();
		BackUp::monostable.beginCycle();
	}
}
void northPoleToMercator(){
	if(ImmediatePart* part = dynamic_cast<ImmediatePart*>(g_selNode.getRaw())){
		convertNorthPoleToMercator(part->coord.refValue());
		Seamless* smls = part->smls;
		smls->ctrf=true;
		smls->doBuild();
		smls->convertToRenderFormat();
		BackUp::monostable.beginCycle();
	}
}
void southPoleToMercator(){
	if(ImmediatePart* part = dynamic_cast<ImmediatePart*>(g_selNode.getRaw())){
		convertSouthPoleToMercator(part->coord.refValue());
		Seamless* smls = part->smls;
		smls->ctrf=true;
		smls->doBuild();
		smls->convertToRenderFormat();
		BackUp::monostable.beginCycle();
	}
}
void poleToPole(){
	if(ImmediatePart* part = dynamic_cast<ImmediatePart*>(g_selNode.getRaw())){
		convertPoleToPole(part->coord.refValue());
		Seamless* smls = part->smls;
		smls->ctrf=true;
		smls->doBuild();
		smls->convertToRenderFormat();
		BackUp::monostable.beginCycle();
	}
}
void copyColorToPolyMap(){
	if(Stem* stem = dynamic_cast<Stem*>(g_selNode.getRaw())){
		if(stem->source.length){
			if(ImmediatePart* bluePart = dynamic_cast<ImmediatePart*>(stem->source.refValue()[0].p)){
				if(stem->part.length){
					if(Art* whitePart = dynamic_cast<Art*>(stem->part.refValue()[0].p)){
						if(ColorEffect* colorEffect = dynamic_cast<ColorEffect*>(whitePart->workEffect)){
							M_Triangle& whiteTri = whitePart->triangle;
							M_Triangle& blueTri = bluePart->triangle;
							M_Vec3f& whiteCoord = whitePart->coord.refValue();
							M_Vec3f& blueCoord = bluePart->coord.refValue();
							M_Color3f& m_color = colorEffect->color.refValue();
							if(whiteTri.length == blueTri.length && whiteCoord.length == blueCoord.length){
								for(int i = whiteTri.getTriangleLength(); --i>=0;){
									m_color[blueTri.refTexIndex(i)] = m_color[whiteTri.refTexIndex(i)];
								}
							}
							Seamless* smls = bluePart->smls;
							smls->doBuild();
							smls->ctrf=true;
							BackUp::monostable.beginCycle();
							return;
						}
					}
				}
			}
		}
		else if(Art* whitePart = dynamic_cast<Art*>(stem->part.refValue()[0].p)){
			if(ColorEffect* colorEffect = dynamic_cast<ColorEffect*>(whitePart->workEffect)){

				M_Triangle& whiteTri = whitePart->triangle;
				M_Int& coordIndex = stem->coordIndex;
				M_Color3f& dstColor = stem->color.refValue();
				M_Color3f& m_color = colorEffect->color.refValue();
				int beginTriOffset = stem->beginTriOffset * 3;
				for(int i = 0; i < coordIndex.length; i++){
					dstColor[coordIndex[i]] = m_color[whiteTri.refTexIndex(beginTriOffset + i)];
				}
				Seamless* smls = whitePart->smls;
				smls->doBuild();
				smls->ctrf=true;
				BackUp::monostable.beginCycle();
				return;
			}
		}
	}
	msG("Stem must have source and destination parts with equal triangle and coord length");
}


Art* storeGeneratedPolyMap(Stem* stem){
	if(stem->part.length){
		Art* whitePart = stem->part[0].p;
		ImmediatePart* bluePart = selectPolyMapPart(whitePart);
		stem->copyPolyMapToPart(bluePart);

		Stem tmpStem;
		tmpStem.copyFields(stem);
		tmpStem.source.refValue().length=0;
		tmpStem.part.refValue().length=0;
		tmpStem.part.refValue()+=bluePart;
		tmpStem.mirror.refValue()=false;
		tmpStem.add.refValue()=false;
		tmpStem.mould.refValue()=false;
		tmpStem.store.refValue()=false;
		tmpStem.doBuild();
		stem->copyPartToPolyMap(bluePart);
		return whitePart;
	}
	return 0;
}


void storeGeneratedPolyMap(){
	if(Stem* stem = dynamic_cast<Stem*>(g_selNode.getRaw())){
		if(stem->source.length){
			if(ImmediatePart* bluePart = dynamic_cast<ImmediatePart*>(stem->source.refValue()[0].p)){
#ifdef PROG_ROB
	dLog("new MoveToNode(" + String(g_stw->selLine) + "),");
	dLog("new RightClick,");
	dLog("new Move_MenuItem(menu_copy_poly_map_to_source),");
	dLog("new Click,");
#endif

				bluePart->deleteGeometry();
				Stem tmpStem;
				tmpStem.copyFields(stem);
				tmpStem.source.refValue().length=0;
				tmpStem.part.refValue().length=0;
				tmpStem.part.refValue()+=bluePart;
				tmpStem.mould.refValue()=false;
				tmpStem.add.refValue()=false;
				tmpStem.doBuild();

				bluePart->smls->doBuild();
				makePointerEqualIndexLocation();
				bluePart->smls->ctrf=true;
				Window::paintAllControlPanels();

			}
			else{
				msG("a blue part must be referenced for the source");
			}
		}
		else{
			if(Art* whitePart = storeGeneratedPolyMap(stem)){
				whitePart->smls->doBuild();
				makePointerEqualIndexLocation();
				whitePart->smls->ctrf=true;
				Window::paintAllControlPanels();
			}
		}
	}
}
void incToNextTriThatSharesActivePoint(){
	if(Art* part=dynamic_cast<Art*>(g_selNode.getRaw())){
		Seamless* smls = part->smls;
		smls->ActTri=findNextTriangleFromVertex(part, smls->actTriPart, smls->ActCtrlPnt, smls->ActTriCorner,  smls->ActTri);
		makePointerEqualIndexLocation();
	}
	else if(Subdivision* subdivision=dynamic_cast<Subdivision*>(g_selNode.getRaw())){
		subdivision->incToNextPolyThatSharesActivePoint();
	}
}
void decToNextTriThatSharesActivePoint(){
	if(Art* part=dynamic_cast<Art*>(g_selNode.getRaw())){
		Seamless* smls = part->smls;
		smls->ActTri=findPrevTriangleFromVertex(part, smls->actTriPart, smls->ActCtrlPnt, smls->ActTriCorner,  smls->ActTri);
		makePointerEqualIndexLocation();
	}
	else if(Subdivision* subdivision=dynamic_cast<Subdivision*>(g_selNode.getRaw())){
		subdivision->decToNextPolyThatSharesActivePoint();
	}
}

void convertToSubdiv(NurbsPatch* patch, M_Vec3f& point, M_Int& coordIndex, int& offSet);

void convertToSubdiv(FuseSurface* fuseSurface, M_Vec3f& point, M_Int& coordIndex, int& offSet){
	if(NurbsPatch* patch=dynamic_cast<NurbsPatch*>(fuseSurface->patch.refValue().p)){
		convertToSubdiv(patch, point, coordIndex, offSet);
	}
}
void convertToSubdiv(Stem* stem, M_Vec3f& point, M_Int& coordIndex, int& offSet){
	M_Node& fuse = stem->fuse.refValue();
	for(int k = fuse.length; --k>=0;){
		convertToSubdiv(dynamic_cast<FuseSurface*>(fuse[k].p), point, coordIndex, offSet);
	}
}
void convertToSubdiv(NurbsPatch* patch, M_Vec3f& point, M_Int& coordIndex, int& offSet){
	point += patch->dpoint;
	int uDim = patch->dim.x;
	int vDim = patch->dim.y;

	for(int v = 0; v < vDim-1; v++){
		for(int u = 0; u < uDim - 1; u+=2){
			coordIndex += offSet + v * uDim + u + 0;
			coordIndex += offSet + v * uDim + u + 2;
			coordIndex += offSet + v * uDim + u + 2 + uDim;
			coordIndex += offSet + v * uDim + u + 0 + uDim;
		}
	}
	offSet += patch->dpoint.length;

	M_Node& m_stem = patch->stem.refValue();
	for(int k = m_stem.length; --k>=0;){
		convertToSubdiv(dynamic_cast<Stem*>(m_stem[k].p), point, coordIndex, offSet);
	}
}
void convertToSubdiv(){
	if(NurbsPatch* patch = dynamic_cast<NurbsPatch*>(g_selNode.getRaw())){
		Subdivision* subdiv = new Subdivision;
		M_Vec3f& point = subdiv->point.refValue();
		M_Int& coordIndex = subdiv->coordIndex.refValue();
		coordIndex.length = 0;
		point.length = 0;

		int offSet = 0;
		convertToSubdiv(patch, point, coordIndex, offSet);
		subdiv->processFields();



		int selLine = g_stw->selLine;
		int parentLine=getParent(selLine);
		Node* parentNode= g_sceneTreeList[parentLine].node.p;
		int fieldIdx=0, element=0;

		getFNodeElementForParent(parentLine, selLine, g_sceneTreeList, fieldIdx, element);

		Field* field=parentNode->getField(fieldIdx);


		if(field->getType()==MFNODE){
			MFNode *mfnode=dynamic_cast<MFNode*>(field);
			mfnode->refValue()[element]=subdiv;
			*mfnode=mfnode->refValue();
		}
		g_stw->commonPostInsertNodeStuff();
	}
}
void compileBuildTransform(){
	int k = g_stw->selLine;
	int indent = g_sceneTreeList[k].indent;
	Seamless* smls = dynamic_cast<Seamless*>(g_sceneTreeList[k].node.p);
	Matrix matrix = scaleMatrix(smls->scale) * rotationMatrix(smls->rotation) * translationMatrix(smls->translation);
	
	M_SArt& skeleton = smls->skeleton.refValue();
	for(int k = skeleton.length; --k>=0;){
		Art* part = skeleton[k];
		part->pivotPoint.refValue() += smls->translation;
	}
	while(++k < g_sceneTreeList.length && g_sceneTreeList[k].indent > indent){
		Node* node = g_sceneTreeList[k].node;
		if(NurbsPatch* patch = dynamic_cast<NurbsPatch*>(node)){
			if(patch->translation==Vec3f()){

				M_Vec3f& point = patch->point.refValue();
				M_Vec3f& dpoint = patch->dpoint.refValue();
				M_Float& weight = patch->weight.refValue();

				int uDim = patch->dim.x;
				int vDim = patch->dim.y;

				for(int i = uDim*vDim; --i>=0;){
					point[i] = matrix * dpoint[i] * weight[i];
				}
				patch->scale.refValue() = Vec3f(1,1,1);
				patch->rotation.refValue() = Rotation();
				patch->translation.refValue() = Vec3f();
			}
			else{
				patch->translation.refValue() += smls->translation;
			}
			patch->processFields();
		}
		else if(NurbsLathe* lathe = dynamic_cast<NurbsLathe*>(node)){
			lathe->translation.refValue() += smls->translation;
			lathe->processFields();
		}
		else if(BendyLathe* lathe = dynamic_cast<BendyLathe*>(node)){
			lathe->translation.refValue() += smls->translation;
			lathe->processFields();
		}
		else if(BezierLatheBuild* lathe = dynamic_cast<BezierLatheBuild*>(node)){
			lathe->translation.refValue() += smls->translation;
			lathe->processFields();
		}
		else if(TorusBuild* torus = dynamic_cast<TorusBuild*>(node)){
			torus->translation.refValue() += smls->translation;
			torus->processFields();
		}
		else if(SphereBuild* sphereBuild = dynamic_cast<SphereBuild*>(node)){
			sphereBuild->translation.refValue() += smls->translation;
			sphereBuild->processFields();
		}
		else if(CylinderBuild* cylinderBuild = dynamic_cast<CylinderBuild*>(node)){
			cylinderBuild->translation.refValue() += smls->translation;
			cylinderBuild->processFields();
		}
		else if(ConeBuild* coneBuild = dynamic_cast<ConeBuild*>(node)){
			coneBuild->translation.refValue() += smls->translation;
			coneBuild->processFields();
		}
		else if(ColorSweep* colorSweep = dynamic_cast<ColorSweep*>(node)){
			if(dynamic_cast<Stem*>(g_sceneTreeList[getParent(k)].node.p) == 0){
				colorSweep->location.refValue() += smls->translation;
			}
		}
	}
	smls->scale.refValue() = Vec3f(1,1,1);
	smls->rotation.refValue() = Rotation();
	smls->translation.refValue() = Vec3f();
	smls->process();
	smls->doBuild();
	smls->ctrf=true;
	BackUp::monostable.beginCycle();
}
void saveRelWinMousePos(){
	Vec2i pos = g_mousePos - Window::activeWindow->pos; 
	dLog("new MoveRelWindow(" + String(pos.x) + ", " + String(pos.y) + "),");
}

extern Coordinate* g_vcSkeletonLinesCoordinate;
extern Material* g_stickManMaterial;

void StickMan::addLineForBone(M_SArt& m_part, int parentLineIndex, M_Vec3f& point, M_Int& coordIndex, bool& beginNewLine){
	for(int k = 0; k < m_part.length; k++){
		Art* part = m_part[k];
		if(beginNewLine){
			if(coordIndex.length > 0){
				coordIndex += -1;
			}
			coordIndex += parentLineIndex;
			beginNewLine = false;
		}
		int last = point.length;
		coordIndex += last;
		partList += part;
		point += part->pivotPointMatrix * Vec3f(0,0,0);

		if(part->part.length){
			addLineForBone(part->part.refValue(), last, point, coordIndex, beginNewLine);
		}
		beginNewLine = true;
	}
	beginNewLine = true;
}


void StickMan::process(Seamless* smls){
	smls->stickMan = this;

	M_Vec3f& point = coord->point.refValue();
	M_Int& coordIndex = ils->coordIndex.refValue();

	point.length = 0;
	coordIndex.length = 0;
	partList.length = 0;

	bool beginNewLine = true;

	M_SArt& skeleton = smls->skeleton.refValue();
	for(int k = 0; k < skeleton.length; k++){
		Art* part = skeleton[k];
		if(part->part.length){
			partList += part;
			point += part->pivotPointMatrix * Vec3f(0,0,0);
			addLineForBone(part->part.refValue(), point.length-1, point, coordIndex, beginNewLine);
		}
	}
	if(partList.length == 0){
		if(smls->partList.length == 1){
			partList+= smls->partList[0];
		}
	}
	//Color* color = new Color;
	//color->color.refValue().length = point.length;
	//ils->color = color;
	//for(int i = color->color.refValue().length; --i >= 0;){
	//	color->color.refValue()[i] = Color3f(1,1,0);
	//}

	if(partList.length >= 2 && g_bReady){
		ils->convertToRenderFormat(shape->render);
	}
}

void Seamless::XRay::operator=(const bool& par){
	refValue() = par;
	if(par){
		home.stickMan = new StickMan;
		home.stickMan->process(&home);
	}
	else{
		delete home.stickMan;
		home.stickMan = 0;
	}
}

StickMan::StickMan():
	ils(new IndexedLineSet),
	coord(new Coordinate),
	shape(new Shape)
{
	Appearance* appearance = new Appearance;
	appearance->material = g_stickManMaterial;
	shape->appearance = appearance;

	ils->coord = coord;
	shape->geometry = ils;
}

void toggleXRayEdit(){
	if(Window::animMode()){
		return;
	}
	g_scene->xRayEdit = !g_scene->xRayEdit;
}

void toggleXRayMode(){
	Seamless* smls = dynamic_cast<Seamless*>(g_selNode.getRaw());
	if(Art* part = dynamic_cast<Art*>(g_selNode.getRaw())){
		smls = part->smls;
	}
	if(smls){
		smls->xRay = !smls->xRay;
	}
	else{
		needPartOrSmlsMsg();
	}
}

void toggleBackfaceSelect(){
	MultiSelect::backfaceIgnore = !MultiSelect::backfaceIgnore;
	if(MultiSelect::backfaceIgnore){
		g_toggleBackfaceSelectButton->srcPos = Vec2i(24 * 5, 0);
	}
	else{
		g_toggleBackfaceSelectButton->srcPos = Vec2i(24 * 6, 0);
	}
}

void toggleAccumulateSelect(){
	MultiSelect::accumulate = !MultiSelect::accumulate;
	if(MultiSelect::accumulate){
		g_toggleAccumulateSelectButton->srcPos = Vec2i(24 * 11, 24);
	}
	else{
		g_toggleAccumulateSelectButton->srcPos = Vec2i(24 * 7, 24 * 3);
		MultiSelect::reset();
		houseKeepVisualClue();
	}
}

void manualCaptureIKPoints(Anim* anim, Art* part, Rotation accumRot){
	if(accumRot.angle != 0 || part->rotation.refValue().angle != 0){
		accumRot = part->rotation.refValue() * accumRot;
	}
	if(part->ik){
		part->ikPos = part->matrix * part->absolutePivotPoint;
		anim->captureSinglePartPose(part, &part->ikPos);
		if(part->pin){
			part->ikRot = part->rotation;
			anim->captureSinglePartPose(part, &part->ikRot);
		}
	}
	M_SArt& m_part = part->part.refValue();
	for(int k = m_part.length; --k>=0;){
		manualCaptureIKPoints(anim, m_part[k].p, part->rotation);
	}
}

void manualCaptureIKPoints(){
	if(Anim* anim = Window::animMode()){
		if(Art* part = dynamic_cast<Art*>(g_selNode.getRaw())){
			Seamless* smls = part->smls;
			M_SArt& skeleton = smls->skeleton.refValue();
			for(int k = skeleton.length; --k>=0;){
				manualCaptureIKPoints(anim, skeleton[k].p, smls->rotation);
			}
		}
	}
}

void manualCaptureIKPoint(){
	if(Anim* anim = Window::animMode()){
		if(Art* part = dynamic_cast<Art*>(g_selNode.getRaw())){
			if(part->ik){
				part->ikPos = part->matrix * part->absolutePivotPoint;
				anim->captureSinglePartPose(part, &part->ikPos);
			}
		}
	}
}

void captureFrame(){
	if(Anim* anim = Window::animMode()){
		if(Art* part = dynamic_cast<Art*>(g_selNode.getRaw())){
			//anim->captureMorph(part->smls);
			anim->captureSinglePartPose(part, &part->rotation);
		}
		else{
			needPartMsg();
		}
	}
}

void addOscToOsc(Oscillator* fromOsc, Oscillator* toOsc){
	toOsc->note.refValue() += fromOsc->note.refValue();
	toOsc->fWeight.refValue() += fromOsc->fWeight.refValue();
	updateNurbsCurveKnotMid0(toOsc->note.length, toOsc->fKnot.refValue());
	toOsc->processFields();
	toOsc->houseKeep();
}
void scaleOsc(Oscillator* osc){
	M_Vec2f& fPoint = osc->note;
	Vec2f pivotPoint;
	Vec2f scale;
	Vec2f translate;
	pivotPoint.x = g_transformCP->pivotPoint.x;
	pivotPoint.y = g_transformCP->pivotPoint.y;
	scale.x = g_transformCP->scale.x;
	scale.y = g_transformCP->scale.y;
	translate.x = g_transformCP->translation.x;
	translate.y = g_transformCP->translation.y;
	for(int i = fPoint.length; --i>=0;){
		fPoint[i] = (fPoint[i] - pivotPoint) * scale + pivotPoint + translate;
	}
	if(osc->delay != 0){
		osc->delay.refValue() = (osc->delay.refValue() - pivotPoint.x) * scale.x + pivotPoint.x + translate.x;
	}
	osc->processFields();
}

void showSingleCage(){
	if(NurbsPatch* patch = dynamic_cast<NurbsPatch*>(g_selNode.getRaw())){
		Seamless* smls = patch->smls;
		M_<NurbsPatch*>& cpNodeList = smls->cpNodeList;
		for(int k = cpNodeList.length; --k >= 0;){
			cpNodeList[k]->cage = false;
		}
		for(int k = MultiSelect::pointList.length; --k >= 0;){
			MultiSelect::pointList[k].p->cage = true;
		}
		g_stw->paint();
		houseKeepVisualClue();
	}
	else if(Oscillator* osc = dynamic_cast<Oscillator*>(g_selNode.getRaw())){
		scaleOsc(osc);
		houseKeepVisualClue();
	}
	else if(Track* track = dynamic_cast<Track*>(g_selNode.getRaw())){
		M_Node& m_node = track->sound.refValue();
		for(int k = m_node.length; --k>=0;){
			if(Oscillator* osc = dynamic_cast<Oscillator*>(m_node[k].p)){
				scaleOsc(osc);
			}
		}
		houseKeepVisualClue();
	}
}

void showAllCages(){
	if(NurbsPatch* patch = dynamic_cast<NurbsPatch*>(g_selNode.getRaw())){
		Seamless* smls = patch->smls;
		M_<NurbsPatch*>& cpNodeList = smls->cpNodeList;
		for(int k = cpNodeList.length; --k >= 0;){
			cpNodeList[k]->cage = true;
		}
		//for(int k = MultiSelect::pointList.length; --k >= 0;){
		//	if(MultiSelect::pointList[k].p != patch){
		//		MultiSelect::pointList[k].p->cage = false;
		//	}
		//}
		//if(MultiSelect::pointList.length > 1){
		//	MultiSelect::pointList.length = 0;
		//}
		g_stw->paint();
		houseKeepVisualClue();
	}
	else if(Oscillator* toOsc = dynamic_cast<Oscillator*>(g_selNode.getRaw())){
		if(Oscillator* fromOsc = dynamic_cast<Oscillator*>(g_copyNode.getRaw())){
			addOscToOsc(fromOsc, toOsc);
		}
	}
}


void hideSelectedCages(){
	if(NurbsPatch* patch = dynamic_cast<NurbsPatch*>(g_selNode.getRaw())){
		Seamless* smls = patch->smls;
		M_<NurbsPatch*>& cpNodeList = smls->cpNodeList;
		for(int k = MultiSelect::pointList.length; --k >= 0;){
			MultiSelect::pointList[k].p->cage = false;
		}
		MultiSelect::pointList.length = 0;
		g_stw->paint();
		houseKeepVisualClue();
	}

}

void showCagesForStem(M_Node& m_stem, bool state){
	for(int k = 0; k < m_stem.length; k++){
		Stem* stem = dynamic_cast<Stem*>(m_stem[k].p);
		M_Node& m_fuse = stem->fuse.refValue();
		for(int k = 0; k < m_fuse.length; k++){
			if(FuseSurface* fuse = dynamic_cast<FuseSurface*>(m_fuse[k].p)){
				if(NurbsPatch* patch = dynamic_cast<NurbsPatch*>(fuse->patch.p)){
					patch->cage = state;
					showCagesForStem(patch->stem.refValue(), state);
				}
			}
		}
	}
}

void showCages(M_Build& m_build, bool state){
	for(int k = m_build.length; --k >= 0;){
		Build* build = m_build[k];

		if(NurbsPatch* patch = dynamic_cast<NurbsPatch*>(build)){
			patch->cage = state;
			showCagesForStem(patch->stem.refValue(), state);
		}
		else if(GroupBuild* groupBuild = dynamic_cast<GroupBuild*>(build)){
			showCages(groupBuild->build.refValue(), state);
		}
	}
}

void showCages(){
	GroupBuild* groupBuild = dynamic_cast<GroupBuild*>(g_sceneTreeList[g_stw->selLine].node.p);
	showCages(groupBuild->build.refValue(), true);
}

void hideCages(){
	GroupBuild* groupBuild = dynamic_cast<GroupBuild*>(g_sceneTreeList[g_stw->selLine].node.p);
	showCages(groupBuild->build.refValue(), false);
}

void NurbsPatch::removeSouthEdge(NurbsPatch* patchA){
	int uDim = dim.x;
	point.refValue().deleteElement(0, uDim);
	weight.refValue().deleteElement(0, uDim);
	dim.refValue().y--;
	updateKnot(dim.refValue().y, vKnot.refValue(), order.refValue().y);
	processFields();

	patchA->dim.refValue().y += dim.y;
	patchA->updateKnot(patchA->dim.refValue().y, patchA->vKnot.refValue(), patchA->order.refValue().y);
}

void NurbsPatch::removeNorthEdge(NurbsPatch* patchA){
	int uDimB = dim.x;
	int vDimB = dim.y;
	point.refValue().deleteElement((vDimB + 1) * uDimB, uDimB);
	weight.refValue().deleteElement((vDimB + 1) * uDimB, uDimB);
	dim.refValue().y--;
	updateKnot(dim.refValue().y, vKnot.refValue(), order.refValue().y);
	processFields();

	patchA->dim.refValue().y += dim.y;
	patchA->updateKnot(patchA->dim.refValue().y, patchA->vKnot.refValue(), patchA->order.refValue().y);
}

void NurbsPatch::removeWestEdge(NurbsPatch* patchA){
	M_Vec3f& point = this->point.refValue();
	M_Float& weight = this->weight.refValue();
	int uDim = dim.x;
	int vDim = dim.y;

	int i = 0;

	for(int v = 0; v < vDim; v++){
		for(int u = 0; u < uDim - 1; u++){
			point[i] = point[v * uDim + u];
			weight[i] = weight[v * uDim + u];
			i++;
		}
	}
	point.length = i;
	weight.length = i;

	patchA->dim.refValue().x--;
	patchA->updateKnot(patchA->dim.refValue().x, patchA->uKnot.refValue(), patchA->order.refValue().x);
	patchA->processFields();

	patchA->dim.refValue().x += dim.x;
	patchA->updateKnot(patchA->dim.refValue().x, patchA->uKnot.refValue(), patchA->order.refValue().x);
}

void NurbsPatch::removeEastEdge(NurbsPatch* patchA){
	M_Vec3f& pointB = point.refValue();
	M_Float& weightB = weight.refValue();
	int uDimB = dim.x;
	int vDimB = dim.y;

	int i = 0;

	for(int v = 0; v < vDimB; v++){
		for(int u = 1; u < uDimB; u++){
			pointB[i] = pointB[v * uDimB + u];
			weightB[i] = weightB[v * uDimB + u];
			i++;
		}
	}
	pointB.length = i;
	weightB.length = i;

	dim.refValue().x--;
	updateKnot(dim.refValue().x, uKnot.refValue(), order.refValue().x);
	processFields();

	patchA->dim.refValue().x += dim.x;
	patchA->updateKnot(patchA->dim.refValue().x, patchA->uKnot.refValue(), patchA->order.refValue().x);
}


void incorporate(){
	if(MultiSelect::pointList.length == 2){
		NurbsPatch* patchA = MultiSelect::pointList[0].p;
		int idxA = MultiSelect::pointList[0].i;
		int uDimA = patchA->dim.x;
		int vDimA = patchA->dim.y;
		NurbsPatch* patchB = MultiSelect::pointList[1].p;
		int uDimB = patchB->dim.x;
		int vDimB = patchB->dim.y;
		int idxB = MultiSelect::pointList[1].i;

		if(NurbsPatch::isEastEdge(idxA, uDimA, vDimA)){
			patchA->rotateIndexOrder270();
		}
		if(NurbsPatch::isSouthEdge(idxA, uDimA, vDimA)){
			patchA->rotateIndexOrder180();
		}
		else if(NurbsPatch::isWestEdge(idxA, uDimA, vDimA)){
			patchA->rotateIndexOrder90();
		}
		if(NurbsPatch::isNorthEdge(idxB, uDimB, vDimB)){
			patchB->rotateIndexOrder180();
		}
		else if(NurbsPatch::isWestEdge(idxB, uDimB, vDimB)){
			patchB->rotateIndexOrder270();
		}
		else if(NurbsPatch::isEastEdge(idxB, uDimB, vDimB)){
			patchB->rotateIndexOrder90();
		}
		patchB->removeSouthEdge(patchA);
		patchB->processFields();
		patchA->point.refValue() += patchB->point;
		patchA->weight.refValue() += patchB->weight;
		patchA->processFields();
		MultiSelect::pointList.length = 0;
		deleteNode(getNodeLine(patchB));
		g_stw->selLine = getNodeLine(patchA);
		g_stw->commonPostInsertNodeStuff();
		//updateBuild3dWindow();
	}
}

void convert_to_0_v_mid_points(NurbsPatch* patch){
	M_Vec3f& dpoint = patch->dpoint.refValue();
	M_Vec3f& point = patch->point.refValue();
	M_Float& weight = patch->weight.refValue();
	M_Vec3f _point = point;
	M_Float _weight = weight;
	int uDim = patch->dim.x;
	int vDim = patch->dim.y;


	int i = 0;

	for(int u = 0; u < uDim; u++){
		point[i] = _point[u];
		weight[i] = _weight[u];
		i++;
	}
	for(int v = 1; v < vDim - 1; v++){
		for(int u = 0; u < uDim; u++){
			point[i] = _point[v * uDim + u];
			weight[i] = _weight[v * uDim + u];
			i++;
		}
		v++;
	}
	for(int v = vDim - 1, u = 0; u < uDim; u++){
		point[i] = _point[v * uDim + u];
		weight[i] = _weight[v * uDim + u];
		i++;
	}

	vDim -= (vDim - 3) / 2;
	patch->dim.refValue().y = vDim;
	point.length = weight.length = vDim * uDim;
	updateNurbsCurveKnotMid0(vDim, patch->vKnot.refValue());
	patch->processFields();
}
void convert_to_1_v_mid_points(NurbsPatch* patch){

	M_Vec3f& dpoint = patch->dpoint.refValue();
	M_Vec3f& point = patch->point.refValue();
	M_Float& weight = patch->weight.refValue();
	M_Vec3f _point = point;
	M_Float _weight = weight;
	int uDim = patch->dim.x;
	int vDim = patch->dim.y;

	point.length += (vDim - 3) * uDim;
	weight.length += (vDim - 3) * uDim;


	int i = 0;

	for(int u = 0; u < uDim; u++){
		point[i] = _point[u];
		weight[i] = _weight[u];
		i++;
	}
	for(int v = 1; v < vDim - 1; v++){
		for(int u = 0; u < uDim; u++){
			point[i] = _point[v * uDim + u];
			weight[i] = _weight[v * uDim + u];
			i++;
		}
		for(int u = 0; u < uDim; u++){
			weight[i] = (_weight[v * uDim + u] + _weight[(v + 1) * uDim + u]) / 2;
			point[i] = (dpoint[v * uDim + u] + dpoint[(v + 1) * uDim + u]) / 2 * weight[i];
			//weight[i] = 1;
			i++;
		}
	}
	i -= uDim;
	for(int v = vDim - 1, u = 0; u < uDim; u++){
		point[i] = _point[v * uDim + u];
		weight[i] = _weight[v * uDim + u];
		i++;
	}


	vDim += vDim - 3;
	patch->dim.refValue().y = vDim;
	updateNurbsCurveKnotMid1(vDim, patch->vKnot.refValue());
	patch->processFields();
}

void convert_to_0_v_mid_points(){
	NurbsPatch* patch = dynamic_cast<NurbsPatch*>(g_selNode.getRaw());
	convert_to_0_v_mid_points(patch);
	resetCPNodeList();
	updateBuild3dWindow();
}

void convert_to_1_v_mid_points(){
	NurbsPatch* patch = dynamic_cast<NurbsPatch*>(g_selNode.getRaw());
	convert_to_1_v_mid_points(patch);
	resetCPNodeList();
	updateBuild3dWindow();
}

void convert_to_0_u_mid_points(){
	NurbsPatch* patch = dynamic_cast<NurbsPatch*>(g_selNode.getRaw());
	patch->rotateIndexOrder90();
	patch->processFields();
	convert_to_0_v_mid_points(patch);
	patch->rotateIndexOrder270();
	patch->processFields();
	resetCPNodeList();
	updateBuild3dWindow();
}


void convert_to_1_u_mid_points(){
	NurbsPatch* patch = dynamic_cast<NurbsPatch*>(g_selNode.getRaw());
	patch->rotateIndexOrder90();
	patch->processFields();
	convert_to_1_v_mid_points(patch);
	patch->rotateIndexOrder270();
	patch->processFields();
	resetCPNodeList();
	updateBuild3dWindow();
}

void calcAllPolyMidPoints(){
	if(NurbsPatch* patch = dynamic_cast<NurbsPatch*>(g_selNode.getRaw())){
		patch->smls->calcAllPolyMidPoints();
	}
	updateBuild3dWindow();
}

void fillEdgesWithCtrlPolys(){
	if(NurbsPatch* patch = dynamic_cast<NurbsPatch*>(g_selNode.getRaw())){
		patch->fillEdgesWithCtrlPolys();
		resetCPNodeList();
		houseKeepVisualClue();
	}
	else if(Composition* composition = dynamic_cast<Composition*>(g_selNode.getRaw())){
		composition->evolveSound();
	}
}

void mergeCtrlPolys(){
	if(NurbsPatch* patch=dynamic_cast<NurbsPatch*>(g_selNode.getRaw())){
		patch->smls->stitchControlPolygons();
	}
}

void mergeTo5CtrlPoly(){
	if(NurbsPatch* patch = dynamic_cast<NurbsPatch*>(g_selNode.getRaw())){
		patch->smls->mergeTo5CtrlPoly();
	}
}

void mergeTo6CtrlPoly(){
	if(NurbsPatch* patch = dynamic_cast<NurbsPatch*>(g_selNode.getRaw())){
		patch->smls->mergeTo6CtrlPoly();
	}
}

void toggleShowControlPolys(){
	g_showControlPolys = !g_showControlPolys;
	if(g_showControlPolys == false){
		MultiSelect::ctrlPolySelList.length = 0;
	}
	houseKeepVisualClue();
}

void linkPointToEdge(){
	if(NurbsPatch* patch = dynamic_cast<NurbsPatch*>(g_selNode.getRaw())){
		if(MultiSelect::pointList.length){
			if(MultiSelect::ctrlPolySelList.length){
				patch->smls->linkPointToEdge(MultiSelect::ctrlPolySelList[0]);
			}
		}
	}
}

void Stem::copyPartToPolyMap(Art* part){
	GeomSource geomSource;
	part->extractGeometryGS(geomSource, g_tmpIndex);

	M_Vec3f& srcCoord = geomSource.coord.refValue();
	M_Vec2f& dstCoord = this->coord.refValue();
	dstCoord.length = srcCoord.length;
	for(int i = srcCoord.length; --i>=0;){
		Vec3f vec3f = srcCoord[i];
		dstCoord[i] = Vec2f(vec3f.x, vec3f.y);
	}

	color.refValue() = geomSource.color;
	texCoord.refValue() = geomSource.texCoord;
	coordIndex.refValue() = geomSource.coordIndex;
	rc.refValue() = geomSource.rc;
}

void mirrorAllCtrlPoints(){
	if(NurbsPatch* nurbsPatch=dynamic_cast<NurbsPatch*>(g_selNode.getRaw())){
		M_Vec3f& point = nurbsPatch->point.refValue();
		M_Float& weight = nurbsPatch->weight.refValue();

		for(int i = point.length; --i >= 0;){
			point[i].x = -point[i].x;
		}

		nurbsPatch->processFields();
		updateBuild3dWindow();
		houseKeepVisualClue();
		BackUp::monostable.beginCycle();
	}
}

void mirrorCtrlPoints(){
	if(NurbsPatch* patch=dynamic_cast<NurbsPatch*>(g_selNode.getRaw())){
		if(MultiSelect::pointList.length){
			if(MultiSelect::pointList[0].p == patch){
				if(patch->mirror){
					if(NurbsPatch* mirrorPatch = patch->mirrorPatch){
						M_Vec3f& point = patch->point.refValue();
						M_Vec3f& mirPoint = mirrorPatch->point.refValue();

						for(int i = point.length; --i >= 0;){
							Vec3f vec = point[i];
							vec.x = -vec.x;
							mirPoint[i] = vec;
						}
						mirrorPatch->processFields();
					}
					else{
						int idx = MultiSelect::pointList[0].i;
						int uDim = patch->dim.x;
						int vDim = patch->dim.y;
						int u = idx % uDim;
						int v = idx / uDim;
						if(u > uDim / 2){
							patch->mirrorModifyWest();
						}
						else{
							patch->mirrorModifyEast();
						}
						patch->processFields();
					}
					updateBuild3dWindow();
					houseKeepVisualClue();
					BackUp::monostable.beginCycle();
				}
			}
		}
	}
	else if(BezierPatch* bezierPatch=dynamic_cast<BezierPatch*>(g_selNode.getRaw())){
		M_Vec3f& point = bezierPatch->point.refValue();
		const M_Vec2f& dim = bezierPatch->dim;
		int uDim=dim[0].x;
		int vDim=dim[0].y;
		for(int v=vDim; --v>=0;){

			point[v * uDim + uDim - 2].y = point[v * uDim + uDim - 1].y;
			point[v * uDim + uDim - 2].z = point[v * uDim + uDim - 1].z;

			for(int u = 0; u < uDim>>1; u++){
				Vec3f p = point[v*uDim + uDim - 1 - u];
				p.x = -p.x;
				point[v*uDim + u] = p;
			}
		}
	}
	else{
		return;
	}
	updateBuild3dWindow();
	houseKeepVisualClue();
	BackUp::monostable.beginCycle();
}

void copyStemMapToStemMap(Stem* selStem, Stem* copyStem, Seamless* smls){
	selStem->coord.refValue() = copyStem->coord;
	selStem->color.refValue() = copyStem->color;
	selStem->texCoord.refValue() = copyStem->texCoord;
	selStem->coordIndex.refValue() = copyStem->coordIndex;
	selStem->rc.refValue() = copyStem->rc;
	selStem->processFields();
	smls->doBuild();
	smls->ctrf=true;
	BackUp::monostable.beginCycle();
}

void pasteToPolyMap(){
	if(Part* whitePart=dynamic_cast<Part*>(g_selNode.getRaw())){
		openSceneTree();
		if(g_copyNode.getRaw()==whitePart){
			Seamless* smls = whitePart->smls;
			NurbsPatch* patch = 0;
			Stem* selStem = 0;
			Stem* copyStem = 0;
			Polymap* polymap = 0;
			ImmediatePart* bluePart = 0;
			if(getPolyMap(patch, selStem, polymap, whitePart, smls->ActTri, bluePart)){
				if(getPolyMap(patch, copyStem, polymap, whitePart, smls->remActTri, bluePart)){
					if(bluePart){
						selStem->copyPartToPolyMap(bluePart);
						updateBuild3dWindow();
						houseKeepVisualClue();
						BackUp::monostable.beginCycle();
					}
					else{
						copyStemMapToStemMap(selStem, copyStem, whitePart->smls);
						makePointerEqualIndexLocation();
					}
				}
			}
		}
	}
	else if(ImmediatePart* bluePart=dynamic_cast<ImmediatePart*>(g_selNode.getRaw())){
		Stem* stem = dynamic_cast<Stem*>(g_selNode.getRaw());
		stem->copyPartToPolyMap(bluePart);
		updateBuild3dWindow();
		houseKeepVisualClue();
		BackUp::monostable.beginCycle();
	}
	if(Stem* selStem = dynamic_cast<Stem*>(g_selNode.getRaw())){
		if(Stem* copyStem=dynamic_cast<Stem*>(g_copyNode.getRaw())){
			copyStemMapToStemMap(selStem, copyStem, getSeamless());
		}
		else if(Art* part=dynamic_cast<Art*>(g_copyNode.getRaw())){
			selStem->copyPartToPolyMap(part);
			updateBuild3dWindow();
			houseKeepVisualClue();
			BackUp::monostable.beginCycle();
		}
	}
}

void pasteCtrlPoints(NurbsPatch* patch, NurbsPatch* copyPatch){
	MultiSelect::pointList.length = 0;
	MultiSelect::ctrlPolySelList.length = 0;
	patch->point.refValue()	=	copyPatch->point;
	patch->weight.refValue()=	copyPatch->weight;
	patch->uKnot.refValue()	=	copyPatch->uKnot;
	patch->vKnot.refValue()	=	copyPatch->vKnot;
	patch->dim.refValue()	=	copyPatch->dim;
	patch->order.refValue()	=	copyPatch->order;
	patch->uJoinEnds.refValue()	=	copyPatch->uJoinEnds;
	patch->processFields();
}

void pasteCtrlPoints(){
	if(NurbsPatch* copyPatch=dynamic_cast<NurbsPatch*>(g_copyNode.getRaw())){
		NurbsPatch* patch=dynamic_cast<NurbsPatch*>(g_selNode.getRaw());
		pasteCtrlPoints(patch, copyPatch);
		updateBuild3dWindow();
		houseKeepVisualClue();
		BackUp::monostable.beginCycle();
	}
}
void pasteCtrlPointsMirror(){
	if(NurbsPatch* copyPatch=dynamic_cast<NurbsPatch*>(g_copyNode.getRaw())){
		NurbsPatch* patch=dynamic_cast<NurbsPatch*>(g_selNode.getRaw());
		pasteCtrlPoints(patch, copyPatch);
		mirrorAllCtrlPoints();
	}
}

void mirrorHalfPolyMap(){
	if(Part* whitePart=dynamic_cast<Part*>(g_selNode.getRaw())){
		if(dynamic_cast<ColorEffect*>(whitePart->workEffect)==0)return;

		Seamless* smls = whitePart->smls;
		NurbsPatch* patch = 0;
		Stem* stem = 0;
		Polymap* polymap = 0;
		ImmediatePart* bluePart = 0;
		if(getPolyMap(patch, stem, polymap, whitePart, smls->ActTri, bluePart)){
			if(bluePart){
			}
			else{
				int selIdx = smls->ActCtrlPnt-stem->beginCoordOffset;
				M_Vec2f& coord = stem->coord.refValue();
				const M_Int& rc = stem->rc;
				const M_Int& coordIndex = stem->coordIndex;
				int half = coord.length / 2;
				if(selIdx < half){
					for(int ci = 0; ci < coordIndex.length; ci++){
						int i = coordIndex[ci];
						int g = i - coord.length;
						if(g >= 0){
							i = rc[g];
						}
						if(i < half){
							Vec2f crd = coord[i];
							crd.x = -(crd.x - .5) + .5;
							coord[i + half] = crd;
						}
					}
				}
				else{
					for(int ci = 0; ci < coordIndex.length; ci++){
						int i = coordIndex[ci];
						int g = i - coord.length;
						if(g >= 0){
							i = rc[g];
						}
						if(i > half){
							Vec2f crd = coord[i];
							crd.x = -(crd.x - .5) + .5;
							coord[i - half] = crd;
						}
					}
				}
				smls->doBuild();
				makePointerEqualIndexLocation();
				smls->ctrf=true;
				Window::paintAllControlPanels();
			}
		}
	}
}

void addMirrorHalfPolyMap(){
	if(Part* whitePart=dynamic_cast<Part*>(g_selNode.getRaw())){
		Seamless* smls = whitePart->smls;
		NurbsPatch* patch = 0;
		Stem* stem = 0;
		Polymap* polymap = 0;
		ImmediatePart* bluePart = 0;
		if(getPolyMap(patch, stem, polymap, whitePart, smls->ActTri, bluePart)){
			if(bluePart){
			}
			else{
				M_Vec2f& coord = stem->coord.refValue();
				for( int i=coord.length; --i>=0; ){
					coord[i].x -= .5;
				}

				bluePart = selectPolyMapPart(whitePart);
				stem->copyPolyMapToPart(bluePart);

				CopyPart cp;
				cp.mirrorX=true;
				cp.add=true;
				cp.dest+=bluePart;
				cp.doBuild();			
				stem->copyPartToPolyMap(bluePart);
				for( int i=coord.length; --i>=0; ){
					coord[i].x += .5;
				}
				commonEndVertexBuild();
			}
		}
	}
}

void deletePolyMap(){
	if(Stem* stem = dynamic_cast<Stem*>(g_selNode.getRaw())){
		stem->coord.refValue().length = 0;
		stem->color.refValue().length = 0;
		stem->texCoord.refValue().length = 0;
		stem->coordIndex.refValue().length = 0;
		stem->rc.refValue().length = 0;
		stem->processFields();
		BackUp::monostable.beginCycle();
		updateBuild3dWindow();
	}
}

void deleteHalfPolyMap(){
	if(Part* whitePart=dynamic_cast<Part*>(g_selNode.getRaw())){
		Seamless* smls = whitePart->smls;
		NurbsPatch* patch = 0;
		Stem* stem = 0;
		Polymap* polymap = 0;
		ImmediatePart* bluePart = 0;
		if(getPolyMap(patch, stem, polymap, whitePart, smls->ActTri, bluePart)){
			if(bluePart){
			}
			else{
				if(Stem* mirrorStem = stem->mirrorStem){
					mirrorStem->coord.refValue().length = 0;
					mirrorStem->color.refValue().length = 0;
					mirrorStem->texCoord.refValue().length = 0;
					mirrorStem->coordIndex.refValue().length = 0;
					mirrorStem->rc.refValue().length = 0;
					mirrorStem->processFields();
				}
				else{
					bluePart = selectPolyMapPart(whitePart);
					stem->copyPolyMapToPart(bluePart);
					int selIdx = smls->ActCtrlPnt-stem->beginCoordOffset;
					Vec2f selVec = stem->coord[selIdx];
					if(selVec.x >= .5){
						bluePart->cutLeft(.5+.0001);
					}
					else{
						bluePart->cutRight(.5-.0001);
					}
					stem->copyPartToPolyMap(bluePart);
				}
				commonEndVertexBuild();
			}
		}
	}
}

void setAllWeightsToOne(){
	if(NurbsPatch* patch=dynamic_cast<NurbsPatch*>(g_selNode.getRaw())){
		M_Float& weight = patch->weight.refValue();
		const M_Vec3f& dpoint = patch->dpoint;
		M_Vec3f& point = patch->point.refValue();
		for(int i = point.length; --i>=0;){
			weight[i] = 1;
			point[i] = dpoint[i];
		}
		patch->processFields();
		Seamless* smls = patch->smls;
		updateBuild3dWindow();
	}
}

inline void addPartToList(M_Art& relevant, Art*& part){
	for(int r=relevant.length; --r>=0;){
		if(relevant[r]==part)return;
	}
	relevant+=part;
}

void Seamless::removeAllPatchesFromPartsNotInList(M_Art& m_part){
	for(int k = partList.length; --k>=0;){
		Art* part = partList[k];
		M_ControlPoint& controlPoint = part->controlPoint;
		int d = 0;
		for(int s = 0; s < controlPoint.length; s++){
			NurbsPatch* patch = controlPoint[s].p;
			if(patch->plidx >= cpNodeList.length || cpNodeList[patch->plidx] != patch){
				addPartToList(m_part, part);
				continue;				
			}
			controlPoint[d++] = controlPoint[s];
		}
		controlPoint.length = d;
	}
}

void deleteNode(){//DELETE_NODE
	g_stw->textCell.cellActive = false;

	M_<NurbsPatchPtr> m_hold;

	for(int k = g_sceneTreeList.length; --k>=0;){
		if(Seamless* smls = dynamic_cast<Seamless*>(g_sceneTreeList[k].node.p)){
			for(int k = 0; k < smls->cpNodeList.length; k++){
				m_hold += smls->cpNodeList[k];
			}
		}
	}
	fastFullBuildForAutoIn(true);

	deleteNode(g_stw->selLine, g_selCount);
	
#ifdef PROG_ROB
	dLog("new MoveToNode(" + String(g_stw->selLine) + "),");
	dLog("new RightClick,");
	dLog("new Move_MenuItem(menu_delete),");
	dLog("new Click,");
#endif

	g_stw->genNodeList();
	M_Art m_part;
	for(int k = g_sceneTreeList.length; --k>=0;){
		if(Seamless* smls = dynamic_cast<Seamless*>(g_sceneTreeList[k].node.p)){
			smls->cpNodeList.length = 0;
			smls->cpNodeListAdd(smls->build.refValue());
			smls->deleteLinksToNonExistentPatches(); //not to be confused with non existing control polys
			smls->deletePolysWithNonExistingPoints();
			smls->removeAllPatchesFromPartsNotInList(m_part);
		}
	}

	for(int k = g_sceneTreeList.length; --k>=0;){
		if(Anim* anim = dynamic_cast<Anim*>(g_sceneTreeList[k].node.p)){
			anim->resetMorphers(m_part);
		}
	}

	g_stw->commonPostInsertNodeStuff(true);
	updateAutoInForClick();
}

void deleteLink(){
	if(MultiSelect::pointList.length == 1){
		NurbsPatch* patch = MultiSelect::pointList[0].p;
		int idx = MultiSelect::pointList[0].i;
		M_Link& link = patch->smls->link;
		for(int i = link.length; --i>=0;){
			if(link[i].patch == patch && link[i].point == idx){
				patch->smls->link.deleteElement(i);
				patch->processFields();
				updateBuild3dWindow();
				break;
			}
		}
	}
}


void pasteColumn(){
	if(NurbsPatch* copyPatch = dynamic_cast<NurbsPatch*>(g_copyNode.getRaw())){
		int copy_uDim = copyPatch->dim.x;
		int copy_vDim = copyPatch->dim.y;
		int column = g_rememberIndex % copy_uDim;
		M_Vec3f& srcPoint = copyPatch->point.refValue();
		if(NurbsPatch* patch = dynamic_cast<NurbsPatch*>(g_selNode.getRaw())){
			int uDim = patch->dim.x;
			int vDim = patch->dim.y;
			int index = patch->index[0] % uDim;
			M_Vec3f& dstPoint = patch->point.refValue();
			for(int v = 0; v < vDim; v++){
				dstPoint[v * uDim + index].y = srcPoint[v * copy_uDim + column].y;
				dstPoint[v * uDim + index].z = srcPoint[v * copy_uDim + column].z;
			}
			patch->processFields();
			updateBuild3dWindow();
		}
	}
}

void rotateCtrlStaves(){
	if(NurbsPatch* patch=dynamic_cast<NurbsPatch*>(g_selNode.getRaw())){
		int uDim = patch->dim.x;
		int vDim = patch->dim.y;
		M_Float& weight = patch->weight.refValue();
		M_Vec3f& point = patch->point.refValue();
		for(int i = 2; --i>=0;){
			for(int u = 0; u < (uDim - 1); u++){
				for(int v = 0; v < vDim; v++){
					point[v * uDim + u] = point[v * uDim + (u + 1)];
					weight[v * uDim + u] = weight[v * uDim + (u + 1)];
				}
			}
			for(int v = 0; v < vDim; v++){
				point[v * uDim + uDim - 1] = point[v * uDim];
				weight[v * uDim + uDim - 1] = weight[v * uDim];
			}
		}
		patch->processFields();
		updateBuild3dWindow();
	}
}

void calcLatheWeights(){
	if(NurbsPatch* patch=dynamic_cast<NurbsPatch*>(g_selNode.getRaw())){
		M_Vec3f& point = patch->point.refValue();
		M_Vec3f& dpoint = patch->dpoint.refValue();
		M_Float& weight = patch->weight.refValue();
		int uDim = patch->dim.x;
		int vDim = patch->dim.y;
		int ctrlStaves = uDim / 2;
		double sinWeight = cos(PI / ctrlStaves);

		for(int r = 0; r < vDim; r++){
			double a = PI / ctrlStaves;
			for(int c = 0; c < uDim; ){
				double b = a * c;
				int i = r * uDim + c++;
				weight[i] = 1;
				point[i] = dpoint[i] * weight[i];

				if(c == uDim){
					break;
				}

				i = r * uDim + c++;
				weight[i] = sinWeight;
				point[i] = dpoint[i] * weight[i];
			}
		}
		patch->processFields();
		updateBuild3dWindow();
	}
}

void deleteMorpherChannels(){
	if(Art* part=dynamic_cast<Art*>(g_selNode.getRaw())){
		if(Anim* anim = Window::animMode()){
			anim->deleteMorpherChannel();
			BackUp::monostable.beginCycle();
		}
	}
	else{
		needPartMsg();
	}
}

void compileMorph(){
	if(Art* part=dynamic_cast<Art*>(g_selNode.getRaw())){
		if(Anim* anim = Window::animMode()){
			anim->compileMorph(part->smls);
		}
	}
	else{
		needPartMsg();
	}
}

void compileOrientationInterpolator(const Art* part, Rotation accumRot, M_<SceneTreeListItem>& sceneTreeList);

void untickIkFields(){
	if(Seamless* smls = dynamic_cast<Seamless*>(g_selNode.getRaw())){
		M_Art& partList = smls->partList;
		for(int k = partList.length; --k >= 0;){
			Art* part = partList[k];
			part->ik = false;
			part->pin = false;
			part->ikPos = Vec3f();
			part->ikRot = Rotation();
			part->fka = 0;
			part->fkb = 0;
		}
	}
}

void transformAnimation(){
	if(Art* part=dynamic_cast<Art*>(g_selNode.getRaw())){
		if(Anim* anim = Window::animMode()){
			anim->transformInterpolators(part);
		}
	}
	else{
		needPartMsg();
	}
}

void toggleAnimNeutral(){
	Anim::autoCapture = !Anim::autoCapture;
}

void take2(Vec2f slider, Vec3f pos, M_Int& index, M_Vec2f& moment, M_Vec3f& relPos);

void beginCompensation(){
	if(Art* part=dynamic_cast<Art*>(g_selNode.getRaw())){
		if(Anim* anim = Window::animMode()){
			if(anim->play == false){
				PosCompensator* interp = 0;
				if(interp = anim->getPosCompensator(part)){
				}
				else if(interp = anim->connectPartToPosCompensator(part)){
				}
				if(interp){
					take2(Vec2f(anim->slider, anim->slider + 1.0 / anim->stepsPerSecond * 2 / anim->period), part->ikPos, interp->index, interp->moment.refValue(), interp->relPos.refValue());
				}
			}
		}
	}
	else{
		needPartMsg();
	}
}

void removeCompensation(){
	if(Art* part=dynamic_cast<Art*>(g_selNode.getRaw())){
		if(Anim* anim = Window::animMode()){
			if(anim->play == false){
				anim->removeCompensation(part);
			}
		}
	}
	else{
		needPartMsg();
	}
}

void PosCompensator::take(float slider, Node* node, Field* field){};
void PosCompensator::deleteSample(float slider){};

void highlightWholeTimeLine(){
	Anim* anim = Window::animMode();
	anim->selectFrom = 0;
	anim->selectTo = 1;
}


void makeTrianglesTiny(){
	if(Art* part=dynamic_cast<Art*>(g_selNode.getRaw())){
		M_Vec3f& coord = part->coord.refValue();
		Seamless* smls = part->smls;
		for(int k = 0; k < coord.length; k+=3){
			Vec3f center = (coord[k] + coord[k + 1] + coord[k + 2]) / 3.0;
			coord[k + 0] = Vec3f(center.x - .0008, center.y, center.z);
			coord[k + 1] = Vec3f(center.x + .0004, center.y, center.z + .0008 * 0.866025404);
			coord[k + 2] = Vec3f(center.x + .0004, center.y, center.z - .0008 * 0.866025404);
		}
		smls->resetRelevantParts();
		smls->process();
		smls->verifyIntegrity();
		smls->ctrf = true;
	}
}
