/* Seamless3d

Copyright (C) 2001-2010 Graham John Perrett (AKA thyme) thyme@seamless3d.com www.seamless3d.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this
software and associated documentation files (the "Software"), to deal in the Software
without restriction, including without limitation the rights to use, copy, modify,
merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to the following
conditions:

The above copyright notice and this permission notice shall be included in all copies or
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT
OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.
*/

#define _WIN32_WINNT 0x0400
#define WINVER 0x0500

#include <windows.h>
#include <direct.h>
#include <time.h>

#include "seamless3d.h"
#include "techuelife_triangle_editor.h"
#include "gui.h"
#include "scene_tree.h"
#include "resource.h"
#include "input.h"
#include "robot_aid.h"
#include "seamless3d_gui_render.h"



//#define GET_WM_CTLCOLOR_HDC(wp, lp, msg)        (HDC)(wp) 


//#define iconLookupCX pixSymbolCX*10
//#define iconLookupCY iconLookupCX

#define CTRLF_DIV	12
#define CTRLF_E		0
#define CTRLF_SPN0	1
#define CTRLF_SPN1	2
#define CTRLF_SPN2	3
#define CTRLF_SPN4	4
#define CTRLF_RSET	5
#define CTRLF_IDX_E	6
#define CTRLF_IDX_S	7
#define CTRLF_IDX_STAT	8
#define CTRLF_LEN_E	9
#define CTRLF_LEN_S	10
#define CTRLF_LEN_STAT	11

#define NAME_CELL_CX	1024

extern const char* msg_canNot_delete_used_else_where;



Ptr<Menu> g_remMenu;
extern double g_time;


int g_numOfVisLines;
bool g_controlPanelOpen=false;
bool g_dontBackup=false;

//unsigned g_lightBlue=RGB(230, 230, 255);
unsigned g_cream=RGB(205, 255, 255);
unsigned g_white=RGB(255, 255, 255);

//unsigned g_fieldColor=RGB(0, 0, 255);
unsigned g_fieldColor=RGB(255, 0, 0);

unsigned g_dstRouteNodeColor=RGB(255, 0, 0);
unsigned g_dstRouteFieldColor=RGB(0, 0, 255);
LPCSTR g_currentType = IDC_ARROW;

Ptr<SceneTree> g_stw;
M_<CallBack*>g_callBackList;

M_Int g_copyIndex;

int g_fromFieldIdx;
SelCount g_selCount;
int g_copySibCount=0;
int g_copyAllCount=0;

M_<SceneTreeListItem> g_sceneTreeList;
M_<SceneTreeListItem> g_sceneTreeCollapsed;
M_<Anim*> g_animBarList;

HBRUSH g_editNodeNameBrush=CreateSolidBrush(0);
void createFunctionListFromSceneTreeList();
void commonEndForSetupAnim();
void replaceNodeRefInScripts(NodePtr from, Node* to);
void renderScene();

extern HWND g_hDlg;

void setTitleName();
bool smlsIsSaved();

extern bool outPutIndices;


bool getJoinIndices=false;
bool getParentJoinIndices=false;
HWND g_hDlg;
M_Int progIndices;
int g_routeIndex=0;

int g_nDocument=0;
int g_nClosed=0;

void setTitleName();
int getFileSize(FILE* Stream);
int search(char* match ,char*& from,char* BufEnd);
void useNodeField(Node* parentSelNode, int selFieldIdx, int selElement, Node* parentCopyNode, int copyFieldIdx, int copyElement, int copyNodeLine, int selCount, int copySibCount, unsigned type);
void createActiveLabel(int collapsedLine);
void resetDefList();
void updateAutoInForClick();

Vec3f getVec3f(char*& p);
Rotation getRotation(char*& p);
bool getBool(char*& p);
int undoUniqueNameCntr=0;
extern NodePtr g_menueParent;
extern int g_menueElementForParent;
int g_sceneTreeNodeIndex;

void replaceNodeRefInSceneTree(NodePtr from, Node* to);

void removePart(Art* pj);
void convertToRenderFormat(IndexedFaceSetPtr indexedFaceSet);
void changeSrcPointersToExistingNames(NodePtr& ptr);
void deleteAllCodeLinesThatRefNonExistentNodes();
bool decodeSmlsVrmlHeader(String& str, int& i);


void copyMultiUrlWithSingleUrl(M_String& m_url, const String& url);
void copySingleUrlWithMultiUrl(String& url, const M_String& m_url);




bool nodeHasRouteTo(Node* from, Node* to){
	for(int i=0; i<from->getFLen(); i++){
		Field* field=from->getField(i);
		const M_Route& route=field->getRoute();
		for(int r=0; r<route.length; r++){
			if(route[r].toNode==to){
				return true;
			}
		}
	}
	return false;
}
bool Node::searchForRefNode(){
	for(int i=0; i<g_sceneTreeList.length; i++){
		Node* node=g_sceneTreeList[i].node;
		if(node==this){
			return true;
		}
	}
	return false;
}
int getNodeLine(Node* node){
	for(int i=0; i<g_sceneTreeList.length; i++){
		if(g_sceneTreeList[i].node==node){
			return i;
		}
	}
	return -1;
}
Node* searchForMatchingName(String& str){
	for(int i=0; i<g_sceneTreeList.length; i++){
		if(g_sceneTreeList[i].node->name==str){
			return g_sceneTreeList[i].node;
		}
	}
	return 0;
}
extern bool g_exitBuildFlag;
int GroupBuild::getIndexForLastBuildNode(M_Node*& r_build){
	int k=0;
	g_exitBuildFlag=false;
	for(; k<build.length; k++){
		GroupBuild* groupBuild=dynamic_cast<GroupBuild*>(&*build[k]);
		if(groupBuild!=nullPtr){
			int g=groupBuild->getIndexForLastBuildNode(r_build);
			if(g_exitBuildFlag){
				k=g;
				break;
			}
		}
		else{
			ExitBuild* exitBuild=dynamic_cast<ExitBuild*>(&*build[k]);
			if(exitBuild!=nullPtr && exitBuild->enabled==true){
				g_exitBuildFlag=true;
			}
		}
		if(g_exitBuildFlag){
			MFNode* m=&build;
			r_build= &m->refValue();
			break;
		}
	}
	return k;
}
int Seamless::getIndexForLastBuildNode(M_Node*& r_build){
	int k=0;
	g_exitBuildFlag=false;
	MFNode* m=&build;
	r_build= &m->refValue();
	for(; k<build.length; k++){
		GroupBuild* groupBuild=dynamic_cast<GroupBuild*>(&*build[k]);
		if(groupBuild!=nullPtr){
			int g=groupBuild->getIndexForLastBuildNode(r_build);
			if(g_exitBuildFlag){
				k=g;
				break;
			}
		}
		else{
			ExitBuild* exitBuild=dynamic_cast<ExitBuild*>(&*build[k]);
			if(exitBuild!=nullPtr && exitBuild->enabled==true){
				g_exitBuildFlag=true;
				r_build= &m->refValue();
				break;
			}
		}
	}
	return k-1;
}
Stem* getStemOwner(){
	Stem* stem=nullPtr;
	if(	g_sceneTreeList[g_stw->selLine-1].node->getType()==SURFACEGENERATOR){
		int indent=g_sceneTreeList[g_stw->selLine-1].indent;
		for(int i=g_stw->selLine-1; --i;){
			if(g_sceneTreeList[i].indent<indent){
				stem=dynamic_cast<Stem*>(g_sceneTreeList[i].node.p);
				break;

			}
		}
	}
	return stem;
}

const Properties& Node::getProperties(Field& field)const{
	for(int i=0; i<getFLen(); i++){
		if(&field==getField(i)){
			return getProperties()[i];
		}
	}
	throw this;
}
void CCLathe::Curve::operator=(const M_Float& par){
	refValue()=par;
}
void CCLathe::Curve::resize(int newLen){
	refValue().length=newLen;
	*this=this->refValue();
}
void ECLatheECSyncBendPointEquals(ALathe& home, const int length, M_Vec2f& point, M_Float& tangent){
	if(tangent.length>length){
		tangent.length=length;
	}
	home.processFields();
}
void ECLathe::Point::operator=(const M_Vec2f& par){
	refValue()=par;
	ECLatheECSyncBendPointEquals(home, length, *this, home.tangent);
}
void BezierLatheBuild::Point::operator=(const M_Vec2f& par){
	refValue()=par;
	home.processFields();
}
int Bezier::getGroupIndices(int& fixSegBeginIndex){
	int idx=index[0];
	int groupIndex=-1;
	fixSegBeginIndex=0;
	for(int i=0; i<fix.length-1; i++){
		if(fix[i]){
			groupIndex++;
			fixSegBeginIndex=i;
		}
		if(i==idx){
			break;
		}
	}
	return groupIndex;
}
int Bezier3f::getGroupIndices(int& fixSegBeginIndex){
	int idx=index[0];
	int groupIndex=-1;
	fixSegBeginIndex=0;
	for(int i=0; i<fix.length-1; i++){
		if(fix[i]){
			groupIndex++;
			fixSegBeginIndex=i;
		}
		if(i==idx){
			break;
		}
	}
	return groupIndex;
}
void BezierLatheBuild::DBands::operator=(const M_Int& par){
	refValue()=par;
	int fixSegBeginIndex=0;
	int groupIndex=home.getGroupIndices(fixSegBeginIndex);
	if(groupIndex>=0){
		M_Int& bands=home.bands;
		int idx=home.index[0];
		bands[groupIndex] = par[idx];
		home.processFields();
	}
}
void BezierLatheBuild::Point::resize(int newLen){
	refValue().length=newLen;
	*this=this->refValue();
}
void ECSyncBend::Point::operator=(const M_Vec2f& par){
	refValue()=par;
	ECLatheECSyncBendPointEquals(home, length, *this, home.tangent);
}
void ECLathe::Point::resize(int newLen){
	refValue().length=newLen;
	*this=this->refValue();
}
void ECSyncBend::Point::resize(int newLen){
	refValue().length=newLen;
	*this=this->refValue();
}
ALathe* getLathe(Stem& stem){
	ALathe* lathe=nullPtr;
	Generator* surface = dynamic_cast<Generator*>(stem.surface.p);
	if(SurfaceGenerator* surfaceGenerator=dynamic_cast<SurfaceGenerator*>(surface)){
		M_Node& generator=surfaceGenerator->generator;
		if(generator.length>0){
			lathe=dynamic_cast<ALathe*>(&*generator[0]);
		}
	}
	return lathe;
}


void Node::genNodeListRecursion(const int parent, int indent, Color3f marker, M_<SceneTreeListItem>& sceneTreeList, M_<Anim*>& animBarList){
	if(GroupBuild* groupBuild = dynamic_cast<GroupBuild*>(this)){
		marker = groupBuild->marker;
	}
	else if(Anim* anim = dynamic_cast<Anim*>(this)){
		animBarList += anim;
	}
	bool ref=searchForRefNode();
	int currentLine = sceneTreeList.length;
	sceneTreeList += SceneTreeListItem(ref, this, parent, indent, marker);

	if(ref==0){
		indent++;
		for(int i=0; i<getFLen(); i++){
			Field* field=getField(i);
			if(field->getType()==MFNODE){
				AGroup* group=0;
				if(group=dynamic_cast<AGroup*>(this)){
					group->sensor=0;
				}
				M_Node &mnode=dynamic_cast<MFNode*>(field)->refValue();
				int length=mnode.length;
				if(length){
					for(int k=0;k<length; k++){
						Node* child=mnode[k];
						if(TouchSensor* sensor=dynamic_cast<TouchSensor*>(child)){
							if(group){
								group->sensor=sensor;
							}
						}
						child->genNodeListRecursion(currentLine, indent, marker, sceneTreeList, animBarList);
					}
				}
			}
			else if(field->getType()==SFNODE){
				NodePtr &snode=dynamic_cast<SFNode*>(field)->refValue();
				if(snode!=NULL){
					snode->genNodeListRecursion(currentLine, indent, marker, sceneTreeList, animBarList);
				}
			}
		}
	}
}

int getChildrenLen(int index){
	int selIndent=g_sceneTreeCollapsed[index].indent;
	int siblings=0;
	while(++index < g_sceneTreeCollapsed.length && g_sceneTreeCollapsed[index].indent > selIndent){
		siblings++;
	}
	return siblings;
}
int getChildrenLenNonCollapsed(int index){
	int selIndent = g_sceneTreeList[index].indent;
	int siblings = 0;
	while(++index < g_sceneTreeList.length && g_sceneTreeList[index].indent > selIndent){
		siblings++;
	}
	return siblings;
}

int getNodeLineForNextSibling(int nodeLine){
	int selIndent = g_sceneTreeList[nodeLine].indent;
	int siblings = 0;
	while(++nodeLine < g_sceneTreeList.length && g_sceneTreeList[nodeLine].indent > selIndent){
	}
	return nodeLine;
}


int getLineForNonCollapsed(int collapsedLine){
	int i=0;
	int j=0;
	for(; j<collapsedLine; i++,j++){
		if(g_sceneTreeList[i].node->collapsed){
			i+=getChildrenLenNonCollapsed(i);
		}
	}
	return i;
}

bool getCollapsedLineNoGlobalChange(int nonCollapsedLine, int& collapsedLine){
	bool visible=false;
	collapsedLine=0;
	int j=0;
	for(; j<nonCollapsedLine; collapsedLine++,j++){
		if(g_sceneTreeCollapsed[collapsedLine].node->collapsed){
			j+=getChildrenLenNonCollapsed(j);
		}
	}

	if(j==nonCollapsedLine){
		return true;
	}
	return false;
}

bool getCollapsedLine(int nonCollapsedLine, int& collapsedLine){
	bool visible=false;
	collapsedLine=0;
	int j=0;
	for(; j<nonCollapsedLine; collapsedLine++,j++){
		if(g_sceneTreeCollapsed[collapsedLine].node->collapsed){
			j+=getChildrenLenNonCollapsed(j);
		}
	}

	if(g_stw && collapsedLine < g_sceneTreeCollapsed.length){
		g_stw->textCell.cellPos = Vec2i(nameOffset - 1 + pixSymbolCX + g_sceneTreeCollapsed[collapsedLine].indent * pixSymbolCX, g_stw->beginYOffset + 1 + (collapsedLine - g_scene->scroll) * pixSymbolCY);
		g_stw->textCell.cellSize = Vec2i(g_stw->size.x - g_stw->textCell.cellPos.x - 15, 16);
	}

	if(j==nonCollapsedLine){
		return true;
	}
	return false;
}

void shortenCollapsedNodes(){
	g_sceneTreeCollapsed.length=0;
	Color3f color;
	for(int src=0,dst=0; src<g_sceneTreeList.length; src++, dst++){
		g_sceneTreeCollapsed+=g_sceneTreeList[src];
		if(g_sceneTreeList[src].node->collapsed){
			src+=getChildrenLenNonCollapsed(src);
		}
	}
	if(g_stw){
		int collapsedLine = 0;
		getCollapsedLine(g_stw->selLine, collapsedLine);
	}
}
void Scene::preSaveProcessCollapse(){
	M_Int& collapse=this->collapse;
	collapse.length=0;
	for(int k=g_sceneTreeList.length; --k>=0;){
		if(g_sceneTreeList[k].node->collapsed){
			collapse+=k;
		}
	}
}
void Scene::postLoadProcessCollapse(){
	M_Int& collapse=this->collapse;
	for(int k=collapse.length; --k>=0;){
		int j=collapse[k];
		if(j<g_sceneTreeList.length){
			g_sceneTreeList[j].node->collapsed=true;
		}
	}
}
void Scene::genNodeList2(){
	g_sceneTreeList.length = 0;
	g_animBarList.length = 0;
	int indent=0;
	genNodeListRecursion(-1, indent, Color3f(), g_sceneTreeList, g_animBarList);
	if(selLine>=g_sceneTreeList.length){
		selLine.refValue()=g_sceneTreeList.length - 1;
		if(selLine < 0){
			selLine = 0;
		}
		wOpen.refValue().length=0;
		wPos.refValue().length=0;
		wSize.refValue().length=0;
		collapse.refValue().length=0;
	}
	shortenCollapsedNodes();
}

void Scene::genNodeList3(){//same as genNodeList2 minus shortenCollapsedNodes()
	g_sceneTreeList.length = 0;
	g_animBarList.length = 0;
	int indent=0;
	genNodeListRecursion(-1, indent, Color3f(), g_sceneTreeList, g_animBarList);
	if(selLine>=g_sceneTreeList.length){
		selLine.refValue()=g_sceneTreeList.length - 1;
		if(selLine < 0){
			selLine = 0;
		}
		wOpen.refValue().length=0;
		wPos.refValue().length=0;
		wSize.refValue().length=0;
		collapse.refValue().length=0;
	}
}

void SceneTree::genNodeList(){
	g_scene->genNodeList2();
	slider.setPixelPos();
	textCell.updateCaret();
}

void GroupBuild::Marker::operator =(const Color3f &par){
	refValue()=par;
	g_stw->genNodeList();
	g_stw->paint();
}
void convertArtToArt(Art* from, Art* to, bool refNodes = true){
	to->name=from->name;
	to->coord.refValue()=from->coord;
	to->normal.refValue()=from->normal;
	to->color=from->color;
	to->texCoord=from->texCoord;
	to->scale.refValue()=from->scale;
	to->translation.refValue()=from->translation;
	to->rotation.refValue()=from->rotation;
	to->coordGroup.refValue()=from->coordGroup;
	to->weight.refValue()=from->weight;
	to->pivotPoint.refValue()=from->pivotPoint;
	if(refNodes){
		to->effect.refValue()=from->effect.refValue();
		to->workEffect=from->workEffect;
		to->morpher.refValue()=from->morpher;
		to->part.refValue()=from->part;
	}
	to->show=from->show;

	to->fromEffectCoord=from->fromEffectCoord;
	to->toEffectCoord=from->toEffectCoord;
	to->fromCoord=from->fromCoord;
	to->toCoord=from->toCoord;
	to->fromPoly=from->fromPoly;
	to->toPoly=from->toPoly;
	to->plidx=from->plidx;
	to->renderFromCoord=from->renderFromCoord;

	to->geometry=from->geometry;

	to->vCoordGroup=from->vCoordGroup;
	to->scratchC=from->scratchC;
	to->partTree=from->partTree;
	to->translatePart=from->translatePart;
	to->matrix=from->matrix;
	to->imatrix=from->imatrix;
	to->pivotPointMatrix=from->pivotPointMatrix;
	to->smls=from->smls;
	to->matrixIndex=from->matrixIndex;

	M_Art& partList = from->smls->partList;
	for(int k=partList.length; --k>=0;){
		Art* p=partList[k];
		M_Triangle& triangle=p->triangle;
		for(int i=triangle.length*3; --i>=0;){
			if(triangle.refPart(i)==from){
				triangle.refPart(i)=to;
			}
		}
		M_Art& relevantPart=p->relevantPart;

		for(int i=relevantPart.length; --i>=0;){
			if(relevantPart[i]==from){
				relevantPart[i]=to;
			}
		}
		to->relevantPart=from->relevantPart;
	}
	to->triangle=from->triangle;
}
ImmediatePart* convertPartToImmediate(Art* part){
	ImmediatePart* immediatePart=new ImmediatePart;
	convertArtToArt(part, immediatePart);
	return immediatePart;
}
Part* convertImmediateToPart(Art* immediate){
	Part* part=new Part;
	convertArtToArt(immediate, part);
	return part;
}
void postLoadProcessNodes(M_<SceneTreeListItem>& list){
	M_Rotation rotList;

	for(int k=0; k<list.length; k++){
		if(list[k].ref == false){
			list[k].node->processFields();
		}
	}
	if(g_scene->version.refValue()<2.117-.0001){
		for(int k=0; k<list.length; k++){
			if(ArtPtr part=dynamic_cast<Art*>(list[k].node.p)){
				Seamless* smls=part->smls;
				if(part->coord.refValue().length || part->triangle.length){	
					Ptr<ImmediatePart> immediate = convertPartToImmediate(part);
					replaceNodeRefInSceneTree(part.p, immediate.p);
					g_scene->genNodeList2();
					smls->process();
					smls->resetRelevantParts();
				}
			}
		}
	}
	for(int k=0; k<list.length; k++){
		if(Seamless* smls=dynamic_cast<Seamless*>(list[k].node.p)){
			smls->version.refValue()=g_version;
			M_Art& partList = smls->partList;

			rotList.length=partList.length;
			for(int k=partList.length; --k>=0;){
				rotList[k]=partList[k]->rotation;
			}
			smls->doBuild();

			for(int k=partList.length; --k>=0;){
				partList[k]->rotation=rotList[k];
			}
			smls->verifyIntegrity();
		}
	}
	g_scene->version.refValue()=g_version;
}
const NodeIconLookUpItem nodeIconLookUp[LOOKUPLEN]={
{Vec2i(16,16), Vec2i(0,96)}, //NullNode+4
{Vec2i(16,16), Vec2i(192,112)}, //Scene
{Vec2i(16,16), Vec2i(448,128)}, //AvatarSocket
{Vec2i(16,16), Vec2i(224,128)}, //Shape
{Vec2i(16,16), Vec2i(0,96)}, //Part
{Vec2i(16,16), Vec2i(256,112)}, //Appearance
{Vec2i(16,16), Vec2i(448,144)}, //IndexedFaceSet
{Vec2i(16,16), Vec2i(256,144)}, //ImageTexture
{Vec2i(16,16), Vec2i(272,112)}, //Material
{Vec2i(16,16), Vec2i(288,96)}, //NavigationInfo
{Vec2i(16,16), Vec2i(304,112)}, //Group
{Vec2i(16,16), Vec2i(32,96)}, //NoOutputGroup
{Vec2i(16,16), Vec2i(368,96)}, //GestureGroup
{Vec2i(16,16), Vec2i(112,96)}, //Mobility
{Vec2i(16,16), Vec2i(48,112)}, //CCLathe
{Vec2i(16,16), Vec2i(80,96)}, //CCBend
{Vec2i(16,16), Vec2i(0,128)}, //SurfaceGenerator
{Vec2i(16,16), Vec2i(16,112)}, //Stem
{Vec2i(16,16), Vec2i(16,112)}, //Polymap
{Vec2i(16,16), Vec2i(16,128)}, //BranchJoin
{Vec2i(16,16), Vec2i(16,128)}, //Generator
{Vec2i(16,16), Vec2i(320,160)}, //GeomSource
{Vec2i(16,16), Vec2i(64,96)}, //SurfaceTransform
{Vec2i(16,16), Vec2i(placeIcon_x, placeIcon_y)}, //Place
{Vec2i(16,16), Vec2i(304,96)}, //DirectionalLight
{Vec2i(16,16), Vec2i(64,144)}, //CCScale
{Vec2i(16,16), Vec2i(96,96)}, //CCTranslate
{Vec2i(16,16), Vec2i(240,112)}, //CCRingScale
{Vec2i(16,16), Vec2i(32,144)}, //BezierLathe
{Vec2i(16,16), Vec2i(48,112)}, //ECLathe
{Vec2i(16,16), Vec2i(64,144)}, //ECScale
{Vec2i(16,16), Vec2i(96,96)}, //ECShift
{Vec2i(16,16), Vec2i(80,96)}, //ECBend
{Vec2i(16,16), Vec2i(80,96)}, //ECSyncBend
{Vec2i(16,16), Vec2i(80,96)}, //ECTwist
{Vec2i(16,16), Vec2i(272,96)}, //Background
{Vec2i(16,16), Vec2i(0,112)}, //Seamless
{Vec2i(16,16), Vec2i(304,112)}, //GroupBuild
{Vec2i(16,16), Vec2i(224,96+16)}, //SCGroupBuild
{Vec2i(16,16), Vec2i(224,96)}, //JoinCloseVertex
{Vec2i(16,16), Vec2i(centerJoint_x,centerJoint_y)}, //CenterJoint
{Vec2i(16,16), Vec2i(64,112)}, //DelPoly
{Vec2i(16,16), Vec2i(48,160)}, //JoinVertex
{Vec2i(16,16), Vec2i(copyPartIcon_x, copyPartIcon_y)}, //CopyPart
{Vec2i(16,16), Vec2i(240,144)}, //Viewpoint
{Vec2i(16,16), Vec2i(336,112)}, //TransferBCyl
{Vec2i(16,16), Vec2i(256,96)}, //TransferOwnership
{Vec2i(16,16), Vec2i(256,96)}, //TransferVertex
{Vec2i(16,16), Vec2i(256,128)}, //Coordinate
{Vec2i(16,16), Vec2i(272,128)}, //Normal
{Vec2i(16,16), Vec2i(288,128)}, //Color
{Vec2i(16,16), Vec2i(304,128)}, //TextureCoordinate
{Vec2i(16,16), Vec2i(32,160)}, //SineTug
{Vec2i(16,16), Vec2i(32,160)}, //SimulSineTug
{Vec2i(16,16), Vec2i(64,128)}, //GenPoly
{Vec2i(16,16), Vec2i(288,112)}, //Script
{Vec2i(16,16), Vec2i(16,128)}, //BranchJoinNegX
{Vec2i(16,16), Vec2i(16,144)}, //BranchJoinPosX
{Vec2i(16,16), Vec2i(16,160)}, //BranchJoinNegY
{Vec2i(16,16), Vec2i(208,96)}, //BranchJoinPosY
{Vec2i(16,16), Vec2i(208,112)}, //BranchJoinNegZ
{Vec2i(16,16), Vec2i(208,128)}, //BranchJoinPosZ
{Vec2i(16,16), Vec2i(16,160)}, //ProjectOntoTri
{Vec2i(16,16), Vec2i(colorSweepIcon_x, colorSweepIcon_y)}, //ColorSweep
{Vec2i(16,16), Vec2i(336,128)}, //EvolveCoord
{Vec2i(16,16), Vec2i(160,96)}, //Function
{Vec2i(16,16), Vec2i(0,176)}, //MemberFunction
{Vec2i(16,16), Vec2i(112,96)}, //CodeLine
{Vec2i(16,16), Vec2i(80,128)}, //MorphPart
{Vec2i(16,16), Vec2i(96,128)}, //Morpher
{Vec2i(16,16), Vec2i(192,128)}, //TextureTransform
{Vec2i(16,16), Vec2i(80,112)}, //ProtoDeclare
{Vec2i(16,16), Vec2i(288,160)}, //ProtoInterface
{Vec2i(16,16), Vec2i(272,160)}, //ProtoBody
{Vec2i(16,16), Vec2i(336,160)}, //ProtoInstance
{Vec2i(16,16), Vec2i(176,144+16)}, //Movie
{Vec2i(16,16), Vec2i(176,144)}, //Video
{Vec2i(16,16), Vec2i(32,192)}, //Anim
{Vec2i(16,16), Vec2i(112,112)}, //TimeSensor
{Vec2i(16,16), Vec2i(96,144)}, //OrientationInterpolator
{Vec2i(16,16), Vec2i(112,144)}, //PositionInterpolator
{Vec2i(16,16), Vec2i(112,144)}, //NurbsPath
{Vec2i(16,16), Vec2i(112,160)}, //ScalarInterpolator
{Vec2i(16,16), Vec2i(112,96)}, //PosCompensator
{Vec2i(16,16), Vec2i(48,144)}, //Collision
{Vec2i(16,16), Vec2i(32,112)}, //Transform
{Vec2i(16,16), Vec2i(80,144)}, //ColorInterpolator
{Vec2i(16,16), Vec2i(96,160)}, //CoordinateInterpolator
{Vec2i(16,16), Vec2i(80,160)}, //NormalInterpolator
{Vec2i(16,16), Vec2i(224,176)}, //ModelSequence
{Vec2i(16,16), Vec2i(192,192)}, //ModelMorpher
{Vec2i(16,16), Vec2i(144,112)}, //WorldInfo
{Vec2i(16,16), Vec2i(320,160)}, //TouchSensor
{Vec2i(16,16), Vec2i(192,128)}, //CCPath
{Vec2i(16,16), Vec2i(0,144)}, //BspTree
{Vec2i(16,16), Vec2i(224,144)}, //Inline
{Vec2i(16,16), Vec2i(416,96)}, //Fog
{Vec2i(16,16), Vec2i(432,96)}, //Fog2
{Vec2i(16,16), Vec2i(384,96)}, //Anchor
{Vec2i(16,16), Vec2i(304,160)}, //AudioClip
{Vec2i(16,16), Vec2i(400,96)}, //Billboard
{Vec2i(16,16), Vec2i(boxBuildIcon_x,boxBuildIcon_y)}, //Box
{Vec2i(16,16), Vec2i(coneBuildIcon_x,coneBuildIcon_y)}, //Cone
{Vec2i(16,16), Vec2i(cylinderBuildIcon_x,cylinderBuildIcon_y)}, //Cylinder
{Vec2i(16,16), Vec2i(160,112)}, //CylinderSensor
{Vec2i(16,16), Vec2i(352,160)}, //ElevationGrid
{Vec2i(16,16), Vec2i(176,128)}, //Extrusion
{Vec2i(16,16), Vec2i(416,144)}, //FontStyle
{Vec2i(16,16), Vec2i(432,144)}, //IndexedLineSet
{Vec2i(16,16), Vec2i(112,96)}, //NurbsSurface
{Vec2i(16,16), Vec2i(224,160)}, //LOD
{Vec2i(16,16), Vec2i(336,96)}, //MovieTexture
{Vec2i(16,16), Vec2i(224,160)}, //PixelTexture
{Vec2i(16,16), Vec2i(320,144)}, //PlaneSensor
{Vec2i(16,16), Vec2i(464,96)}, //PointLight
{Vec2i(16,16), Vec2i(224,160)}, //PointSet
{Vec2i(16,16), Vec2i(160,128)}, //ProximitySensor
{Vec2i(16,16), Vec2i(192,160)}, //Sound
{Vec2i(16,16), Vec2i(sphereBuildIcon_x, sphereBuildIcon_y)}, //Sphere
{Vec2i(16,16), Vec2i(176,96)}, //SphereSensor
{Vec2i(16,16), Vec2i(384,144)}, //SpotLight
{Vec2i(16,16), Vec2i(80,128)}, //Switch
{Vec2i(16,16), Vec2i(400,144)}, //Text
{Vec2i(16,16), Vec2i(128,128)}, //VisibilitySensor
{Vec2i(16,16), Vec2i(0,160)}, //TextureMapper
{Vec2i(16,16), Vec2i(320,96)}, //TextureEffect
{Vec2i(16,16), Vec2i(128,160)}, //ColorEffect
{Vec2i(16,16), Vec2i(112,96)}, //MonoColorEffect
{Vec2i(16,16), Vec2i(128,144)}, //ExitBuild
{Vec2i(16,16), Vec2i(320,112)}, //SplitTri
{Vec2i(16,16), Vec2i(swapDivide_x,swapDivide_y)}, //SwapTriDivide
{Vec2i(16,16), Vec2i(192,144)}, //PlaceTexture
{Vec2i(16,16), Vec2i(448,96)}, //HAnimHumanoid
{Vec2i(16,16), Vec2i(0,96)}, //HAnimJoint
{Vec2i(16,16), Vec2i(96,128)}, //HAnimDisplacer
{Vec2i(16,16), Vec2i(112,96)}, //HAnimSegment
{Vec2i(16,16), Vec2i(112,96)}, //HAnimSite
{Vec2i(16,16), Vec2i(112,96)}, //BooleanFilter
{Vec2i(16,16), Vec2i(112,96)}, //BooleanSequencer
{Vec2i(16,16), Vec2i(112,96)}, //BooleanToggle
{Vec2i(16,16), Vec2i(112,96)}, //BooleanTrigger
{Vec2i(16,16), Vec2i(112,96)}, //IntegerSequencer
{Vec2i(16,16), Vec2i(112,96)}, //IntegerTrigger
{Vec2i(16,16), Vec2i(112,96)}, //TimeTrigger
{Vec2i(16,16), Vec2i(368,96)}, //StaticGroup
{Vec2i(16,16), Vec2i(208,160)}, //FloatAccelerator
{Vec2i(16,16), Vec2i(208,144)}, //RotAccelerator
{Vec2i(16,16), Vec2i(144,96)}, //CopyToMorph
{Vec2i(16,16), Vec2i(416,128)}, //RelateMorpher
{Vec2i(16,16), Vec2i(128,96)}, //ImmediatePart
{Vec2i(16,16), Vec2i(256,160)}, //DelPlane
{Vec2i(16,16), Vec2i(256,160)}, //DelSphere
{Vec2i(16,16), Vec2i(336,144)}, //RadialTug
{Vec2i(16,16), Vec2i(144,144)}, //TapeMeasure
{Vec2i(16,16), Vec2i(400,128)}, //PosAccelerator
{Vec2i(16,16), Vec2i(288,144)}, //AnimByAnkle
{Vec2i(16,16), Vec2i(304,144)}, //AnimByWrist
{Vec2i(16,16), Vec2i(240,160)}, //RotFilter
{Vec2i(16,16), Vec2i(272,144)}, //RotMult
{Vec2i(16,16), Vec2i(432,128)}, //MultiUser
{Vec2i(16,16), Vec2i(240,128)}, //IndexedTriangleSet
{Vec2i(16,16), Vec2i(432,144)}, //Polyline2D

{Vec2i(16,16), Vec2i(sphereBuildIcon_x, sphereBuildIcon_y)}, //SphereBuild
{Vec2i(16,16), Vec2i(cylinderBuildIcon_x, cylinderBuildIcon_y)}, //CylinderBuild
{Vec2i(16,16), Vec2i(coneBuildIcon_x, coneBuildIcon_y)}, //ConeBuild
{Vec2i(16,16), Vec2i(torusBuildIcon_x, torusBuildIcon_y)}, //TorusBuild
{Vec2i(16,16), Vec2i(boxBuildIcon_x, boxBuildIcon_y)}, //BoxBuild
{Vec2i(16,16), Vec2i(rectangleBuildIcon_x, rectangleBuildIcon_y)}, //RectangleBuild
{Vec2i(16,16), Vec2i(bezierLatheBuildIcon_x, bezierLatheBuildIcon_y)}, //BezierLatheBuild

{Vec2i(16,16), Vec2i(16*2,112+16)}, //MirrorEdit
{Vec2i(16,16), Vec2i(16*3,112+16)}, //MirrorGroup

{Vec2i(16,16), Vec2i(128+16*2, 176)}, //VertexRing
{Vec2i(16,16), Vec2i(128+16, 176)}, //RingBuild
{Vec2i(16,16), Vec2i(64,144)}, //BezierScale
{Vec2i(16,16), Vec2i(240,112)}, //BezierCircumScale
{Vec2i(16,16), Vec2i(112,96)}, //BezierPatch
{Vec2i(16,16), Vec2i(112,96)}, //BezierPatchLink
{Vec2i(16,16), Vec2i(464,96)}, //PovLight
{Vec2i(16,16), Vec2i(416,96)}, //PovFog
{Vec2i(16,16), Vec2i(96,96)}, //BezierShift
{Vec2i(16,16), Vec2i(80,96)}, //BezierTwist
{Vec2i(16,16), Vec2i(80,96)}, //BezierTape
{Vec2i(16,16), Vec2i(80,96)}, //BezierBend
{Vec2i(16,16), Vec2i(160,192)}, //NurbsPatch
{Vec2i(16,16), Vec2i(160+16,192)}, //NurbsLathe
{Vec2i(16,16), Vec2i(160+16*3,192)}, //BendyLathe
{Vec2i(16,16), Vec2i(240,192)}, //FuseSurface
{Vec2i(16,16), Vec2i(144,192)}, //Subdivision
{Vec2i(16,16), Vec2i(80,192)}, //Oscillator
{Vec2i(16,16), Vec2i(240,112)}, //Envelope
{Vec2i(16,16), Vec2i(144,192)}, //AmplitudeModulator
{Vec2i(16,16), Vec2i(80,96)}, //FrequencyModulator
{Vec2i(16,16), Vec2i(96,192)}, //Chords
{Vec2i(16,16), Vec2i(96,192)}, //Larynx
{Vec2i(16,16), Vec2i(128,192)}, //Phrase
{Vec2i(16,16), Vec2i(96,192)}, //Harmonic
{Vec2i(16,16), Vec2i(96,192)}, //HarmonicPatch
{Vec2i(16,16), Vec2i(112,192)}, //Sample
{Vec2i(16,16), Vec2i(224,160)}, //WaveToPoly
{Vec2i(16,16), Vec2i(64,192)}, //Track
{Vec2i(16,16), Vec2i(48,192)}, //Composition
{Vec2i(16,16), Vec2i(96,128)}, //CPMorpher

{Vec2i(16,16), Vec2i(112,96)}, //G_S_Bool
{Vec2i(16,16), Vec2i(112,96)}, //G_S_Color
{Vec2i(16,16), Vec2i(112,96)}, //G_S_Float
{Vec2i(16,16), Vec2i(112,96)}, //G_S_Image
{Vec2i(16,16), Vec2i(112,96)}, //G_S_Int32
{Vec2i(16,16), Vec2i(112,96)}, //G_NodePtr
{Vec2i(16,16), Vec2i(112,96)}, //G_Rotation
{Vec2i(16,16), Vec2i(112,96)}, //G_S_String
{Vec2i(16,16), Vec2i(112,96)}, //G_S_Time
{Vec2i(16,16), Vec2i(112,96)}, //G_S_Vec2f
{Vec2i(16,16), Vec2i(112,96)}, //G_S_Vec3f
{Vec2i(16,16), Vec2i(112,96)}, //G_M_Bool
{Vec2i(16,16), Vec2i(112,96)}, //G_M_Color
{Vec2i(16,16), Vec2i(112,96)}, //G_M_Float
{Vec2i(16,16), Vec2i(112,96)}, //G_M_Image
{Vec2i(16,16), Vec2i(112,96)}, //G_M_Int32
{Vec2i(16,16), Vec2i(112,96)}, //G_M_Node
{Vec2i(16,16), Vec2i(112,96)}, //G_M_Rotation
{Vec2i(16,16), Vec2i(112,96)}, //G_M_String
{Vec2i(16,16), Vec2i(112,96)}, //G_M_Time
{Vec2i(16,16), Vec2i(112,96)}, //G_M_Vec2f
{Vec2i(16,16), Vec2i(112,96)}, //G_M_Vec3f
{Vec2i(16,16), Vec2i(240,96)}, //File
{Vec2i(16,16), Vec2i(224,160)}, //FloorHeight
{Vec2i(16,16), Vec2i(224,160)}, //FileAlreadyExists
{Vec2i(16,16), Vec2i(224,160)}, //SaveBeforeClosing
{Vec2i(16,16), Vec2i(224,160)}, //COLOR_SELECTOR
{Vec2i(16,16), Vec2i(224,160)}, //COMMON_SPEC
{Vec2i(16,16), Vec2i(224,160)}, //FreezeTime
{Vec2i(16,16), Vec2i(224,160)}, //Settings
{Vec2i(16,16), Vec2i(224,160)}, //StretchCurve
{Vec2i(24,24), Vec2i(224,160)}, //OpenNetFile
{Vec2i(16,16), Vec2i(224,160)}, //Demo
{Vec2i(16,16), Vec2i(160,192)}, //ConvertToPatch
{Vec2i(16,16), Vec2i(160,192)}, //ReduceCapturedPoses
{Vec2i(16,16), Vec2i(224,160)}, //TransformCP
{Vec2i(16,16), Vec2i(224,160)}, //AnimEditOptions
{Vec2i(16,16), Vec2i(224,160)}, //PasteSection
{Vec2i(16,16), Vec2i(224,160)}, //DeleteSection
};

/*
{Vec2i(16,16), Vec2i(400,160)}, //G_S_Bool
{Vec2i(16,16), Vec2i(448,112)}, //G_S_Color
{Vec2i(16,16), Vec2i(400,112)}, //G_S_Float
{Vec2i(16,16), Vec2i(112,96)}, //G_S_Image
{Vec2i(16,16), Vec2i(384,112)}, //G_S_Int32
{Vec2i(16,16), Vec2i(416,160)}, //G_NodePtr
{Vec2i(16,16), Vec2i(464,112)}, //G_Rotation
{Vec2i(16,16), Vec2i(384,160)}, //G_S_String
{Vec2i(16,16), Vec2i(416,112)}, //G_S_Time
{Vec2i(16,16), Vec2i(432,160)}, //G_S_Vec2f
{Vec2i(16,16), Vec2i(432,112)}, //G_S_Vec3f
{Vec2i(16,16), Vec2i(176,160)}, //G_M_Bool
{Vec2i(16,16), Vec2i(368,144)}, //G_M_Color
{Vec2i(16,16), Vec2i(464,160)}, //G_M_Float
{Vec2i(16,16), Vec2i(112,96)}, //G_M_Image
{Vec2i(16,16), Vec2i(448,160)}, //G_M_Int32
{Vec2i(16,16), Vec2i(176,112)}, //G_M_Node
{Vec2i(16,16), Vec2i(368,128)}, //G_M_Rotation
{Vec2i(16,16), Vec2i(368,112)}, //G_M_String
{Vec2i(16,16), Vec2i(464,144)}, //G_M_Time
{Vec2i(16,16), Vec2i(176,144)}, //G_M_Vec2f
{Vec2i(16,16), Vec2i(368,160)}, //G_M_Vec3f
*/
ProtoDeclare* getSceneProtoDeclare(int selLine){
	int remIndent=g_sceneTreeList[selLine].indent;
	while(selLine>=0){
		if(g_sceneTreeList[selLine].node->getType()==PROTODECLARE){
			return dynamic_cast<ProtoDeclare*>(g_sceneTreeList[selLine].node.p);
		}
		selLine--;
	}
	return 0;
}
/*
void colorBlockColorRgn(M_Properties& prop, RECT rc, int dstWidth, int dstHeight, int* pixBuf){
	int remBottom=rc.bottom;
	int lastYchange=0;
	int i=0;
	int ctr=0;
	int lastGroup=-1;
	bool togleColor=false;
	for(i=0; i<mfield.size(); i++,ctr++){
				if(togleColor!=true){
					g_propWinLine[ctr].backgroundColor=RGB(255, 255, 204);
				}
				else{
					g_propWinLine[ctr].backgroundColor=0xffffff;
				}
		Field* field=mfield[i];
		if((field->isWritable()^field->isReadable())==false&&field->getType()!=MFNODE&&field->getType()!=SFNODE){
			lastGroup=prop[i].groupIndex;
			while(i<mfield.size()&&prop[i].groupIndex==lastGroup){
				i++;
				if(togleColor!=true){
					g_propWinLine[ctr].backgroundColor=RGB(255, 255, 204);
				}
				else{
					g_propWinLine[ctr].backgroundColor=g_white;
				}
				ctr++;
			}
			--ctr;
			--i;
			if(i!=-1){
				rc.bottom=g_propWinLine[i].pos+20;
			}
			if(togleColor^=true){
				paintBlock(0xFFFFCC, rc.left, rc.top, rc.right+1, rc.bottom, dstWidth, dstHeight, pixBuf);
			}
			else{
				paintBlock(0xFFFFFF, rc.left, rc.top, rc.right+1, rc.bottom, dstWidth, dstHeight, pixBuf);
			}
			rc.top=rc.bottom;
		}
	}
	rc.bottom=remBottom;
	if(togleColor^=true){
		paintBlock(0xFFFFCC, rc.left, rc.top, rc.right+1, rc.bottom, dstWidth, dstHeight, pixBuf);
	}
	else{
		paintBlock(0xFFFFFF, rc.left, rc.top, rc.right+1, rc.bottom, dstWidth, dstHeight, pixBuf);
	}
}
*/
bool nodeExist(Node* node){
	for(int k=g_sceneTreeList.length; --k>=0;){
		if(g_sceneTreeList[k].node==node){
			return true;
		}
	}
	return false;
}
void deleteAllRoutesToNonExistant(){
	for(int k=g_sceneTreeList.length; --k>=0;){
		Node* node=g_sceneTreeList[k].node;
		for(int rx=0,i=0; i<node->getFLen(); i++){
			Field* field=node->getField(i);
			const M_Route& route=field->getRoute();
			for(int r=0; r<route.length; ){
				if(nodeExist(route[r].toNode)){
					r++;
					rx++;
				}
				else {
					node->deleteRoute(rx);
				}
			}
		}
	}
}
bool matchName(M_Art& partList, String name, Art*& part){
	for(int k=partList.length; --k>=0;){
		part=partList[k];
		if(part->name==name){
			return true;
		}
	}
	return false;
}
void redirectTransformRoutesToParts(M_Art& partList){
	for(int k=g_sceneTreeList.length; --k>=0;){
		Node* node=g_sceneTreeList[k].node;
		for(int i=0; i<node->getFLen(); i++){
			Field* field=node->getField(i);
			M_Route& route=field->refRoute();
			for(int r=0; r<route.length; r++){
				Art* part=0;
				if(Transform* transform=dynamic_cast<Transform*>(route[r].toNode)){
					if(matchName(partList, transform->name, part)){
						if(route[r].toFieldPtr==&transform->scale){
							route[r]=S_Route( part, &part->scale);
						}
						else if(route[r].toFieldPtr==&transform->rotation){
							route[r]=S_Route( part, &part->rotation);
						}
						else if(route[r].toFieldPtr==&transform->translation){
							route[r]=S_Route( part, &part->translation);
						}
					}
				}
				else if(HAnimJoint* hAnimJoint=dynamic_cast<HAnimJoint*>(route[r].toNode)){
					if(matchName(partList, hAnimJoint->name, part)){
						if(route[r].toFieldPtr==&hAnimJoint->scale){
							route[r]=S_Route( part, &part->scale);
						}
						else if(route[r].toFieldPtr==&hAnimJoint->rotation){
							route[r]=S_Route( part, &part->rotation);
						}
						else if(route[r].toFieldPtr==&hAnimJoint->translation){
							route[r]=S_Route( part, &part->translation);
						}
					}
				}
			}
		}
	}
}

void removeNode(int i, M_Node &mnode, int ammount){
	for(; i<mnode.length-ammount; i++){
		mnode[i]=mnode[i+ammount];
	}
	mnode.length-=ammount;
}
void TextureMapper::Size2d::operator=(const Vec2f& par){
	refValue()=par;
	updateBuild3dWindow();
}
void TextureMapper::Position2d::operator=(const Vec2f& par){
	refValue()=par;
	updateBuild3dWindow();
}
void TextureMapper::Clear::operator=(const bool& par){
	refValue()=false;

	home.position2d.refValue()=Vec2f();
	home.size2d.refValue()=Vec2f();

	updateBuild3dWindow();
}
/*
void updateDisplay(Node* node, const int fieldIndex, const int element){
	Field* field=node->getField(fieldIndex);

	int item=FIELDITEMBASE+fieldIndex*CTRLF_DIV;
	if(node->getProperties()[fieldIndex].groupIndex>=0 &&(element<0||element>=field->size())){
		setDlgItem( g_hDlg, item, "");
	}
	else{
		switch(field->getType()){
		case SFBOOL:	CheckDlgButton(g_hDlg, item, dynamic_cast<SFBool*>(field)->refValue()); break;
		case SFCOLOR:	setDlgItem(g_hDlg, item, dynamic_cast<SFColor*>(field)->refValue()); break;
		case SFFLOAT:	setDlgItem(g_hDlg, item, dynamic_cast<SFFloat*>(field)->refValue()); break;
	//	case SFIMAGE:	setDlgItem(g_hDlg, item, dynamic_cast<SFImage*>(field)->refValue()); break;
		case SFINT32:	setDlgItem(g_hDlg, item, dynamic_cast<SFInt32*>(field)->refValue()); break;
	//	case SFNODE:	setDlgItem(g_hDlg, item, dynamic_cast<SFNode*>(field)->refValue()); break;
		case SFROTATION: setDlgItem(g_hDlg, item, dynamic_cast<SFRotation*>(field)->refValue()); break;
		case SFSTRING:	setDlgItem(g_hDlg, item, dynamic_cast<SFString*>(field)->refValue()); break;
		case SFTIME:	setDlgItem(g_hDlg, item, dynamic_cast<SFTime*>(field)->refValue()); break;
		case SFVEC2F:	setDlgItem(g_hDlg, item, dynamic_cast<SFVec2f*>(field)->refValue()); break;
		case SFVEC3F:	setDlgItem(g_hDlg, item, dynamic_cast<SFVec3f*>(field)->refValue()); break;

		case MFBOOL:	CheckDlgButton(g_hDlg, item, dynamic_cast<MFBool*>(field)->refValue()[element]); break;
		case MFCOLOR:	setDlgItem(g_hDlg, item, dynamic_cast<MFColor*>(field)->refValue()[element]); break;
		case MFFLOAT:	setDlgItem(g_hDlg, item, dynamic_cast<MFFloat*>(field)->refValue()[element]); break;
	//	case MFIMAGE:	setDlgItem(g_hDlg, item, dynamic_cast<MFImage*>(field)->refValue()[element]); break;
		case MFINT32:	setDlgItem(g_hDlg, item, dynamic_cast<MFInt32*>(field)->refValue()[element]); break;
	//	case MFNODE:	setDlgItem(g_hDlg, item, dynamic_cast<MFNode*>(field)->refValue()[element]); break;
		case MFROTATION: setDlgItem(g_hDlg, item, dynamic_cast<MFRotation*>(field)->refValue()[element]); break;
		case MFSTRING:	setDlgItem(g_hDlg, item, dynamic_cast<MFString*>(field)->refValue()[element]); break;
		case MFDOUBLE:	setDlgItem(g_hDlg, item, dynamic_cast<MFTime*>(field)->refValue()[element]); break;
		case MFVEC2F:	setDlgItem(g_hDlg, item, dynamic_cast<MFVec2f*>(field)->refValue()[element]); break;
		case MFVEC3F:	setDlgItem(g_hDlg, item, dynamic_cast<MFVec3f*>(field)->refValue()[element]); break;
		}
	}
}

void Scene::setSceneFromViewpoint(const Viewpoint& viewpoint){ //scene transformation is the camera look at
	g_settings->perspective = viewpoint.perspective;

	scale = Vec3f(viewpoint.zoom, viewpoint.zoom, viewpoint.zoom);
	translation = (viewpoint.orientation.refValue() * Vec3f(0, 0, g_viewDist) - viewpoint.position) / scale.x;
	fieldOfView = viewpoint.fieldOfView;
	//scale = Vec3f(1.815, 1.815, 1.815);
	//translation = Vec3f(0, 0, -4/1.815);
	//translation = Vec3f();
}
void Scene::setViewpointFromScene(Viewpoint& viewpoint){ //set viewpoint from camera look at + g_viewDist
	viewpoint.zoom.refValue() = scale.x;
	viewpoint.perspective.refValue() = g_settings->perspective;
	//viewpoint.position.refValue() = -translation + (-rotation * Vec3f(0, 0, g_viewDist));
	viewpoint.position.refValue() = -g_scene->translation * g_scene->scale.x + Vec3f(0,0,g_viewDist);
	viewpoint.orientation.refValue() = -rotation;
	Window::activeWindow->paint();
}

*/

void Scene::setSceneFromViewpoint(const Viewpoint& viewpoint){ //scene transformation is the camera look at
	Viewpoint::neutral = false;

	g_settings->perspective = viewpoint.perspective;

	scale = Vec3f(viewpoint.zoom, viewpoint.zoom, viewpoint.zoom);
	rotation = -viewpoint.orientation;
	translation = (viewpoint.orientation.refValue() * Vec3f(0, 0, g_viewDist)) - viewpoint.position;
	fieldOfView = viewpoint.fieldOfView;
	//scale = Vec3f(1.65,1.65,1);
	//translation = Vec3f();
	//rotation = Rotation();
}
void Scene::setViewpointFromScene(Viewpoint& viewpoint){ //set viewpoint from camera look at + g_viewDist
	viewpoint.zoom.refValue() = scale.x;
	viewpoint.perspective.refValue() = g_settings->perspective;
	viewpoint.position.refValue() = -translation + (-rotation * Vec3f(0, 0, g_viewDist));
	viewpoint.orientation.refValue() = -rotation;
	Window::activeWindow->paint();
}
/*
void setViewPoint(Node* node){
	Viewpoint* viewpoint=dynamic_cast<Viewpoint*>(node);
	if(viewpoint){
		g_settings->perspective=viewpoint->perspective;
		g_scene->scale=Vec3f(viewpoint->zoom,viewpoint->zoom,viewpoint->zoom);
		g_scene->rotation=-viewpoint->orientation;
		g_scene->translation=(viewpoint->orientation.refValue()*Vec3f(0,0,3))-viewpoint->position;
	}
}

	home.zoom.refValue()=g_scene->scale.x;
	home.perspective.refValue()=g_settings->perspective;
	home.position.refValue()=-g_scene->translation+(-g_scene->rotation*Vec3f(0,0,3));
	home.orientation.refValue()=-g_scene->rotation;
*/
M_<Ptr<Viewpoint> > g_viewpoints;
extern int g_viewIndex;
void countUpViewpoints(){
	g_viewpoints.length=0;
	for(int k=0; k<g_sceneTreeList.length; k++){
		if(Viewpoint* vp=dynamic_cast<Viewpoint*>(g_sceneTreeList[k].node.p)){
			g_viewpoints+=vp;
		}
	}
}

/*
void leftClick(HWND hWnd, int x, int y){

	if(g_dragNode.leftClick(x, y)){
		g_leftButtonDown=true;
		return;
	}
	

	g_selCount.reset();
	int tmpInt=(y-SceneTree::beginYOffset)/pixSymbolCY+g_scrollPosCollapsed;

	if(tmpInt>=g_sceneTreeCollapsed.length)return;
	Node* node=g_sceneTreeCollapsed[tmpInt].node;
	int newCaretPos=(x+charWidth/2-(tLnOffst+g_sceneTreeCollapsed[tmpInt].indent*pixSymbolCX+16+3))/charWidth;
	int indentPix = tLnOffst+g_sceneTreeCollapsed[tmpInt].indent*pixSymbolCX;

	if(node->getType()!=SCENE && x<=indentPix){

		if(node->collapsed || getChildrenLen(tmpInt)){
			node->collapsed=!node->collapsed;
			genNodeList();
			g_textCell.updateCaret();
			PostMessage(hWnd, WM_PAINT,	0,	0	);
		}
		return;
	}
	else if(Art* part=dynamic_cast<Art*>(node)){
		if(	x > indentPix&&x < pixSymbolCX + indentPix-2){

			part->show = part->show!=true;
//			if(part->collapsed){ //interferes with current robot sequence 
//				recurseShow(part, part->show);
//			}
			PostMessage(hWnd, WM_PAINT,	0,	0	);

			part->smls->convertToRenderFormat();
			BackUp::monostable.beginCycle();
			PostMessage(hWndMain, WM_PAINT,	0,	0	);
			return;
		}
	}
	else if(Viewpoint* viewpoint=dynamic_cast<Viewpoint*>(node)){
		setViewPoint(viewpoint);
		countUpViewpoints();
		g_viewIndex=g_viewpoints.length;
		while(g_viewpoints[--g_viewIndex]!=node);
	}
	g_leftButtonDown=true;
	int newSelNodeLine=getLineForNonCollapsed(tmpInt);
	selectNode(newSelNodeLine, node, newCaretPos, tmpInt);
}
*/
void setUpBitmap(HWND hWnd, BitmapStuf& bs){
	HDC hdc=GetDC(hWnd);
	RECT rc;
	GetClientRect(hWnd, &rc);

	bs.hBitmap = CreateCompatibleBitmap(hdc, rc.right-rc.left, rc.bottom-rc.top);

	GetObject(bs.hBitmap, sizeof(BITMAP), &bs.bitMap);
	bs.bitmapInfoHeader.biSize=sizeof(BITMAPINFOHEADER);
	bs.bitmapInfoHeader.biWidth=bs.bitMap.bmWidth; 
	bs.bitmapInfoHeader.biHeight=bs.bitMap.bmHeight;
	bs.bitmapInfoHeader.biPlanes=bs.bitMap.bmPlanes;
	bs.bitmapInfoHeader.biBitCount=32;
	bs.bitmapInfoHeader.biCompression=BI_RGB;
	bs.bitmapInfoHeader.biSizeImage=0; 
	bs.bitmapInfoHeader.biXPelsPerMeter=0; 
	bs.bitmapInfoHeader.biYPelsPerMeter=0; 
	bs.bitmapInfoHeader.biClrUsed=0;
	bs.bitmapInfoHeader.biClrImportant=0;

	bs.pixBuf.length=bs.bitmapInfoHeader.biHeight*bs.bitmapInfoHeader.biWidth;
	
	bs.hdc = CreateCompatibleDC(hdc);
	SelectObject(bs.hdc, bs.hBitmap);
}

/*
bool nodeWindowKeyInput(unsigned wParam){
	if(0x8000&GetKeyState(VK_LCONTROL)){
		switch (wParam){
		case 'C':
			remember();
			break;
		case 'Z':
			unDo();
			return true;
		case 'Y':
			reDo();
			return true;
		default:
			return false;
		}
	}
	return true;
}
bool SceneTree::keyInput(unsigned wParam){
	if(0x8000&GetKeyState(VK_LCONTROL)){
		switch (wParam){
		case 'C':{
			remember();
			break;
		}
		case 'X':{
			if(textCell.highlightPos!=-1 && textCell.caretPos!=textCell.highlightPos){
				if (OpenClipboard(hWndMain)){
					EmptyClipboard(); 
		 
					// Allocate a global memory object for the text. 
			 
					int highlightLen=abs(textCell.caretPos-textCell.highlightPos);
					int index=textCell.highlightPos;
					if(textCell.caretPos<textCell.highlightPos){
						index=textCell.caretPos;
					}

					HGLOBAL hglbCopy = GlobalAlloc(GMEM_MOVEABLE, abs(textCell.caretPos-textCell.highlightPos) + 1);
					if (hglbCopy == NULL) 
					{ 
						CloseClipboard(); 
					}
					else{
			 
						// Lock the handle and copy the text to the buffer. 
				 
						LPVOID lptstrCopy = GlobalLock(hglbCopy); 
						memcpy(lptstrCopy, &textCell.cellStr[index], highlightLen + 1); 

						((char*)lptstrCopy)[highlightLen] = 0;    // null character 
						GlobalUnlock(hglbCopy); 
				 
						// Place the handle on the clipboard. 
				 
						SetClipboardData(CF_TEXT, hglbCopy);
						CloseClipboard();

						textCell.deleteHighlight();
						textCell.updateCaret();
						paint();
					}
				} 
			}
			break;
		}
		case 'Z':
			unDo();
			return true;
		case 'Y':
			reDo();
			return true;
		case 'V':{
			if (!IsClipboardFormatAvailable(CF_TEXT)) 
				break; 
			if (!OpenClipboard(hWndMain)) 
				break; 
	 
			HGLOBAL hglb = GetClipboardData(CF_TEXT); 
			if (hglb != NULL) 
			{ 
				LPTSTR lptstr = (LPTSTR)GlobalLock(hglb); 
				if (lptstr != NULL) 
				{ 
					// Call the application-defined ReplaceSelection 
					// function to insert the text and repaint the 
					// window. 
					textCell.deleteHighlight();

					int remLen=textCell.cellStr.length;
					textCell.cellStr=textCell.cellStr.replace(textCell.caretPos, 0, lptstr);
					
 					GlobalUnlock(hglb); 
					textCell.caretPos+=textCell.cellStr.length-remLen;
					textCell.highlightPos=textCell.caretPos;

				} 
			} 
			CloseClipboard(); 
			textCell.updateCaret();
			paint();
			break;
		}
		default:
			return false;
		}
	}
	else if(0x8000&GetKeyState(VK_LSHIFT)){
		switch (wParam){
		case VK_HOME:{
			textCell.caretPos=0;
			paint();
			textCell.updateCaret();
			break;
		}
		case VK_END:{
			textCell.caretPos=textCell.cellStr.length;
			paint();
			textCell.updateCaret();
			break;
		}
		case VK_UP:
			if(g_selCount>1){
				int collapsedLine;
				if(getCollapsedLine(selLine, collapsedLine)){
					g_selCount--;
					int a=collapsedLine+g_selCount.all;
					if(g_scrollPosCollapsed>collapsedLine&&g_scrollPosCollapsed+g_numOfVisLines>a){
						g_scrollPosCollapsed=a-g_numOfVisLines;
					}
					//SetScrollPos( hSceneTreeWnd, SB_VERT, g_scrollPosCollapsed, TRUE );
					paint();
				}
			}
			break;
		case VK_DOWN:
			int collapsedLine;
			if(getCollapsedLine(selLine, collapsedLine)){
				g_selCount++;
				if(collapsedLine+g_selCount.all>g_scrollPosCollapsed+g_numOfVisLines){
					g_scrollPosCollapsed=collapsedLine+g_selCount.all-g_numOfVisLines;
				}
				textCell.updateCaret();
//				SetScrollPos( hSceneTreeWnd, SB_VERT, g_scrollPosCollapsed, TRUE );
				paint();
			}
			break;
		case VK_RIGHT:{
			textCell.incCaretPos();
			paint();
			break;
		}
		case VK_LEFT:{
			textCell.decCaretPos();
			paint();
			break;
		}
		default:
			return false;
		}
	}
	else {
		switch (wParam){
		case VK_DELETE:{
			if(textCell.deleteHighlight()){
				paint();
			}
			else if(textCell.caretPos < textCell.cellStr.length){
				textCell.cellStr.deleteChar(textCell.caretPos);
				paint();
			}
			break;
		}
		case VK_RIGHT:{
			textCell.incCaretPosNH();
			break;
		}
		case VK_LEFT:{
			textCell.decCaretPosNH();
			break;
		}
		case VK_HOME:{
			textCell.caretPos=textCell.highlightPos=0;
			paint();
			textCell.updateCaret();
			break;
		}
		case VK_END:{
			textCell.caretPos=textCell.highlightPos=textCell.cellStr.length;
			paint();
			textCell.updateCaret();
			break;
		}
		case VK_UP:{
			int collapsedLine;
			if(getCollapsedLine(selLine, collapsedLine)){
				if(collapsedLine>0){
					int x = tLnOffst+g_sceneTreeCollapsed[collapsedLine].indent*pixSymbolCX+16+3 + textCell.caretPos*charWidth;
					collapsedLine--;
					if(static_cast<unsigned>(collapsedLine - 1) < static_cast<unsigned>(g_scrollPosCollapsed)){
						g_scrollPosCollapsed=collapsedLine - 1;
					}

					int _x = tLnOffst+g_sceneTreeCollapsed[collapsedLine].indent*pixSymbolCX+16+3;
					int y = SceneTree::beginYOffset+(collapsedLine-g_scrollPosCollapsed)*pixSymbolCY+12;

					if(_x > x){
						x = _x;
					}
					click(Vec2i(x, y));
//					SetScrollPos( hSceneTreeWnd, SB_VERT, g_scrollPosCollapsed, TRUE );

				}
			}
			else{
//				HideCaret(hSceneTreeWnd);
			}
			break;
		}
		case VK_DOWN:{
			int collapsedLine;
			if(getCollapsedLine(selLine, collapsedLine)){
				if(collapsedLine<g_sceneTreeCollapsed.length-1){
					int x = tLnOffst+g_sceneTreeCollapsed[collapsedLine].indent*pixSymbolCX+16+3 + textCell.caretPos*charWidth;
					collapsedLine++;
					if(collapsedLine + 1 > g_scrollPosCollapsed+g_numOfVisLines){
						g_scrollPosCollapsed=collapsedLine-g_numOfVisLines + 1;
					}

					int _x = tLnOffst+g_sceneTreeCollapsed[collapsedLine].indent*pixSymbolCX+16+3;
					int y = SceneTree::beginYOffset+(collapsedLine-g_scrollPosCollapsed)*pixSymbolCY+12;

					if(_x > x){
						x = _x;
					}
					click(Vec2i(x, y));
//					SetScrollPos( hSceneTreeWnd, SB_VERT, g_scrollPosCollapsed, TRUE );
				}
			}
			else{
//				HideCaret(hSceneTreeWnd);
			}
			break;
		}
		case VK_PRIOR:
			setScrolPos(-g_numOfVisLines);
			break;
		case VK_NEXT:
			setScrolPos(g_numOfVisLines);
			break;
		default:
			return false;
		}
	}
	return true;
}
*/
void setNodeWinTitleBar(Node* node){
	static String str;

	str=node->getTypeStr();
	/*
	if(node->getType()>=g_userProtoBegin){
		str="ProtoInstance ";
		str+=node->getTypeStr();
	}
	else{
		str=node->getTypeStr();
	}
*/
	str+=' ';
	str+=node->name;
	SetWindowText(g_hDlg, str.c_str());

/*
		String str=nodeLookUpStr[windowNode->getType()];
		str+=' ';
		str+=windowNode->name;
		SetWindowText(hWnd, str.c_str());
*/

/*
	str=node->name;
	str+=' ';
	if(node->getType()>=g_userProtoBegin){
		str+="ProtoInstance";
	}
	else{
		str+=node->getTypeStr();
	}
*/
}

/*
void updateVec3f(HWND hDlg,int ID, ArtPtr part, Vec3f val){
	if(part==NULL){
		return;
	}
	SeamlessPtr smls=part->smls;
	buildSceneMatrices( smls->skeleton[0], Matrix4());
	smls->convertToRenderFormat();
	setDlgItem( hDlg, ID, val);
	SendMessage(hWndMain, WM_PAINT,	0,	0	);
	BackUp::monostable.beginCycle();
}
void updateVec2f(HWND hDlg,int ID, ArtPtr part, Vec2f val){
	if(part==NULL){
		return;
	}
	SeamlessPtr smls=part->smls;
	buildSceneMatrices( smls->skeleton[0], Matrix4());
	smls->convertToRenderFormat();
	setDlgItem( hDlg, ID, val);
	SendMessage(hWndMain, WM_PAINT,	0,	0	);
	BackUp::monostable.beginCycle();
}
void buildUpdateVec2f(HWND hDlg,int ID, ArtPtr part, Vec2f val){
	if(part==NULL){
		return;
	}
	SeamlessPtr smls=part->smls;
	smls->doBuild();
	smls->convertToRenderFormat();
	setDlgItem( hDlg, ID, val);
	SendMessage(hWndMain, WM_PAINT,	0,	0	);
	BackUp::monostable.beginCycle();
}
void updateRotation(HWND hDlg,int ID, ArtPtr part, Rotation val){
	if(part==NULL){
		return;
	}
	SeamlessPtr smls=part->smls;
	buildSceneMatrices( smls->skeleton[0], Matrix4());
	smls->convertToRenderFormat();
	setDlgItem( hDlg, ID, val);
	SendMessage(hWndMain, WM_PAINT,	0,	0	);
	BackUp::monostable.beginCycle();
}


void updateSceneTree(ArtPtr part){
	if(part==NULL){
		return;
	}
	SeamlessPtr smls=part->smls;
	buildSceneMatrices( smls->skeleton[0], Matrix4());
	smls->convertToRenderFormat();
	SendMessage(hWndMain, WM_PAINT,	0,	0	);
	BackUp::monostable.beginCycle();
}
*/
void ModelMorpher::updateBuild(){
	M_Node& sequence = this->sequence.refValue();
	for(int k = sequence.length; --k>=0;){
		ModelSequence* modelSequence = dynamic_cast<ModelSequence*>(sequence[k].p);
		M_Node& model = modelSequence->model.refValue();
		for(int k = model.length; --k>=0;){
			Seamless* smls = dynamic_cast<Seamless*>(model[k].p);
			smls->doBuild();
			smls->ctrf = true;
		}
	}
}
void Seamless::updateBuild(){
	if(g_bReady==false)return;
	doBuild();
	convertToRenderFormat();
//	SendMessage(hWndMain, WM_PAINT,	0,	0	);
}
void AGroup::updateBuild(){
	M_Node& children=this->children.refValue();
	for( int k=0; k<children.length; k++){
		children[k]->updateBuild();
	}
}
void ProtoDeclare::updateBuild(){
	M_Node& children=dynamic_cast<ProtoBody*>(protoBody.getRawPtr())->children;
	for( int k=0; k<children.length; k++){
		children[k]->updateBuild();
	}
}

void updateBuild3dWindow(bool ifPartBuild){
	if(Button* button=DownButton::getDownButton()){
		if(ControlPanel* controlPanel=dynamic_cast<ControlPanel*>(&button->win)){
			controlPanel->paint();
			if(dynamic_cast<Demo*>(controlPanel->node.p) || dynamic_cast<Anim*>(controlPanel->node.p)){
				return;
			}
		}
	}
	Node* node=g_stw->cp->node;
	if((ifPartBuild && dynamic_cast<Art*>(node))||
		dynamic_cast<Build*>(node)!=0||
		dynamic_cast<Generator*>(node)!=0||
		dynamic_cast<Effect*>(node)!=0||
		dynamic_cast<SurfaceGenerator*>(node)!=0||
		dynamic_cast<TextureMapper*>(node)!=0||
		dynamic_cast<FuseSurface*>(node)!=0||
		g_stw->selLine!=0 && dynamic_cast<Build*>(g_sceneTreeList[getParent(g_stw->selLine)].node.p)!=0
		){
		//dynamic_cast<Seamless*>(&*g_selNode)!=0){
		g_scene->updateBuild();
		houseKeepVisualClue();
	}
	if(dynamic_cast<OpenNetFile*>(node)){
		return;
	}
	if(g_dontBackup==false){
		BackUp::monostable.beginCycle();
	}
	g_dontBackup=false;
	ControlPanel::paintAllControlPanelsExceptActive();
	
}
/*
void genFullPathForSmls(String fullPath, String inPartFile){
	if (inPartFile==""){
		return;
	}
	//are we starting with a directory?
	else if(inPartFile[0]=='\\'||inPartFile[0]=='/'){
		fullPath=inPartFile;
	}
	else if (inPartFile.length>0&&inPartFile[1]==':'){
		fullPath=inPartFile;
	}
	else{ //must be only file name
		fullPath=g_seamlessDir;
		fullPath+=inPartFile;
	}
}
*/
void selectOne(ArtPtr pj){
	M_SArt& children=pj->part.refValue();
	bool show=pj->show;

	if(pj==g_selNode.getRaw()){
		show=true;
	}
	else {
		show=false;
	}
	int k;
	if(children.length){
		for(k=0;k<pj->part.refValue().length;k++){
			selectOne(children[k]);
		}
	}
}
void selectAll(ArtPtr pj){
	pj->show=true;
	int k;
	if(pj->part.refValue().length){
		for(k=0;k<pj->part.refValue().length;k++){
			selectAll(pj->part.refValue()[k]);
		}
	}
}
int getParent(int selLine){
	return g_sceneTreeList[selLine].parent;
}
int getParent(Node* node){
	return getParent(getNodeLine(node));
}
int getListLength(int selLine, M_<SceneTreeListItem>& nodeList){
	if(selLine >= nodeList.length){
		return 0;
	}
	int i=selLine;
	int selectIndent = nodeList[selLine].indent;
	while(++i < nodeList.length && selectIndent < nodeList[i].indent);
	return i-selLine;
}
void getFNodeElementForParent(int parentLine, int selLine, M_<SceneTreeListItem>& nodeList, int& fieldIdx, int& itemIdx ){
	Node* parentNode=nodeList[parentLine].node;
	int lineCntr=parentLine+1;
	for(fieldIdx=0; fieldIdx<parentNode->getFLen(); fieldIdx++){
		Field* field=parentNode->getField(fieldIdx);
		if( field->getType()==MFNODE){
			M_Node& mnode=dynamic_cast<MFNode*>(field)->refValue();
			for(itemIdx=0; itemIdx<mnode.length; itemIdx++){
				if(lineCntr==selLine){
					return;
				}
				else {
					lineCntr+=getListLength(lineCntr, nodeList);
				}
			}
		}
		else if(field->getType()==SFNODE && (*dynamic_cast<SFNode*>(field)).refValue()!=0){
			if(lineCntr==selLine){
				return;
			}
			lineCntr+=getListLength(lineCntr, nodeList);
		}
	}
	throw parentNode;
}
void insertNodeAbove(int targetIndex, M_Node &mnode, NodePtr insertNode){

	mnode.length=mnode.length+1;

	int i=mnode.length-1;
	do{
		--i;
		mnode[i+1]=mnode[i];
	}while(i!=targetIndex);

	mnode[targetIndex]=insertNode;
}
void insertNodeBelow(int targetIndex, M_Node &mnode, NodePtr insertNode){

	mnode.length=mnode.length+1;

	int i=mnode.length-1;

	while(--i!=targetIndex){
		mnode[i+1]=mnode[i];
	};

	mnode[targetIndex+1]=insertNode;
}
/*
void insertColor(int targetIndex, int deleteLen, M_Color& src, M_Color& dst){
	int srcLen=src.length;

	int diff=deleteLen-srcLen;
	if(diff<0){ //make gap bigger
		diff=-diff;
		dst.length+=diff;

		int lastMove=targetIndex+deleteLen;
		for(int i=dst.length-diff; --i>=lastMove; ){
			dst[i+diff]=dst[i];
		}
	}
	else if(diff>0){	//make gap smaller
		for(int i=targetIndex+src.length+diff; i<dst.length-diff; i++){
			dst[i]=dst[diff+i];
		}
		dst.length-=diff;
	}
	for(int i=src.length; --i>=0; ){
		dst[targetIndex+i]=src[i];
	}
}
*/
bool checkNodeIsFreeOfRouteTies(Node* check){
	for(int k=0; k<g_sceneTreeList.length; k++){
		Node* node=g_sceneTreeList[k].node;
		if(node->getType()==PROTOINTERFACE)continue;
		if(nodeHasRouteTo(node,check) || nodeHasRouteTo(check,node)){
			return false;
		}
	}
	return true;
}
bool g_optimized;
void optimize(AGroup* groupingNode){
	M_Node& children = groupingNode->children;
	{
		bool flag;
		do{
			flag=false;
			for(int len=children.length, k=len; --k>=0;){
				Group* group=dynamic_cast<Group*>(children[k].p);
				Transform* transform=dynamic_cast<Transform*>(children[k].p);
				if(group){
					flag=true;
					M_Node store = group->children;
					insertElements(k, 1, children, store);
				}
				else if(transform){
					if(transform->children.length==0){
						flag=true;
						removeNode(k, children);
					}
					else if(
						transform->center==Vec3f()&&
						transform->scaleOrientation.refValue().angle==0&&
						transform->scale==Vec3f(1,1,1)&&
						transform->rotation.refValue().angle==0&&
						transform->translation==Vec3f()&&
						transform->bboxCenter==Vec3f(0,0,0)&&
						transform->bboxSize==Vec3f(-1,-1,-1)&&
						checkNodeIsFreeOfRouteTies(transform)
						){
						flag=true;
						M_Node store = transform->children;
						insertElements(k, 1, children, store);
					}
				}
			}
			if(flag)g_optimized=true;;
		}while(flag);
	}
	{
		for(int k=children.length; --k>=0;){
			AGroup* groupingNode=dynamic_cast<AGroup*>(children[k].p);
			if(groupingNode){
				optimize(groupingNode);
			}
		}
	}
}
float floatIncPIFrac = PI/(12*12);
const float colorIncrement =.01;//1./100.

float getIncDecRes(){
	if(0x8000&GetKeyState(VK_LCONTROL)){
		return g_settings->incDecResCtrl;
	}
	else{
		return g_settings->incDecRes;
	}
}

/*
int Art::winProc(void* hdlg,unsigned iMessage, unsigned wParam, unsigned lParam){
	HWND hDlg=(HWND)hdlg;
	NMUPDOWN* nmUpDown;
	float tmpFloat;
	Vec3f tmpVec;
	Rotation tmpRot;
	ArtPtr tmpPJ;
	
	switch (iMessage){
	case WM_INITDIALOG:
		initSpin4(hDlg, SCALE_E);
		initSpin3(hDlg, ROTATION_E);
		initSpin3(hDlg, TRANSLATION_E);
		break;
	case WM_ACTIVATE:
	case WM_USER:
		if(
			smls->ActCtrlPnt<fromCoord||
			smls->ActCtrlPnt>=toCoord
			){
			smls->ActCtrlPnt=fromCoord;
			smls->ActTriCorner=0;
			smls->ActTri=fromPoly;
		}
		setDlgItem( hDlg, SCALE_E, scale);
		setDlgItem( hDlg, ROTATION_E, rotation);
		setDlgItem( hDlg, TRANSLATION_E, translation);

		CheckDlgButton(hDlg,ISJOINT_C, isJoint);		
		
		setNodeWinTitleBar(hDlg,this);
		break;
	case WM_NOTIFY:
		if(UDN_DELTAPOS!=((LPNMHDR)lParam)->code)return FALSE;
		nmUpDown=(LPNMUPDOWN) lParam;
		switch (LOWORD(wParam)){
			
			
		case SCALE_XYZ:
			tmpFloat=g_incDecRes*(float)nmUpDown->iDelta;
			scale.x+=tmpFloat;
			scale.y+=tmpFloat;
			scale.z+=tmpFloat;
			updateVec3f(hDlg, SCALE_E, this, scale);
			break;
		case SCALE_X:
			scale.x+=g_incDecRes*(float)nmUpDown->iDelta;
			updateVec3f(hDlg, SCALE_E, this, scale);
			break;
		case SCALE_Y:
			scale.y+=g_incDecRes*(float)nmUpDown->iDelta;
			updateVec3f(hDlg, SCALE_E, this, scale);
			break;
		case SCALE_Z:
			scale.z+=g_incDecRes*(float)nmUpDown->iDelta;
			updateVec3f(hDlg, SCALE_E, this, scale);
			break;
		case ROTATION_X:
			rotation=rotation.multiply(Rotation(((ArtPtr)smls->skeleton[0])->unWindRot(this).multVec(Vec3f(1,0,0)), floatIncPIFrac*(float)nmUpDown->iDelta));
			updateRotation(hDlg, ROTATION_E, this, rotation);
			break;
		case ROTATION_Y:
			rotation=rotation.multiply(Rotation(((ArtPtr)smls->skeleton[0])->unWindRot(this).multVec(Vec3f(0,1,0)), floatIncPIFrac*(float)nmUpDown->iDelta));
			updateRotation(hDlg, ROTATION_E, this, rotation);
			break;
		case ROTATION_Z:
			rotation=rotation.multiply(Rotation(((ArtPtr)smls->skeleton[0])->unWindRot(this).multVec(Vec3f(0,0,1)), floatIncPIFrac*(float)nmUpDown->iDelta));
			updateRotation(hDlg, ROTATION_E, this, rotation);
			break;
		case TRANSLATION_X:
			translation+=((ArtPtr)smls->skeleton[0])->unWindMatrixNoTran( this).multVecMatrix(Vec3f(g_incDecRes*(float)nmUpDown->iDelta,0,0));
			updateVec3f(hDlg, TRANSLATION_E, this, translation);
			break;
		case TRANSLATION_Y:
			translation+=((ArtPtr)smls->skeleton[0])->unWindMatrixNoTran( this).multVecMatrix(Vec3f(0,g_incDecRes*(float)nmUpDown->iDelta,0));
			updateVec3f(hDlg, TRANSLATION_E, this, translation);
			break;
		case TRANSLATION_Z:
			translation+=((ArtPtr)smls->skeleton[0])->unWindMatrixNoTran( this).multVecMatrix(Vec3f(0,0,g_incDecRes*(float)nmUpDown->iDelta));
			updateVec3f(hDlg, TRANSLATION_E, this, translation);
			break;
		default:
			return FALSE;
		}
		break;
	case WM_COMMAND:
		switch LOWORD(wParam){
		case ISJOINT_C:
			legalizeOrderForIsJoint();
			CheckDlgButton(g_hDlg,ISJOINT_C, isJoint^=true);
			g_stw->paint();
			break;
		case CUT_NODE:
			cut();
			break;
		case SAVE_PART_AS:
			SaveFileAs(g_selNode,hDlg);
			break;
		case OPEN_PART:
			loadPartFromFile();
			break;
		case NEW_PART_ACCESSORY:
			BackUp::monostable.beginCycle();
			tmpPJ=new Part;
			g_selNode=tmpPJ;
			tmpPJ->name="noName";
			part+=tmpPJ;
			
			smls->convertToCoordLen();
			smls->process();

			genNodeList();
			legalizeOrderForIsJoint(false);
			tmpPJ->isJoint=false;
			setupNodeWindow(&*g_selNode);
			SetFocus(hSceneTreeWnd);
			g_stw->paint();
			createActiveLabel();
			createActiveLabel();
	  
			break;
		case NEW_PART_JOINT:
			BackUp::monostable.beginCycle();
			tmpPJ=new Part;
			g_selNode=tmpPJ;
			tmpPJ->name="noName";
			part+=tmpPJ;
			
			smls->convertToCoordLen();
			smls->process();

			setupNodeWindow(&*g_selNode);
			SetFocus(hSceneTreeWnd);
			genNodeList();
			g_stw->paint();
			createActiveLabel();
			createActiveLabel();
			break;
		case SCALE_RESET:
			scale=Vec3f(1,1,1);
			updateVec3f(hDlg, SCALE_E, this, scale);
			break;
		case ROTATION_RESET:
			rotation=Rotation(0,1,0,0);
			updateRotation(hDlg, ROTATION_E, this, rotation);
			break;
		case TRANSLATION_RESET:
			translation=Vec3f(0,0,0);
			updateVec3f(hDlg, TRANSLATION_E, this, translation);
			break;
		case SCALE_E:
			getDlgItem(hDlg, SCALE_E, tmpVec);
			if(scale!=tmpVec){
				scale=tmpVec;
				updateSceneTree(this);
			}
			break;
		case TRANSLATION_E:
			getDlgItem(hDlg, TRANSLATION_E, tmpVec);
			if(translation!=tmpVec){
				translation=tmpVec;
				updateSceneTree(this);
			}
			break;
		case ROTATION_E:
			getDlgItem(hDlg, ROTATION_E, tmpRot);
			if(rotation!=tmpRot){
				rotation=tmpRot;
				updateSceneTree(this);
			}
			break;
		default:
			return FALSE;
			
		}//END WM_COMMAND switch
		break;
	default:
		return FALSE;
	}//END iMessage switch
	return TRUE; //return TRUE for a caught message
}
*/
void newX3DNodeNode(NodePtr& dest, NodePtr node){
	BackUp::monostable.beginCycle();
	dest=node;
	g_stw->genNodeList();
	g_stw->paint();
	return;
}


void commonInitInsertNodeStuff(Node* node, Field*& field, Node*& newNode, unsigned fieldIdx, unsigned nodeType){
	g_stw->textCell.cellActive=false;
	String& str=g_loadStr;
	str=nodeLookUpStr[nodeType];
	if(nodeType==PROTODECLARE){
		str="PROTO protoDeclare[]";
	}
	str+="{}";
	M_Node nodes;
	createSmlsFromString(nodes, str);
	newNode=nodes[0];
	if(Viewpoint* viewpoint=dynamic_cast<Viewpoint*>(newNode)){
		viewpoint->capture=true;
	}
	else if(Art* part=dynamic_cast<Art*>(newNode)){
		part->mirror.refValue()=false;
	}
	field = node->getField(fieldIdx);
}
void closeWindowIfNodeDoesNotExistsInScene(ControlPanel* controlPanel){
	if(false==nodeExist(controlPanel->node)){
		Window::closeWindow(controlPanel);
	}
}
void setupSelectedNodeWindow(){
	if(dynamic_cast<Anim*>(g_selNode.getRaw())){
		g_acp->setupNodeWindow(g_selNode.getRaw());
		closeWindowIfNodeDoesNotExistsInScene(g_stw->cp);
	}
	else{
		g_stw->cp->setupNodeWindow(g_selNode.getRaw());
		closeWindowIfNodeDoesNotExistsInScene(g_acp);
	}
}

void Seamless::setPointFrom(){
	pointTotal = 0;
	for(int k = 0; k < cpNodeList.length; k++){
		NurbsPatch* patch = cpNodeList[k];
		patch->fromPoint = pointTotal;
		pointTotal += patch->point.length;
	}
}

void resetCPNodeList(){
	MultiSelect::pointList.length = 0;
	
	for(int k = g_sceneTreeList.length; --k >= 0;){
		if(Seamless* smls = dynamic_cast<Seamless*>(g_sceneTreeList[k].node.p)){
			smls->populateMirrorNodes();
		}
	}
	
}

void SceneTree::commonPostInsertNodeStuff(bool dontUpdateAutoInForClick){

	//int remLen=g_sceneTreeCollapsed.node.length;
	g_stw->genNodeList();

	deleteAllRoutesToNonExistant();
	deleteAllCodeLinesThatRefNonExistentNodes();
	resetCPNodeList();


	createFunctionListFromSceneTreeList();

	if(g_stw->selLine>=g_sceneTreeList.length){
		g_stw->selLine=g_sceneTreeList.length-1;
	}
	g_selNode->getSceneDisplayString(textCell.cellStr="");

	textCell.updateCaret();


	commonEndForSetupAnim();

	if(dontUpdateAutoInForClick == false){
		updateAutoInForClick();
	}

	updateBuild3dWindow(true);
	makePointerEqualIndexLocation();
//	SendMessage(hWndMain, WM_PAINT,	0,	0	);
	if(Art* part = dynamic_cast<Art*>(g_selNode.getRaw())){
		part->smls->populateMirrorNodes();
	}
	else if(Seamless* smls = dynamic_cast<Seamless*>(g_selNode.getRaw())){
		smls->populateMirrorNodes();
	}


	setupSelectedNodeWindow();
	setScrolPos(0);
	g_stw->copyLine = 0;
	remPJ=0;
	return;

}
void useNodes(int targetIndex, M_Node& dst, int copyNodeLine, int copySibCount){
	int srcLen=copySibCount;

	int diff=-srcLen;
	if(diff<0){ //make gap bigger
		diff=-diff;
		dst.length+=diff;

		int lastMove=targetIndex;
		for(int i=dst.length-diff; --i>=lastMove; ){
			dst[i+diff]=dst[i];
		}
	}
	else if(diff>0){	//make gap smaller
		for(int i=targetIndex+srcLen+diff; i<dst.length-diff; i++){
			dst[i]=dst[diff+i];
		}
		dst.length-=diff;
	}
	int selLine=copyNodeLine;
	for(int i=0; i<srcLen; i++){
		Node* copyNode = g_sceneTreeList[selLine++].node;

		if(copyNode->name==""){
			createUniqueName(copyNode);
		}
		dst[targetIndex+i]=copyNode;
	}

	g_copySibCount=0;
	g_copyAllCount=0;
}
void moveNodeField(int selNodeLine, int copyNodeLine, int selCount, int copySibCount, int copyAllCount, Node* parentSelNode, int selFieldIdx, int selElement, Node* parentCopyNode, int copyFieldIdx, int copyElement, unsigned type, bool sameField){

	M_<NurbsPatchPtr> m_hold;
	M_Node store;

	for(int k = g_sceneTreeList.length; --k>=0;){
		if(Seamless* smls = dynamic_cast<Seamless*>(g_sceneTreeList[k].node.p)){
			for(int k = 0; k < smls->cpNodeList.length; k++){
				m_hold += smls->cpNodeList[k];
			}
		}
	}
	fastFullBuildForAutoIn(true);

	if(parentSelNode->getField(selFieldIdx)->getType()==MFNODE){
		if(parentCopyNode->getField(copyFieldIdx)->getType()==MFNODE){
			if(parentSelNode->getField(selFieldIdx)==parentCopyNode->getField(copyFieldIdx)){//nodes in same mfnode
				MFNode& mfnode=*dynamic_cast<MFNode*>(parentSelNode->getField(selFieldIdx));
				M_Node& mnode=mfnode.refValue();
				
				NodePtr srcPtr=mnode[copyElement];
				switch(type){
				case DROP_ABOVE:
					if(selElement==copyElement)break;
					copyElements(copyElement, copySibCount, mnode, store);
					insertElements(selElement, 0, mnode, store);
					if(sameField&&copyElement>selElement){
						copyElement+=copySibCount;
					}
					if(copyNodeLine<selNodeLine){
						selNodeLine-=copyAllCount;
					}
					removeNode(copyElement, mnode, copySibCount);
					
					g_copySibCount=0;
					g_copyAllCount=0;
					mfnode=mnode;
					break;
				case DROP_BELOW:
					if(selElement+1==copyElement)break;
					copyElements(copyElement, copySibCount, mnode, store);
					insertElements(selElement+1, 0, mnode, store);
					if(sameField&&copyElement>selElement){
						copyElement+=copySibCount;
					}
					if(copyNodeLine<selNodeLine){
						selNodeLine-=copyAllCount;
					}
					removeNode(copyElement, mnode, copySibCount);
					
					g_copySibCount=0;
					g_copyAllCount=0;
					mfnode=mnode;
					break;
				case DROP_REPLACE:
					if(selElement==copyElement)break;
					copyElements(copyElement, copySibCount, mnode, store);
					insertElements(selElement, selCount, mnode, store);
					if(sameField&&copyElement>selElement){
						copyElement+=copySibCount-selCount;
					}
					if(copyNodeLine<selNodeLine){
						selNodeLine-=copyAllCount;
					}
					removeNode(copyElement, mnode, copySibCount);
					
					g_copySibCount=0;
					g_copyAllCount=0;
					mfnode=mnode;
					break;
				}
			}
			else {	// src and dest different fields
				MFNode& fsrc=*dynamic_cast<MFNode*>(parentCopyNode->getField(copyFieldIdx));
				MFNode& fdst=*dynamic_cast<MFNode*>(parentSelNode->getField(selFieldIdx));

				M_Node& src=fsrc.refValue();
				M_Node& dst=fdst.refValue();

				NodePtr srcPtr=src[copyElement];
				switch(type){
				case DROP_ABOVE:
					copyElements(copyElement, copySibCount, src, store);
					insertElements(selElement, 0, dst, store);
					if(copyNodeLine<selNodeLine){
						selNodeLine-=copyAllCount;
					}
					removeNode(copyElement, src, copySibCount);
					
					g_copySibCount=0;
					g_copyAllCount=0;
					fsrc=src;
					fdst=dst;
					break;
				case DROP_BELOW:

					copyElements(copyElement, copySibCount, src, store);
					insertElements(selElement+1, 0, dst, store);
					if(copyNodeLine<selNodeLine){
						selNodeLine-=copyAllCount;
					}
					removeNode(copyElement, src, copySibCount);
					
					g_copySibCount=0;
					g_copyAllCount=0;
//					mfnode=mnode;


//					insertNodeBelow(selElement, dst, srcPtr);
//					removeNode(copyElement, src);
					fsrc=src;
					fdst=dst;
					break;
				case DROP_REPLACE:
					dst[selElement]=srcPtr;
					removeNode(copyElement, src);
					fsrc=src;
					fdst=dst;
					break;
				}
				
			}
		}
		else{//dst == MF and src == SF
			MFNode& fdst=*dynamic_cast<MFNode*>(parentSelNode->getField(selFieldIdx));
			SFNode& fsrc=*dynamic_cast<SFNode*>(parentCopyNode->getField(copyFieldIdx));
			M_Node& dst=fdst.refValue();

			switch(type){
				case DROP_ABOVE:
					useNodes(selElement, dst, copyNodeLine, 1);//we are not using any nodes in the end but we can take advantage of this funtionality
					break;
				case DROP_BELOW:
					useNodes(selElement+1, dst, copyNodeLine, 1);
					break;
				case DROP_REPLACE:
					dst[selElement]=fsrc.p;
					break;
			}


			fsrc = 0;
			fdst=dst;
		}
	}
	else if(parentSelNode->getField(selFieldIdx)->getType()==SFNODE){
		SFNode& fdst=*dynamic_cast<SFNode*>(parentSelNode->getField(selFieldIdx));
		NodePtr& dst=fdst.refValue();
		if(parentCopyNode->getField(copyFieldIdx)->getType()==MFNODE){
			MFNode& fsrc=*dynamic_cast<MFNode*>(parentCopyNode->getField(copyFieldIdx));
			M_Node& src=fsrc.refValue();
			dst=src[copyElement];
			removeNode(copyElement, src);
			fsrc=src;
			fdst=dst;
		}
		else{
			SFNode& fsrc=*dynamic_cast<SFNode*>(parentCopyNode->getField(copyFieldIdx));
			NodePtr& src=fsrc.refValue();
			dst=src;
			src=0;
			fsrc=src;
			fdst=dst;
		}
	}


	g_stw->genNodeList();
	M_Art m_part;
	for(int k = g_sceneTreeList.length; --k>=0;){
		if(Seamless* smls = dynamic_cast<Seamless*>(g_sceneTreeList[k].node.p)){
			smls->cpNodeList.length = 0;
			smls->cpNodeListAdd(smls->build.refValue());
			smls->deleteLinksToNonExistentPatches(); //not to be confused with non existing control polys
			smls->deletePolysWithNonExistingPoints();
			smls->removeAllPatchesFromPartsNotInList(m_part);
		}
	}

	for(int k = g_sceneTreeList.length; --k>=0;){
		if(Anim* anim = dynamic_cast<Anim*>(g_sceneTreeList[k].node.p)){
			anim->resetMorphers(m_part);
		}
	}

	g_stw->commonPostInsertNodeStuff(true);
	updateAutoInForClick();
}
void useNodeField(Node* parentSelNode, int selFieldIdx, int selElement, Node* parentCopyNode, int copyFieldIdx, int copyElement, int copyNodeLine, int selCount, int copySibCount, unsigned type){
	NodePtr srcPtr;
	if(parentCopyNode->getField(copyFieldIdx)->getType()==MFNODE){
		srcPtr=dynamic_cast<MFNode*>(parentCopyNode->getField(copyFieldIdx))->refValue()[copyElement];
	}
	else {
		srcPtr=dynamic_cast<SFNode*>(parentCopyNode->getField(copyFieldIdx))->refValue();
	}
	if(parentSelNode->getField(selFieldIdx)->getType()==MFNODE){
		MFNode& mfnode=*dynamic_cast<MFNode*>(parentSelNode->getField(selFieldIdx));
		M_Node& mnode=mfnode.refValue();
		switch(type){
		case DROP_ABOVE:
			useNodes(selElement, mnode, copyNodeLine, copySibCount);
			mfnode=mnode;
			break;
		case DROP_BELOW:
			useNodes(selElement+1, mnode, copyNodeLine, copySibCount);
			mfnode=mnode;
			break;
		case DROP_REPLACE:
			useNodes(selElement+selCount, mnode, copyNodeLine, copySibCount);
			removeNode(selElement, mnode, selCount);
			mfnode=mnode;
			break;
		}
	}
	else { //dest SFNode type
		SFNode& fdst=*dynamic_cast<SFNode*>(parentSelNode->getField(selFieldIdx));
		fdst=srcPtr;
	}
	g_stw->commonPostInsertNodeStuff();
}
void deepCopyNodes(int targetIndex, M_Node& dst, int copyNodeLine, int copySibCount){
	int srcLen=copySibCount;

	int diff=-srcLen;
	if(diff<0){ //make gap bigger
		diff=-diff;
		dst.length+=diff;

		int lastMove=targetIndex;
		for(int i=dst.length-diff; --i>=lastMove; ){
			dst[i+diff]=dst[i];
		}
	}
	else if(diff>0){	//make gap smaller
		for(int i=targetIndex+srcLen+diff; i<dst.length-diff; i++){
			dst[i]=dst[diff+i];
		}
		dst.length-=diff;
	}
	for(int selLine=copyNodeLine, i=0; i<srcLen; i++){
		dst[targetIndex+i]=deepCopy(selLine);
	}
	g_copySibCount=0;
	g_copyAllCount=0;
}

void copyNodeField(Node* parentSelNode, int selFieldIdx, int selElement, unsigned type, int copyNodeLine, int copySibCount){
	if(parentSelNode->getField(selFieldIdx)->getType()==MFNODE){
		MFNode& mfnode=*dynamic_cast<MFNode*>(parentSelNode->getField(selFieldIdx));
		M_Node& mnode=mfnode.refValue();
		switch(type){
		case DROP_ABOVE:
			deepCopyNodes(selElement, mnode, copyNodeLine, copySibCount);
			mfnode=mnode;
			break;
		case DROP_BELOW:
			deepCopyNodes(selElement+1, mnode, copyNodeLine, copySibCount);
			mfnode=mnode;
			break;
		case DROP_REPLACE:
			deepCopyNodes(selElement+g_selCount, mnode, copyNodeLine, copySibCount);
			removeNode(selElement, mnode, g_selCount);
			mfnode=mnode;
			break;
		}
	}
	else { //dest SFNode type
		SFNode& fdst=*dynamic_cast<SFNode*>(parentSelNode->getField(selFieldIdx));
		fdst=deepCopy(copyNodeLine);
	}
	g_stw->commonPostInsertNodeStuff();
}

void importNodeField(Node* parentSelNode, int selFieldIdx, int selElement, M_Node& src, unsigned type){

	MFNode& mfnode=*dynamic_cast<MFNode*>(parentSelNode->getField(selFieldIdx));
	M_Node& mnode=mfnode.refValue();
	switch(type){
	case IMPORT_ABOVE:
		insertElements(selElement, 0, mnode, src);
		mfnode=mnode;
		break;
	case IMPORT_REPLACE:
		insertElements(selElement, g_selCount, mnode, src);
		mfnode=mnode;
		break;
	case IMPORT_BELOW:
		insertElements(selElement+1, 0, mnode, src);
		mfnode=mnode;
		break;
	}
	g_stw->genNodeList();
//	for(int k=src.length; --k>=0;){
//		changeSrcPointersToExistingNames(src[k]);
//	}
}
void convertGroupToSmlsLoop(Art* part, const AGroup* vrmlgroup, int& partCnt, M_Node& unknownList, Vec3f parentCenter, Matrix matrix){
	Art::PartChild& m_part=part->part;
//	part->name=String("part")+=partCnt++;
	partCnt++;
	if(const Transform* transform=dynamic_cast<const Transform*>(vrmlgroup)){
		Rotation scaleOrientation = transform->scaleOrientation;
		Vec3f center = transform->center;
		matrix = translationMatrix(-center) * rotationMatrix(-scaleOrientation) * scaleMatrix(transform->scale) * rotationMatrix(scaleOrientation) *
			rotationMatrix(transform->rotation.refValue()) * translationMatrix(center) * translationMatrix(transform->translation.refValue()) * matrix;

		part->pivotPoint.refValue() = center - parentCenter;
		parentCenter = center;
	}
	if(vrmlgroup->name!=""){
		part->name=vrmlgroup->name;
	}
	for(int shapeNodeCnt=0, i=0; i<vrmlgroup->children.length; i++){
		Node* node = vrmlgroup->children[i];
		if(dynamic_cast<Transform*>(node) || dynamic_cast<Group*>(node)){
			Art* newPart = new ImmediatePart;
			newPart->mirror.refValue() = false;
			replaceNodeRefInScripts(node, newPart);
			m_part+=newPart;
			convertGroupToSmlsLoop( newPart, dynamic_cast<AGroup*>(node), partCnt, unknownList, parentCenter, matrix);
		}
		else if(const Shape* shape=dynamic_cast<const Shape*>(vrmlgroup->children[i].p)){
			if(shapeNodeCnt++==0){
				if(shape->name!=""){
					part->name=shape->name;
				}
				addGeometry( part, shape);
				
				M_Vec3f& coord = part->coord.refValue();
				for(int i = coord.length; --i >= 0;){
					coord[i] = matrix * coord[i];
				}
			}
			else {
				Art* newPart = new ImmediatePart;
				newPart->mirror.refValue() = false;
				replaceNodeRefInScripts(node, newPart);
				if(shape->name!=""){
					newPart->name=shape->name;
				}
				else{
					//newPart->name=String("part")+=partCnt++;
					partCnt++;
				}
				m_part+=newPart;
				addGeometry( newPart, shape);
			}
		}
		else{
			unknownList+=node;
		}
	}
}
void checkForImageTexture(Effect*& effect){
	int indent = g_sceneTreeList[g_stw->selLine].indent;
	for(int k = g_stw->selLine + 1; k < g_sceneTreeList.length && indent < g_sceneTreeList[k].indent; k++){
		ImageTexture* imageTexture=dynamic_cast<ImageTexture*>(g_sceneTreeList[k].node.p);
		if(imageTexture && imageTexture->url.length){
			effect = new TextureEffect;
			copySingleUrlWithMultiUrl(dynamic_cast<TextureEffect*>(effect)->url.refValue(), imageTexture->url);
			return;
		}
	}
	effect = new ColorEffect;
}

void createPartTree(Art* parentPart, HAnimJoint* parentJoint, M_<HAnimJoint*>& joints, M_SArt& segmentPart, M_Node& segmentNode){
	if(parentJoint->name==""){
		parentPart->name = parentJoint->Node::name;
	}
	else{
		parentJoint->Node::name = parentPart->name = parentJoint->name;
		String& str = parentPart->name;
		String& str2 = parentJoint->Node::name;
		for(int i=0; i<str.length; i++){
			if(	str[i]==' '){
				str[i]='_';
				str2[i]='_';
			}
		}
	}
	joints += parentJoint;
	M_Node& children = parentJoint->children.refValue();
	for(int k = 0; k < children.length; k++){
		Node* child = children[k].p;
		if(HAnimJoint* hAnimJoint = dynamic_cast<HAnimJoint*>(child)){
			ImmediatePart* part = new ImmediatePart;
			part->pivotPoint.refValue() = hAnimJoint->center;
			parentPart->part.refValue() += part;
			replaceNodeRefInScripts(hAnimJoint, part);
			createPartTree(part, hAnimJoint, joints, segmentPart, segmentNode);
		}
		else{
			segmentPart+=parentPart;
			segmentNode+=child;
		}
	}
}
void convertHAnimIndices(M_Art& partList, Art* rootPart, M_<HAnimJoint*>& joints, const int coordLen, M_Int& maskIndex){
	for(int i = coordLen; --i>=0;){
		for(int k = joints.length; --k>=0;){
			HAnimJoint* hAnimJoint = joints[k];
			M_Int& skinCoordIndex = hAnimJoint->skinCoordIndex.refValue();
			M_Float& skinCoordWeight = hAnimJoint->skinCoordWeight.refValue();
			for(int s = skinCoordIndex.length; --s>=0;){
				if(i == maskIndex[skinCoordIndex[s]]){
					Art* part = partList[k];
					int coordGroupCntr = 0;
					transferOneVertex(rootPart, part, i);
					//if(skinCoordWeight[s] < 1.0){
					//part->weight.refValue() += skinCoordWeight[s];
					part->weight.refValue().insertElement(0, skinCoordWeight[s]);
					part->coordGroup.refValue() += 1;
					//	part->swapVertex(part->coord.length-1, coordGroupCntr);
					//	coordGroupCntr++;
					//}
					k = 0;
					break;
				}
			}
		}
	}
}
inline void bubbleSortWeightRise(Art* part, M_Float& weight){
	bool flag;
	do{
		flag=false;
		for(int i=0;i<weight.length-1;i++){
			if (weight[i]>weight[i+1]){

				float swap = weight[i+1];
				weight[i+1]=weight[i];
				weight[i]=swap;

				part->swapVertex(i, i+1);

				flag=true;
			}
		}
	}while(flag);
}
inline float round(float par){//standard ecma script funtion
	return floor(par +.5);
}
inline void roundWeights(M_Float& weight){
	for(int i = weight.length; --i>=0;){
		weight[i] = round(weight[i]*10)/10;
	}
}
void roundOffCoordGroups(M_Art& partList){
	for(int k = partList.length; --k>=0;){
		Art* part = partList[k];
		M_Float& weight = part->weight.refValue();
		M_Int& coordGroup = part->coordGroup.refValue();

		roundWeights(weight);
		bubbleSortWeightRise(part, weight);

		for(int i = 1; i < coordGroup.length; i++){
			float w = weight[i];
			if(w == weight[i-1]){
				coordGroup.deleteElement(i);
				weight.deleteElement(i);
				i--;
				coordGroup[i]++;
			}
			//int coordCntr2 = coordCntr + 1;
			//for(int j = i + 1; j < coordGroup.length; j++, coordCntr2++){
			//	if(w == weight[j]){
			//		coordCntr++;
			//		part->swapVertex(coordCntr, coordCntr2);
			//		coordGroup[i]++;
			//		coordGroup.deleteElement(j);
			//		weight.deleteElement(j);
			//		j--;
			//	}
			//}
			//coordCntr++;
		}
		if(weight.length && weight[weight.length-1] == 1){
			coordGroup.length--;
			weight.length--;
		}
	}
}
void convertSegmentToPart(Node* node, Seamless* smls, M_SArt& m_part, M_Node& unknownList){
	if(Shape* shape=dynamic_cast<Shape*>(node)){
		Art* part=new ImmediatePart;
		m_part+=part;
		smls->process();
		addGeometry(part, shape);
	}
	else if(Transform* transform = dynamic_cast<Transform*>(node)){
		Art* part = new ImmediatePart;
		replaceNodeRefInScripts(transform, part);
		part->scale.refValue() = transform->scale;
		part->rotation.refValue() = transform->rotation;
		part->translation.refValue() = transform->translation;
		part->pivotPoint.refValue() = transform->center;
		m_part += part;
		smls->process();
		M_Node& children = transform->children.refValue();

		Shape* shape = 0;
		if(children.length == 1 && (shape = dynamic_cast<Shape*>(children[0].p))){
			addGeometry(part, shape);
		}
		else{
			for(int k = 0; k < children.length; k++){
				convertSegmentToPart(children[k], smls, part->part.refValue(), unknownList);
			}
		}
	}
	else if(AGroup* agroup = dynamic_cast<AGroup*>(node)){
		M_Node& children = agroup->children.refValue();
		for(int k = 0; k < children.length; k++){
			convertSegmentToPart(children[k], smls, m_part, unknownList);
		}
	}
//	redirectTransformRoutesToParts(smls->partList);
}
void addSegments(M_SArt& segmentPart, M_Node& segmentNode, M_Node& unknownList){
	for(int k = 0; k < segmentPart.length; k++){
		Art* part = segmentPart[k];
		convertSegmentToPart(segmentNode[k], part->smls, part->part.refValue(), unknownList);
	}
}
void addMinusOne(M_Int& coordIndex){
	if(coordIndex.length > 0){
		if(coordIndex[coordIndex.length-1]!=-1){
			coordIndex+=-1;
		}
	}
}
Shape* accumShapeNodes(M_Node& skin){
	if(skin.length == 0){
		return 0;
	}
	Shape* shape = dynamic_cast<Shape*>(skin[0].p);
	if(shape == 0){
		return 0;
	}
	IndexedFaceSet* ifs = dynamic_cast<IndexedFaceSet*>(shape->geometry.p);
	if(ifs == 0){
		return 0;
	}

	M_Int& coordIndex = ifs->coordIndex.refValue();
	M_Int& texCoordIndex = ifs->texCoordIndex.refValue();
	M_Int& colorIndex = ifs->colorIndex.refValue();

	for(int k = 1; k < skin.length; k++){
		if(Shape* shape = dynamic_cast<Shape*>(skin[k].p)){
			if(IndexedFaceSet* ifs = dynamic_cast<IndexedFaceSet*>(shape->geometry.p)){
				addMinusOne(coordIndex);
				addMinusOne(texCoordIndex);
				addMinusOne(colorIndex);

				colorIndex.length = texCoordIndex.length = coordIndex.length;

				coordIndex += ifs->coordIndex;
				texCoordIndex += ifs->texCoordIndex;
				colorIndex += ifs->colorIndex;
			}
		}
	}

	return shape;
}
void convertCenterValuesToPivotPoints(M_SArt& m_part, const Vec3f parentPivotPoint){
	for(int k = m_part.length; --k>=0;){
		Art* part = m_part[k].p;
		convertCenterValuesToPivotPoints(part->part.refValue(), part->pivotPoint);
		part->pivotPoint = part->pivotPoint - parentPivotPoint;
	}
}

void NurbsPatch::splitPatchU(int v, int uDim, Vec2f& dim, Vec2f& order){

	int selLine = g_stw->selLine;
	int parentLine = getParent(selLine);
	Node* parentNode = g_sceneTreeList[parentLine].node.p;
	int itemIdx = 0;
	int fieldIdx = 0;
	getFNodeElementForParent(parentLine, selLine, g_sceneTreeList, fieldIdx, itemIdx);
	MFNode* parentField = 0;

	if(Seamless* smls = dynamic_cast<Seamless*>(parentNode)){
		parentField = &smls->build;
	}
	else if(GroupBuild* group = dynamic_cast<GroupBuild*>(parentNode)){
		parentField = &group->build;
	}
	deepCopyNodes(itemIdx + 1, parentField->refValue(), selLine, 1);
	NurbsPatch* newPatch = dynamic_cast<NurbsPatch*>(parentField->refValue()[itemIdx + 1].p);

	int idxUp = v * uDim;
	dim.y = dim.y - v;
	point.refValue().deleteElement(0, idxUp);
	weight.refValue().deleteElement(0, idxUp);
	updateKnot(dim.y, vKnot.refValue(), order.y);
	processFields();

	v++;
	idxUp = v * uDim;
	newPatch->dim.refValue().y = v;
	int numOf = newPatch->point.refValue().length - idxUp;
	newPatch->point.refValue().deleteElement(idxUp, numOf);
	newPatch->weight.refValue().deleteElement(idxUp, numOf);
	newPatch->updateKnot(newPatch->dim.refValue().y, newPatch->vKnot.refValue(), newPatch->order.refValue().y);
	newPatch->processFields();

	//*parentField = parentField->refValue();
	g_stw->commonPostInsertNodeStuff();
}

void NurbsPatch::splitPatchV(int col, int uDim, int vDim, Vec2f& dim, Vec2f& order){

	int selLine = g_stw->selLine;
	int parentLine = getParent(selLine);
	Node* parentNode = g_sceneTreeList[parentLine].node.p;
	int itemIdx = 0;
	int fieldIdx = 0;
	getFNodeElementForParent(parentLine, selLine, g_sceneTreeList, fieldIdx, itemIdx);
	MFNode* parentField = 0;

	if(Seamless* smls = dynamic_cast<Seamless*>(parentNode)){
		parentField = &smls->build;
	}
	else if(GroupBuild* group = dynamic_cast<GroupBuild*>(parentNode)){
		parentField = &group->build;
	}
	deepCopyNodes(itemIdx + 1, parentField->refValue(), selLine, 1);
	NurbsPatch* newPatch = dynamic_cast<NurbsPatch*>(parentField->refValue()[itemIdx + 1].p);

	int newUDim = uDim - col;
	dim.x = newUDim;
	for(int v = 0; v < vDim; v++){
		for(int u = col; u < uDim; u++){
			point.refValue()[v * newUDim + u - col] = point.refValue()[v * uDim + u];
			weight.refValue()[v * newUDim + u - col] = weight.refValue()[v * uDim + u];
		}
	}
	point.refValue().length = newUDim * vDim;
	weight.refValue().length = newUDim * vDim;
	updateKnot(dim.x, uKnot, order.x);
	processFields();


	col++;
	newPatch->dim.refValue().x = col;
	int idxUp = col * vDim;
	int numOf = newPatch->point.refValue().length - idxUp;
	for(int v = 0; v < vDim; v++){
		for(int u = 0; u < col; u++){
			newPatch->point.refValue()[v * col + u] = newPatch->point.refValue()[v * uDim + u];
			newPatch->weight.refValue()[v * col + u] = newPatch->weight.refValue()[v * uDim + u];
		}
	}
	newPatch->point.refValue().length = col * vDim;
	newPatch->weight.refValue().length = col * vDim;
	newPatch->updateKnot(newPatch->dim.refValue().x, newPatch->uKnot.refValue(), newPatch->order.refValue().x);
	newPatch->processFields();


	g_stw->commonPostInsertNodeStuff();
}

void NurbsPatch::splitPatch(){
	if(showAll == false){
		return;
	}
	M_Vec3f& dpoint = this->dpoint.refValue();
	M_Vec3f& point = this->point.refValue();
	M_Float& weight = this->weight.refValue();
	M_Float& uKnot = this->uKnot.refValue();
	Vec2f& dim = this->dim.refValue();
	Vec2f& order = this->order.refValue();
	int uDim = dim.x;
	int vDim = dim.y;
	index[0] = 0;
	Matrix matrix = this->matrix * g_viewMatrix;

	for(int v = 0; v < vDim; v++){
		for(int u = 0; u < uDim - 1; u++){
			int idxA = v * uDim + u;
			int idxB = v * uDim + u + 1;
			Vec2f par = Vec2f(uDim + 1, vDim);
			if(clickedRowColum(idxA, idxB, matrix)){
				splitPatchU(v, uDim, dim, order);
				return;
			}
		}
	}
	for(int u = 0; u < uDim; u++){
		for(int v = 0; v < vDim - 1; v++){
			int idxA = v * uDim + u;
			int idxB = (v + 1) * uDim + u;
			Vec2f par = Vec2f(uDim, vDim + 1);
			if(clickedRowColum(idxA, idxB, matrix)){
				splitPatchV(u, uDim, vDim, dim, order);
				return;
			}
		}
	}
}

void convertHAnimToSmls(){
	BackUp::monostable.beginCycle();

	SeamlessPtr smls = new Seamless;

	Effect* effect;
	checkForImageTexture(effect);
	smls->effect=effect;

	ArtPtr part = new ImmediatePart;//we must hold down the part (smart ptr) or we loose it later when we loose its children when we want to get rid of this part 
	smls->skeleton += part.p;
	HAnimHumanoid* hAnimHumanoid = dynamic_cast<HAnimHumanoid*>(g_selNode.getRaw());
	
	//if(Shape* shape = accumShapeNodes(hAnimHumanoid->skin.refValue())){
	Shape* shape = accumShapeNodes(hAnimHumanoid->skin.refValue());

	hAnimHumanoid->processFields();
	const Appearance* appearance = shape ? dynamic_cast<Appearance*>(shape->appearance.p) : 0;
	IndexedFaceSet* ifs = shape ? dynamic_cast<IndexedFaceSet*>(shape->geometry.p) : 0;

	GeomSourcePtr geomSource=new GeomSource;

	if(ifs){
		ifs->convertToGeomSource(*geomSource.p, appearance, hAnimHumanoid->maskIndex);
	}
	if(shape){
		addGeometry(part, shape);
	}

	M_<HAnimJoint*> joints;

	HAnimJoint* hAnimJoint = dynamic_cast<HAnimJoint*>(hAnimHumanoid->skeleton[0].p);
	part->pivotPoint.refValue() = hAnimJoint->center;

	M_SArt segmentPart;
	M_Node segmentNode;
	createPartTree(part, hAnimJoint, joints, segmentPart, segmentNode);
	smls->process();

	redirectTransformRoutesToParts(smls->partList);
	if(shape){
		convertHAnimIndices(part->smls->partList, part, joints, dynamic_cast<Coordinate*>(hAnimHumanoid->skinCoord.p)->point.length, hAnimHumanoid->maskIndex);
		roundOffCoordGroups(part->smls->partList);
	}
	smls->process();
	M_Node unknownList;
	addSegments(segmentPart, segmentNode, unknownList);
	convertCenterValuesToPivotPoints(smls->skeleton.refValue(), Vec3f());

	Group tempGroup;
	tempGroup.children.refValue()+=smls.p;
	tempGroup.children.refValue()+=unknownList;


	int selFieldIdx=0;
	int selElement=0;
	int copyFieldIdx=0;
	int copyElement=0;
	
	int parentSelLine=getParent(g_stw->selLine);
	
	Node* parentSelNode=g_sceneTreeList[parentSelLine].node;
	
	getFNodeElementForParent(parentSelLine, g_stw->selLine, g_sceneTreeList, selFieldIdx, selElement );

	if(MFNode* mfnode = dynamic_cast<MFNode*>(parentSelNode->getField(selFieldIdx))){
		insertElements(selElement, 1, mfnode->refValue(), tempGroup.children.refValue());
	}
	else if(SFNode* sfnode = dynamic_cast<SFNode*>(parentSelNode->getField(selFieldIdx))){
		*sfnode=smls;
	}
	g_stw->commonPostInsertNodeStuff();
	//}
}

void convertGroupToSmls(){
	Seamless* smls=new Seamless;

	Effect* effect;
	checkForImageTexture(effect);
	smls->effect=effect;

	ArtPtr part=new ImmediatePart;//we must hold down the part (smart ptr) or we loose it later when we loose its children when we want to get rid of this part 
	part->mirror.refValue() = false;
	smls->skeleton+=part.p;
	AGroup* vrmlgroup=dynamic_cast<AGroup*>(g_selNode.getRaw());
	Transform* transform=dynamic_cast<Transform*>(vrmlgroup);
	replaceNodeRefInScripts(vrmlgroup, part);
	//if(transform){
	//	smls->scale.refValue()=transform->scale;
	//	smls->rotation.refValue()=transform->rotation;
	//	smls->translation.refValue()=transform->translation;
	//}
	Shape* shape=dynamic_cast<Shape*>(g_selNode.getRaw());
	if(shape){
		vrmlgroup=new Group;
		vrmlgroup->children.refValue()+=shape;
	}

/*
	M_Node& mnode=vrmlgroup->children;
	for(int i=0; mnode.length; i++){
		if(mnode[i]->getType()==TRANSFORM||mnode[i]->getType()==GROUP){
			vrmlgroup=dynamic_cast<AGroup*>(mnode[i].p);
			break;
		}
		else if(mnode[i]->getType()==SHAPE){
			vrmlgroup = new Transform;
			vrmlgroup->children.refValue()+=mnode[i];
			break;
		}
	}
	if(vrmlgroup==NULL)return;
*/
	int partCnt=0;
	M_Node unknownList;
	convertGroupToSmlsLoop(part, vrmlgroup, partCnt, unknownList, Vec3f(), Matrix());
	if(part->coord.length==0){
		smls->skeleton=part->part;
	}
	smls->process();
	redirectTransformRoutesToParts(smls->partList);


	Group tempGroup;
	tempGroup.children.refValue()+=smls;
	tempGroup.children.refValue()+=unknownList;


	int selFieldIdx=0;
	int selElement=0;
	int copyFieldIdx=0;
	int copyElement=0;
	
	int parentSelLine=getParent(g_stw->selLine);
	
	Node* parentSelNode=g_sceneTreeList[parentSelLine].node;
	
	getFNodeElementForParent(parentSelLine, g_stw->selLine, g_sceneTreeList, selFieldIdx, selElement );

	if(MFNode* mfnode = dynamic_cast<MFNode*>(parentSelNode->getField(selFieldIdx))){
		insertElements(selElement, 1, mfnode->refValue(), tempGroup.children.refValue());
	}
	else if(SFNode* sfnode = dynamic_cast<SFNode*>(parentSelNode->getField(selFieldIdx))){
		*sfnode=smls;
	}
	g_stw->commonPostInsertNodeStuff();
}
bool checkForNodeName(Node* par){
	for(int k=g_sceneTreeList.length; --k>=0; ){
		Node* node = g_sceneTreeList[k].node;
		if(node!=par && node->name == par->name){
			return true;
		}
	}
	return false;
}
bool checkForNodeNameInDefList(Node* par){
	for(int k=g_defLookUp.length; --k>=0; ){
		Node* node = g_defLookUp[k];
		if(node!=par && node->name == par->name){
			return true;
		}
	}
	return false;
}
int getLineByNodeName(String str){
	for(int k=g_sceneTreeList.length; --k>=0; ){
		if(g_sceneTreeList[k].node->name==str){
			return k;
		}
	}
	return -1;
}
void createUniqueName(Node* node){
	String& name=node->name;
	int num=1;
	do{
		name=node->getTypeStr();
		name=name.replace(0, 1, name.substr(0,1).toLowerCase());
		name+=num++;
	}while(checkForNodeName(node));
}
void makeExistingNameUnique(Node* node){
	int mark=node->name.length;
	for(int num = 0; checkForNodeName(node); num++){
		node->name=node->name.replace(mark, mark, num);
	}
}
void addNodeToDefList(Node* node, M_Node& defLookUp){
	for(int k=defLookUp.length; --k>=0;){
		if(defLookUp[k]==node){
			return;
		}
	}
	defLookUp+=node;
}
void makeExistingNameUniqueUsingDefList(Node* node){
	if(node->name!=""){
		int mark=node->name.length;
		for(int num = 0; checkForNodeNameInDefList(node); num++){
			node->name=node->name.replace(mark, mark, num);
		}
	}
	addNodeToDefList(node, g_defLookUp);
}
void removeRouteFromProto(Node* windowNode){
	ProtoInterface* protoInterface=getSceneProtoDeclare(g_stw->selLine)->protoInterface;
	Field* field=windowNode->getField(g_fromFieldIdx);
	for(int k=protoInterface->getFLen(); --k>=0;){
		M_Route& mroute=protoInterface->getField(k)->refRoute();
		for(int i=0; i<mroute.length; ){
			if(mroute[i].toNode==windowNode && mroute[i].toFieldIdx==g_fromFieldIdx){
				mroute.deleteElement(i);
				continue;
			}
			i++;
		}
	}
	windowNode->undoIS(g_fromFieldIdx);
}
bool checkRouteToPart(M_Route& route, Art*& part){
	for(int i=route.length; --i>=0;){
		part=dynamic_cast<Art*>(route[i].toNode);
		if(part!=nullPtr){
			return true;
		}
	}
	return false;
}
void RotAccelerator::calcPose(Rotation rotation){
	M_Vec3f& axis=this->axis.refValue();
	M_Float& accelerate=this->accelerate.refValue();

	int i=index[0];
	Rotation accum=accumRotation(i, delay[i]+attack[i]+sustain[i]+decay[i]);

	accum=-accum*rotation;

	axis[i]=accum.getAxis();

	float desiredDist=accum.angle;

	float _accelerate=accelerate[i];
	float _attack=attack[i];
	float _sustain=sustain[i];
	float _decay=decay[i];

	accelerate[i]=desiredDist*(_attack/(_decay+_attack+_sustain*2))*2/(_attack*_attack);

	Window::activeWindow->paint();

}
void OrientationInterpolator::deleteSample(float slider){
	M_Float& key=this->key.refValue();
	M_Rotation& keyValue=this->keyValue.refValue();

	if(slider == 0){
		keyValue[0] = Rotation();
		return;
	}
	if(slider == 1){
		keyValue[keyValue.length - 1] = Rotation();
		return;
	}


	float tm=slider;

	for(int i=0; i<key.length; i++){
		if(fabs(tm-key[i])<.00001){
			keyValue.deleteElement(i);
			key.deleteElement(i);
			i--;
			if(i<0){
				i=0;
			}
			index[0]=i;
			break;
		}
	}
}

void ScalarInterpolator::deleteSample(float slider){
	M_Float& key=this->key.refValue();
	M_Float& keyValue=this->keyValue.refValue();

	for(int i=0; i<key.length; i++){
		if(fabs(slider-key[i])<.00001){
			keyValue.deleteElement(i);
			key.deleteElement(i);
			i--;
			if(i<0){
				i=0;
			}
			index[0]=i;
			break;
		}
	}
}

void PositionInterpolator::deleteSample(float slider){
	M_Float& key=this->key.refValue();
	M_Vec3f& keyValue=this->keyValue.refValue();

	if(slider == 0){
		keyValue[0] = Vec3f();
		return;
	}
	if(slider == 1){
		keyValue[keyValue.length - 1] = Vec3f();
		return;
	}

	for(int i=0; i<key.length; i++){
		if(fabs(slider-key[i])<.00001){
			keyValue.deleteElement(i);
			key.deleteElement(i);
			i--;
			if(i<0){
				i=0;
			}
			index[0]=i;
			break;
		}
	}
}

void CPMorpher::deleteSample(float slider){
	M_Float& key = this->key;
	M_Vec3f& keyValue = this->keyValue;
	const int numOfChannels = cpIndex.length;
	if(slider == 0){
		for(int i = numOfChannels; --i>=0;){
			keyValue[i] = Vec3f();
		}
		return;
	}
	if(slider == 1){
		for(int i = numOfChannels; --i>=0;){
			keyValue[keyValue.length - i - 1] = Vec3f();
		}
		return;
	}


	for(int i=0; i<key.length; i++){
		if(fabs(slider-key[i]) < .00001){
			keyValue.deleteElement(i * numOfChannels, numOfChannels);
			key.deleteElement(i);
			i--;
			if(i < 0){
				i = 0;
			}
			index[0] = i;
			break;
		}
	}
}
void take2(float slider, Rotation rot, M_Int& index, M_Float& key, M_Rotation& keyValue){
	if(key.length == 0){
		key += slider;
		keyValue += rot;
	}
	else{
		for(int i = 0; i < key.length; i++){
			if(roughEQ(slider, key[i])){
				keyValue[i] = rot;
				index[0] = i;
				break;
			}
			else if(roughGT(key[i], slider)){
				key.insertElement(i, slider);
				keyValue.insertElement(i, rot);
				index[0] = i;
				break;
			}
			else if(i == key.length - 1){
				i++;
				key.insertElement(i, slider);
				keyValue.insertElement(i, rot);
				index[0] = i;
				break;
			}
		}
	}
}

void take2(float slider, Vec3f pos, M_Int& index, M_Float& key, M_Vec3f& keyValue){

	if(key.length == 0){
		key += slider;
		keyValue += pos;
	}
	else{
		for(int i=0; i < key.length; i++){
			if(roughEQ(slider, key[i])){
				keyValue[i] = pos;
				index[0] = i;
				break;
			}
			else if(roughGT(key[i], slider)){
				key.insertElement(i, slider);
				keyValue.insertElement(i, pos);
				index[0] = i;
				break;
			}
			else if(i == key.length - 1){
				i++;
				key.insertElement(i, slider);
				keyValue.insertElement(i, pos);
				index[0] = i;
				break;
			}
		}
	}
}

void take2(Vec2f slider, Vec3f pos, M_Int& index, M_Vec2f& moment, M_Vec3f& relPos){

	if(moment.length == 0){
		moment += slider;
		relPos += pos;
	}
	else{
		for(int i = 0; i < moment.length; i++){
			if(roughEQ(slider.x, moment[i].x)){
				relPos[i] = pos;
				index[0] = i;
				break;
			}
			else if(roughGT(moment[i].x, slider.x)){
				moment.insertElement(i, slider);
				relPos.insertElement(i, pos);
				index[0] = i;
				break;
			}
			else if(i == moment.length - 1){
				i++;
				moment.insertElement(i, slider);
				relPos.insertElement(i, pos);
				index[0] = i;
				break;
			}
		}
	}
}

void take2(float slider, float zoom, M_Int& index, M_Float& key, M_Float& keyValue){

	if(key.length == 0){
		key += slider;
		keyValue += zoom;
	}
	else{
		for(int i=0; i < key.length; i++){
			if(roughEQ(slider, key[i])){
				keyValue[i] = zoom;
				index[0] = i;
				break;
			}
			else if(roughGT(key[i], slider)){
				key.insertElement(i, slider);
				keyValue.insertElement(i, zoom);
				index[0] = i;
				break;
			}
			else if(i == key.length - 1){
				i++;
				key.insertElement(i, slider);
				keyValue.insertElement(i, zoom);
				index[0] = i;
				break;
			}
		}
	}
}

void ScalarInterpolator::take(float slider, Node* node, Field* field){
	M_Float& key = this->key.refValue();
	M_Float& keyValue = this->keyValue.refValue();
	if(key.length == 0){
		keyValue.length = key.length = 2;
		key[1] = 1;
	}
	take2(slider, *dynamic_cast<SFFloat*>(field), index, key, keyValue);
}

void PositionInterpolator::take(float slider, Node* node, Field* field){
	M_Float& key = this->key.refValue();
	M_Vec3f& keyValue = this->keyValue.refValue();
	if(Art* part = dynamic_cast<Art*>(node)){
		if(part->ik){
			if(key.length == 0){
				keyValue += part->ikPos;
				keyValue += part->ikPos;
				key.length = 2;
				key[1] = 1;
			}
		}
	}
	if(key.length == 0){
		keyValue.length = key.length = 2;
		key[1] = 1;
	}
	take2(slider, *dynamic_cast<SFVec3f*>(field), index, key, keyValue);
}

void OrientationInterpolator::take(float slider, Node* node, Field* field){
	M_Float& key = this->key.refValue();
	M_Rotation& keyValue = this->keyValue.refValue();
	if(key.length == 0){
		keyValue.length = key.length = 2;
		key[1] = 1;
	}
	take2(slider, *dynamic_cast<SFRotation*>(field), index, key, keyValue);
}

void checkSurfaceGeneratorBeforeMoveDel(Node* node){
	Stem* stem=dynamic_cast<Stem*>(node);
	if(stem!=nullPtr && stem->sync==true){
		stem->sync=false;
	}
}

template <typename F, typename V>inline void _insertElement( V& _src, F& _dst, bool& move, bool& replace, int& after, int& dstIndex, int& srcIndex){
	if(replace){
		_dst.refValue()[dstIndex]=_src[srcIndex];
	}
	else { //insert
		_dst.refValue().insertElement( dstIndex,	_src[srcIndex]);
		if(g_selNode.getRaw()==g_copyNode.getRaw() && srcIndex>dstIndex){
			srcIndex++;
		}
	}
	if(move){
		_src.deleteElement(srcIndex);
	}
};

void insertElement(bool move, bool replace, int after){
	const M_Properties& m_properties=g_selNode->getProperties();
	M_Int& mdstIndex=*(g_selNode->getIndexPtr());
	M_Int& msrcIndex=*(g_copyNode->getIndexPtr());
	int srcNewSelElement=msrcIndex[g_menuGroupIndex];
	int dstNewSelElement=mdstIndex[g_menuGroupIndex];

	for(int fidx=0; fidx<m_properties.length; fidx++){
		const Properties& s_prop=m_properties[fidx];
		if(s_prop.groupIndex==g_menuGroupIndex){
			
			Field* dst=g_selNode->getField(fidx);
			Field* src=g_copyNode->getField(fidx);
			int srcIndex=g_copyIndex[g_menuGroupIndex]-s_prop.startAtIndex;
			int dstIndex=mdstIndex[g_menuGroupIndex]-s_prop.startAtIndex+after;

//			if(srcProp.startAtIndex>dstIndex || srcProp.startAtIndex>srcIndex) continue;

			switch(dst->getType()){
			case MFBOOL:	_insertElement (dynamic_cast<MFBool*>(src)->refValue(),  *dynamic_cast<MFBool*>(dst), move, replace, after, dstIndex, srcIndex); break;
			case MFCOLOR:	_insertElement (dynamic_cast<MFColor*>(src)->refValue(), *dynamic_cast<MFColor*>(dst), move, replace, after, dstIndex, srcIndex); break;
			case MFFLOAT:	_insertElement (dynamic_cast<MFFloat*>(src)->refValue(), *dynamic_cast<MFFloat*>(dst), move, replace, after, dstIndex, srcIndex); break;
			case MFIMAGE:	_insertElement (dynamic_cast<MFImage*>(src)->refValue(), *dynamic_cast<MFImage*>(dst), move, replace, after, dstIndex, srcIndex); break;
			case MFINT32:	_insertElement (dynamic_cast<MFInt32*>(src)->refValue(), *dynamic_cast<MFInt32*>(dst), move, replace, after, dstIndex, srcIndex); break;
			case MFROTATION: _insertElement(dynamic_cast<MFRotation*>(src)->refValue(), *dynamic_cast<MFRotation*>(dst), move, replace, after, dstIndex, srcIndex); break;
			case MFSTRING:	_insertElement (dynamic_cast<MFString*>(src)->refValue(), *dynamic_cast<MFString*>(dst), move, replace, after, dstIndex, srcIndex); break;
			case MFDOUBLE:	_insertElement (dynamic_cast<MFTime*>(src)->refValue(),	 *dynamic_cast<MFTime*>(dst), move, replace, after, dstIndex, srcIndex); break;
			case MFVEC2F:	_insertElement (dynamic_cast<MFVec2f*>(src)->refValue(), *dynamic_cast<MFVec2f*>(dst), move, replace, after, dstIndex, srcIndex); break;
			case MFVEC3F:	_insertElement (dynamic_cast<MFVec3f*>(src)->refValue(), *dynamic_cast<MFVec3f*>(dst), move, replace, after, dstIndex, srcIndex); break;
			}
			if(srcNewSelElement>=src->size()+s_prop.startAtIndex){
				srcNewSelElement=src->size()+s_prop.startAtIndex-1;
				if(srcNewSelElement<0){
					srcNewSelElement=0;
				}
			}
			if(dstNewSelElement>=dst->size()+s_prop.startAtIndex){
				dstNewSelElement=dst->size()+s_prop.startAtIndex-1;
				if(dstNewSelElement<0){
					dstNewSelElement=0;
				}
			}
		}
	}
	msrcIndex[g_menuGroupIndex]=srcNewSelElement;
	mdstIndex[g_menuGroupIndex]=dstNewSelElement;
	updateBuild3dWindow();
}

bool checkIfPartIsUsed(Node* part){
	for(int k = g_sceneTreeList.length; --k >= 0;){
		if(g_sceneTreeList[k].node == part && g_sceneTreeList[k].ref){
			return true;
		}
	}
	return false;
}
////////////////////////

bool checkIfPartIsUsedInBranch(Art* part, bool& flag){
	if(flag){
		return true;
	}
	if(checkIfPartIsUsed(part)){
		flag = true;
		return true;
	}
	M_SArt& m_part = part->part.refValue();
	for(int k = 0; k < m_part.length; k++){
		if(checkIfPartIsUsedInBranch(m_part[k], flag)){
			return true;
		}
	}
	return flag;
}

bool checkPartCanBeDeleted(Node* node, int nodeLine){
	if(Art* part = dynamic_cast<Art*>(node)){
		if(g_sceneTreeList[nodeLine].ref){
			return true;
		}
		bool flag = false;
		if(checkIfPartIsUsedInBranch(part, flag)){
			msG(msg_canNot_delete_used_else_where);
			return false;
		}
		part->deleteGeometryInBranch();
	}
	return true;
}

void deleteNode(int nodeLine, int selCount){//DELETE_NODE
	Node* selNode = g_sceneTreeList[nodeLine].node;
	if(selNode->getType() == SCENE)return;


	
	int selFieldIdx = 0;
	int selElement = 0;
	
	int parentSelLine = getParent(nodeLine);
	
	NodePtr parentSelNode = g_sceneTreeList[parentSelLine].node;
	checkSurfaceGeneratorBeforeMoveDel(parentSelNode);
	
	getFNodeElementForParent(parentSelLine, nodeLine, g_sceneTreeList, selFieldIdx, selElement);
	
	Field* field = parentSelNode->getField(selFieldIdx);
	
	if(field->getType() == MFNODE){
		MFNode& mfnode = *dynamic_cast<MFNode*>(field);
		M_Node& mnode = mfnode.refValue();

		for(int k = 0, l = nodeLine; k < selCount; k++){
			if(false == checkPartCanBeDeleted(mnode[selElement + k], l)){
				return;
			}
			l = getNodeLineForNextSibling(l);
		}

		removeNode(selElement, mnode, selCount);
		mfnode = mnode;
	}
	else{
		for(int k = 0; k < selCount; k++){
			if(false == checkPartCanBeDeleted(selNode, nodeLine)){
				return;
			}
		}
		SFNode& snode = *dynamic_cast<SFNode*>(field);
		snode = nullPtr;
	}
}

void homeToInit(){//HOME_TO_INIT
	RotAccelerator* rotAccelerator=dynamic_cast<RotAccelerator*>(g_selNode.getRaw());
	if(rotAccelerator==0)return;
	rotAccelerator->calcPose(rotAccelerator->init);
}
	/*
	case HOME_TO_POSE:{
	RotAccelerator* rotAccelerator=dynamic_cast<RotAccelerator*>(&*windowNode);
	if(rotAccelerator==0)break;
	Art* part;
	if(checkRouteToPart(rotAccelerator->value_changed.route, part)){
	rotAccelerator->calcPose(part->rotation);
	}
	}
	break;
	*/

void makeSureBandCountIsLegit(StemPtr stem, int index){
	M_Int& bands=stem->bands.refValue();
	if(bands.length == 1 && stem->beginSP.refValue() && stem->endSP.refValue() && bands[0] < 3){
		bands[0] = 3;
		return;
	}
	if(index == 0 && bands[0] < 2 ){
		bands[0] = 2;
		return;
	}
	if(bands[index] < 2){
		bands[index] = 2;
		return;
	}

}
void Seamless::newBuildNode(Build* node){
	BackUp::monostable.beginCycle();
	build.refValue()+=node;
	
	g_stw->genNodeList();
	g_stw->paint();
	return;
}
void Scene::newChildrenNode(Node* node){
	BackUp::monostable.beginCycle();
	children.refValue()+=node;
	
	g_stw->genNodeList();
	g_stw->paint();
	return;
}
void SurfaceGenerator::newGeneratorNode(Node* node){
	BackUp::monostable.beginCycle();
	generator.refValue()+=node;
	
	g_stw->genNodeList();
	g_stw->paint();
	return;
}

/*
Matrix4 Art::windCursorMatrix(ArtPtr pj){ //same as windMatrix with out scale
	Matrix4 matrix;
	Matrix4 partMatrix;
	
	while(1){
		partMatrix=scaleMatrix(pj->scale.refValue());
		partMatrix=partMatrix*rotationMatrix( pj->rotation.refValue());
		partMatrix=partMatrix*translationMatrix( pj->translation.refValue());
		matrix=matrix*partMatrix;
		if(pj==this)break;
		NodePtr sfnode;
		do{
			sfnode=pj->parent;
		}while(sfnode->getType()!=PART);
		pj=sfnode;
	}
	partMatrix=scaleMatrix(g_scene->scale);
	partMatrix=partMatrix*translationMatrix(g_scene->translation);
	partMatrix=partMatrix*rotationMatrix(g_scene->rotation);
	matrix=matrix*partMatrix;
	
	return matrix;
}
/*
Matrix4 Art::windMatrix(ArtPtr pj, bool viewMat){
	Matrix4 tmpMatrix;
	Matrix4 catMatrix;
	Matrix4 viewMatrix;
	
	
	while(1){
		
		tmpMatrix.setTransform(pj->translation.refValue(), pj->rotation.refValue(), pj->scale.refValue());
		
		catMatrix=catMatrix*tmpMatrix;
		
		if(pj==this)break;
		NodePtr sfnode;
		do{
			sfnode=pj->parent;
		}while(sfnode->getType()!=PART);
		pj=sfnode;
	}
	if(viewMat){
		viewMatrix=scaleMatrix(g_scene->scale) * translationMatrix(g_scene->translation) * rotationMatrix(g_scene->rotation);
		catMatrix=catMatrix*viewMatrix;
	}
	
	
	return catMatrix;
}
*/
void getIRotationLoop(Art* target, Art* part, Rotation rot, bool& found){
	if (found==true)return;
	rot=-part->rotation.refValue()*rot;
	if(part==target){
		found=true;
		return;
	}
	M_SArt& prt=part->part.refValue();
	for(int k=0; k<prt.length; k++){
		getIRotationLoop(target, prt[k],rot,found);
	}
}
Rotation Art::getIRotation(){
	M_SArt& part=smls->skeleton.refValue();
	for(int k=0; k<part.length; k++){
		Rotation rot;
		bool found=false;
		getIRotationLoop(this, part[k],rot,found);
		if(found)return rot;
	}
	throw this;
}

/*
Matrix4 Art::unWindMatrix(ArtPtr pj, bool viewMat){
	Matrix4 tmpMatrix;
	Matrix4 catMatrix;
	Matrix4 viewMatrix;
	
	
	while(1){
		Vec3f translation=pj->translation.refValue();
		Rotation rotation=pj->rotation.refValue();
		Vec3f scale=pj->scale.refValue();

		tmpMatrix=translationMatrix(Vec3f(-translation.x, -translation.y, -translation.z)) * rotationMatrix( -rotation) * scaleMatrix(Vec3f(1./scale.x, 1./scale.y, 1./scale.z));
		
		catMatrix=tmpMatrix*catMatrix;
		
		if(pj==this)break;
		pj=pj->parent;
	}
	if(viewMat){
		viewMatrix=rotationMatrix( -g_scene->rotation) * translationMatrix(Vec3f(-g_scene->translation.x, -g_scene->translation.y, -g_scene->translation.z)) * scaleMatrix(Vec3f(1./g_scene->scale.x, 1./g_scene->scale.y, 1./g_scene->scale.z));
		catMatrix=viewMatrix*catMatrix;
	}
	
	
	return catMatrix;
}
*/
/*
Matrix4 Art::unWindMatrixNoTran(ArtPtr pj, bool viewMat){//should be tidied up to bypass translations
	Matrix4 tmpMatrix;
	Matrix4 catMatrix;
	Matrix4 viewMatrix;
	
	if(pj!=this){
		NodePtr sfnode;
		do{
			sfnode=pj->parent;
		}while(sfnode->getType()!=PART);
		pj=sfnode;
		
		while(1){
			Vec3f translation=pj->translation.refValue();
			Vec3f scale=pj->scale.refValue();
			Rotation rotation=pj->rotation.refValue();
			
			tmpMatrix = translationMatrix(Vec3f(0, 0, 0)) * rotationMatrix( -rotation) * scaleMatrix(Vec3f(1./scale.x, 1./scale.y, 1./scale.z));
			
			catMatrix=tmpMatrix*catMatrix;
			
			if(pj==this)break;
			NodePtr sfnode;
			do{
				sfnode=pj->parent;
			}while(sfnode->getType()!=PART);
			pj=sfnode;
		}
	}
	if(viewMat){
		viewMatrix=rotationMatrix( -g_scene->rotation) * translationMatrix(Vec3f(0, 0, 0)) * scaleMatrix(Vec3f(1./g_scene->scale.x, 1./g_scene->scale.y, 1./g_scene->scale.z));
		catMatrix=viewMatrix*catMatrix;
	}
	
	
	return catMatrix;
}

Rotation Art::unWindRot(ArtPtr pj, bool viewMat){
	Rotation rot;
	
	if(pj!=this){
		NodePtr sfnode;
		do{
			sfnode=pj->parent;
		}while(sfnode->getType()!=PART);
		pj=sfnode;
		
		while(1){
			
			rot=rot.multiply(pj->rotation.refValue());
			//		rot=pj->rotation.multiply(rot);
			
			if(pj==this)break;
			
			NodePtr sfnode;
			do{
				sfnode=pj->parent;
			}while(sfnode->getType()!=PART);
			pj=sfnode;
		}
	}
	if(viewMat){
		rot=rot.multiply(g_scene->rotation);
		//		rot=g_scene->rotation.multiply(rot);
	}
	
	
	return -rot;
}
*/
/*
void Art::convertAllPartsToDX(){
	return;
	int k;
	ArtPtr pj;
	
	if(geometry->coordIndex.refValue().length){
		smls->convertToRenderFormat();
	}
	M_SArt& pp=part.refValue();

	if(pp.length){
		for(k=0;k<pp.length;k++){
			pj=pp[k];
			pj->convertAllPartsToDX();
		}
	}
}
*/
extern int maxBufSize;



int getFileSize(FILE* Stream);


inline void convertToTriangleVertex(Art*& triPart, int& i, M_Art& partList){
	for(int k=0; k<partList.length; k++ ){
		Art* part=partList[k];
		i-=part->coord.refValue().length;
		if(i<0){
			i+=part->coord.refValue().length;
			triPart=part;
			return;
		}
	}
}
inline void convertCoordIndexToSeamlessTriangles(M_Art& partLookUp, const M_Int& coordIndex, const M_Int& texCoordIndex){
	for(int i=0; i<coordIndex.length; ){
		S_Triangle tri;
		
		int linIdx;

		tri.ta=texCoordIndex[i];
		linIdx=coordIndex[i++];
		tri.pa=partLookUp[linIdx];
		tri.ia=linIdx-tri.pa->fromCoord;

		tri.tb=texCoordIndex[i];
		linIdx=coordIndex[i++];
		tri.pb=partLookUp[linIdx];
		tri.ib=linIdx-tri.pb->fromCoord;

		tri.tc=texCoordIndex[i];
		linIdx=coordIndex[i++];
		tri.pc=partLookUp[linIdx];
		tri.ic=linIdx-tri.pc->fromCoord;

		i++;

		Art* owner=tri.getOwner();

		owner->triangle+=tri;

	}

}

/*
inline void convertTexCoordIndexToSeamlessTriangles(M_Art& partLookUp, M_Art& partList, M_Int& coordIndex, M_Int& texCoordIndex, M_Int& newTexCoordIndex){

	newTexCoordIndex.resize(texCoordIndex.length/4*3);
	for(int accumPoly=0, k=0; k<partList.length; k++){
		Art* part=partList[k];
		part->fromPoly=accumPoly;
		accumPoly+=part->triangle.length;
		part->triangle.length=0;
	}

	for(int i=0; i<coordIndex.length; ){
		S_Triangle tri;
		
		int linIdx;


		int tciA=texCoordIndex[i];
		linIdx=coordIndex[i++];
		tri.pa=partLookUp[linIdx];
		tri.ia=linIdx-tri.pa->fromCoord;

		int tciB=texCoordIndex[i];
		linIdx=coordIndex[i++];
		tri.pb=partLookUp[linIdx];
		tri.ib=linIdx-tri.pb->fromCoord;

		int tciC=texCoordIndex[i];
		linIdx=coordIndex[i++];
		tri.pc=partLookUp[linIdx];
		tri.ic=linIdx-tri.pc->fromCoord;

		i++;

		Art* owner=tri.getOwner();

		int j=(owner->fromPoly+owner->triangle.length)*3;
		owner->triangle+=tri;

		newTexCoordIndex[j++]=tciA;
		newTexCoordIndex[j++]=tciB;
		newTexCoordIndex[j  ]=tciC;
	}
}

/*
inline void addRc(Art*& cornerPart, int& cornerIndex, M_Vec2f& texCoord, int& linTexIdx){

	M_RedirectCoord& rc=cornerPart->rc;

	for(int i=0; i<rc.length; i++){
		if(rc[i].p==cornerPart && rc[i].i==cornerIndex && texCoord[linTexIdx] == cornerPart->texCoord[cornerPart->coord.length + i]){
		//	rc+=S_RedirectCoord(cornerPart, cornerIndex);
		//	cornerIndex=cornerPart->texCoord.length;
		//	cornerPart->texCoord+=texCoord[linTexIdx];

			cornerIndex=cornerPart->coord.length + i;
			return;
		}
	}

	rc+=S_RedirectCoord(cornerPart, cornerIndex);

	cornerIndex=cornerPart->texCoord.length;

	cornerPart->texCoord+=texCoord[linTexIdx];
}

inline bool rcAlreadyAdded(Art*& cornerPart, int& cornerIndex){

}
inline bool notAlreadyTaken(M_Int& coordIndexMask, int linTexIdx){
	for(int i=coordIndexMask.length; --i>=0; ){
		if(coordIndexMask[i]==linTexIdx){
			return false;
		}
	}
	return true;
}
void verifyAllVerticesAreReferenced(M_Art& partList);
inline void convertTexCoordIndexToSeamlessTexture(M_Art& partLookUp, IndexedFaceSet* ifs, M_Art& partList, M_Int& newTexCoordIndex){
//	dLog(newTexCoordIndex.toString());

	M_Vec2f& ifsTexCoord=ifs->refTexPoint();

	M_Int coordIndexMask;
	coordIndexMask.resize(ifsTexCoord.length);
	{
		for(int i=coordIndexMask.length; --i>=0; ){
			coordIndexMask[i]=-1;
		}
	}
	{
		for(int k=0; k<partList.length; k++){
			Art* part=partList[k];
			part->texCoord.resize(part->coord.length);
		}
	}
	for(int k=0; k<partList.length; k++){

		Art* part=partList[k];
		M_Triangle& triangle=part->triangle;

		for(int i=0; i<triangle.length*3; i++){
			int linTexIdx=newTexCoordIndex[part->fromPoly*3+i];

			Art*& cornerPart=triangle.refPart(i);
			int& cornerIndex=triangle.refIndex(i);
			if(coordIndexMask[linTexIdx]==-1 && notAlreadyTaken(coordIndexMask, cornerPart->fromCoord+cornerIndex)){
				coordIndexMask[linTexIdx]=cornerPart->fromCoord+cornerIndex;
				cornerPart->texCoord[cornerIndex]=ifsTexCoord[linTexIdx];
			}
			else if(coordIndexMask[linTexIdx]==cornerPart->fromCoord+cornerIndex){
				//good its done already
			}
			else if(coordIndexMask[cornerPart->fromCoord+cornerIndex]==-1 && notAlreadyTaken(coordIndexMask, cornerPart->fromCoord+cornerIndex)){
				coordIndexMask[linTexIdx]=cornerPart->fromCoord+cornerIndex;
				cornerPart->texCoord[cornerIndex]=ifsTexCoord[linTexIdx];
			}
			else {
				addRc(cornerPart, cornerIndex, ifsTexCoord, linTexIdx);
			}
		}
	}
	verifyAllVerticesAreReferenced( partList);
}
*/
void legaliseTriangles(M_Art& partList){

	for(int k=0; k<partList.length; k++){
		Art* part=partList[k];

		M_Triangle& triangle=part->triangle;
		int d=0;
		int numLost=0;
		for(int i=0; i<triangle.length; i++){
			S_Triangle& tri=triangle[i];

			Art* owner=tri.getOwner();

			if(part==owner){
				triangle[d++]=tri;
			}
			else {
				owner->triangle+=tri;
				numLost++;
			}
		}
		triangle.length-=numLost;
	}
}
void setUpInternalEffectPtrs(Seamless* smls);
void creatPartLookUp(M_Art& partLookUp, M_Art& partList);
void findMFField(const String str, const String compair, int& i){
	do{
		str.findWord(compair, i);
		str.stwo_stb(i);
	}while(str[i]!='[');
}

void openBxxAvatarStudio( HWND hWnd );
ProtoDeclare* getAvatarProtoDeclare(){
	for(int k = g_sceneTreeList.length; --k>=0;){
		if(ProtoDeclare* protoDeclare = dynamic_cast<ProtoDeclare*>(g_sceneTreeList[k].node.p)){
			return protoDeclare;
		}
	}
	return 0;
}
Shape* getPointList(IndexedFaceSet*& ifs, Coordinate*& pointList){
	for(int k = g_sceneTreeList.length; --k>=0;){
		if(Shape* shape = dynamic_cast<Shape*>(g_sceneTreeList[k].node.p)){
			if(ifs = dynamic_cast<IndexedFaceSet*>(shape->geometry.p)){
				if(pointList = dynamic_cast<Coordinate*>(ifs->coord.p)){
					if(pointList->name == "PointList"){
						return shape;
					}
				}
			}
		}
	}
	return 0;
}
Script* getWelder(){
	for(int k = g_sceneTreeList.length; --k>=0;){
		if(Script* script = dynamic_cast<Script*>(g_sceneTreeList[k].node.p)){
			return script;
		}
	}
	return 0;
}
M_Int* getVgroups(Script* script){
	M_Properties& m_prop = script->properties;
	for(int k = m_prop.length; --k>=0;){
		if(m_prop[k].name == "vgroups"){
			Field* field = script->getField(k);
			return &dynamic_cast<MFInt32*>(field)->refValue();
		}
	}
	return 0;
}
M_Vec3f* getP(Script* script){
	M_Properties& m_prop = script->properties;
	for(int k = m_prop.length; --k>=0;){
		if(m_prop[k].name == "P"){
			Field* field = script->getField(k);
			return &dynamic_cast<MFVec3f*>(field)->refValue();
		}
	}
	return 0;
}
M_Node* getTS(Script* script){
	M_Properties& m_prop = script->properties;
	for(int k = m_prop.length; --k>=0;){
		if(m_prop[k].name == "ts"){
			Field* field = script->getField(k);
			return &dynamic_cast<MFNode*>(field)->refValue();
		}
	}
	return 0;
}
M_String* getJointNames(ProtoInterface* protoInterface){
	M_Properties& m_prop = protoInterface->properties;
	for(int k = m_prop.length; --k>=0;){
		if(m_prop[k].name == "jointNames"){
			Field* field = protoInterface->getField(k);
			return &dynamic_cast<MFString*>(field)->refValue();
		}
	}
	return 0;
}
void decodeNamesFromScript(M_Art& partList, const String url){
	int i = 0;
	while(-1 != (i = url.indexOf("rot_", i))){
		i += 4;
		int end = url.indexOf("(", i) - 1;
		String name = url.substring(i, end);
		i = url.indexOf("[", i) + 1;
		int k = getInt(url, i);
		partList[k]->name = name;
	}
}
void convertTranslationsToPivotPoints(Seamless* smls);
void convertGestures(Seamless* smls, M_Node& m_ts, M_Node& m_anim, Script* welder);

void createSmlsFromBxxAvStudio(){
	Seamless* smls=new Seamless;
	smls->version=g_version;

	IndexedFaceSet* ifs = 0;
	Coordinate* pointList = 0;
	Shape* shape = getPointList(ifs, pointList);
	ProtoDeclare* protoDeclare = getAvatarProtoDeclare();
	M_String* jointNames = getJointNames(protoDeclare->protoInterface);
	Script* welder = getWelder();
	M_Int& partTree = *getVgroups(welder);
	M_Vec3f& translatePart = *getP(welder);
	M_Node& ts = *getTS(welder);

	M_Int& coordIndex=ifs->coordIndex.refValue();
	M_Int& colorIndex=ifs->colorIndex.refValue();
	M_Int& texCoordIndex=ifs->texCoordIndex.refValue();
	if(ifs->color.getValue()!=0){
		if(colorIndex.length!=coordIndex.length){
			colorIndex=coordIndex;
		}
		ifs->addExtraColor();
	}
	else if(ifs->texCoord.getValue()!=0){
		if(texCoordIndex.length!=coordIndex.length){
			texCoordIndex=coordIndex;
		}
		ifs->addExtraTexCoord();
	}

	Appearance* appearance = dynamic_cast<Appearance*>(shape->appearance.getRawPtr());

//////////////////////////

	ArtPtr pj= new ImmediatePart;
	smls->skeleton.refValue()+=pj;

	M_Vec3f& ifsCoord=ifs->refPoint();

	for(int i=0;i<partTree.length/3;i++){
		ArtPtr newPart=new ImmediatePart;
		if(jointNames){
			newPart->name=(*jointNames)[i];
		}
		newPart->_matrixStack=partTree[i*3];
		newPart->translation.refValue() = translatePart[i];
		newPart->coord.resize(partTree[i*3+2]-partTree[i*3+1]);

		M_Vec3f& coord=newPart->coord.refValue();

		for(int s=partTree[i*3+1], d=0; d<coord.length; s++,d++){
			coord[d]=ifsCoord[s];
		}

		if(newPart->_matrixStack==0){
			pj=newPart;
			smls->skeleton.refValue()[0]=pj;
		}
		else if(newPart->_matrixStack-1==pj->_matrixStack){//make newPart a child pj

			newPart->parent=&*pj;

			pj->part.refValue()+=newPart;
			pj=newPart;
		}
		else if(newPart->_matrixStack<=pj->_matrixStack){//make newPart a sibling
			if(newPart->_matrixStack<pj->_matrixStack){
				//throw;
			}
			int countTo=pj->_matrixStack;
			int countFrom=newPart->_matrixStack;
			for(int k=countFrom;k<countTo;k++){
				pj=pj->parent;
			}
			newPart->parent=pj->parent;
			ArtPtr(pj->parent)->part.refValue()+=newPart;
			pj=newPart;//pj=next sibling
		}
	}
	smls->partList.length=0;
	smls->process();
	if(smls->partList.length){
		if(ifs->color.getValue()!=0){
			ColorEffect* effect=new ColorEffect;
			smls->effect=effect;
			effect->color.refValue()=ifs->refColor();
		}
		else if(ifs->texCoord.getValue()!=0){
			TextureEffect* effect=new TextureEffect;
			smls->effect=effect;
			effect->texCoord.refValue()=ifs->refTexPoint();
			if(appearance->texture.refValue()!=0){
				ImageTexture* textureImage=dynamic_cast<ImageTexture*>(appearance->texture.getRawPtr());
				effect->url.refValue()+=textureImage->url.refValue();
			}
		}
		else throw ifs;
		setUpInternalEffectPtrs(smls);
	}
	else return;

	M_Art partLookUp;
	creatPartLookUp(partLookUp, smls->partList);

	if(ifs->color.getValue()!=0){
		convertCoordIndexToSeamlessTriangles(partLookUp, coordIndex, colorIndex);
	}
	else if(ifs->texCoord.getValue()!=0){
		convertCoordIndexToSeamlessTriangles(partLookUp, coordIndex, texCoordIndex);
	}
	else throw ifs;
	smls->buildPartList();
	if(smls->partList.length > 1){
		if(smls->partList[1]->name=="Center"){
			smls->partList[1]->translation.refValue()=Vec3f();
		}
	}
	convertTranslationsToPivotPoints(smls);

	smls->resetRelevantParts();

	smls->verifyIntegrity();

	if(jointNames==0){
		decodeNamesFromScript(smls->partList, welder->url[0]);
	}

	GestureGroup* gestureGroup = new GestureGroup;
	convertGestures(smls, ts, gestureGroup->children.refValue(), welder);
	g_acp->setupNodeWindow(dynamic_cast<Anim*>(gestureGroup->children.refValue()[0].p));


	M_Node& children = g_scene->children;
	children.length=0;
	children+=smls;
	children+=gestureGroup;

	g_lookupProtoInterface.length=g_userProtoBegin;
	nodeLookUpStr.length=g_userProtoBegin;
	containerFieldLookUpStr.length=g_userProtoBegin;

	//orderPolygonsFromVertices();
//	g_stw->genNodeList();
//	g_stw->paint();

//
	g_scene->wOpen.refValue().length=0;
	g_scene->wPos.refValue().length=0;
	g_scene->wSize.refValue().length=0;

	g_scene->genNodeList2();
	//postLoadProcessNodes(g_sceneTreeList);

	resetLooseEnds();

	commonEndForSetupAnim();

	Window::postLoadProcessWin(*g_scene);
	setTitleName();

	BackUp::deleteBackupFiles();
	BackUp::createUniqueBackupName();
	g_stw->paint();

	Window::activeWindow=g_acp.p;
	Window::paintAllControlPanels();
	g_stw->commonPostInsertNodeStuff();
}


void convertPartFieldIndexToScriptFieldIndex(Art* part, int fieldIndex, Script* script){
	String fieldName=part->name;
	fieldName+=fieldIndex;
	const M_Properties& properties=script->getProperties();
	{
		for(int i=0; i<script->getFLen(); i++){
			Field* field=script->getField(i);
			if(properties[i].name==fieldName){
				return;
			}
		}
	}
	Field* field=part->getField(fieldIndex);
	if(field->getType()==SFROTATION){
		*script+=new SFRotation(Rotation());
	}
	else if(field->getType()==SFVEC3F){
		*script+=new SFVec3f(Vec3f());
	}
	else return;
	String funcName=fieldName;
	script->properties+=Properties(funcName);
	String& url=script->url.refValue()[0];
	url+="\nfunction ";
	url+=fieldName;
	url+="(v){";

	int treePartCntr=0;
	M_Art& partList=part->smls->partList;
	for(int k=0; k<partList.length; k++){
		Art* p=partList[k];
		if(part==p)break;
		if(p->getType()==PART||p->getType()==IMMEDIATEPART){
			treePartCntr+=p->weight.refValue().length+1;
		}
	}
	if(field->getType()==SFROTATION){
		M_Float& weight=part->weight.refValue();
		for(int i=0; i<weight.length; i++){
			url+="o[";
			url+=treePartCntr++;
			url+="]=new SFRotation(v.x, v.y, v. z, v.angle*";
			url+=weight[i];
			url+=");";
		}
		url+="o[";
		url+=treePartCntr;
		url+="]=v;}";
	}
	
	else if(field->getType()==SFVEC3F){

		const Properties& sprop=part->getProperties(*field);

		if(sprop.name=="translation"){
			url+="translatePart[";
			url+=treePartCntr;
			url+="]=pivotPoint[";
			url+=treePartCntr;
			url+="].add(v);}";
		}
		else {
			url+="scalePart[";
			part->smls->addScaleField=true;
			url+=treePartCntr;
			url+="]=v;}";
		}
	}
}

void convertPartToHAnimJoint(Art* parentPart, HAnimJoint* parentJoint, M_Node& joints);


bool isSegmented(Seamless* smls){
	M_SArt& m_part = smls->skeleton.refValue();
	for(int k=0; k<m_part.length; k++){
		if(m_part[k]->sharedVertices()){
			return false;
		}
	}
	return true;
}
void Seamless::addNewSegmentedNode(){
	if(isSegmented(this)){
		outSegmented=true;
		Transform* transform=new Transform;
		outputNode=transform;
		M_SArt& skeleton=this->skeleton.refValue();
		for(int k=0; k<skeleton.length; k++){
			skeleton[k]->processSegmentedProduct(transform->children.refValue());
		}
	}
}
void addFieldToScript(Seamless* smls, Node* node, int fieldIdx){
	Seamless* smls2=0;
	Art* part=dynamic_cast<Art*>(node);
	if(part!=0){
		smls2=part->smls;
	}
	else{
		Morpher* morpher=dynamic_cast<Morpher*>(node);
		if(morpher!=0){
			smls2=morpher->smls;
		}
	}
	smls->outSegmented=false;
	if(smls==smls2){
		if(g_scene->autoOutSegmentedParts){
			smls->addNewSegmentedNode();
		}
		if(smls->outSegmented==false){
			if(g_genHAnimNodes){
				smls->addNewHAnimNode();
			}
			else{
				Script* script;
				if((script=dynamic_cast<Script*>(smls->outputNode.p))==0){
					smls->addScaleField=false;
					script=new Script;
					script->directOutput=true;
					smls->outputNode=script;
					script->name='s';
					makeNodeNameFromAddr(script->name, smls);
					script->url.refValue()+="javascript:\n";
				}
				if(part!=0){
					convertPartFieldIndexToScriptFieldIndex(part,  fieldIdx, script);
				}
			}
		}
	}
}
void convertRoutesToScriptForFunction(Seamless* smls);
void convertRoutesToScript(Seamless* smls){//if a route is found make a script node or hanim node so that it animates
	Art* part;
	for(int k=g_sceneTreeList.length; --k>=0;){
		Node* node=g_sceneTreeList[k].node;
		AnimByAnkle* animByAnkle=dynamic_cast<AnimByAnkle*>(node);
		AnimByWrist* animByWrist=dynamic_cast<AnimByWrist*>(node);
		if(animByAnkle){
			if(	(part=animByAnkle->hip)!=0){
				addFieldToScript(smls, part, part->getFieldIndex(&part->rotation));
			}
			if(	(part=animByAnkle->knee)!=0){
				addFieldToScript(smls, part, part->getFieldIndex(&part->rotation));
			}
			if(	(part=animByAnkle->ankle)!=0){
				addFieldToScript(smls, part, part->getFieldIndex(&part->rotation));
			}
		}
		else if(animByWrist){
			if(	(part=animByWrist->sholder)!=0){
				addFieldToScript(smls, part, part->getFieldIndex(&part->rotation));
			}
			if(	(part=animByWrist->elbow)!=0){
				addFieldToScript(smls, part, part->getFieldIndex(&part->rotation));
			}
			if(	(part=animByWrist->wrist)!=0){
				addFieldToScript(smls, part, part->getFieldIndex(&part->rotation));
			}
		}
		else{
			for(int rx=0,i=0; i<node->getFLen(); i++){
				Field* field=node->getField(i);
				const M_Route& route=field->getRoute();
				for(int r=0; r<route.length; r++){
					const S_Route& sroute=route[r];
					addFieldToScript(smls, sroute.toNode, sroute.toFieldIdx);
				}
			}
		}
	}
	convertRoutesToScriptForFunction(smls);

}
void insertMorphers(Seamless* smls, int itemIdx, M_Node& mnode){
	M_Art& partList=smls->partList;
	for(int k=0; k<partList.length; k++){
		if(partList[k]->getType()==MORPHER){
			insertNodeBelow(itemIdx++, mnode, partList[k]);
		}
	}
}

void genTransformedNormals(M_Art& partList, IndexedFaceSet& ifs);
void addExtraCoords(M_Art& partList, M_Int& coordIndex);
//void resetCoordLen(M_Art& partList);



void String::stb(String& dest, int& position)const{
	char* p=&pointer->rawChar[position];
	int newLength=pointer->rawLen-position-1;
	for(int i=0;i<newLength;i++){
		if(*p>' '&&*p!=','){
			if(*p=='#'){
				while(*p>=' '||*p=='\t'){//serch for new selLine
					dest+=*p;
					p++;
				}
			}
			else{
				break;
			}
		}
		dest+=*p;
		p++;
	}
	position=p-pointer->rawChar;
}
void String::stw(String& dest, int& position)const{
	char* p=&pointer->rawChar[position];
	int newLength=pointer->rawLen-position;
	for(int i=0;i<newLength;i++){
		if(*p<=' '||*p==','){
			position=p-pointer->rawChar;
			return; 
		}
		dest+=*p;
		p++;
	}
	throw this;
}
void StretchCurve::DoIt::operator=(const bool& par){
	Vec2f sc=(home.at+home.add)/home.at;
	EC2* ec=dynamic_cast<EC2*>(g_selNode.getRaw());

	M_Vec2f& point=ec->point;
	M_Float& tangent=ec->tangent;
	for(int i=point.length; --i>=0;){
		point[i]*=sc;
		Vec2f tp=Vec2f(1,0);
		tp=tp.rotate(tangent[i]);
		tp*=sc;
		float hyp=sqrt(tp.x*tp.x+tp.y*tp.y);
		float angle=asin(tp.y/hyp);
		if(tp.x<0){
			if(tp.y>0){
				angle=PI-angle;
			}
			else{
				angle=-PI-angle;
			}
		}
		if(angle>0){
			if(angle>PI){
				angle-=PI;
			}
		}
		else if(angle<-PI){
			angle+=PI;
		}

		tangent[i]=angle;
	}
	g_selNode->processFields();
	g_scene->updateBuild();
	houseKeepVisualClue();
	BackUp::monostable.beginCycle();
}
void Viewpoint::Perspective::operator=(const bool& par){
	refValue()=par;
	if(g_viewIndex < g_viewpoints.length && &home==g_viewpoints[g_viewIndex]){
		g_scene->setSceneFromViewpoint(home);
	}
}

void Viewpoint::Zoom::operator=(const float& par){
	refValue()=par;
	if(g_viewIndex < g_viewpoints.length && &home==g_viewpoints[g_viewIndex] && Viewpoint::neutral == false){
		g_scene->setSceneFromViewpoint(home);
	}
}

void Viewpoint::FieldOfView::operator=(const float& par){
	refValue()=par;
	if(g_viewIndex < g_viewpoints.length && &home==g_viewpoints[g_viewIndex] && Viewpoint::neutral == false){
		g_scene->setSceneFromViewpoint(home);
	}
}

void Viewpoint::Orientation::operator=(const Rotation& par){
	refValue()=par;
	if(g_viewIndex < g_viewpoints.length && &home==g_viewpoints[g_viewIndex] && Viewpoint::neutral == false){
		g_scene->setSceneFromViewpoint(home);
	}
}
void Viewpoint::Position::operator=(const Vec3f& par){
	refValue()=par;
	if(g_viewIndex < g_viewpoints.length && &home==g_viewpoints[g_viewIndex] && Viewpoint::neutral == false){
		g_scene->setSceneFromViewpoint(home);
	}
}
void Viewpoint::Capture::operator=(const bool& par){
	g_scene->setViewpointFromScene(home);
}

void Viewpoint::AnimCapture::operator=(const bool& par){
	g_scene->setViewpointFromScene(home);
	if(Anim* anim = Window::animMode()){
		anim->captureViewpoint(&home);
	}
}


void Viewpoint::gotoFirstViewpoint(){
	g_viewIndex=0;
	countUpViewpoints();
	if(g_viewpoints.length){
		g_scene->setSceneFromViewpoint(*(g_viewpoints[0]));
	}
}
void Viewpoint::gotoPrevViewpoint(){
	countUpViewpoints();
	if(g_viewpoints.length){
		if(--g_viewIndex<0){
			g_viewIndex=g_viewpoints.length-1;
		}
		g_scene->setSceneFromViewpoint(*(g_viewpoints[g_viewIndex]));
	}
	g_stw->paint();
}
void Viewpoint::gotoNextViewpoint(){
	countUpViewpoints();
	if(g_viewpoints.length){
		if(++g_viewIndex>=g_viewpoints.length){
			g_viewIndex=0;
		}
		g_scene->setSceneFromViewpoint(*(g_viewpoints[g_viewIndex]));
	}
	g_stw->paint();
}
void RotAccelerator::CalcForReturn::operator=(const bool& par){
	RotAccelerator* rotAccelerator=dynamic_cast<RotAccelerator*>(g_selNode.getRaw());
	if(rotAccelerator!=0){
		rotAccelerator->calcPose(rotAccelerator->init);
	}
}
void RotAccelerator::CalcForPose::operator=(const bool& par){
	RotAccelerator* rotAccelerator=dynamic_cast<RotAccelerator*>(g_selNode.getRaw());
	if(rotAccelerator!=0){
		Art* part;
		if(checkRouteToPart(rotAccelerator->value_changed.route, part)){
			rotAccelerator->calcPose(part->rotation);
		}
	}
}


void TapeMeasure::Measurement::operator=(const float& par){
	refValue()=par;
	if(g_bReady && home==g_selNode.getRaw()){
		Window::paintAllControlPanels();
	}
}
void showBuildError(String str){
	str="Error in build cycle!\n" + str + " Exception";
	msG(str,MB_OK | MB_ICONEXCLAMATION);
	dLog(str);
}
//void Settings::AutoOpen::operator=(const bool& par){
//	refValue()=par;
//	if(par){
//	MessageWin::msg(
//			"Warning!!!\n\
//			autoOpen causes seamless3d to open the previous open file when launched.\n\
//			This can cause seamless3d to fail to launch due to not being able to open a\n\
//			corupted file.\n\
//			See the Survival Guide before leaving autoOpen enabled!",
//			MB_OK | MB_ICONEXCLAMATION
//		);
//	}
//}


void Settings::SaveGZipped::operator=(const bool& par){
	g_writeGzipped = refValue() = par;
}
int getNodeFlatLen(int listPos){
	int targetIndent = g_sceneTreeCollapsed[listPos].indent;
	int i = listPos;
	while(g_sceneTreeCollapsed.length > ++i && g_sceneTreeCollapsed[i].indent > targetIndent);
	return i - listPos; 
}

void countAllSibNodes(int& sib, int& all){

	all = 0;
	int collapsedLine;
	if(getCollapsedLine(g_stw->selLine, collapsedLine)){
	
		int parentSelLine = getParent(g_stw->selLine);
		Node* parentSelNode = g_sceneTreeList[parentSelLine].node;
		int selFieldIdx, selElement;
		getFNodeElementForParent(parentSelLine, g_stw->selLine, g_sceneTreeList, selFieldIdx, selElement );
		MFNode* mfnode=dynamic_cast<MFNode*>(parentSelNode->getField(selFieldIdx));
		if(mfnode==0||mfnode->length<=1){
			return;
		}
		M_Node& m_node = *mfnode;
		if(selElement + sib>m_node.length){
			sib=m_node.length - selElement;
		}

		for(int s=0; s<sib; s++){
			all+=getNodeFlatLen(collapsedLine+all);
		}
	}
}
void SelCount::operator ++(int){
	sib++;
	countAllSibNodes(sib, *const_cast<int*>(&all));
}
void SelCount::operator --(int){
	if(sib>1){
		sib--;
		countAllSibNodes(sib, *const_cast<int*>(&all));
	}
}
void setValueFieldValueOnly(Field* dest, Field* src){ //would it be any less efficiant if we made all fields have this as a virtual function? (this is not commonly used)
	if(SFBool* tDest = dynamic_cast<SFBool*>(dest)){
		tDest->refValue() = *dynamic_cast<SFBool*>(src);
	}
	else if(SFColor* tDest = dynamic_cast<SFColor*>(dest)){
		tDest->refValue() = *dynamic_cast<SFColor*>(src);
	}
	else if(SFFloat* tDest = dynamic_cast<SFFloat*>(dest)){
		tDest->refValue() = *dynamic_cast<SFFloat*>(src);
	}
	else if(SFImage* tDest = dynamic_cast<SFImage*>(dest)){
		tDest->refValue() = *dynamic_cast<SFImage*>(src);
	}
	else if(SFInt32* tDest = dynamic_cast<SFInt32*>(dest)){
		tDest->refValue() = *dynamic_cast<SFInt32*>(src);
	}
	else if(SFNode* tDest = dynamic_cast<SFNode*>(dest)){
		tDest->refValue() = dynamic_cast<SFNode*>(src)->p;
	}
	else if(SFRotation* tDest = dynamic_cast<SFRotation*>(dest)){
		tDest->refValue() = *dynamic_cast<SFRotation*>(src);
	}
	else if(SFString* tDest = dynamic_cast<SFString*>(dest)){
		tDest->refValue() = *dynamic_cast<SFString*>(src);
	}
	else if(SFTime* tDest = dynamic_cast<SFTime*>(dest)){
		tDest->refValue() = *dynamic_cast<SFTime*>(src);
	}
	else if(SFVec2f* tDest = dynamic_cast<SFVec2f*>(dest)){
		tDest->refValue() = *dynamic_cast<SFVec2f*>(src);
	}
	else if(SFVec3f* tDest = dynamic_cast<SFVec3f*>(dest)){
		tDest->refValue() = *dynamic_cast<SFVec3f*>(src);
	}
	else if(MFBool* tDest = dynamic_cast<MFBool*>(dest)){
		tDest->refValue() = *dynamic_cast<MFBool*>(src);
	}
	else if(MFColor* tDest = dynamic_cast<MFColor*>(dest)){
		tDest->refValue() = *dynamic_cast<MFColor*>(src);
	}
	else if(MFFloat* tDest = dynamic_cast<MFFloat*>(dest)){
		tDest->refValue() = *dynamic_cast<MFFloat*>(src);
	}
	else if(MFImage* tDest = dynamic_cast<MFImage*>(dest)){
		tDest->refValue() = *dynamic_cast<MFImage*>(src);
	}
	else if(MFInt32* tDest = dynamic_cast<MFInt32*>(dest)){
		tDest->refValue() = *dynamic_cast<MFInt32*>(src);
	}
	else if(MFNode* tDest = dynamic_cast<MFNode*>(dest)){
		tDest->refValue() = *dynamic_cast<MFNode*>(src);
	}
	else if(MFRotation* tDest = dynamic_cast<MFRotation*>(dest)){
		tDest->refValue() = *dynamic_cast<MFRotation*>(src);
	}
	else if(MFString* tDest = dynamic_cast<MFString*>(dest)){
		tDest->refValue() = *dynamic_cast<MFString*>(src);
	}
	else if(MFTime* tDest = dynamic_cast<MFTime*>(dest)){
		tDest->refValue() = *dynamic_cast<MFTime*>(src);
	}
	else if(MFVec2f* tDest = dynamic_cast<MFVec2f*>(dest)){
		tDest->refValue() = *dynamic_cast<MFVec2f*>(src);
	}
	else if(MFVec3f* tDest = dynamic_cast<MFVec3f*>(dest)){
		tDest->refValue() = *dynamic_cast<MFVec3f*>(src);
	}
}
void Node::deepCopyFields(Node* from, int& selLine){
	selLine++;
	for(int i=0; i<getFLen(); i++){
		Field* fromField=from->getField(i);
		SFNode* sfnode=dynamic_cast<SFNode*>(fromField);
		MFNode* mfnode=dynamic_cast<MFNode*>(fromField);
		if(mfnode){
			M_Node& dst=dynamic_cast<MFNode*>(getField(i))->refValue();
			for(int k=0; k<mfnode->length; k++){
				Node* node=mfnode->refValue()[k];
				if(g_sceneTreeList[selLine].ref){
					dst+=(*mfnode)[k];
					selLine++;
				}
				else{
					dst+=deepCopy(selLine);
				}
			}
		}
		else if(sfnode && sfnode->p){
			if( g_sceneTreeList[selLine].ref==0){
				dynamic_cast<SFNode*>(getField(i))->refValue()=deepCopy(selLine);
			}
			else{//USEd node
				selLine++;
				goto here;
			}
		}
		else{
here:
			setValueFieldValueOnly(getField(i), fromField);
		}
	}
	collapsed=from->collapsed;
	processFields();
}
void makeExistingPartNamesUniqueUsingDefList(Art* part){
	makeExistingNameUniqueUsingDefList(part);
	if(part->effect.refValue()){
		makeExistingNameUniqueUsingDefList(part->effect);
	}
	M_SArt& m_part = part->part.refValue();
	for(int k=0; k<m_part.length; k++){
		makeExistingPartNamesUniqueUsingDefList(m_part[k]);
	}
}
void makeExistingSmlsNamesUniqueUsingDefList(Seamless* smls){
	makeExistingNameUniqueUsingDefList(smls);
	if(smls->effect.refValue()){
		makeExistingNameUniqueUsingDefList(smls->effect);
	}
	M_SArt& skeleton = smls->skeleton.refValue();
	for(int k=0; k<skeleton.length; k++){
		makeExistingPartNamesUniqueUsingDefList(skeleton[k]);
	}
}
NodePtr deepCopy(int& selLine){
	String& str=g_loadStr;
	Node* node=g_sceneTreeList[selLine].node;
	if(Seamless* smls = dynamic_cast<Seamless*>(node)){
		smls->toString(str="");
		M_Node m_smls;
		createSmlsFromString(m_smls, str);
		resetDefList();
		Seamless* smlsCopy = dynamic_cast<Seamless*>(m_smls[0].p);
		makeExistingSmlsNamesUniqueUsingDefList(smlsCopy);
		smlsCopy->processFields();
		smlsCopy->processBuild();
		smlsCopy->doBuild();
		return smlsCopy;
	}
	else{
		str=node->getTypeStr();
		str+="{}";
		M_Node md;
		createSmlsFromString(md, str);
		NodePtr nodeCopy=md[0];
		nodeCopy->deepCopyFields(node, selLine);
		return nodeCopy;
	}
}
void addNewNodeAsChild(Node* node, unsigned fieldIdx, unsigned nodeType){
	Field* field;
	Node* newNode;
	
	commonInitInsertNodeStuff(node, field, newNode, fieldIdx, nodeType);
		
	if(field->getType()==MFNODE){
		MFNode *mfnode=dynamic_cast<MFNode*>(field);
		mfnode->refValue()+=newNode;
		*mfnode=mfnode->refValue();
	}
	else {
		SFNode* sfnode=dynamic_cast<SFNode*>(field);
		*sfnode=newNode;
	}
	g_stw->commonPostInsertNodeStuff();

}
void addCopyOfAsChild(Node* windowNode, unsigned fieldIdx, int copyNodeLine, int copySibCount){
	g_stw->textCell.cellActive=false;
	Field *field = windowNode->getField(fieldIdx);
	if(field->getType()==MFNODE){
		MFNode &mfnode=*dynamic_cast<MFNode*>(field);
		M_Node& mnode=mfnode.refValue();
		deepCopyNodes(mnode.length, mnode, copyNodeLine, copySibCount);
		mfnode=mnode;
	}
	else {
		SFNode &sfnode=*dynamic_cast<SFNode*>(field);
		sfnode=deepCopy(copyNodeLine);
	}
	g_stw->commonPostInsertNodeStuff();
}
void addNewNodeAsSibling(Node* parent, unsigned fieldIdx, int element, unsigned nodeType, int dropWhere){
	Field* field;
	Node* newNode;
	commonInitInsertNodeStuff(parent, field, newNode, fieldIdx, nodeType);
		
	if(field->getType()==MFNODE){
		MFNode *mfnode=dynamic_cast<MFNode*>(field);
		if(dropWhere==DROP_BELOW){
			insertNodeBelow(element, mfnode->refValue(), newNode);
		}
		else{
			insertNodeAbove(element, mfnode->refValue(), newNode);
		}
		*mfnode=mfnode->refValue();
		g_stw->commonPostInsertNodeStuff();


	}
}
void replaceWithNewNode(Node* parent, unsigned fieldIdx, int element, unsigned nodeType){
	Field* field;
	Node* newNode;
	commonInitInsertNodeStuff(parent, field, newNode, fieldIdx, nodeType);
	if(field->getType()==MFNODE){
		MFNode *mfnode=dynamic_cast<MFNode*>(field);
		mfnode->refValue()[element]=newNode;
		*mfnode=mfnode->refValue();
	}
	else {
		SFNode* sfnode=dynamic_cast<SFNode*>(field);
		*sfnode=newNode;
	}
	g_stw->commonPostInsertNodeStuff();
}
void addCopyOfNodeAsSibling(int selNodeLine, int copyNodeLine, int copySibCount, int insertCopyofWhere){
	g_stw->textCell.cellActive=false;

	int selFieldIdx=0;
	int selElement=0;
	
	int parentSelLine=getParent(selNodeLine);
	
	NodePtr parentSelNode=g_sceneTreeList[parentSelLine].node;
	
	getFNodeElementForParent(parentSelLine, selNodeLine, g_sceneTreeList, selFieldIdx, selElement );
	
	copyNodeField(parentSelNode, selFieldIdx, selElement, insertCopyofWhere, copyNodeLine, copySibCount);
}
void replaceNodeWithMove(int selNodeLine, int copyNodeLine, int selCount, int copySibCount, int copyAllCount, int insertCopyofWhere){
	int selFieldIdx=0;
	int selElement=0;
	int copyFieldIdx=0;
	int copyElement=0;
	
	int parentSelLine=getParent(selNodeLine);
	int parentCopyLine=getParent(copyNodeLine);
	
	NodePtr parentSelNode=g_sceneTreeList[parentSelLine].node;
	NodePtr parentCopyNode=g_sceneTreeList[parentCopyLine].node;
	
	checkSurfaceGeneratorBeforeMoveDel(parentSelNode);
	checkSurfaceGeneratorBeforeMoveDel(parentCopyNode);
	
	getFNodeElementForParent(parentSelLine, selNodeLine, g_sceneTreeList, selFieldIdx, selElement );
	getFNodeElementForParent(parentCopyLine, copyNodeLine, g_sceneTreeList, copyFieldIdx, copyElement );
	
	
	moveNodeField(selNodeLine, copyNodeLine, selCount, copySibCount, copyAllCount, parentSelNode, selFieldIdx, selElement, parentCopyNode, copyFieldIdx, copyElement, insertCopyofWhere, parentCopyNode==parentSelNode&&selFieldIdx==copyFieldIdx);
}
void moveNodeAsChild(Node* selNode, int selFieldIdx, int selNodeLine, int copyNodeLine, int selCount, int copySibCount, int copyAllCount){
	Field *field = selNode->getField(selFieldIdx);
	int parentCopyLine=getParent(copyNodeLine);
	int copyFieldIdx=0;
	int copyElement=0;
	int selElement=0;
	NodePtr parentCopyNode=g_sceneTreeList[parentCopyLine].node;
		
	checkSurfaceGeneratorBeforeMoveDel(parentCopyNode);
	if(selNodeLine>0){
		checkSurfaceGeneratorBeforeMoveDel(g_sceneTreeList[getParent(selNodeLine)].node);
	}
		
	getFNodeElementForParent(parentCopyLine, copyNodeLine, g_sceneTreeList, copyFieldIdx, copyElement );
		
	if(field->getType()==MFNODE){
		selElement=field->size()-1;
	}
	moveNodeField(selNodeLine, copyNodeLine, selCount, copySibCount, copyAllCount, selNode, selFieldIdx, selElement, parentCopyNode, copyFieldIdx, copyElement, DROP_BELOW, selNode==parentCopyNode&&(int)selFieldIdx==copyFieldIdx);
}
void useNodeAsChild(Node* selNode, Node* copyNode, int selFieldIdx, int copyNodeLine, int selCount, int copySibCount){
	g_stw->textCell.cellActive=false;

	Field *field = selNode->getField(selFieldIdx);
	if(field->getType()==MFNODE){
		MFNode &mfnode=*dynamic_cast<MFNode*>(field);
		M_Node& mnode=mfnode.refValue();
		useNodes(mnode.length, mnode, copyNodeLine, copySibCount);
		mfnode=mnode;
	}
	else {
		SFNode &sfnode=*dynamic_cast<SFNode*>(field);
		sfnode=copyNode;
	}
	if(copyNode->name==""){
		createUniqueName(copyNode);
	}
	g_stw->commonPostInsertNodeStuff();
}
void useSiblingNode(Node* selNode, int selNodeLine, int copyNodeLine, int selCount, int copySibCount, int type){
	if(selNode->getType()==SCENE)return;
	g_stw->textCell.cellActive=false;
	int selFieldIdx=0;
	int selElement=0;
	int copyFieldIdx=0;
	int copyElement=0;
	
	int parentSelLine=getParent(selNodeLine);
	int parentCopyLine=getParent(copyNodeLine);
	
	NodePtr parentSelNode=g_sceneTreeList[parentSelLine].node;
	NodePtr parentCopyNode=g_sceneTreeList[parentCopyLine].node;
	
	getFNodeElementForParent(parentSelLine, selNodeLine, g_sceneTreeList, selFieldIdx, selElement );
	getFNodeElementForParent(parentCopyLine, copyNodeLine, g_sceneTreeList, copyFieldIdx, copyElement );
	
	useNodeField(parentSelNode, selFieldIdx, selElement, parentCopyNode, copyFieldIdx, copyElement, copyNodeLine, selCount, copySibCount, type);
}
void updateFuseSurfaceClue();
void resetLooseEnds(){

//	if(g_sceneTreeList.length==0){

	g_stw->genNodeList();
	if(g_stw->selLine!=0&&g_stw->selLine>=g_sceneTreeList.length){
		g_stw->selLine=g_sceneTreeList.length-1;
	}
//	}
	if(g_stw->cp->node==0){
		g_stw->cp->node=g_selNode.getRaw();
	}
	g_stw->reset();

	g_selNode->getSceneDisplayString(g_stw->textCell.cellStr="");
	g_stw->textCell.caretPos=g_stw->textCell.highlightPos=0;
	g_selCount.reset();
	g_visualClue=nullPtr;
	setupSelectedNodeWindow();
	g_stw->keepLeagal();

	makePointerEqualIndexLocation();
	setTitleName();
	g_tb->updateToolBarContext();

	g_stw->paint();
	updateFuseSurfaceClue();
}
void Anim::Slider::operator=(const float& par){
	float v=par;
	if(v<0){
		v=0;
	}
	if(v>1){
		if(home.loop){
			v=0;
		}
		else{
			v=1;
		}
	}
	home.update=true;

	refValue()=v;
}
/*
void Anim::Fraction_changed::operator=(const float& par){
	float step=1.0/home.stepsPerSecond/home.cycleInterval;
	float quantisedPeriod = home.cycleInterval;

	int numOfStepsInSlider = (par+step*.9999)/step;
	float quantisedSlider=numOfStepsInSlider*step;

	if(*this - par > .0001){
		quantisedSlider-=step;
	}
	if(quantisedSlider<0 || quantisedSlider>1){
		quantisedSlider=0;
	}
	home.update=true;

	this->refValue()=quantisedSlider;
}
*/
/*
void renderScene();
void Anim::PastePose::operator=(const bool& par){
	float currentSlider = home.slider;
	home.slider.refValue() = home.copySlider;
	renderScene();
	home.slider.refValue() = currentSlider;
	home.captureMultiplePartPose();
}
*/

bool endsWithSemicolon(const String& str){
	for(int i = str.length; --i>=0;){
		char c = str[i];
		if(c == ';'){
			return true;
		}
		if(c > ' '){
			break;
		}
	}
	return false;
}
bool addTextCell(String& str, bool insertCurly){

	if(g_stw->textCell.cellReadyToEnter){
		g_stw->textCell.cellReadyToEnter = false;
		String& cellStr = g_stw->textCell.cellStr;
		str+=cellStr;
		int i = 0;
		cellStr.stbCpp(i);
		if(cellStr == "" || i==cellStr.length-1){
			str+="\n";
		}
		else if(insertCurly){
			if(cellStr.cmpCppWord("if",i) || cellStr.cmpCppWord("else",i) || cellStr.cmpCppWord("while",i) || cellStr.cmpCppWord("for",i) || cellStr.cmpCppWord("do",i)){
				str+="{}\n";
			}
			else{
				if(endsWithSemicolon(str) == false){
					str+=';';
				}
				str+='\n';
			}
		}
		return true;
	}
	return false;
}
void addTextCellForFunctionHeader(String& str){
	str+=g_stw->textCell.cellStr;
}
//if(ptCursor.x >= textValPos.x && ptCursor.y >= textValPos.y && ptCursor.y < textValPos.y+25){

bool input(){
	return false;
}


CallBack::~CallBack(){
}
void commonDoDrag(const Vec2i ptCursor, Vec2i& pos, Vec2i& dragPos){
	Vec2i diff=ptCursor-dragPos;

	pos+=diff;          // horizontal position
	dragPos+=diff;
}
void Drag::endDrag(const Vec2i ptCursor){
	g_drag=0;
}
void _3DWin::onMouseWheel(const int par){
	if(0x8000&GetKeyState(VK_LCONTROL)){
		g_scene->translation += -g_scene->rotation * Vec3f(0,0, par * 0.2f);
	}
	else{
		float a = 1.0f + fabsf(par) * 0.1f;
		if(par > 0){
			g_scene->setZoom(g_scene->getZoom()*a);
		}
		else{
			g_scene->setZoom(g_scene->getZoom()/a);
		}
#ifdef PROG_ROB
		dLog("new MouseZoom("+String(g_scene->scale.x)+"),");
#endif
	}
	houseKeepVisualClue();
}
int CaretWindow::remFirstChar=0;
Vec2i CaretWindow::remPreviousPos;

void changeCursor(LPCSTR requestType){
	if(g_currentType != requestType){
		SetClassLong(hWndMain, GCL_HCURSOR, (LONG)LoadCursor(0, requestType));
		g_currentType = requestType;
	}
}
void Bezier::snapBorderTangentsForFirstCP(int i){
	if(i<fix.length-1 && fix[i+1]==false && (0x8000&GetKeyState(VK_LSHIFT))==0){
		float xDiffA=0, yDiffA=0, hypA=0, xDiffB=0, yDiffB=0, hypB=0;
		getHypes(i, xDiffA, yDiffA, hypA, xDiffB, yDiffB, hypB);

		float ratioA=hypB/hypA;
		Vec2f newPointA = point[i] + Vec2f(-xDiffA*ratioA, -yDiffA*ratioA);

		point[i+1] = newPointA;
	}
}
void Bezier3f::snapBorderTangentsForFirstCP(int i){
	if(i<fix.length-1 && fix[i+1]==false && (0x8000&GetKeyState(VK_LSHIFT))==0){
		float xDiffA=0, yDiffA=0, zDiffA=0, hypA=0, xDiffB=0, yDiffB=0, zDiffB=0, hypB=0;
		getHypes(i, xDiffA, yDiffA, zDiffA, hypA, xDiffB, yDiffB, zDiffB, hypB);

		float ratioA=hypB/hypA;
		Vec3f newPointA = point[i] + Vec3f(-xDiffA*ratioA, -yDiffA*ratioA, -zDiffA*ratioA);

		point[i+1] = newPointA;
	}
}
void Bezier::snapBorderTangentsForLastCP(int i){
	if(i>0 && fix[i-1]==false && (0x8000&GetKeyState(VK_LSHIFT))==0){
		float xDiffA=0, yDiffA=0, hypA=0, xDiffB=0, yDiffB=0, hypB=0;
		getHypes(i, xDiffA, yDiffA, hypA, xDiffB, yDiffB, hypB);

		float ratioB=hypA/hypB;
		Vec2f newPointB = point[i] + Vec2f(-xDiffB*ratioB, -yDiffB*ratioB);

		point[i-1] = newPointB;
	}
}
void Bezier3f::snapBorderTangentsForLastCP(int i){
	if(i>0 && fix[i-1]==false && (0x8000&GetKeyState(VK_LSHIFT))==0){
		float xDiffA=0, yDiffA=0, zDiffA=0, hypA=0, xDiffB=0, yDiffB=0, zDiffB=0, hypB=0;
		getHypes(i, xDiffA, yDiffA, zDiffA, hypA, xDiffB, yDiffB, zDiffB, hypB);

		float ratioB=hypA/hypB;
		Vec3f newPointB = point[i] + Vec3f(-xDiffB*ratioB, -yDiffB*ratioB, -zDiffB*ratioB);

		point[i-1] = newPointB;
	}
}
void Bezier::matchBorderTangents(const Vec2f vec){
	int i=index[0];
	point[i]=vec;
	if((0x8000&GetKeyState(VK_LSHIFT))==0){
		if(i > 0 && i < point.length-2 && fix[i]){

			Vec2f diff=vec-point[i];

			if(fix[i-1]==false){
				point[i-1]+=diff;
			}
			if(fix[i+1]==false){
				point[i+1]+=diff;
			}
		}
		else if(i < point.length-2 && fix[i+1]){
			snapBorderTangentsForFirstCP(i+1);
		}
		else if(i > 1 && fix[i-1]){
			snapBorderTangentsForLastCP(i-1);
		}
	}
}
void Bezier3f::matchBorderTangents(const Vec3f vec){
	int i=index[0];
	point[i]=vec;
	if((0x8000&GetKeyState(VK_LSHIFT))==0){
		if(i > 0 && i < point.length-2 && fix[i]){

			Vec3f diff=vec-point[i];

			if(fix[i-1]==false){
				point[i-1]+=diff;
			}
			if(fix[i+1]==false){
				point[i+1]+=diff;
			}
		}
		else if(i < point.length-2 && fix[i+1]){
			snapBorderTangentsForFirstCP(i+1);
		}
		else if(i > 1 && fix[i-1]){
			snapBorderTangentsForLastCP(i-1);
		}
	}
}
/*
int RingBuild::getLinearRingIndex(int idx, int nodeIdx){

	M_Node& ring = this->ring.refValue();
	int cntr=0;
	for(int k = 0; k < nodeIdx; k++){
F		if(VertexRing* vertexRing = dynamic_cast<VertexRing*>(ring[k].p)){
			cntr+=vertexRing->radius.length;
		}
	}
	return cntr + idx;
}
int RingBuild::getLinearRingLength(int idx){
	M_Node& ring = this->ring.refValue();
	int cntr=0;
	for(int k = 0; k < ring.length; k++){
		if(VertexRing* vertexRing = dynamic_cast<VertexRing*>(ring[k].p)){
			cntr+=vertexRing->radius.length;
		}
	}
	return cntr + idx;
}

bool RingBuild::getLinearRing(int idx, float& radius, Vec2f& scaleX, Vec2f& scaleZ, Rotation& rotation, Vec3f& translation){
	M_Node& ring = this->ring.refValue();
	int cntr=0;
	for(int k = 0; k < ring.length; k++){
		if(VertexRing* vertexRing = dynamic_cast<VertexRing*>(ring[k].p)){
			for(int i = 0; i < vertexRing->radius.refValue().length; i++){
				radius = vertexRing->radius.refValue()[i];
				scaleX = vertexRing->scaleX.refValue()[i];
				scaleZ = vertexRing->scaleZ.refValue()[i];
				rotation = vertexRing->rotation.refValue()[i];
				translation = vertexRing->translation.refValue()[i];
				if(++cntr == idx){
					return true;
				}
			}
		}
	}
	return false;
}
*/
void Stem::insertBands(int& newLen){
	M_Float& height = this->height.refValue();
	M_Int& bands = this->bands.refValue();

	int idx = index[0];
	if(idx > 0){
		int bandsToInsert = bands[idx-1]/2;
		if(bandsToInsert==0){
			bandsToInsert++;
		}
		bands[idx-1]=bandsToInsert;
		bands.insertElement(idx-1, bandsToInsert);

		float diff = height[idx]-height[idx-1];
		height.insertElement(idx, height[idx]);
		height[idx]=height[idx-1]+diff*.5;
		newLen++;
	}
	else if(height.length > 0){
		int bandsToInsert = bands[0]/2;
		if(bandsToInsert==0){
			bandsToInsert++;
		}
		bands.insertElement(0, bandsToInsert);
		bands[1]=bandsToInsert;

		float diff = height[1]-height[0];
		height.insertElement(1, height[0]+diff*.5);
		newLen++;
	}
	else{
		height+=0;
		height+=1;
		bands+= 16;
		index[0]=1;
	}
	processFields();
}
void Stem::uninsertBands(int& newLen){
	M_Float& height = this->height.refValue();
	M_Float& yDistance = this->yDistance.refValue();
	M_Int& bands = this->bands.refValue();

	int idx = index[0];
	if(height.length>2){
		if(idx == 0){
			bands.deleteElement(0);
			height.deleteElement(0);
		}
		else if(idx == height.length-1){
			bands.length--;
			height.length--;
			newLen--;
			index[0]=--idx;
		}
		else{
			bands[idx-1]+=bands[idx];
			bands.deleteElement(idx);

			height[idx]=height[idx+1];
			height.deleteElement(idx+1);
			newLen--;
			index[0]=idx;
		}
	}
	else{
		height.length=0;
		bands.length=0;
		index[0]=0;
	}
	yDistance.length=0;
	processFields();
}
void TransferOwnership::incNumOfGroup(int& newLen){
	M_Int& numOf = this->numOf.refValue();
	M_Int& qty = this->qty.refValue();

	int idx = index[0];
	if(numOf.length){
		numOf.insertElement(idx, numOf[idx]);
		qty.insertElement(idx, qty[idx]);
		index[0]++;
	}
	else{
		numOf+=0;
		qty+=1;
	}
	newLen=numOf.length;
}
void TransferOwnership::decNumOfGroup(int& newLen){
	if(numOf.length==0) return;
	M_Int& numOf = this->numOf.refValue();
	M_Int& qty = this->qty.refValue();

	int idx = index[0];
	numOf.deleteElement(idx);
	qty.deleteElement(idx);
	if(idx >= numOf.length){
		idx = numOf.length-1;
		if(idx < 0){
			idx = 0;
		}
	}
	index[0]=idx;
	newLen=numOf.length;
}
void RingBuild::addControlPoint(int& newLen){
	M_Vec2f& scaleX = this->scaleX.refValue();
	M_Vec2f& scaleZ = this->scaleZ.refValue();
	M_Rotation& orientation = this->orientation.refValue();
	M_Vec3f& location = this->location.refValue();

	int parentIndex = getParent(g_stw->selLine);
	int nodeIdx = parentIndex - g_stw->selLine;
	RingBuild* ringBuild = dynamic_cast<RingBuild*>(g_sceneTreeList[parentIndex].node.p);

	if(scaleX.length==0){
		scaleX=M_Vec2f(Vec2f(.25, .25), Vec2f(.25, .25));
		scaleZ=M_Vec2f(Vec2f(.25, .25), Vec2f(.25, .25));
		orientation=M_Rotation(Rotation(0,0,1,0), Rotation(0,0,1,0));
		location=M_Vec3f(Vec3f(), Vec3f(0, .5, 0));
	}
	else{

		int idx=index[0];
		if(idx<scaleX.length-1){
			idx++;
		}
		int& numOfRingsToInsert = this->numOfRingsToInsert.refValue();
		int marc=idx-1;

		for(int h=1; h<=numOfRingsToInsert; h++){
			float frac = 1.0f/(numOfRingsToInsert + 1) * h;
			scaleX.insertElement(idx, (scaleX[idx]-scaleX[marc])*Vec2f(frac,frac)+scaleX[marc]);
			scaleZ.insertElement(idx, (scaleZ[idx]-scaleZ[marc])*Vec2f(frac,frac)+scaleZ[marc]);
			orientation.insertElement(idx, orientation[marc].slerp(orientation[idx], frac));
			location.insertElement(idx, (location[idx]-location[marc])*Vec3f(frac,frac,frac)+location[marc]);
			idx++;
		}
		index[0]=--idx;
	}
	newLen=scaleX.length;
}
void RingBuild::removeControlPoint(int& newLen){
	M_Vec2f& scaleX = this->scaleX.refValue();
	M_Vec2f& scaleZ = this->scaleZ.refValue();
	M_Rotation& orientation = this->orientation.refValue();
	M_Vec3f& location = this->location.refValue();

	if(scaleX.length==0){
		return;
	}
	int idx=index[0];

	scaleX.deleteElement(idx);
	scaleZ.deleteElement(idx);
	orientation.deleteElement(idx);
	location.deleteElement(idx);

	if(scaleX.length==1){
		scaleX.length=0;
		scaleZ.length=0;
		orientation.length=0;
		location.length=0;
	}
	if(idx >= scaleX.length){
		idx = scaleX.length-1;
		if(idx < 0){
			idx = 0;
		}
		index[0] = idx;
	}

	newLen=scaleX.length;
}
void FuseSurface::incBaseCoordGroup(int& newLen){
	M_Vec2f& baseCoord = this->baseCoord.refValue();
	if(baseCoord.length){
		for(int i = qty; --i>=0;){
			baseCoord+=baseCoord[baseCoord.length-qty];
		}
		numOf.refValue()++;
		newLen=baseCoord.length;
	}
}
bool FuseSurface::processGroup(int groupIndex, int command, int& newLen){
	if(groupIndex==0){
		if(command==1){
			incBaseCoordGroup(newLen);
		}
		else{
			decBaseCoordGroup(newLen);
		}
		return true;
	}
	else{
		return false;
	}
}
bool TransferOwnership::processGroup(int groupIndex, int command, int& newLen){
	if(groupIndex==0){
		if(command==1){
			incNumOfGroup(newLen);
		}
		else{
			decNumOfGroup(newLen);
		}
		g_stw->paint();
		return true;
	}
	else{
		return false;
	}
}
void FuseSurface::decBaseCoordGroup(int& newLen){
	M_Vec2f& baseCoord = this->baseCoord.refValue();

	if(baseCoord.length==0){
		return;
	}

	if(baseCoord.length>qty){
		baseCoord.length-=qty;
		numOf.refValue()--;
	}
	else{
		baseCoord.length=0;
		numOf.refValue()=1;
		qty.refValue()=0;
	}

	newLen=baseCoord.length;
}

void ColorSweep::Spherical::operator=(const bool& par){
	home.spherical.refValue()=true;
	home.straight.refValue()=false;
	Window::activeWindow->paint();
}
void ColorSweep::Straight::operator=(const bool& par){
	home.spherical.refValue()=false;
	home.straight.refValue()=true;
	Window::activeWindow->paint();
}
/*
void BendyLathe::DragRings::operator=(const bool& par){
	refValue()=par;
	int dpointIdx = home.getFieldIndex(&home.dpoint);
	int radiusIdx = home.getFieldIndex(&home.radius);
	int atIdx = home.getFieldIndex(&home.at);
	M_Properties& properties = const_cast<M_Properties&>(home.getProperties());

	properties[dpointIdx].set_disable(par);
	properties[radiusIdx].set_disable(!par);
	properties[atIdx].set_disable(!par);
	removeTypeFromList<Button>(g_callBackList);
	g_stw->cp->setupNodeWindow(&home);
	DownButton::setDownButton(g_stw->cp->buttonList[0]);
}
*/
void formatTime(String& str, const time_t& rawtime){
	char chars[100];
	tm* ptm = gmtime(&rawtime);
	sprintf(chars, "%i/%.2i/%.2i %.2i:%.2i", ptm->tm_year+1900, ptm->tm_mon+1, ptm->tm_mday, (ptm->tm_hour)%24, ptm->tm_min);
	str+=chars;
}
void gmtTime(String& str){
	time_t rawtime;
	time(&rawtime);
	formatTime(str, rawtime);
	str+=" GMT - ";
}
