/* Seamless3d

Copyright (C) 2001-2010 Graham John Perrett (AKA thyme) thyme@seamless3d.com www.seamless3d.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this
software and associated documentation files (the "Software"), to deal in the Software
without restriction, including without limitation the rights to use, copy, modify,
merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to the following
conditions:

The above copyright notice and this permission notice shall be included in all copies or
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT
OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.
*/

#include "seamless3d.h"

extern const char* tip_cCLathe_curve;
extern const char* tip_cCLathe_distance;
extern const char* tip_cCLathe_baseRad;
extern const char* tip_cCLathe_autoClose;
const char* tip_stem_add = "add generated triangles to existing triangles";
extern const char* tip_stem_slide;
extern const char* tip_stem_orientation;
extern const char* tip_stem_staves;
extern const char* tip_stem_xDistance;
extern const char* tip_stem_bands;
extern const char* tip_stem_yDistance;
extern const char* tip_stem_height;
extern const char* tip_stem_sync;
extern const char* tip_stem_mould;
extern const char* tip_stem_insideOut;
extern const char* tip_stem_joinEnds;
extern const char* tip_stem_endSP;
extern const char* tip_stem_beginSP;
extern const char* tip_rectangleBuild_staves;
extern const char* tip_rectangleBuild_bands;
extern const char* tip_rectangleBuild_endSP;
extern const char* tip_rectangleBuild_beginSP;

struct InitPromptNode{
	const int* list;
	const int numOf;
	InitPromptNode(const int* list, const int numOf):
		list(list),
		numOf(numOf)
	{}
};
#define INIT_PROMPT_NODE(x) InitPromptNode(x, sizeof(x)/sizeof(int))

class BIField{//built in field
public:
	const int type;
	const char* str;
	const unsigned int flags;
	const InitPromptNode initPromptNode;
	const char* tip;
	BIField(const int type, const char* str, unsigned int flags, const InitPromptNode initPromptNode, const char* tip=0):
		type(type),
		str(str),
		flags(flags),
		initPromptNode(initPromptNode),
		tip(tip)
	{}
	BIField(const int type, const char* str, unsigned int flags, const char* tip=0):
		type(type),
		str(str),
		flags(flags),
		initPromptNode(0,0),
		tip(tip)
	{}
};
class BIProto{//built in proto
public:
	const char* ntype;
	const char* containerField;
	const BIField* bifieldRef;
	const int numOf;
	BIProto(const char* ntype, const char* containerField, const int numOf, const BIField* bifieldRef):
		ntype(ntype),
		containerField(containerField),
		numOf(numOf),
		bifieldRef(bifieldRef)
	{}
};
class BIProto0 : public BIProto{//built in proto
public:
	BIProto0(const char* ntype, const char* containerField):
		BIProto(ntype, containerField, 0, 0)
	{}
};
class BIProto1 : public BIProto{//built in proto
public:
	const BIField bifield0;
	BIProto1(const char* ntype, const char* containerField, BIField bifield0):
		BIProto(ntype, containerField, 1, &this->bifield0),
		bifield0(bifield0)
	{}
};
class BIProto2 : public BIProto{//built in proto
public:
	const BIField bifield0;
	const BIField bifield1;
	BIProto2(const char* ntype, const char* containerField, BIField bifield0, BIField bifield1):
		BIProto(ntype, containerField, 2, &this->bifield0),
		bifield0(bifield0),
		bifield1(bifield1)
	{}
};
class BIProto3 : public BIProto{//built in proto
public:
	const BIField bifield0;
	const BIField bifield1;
	const BIField bifield2;
	BIProto3(const char* ntype, const char* containerField, BIField bifield0, BIField bifield1, BIField bifield2):
		BIProto(ntype, containerField, 3, &this->bifield0),
		bifield0(bifield0),
		bifield1(bifield1),
		bifield2(bifield2)
	{}
};
class BIProto4 : public BIProto{//built in proto
public:
	const BIField bifield0;
	const BIField bifield1;
	const BIField bifield2;
	const BIField bifield3;
	BIProto4(const char* ntype, const char* containerField, BIField bifield0, BIField bifield1, BIField bifield2, BIField bifield3):
		BIProto(ntype, containerField, 4, &this->bifield0),
		bifield0(bifield0),
		bifield1(bifield1),
		bifield2(bifield2),
		bifield3(bifield3)
	{}
};
class BIProto5 : public BIProto{//built in proto
public:
	const BIField bifield0;
	const BIField bifield1;
	const BIField bifield2;
	const BIField bifield3;
	const BIField bifield4;
	const BIField* my_bifieldRef;
	BIProto5(const char* ntype, const char* containerField, BIField bifield0, BIField bifield1, BIField bifield2, BIField bifield3, BIField bifield4):
		BIProto(ntype, containerField, 5, &this->bifield0),
		bifield0(bifield0),
		bifield1(bifield1),
		bifield2(bifield2),
		bifield3(bifield3),
		bifield4(bifield4),
		my_bifieldRef(&this->bifield0)
	{}
};
class BIProto6 : public BIProto{//built in proto
public:
	const BIField bifield0;
	const BIField bifield1;
	const BIField bifield2;
	const BIField bifield3;
	const BIField bifield4;
	const BIField bifield5;
	BIProto6(const char* ntype, const char* containerField, BIField bifield0, BIField bifield1, BIField bifield2, BIField bifield3, BIField bifield4, BIField bifield5):
		BIProto(ntype, containerField, 6, &this->bifield0),
		bifield0(bifield0),
		bifield1(bifield1),
		bifield2(bifield2),
		bifield3(bifield3),
		bifield4(bifield4),
		bifield5(bifield5)
	{}
};
class BIProto7 : public BIProto{//built in proto
public:
	const BIField bifield0;
	const BIField bifield1;
	const BIField bifield2;
	const BIField bifield3;
	const BIField bifield4;
	const BIField bifield5;
	const BIField bifield6;
	BIProto7(const char* ntype, const char* containerField, BIField bifield0, BIField bifield1, BIField bifield2, BIField bifield3, BIField bifield4, BIField bifield5, BIField bifield6):
		BIProto(ntype, containerField, 7, &this->bifield0),
		bifield0(bifield0),
		bifield1(bifield1),
		bifield2(bifield2),
		bifield3(bifield3),
		bifield4(bifield4),
		bifield5(bifield5),
		bifield6(bifield6)
	{}
};
class BIProto8 : public BIProto{//built in proto
public:
	const BIField bifield0;
	const BIField bifield1;
	const BIField bifield2;
	const BIField bifield3;
	const BIField bifield4;
	const BIField bifield5;
	const BIField bifield6;
	const BIField bifield7;
	BIProto8(const char* ntype, const char* containerField, BIField bifield0, BIField bifield1, BIField bifield2, BIField bifield3, BIField bifield4, BIField bifield5, BIField bifield6, BIField bifield7):
		BIProto(ntype, containerField, 8, &this->bifield0),
		bifield0(bifield0),
		bifield1(bifield1),
		bifield2(bifield2),
		bifield3(bifield3),
		bifield4(bifield4),
		bifield5(bifield5),
		bifield6(bifield6),
		bifield7(bifield7)
	{}
};
class BIProto9 : public BIProto{//built in proto
public:
	const BIField bifield0;
	const BIField bifield1;
	const BIField bifield2;
	const BIField bifield3;
	const BIField bifield4;
	const BIField bifield5;
	const BIField bifield6;
	const BIField bifield7;
	const BIField bifield8;
	BIProto9(const char* ntype, const char* containerField, BIField bifield0, BIField bifield1, BIField bifield2, BIField bifield3, BIField bifield4, BIField bifield5, BIField bifield6, BIField bifield7, BIField bifield8):
		BIProto(ntype, containerField, 9, &this->bifield0),
		bifield0(bifield0),
		bifield1(bifield1),
		bifield2(bifield2),
		bifield3(bifield3),
		bifield4(bifield4),
		bifield5(bifield5),
		bifield6(bifield6),
		bifield7(bifield7),
		bifield8(bifield8)
	{}
};
class BIProto10 : public BIProto{//built in proto
public:
	const BIField bifield0;
	const BIField bifield1;
	const BIField bifield2;
	const BIField bifield3;
	const BIField bifield4;
	const BIField bifield5;
	const BIField bifield6;
	const BIField bifield7;
	const BIField bifield8;
	const BIField bifield9;
	BIProto10(const char* ntype, const char* containerField, BIField bifield0, BIField bifield1, BIField bifield2, BIField bifield3, BIField bifield4, BIField bifield5, BIField bifield6, BIField bifield7, BIField bifield8, BIField bifield9):
		BIProto(ntype, containerField, 10, &this->bifield0),
		bifield0(bifield0),
		bifield1(bifield1),
		bifield2(bifield2),
		bifield3(bifield3),
		bifield4(bifield4),
		bifield5(bifield5),
		bifield6(bifield6),
		bifield7(bifield7),
		bifield8(bifield8),
		bifield9(bifield9)
	{}
};
class BIProto11 : public BIProto{//built in proto
public:
	const BIField bifield0;
	const BIField bifield1;
	const BIField bifield2;
	const BIField bifield3;
	const BIField bifield4;
	const BIField bifield5;
	const BIField bifield6;
	const BIField bifield7;
	const BIField bifield8;
	const BIField bifield9;
	const BIField bifield10;
	BIProto11(const char* ntype, const char* containerField, BIField bifield0, BIField bifield1, BIField bifield2, BIField bifield3, BIField bifield4, BIField bifield5, BIField bifield6, BIField bifield7, BIField bifield8, BIField bifield9, BIField bifield10):
		BIProto(ntype, containerField, 11, &this->bifield0),
		bifield0(bifield0),
		bifield1(bifield1),
		bifield2(bifield2),
		bifield3(bifield3),
		bifield4(bifield4),
		bifield5(bifield5),
		bifield6(bifield6),
		bifield7(bifield7),
		bifield8(bifield8),
		bifield9(bifield9),
		bifield10(bifield10)
	{}
};
class BIProto12 : public BIProto{//built in proto
public:
	const BIField bifield0;
	const BIField bifield1;
	const BIField bifield2;
	const BIField bifield3;
	const BIField bifield4;
	const BIField bifield5;
	const BIField bifield6;
	const BIField bifield7;
	const BIField bifield8;
	const BIField bifield9;
	const BIField bifield10;
	const BIField bifield11;
	BIProto12(const char* ntype, const char* containerField, BIField bifield0, BIField bifield1, BIField bifield2, BIField bifield3, BIField bifield4, BIField bifield5, BIField bifield6, BIField bifield7, BIField bifield8, BIField bifield9, BIField bifield10, BIField bifield11):
		BIProto(ntype, containerField, 12, &this->bifield0),
		bifield0(bifield0),
		bifield1(bifield1),
		bifield2(bifield2),
		bifield3(bifield3),
		bifield4(bifield4),
		bifield5(bifield5),
		bifield6(bifield6),
		bifield7(bifield7),
		bifield8(bifield8),
		bifield9(bifield9),
		bifield10(bifield10),
		bifield11(bifield11)
	{}
};
class BIProto13 : public BIProto{//built in proto
public:
	const BIField bifield0;
	const BIField bifield1;
	const BIField bifield2;
	const BIField bifield3;
	const BIField bifield4;
	const BIField bifield5;
	const BIField bifield6;
	const BIField bifield7;
	const BIField bifield8;
	const BIField bifield9;
	const BIField bifield10;
	const BIField bifield11;
	const BIField bifield12;
	BIProto13(const char* ntype, const char* containerField, BIField bifield0, BIField bifield1, BIField bifield2, BIField bifield3, BIField bifield4, BIField bifield5, BIField bifield6, BIField bifield7, BIField bifield8, BIField bifield9, BIField bifield10, BIField bifield11, BIField bifield12):
		BIProto(ntype, containerField, 13, &this->bifield0),
		bifield0(bifield0),
		bifield1(bifield1),
		bifield2(bifield2),
		bifield3(bifield3),
		bifield4(bifield4),
		bifield5(bifield5),
		bifield6(bifield6),
		bifield7(bifield7),
		bifield8(bifield8),
		bifield9(bifield9),
		bifield10(bifield10),
		bifield11(bifield11),
		bifield12(bifield12)
	{}
};
class BIProto14 : public BIProto{//built in proto
public:
	const BIField bifield0;
	const BIField bifield1;
	const BIField bifield2;
	const BIField bifield3;
	const BIField bifield4;
	const BIField bifield5;
	const BIField bifield6;
	const BIField bifield7;
	const BIField bifield8;
	const BIField bifield9;
	const BIField bifield10;
	const BIField bifield11;
	const BIField bifield12;
	const BIField bifield13;
	BIProto14(const char* ntype, const char* containerField, BIField bifield0, BIField bifield1, BIField bifield2, BIField bifield3, BIField bifield4, BIField bifield5, BIField bifield6, BIField bifield7, BIField bifield8, BIField bifield9, BIField bifield10, BIField bifield11, BIField bifield12, BIField bifield13):
		BIProto(ntype, containerField, 14, &this->bifield0),
		bifield0(bifield0),
		bifield1(bifield1),
		bifield2(bifield2),
		bifield3(bifield3),
		bifield4(bifield4),
		bifield5(bifield5),
		bifield6(bifield6),
		bifield7(bifield7),
		bifield8(bifield8),
		bifield9(bifield9),
		bifield10(bifield10),
		bifield11(bifield11),
		bifield12(bifield12),
		bifield13(bifield13)
	{}
};
class BIProto15 : public BIProto{//built in proto
public:
	const BIField bifield0;
	const BIField bifield1;
	const BIField bifield2;
	const BIField bifield3;
	const BIField bifield4;
	const BIField bifield5;
	const BIField bifield6;
	const BIField bifield7;
	const BIField bifield8;
	const BIField bifield9;
	const BIField bifield10;
	const BIField bifield11;
	const BIField bifield12;
	const BIField bifield13;
	const BIField bifield14;
	BIProto15(const char* ntype, const char* containerField, BIField bifield0, BIField bifield1, BIField bifield2, BIField bifield3, BIField bifield4, BIField bifield5, BIField bifield6, BIField bifield7, BIField bifield8, BIField bifield9, BIField bifield10, BIField bifield11, BIField bifield12, BIField bifield13, BIField bifield14):
		BIProto(ntype, containerField, 15, &this->bifield0),
		bifield0(bifield0),
		bifield1(bifield1),
		bifield2(bifield2),
		bifield3(bifield3),
		bifield4(bifield4),
		bifield5(bifield5),
		bifield6(bifield6),
		bifield7(bifield7),
		bifield8(bifield8),
		bifield9(bifield9),
		bifield10(bifield10),
		bifield11(bifield11),
		bifield12(bifield12),
		bifield13(bifield13),
		bifield14(bifield14)
	{}
};
class BIProto16 : public BIProto{//built in proto
public:
	const BIField bifield0;
	const BIField bifield1;
	const BIField bifield2;
	const BIField bifield3;
	const BIField bifield4;
	const BIField bifield5;
	const BIField bifield6;
	const BIField bifield7;
	const BIField bifield8;
	const BIField bifield9;
	const BIField bifield10;
	const BIField bifield11;
	const BIField bifield12;
	const BIField bifield13;
	const BIField bifield14;
	const BIField bifield15;
	BIProto16(const char* ntype, const char* containerField, BIField bifield0, BIField bifield1, BIField bifield2, BIField bifield3, BIField bifield4, BIField bifield5, BIField bifield6, BIField bifield7, BIField bifield8, BIField bifield9, BIField bifield10, BIField bifield11, BIField bifield12, BIField bifield13, BIField bifield14, BIField bifield15):
		BIProto(ntype, containerField, 16, &this->bifield0),
		bifield0(bifield0),
		bifield1(bifield1),
		bifield2(bifield2),
		bifield3(bifield3),
		bifield4(bifield4),
		bifield5(bifield5),
		bifield6(bifield6),
		bifield7(bifield7),
		bifield8(bifield8),
		bifield9(bifield9),
		bifield10(bifield10),
		bifield11(bifield11),
		bifield12(bifield12),
		bifield13(bifield13),
		bifield14(bifield14),
		bifield15(bifield15)
	{}
};
class BIProto17 : public BIProto{//built in proto
public:
	const BIField bifield0;
	const BIField bifield1;
	const BIField bifield2;
	const BIField bifield3;
	const BIField bifield4;
	const BIField bifield5;
	const BIField bifield6;
	const BIField bifield7;
	const BIField bifield8;
	const BIField bifield9;
	const BIField bifield10;
	const BIField bifield11;
	const BIField bifield12;
	const BIField bifield13;
	const BIField bifield14;
	const BIField bifield15;
	const BIField bifield16;
	BIProto17(const char* ntype, const char* containerField, BIField bifield0, BIField bifield1, BIField bifield2, BIField bifield3, BIField bifield4, BIField bifield5, BIField bifield6, BIField bifield7, BIField bifield8, BIField bifield9, BIField bifield10, BIField bifield11, BIField bifield12, BIField bifield13, BIField bifield14, BIField bifield15, BIField bifield16):
		BIProto(ntype, containerField, 17, &this->bifield0),
		bifield0(bifield0),
		bifield1(bifield1),
		bifield2(bifield2),
		bifield3(bifield3),
		bifield4(bifield4),
		bifield5(bifield5),
		bifield6(bifield6),
		bifield7(bifield7),
		bifield8(bifield8),
		bifield9(bifield9),
		bifield10(bifield10),
		bifield11(bifield11),
		bifield12(bifield12),
		bifield13(bifield13),
		bifield14(bifield14),
		bifield15(bifield15),
		bifield16(bifield16)
	{}
};
class BIProto18 : public BIProto{//built in proto
public:
	const BIField bifield0;
	const BIField bifield1;
	const BIField bifield2;
	const BIField bifield3;
	const BIField bifield4;
	const BIField bifield5;
	const BIField bifield6;
	const BIField bifield7;
	const BIField bifield8;
	const BIField bifield9;
	const BIField bifield10;
	const BIField bifield11;
	const BIField bifield12;
	const BIField bifield13;
	const BIField bifield14;
	const BIField bifield15;
	const BIField bifield16;
	const BIField bifield17;
	BIProto18(const char* ntype, const char* containerField, BIField bifield0, BIField bifield1, BIField bifield2, BIField bifield3, BIField bifield4, BIField bifield5, BIField bifield6, BIField bifield7, BIField bifield8, BIField bifield9, BIField bifield10, BIField bifield11, BIField bifield12, BIField bifield13, BIField bifield14, BIField bifield15, BIField bifield16, BIField bifield17):
		BIProto(ntype, containerField, 18, &this->bifield0),
		bifield0(bifield0),
		bifield1(bifield1),
		bifield2(bifield2),
		bifield3(bifield3),
		bifield4(bifield4),
		bifield5(bifield5),
		bifield6(bifield6),
		bifield7(bifield7),
		bifield8(bifield8),
		bifield9(bifield9),
		bifield10(bifield10),
		bifield11(bifield11),
		bifield12(bifield12),
		bifield13(bifield13),
		bifield14(bifield14),
		bifield15(bifield15),
		bifield16(bifield16),
		bifield17(bifield17)
	{}
};
class BIProto19 : public BIProto{//built in proto
public:
	const BIField bifield0;
	const BIField bifield1;
	const BIField bifield2;
	const BIField bifield3;
	const BIField bifield4;
	const BIField bifield5;
	const BIField bifield6;
	const BIField bifield7;
	const BIField bifield8;
	const BIField bifield9;
	const BIField bifield10;
	const BIField bifield11;
	const BIField bifield12;
	const BIField bifield13;
	const BIField bifield14;
	const BIField bifield15;
	const BIField bifield16;
	const BIField bifield17;
	const BIField bifield18;
	BIProto19(const char* ntype, const char* containerField, BIField bifield0, BIField bifield1, BIField bifield2, BIField bifield3, BIField bifield4, BIField bifield5, BIField bifield6, BIField bifield7, BIField bifield8, BIField bifield9, BIField bifield10, BIField bifield11, BIField bifield12, BIField bifield13, BIField bifield14, BIField bifield15, BIField bifield16, BIField bifield17, BIField bifield18):
		BIProto(ntype, containerField, 19, &this->bifield0),
		bifield0(bifield0),
		bifield1(bifield1),
		bifield2(bifield2),
		bifield3(bifield3),
		bifield4(bifield4),
		bifield5(bifield5),
		bifield6(bifield6),
		bifield7(bifield7),
		bifield8(bifield8),
		bifield9(bifield9),
		bifield10(bifield10),
		bifield11(bifield11),
		bifield12(bifield12),
		bifield13(bifield13),
		bifield14(bifield14),
		bifield15(bifield15),
		bifield16(bifield16),
		bifield17(bifield17),
		bifield18(bifield18)
	{}
};
class BIProto20 : public BIProto{//built in proto
public:
	const BIField bifield0;
	const BIField bifield1;
	const BIField bifield2;
	const BIField bifield3;
	const BIField bifield4;
	const BIField bifield5;
	const BIField bifield6;
	const BIField bifield7;
	const BIField bifield8;
	const BIField bifield9;
	const BIField bifield10;
	const BIField bifield11;
	const BIField bifield12;
	const BIField bifield13;
	const BIField bifield14;
	const BIField bifield15;
	const BIField bifield16;
	const BIField bifield17;
	const BIField bifield18;
	const BIField bifield19;
	BIProto20(const char* ntype, const char* containerField, BIField bifield0, BIField bifield1, BIField bifield2, BIField bifield3, BIField bifield4, BIField bifield5, BIField bifield6, BIField bifield7, BIField bifield8, BIField bifield9, BIField bifield10, BIField bifield11, BIField bifield12, BIField bifield13, BIField bifield14, BIField bifield15, BIField bifield16, BIField bifield17, BIField bifield18, BIField bifield19):
		BIProto(ntype, containerField, 20, &this->bifield0),
		bifield0(bifield0),
		bifield1(bifield1),
		bifield2(bifield2),
		bifield3(bifield3),
		bifield4(bifield4),
		bifield5(bifield5),
		bifield6(bifield6),
		bifield7(bifield7),
		bifield8(bifield8),
		bifield9(bifield9),
		bifield10(bifield10),
		bifield11(bifield11),
		bifield12(bifield12),
		bifield13(bifield13),
		bifield14(bifield14),
		bifield15(bifield15),
		bifield16(bifield16),
		bifield17(bifield17),
		bifield18(bifield18),
		bifield19(bifield19)
	{}
};
class BIProto21 : public BIProto{//built in proto
public:
	const BIField bifield0;
	const BIField bifield1;
	const BIField bifield2;
	const BIField bifield3;
	const BIField bifield4;
	const BIField bifield5;
	const BIField bifield6;
	const BIField bifield7;
	const BIField bifield8;
	const BIField bifield9;
	const BIField bifield10;
	const BIField bifield11;
	const BIField bifield12;
	const BIField bifield13;
	const BIField bifield14;
	const BIField bifield15;
	const BIField bifield16;
	const BIField bifield17;
	const BIField bifield18;
	const BIField bifield19;
	const BIField bifield20;
	BIProto21(const char* ntype, const char* containerField, BIField bifield0, BIField bifield1, BIField bifield2, BIField bifield3, BIField bifield4, BIField bifield5, BIField bifield6, BIField bifield7, BIField bifield8, BIField bifield9, BIField bifield10, BIField bifield11, BIField bifield12, BIField bifield13, BIField bifield14, BIField bifield15, BIField bifield16, BIField bifield17, BIField bifield18, BIField bifield19, BIField bifield20):
		BIProto(ntype, containerField, 21, &this->bifield0),
		bifield0(bifield0),
		bifield1(bifield1),
		bifield2(bifield2),
		bifield3(bifield3),
		bifield4(bifield4),
		bifield5(bifield5),
		bifield6(bifield6),
		bifield7(bifield7),
		bifield8(bifield8),
		bifield9(bifield9),
		bifield10(bifield10),
		bifield11(bifield11),
		bifield12(bifield12),
		bifield13(bifield13),
		bifield14(bifield14),
		bifield15(bifield15),
		bifield16(bifield16),
		bifield17(bifield17),
		bifield18(bifield18),
		bifield19(bifield19),
		bifield20(bifield20)
	{}
};
class BIProto22 : public BIProto{//built in proto
public:
	const BIField bifield0;
	const BIField bifield1;
	const BIField bifield2;
	const BIField bifield3;
	const BIField bifield4;
	const BIField bifield5;
	const BIField bifield6;
	const BIField bifield7;
	const BIField bifield8;
	const BIField bifield9;
	const BIField bifield10;
	const BIField bifield11;
	const BIField bifield12;
	const BIField bifield13;
	const BIField bifield14;
	const BIField bifield15;
	const BIField bifield16;
	const BIField bifield17;
	const BIField bifield18;
	const BIField bifield19;
	const BIField bifield20;
	const BIField bifield21;
	BIProto22(const char* ntype, const char* containerField, BIField bifield0, BIField bifield1, BIField bifield2, BIField bifield3, BIField bifield4, BIField bifield5, BIField bifield6, BIField bifield7, BIField bifield8, BIField bifield9, BIField bifield10, BIField bifield11, BIField bifield12, BIField bifield13, BIField bifield14, BIField bifield15, BIField bifield16, BIField bifield17, BIField bifield18, BIField bifield19, BIField bifield20, BIField bifield21):
		BIProto(ntype, containerField, 22, &this->bifield0),
		bifield0(bifield0),
		bifield1(bifield1),
		bifield2(bifield2),
		bifield3(bifield3),
		bifield4(bifield4),
		bifield5(bifield5),
		bifield6(bifield6),
		bifield7(bifield7),
		bifield8(bifield8),
		bifield9(bifield9),
		bifield10(bifield10),
		bifield11(bifield11),
		bifield12(bifield12),
		bifield13(bifield13),
		bifield14(bifield14),
		bifield15(bifield15),
		bifield16(bifield16),
		bifield17(bifield17),
		bifield18(bifield18),
		bifield19(bifield19),
		bifield20(bifield20),
		bifield21(bifield21)
	{}
};
class BIProto23 : public BIProto{//built in proto
public:
	const BIField bifield0;
	const BIField bifield1;
	const BIField bifield2;
	const BIField bifield3;
	const BIField bifield4;
	const BIField bifield5;
	const BIField bifield6;
	const BIField bifield7;
	const BIField bifield8;
	const BIField bifield9;
	const BIField bifield10;
	const BIField bifield11;
	const BIField bifield12;
	const BIField bifield13;
	const BIField bifield14;
	const BIField bifield15;
	const BIField bifield16;
	const BIField bifield17;
	const BIField bifield18;
	const BIField bifield19;
	const BIField bifield20;
	const BIField bifield21;
	const BIField bifield22;
	BIProto23(const char* ntype, const char* containerField, BIField bifield0, BIField bifield1, BIField bifield2, BIField bifield3, BIField bifield4, BIField bifield5, BIField bifield6, BIField bifield7, BIField bifield8, BIField bifield9, BIField bifield10, BIField bifield11, BIField bifield12, BIField bifield13, BIField bifield14, BIField bifield15, BIField bifield16, BIField bifield17, BIField bifield18, BIField bifield19, BIField bifield20, BIField bifield21, BIField bifield22):
		BIProto(ntype, containerField, 23, &this->bifield0),
		bifield0(bifield0),
		bifield1(bifield1),
		bifield2(bifield2),
		bifield3(bifield3),
		bifield4(bifield4),
		bifield5(bifield5),
		bifield6(bifield6),
		bifield7(bifield7),
		bifield8(bifield8),
		bifield9(bifield9),
		bifield10(bifield10),
		bifield11(bifield11),
		bifield12(bifield12),
		bifield13(bifield13),
		bifield14(bifield14),
		bifield15(bifield15),
		bifield16(bifield16),
		bifield17(bifield17),
		bifield18(bifield18),
		bifield19(bifield19),
		bifield20(bifield20),
		bifield21(bifield21),
		bifield22(bifield22)
	{}
};
class BIProto24 : public BIProto{//built in proto
public:
	const BIField bifield0;
	const BIField bifield1;
	const BIField bifield2;
	const BIField bifield3;
	const BIField bifield4;
	const BIField bifield5;
	const BIField bifield6;
	const BIField bifield7;
	const BIField bifield8;
	const BIField bifield9;
	const BIField bifield10;
	const BIField bifield11;
	const BIField bifield12;
	const BIField bifield13;
	const BIField bifield14;
	const BIField bifield15;
	const BIField bifield16;
	const BIField bifield17;
	const BIField bifield18;
	const BIField bifield19;
	const BIField bifield20;
	const BIField bifield21;
	const BIField bifield22;
	const BIField bifield23;
	BIProto24(const char* ntype, const char* containerField, BIField bifield0, BIField bifield1, BIField bifield2, BIField bifield3, BIField bifield4, BIField bifield5, BIField bifield6, BIField bifield7, BIField bifield8, BIField bifield9, BIField bifield10, BIField bifield11, BIField bifield12, BIField bifield13, BIField bifield14, BIField bifield15, BIField bifield16, BIField bifield17, BIField bifield18, BIField bifield19, BIField bifield20, BIField bifield21, BIField bifield22, BIField bifield23):
		BIProto(ntype, containerField, 24, &this->bifield0),
		bifield0(bifield0),
		bifield1(bifield1),
		bifield2(bifield2),
		bifield3(bifield3),
		bifield4(bifield4),
		bifield5(bifield5),
		bifield6(bifield6),
		bifield7(bifield7),
		bifield8(bifield8),
		bifield9(bifield9),
		bifield10(bifield10),
		bifield11(bifield11),
		bifield12(bifield12),
		bifield13(bifield13),
		bifield14(bifield14),
		bifield15(bifield15),
		bifield16(bifield16),
		bifield17(bifield17),
		bifield18(bifield18),
		bifield19(bifield19),
		bifield20(bifield20),
		bifield21(bifield21),
		bifield22(bifield22),
		bifield23(bifield23)
	{}
};
class BIProto25 : public BIProto{//built in proto
public:
	const BIField bifield0;
	const BIField bifield1;
	const BIField bifield2;
	const BIField bifield3;
	const BIField bifield4;
	const BIField bifield5;
	const BIField bifield6;
	const BIField bifield7;
	const BIField bifield8;
	const BIField bifield9;
	const BIField bifield10;
	const BIField bifield11;
	const BIField bifield12;
	const BIField bifield13;
	const BIField bifield14;
	const BIField bifield15;
	const BIField bifield16;
	const BIField bifield17;
	const BIField bifield18;
	const BIField bifield19;
	const BIField bifield20;
	const BIField bifield21;
	const BIField bifield22;
	const BIField bifield23;
	const BIField bifield24;
	BIProto25(const char* ntype, const char* containerField, BIField bifield0, BIField bifield1, BIField bifield2, BIField bifield3, BIField bifield4, BIField bifield5, BIField bifield6, BIField bifield7, BIField bifield8, BIField bifield9, BIField bifield10, BIField bifield11, BIField bifield12, BIField bifield13, BIField bifield14, BIField bifield15, BIField bifield16, BIField bifield17, BIField bifield18, BIField bifield19, BIField bifield20, BIField bifield21, BIField bifield22, BIField bifield23, BIField bifield24):
		BIProto(ntype, containerField, 25, &this->bifield0),
		bifield0(bifield0),
		bifield1(bifield1),
		bifield2(bifield2),
		bifield3(bifield3),
		bifield4(bifield4),
		bifield5(bifield5),
		bifield6(bifield6),
		bifield7(bifield7),
		bifield8(bifield8),
		bifield9(bifield9),
		bifield10(bifield10),
		bifield11(bifield11),
		bifield12(bifield12),
		bifield13(bifield13),
		bifield14(bifield14),
		bifield15(bifield15),
		bifield16(bifield16),
		bifield17(bifield17),
		bifield18(bifield18),
		bifield19(bifield19),
		bifield20(bifield20),
		bifield21(bifield21),
		bifield22(bifield22),
		bifield23(bifield23),
		bifield24(bifield24)
	{}
};

class BIProto26 : public BIProto{//built in proto
public:
	const BIField bifield0;
	const BIField bifield1;
	const BIField bifield2;
	const BIField bifield3;
	const BIField bifield4;
	const BIField bifield5;
	const BIField bifield6;
	const BIField bifield7;
	const BIField bifield8;
	const BIField bifield9;
	const BIField bifield10;
	const BIField bifield11;
	const BIField bifield12;
	const BIField bifield13;
	const BIField bifield14;
	const BIField bifield15;
	const BIField bifield16;
	const BIField bifield17;
	const BIField bifield18;
	const BIField bifield19;
	const BIField bifield20;
	const BIField bifield21;
	const BIField bifield22;
	const BIField bifield23;
	const BIField bifield24;
	const BIField bifield25;
	BIProto26(const char* ntype, const char* containerField, BIField bifield0, BIField bifield1, BIField bifield2, BIField bifield3, BIField bifield4, BIField bifield5, BIField bifield6, BIField bifield7, BIField bifield8, BIField bifield9, BIField bifield10, BIField bifield11, BIField bifield12, BIField bifield13, BIField bifield14, BIField bifield15, BIField bifield16, BIField bifield17, BIField bifield18, BIField bifield19, BIField bifield20, BIField bifield21, BIField bifield22, BIField bifield23, BIField bifield24, BIField bifield25):
		BIProto(ntype, containerField, 26, &this->bifield0),
		bifield0(bifield0),
		bifield1(bifield1),
		bifield2(bifield2),
		bifield3(bifield3),
		bifield4(bifield4),
		bifield5(bifield5),
		bifield6(bifield6),
		bifield7(bifield7),
		bifield8(bifield8),
		bifield9(bifield9),
		bifield10(bifield10),
		bifield11(bifield11),
		bifield12(bifield12),
		bifield13(bifield13),
		bifield14(bifield14),
		bifield15(bifield15),
		bifield16(bifield16),
		bifield17(bifield17),
		bifield18(bifield18),
		bifield19(bifield19),
		bifield20(bifield20),
		bifield21(bifield21),
		bifield22(bifield22),
		bifield23(bifield23),
		bifield24(bifield24),
		bifield25(bifield25)
	{}
};
class BIProto27 : public BIProto{//built in proto
public:
	const BIField bifield0;
	const BIField bifield1;
	const BIField bifield2;
	const BIField bifield3;
	const BIField bifield4;
	const BIField bifield5;
	const BIField bifield6;
	const BIField bifield7;
	const BIField bifield8;
	const BIField bifield9;
	const BIField bifield10;
	const BIField bifield11;
	const BIField bifield12;
	const BIField bifield13;
	const BIField bifield14;
	const BIField bifield15;
	const BIField bifield16;
	const BIField bifield17;
	const BIField bifield18;
	const BIField bifield19;
	const BIField bifield20;
	const BIField bifield21;
	const BIField bifield22;
	const BIField bifield23;
	const BIField bifield24;
	const BIField bifield25;
	const BIField bifield26;
	BIProto27(const char* ntype, const char* containerField, BIField bifield0, BIField bifield1, BIField bifield2, BIField bifield3, BIField bifield4, BIField bifield5, BIField bifield6, BIField bifield7, BIField bifield8, BIField bifield9, BIField bifield10, BIField bifield11, BIField bifield12, BIField bifield13, BIField bifield14, BIField bifield15, BIField bifield16, BIField bifield17, BIField bifield18, BIField bifield19, BIField bifield20, BIField bifield21, BIField bifield22, BIField bifield23, BIField bifield24, BIField bifield25, BIField bifield26):
		BIProto(ntype, containerField, 27, &this->bifield0),
		bifield0(bifield0),
		bifield1(bifield1),
		bifield2(bifield2),
		bifield3(bifield3),
		bifield4(bifield4),
		bifield5(bifield5),
		bifield6(bifield6),
		bifield7(bifield7),
		bifield8(bifield8),
		bifield9(bifield9),
		bifield10(bifield10),
		bifield11(bifield11),
		bifield12(bifield12),
		bifield13(bifield13),
		bifield14(bifield14),
		bifield15(bifield15),
		bifield16(bifield16),
		bifield17(bifield17),
		bifield18(bifield18),
		bifield19(bifield19),
		bifield20(bifield20),
		bifield21(bifield21),
		bifield22(bifield22),
		bifield23(bifield23),
		bifield24(bifield24),
		bifield25(bifield25),
		bifield26(bifield26)
	{}
};
class BIProto28 : public BIProto{//built in proto
public:
	const BIField bifield0;
	const BIField bifield1;
	const BIField bifield2;
	const BIField bifield3;
	const BIField bifield4;
	const BIField bifield5;
	const BIField bifield6;
	const BIField bifield7;
	const BIField bifield8;
	const BIField bifield9;
	const BIField bifield10;
	const BIField bifield11;
	const BIField bifield12;
	const BIField bifield13;
	const BIField bifield14;
	const BIField bifield15;
	const BIField bifield16;
	const BIField bifield17;
	const BIField bifield18;
	const BIField bifield19;
	const BIField bifield20;
	const BIField bifield21;
	const BIField bifield22;
	const BIField bifield23;
	const BIField bifield24;
	const BIField bifield25;
	const BIField bifield26;
	const BIField bifield27;
	BIProto28(const char* ntype, const char* containerField, BIField bifield0, BIField bifield1, BIField bifield2, BIField bifield3, BIField bifield4, BIField bifield5, BIField bifield6, BIField bifield7, BIField bifield8, BIField bifield9, BIField bifield10, BIField bifield11, BIField bifield12, BIField bifield13, BIField bifield14, BIField bifield15, BIField bifield16, BIField bifield17, BIField bifield18, BIField bifield19, BIField bifield20, BIField bifield21, BIField bifield22, BIField bifield23, BIField bifield24, BIField bifield25, BIField bifield26, BIField bifield27):
		BIProto(ntype, containerField, 28, &this->bifield0),
		bifield0(bifield0),
		bifield1(bifield1),
		bifield2(bifield2),
		bifield3(bifield3),
		bifield4(bifield4),
		bifield5(bifield5),
		bifield6(bifield6),
		bifield7(bifield7),
		bifield8(bifield8),
		bifield9(bifield9),
		bifield10(bifield10),
		bifield11(bifield11),
		bifield12(bifield12),
		bifield13(bifield13),
		bifield14(bifield14),
		bifield15(bifield15),
		bifield16(bifield16),
		bifield17(bifield17),
		bifield18(bifield18),
		bifield19(bifield19),
		bifield20(bifield20),
		bifield21(bifield21),
		bifield22(bifield22),
		bifield23(bifield23),
		bifield24(bifield24),
		bifield25(bifield25),
		bifield26(bifield26),
		bifield27(bifield27)
	{}
};
class BIProto29 : public BIProto{//built in proto
public:
	const BIField bifield0;
	const BIField bifield1;
	const BIField bifield2;
	const BIField bifield3;
	const BIField bifield4;
	const BIField bifield5;
	const BIField bifield6;
	const BIField bifield7;
	const BIField bifield8;
	const BIField bifield9;
	const BIField bifield10;
	const BIField bifield11;
	const BIField bifield12;
	const BIField bifield13;
	const BIField bifield14;
	const BIField bifield15;
	const BIField bifield16;
	const BIField bifield17;
	const BIField bifield18;
	const BIField bifield19;
	const BIField bifield20;
	const BIField bifield21;
	const BIField bifield22;
	const BIField bifield23;
	const BIField bifield24;
	const BIField bifield25;
	const BIField bifield26;
	const BIField bifield27;
	const BIField bifield28;
	BIProto29(const char* ntype, const char* containerField, BIField bifield0, BIField bifield1, BIField bifield2, BIField bifield3, BIField bifield4, BIField bifield5, BIField bifield6, BIField bifield7, BIField bifield8, BIField bifield9, BIField bifield10, BIField bifield11, BIField bifield12, BIField bifield13, BIField bifield14, BIField bifield15, BIField bifield16, BIField bifield17, BIField bifield18, BIField bifield19, BIField bifield20, BIField bifield21, BIField bifield22, BIField bifield23, BIField bifield24, BIField bifield25, BIField bifield26, BIField bifield27, BIField bifield28):
		BIProto(ntype, containerField, 29, &this->bifield0),
		bifield0(bifield0),
		bifield1(bifield1),
		bifield2(bifield2),
		bifield3(bifield3),
		bifield4(bifield4),
		bifield5(bifield5),
		bifield6(bifield6),
		bifield7(bifield7),
		bifield8(bifield8),
		bifield9(bifield9),
		bifield10(bifield10),
		bifield11(bifield11),
		bifield12(bifield12),
		bifield13(bifield13),
		bifield14(bifield14),
		bifield15(bifield15),
		bifield16(bifield16),
		bifield17(bifield17),
		bifield18(bifield18),
		bifield19(bifield19),
		bifield20(bifield20),
		bifield21(bifield21),
		bifield22(bifield22),
		bifield23(bifield23),
		bifield24(bifield24),
		bifield25(bifield25),
		bifield26(bifield26),
		bifield27(bifield27),
		bifield28(bifield28)
	{}
};
class BIProto30 : public BIProto{//built in proto
public:
	const BIField bifield0;
	const BIField bifield1;
	const BIField bifield2;
	const BIField bifield3;
	const BIField bifield4;
	const BIField bifield5;
	const BIField bifield6;
	const BIField bifield7;
	const BIField bifield8;
	const BIField bifield9;
	const BIField bifield10;
	const BIField bifield11;
	const BIField bifield12;
	const BIField bifield13;
	const BIField bifield14;
	const BIField bifield15;
	const BIField bifield16;
	const BIField bifield17;
	const BIField bifield18;
	const BIField bifield19;
	const BIField bifield20;
	const BIField bifield21;
	const BIField bifield22;
	const BIField bifield23;
	const BIField bifield24;
	const BIField bifield25;
	const BIField bifield26;
	const BIField bifield27;
	const BIField bifield28;
	const BIField bifield29;
	BIProto30(const char* ntype, const char* containerField, BIField bifield0, BIField bifield1, BIField bifield2, BIField bifield3, BIField bifield4, BIField bifield5, BIField bifield6, BIField bifield7, BIField bifield8, BIField bifield9, BIField bifield10, BIField bifield11, BIField bifield12, BIField bifield13, BIField bifield14, BIField bifield15, BIField bifield16, BIField bifield17, BIField bifield18, BIField bifield19, BIField bifield20, BIField bifield21, BIField bifield22, BIField bifield23, BIField bifield24, BIField bifield25, BIField bifield26, BIField bifield27, BIField bifield28, BIField bifield29):
		BIProto(ntype, containerField, 30, &this->bifield0),
		bifield0(bifield0),
		bifield1(bifield1),
		bifield2(bifield2),
		bifield3(bifield3),
		bifield4(bifield4),
		bifield5(bifield5),
		bifield6(bifield6),
		bifield7(bifield7),
		bifield8(bifield8),
		bifield9(bifield9),
		bifield10(bifield10),
		bifield11(bifield11),
		bifield12(bifield12),
		bifield13(bifield13),
		bifield14(bifield14),
		bifield15(bifield15),
		bifield16(bifield16),
		bifield17(bifield17),
		bifield18(bifield18),
		bifield19(bifield19),
		bifield20(bifield20),
		bifield21(bifield21),
		bifield22(bifield22),
		bifield23(bifield23),
		bifield24(bifield24),
		bifield25(bifield25),
		bifield26(bifield26),
		bifield27(bifield27),
		bifield28(bifield28),
		bifield29(bifield29)
	{}
};
class BIProto31 : public BIProto{//built in proto
public:
	const BIField bifield0;
	const BIField bifield1;
	const BIField bifield2;
	const BIField bifield3;
	const BIField bifield4;
	const BIField bifield5;
	const BIField bifield6;
	const BIField bifield7;
	const BIField bifield8;
	const BIField bifield9;
	const BIField bifield10;
	const BIField bifield11;
	const BIField bifield12;
	const BIField bifield13;
	const BIField bifield14;
	const BIField bifield15;
	const BIField bifield16;
	const BIField bifield17;
	const BIField bifield18;
	const BIField bifield19;
	const BIField bifield20;
	const BIField bifield21;
	const BIField bifield22;
	const BIField bifield23;
	const BIField bifield24;
	const BIField bifield25;
	const BIField bifield26;
	const BIField bifield27;
	const BIField bifield28;
	const BIField bifield29;
	const BIField bifield30;
	BIProto31(const char* ntype, const char* containerField, BIField bifield0, BIField bifield1, BIField bifield2, BIField bifield3, BIField bifield4, BIField bifield5, BIField bifield6, BIField bifield7, BIField bifield8, BIField bifield9, BIField bifield10, BIField bifield11, BIField bifield12, BIField bifield13, BIField bifield14, BIField bifield15, BIField bifield16, BIField bifield17, BIField bifield18, BIField bifield19, BIField bifield20, BIField bifield21, BIField bifield22, BIField bifield23, BIField bifield24, BIField bifield25, BIField bifield26, BIField bifield27, BIField bifield28, BIField bifield29, BIField bifield30):
		BIProto(ntype, containerField, 31, &this->bifield0),
		bifield0(bifield0),
		bifield1(bifield1),
		bifield2(bifield2),
		bifield3(bifield3),
		bifield4(bifield4),
		bifield5(bifield5),
		bifield6(bifield6),
		bifield7(bifield7),
		bifield8(bifield8),
		bifield9(bifield9),
		bifield10(bifield10),
		bifield11(bifield11),
		bifield12(bifield12),
		bifield13(bifield13),
		bifield14(bifield14),
		bifield15(bifield15),
		bifield16(bifield16),
		bifield17(bifield17),
		bifield18(bifield18),
		bifield19(bifield19),
		bifield20(bifield20),
		bifield21(bifield21),
		bifield22(bifield22),
		bifield23(bifield23),
		bifield24(bifield24),
		bifield25(bifield25),
		bifield26(bifield26),
		bifield27(bifield27),
		bifield28(bifield28),
		bifield29(bifield29),
		bifield30(bifield30)
	{}
};
class BIProto32 : public BIProto{//built in proto
public:
	const BIField bifield0;
	const BIField bifield1;
	const BIField bifield2;
	const BIField bifield3;
	const BIField bifield4;
	const BIField bifield5;
	const BIField bifield6;
	const BIField bifield7;
	const BIField bifield8;
	const BIField bifield9;
	const BIField bifield10;
	const BIField bifield11;
	const BIField bifield12;
	const BIField bifield13;
	const BIField bifield14;
	const BIField bifield15;
	const BIField bifield16;
	const BIField bifield17;
	const BIField bifield18;
	const BIField bifield19;
	const BIField bifield20;
	const BIField bifield21;
	const BIField bifield22;
	const BIField bifield23;
	const BIField bifield24;
	const BIField bifield25;
	const BIField bifield26;
	const BIField bifield27;
	const BIField bifield28;
	const BIField bifield29;
	const BIField bifield30;
	const BIField bifield31;
	BIProto32(const char* ntype, const char* containerField, BIField bifield0, BIField bifield1, BIField bifield2, BIField bifield3, BIField bifield4, BIField bifield5, BIField bifield6, BIField bifield7, BIField bifield8, BIField bifield9, BIField bifield10, BIField bifield11, BIField bifield12, BIField bifield13, BIField bifield14, BIField bifield15, BIField bifield16, BIField bifield17, BIField bifield18, BIField bifield19, BIField bifield20, BIField bifield21, BIField bifield22, BIField bifield23, BIField bifield24, BIField bifield25, BIField bifield26, BIField bifield27, BIField bifield28, BIField bifield29, BIField bifield30, BIField bifield31):
		BIProto(ntype, containerField, 32, &this->bifield0),
		bifield0(bifield0),
		bifield1(bifield1),
		bifield2(bifield2),
		bifield3(bifield3),
		bifield4(bifield4),
		bifield5(bifield5),
		bifield6(bifield6),
		bifield7(bifield7),
		bifield8(bifield8),
		bifield9(bifield9),
		bifield10(bifield10),
		bifield11(bifield11),
		bifield12(bifield12),
		bifield13(bifield13),
		bifield14(bifield14),
		bifield15(bifield15),
		bifield16(bifield16),
		bifield17(bifield17),
		bifield18(bifield18),
		bifield19(bifield19),
		bifield20(bifield20),
		bifield21(bifield21),
		bifield22(bifield22),
		bifield23(bifield23),
		bifield24(bifield24),
		bifield25(bifield25),
		bifield26(bifield26),
		bifield27(bifield27),
		bifield28(bifield28),
		bifield29(bifield29),
		bifield30(bifield30),
		bifield31(bifield31)
	{}
};
class BIProto33 : public BIProto{//built in proto
public:
	const BIField bifield0;
	const BIField bifield1;
	const BIField bifield2;
	const BIField bifield3;
	const BIField bifield4;
	const BIField bifield5;
	const BIField bifield6;
	const BIField bifield7;
	const BIField bifield8;
	const BIField bifield9;
	const BIField bifield10;
	const BIField bifield11;
	const BIField bifield12;
	const BIField bifield13;
	const BIField bifield14;
	const BIField bifield15;
	const BIField bifield16;
	const BIField bifield17;
	const BIField bifield18;
	const BIField bifield19;
	const BIField bifield20;
	const BIField bifield21;
	const BIField bifield22;
	const BIField bifield23;
	const BIField bifield24;
	const BIField bifield25;
	const BIField bifield26;
	const BIField bifield27;
	const BIField bifield28;
	const BIField bifield29;
	const BIField bifield30;
	const BIField bifield31;
	const BIField bifield32;
	BIProto33(const char* ntype, const char* containerField, BIField bifield0, BIField bifield1, BIField bifield2, BIField bifield3, BIField bifield4, BIField bifield5, BIField bifield6, BIField bifield7, BIField bifield8, BIField bifield9, BIField bifield10, BIField bifield11, BIField bifield12, BIField bifield13, BIField bifield14, BIField bifield15, BIField bifield16, BIField bifield17, BIField bifield18, BIField bifield19, BIField bifield20, BIField bifield21, BIField bifield22, BIField bifield23, BIField bifield24, BIField bifield25, BIField bifield26, BIField bifield27, BIField bifield28, BIField bifield29, BIField bifield30, BIField bifield31, BIField bifield32):
		BIProto(ntype, containerField, 33, &this->bifield0),
		bifield0(bifield0),
		bifield1(bifield1),
		bifield2(bifield2),
		bifield3(bifield3),
		bifield4(bifield4),
		bifield5(bifield5),
		bifield6(bifield6),
		bifield7(bifield7),
		bifield8(bifield8),
		bifield9(bifield9),
		bifield10(bifield10),
		bifield11(bifield11),
		bifield12(bifield12),
		bifield13(bifield13),
		bifield14(bifield14),
		bifield15(bifield15),
		bifield16(bifield16),
		bifield17(bifield17),
		bifield18(bifield18),
		bifield19(bifield19),
		bifield20(bifield20),
		bifield21(bifield21),
		bifield22(bifield22),
		bifield23(bifield23),
		bifield24(bifield24),
		bifield25(bifield25),
		bifield26(bifield26),
		bifield27(bifield27),
		bifield28(bifield28),
		bifield29(bifield29),
		bifield30(bifield30),
		bifield31(bifield31),
		bifield32(bifield32)
	{}
};

class BIProto34 : public BIProto{//built in proto
public:
	const BIField bifield0;
	const BIField bifield1;
	const BIField bifield2;
	const BIField bifield3;
	const BIField bifield4;
	const BIField bifield5;
	const BIField bifield6;
	const BIField bifield7;
	const BIField bifield8;
	const BIField bifield9;
	const BIField bifield10;
	const BIField bifield11;
	const BIField bifield12;
	const BIField bifield13;
	const BIField bifield14;
	const BIField bifield15;
	const BIField bifield16;
	const BIField bifield17;
	const BIField bifield18;
	const BIField bifield19;
	const BIField bifield20;
	const BIField bifield21;
	const BIField bifield22;
	const BIField bifield23;
	const BIField bifield24;
	const BIField bifield25;
	const BIField bifield26;
	const BIField bifield27;
	const BIField bifield28;
	const BIField bifield29;
	const BIField bifield30;
	const BIField bifield31;
	const BIField bifield32;
	const BIField bifield33;
	BIProto34(const char* ntype, const char* containerField, BIField bifield0, BIField bifield1, BIField bifield2, BIField bifield3, BIField bifield4, BIField bifield5, BIField bifield6, BIField bifield7, BIField bifield8, BIField bifield9, BIField bifield10, BIField bifield11, BIField bifield12, BIField bifield13, BIField bifield14, BIField bifield15, BIField bifield16, BIField bifield17, BIField bifield18, BIField bifield19, BIField bifield20, BIField bifield21, BIField bifield22, BIField bifield23, BIField bifield24, BIField bifield25, BIField bifield26, BIField bifield27, BIField bifield28, BIField bifield29, BIField bifield30, BIField bifield31, BIField bifield32, BIField bifield33):
		BIProto(ntype, containerField, 34, &this->bifield0),
		bifield0(bifield0),
		bifield1(bifield1),
		bifield2(bifield2),
		bifield3(bifield3),
		bifield4(bifield4),
		bifield5(bifield5),
		bifield6(bifield6),
		bifield7(bifield7),
		bifield8(bifield8),
		bifield9(bifield9),
		bifield10(bifield10),
		bifield11(bifield11),
		bifield12(bifield12),
		bifield13(bifield13),
		bifield14(bifield14),
		bifield15(bifield15),
		bifield16(bifield16),
		bifield17(bifield17),
		bifield18(bifield18),
		bifield19(bifield19),
		bifield20(bifield20),
		bifield21(bifield21),
		bifield22(bifield22),
		bifield23(bifield23),
		bifield24(bifield24),
		bifield25(bifield25),
		bifield26(bifield26),
		bifield27(bifield27),
		bifield28(bifield28),
		bifield29(bifield29),
		bifield30(bifield30),
		bifield31(bifield31),
		bifield32(bifield32),
		bifield33(bifield33)
	{}
};

class BIProto35 : public BIProto{//built in proto
public:
	const BIField bifield0;
	const BIField bifield1;
	const BIField bifield2;
	const BIField bifield3;
	const BIField bifield4;
	const BIField bifield5;
	const BIField bifield6;
	const BIField bifield7;
	const BIField bifield8;
	const BIField bifield9;
	const BIField bifield10;
	const BIField bifield11;
	const BIField bifield12;
	const BIField bifield13;
	const BIField bifield14;
	const BIField bifield15;
	const BIField bifield16;
	const BIField bifield17;
	const BIField bifield18;
	const BIField bifield19;
	const BIField bifield20;
	const BIField bifield21;
	const BIField bifield22;
	const BIField bifield23;
	const BIField bifield24;
	const BIField bifield25;
	const BIField bifield26;
	const BIField bifield27;
	const BIField bifield28;
	const BIField bifield29;
	const BIField bifield30;
	const BIField bifield31;
	const BIField bifield32;
	const BIField bifield33;
	const BIField bifield34;
	BIProto35(const char* ntype, const char* containerField, BIField bifield0, BIField bifield1, BIField bifield2, BIField bifield3, BIField bifield4, BIField bifield5, BIField bifield6, BIField bifield7, BIField bifield8, BIField bifield9, BIField bifield10, BIField bifield11, BIField bifield12, BIField bifield13, BIField bifield14, BIField bifield15, BIField bifield16, BIField bifield17, BIField bifield18, BIField bifield19, BIField bifield20, BIField bifield21, BIField bifield22, BIField bifield23, BIField bifield24, BIField bifield25, BIField bifield26, BIField bifield27, BIField bifield28, BIField bifield29, BIField bifield30, BIField bifield31, BIField bifield32, BIField bifield33, BIField bifield34):
		BIProto(ntype, containerField, 35, &this->bifield0),
		bifield0(bifield0),
		bifield1(bifield1),
		bifield2(bifield2),
		bifield3(bifield3),
		bifield4(bifield4),
		bifield5(bifield5),
		bifield6(bifield6),
		bifield7(bifield7),
		bifield8(bifield8),
		bifield9(bifield9),
		bifield10(bifield10),
		bifield11(bifield11),
		bifield12(bifield12),
		bifield13(bifield13),
		bifield14(bifield14),
		bifield15(bifield15),
		bifield16(bifield16),
		bifield17(bifield17),
		bifield18(bifield18),
		bifield19(bifield19),
		bifield20(bifield20),
		bifield21(bifield21),
		bifield22(bifield22),
		bifield23(bifield23),
		bifield24(bifield24),
		bifield25(bifield25),
		bifield26(bifield26),
		bifield27(bifield27),
		bifield28(bifield28),
		bifield29(bifield29),
		bifield30(bifield30),
		bifield31(bifield31),
		bifield32(bifield32),
		bifield33(bifield33),
		bifield34(bifield34)
	{}
};

extern int g_userProtoBegin;

static const int promptNode_scene_children[] = {ANCHOR, ANIM, ANIMBYANKLE, ANIMBYWRIST, BACKGROUND, BILLBOARD, BOOLEANFILTER, BOOLEANSEQUENCER, BOOLEANTOGGLE, BOOLEANTRIGGER, BSPTREE, CCPATH, COLLISION, COLORINTERPOLATOR, COMPOSITION, COORDINATEINTERPOLATOR, CODELINE, CYLINDERSENSOR, DIRECTIONALLIGHT, FLOATACCELERATOR, FOG, FOG2, FUNCTION, _GROUP, GESTUREGROUP, HANIMHUMANOID, INLINE, INTEGERSEQUENCER, INTEGERTRIGGER, LOD_TYPE, MODELMORPHER, MOVIE, MULTIUSER, NAVIGATIONINFO, NOOUTPUTGROUP, NORMALINTERPOLATOR, NURBSTRACK, ORIENTATIONINTERPOLATOR, PLANESENSOR, POINTLIGHT, POSACCELERATOR, POSITIONINTERPOLATOR, POVFOG, POVLIGHT, PROTODECLARE, PROXIMITYSENSOR, ROTACCELERATOR, ROTFILTER, ROTMULT, SEAMLESS, SCALARINTERPOLATOR, SHAPE, SCRIPT, SOUND, SPHERESENSOR, SPOTLIGHT, STATICGROUP, SWITCH, TIMESENSOR, TIMETRIGGER, TOUCHSENSOR, TRANSFORM, VIEWPOINT, VISIBILITYSENSOR, WAVETOPOLY, WORLDINFO, 	G_S_BOOL, G_S_COLOR, G_S_FLOAT, G_S_IMAGE, G_S_INT32, G_NODEPTR, G_S_ROTATION, G_S_STRING, G_S_TIME, G_S_VEC2F, G_S_VEC3F, G_M_BOOL, G_M_COLOR, G_M_FLOAT, G_M_IMAGE, G_M_INT32, G_M_NODE, G_M_ROTATION, G_M_STRING, G_M_TIME, G_M_VEC2F, G_M_VEC3F};
static const int promptNode_shape_appearance[] = {APPEARANCE};
static const int promptNode_shape_geometry[] = {BOX, CONE, CYLINDER, ELEVATIONGRID, EXTRUSION, NURBSSURFACE, SPHERE, TEXT, INDEXEDLINESET, INDEXEDFACESET, INDEXEDTRIANGLESET, POINTSET, POLYLINE2D};
static const int promptNode_part_effect[] = {TEXTUREEFFECT, COLOREFFECT, MONOCOLOREFFECT};
static const int promptNode_part_morpher[] = {MORPHER};
static const int promptNode_part_part[] = {PART, IMMEDIATEPART};
static const int promptNode_appearance_material[] = {MATERIAL};
static const int promptNode_appearance_texture[] = {IMAGETEXTURE, MOVIETEXTURE};
static const int promptNode_appearance_textureTransform[] = {TEXTURETRANSFORM};
static const int promptNode_indexedFaceSet_coord[] = {COORDINATE};
static const int promptNode_indexedFaceSet_normal[] = {NORMAL};
static const int promptNode_indexedFaceSet_color[] = {COLOR};
static const int promptNode_indexedFaceSet_texCoord[] = {TEXTURECOORDINATE};
static const int promptNode_gestureGroup_children[] = {ANIM};
static const int promptNode_gestureGroup_mobility[] = {MOBILE};
static const int promptNode_surfaceGenerator_generator[] = {BEZIERBEND, BEZIERLATHE, BEZIERPATCH, BEZIERSCALE, BEZIERCIRCUMSCALE, BEZIERSHIFT, BEZIERTAPE, BEZIERTWIST, CCLATHE, CCBEND, CCSCALE, CCTRANSLATE, CCRINGSCALE, ECBEND, ECLATHE, ECSCALE, ECSHIFT, ECSYNCBEND, ECTWIST, RADIALTUG, NURBSPATCH, SINETUG, SURFACETRANSFORM};
static const int promptNode_stem_source[] = {PART, IMMEDIATEPART};
static const int promptNode_stem_part[] = {PART};
static const int promptNode_stem_setup[] = {MEMBERFUNCTION};
static const int promptNode_stem_textureMapper[] = {COLORSWEEP, TEXTUREMAPPER};
static const int promptNode_stem_surface[] = {SURFACEGENERATOR};
static const int promptNode_stem_fuse[] = {FUSESURFACE};
static const int promptNode_place_part[] = {PART};
static const int promptNode_place_setup[] = {MEMBERFUNCTION};
static const int promptNode_seamless_effect[] = {TEXTUREEFFECT, COLOREFFECT, MONOCOLOREFFECT};
static const int promptNode_seamless_skeleton[] = {PART, IMMEDIATEPART};
static const int promptNode_seamless_build[] = {BENDYLATHE, BEZIERLATHEBUILD, BEZIERPATCH, BOXBUILD, BRANCHJOINNEGX, BRANCHJOINPOSX, BRANCHJOINNEGY, BRANCHJOINPOSY, BRANCHJOINNEGZ, BRANCHJOINPOSZ, CENTERJOINT, COLORSWEEP, CONEBUILD, COPYPART, COPYTOMORPH, CYLINDERBUILD, DELPLANE, DELPOLY, DELSPHERE, EVOLVECOORD, EXITBUILD, GENPOLY, GROUPBUILD, JOINVERTEX, NURBSLATHE, NURBSPATCH, PLACE, PLACETEXTURE, POLYMAP, POLYTONURBS, PROJECTONTOTRI, RECTANGLEBUILD, RELATEMORPHER, RINGBUILD, SCGROUPBUILD, SPHEREBUILD, SPLITTRI, STEM, SUBDIVISION, SWAPTRIDIVIDE, TEXTUREMAPPER, TORUSBUILD, TRANSFERBCYL, TRANSFEROWNERSHIP, TRANSFERVERTEX};
static const int promptNode_seamless_mirrorGroup[] = {MIRRORGROUP};
static const int promptNode_groupBuild_setup[] = {MEMBERFUNCTION};
static const int promptNode_joinCloseVertex_source[] = {PART};
static const int promptNode_joinCloseVertex_part[] = {PART};
static const int promptNode_joinCloseVertex_setup[] = {MEMBERFUNCTION};
static const int promptNode_centerJoint_part[] = {PART};
static const int promptNode_centerJoint_setup[] = {MEMBERFUNCTION};
static const int promptNode_delPoly_part[] = {PART};
static const int promptNode_delPoly_setup[] = {MEMBERFUNCTION};
static const int promptNode_joinVertex_source[] = {PART};
static const int promptNode_joinVertex_part[] = {PART};
static const int promptNode_joinVertex_setup[] = {MEMBERFUNCTION};
static const int promptNode_copyPart_source[] = {PART, IMMEDIATEPART};
static const int promptNode_copyPart_dest[] = {PART};
static const int promptNode_copyPart_setup[] = {MEMBERFUNCTION};
static const int promptNode_transferBCyl_source[] = {PART};
static const int promptNode_transferBCyl_dest[] = {PART};
static const int promptNode_transferBCyl_setup[] = {MEMBERFUNCTION};
static const int promptNode_transferOwnership_source[] = {STEM, PART};
static const int promptNode_transferVertex_source[] = {PART};
static const int promptNode_transferVertex_dest[] = {PART};
static const int promptNode_transferVertex_setup[] = {MEMBERFUNCTION};
static const int promptNode_genPoly_part[] = {PART};
static const int promptNode_genPoly_setup[] = {MEMBERFUNCTION};
static const int promptNode_branchJoinNegX_part[] = {PART};
static const int promptNode_branchJoinNegX_setup[] = {MEMBERFUNCTION};
static const int promptNode_branchJoinNegX_surface[] = {SURFACEGENERATOR};
static const int promptNode_branchJoinPosX_part[] = {PART};
static const int promptNode_branchJoinPosX_setup[] = {MEMBERFUNCTION};
static const int promptNode_branchJoinPosX_surface[] = {SURFACEGENERATOR};
static const int promptNode_branchJoinNegY_part[] = {PART};
static const int promptNode_branchJoinNegY_setup[] = {MEMBERFUNCTION};
static const int promptNode_branchJoinNegY_surface[] = {SURFACEGENERATOR};
static const int promptNode_branchJoinPosY_part[] = {PART};
static const int promptNode_branchJoinPosY_setup[] = {MEMBERFUNCTION};
static const int promptNode_branchJoinPosY_surface[] = {SURFACEGENERATOR};
static const int promptNode_branchJoinNegZ_part[] = {PART};
static const int promptNode_branchJoinNegZ_setup[] = {MEMBERFUNCTION};
static const int promptNode_branchJoinNegZ_surface[] = {SURFACEGENERATOR};
static const int promptNode_branchJoinPosZ_part[] = {PART};
static const int promptNode_branchJoinPosZ_setup[] = {MEMBERFUNCTION};
static const int promptNode_branchJoinPosZ_surface[] = {SURFACEGENERATOR};
static const int promptNode_colorSweep_part[] = {PART, IMMEDIATEPART};
static const int promptNode_colorSweep_setup[] = {MEMBERFUNCTION};
static const int promptNode_evolveCoord_source[] = {PART, IMMEDIATEPART};
static const int promptNode_evolveCoord_part[] = {PART};
static const int promptNode_evolveCoord_setup[] = {MEMBERFUNCTION};
static const int promptNode_function_codeLine[] = {CODELINE};
static const int promptNode_memberFunction_codeLine[] = {CODELINE};
static const int promptNode_codeLine_codeLine[] = {CODELINE};
static const int promptNode_morphPart_part[] = {PART};
static const int promptNode_morphPart_bank[] = {MORPHER};
static const int promptNode_anim_soundTrack[] = {COMPOSITION};
static const int promptNode_anim_video[] = {VIDEO};
static const int promptNode_anim_children[] = {ANIMBYANKLE, ANIMBYWRIST, BOOLEANFILTER, BOOLEANSEQUENCER, BOOLEANTOGGLE, BOOLEANTRIGGER, CCPATH, COLORINTERPOLATOR, COORDINATEINTERPOLATOR, CODELINE, CPMORPHER, FLOATACCELERATOR, FUNCTION, _GROUP, INTEGERSEQUENCER, INTEGERTRIGGER, NORMALINTERPOLATOR, NURBSTRACK, ORIENTATIONINTERPOLATOR, POSACCELERATOR, POSCOMPENSATOR, POSITIONINTERPOLATOR, ROTACCELERATOR, ROTFILTER, ROTMULT, SCALARINTERPOLATOR, SOUND, TIMETRIGGER};
static const int promptNode_anim_onFrame[] = {MEMBERFUNCTION};
static const int promptNode_movie_clip[] = {ANIM, VIDEO};
static const int promptNode_timeSensor_on_fraction_changed[] = {MEMBERFUNCTION};
static const int promptNode_elevationGrid_color[] = {COLOR};
static const int promptNode_elevationGrid_normal[] = {NORMAL};
static const int promptNode_elevationGrid_texCoord[] = {TEXTURECOORDINATE};
static const int promptNode_indexedLineSet_coord[] = {COORDINATE};
static const int promptNode_indexedLineSet_color[] = {COLOR};
static const int promptNode_pointSet_coord[] = {COORDINATE};
static const int promptNode_pointSet_color[] = {COLOR};
static const int promptNode_sound_source[] = {AUDIOCLIP, MOVIETEXTURE};
static const int promptNode_text_fontStyle[] = {FONTSTYLE};
static const int promptNode_exitBuild_setup[] = {MEMBERFUNCTION};
static const int promptNode_splitTri_part[] = {PART};
static const int promptNode_splitTri_setup[] = {MEMBERFUNCTION};
static const int promptNode_swapTriDivide_part[] = {PART};
static const int promptNode_swapTriDivide_setup[] = {MEMBERFUNCTION};
static const int promptNode_placeTexture_part[] = {PART};
static const int promptNode_placeTexture_setup[] = {MEMBERFUNCTION};
static const int promptNode_hAnimHumanoid_skeleton[] = {HANIMJOINT};
static const int promptNode_hAnimHumanoid_joints[] = {HANIMJOINT};
static const int promptNode_hAnimHumanoid_skin[] = {SHAPE};
static const int promptNode_hAnimHumanoid_skinCoord[] = {COORDINATE};
static const int promptNode_hAnimHumanoid_skinNormal[] = {NORMAL};
static const int promptNode_hAnimJoint_children[] = {HANIMJOINT, HANIMSEGMENT, HANIMSITE};
static const int promptNode_hAnimJoint_displacers[] = {HANIMDISPLACER};
static const int promptNode_copyToMorph_part[] = {PART, IMMEDIATEPART};
static const int promptNode_copyToMorph_morpher[] = {MORPHER};
static const int promptNode_copyToMorph_setup[] = {MEMBERFUNCTION};
static const int promptNode_relateMorpher_part[] = {PART};
static const int promptNode_relateMorpher_setup[] = {MEMBERFUNCTION};
static const int promptNode_immediatePart_effect[] = {TEXTUREEFFECT, COLOREFFECT, MONOCOLOREFFECT};
static const int promptNode_immediatePart_part[] = {PART, IMMEDIATEPART};
static const int promptNode_delPlane_part[] = {PART};
static const int promptNode_delPlane_setup[] = {MEMBERFUNCTION};
static const int promptNode_tapeMeasure_part[] = {PART, IMMEDIATEPART};
static const int promptNode_tapeMeasure_setup[] = {MEMBERFUNCTION};
static const int promptNode_animByAnkle_pelvis[] = {PART};
static const int promptNode_animByAnkle_hip[] = {PART};
static const int promptNode_animByAnkle_knee[] = {PART};
static const int promptNode_animByAnkle_ankle[] = {PART};
static const int promptNode_animByWrist_sholder[] = {PART};
static const int promptNode_animByWrist_elbow[] = {PART};
static const int promptNode_animByWrist_wrist[] = {PART};
static const int promptNode_indexedTriangleSet_coord[] = {COORDINATE};
static const int promptNode_indexedTriangleSet_normal[] = {NORMAL};
static const int promptNode_indexedTriangleSet_color[] = {COLOR};
static const int promptNode_indexedTriangleSet_texCoord[] = {TEXTURECOORDINATE};
static const int promptNode_sphereBuild_part[] = {PART};
static const int promptNode_sphereBuild_setup[] = {MEMBERFUNCTION};
static const int promptNode_cylinderBuild_part[] = {PART};
static const int promptNode_cylinderBuild_setup[] = {MEMBERFUNCTION};
static const int promptNode_coneBuild_part[] = {PART};
static const int promptNode_coneBuild_setup[] = {MEMBERFUNCTION};
static const int promptNode_torusBuild_part[] = {PART};
static const int promptNode_torusBuild_setup[] = {MEMBERFUNCTION};
static const int promptNode_boxBuild_part[] = {PART};
static const int promptNode_boxBuild_setup[] = {MEMBERFUNCTION};
static const int promptNode_rectangleBuild_part[] = {PART};
static const int promptNode_rectangleBuild_setup[] = {MEMBERFUNCTION};
static const int promptNode_bezierLatheBuild_part[] = {PART};
static const int promptNode_bezierLatheBuild_setup[] = {MEMBERFUNCTION};
static const int promptNode_mirrorGroup_mirrorEdit[] = {MIRROREDIT};
static const int promptNode_mirrorEdit_part[] = {PART, IMMEDIATEPART};
static const int promptNode_polyToNurbs_mimic[] = {NURBSPATCH, POLYMAP};
static const int promptNode_polyToNurbs_original[] = {PART, IMMEDIATEPART};
static const int promptNode_projectOntoTri_feedback[] = {FEEDBACK};
static const int promptNode_feedback_source[] = {PART, IMMEDIATEPART};
//static const int promptNode_nurbsPatch_fuseSurface[] = {FUSESURFACE};
static const int promptNode_nurbsPatch_stem[] = {STEM};
static const int promptNode_fuseSurface_patch[] = {NURBSPATCH};
static const int promptNode_envelope_chords[] = {CHORDS};
static const int promptNode_larynx_chords[] = {CHORDS, ENVELOPE};
static const int promptNode_chords_phaseRef[] = {CHORDS};
static const int promptNode_oscillator_larynx[] = {LARYNX};
static const int promptNode_oscillator_timbre[] = {HARMONIC, HARMONICPATCH, SAMPLE, NURBSPATCH};
static const int promptNode_chords_amplitudeModulator[] = {AMPLITUDEMODULATOR, FREQUENCYMODULATOR};
static const int promptNode_chords_envelope[] = {ENVELOPE};
static const int promptNode_modulator_oscillator[] = {OSCILLATOR};
static const int promptNode_track_sound[] = {OSCILLATOR, PHRASE};
static const int promptNode_composition_track[] = {TRACK};
static const int promptNode_modelSequence_seamless[] = {SEAMLESS};
static const int promptNode_modelMorpher_sequence[] = {MODELSEQUENCE};
static const int promptNode_fuseSurface_stem[] = {STEM};
static const int promptNode_settings_commonSpec[] = {COMMON_SPEC};

static const BIProto0 nullNode_ = BIProto0(
	"NullNode", 0
);
static const BIProto35 scene_ = BIProto35(
	"Scene", 0,
	BIField(SFFLOAT,	"version 0", UNREADABLE | UNWRITEABLE | DISABLE),
	BIField(SFBOOL,	"output FALSE", PUSH),
	BIField(SFBOOL,	"openOutputFolder FALSE", PUSH),
	BIField(SFBOOL,	"outputGZipped TRUE", 0),
	BIField(SFBOOL,	"outputVRML TRUE", 0),
	BIField(SFBOOL,	"outputX3DVRMLClassic FALSE", 0),
	BIField(SFBOOL,	"outputX3DXML FALSE", 0),
	BIField(SFBOOL,	"outputObj FALSE", 0),
	BIField(SFBOOL,	"outputSculpties FALSE", DISABLE),
	BIField(SFBOOL,	"outputPovRay FALSE", 0),
	BIField(SFBOOL,	"outputPovRayMovie FALSE", DISABLE),

	BIField(SFBOOL,	"movieAVI FALSE", 0),
	BIField(SFBOOL,	"movieMPG FALSE", 0),
	BIField(SFBOOL,	"movieMP4 FALSE", 0),
	BIField(SFBOOL,	"movieFLV FALSE", 0),

	BIField(SFBOOL,	"genHAnimNodes FALSE", 0),
	BIField(SFBOOL,	"genIdxTriSetNodes FALSE", 0),
	BIField(SFBOOL,	"genContactSBVT TRUE", 0),
	BIField(SFBOOL,	"genNormals TRUE", 0),
	BIField(SFBOOL,	"genColorIndex FALSE", 0),
	BIField(SFBOOL,	"autoOutSegmentedParts FALSE", 0),
	BIField(SFFLOAT,"roundCoordsOutput 3", 0),
	BIField(SFBOOL,	"outScriptAsRoutes FALSE", 0),
	BIField(SFBOOL,	"outputNodeBind FALSE", PUSH),

	BIField(SFCOLOR,	"backgroundColor 0 0 0", 0),
	BIField(SFCOLOR,	"scene3DFontColor 1 1 0", 0),

	BIField(SFBOOL,		"xRayEdit FALSE", UNREADABLE | UNWRITEABLE | DISABLE),
	BIField(SFBOOL,		"wireframe TRUE", UNREADABLE | UNWRITEABLE | DISABLE),

	BIField(SFINT32, 	"selLine 0", UNREADABLE | UNWRITEABLE | DISABLE),
	BIField(SFINT32,	"scroll 0", UNREADABLE | UNWRITEABLE | DISABLE),
	BIField(MFINT32,	"collapse []", UNREADABLE | UNWRITEABLE | DISABLE),
	BIField(MFINT32,	"wOpen []", UNREADABLE | UNWRITEABLE | DISABLE),
	BIField(MFVEC2F,	"wPos []", UNREADABLE | UNWRITEABLE | DISABLE),
	BIField(MFVEC2F,	"wSize []", UNREADABLE | UNWRITEABLE | DISABLE),
	BIField(MFNODE,	"children []", 0,	INIT_PROMPT_NODE(promptNode_scene_children))
);
static const BIProto8 avatarSocket_ = BIProto8(
	"AvatarSocket", 0,
	BIField(SFROTATION,	"orientation 0 1 0 0", 0),
	BIField(SFVEC3F,	"position 0 0 0", 0),
	BIField(MFNODE,	"children []", 0),
	BIField(SFSTRING,	"nickname \"\"", 0),
	BIField(SFSTRING,	"avatarUrl \"\"", 0),
	BIField(SFBOOL,	"male FALSE", 0),
	BIField(SFSTRING,	"voice \"\"", 0),
	BIField(SFINT32,	"id 0", UNREADABLE | UNWRITEABLE | DISABLE)
);
static const BIProto2 shape_ = BIProto2(
	"Shape", "children",
	BIField(SFNODE,	"appearance NULL", 0,	INIT_PROMPT_NODE(promptNode_shape_appearance)),
	BIField(SFNODE,	"geometry NULL", 0,	INIT_PROMPT_NODE(promptNode_shape_geometry))
);
static const BIProto19 part_ = BIProto19(
	"Part", "_part",
	BIField(SFVEC3F,	"scale 1 1 1", 0),
	BIField(SFROTATION,	"rotation 0 0 1 0", 0),
	BIField(SFVEC3F,	"translation 0 0 0", 0),
	BIField(SFBOOL,		"track TRUE", DISABLE),
	BIField(SFBOOL,		"mirror TRUE", 0),
	BIField(SFBOOL,		"ik FALSE", 0),
	BIField(SFBOOL,		"pin FALSE", 0),
	BIField(SFROTATION,	"ikRot 0 0 1 0", 0),
	BIField(SFVEC3F,	"ikPos 0 0 0", 0),
	BIField(SFFLOAT,	"fka 0", 0), //ik blend angle
	BIField(SFFLOAT,	"fkb 0", 0), //ik blend interpolation
	BIField(MFVEC3F,	"coord []", DISABLE_LEN | DO_NOT_SAVE),
	BIField(MFVEC3F,	"normal []", DISABLE | DISABLE_LEN | DO_NOT_SAVE),
	BIField(MFINT32,	"coordGroup []{beginGroup TRUE rangeFrom 0 initElVal \"10\"}", SAME_AS_LAST | DISABLE),
	BIField(MFFLOAT,	"weight []{initElVal \"1\"}", SAME_AS_LAST | DISABLE),
	BIField(SFVEC3F,	"pivotPoint 0 0 0", 0),
	BIField(SFNODE,		"effect NULL", 0,	INIT_PROMPT_NODE(promptNode_part_effect)),
	BIField(MFNODE,	"morpher []", BLOCK_USE,	INIT_PROMPT_NODE(promptNode_part_morpher)),
	BIField(MFNODE,	"part []", BLOCK_USE,	INIT_PROMPT_NODE(promptNode_part_part))
);
static const BIProto3 appearance_ = BIProto3(
	"Appearance", "appearance",
	BIField(SFNODE,	"material NULL", 0,	INIT_PROMPT_NODE(promptNode_appearance_material)),
	BIField(SFNODE,	"texture NULL", 0,	INIT_PROMPT_NODE(promptNode_appearance_texture)),
	BIField(SFNODE,	"textureTransform NULL", 0,	INIT_PROMPT_NODE(promptNode_appearance_textureTransform))
);
static const BIProto18 indexedFaceSet_ = BIProto18(
	"IndexedFaceSet", "geometry",
	BIField(MFINT32,	"set_coordIndex", UNREADABLE),
	BIField(MFINT32,	"set_normalIndex", UNREADABLE),
	BIField(MFINT32,	"set_colorIndex", UNREADABLE),
	BIField(MFINT32,	"set_texCoordIndex", UNREADABLE),
	BIField(SFNODE,	"coord NULL", 0,	INIT_PROMPT_NODE(promptNode_indexedFaceSet_coord)),
	BIField(SFNODE,	"normal NULL", 0,	INIT_PROMPT_NODE(promptNode_indexedFaceSet_normal)),
	BIField(SFNODE,	"color NULL", 0,	INIT_PROMPT_NODE(promptNode_indexedFaceSet_color)),
	BIField(SFNODE,	"texCoord NULL", 0,	INIT_PROMPT_NODE(promptNode_indexedFaceSet_texCoord)),
	BIField(SFBOOL,	"ccw TRUE", UNREADABLE | UNWRITEABLE),
	BIField(MFINT32,	"coordIndex []{rangeFrom -1}", UNREADABLE | UNWRITEABLE),
	BIField(MFINT32,	"normalIndex []{rangeFrom -1}", UNREADABLE | UNWRITEABLE),
	BIField(MFINT32,	"colorIndex []{rangeFrom -1}", UNREADABLE | UNWRITEABLE),
	BIField(MFINT32,	"texCoordIndex []{rangeFrom -1}", UNREADABLE | UNWRITEABLE),
	BIField(SFBOOL,	"normalPerVertex TRUE", UNREADABLE | UNWRITEABLE),
	BIField(SFBOOL,	"colorPerVertex TRUE", UNREADABLE | UNWRITEABLE),
	BIField(SFBOOL,	"convex TRUE", UNREADABLE | UNWRITEABLE),
	BIField(SFFLOAT,	"creaseAngle 0", UNREADABLE | UNWRITEABLE),
	BIField(SFBOOL,	"solid TRUE", UNREADABLE | UNWRITEABLE)
);
static const BIProto3 imageTexture_ = BIProto3(
	"ImageTexture", "texture",
	BIField(MFSTRING,	"url []", 0),
	BIField(SFBOOL,	"repeatS TRUE", UNREADABLE | UNWRITEABLE),
	BIField(SFBOOL,	"repeatT TRUE", UNREADABLE | UNWRITEABLE)
);
static const BIProto6 material_ = BIProto6(
	"Material", "material",
	BIField(SFFLOAT,	"ambientIntensity .2{rangeFrom 0 rangeTo 1}", 0),
	BIField(SFCOLOR,	"diffuseColor .8 .8 .8", 0),
	BIField(SFCOLOR,	"emissiveColor 0 0 0", 0),
	BIField(SFFLOAT,	"shininess .2{rangeFrom 0 rangeTo 1}", 0),
	BIField(SFCOLOR,	"specularColor 0 0 0", 0),
	BIField(SFFLOAT,	"transparency 0{rangeFrom 0 rangeTo 1}", 0)
);
static const BIProto7 navigationInfo_ = BIProto7(
	"NavigationInfo", "navigationInfo",
	BIField(SFBOOL,	"set_bind", UNREADABLE),
	BIField(MFFLOAT,	"avatarSize [.25 1.6 .75 ]", 0),
	BIField(SFBOOL,	"headlight TRUE", 0),
	BIField(SFFLOAT,	"speed 1", 0),
	BIField(MFSTRING,	"type [\"WALK\" \"ANY\"]", 0),
	BIField(SFFLOAT,	"visibilityLimit 0", 0),
	BIField(SFBOOL,	"isBound", UNWRITEABLE)
);
static const BIProto5 group_ = BIProto5(
	"Group", "children",
	BIField(MFNODE,	"addChildren", UNREADABLE),
	BIField(MFNODE,	"removeChildren", UNREADABLE),
	BIField(MFNODE,	"children []", 0,	INIT_PROMPT_NODE(promptNode_scene_children)),
	BIField(SFVEC3F,	"bboxCenter 0 0 0", UNREADABLE | UNWRITEABLE),
	BIField(SFVEC3F,	"bboxSize -1 -1 -1", UNREADABLE | UNWRITEABLE)
);
static const BIProto5 noOutputGroup_ = BIProto5(
	"NoOutputGroup", "children",
	BIField(MFNODE,	"addChildren", UNREADABLE),
	BIField(MFNODE,	"removeChildren", UNREADABLE),
	BIField(MFNODE,	"children []", 0,	INIT_PROMPT_NODE(promptNode_scene_children)),
	BIField(SFVEC3F,	"bboxCenter 0 0 0", UNREADABLE | UNWRITEABLE | DISABLE),
	BIField(SFVEC3F,	"bboxSize -1 -1 -1", UNREADABLE | UNWRITEABLE | DISABLE)
);
static const BIProto8 gestureGroup_ = BIProto8(
	"GestureGroup", "children",
	BIField(MFNODE,	"addChildren", UNREADABLE),
	BIField(MFNODE,	"removeChildren", UNREADABLE),
	BIField(SFROTATION,	"rotation 0 0 1 0", 0),
	BIField(SFVEC3F,	"translation 0 0 0", 0),
	BIField(SFVEC3F,	"bboxCenter 0 0 0", UNREADABLE | UNWRITEABLE | DISABLE),
	BIField(SFVEC3F,	"bboxSize -1 -1 -1", UNREADABLE | UNWRITEABLE | DISABLE),
	BIField(MFNODE,	"children []", 0,	INIT_PROMPT_NODE(promptNode_gestureGroup_children)),
	BIField(MFNODE,	"mobility []", 0,	INIT_PROMPT_NODE(promptNode_gestureGroup_mobility))
);
static const BIProto3 mobility_ = BIProto3(
	"Mobility", "mobility",
	BIField(MFNODE,	"children []", 0,	INIT_PROMPT_NODE(promptNode_scene_children)),
	BIField(SFNODE,	"onRot NULL", 0,	INIT_PROMPT_NODE(promptNode_timeSensor_on_fraction_changed)),
	BIField(SFNODE,	"onTran NULL", 0,	INIT_PROMPT_NODE(promptNode_timeSensor_on_fraction_changed))
);
static const BIProto4 cCLathe_ = BIProto4(
	"CCLathe", "generator",
	BIField(MFFLOAT,	"curve []{beginGroup TRUE initElVal \"[-1.5708 1.5708 ]\"}", SAME_AS_LAST,	tip_cCLathe_curve),
	BIField(MFFLOAT,	"distance []{startAtIndex 1 initElVal \"1.5708\"}", SAME_AS_LAST,	tip_cCLathe_distance),
	BIField(SFFLOAT,	"baseRad 0", 0,	tip_cCLathe_baseRad),
	BIField(SFBOOL,	"autoClose FALSE", 0,	tip_cCLathe_autoClose)
);
static const BIProto5 cCBend_ = BIProto5(
	"CCBend", "generator",
	BIField(SFVEC3F,	"location 0 0 0", 0),
	BIField(SFROTATION,	"orientation 0 1 0 0", 0),
	BIField(MFFLOAT,	"curve []{beginGroup TRUE initElVal \"[0 .0001 ]\"}", SAME_AS_LAST),
	BIField(MFFLOAT,	"distance []{startAtIndex 1 initElVal \"1.5708\"}", SAME_AS_LAST),
	BIField(MFFLOAT,	"yAngle []{startAtIndex 1 initElVal \"0\"}", SAME_AS_LAST)
);
static const BIProto1 surfaceGenerator_ = BIProto1(
	"SurfaceGenerator", "surface",
	BIField(MFNODE,	"generator []", 0,	INIT_PROMPT_NODE(promptNode_surfaceGenerator_generator))
);
static const BIProto30 stem_ = BIProto30(
	"Stem", "build",
	BIField(SFBOOL,	"add TRUE", 0, tip_stem_add),
	BIField(SFBOOL,	"mirror FALSE", 0),
	BIField(SFBOOL,	"store FALSE", 0),

	BIField(SFVEC2F,	"slide 0 0", 0,	tip_stem_slide),
	BIField(SFROTATION,	"orientation 0 1 0 0", DISABLE,	tip_stem_orientation),
	BIField(SFINT32,	"staves 16{rangeFrom 1}", 0,	tip_stem_staves),
	BIField(SFFLOAT,	"xDistance 6.283185", 0,	tip_stem_xDistance),
	BIField(MFFLOAT,	"height []{beginGroup TRUE initElVal \"[0 1]\"}", SAME_AS_LAST,	tip_stem_height),
	BIField(MFFLOAT,	"yDistance []{startAtIndex 1 initElVal \"1\"}", SAME_AS_LAST ,	tip_stem_yDistance),
	BIField(MFINT32,	"bands []{startAtIndex 1 rangeFrom 1 initElVal \"16\"}", SAME_AS_LAST,	tip_stem_bands),
	//BIField(MFINT32,	"ownership []{beginGroup TRUE initElVal \"1\" rangeFrom 0}", SAME_AS_LAST),
	BIField(SFBOOL,	"sync FALSE", 0,	tip_stem_sync),
	BIField(SFBOOL,	"mould TRUE", 0,	tip_stem_mould),
	BIField(SFBOOL,	"northPole FALSE", 0),
	BIField(SFBOOL,	"southPole FALSE", 0),
	BIField(SFBOOL,	"insideOut FALSE", 0,	tip_stem_insideOut),
	BIField(SFBOOL,	"xCentred TRUE", 0),
	BIField(SFBOOL,	"xJoinEnds TRUE", 0,	tip_stem_joinEnds),
	BIField(SFBOOL,	"endSP FALSE", 0,	tip_stem_endSP),
	BIField(SFBOOL,	"beginSP FALSE", 0,	tip_stem_beginSP),

	BIField(MFVEC2F,	"coord []", DISABLE),
	BIField(MFCOLOR,	"color []", DISABLE),
	BIField(MFVEC2F,	"texCoord []", DISABLE),
	BIField(MFINT32,	"coordIndex []{rangeFrom 0}", UNREADABLE | UNWRITEABLE | DISABLE),
	BIField(MFINT32,	"rc []", UNREADABLE | UNWRITEABLE | DISABLE),

	BIField(MFNODE,	"source []", BLOCK_NEW | BLUE_DOT,	INIT_PROMPT_NODE(promptNode_stem_source)),
	BIField(MFNODE,	"part []", BLOCK_NEW | RED_DOT,	INIT_PROMPT_NODE(promptNode_stem_part)),
	BIField(SFNODE,	"setup NULL", 0,	INIT_PROMPT_NODE(promptNode_stem_setup)),
	BIField(MFNODE,	"textureMapper []", 0,	INIT_PROMPT_NODE(promptNode_stem_textureMapper)),
	BIField(SFNODE,	"surface NULL", 0,	INIT_PROMPT_NODE(promptNode_stem_surface)),
	BIField(MFNODE,	"fuse []", 0,	INIT_PROMPT_NODE(promptNode_stem_fuse))
);

static const BIProto10 polymap_ = BIProto10(
	"Polymap", "build",
	BIField(SFBOOL,	"add TRUE", 0, tip_stem_add),

	BIField(MFVEC2F,	"coord []", DISABLE),
	BIField(MFCOLOR,	"color []", DISABLE),
	BIField(MFVEC2F,	"texCoord []", DISABLE),
	BIField(MFINT32,	"coordIndex []{rangeFrom 0}", UNREADABLE | UNWRITEABLE | DISABLE),
	BIField(MFINT32,	"rc []", UNREADABLE | UNWRITEABLE | DISABLE),
	BIField(MFINT32,	"patchOwner []", DISABLE),
	BIField(MFNODE,	"surface []", 0,	INIT_PROMPT_NODE(promptNode_surfaceGenerator_generator)),
	BIField(SFNODE,	"part NULL", BLOCK_NEW ,	INIT_PROMPT_NODE(promptNode_stem_part)),
	BIField(SFNODE,	"setup NULL", 0,	INIT_PROMPT_NODE(promptNode_stem_setup))
);

static const BIProto0 branchJoin_ = BIProto0(
	"BranchJoin", "build"
);
static const BIProto0 generator_ = BIProto0(
	"Generator", 0
);
static const BIProto6 geomSource_ = BIProto6(
	"GeomSource", "source",
	BIField(MFVEC3F,	"coord []", 0),
	BIField(MFVEC3F,	"normal []", 0),
	BIField(MFCOLOR,	"color []", 0),
	BIField(MFVEC2F,	"texCoord []", 0),
	BIField(MFINT32,	"coordIndex []{rangeFrom 0}", UNREADABLE | UNWRITEABLE),
	BIField(MFINT32,	"rc []", UNREADABLE | UNWRITEABLE)
);
static const BIProto3 surfaceTransform_ = BIProto3(
	"SurfaceTransform", "generator",
	BIField(SFVEC3F,	"scale 1 1 1", 0),
	BIField(SFROTATION,	"rotation 0 1 0 0", 0),
	BIField(SFVEC3F,	"translation 0 0 0", 0)
);
static const BIProto12 place_ = BIProto12(
	"Place", "build",
	BIField(SFVEC3F,	"scale 1 1 1", 0),
	BIField(SFROTATION,	"rotation 0 1 0 0", 0),
	BIField(SFINT32,	"tranVertex 0{rangeFrom -1}", 0),
	BIField(SFVEC3F,	"multiply 0 0 0", DISABLE),
	BIField(SFBOOL,	"alignX FALSE", DO_NOT_SAVE),
	BIField(SFBOOL,	"alignY FALSE", DO_NOT_SAVE),
	BIField(SFBOOL,	"alignZ FALSE", DO_NOT_SAVE),
	BIField(SFVEC3F,	"translation 0 0 0", 0),
	BIField(SFBOOL,	"descendants TRUE", 0),
	BIField(SFBOOL,	"staticTransform TRUE", 0),
	BIField(MFNODE,	"part []", BLOCK_NEW,	INIT_PROMPT_NODE(promptNode_place_part)),
	BIField(SFNODE,	"setup NULL", 0,	INIT_PROMPT_NODE(promptNode_place_setup))
);
static const BIProto6 directionalLight_ = BIProto6(
	"DirectionalLight", "children",
	BIField(SFFLOAT,	"ambientIntensity 0", 0),
	BIField(SFCOLOR,	"color 1 1 1", 0),
	BIField(SFVEC3F,	"direction 0 0 -1", 0),
	BIField(SFBOOL,	"global FALSE", 0),
	BIField(SFFLOAT,	"intensity 1", 0),
	BIField(SFBOOL,	"on TRUE", 0)
);
static const BIProto13 cCScale_ = BIProto13(
	"CCScale", "generator",
	BIField(SFVEC3F,	"location 0 0 0", 0),
	BIField(SFROTATION,	"orientation 0 1 0 0", 0),
	BIField(SFFLOAT,	"baseRad 1", 0),
	BIField(MFFLOAT,	"curve []{beginGroup TRUE initElVal \"[0 .0001 ]\"}", SAME_AS_LAST),
	BIField(MFFLOAT,	"distance []{startAtIndex 1 initElVal \"[1.5708]\"}", SAME_AS_LAST),
	BIField(SFBOOL,	"bNegX FALSE", 0),
	BIField(SFBOOL,	"bPosX FALSE", 0),
	BIField(SFBOOL,	"bNegZ FALSE", 0),
	BIField(SFBOOL,	"bPosZ FALSE", 0),
	BIField(SFFLOAT,	"negX 1", 0),
	BIField(SFFLOAT,	"posX 1", 0),
	BIField(SFFLOAT,	"negZ 1", 0),
	BIField(SFFLOAT,	"posZ 1", 0)
);
static const BIProto5 cCTranslate_ = BIProto5(
	"CCTranslate", "generator",
	BIField(SFVEC3F,	"location 0 0 0", 0),
	BIField(SFROTATION,	"orientation 0 1 0 0", 0),
	BIField(SFFLOAT,	"baseRad 0", 0),
	BIField(MFFLOAT,	"curve []{beginGroup TRUE initElVal \"[-1.5708 1.5708 ]\"}", SAME_AS_LAST),
	BIField(MFFLOAT,	"distance []{startAtIndex 1 initElVal \"1.5708\"}", SAME_AS_LAST)
);
static const BIProto11 cCRingScale_ = BIProto11(
	"CCRingScale", "generator",
	BIField(SFVEC3F,	"location 0 0 0", 0),
	BIField(SFROTATION,	"orientation 0 1 0 0", 0),
	BIField(SFFLOAT,	"centerOffset 0", 0),
	BIField(SFFLOAT,	"baseRad 1", 0),
	BIField(MFFLOAT,	"curve []{beginGroup TRUE initElVal \"[-1.5708 1.5708 ]\"}", SAME_AS_LAST),
	BIField(MFFLOAT,	"distance []{startAtIndex 1 initElVal \"1.5708\"}", SAME_AS_LAST),
	BIField(SFFLOAT,	"autoEnd 1", 0),
	BIField(SFFLOAT,	"negX 1", 0),
	BIField(SFFLOAT,	"posX 1", 0),
	BIField(SFFLOAT,	"negZ 1", 0),
	BIField(SFFLOAT,	"posZ 1", 0)
);
static const BIProto3 bezierLathe_ = BIProto3(
	"BezierLathe", "generator",
	BIField(MFVEC2F,	"point []{beginGroup TRUE}", 0),
	BIField(MFBOOL,	"fix []", DO_NOT_SAVE),
	BIField(MFINT32,	"nopfs []", DISABLE)
);
static const BIProto2 eCLathe_ = BIProto2(
	"ECLathe", "generator",
	BIField(MFVEC2F,	"point []{beginGroup TRUE initElVal \"[0 0 1 1]\"}", SAME_AS_LAST),
	BIField(MFFLOAT,	"tangent []{initElVal \"[0 1.570796]\"}", SAME_AS_LAST)
);
static const BIProto12 eCScale_ = BIProto12(
	"ECScale", "generator",
	BIField(SFVEC3F,	"location 0 0 0", 0),
	BIField(SFROTATION,	"orientation 0 1 0 0", 0),
	BIField(MFVEC2F,	"point []{beginGroup TRUE initElVal \"[0 0 .1 .5]\"}", SAME_AS_LAST),
	BIField(MFFLOAT,	"tangent []{initElVal \"[1.570796 1.089909]\"}", SAME_AS_LAST),
	BIField(SFBOOL,	"bNegX FALSE", 0),
	BIField(SFBOOL,	"bPosX FALSE", 0),
	BIField(SFBOOL,	"bNegZ FALSE", 0),
	BIField(SFBOOL,	"bPosZ FALSE", 0),
	BIField(SFFLOAT,	"negX 1", 0),
	BIField(SFFLOAT,	"posX 1", 0),
	BIField(SFFLOAT,	"negZ 1", 0),
	BIField(SFFLOAT,	"posZ 1", 0)
);
static const BIProto6 eCShift_ = BIProto6(
	"ECShift", "generator",
	BIField(SFVEC3F,	"location 0 0 0", 0),
	BIField(SFROTATION,	"orientation 0 1 0 0", 0),
	BIField(MFVEC2F,	"point []{beginGroup TRUE initElVal \"[0 0 .1 .5]\"}", SAME_AS_LAST),
	BIField(MFFLOAT,	"tangent []{initElVal \"[1.570796 1.089909]\"}", SAME_AS_LAST),
	BIField(SFFLOAT,	"negDecay 1", 0),
	BIField(SFFLOAT,	"posDecay 1", 0)
);
static const BIProto5 eCBend_ = BIProto5(
	"ECBend", "generator",
	BIField(SFVEC3F,	"location 0 0 0", 0),
	BIField(SFROTATION,	"orientation 0 1 0 0", 0),
	BIField(MFVEC2F,	"point []{beginGroup TRUE initElVal \"[0 0 .1 .5]\"}", SAME_AS_LAST),
	BIField(MFFLOAT,	"tangent []{initElVal \"[1.570796 1.089909]\"}", SAME_AS_LAST),
	BIField(MFFLOAT,	"yAngle []{startAtIndex 1 initElVal \"0\"}", SAME_AS_LAST)
);
static const BIProto2 eCSyncBend_ = BIProto2(
	"ECSyncBend", "generator",
	BIField(MFVEC2F,	"point []{beginGroup TRUE initElVal \"[0 0 .1 .5]\"}", SAME_AS_LAST),
	BIField(MFFLOAT,	"tangent []{initElVal \"[1.570796 1.089909]\"}", SAME_AS_LAST)
);
static const BIProto4 eCTwist_ = BIProto4(
	"ECTwist", 0,
	BIField(SFVEC3F,	"location 0 0 0", 0),
	BIField(SFROTATION,	"orientation 0 1 0 0", 0),
	BIField(MFVEC2F,	"point []{beginGroup TRUE initElVal \"[0 0 1 2]\"}", SAME_AS_LAST),
	BIField(MFFLOAT,	"tangent []{initElVal \"[.995382 1.29955]\"}", SAME_AS_LAST)
);
static const BIProto12 background_ = BIProto12(
	"Background", "children",
	BIField(SFBOOL,	"set_bind", UNREADABLE),
	BIField(MFFLOAT,	"groundAngle []", 0),
	BIField(MFCOLOR,	"groundColor []", 0),
	BIField(MFFLOAT,	"skyAngle []", 0),
	BIField(MFCOLOR,	"skyColor 0 0 0", 0),
	BIField(MFSTRING,	"backUrl []", 0),
	BIField(MFSTRING,	"bottomUrl []", 0),
	BIField(MFSTRING,	"frontUrl []", 0),
	BIField(MFSTRING,	"leftUrl []", 0),
	BIField(MFSTRING,	"rightUrl []", 0),
	BIField(MFSTRING,	"topUrl []", 0),
	BIField(SFBOOL,	"isBound", UNWRITEABLE)
);
static const BIProto17 seamless_ = BIProto17(
	"Seamless", "children",
	BIField(SFFLOAT,	"version 0", UNREADABLE | UNWRITEABLE | DISABLE),
	BIField(SFVEC3F,	"scale 1 1 1", 0),
	BIField(SFROTATION,	"rotation 0 1 0 0", 0),
	BIField(SFVEC3F,	"translation 0 0 0", 0),
	BIField(SFBOOL,	"track TRUE", DISABLE),
	BIField(SFBOOL,	"xRay FALSE", 0),
	BIField(SFBOOL,	"autoIn FALSE", 0),
	BIField(SFBOOL,	"in FALSE", DISABLE),
	BIField(SFBOOL,	"animCP FALSE", 0),
	BIField(SFBOOL,	"buildNormals FALSE", DISABLE),
	BIField(SFBOOL,	"surfaceNormals FALSE", 0),
	BIField(SFBOOL,	"ignoreWireFrame FALSE", 0),
	BIField(MFINT32,"allot []", DISABLE), //0 1 0 27, 1 1 27 27 partIndex, number of groups, point index, num of points (point index & num of points = group)
	BIField(MFINT32,"stitchIndex []", DISABLE), //destination, source
	BIField(SFNODE,	"effect NULL", 0,	INIT_PROMPT_NODE(promptNode_seamless_effect)),
	BIField(MFNODE,	"skeleton []", BLOCK_USE,	INIT_PROMPT_NODE(promptNode_seamless_skeleton)),
	BIField(MFNODE,	"build []", 0,	INIT_PROMPT_NODE(promptNode_seamless_build))
	//BIField(SFNODE,	"mirrorGroup NULL", 0,	INIT_PROMPT_NODE(promptNode_seamless_mirrorGroup))
);
static const BIProto3 groupBuild_ = BIProto3(
	"GroupBuild", "children",
	BIField(SFCOLOR,"marker 1 1 1", 0),
	BIField(SFNODE,	"setup NULL", 0,	INIT_PROMPT_NODE(promptNode_groupBuild_setup)),
	BIField(MFNODE,	"build []", 0,		INIT_PROMPT_NODE(promptNode_seamless_build))
);
static const BIProto3 sCGroupBuild_ = BIProto3(//Script Controlled GroupBuild
	"GroupBuild", "children",
	BIField(SFCOLOR,"marker 1 1 1", 0),
	BIField(SFNODE,	"setup NULL", 0,	INIT_PROMPT_NODE(promptNode_groupBuild_setup)),
	BIField(MFNODE,	"build []", 0,		INIT_PROMPT_NODE(promptNode_seamless_build))
);
static const BIProto6 joinCloseVertex_ = BIProto6(
	"JoinCloseVertex", "build",
	BIField(SFBOOL,	"srcColor TRUE", 0),
	BIField(SFBOOL,	"dstColor TRUE", 0),
	BIField(SFFLOAT,	"joinRange .001", 0),
	BIField(MFNODE,	"source []", BLOCK_NEW | BLUE_DOT,	INIT_PROMPT_NODE(promptNode_joinCloseVertex_source)),
	BIField(MFNODE,	"part []", BLOCK_NEW | RED_DOT,	INIT_PROMPT_NODE(promptNode_joinCloseVertex_part)),
	BIField(SFNODE,	"setup NULL", 0,	INIT_PROMPT_NODE(promptNode_joinCloseVertex_setup))
);
static const BIProto5 centerJoint_ = BIProto5(
	"CenterJoint", "build",
	BIField(SFINT32,	"fromVertex 0{rangeFrom 0}", 0),
	BIField(SFINT32,	"numOf 0{rangeFrom 0}", 0),
	BIField(SFVEC3F,	"translation 0 0 0", 0),
	BIField(SFNODE,	"part NULL", BLOCK_NEW,	INIT_PROMPT_NODE(promptNode_centerJoint_part)),
	BIField(SFNODE,	"setup NULL", 0,	INIT_PROMPT_NODE(promptNode_centerJoint_setup))
);
static const BIProto3 delPoly_ = BIProto3(
	"DelPoly", "build",
	BIField(MFINT32,	"poly []", 0),
	BIField(SFNODE,	"part NULL", BLOCK_NEW,	INIT_PROMPT_NODE(promptNode_delPoly_part)),
	BIField(SFNODE,	"setup NULL", 0,	INIT_PROMPT_NODE(promptNode_delPoly_setup))
);
static const BIProto10 joinVertex_ = BIProto10(
	"JoinVertex", "build",
	BIField(SFBOOL,	"move FALSE", 0),
	BIField(SFBOOL,	"srcColor TRUE", 0),
	BIField(SFBOOL,	"dstColor TRUE", 0),
	BIField(MFINT32,	"srcCoord []{beginGroup TRUE}", 0),
	BIField(MFINT32,	"dstCoord []", 0),
	BIField(SFBOOL,	"in FALSE", 0),
	BIField(SFFLOAT,	"range .001", 0),
	BIField(SFNODE,	"source NULL", BLOCK_NEW | BLUE_DOT,	INIT_PROMPT_NODE(promptNode_joinVertex_source)),
	BIField(SFNODE,	"part NULL", BLOCK_NEW | RED_DOT,	INIT_PROMPT_NODE(promptNode_joinVertex_part)),
	BIField(SFNODE,	"setup NULL", 0,	INIT_PROMPT_NODE(promptNode_joinVertex_setup))
);
static const BIProto14 copyPart_ = BIProto14(
	"CopyPart", "build",
	BIField(SFBOOL,	"mirrorX FALSE", 0),
	BIField(SFBOOL,	"mirrorY FALSE", 0),
	BIField(SFBOOL,	"mirrorZ FALSE", 0),
	BIField(SFBOOL,	"add FALSE", 0),
	BIField(SFBOOL,	"insideOut FALSE", 0),
	BIField(SFBOOL,	"staticTransform FALSE", DISABLE),
	BIField(SFBOOL,	"copyTransform FALSE", DISABLE),
	BIField(SFBOOL,	"resetTransform FALSE", DISABLE),
	BIField(SFVEC3F,	"scale 1 1 1", 0),
	BIField(SFROTATION,	"rotation 0 0 1 0", 0),
	BIField(SFVEC3F,	"translation 0 0 0", 0),
	BIField(MFNODE,	"source []", BLOCK_NEW | BLUE_DOT,	INIT_PROMPT_NODE(promptNode_copyPart_source)),
	BIField(MFNODE,	"dest []", BLOCK_NEW | RED_DOT,	INIT_PROMPT_NODE(promptNode_copyPart_dest)),
	BIField(SFNODE,	"setup NULL", 0,	INIT_PROMPT_NODE(promptNode_copyPart_setup))
);
static const BIProto14 viewpoint_ = BIProto14(
	"Viewpoint", "children",
	BIField(SFBOOL,	"set_bind", UNREADABLE),
	BIField(SFBOOL,	"perspective TRUE", DO_NOT_OUTPUT),
	BIField(SFFLOAT,	"zoom 1", DO_NOT_OUTPUT),
	BIField(SFFLOAT,	"fieldOfView .785398", 0),
	BIField(SFROTATION,	"orientation 0 0 1 0", 0),
	BIField(SFVEC3F,	"position 0 0 10", 0),
	BIField(SFBOOL,	"track TRUE", DISABLE),
	BIField(SFBOOL,	"jump TRUE", DISABLE),
	BIField(SFSTRING,	"description \"\"", UNREADABLE | UNWRITEABLE | DISABLE),
	BIField(SFBOOL,	"capture FALSE", PUSH),
	BIField(SFBOOL,	"animCapture FALSE", PUSH),
	BIField(SFBOOL,	"animDelete FALSE", PUSH),
	BIField(SFTIME,	"bindTime", UNWRITEABLE),
	BIField(SFBOOL,	"isBound", UNWRITEABLE)
);
static const BIProto14 transferBCyl_ = BIProto14(
	"TransferBCyl", "build",
	BIField(SFVEC3F,	"location 0 0 0", 0),
	BIField(SFROTATION,	"orientation 0 0 1 0", 0),
	BIField(SFFLOAT,	"height .5{rangeFrom 0}", 0),
	BIField(SFFLOAT,	"radius .25{rangeFrom 0}", 0),
	BIField(SFBOOL,		"indexPosition TRUE", 0),
	BIField(SFBOOL,		"insert FALSE", 0),
	BIField(SFFLOAT,	"groupingHeight 0{rangeFrom 0}", 0),
	BIField(SFINT32,	"numOfGroups 3{rangeFrom 0}", 0),
	BIField(SFINT32,	"group -1{rangeFrom -1}", 0),
	BIField(MFINT32,	"vertex []", 0),
	BIField(SFBOOL,	"in TRUE", 0),
	BIField(SFNODE,	"source NULL", BLOCK_NEW | BLUE_DOT,	INIT_PROMPT_NODE(promptNode_transferBCyl_source)),
	BIField(SFNODE,	"dest NULL", BLOCK_NEW | RED_DOT,	INIT_PROMPT_NODE(promptNode_transferBCyl_dest)),
	BIField(SFNODE,	"setup NULL", 0,	INIT_PROMPT_NODE(promptNode_transferBCyl_setup))
);
static const BIProto6 transferOwnership_ = BIProto6(
	"TransferOwnership", "build",
	BIField(MFINT32,	"numOf []{beginGroup TRUE initElVal \"0\"}", SAME_AS_LAST),
	BIField(MFINT32,	"qty []{initElVal \"1\"}", SAME_AS_LAST),
	BIField(SFBOOL,	"beginSP FALSE", 0,	tip_rectangleBuild_beginSP),
	BIField(SFNODE,	"source NULL", BLOCK_NEW | BLUE_DOT,	INIT_PROMPT_NODE(promptNode_transferOwnership_source)),
	BIField(MFNODE,	"dest NULL", BLOCK_NEW | RED_DOT,	INIT_PROMPT_NODE(promptNode_transferVertex_dest)),
	BIField(SFNODE,	"setup NULL", 0,	INIT_PROMPT_NODE(promptNode_transferVertex_setup))
);
static const BIProto7 transferVertex_ = BIProto7(
	"TransferVertex", "build",
	BIField(SFINT32,	"fromVertex 0{rangeFrom 0}", 0),
	BIField(SFINT32,	"numOf -1{rangeFrom -1}", 0),
	BIField(SFINT32,	"qty 1", 0),
	BIField(SFBOOL,	"beginSP FALSE", 0,	tip_rectangleBuild_beginSP),
	BIField(SFNODE,	"source NULL", BLOCK_NEW | BLUE_DOT,	INIT_PROMPT_NODE(promptNode_transferVertex_source)),
	BIField(SFNODE,	"dest NULL", BLOCK_NEW | RED_DOT,	INIT_PROMPT_NODE(promptNode_transferVertex_dest)),
	BIField(SFNODE,	"setup NULL", 0,	INIT_PROMPT_NODE(promptNode_transferVertex_setup))
);
static const BIProto1 coordinate_ = BIProto1(
	"Coordinate", "coord",
	BIField(MFVEC3F,	"point []", 0)
);
static const BIProto1 normal_ = BIProto1(
	"Normal", "normal",
	BIField(MFVEC3F,	"vector []", 0)
);
static const BIProto1 color_ = BIProto1(
	"Color", "color",
	BIField(MFCOLOR,	"color []", 0)
);
static const BIProto1 textureCoordinate_ = BIProto1(
	"TextureCoordinate", "texCoord",
	BIField(MFVEC2F,	"point []", 0)
);
static const BIProto6 sineTug_ = BIProto6(
	"SineTug", "generator",
	BIField(MFROTATION,	"orientation []{beginGroup TRUE}", SAME_AS_LAST),
	BIField(MFVEC3F,	"location []", SAME_AS_LAST),
	BIField(MFFLOAT,	"radius []{initElVal \"1\"}", SAME_AS_LAST),
	BIField(MFFLOAT,	"distance []", SAME_AS_LAST),
	BIField(MFFLOAT,	"decay []{rangeFrom 0 initElVal \"1\"}", SAME_AS_LAST),
	BIField(MFFLOAT,	"xScale []{initElVal \"1\"}", SAME_AS_LAST)
);
static const BIProto4 simulSineTug_ = BIProto4(
	"SimulSineTug", "generator",
	BIField(SFVEC3F,	"location 0 0 0", 0),
	BIField(MFROTATION,	"orientation []{beginGroup TRUE}", SAME_AS_LAST),
	BIField(MFFLOAT,	"distance []", SAME_AS_LAST),
	BIField(SFFLOAT,	"radius 0", 0)
);
static const BIProto3 genPoly_ = BIProto3(
	"GenPoly", "build",
	BIField(MFINT32,	"vertex []{initElVal \"[0 0 0]\" incElVal \"[0 0 0]\"}", 0),
	BIField(SFNODE,	"part NULL", BLOCK_NEW,	INIT_PROMPT_NODE(promptNode_genPoly_part)),
	BIField(SFNODE,	"setup NULL", 0,	INIT_PROMPT_NODE(promptNode_genPoly_setup))
);
static const BIProto3 script_ = BIProto3(
	"Script", "children",
	BIField(MFSTRING,	"url []", 0),
	BIField(SFBOOL,	"directOutput FALSE", UNREADABLE | UNWRITEABLE),
	BIField(SFBOOL,	"mustEvaluate FALSE", UNREADABLE | UNWRITEABLE)
);
static const BIProto10 branchJoinNegX_ = BIProto10(
	"BranchJoinNegX", "build",
	BIField(SFROTATION,	"orientation 0 1 0 0", 0),
	BIField(SFVEC3F,	"location 0 0 0", 0),
	BIField(SFBOOL,	"descendants FALSE", 0),
	BIField(SFFLOAT,"decay 0", 0),
	BIField(SFBOOL,	"sawn FALSE", 0),
	BIField(SFBOOL,	"backward FALSE", 0),
	BIField(SFINT32,"numOf -1{rangeFrom -1}", 0),
	BIField(MFNODE,	"part []", BLOCK_NEW,	INIT_PROMPT_NODE(promptNode_branchJoinNegX_part)),
	BIField(SFNODE,	"setup NULL", 0,	INIT_PROMPT_NODE(promptNode_branchJoinNegX_setup)),
	BIField(SFNODE,	"surface NULL", 0,	INIT_PROMPT_NODE(promptNode_branchJoinNegX_surface))
);
static const BIProto10 branchJoinPosX_ = BIProto10(
	"BranchJoinPosX", "build",
	BIField(SFROTATION,	"orientation 0 1 0 0", 0),
	BIField(SFVEC3F,	"location 0 0 0", 0),
	BIField(SFBOOL,	"descendants FALSE", 0),
	BIField(SFFLOAT,"decay 0", 0),
	BIField(SFBOOL,	"sawn FALSE", 0),
	BIField(SFBOOL,	"backward FALSE", 0),
	BIField(SFINT32,"numOf -1{rangeFrom -1}", 0),
	BIField(MFNODE,	"part []", BLOCK_NEW,	INIT_PROMPT_NODE(promptNode_branchJoinPosX_part)),
	BIField(SFNODE,	"setup NULL", 0,	INIT_PROMPT_NODE(promptNode_branchJoinPosX_setup)),
	BIField(SFNODE,	"surface NULL", 0,	INIT_PROMPT_NODE(promptNode_branchJoinPosX_surface))
);
static const BIProto10 branchJoinNegY_ = BIProto10(
	"BranchJoinNegY", "build",
	BIField(SFROTATION,	"orientation 0 1 0 0", 0),
	BIField(SFVEC3F,	"location 0 0 0", 0),
	BIField(SFBOOL,	"descendants FALSE", 0),
	BIField(SFFLOAT,"decay 0", 0),
	BIField(SFBOOL,	"sawn FALSE", 0),
	BIField(SFBOOL,	"backward FALSE", 0),
	BIField(SFINT32,"numOf -1{rangeFrom -1}", 0),
	BIField(MFNODE,	"part []", BLOCK_NEW,	INIT_PROMPT_NODE(promptNode_branchJoinNegY_part)),
	BIField(SFNODE,	"setup NULL", 0,	INIT_PROMPT_NODE(promptNode_branchJoinNegY_setup)),
	BIField(SFNODE,	"surface NULL", 0,	INIT_PROMPT_NODE(promptNode_branchJoinNegY_surface))
);
static const BIProto10 branchJoinPosY_ = BIProto10(
	"BranchJoinPosY", "build",
	BIField(SFROTATION,	"orientation 0 1 0 0", 0),
	BIField(SFVEC3F,	"location 0 0 0", 0),
	BIField(SFBOOL,	"descendants FALSE", 0),
	BIField(SFFLOAT,"decay 0", 0),
	BIField(SFBOOL,	"sawn FALSE", 0),
	BIField(SFBOOL,	"backward FALSE", 0),
	BIField(SFINT32,"numOf -1{rangeFrom -1}", 0),
	BIField(MFNODE,	"part []", BLOCK_NEW,	INIT_PROMPT_NODE(promptNode_branchJoinPosY_part)),
	BIField(SFNODE,	"setup NULL", 0,	INIT_PROMPT_NODE(promptNode_branchJoinPosY_setup)),
	BIField(SFNODE,	"surface NULL", 0,	INIT_PROMPT_NODE(promptNode_branchJoinPosY_surface))
);
static const BIProto10 branchJoinNegZ_ = BIProto10(
	"BranchJoinNegZ", "build",
	BIField(SFROTATION,	"orientation 0 1 0 0", 0),
	BIField(SFVEC3F,	"location 0 0 0", 0),
	BIField(SFBOOL,	"descendants FALSE", 0),
	BIField(SFFLOAT,"decay 0", 0),
	BIField(SFBOOL,	"sawn FALSE", 0),
	BIField(SFBOOL,	"backward FALSE", 0),
	BIField(SFINT32,"numOf -1{rangeFrom -1}", 0),
	BIField(MFNODE,	"part []", BLOCK_NEW,	INIT_PROMPT_NODE(promptNode_branchJoinNegZ_part)),
	BIField(SFNODE,	"setup NULL", 0,	INIT_PROMPT_NODE(promptNode_branchJoinNegZ_setup)),
	BIField(SFNODE,	"surface NULL", 0,	INIT_PROMPT_NODE(promptNode_branchJoinNegZ_surface))
);
static const BIProto10 branchJoinPosZ_ = BIProto10(
	"BranchJoinPosZ", "build",
	BIField(SFROTATION,	"orientation 0 1 0 0", 0),
	BIField(SFVEC3F,	"location 0 0 0", 0),
	BIField(SFBOOL,	"descendants FALSE", 0),
	BIField(SFFLOAT,"decay 0", 0),
	BIField(SFBOOL,	"sawn FALSE", 0),
	BIField(SFBOOL,	"backward FALSE", 0),
	BIField(SFINT32,"numOf -1{rangeFrom -1}", 0),
	BIField(MFNODE,	"part []", BLOCK_NEW,	INIT_PROMPT_NODE(promptNode_branchJoinPosZ_part)),
	BIField(SFNODE,	"setup NULL", 0,	INIT_PROMPT_NODE(promptNode_branchJoinPosZ_setup)),
	BIField(SFNODE,	"surface NULL", 0,	INIT_PROMPT_NODE(promptNode_branchJoinPosZ_surface))
);
static const BIProto11 projectOntoTri_ = BIProto11(
	"ProjectOntoTri", "build",
	BIField(SFROTATION,	"orientation 0 1 0 0", 0),
	BIField(SFVEC3F,	"location 0 0 0", 0),
	BIField(SFBOOL,	"descendants FALSE", 0),
	BIField(SFFLOAT,"decay 0", 0),
	BIField(SFBOOL,	"sawn FALSE", 0),
	BIField(SFBOOL,	"backward FALSE", 0),
	BIField(SFINT32,"numOf -1{rangeFrom -1}", 0),
	BIField(MFNODE,	"part []", BLOCK_NEW,	INIT_PROMPT_NODE(promptNode_stem_part)),
	BIField(SFNODE,	"surface NULL", 0,	INIT_PROMPT_NODE(promptNode_stem_part)),
	BIField(SFNODE,	"feedback NULL", 0,	INIT_PROMPT_NODE(promptNode_projectOntoTri_feedback)),
	BIField(SFNODE,	"setup NULL", 0,	INIT_PROMPT_NODE(promptNode_stem_setup))
);
static const BIProto11 colorSweep_ = BIProto11(
	"ColorSweep", "build",
	BIField(SFROTATION,	"orientation 0 1 0 0", 0),
	BIField(SFVEC3F,	"location 0 0 0", 0),
	BIField(MFCOLOR,	"color []{beginGroup TRUE initElVal \"[1 1 1 1 1 1]\"}", SAME_AS_LAST),
	BIField(MFFLOAT,	"distance []{startAtIndex 1 initElVal \"1\"}", SAME_AS_LAST),
	BIField(MFFLOAT,	"opacity []{initElVal \"[1 1]\"}", SAME_AS_LAST),
	BIField(SFBOOL,	"spherical FALSE", 0),
	BIField(SFBOOL,	"mirror FALSE", DISABLE),
	BIField(SFBOOL,	"cylindrical FALSE", DISABLE),
	BIField(SFBOOL,	"straight FALSE", 0),
	BIField(MFNODE,	"part []", BLOCK_NEW,	INIT_PROMPT_NODE(promptNode_colorSweep_part)),
	BIField(SFNODE,	"setup NULL", 0,	INIT_PROMPT_NODE(promptNode_colorSweep_setup))
);
static const BIProto7 evolveCoord_ = BIProto7(
	"EvolveCoord", "build",
	BIField(SFVEC3F,	"location 0 0 0", 0),
	BIField(SFROTATION,	"orientation 0 0 1 0", 0),
	BIField(MFVEC2F,	"point []{beginGroup TRUE initElVal \"[0 0 1 2]\"}", SAME_AS_LAST),
	BIField(MFFLOAT,	"tangent []{initElVal \"[.995382 1.29955]\"}", SAME_AS_LAST),
	BIField(MFNODE,	"source []", BLOCK_NEW | BLUE_DOT,	INIT_PROMPT_NODE(promptNode_evolveCoord_source)),
	BIField(MFNODE,	"part []", BLOCK_NEW | RED_DOT,	INIT_PROMPT_NODE(promptNode_evolveCoord_part)),
	BIField(SFNODE,	"setup NULL", 0,	INIT_PROMPT_NODE(promptNode_evolveCoord_setup))
);
static const BIProto1 function_ = BIProto1(
	"Function", 0,
	BIField(MFNODE,	"codeLine []", BLOCK_USE,	INIT_PROMPT_NODE(promptNode_function_codeLine))
);
static const BIProto1 memberFunction_ = BIProto1(
	"MemberFunction", 0,
	BIField(MFNODE,	"codeLine []", BLOCK_USE,	INIT_PROMPT_NODE(promptNode_memberFunction_codeLine))
);
static const BIProto1 codeLine_ = BIProto1(
	"CodeLine", 0,
	BIField(MFNODE,	"codeLine []", BLOCK_USE,	INIT_PROMPT_NODE(promptNode_codeLine_codeLine))
);
static const BIProto6 morphPart_ = BIProto6(
	"MorphPart", 0,
	BIField(SFFLOAT,	"set_fraction", UNREADABLE),
	BIField(MFVEC3F,	"value_changed", UNWRITEABLE),
	BIField(MFFLOAT,	"key []{beginGroup TRUE}", 0),
	BIField(MFFLOAT,	"weight []", 0),
	BIField(SFNODE,	"part NULL", BLOCK_NEW,	INIT_PROMPT_NODE(promptNode_morphPart_part)),
	BIField(MFNODE,	"bank []", 0,	INIT_PROMPT_NODE(promptNode_morphPart_bank))
);
static const BIProto2 morpher_ = BIProto2(
	"Morpher", 0,
	BIField(SFFLOAT,	"weight", UNREADABLE),
	BIField(MFVEC3F,	"displacement []", 0)
);
static const BIProto4 textureTransform_ = BIProto4(
	"TextureTransform", "textureTransform",
	BIField(SFVEC2F,	"center 0 0", 0),
	BIField(SFVEC2F,	"scale 1 1", 0),
	BIField(SFFLOAT,	"rotation 0", 0),
	BIField(SFVEC2F,	"translation 0 0", 0)
);
static const BIProto2 protoDeclare_ = BIProto2(
	"ProtoDeclare", 0,
	BIField(SFNODE,	"protoInterface NULL", 0),
	BIField(SFNODE,	"protoBody NULL", 0)
);
static const BIProto0 protoInterface_ = BIProto0(
	"ProtoInterface", 0
);
static const BIProto1 protoBody_ = BIProto1(
	"ProtoBody", 0,
	BIField(MFNODE,	"children []", 0,	INIT_PROMPT_NODE(promptNode_scene_children))
);
static const BIProto0 protoInstance_ = BIProto0(
	"ProtoInstance", 0
);

static const BIProto1 movie_ = BIProto1(
	"Movie", "children",
	BIField(MFNODE,	"clip []", 0,	INIT_PROMPT_NODE(promptNode_movie_clip))
);

static const BIProto2 video_ = BIProto2(
	"Video", "clip",
	BIField(SFBOOL,	"open FALSE", DO_NOT_SAVE | PUSH | DO_NOT_DISPLY_NAME, "open png file sequence"),
	BIField(SFSTRING,	"file \"\"", DISABLE)
);

static const BIProto11 timeSensor_ = BIProto11(
	"TimeSensor", "children",
	BIField(SFTIME,	"cycleInterval 1", 0),
	BIField(SFBOOL,	"enabled TRUE", 0),
	BIField(SFBOOL,	"loop FALSE", 0),
	BIField(SFTIME,	"startTime 0", 0),
	BIField(SFTIME,	"stopTime 0", 0),
	BIField(SFTIME,	"cycleTime", UNWRITEABLE),
	BIField(SFFLOAT,	"fraction_changed", UNWRITEABLE),
	BIField(SFBOOL,	"isActive", UNWRITEABLE),
	BIField(SFTIME,	"time", UNWRITEABLE),
	BIField(MFNODE,	"children []", 0),
	BIField(SFNODE,	"on_fraction_changed NULL", 0,	INIT_PROMPT_NODE(promptNode_timeSensor_on_fraction_changed))
);
static const BIProto4 orientationInterpolator_ = BIProto4(
	"OrientationInterpolator", "children",
	BIField(SFFLOAT,	"set_fraction", UNREADABLE),
	BIField(MFFLOAT,	"key []{beginGroup TRUE}", 0),
	BIField(MFROTATION,	"keyValue []", 0),
	BIField(SFROTATION,	"value_changed", UNWRITEABLE)
);
static const BIProto4 positionInterpolator_ = BIProto4(
	"PositionInterpolator", "children",
	BIField(SFFLOAT,	"set_fraction", UNREADABLE),
	BIField(MFFLOAT,	"key []{beginGroup TRUE}", 0),
	BIField(MFVEC3F,	"keyValue []", 0),
	BIField(SFVEC3F,	"value_changed", UNWRITEABLE)
);
static const BIProto5 nurbsTrack_ = BIProto5(
	"NurbsTrack", "children",
	BIField(SFFLOAT,	"frame", UNREADABLE),
	BIField(MFFLOAT,	"key []{beginGroup TRUE}", 0),
	BIField(MFVEC3F,	"keyValue []{beginGroup TRUE}", 0),
	BIField(MFINT32,	"dim []{beginGroup TRUE}", 0),
	BIField(SFVEC3F,	"pos", UNWRITEABLE)
);

/*
static const BIProto4 animCurver_ = BIProto4(
	"AnimCurver", "children",
	BIField(MFVEC3F,	"point []", 0),
	BIField(SFFLOAT, "curveDistance 0", 0),
	BIField(SFNODE,	"source NULL", BLOCK_NEW | BLUE_DOT,	INIT_PROMPT_NODE(promptNode_joinVertex_source)),
	BIField(SFNODE,	"dest NULL", BLOCK_NEW | BLUE_DOT,	INIT_PROMPT_NODE(promptNode_joinVertex_source))
);
*/

static const BIProto4 scalarInterpolator_ = BIProto4(
	"ScalarInterpolator", "children",
	BIField(SFFLOAT,	"set_fraction", UNREADABLE),
	BIField(MFFLOAT,	"key []{beginGroup TRUE}", 0),
	BIField(MFFLOAT,	"keyValue []", 0),
	BIField(SFFLOAT,	"value_changed", UNWRITEABLE)
);
static const BIProto2 posCompensator_ = BIProto2(
	"PosCompensator", "children",
	BIField(MFVEC2F,	"moment []{beginGroup TRUE}", 0),
	BIField(MFVEC3F,	"relPos []", 0)
);
static const BIProto8 collision_ = BIProto8(
	"Collision", "children",
	BIField(MFNODE,	"addChildren", UNREADABLE),
	BIField(MFNODE,	"removeChildren", UNREADABLE),
	BIField(MFNODE,	"children []", 0,	INIT_PROMPT_NODE(promptNode_scene_children)),
	BIField(SFBOOL,	"collide TRUE", 0),
	BIField(SFVEC3F,	"bboxCenter 0 0 0", UNREADABLE | UNWRITEABLE),
	BIField(SFVEC3F,	"bboxSize -1 -1 -1", UNREADABLE | UNWRITEABLE),
	BIField(SFNODE,	"proxy NULL", UNREADABLE | UNWRITEABLE),
	BIField(SFTIME,	"collideTime", UNWRITEABLE)
);
static const BIProto10 transform_ = BIProto10(
	"Transform", "children",
	BIField(MFNODE,	"addChildren", UNREADABLE),
	BIField(MFNODE,	"removeChildren", UNREADABLE),
	BIField(SFVEC3F,	"center 0 0 0", 0),
	BIField(SFROTATION,	"scaleOrientation 0 0 1 0", 0),
	BIField(SFVEC3F,	"scale 1 1 1", 0),
	BIField(SFROTATION,	"rotation 0 0 1 0", 0),
	BIField(SFVEC3F,	"translation 0 0 0", 0),
	BIField(SFVEC3F,	"bboxCenter 0 0 0", UNREADABLE | UNWRITEABLE),
	BIField(SFVEC3F,	"bboxSize -1 -1 -1", UNREADABLE | UNWRITEABLE),
	BIField(MFNODE,	"children []", 0,	INIT_PROMPT_NODE(promptNode_scene_children))
);
static const BIProto4 colorInterpolator_ = BIProto4(
	"ColorInterpolator", "children",
	BIField(SFFLOAT,	"set_fraction", UNREADABLE),
	BIField(MFFLOAT,	"key []{beginGroup TRUE}", 0),
	BIField(MFCOLOR,	"keyValue []", 0),
	BIField(SFCOLOR,	"value_changed", UNWRITEABLE)
);
static const BIProto4 coordinateInterpolator_ = BIProto4(
	"CoordinateInterpolator", "children",
	BIField(SFFLOAT,	"set_fraction", UNREADABLE),
	BIField(MFFLOAT,	"key []", 0),
	BIField(MFVEC3F,	"keyValue []", 0),
	BIField(MFVEC3F,	"value_changed", UNWRITEABLE)
);
static const BIProto4 normalInterpolator_ = BIProto4(
	"NormalInterpolator", "children",
	BIField(SFFLOAT,	"set_fraction", UNREADABLE),
	BIField(MFFLOAT,	"key []", 0),
	BIField(MFVEC3F,	"keyValue []", 0),
	BIField(MFVEC3F,	"value_changed", UNWRITEABLE)
);
static const BIProto3 modelSequence_ = BIProto3(
	"ModelSequence", "modelSequence",
	BIField(SFFLOAT,	"frame", UNREADABLE),
	BIField(MFFLOAT,	"key []", 0),
	BIField(MFNODE,	"model []", 0, INIT_PROMPT_NODE(promptNode_modelSequence_seamless))
);
static const BIProto1 modelMorpher_ = BIProto1(
	"ModelMorpher", "children",
	BIField(MFNODE,	"sequence []", 0, INIT_PROMPT_NODE(promptNode_modelMorpher_sequence))
);
static const BIProto5 worldInfo_ = BIProto5(
	"WorldInfo", "children",
	BIField(SFSTRING,	"title \"\"", UNREADABLE | UNWRITEABLE),
	BIField(SFSTRING,	"author \"\"", UNREADABLE | UNWRITEABLE | DO_NOT_OUTPUT),
	BIField(SFSTRING,	"website \"\"", UNREADABLE | UNWRITEABLE | DO_NOT_OUTPUT),
	BIField(MFSTRING,	"info []", UNREADABLE | UNWRITEABLE),
	BIField(SFBOOL,		"netSave FALSE", UNREADABLE | UNWRITEABLE | DO_NOT_OUTPUT)
);
static const BIProto9 touchSensor_ = BIProto9(
	"TouchSensor", "children",
	BIField(SFSTRING,	"description \"\"", 0),
	BIField(SFBOOL,	"enabled TRUE", 0),
	BIField(SFVEC3F,	"hitNormal_changed", UNWRITEABLE),
	BIField(SFVEC3F,	"hitPoint_changed", UNWRITEABLE),
	BIField(SFVEC2F,	"hitTexCoord_changed", UNWRITEABLE),
	BIField(SFBOOL,	"isActive", UNWRITEABLE),
	BIField(SFBOOL,	"isOver", UNWRITEABLE),
	BIField(SFTIME,	"touchTime", UNWRITEABLE),
	BIField(SFNODE,	"onTouchTime NULL", DO_NOT_OUTPUT,	INIT_PROMPT_NODE(promptNode_timeSensor_on_fraction_changed))
);
static const BIProto9 cCPath_ = BIProto9(
	"CCPath", 0,
	BIField(SFFLOAT,	"set_fraction", UNREADABLE),
	BIField(SFFLOAT,	"fraction_changed", UNWRITEABLE),
	BIField(SFVEC3F,	"location 0 0 0", 0),
	BIField(SFROTATION,	"orientation 0 1 0 0", 0),
	BIField(MFFLOAT,	"curve []{beginGroup TRUE}", 0),
	BIField(MFFLOAT,	"distance []{startAtIndex 1}", 0),
	BIField(MFFLOAT,	"zAngle []{startAtIndex 1}", 0),
	BIField(SFROTATION,	"rot_changed", UNWRITEABLE),
	BIField(SFVEC3F,	"pos_changed", UNWRITEABLE)
);
static const BIProto4 bspTree_ = BIProto4(
	"BspTree", 0,
	BIField(SFROTATION,	"plane 0 1 0 0", 0),
	BIField(SFNODE,	"front NULL", 0,	INIT_PROMPT_NODE(promptNode_scene_children)),
	BIField(SFNODE,	"overlap NULL", 0,	INIT_PROMPT_NODE(promptNode_scene_children)),
	BIField(SFNODE,	"back NULL", 0,	INIT_PROMPT_NODE(promptNode_scene_children))
);
static const BIProto3 inline_ = BIProto3(
	"Inline", "children",
	BIField(MFSTRING,	"url []", 0),
	BIField(SFVEC3F,	"bboxCenter 0 0 0", UNREADABLE | UNWRITEABLE),
	BIField(SFVEC3F,	"bboxSize -1 -1 -1", UNREADABLE | UNWRITEABLE)
);
static const BIProto5 fog_ = BIProto5(
	"Fog", "children",
	BIField(SFCOLOR,	"color 1 1 1", 0),
	BIField(SFSTRING,	"fogType \"LINEAR\"", 0),
	BIField(SFFLOAT,	"visibilityRange 0", 0),
	BIField(SFBOOL,	"set_bind", UNREADABLE),
	BIField(SFBOOL,	"isBound", UNWRITEABLE)
);
static const BIProto7 fog2_ = BIProto7(
	"Fog2", "children",
	BIField(SFCOLOR,	"color 1 1 1", 0),
	BIField(SFSTRING,	"fogType \"LINEAR\"", 0),
	BIField(SFFLOAT,	"visibilityRange 0", 0),
	BIField(SFFLOAT,	"visibilityStart 0", 0),
	BIField(SFFLOAT,	"density 0", 0),
	BIField(SFBOOL,	"set_bind", UNREADABLE),
	BIField(SFBOOL,	"isBound", UNWRITEABLE)
);
static const BIProto8 anchor_ = BIProto8(
	"Anchor", "children",
	BIField(MFNODE,	"addChildren", UNREADABLE),
	BIField(MFNODE,	"removeChildren", UNREADABLE),
	BIField(MFNODE,	"children []", 0,	INIT_PROMPT_NODE(promptNode_scene_children)),
	BIField(SFSTRING,	"description \"\"", 0),
	BIField(MFSTRING,	"parameter []", 0),
	BIField(MFSTRING,	"url []", 0),
	BIField(SFVEC3F,	"bboxCenter 0 0 0", UNREADABLE | UNWRITEABLE),
	BIField(SFVEC3F,	"bboxSize -1 -1 -1", UNREADABLE | UNWRITEABLE)
);
static const BIProto8 audioClip_ = BIProto8(
	"AudioClip", "children",
	BIField(SFSTRING,	"description \"\"", 0),
	BIField(SFBOOL,	"loop FALSE", 0),
	BIField(SFFLOAT,	"pitch 1", 0),
	BIField(SFTIME,	"startTime 0", 0),
	BIField(SFTIME,	"stopTime 0", 0),
	BIField(MFSTRING,	"url []", 0),
	BIField(SFTIME,	"duration_changed", UNWRITEABLE),
	BIField(SFBOOL,	"isActive", UNWRITEABLE)
);
static const BIProto6 billboard_ = BIProto6(
	"Billboard", "children",
	BIField(MFNODE,	"addChildren", UNREADABLE),
	BIField(MFNODE,	"removeChildren", UNREADABLE),
	BIField(SFVEC3F,	"axisOfRotation 0 1 0", 0),
	BIField(MFNODE,	"children []", 0,	INIT_PROMPT_NODE(promptNode_scene_children)),
	BIField(SFVEC3F,	"bboxCenter 0 0 0", UNREADABLE | UNWRITEABLE),
	BIField(SFVEC3F,	"bboxSize -1 -1 -1", UNREADABLE | UNWRITEABLE)
);
static const BIProto2 box_ = BIProto2(
	"Box", "geometry",
	BIField(SFVEC3F,	"size 2 2 2", UNREADABLE | UNWRITEABLE),
	BIField(SFBOOL,	"solid TRUE", UNREADABLE | UNWRITEABLE)
);
static const BIProto5 cone_ = BIProto5(
	"Cone", "geometry",
	BIField(SFFLOAT,	"height 2", UNREADABLE | UNWRITEABLE),
	BIField(SFFLOAT,	"bottomRadius 1", UNREADABLE | UNWRITEABLE),
	BIField(SFBOOL,	"side TRUE", UNREADABLE | UNWRITEABLE),
	BIField(SFBOOL,	"bottom TRUE", UNREADABLE | UNWRITEABLE),
	BIField(SFBOOL,	"solid TRUE", UNREADABLE | UNWRITEABLE)
);
static const BIProto6 cylinder_ = BIProto6(
	"Cylinder", "geometry",
	BIField(SFFLOAT,	"height 2", UNREADABLE | UNWRITEABLE),
	BIField(SFFLOAT,	"radius 1", UNREADABLE | UNWRITEABLE),
	BIField(SFBOOL,	"side TRUE", UNREADABLE | UNWRITEABLE),
	BIField(SFBOOL,	"top TRUE", UNREADABLE | UNWRITEABLE),
	BIField(SFBOOL,	"bottom TRUE", UNREADABLE | UNWRITEABLE),
	BIField(SFBOOL,	"solid TRUE", UNREADABLE | UNWRITEABLE)
);
static const BIProto9 cylinderSensor_ = BIProto9(
	"CylinderSensor", "geometry",
	BIField(SFBOOL,	"autoOffset TRUE", 0),
	BIField(SFFLOAT,	"diskAngle .262", 0),
	BIField(SFBOOL,	"enabled TRUE", 0),
	BIField(SFFLOAT,	"maxAngle -1", 0),
	BIField(SFFLOAT,	"minAngle 0", 0),
	BIField(SFFLOAT,	"offset 0", 0),
	BIField(SFBOOL,	"isActive", UNWRITEABLE),
	BIField(SFROTATION,	"rotation_changed", UNWRITEABLE),
	BIField(SFVEC3F,	"trackPoint_changed", UNWRITEABLE)
);
static const BIProto14 elevationGrid_ = BIProto14(
	"ElevationGrid", "geometry",
	BIField(MFFLOAT,	"set_height", UNREADABLE),
	BIField(SFNODE,	"color NULL", 0,	INIT_PROMPT_NODE(promptNode_elevationGrid_color)),
	BIField(SFNODE,	"normal NULL", 0,	INIT_PROMPT_NODE(promptNode_elevationGrid_normal)),
	BIField(SFNODE,	"texCoord NULL", 0,	INIT_PROMPT_NODE(promptNode_elevationGrid_texCoord)),
	BIField(MFFLOAT,	"height []", UNREADABLE | UNWRITEABLE),
	BIField(SFBOOL,	"ccw TRUE", UNREADABLE | UNWRITEABLE),
	BIField(SFBOOL,	"colorPerVertex TRUE", UNREADABLE | UNWRITEABLE),
	BIField(SFFLOAT,	"creaseAngle 0", UNREADABLE | UNWRITEABLE),
	BIField(SFBOOL,	"normalPerVertex TRUE", UNREADABLE | UNWRITEABLE),
	BIField(SFBOOL,	"solid TRUE", UNREADABLE | UNWRITEABLE),
	BIField(SFINT32,	"xDimension 0{rangeFrom 0}", UNREADABLE | UNWRITEABLE),
	BIField(SFFLOAT,	"xSpacing 1", UNREADABLE | UNWRITEABLE),
	BIField(SFINT32,	"zDimension 0{rangeFrom 0}", UNREADABLE | UNWRITEABLE),
	BIField(SFFLOAT,	"zSpacing 1", UNREADABLE | UNWRITEABLE)
);
static const BIProto14 extrusion_ = BIProto14(
	"Extrusion", "geometry",
	BIField(MFVEC2F,	"set_crossSection", UNREADABLE),
	BIField(MFROTATION,	"set_orientation", UNREADABLE),
	BIField(MFVEC2F,	"set_scale", UNREADABLE),
	BIField(MFVEC3F,	"set_spine", UNREADABLE),
	BIField(SFBOOL,	"beginCap TRUE", UNREADABLE | UNWRITEABLE),
	BIField(SFBOOL,	"ccw TRUE", UNREADABLE | UNWRITEABLE),
	BIField(SFBOOL,	"convex TRUE", UNREADABLE | UNWRITEABLE),
	BIField(SFFLOAT,	"creaseAngle 0", UNREADABLE | UNWRITEABLE),
	BIField(MFVEC2F,	"crossSection [1 1 1 -1 -1 -1 -1 1 1 1 ]", UNREADABLE | UNWRITEABLE),
	BIField(SFBOOL,	"endCap TRUE", UNREADABLE | UNWRITEABLE),
	BIField(MFROTATION,	"orientation 0 0 1 0", UNREADABLE | UNWRITEABLE),
	BIField(MFVEC2F,	"scale 1 1", UNREADABLE | UNWRITEABLE),
	BIField(SFBOOL,	"solid TRUE", UNREADABLE | UNWRITEABLE),
	BIField(MFVEC3F,	"spine [0 0 0 0 1 0 ]", UNREADABLE | UNWRITEABLE)
);
static const BIProto9 fontStyle_ = BIProto9(
	"FontStyle", "fontStyle",
	BIField(MFSTRING,	"family \"SERIF\"", UNREADABLE | UNWRITEABLE),
	BIField(SFBOOL,	"horizontal TRUE", UNREADABLE | UNWRITEABLE),
	BIField(MFSTRING,	"justify \"BEGIN\"", UNREADABLE | UNWRITEABLE),
	BIField(SFSTRING,	"language \"\"", UNREADABLE | UNWRITEABLE),
	BIField(SFBOOL,	"leftToRight TRUE", UNREADABLE | UNWRITEABLE),
	BIField(SFFLOAT,	"size 1", UNREADABLE | UNWRITEABLE),
	BIField(SFFLOAT,	"spacing 1", UNREADABLE | UNWRITEABLE),
	BIField(SFSTRING,	"style \"PLAIN\"", UNREADABLE | UNWRITEABLE),
	BIField(SFBOOL,	"topToBottom TRUE", UNREADABLE | UNWRITEABLE)
);
static const BIProto7 indexedLineSet_ = BIProto7(
	"IndexedLineSet", "geometry",
	BIField(MFINT32,	"set_colorIndex", UNREADABLE),
	BIField(MFINT32,	"set_coordIndex", UNREADABLE),
	BIField(SFNODE,	"coord NULL", 0,	INIT_PROMPT_NODE(promptNode_indexedLineSet_coord)),
	BIField(SFNODE,	"color NULL", 0,	INIT_PROMPT_NODE(promptNode_indexedLineSet_color)),
	BIField(MFINT32,	"colorIndex []{rangeFrom -1}", UNREADABLE | UNWRITEABLE),
	BIField(SFBOOL,	"colorPerVertex TRUE", UNREADABLE | UNWRITEABLE),
	BIField(MFINT32,	"coordIndex []{rangeFrom -1}", UNREADABLE | UNWRITEABLE)
);
static const BIProto13 nurbsSurface_ = BIProto13(
	"NurbsSurface", "geometry",
	BIField(MFVEC3F,	"controlPoint []{beginGroup TRUE}", 0),
	BIField(MFFLOAT,	"weight []", 0),
	BIField(MFFLOAT,	"uKnot []{beginGroup TRUE}", DISABLE),
	BIField(MFFLOAT,	"vKnot []{beginGroup TRUE}", DISABLE),
	BIField(SFNODE,		"texCoord NULL", 0,	INIT_PROMPT_NODE(promptNode_indexedFaceSet_texCoord)),
	BIField(SFFLOAT,	"uDimension 0", 0),
	BIField(SFFLOAT,	"vDimension 0", 0),
	BIField(SFFLOAT,	"uOrder 3", 0),
	BIField(SFFLOAT,	"vOrder 3", 0),
	BIField(SFFLOAT,	"uTessellation 0", 0),
	BIField(SFFLOAT,	"vTessellation 0", 0),
	BIField(SFBOOL,		"ccw TRUE", UNREADABLE | UNWRITEABLE),
	BIField(SFBOOL,		"solid TRUE", 0)
);
static const BIProto7 lOD_ = BIProto7(
	"LOD", "children",
	BIField(MFNODE,	"addChildren", UNREADABLE),
	BIField(MFNODE,	"removeChildren", UNREADABLE),
	BIField(SFVEC3F,	"bboxCenter 0 0 0", UNREADABLE | UNWRITEABLE),
	BIField(SFVEC3F,	"bboxSize -1 -1 -1", UNREADABLE | UNWRITEABLE),
	BIField(SFVEC3F,	"center 0 0 0", UNREADABLE | UNWRITEABLE),
	BIField(MFFLOAT,	"range []", UNREADABLE | UNWRITEABLE),
	BIField(MFNODE,	"children []", 0,	INIT_PROMPT_NODE(promptNode_scene_children))
);
static const BIProto9 movieTexture_ = BIProto9(
	"MovieTexture", "texture",
	BIField(SFBOOL,	"loop FALSE", 0),
	BIField(SFFLOAT,	"speed 1", 0),
	BIField(SFTIME,	"startTime 0", 0),
	BIField(SFTIME,	"stopTime 0", 0),
	BIField(MFSTRING,	"url []", 0),
	BIField(SFBOOL,	"repeatS TRUE", UNREADABLE | UNWRITEABLE),
	BIField(SFBOOL,	"repeatT TRUE", UNREADABLE | UNWRITEABLE),
	BIField(SFTIME,	"duration_changed", UNWRITEABLE),
	BIField(SFBOOL,	"isActive", UNWRITEABLE)
);
static const BIProto2 pixelTexture_ = BIProto2(
	"PixelTexture", "texture",
	BIField(SFBOOL,	"repeatS TRUE", UNREADABLE | UNWRITEABLE),
	BIField(SFBOOL,	"repeatT TRUE", UNREADABLE | UNWRITEABLE)
);
static const BIProto8 planeSensor_ = BIProto8(
	"PlaneSensor", "children",
	BIField(SFBOOL,	"autoOffset TRUE", 0),
	BIField(SFBOOL,	"enabled TRUE", 0),
	BIField(SFVEC2F,	"maxPosition -1 -1", 0),
	BIField(SFVEC2F,	"minPosition 0 0", 0),
	BIField(SFVEC3F,	"offset 0 0 0", 0),
	BIField(SFBOOL,	"isActive", UNWRITEABLE),
	BIField(SFVEC3F,	"trackPoint_changed", UNWRITEABLE),
	BIField(SFVEC3F,	"translation_changed", UNWRITEABLE)
);
static const BIProto8 pointLight_ = BIProto8(
	"PointLight", "children",
	BIField(SFFLOAT,	"ambientIntensity 0", 0),
	BIField(SFVEC3F,	"attenuation 1 0 0", 0),
	BIField(SFCOLOR,	"color 1 1 1", 0),
	BIField(SFBOOL,	"global TRUE", 0),
	BIField(SFFLOAT,	"intensity 1", 0),
	BIField(SFVEC3F,	"location 0 0 0", 0),
	BIField(SFBOOL,	"on TRUE", 0),
	BIField(SFFLOAT,	"radius 100", 0)
);
static const BIProto2 pointSet_ = BIProto2(
	"PointSet", "children",
	BIField(SFNODE,	"coord NULL", 0,	INIT_PROMPT_NODE(promptNode_pointSet_coord)),
	BIField(SFNODE,	"color NULL", 0,	INIT_PROMPT_NODE(promptNode_pointSet_color))
);
static const BIProto8 proximitySensor_ = BIProto8(
	"ProximitySensor", "children",
	BIField(SFVEC3F,	"center 0 0 0", 0),
	BIField(SFVEC3F,	"size 0 0 0", 0),
	BIField(SFBOOL,	"enabled TRUE", 0),
	BIField(SFBOOL,	"isActive", UNWRITEABLE),
	BIField(SFVEC3F,	"position_changed", UNWRITEABLE),
	BIField(SFROTATION,	"orientation_changed", UNWRITEABLE),
	BIField(SFTIME,	"enterTime", UNWRITEABLE),
	BIField(SFTIME,	"exitTime", UNWRITEABLE)
);
static const BIProto10 sound_ = BIProto10(
	"Sound", "children",
	BIField(SFVEC3F,	"direction 0 0 1", 0),
	BIField(SFFLOAT,	"intensity 1", 0),
	BIField(SFVEC3F,	"location 0 0 0", 0),
	BIField(SFFLOAT,	"maxBack 10", 0),
	BIField(SFFLOAT,	"maxFront 10", 0),
	BIField(SFFLOAT,	"minBack 1", 0),
	BIField(SFFLOAT,	"minFront 1", 0),
	BIField(SFFLOAT,	"priority 0", 0),
	BIField(SFNODE,	"source NULL", 0,	INIT_PROMPT_NODE(promptNode_sound_source)),
	BIField(SFBOOL,	"spatialize TRUE", UNREADABLE | UNWRITEABLE)
);
static const BIProto1 sphere_ = BIProto1(
	"Sphere", "geometry",
	BIField(SFFLOAT,	"radius 1", UNREADABLE | UNWRITEABLE)
);
static const BIProto6 sphereSensor_ = BIProto6(
	"SphereSensor", "children",
	BIField(SFBOOL,	"autoOffset TRUE", 0),
	BIField(SFBOOL,	"enabled TRUE", 0),
	BIField(SFROTATION,	"offset 0 1 0 0", 0),
	BIField(SFBOOL,	"isActive", UNWRITEABLE),
	BIField(SFROTATION,	"rotation_changed", UNWRITEABLE),
	BIField(SFVEC3F,	"trackPoint_changed", UNWRITEABLE)
);
static const BIProto11 spotLight_ = BIProto11(
	"SpotLight", "children",
	BIField(SFFLOAT,	"ambientIntensity 0", 0),
	BIField(SFVEC3F,	"attenuation 1 0 0", 0),
	BIField(SFFLOAT,	"beamWidth 1.570796", 0),
	BIField(SFCOLOR,	"color 1 1 1", 0),
	BIField(SFFLOAT,	"cutOffAngle .785398", 0),
	BIField(SFVEC3F,	"direction 0 0 -1", 0),
	BIField(SFBOOL,	"global TRUE", 0),
	BIField(SFFLOAT,	"intensity 1", 0),
	BIField(SFVEC3F,	"location 0 0 0", 0),
	BIField(SFBOOL,	"on TRUE", 0),
	BIField(SFFLOAT,	"radius 100", 0)
);
static const BIProto6 switch_ = BIProto6(
	"Switch", "children",
	BIField(MFNODE,	"addChildren", UNREADABLE),
	BIField(MFNODE,	"removeChildren", UNREADABLE),
	BIField(SFINT32,	"whichChoice -1", 0),
	BIField(SFVEC3F,	"bboxCenter 0 0 0", UNREADABLE | UNWRITEABLE),
	BIField(SFVEC3F,	"bboxSize -1 -1 -1", UNREADABLE | UNWRITEABLE),
	BIField(MFNODE,	"children []", 0,	INIT_PROMPT_NODE(promptNode_scene_children))
);
static const BIProto4 text_ = BIProto4(
	"Text", "geometry",
	BIField(MFSTRING,	"string []", 0),
	BIField(MFFLOAT,	"length []", 0),
	BIField(SFFLOAT,	"maxExtent 0", 0),
	BIField(SFNODE,	"fontStyle NULL", 0,	INIT_PROMPT_NODE(promptNode_text_fontStyle))
);
static const BIProto6 visibilitySensor_ = BIProto6(
	"VisibilitySensor", "children",
	BIField(SFVEC3F,	"center 0 0 0", 0),
	BIField(SFBOOL,	"enabled TRUE", 0),
	BIField(SFVEC3F,	"size 0 0 0", 0),
	BIField(SFTIME,	"enterTime", UNWRITEABLE),
	BIField(SFTIME,	"exitTime", UNWRITEABLE),
	BIField(SFBOOL,	"isActive", UNWRITEABLE)
);
static const BIProto10 textureMapper_ = BIProto10(
	"TextureMapper", "textureMapper",
	BIField(SFROTATION,	"orientation 0 1 0 0", 0),
	BIField(SFBOOL,	"autoSize3d FALSE", 0),
	BIField(SFVEC2F,	"size3d 1 1", 0),
	BIField(SFVEC3F,	"position3d 0 0 0", 0),
	BIField(SFVEC2F,	"size2d 1 1", 0),
	BIField(SFVEC2F,	"position2d 0 0", 0),
	BIField(SFBOOL,	"flat TRUE", 0),
	BIField(SFBOOL,	"spherical FALSE", 0),
	BIField(SFBOOL,	"cylindrical FALSE", 0),
	BIField(SFBOOL,	"clear FALSE", PUSH)
);
static const BIProto23 textureEffect_ = BIProto23(
	"TextureEffect", "effect",
	BIField(SFCOLOR,	"ambient .2 .2 .2", 0),
	BIField(SFCOLOR,	"diffuse .8 .8 .8", 0),
	BIField(SFFLOAT,	"brilliance 1", 0),
	BIField(SFCOLOR,	"emissive 0 0 0", 0),
	BIField(SFFLOAT,	"shininess .2{rangeFrom 0 rangeTo 1}", 0),
	BIField(SFCOLOR,	"specular 0 0 0", 0),
	BIField(SFBOOL,		"backfaceRender FALSE", 0),
	BIField(SFFLOAT,	"phong 0", 0),
	BIField(SFFLOAT,	"phong_size 40", 0),
	BIField(SFFLOAT,	"roughness .05", 0),
	BIField(SFFLOAT,	"crand 0", 0),
	BIField(SFBOOL,		"conserve_energy FALSE", 0),
	BIField(SFCOLOR,	"reflection 0 0 0", 0),
	BIField(SFBOOL,		"fresnel FALSE", 0),
	BIField(SFFLOAT,	"falloff 0", 0),
	BIField(SFFLOAT,	"exponent 1", 0),
	BIField(SFFLOAT,	"metallic 1", 0),
	BIField(SFFLOAT,	"iridescence 0", 0),
	BIField(SFFLOAT,	"thickness 0", 0),
	BIField(SFFLOAT,	"turbulence 0", 0),
	BIField(SFFLOAT,	"transparency 0{rangeFrom 0 rangeTo 1}", 0),
	BIField(SFSTRING,	"url \"\"", 0),
	BIField(MFVEC2F,	"texCoord []", DISABLE_LEN)
);
static const BIProto22 colorEffect_ = BIProto22(
	"ColorEffect", "effect",
	BIField(SFCOLOR,	"ambient .2 .2 .2", 0),
	BIField(SFCOLOR,	"diffuse .8 .8 .8", 0),
	BIField(SFFLOAT,	"brilliance 1", 0),
	BIField(SFCOLOR,	"emissive 0 0 0", 0),
	BIField(SFFLOAT,	"shininess .2{rangeFrom 0 rangeTo 1}", 0),
	BIField(SFCOLOR,	"specular 0 0 0", 0),
	BIField(SFBOOL,		"backfaceRender FALSE", 0),
	BIField(SFFLOAT,	"phong 0", 0),
	BIField(SFFLOAT,	"phong_size 40", 0),
	BIField(SFFLOAT,	"roughness .05", 0),
	BIField(SFFLOAT,	"crand 0", 0),
	BIField(SFBOOL,		"conserve_energy FALSE", 0),
	BIField(SFCOLOR,	"reflection 0 0 0", 0),
	BIField(SFBOOL,		"fresnel FALSE", 0),
	BIField(SFFLOAT,	"falloff 0", 0),
	BIField(SFFLOAT,	"exponent 1", 0),
	BIField(SFFLOAT,	"metallic 1", 0),
	BIField(SFFLOAT,	"iridescence 0", 0),
	BIField(SFFLOAT,	"thickness 0", 0),
	BIField(SFFLOAT,	"turbulence 0", 0),
	BIField(SFFLOAT,	"transparency 0{rangeFrom 0 rangeTo 1}", 0),
	BIField(MFCOLOR,	"color []", DISABLE_LEN)
);
static const BIProto22 monoColorEffect_ = BIProto22(
	"MonoColorEffect", "effect",
	BIField(SFCOLOR,	"ambient .2 .2 .2", 0),
	BIField(SFCOLOR,	"diffuse .8 .8 .8", 0),
	BIField(SFFLOAT,	"brilliance 0", 0),
	BIField(SFCOLOR,	"emissive 0 0 0", 0),
	BIField(SFFLOAT,	"shininess .2{rangeFrom 0 rangeTo 1}", 0),
	BIField(SFCOLOR,	"specular 0 0 0", 0),
	BIField(SFBOOL,		"backfaceRender FALSE", 0),
	BIField(SFFLOAT,	"phong 0", 0),
	BIField(SFFLOAT,	"phong_size 40", 0),
	BIField(SFFLOAT,	"roughness .05", 0),
	BIField(SFFLOAT,	"crand 0", 0),
	BIField(SFBOOL,		"conserve_energy FALSE", 0),
	BIField(SFCOLOR,	"reflection 0 0 0", 0),
	BIField(SFBOOL,		"fresnel FALSE", 0),
	BIField(SFFLOAT,	"falloff 0", 0),
	BIField(SFFLOAT,	"exponent 1", 0),
	BIField(SFFLOAT,	"metallic 1", 0),
	BIField(SFFLOAT,	"iridescence 0", 0),
	BIField(SFFLOAT,	"thickness 0", 0),
	BIField(SFFLOAT,	"turbulence 0", 0),
	BIField(SFFLOAT,	"transparency 0{rangeFrom 0 rangeTo 1}", 0),
	BIField(MFCOLOR,	"color []", DISABLE_LEN)
);
static const BIProto2 exitBuild_ = BIProto2(
	"ExitBuild", "build",
	BIField(SFBOOL,	"enabled TRUE", 0),
	BIField(SFNODE,	"setup NULL", 0,	INIT_PROMPT_NODE(promptNode_exitBuild_setup))
);
static const BIProto3 splitTri_ = BIProto3(
	"SplitTri", "build",
	BIField(MFINT32,	"corner []", 0),
	BIField(SFNODE,	"part NULL", BLOCK_NEW,	INIT_PROMPT_NODE(promptNode_splitTri_part)),
	BIField(SFNODE,	"setup NULL", 0,	INIT_PROMPT_NODE(promptNode_splitTri_setup))
);
static const BIProto3 swapTriDivide_ = BIProto3(
	"SwapTriDivide", "build",
	BIField(MFINT32,	"corner []{initElVal \"[0 0]\" incElVal \"[0 0]\"}", 0),
	BIField(SFNODE,	"part NULL", BLOCK_NEW,	INIT_PROMPT_NODE(promptNode_swapTriDivide_part)),
	BIField(SFNODE,	"setup NULL", 0,	INIT_PROMPT_NODE(promptNode_swapTriDivide_setup))
);
static const BIProto5 placeTexture_ = BIProto5(
	"PlaceTexture", "build",
	BIField(SFVEC2F,	"scale 1 1", 0),
	BIField(SFFLOAT,	"rotation 0", 0),
	BIField(SFVEC2F,	"translation 0 0", 0),
	BIField(MFNODE,	"part []", BLOCK_NEW,	INIT_PROMPT_NODE(promptNode_placeTexture_part)),
	BIField(SFNODE,	"setup NULL", 0,	INIT_PROMPT_NODE(promptNode_placeTexture_setup))
);
static const BIProto18 hAnimHumanoid_ = BIProto18(
	"HAnimHumanoid", "children",
	BIField(SFSTRING,	"version \"\"", 0),
	BIField(SFSTRING,	"name \"\"", 0),
	BIField(MFSTRING,	"info []", 0),
	BIField(SFVEC3F,	"center 0 0 0", 0),
	BIField(SFROTATION,	"scaleOrientation 0 0 1 0", 0),
	BIField(SFVEC3F,	"scale 1 1 1", 0),
	BIField(SFROTATION,	"rotation 0 0 1 0", 0),
	BIField(SFVEC3F,	"translation 0 0 0", 0),
	BIField(SFVEC3F,	"bboxCenter 0 0 0", UNREADABLE | UNWRITEABLE),
	BIField(SFVEC3F,	"bboxSize -1 -1 -1", UNREADABLE | UNWRITEABLE),
	BIField(MFNODE,	"skin []", 0,	INIT_PROMPT_NODE(promptNode_hAnimHumanoid_skin)),
	BIField(SFNODE,	"skinCoord NULL", 0,	INIT_PROMPT_NODE(promptNode_hAnimHumanoid_skinCoord)),
	BIField(SFNODE,	"skinNormal NULL", 0,	INIT_PROMPT_NODE(promptNode_hAnimHumanoid_skinNormal)),
	BIField(MFNODE,	"skeleton []", 0,	INIT_PROMPT_NODE(promptNode_hAnimHumanoid_skeleton)),
	BIField(MFNODE,	"joints []", BLOCK_NEW,	INIT_PROMPT_NODE(promptNode_hAnimHumanoid_joints)),
	BIField(MFNODE,	"segments []", 0),
	BIField(MFNODE,	"sites []", 0),
	BIField(MFNODE,	"viewpoints []", 0)
);
static const BIProto19 hAnimJoint_ = BIProto19(
	"HAnimJoint", "",	//add back "children" when contact does no longer needs it specified for xml
	BIField(MFNODE,	"addChildren", UNREADABLE),
	BIField(MFNODE,	"removeChildren", UNREADABLE),
	BIField(SFSTRING,	"name \"\"", 0),
	BIField(SFVEC3F,	"center 0 0 0", 0),
	BIField(SFROTATION,	"scaleOrientation 0 0 1 0", 0),
	BIField(SFVEC3F,	"scale 1 1 1", 0),
	BIField(SFROTATION,	"rotation 0 0 1 0", 0),
	BIField(SFVEC3F,	"translation 0 0 0", 0),
	BIField(SFVEC3F,	"bboxCenter 0 0 0", UNREADABLE | UNWRITEABLE),
	BIField(SFVEC3F,	"bboxSize -1 -1 -1", UNREADABLE | UNWRITEABLE),
	BIField(MFVEC3F,	"skinCoordLocal []{beginGroup TRUE}", 0),
	BIField(MFINT32,	"skinCoordIndex []{beginGroup TRUE}", 0),
	BIField(MFFLOAT,	"skinCoordWeight []", 0),
	BIField(SFROTATION,	"limitOrientation 0 0 1 0", 0),
	BIField(MFFLOAT,	"llimit []", 0),
	BIField(MFFLOAT,	"stiffness [0 0 0]", 0),
	BIField(MFFLOAT,	"ulimit []", 0),
	BIField(MFNODE,	"children []", BLOCK_USE,	INIT_PROMPT_NODE(promptNode_hAnimJoint_children)),
	BIField(MFNODE,	"displacers []", BLOCK_USE,	INIT_PROMPT_NODE(promptNode_hAnimJoint_displacers))
);
static const BIProto4 hAnimDisplacer_ = BIProto4(
	"HAnimDisplacer", "displacers",
	BIField(MFINT32,	"coordIndex []{beginGroup TRUE}", 0),
	BIField(MFVEC3F,	"displacements []", 0),
	BIField(SFSTRING,	"name \"\"", 0),
	BIField(SFFLOAT,	"weight 0", 0)
);
static const BIProto11 hAnimSegment_ = BIProto11(
	"HAnimSegment", "children",
	BIField(MFNODE,		"addChildren", UNREADABLE),
	BIField(MFNODE,		"removeChildren", UNREADABLE),
	BIField(SFVEC3F,	"centerOfMass 0 0 0", 0),
	BIField(MFNODE,		"children []", 0,	INIT_PROMPT_NODE(promptNode_scene_children)),
	BIField(MFVEC3F,	"coord []", 0),
	BIField(MFNODE,		"displacers []", BLOCK_USE,	INIT_PROMPT_NODE(promptNode_hAnimJoint_displacers)),
	BIField(SFFLOAT,	"mass 0", 0),
	BIField(MFFLOAT,	"momentsOfInertia [0 0 0 0 0 0 0 0 0]", 0),
	BIField(SFSTRING,	"name \"\"", 0),
	BIField(SFVEC3F,	"bboxCenter 0 0 0", UNREADABLE | UNWRITEABLE),
	BIField(SFVEC3F,	"bboxSize -1 -1 -1", UNREADABLE | UNWRITEABLE)
);
static const BIProto11 hAnimSite_ = BIProto11(
	"HAnimSite", "children",
	BIField(MFNODE,		"addChildren", UNREADABLE),
	BIField(MFNODE,		"removeChildren", UNREADABLE),
	BIField(MFNODE,		"children []", 0,	INIT_PROMPT_NODE(promptNode_scene_children)),
	BIField(SFSTRING,	"name \"\"", 0),
	BIField(SFVEC3F,	"center 0 0 0", 0),
	BIField(SFROTATION,	"scaleOrientation 0 0 1 0", 0),
	BIField(SFVEC3F,	"scale 1 1 1", 0),
	BIField(SFROTATION,	"rotation 0 0 1 0", 0),
	BIField(SFVEC3F,	"translation 0 0 0", 0),
	BIField(SFVEC3F,	"bboxCenter 0 0 0", UNREADABLE | UNWRITEABLE),
	BIField(SFVEC3F,	"bboxSize -1 -1 -1", UNREADABLE | UNWRITEABLE)
);
static const BIProto4 booleanFilter_ = BIProto4(
	"BooleanFilter", "children",
	BIField(SFBOOL,	"set_boolean", UNREADABLE),
	BIField(SFBOOL,	"inputFalse", UNWRITEABLE),
	BIField(SFBOOL,	"inputNegate", UNWRITEABLE),
	BIField(SFBOOL,	"inputTrue", UNWRITEABLE)
);
static const BIProto6 booleanSequencer_ = BIProto6(
	"BooleanSequencer", "children",
	BIField(SFBOOL,	"next", UNREADABLE),
	BIField(SFBOOL,	"previous", UNREADABLE),
	BIField(SFFLOAT,	"set_fraction", UNREADABLE),
	BIField(MFFLOAT,	"key []{beginGroup TRUE}", 0),
	BIField(MFBOOL,	"keyValue []", 0),
	BIField(SFBOOL,	"value_changed", UNWRITEABLE)
);
static const BIProto2 booleanToggle_ = BIProto2(
	"BooleanToggle", "children",
	BIField(SFBOOL,	"set_boolean", UNREADABLE),
	BIField(SFBOOL,	"toggle FALSE", 0)
);
static const BIProto2 booleanTrigger_ = BIProto2(
	"BooleanTrigger", "children",
	BIField(SFTIME,	"set_triggerTime", UNREADABLE),
	BIField(SFBOOL,	"triggerTrue", UNWRITEABLE)
);
static const BIProto6 integerSequencer_ = BIProto6(
	"IntegerSequencer", "children",
	BIField(SFBOOL,	"next", UNREADABLE),
	BIField(SFBOOL,	"previous", UNREADABLE),
	BIField(SFFLOAT,	"set_fraction", UNREADABLE),
	BIField(MFFLOAT,	"key []{beginGroup TRUE}", 0),
	BIField(MFINT32,	"keyValue []", 0),
	BIField(SFINT32,	"value_changed", UNWRITEABLE)
);
static const BIProto3 integerTrigger_ = BIProto3(
	"IntegerTrigger", "children",
	BIField(SFBOOL,	"set_boolean", UNREADABLE),
	BIField(SFINT32,	"integerKey -1", 0),
	BIField(SFINT32,	"triggerValue", UNWRITEABLE)
);
static const BIProto2 timeTrigger_ = BIProto2(
	"TimeTrigger", "children",
	BIField(SFBOOL,	"set_boolean", UNREADABLE),
	BIField(SFTIME,	"triggerTime", UNWRITEABLE)
);
static const BIProto3 staticGroup_ = BIProto3(
	"StaticGroup", "children",
	BIField(MFNODE,	"children []", UNREADABLE | UNWRITEABLE,	INIT_PROMPT_NODE(promptNode_scene_children)),
	BIField(SFVEC3F,	"bboxCenter 0 0 0", UNREADABLE | UNWRITEABLE),
	BIField(SFVEC3F,	"bboxSize -1 -1 -1", UNREADABLE | UNWRITEABLE)
);
static const BIProto10 floatAccelerator_ = BIProto10(
	"FloatAccelerator", 0,
	BIField(SFFLOAT,	"set_fraction", UNREADABLE),
	BIField(SFFLOAT,	"fraction_changed", UNWRITEABLE),
	BIField(SFFLOAT,	"value_changed", UNWRITEABLE),
	BIField(SFFLOAT,	"fracMult 1", 0),
	BIField(SFFLOAT,	"init 0", 0),
	BIField(MFFLOAT,	"delay []{beginGroup TRUE initElVal \"1\"}", SAME_AS_LAST),
	BIField(MFFLOAT,	"accelerate []{initElVal \"1\"}", SAME_AS_LAST),
	BIField(MFFLOAT,	"attack []{initElVal \"1\"}", SAME_AS_LAST),
	BIField(MFFLOAT,	"sustain []{initElVal \"0\"}", SAME_AS_LAST),
	BIField(MFFLOAT,	"decay []{initElVal \"1\"}", SAME_AS_LAST)
);
static const BIProto13 rotAccelerator_ = BIProto13(
	"RotAccelerator", 0,
	BIField(SFFLOAT,	"set_fraction", UNREADABLE),
	BIField(SFFLOAT,	"fraction_changed", UNWRITEABLE),
	BIField(SFROTATION,	"value_changed", UNWRITEABLE),
	BIField(SFFLOAT,	"fracMult 1", 0),
	BIField(SFROTATION,	"init 0 0 1 0", 0),
	BIField(MFVEC3F,	"axis []{beginGroup TRUE initElVal \"0 1 0\"}", SAME_AS_LAST | AXIS),
	BIField(MFFLOAT,	"delay []{initElVal \"1\"}", SAME_AS_LAST),
	BIField(MFFLOAT,	"accelerate []{initElVal \"1\"}", SAME_AS_LAST),
	BIField(MFFLOAT,	"attack []{initElVal \"1\"}", SAME_AS_LAST),
	BIField(MFFLOAT,	"sustain []{initElVal \"0\"}", SAME_AS_LAST),
	BIField(MFFLOAT,	"decay []{initElVal \"1\"}", SAME_AS_LAST),
	BIField(SFBOOL,	"calcForReturn FALSE", PUSH),
	BIField(SFBOOL,	"calcForPose FALSE", PUSH)
);
static const BIProto3 copyToMorph_ = BIProto3(
	"CopyToMorph", "build",
	BIField(SFNODE,	"part NULL", BLOCK_NEW,	INIT_PROMPT_NODE(promptNode_copyToMorph_part)),
	BIField(SFNODE,	"morpher NULL", BLOCK_NEW,	INIT_PROMPT_NODE(promptNode_copyToMorph_morpher)),
	BIField(SFNODE,	"setup NULL", 0,	INIT_PROMPT_NODE(promptNode_copyToMorph_setup))
);
static const BIProto2 relateMorpher_ = BIProto2(
	"RelateMorpher", "build",
	BIField(SFNODE,	"part NULL", BLOCK_NEW,	INIT_PROMPT_NODE(promptNode_relateMorpher_part)),
	BIField(SFNODE,	"setup NULL", 0,	INIT_PROMPT_NODE(promptNode_relateMorpher_setup))
);
static const BIProto19 immediatePart_ = BIProto19(
	"ImmediatePart", "_part",
	BIField(SFVEC3F,	"scale 1 1 1", 0),
	BIField(SFROTATION,	"rotation 0 1 0 0", 0),
	BIField(SFVEC3F,	"translation 0 0 0", 0),
	BIField(SFBOOL,		"track TRUE", DISABLE),
	BIField(SFBOOL,		"mirror TRUE", 0),
	BIField(SFBOOL,		"ik FALSE", 0),
	BIField(SFBOOL,		"pin FALSE", 0),
	BIField(SFROTATION,	"ikRot 0 0 1 0", 0),
	BIField(SFVEC3F,	"ikPos 0 0 0", 0),
	BIField(SFFLOAT,	"fka 0", 0), //ik blend angle
	BIField(SFFLOAT,	"fkb 0", 0), //ik blend interpolation
	BIField(MFVEC3F,	"coord []", DISABLE_LEN),
	BIField(MFVEC3F,	"normal []", DISABLE | DISABLE_LEN),
	BIField(MFINT32,	"coordGroup []{beginGroup TRUE rangeFrom 0 initElVal \"10\"}", SAME_AS_LAST | DISABLE),
	BIField(MFFLOAT,	"weight []{initElVal \"1\"}", SAME_AS_LAST | DISABLE),
	BIField(SFVEC3F,	"pivotPoint 0 0 0", 0),
	BIField(SFNODE,	"effect NULL", 0,	INIT_PROMPT_NODE(promptNode_part_effect)),
	BIField(MFNODE,	"morpher []", BLOCK_USE,	INIT_PROMPT_NODE(promptNode_part_morpher)),
	BIField(MFNODE,	"part []", BLOCK_USE,	INIT_PROMPT_NODE(promptNode_part_part))
);
static const BIProto12 delPlane_ = BIProto12(
	"DelPlane", "build",
	BIField(SFROTATION,	"orientation 0 1 0 0", 0),
	BIField(SFVEC3F,	"location 0 0 0", 0),
	BIField(SFBOOL,	"negX FALSE", 0),
	BIField(SFBOOL,	"posX FALSE", 0),
	BIField(SFBOOL,	"negY FALSE", 0),
	BIField(SFBOOL,	"posY FALSE", 0),
	BIField(SFBOOL,	"negZ FALSE", 0),
	BIField(SFBOOL,	"posZ FALSE", 0),
	BIField(SFBOOL,	"full TRUE", 0),
	BIField(SFBOOL,	"align FALSE", 0),
	BIField(MFNODE,	"part []", BLOCK_NEW,	INIT_PROMPT_NODE(promptNode_delPlane_part)),
	BIField(SFNODE,	"setup NULL", 0,	INIT_PROMPT_NODE(promptNode_delPlane_setup))
);
static const BIProto4 delSphere_ = BIProto4(
	"DelPlane", "build",
	BIField(SFVEC3F,	"location 0 0 0", 0),
	BIField(SFFLOAT,	"radius .25", 0),
	BIField(MFNODE,	"part []", BLOCK_NEW,	INIT_PROMPT_NODE(promptNode_delPlane_part)),
	BIField(SFNODE,	"setup NULL", 0,	INIT_PROMPT_NODE(promptNode_delPlane_setup))
);
static const BIProto4 radialTug_ = BIProto4(
	"RadialTug", "build",
	BIField(MFVEC3F,	"location []{beginGroup TRUE}", 0),
	BIField(MFVEC3F,	"displacement []", 0),
	BIField(MFFLOAT,	"slope []{rangeFrom .0001 initElVal \"1\"}", SAME_AS_LAST),
	BIField(MFFLOAT,	"radius []{initElVal \"1\"}", SAME_AS_LAST)
);
static const BIProto5 tapeMeasure_ = BIProto5(
	"TapeMeasure", "build",
	BIField(SFFLOAT,	"measurement 0", 0),
	BIField(SFINT32,	"baseVertex 0{rangeFrom 0}", 0),
	BIField(SFINT32,	"staves 0{rangeFrom 0}", 0),
	BIField(SFNODE,	"part NULL", BLOCK_NEW,	INIT_PROMPT_NODE(promptNode_tapeMeasure_part)),
	BIField(SFNODE,	"setup NULL", 0,	INIT_PROMPT_NODE(promptNode_tapeMeasure_setup))
);
static const BIProto11 posAccelerator_ = BIProto11(
	"PosAccelerator", 0,
	BIField(SFFLOAT,	"set_fraction", UNREADABLE),
	BIField(SFFLOAT,	"fraction_changed", UNWRITEABLE),
	BIField(SFVEC3F,	"value_changed", UNWRITEABLE),
	BIField(SFFLOAT,	"fracMult 1", 0),
	BIField(SFVEC3F,	"init 0 0 0", 0),
	BIField(MFVEC3F,	"direction []{beginGroup TRUE initElVal \"0 1 0\"}", SAME_AS_LAST | AXIS),
	BIField(MFFLOAT,	"delay []{initElVal \"1\"}", SAME_AS_LAST),
	BIField(MFFLOAT,	"accelerate []{initElVal \"1\"}", SAME_AS_LAST),
	BIField(MFFLOAT,	"attack []{initElVal \"1\"}", SAME_AS_LAST),
	BIField(MFFLOAT,	"sustain []{initElVal \"0\"}", SAME_AS_LAST),
	BIField(MFFLOAT,	"decay []{initElVal \"1\"}", SAME_AS_LAST)
);
static const BIProto8 animByAnkle_ = BIProto8(
	"AnimByAnkle", 0,
	BIField(SFVEC3F,	"position", UNREADABLE),
	BIField(SFROTATION,	"catAnkle", UNREADABLE),
	BIField(SFBOOL,	"outAnkle TRUE", 0),
	BIField(SFFLOAT,	"swivel 0", 0),
	BIField(SFNODE,	"pelvis NULL", BLOCK_NEW,	INIT_PROMPT_NODE(promptNode_animByAnkle_pelvis)),
	BIField(SFNODE,	"hip NULL", BLOCK_NEW,	INIT_PROMPT_NODE(promptNode_animByAnkle_hip)),
	BIField(SFNODE,	"knee NULL", BLOCK_NEW,	INIT_PROMPT_NODE(promptNode_animByAnkle_knee)),
	BIField(SFNODE,	"ankle NULL", BLOCK_NEW,	INIT_PROMPT_NODE(promptNode_animByAnkle_ankle))
);
static const BIProto7 animByWrist_ = BIProto7(
	"AnimByWrist", 0,
	BIField(SFVEC3F,	"position", UNREADABLE),
	BIField(SFROTATION,	"catWrist", UNREADABLE),
	BIField(SFBOOL,	"outWrist TRUE", 0),
	BIField(SFFLOAT,	"swivel 0", 0),
	BIField(SFNODE,	"sholder NULL", BLOCK_NEW,	INIT_PROMPT_NODE(promptNode_animByWrist_sholder)),
	BIField(SFNODE,	"elbow NULL", BLOCK_NEW,	INIT_PROMPT_NODE(promptNode_animByWrist_elbow)),
	BIField(SFNODE,	"wrist NULL", BLOCK_NEW,	INIT_PROMPT_NODE(promptNode_animByWrist_wrist))
);
static const BIProto4 rotFilter_ = BIProto4(
	"RotFilter", 0,
	BIField(SFROTATION,	"inRot", UNREADABLE),
	BIField(SFVEC3F,	"axis 1 1 1", 0),
	BIField(SFFLOAT,	"angle 1", 0),
	BIField(SFROTATION,	"outRot", UNWRITEABLE)
);
static const BIProto3 rotMult_ = BIProto3(
	"RotMult", 0,
	BIField(SFROTATION,	"inA", UNREADABLE),
	BIField(SFROTATION,	"inB", UNREADABLE),
	BIField(SFROTATION,	"catenation", UNWRITEABLE)
);
static const BIProto1 multiUser_ = BIProto1(
	"MultiUser", 0,
	BIField(MFSTRING,	"ip []", 0)
);
static const BIProto10 indexedTriangleSet_ = BIProto10(
	"IndexedTriangleSet", "geometry",
	BIField(MFINT32,	"set_index", UNREADABLE),
	BIField(SFNODE,	"coord NULL", 0,	INIT_PROMPT_NODE(promptNode_indexedTriangleSet_coord)),
	BIField(SFNODE,	"normal NULL", 0,	INIT_PROMPT_NODE(promptNode_indexedTriangleSet_normal)),
	BIField(SFNODE,	"color NULL", 0,	INIT_PROMPT_NODE(promptNode_indexedTriangleSet_color)),
	BIField(SFNODE,	"texCoord NULL", 0,	INIT_PROMPT_NODE(promptNode_indexedTriangleSet_texCoord)),
	BIField(SFBOOL,	"ccw TRUE", 0),
	BIField(SFBOOL,	"normalPerVertex TRUE", 0),
	BIField(SFBOOL,	"colorPerVertex TRUE", 0),
	BIField(SFBOOL,	"solid TRUE", 0),
	BIField(MFINT32,	"index []{rangeFrom 0 initElVal \"[0 0 0]\" incElVal \"[0 0 0]\"}", 0)
);
static const BIProto1 polyline2D_ = BIProto1(
	"Polyline2D", "geometry",
	BIField(MFVEC2F,	"lineSegments []", 0)
);
static const BIProto12 sphereBuild_ = BIProto12(
	"SphereBuild", 0,
	BIField(SFFLOAT,	"radius .25{rangeFrom 0}", 0),
	BIField(SFINT32,	"staves 16{rangeFrom 2}", 0),
	BIField(SFINT32,	"bands 8{rangeFrom 1}", 0),
	BIField(SFCOLOR,	"color 1 1 1", 0),
	BIField(SFBOOL,	"yCentred TRUE", 0),
	BIField(SFVEC3F,	"scale 1 1 1", 0),
	BIField(SFROTATION,	"rotation 0 1 0 0", 0),
	BIField(SFVEC3F,	"translation 0 0 0", 0),
	BIField(SFBOOL, 	"descendants FALSE", DISABLE),
	BIField(SFNODE,	"part NULL", BLOCK_NEW,	INIT_PROMPT_NODE(promptNode_sphereBuild_part)),
	BIField(SFNODE,	"textureMapper NULL", 0,	INIT_PROMPT_NODE(promptNode_stem_textureMapper)),
	BIField(SFNODE,	"setup NULL", 0,	INIT_PROMPT_NODE(promptNode_sphereBuild_setup))
);
static const BIProto19 cylinderBuild_ = BIProto19(
	"CylinderBuild", 0,
	BIField(SFFLOAT,	"height .5", 0),
	BIField(SFFLOAT,	"radius .25", 0),
	BIField(SFINT32,	"staves 16{rangeFrom 2}", 0),
	BIField(SFINT32,	"bands 1{rangeFrom 1}", 0),
	BIField(SFBOOL,	"topCap TRUE", 0),
	BIField(SFBOOL,	"baseCap TRUE", 0),
	BIField(SFFLOAT,	"topHeight 0{rangeFrom 0}", 0),
	BIField(SFFLOAT,	"baseHeight 0{rangeFrom 0}", 0),
	BIField(SFINT32,	"topBands 1{rangeFrom 1}", 0),
	BIField(SFINT32,	"baseBands 1{rangeFrom 1}", 0),
	BIField(SFCOLOR,	"color 1 1 1", 0),
	BIField(SFBOOL,	"yCentred FALSE", 0),
	BIField(SFVEC3F,	"scale 1 1 1", 0),
	BIField(SFROTATION,	"rotation 0 1 0 0", 0),
	BIField(SFVEC3F,	"translation 0 0 0", 0),
	BIField(SFBOOL, 	"descendants FALSE", DISABLE),
	BIField(SFNODE,	"part NULL", BLOCK_NEW,	INIT_PROMPT_NODE(promptNode_cylinderBuild_part)),
	BIField(SFNODE,	"textureMapper NULL", 0,	INIT_PROMPT_NODE(promptNode_stem_textureMapper)),
	BIField(SFNODE,	"setup NULL", 0,	INIT_PROMPT_NODE(promptNode_cylinderBuild_setup))
);
static const BIProto18 coneBuild_ = BIProto18(
	"ConeBuild", 0,
	BIField(SFFLOAT,	"height .5{rangeFrom 0}", 0),
	BIField(SFFLOAT,	"topRad 0{rangeFrom 0}", 0),
	BIField(SFFLOAT,	"baseRad .25{rangeFrom 0}", 0),
	BIField(SFINT32,	"staves 16{rangeFrom 2}", 0),
	BIField(SFINT32,	"bands 7{rangeFrom 1}", 0),
	BIField(SFBOOL,	"topCap TRUE", 0),
	BIField(SFBOOL,	"baseCap TRUE", 0),
	BIField(SFINT32,	"topBands 1{rangeFrom 1}", 0),
	BIField(SFINT32,	"baseBands 1{rangeFrom 1}", 0),
	BIField(SFCOLOR,	"color 1 1 1", 0),
	BIField(SFBOOL,	"yCentred FALSE", 0),
	BIField(SFVEC3F,	"scale 1 1 1", 0),
	BIField(SFROTATION,	"rotation 0 1 0 0", 0),
	BIField(SFVEC3F,	"translation 0 0 0", 0),
	BIField(SFBOOL, 	"descendants FALSE", DISABLE),
	BIField(SFNODE,	"part NULL", BLOCK_NEW,	INIT_PROMPT_NODE(promptNode_coneBuild_part)),
	BIField(SFNODE,	"textureMapper NULL", 0,	INIT_PROMPT_NODE(promptNode_stem_textureMapper)),
	BIField(SFNODE,	"setup NULL", 0,	INIT_PROMPT_NODE(promptNode_coneBuild_setup))
);
static const BIProto13 torusBuild_ = BIProto13(
	"TorusBuild", 0,
	BIField(SFFLOAT,	"holeRad .1", 0),
	BIField(SFFLOAT,	"tubeRad .1", 0),
	BIField(SFINT32,	"staves 32{rangeFrom 2}", 0),
	BIField(SFINT32,	"bands 16{rangeFrom 1}", 0),
	BIField(SFCOLOR,	"color 1 1 1", 0),
	BIField(SFBOOL,	"yCentred FALSE", 0),
	BIField(SFVEC3F,	"scale 1 1 1", 0),
	BIField(SFROTATION,	"rotation 0 1 0 0", 0),
	BIField(SFVEC3F,	"translation 0 0 0", 0),
	BIField(SFBOOL, 	"descendants FALSE", DISABLE),
	BIField(SFNODE,	"part NULL", BLOCK_NEW,	INIT_PROMPT_NODE(promptNode_torusBuild_part)),
	BIField(SFNODE,	"textureMapper NULL", 0,	INIT_PROMPT_NODE(promptNode_stem_textureMapper)),
	BIField(SFNODE,	"setup NULL", 0,	INIT_PROMPT_NODE(promptNode_torusBuild_setup))
);
static const BIProto9 boxBuild_ = BIProto9(
	"BoxBuild", 0,
	BIField(SFCOLOR,	"color 1 1 1", 0),
	BIField(SFBOOL,	"yCentred FALSE", 0),
	BIField(SFVEC3F,	"scale .25 .25 .25", 0),
	BIField(SFROTATION,	"rotation 0 1 0 0", 0),
	BIField(SFVEC3F,	"translation 0 0 0", 0),
	BIField(SFBOOL, 	"descendants FALSE", DISABLE),
	BIField(SFNODE,	"part NULL", BLOCK_NEW,	INIT_PROMPT_NODE(promptNode_boxBuild_part)),
	BIField(MFNODE,	"textureMapper []", 0,	INIT_PROMPT_NODE(promptNode_stem_textureMapper)),
	BIField(SFNODE,	"setup NULL", 0,	INIT_PROMPT_NODE(promptNode_boxBuild_setup))
);
static const BIProto18 rectangleBuild_ = BIProto18(
	"RectangleBuild", 0,
	BIField(SFFLOAT,	"xDistance 1", 0),
	BIField(SFFLOAT,	"yDistance 1", 0),
	BIField(SFINT32,	"staves 4{rangeFrom 1}", 0,	tip_rectangleBuild_staves),
	BIField(SFINT32,	"bands 4{rangeFrom 1}", 0,	tip_rectangleBuild_bands),
	BIField(SFBOOL,	"endSP FALSE", 0,	tip_rectangleBuild_endSP),
	BIField(SFBOOL,	"beginSP FALSE", 0,	tip_rectangleBuild_beginSP),
	BIField(SFCOLOR,	"color 1 1 1", 0),
	BIField(SFBOOL,	"xCentred TRUE", 0),
	BIField(SFBOOL,	"yCentred FALSE", 0),
	BIField(SFBOOL,	"xNegSide FALSE", 0),
	BIField(SFBOOL,	"yNegSide FALSE", 0),
	BIField(SFVEC3F,	"scale 1 1 1", 0),
	BIField(SFROTATION,	"rotation 0 1 0 0", 0),
	BIField(SFVEC3F,	"translation 0 0 0", 0),
	BIField(SFBOOL, 	"descendants FALSE", DISABLE),
	BIField(SFNODE,	"part NULL", BLOCK_NEW,	INIT_PROMPT_NODE(promptNode_rectangleBuild_part)),
	BIField(SFNODE,	"textureMapper NULL", 0,	INIT_PROMPT_NODE(promptNode_stem_textureMapper)),
	BIField(SFNODE,	"setup NULL", 0,	INIT_PROMPT_NODE(promptNode_rectangleBuild_setup))
);
static const BIProto14 bezierLatheBuild_ = BIProto14(
	"BezierLatheBuild", 0,
	BIField(MFVEC2F,	"point []{beginGroup TRUE}", 0),
	BIField(MFBOOL,		"fix []", DO_NOT_SAVE),
	BIField(MFINT32,	"dbands 16{rangeFrom 1}", DO_NOT_SAVE),
	BIField(MFINT32,	"nopfs []", DISABLE),
	BIField(MFINT32,	"bands 16{rangeFrom 1 beginGroup TRUE}", 0),
	BIField(SFINT32,	"staves 16{rangeFrom 2}", 0),
	BIField(SFCOLOR,	"color 1 1 1", 0),
	BIField(SFVEC3F,	"scale 1 1 1", 0),
	BIField(SFROTATION,	"rotation 0 1 0 0", 0),
	BIField(SFVEC3F,	"translation 0 0 0", 0),
	BIField(SFBOOL, 	"descendants FALSE", DISABLE),
	BIField(SFNODE,	"part NULL", BLOCK_NEW,	INIT_PROMPT_NODE(promptNode_bezierLatheBuild_part)),
	BIField(MFNODE,	"textureMapper []", 0,	INIT_PROMPT_NODE(promptNode_stem_textureMapper)),
	BIField(SFNODE,	"setup NULL", 0,	INIT_PROMPT_NODE(promptNode_bezierLatheBuild_setup))
);
static const BIProto1 mirrorEdit_ = BIProto1(
	"MirrorEdit", 0,
	BIField(MFNODE,	"part []", BLOCK_NEW,	INIT_PROMPT_NODE(promptNode_mirrorEdit_part))
);
static const BIProto1 mirrorGroup_ = BIProto1(
	"MirrorGroup", 0,
	BIField(MFNODE,	"mirrorEdit []", BLOCK_USE,	INIT_PROMPT_NODE(promptNode_mirrorGroup_mirrorEdit))
);
static const BIProto15 polyToNurbs_ = BIProto15(
	"PolyToNurbs", 0,
	BIField(SFFLOAT,	"sumDiff 0", 0),
	BIField(SFFLOAT,	"averageDiff 0", 0),
	BIField(SFBOOL,		"bond FALSE", 0),
	BIField(SFFLOAT,	"coordChange .01", 0),
	BIField(SFBOOL,		"evolve FALSE", 0),
	BIField(SFFLOAT,	"pointChange .001", 0),
	BIField(SFBOOL,		"spread FALSE", PUSH | DO_NOT_SAVE),
	BIField(SFINT32,	"spreadRes 100", 0),
	BIField(SFINT32,	"offset, 0", DO_NOT_SAVE),

	BIField(SFINT32,	"selOriginalVertex, 0", 0),
	BIField(SFINT32,	"selMimicVertex, 0{rangeFrom 0}", 0),
	BIField(SFBOOL,		"mimicOriginalIndex FALSE", DO_NOT_SAVE),

	BIField(SFNODE,		"original NULL", BLOCK_NEW,	INIT_PROMPT_NODE(promptNode_polyToNurbs_original)),
	BIField(SFNODE,		"mimic NULL", 0,	INIT_PROMPT_NODE(promptNode_polyToNurbs_mimic)),
	BIField(SFNODE,		"setup NULL", 0,	INIT_PROMPT_NODE(promptNode_sphereBuild_setup))
);
static const BIProto15 ringBuild_ = BIProto15(
	"RingBuild", 0,
	BIField(SFINT32,"numOfRingsToInsert 1{rangeFrom 1}", 0),
	BIField(MFVEC2F,	"scaleX []{beginGroup TRUE initElVal \".25 .25\"}", SAME_AS_LAST),
	BIField(MFVEC2F,	"scaleZ []{initElVal \".25 .25\"}", SAME_AS_LAST),
	BIField(MFROTATION,	"orientation []{initElVal \"0 0 1 0\"}", SAME_AS_LAST),
	BIField(MFVEC3F,	"location []{endGroup TRUE initElVal \"0 0 0\"}", SAME_AS_LAST),

	BIField(MFINT32,	"ownership []{beginGroup TRUE initElVal \"1\" rangeFrom 0}", SAME_AS_LAST),

	BIField(SFINT32,"staves 16{rangeFrom 1}", 0,	tip_stem_staves),
	//BIField(SFBOOL,	"insideOut FALSE", 0,	tip_stem_insideOut),

	BIField(SFBOOL,	"dragLocationY TRUE", 0),
	BIField(SFBOOL,	"dragLocationXZ TRUE", 0),
	BIField(SFBOOL,	"dragOrientation TRUE", 0),

	BIField(SFVEC3F,	"scale 1 1 1", 0),
	BIField(SFROTATION,	"rotation 0 0 1 0", 0),
	BIField(SFVEC3F,	"translation 0 0 0", 0),
	BIField(MFNODE,	"part []", BLOCK_NEW,	INIT_PROMPT_NODE(promptNode_sphereBuild_part)),
	BIField(SFNODE,	"setup NULL", 0,	INIT_PROMPT_NODE(promptNode_sphereBuild_setup))
);
static const BIProto13 bezierScale_ = BIProto13(
	"BezierScale", "children",
	BIField(SFVEC3F,	"location 0 0 0", 0),
	BIField(SFROTATION,	"orientation 0 1 0 0", 0),
	BIField(MFVEC2F,	"point []{beginGroup TRUE}", 0),
	BIField(MFBOOL,		"fix []", DO_NOT_SAVE),
	BIField(MFINT32,	"nopfs []", DISABLE),
	BIField(SFBOOL,		"bNegX FALSE", 0),
	BIField(SFBOOL,		"bPosX FALSE", 0),
	BIField(SFBOOL,		"bNegZ FALSE", 0),
	BIField(SFBOOL,		"bPosZ FALSE", 0),
	BIField(SFFLOAT,	"negX 1", 0),
	BIField(SFFLOAT,	"posX 1", 0),
	BIField(SFFLOAT,	"negZ 1", 0),
	BIField(SFFLOAT,	"posZ 1", 0)
);
static const BIProto7 bezierCircumScale_ = BIProto7(
	"BezierCircumScale", "children",
	BIField(SFVEC3F,	"location 0 0 0", 0),
	BIField(SFROTATION,	"orientation 0 1 0 0", 0),
	BIField(MFVEC2F,	"point []{beginGroup TRUE}", 0),
	BIField(MFBOOL,		"fix []", DO_NOT_SAVE),
	BIField(MFINT32,	"nopfs []", DISABLE),
	BIField(SFFLOAT,	"centerOffset 0", 0),
	BIField(SFFLOAT,	"level 1", 0)
);
/*
static const BIProto11 bezierPatch_ = BIProto11(
	"BezierPatch", "build",
	BIField(MFVEC3F,	"point []{beginGroup TRUE}", 0),
	BIField(MFBOOL,		"fix []", DO_NOT_SAVE),
	BIField(MFINT32,	"fixIdx []{beginGroup TRUE}", 0),
	BIField(SFINT32,	"numOfU 2", 0),
	BIField(SFINT32,	"numOfV 2", 0),
	BIField(SFINT32,	"staves 10", 0),
	BIField(SFINT32,	"bands 10", 0),
	BIField(MFNODE,		"source []", BLOCK_NEW,	INIT_PROMPT_NODE(promptNode_stem_source)),
	BIField(MFNODE,		"part []", BLOCK_NEW,	INIT_PROMPT_NODE(promptNode_stem_part)),
	BIField(MFNODE,		"task []", BLOCK_USE,	INIT_PROMPT_NODE(promptNode_bezierPatch_task)),
	BIField(SFNODE,		"setup NULL", 0,	INIT_PROMPT_NODE(promptNode_stem_setup))
);
*/
static const BIProto5 bezierPatch_ = BIProto5(
	"BezierPatch",		"build",
	BIField(MFVEC3F,	"point []{beginGroup TRUE}", 0),
	BIField(MFFLOAT,	"weight []", 0),
//	BIField(MFINT32,	"fixIdx []{beginGroup TRUE}", DISABLE),
	BIField(MFINT32,	"link []{beginGroup TRUE}", DISABLE),

	BIField(MFVEC2F,	"dim []{beginGroup TRUE}", 0),
	BIField(MFVEC2F,	"zone []", 0)
//	BIField(MFINT32,	"staves []", 0),
//	BIField(MFINT32,	"bands []", 0),
//	BIField(MFINT32,	"fixIdxIdx []", DISABLE)

//	BIField(SFBOOL,	"mimicX []{beginGroup TRUE}", 0),
//	BIField(SFBOOL,	"mimicY []", 0),
//	BIField(SFBOOL,	"mimicZ []", 0),
//	BIField(MFINT32,	"mimicSrc []", 0),
//	BIField(MFINT32,	"mimicDst []", 0)

//	BIField(MFNODE,		"source []", BLOCK_NEW,	INIT_PROMPT_NODE(promptNode_stem_source)),
//	BIField(MFNODE,		"part []", BLOCK_NEW,	INIT_PROMPT_NODE(promptNode_stem_part)),
//	BIField(MFNODE,		"task []", BLOCK_USE,	INIT_PROMPT_NODE(promptNode_bezierPatch_task)),
//	BIField(SFNODE,		"setup NULL", 0,	INIT_PROMPT_NODE(promptNode_stem_setup))
);
static const BIProto3 feedback_ = BIProto3(
	"Feedback", "feedback",
	BIField(SFVEC2F,	"multiply .5 .5", 0),
	BIField(SFINT32,	"from 0", 0),
	BIField(SFNODE,		"source NULL", 0,	INIT_PROMPT_NODE(promptNode_feedback_source))
);
static const BIProto11 povLight_ = BIProto11(
	"PovLight", "children",
	BIField(SFVEC3F,	"location 0 0 100", 0),
	BIField(SFCOLOR,	"color 1 1 1", 0),
	BIField(SFBOOL,		"pointlight FALSE", 0),
	BIField(SFBOOL,		"spotlight FALSE", 0),
	BIField(SFBOOL,		"shadowless FALSE", 0),
	BIField(SFFLOAT,	"radius 0.523598776", 0),
	BIField(SFFLOAT,	"falloff 0.785398163", 0),
	BIField(SFFLOAT,	"tightness 0", 0),
    BIField(SFVEC3F,	"point_at 0 0 0", 0),
	BIField(SFFLOAT,	"fade_distance 0", 0),
    BIField(SFFLOAT,	"fade_power 0", 0)
);
static const BIProto9 povFog_ = BIProto9(
	"PovFog", "children",
	BIField(SFFLOAT,	"lambda 2", 0),
	BIField(SFINT32,	"fog_type 1", 0),
	BIField(SFFLOAT,	"fog_offset 0", 0),
	BIField(SFFLOAT,	"fog_alt 0", 0),
	BIField(SFINT32,	"octaves 6", 0),
	BIField(SFFLOAT,	"omega 5", 0),
	BIField(SFVEC3F,	"turbulence 0 0 0", 0),
	BIField(SFFLOAT,	"turb_depth .5", 0),
	BIField(SFVEC3F,	"up 0 1 0", 0)
);
static const BIProto7 bezierShift_ = BIProto7(
	"BezierShift", "children",
	BIField(SFVEC3F,	"location 0 0 0", 0),
	BIField(SFROTATION,	"orientation 0 1 0 0", 0),
	BIField(MFVEC2F,	"point []{beginGroup TRUE}", 0),
	BIField(MFBOOL,		"fix []", DO_NOT_SAVE),
	BIField(MFINT32,	"nopfs []", DISABLE),
	BIField(SFFLOAT,	"negDecay 1", 0),
	BIField(SFFLOAT,	"posDecay 1", 0)
);
static const BIProto5 bezierTwist_ = BIProto5(
	"BezierTwist", "children",
	BIField(SFVEC3F,	"location 0 0 0", 0),
	BIField(SFROTATION,	"orientation 0 1 0 0", 0),
	BIField(MFVEC2F,	"point []{beginGroup TRUE}", 0),
	BIField(MFBOOL,		"fix []", DO_NOT_SAVE),
	BIField(MFINT32,	"nopfs []", DISABLE)
);
static const BIProto3 bezierTape_ = BIProto3(
	"BezierTape", "children",
	BIField(MFVEC2F,	"point []{beginGroup TRUE}", 0),
	BIField(MFBOOL,		"fix []", DO_NOT_SAVE),
	BIField(MFINT32,	"nopfs []", DISABLE)
);
static const BIProto5 bezierBend_ = BIProto5(
	"BezierBend", "children",
	BIField(SFVEC3F,	"location 0 0 0", 0),
	BIField(SFROTATION,	"orientation 0 1 0 0", 0),
	BIField(MFVEC3F,	"point []{beginGroup TRUE}", 0),
	BIField(MFBOOL,		"fix []", DO_NOT_SAVE),
	BIField(MFINT32,	"nopfs []", DISABLE)
);
static const BIProto22 nurbsPatch_ = BIProto22(
	"NurbsPatch", "build",
	BIField(MFVEC3F,	"point []{beginGroup TRUE}", DISABLE),
	BIField(MFVEC3F,	"point []", DISABLE_LEN | DO_NOT_SAVE),
	BIField(MFFLOAT,	"weight []", 0),
	BIField(MFFLOAT,	"uKnot []{beginGroup TRUE}", DISABLE),
	BIField(MFFLOAT,	"vKnot []{beginGroup TRUE}", DISABLE),
	BIField(SFVEC2F,	"order 3 3", DISABLE),

	BIField(SFBOOL,		"cage TRUE", DISABLE),
	BIField(SFBOOL,		"showAll FALSE", 0),
	BIField(SFBOOL,		"mirror FALSE", 0),
	BIField(SFVEC2F,	"dim 2 2", 0),

	BIField(SFBOOL,		"xTug TRUE", 0),
	BIField(SFBOOL,		"yTug TRUE", 0),
	BIField(SFBOOL,		"zTug TRUE", 0),

	BIField(SFBOOL,		"endSP FALSE", 0),
	BIField(SFBOOL,		"uJoinEnds TRUE", 0),
	BIField(SFBOOL,		"beginSP FALSE", 0),

	BIField(SFVEC3F,	"scale 1 1 1", DISABLE),
	BIField(SFROTATION,	"rotation 0 1 0 0", DISABLE),
	BIField(SFVEC3F,	"translation 0 0 0", DISABLE),
	BIField(SFNODE,		"fuseSurface NULL", 0),	//INIT_PROMPT_NODE(promptNode_nurbsPatch_fuseSurface)),
	BIField(MFNODE,		"stem NULL", 0,	INIT_PROMPT_NODE(promptNode_nurbsPatch_stem)),
	BIField(SFNODE, 	"setup NULL", 0,	INIT_PROMPT_NODE(promptNode_bezierLatheBuild_setup))
);

static const BIProto17 nurbsLathe_ = BIProto17(
	"NurbsLathe", 0,
	BIField(MFVEC2F,	"point []{beginGroup TRUE}", DISABLE),
	BIField(MFVEC2F,	"point []", DO_NOT_SAVE),
	BIField(MFFLOAT,	"weight []", 0),
	BIField(MFFLOAT,	"knot []{beginGroup TRUE}", DISABLE),
	BIField(SFINT32,	"order 3{rangeFrom 1}", DISABLE),
	BIField(SFINT32,	"staves 16{rangeFrom 2}", 0),
	BIField(SFINT32,	"bands 16{rangeFrom 1}", 0),
	//BIField(SFINT32,	"ctrlStaves 4{rangeFrom 3}", DO_NOT_SAVE),
	BIField(SFCOLOR,	"color 1 1 1", 0),
	BIField(SFVEC3F,	"scale 1 1 1", 0),
	BIField(SFROTATION,	"rotation 0 1 0 0", 0),
	BIField(SFVEC3F,	"translation 0 0 0", 0),
	BIField(SFBOOL,	"endSP TRUE", 0,	tip_rectangleBuild_endSP),
	BIField(SFBOOL,	"beginSP TRUE", 0,	tip_rectangleBuild_beginSP),
	BIField(SFBOOL, 	"descendants FALSE", DISABLE),
	BIField(SFNODE, 	"part NULL", BLOCK_NEW,	INIT_PROMPT_NODE(promptNode_bezierLatheBuild_part)),
	BIField(SFNODE,		"textureMapper NULL", 0,	INIT_PROMPT_NODE(promptNode_stem_textureMapper)),
	BIField(SFNODE, 	"setup NULL", 0,	INIT_PROMPT_NODE(promptNode_bezierLatheBuild_setup))
);
static const BIProto18 bendyLathe_ = BIProto18(
	"BendyLathe", 0,
	BIField(SFBOOL, 	"showRingControlPoints FALSE", 0),
	BIField(MFVEC3F,	"point []{beginGroup TRUE}", DISABLE),
	BIField(MFVEC3F,	"point []", DO_NOT_SAVE),
	BIField(MFFLOAT,	"weight []", DISABLE),
	BIField(MFFLOAT,	"knot []{beginGroup TRUE}", DISABLE),
	BIField(SFINT32,	"order 3{rangeFrom 1}", DISABLE),
	BIField(MFFLOAT,	"radius []{beginGroup TRUE}", 0),
	BIField(MFFLOAT,	"at []", 0),
	BIField(SFINT32,	"staves 16{rangeFrom 2}", 0),
	BIField(SFINT32,	"bands 25{rangeFrom 1}", 0),
	//BIField(SFINT32,	"ctrlStaves 4{rangeFrom 3}", DO_NOT_SAVE),
	BIField(SFCOLOR,	"color 1 1 1", 0),
	BIField(SFVEC3F,	"scale 1 1 1", 0),
	BIField(SFROTATION,	"rotation 0 1 0 0", 0),
	BIField(SFVEC3F,	"translation 0 0 0", 0),
	BIField(SFBOOL, 	"descendants FALSE", DISABLE),
	BIField(SFNODE, 	"part NULL", BLOCK_NEW,	INIT_PROMPT_NODE(promptNode_bezierLatheBuild_part)),
	BIField(SFNODE,		"textureMapper NULL", 0,	INIT_PROMPT_NODE(promptNode_stem_textureMapper)),
	BIField(SFNODE, 	"setup NULL", 0,	INIT_PROMPT_NODE(promptNode_bezierLatheBuild_setup))
);
static const BIProto22 fuseSurface_ = BIProto22(
	"FuseSurface",		"fuseSurface",
	BIField(MFVEC2F,	"baseCoord []{beginGroup TRUE initElVal \"[.01]\"}", SAME_AS_LAST),

	BIField(SFFLOAT,	"diff 0", DO_NOT_SAVE),
	BIField(SFFLOAT,	"sumDiff 0", DO_NOT_SAVE),
	BIField(SFFLOAT,	"averageDiff 0", DO_NOT_SAVE),

	BIField(SFFLOAT,	"vSpan .5", 0),
	BIField(SFINT32,	"numOf 1", 0),
	BIField(SFINT32,	"qty 0", 0),
	BIField(SFROTATION,	"orientation 0 1 0 0", DISABLE | DO_NOT_SAVE),
	BIField(SFBOOL,		"bond TRUE", 0),
	BIField(SFFLOAT,	"coordChange .01", 0),
	BIField(SFBOOL,		"spread FALSE", PUSH | DO_NOT_SAVE),
	BIField(SFINT32,	"spreadRes 100", 0),
	BIField(SFBOOL,		"polyMapAid FALSE", 0),
	BIField(SFBOOL,		"polyMapAidMirror FALSE", DISABLE | DO_NOT_SAVE),

	BIField(MFINT32,	"srcMatch []{beginGroup TRUE}", DISABLE_LEN),
	BIField(MFINT32,	"dstMatch []", 0),
	BIField(SFBOOL,		"in FALSE", 0),
	BIField(SFFLOAT,	"range .001", 0),
	BIField(SFBOOL,		"snap FALSE", 0),
	BIField(SFBOOL,		"evolve FALSE", 0),
	BIField(SFFLOAT,	"pointChange .001", 0),

	BIField(SFNODE,		"patch NULL", 0,		INIT_PROMPT_NODE(promptNode_fuseSurface_patch))
);
static const BIProto5 subdivision_ = BIProto5(
	"Subdivision", "build",
	BIField(MFVEC3F,	"point []{beginGroup TRUE}", 0),
	BIField(SFINT32,	"cycle 2", 0),
	BIField(MFINT32,	"coordIndex []{beginGroup TRUE initElVal \"[0 1 2 3]\"}", SAME_AS_LAST),
//	BIField(MFINT32,	"coordIndex []{beginGroup TRUE", DISABLE_LEN),
//	BIField(SFNODE,	"part NULL", BLOCK_NEW,	INIT_PROMPT_NODE(promptNode_stem_part)),
	BIField(SFNODE,	"part NULL", BLOCK_NEW,	INIT_PROMPT_NODE(promptNode_centerJoint_part)),
	BIField(SFNODE,	"setup NULL", 0,	INIT_PROMPT_NODE(promptNode_stem_setup))
);
static const BIProto13 oscillator_ = BIProto13(
	"Oscillator", "sound",
	BIField(MFVEC2F,	"note []{beginGroup TRUE}", DISABLE),
	BIField(MFVEC2F,	"note []", DO_NOT_SAVE),
	BIField(MFFLOAT,	"fWeight []", DISABLE),
	BIField(MFFLOAT,	"fKnot []{beginGroup TRUE}", DISABLE),
	BIField(SFINT32,	"fOrder 3", DISABLE),
	BIField(SFFLOAT,	"volume 1", 0),
	BIField(SFFLOAT,	"delay 0", 0),
	BIField(SFINT32,	"staves 32", 0),
	BIField(MFINT32,	"limit []", 0),
	BIField(MFVEC2F,	"vCoord []{beginGroup TRUE}", DO_NOT_SAVE),
	BIField(MFVEC3F,	"coord []{endGroup TRUE}", DO_NOT_SAVE),
	BIField(SFNODE,		"larynx NULL", 0,	INIT_PROMPT_NODE(promptNode_oscillator_larynx)),
	BIField(MFNODE,		"timbre NULL", 0,	INIT_PROMPT_NODE(promptNode_oscillator_timbre))
);

static const BIProto2 envelope_ = BIProto2(
	"Envelope", "envelope",
	BIField(MFVEC2F,	"point []", 0),
	BIField(MFNODE,		"chords NULL", 0,	INIT_PROMPT_NODE(promptNode_envelope_chords))
);

static const BIProto2 amplitudeModulator_ = BIProto2(
	"AmplitudeModulator", "modulator",
	BIField(MFVEC2F,	"wet []", 0),
	BIField(SFNODE,		"oscillator NULL", 0,	INIT_PROMPT_NODE(promptNode_modulator_oscillator))
);

static const BIProto2 frequencyModulator_ = BIProto2(
	"FrequencyModulator", "modulator",
	BIField(MFVEC2F,	"wet []", 0),
	BIField(SFNODE,		"oscillator NULL", 0,	INIT_PROMPT_NODE(promptNode_modulator_oscillator))
);

static const BIProto14 chords_ = BIProto14(
	"Chords", "chords",
	BIField(MFVEC2F,	"spectrum []{beginGroup TRUE}", 0),
	BIField(MFFLOAT,	"multiple []{beginGroup TRUE}", 0),
	BIField(MFFLOAT,	"amplitude []", 0),
	BIField(MFFLOAT,	"phase  []", 0),
	BIField(SFBOOL,		"sawTooth true", 0),
	BIField(SFBOOL,		"square false", 0),
	BIField(SFFLOAT,	"randomize 0", 0),
	BIField(SFINT32,	"harmonics 92", 0),
	BIField(SFFLOAT,	"increment 1", 0),
	BIField(SFBOOL,		"enable TRUE", 0),
	BIField(SFFLOAT,	"interval 0", 0),
	BIField(SFFLOAT,	"level 1", 0),
	//BIField(SFNODE,		"envelope NULL", 0,	INIT_PROMPT_NODE(promptNode_chords_envelope)),
	BIField(SFNODE,		"modulator NULL", 0,	INIT_PROMPT_NODE(promptNode_chords_amplitudeModulator)),
	BIField(SFNODE,		"phaseRef NULL", 0,	INIT_PROMPT_NODE(promptNode_chords_phaseRef))
);

static const BIProto1 larynx_ = BIProto1(
	"Larynx", "larynx",
	BIField(MFNODE,		"chords NULL", 0,	INIT_PROMPT_NODE(promptNode_larynx_chords))
);

static const BIProto3 phrase_ = BIProto3(
	"Phrase", "sound",
	BIField(SFFLOAT,	"volume 1", 0),
	BIField(SFFLOAT,	"delay 0", 0),
	BIField(MFNODE,		"sound []", 0,	INIT_PROMPT_NODE(promptNode_track_sound))
);
static const BIProto7 harmonic_ = BIProto7(
	"Harmonic", "timbre",
	BIField(MFVEC2F,	"point []{beginGroup TRUE}", DISABLE),
	BIField(MFVEC2F,	"point []", DO_NOT_SAVE),
	BIField(MFFLOAT,	"weight []", 0),
	BIField(MFFLOAT,	"knot []{beginGroup TRUE}", 0),
	BIField(SFINT32,	"order 3", 0),
	//BIField(MFINT32,	"numOf []{beginGroup TRUE initElVal \"[1]\"}", SAME_AS_LAST),
	BIField(MFFLOAT,	"multiply []{beginGroup TRUE initElVal \"[1]\"}", SAME_AS_LAST),
	BIField(MFFLOAT,	"phase []{initElVal \"[0]\"}", SAME_AS_LAST)
);
static const BIProto11 harmonicPatch_ = BIProto11(
	"HarmonicPatch", "timbre",
	BIField(MFVEC3F,	"point []{beginGroup TRUE}", DISABLE),
	BIField(MFVEC3F,	"point []", DISABLE_LEN | DO_NOT_SAVE),
	BIField(MFFLOAT,	"weight []", 0),
	BIField(MFFLOAT,	"uKnot []{beginGroup TRUE}", 0),
	BIField(MFFLOAT,	"vKnot []{beginGroup TRUE}", 0),
	BIField(SFVEC2F,	"dim 2 2", 0),
	BIField(SFVEC2F,	"order 3 3", 0),
	BIField(SFINT32,	"staves 100", SAME_AS_LAST),
	BIField(MFFLOAT,	"vCoord []{beginGroup TRUE}", SAME_AS_LAST),
	BIField(MFVEC3F,	"coord []{endGroup TRUE}", SAME_AS_LAST),
	BIField(SFINT32,	"harmonics 30", SAME_AS_LAST)
);
static const BIProto1 sample_ = BIProto1(
	"SampleRate", "timbre",
	BIField(SFSTRING,	"url \"\"", 0)
);
static const BIProto13 waveToPoly_ = BIProto13(
	"WaveToPoly", "children",
	BIField(SFBOOL,	"open FALSE", DO_NOT_SAVE | PUSH | DO_NOT_DISPLY_NAME, "open wav file"),
	BIField(SFSTRING,	"file \"\"", DISABLE),
	BIField(SFINT32,	"viewFrom 0", 0),
	BIField(SFINT32,	"zoom 1{rangeFrom 1}", DO_NOT_DISPLY_NAME | DO_NOT_DISPLY_VALUE, "zoom"),
	BIField(SFBOOL,	"convert FALSE", DO_NOT_SAVE | PUSH , 0),
	BIField(SFINT32,	"cMark 0", DO_NOT_DISPLY_NAME, "dec/inc cycle"),
	BIField(SFINT32,	"nextCycle 0", DO_NOT_DISPLY_NAME, "next cycle"),
	BIField(SFBOOL,	"addMark FALSE", DO_NOT_SAVE | PUSH | DO_NOT_DISPLY_NAME, "add cycle mark"),
	BIField(SFBOOL,	"outSections FALSE", DO_NOT_SAVE | PUSH | DO_NOT_DISPLY_NAME, "out sections"),
	BIField(MFINT32,	"cycle []", DISABLE),
	BIField(SFBOOL,	"sawTooth TRUE", 0),
	BIField(SFNODE,	"part NULL", BLOCK_NEW,	INIT_PROMPT_NODE(promptNode_part_part)),
	BIField(SFNODE,	"osc NULL", BLOCK_NEW,	INIT_PROMPT_NODE(promptNode_track_sound))
);
static const BIProto7 track_ = BIProto7(
	"Track", "track",
	//BIField(SFBOOL,		"output FALSE", PUSH | DISABLE),
	BIField(SFINT32,	"voices 1", 0),
	BIField(SFFLOAT, 	"detune 0", 0),
	BIField(SFINT32,	"echos 0", 0),
	BIField(SFFLOAT,	"echoDelay 0", 0),
	BIField(SFFLOAT,	"decay 0", 0),
	BIField(SFFLOAT,	"volume 1", 0),
	BIField(MFNODE,		"sound []", 0,	INIT_PROMPT_NODE(promptNode_track_sound))
);

static const BIProto5 composition_ = BIProto5(
	"Composition", "children",
	BIField(SFBOOL,		"output FALSE", PUSH),
	BIField(SFFLOAT,	"volume 1", 0),
	BIField(SFFLOAT,	"delay 0", 0),
	BIField(SFFLOAT,	"duration 10", 0),
	BIField(MFNODE,		"track []", 0,	INIT_PROMPT_NODE(promptNode_composition_track))
);

static const BIProto3 cPMorpher_ = BIProto3(
	"CPMorpher", "children",
	BIField(MFFLOAT,	"key []", 0),
	BIField(MFVEC3F,	"keyValue []", 0),
	BIField(MFINT32,	"cpIndex []", 0)
);

static const BIProto1 g_S_Bool_ = BIProto1(
	"G_S_Bool", 0,
	BIField(SFBOOL,	"value FALSE", 0)
);
static const BIProto1 g_S_Color_ = BIProto1(
	"G_S_Color", 0,
	BIField(SFCOLOR,	"value 1 1 1", 0)
);
static const BIProto1 g_S_Float_ = BIProto1(
	"G_S_Float", 0,
	BIField(SFFLOAT,	"value 0", 0)
);
static const BIProto0 g_S_Image_ = BIProto0(
	"G_S_Image", 0
);
static const BIProto1 g_S_Int32_ = BIProto1(
	"G_S_Int32", 0,
	BIField(SFINT32,	"value 0", 0)
);
static const BIProto1 g_NodePtr_ = BIProto1(
	"G_NodePtr", 0,
	BIField(SFNODE,	"value NULL", 0)
);
static const BIProto1 g_S_Rotation_ = BIProto1(
	"G_S_Rotation", 0,
	BIField(SFROTATION,	"value 0 1 0 0", 0)
);
static const BIProto1 g_S_String_ = BIProto1(
	"G_S_String", 0,
	BIField(SFSTRING,	"value \"\"", 0)
);
static const BIProto1 g_S_Time_ = BIProto1(
	"G_S_Time", 0,
	BIField(SFTIME,	"value 0", 0)
);
static const BIProto1 g_S_Vec2f_ = BIProto1(
	"G_S_Vec2f", 0,
	BIField(SFVEC2F,	"value 0 0", 0)
);
static const BIProto1 g_S_Vec3f_ = BIProto1(
	"G_S_Vec3f", 0,
	BIField(SFVEC3F,	"value 0 0 0", 0)
);
static const BIProto1 g_M_Bool_ = BIProto1(
	"G_M_Bool", 0,
	BIField(MFBOOL,	"value []", 0)
);
static const BIProto1 g_M_Color_ = BIProto1(
	"G_M_Color", 0,
	BIField(MFCOLOR,	"value []", 0)
);
static const BIProto1 g_M_Float_ = BIProto1(
	"G_M_Float", 0,
	BIField(MFFLOAT,	"value []", 0)
);
static const BIProto0 g_M_Image_ = BIProto0(
	"G_M_Image", 0
);
static const BIProto1 g_M_Int32_ = BIProto1(
	"G_M_Int32", 0,
	BIField(MFINT32,	"value []", 0)
);
static const BIProto1 g_M_Node_ = BIProto1(
	"G_M_Node", 0,
	BIField(MFNODE,	"value []", 0)
);
static const BIProto1 g_M_Rotation_ = BIProto1(
	"G_M_Rotation", 0,
	BIField(MFROTATION,	"value []", 0)
);
static const BIProto1 g_M_String_ = BIProto1(
	"G_M_String", 0,
	BIField(MFSTRING,	"value []", 0)
);
static const BIProto1 g_M_Time_ = BIProto1(
	"G_M_Time", 0,
	BIField(MFDOUBLE,	"value []", 0)
);
static const BIProto1 g_M_Vec2f_ = BIProto1(
	"G_M_Vec2f", 0,
	BIField(MFVEC2F,	"value []", 0)
);
static const BIProto1 g_M_Vec3f_ = BIProto1(
	"G_M_Vec3f", 0,
	BIField(MFVEC3F,	"value []", 0)
);

static const BIProto4 file_ = BIProto4(
	"File", 0,
	BIField(SFBOOL,		"sortByDate FALSE", 0),
	BIField(SFSTRING,	"filename \"\"", 0),
	BIField(SFBOOL,		"open FALSE", PUSH),
	BIField(MFNODE,		"folder []", 0)
);
static const BIProto2 floorHeight_ = BIProto2(
	"FloorHeight", 0,
	BIField(SFBOOL,		"zero_height FALSE", PUSH),
	BIField(SFBOOL,		"blaxxun_height FALSE", PUSH)
);
static const BIProto4 questionMessage3Button_ = BIProto4(
	"QuestionMessage3Button", 0,
	BIField(SFSTRING,	"msg \"\"", 0),
	BIField(SFBOOL,		"yes FALSE", PUSH),
	BIField(SFBOOL,		"no FALSE", PUSH),
	BIField(SFBOOL,		"cancel FALSE", PUSH)
);
static const BIProto6 recentFiles_ = BIProto6(
	"RecentFiles", 0,
	BIField(SFBOOL,		"sel0 FALSE", LINK | DISABLE),
	BIField(SFBOOL,		"sel1 FALSE", LINK | DISABLE),
	BIField(SFBOOL,		"sel2 FALSE", LINK | DISABLE),
	BIField(SFBOOL,		"sel3 FALSE", LINK | DISABLE),
	BIField(SFBOOL,		"sel4 FALSE", LINK | DISABLE),
	BIField(SFBOOL,		"sel5 FALSE", LINK | DISABLE)
);

static const BIProto13 palette_ = BIProto13(
	"Palette", 0,
	BIField(SFBOOL,		"insert FALSE", PUSH | DO_NOT_SAVE),
	BIField(SFBOOL,		"delete FALSE", PUSH | DO_NOT_SAVE),
	BIField(SFINT32,	"coords 0", DO_NOT_SAVE),
	BIField(SFFLOAT,	"weight 0", DO_NOT_SAVE),
	BIField(SFBOOL,		"autoWeightDivide TRUE", DO_NOT_SAVE),
	BIField(SFCOLOR,	"color 1 1 1", DO_NOT_DISPLY_NAME),
	BIField(SFBOOL,		"define FALSE", DO_NOT_SAVE),
	BIField(SFBOOL,		"dropper FALSE", DO_NOT_SAVE),
	BIField(SFBOOL,		"rectangle FALSE", DO_NOT_SAVE),
	BIField(SFBOOL,		"fill FALSE", DO_NOT_SAVE),
	BIField(SFBOOL,		"vertex FALSE", DO_NOT_SAVE),
	BIField(SFBOOL,		"triangle FALSE", DO_NOT_SAVE),
	BIField(MFCOLOR,	"customColor []", 0)
);

static const BIProto3 commonSpec_ = BIProto3(
	"CommonSpec", 0,
	BIField(SFBOOL,		"mirror FALSE", DISABLE),
	BIField(MFCOLOR,	"color []", 0),
	BIField(MFSTRING,	"alternative []", 0)
);

static const BIProto10 freezeTime_ = BIProto10(
	"FreezeTime", 0,
	BIField(SFTIME,	"time 0", 0),
	BIField(SFFLOAT,	"multiply 1", 0),
	BIField(SFFLOAT,	"step .25", 0),
	BIField(SFBOOL,		"play FALSE", PUSH),
	BIField(SFFLOAT,	"period 6", 0),
	BIField(SFBOOL,	"repeat FALSE", 0),
	BIField(SFBOOL,	"deleteSample FALSE", PUSH),
	BIField(SFBOOL,	"samplePose FALSE", PUSH),
	BIField(SFBOOL,	"retainPose FALSE", 0),
	BIField(SFBOOL,	"freeze FALSE", 0)
);

static const BIProto33 settings_ = BIProto33(
	"Settings", 0,
	//BIField(SFBOOL,		"autoOpen FALSE", 0),
	BIField(SFBOOL,		"screenshotWindows FALSE", 0),
	BIField(SFBOOL,		"screenshotMouse FALSE", 0),
	BIField(SFBOOL,		"saveGZipped TRUE", 0),
	BIField(SFBOOL,		"invertWindowColor FALSE", 0),
	BIField(SFBOOL,		"showToolBar TRUE", 0),
	BIField(SFBOOL,		"extendNodeBar FALSE", 0),
	BIField(SFBOOL,		"loadColorCodePresets FALSE", PUSH),
	BIField(MFSTRING,	"searchEngine []", DISABLE_LEN),
	BIField(SFBOOL,		"absoluteRot90 FALSE", DISABLE | DO_NOT_SAVE),//phase out
	BIField(SFBOOL,		"backfaceIgnore TRUE", 0),
	BIField(SFBOOL,		"perspective FALSE", 0),
	BIField(SFFLOAT,	"incDecRes .01", 0),
	BIField(SFFLOAT,	"incDecResCtrl .001", 0),
	BIField(SFFLOAT,	"backupDelay .4", 0),
	BIField(SFINT32,	"freeCPU 0", 0),
	BIField(SFBOOL,		"record FALSE", 0),
	BIField(SFBOOL,		"shiftDragTranslate FALSE", DISABLE),//phase out?
	BIField(SFFLOAT,	"tugSlope 1", 0),
	BIField(SFFLOAT,	"tugRadius 0", 0),
	BIField(SFBOOL,		"multiPartTug TRUE", 0),
	BIField(SFSTRING,	"htmlBrowser \"\"", 0),
	BIField(SFFLOAT,	"version 0", UNREADABLE | UNWRITEABLE | DISABLE),
	BIField(MFSTRING,	"file []", UNREADABLE | UNWRITEABLE | DISABLE),
	BIField(SFBOOL,		"sortByDate FALSE", UNREADABLE | UNWRITEABLE | DISABLE),
	BIField(SFVEC2F,	"win3dPos 0 0", UNREADABLE | UNWRITEABLE | DISABLE),
	BIField(SFVEC2F,	"win3dSize 800 522", UNREADABLE | UNWRITEABLE),
	BIField(SFVEC2F,	"sceneTreePos 500 25", UNREADABLE | UNWRITEABLE | DISABLE),
	BIField(SFVEC2F,	"sceneTreeSize 300 480", UNREADABLE | UNWRITEABLE | DISABLE),
	BIField(SFVEC2F,	"nodeWinPos 400 400", UNREADABLE | UNWRITEABLE | DISABLE | DO_NOT_SAVE),
	BIField(SFVEC2F,	"toolBarPos 0 450", UNREADABLE | UNWRITEABLE | DISABLE),
	BIField(SFFLOAT,	"roundCoordsSave 6", 0),
	BIField(SFNODE,		"palette NULL", 0),
	BIField(MFNODE,		"commonSpec []", 0,	INIT_PROMPT_NODE(promptNode_settings_commonSpec))
);

static const BIProto3 stretchCurve_ = BIProto3(
	"StretchCurve", 0,
	BIField(SFBOOL,	"doIt FALSE", PUSH),
	BIField(SFVEC2F,	"at 1 1", 0),
	BIField(SFVEC2F,	"add 0 .01", 0)
);
static const BIProto12 openNetFile_ = BIProto12(
	"OpenNetFile", 0,
	BIField(SFBOOL,		"worldList FALSE", 0),
	BIField(SFSTRING,	"worldUrl \"\"", 0),
	BIField(SFSTRING,	"serverUrl \"\"", 0),
	BIField(SFSTRING,	"nickname \"\"", 0),
	BIField(SFSTRING,	"avatarUrl \"\"", 0),
	BIField(SFFLOAT,	"landOffset 0", 0),
	BIField(SFBOOL,		"male FALSE", 0),
	BIField(SFSTRING,	"voice \"\"", 0),
	BIField(SFSTRING,	"alarmFile \"\"", 0),
	BIField(SFBOOL,		"go FALSE", PUSH),
	BIField(SFVEC2F,	"chatPos 175, 83", DISABLE),
	BIField(SFVEC2F,	"chatSize 400, 132", DISABLE)
);
static const BIProto5 demo_ = BIProto5(
	"Demo", 0,
	BIField(SFINT32,	"operation 0", 0),
	BIField(SFBOOL,	"play FALSE", PUSH),
	BIField(SFBOOL,	"stop FALSE", PUSH),
	BIField(SFBOOL,	"express FALSE", 0),
	BIField(SFSTRING,	"status \"\"", 0)
);
static const BIProto2 convertToPatch_ = BIProto2(
	"ConvertToPatch", 0,
	BIField(SFINT32,	"control_staves 4{rangeFrom 1}", 0),
	BIField(SFINT32,	"control_bands 4{rangeFrom 1}", 0)
);
static const BIProto2 reduceCapturedPoses_ = BIProto2(
	"ReduceCapturedPoses", 0,
	BIField(SFINT32,	"keepCaptureForEvery 2", 0),
	BIField(SFBOOL,	"reduce FALSE", PUSH)
);

static const BIProto11 transformCP_ = BIProto11(
	"Selection", 0,
	BIField(SFBOOL,		"selectRing FALSE", 0),
	BIField(SFBOOL,		"selectRings FALSE", 0),
	BIField(SFBOOL,		"sameScreenPosition FALSE", 0),
	BIField(SFBOOL,		"rotAxisX FALSE", 0),
	BIField(SFBOOL,		"rotAxisY FALSE", 0),
	BIField(SFBOOL,		"rotAxisZ FALSE", 0),
	BIField(SFVEC3F,	"pivotPoint 0 0 0", 0),
	BIField(SFROTATION,	"scaleOrientation 0 0 1 0", 0),
	BIField(SFVEC3F,	"scale 1 1 1", 0),
	BIField(SFROTATION,	"rotation 0 0 1 0", 0),
	BIField(SFVEC3F,	"translation 0 0 0", 0)
);

static const BIProto18 anim_ = BIProto18(
	"Anim", "children",
	BIField(SFFLOAT,	"slider 0", DO_NOT_OUTPUT | DO_NOT_DISPLY_NAME | DO_NOT_RESIZE, "drag slider to move to a point in time"),
	BIField(SFINT32,	"timeLine 0", DO_NOT_OUTPUT | UNREADABLE | UNWRITEABLE | DO_NOT_DISPLY_NAME | DO_NOT_DISPLY_VALUE, "click time-line to move to a point in time"),
	BIField(SFINT32,	"select 0", DO_NOT_OUTPUT | UNREADABLE | UNWRITEABLE | DO_NOT_DISPLY_NAME | DO_NOT_DISPLY_VALUE, "drag to select section"),
	BIField(SFBOOL,	"play FALSE",  PUSH | DO_NOT_DISPLY_NAME, "play/stop"),
	BIField(SFBOOL,	"pause TRUE", PUSH | DO_NOT_DISPLY_NAME, "pause/resume"),
	BIField(SFBOOL,	"loop FALSE", PUSH | DO_NOT_DISPLY_NAME, "loop/play once"),
	BIField(SFTIME,	"zoom 4", DO_NOT_DISPLY_NAME | DO_NOT_RESIZE | DO_NOT_DISPLY_VALUE, "zoom time line"),
	BIField(SFTIME,	"period 1", DO_NOT_DISPLY_NAME | DO_NOT_RESIZE | DO_NOT_DISPLY_VALUE, "add second to timeline"),
	BIField(SFINT32,	"stepsPerSecond 25{rangeFrom 1 rangeTo 70}", DO_NOT_OUTPUT | DO_NOT_DISPLY_NAME | DO_NOT_RESIZE, "steps per second"),
	BIField(SFFLOAT,	"speed 1", DO_NOT_OUTPUT | DO_NOT_DISPLY_NAME | DO_NOT_RESIZE, "speed"),
	
	BIField(SFFLOAT,	"playFrom FALSE", DO_NOT_DISPLY_NAME | DO_NOT_RESIZE | DO_NOT_DISPLY_VALUE, "mark where to play from"),
	BIField(SFBOOL,	"copy FALSE", PUSH | DO_NOT_DISPLY_NAME, "copy"),
	BIField(SFBOOL,	"paste FALSE", PUSH | DO_NOT_DISPLY_NAME, "paste"),
	BIField(SFBOOL,	"remove FALSE", PUSH | DO_NOT_DISPLY_NAME, "delete"),
	BIField(SFNODE,	"soundTrack NULL", DO_NOT_OUTPUT,	INIT_PROMPT_NODE(promptNode_anim_soundTrack)),
	BIField(SFNODE,	"video NULL", DO_NOT_OUTPUT,	INIT_PROMPT_NODE(promptNode_anim_video)),
	BIField(MFNODE,	"children []", 0,	INIT_PROMPT_NODE(promptNode_anim_children)),
	BIField(SFNODE,	"onFrame NULL", DO_NOT_OUTPUT,	INIT_PROMPT_NODE(promptNode_anim_onFrame))
);

static const BIProto15 timelineFilter_ = BIProto15(
	"TimelineFilter", "children",
	BIField(SFBOOL,	"morph TRUE", 0),
	BIField(SFBOOL,	"rotation TRUE", 0),
	BIField(SFBOOL,	"translation TRUE", 0),
	BIField(SFBOOL,	"posCompensation TRUE", 0),
	BIField(SFBOOL,	"fka TRUE", 0),
	BIField(SFBOOL,	"fkb TRUE", 0),
	BIField(SFBOOL,	"zoom TRUE", 0),
	BIField(SFBOOL,	"fieldOfView TRUE", 0),
	BIField(SFBOOL,	"orientation TRUE", 0),
	BIField(SFBOOL,	"position TRUE", 0),
	BIField(SFBOOL,	"tickMorphFieldOnly FALSE", PUSH),
	BIField(SFBOOL,	"tickPartFieldsOnly FALSE", PUSH),
	BIField(SFBOOL,	"tickViewpointFieldsOnly FALSE", PUSH),
	BIField(SFBOOL,	"selectAllFields FALSE", PUSH),
	BIField(SFBOOL,	"ok FALSE", PUSH)
);


static const BIProto16 pasteOptions_ = BIProto16(
	"PasteOptions", "children",
	BIField(SFBOOL,	"filter FALSE", PUSH),
	BIField(SFBOOL,	"insert FALSE", 0),
	BIField(SFBOOL,	"beginEdge TRUE", 0),
	BIField(SFBOOL,	"endEdge TRUE", 0),
	BIField(SFINT32, "steps 0", 0),
	BIField(SFBOOL,	"quantize TRUE", 0),
	BIField(SFBOOL,	"transform FALSE", 0),
	BIField(SFBOOL,	"calcRot FALSE", 0),
	BIField(SFVEC3F,	"pivotPoint 0 0 0", 0),
	BIField(SFVEC3F,	"scale 1 1 1", 0),
	BIField(SFROTATION,	"rotate 0 0 1 0", 0),
	BIField(SFVEC3F,	"translate 0 0 0", 0),
	BIField(SFBOOL,	"mirror FALSE", 0),
	BIField(SFBOOL,	"curve FALSE", 0),
	BIField(SFFLOAT, "curveDistance 0", 0),
	BIField(SFBOOL,	"ok FALSE", PUSH)
);

static const BIProto5 deleteOptions_ = BIProto5(
	"DeleteOptions", "children",
	BIField(SFBOOL,	"filter FALSE", PUSH),
	BIField(SFBOOL,	"time FALSE", 0),
	BIField(SFBOOL,	"beginEdge TRUE", 0),
	BIField(SFBOOL,	"endEdge TRUE", 0),
	BIField(SFBOOL,	"ok FALSE", PUSH)
);

static const BIProto* builtInProtoList[]={
	&nullNode_,
	&scene_,
	&avatarSocket_,
	&shape_,
	&part_,
	&appearance_,
	&indexedFaceSet_,
	&imageTexture_,
	&material_,
	&navigationInfo_,
	&group_,
	&noOutputGroup_,
	&gestureGroup_,
	&mobility_,
	&cCLathe_,
	&cCBend_,
	&surfaceGenerator_,
	&stem_,
	&polymap_,
	&branchJoin_,
	&generator_,
	&geomSource_,
	&surfaceTransform_,
	&place_,
	&directionalLight_,
	&cCScale_,
	&cCTranslate_,
	&cCRingScale_,
	&bezierLathe_,
	&eCLathe_,
	&eCScale_,
	&eCShift_,
	&eCBend_,
	&eCSyncBend_,
	&eCTwist_,
	&background_,
	&seamless_,
	&groupBuild_,
	&sCGroupBuild_,
	&joinCloseVertex_,
	&centerJoint_,
	&delPoly_,
	&joinVertex_,
	&copyPart_,
	&viewpoint_,
	&transferBCyl_,
	&transferOwnership_,
	&transferVertex_,
	&coordinate_,
	&normal_,
	&color_,
	&textureCoordinate_,
	&sineTug_,
	&simulSineTug_,
	&genPoly_,
	&script_,
	&branchJoinNegX_,
	&branchJoinPosX_,
	&branchJoinNegY_,
	&branchJoinPosY_,
	&branchJoinNegZ_,
	&branchJoinPosZ_,
	&projectOntoTri_,
	&colorSweep_,
	&evolveCoord_,
	&function_,
	&memberFunction_,
	&codeLine_,
	&morphPart_,
	&morpher_,
	&textureTransform_,
	&protoDeclare_,
	&protoInterface_,
	&protoBody_,
	&protoInstance_,
	&movie_,
	&video_,
	&anim_,
	&timeSensor_,
	&orientationInterpolator_,
	&positionInterpolator_,
	&nurbsTrack_,
	&scalarInterpolator_,
	&posCompensator_,
	&collision_,
	&transform_,
	&colorInterpolator_,
	&coordinateInterpolator_,
	&normalInterpolator_,
	&modelSequence_,
	&modelMorpher_,
	&worldInfo_,
	&touchSensor_,
	&cCPath_,
	&bspTree_,
	&inline_,
	&fog_,
	&fog2_,
	&anchor_,
	&audioClip_,
	&billboard_,
	&box_,
	&cone_,
	&cylinder_,
	&cylinderSensor_,
	&elevationGrid_,
	&extrusion_,
	&fontStyle_,
	&indexedLineSet_,
	&nurbsSurface_,
	&lOD_,
	&movieTexture_,
	&pixelTexture_,
	&planeSensor_,
	&pointLight_,
	&pointSet_,
	&proximitySensor_,
	&sound_,
	&sphere_,
	&sphereSensor_,
	&spotLight_,
	&switch_,
	&text_,
	&visibilitySensor_,
	&textureMapper_,
	&textureEffect_,
	&colorEffect_,
	&monoColorEffect_,
	&exitBuild_,
	&splitTri_,
	&swapTriDivide_,
	&placeTexture_,
	&hAnimHumanoid_,
	&hAnimJoint_,
	&hAnimDisplacer_,
	&hAnimSegment_,
	&hAnimSite_,
	&booleanFilter_,
	&booleanSequencer_,
	&booleanToggle_,
	&booleanTrigger_,
	&integerSequencer_,
	&integerTrigger_,
	&timeTrigger_,
	&staticGroup_,
	&floatAccelerator_,
	&rotAccelerator_,
	&copyToMorph_,
	&relateMorpher_,
	&immediatePart_,
	&delPlane_,
	&delSphere_,
	&radialTug_,
	&tapeMeasure_,
	&posAccelerator_,
	&animByAnkle_,
	&animByWrist_,
	&rotFilter_,
	&rotMult_,
	&multiUser_,
	&indexedTriangleSet_,
	&polyline2D_,
	&sphereBuild_,
	&cylinderBuild_,
	&coneBuild_,
	&torusBuild_,
	&boxBuild_,
	&rectangleBuild_,
	&bezierLatheBuild_,
	&mirrorEdit_,
	&mirrorGroup_,
	&polyToNurbs_,
	&ringBuild_,
	&bezierScale_,
	&bezierCircumScale_,
	&bezierPatch_,
	&feedback_,
	&povLight_,
	&povFog_,
	&bezierShift_,
	&bezierTwist_,
	&bezierTape_,
	&bezierBend_,
	&nurbsPatch_,
	&nurbsLathe_,
	&bendyLathe_,
	&fuseSurface_,
	&subdivision_,
	&oscillator_,
	&envelope_,
	&amplitudeModulator_,
	&frequencyModulator_,
	&chords_,
	&larynx_,
	&phrase_,
	&harmonic_,
	&harmonicPatch_,
	&sample_,
	&waveToPoly_,
	&track_,
	&composition_,
	&cPMorpher_,
	&g_S_Bool_,
	&g_S_Color_,
	&g_S_Float_,
	&g_S_Image_,
	&g_S_Int32_,
	&g_NodePtr_,
	&g_S_Rotation_,
	&g_S_String_,
	&g_S_Time_,
	&g_S_Vec2f_,
	&g_S_Vec3f_,
	&g_M_Bool_,
	&g_M_Color_,
	&g_M_Float_,
	&g_M_Image_,
	&g_M_Int32_,
	&g_M_Node_,
	&g_M_Rotation_,
	&g_M_String_,
	&g_M_Time_,
	&g_M_Vec2f_,
	&g_M_Vec3f_,
	&file_,
	&floorHeight_,
	&questionMessage3Button_,
	&recentFiles_,
	&palette_,
	&commonSpec_,
	&freezeTime_,
	&settings_,
	&stretchCurve_,
	&openNetFile_,
	&demo_,
	&convertToPatch_,
	&reduceCapturedPoses_,
	&transformCP_,
	&timelineFilter_,
	&pasteOptions_,
	&deleteOptions_,
};
const char* constBuiltInNodeNameList[]={
	"NullNode",
	"Scene",
	"AvatarSocket",
	"Shape",
	"Part",
	"Appearance",
	"IndexedFaceSet",
	"ImageTexture",
	"Material",
	"NavigationInfo",
	"Group",
	"NoOutputGroup",
	"GestureGroup",
	"Mobility",
	"CCLathe",
	"CCBend",
	"SurfaceGenerator",
	"Stem",
	"Polymap",
	"BranchJoin",
	"Generator",
	"GeomSource",
	"SurfaceTransform",
	"Place",
	"DirectionalLight",
	"CCScale",
	"CCTranslate",
	"CCRingScale",
	"BezierLathe",
	"ECLathe",
	"ECScale",
	"ECShift",
	"ECBend",
	"ECSyncBend",
	"ECTwist",
	"Background",
	"Seamless",
	"GroupBuild",
	"SCGroupBuild",
	"JoinCloseVertex",
	"CenterJoint",
	"DelPoly",
	"JoinVertex",
	"CopyPart",
	"Viewpoint",
	"TransferBCyl",
	"TransferOwnership",
	"TransferVertex",
	"Coordinate",
	"Normal",
	"Color",
	"TextureCoordinate",
	"SineTug",
	"SimulSineTug",
	"GenPoly",
	"Script",
	"BranchJoinNegX",
	"BranchJoinPosX",
	"BranchJoinNegY",
	"BranchJoinPosY",
	"BranchJoinNegZ",
	"BranchJoinPosZ",
	"ProjectOntoTri",
	"ColorSweep",
	"EvolveCoord",
	"Function",
	"MemberFunction",
	"CodeLine",
	"MorphPart",
	"Morpher",
	"TextureTransform",
	"ProtoDeclare",
	"ProtoInterface",
	"ProtoBody",
	"ProtoInstance",
	"Movie",
	"Video",
	"Anim",
	"TimeSensor",
	"OrientationInterpolator",
	"PositionInterpolator",
	"NurbsTrack",
	"ScalarInterpolator",
	"PosCompensator",
	"Collision",
	"Transform",
	"ColorInterpolator",
	"CoordinateInterpolator",
	"NormalInterpolator",
	"ModelSequence",
	"ModelMorpher",
	"WorldInfo",
	"TouchSensor",
	"CCPath",
	"BspTree",
	"Inline",
	"Fog", 
	"Fog2",
	"Anchor",
	"AudioClip",
	"Billboard",
	"Box",
	"Cone",
	"Cylinder",
	"CylinderSensor",
	"ElevationGrid",
	"Extrusion",
	"FontStyle",
	"IndexedLineSet",
	"NurbsSurface",
	"LOD",
	"MovieTexture",
	"PixelTexture",
	"PlaneSensor",
	"PointLight",
	"PointSet",
	"ProximitySensor",
	"Sound",
	"Sphere",
	"SphereSensor",
	"SpotLight",
	"Switch",
	"Text",
	"VisibilitySensor",
	"TextureMapper",
	"TextureEffect",
	"ColorEffect",
	"MonoColorEffect",
	"ExitBuild",
	"SplitTri",
	"SwapTriDivide",
	"PlaceTexture",
	"HAnimHumanoid",
	"HAnimJoint",
	"HAnimDisplacer",
	"HAnimSegment",
	"HAnimSite",
	"BooleanFilter",
	"BooleanSequencer",
	"BooleanToggle",
	"BooleanTrigger",
	"IntegerSequencer",
	"IntegerTrigger",
	"TimeTrigger",
	"StaticGroup",
	"FloatAccelerator",
	"RotAccelerator",
	"CopyToMorph",
	"RelateMorpher",
	"ImmediatePart",
	"DelPlane",
	"DelSphere",
	"RadialTug",
	"TapeMeasure",
	"PosAccelerator",
	"AnimByAnkle",
	"AnimByWrist",
	"RotFilter",
	"RotMult",
	"MultiUser",
	"IndexedTriangleSet",
	"Polyline2D",
	"SphereBuild",
	"CylinderBuild",
	"ConeBuild",
	"TorusBuild",
	"BoxBuild",
	"RectangleBuild",
	"BezierLatheBuild",
	"MirrorEdit",
	"MirrorGroup",
	"PolyToNurbs",
	"RingBuild",
	"BezierScale",
	"BezierCircumScale",
	"BezierPatch",
	"Feedback",
	"PovLight",
	"PovFog",
	"BezierShift",
	"BezierTwist",
	"BezierTape",
	"BezierBend",
	"NurbsPatch",
	"NurbsLathe",
	"BendyLathe",
	"FuseSurface",
	"Subdivision",
	"Oscillator",
	"Envelope",
	"AmplitudeModulator",
	"FrequencyModulator",
	"Chords",
	"Larynx",
	"Phrase",
	"Harmonic",
	"HarmonicPatch",
	"Sample",
	"WaveToPoly",
	"Track",
	"Composition",
	"CPMorpher",
	"G_S_Bool",
	"G_S_Color",
	"G_S_Float",
	"G_S_Image",
	"G_S_Int32",
	"G_NodePtr",
	"G_S_Rotation",
	"G_S_String",
	"G_S_Time",
	"G_S_Vec2f",
	"G_S_Vec3f",
	"G_M_Bool",
	"G_M_Color",
	"G_M_Float",
	"G_M_Image",
	"G_M_Int32",
	"G_M_Node",
	"G_M_Rotation",
	"G_M_String",
	"G_M_Time",
	"G_M_Vec2f",
	"G_M_Vec3f",
	"File",
	"Floor Height",
	"Message from thyme's code",
	"Recent Files",
	"Palette",
	"CommonSpec",
	"FreezeTime",
	"Settings",
	"StretchCurve",
	"OpenNetFile",
	"Demo",
	"ConvertToPatch",
	"ReduceCapturedPoses",
	"Control Point Selection",
	"TimelineFilter",
	"PasteOptions",
	"DeleteOptions",
};

void DynamicMField::read2(Field *&field, Properties& sprop, bool& wasMF, const BIField& bi){
	if(bi.type==MFBOOL){
		wasMF=true;
		field=new MFBool(M_Bool());
	}
	else if(bi.type==MFCOLOR){
		wasMF=true;
		field=new MFColor(M_Color3f());
	}
	else if(bi.type==MFFLOAT){
		wasMF=true;
		field=new MFFloat(M_Float());
	}
	else if(bi.type==MFINT32){
		wasMF=true;
		field=new MFInt32(M_Int());
	}
	else if(bi.type==MFNODE){
		wasMF=true;
		field=new MFNode(M_Node());
	}
	else if(bi.type==MFROTATION){
		wasMF=true;
		field=new MFRotation(M_Rotation());
	}
	else if(bi.type==MFSTRING){
		wasMF=true;
		field=new MFString(M_String());
	}
	else if(bi.type==MFDOUBLE){
		wasMF=true;
		field=new MFTime(M_Double());
	}
	else if(bi.type==MFVEC2F){
		wasMF=true;
		field=new MFVec2f(M_Vec2f());
	}
	else if(bi.type==MFVEC3F){
		wasMF=true;
		field=new MFVec3f(M_Vec3f());
	}
	else if(bi.type==SFBOOL){
		field=new SFBool(false);
	}
	else if(bi.type==SFCOLOR){
		field=new SFColor(Color3f());
	}
	else if(bi.type==SFFLOAT){
		field=new SFFloat(0);
	}
	else if(bi.type==SFINT32){
		field=new SFInt32(0);
	}
	else if(bi.type==SFNODE){
		field=new SFNode(NodePtr());
	}
	else if(bi.type==SFROTATION){
		field=new SFRotation(Rotation());
	}
	else if(bi.type==SFSTRING){
		field=new SFString(String());
	}
	else if(bi.type==SFTIME){
		field=new SFTime(0);
	}
	else if(bi.type==SFVEC2F){
		field=new SFVec2f(Vec2f());
	}
	else if(bi.type==SFVEC3F){
		field=new SFVec3f(Vec3f());
	}
	else throw;

}
void DynamicMField::read(M_Properties& properties, const BIProto& binode, String& str){
	bool groupHasBegan=false;
	int groupCntr=-1;
	for(int k=0; k<binode.numOf; k++){
		Properties sprop;
		bool wasMF=false;
		const BIField* biPtr=binode.bifieldRef;
		const BIField& bi=biPtr[k];
		sprop.flags=bi.flags;
		sprop.promptNode.length=bi.initPromptNode.numOf;
		for(int j=0; j<bi.initPromptNode.numOf; j++){
			sprop.promptNode[j]=bi.initPromptNode.list[j];
		}
		str=bi.str;
		int i=0;
		sprop.name=str.getWord(i);

		Field* field=0;
		sprop.tip=bi.tip;

		read2(field, sprop, wasMF, bi);

		if(sprop.initializeOnly() || sprop.inputOutput()){
			field->read(str,i);
		}
		*this+=field;

		bool beginGroup=false;
		bool endGroup=false;
		str.stb(i);
		if(i<str.length){
			sprop.read(*field, str, i, beginGroup, endGroup);
		}
		if( wasMF&&(sprop.isReadable()^sprop.isWritable())==false&&field->getType()!=MFNODE){
			if(groupHasBegan==false&&beginGroup==false){
				sprop.groupIndex=++groupCntr;
			}
			else if(beginGroup){
				sprop.groupIndex=++groupCntr;
				groupHasBegan=true;
			}
			else if(groupHasBegan){
				sprop.groupIndex=groupCntr;
			}
			if(endGroup){
				groupHasBegan=false;
			}
		}
		properties+=sprop;
	}
}



M_ProtoInterface g_lookupProtoInterface;
extern String g_protoInterfaceListStr;
extern String g_protoInterfaceTips;
extern String g_protoInterfaceNodePromptStr;
void initLookup(){
///////////////

	static String str;
	nodeLookUpStr.length=0;
	const int numOfBuiltInNodes = sizeof(constBuiltInNodeNameList)/sizeof(const char*);
	for(int i=0; i<numOfBuiltInNodes; i++){
		nodeLookUpStr+=constBuiltInNodeNameList[i];
	}
	g_lookupProtoInterface.length=numOfBuiltInNodes;
	containerFieldLookUpStr.length=numOfBuiltInNodes;

	for(int k=0; k<numOfBuiltInNodes; k++){
		ProtoInterface* protoInterface=new ProtoInterface;
		g_lookupProtoInterface[k]=protoInterface;
		if(builtInProtoList[k]->containerField){
			containerFieldLookUpStr[k]=builtInProtoList[k]->containerField;
		}
		protoInterface->read(protoInterface->properties, *(builtInProtoList[k]), str);
	}
	g_userProtoBegin=g_lookupProtoInterface.length;


/////////////////
	/*
	str="";
	for(int i=0; i<g_lookupProtoInterface.length;i++){
		Node* node=g_lookupProtoInterface[i];
		node->name=nodeLookUpStr[i];
		node->name[0]+=0x20;
		node->name+="_";

		str+="static const BIProto";
		str+=node->getFLen();
		str+=' ';
		str+=node->name;
		str+=" = BIProto";
		str+=node->getFLen();
		str+="(\n\t\"";
		str+=nodeLookUpStr[i];
		str+="\", ";
		if(containerFieldLookUpStr[i]==""){
			str+="0";
		}
		else{
			str+='\"';
			str+=containerFieldLookUpStr[i];
			str+='\"';
		}
		if(node->getFLen()){
			str+=",\n";

			str+=node->toCPPString();
		}
		else{
			str+='\n';
		}
		str+=");\n";
	}
	bool store = g_writeGzipped;
	g_writeGzipped=false;
	str.write("proto_interface_data.cpp");
	g_protoInterfaceListStr.write("proto_interface_list.cpp");
	g_protoInterfaceTips.write("proto_interface_tips.cpp");
	g_protoInterfaceNodePromptStr.write("proto_interface_prompts.cpp");
	g_writeGzipped=store;
	*/
}
