/* Seamless3d

Copyright (C) 2001-2010 Graham John Perrett (AKA thyme) thyme@seamless3d.com www.seamless3d.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this
software and associated documentation files (the "Software"), to deal in the Software
without restriction, including without limitation the rights to use, copy, modify,
merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to the following
conditions:

The above copyright notice and this permission notice shall be included in all copies or
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT
OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.
*/


#ifdef DX8
#include <d3dx8.h>
#endif
#ifdef DX9
#include <d3dx9.h>
#endif

#include "seamless3d.h"

#include "techuelife_triangle_editor.h"
#include "gui.h"
#include "scene_tree.h"
#include "input.h"
#include "render_dx.h"
#include "seamless3d_gui_render.h"
#include "robot_aid.h"

#ifdef DX8
struct IDirect3DTexture8;
typedef IDirect3DTexture8 IDirect3DTexture;
typedef D3DLIGHT8 D3DLIGHT;
typedef D3DMATERIAL8 D3DMATERIAL;
typedef D3DVIEWPORT8 D3DVIEWPORT;
typedef D3DADAPTER_IDENTIFIER8 D3DADAPTER_IDENTIFIER;
typedef D3DCAPS8 D3DCAPS;

LPDIRECT3D8             g_pD3D       = NULL; // Used to create the D3DDevice
LPDIRECT3DDEVICE8       g_pd3dDevice = NULL; // Our rendering device
LPDIRECT3DVERTEXBUFFER8 g_pVB        = NULL; // Buffer to hold vertices
LPDIRECT3DINDEXBUFFER8  m_pIB		 = NULL;
IDirect3DSurface8* g_textureSurface = 0;
IDirect3DSurface8* g_pBackSurface = 0;
IDirect3DSurface8* g_offScreenSurface = 0;
LPD3DXRENDERTOSURFACE g_renderToSurface = NULL;
#endif
#ifdef DX9
struct IDirect3DTexture9;
typedef IDirect3DTexture9 IDirect3DTexture;
typedef D3DLIGHT9 D3DLIGHT;
typedef D3DMATERIAL9 D3DMATERIAL;
typedef D3DVIEWPORT9 D3DVIEWPORT;
typedef D3DADAPTER_IDENTIFIER9 D3DADAPTER_IDENTIFIER;
typedef D3DCAPS9 D3DCAPS;

LPDIRECT3D9             g_pD3D       = NULL; // Used to create the D3DDevice
LPDIRECT3DDEVICE9       g_pd3dDevice = NULL; // Our rendering device
LPDIRECT3DVERTEXBUFFER9 g_pVB        = NULL; // Buffer to hold vertices
LPDIRECT3DINDEXBUFFER9  m_pIB		 = NULL;
IDirect3DSurface9* g_pBackSurface=0;
#endif

extern double g_time;

//bool g_alphaEnable = false;
//bool g_alphaTexture = false;
bool g_backfaceRender = false;
bool g_wireFrameMode=true;
bool g_windowSizeWindowBool=false;
extern bool g_minimized;
extern Group* g_vcSelectRectGroup;
double g_windowSizeWindowTime=0;

extern bool g_multiUserMode;
extern bool g_browseMode;
extern bool g_screenshotEarly;
extern bool g_screenshotLate;

extern Pix24* g_frameBuffer;

extern TransformPtr g_vcTransferBCylBase;

DXRender* g_redRenderPoint=0;
DXRender* g_lastRenderPoint=0;
DXRender* g_nextRenderPoint=0;
DXRender* g_yellowRenderPoint=0;

extern IndexedLineSet* g_triClue;
extern Coordinate* g_triClueCoord;
extern Color* g_triClueColor;
extern Shape* g_triClueShape;

//extern Ptr<ControlPanel> g_rcp;


D3DSURFACE_DESC	g_surfaceBuffer;
D3DSURFACE_DESC	m_d3dsdBackBuffer;
D3DLIGHT g_light;
BOOL g_headlight=true;
void getInput();

D3DXMATRIX matProj;
D3DXMATRIX matView;

ID3DXFont* ppFont=0;

struct RenderLast{
	Seamless* smls;
	Matrix matrix;
	Vec3f parentColScale;
	RenderLast(Seamless* smls, Matrix matrix, Vec3f parentColScale):
		smls(smls),
		matrix(matrix),
		parentColScale(parentColScale)
	{}
	RenderLast():
		smls(0)
	{}

};

struct TransparencyList{
	Node* node;
	Matrix matrix;
	Vec3f parentColScale;
	TransparencyList(Node* node, Matrix matrix, Vec3f parentColScale):
		node(node),
		matrix(matrix),
		parentColScale(parentColScale)
	{}
	TransparencyList():
		node(0)
	{}

};

struct AlphaChanList{
	Node* node;
	Matrix matrix;
	Vec3f parentColScale;
	AlphaChanList(Node* node, Matrix matrix, Vec3f parentColScale):
		node(node),
		matrix(matrix),
		parentColScale(parentColScale)
	{}
	AlphaChanList():
		node(0)
	{}

};


M_<RenderLast> g_skelList;
M_<TransparencyList> g_transparentList;
M_<AlphaChanList> g_alphaChanList;
bool g_renderLater = false;

M_Node renderProcessedQue;

D3DPRESENT_PARAMETERS* g_d3dpp=0;

extern Ptr<PartClue> g_vcArtTriBase;
extern TransformPtr g_vcOscBase;
CommonSpec* getCommonSpec(const Art* part);
extern bool g_recordSoundMovie;

extern IndexedLineSet* g_nurbsIls;
extern Coordinate* g_nurbsIlsCoord;
extern Color* g_nurbsIlsColor;
extern Shape* g_nurbsIlsShape;
extern TransformPtr g_nurbsIlsBase;

bool backFace(const Vec3f& v0, const Vec3f& v1, const Vec3f& v2);
extern M_Matrix m_matrix;
extern M_Int m_smi;
extern M_Vec3f g_m_coord;

//-----------------------------------------------------------------------------
// Global variables
//-----------------------------------------------------------------------------


DWORD        m_dwNumIndices=6;

HRESULT InitD3D( HWND hWnd );
/*
HRESULT resizeBackBuffer(){
	RECT rect;
	GetClientRect(hWndMain, &rect);
	g_d3dpp->BackBufferWidth=rect.right;
	g_d3dpp->BackBufferHeight=rect.bottom;
	HRESULT hresult=g_pd3dDevice->Reset(g_d3dpp);

	if(hresult==S_OK){
		return S_OK;
	}
	else if(hresult==D3DERR_DEVICELOST){
		hresult=g_pd3dDevice->Reset(g_d3dpp);

		if(hresult==S_OK){
			return S_OK;
		}
		else if(hresult==D3DERR_DEVICELOST){
			return S_OK;
		}
	}

    // Turn on the zbuffer
    g_pd3dDevice->SetRenderState( D3DRS_ZENABLE, TRUE );
	
	if(g_settings->backfaceIgnore.refValue()){
		g_pd3dDevice->SetRenderState( D3DRS_CULLMODE, D3DCULL_CCW );
	}
	else{
		g_pd3dDevice->SetRenderState( D3DRS_CULLMODE, D3DCULL_NONE );
	}

	return S_OK;
}
*/
void resetImageTextures(){
	for(int i=g_sceneTreeList.length; --i>=0;){
		if(ImageTexture* imageTexture=dynamic_cast<ImageTexture*>(g_sceneTreeList[i].node.p)){
			if(imageTexture->dxTexture){
				imageTexture->dxTexture->Release();
				imageTexture->dxTexture=0;
			}
			imageTexture->update=true;
		}
		else if(TextureEffect* textureEffect=dynamic_cast<TextureEffect*>(g_sceneTreeList[i].node.p)){
			if(textureEffect->dxTexture){
				textureEffect->dxTexture->Release();
				textureEffect->dxTexture=0;
			}
			textureEffect->url.update=true;
		}
		else if(Shape* shape=dynamic_cast<Shape*>(g_sceneTreeList[i].node.p)){
			DXRender*& render=shape->render;
			if(render != NULL){
				if( render->pVB != NULL ){
					render->pVB->Release();
				}
				if( render->pIB != NULL ){
					render->pIB->Release();
				}
			}
			delete render;
			render=0;

			if(IndexedFaceSet* indexedFaceSet=dynamic_cast<IndexedFaceSet*>(shape->geometry.p)){
				if(Coordinate* crd=dynamic_cast<Coordinate*>(indexedFaceSet->coord.p)){
					crd->renderProcessed=false;
				}
			}
		}
	}
}

bool windowSizeHasChanged(){
	static bool flag=false;
	RECT rect;
	GetClientRect(
	  hWndMain,      // handle to window
	  &rect   // address of structure for client coordinates
	);
	if((m_d3dsdBackBuffer.Width!=(unsigned)rect.right||m_d3dsdBackBuffer.Height!=(unsigned)rect.bottom)&&flag==false){
		flag=true;
//		msG("Close down Seamless3d and reopen it for the new window size to properly take effect");
//		msG("Seamless3d is about update settings for this 3d window.\n If strange behavour is observed close down Seamless3d and reopen it for the new window size to properly take effect and plase send a description to thyme@seamless3d.com of what happened");

		cleanupForRefreshSize();

		InitD3D(hWndMain);

		g_stw->checkWinVis();
		g_stw->cp->checkWinVis();

		resetImageTextures();

		//POINT dstPoint;
		//dstPoint.x=g_settings->toolBarPos.refValue().x,
		//dstPoint.y=g_settings->toolBarPos.refValue().y;
		//if(dstPoint.x <= TOOLBAR_CX || dstPoint.x >= static_cast<int>(m_d3dsdBackBuffer.Width) || dstPoint.y <= TOOLBAR_CY_ALL || dstPoint.y >= static_cast<int>(m_d3dsdBackBuffer.Height)){
		g_tb->pos=Vec2i(0, rect.bottom - TOOLBAR_CY_ALL);
			//g_settings->toolBarPos=Vec2f(0, 0);
		//}

		flag=false;
		return true;
	}
	return false;
}

//-----------------------------------------------------------------------------
// Name: Cleanup()
// Desc: Releases all previously initialized objects
//-----------------------------------------------------------------------------
void cleanupForRefreshSize(){
	g_remMenu=0;

	Window::releaseAllControlPanels();

	g_stw->cp->releaseWindow();
	g_stw->releaseWindow();
	g_tb->releaseWindow();
	g_tt->releaseWindow();
	g_msg->releaseWindow();
	g_cw->releaseWindow();

	if(ppFont){
		ppFont->Release();
	}
	if(g_pd3dDevice){
        g_pd3dDevice->Release();
	}
	if(g_pD3D){
        g_pD3D->Release();
	}
}
//-----------------------------------------------------------------------------
// Name: Cleanup()
// Desc: Releases all previously initialized objects
//-----------------------------------------------------------------------------
void Cleanup(){
	Window::releaseAllControlPanels();

	g_remMenu=0;
	g_stw=0;
	g_tb=0;
	g_tt=0;
	g_msg=0;
	g_cw=0;
//	g_rcp=0;

	if( ppFont != NULL){
		ppFont->Release();
	}
	if( g_pd3dDevice != NULL ){
        g_pd3dDevice->Release();
	}
	if( g_pD3D != NULL ){
        g_pD3D->Release();
	}
}

void SetupMatrices(){
	
    // Set up our view matrix. A view matrix can be defined given an eye point,
    // a point to lookat, and a direction for which way is up. Here, we set the
    // eye five units back along the z-axis and up three units, look at the
    // origin, and define "up" to be in the y-direction.
    
    D3DXMatrixLookAtRH( &matView, 
		&D3DXVECTOR3( 0, 0, g_viewDist),
		&D3DXVECTOR3( 0.0f, 0.0f, 0.0f),
		&D3DXVECTOR3( 0.0f, 1.0f, 0.0f ) );
    g_pd3dDevice->SetTransform( D3DTS_VIEW, &matView );

	//D3DXMatrixLookAtRH( &matView, 
	//	&D3DXVECTOR3( -44.064507,1.872501,19.52692),
	//	&D3DXVECTOR3( -42.770695,1.872501,16.820251),
	//	&D3DXVECTOR3( 0.0f, 1.0f, 0.0f ) );
 //   g_pd3dDevice->SetTransform( D3DTS_VIEW, &matView );
	
	
//	g_pd3dDevice->GetBackBuffer( 0, D3DBACKBUFFER_TYPE_MONO, &g_pBackSurface);
    g_pBackSurface->GetDesc( &m_d3dsdBackBuffer );
    //g_pBackSurface->Release();
	
	
	// For the projection matrix, we set up a perspective transform (which
    // transforms geometry from 3D view space to 2D viewport space, with
    // a perspective divide making objects smaller in the distance). To build
    // a perpsective transform, we need the field of view (1/4 pi is common),
    // the aspect ratio, and the near and far clipping planes (which define at
    // what distances geometry should be no longer be rendered).
	
	
	if(g_settings->perspective){
		D3DXMatrixPerspectiveFovRH( &matProj, g_scene->fieldOfView, (float)m_d3dsdBackBuffer.Width/(float)m_d3dsdBackBuffer.Height, 1, 400.0f );
		//D3DXMatrixPerspectiveFovRH( &matProj, g_scene->fieldOfView, 1, 1, 400.0f );
	}
	else{
		D3DXMatrixOrthoRH( &matProj, 4, 4.*((float)m_d3dsdBackBuffer.Height/(float)m_d3dsdBackBuffer.Width), 1, 400.0f );
	}
	
	g_pd3dDevice->SetTransform( D3DTS_PROJECTION, &matProj );
}

//-----------------------------------------------------------------------------
// Name: SetupLights()
// Desc: Sets up the lights and materials for the scene.
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
// Name: SetupLights()
// Desc: Sets up the lights and materials for the scene.
//-----------------------------------------------------------------------------
VOID SetupLights(){
    // Set up a material. The material here just has the diffuse and ambient
    // colors set to yellow. Note that only one material can be used at a time.
    D3DMATERIAL mtrl;
    ZeroMemory( &mtrl, sizeof(D3DMATERIAL) );
    mtrl.Diffuse.r = mtrl.Ambient.r = 1.0f;
    mtrl.Diffuse.g = mtrl.Ambient.g = 1.0f;
    mtrl.Diffuse.b = mtrl.Ambient.b = 1.0f;
    mtrl.Diffuse.a = mtrl.Ambient.a = 1.0f;
	mtrl.Specular.r=1;
	mtrl.Specular.g=1;
	mtrl.Specular.b=1;
	mtrl.Specular.a=1;
	mtrl.Power=.5;
    


	//mtrl.Emissive.r = .3;
	//mtrl.Emissive.g = .3;
	//mtrl.Emissive.b = .3;
    g_pd3dDevice->SetMaterial( &mtrl );
	
    // Set up a white, directional light, with an oscillating direction.
    // Note that many lights may be active at a time (but each one slows down
    // the rendering of our scene). However, here we are just using one. Also,
    // we need to set the D3DRS_LIGHTING renderstate to enable lighting
    g_pd3dDevice->LightEnable( 0, TRUE );
    g_pd3dDevice->SetRenderState( D3DRS_LIGHTING, TRUE );
	
    // Finally, turn on some ambient light.
    g_pd3dDevice->SetRenderState( D3DRS_AMBIENT, 0x00000000 );
	//g_pd3dDevice->SetRenderState(D3DRS_COLORWRITEENABLE  , D3DCOLORWRITEENABLE_RED|D3DCOLORWRITEENABLE_GREEN|D3DCOLORWRITEENABLE_BLUE |D3DCOLORWRITEENABLE_ALPHA );

//	g_pd3dDevice->SetRenderState( D3DRS_ALPHABLENDENABLE    , TRUE );
//	g_pd3dDevice->SetRenderState(D3DRS_SRCBLEND, D3DBLEND_SRCALPHA );
//	g_pd3dDevice->SetRenderState(D3DRS_DESTBLEND, D3DBLEND_INVSRCALPHA );

}

void outputText( DWORD x, DWORD y, char* str, unsigned fontColor){
	RECT rect;
	int strLen=strlen(str);
	rect.left=x;
	rect.top=y;
	rect.right=x+strLen*10;
	rect.bottom=y+18;
#ifdef DX8
	ppFont->DrawText(str, strLen, &rect, 0, fontColor);
#endif
#ifdef DX9
	ppFont->DrawText(0, str, strLen, &rect, 0, fontColor);
#endif
	
}
void outputColorText(DWORD x, DWORD y, char* str ,Color3f color,unsigned defaultFontColor){
	RECT rect;
	outputText( x, y, str ,defaultFontColor);
	char buf[5];
	x+=strlen(str)*10;
	sprintf(buf,"%.02f", color.r);
	rect.left=x;
	rect.top=y;
	rect.right=x+4*10;
	rect.bottom=y+18;
#ifdef DX8
	ppFont->DrawText(buf, 4, &rect, 0, 0xffff2020);
#endif
#ifdef DX9
	ppFont->DrawText(0, buf, 4, &rect, 0, 0xffff2020);
#endif

	x+=5*10;
	sprintf(buf,"%.02f", color.g);
	rect.left=x;
	rect.top=y;
	rect.right=x+4*10;
	rect.bottom=y+18;

#ifdef DX8
	ppFont->DrawText(buf, 4, &rect, 0, 0xff20ff20);
#endif
#ifdef DX9
	ppFont->DrawText(0, buf, 4, &rect, 0, 0xff20ff20);
#endif

	x+=5*10;
	sprintf(buf,"%.02f", color.b);
	rect.left=x;
	rect.top=y;
	rect.right=x+4*10;
	rect.bottom=y+18;
#ifdef DX8
	ppFont->DrawText(buf, 4, &rect, 0, 0xff6060ff);
#endif
#ifdef DX9
	ppFont->DrawText(0, buf, 4, &rect, 0, 0xff6060ff);
#endif
}

extern unsigned g_state;

void AppShowStats(){
	if(g_remMenu && g_remMenu->pos==Vec2i()){
		return;
	}

//	ppFont->Begin();
    static FLOAT fFPS      = 0.0f;
    static FLOAT fLastTime = 0.0f;
    static DWORD dwFrames  = 0L;
	
	// Keep track of the time lapse and frame count
	++dwFrames;
	
	// Update the frame rate once per second
	if( g_time - fLastTime > 1.0f )
    {
        fFPS      = dwFrames / (g_time - fLastTime);
        fLastTime = g_time;
        dwFrames  = 0L;
    }
	unsigned fontColor=g_scene->scene3DFontColor.refValue().colorToHex();
	
    // Setup the text buffer to write out
    CHAR buffer[0x200];
	
    sprintf( buffer, "FPS %7.02f" , fFPS,fontColor);
    outputText(  0, 0, buffer ,fontColor);
	float tugRadius=g_settings->tugRadius.refValue();
	if(tugRadius>0){
		sprintf( buffer, "tugRadius %7.03f" ,tugRadius );
		outputText( 0, 200, buffer ,fontColor);
	}
	if(g_dragActuallyHappned){
		sprintf( buffer, "tug");
		outputText( 0, 220, buffer, fontColor);
	}

	/*
	if(mouse3dMode){
		outputText( 0, 160, "mouse3dMode TRUE"  );
	}
	else {
		outputText( 0, 160, "mouse3dMode FALSE"  );
	}
	if(g_MousebActive){
		outputText( 0, 180, "g_MousebActive TRUE"  );
	}
	else {
		outputText( 0, 180, "g_MousebActive FALSE"  );
	}
		
	if(DIERR_INPUTLOST==g_state){
		outputText( 0, 280, "DIERR_INPUTLOST");
	}
	else if(DIERR_INVALIDPARAM ==g_state){
		outputText( 0, 280, "DIERR_INVALIDPARAM ");
	}
	else if(DIERR_NOTACQUIRED==g_state){
		outputText( 0, 280, "DIERR_NOTACQUIRED");
	}
	else if(DIERR_NOTINITIALIZED==g_state){
		outputText( 0, 280, "DIERR_NOTINITIALIZED");
	}
	else if(E_PENDING ==g_state){
		outputText( 0, 280, "E_PENDING ");
	}
	else {
		outputText( 0, 280, "whate ever ");
	}
*/
	Art* part=dynamic_cast<Art*>(g_selNode.getRaw());
	if(part){
		Seamless* smls = part->smls;
		if(smls->workAnimCP){
			return;
		}
		if(smls->actTriPart==0){
			//ppFont->End();
			return;
		}


		M_Triangle& triangle= smls->actTriPart->triangle;

		const M_Vec3f& coord=part->coord;
		//const M_Vec3f& normal=part->normal;
		if(smls->ActCtrlPnt >= coord.length){
			return;
		}


		if(part->coord.length){
			if(part->workEffect && smls->actTriPart){
				ColorEffect* colorEffect= dynamic_cast<ColorEffect*>(part->workEffect);
				if(colorEffect!=0){
					const M_Color3f& color=colorEffect->color;
					int ci=triangle.refTexIndex(smls->ActTri*3+smls->ActTriCorner);
					if(ci<color.length){
						outputColorText(0, 20, "color", color[ci], fontColor);
					}
				}
			}
			//Vec3f spoint=normal[part->smls->ActCtrlPnt];
			//sprintf( buffer, "normal %7.03f %7.03f %7.03f" ,spoint.x,spoint.y,spoint.z );
			Vec3f spoint=part->matrix*coord[part->smls->ActCtrlPnt];
			sprintf(buffer, "coord %7.03f %7.03f %7.03f" ,spoint.x,spoint.y,spoint.z );
			outputText(0, 40, buffer, fontColor);

			sprintf(buffer, "coordIndex %i" ,smls->ActCtrlPnt);
			outputText(0, 60, buffer, fontColor);
			
			static String str;
			str="triangle ";
			str+=smls->actTriPart->name;
			str+=": %i";
			sprintf( buffer, str.c_str() ,smls->ActTri);
			outputText( 0, 80, buffer, fontColor);
			
			char Corner=part->smls->ActTriCorner;
			Corner+='A';
			strcpy(buffer,"triangleCorner A");
			buffer[15]=Corner;
			outputText( 0, 100, buffer, fontColor);
			
/*
			if(g_localControl){
				sprintf( buffer, "transformMode");
				outputText( 0, 240, buffer, fontColor);
			}
*/			
		}
	}

//	ppFont->End();
}

void callBindScript();

inline void manageDownload(){
	if(g_downloadQue.length){
		if(g_downloadQue.length){
			if(g_downloadQue[0]->download()){
				for(int i=0; i<g_downloadQue.length-1; i++){
					g_downloadQue[i]=g_downloadQue[i+1];
				}
				g_downloadQue.length--;
			}
		}
	}
}
void checkKeys();
void Window::render(){
	RECT srcRect;
	srcRect.top=0;
	srcRect.left=0;
	srcRect.right=size.x;
	srcRect.bottom=size.y;

	POINT dstPoint={pos.x, pos.y};
	if(dstPoint.x<0){
		srcRect.left=-dstPoint.x;
		dstPoint.x=0;
	}
	if(dstPoint.x + size.x >= static_cast<int>(m_d3dsdBackBuffer.Width)){
		srcRect.right=size.x - ((dstPoint.x + size.x) - m_d3dsdBackBuffer.Width);
	}
	if(dstPoint.y<0){
		srcRect.top=-dstPoint.y;
		dstPoint.y=0;
	}
	if(dstPoint.y + size.y >= static_cast<int>(m_d3dsdBackBuffer.Height)){
		srcRect.bottom=size.y - ((dstPoint.y + size.y) - m_d3dsdBackBuffer.Height);
	}

	if(workSurface==0){
		initWindow();
	}
#ifdef DX8
	g_pd3dDevice->CopyRects(workSurface, &srcRect, 1, g_pBackSurface, &dstPoint);
#endif
#ifdef DX9
	g_pd3dDevice->UpdateSurface(workSurface, &srcRect, g_pBackSurface, &dstPoint);
#endif
}

// =====================================================================================
// Function Name: validateDevice
//		 Purpose: Ensures we have control of the display adapter in the multitasking
//				  win32 environment. adapted from 32Bits.co.uk
// =====================================================================================

inline HRESULT validateDevice(){
	/*
	HRESULT rslt=0;
	// Test the current state of the device
	rslt=g_pd3dDevice->TestCooperativeLevel();
	if(FAILED(rslt)){
		// if the device is lost, then return a fail
		if(rslt==D3DERR_DEVICELOST){
			return E_FAIL;
		}

		// if the device is ready to be reset, then try it
		if(rslt==D3DERR_DEVICENOTRESET){
			// Release the back surface so it can be re-created
			if(g_pBackSurface){
				g_pBackSurface->Release();
			}
			
			// reset the device
			rslt=g_pd3dDevice->Reset(g_d3dpp);
			if(FAILED(rslt)){
				// If the device was not reset, then exit
				PostQuitMessage(0);
				return E_FAIL;
			}

			// reacquire a pointer to the new back buffer
			rslt=g_pd3dDevice->GetBackBuffer(0,D3DBACKBUFFER_TYPE_MONO, &g_pBackSurface);
			if(FAILED(rslt)){
				PostQuitMessage(0);
				return E_FAIL;
			}
		}

	}
	*/
	return S_OK;
}


unsigned onCursorImage[]={
	0xff0000,
	0,
	0,
	0,
	0,
	0,
	0,
	0,
	0,
	0,
	0,
	0,
	0,
	0,
	0,
	0,
	0xff0000,
};
unsigned onCursorImageInverted[]={
	0x00ffff,
	0xffffff,
	0xffffff,
	0xffffff,
	0xffffff,
	0xffffff,
	0xffffff,
	0xffffff,
	0xffffff,
	0xffffff,
	0xffffff,
	0xffffff,
	0xffffff,
	0xffffff,
	0xffffff,
	0xffffff,
	0x00ffff,
};
#define ON_CURSOR_IMAGE_LEN (sizeof(onCursorImage)/sizeof(unsigned))
unsigned offCursorImage[ON_CURSOR_IMAGE_LEN];
inline void CaretWindow::doWinkOff(const Vec2i& c, RECT& srcRect)const{
	if(static_cast<unsigned>(c.x) < static_cast<unsigned>(win.size.x) && static_cast<unsigned>(c.y+ON_CURSOR_IMAGE_LEN) < static_cast<unsigned>(win.size.y)){

		const int* p=win.pixBuf.getArray();
		for(int i=ON_CURSOR_IMAGE_LEN; --i>=0;){
			offCursorImage[i]=p[(c.y+i)*win.size.x+c.x];
		}

		RECT dstRect;
		dstRect.left=c.x;
		dstRect.top=c.y;
		dstRect.right=c.x+1;
		dstRect.bottom=c.y+ON_CURSOR_IMAGE_LEN;


		D3DXLoadSurfaceFromMemory(
			win.workSurface,
			0,
			&dstRect,
			offCursorImage,
			g_d3dpp->BackBufferFormat,
			srcRect.right*4,
			0,
			&srcRect,
			D3DX_DEFAULT,
			0x0
		);
	}
}
inline void CaretWindow::doWinkOn(const Vec2i& c, RECT& srcRect)const{
	if(static_cast<unsigned>(c.x) < static_cast<unsigned>(win.size.x) && static_cast<unsigned>(c.y+ON_CURSOR_IMAGE_LEN) < static_cast<unsigned>(win.size.y)){
		RECT dstRect;
		dstRect.left=c.x;
		dstRect.top=c.y;
		dstRect.right=c.x+1;
		dstRect.bottom=c.y+ON_CURSOR_IMAGE_LEN;

		unsigned * oCI = onCursorImage;
		if(g_settings->invertWindowColor){
			oCI = onCursorImageInverted;
		}

		D3DXLoadSurfaceFromMemory(
			win.workSurface,
			0,
			&dstRect,
			oCI,
			g_d3dpp->BackBufferFormat,
			srcRect.right*4,
			0,
			&srcRect,
			D3DX_DEFAULT,
			0x0
		);
	}
}
void CaretWindow::doCallBack(){//renderCaret
	if(win.workSurface==0){
		return;
	}

	RECT srcRect;
	srcRect.left=0;
	srcRect.top=0;
	srcRect.right=1;
	srcRect.bottom=ON_CURSOR_IMAGE_LEN;

	Vec2i currentPos=getCaretPos();
	//if(remPreviousPos!=currentPos || firstVisChar!=remFirstChar){
	if((remPreviousPos!=currentPos || firstVisChar!=remFirstChar)  && remPreviousPos.y >= 0){//remPreviousPos.y >= 0 seems to stop crashes when clicking on scene tree node
		winkOn=true;
		lastWinkTime=g_time;
		doWinkOff(remPreviousPos, srcRect);
		doWinkOn(currentPos, srcRect);

	}
	else if(g_time > lastWinkTime+.5){
		lastWinkTime=g_time;
		winkOn=!winkOn;
		if(winkOn){
			doWinkOn(currentPos, srcRect);
		}
		else{
			doWinkOff(currentPos, srcRect);
		}
	}
	remPreviousPos=currentPos;
	remFirstChar=firstVisChar;
}
void CaretWindow::doNoThinkWink(){
	RECT srcRect;
	srcRect.left=0;
	srcRect.top=0;
	srcRect.right=1;
	srcRect.bottom=ON_CURSOR_IMAGE_LEN;

	Vec2i currentPos=getCaretPos();
	if(winkOn){
		doWinkOn(currentPos, srcRect);
	}
	else{
		doWinkOff(currentPos, srcRect);
	}
}
void CaretWindow::doNoThinkWinkOff(){
	RECT srcRect;
	srcRect.left=0;
	srcRect.top=0;
	srcRect.right=1;
	srcRect.bottom=ON_CURSOR_IMAGE_LEN;

	Vec2i currentPos=getCaretPos();
	doWinkOff(currentPos, srcRect);
}
/*
inline HRESULT checkDXDeviceIsValid(){
	HRESULT  result=g_pd3dDevice->TestCooperativeLevel();
	if(FAILED(result)){
		dLog("TestCooperativeLevel() failed");
		cleanupForRefreshSize();

		InitD3D(hWndMain);
		resetImageTextures();
	}
	return S_OK;
}
http://www.seamless3d.com/wrl/techuelife_island/gm_tree_s9b.smls
*/

int g_waitMiliseconds = 0;
void Settings::FreeCPU::operator=(const int& par){
	raw = par;
	if(raw>100){
		raw=100;
	}
	if(raw<0){
		raw = 0;
	}
	g_waitMiliseconds = raw;
}

void scriptBindAnimation();
void doChat();
extern bool g_loggedIn;
bool tugFlag2=false;
extern bool g_debugFlag;
extern bool g_fogExist;

void saveScreenshotToMem(unsigned int* pBits, unsigned int screenWidth, unsigned int width, unsigned int height, Pix24* lpDest);
void saveFrameToMem(unsigned int* pBits, unsigned int screenWidth, unsigned int width, unsigned int height);
void updateBuffer();

void outputScreenShotForDemo1();
void outputScreenShotForDemo2();

void renderScene(){
	bool fogExist = g_fogExist;
	g_fogExist=false;

	if(g_minimized){
		return;
	}
	/*
	if(!g_pd3dDevice || E_FAIL==checkDXDeviceIsValid()){
		return;
	}
*/
	if(g_windowSizeWindowBool && g_windowSizeWindowTime+1<g_time){
		windowSizeHasChanged();
		g_windowSizeWindowBool=false;
		if(g_pd3dDevice==0){
			g_windowSizeWindowBool=false;
		}
	}


	if(g_multiUserMode){
		checkKeys();
	}
	manageDownload();

	if(g_loggedIn){
		doChat();
	}

#ifdef ROBOT_SYNC
	InputEvent::runRobotSequence();
#endif

	renderProcessedQue.length=0;

#ifdef FRAME_TIME_SYNC
	g_time += FRAME_TIME_SYNC;
#else
	g_time = timeGetTime() * MILISECONDS_TO_SECONDS;
#endif
	getInput();
	if(g_wireFrameMode){
		g_pd3dDevice->SetRenderState( D3DRS_FILLMODE, D3DFILL_WIREFRAME );
	}
	else{
		g_pd3dDevice->SetRenderState( D3DRS_FILLMODE, D3DFILL_SOLID );
	}
	Seamless* smls=0;
	Art* part=dynamic_cast<Art*>(g_selNode.getRaw());
	if(part){
		smls=part->smls;

		if(smls->animCP == false){
			if(tugFlag || tugFlag!=tugFlag2){
				if(dynamic_cast<AnimClue*>(g_visualClue)){
					//smls->morphFlag = true;
				}
				else{
					smls->updateAllVertexInBuffers();
				}
			}
			tugFlag2 = tugFlag;
		}
	}

#ifdef DX8
	if(D3D_OK==g_pd3dDevice->GetBackBuffer(0, D3DBACKBUFFER_TYPE_MONO, &g_pBackSurface)){
#endif
#ifdef DX9
	if(D3D_OK==g_pd3dDevice->GetBackBuffer( 0, 0, D3DBACKBUFFER_TYPE_MONO, &g_pBackSurface)){
#endif
		// Clear the backbuffer and the zbuffer
		g_pd3dDevice->Clear( 0, NULL, D3DCLEAR_TARGET|D3DCLEAR_ZBUFFER,	g_scene->backgroundColor.refValue().colorToHex() , 1.0f, 0 );
		
		// Begin the scene
		g_pd3dDevice->BeginScene();
		
		if(S_OK==validateDevice()){
		
	
			// Setup the lights and materials
			SetupLights();
			
			// Setup the world, view, and projection matrices
			SetupMatrices();
			
		//	Matrix viewMatrix=translationMatrix(g_scene->translation) * rotationMatrix(g_scene->rotation) * scaleMatrix(g_scene->scale);
		//	Matrix viewMatrix=translationMatrix(g_scene->translation) * rotationMatrix(g_scene->rotation) * scaleMatrix(g_scene->scale);

			Matrix viewMatrix = g_scene->getViewMatrix();

			//Matrix viewMatrix=scaleMatrix(g_scene->scale) * translationMatrix(g_scene->translation) * rotationMatrix(g_scene->rotation);


			{
				for(int i = 0; i < g_animBarList.length; i++){
					g_animBarList[i]->renderNodeFirst();
				}
			}

			M_Node& children=g_scene->children;
			float zoom=g_scene->getZoom();
			Vec3f scale=Vec3f(zoom, zoom, zoom);
			{
				for(int i = 0; i < children.length; i++){
					children[i]->renderNode(viewMatrix, scale);
				}
			}

			//if(g_wireFrameMode){
			//	g_pd3dDevice->SetRenderState( D3DRS_FILLMODE, D3DFILL_SOLID );
			//}

			if(g_screenshotEarly){
				D3DLOCKED_RECT lockRect;

				HRESULT rs = g_pBackSurface->LockRect(&lockRect, 0, D3DLOCK_READONLY);
				if(D3D_OK != rs){
					throw;
				}
				unsigned int screenWidth = lockRect.Pitch / 4;
				unsigned int width = m_d3dsdBackBuffer.Width;
				unsigned int height = m_d3dsdBackBuffer.Height;
				
				unsigned int* pBits = reinterpret_cast<unsigned int*>(lockRect.pBits);

				if(g_recordSoundMovie){
					updateBuffer();
					saveFrameToMem(pBits, screenWidth, width, height);
				}
				else{
					saveScreenshotToMem(pBits, screenWidth, width, height, g_frameBuffer);
				}
				g_pBackSurface->UnlockRect();
			}

			if(g_wireFrameMode){
				if(g_screenshotEarly == false){
					g_pd3dDevice->SetRenderState( D3DRS_FILLMODE, D3DFILL_SOLID );
				}
			}
			if(g_visualClue){
				if(g_visualClue == g_vcOscBase){
					g_pd3dDevice->SetRenderState( D3DRS_FILLMODE, D3DFILL_WIREFRAME );
					g_visualClue->renderNode(viewMatrix, scale);
					g_pd3dDevice->SetRenderState( D3DRS_FILLMODE, D3DFILL_SOLID );
				}
				else{
					g_visualClue->renderNode(viewMatrix, scale);
				}
			}
			if(g_transparentList.length){
				g_pd3dDevice->SetRenderState(D3DRS_ALPHABLENDENABLE, TRUE);
				g_renderLater = true;
				for(int k = g_transparentList.length; --k>=0;){
					TransparencyList& rl = g_transparentList[k];
					rl.node->renderNode(rl.matrix, rl.parentColScale);
				}
				g_renderLater = false;
				g_transparentList.length = 0;
				g_pd3dDevice->SetRenderState(D3DRS_ALPHABLENDENABLE, FALSE);
			}
			if(g_alphaChanList.length){
				g_pd3dDevice->SetRenderState(D3DRS_ALPHABLENDENABLE, TRUE);
				g_pd3dDevice->SetTextureStageState( 0, D3DTSS_ALPHAOP, D3DTA_TEXTURE);
				g_renderLater = true;
				for(int k = g_alphaChanList.length; --k>=0;){
					AlphaChanList& rl = g_alphaChanList[k];
					rl.node->renderNode(rl.matrix, rl.parentColScale);
				}
				g_renderLater = false;
				g_alphaChanList.length = 0;
				g_pd3dDevice->SetRenderState(D3DRS_ALPHABLENDENABLE, FALSE);
				g_pd3dDevice->SetTextureStageState( 0, D3DTSS_ALPHAOP, D3DTOP_DISABLE);
			}
			if(g_skelList.length){
				g_pd3dDevice->SetRenderState( D3DRS_ZENABLE, FALSE);
				for(int k = g_skelList.length; --k>=0;){
					RenderLast& rl = g_skelList[k];
					rl.smls->stickMan->renderStickMan(rl.smls, rl.matrix, rl.parentColScale);
				}
				g_skelList.length = 0;
				g_pd3dDevice->SetRenderState( D3DRS_ZENABLE, TRUE);
			}
			{
				for(int i=renderProcessedQue.length; --i>=0;){
					renderProcessedQue[i]->renderProcessed=true;
				}
			}
			if(g_wireFrameMode){
				g_pd3dDevice->SetRenderState( D3DRS_FILLMODE, D3DFILL_WIREFRAME );
			}

			//g_stw->render();
			//g_pw->render();
			//g_tb->render();
			Window::renderVisible();

			//g_activeWindow->renderCaret();

		
			for(int i=0; i<g_callBackList.length; i++){
				g_callBackList[i]->doCallBack();
			}
			//if(g_acp && g_acp->visible){
			if(g_acp && g_acp->visible){
				g_acp->paint();
			}

			
		//    g_pd3dDevice->EndScene(); g_pd3dDevice->Present( NULL, NULL, NULL, NULL );return;

			AppShowStats();
			
#ifdef DEMO_TO_MOVIE
outputScreenShotForDemo1();
#endif

#if RECORD_SOUNDMOVIE	
			if(g_screenshotLate){
				D3DLOCKED_RECT lockRect;

				// Get a pointer to the back buffer (remember, page flipping has taken place)
				//HRESULT rs = g_pd3dDevice->GetBackBuffer(0, D3DBACKBUFFER_TYPE_MONO, &g_pBackSurface);
				//if(D3D_OK != rs){
				//	throw;
				//}


				HRESULT rs = g_pBackSurface->LockRect(&lockRect, 0, D3DLOCK_READONLY);
				if(D3D_OK != rs){
					throw;
				}
				unsigned int screenWidth = lockRect.Pitch / 4;
				unsigned int width = m_d3dsdBackBuffer.Width;
				unsigned int height = m_d3dsdBackBuffer.Height;
				
				unsigned int* pBits = reinterpret_cast<unsigned int*>(lockRect.pBits);

				//saveBackBuffer(pBits, pixels, screenWidth, width, height);
				
		//		if( D3DERR_INVALIDCALL == D3DXSaveSurfaceToFile("screen.bmp",D3DXIFF_BMP,g_pBackSurface,NULL,NULL)){
		//			throw;
		//		}
				if(g_recordSoundMovie){
					updateBuffer();
					saveFrameToMem(pBits, screenWidth, width, height);
				}
				else{
					saveScreenshotToMem(pBits, screenWidth, width, height, g_frameBuffer);
				}
				g_pBackSurface->UnlockRect();
			}
#endif

#ifdef DEMO_TO_MOVIE
outputScreenShotForDemo2();
#endif

		}
		if(fogExist == true && g_fogExist==false){
			g_pd3dDevice->SetRenderState(D3DRS_FOGENABLE, FALSE);
		}

		g_pd3dDevice->EndScene();

		// Present the backbuffer contents to the display
		g_pd3dDevice->Present(NULL, NULL, hWndMain, NULL );
		g_pBackSurface->Release();
	}
	InputEvent::renderCounter++;
	setMathHiPresision();
	InputEvent::runRobotSequence();
	if(g_waitMiliseconds > 0){
		Sleep(g_waitMiliseconds);
	}




	//get2dMousePosFrom3d();
	//g_time+=1.0/12.0;
#ifdef BIND_DEVELOPER_SCRIPT
	scriptBindAnimation();
#endif
}
/*
void Render(){
	if(D3D_OK==g_pd3dDevice->GetBackBuffer(0, D3DBACKBUFFER_TYPE_MONO, &g_pBackSurface)){
		// Clear the backbuffer and the zbuffer
		g_pd3dDevice->Clear( 0, NULL, D3DCLEAR_TARGET|D3DCLEAR_ZBUFFER,	g_settings->backgroundColor.refValue().colorToHex() , 1.0f, 0 );
		
		// Begin the scene
		g_pd3dDevice->BeginScene();
		
		


		renderToolbar();

		
		g_pBackSurface->Release();




		g_pd3dDevice->EndScene();

		g_pd3dDevice->Present( NULL, NULL, NULL, NULL );
	}
}
*/
/*
//-----------------------------------------------------------------------------
// Name: MsgProc()
// Desc: The window's message handler
//-----------------------------------------------------------------------------
LRESULT WINAPI MsgProc( HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam )
{
    switch( msg )
    {
	case WM_DESTROY:
		PostQuitMessage( 0 );
		return 0;
    }
	
    return DefWindowProc( hWnd, msg, wParam, lParam );
}
*/


/*
//-----------------------------------------------------------------------------
// Name: WinMain()
// Desc: The application's entry point
//-----------------------------------------------------------------------------
INT WINAPI WinMain( HINSTANCE hInst, HINSTANCE, LPSTR, INT )
{
    // Register the window class
    WNDCLASSEX wc = { sizeof(WNDCLASSEX), CS_CLASSDC, MsgProc, 0L, 0L,
		GetModuleHandle(NULL), NULL, NULL, NULL, NULL,
		"D3D Tutorial", NULL };
    RegisterClassEx( &wc );
	
    // Create the application's window
    HWND hWnd = CreateWindow( "D3D Tutorial", "D3D Tutorial 04: Lights",
		WS_OVERLAPPEDWINDOW, 100, 100, 500, 500,
		GetDesktopWindow(), NULL, wc.hInstance, NULL );
	
    // Initialize Direct3D
    if( SUCCEEDED( InitD3D( hWnd ) ) )
    {
        // Create the geometry
        if( SUCCEEDED( InitGeometry() ) )
        {
            // Show the window
            ShowWindow( hWnd, SW_SHOWDEFAULT );
            UpdateWindow( hWnd );
			
            // Enter the message loop
            MSG msg;
            ZeroMemory( &msg, sizeof(msg) );
            while( msg.message!=WM_QUIT )
            {
                if( PeekMessage( &msg, NULL, 0U, 0U, PM_REMOVE ) )
                {
                    TranslateMessage( &msg );
                    DispatchMessage( &msg );
                }
                else
                    Render();
            }
        }
    }
	
    // Clean up everything and exit the app
    Cleanup();
    UnregisterClass( "D3D Tutorial", wc.hInstance );
    return 0;
}

*/
void IndexedFaceSet::renderNode(Matrix parentMatrix, Vec3f parentColScale, GeomSource& geomSource, DXRender*& render, const Appearance* appearance){
	if(coordIndex.refValue().length == 0)return;
	Coordinate* crd = dynamic_cast<Coordinate*>(coord.p);
	Normal* nrm = dynamic_cast<Normal*>(normal.p);
	Color* color = dynamic_cast<Color*>(this->color.p);
	TextureCoordinate* txc = dynamic_cast<TextureCoordinate*>(texCoord.p);
	Material* mat = 0;
	if(appearance){
		mat = dynamic_cast<Material*>(appearance->material.p);
	}
	
	if(render==0){
		goto here;
	}

	if(crd == 0){
		return;
	}
	if(crd->renderProcessed == false){
		renderProcessedQue += crd;
		goto here;
	}
	else if(nrm && nrm->renderProcessed == false){
		renderProcessedQue += nrm;
		goto here;
	}
	else if(color && color->renderProcessed == false){
		renderProcessedQue += color;
		goto here;
	}
	else if(color && mat && mat->renderProcessed == false){
		renderProcessedQue += mat;
		goto here;
	}
	else if(txc && txc->renderProcessed == false){
		renderProcessedQue+=txc;
		goto here;
	}
	else if(renderProcessed == false){
here:
		renderProcessedQue+=this;
		try{
			convertToGeomSource(geomSource,appearance, g_tmpIndex);
		}
		catch(...){
			delete render;
			render=0;
			return;
		}
		geomSource.convertToRenderFormat(render, appearance);
	}
	if(render==0){
		return;
	}
	if(appearance){
		TextureTransform* textureTransform=dynamic_cast<TextureTransform*>(appearance->textureTransform.p);
		if(textureTransform && textureTransform->renderProcessed==false){
			geomSource.transformRenderTexture(render, textureTransform);
			renderProcessedQue+=textureTransform;
		}
	}
	geomSource.renderNode(parentMatrix, parentColScale, render, appearance);
}

void IndexedTriangleSet::renderNode(Matrix parentMatrix, Vec3f parentColScale, GeomSource& geomSource, DXRender*& render, const Appearance* appearance){
	if(index.refValue().length==0)return;
	Coordinate* crd=dynamic_cast<Coordinate*>(coord.p);
	Normal* nrm=dynamic_cast<Normal*>(normal.p);
	Color* color=dynamic_cast<Color*>(this->color.p);
	TextureCoordinate* txc=dynamic_cast<TextureCoordinate*>(texCoord.p);
	if(crd==0){
		return;
	}
	if(crd->renderProcessed==false){
		renderProcessedQue+=crd;
		goto here;
	}
	else if(nrm!=0&&nrm->renderProcessed==false){
		renderProcessedQue+=nrm;
		goto here;
	}
	else if(color!=0&&color->renderProcessed==false){
		renderProcessedQue+=color;
		goto here;
	}
	else if(txc!=0&&txc->renderProcessed==false){
		renderProcessedQue+=txc;
		goto here;
	}
	else if(renderProcessed==false){
here:
		renderProcessedQue+=this;
		try{
			convertToGeomSource(geomSource,appearance, g_tmpIndex);
		}
		catch(...){
			delete render;
			render=0;
			return;
		}
		geomSource.convertToRenderFormat(render, appearance);
	}
	if(render==0){
		return;
	}
	if(appearance){
		TextureTransform* textureTransform=dynamic_cast<TextureTransform*>(appearance->textureTransform.p);
		if(textureTransform && textureTransform->renderProcessed==false){
			geomSource.transformRenderTexture(render, textureTransform);
			renderProcessedQue+=textureTransform;
		}
	}
	geomSource.renderNode(parentMatrix, parentColScale, render, appearance);
}
void Polyline2D::convertToRenderFormat(	DXRender*& render){
	try{
		if(render==0)render=new DXRender;

		if( render->pVB != NULL ){
			render->pVB->Release();
		}

		// Create the vertex buffer.
#ifdef DX8
		if( FAILED( g_pd3dDevice->CreateVertexBuffer( lineSegments.length*sizeof(VERTEX),	0, D3DFVF_VERTEX, D3DPOOL_DEFAULT, &render->pVB ) ) ){
#endif
#ifdef DX9
		if( FAILED( g_pd3dDevice->CreateVertexBuffer( lineSegments.length*sizeof(VERTEX), 0, D3DFVF_VERTEX, D3DPOOL_DEFAULT, &render->pVB, 0) ) ){
#endif
			throw this;
		}
		// Fill the vertex buffer.
		// here, including the normals, which are used for lighting.
		VERTEX* pVertices;
#ifdef DX8
		if( FAILED( render->pVB->Lock( 0, 0, (BYTE**)&pVertices, 0 ) ) ){
#endif
#ifdef DX9
		if( FAILED( render->pVB->Lock( 0, 0, (void**)&pVertices, 0 ) ) ){
#endif
			throw this;
		}
		for( int i=lineSegments.length;--i>=0; ){
			pVertices[i].position.x=lineSegments[i].x;
			pVertices[i].position.y=lineSegments[i].y;
			pVertices[i].position.z=0;
		}

		render->pVB->Unlock();
	}
	catch(Node*){
		dLog("Unable to render");
		delete render;
		render=0;
	}
}
void Polyline2D::renderNode(Matrix parentMatrix, Vec3f parentColScale, GeomSource& geomSource, DXRender*& render, const Appearance* appearance){
	if(lineSegments.length==0)return;
	else if(renderProcessed==false){
		renderProcessedQue+=this;
		this->convertToRenderFormat(render);
	}
	float a=(parentColScale.x+parentColScale.y+parentColScale.z)*(1./3.);
	g_light.Diffuse.r  = a;
	g_light.Diffuse.g  = a;
	g_light.Diffuse.b  = a;
	g_pd3dDevice->SetLight( 0, &g_light );


	g_pd3dDevice->SetTransform( D3DTS_WORLD, (D3DMATRIX*)&parentMatrix );

	if(appearance){
		if(Material* material=dynamic_cast<Material*>(appearance->material.refValue().getRawPtr())){
			Color3f diffuseColor=material->diffuseColor;
			Color3f emissiveColor=material->emissiveColor;
			D3DMATERIAL mtrl;
			ZeroMemory( &mtrl, sizeof(D3DMATERIAL) );
			mtrl.Diffuse.r = mtrl.Ambient.r = diffuseColor.r;
			mtrl.Diffuse.g = mtrl.Ambient.g = diffuseColor.g;
			mtrl.Diffuse.b = mtrl.Ambient.b = diffuseColor.b;
			mtrl.Diffuse.a = mtrl.Ambient.a = material->transparency;
			mtrl.Emissive.r = emissiveColor.r;
			mtrl.Emissive.g = emissiveColor.g;
			mtrl.Emissive.b = emissiveColor.b;
			g_pd3dDevice->SetMaterial( &mtrl );
		}
	}

	g_pd3dDevice->SetTexture( 0, NULL );
#ifdef DX8
	g_pd3dDevice->SetStreamSource( 0, render->pVB, sizeof(VERTEX) );
	g_pd3dDevice->SetVertexShader( D3DFVF_VERTEX );
#endif
#ifdef DX9
	g_pd3dDevice->SetStreamSource( 0, render->pVB, 0, sizeof(VERTEX));
	g_pd3dDevice->SetFVF( D3DFVF_VERTEX );
#endif
	g_pd3dDevice->DrawPrimitive( D3DPT_LINESTRIP, 0, lineSegments.length-1);
}
void IndexedLineSet::updateAllVertexInBuffers(DXRender*& render){
	Coordinate* coord=dynamic_cast<Coordinate*>(this->coord.p);
	Color* color=dynamic_cast<Color*>(this->color.p);
	if(coord==0)return;
	M_Vec3f& point=coord->point;

	try{
		// Fill the vertex buffer.
		// here, including the normals, which are used for lighting.
		if(color && color->color.length == point.length){
			M_Color3f& m_color = color->color;

			COLOR_VERTEX_NO_NORMAL* pVertices;

#ifdef DX8
			if(FAILED(render->pVB->Lock( 0, 0, (BYTE**)&pVertices, 0 ) ) ){
#endif
#ifdef DX9
			if(FAILED(render->pVB->Lock( 0, 0, (void**)&pVertices, 0 ) ) ){
#endif
				throw this;
			}
			for(int i = point.length; --i >= 0;){
				pVertices[i].position = *(D3DXVECTOR3*)&point[i];
			}
		}
		else{
			VERTEX* pVertices;
#ifdef DX8
			if(FAILED(render->pVB->Lock( 0, 0, (BYTE**)&pVertices, 0 ) ) ){
#endif
#ifdef DX9
			if(FAILED(render->pVB->Lock( 0, 0, (void**)&pVertices, 0 ) ) ){
#endif
				throw this;
			}
			for(int i = point.length; --i >= 0;){
				pVertices[i].position = *(D3DXVECTOR3*)&point[i];
			}
		}

		render->pVB->Unlock();
	}
	catch(Node*){
		dLog("Unable to render");
		delete render;
		render=0;
	}
}

void IndexedLineSet::convertToRenderFormat(	DXRender*& render){
	Coordinate* coord=dynamic_cast<Coordinate*>(this->coord.p);
	Color* color=dynamic_cast<Color*>(this->color.p);
	if(coord==0)return;
	M_Vec3f& point=coord->point;

	indexCount=0;
	for(int i=0, j=1; j<coordIndex.length; ){
		if(coordIndex[j]==-1 && ++j<coordIndex.length){
			j++;
		}
		indexCount+=2;
		j++;
	}

	try{
		if(render==0)render=new DXRender;

		if( render->pVB != NULL ){
			render->pVB->Release();
		}
		if( render->pIB != NULL )
			render->pIB->Release();

		{
			WORD* pIndices;
			
	#ifdef DX8			
				if( FAILED( g_pd3dDevice->CreateIndexBuffer( indexCount*sizeof(WORD), 0, D3DFMT_INDEX16, D3DPOOL_DEFAULT, &render->pIB ) ) ){
					throw this;
				}		
				if( FAILED( render->pIB->Lock( 0, indexCount*sizeof(WORD), (BYTE**)&pIndices, 0 ) ) ){
					throw this;
				}
	#else
			if( FAILED( g_pd3dDevice->CreateIndexBuffer( indexCount*sizeof(WORD), 0, D3DFMT_INDEX16, D3DPOOL_DEFAULT, &render->pIB, 0) ) ){
				throw this;
			}
			if( FAILED( render->pIB->Lock( 0, indexCount*sizeof(WORD), (void**)&pIndices, 0 ) ) ){
				throw this;
			}
	#endif
			{

				for(int i=0, j=1, q=coordIndex[0]; j<coordIndex.length; ){
					if(coordIndex[j]==-1){
						if(++j==coordIndex.length){
							break;
						}
						q=coordIndex[j++];
					}
					pIndices[i++]=q;
					pIndices[i++]=q=coordIndex[j++];
				}
				if( FAILED( render->pIB->Unlock() ) )
					throw this;
			}

		}
		// Create the vertex buffer.
		if(color && color->color.length == point.length){
			M_Color3f& m_color = color->color;
			// Create the vertex buffer.
#ifdef DX8			
			if( FAILED( g_pd3dDevice->CreateVertexBuffer( point.length*sizeof(COLOR_VERTEX_NO_NORMAL), 0, D3DFVF_COLOR_VERTEX_NO_NORMAL, D3DPOOL_DEFAULT, &render->pVB ) ) ){
				throw this;
			}
			// Fill the vertex buffer.
			// here, including the normals, which are used for lighting.
			COLOR_VERTEX_NO_NORMAL* pVertices;
			if( FAILED( render->pVB->Lock( 0, 0, (BYTE**)&pVertices, 0 ) ) ){
				throw this;
			}
#else
			if( FAILED( g_pd3dDevice->CreateVertexBuffer( point.length*sizeof(COLOR_VERTEX_NO_NORMAL), 0, D3DFVF_COLOR_VERTEX_NO_NORMAL, D3DPOOL_DEFAULT, &render->pVB, 0) ) ){
				throw this;
			}
			// Fill the vertex buffer.
			// here, including the normals, which are used for lighting.
			COLOR_VERTEX_NO_NORMAL* pVertices;
			if( FAILED( render->pVB->Lock( 0, 0, (void**)&pVertices, 0 ) ) ){
				throw this;
			}
#endif
			for(int i = point.length; --i >= 0;){
				pVertices[i].position = *(D3DXVECTOR3*)&point[i];
				pVertices[i].color = m_color[i].colorToHex();
			}
			render->vertexCount = point.length;
		}
		else{
#ifdef DX8
			if( FAILED( g_pd3dDevice->CreateVertexBuffer( point.length*sizeof(VERTEX), 0, D3DFVF_VERTEX, D3DPOOL_DEFAULT, &render->pVB ) ) ){
#else
			if( FAILED( g_pd3dDevice->CreateVertexBuffer( point.length*sizeof(VERTEX), 0, D3DFVF_VERTEX, D3DPOOL_DEFAULT, &render->pVB, 0) ) ){
#endif
				throw this;
			}
		// Fill the vertex buffer.
		// here, including the normals, which are used for lighting.
			VERTEX* pVertices;
#ifdef DX8
			if( FAILED( render->pVB->Lock( 0, 0, (BYTE**)&pVertices, 0 ) ) ){
#endif
#ifdef DX9
			if( FAILED( render->pVB->Lock( 0, 0, (void**)&pVertices, 0 ) ) ){
#endif
				throw this;
			}
			for( int i=point.length;--i>=0; ){
				pVertices[i].position = *(D3DXVECTOR3*)&point[i];
			}
		}

		render->pVB->Unlock();
	}
	catch(Node*){
		dLog("Unable to render");
		delete render;
		render=0;
	}
}
void IndexedLineSet::renderNode(Matrix parentMatrix, Vec3f parentColScale, GeomSource& geomSource, DXRender*& render, const Appearance* appearance){
	Coordinate* coord=dynamic_cast<Coordinate*>(this->coord.p);
	if(coord==0)return;
	M_Vec3f& point=coord->point;
	if(point.length==0)return;
	Color* color = dynamic_cast<Color*>(this->color.p);

	if(render==0){
		goto here;
	}

	if(coord==0){
		return;
	}
	if(coord->renderProcessed==false){
		renderProcessedQue+=coord;
		goto here;
	}
	else if(color!=0&&color->renderProcessed==false){
		renderProcessedQue+=color;
		goto here;
	}
	else if(renderProcessed==false){
here:
		renderProcessedQue+=this;
		try{
			this->convertToRenderFormat(render);
		}
		catch(...){
			delete render;
			render=0;
			return;
		}
	}
	if(render==0){
		return;
	}


	g_pd3dDevice->SetTransform( D3DTS_WORLD, (D3DMATRIX*)&parentMatrix );

	if(color && color->color.length == point.length){
		M_Color3f& m_color = color->color;
		g_pd3dDevice->SetRenderState( D3DRS_LIGHTING, FALSE);
		
		if(noZbuf){
			g_pd3dDevice->SetRenderState( D3DRS_ZENABLE, FALSE);
		}
		g_pd3dDevice->SetTexture( 0, NULL );
#ifdef DX8
		g_pd3dDevice->SetStreamSource( 0, render->pVB, sizeof(COLOR_VERTEX_NO_NORMAL) );
		g_pd3dDevice->SetVertexShader( D3DFVF_COLOR_VERTEX_NO_NORMAL );
		g_pd3dDevice->SetIndices( render->pIB, 0 );
		g_pd3dDevice->DrawIndexedPrimitive( D3DPT_LINELIST, 0, point.length, 0, indexCount>>1);
#endif
#ifdef DX9
		g_pd3dDevice->SetStreamSource( 0, render->pVB, 0, sizeof(COLOR_VERTEX_NO_NORMAL));
		g_pd3dDevice->SetFVF( D3DFVF_COLOR_VERTEX_NO_NORMAL );
		g_pd3dDevice->SetIndices( render->pIB);
		g_pd3dDevice->DrawIndexedPrimitive( D3DPT_LINELIST, 0, 0, point.length, 0, indexCount>>1);
#endif
		if(noZbuf){
			g_pd3dDevice->SetRenderState( D3DRS_ZENABLE, TRUE);
		}
		
		g_pd3dDevice->SetRenderState( D3DRS_LIGHTING, TRUE);
	}
	else{
		if(appearance!=0){
			if(Material* material=dynamic_cast<Material*>(appearance->material.refValue().getRawPtr())){
				Color3f diffuseColor=material->diffuseColor;
				D3DMATERIAL mtrl;
				ZeroMemory( &mtrl, sizeof(D3DMATERIAL) );
				mtrl.Diffuse.r = mtrl.Ambient.r = diffuseColor.r;
				mtrl.Diffuse.g = mtrl.Ambient.g = diffuseColor.g;
				mtrl.Diffuse.b = mtrl.Ambient.b = diffuseColor.b;
				mtrl.Diffuse.a = mtrl.Ambient.a = material->transparency;
				mtrl.Emissive.r = diffuseColor.r;
				mtrl.Emissive.g = diffuseColor.g;
				mtrl.Emissive.b = diffuseColor.b;
				g_pd3dDevice->SetMaterial( &mtrl );
			}
		}
		if(noZbuf){
			g_pd3dDevice->SetRenderState( D3DRS_ZENABLE, FALSE);
		}

		g_pd3dDevice->SetTexture( 0, NULL );
#ifdef DX8
		g_pd3dDevice->SetStreamSource( 0, render->pVB, sizeof(VERTEX) );
		g_pd3dDevice->SetVertexShader( D3DFVF_VERTEX );
		g_pd3dDevice->SetIndices( render->pIB, 0 );
		g_pd3dDevice->DrawIndexedPrimitive( D3DPT_LINELIST, 0, point.length, 0, indexCount>>1);
#endif
#ifdef DX9
		g_pd3dDevice->SetStreamSource( 0, render->pVB, 0, sizeof(VERTEX));
		g_pd3dDevice->SetFVF( D3DFVF_VERTEX );
		g_pd3dDevice->SetIndices( render->pIB);
		g_pd3dDevice->DrawIndexedPrimitive( D3DPT_LINELIST, 0, 0, point.length, 0, indexCount>>1);
#endif
		if(noZbuf){
			g_pd3dDevice->SetRenderState( D3DRS_ZENABLE, TRUE);
		}
	}

	
//	g_pd3dDevice->DrawPrimitive( D3DPT_LINESTRIP, 0, point.length-1);

}
Node* Shape::AppearanceField::operator = (Node* par){
	refValue()=par;
	if(home.render){
		if( home.render->pVB ){
			home.render->pVB->Release();
		}
		if( home.render->pIB){
			home.render->pIB->Release();
		}
	}
	home.render=0;
	home.geomSource.reset();
	return par;
}
Node* Shape::GeometryField::operator = (Node* par){
	refValue()=par;
	if(home.render){
		if( home.render->pVB ){
			home.render->pVB->Release();
		}
		if( home.render->pIB){
			home.render->pIB->Release();
		}
	}
	home.render=0;
	home.geomSource.reset();
	return par;
}
void Shape::renderNode(Matrix parentMatrix, Vec3f parentColScale){
	
	if(Appearance* appearance = dynamic_cast<Appearance*>(this->appearance.p)){
		Material* material = dynamic_cast<Material*>(appearance->material.p);
		if(ImageTexture* texture=dynamic_cast<ImageTexture*>(appearance->texture.p)){
			
			if(g_renderLater == false){
				if(texture->alphaTexture){
					g_alphaChanList += AlphaChanList(this, parentMatrix, parentColScale);
					return;
				}
				if(material && material->transparency > 0){
					g_transparentList += TransparencyList(this, parentMatrix, parentColScale);
					return;
				}
			}

			if(texture->update){
				const M_String& url=texture->url;
				if(url.length>0){
					if( FAILED( D3DXCreateTextureFromFile(  g_pd3dDevice, url[0].c_str(), &texture->dxTexture ) ) ){
						//dLog (String("failed to set Texture ")+texture->url.refValue()[0]);
					}
					else{
						dLog ("set Texture from local file");
						texture->dxTexture->GetSurfaceLevel(0, &g_textureSurface);
						g_textureSurface->GetDesc(&g_surfaceBuffer);
						texture->alphaTexture = false;
						if(g_surfaceBuffer.Format == D3DFMT_A8R8G8B8){
							texture->alphaTexture = true;
						}
						texture->update=false;
					}
				}
			}
		}
		else{
			if(g_renderLater == false){
				if(material && material->transparency > 0){
					g_transparentList += TransparencyList(this, parentMatrix, parentColScale);
					return;
				}
			}
		}
	}
	else{
		return;
	}
	if(Geometry* geometry=dynamic_cast<Geometry*>(this->geometry.getRawPtr())){
		geometry->renderNode(parentMatrix, parentColScale, geomSource, render, appearance);
	}
}

inline Matrix dxWindMatrix(Matrix& parentMatrix, Vec3f& scale, Rotation& rotation, Vec3f& translation){

	D3DXMATRIX sclMatrix;
	D3DXMATRIX rotMatrix;
	D3DXMATRIX trnMatrix;
	Matrix outputMatrix;

	D3DXMatrixScaling(&sclMatrix, scale.x , scale.y, scale.z);
	D3DXMatrixRotationAxis(&rotMatrix, &D3DXVECTOR3( rotation.x, rotation.y, rotation.z), rotation.angle);
	D3DXMatrixTranslation(&trnMatrix, translation.x , translation.y, translation.z);

	D3DXMatrixMultiply( (D3DXMATRIX*)&outputMatrix, &sclMatrix, &rotMatrix);
	D3DXMatrixMultiply( (D3DXMATRIX*)&outputMatrix, (D3DXMATRIX*)&outputMatrix, &trnMatrix);

	D3DXMatrixMultiply( (D3DXMATRIX*)&outputMatrix, (D3DXMATRIX*)&outputMatrix, (D3DXMATRIX*)&parentMatrix);
	return outputMatrix;
}

inline Matrix Scene::dxWindMatrix(Matrix& parentMatrix){

	D3DXMATRIX sclMatrix;
	D3DXMATRIX rotMatrix;
	D3DXMATRIX trnMatrix;
	Matrix outputMatrix;

	D3DXMatrixScaling(&sclMatrix, scale.x , scale.y, scale.z);
	D3DXMatrixRotationAxis(&rotMatrix, &D3DXVECTOR3( rotation.x, rotation.y, rotation.z), rotation.angle);
	D3DXMatrixTranslation(&trnMatrix, translation.x , translation.y, translation.z);

	D3DXMatrixMultiply( (D3DXMATRIX*)&outputMatrix, &sclMatrix, &rotMatrix);
	D3DXMatrixMultiply( (D3DXMATRIX*)&outputMatrix, (D3DXMATRIX*)&outputMatrix, &trnMatrix);

	D3DXMatrixMultiply( (D3DXMATRIX*)&outputMatrix, (D3DXMATRIX*)&outputMatrix, (D3DXMATRIX*)&parentMatrix);
	return outputMatrix;
}
inline Matrix Transform::dxWindMatrix(Matrix& parentMatrix){
//   P' = T  C  R  SR  S  -SR  -C  P

	return translationMatrix(-center) * rotationMatrix(-scaleOrientation) * scaleMatrix(scale) * rotationMatrix(scaleOrientation) *
		rotationMatrix(rotation) * translationMatrix(center) * translationMatrix(translation) * parentMatrix;

/*
	D3DXMATRIX sclMatrix;
	D3DXMATRIX rotMatrix;
	D3DXMATRIX trnMatrix;
	Matrix outputMatrix;

	Vec3f& s_scale=scale;
	Rotation& s_rotation=rotation;
	Vec3f& s_translation=translation;
	Vec3f& s_center=center;
	Vec3f& s_scaleOrientation=scaleOrientation;

	D3DXMatrixScaling(&sclMatrix, s_scale.x , s_scale.y, s_scale.z);
	D3DXMatrixRotationAxis(&rotMatrix, &D3DXVECTOR3( s_rotation.x, s_rotation.y, s_rotation.z), s_rotation.angle);
	D3DXMatrixTranslation(&trnMatrix, s_translation.x , s_translation.y, s_translation.z);

	D3DXMatrixMultiply( (D3DXMATRIX*)&outputMatrix, &sclMatrix, &rotMatrix);
	D3DXMatrixMultiply( (D3DXMATRIX*)&outputMatrix, (D3DXMATRIX*)&outputMatrix, &trnMatrix);

	D3DXMatrixMultiply( (D3DXMATRIX*)&outputMatrix, (D3DXMATRIX*)&outputMatrix, (D3DXMATRIX*)&parentMatrix);
	return outputMatrix;
*/
}
void Inline::renderNode(Matrix parentMatrix, Vec3f parentColScale){
	Matrix thisMatrix=parentMatrix;
	for(int k=0;k<children.length;k++){
		children[k]->renderNode(thisMatrix,parentColScale);
	}
}
void Scene::renderNode(Matrix parentMatrix, Vec3f parentColScale){
	parentColScale.x*=scale.x;
	parentColScale.y*=scale.y;
	parentColScale.z*=scale.z;

	Matrix thisMatrix=dxWindMatrix(parentMatrix);
	M_Node& children=this->children.refValue();
	for(int k=0;k<children.length;k++){
		children[k]->renderNode(thisMatrix,parentColScale);
	}
}
void Transform::renderNode(Matrix parentMatrix, Vec3f parentColScale){
	Vec3f &s_scale=scale.refValue();
	parentColScale.x*=s_scale.x;
	parentColScale.y*=s_scale.y;
	parentColScale.z*=s_scale.z;

	Matrix thisMatrix=dxWindMatrix(parentMatrix);
	for(int k=0;k<children.refValue().length;k++){
		children.refValue()[k]->renderNode(thisMatrix,parentColScale);
	}
}
void ProtoDeclare::renderNode(Matrix parentMatrix, Vec3f parentColScale){
	M_Node& children=protoBody->children.refValue();
	Matrix thisMatrix=parentMatrix;
	for(int k=0;k<children.length;k++){
		children[k]->renderNode(thisMatrix,parentColScale);
	}
}
void AGroup::renderNode(Matrix parentMatrix, Vec3f parentColScale){
	Matrix thisMatrix=parentMatrix;
	M_Node& children=this->children.refValue();
	for(int k=0;k<children.length;k++){
		children[k]->renderNode(thisMatrix,parentColScale);
	}
}

void NoOutputGroup::renderNode(Matrix parentMatrix, Vec3f parentColScale){
	if(g_screenshotEarly){
		return;
	}
	Matrix thisMatrix=parentMatrix;
	M_Node& children=this->children.refValue();
	for(int k=0;k<children.length;k++){
		children[k]->renderNode(thisMatrix,parentColScale);
	}
}

void GeomSource::transformRenderTexture(DXRender*& render, TextureTransform* textureTransform){
	TEXTUREVERTEX* pVertices;
#ifdef DX8
		if( FAILED( render->pVB->Lock( 0, 0, (BYTE**)&pVertices, 0 ) ) ){
#endif
#ifdef DX9
		if( FAILED( render->pVB->Lock( 0, 0, (void**)&pVertices, 0 ) ) ){
#endif
		throw this;
	}
/*
	Matrix3 matrix=
	translationMatrix(-textureTransform->center)*
	scaleMatrix(textureTransform->scale)*
	rotationMatrix(textureTransform->rotation)*
	translationMatrix(textureTransform->center)*
	translationMatrix(textureTransform->translation);
*/
	Matrix3 matrix=
	translationMatrix(textureTransform->translation)*
	translationMatrix(textureTransform->center)*
	rotationMatrix(textureTransform->rotation)*
	scaleMatrix(textureTransform->scale)*
	translationMatrix(-textureTransform->center);

/*	for(int i=texCoord.length; --i>=0;){
		int ci=getIndex(i);
		Vec2f tc=matrix*texCoord[i];

		pVertices[i].tu = tc.x;
		pVertices[i].tv = 1-tc.y;
	}*/
	for(int i=texCoord.length; --i>=0;){
		int ci=getIndex(i);
		Vec2f tc=matrix*Vec2f(texCoord[i].x, texCoord[i].y);

		//Vec2f tc=matrix*Vec2f(texCoord[i].y, texCoord[i].x);
		//Vec2f tc=matrix*Vec2f(texCoord[i].y, texCoord[i].x);

		pVertices[i].tu = tc.x;
		pVertices[i].tv = 1-tc.y;
	}
	render->pVB->Unlock();
}
void GeomSource::renderNode(Matrix parentMatrix, Vec3f parentColScale, DXRender*& render, const Appearance* appearance){
	if(coordIndex.length==0)return;
	if(coord.length==0)return;
	if(render==0)return;
	if(render->pIB==NULL){
		return;
	}
	float a=(parentColScale.x+parentColScale.y+parentColScale.z)*(1./3.);
	g_light.Diffuse.r  = a;
	g_light.Diffuse.g  = a;
	g_light.Diffuse.b  = a;
	g_pd3dDevice->SetLight( 0, &g_light );

	g_pd3dDevice->SetTransform( D3DTS_WORLD, (D3DMATRIX*)&parentMatrix );

	if(texCoord.length){
		if(appearance){
			if(Material* material=dynamic_cast<Material*>(appearance->material.p)){
				float transparency = material->transparency;
				Color3f diffuseColor=material->diffuseColor;
				Color3f emissiveColor=material->emissiveColor;
				D3DMATERIAL mtrl;
				ZeroMemory( &mtrl, sizeof(D3DMATERIAL) );
				mtrl.Diffuse.r = mtrl.Ambient.r = diffuseColor.r;
				mtrl.Diffuse.g = mtrl.Ambient.g = diffuseColor.g;
				mtrl.Diffuse.b = mtrl.Ambient.b = diffuseColor.b;
				mtrl.Diffuse.a = mtrl.Ambient.a = transparency;
				mtrl.Emissive.r = emissiveColor.r;
				mtrl.Emissive.g = emissiveColor.g;
				mtrl.Emissive.b = emissiveColor.b;
				g_pd3dDevice->SetMaterial( &mtrl );
			}
			ImageTexture* texture=dynamic_cast<ImageTexture*>(appearance->texture.p);
			if(texture && texture->url.refValue().length){
 				if(texture->dxTexture==NULL && !texture->hInternetOpenUrl && g_online){
					texture->openUrl();
				}
#ifdef DX8
				g_pd3dDevice->SetTexture( 0, (IDirect3DBaseTexture8 *)texture->dxTexture );
				g_pd3dDevice->SetTextureStageState( 0, D3DTSS_COLORARG1, D3DTA_TEXTURE );
				//g_pd3dDevice->SetTextureStageState( 0, D3DTSS_ALPHAOP,   D3DTOP_DISABLE );
				//g_pd3dDevice->SetTextureStageState( 0, D3DTSS_ALPHAARG1, D3DTA_TEXTURE);

				g_pd3dDevice->SetTextureStageState( 0, D3DTSS_MAGFILTER , D3DTEXF_LINEAR  );

				g_pd3dDevice->SetStreamSource( 0, render->pVB, sizeof(TEXTUREVERTEX) );
				g_pd3dDevice->SetVertexShader( TEXTURE_CUSTOMVERTEX );
#endif
#ifdef DX9
				g_pd3dDevice->SetTexture( 0, texture->dxTexture );
				g_pd3dDevice->SetTextureStageState( 0, D3DTSS_COLORARG1, D3DTA_TEXTURE );
				g_pd3dDevice->SetTextureStageState( 0, D3DTSS_ALPHAOP,   D3DTOP_DISABLE );
				g_pd3dDevice->SetSamplerState( 0, D3DSAMP_MAGFILTER, D3DTEXF_LINEAR );

				g_pd3dDevice->SetStreamSource( 0, render->pVB, 0, sizeof(TEXTUREVERTEX) );
				g_pd3dDevice->SetFVF( TEXTURE_CUSTOMVERTEX );
#endif
			}
		}
	} 
	else if(color.length){
		if(appearance){
			if(Material* material=dynamic_cast<Material*>(appearance->material.refValue().getRawPtr())){
				float transparency = material->transparency;
				transparency = 1;
				Color3f diffuse = material->diffuseColor;
				Color3f emissive = material->emissiveColor;
				Color3f specular = material->specularColor;

				D3DMATERIAL mtrl;
				ZeroMemory( &mtrl, sizeof(D3DMATERIAL) );
				mtrl.Diffuse.r = 1;
				mtrl.Diffuse.g = 1;
				mtrl.Diffuse.b = 1;
				mtrl.Diffuse.a = transparency;

				mtrl.Ambient.r = diffuse.r;
				mtrl.Ambient.g = diffuse.g;
				mtrl.Ambient.b = diffuse.b;
				mtrl.Ambient.a = transparency;

				mtrl.Emissive.r = emissive.r;
				mtrl.Emissive.g = emissive.g;
				mtrl.Emissive.b = emissive.b;
				mtrl.Emissive.a = transparency;

				mtrl.Specular.r = specular.r;
				mtrl.Specular.g = specular.g;
				mtrl.Specular.b = specular.b;
				mtrl.Specular.a = transparency;

				mtrl.Power=.5;
				g_pd3dDevice->SetMaterial( &mtrl );
			}
		}

		g_pd3dDevice->SetTexture( 0, NULL );

		#ifdef DX8
				g_pd3dDevice->SetStreamSource( 0, render->pVB, sizeof(COLOR_VERTEX) );
				g_pd3dDevice->SetVertexShader( D3DFVF_COLOR_VERTEX );
		#endif
		#ifdef DX9
				g_pd3dDevice->SetStreamSource( 0, render->pVB, 0, sizeof(COLOR_VERTEX) );
				g_pd3dDevice->SetFVF( D3DFVF_COLOR_VERTEX );
		#endif
	}
	else{
		if(appearance){
			if(Material* material=dynamic_cast<Material*>(appearance->material.refValue().getRawPtr())){
				float transparency = material->transparency;
				transparency = 1 - transparency; //why do we have to invert for material but not for ColorEffect?
				Color3f diffuse = material->diffuseColor;
				Color3f emissive = material->emissiveColor;
				Color3f specular = material->specularColor;

				D3DMATERIAL mtrl;
				ZeroMemory( &mtrl, sizeof(D3DMATERIAL) );
				mtrl.Diffuse.r = diffuse.r;
				mtrl.Diffuse.g = diffuse.g;
				mtrl.Diffuse.b = diffuse.b;
				mtrl.Diffuse.a = transparency;

				mtrl.Ambient.r = diffuse.r;
				mtrl.Ambient.g = diffuse.g;
				mtrl.Ambient.b = diffuse.b;
				mtrl.Ambient.a = transparency;

				mtrl.Emissive.r = emissive.r;
				mtrl.Emissive.g = emissive.g;
				mtrl.Emissive.b = emissive.b;
				mtrl.Emissive.a = transparency;

				mtrl.Specular.r = specular.r;
				mtrl.Specular.g = specular.g;
				mtrl.Specular.b = specular.b;
				mtrl.Specular.a = transparency;

				mtrl.Power=.5;
				g_pd3dDevice->SetMaterial( &mtrl );


/*
				if(g_alphaEnable != (material->transparency > 0)){
					if((material->transparency > 0)){
						g_alphaEnable = true;
						g_pd3dDevice->SetRenderState(D3DRS_ALPHABLENDENABLE, TRUE);
					}
					else{
						g_alphaEnable = false;
						g_pd3dDevice->SetRenderState(D3DRS_ALPHABLENDENABLE, FALSE);
					}
				}
				float transparency = material->transparency;
				Color3f diffuseColor=material->diffuseColor;
				Color3f emissiveColor=material->emissiveColor;
				D3DMATERIAL mtrl;
				ZeroMemory( &mtrl, sizeof(D3DMATERIAL) );
				mtrl.Diffuse.r = mtrl.Ambient.r = diffuseColor.r;
				mtrl.Diffuse.g = mtrl.Ambient.g = diffuseColor.g;
				mtrl.Diffuse.b = mtrl.Ambient.b = diffuseColor.b;
				mtrl.Diffuse.a = mtrl.Ambient.a = transparency;
				mtrl.Specular.r = mtrl.Emissive.r = emissiveColor.r;
				mtrl.Specular.g = mtrl.Emissive.g = emissiveColor.g;
				mtrl.Specular.b = mtrl.Emissive.b = emissiveColor.b;
				mtrl.Specular.a = mtrl.Emissive.a = transparency;
				mtrl.Power=.5;
				g_pd3dDevice->SetMaterial( &mtrl );
				*/
			}
		}
		g_pd3dDevice->SetTexture( 0, NULL );
#ifdef DX8
		g_pd3dDevice->SetStreamSource( 0, render->pVB, sizeof(SIMPLEVERTEX) );
		g_pd3dDevice->SetVertexShader( D3DFVF_SIMPLEVERTEX );
#else
		g_pd3dDevice->SetStreamSource( 0, render->pVB, 0, sizeof(SIMPLEVERTEX) );
		g_pd3dDevice->SetFVF( D3DFVF_SIMPLEVERTEX );
#endif
	}
#ifdef DX8
	g_pd3dDevice->SetIndices( render->pIB, 0 );
	g_pd3dDevice->DrawIndexedPrimitive( D3DPT_TRIANGLELIST, 0, render->vertexCount, 0, render->indexCount/3 );
#else
	g_pd3dDevice->SetIndices( render->pIB);
	g_pd3dDevice->DrawIndexedPrimitive( D3DPT_TRIANGLELIST, 0, 0, render->vertexCount, 0, render->indexCount/3 );
#endif
}
void GeomRender::renderNode(Matrix parentMatrix, Vec3f parentColScale){
	if(render==0){
		convertToRenderFormat(render, appearance);
	}
	GeomSource::renderNode(parentMatrix, parentColScale, render, appearance);
}

void GeomSource::convertToRenderFormat(DXRender*& render, const Appearance* appearance){
	try{
		if(render==0)render=new DXRender;

		if(normal.length!=coord.length){
			genNormals();
		}

		int indexCount=coordIndex.length;
		render->indexCount=indexCount;
		if( render->pVB != NULL )
			render->pVB->Release();

		if( render->pIB != NULL )
			render->pIB->Release();


		{
			WORD* pIndices;
#ifdef DX8			
			if( FAILED( g_pd3dDevice->CreateIndexBuffer( indexCount*sizeof(WORD), 0, D3DFMT_INDEX16, D3DPOOL_DEFAULT, &render->pIB ) ) ){
				throw this;
			}
			if( FAILED( render->pIB->Lock( 0, indexCount*sizeof(WORD), (BYTE**)&pIndices, 0 ) ) ){
				throw this;
			}
#else
			if( FAILED( g_pd3dDevice->CreateIndexBuffer(indexCount*sizeof(WORD), 0, D3DFMT_INDEX16, D3DPOOL_DEFAULT, &render->pIB, 0 ) ) ){
				throw this;
			}
			if( FAILED( render->pIB->Lock( 0, indexCount*sizeof(WORD), (void**)&pIndices, 0 ) ) ){
				throw this;
			}
#endif
			for(int i=indexCount; (i-=3)>=0; ){
				pIndices[i]  =coordIndex[i+2];
				pIndices[i+1]=coordIndex[i+1];
				pIndices[i+2]=coordIndex[i];
			}
			if( FAILED( render->pIB->Unlock() ) )
				throw this;
		}
		if(texCoord.length > 0){
			if(texCoord.length<coord.length)throw this;
			// Create the vertex buffer.
#ifdef DX8			
			if( FAILED( g_pd3dDevice->CreateVertexBuffer( texCoord.length*sizeof(TEXTUREVERTEX), 0, TEXTURE_CUSTOMVERTEX, D3DPOOL_DEFAULT, &render->pVB ) ) ){
				throw this;
			}
			TEXTUREVERTEX* pVertices;
			if( FAILED( render->pVB->Lock( 0, 0, (BYTE**)&pVertices, 0 ) ) ){
				throw this;
			}
#else
			if( FAILED( g_pd3dDevice->CreateVertexBuffer( texCoord.length*sizeof(TEXTUREVERTEX), 0, TEXTURE_CUSTOMVERTEX, D3DPOOL_DEFAULT, &render->pVB, 0) ) ){
				throw this;
			}
			TEXTUREVERTEX* pVertices;
			if( FAILED( render->pVB->Lock( 0, 0, (void**)&pVertices, 0 ) ) ){
				throw this;
			}
#endif
			for(int i=texCoord.length; --i>=0;){
				int ci=getIndex(i);
				pVertices[i].position = *(D3DXVECTOR3*)&coord[ci];
				pVertices[i].normal = *(D3DXVECTOR3*)&normal[ci];
				pVertices[i].tu = texCoord[i].x;
				pVertices[i].tv = 1 - texCoord[i].y;
			}
			render->vertexCount=texCoord.length;
		}
		else if(color.length > 0){
			if(color.length<coord.length)throw this;
			// Create the vertex buffer.
#ifdef DX8			
			if( FAILED( g_pd3dDevice->CreateVertexBuffer( color.length*sizeof(COLOR_VERTEX), 0, D3DFVF_COLOR_VERTEX, D3DPOOL_DEFAULT, &render->pVB ) ) ){
				throw this;
			}
			// Fill the vertex buffer.
			// here, including the normals, which are used for lighting.
			COLOR_VERTEX* pVertices;
			if( FAILED( render->pVB->Lock( 0, 0, (BYTE**)&pVertices, 0 ) ) ){
				throw this;
			}
#else
			if( FAILED( g_pd3dDevice->CreateVertexBuffer( color.length*sizeof(COLOR_VERTEX), 0, D3DFVF_COLOR_VERTEX, D3DPOOL_DEFAULT, &render->pVB, 0) ) ){
				throw this;
			}
			// Fill the vertex buffer.
			// here, including the normals, which are used for lighting.
			COLOR_VERTEX* pVertices;
			if( FAILED( render->pVB->Lock( 0, 0, (void**)&pVertices, 0 ) ) ){
				throw this;
			}
#endif
			float alpha = 1;
			if(appearance){
				if(Material* mat = dynamic_cast<Material*>(appearance->material.p)){
					alpha = 1.0f - mat->transparency;
				}
			}

			for(int i=color.length; --i>=0;){
				int ci=getIndex(i);
				pVertices[i].position = *(D3DXVECTOR3*)&coord[ci];
				pVertices[i].normal = *(D3DXVECTOR3*)&normal[ci];
				pVertices[i].color = color[i].colorToHex(alpha);
			}
			render->vertexCount = color.length;
		}
		else{
			// Create the vertex buffer.
#ifdef DX8			
			if( FAILED( g_pd3dDevice->CreateVertexBuffer( coord.length*sizeof(SIMPLEVERTEX), 0, D3DFVF_SIMPLEVERTEX, D3DPOOL_DEFAULT, &render->pVB ) ) ){
				throw this;
			}
			// Fill the vertex buffer.
			// here, including the normals, which are used for lighting.
			SIMPLEVERTEX* pVertices;
			if( FAILED( render->pVB->Lock( 0, 0, (BYTE**)&pVertices, 0 ) ) ){
				throw this;
			}
#else
			if( FAILED( g_pd3dDevice->CreateVertexBuffer( coord.length*sizeof(SIMPLEVERTEX), 0, D3DFVF_SIMPLEVERTEX, D3DPOOL_DEFAULT, &render->pVB, 0) ) ){
				throw this;
			}
			// Fill the vertex buffer.
			// here, including the normals, which are used for lighting.
			SIMPLEVERTEX* pVertices;
			if( FAILED( render->pVB->Lock( 0, 0, (void**)&pVertices, 0 ) ) ){
				throw this;
			}
#endif
			for( int i=coord.length;--i>=0; ){
				pVertices[i].position = *(D3DXVECTOR3*)&coord[i];
				pVertices[i].normal = *(D3DXVECTOR3*)&normal[i];
			}
			render->vertexCount = coord.length;
		}
		render->pVB->Unlock();
	}
	catch(Node*){
		dLog("*Unable To Render");
		delete render;
		render=0;
	}
}

void Seamless::addExtraCoordsRender(M_Int& coordIndex){//and populate matrixIndex
	if(partList.length==0) return;
	Effect* effect=partList[0]->workEffect;
	if(effect==0) return;
	static M_Int lookUpCoord;
	lookUpCoord.length=effect->getEffectLength();
	{
		for(int i=lookUpCoord.length; --i>=0; ){
			lookUpCoord[i]=-1;
		}
	}
	{
		for(int mi=0, j=0, k=0; k<partList.length; k++,mi++){
			Art* part=partList[k];
			part->rc.length=0;
			M_Int& matrixIndex=part->matrixIndex;
			matrixIndex.length=part->coord.length;
			const M_Int& coordGroup=part->coordGroup;
			{
				int j=0;
				for(int g=0; g<coordGroup.length; g++, mi++ ){
					for(int i=0; i<coordGroup[g]&&j<matrixIndex.length; i++){
						matrixIndex[j++]=mi;
					}
				}
				while(j<matrixIndex.length){
					matrixIndex[j++]=mi;
				}

			}
		}
	}
	if(ColorEffect* colorEffect=dynamic_cast<ColorEffect*>(effect)){
		const M_Color3f& color=colorEffect->color;

		for(int j=0, k=0; k<partList.length; k++){
			Art* part=partList[k];
			M_Triangle& triangle=part->triangle;
			int triLen=triangle.length*3;
			for(int o=3, i=0; i<triLen; i++){
				int linTci=triangle.refTexIndex(i);
				int ci=triangle.refIndex(i);
				Art* cornerPart=triangle.refPart(i);
				M_Int& localCoordLookUp=cornerPart->partTree;
				if( lookUpCoord[linTci]==-1){
					if( localCoordLookUp[ci]==-1){
						localCoordLookUp[ci]=linTci;
						cornerPart->color[ci]=color[linTci];
						lookUpCoord[linTci]=ci;
						coordIndex[j++]=ci;
					}
					else {
						coordIndex[j++]=lookUpCoord[linTci]=cornerPart->coord.length;
						cornerPart->color+=color[linTci];
						cornerPart->coord.refValue()+=cornerPart->coord[ci];
						cornerPart->normal.refValue()+=cornerPart->normal[ci];
						cornerPart->matrixIndex+=cornerPart->matrixIndex[ci];
						cornerPart->rc+=ci;
					}
				}
				else {
					coordIndex[j++]=lookUpCoord[linTci];
				}
				if(--o==0){
					o=3;
					j++;//coordIndex[j++]=-1;
				}
			}
		}
	}
	else if(TextureEffect* texEffect=dynamic_cast<TextureEffect*>(effect)){
		M_Vec2f& texCoord=texEffect->texCoord.refValue();

		for(int j=0, k=0; k<partList.length; k++){
			Art* part=partList[k];
			M_Triangle& triangle=part->triangle;
			int triLen=triangle.length*3;
			for(int o=3, i=0; i<triLen; i++){
				int linTci=triangle.refTexIndex(i);
				int ci=triangle.refIndex(i);
				Art* cornerPart=triangle.refPart(i);
				M_Int& localCoordLookUp=cornerPart->partTree;
				if( lookUpCoord[linTci]==-1){
					if( localCoordLookUp[ci]==-1){
						localCoordLookUp[ci]=linTci;
						cornerPart->texCoord[ci]=texCoord[linTci];
						lookUpCoord[linTci]=ci;
						coordIndex[j++]=ci;
					}
					else {
						coordIndex[j++]=lookUpCoord[linTci]=cornerPart->coord.length;
						cornerPart->texCoord+=texCoord[linTci];
						cornerPart->coord.refValue()+=cornerPart->coord[ci];
						cornerPart->normal.refValue()+=cornerPart->normal[ci];
						cornerPart->matrixIndex+=cornerPart->matrixIndex[ci];
						cornerPart->rc+=ci;
					}
				}
				else {
					coordIndex[j++]=lookUpCoord[linTci];
				}
				if(--o==0){
					o=3;
					j++;//coordIndex[j++]=-1;
				}
			}
		}
	}
	else if(MonoColorEffect* monoColorEffect=dynamic_cast<MonoColorEffect*>(effect)){
		const M_Color3f& color=colorEffect->color;

		for(int j=0, k=0; k<partList.length; k++){
			Art* part=partList[k];
			M_Triangle& triangle=part->triangle;
			int triLen=triangle.length*3;
			for(int o=3, i=0; i<triLen; i++){
				int linTci=triangle.refTexIndex(i);
				int ci=triangle.refIndex(i);
				Art* cornerPart=triangle.refPart(i);
				M_Int& localCoordLookUp=cornerPart->partTree;
				if( lookUpCoord[linTci]==-1){
					if( localCoordLookUp[ci]==-1){
						localCoordLookUp[ci]=linTci;
						cornerPart->color[ci]=color[linTci];
						lookUpCoord[linTci]=ci;
						coordIndex[j++]=ci;
					}
					else {
						coordIndex[j++]=lookUpCoord[linTci]=cornerPart->coord.length;
						cornerPart->color+=color[linTci];
						cornerPart->coord.refValue()+=cornerPart->coord[ci];
						cornerPart->normal.refValue()+=cornerPart->normal[ci];
						cornerPart->matrixIndex+=cornerPart->matrixIndex[ci];
						cornerPart->rc+=ci;
					}
				}
				else {
					coordIndex[j++]=lookUpCoord[linTci];
				}
				if(--o==0){
					o=3;
					j++;//coordIndex[j++]=-1;
				}
			}
		}
	}
	{
		int linCoordLen=0;
		for(int k=0; k<partList.length; k++){
			Art* part=partList[k];
			part->fromCoord=linCoordLen;
			linCoordLen+=part->coord.length;
		}
	}
	{
		for(int j=0, k=0; k<partList.length; k++){
			Art* part=partList[k];
			M_Triangle& triangle=part->triangle;
			int triLen=triangle.length*3;
			for(int i=0; i<triLen; ){
				coordIndex[j++]+=triangle.refPart(i++)->fromCoord;
				coordIndex[j++]+=triangle.refPart(i++)->fromCoord;
				coordIndex[j++]+=triangle.refPart(i++)->fromCoord;
				j++;
			}
		}
	}
}

void Seamless::addExtraCoordsRenderForGeomSource(M_Int& coordIndex){
	if(partList.length==0) return;
	Effect* effect=partList[0]->workEffect;
	if(effect==0) return;
	static M_Int lookUpCoord;
	lookUpCoord.length=effect->getEffectLength();
	{
		for(int i=lookUpCoord.length; --i>=0; ){
			lookUpCoord[i]=-1;
		}
	}
	{
		for(int mi=0, j=0, k=0; k<partList.length; k++,mi++){
			Art* part=partList[k];
			part->rc.length=0;
			M_Int& matrixIndex=part->matrixIndex;
			matrixIndex.length=part->coord.length;
			const M_Int& coordGroup=part->coordGroup;
			{
				int j=0;
				for(int g=0; g<coordGroup.length; g++, mi++ ){
					for(int i=0; i<coordGroup[g]&&j<matrixIndex.length; i++){
						matrixIndex[j++]=mi;
					}
				}
				while(j<matrixIndex.length){
					matrixIndex[j++]=mi;
				}

			}
		}
	}
	if(ColorEffect* colorEffect=dynamic_cast<ColorEffect*>(effect)){
		const M_Color3f& color=colorEffect->color;

		for(int j=0, k=0; k<partList.length; k++){
			Art* part=partList[k];
			M_Triangle& triangle=part->triangle;
			int triLen=triangle.length*3;
			for(int i=0; i<triLen; i++){
				int linTci=triangle.refTexIndex(i);
				int ci=triangle.refIndex(i);
				Art* cornerPart=triangle.refPart(i);
				M_Int& localCoordLookUp=cornerPart->partTree;
				if( lookUpCoord[linTci]==-1){
					if( localCoordLookUp[ci]==-1){
						localCoordLookUp[ci]=linTci;
						cornerPart->color[ci]=color[linTci];
						lookUpCoord[linTci]=ci;
						coordIndex[j++]=ci;
					}
					else {
						coordIndex[j++]=lookUpCoord[linTci]=cornerPart->coord.length;
						cornerPart->color+=color[linTci];
						cornerPart->coord.refValue()+=cornerPart->coord[ci];
						cornerPart->normal.refValue()+=cornerPart->normal[ci];
						cornerPart->matrixIndex+=cornerPart->matrixIndex[ci];
						cornerPart->rc+=ci;
					}
				}
				else {
					coordIndex[j++]=lookUpCoord[linTci];
				}
			}
		}
	}
	else if(TextureEffect* texEffect=dynamic_cast<TextureEffect*>(effect)){
		M_Vec2f& texCoord=texEffect->texCoord.refValue();

		for(int j=0, k=0; k<partList.length; k++){
			Art* part=partList[k];
			M_Triangle& triangle=part->triangle;
			int triLen=triangle.length*3;
			for(int i=0; i<triLen; i++){
				int linTci=triangle.refTexIndex(i);
				int ci=triangle.refIndex(i);
				Art* cornerPart=triangle.refPart(i);
				M_Int& localCoordLookUp=cornerPart->partTree;
				if( lookUpCoord[linTci]==-1){
					if( localCoordLookUp[ci]==-1){
						localCoordLookUp[ci]=linTci;
						cornerPart->texCoord[ci]=texCoord[linTci];
						lookUpCoord[linTci]=ci;
						coordIndex[j++]=ci;
					}
					else {
						coordIndex[j++]=lookUpCoord[linTci]=cornerPart->coord.length;
						cornerPart->texCoord+=texCoord[linTci];
						cornerPart->coord.refValue()+=cornerPart->coord[ci];
						cornerPart->normal.refValue()+=cornerPart->normal[ci];
						cornerPart->matrixIndex+=cornerPart->matrixIndex[ci];
						cornerPart->rc+=ci;
					}
				}
				else {
					coordIndex[j++]=lookUpCoord[linTci];
				}
			}
		}
	}
	if(MonoColorEffect* colorEffect=dynamic_cast<MonoColorEffect*>(effect)){
		const M_Color3f& color=colorEffect->color;

		for(int j=0, k=0; k<partList.length; k++){
			Art* part=partList[k];
			M_Triangle& triangle=part->triangle;
			int triLen=triangle.length*3;
			for(int i=0; i<triLen; i++){
				int linTci=triangle.refTexIndex(i);
				int ci=triangle.refIndex(i);
				Art* cornerPart=triangle.refPart(i);
				M_Int& localCoordLookUp=cornerPart->partTree;
				if( lookUpCoord[linTci]==-1){
					if( localCoordLookUp[ci]==-1){
						localCoordLookUp[ci]=linTci;
						cornerPart->color[ci]=color[linTci];
						lookUpCoord[linTci]=ci;
						coordIndex[j++]=ci;
					}
					else {
						coordIndex[j++]=lookUpCoord[linTci]=cornerPart->coord.length;
						cornerPart->color+=color[linTci];
						cornerPart->coord.refValue()+=cornerPart->coord[ci];
						cornerPart->normal.refValue()+=cornerPart->normal[ci];
						cornerPart->matrixIndex+=cornerPart->matrixIndex[ci];
						cornerPart->rc+=ci;
					}
				}
				else {
					coordIndex[j++]=lookUpCoord[linTci];
				}
			}
		}
	}
	{
		int linCoordLen=0;
		for(int k=0; k<partList.length; k++){
			Art* part=partList[k];
			part->fromCoord=linCoordLen;
			linCoordLen+=part->coord.length;
		}
	}
	{
		for(int j=0, k=0; k<partList.length; k++){
			Art* part=partList[k];
			M_Triangle& triangle=part->triangle;
			int triLen=triangle.length*3;
			for(int i=0; i<triLen; ){
				coordIndex[j++]+=triangle.refPart(i++)->fromCoord;
				coordIndex[j++]+=triangle.refPart(i++)->fromCoord;
				coordIndex[j++]+=triangle.refPart(i++)->fromCoord;
			}
		}
	}
}

void resetCoordLen(M_Art& partList);
void setUpInternalEffectPtrsTmpForRenderingTransparancy(Seamless* smls);
void setUpInternalEffectPtrs(Seamless* smls);
void setUpInternalEffectPtrs2(Seamless* smls);

void RingBuild::resetCo(Art* part, int& co){
	int i = index[1];
	if(this->part.length && i < this->part.length){
		if(part == this->part[i].p){
			co = 0;
		}
	}
}
void TransferOwnership::resetCo(Art* part, int& co){
	int i = index[0];
	if(dest.length && i < dest.length){
		if(part == dest[i].p){
			co = 0;
		}
	}
}

void Seamless::convertToOwnershipAidRenderFormat(){
	if(g_bReady==false)return;
	if(partList.length==0||effect==0){
		if(render!=0){
			if( render->pVB != NULL ){
				render->pVB->Release();
			}
			if( render->pIB != NULL ){
				render->pIB->Release();
			}
			render=0;
		}
		return;
	}
	if(render == NULL )render=new DXRender;
	int j=0;
	if(geometry == 0){
		geometry = new IndexedFaceSet;
	}
	IndexedFaceSet* ifs=geometry.getRawPtr();
	ifs->coord=new Coordinate();
	ifs->normal=new Normal;
	ifs->color=new Color;
	int coordLen=0;

	M_Int& coordIndex=ifs->coordIndex.refValue();

	catTranslationMatrix();
	ifs->genTransformedNormals(partList);
	catMatrix();
	{
		if(partList.length==0) return;
		Effect* effect = partList[0]->workEffect;
		if(effect==0) return;
		static M_Int lookUpCoord;
		lookUpCoord.length=effect->getEffectLength();
		{
			for(int i=lookUpCoord.length; --i>=0; ){
				lookUpCoord[i]=-1;
			}
		}
		{
			for(int mi=0, j=0, k=0; k<partList.length; k++,mi++){
				Art* part=partList[k];
				part->rc.length=0;
				M_Int& matrixIndex=part->matrixIndex;
				matrixIndex.length=part->coord.length;
				const M_Int& coordGroup=part->coordGroup;
				{
					int j=0;
					for(int g=0; g<coordGroup.length; g++, mi++ ){
						for(int i=0; i<coordGroup[g]&&j<matrixIndex.length; i++){
							matrixIndex[j++]=mi;
						}
					}
					while(j<matrixIndex.length){
						matrixIndex[j++]=mi;
					}

				}
			}
		}
	}
	int indexCount=coordIndex.length/4*3;
	render->indexCount=indexCount;

	if(!indexCount){
		return;
	}
	if( render->pVB != NULL )
		render->pVB->Release();

	if( render->pIB != NULL )
		render->pIB->Release();

	{
        WORD* pIndices;
		
#ifdef DX8			
			if( FAILED( g_pd3dDevice->CreateIndexBuffer( indexCount*sizeof(WORD), 0, D3DFMT_INDEX16, D3DPOOL_DEFAULT, &render->pIB ) ) ){
				throw this;
			}		
			if( FAILED( render->pIB->Lock( 0, indexCount*sizeof(WORD), (BYTE**)&pIndices, 0 ) ) ){
				throw this;
			}
#else
		if( FAILED( g_pd3dDevice->CreateIndexBuffer( indexCount*sizeof(WORD), 0, D3DFMT_INDEX16, D3DPOOL_DEFAULT, &render->pIB, 0) ) ){
			throw this;
		}
		if( FAILED( render->pIB->Lock( 0, indexCount*sizeof(WORD), (void**)&pIndices, 0 ) ) ){
            throw this;
		}
#endif
		{
			for(int t=2, h=0, j=0, k=0; k<partList.length; k++){
				Art* part=partList[k];
				coordLen+=part->coord.length;
				int triLen=part->triangle.length;
				if(part->show){
					for(int i=triLen; --i>=0; ){

						pIndices[j++]=coordIndex[t];
						--t;
						pIndices[j++]=coordIndex[t];
						--t;
						pIndices[j++]=coordIndex[t];
						t+=6;
					}
				}
				else{
					t+=triLen*4;
				}
			}
			if( FAILED( render->pIB->Unlock() ) )
				throw this;
		}
	}
	render->type=COLOR_CUSTOMVERTEX_0WEIGHT;
	COLORVERTEX_0WEIGHT* pVertices;
#ifdef DX8
	if( FAILED( g_pd3dDevice->CreateVertexBuffer( coordLen*sizeof(COLORVERTEX_0WEIGHT),	0, COLOR_CUSTOMVERTEX_0WEIGHT, D3DPOOL_DEFAULT, &render->pVB ) ) ){
		throw this;
	}
	if( FAILED( render->pVB->Lock( 0, 0, (BYTE**)&pVertices, 0 ) ) ){
		throw this;
	}
#else
	if( FAILED( g_pd3dDevice->CreateVertexBuffer( coordLen*sizeof(COLORVERTEX_0WEIGHT), 0, COLOR_CUSTOMVERTEX_0WEIGHT, D3DPOOL_DEFAULT, &render->pVB, 0) ) ){
		throw this;
	}
	if( FAILED( render->pVB->Lock( 0, 0, (void**)&pVertices, 0 ) ) ){
		throw this;
	}
#endif
	{
		setUpInternalEffectPtrsTmpForRenderingTransparancy(this);


		for(int j=0, smi=0, k=0, co=0; k<partList.length;k++){

			Art* part = partList[k];

			const CommonSpec* commonSpec = getCommonSpec(part);

			part->renderFromCoord = j;
			float alpha = 1.0f - reinterpret_cast<ColorEffect*>(part->workEffect)->transparency;

			const M_Vec3f& coord=part->coord;
			const M_Vec3f& normal=part->normal;
			const M_Color3f& color=part->color;
			const M_Int& coordGroup=part->coordGroup;
			const M_Float& weight=part->weight;
			const M_Int& matrixIndex=part->matrixIndex;
			Vec3f absolutePivotPoint;
			if(workAnimCP == false){
				 absolutePivotPoint = part->absolutePivotPoint;
			}


			int i=0;

			unsigned intColor = 0xFFFFFFFF;
			if((co&0x1f) == 0){
				co = 1;
			}
			if(part == g_selNode.getRaw()){
				co = 0;
			}
			else if(MarkPart* markPart = dynamic_cast<MarkPart*>(g_selNode.getRaw())){
				markPart->resetCo(part, co);
			}

			for(int cg=0; cg < coordGroup.length; cg++){
				intColor = 0x808080;
				if(commonSpec && cg < commonSpec->color.length){
					intColor = commonSpec->color[cg].colorToHex();
				}
				for(int c = coordGroup[cg]; --c >= 0 && i < part->coord.length; i++, j++){
					pVertices[j].position = *(D3DXVECTOR3*) &(coord[i] - absolutePivotPoint);
					pVertices[j].normal = *(D3DXVECTOR3*) &normal[i];
					pVertices[j].matrixIndices = matrixIndex[i];
					pVertices[j].color = intColor;
				}
				smi++;
				co++;
				if((co&0x1f) == 0){
					co = 1;
				}
			}
			intColor = 0x808080;
			if(commonSpec && commonSpec->color.length){
				intColor = commonSpec->color[commonSpec->color.length-1].colorToHex();
			}
			for(;i<part->coord.length;i++, j++){
				pVertices[j].position = *(D3DXVECTOR3*) &(coord[i] - absolutePivotPoint);
				pVertices[j].normal = *(D3DXVECTOR3*) &normal[i];
				pVertices[j].matrixIndices = matrixIndex[i];
				pVertices[j].color = intColor;
			}
			smi++;
			co++;
		}
		setUpInternalEffectPtrs(this);
	}
	render->vertexCount=coordLen;
	resetCoordLen( partList);
	render->pVB->Unlock();
	ctrf=false;
}
extern SeamlessPtr g_vcTransferBCyl;
void Seamless::convertToRenderFormat(){
	if(g_bReady==false)return;
	if(partList.length==0||effect==0){
		if(render!=0){
			if( render->pVB != NULL ){
				render->pVB->Release();
			}
			if( render->pIB != NULL ){
				render->pIB->Release();
			}
			render=0;
		}
		return;
	}
	if(partList.length==0)return;
	if(partList[0]->workEffect==0)return;

	if(render == NULL)render=new DXRender;

	setUpInternalEffectPtrs2(this);
	setUpInternalEffectPtrs(this);

	if(workAnimCP == false && g_colorOwnership){
		if(this!=g_vcTransferBCyl){
			convertToOwnershipAidRenderFormat();
			return;
		}
	}

	int j=0;
	if(geometry==0){
		geometry=new IndexedFaceSet;
	}
	IndexedFaceSet* ifs = geometry.getRawPtr();
	ifs->coord = new Coordinate;
	ifs->normal = new Normal;
	if(effect.refValue()->getType()==COLOREFFECT || effect.refValue()->getType()==MONOCOLOREFFECT){
		ifs->color = new Color;
	}
	else{
		ifs->texCoord = new TextureCoordinate;
	}
	int coordLen=0;

	M_Int& coordIndex=ifs->coordIndex.refValue();

	catTranslationMatrix();
	ifs->genTransformedNormals(partList);
	catMatrix();

	
	addExtraCoordsRender(coordIndex);

	int indexCount=coordIndex.length/4*3;
	render->indexCount=indexCount;

	if(!indexCount){
		return;
	}
	if( render->pVB != NULL )
		render->pVB->Release();

	if( render->pIB != NULL )
		render->pIB->Release();

	{
        WORD* pIndices;
		
#ifdef DX8			
			if( FAILED( g_pd3dDevice->CreateIndexBuffer( indexCount*sizeof(WORD), 0, D3DFMT_INDEX16, D3DPOOL_DEFAULT, &render->pIB ) ) ){
				throw this;
			}		
			if( FAILED( render->pIB->Lock( 0, indexCount*sizeof(WORD), (BYTE**)&pIndices, 0 ) ) ){
				throw this;
			}
#else
		if( FAILED( g_pd3dDevice->CreateIndexBuffer( indexCount*sizeof(WORD), 0, D3DFMT_INDEX16, D3DPOOL_DEFAULT, &render->pIB, 0) ) ){
			throw this;
		}
		if( FAILED( render->pIB->Lock( 0, indexCount*sizeof(WORD), (void**)&pIndices, 0 ) ) ){
            throw this;
		}
#endif
		for(int k = 0, fromPoly = 0; k < partList.length; k++){
			Art* part = partList[k];
			part->fromPoly = fromPoly;
			fromPoly += part->triangle.length;
		}
		for(int e = 0, j = 0; e < effectList.length; e++){
			Effect* effect = effectList[e];
			M_Art& partList = effect->partList;
			for(int k=0; k < partList.length; k++){
				Art* part=partList[k];
				coordLen+=part->coord.length;
				int triLen=part->triangle.length;
				if(part->show){
					for(int i = triLen, t = part->fromPoly * 4 + 2; --i>=0; ){
						pIndices[j++]=coordIndex[t];
						--t;
						pIndices[j++]=coordIndex[t];
						--t;
						pIndices[j++]=coordIndex[t];
						t+=6;
					}
				}
			}
			if( FAILED( render->pIB->Unlock() ) )
				throw this;
		}
		//{
		//	for(int t=2, h=0, j=0, k=0; k<partList.length; k++){
		//		Art* part=partList[k];
		//		coordLen+=part->coord.length;
		//		int triLen=part->triangle.length;
		//		if(part->show){
		//			for(int i=triLen; --i>=0; ){

		//				pIndices[j++]=coordIndex[t];
		//				--t;
		//				pIndices[j++]=coordIndex[t];
		//				--t;
		//				pIndices[j++]=coordIndex[t];
		//				t+=6;
		//			}
		//		}
		//		else{
		//			t+=triLen*4;
		//		}
		//	}
		//	if( FAILED( render->pIB->Unlock() ) )
		//		throw this;
		//}
	}
	if(effect.refValue()->getType()==COLOREFFECT){
		// Create the vertex buffer.
		render->type=COLOR_CUSTOMVERTEX_0WEIGHT;
		COLORVERTEX_0WEIGHT* pVertices;
#ifdef DX8
		if( FAILED( g_pd3dDevice->CreateVertexBuffer( coordLen*sizeof(COLORVERTEX_0WEIGHT),	0, COLOR_CUSTOMVERTEX_0WEIGHT, D3DPOOL_DEFAULT, &render->pVB ) ) ){
			throw this;
		}
		if( FAILED( render->pVB->Lock( 0, 0, (BYTE**)&pVertices, 0 ) ) ){
			throw this;
		}
#else
		if( FAILED( g_pd3dDevice->CreateVertexBuffer( coordLen*sizeof(COLORVERTEX_0WEIGHT), 0, COLOR_CUSTOMVERTEX_0WEIGHT, D3DPOOL_DEFAULT, &render->pVB, 0) ) ){
			throw this;
		}
		if( FAILED( render->pVB->Lock( 0, 0, (void**)&pVertices, 0 ) ) ){
			throw this;
		}
#endif
		{
			setUpInternalEffectPtrsTmpForRenderingTransparancy(this);

			for(int j=0, smi=0,k=0;k<partList.length;k++){

				Art* part=partList[k];

				part->renderFromCoord = j;
				float alpha = 1.0f - reinterpret_cast<ColorEffect*>(part->workEffect)->transparency;

				const M_Vec3f& coord=part->coord;
				const M_Vec3f& normal=part->normal;
				const M_Color3f& color=part->color;
				const M_Int& coordGroup=part->coordGroup;
				const M_Int& matrixIndex=part->matrixIndex;
				Vec3f absolutePivotPoint;
				if(workAnimCP == false){
					 absolutePivotPoint = part->absolutePivotPoint;
				}

				int i = 0;

				for(int cg=0;cg<coordGroup.length; cg++){
					for(int c=coordGroup[cg];--c>=0&&i<part->coord.length;i++, j++){
						pVertices[j].position = *(D3DXVECTOR3*) &(coord[i] - absolutePivotPoint);
						pVertices[j].normal = *(D3DXVECTOR3*) &normal[i];
						pVertices[j].matrixIndices = matrixIndex[i];
						pVertices[j].color = color[i].colorToHex();
					}
					smi++;
				}
				for(;i<part->coord.length;i++, j++){
					pVertices[j].position = *(D3DXVECTOR3*) &(coord[i] - absolutePivotPoint);
					pVertices[j].normal = *(D3DXVECTOR3*) &normal[i];
					pVertices[j].matrixIndices = matrixIndex[i];
					pVertices[j].color = color[i].colorToHex(alpha);
				}
				smi++;
			}
			setUpInternalEffectPtrs(this);
		}
	}
	else if(effect.refValue()->getType()==MONOCOLOREFFECT){
		// Create the vertex buffer.
		render->type=COLOR_CUSTOMVERTEX_0WEIGHT;
		COLORVERTEX_0WEIGHT* pVertices;
#ifdef DX8
		if( FAILED( g_pd3dDevice->CreateVertexBuffer( coordLen*sizeof(COLORVERTEX_0WEIGHT),	0, COLOR_CUSTOMVERTEX_0WEIGHT, D3DPOOL_DEFAULT, &render->pVB ) ) ){
			throw this;
		}
		if( FAILED( render->pVB->Lock( 0, 0, (BYTE**)&pVertices, 0 ) ) ){
			throw this;
		}
#else
		if( FAILED( g_pd3dDevice->CreateVertexBuffer( coordLen*sizeof(COLORVERTEX_0WEIGHT), 0, COLOR_CUSTOMVERTEX_0WEIGHT, D3DPOOL_DEFAULT, &render->pVB, 0) ) ){
			throw this;
		}
		if( FAILED( render->pVB->Lock( 0, 0, (void**)&pVertices, 0 ) ) ){
			throw this;
		}
#endif
		{
			buildPartList(); //why do we need this?
			setUpInternalEffectPtrs(this); //why do we need this?

			setUpInternalEffectPtrsTmpForRenderingTransparancy(this);

			for(int j=0, smi=0,k=0;k<partList.length;k++){

				Art* part=partList[k];

				part->renderFromCoord = j;
				if(MonoColorEffect* monoColorEffect=dynamic_cast<MonoColorEffect*>(part->workEffect)){

					float alpha = 1.0f - monoColorEffect->transparency;

					const M_Vec3f& coord=part->coord;
					const M_Vec3f& normal=part->normal;
					const Color3f color=monoColorEffect->diffuse;
					const M_Int& coordGroup=part->coordGroup;
					const M_Int& matrixIndex=part->matrixIndex;
					Vec3f absolutePivotPoint;
					if(workAnimCP == false){
						 absolutePivotPoint = part->absolutePivotPoint;
					}

					int i=0;

					for(int cg=0;cg<coordGroup.length; cg++){
						for(int c=coordGroup[cg];--c>=0&&i<part->coord.length;i++, j++){
							pVertices[j].position = *(D3DXVECTOR3*) &(coord[i] - absolutePivotPoint);
							pVertices[j].normal = *(D3DXVECTOR3*) &normal[i];
							pVertices[j].matrixIndices = matrixIndex[i];
							pVertices[j].color = color.colorToHex();
						}
						smi++;
					}
					for(;i<part->coord.length;i++, j++){
						pVertices[j].position = *(D3DXVECTOR3*) &(coord[i] - absolutePivotPoint);
						pVertices[j].normal = *(D3DXVECTOR3*) &normal[i];
						pVertices[j].matrixIndices = matrixIndex[i];
						pVertices[j].color = color.colorToHex(alpha);
					}
					smi++;
				}
			}
			setUpInternalEffectPtrs(this);
		}
	}
	else if(effect.refValue()->getType()==TEXTUREEFFECT){
		// Create the vertex buffer.
		render->type=TEXTURE_CUSTOMVERTEX_0WEIGHT;
		TEXTUREVERTEX_0WEIGHT* pVertices;
#ifdef DX8
		if( FAILED( g_pd3dDevice->CreateVertexBuffer( coordLen*sizeof(TEXTUREVERTEX_0WEIGHT), 0, TEXTURE_CUSTOMVERTEX_0WEIGHT, D3DPOOL_DEFAULT, &render->pVB ) ) ){
			throw this;
		}
		if( FAILED( render->pVB->Lock( 0, 0, (BYTE**)&pVertices, 0 ) ) ){
			throw this;
		}
#else
		if( FAILED( g_pd3dDevice->CreateVertexBuffer( coordLen*sizeof(TEXTUREVERTEX_0WEIGHT), 0, TEXTURE_CUSTOMVERTEX_0WEIGHT, D3DPOOL_DEFAULT, &render->pVB, 0) ) ){
			throw this;
		}
		if( FAILED( render->pVB->Lock( 0, 0, (void**)&pVertices, 0 ) ) ){
			throw this;
		}
#endif
		{
			for(int j=0, smi=0,k=0;k<partList.length;k++){

				Art* part=partList[k];

				part->renderFromCoord=j;

				const M_Vec3f& coord=part->coord;
				const M_Vec3f& normal=part->normal;
				const M_Vec2f& texCoord=part->texCoord;
				const M_Int& coordGroup=part->coordGroup.refValue();
				const M_Int& matrixIndex=part->matrixIndex;
				Vec3f absolutePivotPoint;
				if(workAnimCP == false){
					 absolutePivotPoint = part->absolutePivotPoint;
				}

				int i=0;

				for(int cg=0;cg<coordGroup.length; cg++){
					for(int c=coordGroup[cg];--c>=0&&i<part->coord.length;i++, j++){
						pVertices[j].position = *(D3DXVECTOR3*) &(coord[i] - absolutePivotPoint);
						pVertices[j].normal = *(D3DXVECTOR3*) &normal[i];
						pVertices[j].matrixIndices = matrixIndex[i];
						pVertices[j].tu = texCoord[i].x;
						pVertices[j].tv =1-texCoord[i].y;
					}
					smi++;
				}
				for(;i<part->coord.length;i++, j++){
					pVertices[j].position = *(D3DXVECTOR3*) &(coord[i] - absolutePivotPoint);
					pVertices[j].normal = *(D3DXVECTOR3*) &normal[i];
					pVertices[j].matrixIndices = matrixIndex[i];
					pVertices[j].tu = texCoord[i].x;
					pVertices[j].tv =1-texCoord[i].y;
				}
				smi++;
			}
		}
	}
	else throw this;
	render->vertexCount=coordLen;
	resetCoordLen( partList);
	render->pVB->Unlock();
	ctrf=false;
}

void Seamless::convertToSimpleRenderFormat(){
	if(partList.length==0||effect==0){
		if(render!=0){
			if( render->pVB != NULL ){
				render->pVB->Release();
			}
			if( render->pIB != NULL ){
				render->pIB->Release();
			}
			render=0;
		}
		return;
	}
	if(partList[0]->workEffect==0)return;
	if(render==NULL )render=new DXRender;
	int j=0;
	M_Triangle& triangle=partList[0]->triangle;
	int indexCount=triangle.length*3;
	render->indexCount=indexCount;
	int coordLen=0;

	if( render->pVB != NULL )
		render->pVB->Release();

	if( render->pIB != NULL )
		render->pIB->Release();

	{
        WORD* pIndices;
		
#ifdef DX8
		if( FAILED( g_pd3dDevice->CreateIndexBuffer( indexCount*sizeof(WORD), 0, D3DFMT_INDEX16, D3DPOOL_DEFAULT, &render->pIB ) ) ){
			throw this;
		}		
		if( FAILED( render->pIB->Lock( 0, indexCount*sizeof(WORD), (BYTE**)&pIndices, 0 ) ) ){
            throw this;
		}

#else
		if( FAILED( g_pd3dDevice->CreateIndexBuffer( indexCount*sizeof(WORD), 0, D3DFMT_INDEX16, D3DPOOL_DEFAULT, &render->pIB, 0) ) ){
			throw this;
		}
		if( FAILED( render->pIB->Lock( 0, indexCount*sizeof(WORD), (void**)&pIndices, 0 ) ) ){
            throw this;
		}
#endif
		{
			for(int t=2, h=0, j=0, k=0; k<partList.length; k++){
				Art* part=partList[k];
				coordLen+=part->coord.length;
				int triLen=part->triangle.length;
				if(part->show){
					for(int i=triLen; --i>=0; ){

						pIndices[j++]=triangle.refIndex(t);
						--t;
						pIndices[j++]=triangle.refIndex(t);
						--t;
						pIndices[j++]=triangle.refIndex(t);
						t+=5;
					}
				}
				else{
					t+=triLen*3;
				}
			}
			if( FAILED( render->pIB->Unlock() ) )
				throw this;
		}
	}
	// Create the vertex buffer.
	render->type=D3DFVF_SIMPLEVERTEX;
	SIMPLEVERTEX* pVertices;
#ifdef DX8
	if( FAILED( g_pd3dDevice->CreateVertexBuffer( coordLen*sizeof(SIMPLEVERTEX), 0, D3DFVF_SIMPLEVERTEX, D3DPOOL_DEFAULT, &render->pVB ) ) ){
		throw this;
	}
	if( FAILED( render->pVB->Lock( 0, 0, (BYTE**)&pVertices, 0 ) ) ){
		throw this;
	}
#else
	if( FAILED( g_pd3dDevice->CreateVertexBuffer( coordLen*sizeof(SIMPLEVERTEX), 0, D3DFVF_SIMPLEVERTEX, D3DPOOL_DEFAULT, &render->pVB, 0) ) ){
		throw this;
	}
	if( FAILED( render->pVB->Lock( 0, 0, (void**)&pVertices, 0 ) ) ){
		throw this;
	}
#endif
	{
		for(int j=0, smi=0,k=0;k<partList.length;k++){

			Art* part=partList[k];
			part->genNormals();

			part->renderFromCoord=j;

			const M_Vec3f& coord=part->coord;
			const M_Vec3f& normal=part->normal;
			Vec3f absolutePivotPoint;
			if(workAnimCP == false){
				 absolutePivotPoint = part->absolutePivotPoint;
			}

			for(int i=0; i<coord.length; i++,j++){
				pVertices[j].position = *(D3DXVECTOR3*) &(coord[i] - absolutePivotPoint);
				pVertices[j].normal = *(D3DXVECTOR3*) &normal[i];
			}
		}
	}
	render->vertexCount=coordLen;
	resetCoordLen( partList);
	render->pVB->Unlock();
	ctrf=false;
}

void HAnimJoint::setJointIndices(int& smi, int midx, COLORVERTEX_1WEIGHT* pVertices, M_Int& maskIndex){

	M_Int& skinCoordIndex=this->skinCoordIndex.refValue();

	int previousMatrixIndex=midx-1;
	if(previousMatrixIndex<0){
		previousMatrixIndex=0;
	}
	for(int i=skinCoordIndex.length; --i>=0;){
		int si = maskIndex[skinCoordIndex[i]];
		pVertices[si].matrixIndices=(previousMatrixIndex<<8)+smi;
		pVertices[si].blend0=skinCoordWeight[i];
	}
	smi++;
	midx=smi;
	for(int k=0;k<children.length;k++){
		HAnimJoint* joint=dynamic_cast<HAnimJoint*>(children[k].getRawPtr());
		if(joint!=NULL){
			joint->setJointIndices(smi, midx, pVertices, maskIndex);
		}
	}
}
void HAnimJoint::setJointIndices(int& smi, int midx, TEXTUREVERTEX_1WEIGHT* pVertices, M_Int& maskIndex){

	M_Int& skinCoordIndex=this->skinCoordIndex.refValue();

	int previousMatrixIndex=midx-1;
	if(previousMatrixIndex<0){
		previousMatrixIndex=0;
	}
	for(int i=skinCoordIndex.length; --i>=0;){
		int si = maskIndex[skinCoordIndex[i]];
		pVertices[si].matrixIndices=(previousMatrixIndex<<8)+smi;
		pVertices[si].blend0=skinCoordWeight[i];
	}
	smi++;
	midx=smi;
	for(int k=0;k<children.length;k++){
		HAnimJoint* joint=dynamic_cast<HAnimJoint*>(children[k].getRawPtr());
		if(joint!=NULL){
			joint->setJointIndices(smi, midx, pVertices, maskIndex);
		}
	}
}
/*
void HAnimJoint::setDisplacerBTS( const M_Int& coordMask){
	for(int d=displacers.length; --d>=0;){
		HAnimDisplacer* displacer=dynamic_cast<HAnimDisplacer*>(displacers[d].getRawPtr());

		const M_Int& coordIndex=displacer->coordIndex;
		M_Int& btsRcR=displacer->btsRcR;
		M_Int& btsRcU=displacer->btsRcU;

		btsRcR.length=0;
		btsRcU.length=0;

		for(int i=coordIndex.length; --i>=0; ){
			int original=coordIndex[i];
			int bci=original;
			while(true){
				bci=coordMask[bci];

				if(bci==-1) break;
				btsRcR+=bci;
				btsRcU+=original;
			}
		}
	}
	for(int k=0;k<children.length;k++){
		HAnimJoint* joint=dynamic_cast<HAnimJoint*>(children[k].getRawPtr());
		if(joint!=NULL){
			joint->setDisplacerBTS(coordMask);
		}
	}
}
void HAnimHumanoid::reOrderFieldIndices(IndexedFaceSet& ifs, GeomSource& geomSource, M_Int& maskIndex){
	ifs.coordIndex = 
	if(Coordinate* coordinate = dynamic_cast<Coordinate*>(ifs.coord.p)){
		coordinate->point.refValue()=geomSource.coord;
	}
	if(Normal* normal = dynamic_cast<Normal*>(ifs.normal.p)){
		normal->vector.refValue()=geomSource.normal;
	}
	if(TextureCoordinate* textureCoordinate = dynamic_cast<TextureCoordinate*>(ifs.texCoord.p)){
		textureCoordinate->point.refValue()=geomSource.texCoord;
	}
	if(Color* color = dynamic_cast<Color*>(ifs.color.p)){
		color->color.refValue()=geomSource.color;
	}
	for(int k = joints.length; --k>=0;){
		HAnimJoint* joint = dynamic_cast<HAnimJoint*>(joints[k].p);
		M_Int& skinCoordIndex = joint->skinCoordIndex.refValue();
		for(int i = skinCoordIndex.length; --i>=0;){
			skinCoordIndex[i] = maskIndex[skinCoordIndex[i]];
		}
	}
}
*/
void HAnimHumanoid::convertToRenderFormat(DXRender*& render, const Appearance* appearance){
	if(skeleton.length==0){
		if(render!=0){
			if( render->pVB != NULL ){
				render->pVB->Release();
			}
			if( render->pIB != NULL ){
				render->pIB->Release();
			}
			render=0;
		}
		return;
	}
	if(skin.length==0)return;
	Shape* shape=dynamic_cast<Shape*>(skin[0].getRawPtr());
	if(shape==0)return;
	IndexedFaceSet* ifs=dynamic_cast<IndexedFaceSet*>(shape->geometry.getRawPtr());
	IndexedTriangleSet* its=dynamic_cast<IndexedTriangleSet*>(shape->geometry.getRawPtr());
	if(ifs==0 && its ==0)return;

	Coordinate* skinCoord=dynamic_cast<Coordinate*>(this->skinCoord.getRawPtr());
	if(skinCoord==0)return;

	if(render==NULL )render=new DXRender;
	int j=0;

	GeomSourcePtr geomSource=new GeomSource;
	if(ifs){
		ifs->convertToGeomSource(*geomSource.p,appearance, maskIndex);
	}
	else{
		its->convertToGeomSource(*geomSource.p,appearance, maskIndex);
	}
	//reOrderFieldIndices(*ifs, *geomSource, maskIndex);

	const M_Int& coordIndex=geomSource->coordIndex;
	int indexCount=coordIndex.length;

	render->indexCount=indexCount;

	if(!indexCount){
		return;
	}
	if( render->pVB != NULL )
		render->pVB->Release();

	if( render->pIB != NULL )
		render->pIB->Release();

	{
        WORD* pIndices;
#ifdef DX8
		if( FAILED( g_pd3dDevice->CreateIndexBuffer( indexCount*sizeof(WORD), 0, D3DFMT_INDEX16, D3DPOOL_DEFAULT, &render->pIB ) ) ){
			throw this;
		}		
		if( FAILED( render->pIB->Lock( 0, indexCount*sizeof(WORD), (BYTE**)&pIndices, 0 ) ) ){
            throw this;
		}
#else
		if( FAILED( g_pd3dDevice->CreateIndexBuffer( indexCount*sizeof(WORD), 0, D3DFMT_INDEX16, D3DPOOL_DEFAULT, &render->pIB, 0) ) ){
			throw this;
		}		
		if( FAILED( render->pIB->Lock( 0, indexCount*sizeof(WORD), (void**)&pIndices, 0 ) ) ){
            throw this;
		}
#endif
		{
			for(int t=2, j=0;  j<coordIndex.length; ){
				pIndices[j++]=coordIndex[t];
				--t;
				pIndices[j++]=coordIndex[t];
				--t;
				pIndices[j++]=coordIndex[t];
				t+=5;
			}
			if( FAILED( render->pIB->Unlock() ) )
				throw this;
		}
	}
	const M_Vec3f& coord=geomSource->coord;
	const int coordLen=coord.length;
	const M_Vec3f& normal=geomSource->normal;
	const M_Int& rc=geomSource->rc;
	const M_Color3f& color=geomSource->color;
	const M_Vec2f& texCoord=geomSource->texCoord;
	if(normal.length<coord.length){
		geomSource->genNormals();
	}
	int renderVertexLen=coord.length+rc.length;
	
	{
		coordMask.length=renderVertexLen;
		{
			for(int i=renderVertexLen; --i>=0; ){
				coordMask[i]=-1;
			}
		}
		{
			for(int i=rc.length; --i>=0; ){
				int bci=rc[i];
				while(true){
					int rci=bci;
					bci=coordMask[bci];

					if(bci==-1){
						coordMask[rci]=i+coord.length;
						break;
					}
				}
			}
		}
	}


	if(color.length>0){
		// Create the vertex buffer.
		render->type=COLOR_CUSTOMVERTEX_1WEIGHT;
		COLORVERTEX_1WEIGHT* pVertices;
#ifdef DX8
		if( FAILED( g_pd3dDevice->CreateVertexBuffer( renderVertexLen*sizeof(COLORVERTEX_1WEIGHT), 0, COLOR_CUSTOMVERTEX_1WEIGHT, D3DPOOL_DEFAULT, &render->pVB ) ) ){
			throw this;
		}
		if( FAILED( render->pVB->Lock( 0, 0, (BYTE**)&pVertices, 0 ) ) ){
			throw this;
		}
#else
		if( FAILED( g_pd3dDevice->CreateVertexBuffer( renderVertexLen*sizeof(COLORVERTEX_1WEIGHT), 0, COLOR_CUSTOMVERTEX_1WEIGHT, D3DPOOL_DEFAULT, &render->pVB, 0) ) ){
			throw this;
		}
		if( FAILED( render->pVB->Lock( 0, 0, (void**)&pVertices, 0 ) ) ){
			throw this;
		}
#endif
		{
			int smi=0;
			for(int k=0;k<skeleton.length;k++){
				dynamic_cast<HAnimJoint*>(skeleton[k].getRawPtr())->setJointIndices(smi,0, pVertices, maskIndex);
			}

			int i=0;
			for(i=coord.length; --i>=0;){
				pVertices[i].position = *(D3DXVECTOR3*) &coord[i];
				pVertices[i].normal = *(D3DXVECTOR3*) &normal[i];
				pVertices[i].color = color[i].colorToHex();
			}
			int j=renderVertexLen;
			for(i=rc.length; --i>=0;){
				int rci=rc[i];
				--j;
				pVertices[j].position = *(D3DXVECTOR3*) &coord[rci];
				pVertices[j].normal = *(D3DXVECTOR3*) &normal[rci];
				pVertices[j].color = color[i+coordLen].colorToHex();
				pVertices[j].matrixIndices=pVertices[rci].matrixIndices;
				pVertices[j].blend0=pVertices[rci].blend0;
			}
		}
	}
	else if(texCoord.length>0){
		// Create the vertex buffer.
		render->type=TEXTURE_CUSTOMVERTEX_1WEIGHT;
		TEXTUREVERTEX_1WEIGHT* pVertices;
#ifdef DX8
		if( FAILED( g_pd3dDevice->CreateVertexBuffer( renderVertexLen*sizeof(TEXTUREVERTEX_1WEIGHT), 0, TEXTURE_CUSTOMVERTEX_1WEIGHT, D3DPOOL_DEFAULT, &render->pVB ) ) ){
			throw this;
		}
		if( FAILED( render->pVB->Lock( 0, 0, (BYTE**)&pVertices, 0 ) ) ){
			throw this;
		}
#else
		if( FAILED( g_pd3dDevice->CreateVertexBuffer( renderVertexLen*sizeof(TEXTUREVERTEX_1WEIGHT), 0, TEXTURE_CUSTOMVERTEX_1WEIGHT, D3DPOOL_DEFAULT, &render->pVB, 0) ) ){
			throw this;
		}
		if( FAILED( render->pVB->Lock( 0, 0, (void**)&pVertices, 0 ) ) ){
			throw this;
		}
#endif
		{
			int smi=0;
			for(int k=0;k<skeleton.length;k++){
				dynamic_cast<HAnimJoint*>(skeleton[k].getRawPtr())->setJointIndices(smi,0, pVertices, maskIndex);
			}

			int i=0;
			for(i=coord.length; --i>=0;){
				pVertices[i].position = *(D3DXVECTOR3*) &coord[i];
				pVertices[i].normal = *(D3DXVECTOR3*) &normal[i];
				pVertices[i].tu = texCoord[i].x;
				pVertices[i].tv =1-texCoord[i].y;
			}
			int j=renderVertexLen;
			for(i=rc.length; --i>=0;){
				int rci=rc[i];
				int h=i+coordLen;
				--j;
				pVertices[j].position = *(D3DXVECTOR3*) &coord[rci];
				pVertices[j].normal = *(D3DXVECTOR3*) &normal[rci];
				pVertices[j].tu =  texCoord[h].x;
				pVertices[j].tv =1-texCoord[h].y;
				pVertices[j].matrixIndices=pVertices[rci].matrixIndices;
				pVertices[j].blend0=pVertices[rci].blend0;
			}
		}
	}
	render->vertexCount=renderVertexLen;
	render->pVB->Unlock();
}
void HAnimDisplacer::Weight::operator =(const float& interp){
	HAnimHumanoid& humanoid = *(home.humanoid);
	DXRender* render=humanoid.render;
	if(render==0)return;
	Coordinate* skinCoord=dynamic_cast<Coordinate*>(humanoid.skinCoord.getRawPtr());
	if(skinCoord==0)return;
	const M_Vec3f& point=skinCoord->point;
	const M_Vec3f& displacements=home.displacements;
	const M_Int& coordIndex=home.coordIndex;
	const M_Int& coordMask=humanoid.coordMask;


	switch(render->type){
	case COLOR_CUSTOMVERTEX_1WEIGHT:{
		COLORVERTEX_1WEIGHT* pVertices;
#ifdef DX8
		if( FAILED( render->pVB->Lock( 0, 0, (BYTE**)&pVertices, 0 ) ) ){
			throw humanoid;
		}
#else
		if( FAILED( render->pVB->Lock( 0, 0, (void**)&pVertices, 0 ) ) ){
			throw humanoid;
		}
#endif
		for(int i=coordIndex.length; --i>=0; ){
			int original=coordIndex[i];
			int ci=original;
			do{
				pVertices[ci].position=(D3DXVECTOR3&) (point[original]+displacements[i]*Vec3f(interp,interp,interp));
				ci=coordMask[ci];
			}while(ci!=-1);
		}
		break;
	}
	case TEXTURE_CUSTOMVERTEX_1WEIGHT:{
		TEXTUREVERTEX_1WEIGHT* pVertices;
#ifdef DX8
		if( FAILED( render->pVB->Lock( 0, 0, (BYTE**)&pVertices, 0 ) ) ){
			throw humanoid;
		}
#else
		if( FAILED( render->pVB->Lock( 0, 0, (void**)&pVertices, 0 ) ) ){
			throw humanoid;
		}
#endif
		for(int i=coordIndex.length; --i>=0; ){
			int original=coordIndex[i];
			int ci=original;
			do{
				pVertices[ci].position=(D3DXVECTOR3&) (point[original]+displacements[i]*Vec3f(interp,interp,interp));
				ci=coordMask[ci];
			}while(ci!=-1);
		}
		break;
	}
	default:
		throw this;
	}
	render->pVB->Unlock();
}


void Morpher::Weight::operator =(const float& interp){
	const Art& part=*(home.parent);
	const Seamless& smls=*(part.smls);
	const M_Vec3f& coord=part.coord;
	const M_Vec3f& displacement=home.displacement;
	DXRender* render=smls.render;
	if(render==0)return;
	switch(render->type){
	case COLOR_CUSTOMVERTEX_0WEIGHT:{
		COLORVERTEX_0WEIGHT* pVertices;
#ifdef DX8
		if( FAILED( render->pVB->Lock( 0, 0, (BYTE**)&pVertices, 0 ) ) ){
			throw this;
		}
#else
		if( FAILED( render->pVB->Lock( 0, 0, (void**)&pVertices, 0 ) ) ){
			throw this;
		}
#endif

		int toCoord=part.renderFromCoord+part.coord.length;
		int renderFromCoord=part.renderFromCoord;

		for(int j=renderFromCoord+part.coord.length, i=part.coord.length; --i>=0; ){
			pVertices[--j].position=(D3DXVECTOR3&) (coord[i]+displacement[i]*Vec3f(interp,interp,interp));
		}
		int mark=renderFromCoord+part.coord.length;
		const M_Int& rc=part.rc;
		for(int i=part.rc.length; --i>=0; ){
			pVertices[mark+i].position=pVertices[renderFromCoord+rc[i]].position;
		}
		break;
	}
	case TEXTURE_CUSTOMVERTEX_0WEIGHT:{
		TEXTUREVERTEX_0WEIGHT* pVertices;
#ifdef DX8
		if( FAILED( render->pVB->Lock( 0, 0, (BYTE**)&pVertices, 0 ) ) ){
			throw this;
		}
#else
		if( FAILED( render->pVB->Lock( 0, 0, (void**)&pVertices, 0 ) ) ){
			throw this;
		}
#endif

		int toCoord=part.renderFromCoord+part.coord.length;
		int renderFromCoord=part.renderFromCoord;

		for(int j=renderFromCoord+part.coord.length, i=part.coord.length; --i>=0; ){
			pVertices[--j].position=(D3DXVECTOR3&) (coord[i]+displacement[i]*Vec3f(interp,interp,interp));
		}
		int mark=renderFromCoord+part.coord.length;
		const M_Int& rc=part.rc;
		for(int i=part.rc.length; --i>=0; ){
			pVertices[mark+i].position=pVertices[renderFromCoord+rc[i]].position;
		}
		break;
	}
	default:
		throw this;
	}

	render->pVB->Unlock();
}


template<typename T> void dragPartForMorph2(DXRender*& render, Art*& part, Vec3f newPt, Vec3f diff){
	T* pVertices;
#ifdef DX8
	if( FAILED( render->pVB->Lock( 0, 0, (BYTE**)&pVertices, 0 ) ) ){
		throw;
	}
#else
	if( FAILED( render->pVB->Lock( 0, 0, (void**)&pVertices, 0 ) ) ){
		throw;
	}
#endif
	for(int i = MultiSelect::coordList.length; --i >= 0;){
		Art* part = MultiSelect::coordList[i].p;
		int idx = MultiSelect::coordList[i].i;
		M_Vec3f& coord = part->coord.refValue();
		int renderFromCoord = part->renderFromCoord;
		(*(Vec3f*)&pVertices[renderFromCoord + idx].position) = (g_m_coord[part->fromCoord + idx] = MultiSelect::beginDragPos[i] + diff) - part->absolutePivotPoint;

		int mark = renderFromCoord + part->coord.length; //if need be we could make this more efficiant by flagging each part and doing this after each element in the coordList has been processed
		const M_Int& rc = part->rc;
		for(int i = rc.length; --i>=0; ){
			pVertices[mark + i].position = pVertices[renderFromCoord + rc[i]].position;
		}
	}

	//smls->morphFlag = true;
	render->pVB->Unlock();
}


void AnimClue::dragPartForMorph(Vec3f pt){
	if(Art* part = dynamic_cast<Art*>(g_selNode.getRaw())){
		Seamless* smls = part->smls;
		g_dragActuallyHappned = true;
		DXRender* render=smls->render;
		const M_Int& coordGroup = part->coordGroup;
		const M_Float& weight = part->weight;

		smls->catIMatrix();
		Matrix imatrix = part->plidx > 0 ? smls->partList[part->plidx - 1]->imatrix : smls->imatrix;

		Vec3f translation = part->translation.refValue();
		Rotation rotation = part->rotation.refValue();
		Vec3f scale = part->scale.refValue();
		Vec3f absolutePivotPoint = part->absolutePivotPoint;
		
		int idx = part->index[0];
		int coordsInGroups = 0;
		float w = 1;

		for(int cg = 0; cg < coordGroup.length; cg++){
			coordsInGroups += coordGroup[cg];
			if(idx < coordsInGroups){
				w = weight[cg];
				break;
			}
		}

		rotation.angle = rotation.angle * w;

		imatrix = imatrix * translationMatrix(-translation - absolutePivotPoint) * rotationMatrix( -rotation) * scaleMatrix(Vec3f(1./scale.x, 1./scale.y, 1./scale.z)) * translationMatrix(absolutePivotPoint);

		Vec3f newPt = g_scene->getIViewMatrix() * g_catchIMatrix * imatrix * pt;
		int i = part->index[0];
		Vec3f diff = newPt - MultiSelect::beginDragPos[MultiSelect::clickedIndex];

		switch(render->type){
			case COLOR_CUSTOMVERTEX_0WEIGHT:{
				dragPartForMorph2<COLORVERTEX_0WEIGHT>(render, part, newPt, diff);
				break;
			}
			case TEXTURE_CUSTOMVERTEX_0WEIGHT:{
				dragPartForMorph2<TEXTUREVERTEX_0WEIGHT>(render, part, newPt, diff);
				break;
			}
		}
	}
}

void renderMorphPart(const Art& part, const M_Vec3f& srcCoordA, const M_Vec3f& srcCoordB, float interp){
	Seamless* smls=part.smls;
	DXRender* render=smls->render;
	if(render==0)return;
	if(render->type==COLOR_CUSTOMVERTEX_0WEIGHT){
		COLORVERTEX_0WEIGHT* pVertices;
#ifdef DX8
		if( FAILED( render->pVB->Lock( 0, 0, (BYTE**)&pVertices, 0 ) ) ){
			throw;
		}
#else
		if( FAILED( render->pVB->Lock( 0, 0, (void**)&pVertices, 0 ) ) ){
			throw;
		}
#endif

		int toCoord=part.renderFromCoord+part.coord.length;

		int renderFromCoord=part.renderFromCoord;
		for(int j=renderFromCoord+part.coord.length, i=part.coord.length; --i>=0; ){
			pVertices[--j].position=*(D3DXVECTOR3*) &(srcCoordA[i]+(srcCoordB[i]-srcCoordA[i])*Vec3f(interp,interp,interp));
		}
		int mark=renderFromCoord+part.coord.length;
		const M_Int& rc=part.rc;
		for(int i=part.rc.length; --i>=0; ){
			pVertices[mark+i].position=pVertices[renderFromCoord+rc[i]].position;
		}
		render->pVB->Unlock();
	}
	else if(render->type==TEXTURE_CUSTOMVERTEX_0WEIGHT){
		TEXTUREVERTEX_0WEIGHT* pVertices;
#ifdef DX8
		if( FAILED( render->pVB->Lock( 0, 0, (BYTE**)&pVertices, 0 ) ) ){
			throw;
		}
#else
		if( FAILED( render->pVB->Lock( 0, 0, (void**)&pVertices, 0 ) ) ){
			throw;
		}
#endif

		int toCoord=part.renderFromCoord+part.coord.length;

		int renderFromCoord=part.renderFromCoord;
		for(int j=renderFromCoord+part.coord.length, i=part.coord.length; --i>=0; ){
			pVertices[--j].position=*(D3DXVECTOR3*) &(srcCoordA[i]+(srcCoordB[i]-srcCoordA[i])*Vec3f(interp,interp,interp));
		}
		int mark=renderFromCoord+part.coord.length;
		const M_Int& rc=part.rc;
		for(int i=part.rc.length; --i>=0; ){
			pVertices[mark+i].position=pVertices[renderFromCoord+rc[i]].position;
		}
		render->pVB->Unlock();
	}
	else throw;
}


void Seamless::updateAllVertexInBuffers(){
	ctrf = true;
	return;
	setMathHiPresision();
	if(workAnimCP == false && g_colorOwnership){
		convertToOwnershipAidRenderFormat();
		return;
	}

	if(partList.length==0 || partList[0]->workEffect==0)return;
	if(render==0 )return;
	int j=0;

	static GeomSource* geomSource=new GeomSource;

	int coordLen=0;

	M_Int& coordIndex=geomSource->coordIndex.refValue();

	catTranslationMatrix();
	geomSource->genTransformedNormals(partList);
	catMatrix();

	addExtraCoordsRenderForGeomSource(coordIndex);

	int indexCount=coordIndex.length;
	render->indexCount=indexCount;

	if(!indexCount){
		return;
	}
	{
		for(int k=0; k<partList.length; k++){
			coordLen+=partList[k]->coord.length;
		}
	}
	if(dynamic_cast<ColorEffect*>(effect.p)){
		COLORVERTEX_0WEIGHT* pVertices;
#ifdef DX8
		if( FAILED( render->pVB->Lock( 0, 0, (BYTE**)&pVertices, 0 ) ) ){
			throw this;
		}
#else
		if( FAILED( render->pVB->Lock( 0, 0, (void**)&pVertices, 0 ) ) ){
			throw this;
		}
#endif
		for(int j=0, smi=0,k=0;k<partList.length;k++){

			Art* part=partList[k];

			part->renderFromCoord=j;

			const M_Vec3f& coord=part->coord;
			const M_Vec3f& normal=part->normal;
			const M_Color3f& color=part->color;
			M_Int& coordGroup=part->coordGroup.refValue();
			M_Int& matrixIndex=part->matrixIndex;
			Vec3f absolutePivotPoint;
			if(workAnimCP == false){
				 absolutePivotPoint = part->absolutePivotPoint;
			}

			int i=0;

			for(int cg=0;cg<coordGroup.length; cg++){
				for(int c=coordGroup[cg];--c>=0&&i<part->coord.length;i++, j++){
					pVertices[j].position = *(D3DXVECTOR3*) &(coord[i] - absolutePivotPoint);
					pVertices[j].normal = *(D3DXVECTOR3*) &normal[i];
					pVertices[j].matrixIndices = matrixIndex[i];
					pVertices[j].color = color[i].colorToHex();
				}
				smi++;
			}
			for(;i<part->coord.length;i++, j++){
				pVertices[j].position = *(D3DXVECTOR3*) &(coord[i] - absolutePivotPoint);
				pVertices[j].normal = *(D3DXVECTOR3*) &normal[i];
				pVertices[j].matrixIndices = matrixIndex[i];
				pVertices[j].color = color[i].colorToHex();
			}
			smi++;
		}
	}
	else if(dynamic_cast<TextureEffect*>(effect.p)){
		TEXTUREVERTEX_0WEIGHT* pVertices;
#ifdef DX8
		if( FAILED( render->pVB->Lock( 0, 0, (BYTE**)&pVertices, 0 ) ) ){
			throw this;
		}
#else
		if( FAILED( render->pVB->Lock( 0, 0, (void**)&pVertices, 0 ) ) ){
			throw this;
		}
#endif
		for(int j=0, smi=0,k=0;k<partList.length;k++){

			Art* part=partList[k];

			part->renderFromCoord=j;

			const M_Vec3f& coord=part->coord;
			const M_Vec3f& normal=part->normal;
			const M_Vec2f& texCoord=part->texCoord;
			M_Int& coordGroup=part->coordGroup.refValue();
			M_Int& matrixIndex=part->matrixIndex;
			Vec3f absolutePivotPoint;
			if(workAnimCP == false){
				 absolutePivotPoint = part->absolutePivotPoint;
			}

			int i=0;

			for(int cg=0;cg<coordGroup.length; cg++){
				for(int c=coordGroup[cg];--c>=0&&i<part->coord.length;i++, j++){
					pVertices[j].position = *(D3DXVECTOR3*) &(coord[i] - absolutePivotPoint);
					pVertices[j].normal = *(D3DXVECTOR3*) &normal[i];
					pVertices[j].matrixIndices = matrixIndex[i];
					pVertices[j].tu = texCoord[i].x;
					pVertices[j].tv =1-texCoord[i].y;
				}
				smi++;
			}
			for(;i<part->coord.length;i++, j++){
				pVertices[j].position = *(D3DXVECTOR3*) &(coord[i] - absolutePivotPoint);
				pVertices[j].normal = *(D3DXVECTOR3*) &normal[i];
				pVertices[j].matrixIndices = matrixIndex[i];
				pVertices[j].tu = texCoord[i].x;
				pVertices[j].tv =1-texCoord[i].y;
			}
			smi++;
		}
	}
	else if(dynamic_cast<MonoColorEffect*>(effect.p)){
		COLORVERTEX_0WEIGHT* pVertices;
#ifdef DX8
		if( FAILED( render->pVB->Lock( 0, 0, (BYTE**)&pVertices, 0 ) ) ){
			throw this;
		}
#else
		if( FAILED( render->pVB->Lock( 0, 0, (void**)&pVertices, 0 ) ) ){
			throw this;
		}
#endif
		for(int j=0, smi=0,k=0;k<partList.length;k++){

			Art* part=partList[k];

			part->renderFromCoord=j;

			const M_Vec3f& coord=part->coord;
			const M_Vec3f& normal=part->normal;
			const Color3f color=dynamic_cast<MonoColorEffect*>(part->workEffect)->diffuse;
			M_Int& coordGroup=part->coordGroup.refValue();
			M_Int& matrixIndex=part->matrixIndex;
			Vec3f absolutePivotPoint;
			if(workAnimCP == false){
				 absolutePivotPoint = part->absolutePivotPoint;
			}

			int i=0;

			for(int cg=0;cg<coordGroup.length; cg++){
				for(int c=coordGroup[cg];--c>=0&&i<part->coord.length;i++, j++){
					pVertices[j].position = *(D3DXVECTOR3*) &(coord[i] - absolutePivotPoint);
					pVertices[j].normal = *(D3DXVECTOR3*) &normal[i];
					pVertices[j].matrixIndices = matrixIndex[i];
					pVertices[j].color = color.colorToHex();
				}
				smi++;
			}
			for(;i<part->coord.length;i++, j++){
				pVertices[j].position = *(D3DXVECTOR3*) &(coord[i] - absolutePivotPoint);
				pVertices[j].normal = *(D3DXVECTOR3*) &normal[i];
				pVertices[j].matrixIndices = matrixIndex[i];
				pVertices[j].color = color.colorToHex();
			}
			smi++;
		}
	}

	/*
	else if(render->type==TEXTURE_CUSTOMVERTEX_0WEIGHT){;
		TEXTUREVERTEX_0WEIGHT* pVertices;
		if( FAILED( render->pVB->Lock( 0, 0, (BYTE**)&pVertices, 0 ) ) ){
			throw this;
		}
		for(int j=0, smi=0,k=0;k<partList.length;k++){

			Art* part=partList[k];
			coordLen=part->coord.length;

			part->renderFromCoord=j;

			const M_Vec3f& coord=part->coord;
			const M_Vec3f& normal=part->normal;
			const M_Vec2f& texCoord=part->texCoord;
			M_Int& coordGroup=part->coordGroup.refValue();
			M_Int& matrixIndex=part->matrixIndex;

			int i=0;

			for(int cg=0;cg<coordGroup.length; cg++){
				for(int c=coordGroup[cg];--c>=0&&i<part->coord.length;i++, j++){
					pVertices[j].position = *(D3DXVECTOR3*) &coord[i];
					pVertices[j].normal = *(D3DXVECTOR3*) &normal[i];
					pVertices[j].matrixIndices = matrixIndex[i];
					pVertices[j].tu = texCoord[i].x;
					pVertices[j].tv =1-texCoord[i].y;
				}
				smi++;
			}
			for(;i<coordLen;i++, j++){
				pVertices[j].position = *(D3DXVECTOR3*) &coord[i];
				pVertices[j].normal = *(D3DXVECTOR3*) &normal[i];
				pVertices[j].matrixIndices = matrixIndex[i];
				pVertices[j].tu = texCoord[i].x;
				pVertices[j].tv =1-texCoord[i].y;
			}
			smi++;
		}
	}
	*/
	else throw this;
	render->vertexCount=coordLen;
	resetCoordLen( partList);
	render->pVB->Unlock();
}


void checkFieldsMatch(IndexedFaceSet &geometry){
	Coordinate* coord=dynamic_cast<Coordinate*>(geometry.coord.getRawPtr());
	Normal* normal=dynamic_cast<Normal*>(geometry.normal.getRawPtr());
	Color* color=dynamic_cast<Color*>(geometry.color.getRawPtr());
	TextureCoordinate* texCoord=dynamic_cast<TextureCoordinate*>(geometry.texCoord.getRawPtr());
	if(coord!=NULL){
		if(normal!=NULL && coord->point.refValue().length != normal->vector.refValue().length){
			dLog("coord to normal mismatch");
			normal->vector.resize(coord->point.refValue().length);
			geometry.genNormals();
			//throw geometry;
		}
		if(color!=NULL && coord->point.refValue().length != color->color.refValue().length){
			dLog("**error!** coord to color mismatch");
			color->color.resize(coord->point.refValue().length);
			//throw geometry;
		}
		if(texCoord!=NULL && coord->point.refValue().length != texCoord->point.refValue().length){
			dLog("**error!** coord to texCoord mismatch");
			texCoord->point.resize(coord->point.refValue().length);
			//throw geometry;
		}

	}
	else{
		if(normal!=NULL){
			dLog("**error!** normal != NULL");
			normal=NULL;
//			throw geometry;
		}
		if(color!=NULL){
			dLog("**error!** color != NULL");
			color=NULL;
//			throw geometry;
		}
		if(texCoord!=NULL){
			dLog("**error!** texCoord != NULL");
			texCoord=NULL;
//			throw geometry;
		}
	}
}
/*
void Art::genDynamicTransforms(Matrix parentMatrix, int& smi){
	
	pivotPointMatrix = translationMatrix(absolutePivotPoint) * translationMatrix(translation)*parentMatrix;

	for(int cg=0;cg<coordGroup.length; cg++){
		g_pd3dDevice->SetTransform( D3DTS_WORLDMATRIX(smi), (D3DMATRIX*)&(translationMatrix(-absolutePivotPoint)*scaleMatrix(scale)*rotationMatrix(Rotation(rotation.x,rotation.y,rotation.z, rotation.angle*weight[cg]))*pivotPointMatrix) );
		smi++;
	}

	actualPivotPointMatrix=scaleMatrix(scale)*rotationMatrix(rotation)*pivotPointMatrix;
	pivotPointMatrix=translationMatrix(-absolutePivotPoint)*actualPivotPointMatrix;


	g_pd3dDevice->SetTransform( D3DTS_WORLDMATRIX(smi), (D3DMATRIX*)&pivotPointMatrix );
	smi++;
	for(int k=0;k<part.length;k++){
		part.refValue()[k]->genDynamicTransforms(pivotPointMatrix, smi);
	}
}
*/

void Art::genDynamicTransforms(Matrix parentMatrix, int& smi){
	
	pivotPointMatrix = translationMatrix(translation + pivotPoint) * parentMatrix;

	for(int cg = 0; cg < coordGroup.length; cg++){
		g_pd3dDevice->SetTransform( D3DTS_WORLDMATRIX(smi), (D3DMATRIX*)&(scaleMatrix(scale) * rotationMatrix(Rotation(rotation.x, rotation.y, rotation.z, rotation.angle * weight[cg])) * pivotPointMatrix) );
		smi++;
	}

	pivotPointMatrix = scaleMatrix(scale) * rotationMatrix(rotation) * pivotPointMatrix;

	g_pd3dDevice->SetTransform( D3DTS_WORLDMATRIX(smi), (D3DMATRIX*)&pivotPointMatrix );
	smi++;
	for(int k = 0; k < part.length; k++){
		part.refValue()[k]->genDynamicTransforms(pivotPointMatrix, smi);
	}
}

void Art::genDynamicTransformsCP(Matrix parentMatrix, Matrix parentMatrixCP, int& smi){
	
	pivotPointMatrix = translationMatrix(translation + pivotPoint) * parentMatrix;

	for(int cg=0;cg<coordGroup.length; cg++){
		g_pd3dDevice->SetTransform( D3DTS_WORLDMATRIX(smi), (D3DMATRIX*)&parentMatrixCP);
		smi++;
	}

	pivotPointMatrix = scaleMatrix(scale) * rotationMatrix(rotation) * pivotPointMatrix;

	g_pd3dDevice->SetTransform( D3DTS_WORLDMATRIX(smi), (D3DMATRIX*)&parentMatrixCP);
	smi++;
	for(int k = 0; k < part.length; k++){
		part.refValue()[k]->genDynamicTransformsCP(pivotPointMatrix, parentMatrixCP, smi);
	}
}

void HAnimJoint::genDynamicTransforms(Matrix parentMatrix, int& smi){
	Matrix matrix=  
	translationMatrix(-center)*
	rotationMatrix(-scaleOrientation)*
	scaleMatrix(scale)*
	rotationMatrix(scaleOrientation)*
	rotationMatrix(rotation)*
	translationMatrix(center)*
	translationMatrix(translation)*
	parentMatrix;
	
	g_pd3dDevice->SetTransform( D3DTS_WORLDMATRIX(smi++), (D3DMATRIX*)&matrix);

	for(int k=0;k<children.length;k++){
		HAnimJoint* joint=dynamic_cast<HAnimJoint*>(children[k].getRawPtr());
		if(joint!=NULL){
			joint->genDynamicTransforms(matrix, smi);
		}
	}
}
extern Shape* g_vcEllipticalPolyline2D;
extern Shape* g_transformLeverClue;
extern ShapePtr g_vcFuseSurfaceFlatStageAidShape;
/*
void renderPivotPoint(Art* part, D3DXMATRIX& tranMatrix, Vec3f& pivotPointVec){
	g_pd3dDevice->SetRenderState( D3DRS_FILLMODE, D3DFILL_SOLID );
	pivotPointVec  = part->actualPivotPointMatrix * Vec3f(0,0,0);
	D3DXMatrixTranslation( &tranMatrix, pivotPointVec.x, pivotPointVec.y, pivotPointVec.z);

	g_pd3dDevice->SetTransform( D3DTS_WORLD, &tranMatrix);

#ifdef DX8
	g_pd3dDevice->SetStreamSource( 0, g_yellowRenderPoint->pVB, sizeof(COLOR_VERTEX) );
	g_pd3dDevice->SetVertexShader( D3DFVF_COLOR_VERTEX );
	g_pd3dDevice->SetIndices( g_yellowRenderPoint->pIB, 0 );
	g_pd3dDevice->DrawIndexedPrimitive( D3DPT_TRIANGLELIST, 0, g_yellowPoint->coord.length, 0, g_yellowPoint->coordIndex.refValue().length/3 );
#else
	g_pd3dDevice->SetStreamSource( 0, g_yellowRenderPoint->pVB, 0, sizeof(COLOR_VERTEX) );
	g_pd3dDevice->SetFVF( D3DFVF_COLOR_VERTEX );
	g_pd3dDevice->SetIndices( g_yellowRenderPoint->pIB);
	g_pd3dDevice->DrawIndexedPrimitive( D3DPT_TRIANGLELIST, 0, 0, g_yellowPoint->coord.length, 0, g_yellowPoint->coordIndex.refValue().length/3 );
#endif
}
void CenterJointClue::renderNode(Matrix matrix, Vec3f parentColScale){
	if(CenterJoint* cj=dynamic_cast<CenterJoint*>(g_sceneTreeList[g_scene->selLine].node.p)){
		if(Art* part=dynamic_cast<Art*>(cj->part.p)){
			if(g_wireFrameMode){
				g_pd3dDevice->SetRenderState( D3DRS_FILLMODE, D3DFILL_SOLID );
			}
			g_light.Diffuse.r  = 1;
			g_light.Diffuse.g  = 1;
			g_light.Diffuse.b  = 1;
			g_pd3dDevice->SetLight( 0, &g_light );

			D3DMATERIAL mtrl;
			ZeroMemory( &mtrl, sizeof(D3DMATERIAL) );
			mtrl.Diffuse.r = mtrl.Ambient.r = 1;
			mtrl.Diffuse.g = mtrl.Ambient.g = 1;
			mtrl.Diffuse.b = mtrl.Ambient.b = 1;
			mtrl.Diffuse.a = mtrl.Ambient.a = 1;
			g_pd3dDevice->SetMaterial( &mtrl );

			Vec3f pivotPointVec;
			D3DXMATRIX tranMatrix;
			renderPivotPoint(part, tranMatrix, pivotPointVec);
			if(g_wireFrameMode){
				g_pd3dDevice->SetRenderState( D3DRS_FILLMODE, D3DFILL_WIREFRAME );
			}
		}
	}
}
void AnimClue::renderNode(Matrix matrix, Vec3f parentColScale){
	if(Art* part=dynamic_cast<Art*>(g_sceneTreeList[g_scene->selLine].node.p)){
		if(g_wireFrameMode){
			g_pd3dDevice->SetRenderState( D3DRS_FILLMODE, D3DFILL_SOLID );
		}
		Vec3f pivotPointVec;
		D3DXMATRIX tranMatrix;

		renderPivotPoint(part, tranMatrix, pivotPointVec);
		if(dragTranPointMode==false){

			Matrix cpMatrix0;
			D3DXMATRIX& cpMatrix=(D3DXMATRIX&)cpMatrix0;
			D3DXMatrixTranslation( &cpMatrix, rotCtrlPnt.x, rotCtrlPnt.y, rotCtrlPnt.z);
			g_pd3dDevice->SetTransform(D3DTS_WORLD, &cpMatrix);

	////////////////

		#ifdef DX8
			g_pd3dDevice->SetStreamSource( 0, g_redRenderPoint->pVB, sizeof(COLOR_VERTEX) );
			g_pd3dDevice->SetVertexShader( D3DFVF_COLOR_VERTEX );
			g_pd3dDevice->SetIndices( g_redRenderPoint->pIB, 0 );
			g_pd3dDevice->DrawIndexedPrimitive( D3DPT_TRIANGLELIST, 0, g_redPoint->coord.length, 0, g_redPoint->coordIndex.refValue().length/3 );
		#else
			g_pd3dDevice->SetStreamSource( 0, g_redRenderPoint->pVB, 0, sizeof(COLOR_VERTEX) );
			g_pd3dDevice->SetFVF( D3DFVF_COLOR_VERTEX );
			g_pd3dDevice->SetIndices( g_redRenderPoint->pIB);
			g_pd3dDevice->DrawIndexedPrimitive( D3DPT_TRIANGLELIST, 0, 0, g_redPoint->coord.length, 0, g_redPoint->coordIndex.refValue().length/3 );
		#endif

			IndexedLineSet* vcEllipticalPolyline2D=dynamic_cast<IndexedLineSet*>(g_vcEllipticalPolyline2D->geometry.p);
			M_Vec3f& coord=dynamic_cast<Coordinate*>(vcEllipticalPolyline2D->coord.p)->point;

			coord[0]=pivotPointVec;
			coord[1]=rotCtrlPnt;

			Matrix lineMatrix0;
			D3DXMATRIX& lineMatrix=(D3DXMATRIX&)lineMatrix0;
			g_pd3dDevice->SetTransform( D3DTS_WORLD, &lineMatrix);
			vcEllipticalPolyline2D->updateAllVertexInBuffers(g_vcEllipticalPolyline2D->render);

			g_vcEllipticalPolyline2D->renderNode(lineMatrix0, parentColScale);

		}


		if(g_wireFrameMode){
			g_pd3dDevice->SetRenderState( D3DRS_FILLMODE, D3DFILL_WIREFRAME );
		}
	}
}
*/

void PartCageClue::renderNode(Matrix parentMatrix, Vec3f parentColScale){
	g_nurbsIlsBase->renderNode(parentMatrix, parentColScale);

	//if(g_wireFrameMode){
	//	g_pd3dDevice->SetRenderState( D3DRS_FILLMODE, D3DFILL_SOLID );
	//}

	g_light.Diffuse.r  = 1;
	g_light.Diffuse.g  = 1;
	g_light.Diffuse.b  = 1;
	g_pd3dDevice->SetLight( 0, &g_light );

	D3DMATERIAL mtrl;
	ZeroMemory( &mtrl, sizeof(D3DMATERIAL) );
	mtrl.Diffuse.r = mtrl.Ambient.r = 1;
	mtrl.Diffuse.g = mtrl.Ambient.g = 1;
	mtrl.Diffuse.b = mtrl.Ambient.b = 1;
	mtrl.Diffuse.a = mtrl.Ambient.a = 1;
	g_pd3dDevice->SetMaterial( &mtrl );

	for(int k = MultiSelect::pointList.length; --k >= 0;){
		NurbsPatch* patch = MultiSelect::pointList[k].p;
		Vec3f point = patch->dpoint.refValue()[MultiSelect::pointList[k].i];
		Vec3f pointTran = parentMatrix * point;

		D3DXMATRIX tranMatrix;
		D3DXMatrixTranslation( &tranMatrix, pointTran.x, pointTran.y, pointTran.z);

		g_pd3dDevice->SetTransform( D3DTS_WORLD, &tranMatrix);

#ifdef DX8
		g_pd3dDevice->SetStreamSource( 0, g_redRenderPoint->pVB, sizeof(COLOR_VERTEX) );
		g_pd3dDevice->SetVertexShader( D3DFVF_COLOR_VERTEX );
		g_pd3dDevice->SetIndices( g_redRenderPoint->pIB, 0 );
		g_pd3dDevice->DrawIndexedPrimitive( D3DPT_TRIANGLELIST, 0, g_redPoint->coord.length, 0, g_redPoint->coordIndex.refValue().length/3 );
#else
		g_pd3dDevice->SetStreamSource( 0, g_redRenderPoint->pVB, 0, sizeof(COLOR_VERTEX) );
		g_pd3dDevice->SetFVF( D3DFVF_COLOR_VERTEX );
		g_pd3dDevice->SetIndices( g_redRenderPoint->pIB);
		g_pd3dDevice->DrawIndexedPrimitive( D3DPT_TRIANGLELIST, 0, 0, g_redPoint->coord.length, 0, g_redPoint->coordIndex.refValue().length/3 );
#endif
	}

	//if(g_wireFrameMode){
	//	g_pd3dDevice->SetRenderState( D3DRS_FILLMODE, D3DFILL_WIREFRAME );
	//}
}


extern Ptr<PartCageClue> g_partCageClue;
extern ShapePtr g_vcSelectRectShape;
extern TransformPtr g_nurbsIlsBase;

void TransformCP::renderNode(Matrix parentMatrix, Vec3f parentColScale){
	g_pd3dDevice->SetRenderState( D3DRS_ZENABLE, FALSE);

	//if(g_wireFrameMode){
	//	g_pd3dDevice->SetRenderState( D3DRS_FILLMODE, D3DFILL_SOLID );
	//}
	g_light.Diffuse.r  = 1;
	g_light.Diffuse.g  = 1;
	g_light.Diffuse.b  = 1;
	g_pd3dDevice->SetLight( 0, &g_light );

	D3DMATERIAL mtrl;
	ZeroMemory( &mtrl, sizeof(D3DMATERIAL) );
	mtrl.Diffuse.r = mtrl.Ambient.r = 1;
	mtrl.Diffuse.g = mtrl.Ambient.g = 1;
	mtrl.Diffuse.b = mtrl.Ambient.b = 1;
	mtrl.Diffuse.a = mtrl.Ambient.a = 1;
	g_pd3dDevice->SetMaterial( &mtrl );

	Vec3f pivotPointVec = parentMatrix * (pivotPoint + translation);
	D3DXMATRIX tranMatrix;
	D3DXMatrixTranslation( &tranMatrix, pivotPointVec.x, pivotPointVec.y, pivotPointVec.z);

	g_pd3dDevice->SetTransform( D3DTS_WORLD, &tranMatrix);

#ifdef DX8
	g_pd3dDevice->SetStreamSource( 0, g_lastRenderPoint->pVB, sizeof(COLOR_VERTEX) );
	g_pd3dDevice->SetVertexShader( D3DFVF_COLOR_VERTEX );
	g_pd3dDevice->SetIndices( g_lastRenderPoint->pIB, 0 );
	g_pd3dDevice->DrawIndexedPrimitive( D3DPT_TRIANGLELIST, 0, g_yellowPoint->coord.length, 0, g_yellowPoint->coordIndex.refValue().length/3 );
#else
	g_pd3dDevice->SetStreamSource( 0, lastRenderPoint->pVB, 0, sizeof(COLOR_VERTEX) );
	g_pd3dDevice->SetFVF( D3DFVF_COLOR_VERTEX );
	g_pd3dDevice->SetIndices( g_lastRenderPoint->pIB);
	g_pd3dDevice->DrawIndexedPrimitive( D3DPT_TRIANGLELIST, 0, 0, g_yellowPoint->coord.length, 0, g_yellowPoint->coordIndex.refValue().length/3 );
#endif


	//if(dragTranPointMode==false){
	{
		Matrix cpMatrix0;
		D3DXMATRIX& cpMatrix=(D3DXMATRIX&)cpMatrix0;
		D3DXMatrixTranslation( &cpMatrix, rotCtrlPnt.x, rotCtrlPnt.y, rotCtrlPnt.z);
		g_pd3dDevice->SetTransform(D3DTS_WORLD, &cpMatrix);

////////////////

	#ifdef DX8
		g_pd3dDevice->SetStreamSource( 0, g_nextRenderPoint->pVB, sizeof(COLOR_VERTEX) );
		g_pd3dDevice->SetVertexShader( D3DFVF_COLOR_VERTEX );
		g_pd3dDevice->SetIndices( g_nextRenderPoint->pIB, 0 );
		g_pd3dDevice->DrawIndexedPrimitive( D3DPT_TRIANGLELIST, 0, g_nextPoint->coord.length, 0, g_nextPoint->coordIndex.refValue().length/3 );
	#else
		g_pd3dDevice->SetStreamSource( 0, g_nextRenderPoint->pVB, 0, sizeof(COLOR_VERTEX) );
		g_pd3dDevice->SetFVF( D3DFVF_COLOR_VERTEX );
		g_pd3dDevice->SetIndices( g_nextRenderPoint->pIB);
		g_pd3dDevice->DrawIndexedPrimitive( D3DPT_TRIANGLELIST, 0, 0, g_nextPoint->coord.length, 0, g_nextPoint->coordIndex.refValue().length/3 );
	#endif

		IndexedLineSet* vcEllipticalPolyline2D = dynamic_cast<IndexedLineSet*>(g_transformLeverClue->geometry.p);
		M_Vec3f& coord=dynamic_cast<Coordinate*>(vcEllipticalPolyline2D->coord.p)->point;

		coord[0] = pivotPointVec;
		coord[1] = rotCtrlPnt;

		Matrix lineMatrix0;
		D3DXMATRIX& lineMatrix = (D3DXMATRIX&)lineMatrix0;
		g_pd3dDevice->SetTransform(D3DTS_WORLD, &lineMatrix);
		vcEllipticalPolyline2D->updateAllVertexInBuffers(g_transformLeverClue->render);

		g_transformLeverClue->renderNode(lineMatrix0, parentColScale);

	}

	//if(g_wireFrameMode){
	//	g_pd3dDevice->SetRenderState( D3DRS_FILLMODE, D3DFILL_WIREFRAME );
	//}
	g_pd3dDevice->SetRenderState( D3DRS_ZENABLE, TRUE);
}

void renderPivotPoint(Seamless* smls, Node* selNode){

	M_Art& partList = smls->partList;

	//if(g_wireFrameMode){
	//	g_pd3dDevice->SetRenderState( D3DRS_FILLMODE, D3DFILL_SOLID );
	//}

	for(int k = 0; k < partList.length; k++){
		Art* part = partList[k];

		if(part->ik){
			Vec3f ikPos  = g_scene->getViewMatrix() * part->ikPos;
			D3DXMATRIX tranMatrix;
			D3DXMatrixTranslation( &tranMatrix, ikPos.x, ikPos.y, ikPos.z);

			g_pd3dDevice->SetTransform( D3DTS_WORLD, &tranMatrix);

#ifdef DX8
			g_pd3dDevice->SetStreamSource( 0, g_nextRenderPoint->pVB, sizeof(COLOR_VERTEX) );
			g_pd3dDevice->SetVertexShader( D3DFVF_COLOR_VERTEX );
			g_pd3dDevice->SetIndices( g_nextRenderPoint->pIB, 0 );
			g_pd3dDevice->DrawIndexedPrimitive( D3DPT_TRIANGLELIST, 0, g_yellowPoint->coord.length, 0, g_yellowPoint->coordIndex.refValue().length/3 );
#else
			g_pd3dDevice->SetStreamSource( 0, g_nextRenderPoint->pVB, 0, sizeof(COLOR_VERTEX) );
			g_pd3dDevice->SetFVF( D3DFVF_COLOR_VERTEX );
			g_pd3dDevice->SetIndices( g_nextRenderPoint->pIB);
			g_pd3dDevice->DrawIndexedPrimitive( D3DPT_TRIANGLELIST, 0, 0, g_yellowPoint->coord.length, 0, g_yellowPoint->coordIndex.refValue().length/3 );
#endif
		}
		else if(part != selNode){
			Vec3f pivotPointVec  = part->pivotPointMatrix * Vec3f(0,0,0);
			D3DXMATRIX tranMatrix;
			D3DXMatrixTranslation( &tranMatrix, pivotPointVec.x, pivotPointVec.y, pivotPointVec.z);

			g_pd3dDevice->SetTransform( D3DTS_WORLD, &tranMatrix);

#ifdef DX8
			g_pd3dDevice->SetStreamSource( 0, g_yellowRenderPoint->pVB, sizeof(COLOR_VERTEX) );
			g_pd3dDevice->SetVertexShader( D3DFVF_COLOR_VERTEX );
			g_pd3dDevice->SetIndices( g_yellowRenderPoint->pIB, 0 );
			g_pd3dDevice->DrawIndexedPrimitive( D3DPT_TRIANGLELIST, 0, g_yellowPoint->coord.length, 0, g_yellowPoint->coordIndex.refValue().length/3 );
#else
			g_pd3dDevice->SetStreamSource( 0, g_yellowRenderPoint->pVB, 0, sizeof(COLOR_VERTEX) );
			g_pd3dDevice->SetFVF( D3DFVF_COLOR_VERTEX );
			g_pd3dDevice->SetIndices( g_yellowRenderPoint->pIB);
			g_pd3dDevice->DrawIndexedPrimitive( D3DPT_TRIANGLELIST, 0, 0, g_yellowPoint->coord.length, 0, g_yellowPoint->coordIndex.refValue().length/3 );
#endif
		}

	}

	if(Art* part = dynamic_cast<Art*>(selNode)){
		if(part->smls = smls){
			Vec3f pivotPointVec  = part->pivotPointMatrix * Vec3f(0,0,0);
			D3DXMATRIX tranMatrix;
			D3DXMatrixTranslation( &tranMatrix, pivotPointVec.x, pivotPointVec.y, pivotPointVec.z);

			g_pd3dDevice->SetTransform( D3DTS_WORLD, &tranMatrix);
#ifdef DX8
			g_pd3dDevice->SetStreamSource( 0, g_redRenderPoint->pVB, sizeof(COLOR_VERTEX) );
			g_pd3dDevice->SetVertexShader( D3DFVF_COLOR_VERTEX );
			g_pd3dDevice->SetIndices( g_redRenderPoint->pIB, 0 );
			g_pd3dDevice->DrawIndexedPrimitive( D3DPT_TRIANGLELIST, 0, g_redPoint->coord.length, 0, g_redPoint->coordIndex.refValue().length/3 );
#else
			g_pd3dDevice->SetStreamSource( 0, g_redRenderPoint->pVB, 0, sizeof(COLOR_VERTEX) );
			g_pd3dDevice->SetFVF( D3DFVF_COLOR_VERTEX );
			g_pd3dDevice->SetIndices( g_redRenderPoint->pIB);
			g_pd3dDevice->DrawIndexedPrimitive( D3DPT_TRIANGLELIST, 0, 0, g_redPoint->coord.length, 0, g_redPoint->coordIndex.refValue().length/3 );
#endif
		}
	}

	//if(g_wireFrameMode){
	//	g_pd3dDevice->SetRenderState( D3DRS_FILLMODE, D3DFILL_WIREFRAME );
	//}
}


void renderPivotPoint(Art* part, D3DXMATRIX& tranMatrix, Vec3f& pivotPointVec){
	pivotPointVec  = part->pivotPointMatrix * Vec3f(0,0,0);
	if(part->smls->xRay) return;
	//g_pd3dDevice->SetRenderState( D3DRS_FILLMODE, D3DFILL_SOLID );
	D3DXMatrixTranslation( &tranMatrix, pivotPointVec.x, pivotPointVec.y, pivotPointVec.z);

	g_pd3dDevice->SetTransform( D3DTS_WORLD, &tranMatrix);

#ifdef DX8
	g_pd3dDevice->SetStreamSource( 0, g_yellowRenderPoint->pVB, sizeof(COLOR_VERTEX) );
	g_pd3dDevice->SetVertexShader( D3DFVF_COLOR_VERTEX );
	g_pd3dDevice->SetIndices( g_yellowRenderPoint->pIB, 0 );
	g_pd3dDevice->DrawIndexedPrimitive( D3DPT_TRIANGLELIST, 0, g_yellowPoint->coord.length, 0, g_yellowPoint->coordIndex.refValue().length/3 );
#else
	g_pd3dDevice->SetStreamSource( 0, g_yellowRenderPoint->pVB, 0, sizeof(COLOR_VERTEX) );
	g_pd3dDevice->SetFVF( D3DFVF_COLOR_VERTEX );
	g_pd3dDevice->SetIndices( g_yellowRenderPoint->pIB);
	g_pd3dDevice->DrawIndexedPrimitive( D3DPT_TRIANGLELIST, 0, 0, g_yellowPoint->coord.length, 0, g_yellowPoint->coordIndex.refValue().length/3 );
#endif
}


extern Coordinate* g_vcSkeletonLinesCoordinate;

void StickMan::renderStickMan(Seamless* smls, Matrix& matrix, Vec3f parentColScale){
	g_light.Diffuse.r  = 1;
	g_light.Diffuse.g  = 1;
	g_light.Diffuse.b  = 1;
	g_pd3dDevice->SetLight( 0, &g_light );

	D3DMATERIAL mtrl;
	ZeroMemory( &mtrl, sizeof(D3DMATERIAL) );
	mtrl.Diffuse.r = mtrl.Ambient.r = 1;
	mtrl.Diffuse.g = mtrl.Ambient.g = 1;
	mtrl.Diffuse.b = mtrl.Ambient.b = 1;
	mtrl.Diffuse.a = mtrl.Ambient.a = 1;
	g_pd3dDevice->SetMaterial( &mtrl );

	if(partList.length > 1){
		if(shape->render == 0){
			ils->convertToRenderFormat(shape->render);
		}

		M_Vec3f& point = coord->point;
		try{
			// Fill the vertex buffer.
			// here, including the normals, which are used for lighting.
			VERTEX* pVertices;
#ifdef DX8
			if( FAILED( shape->render->pVB->Lock( 0, 0, (BYTE**)&pVertices, 0 ) ) ){
#endif
#ifdef DX9
			if( FAILED( shape->render->pVB->Lock( 0, 0, (void**)&pVertices, 0 ) ) ){
#endif
				throw shape;
			}
			for( int i = point.length; --i >= 0; ){
				Vec3f vec = partList[i]->pivotPointMatrix * Vec3f(0,0,0);
				pVertices[i].position = *(D3DXVECTOR3*)&vec;
			}

			shape->render->pVB->Unlock();
		}
		catch(Node*){
			dLog("Unable to render");
			delete shape->render;
			shape->render = 0;
		}
		shape->renderNode(Matrix(), parentColScale);

		g_light.Diffuse.r  = 1;
		g_light.Diffuse.g  = 1;
		g_light.Diffuse.b  = 1;
		g_pd3dDevice->SetLight( 0, &g_light );

		D3DMATERIAL mtrl;
		ZeroMemory( &mtrl, sizeof(D3DMATERIAL) );
		mtrl.Diffuse.r = mtrl.Ambient.r = 1;
		mtrl.Diffuse.g = mtrl.Ambient.g = 1;
		mtrl.Diffuse.b = mtrl.Ambient.b = 1;
		mtrl.Diffuse.a = mtrl.Ambient.a = 1;
		g_pd3dDevice->SetMaterial( &mtrl );
	}
	if(g_scene->xRayEdit){
		renderPivotPoint(smls, g_selNode.getRaw());
	}
	else{
		renderPivotPoint(smls, 0);
	}
}

void CenterJointClue::renderNode(Matrix matrix, Vec3f parentColScale){
	if(CenterJoint* cj=dynamic_cast<CenterJoint*>(g_sceneTreeList[g_scene->selLine].node.p)){
		if(Art* part=dynamic_cast<Art*>(cj->part.p)){
			//if(g_wireFrameMode){
			//	g_pd3dDevice->SetRenderState( D3DRS_FILLMODE, D3DFILL_SOLID );
			//}
			g_light.Diffuse.r  = 1;
			g_light.Diffuse.g  = 1;
			g_light.Diffuse.b  = 1;
			g_pd3dDevice->SetLight( 0, &g_light );

			D3DMATERIAL mtrl;
			ZeroMemory( &mtrl, sizeof(D3DMATERIAL) );
			mtrl.Diffuse.r = mtrl.Ambient.r = 1;
			mtrl.Diffuse.g = mtrl.Ambient.g = 1;
			mtrl.Diffuse.b = mtrl.Ambient.b = 1;
			mtrl.Diffuse.a = mtrl.Ambient.a = 1;
			g_pd3dDevice->SetMaterial( &mtrl );

			Vec3f pivotPointVec;
			D3DXMATRIX tranMatrix;
			renderPivotPoint(part, tranMatrix, pivotPointVec);
			//if(g_wireFrameMode){
			//	g_pd3dDevice->SetRenderState( D3DRS_FILLMODE, D3DFILL_WIREFRAME );
			//}
		}
	}
}
void TransferOwnershipClue::renderNode(Matrix matrix, Vec3f parentColScale){
	if(TransferOwnership* transferOwnership=dynamic_cast<TransferOwnership*>(g_sceneTreeList[g_scene->selLine].node.p)){
		if(Art* part = transferOwnership->getSelectedPart()){
			//if(g_wireFrameMode){
			//	g_pd3dDevice->SetRenderState( D3DRS_FILLMODE, D3DFILL_SOLID );
			//}
			g_light.Diffuse.r  = 1;
			g_light.Diffuse.g  = 1;
			g_light.Diffuse.b  = 1;
			g_pd3dDevice->SetLight( 0, &g_light );

			D3DMATERIAL mtrl;
			ZeroMemory( &mtrl, sizeof(D3DMATERIAL) );
			mtrl.Diffuse.r = mtrl.Ambient.r = 1;
			mtrl.Diffuse.g = mtrl.Ambient.g = 1;
			mtrl.Diffuse.b = mtrl.Ambient.b = 1;
			mtrl.Diffuse.a = mtrl.Ambient.a = 1;
			g_pd3dDevice->SetMaterial( &mtrl );

			Vec3f pivotPointVec;
			D3DXMATRIX tranMatrix;
			renderPivotPoint(part, tranMatrix, pivotPointVec);
			//if(g_wireFrameMode){
			//	g_pd3dDevice->SetRenderState( D3DRS_FILLMODE, D3DFILL_WIREFRAME );
			//}
		}
	}
}
void TransferBCylClue::renderNode(Matrix matrix, Vec3f parentColScale){
	if(TransferBCyl* transferBCyl=dynamic_cast<TransferBCyl*>(g_sceneTreeList[g_scene->selLine].node.p)){
		if(Art* part=dynamic_cast<Art*>(transferBCyl->dest.p)){
			//if(g_wireFrameMode){
			//	g_pd3dDevice->SetRenderState( D3DRS_FILLMODE, D3DFILL_SOLID );
			//}
			g_light.Diffuse.r  = 1;
			g_light.Diffuse.g  = 1;
			g_light.Diffuse.b  = 1;
			g_pd3dDevice->SetLight( 0, &g_light );

			D3DMATERIAL mtrl;
			ZeroMemory( &mtrl, sizeof(D3DMATERIAL) );
			mtrl.Diffuse.r = mtrl.Ambient.r = 1;
			mtrl.Diffuse.g = mtrl.Ambient.g = 1;
			mtrl.Diffuse.b = mtrl.Ambient.b = 1;
			mtrl.Diffuse.a = mtrl.Ambient.a = 1;
			g_pd3dDevice->SetMaterial( &mtrl );

			Vec3f pivotPointVec;
			D3DXMATRIX tranMatrix;
			renderPivotPoint(part, tranMatrix, pivotPointVec);
		}
		g_vcTransferBCylBase->renderNode(matrix, parentColScale);
		//if(g_wireFrameMode){
		//	g_pd3dDevice->SetRenderState( D3DRS_FILLMODE, D3DFILL_WIREFRAME );
		//}
	}
}




void AnimClue::renderNode(Matrix matrix, Vec3f parentColScale){
	if(Art* part=dynamic_cast<Art*>(g_sceneTreeList[g_scene->selLine].node.p)){
		if(part->smls->workAnimCP){
			g_partCageClue->renderNode(matrix, parentColScale);
		}
		//if(g_wireFrameMode){
		//	g_pd3dDevice->SetRenderState( D3DRS_FILLMODE, D3DFILL_SOLID );
		//}
		Vec3f pivotPointVec;
		D3DXMATRIX tranMatrix;

		g_light.Diffuse.r  = 1;
		g_light.Diffuse.g  = 1;
		g_light.Diffuse.b  = 1;
		g_pd3dDevice->SetLight( 0, &g_light );

		D3DMATERIAL mtrl;
		ZeroMemory( &mtrl, sizeof(D3DMATERIAL) );
		mtrl.Diffuse.r = mtrl.Ambient.r = 1;
		mtrl.Diffuse.g = mtrl.Ambient.g = 1;
		mtrl.Diffuse.b = mtrl.Ambient.b = 1;
		mtrl.Diffuse.a = mtrl.Ambient.a = 1;
		g_pd3dDevice->SetMaterial( &mtrl );

#ifdef DX8
		g_pd3dDevice->SetTexture( 0, NULL );
		g_pd3dDevice->SetVertexShader( COLOR_CUSTOMVERTEX_0WEIGHT );
#else
		g_pd3dDevice->SetTexture( 0, NULL );
		g_pd3dDevice->SetFVF( COLOR_CUSTOMVERTEX_0WEIGHT );
#endif

		renderPivotPoint(part, tranMatrix, pivotPointVec);
		//if(dragTranPointMode==false){
		{

			Matrix cpMatrix0;
			D3DXMATRIX& cpMatrix=(D3DXMATRIX&)cpMatrix0;
			D3DXMatrixTranslation( &cpMatrix, rotCtrlPnt.x, rotCtrlPnt.y, rotCtrlPnt.z);
			g_pd3dDevice->SetTransform(D3DTS_WORLD, &cpMatrix);

	////////////////

		#ifdef DX8
			g_pd3dDevice->SetStreamSource( 0, g_redRenderPoint->pVB, sizeof(COLOR_VERTEX) );
			g_pd3dDevice->SetVertexShader( D3DFVF_COLOR_VERTEX );
			g_pd3dDevice->SetIndices( g_redRenderPoint->pIB, 0 );
			g_pd3dDevice->DrawIndexedPrimitive( D3DPT_TRIANGLELIST, 0, g_redPoint->coord.length, 0, g_redPoint->coordIndex.refValue().length/3 );
		#else
			g_pd3dDevice->SetStreamSource( 0, g_redRenderPoint->pVB, 0, sizeof(COLOR_VERTEX) );
			g_pd3dDevice->SetFVF( D3DFVF_COLOR_VERTEX );
			g_pd3dDevice->SetIndices( g_redRenderPoint->pIB);
			g_pd3dDevice->DrawIndexedPrimitive( D3DPT_TRIANGLELIST, 0, 0, g_redPoint->coord.length, 0, g_redPoint->coordIndex.refValue().length/3 );
		#endif


			Seamless* smls = part->smls;
			if(smls->workAnimCP == false && MultiSelect::coordList.length){
				g_viewMatrix = g_scene->getViewMatrix();
				if(g_dragActuallyHappned == false){
					smls->genDynamicMatricesAndPopulate();
				}
				for(int i = MultiSelect::coordList.length; --i>=0;){
					Art* part = MultiSelect::coordList[i].p;
					int cdx = MultiSelect::coordList[i].i;
					Vec3f crd = m_matrix[m_smi[part->fromCoord + cdx]] *  g_viewMatrix * (g_m_coord[part->fromCoord + cdx] - part->absolutePivotPoint);

					//DXRender* render = part->smls->render;
					//if(render==0)return;

					//switch(render->type){
					//	case COLOR_CUSTOMVERTEX_0WEIGHT:{
					//		setCrd<COLORVERTEX_0WEIGHT>(render, part, cdx, crd);
					//		break;
					//	}
					//	case TEXTURE_CUSTOMVERTEX_0WEIGHT:{
					//		setCrd<TEXTUREVERTEX_0WEIGHT>(render, part, cdx, crd);
					//		break;
					//	}
					//}

#ifdef DX8
					D3DXMatrixTranslation( &tranMatrix,crd.x,crd.y,crd.z);
					g_pd3dDevice->SetTransform( D3DTS_WORLD, &tranMatrix );
					g_pd3dDevice->SetStreamSource( 0, g_redRenderPoint->pVB, sizeof(COLOR_VERTEX) );
					g_pd3dDevice->SetVertexShader( D3DFVF_COLOR_VERTEX );
					g_pd3dDevice->SetIndices( g_redRenderPoint->pIB, 0 );
					g_pd3dDevice->DrawIndexedPrimitive( D3DPT_TRIANGLELIST, 0, g_redPoint->coord.length, 0, g_redPoint->coordIndex.refValue().length/3 );

#else
					D3DXMatrixTranslation( &tranMatrix,crd.x,crd.y,crd.z);
					g_pd3dDevice->SetTransform( D3DTS_WORLD, &tranMatrix );
					g_pd3dDevice->SetStreamSource( 0, g_redRenderPoint->pVB, 0, sizeof(COLOR_VERTEX) );
					g_pd3dDevice->SetFVF( D3DFVF_COLOR_VERTEX );
					g_pd3dDevice->SetIndices( g_redRenderPoint->pIB);
					g_pd3dDevice->DrawIndexedPrimitive( D3DPT_TRIANGLELIST, 0, 0, g_redPoint->coord.length, 0, g_redPoint->coordIndex.refValue().length/3 );

#endif
				}
			}
			if(MultiSelect::selectRectMode && smls->animCP == false){
				g_vcSelectRectShape->renderNode(Matrix(), parentColScale);
			}
		}

		IndexedLineSet* vcEllipticalPolyline2D=dynamic_cast<IndexedLineSet*>(g_vcEllipticalPolyline2D->geometry.p);
		M_Vec3f& coord=dynamic_cast<Coordinate*>(vcEllipticalPolyline2D->coord.p)->point;

		coord[0] = pivotPointVec;
		coord[1] = rotCtrlPnt;

		Matrix lineMatrix0;
		D3DXMATRIX& lineMatrix = (D3DXMATRIX&)lineMatrix0;
		g_pd3dDevice->SetTransform(D3DTS_WORLD, &lineMatrix);
		vcEllipticalPolyline2D->updateAllVertexInBuffers(g_vcEllipticalPolyline2D->render);

		g_vcEllipticalPolyline2D->renderNode(lineMatrix0, parentColScale);

		//if(g_wireFrameMode){
		//	g_pd3dDevice->SetRenderState(D3DRS_FILLMODE, D3DFILL_WIREFRAME);
		//}
	}
}

void PartClue::renderNode(Matrix matrix, Vec3f parentColScale){
	D3DXMATRIX tranMatrix;
	Vec3f pivotPointVec;

	if(Art* part=dynamic_cast<Art*>(g_selNode.getRaw())){
		g_light.Diffuse.r  = 1;
		g_light.Diffuse.g  = 1;
		g_light.Diffuse.b  = 1;
		g_pd3dDevice->SetLight( 0, &g_light );

		D3DMATERIAL mtrl;
		ZeroMemory( &mtrl, sizeof(D3DMATERIAL) );
		mtrl.Diffuse.r = mtrl.Ambient.r = 1;
		mtrl.Diffuse.g = mtrl.Ambient.g = 1;
		mtrl.Diffuse.b = mtrl.Ambient.b = 1;
		mtrl.Diffuse.a = mtrl.Ambient.a = 1;
		g_pd3dDevice->SetMaterial( &mtrl );
#ifdef DX8
		g_pd3dDevice->SetTexture( 0, NULL );
		g_pd3dDevice->SetVertexShader( COLOR_CUSTOMVERTEX_0WEIGHT );
#else
		g_pd3dDevice->SetTexture( 0, NULL );
		g_pd3dDevice->SetFVF( COLOR_CUSTOMVERTEX_0WEIGHT );
#endif

		Seamless& smls=*(part->smls);
//		renderPivotPoint(part, tranMatrix, pivotPointVec);
		if(MultiSelect::coordList.length){
			for(int i = MultiSelect::coordList.length; --i>=0;){
				Art* part = MultiSelect::coordList[i].p;
				int cdx = MultiSelect::coordList[i].i;
				Vec3f crd  = part->pivotPointMatrix * (part->coord[cdx] - part->absolutePivotPoint);
#ifdef DX8
				D3DXMatrixTranslation( &tranMatrix,crd.x,crd.y,crd.z);
				g_pd3dDevice->SetTransform( D3DTS_WORLD, &tranMatrix );
				g_pd3dDevice->SetStreamSource( 0, g_redRenderPoint->pVB, sizeof(COLOR_VERTEX) );
				g_pd3dDevice->SetVertexShader( D3DFVF_COLOR_VERTEX );
				g_pd3dDevice->SetIndices( g_redRenderPoint->pIB, 0 );
				g_pd3dDevice->DrawIndexedPrimitive( D3DPT_TRIANGLELIST, 0, g_redPoint->coord.length, 0, g_redPoint->coordIndex.refValue().length/3 );

#else
				D3DXMatrixTranslation( &tranMatrix,crd.x,crd.y,crd.z);
				g_pd3dDevice->SetTransform( D3DTS_WORLD, &tranMatrix );
				g_pd3dDevice->SetStreamSource( 0, g_redRenderPoint->pVB, 0, sizeof(COLOR_VERTEX) );
				g_pd3dDevice->SetFVF( D3DFVF_COLOR_VERTEX );
				g_pd3dDevice->SetIndices( g_redRenderPoint->pIB);
				g_pd3dDevice->DrawIndexedPrimitive( D3DPT_TRIANGLELIST, 0, 0, g_redPoint->coord.length, 0, g_redPoint->coordIndex.refValue().length/3 );

#endif
			}
		}


		if(smls.actTriPart && smls.ActTri < smls.actTriPart->triangle.length){
			const M_Vec3f& coord=part->coord;
			M_Triangle& triangle=smls.actTriPart->triangle;
			S_Triangle& tri=triangle[smls.ActTri];
		
			int a=(smls.ActTriCorner);
			int b=(smls.ActTriCorner+5)%3;
			int c=(smls.ActTriCorner+1)%3;
			Art* partA=tri.refPart(a);
			Art* partB=tri.refPart(b);
			Art* partC=tri.refPart(c);
			int vecIndexA=tri.refIndex(a);
			int vecIndexB=tri.refIndex(b);
			int vecIndexC=tri.refIndex(c);

			Vec3f selCoord  = partA->pivotPointMatrix * (partA->coord.refValue()[vecIndexA] - partA->absolutePivotPoint);
			Vec3f lastCoord = partB->pivotPointMatrix * (partB->coord.refValue()[vecIndexB] - partB->absolutePivotPoint);
			Vec3f nextCoord = partC->pivotPointMatrix * (partC->coord.refValue()[vecIndexC] - partC->absolutePivotPoint);



			{
				M_Vec3f& coord = g_triClueCoord->point;

				coord[0] = selCoord;
				coord[1] = nextCoord;

				coord[2] = nextCoord;
				coord[3] = lastCoord;

				coord[4] = lastCoord;
				coord[5] = selCoord;

				Matrix lineMatrix0;
				D3DXMATRIX& lineMatrix = (D3DXMATRIX&)lineMatrix0;
				g_pd3dDevice->SetTransform(D3DTS_WORLD, &lineMatrix);
				g_triClue->updateAllVertexInBuffers(g_triClueShape->render);

				g_triClueShape->renderNode(lineMatrix0, parentColScale);
			}
/*
#ifdef DX8
			D3DXMatrixTranslation( &tranMatrix,selCoord.x,selCoord.y,selCoord.z);
			g_pd3dDevice->SetTransform( D3DTS_WORLD, &tranMatrix );
			g_pd3dDevice->SetStreamSource( 0, g_redRenderPoint->pVB, sizeof(COLOR_VERTEX) );
			g_pd3dDevice->SetVertexShader( D3DFVF_COLOR_VERTEX );
			g_pd3dDevice->SetIndices( g_redRenderPoint->pIB, 0 );
			g_pd3dDevice->DrawIndexedPrimitive( D3DPT_TRIANGLELIST, 0, g_redPoint->coord.length, 0, g_redPoint->coordIndex.refValue().length/3 );

			D3DXMatrixTranslation( &tranMatrix,lastCoord.x,lastCoord.y,lastCoord.z);
			g_pd3dDevice->SetTransform( D3DTS_WORLD, &tranMatrix );
			g_pd3dDevice->SetStreamSource( 0, g_lastRenderPoint->pVB, sizeof(COLOR_VERTEX) );
			g_pd3dDevice->SetVertexShader( D3DFVF_COLOR_VERTEX );
			g_pd3dDevice->SetIndices( g_lastRenderPoint->pIB, 0 );
			g_pd3dDevice->DrawIndexedPrimitive( D3DPT_TRIANGLELIST, 0, g_lastPoint->coord.length, 0, g_lastPoint->coordIndex.refValue().length/3 );
			
			D3DXMatrixTranslation( &tranMatrix, nextCoord.x, nextCoord.y, nextCoord.z);
			g_pd3dDevice->SetTransform( D3DTS_WORLD, &tranMatrix );
			g_pd3dDevice->SetStreamSource( 0, g_nextRenderPoint->pVB, sizeof(COLOR_VERTEX) );
			g_pd3dDevice->SetVertexShader( D3DFVF_COLOR_VERTEX );
			g_pd3dDevice->SetIndices( g_nextRenderPoint->pIB, 0 );
			g_pd3dDevice->DrawIndexedPrimitive( D3DPT_TRIANGLELIST, 0, g_nextPoint->coord.length, 0, g_nextPoint->coordIndex.length/3 );
#else
			D3DXMatrixTranslation( &tranMatrix,selCoord.x,selCoord.y,selCoord.z);
			g_pd3dDevice->SetTransform( D3DTS_WORLD, &tranMatrix );
			g_pd3dDevice->SetStreamSource( 0, g_redRenderPoint->pVB, 0, sizeof(COLOR_VERTEX) );
			g_pd3dDevice->SetFVF( D3DFVF_COLOR_VERTEX );
			g_pd3dDevice->SetIndices( g_redRenderPoint->pIB);
			g_pd3dDevice->DrawIndexedPrimitive( D3DPT_TRIANGLELIST, 0, 0, g_redPoint->coord.length, 0, g_redPoint->coordIndex.refValue().length/3 );

			D3DXMatrixTranslation( &tranMatrix,lastCoord.x,lastCoord.y,lastCoord.z);
			g_pd3dDevice->SetTransform( D3DTS_WORLD, &tranMatrix );
			g_pd3dDevice->SetStreamSource( 0, g_lastRenderPoint->pVB, 0, sizeof(COLOR_VERTEX) );
			g_pd3dDevice->SetFVF( D3DFVF_COLOR_VERTEX );
			g_pd3dDevice->SetIndices( g_lastRenderPoint->pIB);
			g_pd3dDevice->DrawIndexedPrimitive( D3DPT_TRIANGLELIST, 0, 0, g_lastPoint->coord.length, 0, g_lastPoint->coordIndex.refValue().length/3 );		
		
			D3DXMatrixTranslation( &tranMatrix, nextCoord.x, nextCoord.y, nextCoord.z);
			g_pd3dDevice->SetTransform( D3DTS_WORLD, &tranMatrix );
			g_pd3dDevice->SetStreamSource( 0, g_nextRenderPoint->pVB, 0, sizeof(COLOR_VERTEX) );
			g_pd3dDevice->SetFVF( D3DFVF_COLOR_VERTEX );
			g_pd3dDevice->SetIndices( g_nextRenderPoint->pIB);
			g_pd3dDevice->DrawIndexedPrimitive( D3DPT_TRIANGLELIST, 0, 0, g_nextPoint->coord.length, 0, g_nextPoint->coordIndex.length/3 );
#endif
*/

			//if(MultiSelect::coordList.length == 0)
			{
/*
				M_Triangle& triangle=smls.actTriPart->triangle;
				S_Triangle& tri=triangle[smls.ActTri];
			
				int a=(smls.ActTriCorner);
				int b=(smls.ActTriCorner+5)%3;
				int c=(smls.ActTriCorner+1)%3;
				Art* partA=tri.refPart(a);
				Art* partB=tri.refPart(b);
				Art* partC=tri.refPart(c);
				int vecIndexA=tri.refIndex(a);
				int vecIndexB=tri.refIndex(b);
				int vecIndexC=tri.refIndex(c);

				Vec3f selCoord  = partA->pivotPointMatrix * (partA->coord.refValue()[vecIndexA] - partA->absolutePivotPoint);
				Vec3f lastCoord = partB->pivotPointMatrix * (partB->coord.refValue()[vecIndexB] - partB->absolutePivotPoint);
				Vec3f nextCoord = partC->pivotPointMatrix * (partC->coord.refValue()[vecIndexC] - partC->absolutePivotPoint);
#ifdef DX8
				D3DXMatrixTranslation( &tranMatrix,selCoord.x,selCoord.y,selCoord.z);
				g_pd3dDevice->SetTransform( D3DTS_WORLD, &tranMatrix );
				g_pd3dDevice->SetStreamSource( 0, g_redRenderPoint->pVB, sizeof(COLOR_VERTEX) );
				g_pd3dDevice->SetVertexShader( D3DFVF_COLOR_VERTEX );
				g_pd3dDevice->SetIndices( g_redRenderPoint->pIB, 0 );
				g_pd3dDevice->DrawIndexedPrimitive( D3DPT_TRIANGLELIST, 0, g_redPoint->coord.length, 0, g_redPoint->coordIndex.refValue().length/3 );

				D3DXMatrixTranslation( &tranMatrix,lastCoord.x,lastCoord.y,lastCoord.z);
				g_pd3dDevice->SetTransform( D3DTS_WORLD, &tranMatrix );
				g_pd3dDevice->SetStreamSource( 0, g_lastRenderPoint->pVB, sizeof(COLOR_VERTEX) );
				g_pd3dDevice->SetVertexShader( D3DFVF_COLOR_VERTEX );
				g_pd3dDevice->SetIndices( g_lastRenderPoint->pIB, 0 );
				g_pd3dDevice->DrawIndexedPrimitive( D3DPT_TRIANGLELIST, 0, g_lastPoint->coord.length, 0, g_lastPoint->coordIndex.refValue().length/3 );
				
				D3DXMatrixTranslation( &tranMatrix, nextCoord.x, nextCoord.y, nextCoord.z);
				g_pd3dDevice->SetTransform( D3DTS_WORLD, &tranMatrix );
				g_pd3dDevice->SetStreamSource( 0, g_nextRenderPoint->pVB, sizeof(COLOR_VERTEX) );
				g_pd3dDevice->SetVertexShader( D3DFVF_COLOR_VERTEX );
				g_pd3dDevice->SetIndices( g_nextRenderPoint->pIB, 0 );
				g_pd3dDevice->DrawIndexedPrimitive( D3DPT_TRIANGLELIST, 0, g_nextPoint->coord.length, 0, g_nextPoint->coordIndex.length/3 );
#else
				D3DXMatrixTranslation( &tranMatrix,selCoord.x,selCoord.y,selCoord.z);
				g_pd3dDevice->SetTransform( D3DTS_WORLD, &tranMatrix );
				g_pd3dDevice->SetStreamSource( 0, g_redRenderPoint->pVB, 0, sizeof(COLOR_VERTEX) );
				g_pd3dDevice->SetFVF( D3DFVF_COLOR_VERTEX );
				g_pd3dDevice->SetIndices( g_redRenderPoint->pIB);
				g_pd3dDevice->DrawIndexedPrimitive( D3DPT_TRIANGLELIST, 0, 0, g_redPoint->coord.length, 0, g_redPoint->coordIndex.refValue().length/3 );

				D3DXMatrixTranslation( &tranMatrix,lastCoord.x,lastCoord.y,lastCoord.z);
				g_pd3dDevice->SetTransform( D3DTS_WORLD, &tranMatrix );
				g_pd3dDevice->SetStreamSource( 0, g_lastRenderPoint->pVB, 0, sizeof(COLOR_VERTEX) );
				g_pd3dDevice->SetFVF( D3DFVF_COLOR_VERTEX );
				g_pd3dDevice->SetIndices( g_lastRenderPoint->pIB);
				g_pd3dDevice->DrawIndexedPrimitive( D3DPT_TRIANGLELIST, 0, 0, g_lastPoint->coord.length, 0, g_lastPoint->coordIndex.refValue().length/3 );		
			
				D3DXMatrixTranslation( &tranMatrix, nextCoord.x, nextCoord.y, nextCoord.z);
				g_pd3dDevice->SetTransform( D3DTS_WORLD, &tranMatrix );
				g_pd3dDevice->SetStreamSource( 0, g_nextRenderPoint->pVB, 0, sizeof(COLOR_VERTEX) );
				g_pd3dDevice->SetFVF( D3DFVF_COLOR_VERTEX );
				g_pd3dDevice->SetIndices( g_nextRenderPoint->pIB);
				g_pd3dDevice->DrawIndexedPrimitive( D3DPT_TRIANGLELIST, 0, 0, g_nextPoint->coord.length, 0, g_nextPoint->coordIndex.length/3 );
#endif
*/
				if(g_vcArtTriBase->circle){
					Vec3f radSizeCrd  = selCoord;
					radSizeCrd.x += g_scene->getZoom() * g_settings->tugRadius;
	#ifdef DX8
					D3DXMatrixTranslation(&tranMatrix, radSizeCrd.x, radSizeCrd.y, radSizeCrd.z);
					g_pd3dDevice->SetTransform(D3DTS_WORLD, &tranMatrix);
					g_pd3dDevice->SetStreamSource(0, g_redRenderPoint->pVB, sizeof(COLOR_VERTEX));
					g_pd3dDevice->SetVertexShader(D3DFVF_COLOR_VERTEX);
					g_pd3dDevice->SetIndices(g_lastRenderPoint->pIB, 0);
					g_pd3dDevice->DrawIndexedPrimitive(D3DPT_TRIANGLELIST, 0, g_redPoint->coord.length, 0, g_redPoint->coordIndex.refValue().length/3 );

	#else
					D3DXMatrixTranslation(&tranMatrix,radSizeCrd.x,radSizeCrd.y,radSizeCrd.z);
					g_pd3dDevice->SetTransform(D3DTS_WORLD, &tranMatrix );
					g_pd3dDevice->SetStreamSource(0, g_redRenderPoint->pVB, 0, sizeof(COLOR_VERTEX) );
					g_pd3dDevice->SetFVF(D3DFVF_COLOR_VERTEX );
					g_pd3dDevice->SetIndices(g_redRenderPoint->pIB);
					g_pd3dDevice->DrawIndexedPrimitive(D3DPT_TRIANGLELIST, 0, 0, g_redPoint->coord.length, 0, g_redPoint->coordIndex.refValue().length/3 );
	#endif
					//////

					IndexedLineSet* vcEllipticalPolyline2D=dynamic_cast<IndexedLineSet*>(g_vcEllipticalPolyline2D->geometry.p);
					M_Vec3f& coord=dynamic_cast<Coordinate*>(vcEllipticalPolyline2D->coord.p)->point;

					Matrix lineMatrix0;
					D3DXMATRIX& lineMatrix=(D3DXMATRIX&)lineMatrix0;

					D3DXMatrixTranslation( &lineMatrix,selCoord.x,selCoord.y,selCoord.z);

					g_pd3dDevice->SetTransform( D3DTS_WORLD, &lineMatrix);
					vcEllipticalPolyline2D->updateAllVertexInBuffers(g_vcEllipticalPolyline2D->render);

					g_vcEllipticalPolyline2D->renderNode(lineMatrix0, parentColScale);

				}
				if(g_vcArtTriBase->fuseSurface){
					g_vcFuseSurfaceFlatStageAidShape->renderNode(matrix, parentColScale);
				}
			}
		}
		if(MultiSelect::selectRectMode){
			g_vcSelectRectGroup->renderNode(matrix, parentColScale);
		}
		if(MultiSelect::triList.length){
			transpSmls->renderNode(matrix, parentColScale);
		}
	}
	//if(g_wireFrameMode==true){
	//	g_pd3dDevice->SetRenderState( D3DRS_FILLMODE, D3DFILL_WIREFRAME );
	//}
}
void setBackface(bool backfaceRender){
	if(backfaceRender != g_backfaceRender){
		g_backfaceRender = backfaceRender;
		if(backfaceRender){
			g_pd3dDevice->SetRenderState( D3DRS_CULLMODE, D3DCULL_NONE );
		}
		else{
			g_pd3dDevice->SetRenderState( D3DRS_CULLMODE, D3DCULL_CCW );
		}
	}
}
void ColorEffect::setUpVertexShaders(Matrix& parentMatrix, Seamless& smls){
	DXRender* render=smls.render;
	M_SArt& skeleton=smls.skeleton.refValue();

	if(smls.workAnimCP){
		for(int i = 0, smi = 0; i < skeleton.length; i++){
			skeleton[i]->genDynamicTransformsCP(parentMatrix, parentMatrix, smi);
		}
	}
	else{
		for(int i = 0, smi = 0; i < skeleton.length; i++){
			skeleton[i]->genDynamicTransforms(parentMatrix, smi);
		}
	}
#ifdef DX8
		g_pd3dDevice->SetTexture( 0, NULL );
		g_pd3dDevice->SetStreamSource( 0, render->pVB, sizeof(COLORVERTEX_0WEIGHT) );
		g_pd3dDevice->SetVertexShader( COLOR_CUSTOMVERTEX_0WEIGHT );
#else
		g_pd3dDevice->SetTexture( 0, NULL );
		g_pd3dDevice->SetStreamSource( 0, render->pVB, 0, sizeof(COLORVERTEX_0WEIGHT) );
		g_pd3dDevice->SetFVF( COLOR_CUSTOMVERTEX_0WEIGHT );
#endif

}

void ColorEffect::renderMaterial(int triFrom, Matrix& parentMatrix, Seamless& smls, Matrix& originalMatrix, Vec3f& originalColScale){
	if(g_renderLater == false){
		if(transparency > 0){
			g_transparentList += TransparencyList(&smls, originalMatrix, originalColScale);
			return;
		}
	}
	else{
		if(false == (transparency > 0)){
			return;
		}
	}



	setBackface(backfaceRender);

	DXRender* render=smls.render;
	M_SArt& skeleton=smls.skeleton.refValue();


	
//	DXRender* render=smls.render;


    D3DMATERIAL mtrl;
    ZeroMemory( &mtrl, sizeof(D3DMATERIAL) );
    mtrl.Diffuse.r = 1;
    mtrl.Diffuse.g = 1;
    mtrl.Diffuse.b = 1;
    mtrl.Diffuse.a = transparency;

    mtrl.Ambient.r = ambient.r;
    mtrl.Ambient.g = ambient.g;
    mtrl.Ambient.b = ambient.b;
    mtrl.Ambient.a = transparency;

    mtrl.Emissive.r = emissive.r;
    mtrl.Emissive.g = emissive.g;
    mtrl.Emissive.b = emissive.b;
    mtrl.Emissive.a = transparency;

	mtrl.Specular.r = specular.r;
	mtrl.Specular.g = specular.g;
	mtrl.Specular.b = specular.b;
	mtrl.Specular.a = transparency;

	mtrl.Power=.5;
	g_pd3dDevice->SetMaterial( &mtrl );
#ifdef DX8
	g_pd3dDevice->SetIndices( render->pIB, 0 );
	g_pd3dDevice->DrawIndexedPrimitive( D3DPT_TRIANGLELIST, 0, render->vertexCount, triFrom*3, triLen );
#else
	g_pd3dDevice->SetIndices( render->pIB );
	g_pd3dDevice->DrawIndexedPrimitive( D3DPT_TRIANGLELIST, 0, 0, render->vertexCount, triFrom*3, triLen );
#endif

}

void TextureEffect::setUpVertexShaders(Matrix& parentMatrix, Seamless& smls){
	DXRender* render=smls.render;
	M_SArt& skeleton=smls.skeleton.refValue();

	if(smls.workAnimCP){
		for(int i = 0, smi = 0; i < skeleton.length; i++){
			skeleton[i]->genDynamicTransformsCP(parentMatrix, parentMatrix, smi);
		}
	}
	else{
		for(int i = 0, smi = 0; i < skeleton.length; i++){
			skeleton[i]->genDynamicTransforms(parentMatrix, smi);
		}
	}
	if(smls.workAnimCP == false && g_colorOwnership){ //identical code to colourEffect
#ifdef DX8
		g_pd3dDevice->SetTexture( 0, NULL );
		g_pd3dDevice->SetStreamSource( 0, render->pVB, sizeof(COLORVERTEX_0WEIGHT) );
		g_pd3dDevice->SetVertexShader( COLOR_CUSTOMVERTEX_0WEIGHT );
#else
		g_pd3dDevice->SetTexture( 0, NULL );
		g_pd3dDevice->SetStreamSource( 0, render->pVB, 0, sizeof(COLORVERTEX_0WEIGHT) );
		g_pd3dDevice->SetFVF( COLOR_CUSTOMVERTEX_0WEIGHT );
#endif
	}
	else{
#ifdef DX8
		g_pd3dDevice->SetTextureStageState( 0, D3DTSS_COLORARG1, D3DTA_TEXTURE );
		g_pd3dDevice->SetTextureStageState( 0, D3DTSS_MAGFILTER , D3DTEXF_LINEAR  );

		g_pd3dDevice->SetStreamSource( 0, render->pVB, sizeof(TEXTUREVERTEX_0WEIGHT) );
		g_pd3dDevice->SetVertexShader( TEXTURE_CUSTOMVERTEX_0WEIGHT );



#else
		g_pd3dDevice->SetTexture( 0, dxTexture );
		g_pd3dDevice->SetTexture( 0, dxTexture );
		g_pd3dDevice->SetTextureStageState( 0, D3DTSS_COLORARG1, D3DTA_TEXTURE );
		g_pd3dDevice->SetTextureStageState( 0, D3DTSS_ALPHAOP,   D3DTOP_DISABLE );
		g_pd3dDevice->SetSamplerState( 0, D3DSAMP_MAGFILTER, D3DTEXF_LINEAR );

		g_pd3dDevice->SetStreamSource( 0, render->pVB, 0, sizeof(TEXTUREVERTEX_0WEIGHT) );
		g_pd3dDevice->SetFVF( TEXTURE_CUSTOMVERTEX_0WEIGHT );


#endif
	}
}
void MonoColorEffect::setUpVertexShaders(Matrix& parentMatrix, Seamless& smls){
}
void TextureEffect::renderMaterial(int triFrom, Matrix& parentMatrix, Seamless& smls, Matrix& originalMatrix, Vec3f& originalColScale){
	if(g_renderLater == false){
		if(alphaTexture){
			g_alphaChanList += AlphaChanList(&smls, originalMatrix, originalColScale);
			return;
		}
		if(transparency > 0){
			g_transparentList += TransparencyList(&smls, originalMatrix, originalColScale);
			return;
		}
	}
	else{
		if(false == (alphaTexture || transparency > 0)){
			return;
		}
	}
	setBackface(backfaceRender);

	DXRender* render=smls.render;
	M_SArt& skeleton=smls.skeleton.refValue();

	if(g_browseMode == false && url.update && url.refValue().length){
		if( FAILED( D3DXCreateTextureFromFile(  g_pd3dDevice, url.refValue().c_str(), &dxTexture ) ) ){
			//dLog ("failed to set Texture from local file");
		}
		else{
			dLog ("set Texture from local file");
			url.update = false;
			dxTexture->GetSurfaceLevel(0, &g_textureSurface);
			g_textureSurface->GetDesc(&g_surfaceBuffer);
			alphaTexture = false;
			if(g_surfaceBuffer.Format == D3DFMT_A8R8G8B8){
				alphaTexture = true;
			}
		}
	}
    D3DMATERIAL mtrl;
    ZeroMemory( &mtrl, sizeof(D3DMATERIAL) );
    mtrl.Diffuse.r = 1;
    mtrl.Diffuse.g = 1;
    mtrl.Diffuse.b = 1;
    mtrl.Diffuse.a = transparency;

    mtrl.Ambient.r = ambient.r;
    mtrl.Ambient.g = ambient.g;
    mtrl.Ambient.b = ambient.b;
    mtrl.Ambient.a = transparency;

    mtrl.Emissive.r = emissive.r;
    mtrl.Emissive.g = emissive.g;
    mtrl.Emissive.b = emissive.b;
    mtrl.Emissive.a = transparency;

	mtrl.Specular.r=specular.r;
	mtrl.Specular.g=specular.g;
	mtrl.Specular.b=specular.b;
	mtrl.Specular.a=transparency;

	mtrl.Power=.5;
	g_pd3dDevice->SetMaterial( &mtrl );


//	if(dxTexture==NULL&&!downloading && url.length){
//		//addFileToDownLoadList(url.refValue()[0], this, true);
//	}
 	if(dxTexture==NULL && !hInternetOpenUrl && g_online){
		openUrl();
	}
	if(smls.workAnimCP == false && g_colorOwnership){ //identical code to colourEffect
	#ifdef DX8
		g_pd3dDevice->SetIndices( render->pIB, 0 );
		g_pd3dDevice->DrawIndexedPrimitive( D3DPT_TRIANGLELIST, 0, render->vertexCount, triFrom*3, triLen );


	#else
		g_pd3dDevice->SetIndices( render->pIB );
		g_pd3dDevice->DrawIndexedPrimitive( D3DPT_TRIANGLELIST, 0, 0, render->vertexCount, triFrom*3, triLen );
	#endif
	}
	else{

	#ifdef DX8
		g_pd3dDevice->SetTexture( 0, (IDirect3DBaseTexture8 *)dxTexture );

		g_pd3dDevice->SetIndices( render->pIB, 0 );
		g_pd3dDevice->DrawIndexedPrimitive( D3DPT_TRIANGLELIST, 0, render->vertexCount, triFrom*3, triLen );


	#else

		g_pd3dDevice->SetIndices( render->pIB);
		g_pd3dDevice->DrawIndexedPrimitive( D3DPT_TRIANGLELIST, 0, 0, render->vertexCount, triFrom*3, triLen );

	#endif
	}
}
void MonoColorEffect::renderMaterial(int triFrom, Matrix& parentMatrix, Seamless& smls, Matrix& originalMatrix, Vec3f& originalColScale){
	setBackface(backfaceRender);

	DXRender* render=smls.render;
	M_SArt& skeleton=smls.skeleton.refValue();

	for(int i=0, smi=0; i<skeleton.length; i++){
		skeleton[i]->genDynamicTransforms(parentMatrix, smi);
	}

    D3DMATERIAL mtrl;
    ZeroMemory( &mtrl, sizeof(D3DMATERIAL) );
    mtrl.Diffuse.r = 1;
    mtrl.Diffuse.g = 1;
    mtrl.Diffuse.b = 1;
    mtrl.Diffuse.a = transparency;

    mtrl.Ambient.r = ambient.r;
    mtrl.Ambient.g = ambient.g;
    mtrl.Ambient.b = ambient.b;
    mtrl.Ambient.a = transparency;

    mtrl.Emissive.r = emissive.r;
    mtrl.Emissive.g = emissive.g;
    mtrl.Emissive.b = emissive.b;
    mtrl.Emissive.a = transparency;

	mtrl.Specular.r=specular.r;
	mtrl.Specular.g=specular.g;
	mtrl.Specular.b=specular.b;
	mtrl.Specular.a=transparency;

	mtrl.Power=.5;
	g_pd3dDevice->SetMaterial( &mtrl );
#ifdef DX8
	g_pd3dDevice->SetTexture( 0, NULL );
	g_pd3dDevice->SetStreamSource( 0, render->pVB, sizeof(COLORVERTEX_0WEIGHT) );
	g_pd3dDevice->SetVertexShader( COLOR_CUSTOMVERTEX_0WEIGHT );

	g_pd3dDevice->SetIndices( render->pIB, 0 );
	g_pd3dDevice->DrawIndexedPrimitive( D3DPT_TRIANGLELIST, 0, render->vertexCount, 0, render->indexCount/3 );


#else
	g_pd3dDevice->SetTexture( 0, NULL );
	g_pd3dDevice->SetStreamSource( 0, render->pVB, 0, sizeof(COLORVERTEX_0WEIGHT) );
	g_pd3dDevice->SetFVF( COLOR_CUSTOMVERTEX_0WEIGHT );

	g_pd3dDevice->SetIndices( render->pIB );
	g_pd3dDevice->DrawIndexedPrimitive( D3DPT_TRIANGLELIST, 0, 0, render->vertexCount, 0, render->indexCount/3 );

#endif
}


template<class T> inline void  do_polyFrameCommon(T pVertices, const int& renderFromCoord, const M_Vec3f& coord, const M_Int& rc, const Vec3f& absolutePivotPoint, const int& i, const int& portionLen, const M_Vec3f& keyValue, const float& interp, const M_Int& cpIndex){
	int toCoord = renderFromCoord + coord.length;
	int coordLen = coord.length;

	for(int j = 0; j < portionLen; j++){
		int h = i * portionLen + j;
		Vec3f result = keyValue[h] + (keyValue[(i + 1) * portionLen + j] - keyValue[h]) * interp;
		int idx = cpIndex[j];
		if(idx < coordLen){
			pVertices[renderFromCoord + idx].position = (D3DXVECTOR3&) (coord[idx] - absolutePivotPoint + result);
		}
	}

	int mark = renderFromCoord + coord.length;
	for(int i = rc.length; --i>=0; ){
		pVertices[mark + i].position = pVertices[renderFromCoord + rc[i]].position;
	}
}

void CPMorpher::do_polyFrame(const Art* part){
	const M_Float& key = this->key.refValue();
	const M_Vec3f& keyValue = this->keyValue.refValue();
	if(!key.length)return;

//	for(int k = controlPoint.length; --k >= 0;){
//	}
	int portionLen = cpIndex.length;

	float frac = frame;
	int i = 0;
	if(key.length < 2)goto here;
	for(; i<key.length-1; i++){
		if(key[i] > frac){
			i--;
			break;
		}
	}
	if(i < 0){
here:
		if(key.length > 0){
			//value_changed.sendOutRoute(keyValue[0]);
		}
		return;
	}
	if(i == key.length - 1){
		i = key.length - 2;
	}
	float diff = key[i + 1] - key[i];

	float interp = 1.0 / diff * (frac - key[i]);
	if(interp < 0){
		interp =- interp;
	}



	const Seamless* smls = part->smls;
	const M_Vec3f& coord = part->coord;
	const M_Int& rc = part->rc;
	DXRender* render = smls->render;
	if(render==0)return;
	switch(render->type){
	case COLOR_CUSTOMVERTEX_0WEIGHT:{
		COLORVERTEX_0WEIGHT* pVertices;
#ifdef DX8
		if( FAILED( render->pVB->Lock( 0, 0, (BYTE**)&pVertices, 0 ) ) ){
			throw this;
		}
#else
		if( FAILED( render->pVB->Lock( 0, 0, (void**)&pVertices, 0 ) ) ){
			throw this;
		}
#endif
		do_polyFrameCommon<COLORVERTEX_0WEIGHT*>(pVertices, part->renderFromCoord, coord, rc, part->absolutePivotPoint, i, portionLen, keyValue, interp, cpIndex);
		break;
	}
	case TEXTURE_CUSTOMVERTEX_0WEIGHT:{
		TEXTUREVERTEX_0WEIGHT* pVertices;
#ifdef DX8
		if( FAILED( render->pVB->Lock( 0, 0, (BYTE**)&pVertices, 0 ) ) ){
			throw this;
		}
#else
		if( FAILED( render->pVB->Lock( 0, 0, (void**)&pVertices, 0 ) ) ){
			throw this;
		}
#endif
		do_polyFrameCommon<TEXTUREVERTEX_0WEIGHT*>(pVertices, part->renderFromCoord, coord, rc, part->absolutePivotPoint, i, portionLen, keyValue, interp, cpIndex);
		break;
	}
	default:
		throw this;
	}

	render->pVB->Unlock();
}


void Seamless::renderNode(Matrix originalMatrix, Vec3f originalColScale){
	if(ignoreWireFrame==true && g_wireFrameMode){
		g_pd3dDevice->SetRenderState( D3DRS_FILLMODE, D3DFILL_SOLID );
	}
	if(ikProcessFlag){
		if((g_visualClue == g_partCageClue && g_partCageClue->smls == this) || (g_visualClue == g_vcArtTriBase && g_vcArtTriBase->smls == this)){
		}
		else{
			calcIK();
		}
		ikProcessFlag = false;
	}

	if(animCP){
		if(animCPBuildFlag){
			doBuild();
		}
	}
	else if(morphFlag){
		int j = 0;
		if(cpMorpherPartLastLen != cpMorpherPart.length){
			for(int i = cpMorpherPartLastLen; i < cpMorpherPart.length;){
				cpMorpherPart[j++] = cpMorpherPart[i++];
			}

			cpMorpherPart.length = cpMorpherPartLastLen = j;
		}

		for(int k = cpMorpherPartLastLen; --k >= 0;){
			CPMorpherPart& cpmp = cpMorpherPart[k];
			cpmp.morpher->do_polyFrame(cpmp.part);
		}
	}

	if(partList.length==0||this->effect==0)return;
	Vec3f scale=this->scale.refValue();
	Vec3f parentColScale;
	parentColScale.x = originalColScale.x * scale.x;
	parentColScale.y = originalColScale.y * scale.y;
	parentColScale.z = originalColScale.z * scale.z;

	Matrix parentMatrix = dxWindMatrix(originalMatrix, scale, rotation.refValue(), translation.refValue());

	float a=(parentColScale.x+parentColScale.y+parentColScale.z)*(1./3.);
	g_light.Diffuse.r  = a;
	g_light.Diffuse.g  = a;
	g_light.Diffuse.b  = a;
	g_pd3dDevice->SetLight( 0, &g_light );

/*
	if(this->effect!=NULL){
		int fred=this->effect.getType();
		if(this->effect.getType()==TEXTUREEFFECT){
			throw this;

		}
	}
	*/
	//TextureEffect* effect=dynamic_cast<TextureEffect*>(this->effect.getRawPtr());
	if(render==0||ctrf){
		convertToRenderFormat();
		ctrf = false;
		if(render==0){
			return;
		}
	}
	if(render->pIB){
		if(effect!=0){
			dynamic_cast<Effect*>(effect.p)->setUpVertexShaders(parentMatrix, *this);
		}
		g_pd3dDevice->SetRenderState( D3DRS_VERTEXBLEND, D3DVBF_0WEIGHTS );
		g_pd3dDevice->SetRenderState( D3DRS_INDEXEDVERTEXBLENDENABLE, TRUE);

		for(int k = 0, triFrom = 0; k < effectList.length; k++){
			Effect* effect = effectList[k];
			effect->renderMaterial(triFrom, parentMatrix, *this, originalMatrix, originalColScale);
			triFrom += effect->triLen;
		}
		if(ignoreWireFrame==true && g_wireFrameMode){
			g_pd3dDevice->SetRenderState( D3DRS_FILLMODE, D3DFILL_WIREFRAME );
		}

		g_pd3dDevice->SetRenderState( D3DRS_VERTEXBLEND, D3DVBF_DISABLE );
		g_pd3dDevice->SetRenderState( D3DRS_INDEXEDVERTEXBLENDENABLE, FALSE);
	}
	else if(stickMan){
		M_SArt& skeleton=this->skeleton.refValue();
		for(int i=0, smi=0; i<skeleton.length; i++){
			skeleton[i]->genDynamicTransforms(parentMatrix, smi);
		}
	}

	if(stickMan){
		g_skelList+= RenderLast(this, matrix, parentColScale);
	}

	morphFlag = false;
}
Seamless* getSeamless(int& selLine){
	selLine = g_stw->selLine;
	while(selLine != 0){
		if(Seamless* smls = dynamic_cast<Seamless*>(g_sceneTreeList[selLine].node.p)){
			return smls;
		}
		selLine=getParent(selLine);
	}
	return 0;
}

Seamless* getSeamless2(){
	int selLine = g_stw->selLine;
	while(selLine != 0){
		if(Seamless* smls = dynamic_cast<Seamless*>(g_sceneTreeList[selLine].node.p)){
			return smls;
		}
		selLine=getParent(selLine);
	}
	return 0;
}

Seamless* getModelMorpher(ModelMorpher* modelMorpher){
	int selLine = g_stw->selLine;
	if(Seamless* selSmls = getSeamless(selLine)){
		selLine = getParent(selLine);
		if(dynamic_cast<ModelSequence*>(g_sceneTreeList[selLine].node.p)){
			if(modelMorpher == g_sceneTreeList[getParent(selLine)].node){
				return selSmls;
			}
		}
	}
	return 0;
}
void ModelMorpher::renderNode(Matrix matrix, Vec3f parentColScale){
	
	if(Seamless* selSmls = getModelMorpher(this)){
		selSmls->renderNode(matrix, parentColScale);
	}
	else{
		if(smls){
			smls->renderNode(matrix, parentColScale);
		}
	}
}
void HAnimHumanoid::renderNode(Matrix matrix, Vec3f parentColScale){
	if(skeleton.length==0)return;
	Vec3f scale=this->scale.refValue();
	parentColScale.x*=scale.x;
	parentColScale.y*=scale.y;
	parentColScale.z*=scale.z;


	float a=(parentColScale.x+parentColScale.y+parentColScale.z)*(1./3.);
	g_light.Diffuse.r  = a;
	g_light.Diffuse.g  = a;
	g_light.Diffuse.b  = a;
	g_pd3dDevice->SetLight( 0, &g_light );

	if(skin.length==0)return;
	Shape* shape = dynamic_cast<Shape*>(skin[0].p);
	Appearance* appearance=dynamic_cast<Appearance*>(shape->appearance.p);
	if(render==0){
		convertToRenderFormat(render, appearance);
		if(render==0){
			return;
		}
	}
	if(render->pIB==0){
		return;
	}
	g_pd3dDevice->SetRenderState( D3DRS_VERTEXBLEND, D3DVBF_1WEIGHTS );
	g_pd3dDevice->SetRenderState( D3DRS_INDEXEDVERTEXBLENDENABLE, TRUE);

	for(int smi=0,k=0;k<skeleton.length;k++){
		dynamic_cast<HAnimJoint*>(skeleton[k].getRawPtr())->genDynamicTransforms(matrix, smi);
	}
	if(render->type==COLOR_CUSTOMVERTEX_1WEIGHT){
		g_pd3dDevice->SetTexture( 0, NULL );
#ifdef DX8
		g_pd3dDevice->SetStreamSource( 0, render->pVB, sizeof(COLORVERTEX_1WEIGHT));
		g_pd3dDevice->SetVertexShader( COLOR_CUSTOMVERTEX_1WEIGHT );
#else
		g_pd3dDevice->SetStreamSource( 0, render->pVB, 0, sizeof(COLORVERTEX_1WEIGHT));
		g_pd3dDevice->SetFVF( COLOR_CUSTOMVERTEX_1WEIGHT );
#endif
	}
	else if(render->type==TEXTURE_CUSTOMVERTEX_1WEIGHT&&skin.length){

		if(appearance!=NULL){
			ImageTexture* texture=dynamic_cast<ImageTexture*>(appearance->texture.getRawPtr());
			if(texture!=0){
				if (texture->update){
					const M_String& url=texture->url;
					if(url.length>0){
						if( FAILED( D3DXCreateTextureFromFile(  g_pd3dDevice, url[0].c_str(), &texture->dxTexture ) ) ){
							//dLog ("failed to set Texture from local file");
						}
						else{
							dLog ("set Texture from local file");
							texture->update=false;
						}
					}
				}
#ifdef DX8
				g_pd3dDevice->SetTexture( 0, (IDirect3DBaseTexture8 *)texture->dxTexture );
				g_pd3dDevice->SetTextureStageState( 0, D3DTSS_COLORARG1, D3DTA_TEXTURE );
				g_pd3dDevice->SetTextureStageState( 0, D3DTSS_ALPHAOP,   D3DTOP_DISABLE );
				g_pd3dDevice->SetTextureStageState( 0, D3DTSS_MAGFILTER , D3DTEXF_LINEAR  );

				g_pd3dDevice->SetStreamSource( 0, render->pVB, sizeof(TEXTUREVERTEX_1WEIGHT) );
				g_pd3dDevice->SetVertexShader( TEXTURE_CUSTOMVERTEX_1WEIGHT );
#else
				g_pd3dDevice->SetTexture( 0, texture->dxTexture );
				g_pd3dDevice->SetTextureStageState( 0, D3DTSS_COLORARG1, D3DTA_TEXTURE );
				g_pd3dDevice->SetTextureStageState( 0, D3DTSS_ALPHAOP,   D3DTOP_DISABLE );
				g_pd3dDevice->SetSamplerState( 0, D3DSAMP_MAGFILTER, D3DTEXF_LINEAR );

				g_pd3dDevice->SetStreamSource( 0, render->pVB, 0, sizeof(TEXTUREVERTEX_1WEIGHT) );
				g_pd3dDevice->SetFVF( TEXTURE_CUSTOMVERTEX_1WEIGHT );
#endif
			}
		}
	}
	else throw this;

#ifdef DX8
	g_pd3dDevice->SetIndices( render->pIB, 0 );
	g_pd3dDevice->DrawIndexedPrimitive( D3DPT_TRIANGLELIST, 0, render->vertexCount, 0, render->indexCount/3 );
#else
	g_pd3dDevice->SetIndices( render->pIB);
	g_pd3dDevice->DrawIndexedPrimitive( D3DPT_TRIANGLELIST, 0, 0, render->vertexCount, 0, render->indexCount/3 );
#endif
	g_pd3dDevice->SetRenderState( D3DRS_VERTEXBLEND, D3DVBF_DISABLE );
	g_pd3dDevice->SetRenderState( D3DRS_INDEXEDVERTEXBLENDENABLE, FALSE);
}
void Background::renderNode(Matrix parentMatrix, Vec3f parentColScale){
	if(skyColor.refValue().length){
		g_scene->backgroundColor=skyColor.refValue()[0];
	}
}
int g_lightStack;
void DirectionalLight::renderNode(Matrix parentMatrix, Vec3f parentColScale){
	return;
	D3DLIGHT light;
    ZeroMemory( &light, sizeof(D3DLIGHT) );

	Color3f colorVal = color.refValue();
	Vec3f directionVal = direction.refValue();
	float intensityVal = intensity.refValue();
	float ambientIntensityVal = ambientIntensity.refValue();

    light.Type       = D3DLIGHT_DIRECTIONAL;
    light.Diffuse.r  = colorVal.r*intensityVal;
    light.Diffuse.g  = colorVal.g*intensityVal;
    light.Diffuse.b  = colorVal.b*intensityVal;
	light.Direction.x = directionVal.x;
	light.Direction.y = directionVal.y;
	light.Direction.z = directionVal.z;
    light.Ambient.r  = ambientIntensityVal;
    light.Ambient.g  = ambientIntensityVal;
    light.Ambient.b  = ambientIntensityVal;
    g_pd3dDevice->SetLight( g_lightStack, &light );
    g_pd3dDevice->LightEnable( g_lightStack++, on.refValue() );
}
void NavigationInfo::renderNode(Matrix parentMatrix, Vec3f parentColScale){
	g_headlight=headlight.refValue();
}

///////////////////////////////////////////////

bool checkForRenderCaperbilities(){	//checkForDx8point1
    // Simply see if dpnhpast.dll exists.
    HINSTANCE            hDPNHPASTDLL       = NULL;

    hDPNHPASTDLL = LoadLibrary( "dpnhpast.dll" );
    if( hDPNHPASTDLL == NULL )
    {
	    FreeLibrary( hDPNHPASTDLL );
        dLog( "dx8.1 not installed" );
		MessageBox(NULL,"You need DirectX 8.1 or higher to run Seamless3D",NULL,MB_OK|MB_ICONERROR);
		return false;
    }
    FreeLibrary( hDPNHPASTDLL );
	return true;
}
Vec2i get2dMousePosFrom3d(Vec3f par){

	Matrix viewMatrix;//u=scaleMatrix(g_scene->scale) * translationMatrix(g_scene->translation) * rotationMatrix(g_scene->rotation);

	Matrix worldMatrix=viewMatrix;


	D3DXVECTOR3 pOut;

	D3DXVECTOR3 pV;
	pV.x=par.x;
	pV.y=par.y;
	pV.z=par.z;


	D3DVIEWPORT Viewport;
	Viewport.X=0;
	Viewport.Y=0;
	Viewport.Width = m_d3dsdBackBuffer.Width;
	Viewport.Height = m_d3dsdBackBuffer.Height;
	Viewport.MinZ=0;
	Viewport.MaxZ=1;

	D3DXVec3Project(
		&pOut,
		&pV,
		&Viewport,
		&matProj,
		&matView,
		(D3DXMATRIX*)&worldMatrix
	);
	return Vec2i(pOut.x, pOut.y);
}

Vec2i get2dMousePosFrom2d(Vec2f par){
	return get2dMousePosFrom3d(Vec3f(par.x, par.y, 0));
}

void mouse2dSync(HWND &hWnd){
	if(g_pd3dDevice==0)return;
	Vec3f v;
	POINT ptCursor;
	D3DXMATRIX m;
	Vec3f vPickRayDir;
	//Vec3f vPickRayOrig;
	GetCursorPos( &ptCursor );
    ScreenToClient( hWnd, &ptCursor );
	g_pd3dDevice->GetTransform( D3DTS_PROJECTION, &matProj );
	if(g_settings->perspective.refValue()){
		// Compute the vector of the pick ray in screen space
		v.x =  ( ( ( 6.0f * ptCursor.x ) / m_d3dsdBackBuffer.Width  ) - 3 ) / matProj._11;
		v.y = -( ( ( 6.0f * ptCursor.y ) / m_d3dsdBackBuffer.Height ) - 3 ) / matProj._22;
		v.z =  1.0f;
	}
	else{
		v.x =  ( ( ( 6.0f * ptCursor.x ) / m_d3dsdBackBuffer.Width  ) - 3 ) / (matProj._11*3);
		v.y = -( ( ( 6.0f * ptCursor.y ) / m_d3dsdBackBuffer.Height ) - 3 ) / (matProj._22*3);
		v.z =  1.0f;
	}
	  
	// Get the inverse view matrix
	g_pd3dDevice->GetTransform( D3DTS_VIEW, &matView );
	D3DXMatrixInverse( &m, NULL, &matView );
		
	  // Transform the screen space pick ray into 3D space
	vPickRayDir.x  = v.x*m._11 + v.y*m._21 + v.z*m._31;
	vPickRayDir.y  = v.x*m._12 + v.y*m._22 + v.z*m._32;
	vPickRayDir.z  = v.x*m._13 + v.y*m._23 + v.z*m._33;
	//vPickRayOrig.x = m._41;
	//vPickRayOrig.y = m._42;
	//vPickRayOrig.z = m._43;
			
	g_pointer.translation.x=vPickRayDir.x;
	g_pointer.translation.y=vPickRayDir.y;
	g_pointer.translation.z=0;
}
void toggleWireFrame(){
#ifdef PROG_ROB
			dLog("new ClickButton(toggleWireFrame),");
#endif
	if(g_wireFrameMode=!g_wireFrameMode){
		g_pd3dDevice->SetRenderState( D3DRS_FILLMODE, D3DFILL_WIREFRAME );
	}
	else{
		g_pd3dDevice->SetRenderState( D3DRS_FILLMODE, D3DFILL_SOLID );
	}
	g_scene->wireframe=g_wireFrameMode;
	//SendMessage(hWndMain, WM_PAINT,	0,	0	);
}
void destroyRender(){}

BOOL IsDepthFormatOk( D3DFORMAT DepthFormat, D3DFORMAT AdapterFormat, 
                      D3DFORMAT BackBufferFormat ) {
    // Verify that the depth format exists.
    HRESULT hr = g_pD3D->CheckDeviceFormat( D3DADAPTER_DEFAULT,
                                          D3DDEVTYPE_HAL,
                                          AdapterFormat,
                                          D3DUSAGE_DEPTHSTENCIL,
                                          D3DRTYPE_SURFACE,
                                          DepthFormat);

    if( FAILED( hr ) ) return FALSE;

    // Verify that the depth format is compatible.
    hr = g_pD3D->CheckDepthStencilMatch( D3DADAPTER_DEFAULT,
                                       D3DDEVTYPE_HAL,
                                       AdapterFormat,
                                       BackBufferFormat,
                                       DepthFormat);
    return SUCCEEDED( hr );
}

HRESULT Window::loadSurfaceFromMemory(){
	if(workSurface){
		RECT srcRect;
		srcRect.top=0;
		srcRect.left=0;
		srcRect.right=size.x;
		srcRect.bottom=size.y;


		pixBuf.length=size.x*size.y;

		if(D3D_OK != D3DXLoadSurfaceFromMemory(
			workSurface,
			0,
			&srcRect,
			pixBuf.getArray(),
			g_d3dpp->BackBufferFormat,
			srcRect.right*4,
			0,
			&srcRect,
			D3DX_DEFAULT,
			0x0
			)){
			return E_FAIL;
		}
	}
    return S_OK;
}

void Window::releaseWindow(){
	if(this){
		if(workSurface){
			workSurface->Release();
			workSurface=0;
		}
	}
/*
	int d=0;
	int s=0;
	while(s<visWinList.length){
		if(this==visWinList[s]){
			if(workSurface){
				workSurface->Release();
				workSurface=0;
			}
			s++;
			continue;
		}
		visWinList[d++]=visWinList[s++];
	}
	visWinList.length=d;
	*/
};
HRESULT Window::initWindow(){
	releaseWindow();
	if(g_d3dpp){
#ifdef DX8
		if(D3DERR_INVALIDCALL==g_pd3dDevice->CreateImageSurface(size.x, size.y, g_d3dpp->BackBufferFormat, &workSurface)){
#else
		if(D3DERR_INVALIDCALL==g_pd3dDevice->CreateOffscreenPlainSurface(size.x, size.y, g_d3dpp->BackBufferFormat, D3DPOOL_SYSTEMMEM, &workSurface, NULL)){
#endif
			return E_FAIL;
		}
//	if( D3D_OK!=D3DXLoadSurfaceFromFile(g_workSurface, NULL, NULL, "something.bmp", NULL, D3DX_FILTER_NONE, 0x0, NULL)){
//			return E_FAIL;
//	}
		if(E_FAIL==loadSurfaceFromMemory()){
			return E_FAIL;
		}
	}
    return S_OK;
}

void setFog(unsigned int color, float visibilityStart, float visibilityRange, float density, const char* fogType){
	// Enable fog blending.
	g_pd3dDevice->SetRenderState(D3DRS_FOGENABLE, TRUE);

	// Set the fog color.
	g_pd3dDevice->SetRenderState(D3DRS_FOGCOLOR, color);

	if(fogType[1] == 'I' ) {
		g_pd3dDevice->SetRenderState(D3DRS_FOGTABLEMODE, D3DFOG_LINEAR);
		g_pd3dDevice->SetRenderState(D3DRS_FOGSTART, *(DWORD *)(&visibilityStart));
		g_pd3dDevice->SetRenderState(D3DRS_FOGEND,   *(DWORD *)(&visibilityRange));
	}
	else{
		g_pd3dDevice->SetRenderState(D3DRS_FOGTABLEMODE, D3DFOG_EXP);
		g_pd3dDevice->SetRenderState(D3DRS_FOGDENSITY, *(DWORD *)(&density));
	}
}

//-----------------------------------------------------------------------------
// Name: InitD3D()
// Desc: Initializes Direct3D
//-----------------------------------------------------------------------------
HRESULT InitD3D( HWND hWnd ){
#ifdef DX8
	if( NULL == ( g_pD3D = Direct3DCreate8( D3D_SDK_VERSION ) ) ){
        return E_FAIL;
	}
#else
	if( NULL == ( g_pD3D = Direct3DCreate9( D3D_SDK_VERSION ) ) ){
        return E_FAIL;
	}
#endif

    // Create the D3D object.
	
    // Get the current desktop display mode, so we can set up a back
    // buffer of the same format
    D3DDISPLAYMODE d3ddm;
	if( FAILED( g_pD3D->GetAdapterDisplayMode( D3DADAPTER_DEFAULT, &d3ddm ) ) ){
        return E_FAIL;
	}
	char debugStr[0x200];
	sprintf(debugStr,"Resolution: Width=%i Height=%i, format=0x%X",d3ddm.Width,d3ddm.Height, d3ddm.Format);
	dLog(debugStr);
	switch(d3ddm.Format){
	case D3DFMT_R8G8B8:
		dLog("D3DFMT_R8G8B8");
		break;
	case D3DFMT_A8R8G8B8:
		dLog("D3DFMT_A8R8G8B8");
		break;
	case D3DFMT_X8R8G8B8:
		dLog("D3DFMT_X8R8G8B8");
		break;
	case D3DFMT_R5G6B5:
		dLog("D3DFMT_R5G6B5");
		break;
	case D3DFMT_X1R5G5B5:
		dLog("D3DFMT_X1R5G5B5");
		break;
	case D3DFMT_A4R4G4B4:
		dLog("D3DFMT_A4R4G4B4");
		break;
	case D3DFMT_A8:
		dLog("D3DFMT_A8");
		break;
	}
	
    // Set up the structure used to create the D3DDevice. Since we are now
    // using more complex geometry, we will create a device with a zbuffer.
	g_d3dpp=new D3DPRESENT_PARAMETERS;
    ZeroMemory( g_d3dpp, sizeof(D3DPRESENT_PARAMETERS) );

	RECT rect;
	GetClientRect(hWndMain, &rect);
	g_3d->size = Vec2i(rect.right, rect.bottom);
	g_d3dpp->BackBufferWidth = rect.right;
	g_d3dpp->BackBufferHeight = rect.bottom;
//	g_d3dpp->BackBufferWidth	=	g_settings->win3dSize.x;//dont think we need to set the size of the window since they get filled with data
//	g_d3dpp->BackBufferHeight	=	g_settings->win3dSize.y;

    g_d3dpp->BackBufferFormat = d3ddm.Format;
	//g_d3dpp->BackBufferFormat = D3DFMT_X8R8G8B8;

	g_d3dpp->hDeviceWindow=hWnd;

    g_d3dpp->Windowed = TRUE;

	g_d3dpp->BackBufferCount=1;

    g_d3dpp->SwapEffect = D3DSWAPEFFECT_DISCARD;

/*	if( SUCCEEDED(g_pD3D->CheckDeviceMultiSampleType(D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, g_d3dpp->BackBufferFormat, g_d3dpp->Windowed, D3DMULTISAMPLE_2_SAMPLES))){
		g_d3dpp->MultiSampleType = D3DMULTISAMPLE_2_SAMPLES;
	}
	else{
		g_d3dpp->MultiSampleType=D3DMULTISAMPLE_NONE;
	}*/
	g_d3dpp->MultiSampleType=D3DMULTISAMPLE_NONE;

    g_d3dpp->EnableAutoDepthStencil = TRUE;

    g_d3dpp->AutoDepthStencilFormat = D3DFMT_D16;

	g_d3dpp->FullScreen_RefreshRateInHz=0;

	g_d3dpp->Flags = D3DPRESENTFLAG_LOCKABLE_BACKBUFFER;

#ifdef DX8
	g_d3dpp->FullScreen_PresentationInterval = 0;
#else
	g_d3dpp->PresentationInterval=D3DPRESENT_INTERVAL_IMMEDIATE;
#endif
    // Create the D3DDevice
    //if( FAILED( g_pD3D->CreateDevice( D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, hWnd, D3DCREATE_HARDWARE_VERTEXPROCESSING, g_d3dpp, &g_pd3dDevice ) ) ){
    //    return E_FAIL;
    //}
	if(FAILED(g_pD3D->CreateDevice(D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, hWnd, D3DCREATE_SOFTWARE_VERTEXPROCESSING, g_d3dpp, &g_pd3dDevice ) ) ){
        return E_FAIL;
    }
    // Turn off culling
    //g_pd3dDevice->SetRenderState( D3DRS_CULLMODE, D3DCULL_NONE );
    //g_pd3dDevice->SetRenderState( D3DRS_FILLMODE, D3DFILL_WIREFRAME );
//		g_pd3dDevice->SetRenderState( D3DRS_VERTEXBLEND, D3DVBF_1WEIGHTS );
//		g_pd3dDevice->SetRenderState( D3DRS_VERTEXBLEND, D3DVBF_0WEIGHTS );
//		g_pd3dDevice->SetRenderState( D3DRS_INDEXEDVERTEXBLENDENABLE, TRUE);


	
    // Turn on the zbuffer
    g_pd3dDevice->SetRenderState( D3DRS_ZENABLE, TRUE );
	
	if(g_settings->backfaceIgnore.refValue()){
		g_pd3dDevice->SetRenderState( D3DRS_CULLMODE, D3DCULL_CCW );
	}
	else{
		g_pd3dDevice->SetRenderState( D3DRS_CULLMODE, D3DCULL_NONE );
	}

	Window::reInitAllVisable();

/*
	if(Window::initializeAllControlPanels()==false){
		return E_FAIL;
	}
	if(E_FAIL==g_stw->initWindow()){
		return E_FAIL;
	}
	/*
	if(E_FAIL==g_stw->cp.initWindow()){
		return E_FAIL;
	}
	if(E_FAIL==g_tb->initWindow()){
		return E_FAIL;
	}
	if(E_FAIL==g_tt->initWindow()){
		return E_FAIL;
	}
	*/
	/*
	for(int i=g_menuList.length; --i>=0;){
		g_menuList[i]->size=Vec2i(10,10);
		if(E_FAIL==(g_menuList[i]->initWindow())){
			return E_FAIL;
		}
	}
	*/
#ifdef DX8
    LOGFONT lf;
    ZeroMemory(&lf,sizeof(LOGFONT));
	
    lf.lfHeight = 20;
    lf.lfWeight = 700;
    lf.lfCharSet = ANSI_CHARSET;
    lf.lfOutPrecision = OUT_DEFAULT_PRECIS;
    lf.lfClipPrecision = CLIP_DEFAULT_PRECIS;
    lf.lfQuality = PROOF_QUALITY;
    lf.lfPitchAndFamily = DEFAULT_PITCH | FF_SWISS;
    strcpy(lf.lfFaceName,TEXT("Arial"));
	if(D3DXCreateFontIndirect( g_pd3dDevice, &lf, &ppFont)!=D3D_OK)
		return E_FAIL;
#else
	D3DXFONT_DESC dfd;
	ZeroMemory(&dfd,sizeof(D3DXFONT_DESC));
	dfd.Height = 20;
	dfd.Weight = 700;
	dfd.CharSet = ANSI_CHARSET;
	dfd.OutputPrecision = OUT_DEFAULT_PRECIS;
	dfd.Quality = PROOF_QUALITY;
	dfd.PitchAndFamily = DEFAULT_PITCH | FF_SWISS;
	strcpy(dfd.FaceName,TEXT("Arial"));
	if(D3DXCreateFontIndirect( g_pd3dDevice, &dfd, &ppFont)!=D3D_OK)
		return E_FAIL;
#endif

	
    D3DXVECTOR3 vecDir;
    ZeroMemory( &g_light, sizeof(D3DLIGHT) );
    g_light.Type       = D3DLIGHT_DIRECTIONAL;
    g_light.Diffuse.r  = 1;
    g_light.Diffuse.g  = 1;
    g_light.Diffuse.b  = 1;
    vecDir = D3DXVECTOR3(0,	0, -1);
    D3DXVec3Normalize( (D3DXVECTOR3*)&g_light.Direction, &vecDir );
    g_light.Range       = 1000.0f;
    g_pd3dDevice->SetLight( 0, &g_light );

	g_tb->paint();

	g_pd3dDevice->SetRenderState( D3DRS_ALPHABLENDENABLE, FALSE);
	g_pd3dDevice->SetRenderState(D3DRS_SRCBLEND, D3DBLEND_SRCALPHA );
	g_pd3dDevice->SetRenderState(D3DRS_DESTBLEND, D3DBLEND_INVSRCALPHA );


	if(S_OK != D3DXCreateRenderToSurface(g_pd3dDevice, d3ddm.Width, d3ddm.Height, d3ddm.Format, TRUE, D3DFMT_X8R8G8B8, &g_renderToSurface)){
		return E_FAIL;
	}
	//g_renderToSurface->GetSurfaceLevel( 0, &g_offScreenSurface );


	return S_OK;
}
//-----------------------------------------------------------------------------
// Name: InitD3D()
// Desc: Initializes Direct3D
//-----------------------------------------------------------------------------
HRESULT InitRender( HWND hWnd ){
	return InitD3D( hWnd );

	D3DADAPTER_IDENTIFIER pIdentifier;

	dLog("InitRender");


    // Create the D3D object.
#ifdef DX8
	if( NULL == ( g_pD3D = Direct3DCreate8( D3D_SDK_VERSION ) ) ){
        return E_FAIL;
	}
#else
	if( NULL == ( g_pD3D = Direct3DCreate9( D3D_SDK_VERSION ) ) ){
        return E_FAIL;
	}
#endif
	dLog("Direct3DCreate ok");
	

	if( D3D_OK != ( g_pD3D->GetAdapterIdentifier(D3DADAPTER_DEFAULT,0, &pIdentifier ) ) )
        return E_FAIL;
	dLog(pIdentifier.Driver);
	dLog(pIdentifier.Description);
    
	// Get the current desktop display mode, so we can set up a back
    // buffer of the same format
    D3DDISPLAYMODE d3ddm;
    if( FAILED( g_pD3D->GetAdapterDisplayMode( D3DADAPTER_DEFAULT, &d3ddm ) ) )
        return E_FAIL;
	char debugStr[0x200];
	sprintf(debugStr,"Resolution: Width=%i Height=%i, format=0x%X",d3ddm.Width,d3ddm.Height, d3ddm.Format);
	dLog(debugStr);
	switch(d3ddm.Format){
	case D3DFMT_R8G8B8:
		dLog("D3DFMT_R8G8B8");
		break;
	case D3DFMT_A8R8G8B8:
		dLog("D3DFMT_A8R8G8B8");
		break;
	case D3DFMT_X8R8G8B8:
		dLog("D3DFMT_X8R8G8B8");
		break;
	case D3DFMT_R5G6B5:
		dLog("D3DFMT_R5G6B5");
		break;
	case D3DFMT_X1R5G5B5:
		dLog("D3DFMT_X1R5G5B5");
		break;
	case D3DFMT_A4R4G4B4:
		dLog("D3DFMT_A4R4G4B4");
		break;
	case D3DFMT_A8:
		dLog("D3DFMT_A8");
		break;
	}
    // Set up the structure used to create the D3DDevice. Since we are now
    // using more complex geometry, we will create a device with a zbuffer.
	g_d3dpp=new D3DPRESENT_PARAMETERS;
	ZeroMemory( g_d3dpp, sizeof(D3DPRESENT_PARAMETERS) );
    g_d3dpp->Windowed = TRUE;
    g_d3dpp->SwapEffect = D3DSWAPEFFECT_DISCARD;
    g_d3dpp->BackBufferFormat = d3ddm.Format;
    g_d3dpp->EnableAutoDepthStencil = TRUE;
    g_d3dpp->AutoDepthStencilFormat = D3DFMT_D16;
	
    // Create the D3DDevice
	HDC hdc = GetDC(hWndMain);

	D3DCAPS d3dcaps;
    g_pD3D->GetDeviceCaps(D3DADAPTER_DEFAULT ,D3DDEVTYPE_HAL,&d3dcaps);


	if(IsDepthFormatOk( D3DFMT_D16, d3ddm.Format, d3ddm.Format)){
		dLog("DepthFormatOk");
	}
	
    if( SUCCEEDED( g_pD3D->CreateDevice( D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, hWnd,
		D3DCREATE_SOFTWARE_VERTEXPROCESSING,
		g_d3dpp, &g_pd3dDevice ) ) ){

		dLog("D3DDEVTYPE_HAL D3DCREATE_SOFTWARE_VERTEXPROCESSING");
	}
	else if( SUCCEEDED( g_pD3D->CreateDevice( D3DADAPTER_DEFAULT, D3DDEVTYPE_REF, hWnd,
		//	D3DCREATE_HARDWARE_VERTEXPROCESSING,
			D3DCREATE_SOFTWARE_VERTEXPROCESSING,
			g_d3dpp, &g_pd3dDevice ) ) )	{

		dLog("D3DDEVTYPE_REF D3DCREATE_SOFTWARE_VERTEXPROCESSING");
    }
	else{
		return E_FAIL;
	}
	dLog("CreateDevice ok");
	
    // Turn off culling
    g_pd3dDevice->SetRenderState(D3DRS_ALPHABLENDENABLE , FALSE);

 	// Set the source blend state.
	g_pd3dDevice->SetRenderState(D3DRS_SRCBLEND, D3DBLEND_DESTALPHA );
 
	// Set the destination blend state.
	g_pd3dDevice->SetRenderState(D3DRS_DESTBLEND, D3DBLEND_INVSRCALPHA );
	
	if(g_settings->backfaceIgnore.refValue()){
		g_pd3dDevice->SetRenderState( D3DRS_CULLMODE, D3DCULL_CCW );
	}
	else{
		g_pd3dDevice->SetRenderState( D3DRS_CULLMODE, D3DCULL_NONE );
	}

	if(g_wireFrameMode){
		g_pd3dDevice->SetRenderState( D3DRS_FILLMODE, D3DFILL_WIREFRAME );
	}
	else{
		g_pd3dDevice->SetRenderState( D3DRS_FILLMODE, D3DFILL_SOLID );
	}

    // Turn on the zbuffer
    g_pd3dDevice->SetRenderState( D3DRS_ZENABLE, TRUE );
	g_pd3dDevice->SetRenderState( D3DRS_ALPHABLENDENABLE    , FALSE);

/*
	if(smls->appearance != NULL && smls->appearance->texture != NULL && smls->appearance->texture->url.length){

		if( FAILED( D3DXCreateTextureFromFile(  g_pd3dDevice, smls->appearance->texture->url.c_str(), &smls->appearance->texture->dxTexture ) ) ){
			dLog ("failed to set Texture from local file");
		}
	}
*/
    return S_OK;
}
Effect* workOutEffectType(const Appearance*& appearance);

//void Cone::renderNode(Matrix parentMatrix, Vec3f parentColScale, GeomSource*& geomSource, DXRender*& render, Appearance* appearance);
void Cone::renderNode(Matrix parentMatrix, Vec3f parentColScale, GeomSource& geomSource, DXRender*& render, const Appearance* appearance){
	if(renderProcessed==false){
		genCone(height, bottomRadius, side, bottom, 16, geomSource, workOutEffectType(appearance));
		geomSource.convertToRenderFormat(render, appearance);
		renderProcessedQue+=this;
	}
	geomSource.renderNode(parentMatrix, parentColScale, render, appearance);
}
void Sphere::renderNode(Matrix parentMatrix, Vec3f parentColScale, GeomSource& geomSource, DXRender*& render, const Appearance* appearance){
	if(renderProcessed==false || render==0){
		genSphere(radius, 12, 6, geomSource, workOutEffectType(appearance));
		geomSource.convertToRenderFormat(render, appearance);
		renderProcessedQue+=this;
	}
	geomSource.renderNode(parentMatrix, parentColScale, render, appearance);
}
void Cylinder::renderNode(Matrix parentMatrix, Vec3f parentColScale, GeomSource& geomSource, DXRender*& render, const Appearance* appearance){
	if(renderProcessed==false || render==0){
		genCylinder( height, radius, side, top, bottom, 12, 1, geomSource, workOutEffectType(appearance));
		geomSource.convertToRenderFormat(render, appearance);
		renderProcessedQue+=this;
	}
	geomSource.renderNode(parentMatrix, parentColScale, render, appearance);
}
void Box::renderNode(Matrix parentMatrix, Vec3f parentColScale, GeomSource& geomSource, DXRender*& render, const Appearance* appearance){
	if(renderProcessed==false || render==0){
		genBox( size, 1, 1, geomSource, workOutEffectType(appearance));
		geomSource.convertToRenderFormat(render, appearance);
		renderProcessedQue+=this;
	}
	geomSource.renderNode(parentMatrix, parentColScale, render, appearance);
}
void Text::renderNode(Matrix parentMatrix, Vec3f parentColScale, GeomSource& geomSource, DXRender*& render, const Appearance* appearance){}
void PointSet::renderNode(Matrix parentMatrix, Vec3f parentColScale, GeomSource& geomSource, DXRender*& render, const Appearance* appearance){}
void ElevationGrid::renderNode(Matrix parentMatrix, Vec3f parentColScale, GeomSource& geomSource, DXRender*& render, const Appearance* appearance){}
void Script::renderNode(Matrix parentMatrix, Vec3f parentColScale){
}
void convertVisualCluesToRenderFormat(){
	g_redPoint->convertToRenderFormat(g_redRenderPoint, 0);
	g_nextPoint->convertToRenderFormat(g_nextRenderPoint, 0);
	g_lastPoint->convertToRenderFormat(g_lastRenderPoint, 0);
	g_yellowPoint->convertToRenderFormat(g_yellowRenderPoint, 0);
	g_triClue->convertToRenderFormat(g_triClueShape->render);
}
bool createTextureFromFileInMemory(M_<char>& internetBuf, IDirect3DTexture*&  dxTexture, bool& alphaTexture){
	if( FAILED( D3DXCreateTextureFromFileInMemory(  g_pd3dDevice, internetBuf.getArray(), internetBuf.length, &dxTexture ) ) ){
		return false;
	}
	dxTexture->GetSurfaceLevel(0, &g_textureSurface);
	g_textureSurface->GetDesc(&g_surfaceBuffer);
	alphaTexture = false;
	if(g_surfaceBuffer.Format == D3DFMT_A8R8G8B8){
		alphaTexture = true;
	}
	return true;
}
void Window::setWinSize(Vec2i& par){
	pixBuf.length = par.x * par.y;
/*
	if(workSurface){
		workSurface->Release();
	}
	*/
	if(E_FAIL==initWindow()){
		throw;
	}
}
void SceneTree::resizePageUpPageDown(){
	pageUp.dstPos=topScroll.dstPos+Vec2i(5, slider.bSize.y);
	pageUp.bSize=Vec2i(6, slider.dstPos.y - topScroll.dstPos.y - topScroll.bSize.y);
	pageDown.dstPos=slider.dstPos+Vec2i(5, slider.bSize.y);
	pageDown.bSize=Vec2i(6, bottomScroll.dstPos.y - pageDown.dstPos.y);
}

void SceneTree::setWinSize(Vec2i& raw){
	if(raw.x<100){
		raw.x=100;
	}
	if(raw.y<100){
		raw.y=100;
	}
	pixBuf.length = raw.x * raw.y;
	/*
	if(workSurface){
		workSurface->Release();
	}
	*/
	if(E_FAIL==(initWindow())){
		throw;
	}
	topScroll.dstPos=Vec2i(size.x-20, title_bar_height);
	bottomScroll.dstPos=Vec2i(size.x-20, size.y-19);
	slider.setPixelPos();
	resizePageUpPageDown();
	numOfVisLines=(size.y-beginYOffset)/pixSymbolCY;
	textCell.cellSize = Vec2i(size.x - textCell.cellPos.x - 15, 16);
}
Window::~Window(){
	if(workSurface){
		releaseWindow();
	}
}
void Window::releaseAllControlPanels(){
	for(int i=visWinList.length; --i>=0;){
		if(ControlPanel* controlPanel=dynamic_cast<ControlPanel*>(visWinList[i].p)){
			if(controlPanel->workSurface){
				controlPanel->workSurface->Release();
				controlPanel->workSurface=0;
			}
		}
	}
}
void Window::reInitAllVisable(){
	for(int i=visWinList.length; --i>=0;){
		visWinList[i]->initWindow();
	}
}

void convertBMPToPng(const char* filename, int width, int height, M_<char>& header, Pix24* pixels){
	M_<char> bytes;
	bytes += header;
	bytes.length = header.length * 2 + height * width * 3;
	Pix24* destPixels = reinterpret_cast<Pix24*>(bytes.getArray() + header.length);
	for(int i = 0; i < height * width; i++){
		destPixels[i] = pixels[i];
	}
	IDirect3DTexture*  dxTexture = 0;
	if(FAILED(D3DXCreateTextureFromFileInMemory(g_pd3dDevice, bytes.getArray(), bytes.length, &dxTexture))){
		return;
	}
	if(FAILED(D3DXSaveTextureToFile(filename, D3DXIFF_DDS, dxTexture, 0))){
		return;
	}
	dxTexture->Release();
}
