/* Seamless3d

Copyright (C) 2001-2010 Graham John Perrett (AKA thyme) thyme@seamless3d.com www.seamless3d.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this
software and associated documentation files (the "Software"), to deal in the Software
without restriction, including without limitation the rights to use, copy, modify,
merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to the following
conditions:

The above copyright notice and this permission notice shall be included in all copies or
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT
OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.
*/

//key template names
#define MOV_MEM_ESP 3
#define REF_SIZE 4

class Component_X;
class Component_Y;
class Component_Z;
struct ColorFormat;

enum ValueType{
	S_BOOL,
	S_COLOR3F,
	S_FLOAT,
	S_IMAGE,
	S_INT32,
	NODEPTR,
	S_ROTATION,
	S_STRING,
	S_DOUBLE,
	S_VEC2F,
	S_VEC3F,
	M_BOOL,
	M_COLOR,
	M_FLOAT,
	M_IMAGE,
	M_INT32,
	M_NODE,
	M_ROTATION,
	M_STRING,
	M_TIME,
	M_VEC2F,
	M_VEC3F,
	_VOID,
};
#define VARABLETYPE_LOOKUPLEN 23
#define FIRST_VALUE_TYPE S_BOOL
#define FIRST_M_VALUE_TYPE M_BOOL
#define LAST_VALUE_TYPE M_VEC3F

template <class T,class S> inline int templateType();

template <> inline int templateType<S_Bool,Single>(){
	return S_BOOL;
}
template <> inline int templateType<Color3f,Single>(){
	return S_COLOR3F;
}
template <> inline int templateType<S_Float,Single>(){
	return S_FLOAT;
}
template <> inline int templateType<Image,Single>(){
	return S_IMAGE;
}
template <> inline int templateType<S_Int32,Single>(){
	return S_INT32;
}
template <> inline int templateType<NodePtr,Single>(){
	return NODEPTR;
}
template <> inline int templateType<Rotation,Single>(){
	return S_ROTATION;
}
template <> inline int templateType<String,Single>(){
	return S_STRING;
}
template <> inline int templateType<double,Single>(){
	return S_DOUBLE;
}
template <> inline int templateType<Vec2f,Single>(){
	return S_VEC2F;
}
template <> inline int templateType<Vec3f,Single>(){
	return S_VEC3F;
}
template <> inline int templateType<S_Bool,Multiple>(){
	return M_BOOL;
}
template <> inline int templateType<Color3f,Multiple>(){
	return M_COLOR;
}
template <> inline int templateType<S_Float,Multiple>(){
	return M_FLOAT;
}
template <> inline int templateType<Image,Multiple>(){
	return M_IMAGE;
}
template <> inline int templateType<S_Int32,Multiple>(){
	return M_INT32;
}
template <> inline int templateType<NodePtr,Multiple>(){
	return M_NODE;
}
template <> inline int templateType<Rotation,Multiple>(){
	return M_ROTATION;
}
template <> inline int templateType<String,Multiple>(){
	return M_STRING;
}
template <> inline int templateType<double,Multiple>(){
	return M_TIME;
}
template <> inline int templateType<Vec2f,Multiple>(){
	return M_VEC2F;
}
template <> inline int templateType<Vec3f,Multiple>(){
	return M_VEC3F;
}

class Asm2CppAddy {
public:
	Asm2CppAddy():
		addy(0),
		byteIndex(0)
	{}
	Asm2CppAddy(void* i_addy, int i_byteIndex):
		addy(i_addy),
		byteIndex(i_byteIndex)
	{}
	void* addy;
	int byteIndex;
};

class ACall {
public:
	ACall():
		function(0),
		byteIndex(0)
	{}
	ACall(Function* i_function, int i_byteIndex);
	Function* function;
	int byteIndex;
};
class Local2ndCompile {
public:
	Local2ndCompile():
		variableIndex(0),
		byteIndex(0),
		callOffset(0)
	{}
	Local2ndCompile(int i_variableIndex, int i_callOffset, int i_byteIndex):
		variableIndex(i_variableIndex),
		callOffset(i_callOffset),
		byteIndex(i_byteIndex)
	{}
	int variableIndex;
	int callOffset;
	int byteIndex;
};
typedef M_ <L_Variable*> M_L_;

typedef V_<S_Bool,Single,Global> G_S_Bool;
typedef V_<Color3f,Single,Global> G_S_Color;
typedef V_<S_Float,Single,Global> G_S_Float;
typedef V_<Image,Single,Global> G_S_Image;
typedef V_<S_Int32,Single,Global> G_S_Int32;
typedef V_<NodePtr,Single,Global> G_NodePtr;
typedef V_<Rotation,Single,Global> G_S_Rotation;
typedef V_<String,Single,Global> G_S_String;
typedef V_<double,Single,Global> G_S_Time;
typedef V_<Vec2f,Single,Global> G_S_Vec2f;
typedef V_<Vec3f,Single,Global> G_S_Vec3f;
typedef V_<S_Bool,Multiple,Global> G_M_Bool;
typedef V_<Color3f,Multiple,Global> G_M_Color;
typedef V_<S_Float,Multiple,Global> G_M_Float;
typedef V_<Image,Multiple,Global> G_M_Image;
typedef V_<S_Int32,Multiple,Global> G_M_Int;
typedef V_<NodePtr,Multiple,Global> G_M_Node;
typedef V_<Rotation,Multiple,Global> G_M_Rotation;
typedef V_<String,Multiple,Global> G_M_String;
typedef V_<double,Multiple,Global> G_M_Time;
typedef V_<Vec2f,Multiple,Global> G_M_Vec2f;
typedef V_<Vec3f,Multiple,Global> G_M_Vec3f;

typedef V_<S_Bool,Single,Local> L_S_Bool;
typedef V_<Color3f,Single,Local> L_S_Color;
typedef V_<S_Float,Single,Local> L_S_Float;
typedef V_<Image,Single,Local> L_S_Image;
typedef V_<S_Int32,Single,Local> L_S_Int32;
typedef V_<NodePtr,Single,Local> L_NodePtr;
typedef V_<Rotation,Single,Local> L_S_Rotation;
typedef V_<String,Single,Local> L_S_String;
typedef V_<double,Single,Local> L_S_Time;
typedef V_<Vec2f,Single,Local> L_S_Vec2f;
typedef V_<Vec3f,Single,Local> L_S_Vec3f;
typedef V_<S_Bool,Multiple,Local> L_M_Bool;
typedef V_<Color3f,Multiple,Local> L_M_Color;
typedef V_<S_Float,Multiple,Local> L_M_Float;
typedef V_<Image,Multiple,Local> L_M_Image;
typedef V_<S_Int32,Multiple,Local> L_M_Int32;
typedef V_<NodePtr,Multiple,Local> L_M_Node;
typedef V_<Rotation,Multiple,Local> L_M_Rotation;
typedef V_<String,Multiple,Local> L_M_String;
typedef V_<double,Multiple,Local> L_M_Time;
typedef V_<Vec2f,Multiple,Local> L_M_Vec2f;
typedef V_<Vec3f,Multiple,Local> L_M_Vec3f;


class ConstData{
public:
	ConstData():
		constIdx(0),
		codeIdx(0)
	{}
	ConstData(int i_constIdx,int i_codeIdx):
		constIdx(i_constIdx),
		codeIdx(i_codeIdx)
	{}
	int constIdx;
	int codeIdx;
};

class AReturn {
public:
	AReturn():
		endAddress(0),
		returnValOffset(0)
	{
	}
	AReturn(int& i_endAddress):
		endAddress(i_endAddress)
	{
	}
	int endAddress;
	int returnValOffset;
};
class Local2ndCompile;
typedef M_ <Local2ndCompile> M_Local;
extern M_Local g_local2ndCompile;
typedef M_ <ConstData> M_ConstData;
extern M_ConstData g_const2ndCompile;
class Initialize {
public:
	Initialize():
		initTrailer(0)
	{
	}
	Operation* initTrailer;
	virtual String toTrailString()=0;
};

extern "C" unsigned char* CopyScratchPadToGlobal_ele_S_Bool;
extern "C" unsigned char* CopyScratchPadToGlobal_ele_S_Color;
extern "C" unsigned char* CopyScratchPadToGlobal_ele_S_Float;
extern "C" unsigned char* CopyScratchPadToGlobal_ele_S_Image;
extern "C" unsigned char* CopyScratchPadToGlobal_ele_S_Int32;
extern "C" unsigned char* CopyScratchPadToGlobal_ele_NodePtr;
extern "C" unsigned char* CopyScratchPadToGlobal_ele_S_Rotation;
extern "C" unsigned char* CopyScratchPadToGlobal_ele_S_String;
extern "C" unsigned char* CopyScratchPadToGlobal_ele_S_Time;
extern "C" unsigned char* CopyScratchPadToGlobal_ele_S_Vec2f;
extern "C" unsigned char* CopyScratchPadToGlobal_ele_S_Vec3f;
extern "C" unsigned char* CopyScratchPadToGlobal_ele_M_Bool;
extern "C" unsigned char* CopyScratchPadToGlobal_ele_M_Color;
extern "C" unsigned char* CopyScratchPadToGlobal_ele_M_Float;
extern "C" unsigned char* CopyScratchPadToGlobal_ele_M_Image;
extern "C" unsigned char* CopyScratchPadToGlobal_ele_M_Int32;
extern "C" unsigned char* CopyScratchPadToGlobal_ele_M_Node;
extern "C" unsigned char* CopyScratchPadToGlobal_ele_M_Rotation;
extern "C" unsigned char* CopyScratchPadToGlobal_ele_M_String;
extern "C" unsigned char* CopyScratchPadToGlobal_ele_M_Time;
extern "C" unsigned char* CopyScratchPadToGlobal_ele_M_Vec2f;
extern "C" unsigned char* CopyScratchPadToGlobal_ele_M_Vec3f;

template <class T> inline unsigned char** CopyScratchPadToGlobal_insert();
template <> inline unsigned char** CopyScratchPadToGlobal_insert<S_Bool>(){ return &CopyScratchPadToGlobal_ele_S_Bool;}
template <> inline unsigned char** CopyScratchPadToGlobal_insert<Color3f>(){ return &CopyScratchPadToGlobal_ele_S_Color;}
template <> inline unsigned char** CopyScratchPadToGlobal_insert<S_Float>(){ return &CopyScratchPadToGlobal_ele_S_Float;}
template <> inline unsigned char** CopyScratchPadToGlobal_insert<Image>(){ return &CopyScratchPadToGlobal_ele_S_Image;}
template <> inline unsigned char** CopyScratchPadToGlobal_insert<S_Int32>(){ return &CopyScratchPadToGlobal_ele_S_Int32;}
template <> inline unsigned char** CopyScratchPadToGlobal_insert<NodePtr>(){ return &CopyScratchPadToGlobal_ele_NodePtr;}
template <> inline unsigned char** CopyScratchPadToGlobal_insert<Rotation>(){ return &CopyScratchPadToGlobal_ele_S_Rotation;}
template <> inline unsigned char** CopyScratchPadToGlobal_insert<String>(){ return &CopyScratchPadToGlobal_ele_S_String;}
template <> inline unsigned char** CopyScratchPadToGlobal_insert<double>(){ return &CopyScratchPadToGlobal_ele_S_Time;}
template <> inline unsigned char** CopyScratchPadToGlobal_insert<Vec2f>(){ return &CopyScratchPadToGlobal_ele_S_Vec2f;}
template <> inline unsigned char** CopyScratchPadToGlobal_insert<Vec3f>(){ return &CopyScratchPadToGlobal_ele_S_Vec3f;}
template <> inline unsigned char** CopyScratchPadToGlobal_insert<M_Bool>(){ return &CopyScratchPadToGlobal_ele_M_Bool;}
template <> inline unsigned char** CopyScratchPadToGlobal_insert<M_Color3f>(){ return &CopyScratchPadToGlobal_ele_M_Color;}
template <> inline unsigned char** CopyScratchPadToGlobal_insert<M_Float>(){ return &CopyScratchPadToGlobal_ele_M_Float;}
template <> inline unsigned char** CopyScratchPadToGlobal_insert<M_Image>(){ return &CopyScratchPadToGlobal_ele_M_Image;}
template <> inline unsigned char** CopyScratchPadToGlobal_insert<M_Int>(){ return &CopyScratchPadToGlobal_ele_M_Int32;}
template <> inline unsigned char** CopyScratchPadToGlobal_insert<M_Node>(){ return &CopyScratchPadToGlobal_ele_M_Node;}
template <> inline unsigned char** CopyScratchPadToGlobal_insert<M_Rotation>(){ return &CopyScratchPadToGlobal_ele_M_Rotation;}
template <> inline unsigned char** CopyScratchPadToGlobal_insert<M_String>(){ return &CopyScratchPadToGlobal_ele_M_String;}
template <> inline unsigned char** CopyScratchPadToGlobal_insert<M_Double>(){ return &CopyScratchPadToGlobal_ele_M_Time;}
template <> inline unsigned char** CopyScratchPadToGlobal_insert<M_Vec2f>(){ return &CopyScratchPadToGlobal_ele_M_Vec2f;}
template <> inline unsigned char** CopyScratchPadToGlobal_insert<M_Vec3f>(){ return &CopyScratchPadToGlobal_ele_M_Vec3f;}


extern "C" unsigned char* subScratchPadToGlobal_ele_S_Bool;
extern "C" unsigned char* subScratchPadToGlobal_ele_S_Color;
extern "C" unsigned char* subScratchPadToGlobal_ele_S_Float;
extern "C" unsigned char* subScratchPadToGlobal_ele_S_Image;
extern "C" unsigned char* subScratchPadToGlobal_ele_S_Int32;
extern "C" unsigned char* subScratchPadToGlobal_ele_NodePtr;
extern "C" unsigned char* subScratchPadToGlobal_ele_S_Rotation;
extern "C" unsigned char* subScratchPadToGlobal_ele_S_String;
extern "C" unsigned char* subScratchPadToGlobal_ele_S_Time;
extern "C" unsigned char* subScratchPadToGlobal_ele_S_Vec2f;
extern "C" unsigned char* subScratchPadToGlobal_ele_S_Vec3f;
extern "C" unsigned char* subScratchPadToGlobal_ele_M_Bool;
extern "C" unsigned char* subScratchPadToGlobal_ele_M_Color;
extern "C" unsigned char* subScratchPadToGlobal_ele_M_Float;
extern "C" unsigned char* subScratchPadToGlobal_ele_M_Image;
extern "C" unsigned char* subScratchPadToGlobal_ele_M_Int32;
extern "C" unsigned char* subScratchPadToGlobal_ele_M_Node;
extern "C" unsigned char* subScratchPadToGlobal_ele_M_Rotation;
extern "C" unsigned char* subScratchPadToGlobal_ele_M_String;
extern "C" unsigned char* subScratchPadToGlobal_ele_M_Time;
extern "C" unsigned char* subScratchPadToGlobal_ele_M_Vec2f;
extern "C" unsigned char* subScratchPadToGlobal_ele_M_Vec3f;

template <class T> inline unsigned char** subScratchPadToGlobal_insert();
template <> inline unsigned char** subScratchPadToGlobal_insert<S_Bool>(){ return &subScratchPadToGlobal_ele_S_Bool;}
template <> inline unsigned char** subScratchPadToGlobal_insert<Color3f>(){ return &subScratchPadToGlobal_ele_S_Color;}
template <> inline unsigned char** subScratchPadToGlobal_insert<S_Float>(){ return &subScratchPadToGlobal_ele_S_Float;}
template <> inline unsigned char** subScratchPadToGlobal_insert<Image>(){ return &subScratchPadToGlobal_ele_S_Image;}
template <> inline unsigned char** subScratchPadToGlobal_insert<S_Int32>(){ return &subScratchPadToGlobal_ele_S_Int32;}
template <> inline unsigned char** subScratchPadToGlobal_insert<NodePtr>(){ return &subScratchPadToGlobal_ele_NodePtr;}
template <> inline unsigned char** subScratchPadToGlobal_insert<Rotation>(){ return &subScratchPadToGlobal_ele_S_Rotation;}
template <> inline unsigned char** subScratchPadToGlobal_insert<String>(){ return &subScratchPadToGlobal_ele_S_String;}
template <> inline unsigned char** subScratchPadToGlobal_insert<double>(){ return &subScratchPadToGlobal_ele_S_Time;}
template <> inline unsigned char** subScratchPadToGlobal_insert<Vec2f>(){ return &subScratchPadToGlobal_ele_S_Vec2f;}
template <> inline unsigned char** subScratchPadToGlobal_insert<Vec3f>(){ return &subScratchPadToGlobal_ele_S_Vec3f;}
template <> inline unsigned char** subScratchPadToGlobal_insert<M_Bool>(){ return &subScratchPadToGlobal_ele_M_Bool;}
template <> inline unsigned char** subScratchPadToGlobal_insert<M_Color3f>(){ return &subScratchPadToGlobal_ele_M_Color;}
template <> inline unsigned char** subScratchPadToGlobal_insert<M_Float>(){ return &subScratchPadToGlobal_ele_M_Float;}
template <> inline unsigned char** subScratchPadToGlobal_insert<M_Image>(){ return &subScratchPadToGlobal_ele_M_Image;}
template <> inline unsigned char** subScratchPadToGlobal_insert<M_Int>(){ return &subScratchPadToGlobal_ele_M_Int32;}
template <> inline unsigned char** subScratchPadToGlobal_insert<M_Node>(){ return &subScratchPadToGlobal_ele_M_Node;}
template <> inline unsigned char** subScratchPadToGlobal_insert<M_Rotation>(){ return &subScratchPadToGlobal_ele_M_Rotation;}
template <> inline unsigned char** subScratchPadToGlobal_insert<M_String>(){ return &subScratchPadToGlobal_ele_M_String;}
template <> inline unsigned char** subScratchPadToGlobal_insert<M_Double>(){ return &subScratchPadToGlobal_ele_M_Time;}
template <> inline unsigned char** subScratchPadToGlobal_insert<M_Vec2f>(){ return &subScratchPadToGlobal_ele_M_Vec2f;}
template <> inline unsigned char** subScratchPadToGlobal_insert<M_Vec3f>(){ return &subScratchPadToGlobal_ele_M_Vec3f;}

extern "C" unsigned char* addScratchPadToGlobal_ele_S_Bool;
extern "C" unsigned char* addScratchPadToGlobal_ele_S_Color;
extern "C" unsigned char* addScratchPadToGlobal_ele_S_Float;
extern "C" unsigned char* addScratchPadToGlobal_ele_S_Image;
extern "C" unsigned char* addScratchPadToGlobal_ele_S_Int32;
extern "C" unsigned char* addScratchPadToGlobal_ele_NodePtr;
extern "C" unsigned char* addScratchPadToGlobal_ele_S_Rotation;
extern "C" unsigned char* addScratchPadToGlobal_ele_S_String;
extern "C" unsigned char* addScratchPadToGlobal_ele_S_Time;
extern "C" unsigned char* addScratchPadToGlobal_ele_S_Vec2f;
extern "C" unsigned char* addScratchPadToGlobal_ele_S_Vec3f;
extern "C" unsigned char* addScratchPadToGlobal_ele_M_Bool;
extern "C" unsigned char* addScratchPadToGlobal_ele_M_Color;
extern "C" unsigned char* addScratchPadToGlobal_ele_M_Float;
extern "C" unsigned char* addScratchPadToGlobal_ele_M_Image;
extern "C" unsigned char* addScratchPadToGlobal_ele_M_Int32;
extern "C" unsigned char* addScratchPadToGlobal_ele_M_Node;
extern "C" unsigned char* addScratchPadToGlobal_ele_M_Rotation;
extern "C" unsigned char* addScratchPadToGlobal_ele_M_String;
extern "C" unsigned char* addScratchPadToGlobal_ele_M_Time;
extern "C" unsigned char* addScratchPadToGlobal_ele_M_Vec2f;
extern "C" unsigned char* addScratchPadToGlobal_ele_M_Vec3f;

template <class T> inline unsigned char** addScratchPadToGlobal_insert();
template <> inline unsigned char** addScratchPadToGlobal_insert<S_Bool>(){ return &addScratchPadToGlobal_ele_S_Bool;}
template <> inline unsigned char** addScratchPadToGlobal_insert<Color3f>(){ return &addScratchPadToGlobal_ele_S_Color;}
template <> inline unsigned char** addScratchPadToGlobal_insert<S_Float>(){ return &addScratchPadToGlobal_ele_S_Float;}
template <> inline unsigned char** addScratchPadToGlobal_insert<Image>(){ return &addScratchPadToGlobal_ele_S_Image;}
template <> inline unsigned char** addScratchPadToGlobal_insert<S_Int32>(){ return &addScratchPadToGlobal_ele_S_Int32;}
template <> inline unsigned char** addScratchPadToGlobal_insert<NodePtr>(){ return &addScratchPadToGlobal_ele_NodePtr;}
template <> inline unsigned char** addScratchPadToGlobal_insert<Rotation>(){ return &addScratchPadToGlobal_ele_S_Rotation;}
template <> inline unsigned char** addScratchPadToGlobal_insert<String>(){ return &addScratchPadToGlobal_ele_S_String;}
template <> inline unsigned char** addScratchPadToGlobal_insert<double>(){ return &addScratchPadToGlobal_ele_S_Time;}
template <> inline unsigned char** addScratchPadToGlobal_insert<Vec2f>(){ return &addScratchPadToGlobal_ele_S_Vec2f;}
template <> inline unsigned char** addScratchPadToGlobal_insert<Vec3f>(){ return &addScratchPadToGlobal_ele_S_Vec3f;}
template <> inline unsigned char** addScratchPadToGlobal_insert<M_Bool>(){ return &addScratchPadToGlobal_ele_M_Bool;}
template <> inline unsigned char** addScratchPadToGlobal_insert<M_Color3f>(){ return &addScratchPadToGlobal_ele_M_Color;}
template <> inline unsigned char** addScratchPadToGlobal_insert<M_Float>(){ return &addScratchPadToGlobal_ele_M_Float;}
template <> inline unsigned char** addScratchPadToGlobal_insert<M_Image>(){ return &addScratchPadToGlobal_ele_M_Image;}
template <> inline unsigned char** addScratchPadToGlobal_insert<M_Int>(){ return &addScratchPadToGlobal_ele_M_Int32;}
template <> inline unsigned char** addScratchPadToGlobal_insert<M_Node>(){ return &addScratchPadToGlobal_ele_M_Node;}
template <> inline unsigned char** addScratchPadToGlobal_insert<M_Rotation>(){ return &addScratchPadToGlobal_ele_M_Rotation;}
template <> inline unsigned char** addScratchPadToGlobal_insert<M_String>(){ return &addScratchPadToGlobal_ele_M_String;}
template <> inline unsigned char** addScratchPadToGlobal_insert<M_Double>(){ return &addScratchPadToGlobal_ele_M_Time;}
template <> inline unsigned char** addScratchPadToGlobal_insert<M_Vec2f>(){ return &addScratchPadToGlobal_ele_M_Vec2f;}
template <> inline unsigned char** addScratchPadToGlobal_insert<M_Vec3f>(){ return &addScratchPadToGlobal_ele_M_Vec3f;}

extern "C" unsigned char* CopyScratchPadToLocal_ele_S_Bool;
extern "C" unsigned char* CopyScratchPadToLocal_ele_S_Color;
extern "C" unsigned char* CopyScratchPadToLocal_ele_S_Float;
extern "C" unsigned char* CopyScratchPadToLocal_ele_S_Image;
extern "C" unsigned char* CopyScratchPadToLocal_ele_S_Int32;
extern "C" unsigned char* CopyScratchPadToLocal_ele_NodePtr;
extern "C" unsigned char* CopyScratchPadToLocal_ele_S_Rotation;
extern "C" unsigned char* CopyScratchPadToLocal_ele_S_String;
extern "C" unsigned char* CopyScratchPadToLocal_ele_S_Time;
extern "C" unsigned char* CopyScratchPadToLocal_ele_S_Vec2f;
extern "C" unsigned char* CopyScratchPadToLocal_ele_S_Vec3f;
extern "C" unsigned char* CopyScratchPadToLocal_ele_M_Bool;
extern "C" unsigned char* CopyScratchPadToLocal_ele_M_Color;
extern "C" unsigned char* CopyScratchPadToLocal_ele_M_Float;
extern "C" unsigned char* CopyScratchPadToLocal_ele_M_Image;
extern "C" unsigned char* CopyScratchPadToLocal_ele_M_Int32;
extern "C" unsigned char* CopyScratchPadToLocal_ele_M_Node;
extern "C" unsigned char* CopyScratchPadToLocal_ele_M_Rotation;
extern "C" unsigned char* CopyScratchPadToLocal_ele_M_String;
extern "C" unsigned char* CopyScratchPadToLocal_ele_M_Time;
extern "C" unsigned char* CopyScratchPadToLocal_ele_M_Vec2f;
extern "C" unsigned char* CopyScratchPadToLocal_ele_M_Vec3f;

template <class T> inline unsigned char** CopyScratchPadToLocal_insert();
template <> inline unsigned char** CopyScratchPadToLocal_insert<S_Bool>(){ return &CopyScratchPadToLocal_ele_S_Bool;}
template <> inline unsigned char** CopyScratchPadToLocal_insert<Color3f>(){ return &CopyScratchPadToLocal_ele_S_Color;}
template <> inline unsigned char** CopyScratchPadToLocal_insert<S_Float>(){ return &CopyScratchPadToLocal_ele_S_Float;}
template <> inline unsigned char** CopyScratchPadToLocal_insert<Image>(){ return &CopyScratchPadToLocal_ele_S_Image;}
template <> inline unsigned char** CopyScratchPadToLocal_insert<S_Int32>(){ return &CopyScratchPadToLocal_ele_S_Int32;}
template <> inline unsigned char** CopyScratchPadToLocal_insert<NodePtr>(){ return &CopyScratchPadToLocal_ele_NodePtr;}
template <> inline unsigned char** CopyScratchPadToLocal_insert<Rotation>(){ return &CopyScratchPadToLocal_ele_S_Rotation;}
template <> inline unsigned char** CopyScratchPadToLocal_insert<String>(){ return &CopyScratchPadToLocal_ele_S_String;}
template <> inline unsigned char** CopyScratchPadToLocal_insert<double>(){ return &CopyScratchPadToLocal_ele_S_Time;}
template <> inline unsigned char** CopyScratchPadToLocal_insert<Vec2f>(){ return &CopyScratchPadToLocal_ele_S_Vec2f;}
template <> inline unsigned char** CopyScratchPadToLocal_insert<Vec3f>(){ return &CopyScratchPadToLocal_ele_S_Vec3f;}
template <> inline unsigned char** CopyScratchPadToLocal_insert<M_Bool>(){ return &CopyScratchPadToLocal_ele_M_Bool;}
template <> inline unsigned char** CopyScratchPadToLocal_insert<M_Color3f>(){ return &CopyScratchPadToLocal_ele_M_Color;}
template <> inline unsigned char** CopyScratchPadToLocal_insert<M_Float>(){ return &CopyScratchPadToLocal_ele_M_Float;}
template <> inline unsigned char** CopyScratchPadToLocal_insert<M_Image>(){ return &CopyScratchPadToLocal_ele_M_Image;}
template <> inline unsigned char** CopyScratchPadToLocal_insert<M_Int>(){ return &CopyScratchPadToLocal_ele_M_Int32;}
template <> inline unsigned char** CopyScratchPadToLocal_insert<M_Node>(){ return &CopyScratchPadToLocal_ele_M_Node;}
template <> inline unsigned char** CopyScratchPadToLocal_insert<M_Rotation>(){ return &CopyScratchPadToLocal_ele_M_Rotation;}
template <> inline unsigned char** CopyScratchPadToLocal_insert<M_String>(){ return &CopyScratchPadToLocal_ele_M_String;}
template <> inline unsigned char** CopyScratchPadToLocal_insert<M_Double>(){ return &CopyScratchPadToLocal_ele_M_Time;}
template <> inline unsigned char** CopyScratchPadToLocal_insert<M_Vec2f>(){ return &CopyScratchPadToLocal_ele_M_Vec2f;}
template <> inline unsigned char** CopyScratchPadToLocal_insert<M_Vec3f>(){ return &CopyScratchPadToLocal_ele_M_Vec3f;}


extern "C" unsigned char* subScratchPadToLocal_ele_S_Bool;
extern "C" unsigned char* subScratchPadToLocal_ele_S_Color;
extern "C" unsigned char* subScratchPadToLocal_ele_S_Float;
extern "C" unsigned char* subScratchPadToLocal_ele_S_Image;
extern "C" unsigned char* subScratchPadToLocal_ele_S_Int32;
extern "C" unsigned char* subScratchPadToLocal_ele_NodePtr;
extern "C" unsigned char* subScratchPadToLocal_ele_S_Rotation;
extern "C" unsigned char* subScratchPadToLocal_ele_S_String;
extern "C" unsigned char* subScratchPadToLocal_ele_S_Time;
extern "C" unsigned char* subScratchPadToLocal_ele_S_Vec2f;
extern "C" unsigned char* subScratchPadToLocal_ele_S_Vec3f;
extern "C" unsigned char* subScratchPadToLocal_ele_M_Bool;
extern "C" unsigned char* subScratchPadToLocal_ele_M_Color;
extern "C" unsigned char* subScratchPadToLocal_ele_M_Float;
extern "C" unsigned char* subScratchPadToLocal_ele_M_Image;
extern "C" unsigned char* subScratchPadToLocal_ele_M_Int32;
extern "C" unsigned char* subScratchPadToLocal_ele_M_Node;
extern "C" unsigned char* subScratchPadToLocal_ele_M_Rotation;
extern "C" unsigned char* subScratchPadToLocal_ele_M_String;
extern "C" unsigned char* subScratchPadToLocal_ele_M_Time;
extern "C" unsigned char* subScratchPadToLocal_ele_M_Vec2f;
extern "C" unsigned char* subScratchPadToLocal_ele_M_Vec3f;

template <class T> inline unsigned char** subScratchPadToLocal_insert();
template <> inline unsigned char** subScratchPadToLocal_insert<S_Bool>(){ return &subScratchPadToLocal_ele_S_Bool;}
template <> inline unsigned char** subScratchPadToLocal_insert<Color3f>(){ return &subScratchPadToLocal_ele_S_Color;}
template <> inline unsigned char** subScratchPadToLocal_insert<S_Float>(){ return &subScratchPadToLocal_ele_S_Float;}
template <> inline unsigned char** subScratchPadToLocal_insert<Image>(){ return &subScratchPadToLocal_ele_S_Image;}
template <> inline unsigned char** subScratchPadToLocal_insert<S_Int32>(){ return &subScratchPadToLocal_ele_S_Int32;}
template <> inline unsigned char** subScratchPadToLocal_insert<NodePtr>(){ return &subScratchPadToLocal_ele_NodePtr;}
template <> inline unsigned char** subScratchPadToLocal_insert<Rotation>(){ return &subScratchPadToLocal_ele_S_Rotation;}
template <> inline unsigned char** subScratchPadToLocal_insert<String>(){ return &subScratchPadToLocal_ele_S_String;}
template <> inline unsigned char** subScratchPadToLocal_insert<double>(){ return &subScratchPadToLocal_ele_S_Time;}
template <> inline unsigned char** subScratchPadToLocal_insert<Vec2f>(){ return &subScratchPadToLocal_ele_S_Vec2f;}
template <> inline unsigned char** subScratchPadToLocal_insert<Vec3f>(){ return &subScratchPadToLocal_ele_S_Vec3f;}
template <> inline unsigned char** subScratchPadToLocal_insert<M_Bool>(){ return &subScratchPadToLocal_ele_M_Bool;}
template <> inline unsigned char** subScratchPadToLocal_insert<M_Color3f>(){ return &subScratchPadToLocal_ele_M_Color;}
template <> inline unsigned char** subScratchPadToLocal_insert<M_Float>(){ return &subScratchPadToLocal_ele_M_Float;}
template <> inline unsigned char** subScratchPadToLocal_insert<M_Image>(){ return &subScratchPadToLocal_ele_M_Image;}
template <> inline unsigned char** subScratchPadToLocal_insert<M_Int>(){ return &subScratchPadToLocal_ele_M_Int32;}
template <> inline unsigned char** subScratchPadToLocal_insert<M_Node>(){ return &subScratchPadToLocal_ele_M_Node;}
template <> inline unsigned char** subScratchPadToLocal_insert<M_Rotation>(){ return &subScratchPadToLocal_ele_M_Rotation;}
template <> inline unsigned char** subScratchPadToLocal_insert<M_String>(){ return &subScratchPadToLocal_ele_M_String;}
template <> inline unsigned char** subScratchPadToLocal_insert<M_Double>(){ return &subScratchPadToLocal_ele_M_Time;}
template <> inline unsigned char** subScratchPadToLocal_insert<M_Vec2f>(){ return &subScratchPadToLocal_ele_M_Vec2f;}
template <> inline unsigned char** subScratchPadToLocal_insert<M_Vec3f>(){ return &subScratchPadToLocal_ele_M_Vec3f;}


extern "C" unsigned char* addScratchPadToLocal_ele_S_Bool;
extern "C" unsigned char* addScratchPadToLocal_ele_S_Color;
extern "C" unsigned char* addScratchPadToLocal_ele_S_Float;
extern "C" unsigned char* addScratchPadToLocal_ele_S_Image;
extern "C" unsigned char* addScratchPadToLocal_ele_S_Int32;
extern "C" unsigned char* addScratchPadToLocal_ele_NodePtr;
extern "C" unsigned char* addScratchPadToLocal_ele_S_Rotation;
extern "C" unsigned char* addScratchPadToLocal_ele_S_String;
extern "C" unsigned char* addScratchPadToLocal_ele_S_Time;
extern "C" unsigned char* addScratchPadToLocal_ele_S_Vec2f;
extern "C" unsigned char* addScratchPadToLocal_ele_S_Vec3f;
extern "C" unsigned char* addScratchPadToLocal_ele_M_Bool;
extern "C" unsigned char* addScratchPadToLocal_ele_M_Color;
extern "C" unsigned char* addScratchPadToLocal_ele_M_Float;
extern "C" unsigned char* addScratchPadToLocal_ele_M_Image;
extern "C" unsigned char* addScratchPadToLocal_ele_M_Int32;
extern "C" unsigned char* addScratchPadToLocal_ele_M_Node;
extern "C" unsigned char* addScratchPadToLocal_ele_M_Rotation;
extern "C" unsigned char* addScratchPadToLocal_ele_M_String;
extern "C" unsigned char* addScratchPadToLocal_ele_M_Time;
extern "C" unsigned char* addScratchPadToLocal_ele_M_Vec2f;
extern "C" unsigned char* addScratchPadToLocal_ele_M_Vec3f;

template <class T> inline unsigned char** addScratchPadToLocal_insert();
template <> inline unsigned char** addScratchPadToLocal_insert<S_Bool>(){ return &addScratchPadToLocal_ele_S_Bool;}
template <> inline unsigned char** addScratchPadToLocal_insert<Color3f>(){ return &addScratchPadToLocal_ele_S_Color;}
template <> inline unsigned char** addScratchPadToLocal_insert<S_Float>(){ return &addScratchPadToLocal_ele_S_Float;}
template <> inline unsigned char** addScratchPadToLocal_insert<Image>(){ return &addScratchPadToLocal_ele_S_Image;}
template <> inline unsigned char** addScratchPadToLocal_insert<S_Int32>(){ return &addScratchPadToLocal_ele_S_Int32;}
template <> inline unsigned char** addScratchPadToLocal_insert<NodePtr>(){ return &addScratchPadToLocal_ele_NodePtr;}
template <> inline unsigned char** addScratchPadToLocal_insert<Rotation>(){ return &addScratchPadToLocal_ele_S_Rotation;}
template <> inline unsigned char** addScratchPadToLocal_insert<String>(){ return &addScratchPadToLocal_ele_S_String;}
template <> inline unsigned char** addScratchPadToLocal_insert<double>(){ return &addScratchPadToLocal_ele_S_Time;}
template <> inline unsigned char** addScratchPadToLocal_insert<Vec2f>(){ return &addScratchPadToLocal_ele_S_Vec2f;}
template <> inline unsigned char** addScratchPadToLocal_insert<Vec3f>(){ return &addScratchPadToLocal_ele_S_Vec3f;}
template <> inline unsigned char** addScratchPadToLocal_insert<M_Bool>(){ return &addScratchPadToLocal_ele_M_Bool;}
template <> inline unsigned char** addScratchPadToLocal_insert<M_Color3f>(){ return &addScratchPadToLocal_ele_M_Color;}
template <> inline unsigned char** addScratchPadToLocal_insert<M_Float>(){ return &addScratchPadToLocal_ele_M_Float;}
template <> inline unsigned char** addScratchPadToLocal_insert<M_Image>(){ return &addScratchPadToLocal_ele_M_Image;}
template <> inline unsigned char** addScratchPadToLocal_insert<M_Int>(){ return &addScratchPadToLocal_ele_M_Int32;}
template <> inline unsigned char** addScratchPadToLocal_insert<M_Node>(){ return &addScratchPadToLocal_ele_M_Node;}
template <> inline unsigned char** addScratchPadToLocal_insert<M_Rotation>(){ return &addScratchPadToLocal_ele_M_Rotation;}
template <> inline unsigned char** addScratchPadToLocal_insert<M_String>(){ return &addScratchPadToLocal_ele_M_String;}
template <> inline unsigned char** addScratchPadToLocal_insert<M_Double>(){ return &addScratchPadToLocal_ele_M_Time;}
template <> inline unsigned char** addScratchPadToLocal_insert<M_Vec2f>(){ return &addScratchPadToLocal_ele_M_Vec2f;}
template <> inline unsigned char** addScratchPadToLocal_insert<M_Vec3f>(){ return &addScratchPadToLocal_ele_M_Vec3f;}


extern "C" unsigned char* copyGlobalToScratchPad_ele_S_Bool;
extern "C" unsigned char* copyGlobalToScratchPad_ele_S_Color;
extern "C" unsigned char* copyGlobalToScratchPad_ele_S_Float;
extern "C" unsigned char* copyGlobalToScratchPad_ele_S_Image;
extern "C" unsigned char* copyGlobalToScratchPad_ele_S_Int32;
extern "C" unsigned char* copyGlobalToScratchPad_ele_NodePtr;
extern "C" unsigned char* copyGlobalToScratchPad_ele_S_Rotation;
extern "C" unsigned char* copyGlobalToScratchPad_ele_S_String;
extern "C" unsigned char* copyGlobalToScratchPad_ele_S_Time;
extern "C" unsigned char* copyGlobalToScratchPad_ele_S_Vec2f;
extern "C" unsigned char* copyGlobalToScratchPad_ele_S_Vec3f;
extern "C" unsigned char* copyGlobalToScratchPad_ele_M_Bool;
extern "C" unsigned char* copyGlobalToScratchPad_ele_M_Color;
extern "C" unsigned char* copyGlobalToScratchPad_ele_M_Float;
extern "C" unsigned char* copyGlobalToScratchPad_ele_M_Image;
extern "C" unsigned char* copyGlobalToScratchPad_ele_M_Int32;
extern "C" unsigned char* copyGlobalToScratchPad_ele_M_Node;
extern "C" unsigned char* copyGlobalToScratchPad_ele_M_Rotation;
extern "C" unsigned char* copyGlobalToScratchPad_ele_M_String;
extern "C" unsigned char* copyGlobalToScratchPad_ele_M_Time;
extern "C" unsigned char* copyGlobalToScratchPad_ele_M_Vec2f;
extern "C" unsigned char* copyGlobalToScratchPad_ele_M_Vec3f;

template <class T> inline unsigned char** copyGlobalToScratchPad_insert();
template <> inline unsigned char** copyGlobalToScratchPad_insert<S_Bool>(){ return &copyGlobalToScratchPad_ele_S_Bool;}
template <> inline unsigned char** copyGlobalToScratchPad_insert<Color3f>(){ return &copyGlobalToScratchPad_ele_S_Color;}
template <> inline unsigned char** copyGlobalToScratchPad_insert<S_Float>(){ return &copyGlobalToScratchPad_ele_S_Float;}
template <> inline unsigned char** copyGlobalToScratchPad_insert<Image>(){ return &copyGlobalToScratchPad_ele_S_Image;}
template <> inline unsigned char** copyGlobalToScratchPad_insert<S_Int32>(){ return &copyGlobalToScratchPad_ele_S_Int32;}
template <> inline unsigned char** copyGlobalToScratchPad_insert<NodePtr>(){ return &copyGlobalToScratchPad_ele_NodePtr;}
template <> inline unsigned char** copyGlobalToScratchPad_insert<Rotation>(){ return &copyGlobalToScratchPad_ele_S_Rotation;}
template <> inline unsigned char** copyGlobalToScratchPad_insert<String>(){ return &copyGlobalToScratchPad_ele_S_String;}
template <> inline unsigned char** copyGlobalToScratchPad_insert<double>(){ return &copyGlobalToScratchPad_ele_S_Time;}
template <> inline unsigned char** copyGlobalToScratchPad_insert<Vec2f>(){ return &copyGlobalToScratchPad_ele_S_Vec2f;}
template <> inline unsigned char** copyGlobalToScratchPad_insert<Vec3f>(){ return &copyGlobalToScratchPad_ele_S_Vec3f;}
template <> inline unsigned char** copyGlobalToScratchPad_insert<M_Bool>(){ return &copyGlobalToScratchPad_ele_M_Bool;}
template <> inline unsigned char** copyGlobalToScratchPad_insert<M_Color3f>(){ return &copyGlobalToScratchPad_ele_M_Color;}
template <> inline unsigned char** copyGlobalToScratchPad_insert<M_Float>(){ return &copyGlobalToScratchPad_ele_M_Float;}
template <> inline unsigned char** copyGlobalToScratchPad_insert<M_Image>(){ return &copyGlobalToScratchPad_ele_M_Image;}
template <> inline unsigned char** copyGlobalToScratchPad_insert<M_Int>(){ return &copyGlobalToScratchPad_ele_M_Int32;}
template <> inline unsigned char** copyGlobalToScratchPad_insert<M_Node>(){ return &copyGlobalToScratchPad_ele_M_Node;}
template <> inline unsigned char** copyGlobalToScratchPad_insert<M_Rotation>(){ return &copyGlobalToScratchPad_ele_M_Rotation;}
template <> inline unsigned char** copyGlobalToScratchPad_insert<M_String>(){ return &copyGlobalToScratchPad_ele_M_String;}
template <> inline unsigned char** copyGlobalToScratchPad_insert<M_Double>(){ return &copyGlobalToScratchPad_ele_M_Time;}
template <> inline unsigned char** copyGlobalToScratchPad_insert<M_Vec2f>(){ return &copyGlobalToScratchPad_ele_M_Vec2f;}
template <> inline unsigned char** copyGlobalToScratchPad_insert<M_Vec3f>(){ return &copyGlobalToScratchPad_ele_M_Vec3f;}


extern "C" unsigned char* addAssignElement2ValDirect_ele_S_Bool;
extern "C" unsigned char* addAssignElement2ValDirect_ele_S_Color;
extern "C" unsigned char* addAssignElement2ValDirect_ele_S_Float;
extern "C" unsigned char* addAssignElement2ValDirect_ele_S_Image;
extern "C" unsigned char* addAssignElement2ValDirect_ele_S_Int32;
extern "C" unsigned char* addAssignElement2ValDirect_ele_NodePtr;
extern "C" unsigned char* addAssignElement2ValDirect_ele_S_Rotation;
extern "C" unsigned char* addAssignElement2ValDirect_ele_S_String;
extern "C" unsigned char* addAssignElement2ValDirect_ele_S_Time;
extern "C" unsigned char* addAssignElement2ValDirect_ele_S_Vec2f;
extern "C" unsigned char* addAssignElement2ValDirect_ele_S_Vec3f;
extern "C" unsigned char* addAssignElement2ValDirect_ele_M_Bool;
extern "C" unsigned char* addAssignElement2ValDirect_ele_M_Color;
extern "C" unsigned char* addAssignElement2ValDirect_ele_M_Float;
extern "C" unsigned char* addAssignElement2ValDirect_ele_M_Image;
extern "C" unsigned char* addAssignElement2ValDirect_ele_M_Int32;
extern "C" unsigned char* addAssignElement2ValDirect_ele_M_Node;
extern "C" unsigned char* addAssignElement2ValDirect_ele_M_Rotation;
extern "C" unsigned char* addAssignElement2ValDirect_ele_M_String;
extern "C" unsigned char* addAssignElement2ValDirect_ele_M_Time;
extern "C" unsigned char* addAssignElement2ValDirect_ele_M_Vec2f;
extern "C" unsigned char* addAssignElement2ValDirect_ele_M_Vec3f;

template <class T> inline unsigned char** addAssignElement2ValDirect_insert();
template <> inline unsigned char** addAssignElement2ValDirect_insert<S_Bool>(){ return &addAssignElement2ValDirect_ele_S_Bool;}
template <> inline unsigned char** addAssignElement2ValDirect_insert<Color3f>(){ return &addAssignElement2ValDirect_ele_S_Color;}
template <> inline unsigned char** addAssignElement2ValDirect_insert<S_Float>(){ return &addAssignElement2ValDirect_ele_S_Float;}
template <> inline unsigned char** addAssignElement2ValDirect_insert<Image>(){ return &addAssignElement2ValDirect_ele_S_Image;}
template <> inline unsigned char** addAssignElement2ValDirect_insert<S_Int32>(){ return &addAssignElement2ValDirect_ele_S_Int32;}
template <> inline unsigned char** addAssignElement2ValDirect_insert<NodePtr>(){ return &addAssignElement2ValDirect_ele_NodePtr;}
template <> inline unsigned char** addAssignElement2ValDirect_insert<Rotation>(){ return &addAssignElement2ValDirect_ele_S_Rotation;}
template <> inline unsigned char** addAssignElement2ValDirect_insert<String>(){ return &addAssignElement2ValDirect_ele_S_String;}
template <> inline unsigned char** addAssignElement2ValDirect_insert<double>(){ return &addAssignElement2ValDirect_ele_S_Time;}
template <> inline unsigned char** addAssignElement2ValDirect_insert<Vec2f>(){ return &addAssignElement2ValDirect_ele_S_Vec2f;}
template <> inline unsigned char** addAssignElement2ValDirect_insert<Vec3f>(){ return &addAssignElement2ValDirect_ele_S_Vec3f;}
template <> inline unsigned char** addAssignElement2ValDirect_insert<M_Bool>(){ return &addAssignElement2ValDirect_ele_M_Bool;}
template <> inline unsigned char** addAssignElement2ValDirect_insert<M_Color3f>(){ return &addAssignElement2ValDirect_ele_M_Color;}
template <> inline unsigned char** addAssignElement2ValDirect_insert<M_Float>(){ return &addAssignElement2ValDirect_ele_M_Float;}
template <> inline unsigned char** addAssignElement2ValDirect_insert<M_Image>(){ return &addAssignElement2ValDirect_ele_M_Image;}
template <> inline unsigned char** addAssignElement2ValDirect_insert<M_Int>(){ return &addAssignElement2ValDirect_ele_M_Int32;}
template <> inline unsigned char** addAssignElement2ValDirect_insert<M_Node>(){ return &addAssignElement2ValDirect_ele_M_Node;}
template <> inline unsigned char** addAssignElement2ValDirect_insert<M_Rotation>(){ return &addAssignElement2ValDirect_ele_M_Rotation;}
template <> inline unsigned char** addAssignElement2ValDirect_insert<M_String>(){ return &addAssignElement2ValDirect_ele_M_String;}
template <> inline unsigned char** addAssignElement2ValDirect_insert<M_Double>(){ return &addAssignElement2ValDirect_ele_M_Time;}
template <> inline unsigned char** addAssignElement2ValDirect_insert<M_Vec2f>(){ return &addAssignElement2ValDirect_ele_M_Vec2f;}
template <> inline unsigned char** addAssignElement2ValDirect_insert<M_Vec3f>(){ return &addAssignElement2ValDirect_ele_M_Vec3f;}


extern "C" unsigned char* CopyLocalToScratchPad_ele_S_Bool;
extern "C" unsigned char* CopyLocalToScratchPad_ele_S_Color;
extern "C" unsigned char* CopyLocalToScratchPad_ele_S_Float;
extern "C" unsigned char* CopyLocalToScratchPad_ele_S_Image;
extern "C" unsigned char* CopyLocalToScratchPad_ele_S_Int32;
extern "C" unsigned char* CopyLocalToScratchPad_ele_NodePtr;
extern "C" unsigned char* CopyLocalToScratchPad_ele_S_Rotation;
extern "C" unsigned char* CopyLocalToScratchPad_ele_S_String;
extern "C" unsigned char* CopyLocalToScratchPad_ele_S_Time;
extern "C" unsigned char* CopyLocalToScratchPad_ele_S_Vec2f;
extern "C" unsigned char* CopyLocalToScratchPad_ele_S_Vec3f;
extern "C" unsigned char* CopyLocalToScratchPad_ele_M_Bool;
extern "C" unsigned char* CopyLocalToScratchPad_ele_M_Color;
extern "C" unsigned char* CopyLocalToScratchPad_ele_M_Float;
extern "C" unsigned char* CopyLocalToScratchPad_ele_M_Image;
extern "C" unsigned char* CopyLocalToScratchPad_ele_M_Int32;
extern "C" unsigned char* CopyLocalToScratchPad_ele_M_Node;
extern "C" unsigned char* CopyLocalToScratchPad_ele_M_Rotation;
extern "C" unsigned char* CopyLocalToScratchPad_ele_M_String;
extern "C" unsigned char* CopyLocalToScratchPad_ele_M_Time;
extern "C" unsigned char* CopyLocalToScratchPad_ele_M_Vec2f;
extern "C" unsigned char* CopyLocalToScratchPad_ele_M_Vec3f;

template <class T> inline unsigned char** CopyLocalToScratchPad_insert();
template <> inline unsigned char** CopyLocalToScratchPad_insert<S_Bool>(){ return &CopyLocalToScratchPad_ele_S_Bool;}
template <> inline unsigned char** CopyLocalToScratchPad_insert<Color3f>(){ return &CopyLocalToScratchPad_ele_S_Color;}
template <> inline unsigned char** CopyLocalToScratchPad_insert<S_Float>(){ return &CopyLocalToScratchPad_ele_S_Float;}
template <> inline unsigned char** CopyLocalToScratchPad_insert<Image>(){ return &CopyLocalToScratchPad_ele_S_Image;}
template <> inline unsigned char** CopyLocalToScratchPad_insert<S_Int32>(){ return &CopyLocalToScratchPad_ele_S_Int32;}
template <> inline unsigned char** CopyLocalToScratchPad_insert<NodePtr>(){ return &CopyLocalToScratchPad_ele_NodePtr;}
template <> inline unsigned char** CopyLocalToScratchPad_insert<Rotation>(){ return &CopyLocalToScratchPad_ele_S_Rotation;}
template <> inline unsigned char** CopyLocalToScratchPad_insert<String>(){ return &CopyLocalToScratchPad_ele_S_String;}
template <> inline unsigned char** CopyLocalToScratchPad_insert<double>(){ return &CopyLocalToScratchPad_ele_S_Time;}
template <> inline unsigned char** CopyLocalToScratchPad_insert<Vec2f>(){ return &CopyLocalToScratchPad_ele_S_Vec2f;}
template <> inline unsigned char** CopyLocalToScratchPad_insert<Vec3f>(){ return &CopyLocalToScratchPad_ele_S_Vec3f;}
template <> inline unsigned char** CopyLocalToScratchPad_insert<M_Bool>(){ return &CopyLocalToScratchPad_ele_M_Bool;}
template <> inline unsigned char** CopyLocalToScratchPad_insert<M_Color3f>(){ return &CopyLocalToScratchPad_ele_M_Color;}
template <> inline unsigned char** CopyLocalToScratchPad_insert<M_Float>(){ return &CopyLocalToScratchPad_ele_M_Float;}
template <> inline unsigned char** CopyLocalToScratchPad_insert<M_Image>(){ return &CopyLocalToScratchPad_ele_M_Image;}
template <> inline unsigned char** CopyLocalToScratchPad_insert<M_Int>(){ return &CopyLocalToScratchPad_ele_M_Int32;}
template <> inline unsigned char** CopyLocalToScratchPad_insert<M_Node>(){ return &CopyLocalToScratchPad_ele_M_Node;}
template <> inline unsigned char** CopyLocalToScratchPad_insert<M_Rotation>(){ return &CopyLocalToScratchPad_ele_M_Rotation;}
template <> inline unsigned char** CopyLocalToScratchPad_insert<M_String>(){ return &CopyLocalToScratchPad_ele_M_String;}
template <> inline unsigned char** CopyLocalToScratchPad_insert<M_Double>(){ return &CopyLocalToScratchPad_ele_M_Time;}
template <> inline unsigned char** CopyLocalToScratchPad_insert<M_Vec2f>(){ return &CopyLocalToScratchPad_ele_M_Vec2f;}
template <> inline unsigned char** CopyLocalToScratchPad_insert<M_Vec3f>(){ return &CopyLocalToScratchPad_ele_M_Vec3f;}


extern "C" unsigned char* CopyImmediateToScratchPad_ele_S_Bool;
extern "C" unsigned char* CopyImmediateToScratchPad_ele_S_Color;
extern "C" unsigned char* CopyImmediateToScratchPad_ele_S_Float;
extern "C" unsigned char* CopyImmediateToScratchPad_ele_S_Image;
extern "C" unsigned char* CopyImmediateToScratchPad_ele_S_Int32;
extern "C" unsigned char* CopyImmediateToScratchPad_ele_NodePtr;
extern "C" unsigned char* CopyImmediateToScratchPad_ele_S_Rotation;
extern "C" unsigned char* CopyImmediateToScratchPad_ele_S_String;
extern "C" unsigned char* CopyImmediateToScratchPad_ele_S_Time;
extern "C" unsigned char* CopyImmediateToScratchPad_ele_S_Vec2f;
extern "C" unsigned char* CopyImmediateToScratchPad_ele_S_Vec3f;
extern "C" unsigned char* CopyImmediateToScratchPad_ele_M_Bool;
extern "C" unsigned char* CopyImmediateToScratchPad_ele_M_Color;
extern "C" unsigned char* CopyImmediateToScratchPad_ele_M_Float;
extern "C" unsigned char* CopyImmediateToScratchPad_ele_M_Image;
extern "C" unsigned char* CopyImmediateToScratchPad_ele_M_Int32;
extern "C" unsigned char* CopyImmediateToScratchPad_ele_M_Node;
extern "C" unsigned char* CopyImmediateToScratchPad_ele_M_Rotation;
extern "C" unsigned char* CopyImmediateToScratchPad_ele_M_String;
extern "C" unsigned char* CopyImmediateToScratchPad_ele_M_Time;
extern "C" unsigned char* CopyImmediateToScratchPad_ele_M_Vec2f;
extern "C" unsigned char* CopyImmediateToScratchPad_ele_M_Vec3f;

extern "C" unsigned char* subtraction_ele_S_Bool;
extern "C" unsigned char* subtraction_ele_S_Color;
extern "C" unsigned char* subtraction_ele_S_Float;
extern "C" unsigned char* subtraction_ele_S_Image;
extern "C" unsigned char* subtraction_ele_S_Int32;
extern "C" unsigned char* subtraction_ele_NodePtr;
extern "C" unsigned char* subtraction_ele_S_Rotation;
extern "C" unsigned char* subtraction_ele_S_String;
extern "C" unsigned char* subtraction_ele_S_Time;
extern "C" unsigned char* subtraction_ele_S_Vec2f;
extern "C" unsigned char* subtraction_ele_S_Vec3f;
extern "C" unsigned char* subtraction_ele_M_Bool;
extern "C" unsigned char* subtraction_ele_M_Color;
extern "C" unsigned char* subtraction_ele_M_Float;
extern "C" unsigned char* subtraction_ele_M_Image;
extern "C" unsigned char* subtraction_ele_M_Int32;
extern "C" unsigned char* subtraction_ele_M_Node;
extern "C" unsigned char* subtraction_ele_M_Rotation;
extern "C" unsigned char* subtraction_ele_M_String;
extern "C" unsigned char* subtraction_ele_M_Time;
extern "C" unsigned char* subtraction_ele_M_Vec2f;
extern "C" unsigned char* subtraction_ele_M_Vec3f;

template <class T> inline unsigned char** subtraction_insert();
template <> inline unsigned char** subtraction_insert<S_Bool>(){ return &subtraction_ele_S_Bool;}
template <> inline unsigned char** subtraction_insert<Color3f>(){ return &subtraction_ele_S_Color;}
template <> inline unsigned char** subtraction_insert<S_Float>(){ return &subtraction_ele_S_Float;}
template <> inline unsigned char** subtraction_insert<Image>(){ return &subtraction_ele_S_Image;}
template <> inline unsigned char** subtraction_insert<S_Int32>(){ return &subtraction_ele_S_Int32;}
template <> inline unsigned char** subtraction_insert<NodePtr>(){ return &subtraction_ele_NodePtr;}
template <> inline unsigned char** subtraction_insert<Rotation>(){ return &subtraction_ele_S_Rotation;}
template <> inline unsigned char** subtraction_insert<String>(){ return &subtraction_ele_S_String;}
template <> inline unsigned char** subtraction_insert<double>(){ return &subtraction_ele_S_Time;}
template <> inline unsigned char** subtraction_insert<Vec2f>(){ return &subtraction_ele_S_Vec2f;}
template <> inline unsigned char** subtraction_insert<Vec3f>(){ return &subtraction_ele_S_Vec3f;}
template <> inline unsigned char** subtraction_insert<M_Bool>(){ return &subtraction_ele_M_Bool;}
template <> inline unsigned char** subtraction_insert<M_Color3f>(){ return &subtraction_ele_M_Color;}
template <> inline unsigned char** subtraction_insert<M_Float>(){ return &subtraction_ele_M_Float;}
template <> inline unsigned char** subtraction_insert<M_Image>(){ return &subtraction_ele_M_Image;}
template <> inline unsigned char** subtraction_insert<M_Int>(){ return &subtraction_ele_M_Int32;}
template <> inline unsigned char** subtraction_insert<M_Node>(){ return &subtraction_ele_M_Node;}
template <> inline unsigned char** subtraction_insert<M_Rotation>(){ return &subtraction_ele_M_Rotation;}
template <> inline unsigned char** subtraction_insert<M_String>(){ return &subtraction_ele_M_String;}
template <> inline unsigned char** subtraction_insert<M_Double>(){ return &subtraction_ele_M_Time;}
template <> inline unsigned char** subtraction_insert<M_Vec2f>(){ return &subtraction_ele_M_Vec2f;}
template <> inline unsigned char** subtraction_insert<M_Vec3f>(){ return &subtraction_ele_M_Vec3f;}

extern "C" unsigned char* addition_ele_S_Bool;
extern "C" unsigned char* addition_ele_S_Color;
extern "C" unsigned char* addition_ele_S_Float;
extern "C" unsigned char* addition_ele_S_Image;
extern "C" unsigned char* addition_ele_S_Int32;
extern "C" unsigned char* addition_ele_NodePtr;
extern "C" unsigned char* addition_ele_S_Rotation;
extern "C" unsigned char* addition_ele_S_String;
extern "C" unsigned char* addition_ele_S_Time;
extern "C" unsigned char* addition_ele_S_Vec2f;
extern "C" unsigned char* addition_ele_S_Vec3f;
extern "C" unsigned char* addition_ele_M_Bool;
extern "C" unsigned char* addition_ele_M_Color;
extern "C" unsigned char* addition_ele_M_Float;
extern "C" unsigned char* addition_ele_M_Image;
extern "C" unsigned char* addition_ele_M_Int32;
extern "C" unsigned char* addition_ele_M_Node;
extern "C" unsigned char* addition_ele_M_Rotation;
extern "C" unsigned char* addition_ele_M_String;
extern "C" unsigned char* addition_ele_M_Time;
extern "C" unsigned char* addition_ele_M_Vec2f;
extern "C" unsigned char* addition_ele_M_Vec3f;

template <class T> inline unsigned char** addition_insert();
template <> inline unsigned char** addition_insert<S_Bool>(){ return &addition_ele_S_Bool;}
template <> inline unsigned char** addition_insert<Color3f>(){ return &addition_ele_S_Color;}
template <> inline unsigned char** addition_insert<S_Float>(){ return &addition_ele_S_Float;}
template <> inline unsigned char** addition_insert<Image>(){ return &addition_ele_S_Image;}
template <> inline unsigned char** addition_insert<S_Int32>(){ return &addition_ele_S_Int32;}
template <> inline unsigned char** addition_insert<NodePtr>(){ return &addition_ele_NodePtr;}
template <> inline unsigned char** addition_insert<Rotation>(){ return &addition_ele_S_Rotation;}
template <> inline unsigned char** addition_insert<String>(){ return &addition_ele_S_String;}
template <> inline unsigned char** addition_insert<double>(){ return &addition_ele_S_Time;}
template <> inline unsigned char** addition_insert<Vec2f>(){ return &addition_ele_S_Vec2f;}
template <> inline unsigned char** addition_insert<Vec3f>(){ return &addition_ele_S_Vec3f;}
template <> inline unsigned char** addition_insert<M_Bool>(){ return &addition_ele_M_Bool;}
template <> inline unsigned char** addition_insert<M_Color3f>(){ return &addition_ele_M_Color;}
template <> inline unsigned char** addition_insert<M_Float>(){ return &addition_ele_M_Float;}
template <> inline unsigned char** addition_insert<M_Image>(){ return &addition_ele_M_Image;}
template <> inline unsigned char** addition_insert<M_Int>(){ return &addition_ele_M_Int32;}
template <> inline unsigned char** addition_insert<M_Node>(){ return &addition_ele_M_Node;}
template <> inline unsigned char** addition_insert<M_Rotation>(){ return &addition_ele_M_Rotation;}
template <> inline unsigned char** addition_insert<M_String>(){ return &addition_ele_M_String;}
template <> inline unsigned char** addition_insert<M_Double>(){ return &addition_ele_M_Time;}
template <> inline unsigned char** addition_insert<M_Vec2f>(){ return &addition_ele_M_Vec2f;}
template <> inline unsigned char** addition_insert<M_Vec3f>(){ return &addition_ele_M_Vec3f;}

extern "C" unsigned char* division_ele_S_Bool;
extern "C" unsigned char* division_ele_S_Color;
extern "C" unsigned char* division_ele_S_Float;
extern "C" unsigned char* division_ele_S_Image;
extern "C" unsigned char* division_ele_S_Int32;
extern "C" unsigned char* division_ele_NodePtr;
extern "C" unsigned char* division_ele_S_Rotation;
extern "C" unsigned char* division_ele_S_String;
extern "C" unsigned char* division_ele_S_Time;
extern "C" unsigned char* division_ele_S_Vec2f;
extern "C" unsigned char* division_ele_S_Vec3f;
extern "C" unsigned char* division_ele_M_Bool;
extern "C" unsigned char* division_ele_M_Color;
extern "C" unsigned char* division_ele_M_Float;
extern "C" unsigned char* division_ele_M_Image;
extern "C" unsigned char* division_ele_M_Int32;
extern "C" unsigned char* division_ele_M_Node;
extern "C" unsigned char* division_ele_M_Rotation;
extern "C" unsigned char* division_ele_M_String;
extern "C" unsigned char* division_ele_M_Time;
extern "C" unsigned char* division_ele_M_Vec2f;
extern "C" unsigned char* division_ele_M_Vec3f;

template <class T> inline unsigned char** division_insert();
template <> inline unsigned char** division_insert<S_Bool>(){ return &division_ele_S_Bool;}
template <> inline unsigned char** division_insert<Color3f>(){ return &division_ele_S_Color;}
template <> inline unsigned char** division_insert<S_Float>(){ return &division_ele_S_Float;}
template <> inline unsigned char** division_insert<Image>(){ return &division_ele_S_Image;}
template <> inline unsigned char** division_insert<S_Int32>(){ return &division_ele_S_Int32;}
template <> inline unsigned char** division_insert<NodePtr>(){ return &division_ele_NodePtr;}
template <> inline unsigned char** division_insert<Rotation>(){ return &division_ele_S_Rotation;}
template <> inline unsigned char** division_insert<String>(){ return &division_ele_S_String;}
template <> inline unsigned char** division_insert<double>(){ return &division_ele_S_Time;}
template <> inline unsigned char** division_insert<Vec2f>(){ return &division_ele_S_Vec2f;}
template <> inline unsigned char** division_insert<Vec3f>(){ return &division_ele_S_Vec3f;}
template <> inline unsigned char** division_insert<M_Bool>(){ return &division_ele_M_Bool;}
template <> inline unsigned char** division_insert<M_Color3f>(){ return &division_ele_M_Color;}
template <> inline unsigned char** division_insert<M_Float>(){ return &division_ele_M_Float;}
template <> inline unsigned char** division_insert<M_Image>(){ return &division_ele_M_Image;}
template <> inline unsigned char** division_insert<M_Int>(){ return &division_ele_M_Int32;}
template <> inline unsigned char** division_insert<M_Node>(){ return &division_ele_M_Node;}
template <> inline unsigned char** division_insert<M_Rotation>(){ return &division_ele_M_Rotation;}
template <> inline unsigned char** division_insert<M_String>(){ return &division_ele_M_String;}
template <> inline unsigned char** division_insert<M_Double>(){ return &division_ele_M_Time;}
template <> inline unsigned char** division_insert<M_Vec2f>(){ return &division_ele_M_Vec2f;}
template <> inline unsigned char** division_insert<M_Vec3f>(){ return &division_ele_M_Vec3f;}

extern "C" unsigned char* modulo_ele_Int;

extern "C" unsigned char* multiplication_ele_S_Bool;
extern "C" unsigned char* multiplication_ele_S_Color;
extern "C" unsigned char* multiplication_ele_S_Float;
extern "C" unsigned char* multiplication_ele_S_Image;
extern "C" unsigned char* multiplication_ele_S_Int32;
extern "C" unsigned char* multiplication_ele_NodePtr;
extern "C" unsigned char* multiplication_ele_S_Rotation;
extern "C" unsigned char* multiplication_ele_S_String;
extern "C" unsigned char* multiplication_ele_S_Time;
extern "C" unsigned char* multiplication_ele_S_Vec2f;
extern "C" unsigned char* multiplication_ele_S_Vec3f;
extern "C" unsigned char* multiplication_ele_M_Bool;
extern "C" unsigned char* multiplication_ele_M_Color;
extern "C" unsigned char* multiplication_ele_M_Float;
extern "C" unsigned char* multiplication_ele_M_Image;
extern "C" unsigned char* multiplication_ele_M_Int32;
extern "C" unsigned char* multiplication_ele_M_Node;
extern "C" unsigned char* multiplication_ele_M_Rotation;
extern "C" unsigned char* multiplication_ele_M_String;
extern "C" unsigned char* multiplication_ele_M_Time;
extern "C" unsigned char* multiplication_ele_M_Vec2f;
extern "C" unsigned char* multiplication_ele_M_Vec3f;

template <class T> inline unsigned char** multiplication_insert();
template <> inline unsigned char** multiplication_insert<S_Bool>(){ return &multiplication_ele_S_Bool;}
template <> inline unsigned char** multiplication_insert<Color3f>(){ return &multiplication_ele_S_Color;}
template <> inline unsigned char** multiplication_insert<S_Float>(){ return &multiplication_ele_S_Float;}
template <> inline unsigned char** multiplication_insert<Image>(){ return &multiplication_ele_S_Image;}
template <> inline unsigned char** multiplication_insert<S_Int32>(){ return &multiplication_ele_S_Int32;}
template <> inline unsigned char** multiplication_insert<NodePtr>(){ return &multiplication_ele_NodePtr;}
template <> inline unsigned char** multiplication_insert<Rotation>(){ return &multiplication_ele_S_Rotation;}
template <> inline unsigned char** multiplication_insert<String>(){ return &multiplication_ele_S_String;}
template <> inline unsigned char** multiplication_insert<double>(){ return &multiplication_ele_S_Time;}
template <> inline unsigned char** multiplication_insert<Vec2f>(){ return &multiplication_ele_S_Vec2f;}
template <> inline unsigned char** multiplication_insert<Vec3f>(){ return &multiplication_ele_S_Vec3f;}
template <> inline unsigned char** multiplication_insert<M_Bool>(){ return &multiplication_ele_M_Bool;}
template <> inline unsigned char** multiplication_insert<M_Color3f>(){ return &multiplication_ele_M_Color;}
template <> inline unsigned char** multiplication_insert<M_Float>(){ return &multiplication_ele_M_Float;}
template <> inline unsigned char** multiplication_insert<M_Image>(){ return &multiplication_ele_M_Image;}
template <> inline unsigned char** multiplication_insert<M_Int>(){ return &multiplication_ele_M_Int32;}
template <> inline unsigned char** multiplication_insert<M_Node>(){ return &multiplication_ele_M_Node;}
template <> inline unsigned char** multiplication_insert<M_Rotation>(){ return &multiplication_ele_M_Rotation;}
template <> inline unsigned char** multiplication_insert<M_String>(){ return &multiplication_ele_M_String;}
template <> inline unsigned char** multiplication_insert<M_Double>(){ return &multiplication_ele_M_Time;}
template <> inline unsigned char** multiplication_insert<M_Vec2f>(){ return &multiplication_ele_M_Vec2f;}
template <> inline unsigned char** multiplication_insert<M_Vec3f>(){ return &multiplication_ele_M_Vec3f;}

extern "C" unsigned char* preDecLocal_ele_S_Bool;
extern "C" unsigned char* preDecLocal_ele_S_Color;
extern "C" unsigned char* preDecLocal_ele_S_Float;
extern "C" unsigned char* preDecLocal_ele_S_Image;
extern "C" unsigned char* preDecLocal_ele_S_Int32;
extern "C" unsigned char* preDecLocal_ele_NodePtr;
extern "C" unsigned char* preDecLocal_ele_S_Rotation;
extern "C" unsigned char* preDecLocal_ele_S_String;
extern "C" unsigned char* preDecLocal_ele_S_Time;
extern "C" unsigned char* preDecLocal_ele_S_Vec2f;
extern "C" unsigned char* preDecLocal_ele_S_Vec3f;
extern "C" unsigned char* preDecLocal_ele_M_Bool;
extern "C" unsigned char* preDecLocal_ele_M_Color;
extern "C" unsigned char* preDecLocal_ele_M_Float;
extern "C" unsigned char* preDecLocal_ele_M_Image;
extern "C" unsigned char* preDecLocal_ele_M_Int32;
extern "C" unsigned char* preDecLocal_ele_M_Node;
extern "C" unsigned char* preDecLocal_ele_M_Rotation;
extern "C" unsigned char* preDecLocal_ele_M_String;
extern "C" unsigned char* preDecLocal_ele_M_Time;
extern "C" unsigned char* preDecLocal_ele_M_Vec2f;
extern "C" unsigned char* preDecLocal_ele_M_Vec3f;

template <class T> inline unsigned char** preDecLocal_insert();
template <> inline unsigned char** preDecLocal_insert<S_Bool>(){ return &preDecLocal_ele_S_Bool;}
template <> inline unsigned char** preDecLocal_insert<Color3f>(){ return &preDecLocal_ele_S_Color;}
template <> inline unsigned char** preDecLocal_insert<S_Float>(){ return &preDecLocal_ele_S_Float;}
template <> inline unsigned char** preDecLocal_insert<Image>(){ return &preDecLocal_ele_S_Image;}
template <> inline unsigned char** preDecLocal_insert<S_Int32>(){ return &preDecLocal_ele_S_Int32;}
template <> inline unsigned char** preDecLocal_insert<NodePtr>(){ return &preDecLocal_ele_NodePtr;}
template <> inline unsigned char** preDecLocal_insert<Rotation>(){ return &preDecLocal_ele_S_Rotation;}
template <> inline unsigned char** preDecLocal_insert<String>(){ return &preDecLocal_ele_S_String;}
template <> inline unsigned char** preDecLocal_insert<double>(){ return &preDecLocal_ele_S_Time;}
template <> inline unsigned char** preDecLocal_insert<Vec2f>(){ return &preDecLocal_ele_S_Vec2f;}
template <> inline unsigned char** preDecLocal_insert<Vec3f>(){ return &preDecLocal_ele_S_Vec3f;}
template <> inline unsigned char** preDecLocal_insert<M_Bool>(){ return &preDecLocal_ele_M_Bool;}
template <> inline unsigned char** preDecLocal_insert<M_Color3f>(){ return &preDecLocal_ele_M_Color;}
template <> inline unsigned char** preDecLocal_insert<M_Float>(){ return &preDecLocal_ele_M_Float;}
template <> inline unsigned char** preDecLocal_insert<M_Image>(){ return &preDecLocal_ele_M_Image;}
template <> inline unsigned char** preDecLocal_insert<M_Int>(){ return &preDecLocal_ele_M_Int32;}
template <> inline unsigned char** preDecLocal_insert<M_Node>(){ return &preDecLocal_ele_M_Node;}
template <> inline unsigned char** preDecLocal_insert<M_Rotation>(){ return &preDecLocal_ele_M_Rotation;}
template <> inline unsigned char** preDecLocal_insert<M_String>(){ return &preDecLocal_ele_M_String;}
template <> inline unsigned char** preDecLocal_insert<M_Double>(){ return &preDecLocal_ele_M_Time;}
template <> inline unsigned char** preDecLocal_insert<M_Vec2f>(){ return &preDecLocal_ele_M_Vec2f;}
template <> inline unsigned char** preDecLocal_insert<M_Vec3f>(){ return &preDecLocal_ele_M_Vec3f;}

extern "C" unsigned char* preIncLocal_ele_S_Bool;
extern "C" unsigned char* preIncLocal_ele_S_Color;
extern "C" unsigned char* preIncLocal_ele_S_Float;
extern "C" unsigned char* preIncLocal_ele_S_Image;
extern "C" unsigned char* preIncLocal_ele_S_Int32;
extern "C" unsigned char* preIncLocal_ele_NodePtr;
extern "C" unsigned char* preIncLocal_ele_S_Rotation;
extern "C" unsigned char* preIncLocal_ele_S_String;
extern "C" unsigned char* preIncLocal_ele_S_Time;
extern "C" unsigned char* preIncLocal_ele_S_Vec2f;
extern "C" unsigned char* preIncLocal_ele_S_Vec3f;
extern "C" unsigned char* preIncLocal_ele_M_Bool;
extern "C" unsigned char* preIncLocal_ele_M_Color;
extern "C" unsigned char* preIncLocal_ele_M_Float;
extern "C" unsigned char* preIncLocal_ele_M_Image;
extern "C" unsigned char* preIncLocal_ele_M_Int32;
extern "C" unsigned char* preIncLocal_ele_M_Node;
extern "C" unsigned char* preIncLocal_ele_M_Rotation;
extern "C" unsigned char* preIncLocal_ele_M_String;
extern "C" unsigned char* preIncLocal_ele_M_Time;
extern "C" unsigned char* preIncLocal_ele_M_Vec2f;
extern "C" unsigned char* preIncLocal_ele_M_Vec3f;

template <class T> inline unsigned char** preIncLocal_insert();
template <> inline unsigned char** preIncLocal_insert<S_Bool>(){ return &preIncLocal_ele_S_Bool;}
template <> inline unsigned char** preIncLocal_insert<Color3f>(){ return &preIncLocal_ele_S_Color;}
template <> inline unsigned char** preIncLocal_insert<S_Float>(){ return &preIncLocal_ele_S_Float;}
template <> inline unsigned char** preIncLocal_insert<Image>(){ return &preIncLocal_ele_S_Image;}
template <> inline unsigned char** preIncLocal_insert<S_Int32>(){ return &preIncLocal_ele_S_Int32;}
template <> inline unsigned char** preIncLocal_insert<NodePtr>(){ return &preIncLocal_ele_NodePtr;}
template <> inline unsigned char** preIncLocal_insert<Rotation>(){ return &preIncLocal_ele_S_Rotation;}
template <> inline unsigned char** preIncLocal_insert<String>(){ return &preIncLocal_ele_S_String;}
template <> inline unsigned char** preIncLocal_insert<double>(){ return &preIncLocal_ele_S_Time;}
template <> inline unsigned char** preIncLocal_insert<Vec2f>(){ return &preIncLocal_ele_S_Vec2f;}
template <> inline unsigned char** preIncLocal_insert<Vec3f>(){ return &preIncLocal_ele_S_Vec3f;}
template <> inline unsigned char** preIncLocal_insert<M_Bool>(){ return &preIncLocal_ele_M_Bool;}
template <> inline unsigned char** preIncLocal_insert<M_Color3f>(){ return &preIncLocal_ele_M_Color;}
template <> inline unsigned char** preIncLocal_insert<M_Float>(){ return &preIncLocal_ele_M_Float;}
template <> inline unsigned char** preIncLocal_insert<M_Image>(){ return &preIncLocal_ele_M_Image;}
template <> inline unsigned char** preIncLocal_insert<M_Int>(){ return &preIncLocal_ele_M_Int32;}
template <> inline unsigned char** preIncLocal_insert<M_Node>(){ return &preIncLocal_ele_M_Node;}
template <> inline unsigned char** preIncLocal_insert<M_Rotation>(){ return &preIncLocal_ele_M_Rotation;}
template <> inline unsigned char** preIncLocal_insert<M_String>(){ return &preIncLocal_ele_M_String;}
template <> inline unsigned char** preIncLocal_insert<M_Double>(){ return &preIncLocal_ele_M_Time;}
template <> inline unsigned char** preIncLocal_insert<M_Vec2f>(){ return &preIncLocal_ele_M_Vec2f;}
template <> inline unsigned char** preIncLocal_insert<M_Vec3f>(){ return &preIncLocal_ele_M_Vec3f;}

extern "C" unsigned char* preDecGlobal_ele_S_Bool;
extern "C" unsigned char* preDecGlobal_ele_S_Color;
extern "C" unsigned char* preDecGlobal_ele_S_Float;
extern "C" unsigned char* preDecGlobal_ele_S_Image;
extern "C" unsigned char* preDecGlobal_ele_S_Int32;
extern "C" unsigned char* preDecGlobal_ele_NodePtr;
extern "C" unsigned char* preDecGlobal_ele_S_Rotation;
extern "C" unsigned char* preDecGlobal_ele_S_String;
extern "C" unsigned char* preDecGlobal_ele_S_Time;
extern "C" unsigned char* preDecGlobal_ele_S_Vec2f;
extern "C" unsigned char* preDecGlobal_ele_S_Vec3f;
extern "C" unsigned char* preDecGlobal_ele_M_Bool;
extern "C" unsigned char* preDecGlobal_ele_M_Color;
extern "C" unsigned char* preDecGlobal_ele_M_Float;
extern "C" unsigned char* preDecGlobal_ele_M_Image;
extern "C" unsigned char* preDecGlobal_ele_M_Int32;
extern "C" unsigned char* preDecGlobal_ele_M_Node;
extern "C" unsigned char* preDecGlobal_ele_M_Rotation;
extern "C" unsigned char* preDecGlobal_ele_M_String;
extern "C" unsigned char* preDecGlobal_ele_M_Time;
extern "C" unsigned char* preDecGlobal_ele_M_Vec2f;
extern "C" unsigned char* preDecGlobal_ele_M_Vec3f;

template <class T> inline unsigned char** preDecGlobal_insert();
template <> inline unsigned char** preDecGlobal_insert<S_Bool>(){ return &preDecGlobal_ele_S_Bool;}
template <> inline unsigned char** preDecGlobal_insert<Color3f>(){ return &preDecGlobal_ele_S_Color;}
template <> inline unsigned char** preDecGlobal_insert<S_Float>(){ return &preDecGlobal_ele_S_Float;}
template <> inline unsigned char** preDecGlobal_insert<Image>(){ return &preDecGlobal_ele_S_Image;}
template <> inline unsigned char** preDecGlobal_insert<S_Int32>(){ return &preDecGlobal_ele_S_Int32;}
template <> inline unsigned char** preDecGlobal_insert<NodePtr>(){ return &preDecGlobal_ele_NodePtr;}
template <> inline unsigned char** preDecGlobal_insert<Rotation>(){ return &preDecGlobal_ele_S_Rotation;}
template <> inline unsigned char** preDecGlobal_insert<String>(){ return &preDecGlobal_ele_S_String;}
template <> inline unsigned char** preDecGlobal_insert<double>(){ return &preDecGlobal_ele_S_Time;}
template <> inline unsigned char** preDecGlobal_insert<Vec2f>(){ return &preDecGlobal_ele_S_Vec2f;}
template <> inline unsigned char** preDecGlobal_insert<Vec3f>(){ return &preDecGlobal_ele_S_Vec3f;}
template <> inline unsigned char** preDecGlobal_insert<M_Bool>(){ return &preDecGlobal_ele_M_Bool;}
template <> inline unsigned char** preDecGlobal_insert<M_Color3f>(){ return &preDecGlobal_ele_M_Color;}
template <> inline unsigned char** preDecGlobal_insert<M_Float>(){ return &preDecGlobal_ele_M_Float;}
template <> inline unsigned char** preDecGlobal_insert<M_Image>(){ return &preDecGlobal_ele_M_Image;}
template <> inline unsigned char** preDecGlobal_insert<M_Int>(){ return &preDecGlobal_ele_M_Int32;}
template <> inline unsigned char** preDecGlobal_insert<M_Node>(){ return &preDecGlobal_ele_M_Node;}
template <> inline unsigned char** preDecGlobal_insert<M_Rotation>(){ return &preDecGlobal_ele_M_Rotation;}
template <> inline unsigned char** preDecGlobal_insert<M_String>(){ return &preDecGlobal_ele_M_String;}
template <> inline unsigned char** preDecGlobal_insert<M_Double>(){ return &preDecGlobal_ele_M_Time;}
template <> inline unsigned char** preDecGlobal_insert<M_Vec2f>(){ return &preDecGlobal_ele_M_Vec2f;}
template <> inline unsigned char** preDecGlobal_insert<M_Vec3f>(){ return &preDecGlobal_ele_M_Vec3f;}

extern "C" unsigned char* preIncGlobal_ele_S_Bool;
extern "C" unsigned char* preIncGlobal_ele_S_Color;
extern "C" unsigned char* preIncGlobal_ele_S_Float;
extern "C" unsigned char* preIncGlobal_ele_S_Image;
extern "C" unsigned char* preIncGlobal_ele_S_Int32;
extern "C" unsigned char* preIncGlobal_ele_NodePtr;
extern "C" unsigned char* preIncGlobal_ele_S_Rotation;
extern "C" unsigned char* preIncGlobal_ele_S_String;
extern "C" unsigned char* preIncGlobal_ele_S_Time;
extern "C" unsigned char* preIncGlobal_ele_S_Vec2f;
extern "C" unsigned char* preIncGlobal_ele_S_Vec3f;
extern "C" unsigned char* preIncGlobal_ele_M_Bool;
extern "C" unsigned char* preIncGlobal_ele_M_Color;
extern "C" unsigned char* preIncGlobal_ele_M_Float;
extern "C" unsigned char* preIncGlobal_ele_M_Image;
extern "C" unsigned char* preIncGlobal_ele_M_Int32;
extern "C" unsigned char* preIncGlobal_ele_M_Node;
extern "C" unsigned char* preIncGlobal_ele_M_Rotation;
extern "C" unsigned char* preIncGlobal_ele_M_String;
extern "C" unsigned char* preIncGlobal_ele_M_Time;
extern "C" unsigned char* preIncGlobal_ele_M_Vec2f;
extern "C" unsigned char* preIncGlobal_ele_M_Vec3f;

template <class T> inline unsigned char** preIncGlobal_insert();
template <> inline unsigned char** preIncGlobal_insert<S_Bool>(){ return &preIncGlobal_ele_S_Bool;}
template <> inline unsigned char** preIncGlobal_insert<Color3f>(){ return &preIncGlobal_ele_S_Color;}
template <> inline unsigned char** preIncGlobal_insert<S_Float>(){ return &preIncGlobal_ele_S_Float;}
template <> inline unsigned char** preIncGlobal_insert<Image>(){ return &preIncGlobal_ele_S_Image;}
template <> inline unsigned char** preIncGlobal_insert<S_Int32>(){ return &preIncGlobal_ele_S_Int32;}
template <> inline unsigned char** preIncGlobal_insert<NodePtr>(){ return &preIncGlobal_ele_NodePtr;}
template <> inline unsigned char** preIncGlobal_insert<Rotation>(){ return &preIncGlobal_ele_S_Rotation;}
template <> inline unsigned char** preIncGlobal_insert<String>(){ return &preIncGlobal_ele_S_String;}
template <> inline unsigned char** preIncGlobal_insert<double>(){ return &preIncGlobal_ele_S_Time;}
template <> inline unsigned char** preIncGlobal_insert<Vec2f>(){ return &preIncGlobal_ele_S_Vec2f;}
template <> inline unsigned char** preIncGlobal_insert<Vec3f>(){ return &preIncGlobal_ele_S_Vec3f;}
template <> inline unsigned char** preIncGlobal_insert<M_Bool>(){ return &preIncGlobal_ele_M_Bool;}
template <> inline unsigned char** preIncGlobal_insert<M_Color3f>(){ return &preIncGlobal_ele_M_Color;}
template <> inline unsigned char** preIncGlobal_insert<M_Float>(){ return &preIncGlobal_ele_M_Float;}
template <> inline unsigned char** preIncGlobal_insert<M_Image>(){ return &preIncGlobal_ele_M_Image;}
template <> inline unsigned char** preIncGlobal_insert<M_Int>(){ return &preIncGlobal_ele_M_Int32;}
template <> inline unsigned char** preIncGlobal_insert<M_Node>(){ return &preIncGlobal_ele_M_Node;}
template <> inline unsigned char** preIncGlobal_insert<M_Rotation>(){ return &preIncGlobal_ele_M_Rotation;}
template <> inline unsigned char** preIncGlobal_insert<M_String>(){ return &preIncGlobal_ele_M_String;}
template <> inline unsigned char** preIncGlobal_insert<M_Double>(){ return &preIncGlobal_ele_M_Time;}
template <> inline unsigned char** preIncGlobal_insert<M_Vec2f>(){ return &preIncGlobal_ele_M_Vec2f;}
template <> inline unsigned char** preIncGlobal_insert<M_Vec3f>(){ return &preIncGlobal_ele_M_Vec3f;}

extern "C" unsigned char* copyGlobalElement_toScratchPad_ele_M_Bool;
extern "C" unsigned char* copyGlobalElement_toScratchPad_ele_M_Color;
extern "C" unsigned char* copyGlobalElement_toScratchPad_ele_M_Float;
extern "C" unsigned char* copyGlobalElement_toScratchPad_ele_M_Image;
extern "C" unsigned char* copyGlobalElement_toScratchPad_ele_M_Int32;
extern "C" unsigned char* copyGlobalElement_toScratchPad_ele_M_Node;
extern "C" unsigned char* copyGlobalElement_toScratchPad_ele_M_Rotation;
extern "C" unsigned char* copyGlobalElement_toScratchPad_ele_M_String;
extern "C" unsigned char* copyGlobalElement_toScratchPad_ele_M_Time;
extern "C" unsigned char* copyGlobalElement_toScratchPad_ele_M_Vec2f;
extern "C" unsigned char* copyGlobalElement_toScratchPad_ele_M_Vec3f;

template <class T> inline unsigned char** copyGlobalElement_toScratchPad_insert();
template <> inline unsigned char** copyGlobalElement_toScratchPad_insert<M_Bool>(){ return &copyGlobalElement_toScratchPad_ele_M_Bool;}
template <> inline unsigned char** copyGlobalElement_toScratchPad_insert<M_Color3f>(){ return &copyGlobalElement_toScratchPad_ele_M_Color;}
template <> inline unsigned char** copyGlobalElement_toScratchPad_insert<M_Float>(){ return &copyGlobalElement_toScratchPad_ele_M_Float;}
template <> inline unsigned char** copyGlobalElement_toScratchPad_insert<M_Image>(){ return &copyGlobalElement_toScratchPad_ele_M_Image;}
template <> inline unsigned char** copyGlobalElement_toScratchPad_insert<M_Int>(){ return &copyGlobalElement_toScratchPad_ele_M_Int32;}
template <> inline unsigned char** copyGlobalElement_toScratchPad_insert<M_Node>(){ return &copyGlobalElement_toScratchPad_ele_M_Node;}
template <> inline unsigned char** copyGlobalElement_toScratchPad_insert<M_Rotation>(){ return &copyGlobalElement_toScratchPad_ele_M_Rotation;}
template <> inline unsigned char** copyGlobalElement_toScratchPad_insert<M_String>(){ return &copyGlobalElement_toScratchPad_ele_M_String;}
template <> inline unsigned char** copyGlobalElement_toScratchPad_insert<M_Double>(){ return &copyGlobalElement_toScratchPad_ele_M_Time;}
template <> inline unsigned char** copyGlobalElement_toScratchPad_insert<M_Vec2f>(){ return &copyGlobalElement_toScratchPad_ele_M_Vec2f;}
template <> inline unsigned char** copyGlobalElement_toScratchPad_insert<M_Vec3f>(){ return &copyGlobalElement_toScratchPad_ele_M_Vec3f;}

extern "C" unsigned char* copyLocalElement_toScratchPad_ele_M_Bool;
extern "C" unsigned char* copyLocalElement_toScratchPad_ele_M_Color;
extern "C" unsigned char* copyLocalElement_toScratchPad_ele_M_Float;
extern "C" unsigned char* copyLocalElement_toScratchPad_ele_M_Image;
extern "C" unsigned char* copyLocalElement_toScratchPad_ele_M_Int32;
extern "C" unsigned char* copyLocalElement_toScratchPad_ele_M_Node;
extern "C" unsigned char* copyLocalElement_toScratchPad_ele_M_Rotation;
extern "C" unsigned char* copyLocalElement_toScratchPad_ele_M_String;
extern "C" unsigned char* copyLocalElement_toScratchPad_ele_M_Time;
extern "C" unsigned char* copyLocalElement_toScratchPad_ele_M_Vec2f;
extern "C" unsigned char* copyLocalElement_toScratchPad_ele_M_Vec3f;

template <class T> inline unsigned char** copyLocalElement_toScratchPad_insert();
template <> inline unsigned char** copyLocalElement_toScratchPad_insert<M_Bool>(){ return &copyLocalElement_toScratchPad_ele_M_Bool;}
template <> inline unsigned char** copyLocalElement_toScratchPad_insert<M_Color3f>(){ return &copyLocalElement_toScratchPad_ele_M_Color;}
template <> inline unsigned char** copyLocalElement_toScratchPad_insert<M_Float>(){ return &copyLocalElement_toScratchPad_ele_M_Float;}
template <> inline unsigned char** copyLocalElement_toScratchPad_insert<M_Image>(){ return &copyLocalElement_toScratchPad_ele_M_Image;}
template <> inline unsigned char** copyLocalElement_toScratchPad_insert<M_Int>(){ return &copyLocalElement_toScratchPad_ele_M_Int32;}
template <> inline unsigned char** copyLocalElement_toScratchPad_insert<M_Node>(){ return &copyLocalElement_toScratchPad_ele_M_Node;}
template <> inline unsigned char** copyLocalElement_toScratchPad_insert<M_Rotation>(){ return &copyLocalElement_toScratchPad_ele_M_Rotation;}
template <> inline unsigned char** copyLocalElement_toScratchPad_insert<M_String>(){ return &copyLocalElement_toScratchPad_ele_M_String;}
template <> inline unsigned char** copyLocalElement_toScratchPad_insert<M_Double>(){ return &copyLocalElement_toScratchPad_ele_M_Time;}
template <> inline unsigned char** copyLocalElement_toScratchPad_insert<M_Vec2f>(){ return &copyLocalElement_toScratchPad_ele_M_Vec2f;}
template <> inline unsigned char** copyLocalElement_toScratchPad_insert<M_Vec3f>(){ return &copyLocalElement_toScratchPad_ele_M_Vec3f;}

extern "C" unsigned char* lessThan_ele_S_Bool;
extern "C" unsigned char* lessThan_ele_S_Color;
extern "C" unsigned char* lessThan_ele_S_Float;
extern "C" unsigned char* lessThan_ele_S_Image;
extern "C" unsigned char* lessThan_ele_S_Int32;
extern "C" unsigned char* lessThan_ele_NodePtr;
extern "C" unsigned char* lessThan_ele_S_Rotation;
extern "C" unsigned char* lessThan_ele_S_String;
extern "C" unsigned char* lessThan_ele_S_Time;
extern "C" unsigned char* lessThan_ele_S_Vec2f;
extern "C" unsigned char* lessThan_ele_S_Vec3f;
extern "C" unsigned char* lessThan_ele_M_Bool;
extern "C" unsigned char* lessThan_ele_M_Color;
extern "C" unsigned char* lessThan_ele_M_Float;
extern "C" unsigned char* lessThan_ele_M_Image;
extern "C" unsigned char* lessThan_ele_M_Int32;
extern "C" unsigned char* lessThan_ele_M_Node;
extern "C" unsigned char* lessThan_ele_M_Rotation;
extern "C" unsigned char* lessThan_ele_M_String;
extern "C" unsigned char* lessThan_ele_M_Time;
extern "C" unsigned char* lessThan_ele_M_Vec2f;
extern "C" unsigned char* lessThan_ele_M_Vec3f;

template <class T> inline unsigned char** lessThan_insert();
template <> inline unsigned char** lessThan_insert<S_Bool>(){ return &lessThan_ele_S_Bool;}
template <> inline unsigned char** lessThan_insert<Color3f>(){ return &lessThan_ele_S_Color;}
template <> inline unsigned char** lessThan_insert<S_Float>(){ return &lessThan_ele_S_Float;}
template <> inline unsigned char** lessThan_insert<Image>(){ return &lessThan_ele_S_Image;}
template <> inline unsigned char** lessThan_insert<S_Int32>(){ return &lessThan_ele_S_Int32;}
template <> inline unsigned char** lessThan_insert<NodePtr>(){ return &lessThan_ele_NodePtr;}
template <> inline unsigned char** lessThan_insert<Rotation>(){ return &lessThan_ele_S_Rotation;}
template <> inline unsigned char** lessThan_insert<String>(){ return &lessThan_ele_S_String;}
template <> inline unsigned char** lessThan_insert<double>(){ return &lessThan_ele_S_Time;}
template <> inline unsigned char** lessThan_insert<Vec2f>(){ return &lessThan_ele_S_Vec2f;}
template <> inline unsigned char** lessThan_insert<Vec3f>(){ return &lessThan_ele_S_Vec3f;}
template <> inline unsigned char** lessThan_insert<M_Bool>(){ return &lessThan_ele_M_Bool;}
template <> inline unsigned char** lessThan_insert<M_Color3f>(){ return &lessThan_ele_M_Color;}
template <> inline unsigned char** lessThan_insert<M_Float>(){ return &lessThan_ele_M_Float;}
template <> inline unsigned char** lessThan_insert<M_Image>(){ return &lessThan_ele_M_Image;}
template <> inline unsigned char** lessThan_insert<M_Int>(){ return &lessThan_ele_M_Int32;}
template <> inline unsigned char** lessThan_insert<M_Node>(){ return &lessThan_ele_M_Node;}
template <> inline unsigned char** lessThan_insert<M_Rotation>(){ return &lessThan_ele_M_Rotation;}
template <> inline unsigned char** lessThan_insert<M_String>(){ return &lessThan_ele_M_String;}
template <> inline unsigned char** lessThan_insert<M_Double>(){ return &lessThan_ele_M_Time;}
template <> inline unsigned char** lessThan_insert<M_Vec2f>(){ return &lessThan_ele_M_Vec2f;}
template <> inline unsigned char** lessThan_insert<M_Vec3f>(){ return &lessThan_ele_M_Vec3f;}

extern "C" unsigned char* lessEqual_ele_S_Bool;
extern "C" unsigned char* lessEqual_ele_S_Color;
extern "C" unsigned char* lessEqual_ele_S_Float;
extern "C" unsigned char* lessEqual_ele_S_Image;
extern "C" unsigned char* lessEqual_ele_S_Int32;
extern "C" unsigned char* lessEqual_ele_NodePtr;
extern "C" unsigned char* lessEqual_ele_S_Rotation;
extern "C" unsigned char* lessEqual_ele_S_String;
extern "C" unsigned char* lessEqual_ele_S_Time;
extern "C" unsigned char* lessEqual_ele_S_Vec2f;
extern "C" unsigned char* lessEqual_ele_S_Vec3f;
extern "C" unsigned char* lessEqual_ele_M_Bool;
extern "C" unsigned char* lessEqual_ele_M_Color;
extern "C" unsigned char* lessEqual_ele_M_Float;
extern "C" unsigned char* lessEqual_ele_M_Image;
extern "C" unsigned char* lessEqual_ele_M_Int32;
extern "C" unsigned char* lessEqual_ele_M_Node;
extern "C" unsigned char* lessEqual_ele_M_Rotation;
extern "C" unsigned char* lessEqual_ele_M_String;
extern "C" unsigned char* lessEqual_ele_M_Time;
extern "C" unsigned char* lessEqual_ele_M_Vec2f;
extern "C" unsigned char* lessEqual_ele_M_Vec3f;

template <class T> inline unsigned char** lessEqual_insert();
template <> inline unsigned char** lessEqual_insert<S_Bool>(){ return &lessEqual_ele_S_Bool;}
template <> inline unsigned char** lessEqual_insert<Color3f>(){ return &lessEqual_ele_S_Color;}
template <> inline unsigned char** lessEqual_insert<S_Float>(){ return &lessEqual_ele_S_Float;}
template <> inline unsigned char** lessEqual_insert<Image>(){ return &lessEqual_ele_S_Image;}
template <> inline unsigned char** lessEqual_insert<S_Int32>(){ return &lessEqual_ele_S_Int32;}
template <> inline unsigned char** lessEqual_insert<NodePtr>(){ return &lessEqual_ele_NodePtr;}
template <> inline unsigned char** lessEqual_insert<Rotation>(){ return &lessEqual_ele_S_Rotation;}
template <> inline unsigned char** lessEqual_insert<String>(){ return &lessEqual_ele_S_String;}
template <> inline unsigned char** lessEqual_insert<double>(){ return &lessEqual_ele_S_Time;}
template <> inline unsigned char** lessEqual_insert<Vec2f>(){ return &lessEqual_ele_S_Vec2f;}
template <> inline unsigned char** lessEqual_insert<Vec3f>(){ return &lessEqual_ele_S_Vec3f;}
template <> inline unsigned char** lessEqual_insert<M_Bool>(){ return &lessEqual_ele_M_Bool;}
template <> inline unsigned char** lessEqual_insert<M_Color3f>(){ return &lessEqual_ele_M_Color;}
template <> inline unsigned char** lessEqual_insert<M_Float>(){ return &lessEqual_ele_M_Float;}
template <> inline unsigned char** lessEqual_insert<M_Image>(){ return &lessEqual_ele_M_Image;}
template <> inline unsigned char** lessEqual_insert<M_Int>(){ return &lessEqual_ele_M_Int32;}
template <> inline unsigned char** lessEqual_insert<M_Node>(){ return &lessEqual_ele_M_Node;}
template <> inline unsigned char** lessEqual_insert<M_Rotation>(){ return &lessEqual_ele_M_Rotation;}
template <> inline unsigned char** lessEqual_insert<M_String>(){ return &lessEqual_ele_M_String;}
template <> inline unsigned char** lessEqual_insert<M_Double>(){ return &lessEqual_ele_M_Time;}
template <> inline unsigned char** lessEqual_insert<M_Vec2f>(){ return &lessEqual_ele_M_Vec2f;}
template <> inline unsigned char** lessEqual_insert<M_Vec3f>(){ return &lessEqual_ele_M_Vec3f;}

extern "C" unsigned char* greaterThan_ele_S_Bool;
extern "C" unsigned char* greaterThan_ele_S_Color;
extern "C" unsigned char* greaterThan_ele_S_Float;
extern "C" unsigned char* greaterThan_ele_S_Image;
extern "C" unsigned char* greaterThan_ele_S_Int32;
extern "C" unsigned char* greaterThan_ele_NodePtr;
extern "C" unsigned char* greaterThan_ele_S_Rotation;
extern "C" unsigned char* greaterThan_ele_S_String;
extern "C" unsigned char* greaterThan_ele_S_Time;
extern "C" unsigned char* greaterThan_ele_S_Vec2f;
extern "C" unsigned char* greaterThan_ele_S_Vec3f;
extern "C" unsigned char* greaterThan_ele_M_Bool;
extern "C" unsigned char* greaterThan_ele_M_Color;
extern "C" unsigned char* greaterThan_ele_M_Float;
extern "C" unsigned char* greaterThan_ele_M_Image;
extern "C" unsigned char* greaterThan_ele_M_Int32;
extern "C" unsigned char* greaterThan_ele_M_Node;
extern "C" unsigned char* greaterThan_ele_M_Rotation;
extern "C" unsigned char* greaterThan_ele_M_String;
extern "C" unsigned char* greaterThan_ele_M_Time;
extern "C" unsigned char* greaterThan_ele_M_Vec2f;
extern "C" unsigned char* greaterThan_ele_M_Vec3f;

template <class T> inline unsigned char** greaterThan_insert();
template <> inline unsigned char** greaterThan_insert<S_Bool>(){ return &greaterThan_ele_S_Bool;}
template <> inline unsigned char** greaterThan_insert<Color3f>(){ return &greaterThan_ele_S_Color;}
template <> inline unsigned char** greaterThan_insert<S_Float>(){ return &greaterThan_ele_S_Float;}
template <> inline unsigned char** greaterThan_insert<Image>(){ return &greaterThan_ele_S_Image;}
template <> inline unsigned char** greaterThan_insert<S_Int32>(){ return &greaterThan_ele_S_Int32;}
template <> inline unsigned char** greaterThan_insert<NodePtr>(){ return &greaterThan_ele_NodePtr;}
template <> inline unsigned char** greaterThan_insert<Rotation>(){ return &greaterThan_ele_S_Rotation;}
template <> inline unsigned char** greaterThan_insert<String>(){ return &greaterThan_ele_S_String;}
template <> inline unsigned char** greaterThan_insert<double>(){ return &greaterThan_ele_S_Time;}
template <> inline unsigned char** greaterThan_insert<Vec2f>(){ return &greaterThan_ele_S_Vec2f;}
template <> inline unsigned char** greaterThan_insert<Vec3f>(){ return &greaterThan_ele_S_Vec3f;}
template <> inline unsigned char** greaterThan_insert<M_Bool>(){ return &greaterThan_ele_M_Bool;}
template <> inline unsigned char** greaterThan_insert<M_Color3f>(){ return &greaterThan_ele_M_Color;}
template <> inline unsigned char** greaterThan_insert<M_Float>(){ return &greaterThan_ele_M_Float;}
template <> inline unsigned char** greaterThan_insert<M_Image>(){ return &greaterThan_ele_M_Image;}
template <> inline unsigned char** greaterThan_insert<M_Int>(){ return &greaterThan_ele_M_Int32;}
template <> inline unsigned char** greaterThan_insert<M_Node>(){ return &greaterThan_ele_M_Node;}
template <> inline unsigned char** greaterThan_insert<M_Rotation>(){ return &greaterThan_ele_M_Rotation;}
template <> inline unsigned char** greaterThan_insert<M_String>(){ return &greaterThan_ele_M_String;}
template <> inline unsigned char** greaterThan_insert<M_Double>(){ return &greaterThan_ele_M_Time;}
template <> inline unsigned char** greaterThan_insert<M_Vec2f>(){ return &greaterThan_ele_M_Vec2f;}
template <> inline unsigned char** greaterThan_insert<M_Vec3f>(){ return &greaterThan_ele_M_Vec3f;}

extern "C" unsigned char* greaterEqual_ele_S_Bool;
extern "C" unsigned char* greaterEqual_ele_S_Color;
extern "C" unsigned char* greaterEqual_ele_S_Float;
extern "C" unsigned char* greaterEqual_ele_S_Image;
extern "C" unsigned char* greaterEqual_ele_S_Int32;
extern "C" unsigned char* greaterEqual_ele_NodePtr;
extern "C" unsigned char* greaterEqual_ele_S_Rotation;
extern "C" unsigned char* greaterEqual_ele_S_String;
extern "C" unsigned char* greaterEqual_ele_S_Time;
extern "C" unsigned char* greaterEqual_ele_S_Vec2f;
extern "C" unsigned char* greaterEqual_ele_S_Vec3f;
extern "C" unsigned char* greaterEqual_ele_M_Bool;
extern "C" unsigned char* greaterEqual_ele_M_Color;
extern "C" unsigned char* greaterEqual_ele_M_Float;
extern "C" unsigned char* greaterEqual_ele_M_Image;
extern "C" unsigned char* greaterEqual_ele_M_Int32;
extern "C" unsigned char* greaterEqual_ele_M_Node;
extern "C" unsigned char* greaterEqual_ele_M_Rotation;
extern "C" unsigned char* greaterEqual_ele_M_String;
extern "C" unsigned char* greaterEqual_ele_M_Time;
extern "C" unsigned char* greaterEqual_ele_M_Vec2f;
extern "C" unsigned char* greaterEqual_ele_M_Vec3f;

template <class T> inline unsigned char** greaterEqual_insert();
template <> inline unsigned char** greaterEqual_insert<S_Bool>(){ return &greaterEqual_ele_S_Bool;}
template <> inline unsigned char** greaterEqual_insert<Color3f>(){ return &greaterEqual_ele_S_Color;}
template <> inline unsigned char** greaterEqual_insert<S_Float>(){ return &greaterEqual_ele_S_Float;}
template <> inline unsigned char** greaterEqual_insert<Image>(){ return &greaterEqual_ele_S_Image;}
template <> inline unsigned char** greaterEqual_insert<S_Int32>(){ return &greaterEqual_ele_S_Int32;}
template <> inline unsigned char** greaterEqual_insert<NodePtr>(){ return &greaterEqual_ele_NodePtr;}
template <> inline unsigned char** greaterEqual_insert<Rotation>(){ return &greaterEqual_ele_S_Rotation;}
template <> inline unsigned char** greaterEqual_insert<String>(){ return &greaterEqual_ele_S_String;}
template <> inline unsigned char** greaterEqual_insert<double>(){ return &greaterEqual_ele_S_Time;}
template <> inline unsigned char** greaterEqual_insert<Vec2f>(){ return &greaterEqual_ele_S_Vec2f;}
template <> inline unsigned char** greaterEqual_insert<Vec3f>(){ return &greaterEqual_ele_S_Vec3f;}
template <> inline unsigned char** greaterEqual_insert<M_Bool>(){ return &greaterEqual_ele_M_Bool;}
template <> inline unsigned char** greaterEqual_insert<M_Color3f>(){ return &greaterEqual_ele_M_Color;}
template <> inline unsigned char** greaterEqual_insert<M_Float>(){ return &greaterEqual_ele_M_Float;}
template <> inline unsigned char** greaterEqual_insert<M_Image>(){ return &greaterEqual_ele_M_Image;}
template <> inline unsigned char** greaterEqual_insert<M_Int>(){ return &greaterEqual_ele_M_Int32;}
template <> inline unsigned char** greaterEqual_insert<M_Node>(){ return &greaterEqual_ele_M_Node;}
template <> inline unsigned char** greaterEqual_insert<M_Rotation>(){ return &greaterEqual_ele_M_Rotation;}
template <> inline unsigned char** greaterEqual_insert<M_String>(){ return &greaterEqual_ele_M_String;}
template <> inline unsigned char** greaterEqual_insert<M_Double>(){ return &greaterEqual_ele_M_Time;}
template <> inline unsigned char** greaterEqual_insert<M_Vec2f>(){ return &greaterEqual_ele_M_Vec2f;}
template <> inline unsigned char** greaterEqual_insert<M_Vec3f>(){ return &greaterEqual_ele_M_Vec3f;}

extern "C" unsigned char* equality_ele_S_Bool;
extern "C" unsigned char* equality_ele_S_Color;
extern "C" unsigned char* equality_ele_S_Float;
extern "C" unsigned char* equality_ele_S_Image;
extern "C" unsigned char* equality_ele_S_Int32;
extern "C" unsigned char* equality_ele_NodePtr;
extern "C" unsigned char* equality_ele_S_Rotation;
extern "C" unsigned char* equality_ele_S_String;
extern "C" unsigned char* equality_ele_S_Time;
extern "C" unsigned char* equality_ele_S_Vec2f;
extern "C" unsigned char* equality_ele_S_Vec3f;
extern "C" unsigned char* equality_ele_M_Bool;
extern "C" unsigned char* equality_ele_M_Color;
extern "C" unsigned char* equality_ele_M_Float;
extern "C" unsigned char* equality_ele_M_Image;
extern "C" unsigned char* equality_ele_M_Int32;
extern "C" unsigned char* equality_ele_M_Node;
extern "C" unsigned char* equality_ele_M_Rotation;
extern "C" unsigned char* equality_ele_M_String;
extern "C" unsigned char* equality_ele_M_Time;
extern "C" unsigned char* equality_ele_M_Vec2f;
extern "C" unsigned char* equality_ele_M_Vec3f;

template <class T> inline unsigned char** equality_insert();
template <> inline unsigned char** equality_insert<S_Bool>(){ return &equality_ele_S_Bool;}
template <> inline unsigned char** equality_insert<Color3f>(){ return &equality_ele_S_Color;}
template <> inline unsigned char** equality_insert<S_Float>(){ return &equality_ele_S_Float;}
template <> inline unsigned char** equality_insert<Image>(){ return &equality_ele_S_Image;}
template <> inline unsigned char** equality_insert<S_Int32>(){ return &equality_ele_S_Int32;}
template <> inline unsigned char** equality_insert<NodePtr>(){ return &equality_ele_NodePtr;}
template <> inline unsigned char** equality_insert<Rotation>(){ return &equality_ele_S_Rotation;}
template <> inline unsigned char** equality_insert<String>(){ return &equality_ele_S_String;}
template <> inline unsigned char** equality_insert<double>(){ return &equality_ele_S_Time;}
template <> inline unsigned char** equality_insert<Vec2f>(){ return &equality_ele_S_Vec2f;}
template <> inline unsigned char** equality_insert<Vec3f>(){ return &equality_ele_S_Vec3f;}
template <> inline unsigned char** equality_insert<M_Bool>(){ return &equality_ele_M_Bool;}
template <> inline unsigned char** equality_insert<M_Color3f>(){ return &equality_ele_M_Color;}
template <> inline unsigned char** equality_insert<M_Float>(){ return &equality_ele_M_Float;}
template <> inline unsigned char** equality_insert<M_Image>(){ return &equality_ele_M_Image;}
template <> inline unsigned char** equality_insert<M_Int>(){ return &equality_ele_M_Int32;}
template <> inline unsigned char** equality_insert<M_Node>(){ return &equality_ele_M_Node;}
template <> inline unsigned char** equality_insert<M_Rotation>(){ return &equality_ele_M_Rotation;}
template <> inline unsigned char** equality_insert<M_String>(){ return &equality_ele_M_String;}
template <> inline unsigned char** equality_insert<M_Double>(){ return &equality_ele_M_Time;}
template <> inline unsigned char** equality_insert<M_Vec2f>(){ return &equality_ele_M_Vec2f;}
template <> inline unsigned char** equality_insert<M_Vec3f>(){ return &equality_ele_M_Vec3f;}

extern "C" unsigned char* nonEquality_ele_S_Bool;
extern "C" unsigned char* nonEquality_ele_S_Color;
extern "C" unsigned char* nonEquality_ele_S_Float;
extern "C" unsigned char* nonEquality_ele_S_Image;
extern "C" unsigned char* nonEquality_ele_S_Int32;
extern "C" unsigned char* nonEquality_ele_NodePtr;
extern "C" unsigned char* nonEquality_ele_S_Rotation;
extern "C" unsigned char* nonEquality_ele_S_String;
extern "C" unsigned char* nonEquality_ele_S_Time;
extern "C" unsigned char* nonEquality_ele_S_Vec2f;
extern "C" unsigned char* nonEquality_ele_S_Vec3f;
extern "C" unsigned char* nonEquality_ele_M_Bool;
extern "C" unsigned char* nonEquality_ele_M_Color;
extern "C" unsigned char* nonEquality_ele_M_Float;
extern "C" unsigned char* nonEquality_ele_M_Image;
extern "C" unsigned char* nonEquality_ele_M_Int32;
extern "C" unsigned char* nonEquality_ele_M_Node;
extern "C" unsigned char* nonEquality_ele_M_Rotation;
extern "C" unsigned char* nonEquality_ele_M_String;
extern "C" unsigned char* nonEquality_ele_M_Time;
extern "C" unsigned char* nonEquality_ele_M_Vec2f;
extern "C" unsigned char* nonEquality_ele_M_Vec3f;

template <class T> inline unsigned char** nonEquality_insert();
template <> inline unsigned char** nonEquality_insert<S_Bool>(){ return &nonEquality_ele_S_Bool;}
template <> inline unsigned char** nonEquality_insert<Color3f>(){ return &nonEquality_ele_S_Color;}
template <> inline unsigned char** nonEquality_insert<S_Float>(){ return &nonEquality_ele_S_Float;}
template <> inline unsigned char** nonEquality_insert<Image>(){ return &nonEquality_ele_S_Image;}
template <> inline unsigned char** nonEquality_insert<S_Int32>(){ return &nonEquality_ele_S_Int32;}
template <> inline unsigned char** nonEquality_insert<NodePtr>(){ return &nonEquality_ele_NodePtr;}
template <> inline unsigned char** nonEquality_insert<Rotation>(){ return &nonEquality_ele_S_Rotation;}
template <> inline unsigned char** nonEquality_insert<String>(){ return &nonEquality_ele_S_String;}
template <> inline unsigned char** nonEquality_insert<double>(){ return &nonEquality_ele_S_Time;}
template <> inline unsigned char** nonEquality_insert<Vec2f>(){ return &nonEquality_ele_S_Vec2f;}
template <> inline unsigned char** nonEquality_insert<Vec3f>(){ return &nonEquality_ele_S_Vec3f;}
template <> inline unsigned char** nonEquality_insert<M_Bool>(){ return &nonEquality_ele_M_Bool;}
template <> inline unsigned char** nonEquality_insert<M_Color3f>(){ return &nonEquality_ele_M_Color;}
template <> inline unsigned char** nonEquality_insert<M_Float>(){ return &nonEquality_ele_M_Float;}
template <> inline unsigned char** nonEquality_insert<M_Image>(){ return &nonEquality_ele_M_Image;}
template <> inline unsigned char** nonEquality_insert<M_Int>(){ return &nonEquality_ele_M_Int32;}
template <> inline unsigned char** nonEquality_insert<M_Node>(){ return &nonEquality_ele_M_Node;}
template <> inline unsigned char** nonEquality_insert<M_Rotation>(){ return &nonEquality_ele_M_Rotation;}
template <> inline unsigned char** nonEquality_insert<M_String>(){ return &nonEquality_ele_M_String;}
template <> inline unsigned char** nonEquality_insert<M_Double>(){ return &nonEquality_ele_M_Time;}
template <> inline unsigned char** nonEquality_insert<M_Vec2f>(){ return &nonEquality_ele_M_Vec2f;}
template <> inline unsigned char** nonEquality_insert<M_Vec3f>(){ return &nonEquality_ele_M_Vec3f;}

extern "C" unsigned char* and_ele;
extern "C" unsigned char* or_ele;
extern "C" unsigned char* callVal_ele;
extern "C" unsigned char* if_ele;
extern "C" unsigned char* not_if_ele;
extern "C" unsigned char* goto_ele;

extern "C" unsigned char* float2Int_ele;
extern "C" unsigned char* float2Double_ele;
extern "C" unsigned char* int2Float_ele;
extern "C" unsigned char* int2Double_ele;
extern "C" unsigned char* double2Float_ele;
extern "C" unsigned char* double2Int_ele;

template <class A, class B> inline unsigned char** typeToType_insert();
template <> inline unsigned char** typeToType_insert<S_Float,S_Int32>(){ return &float2Int_ele;}
template <> inline unsigned char** typeToType_insert<S_Float,double>(){ return &float2Double_ele;}

template <> inline unsigned char** typeToType_insert<S_Int32,S_Float>(){ return &int2Float_ele;}
template <> inline unsigned char** typeToType_insert<S_Int32,double>(){ return &int2Double_ele;}

template <> inline unsigned char** typeToType_insert<double,S_Float>(){ return &double2Float_ele;}
template <> inline unsigned char** typeToType_insert<double,S_Int32>(){ return &double2Int_ele;}


extern "C" unsigned char* swapOrder_ele_S_Rotation;
extern "C" unsigned char* swapOrder_ele_S_Vec2f;
extern "C" unsigned char* swapOrder_ele_S_Vec3f;

extern "C" unsigned char* copyScratchPadToGlobalComponent_x_ele_S_Vec3f;
extern "C" unsigned char* copyScratchPadToGlobalComponent_y_ele_S_Vec3f;
extern "C" unsigned char* copyScratchPadToGlobalComponent_z_ele_S_Vec3f;

extern "C" unsigned char* copyScratchPadToLocalComponent_x_ele_S_Vec3f;
extern "C" unsigned char* copyScratchPadToLocalComponent_y_ele_S_Vec3f;
extern "C" unsigned char* copyScratchPadToLocalComponent_z_ele_S_Vec3f;

extern "C" unsigned char* copyGlobalComponent_x_toScratchPad_ele_S_Vec2f;
extern "C" unsigned char* copyGlobalComponent_y_toScratchPad_ele_S_Vec2f;

extern "C" unsigned char* copyGlobalComponent_x_toScratchPad_ele_S_Vec3f;
extern "C" unsigned char* copyGlobalComponent_y_toScratchPad_ele_S_Vec3f;
extern "C" unsigned char* copyGlobalComponent_z_toScratchPad_ele_S_Vec3f;

extern "C" unsigned char* copyLocalComponent_x_toScratchPad_ele_S_Rotation;
extern "C" unsigned char* copyLocalComponent_y_toScratchPad_ele_S_Rotation;
extern "C" unsigned char* copyLocalComponent_z_toScratchPad_ele_S_Rotation;
extern "C" unsigned char* copyLocalComponent_angle_toScratchPad_ele_S_Rotation;

extern "C" unsigned char* copyLocalComponent_x_toScratchPad_ele_S_Vec2f;
extern "C" unsigned char* copyLocalComponent_y_toScratchPad_ele_S_Vec2f;

extern "C" unsigned char* copyLocalComponent_x_toScratchPad_ele_S_Vec3f;
extern "C" unsigned char* copyLocalComponent_y_toScratchPad_ele_S_Vec3f;
extern "C" unsigned char* copyLocalComponent_z_toScratchPad_ele_S_Vec3f;

extern "C" unsigned char* copyScratchPadToGlobalComponent_x_ele_M_Vec3f;
extern "C" unsigned char* copyScratchPadToGlobalComponent_y_ele_M_Vec3f;
extern "C" unsigned char* copyScratchPadToGlobalComponent_z_ele_M_Vec3f;

template <class T> inline unsigned char** copyScratchPadToGlobalComponent_M_Vec3f_insert();
template <> inline unsigned char** copyScratchPadToGlobalComponent_M_Vec3f_insert<Component_X>(){ return &copyScratchPadToGlobalComponent_x_ele_M_Vec3f;}
template <> inline unsigned char** copyScratchPadToGlobalComponent_M_Vec3f_insert<Component_Y>(){ return &copyScratchPadToGlobalComponent_y_ele_M_Vec3f;}
template <> inline unsigned char** copyScratchPadToGlobalComponent_M_Vec3f_insert<Component_Z>(){ return &copyScratchPadToGlobalComponent_z_ele_M_Vec3f;}

extern "C" unsigned char* addScratchPadToGlobalComponent_x_ele_G_M_Vec3f;
extern "C" unsigned char* addScratchPadToGlobalComponent_y_ele_G_M_Vec3f;
extern "C" unsigned char* addScratchPadToGlobalComponent_z_ele_G_M_Vec3f;

template <class T> inline unsigned char** addScratchPadToGlobalComponent_M_Vec3f_insert();
template <> inline unsigned char** addScratchPadToGlobalComponent_M_Vec3f_insert<Component_X>(){ return &addScratchPadToGlobalComponent_x_ele_G_M_Vec3f;}
template <> inline unsigned char** addScratchPadToGlobalComponent_M_Vec3f_insert<Component_Y>(){ return &addScratchPadToGlobalComponent_y_ele_G_M_Vec3f;}
template <> inline unsigned char** addScratchPadToGlobalComponent_M_Vec3f_insert<Component_Z>(){ return &addScratchPadToGlobalComponent_z_ele_G_M_Vec3f;}

extern "C" unsigned char* copyGlobalComponent_x_toScratchPad_ele_M_Vec3f;
extern "C" unsigned char* copyGlobalComponent_y_toScratchPad_ele_M_Vec3f;
extern "C" unsigned char* copyGlobalComponent_z_toScratchPad_ele_M_Vec3f;

extern "C" unsigned char* subtraction_ele_S_Float_S_Int32;
extern "C" unsigned char* subtraction_ele_S_Float_S_Time;
extern "C" unsigned char* subtraction_ele_S_Int32_S_Float;
extern "C" unsigned char* subtraction_ele_S_Int32_S_Time;
extern "C" unsigned char* subtraction_ele_S_Time_S_Float;
extern "C" unsigned char* subtraction_ele_S_Time_S_Int32;
extern "C" unsigned char* addition_ele_S_Float_S_Int32;
extern "C" unsigned char* addition_ele_S_Float_S_Time;
extern "C" unsigned char* addition_ele_S_Int32_S_Float;
extern "C" unsigned char* addition_ele_S_Int32_S_Time;
extern "C" unsigned char* addition_ele_S_Time_S_Float;
extern "C" unsigned char* addition_ele_S_Time_S_Int32;

extern "C" unsigned char* division_ele_S_Float_S_Int32;
extern "C" unsigned char* division_ele_S_Float_S_Time;
extern "C" unsigned char* division_ele_S_Int32_S_Float;
extern "C" unsigned char* division_ele_S_Int32_S_Time;
extern "C" unsigned char* division_ele_S_Time_S_Float;
extern "C" unsigned char* division_ele_S_Time_S_Int32;
extern "C" unsigned char* multiplication_ele_S_Float_S_Int32;
extern "C" unsigned char* multiplication_ele_S_Float_S_Time;
extern "C" unsigned char* multiplication_ele_S_Int32_S_Float;
extern "C" unsigned char* multiplication_ele_S_Int32_S_Time;
extern "C" unsigned char* multiplication_ele_S_Time_S_Float;
extern "C" unsigned char* multiplication_ele_S_Time_S_Int32;

extern "C" unsigned char* s_RotationMultiplyS_Vec3f_ele;

extern "C" unsigned char* copyScratchPadToGlobal_ele_S_Float_with_S_Int32;
extern "C" unsigned char* copyScratchPadToGlobal_ele_S_Float_with_S_Time;

extern "C" unsigned char* copyScratchPadToGlobal_ele_S_Int32_with_S_Float;
extern "C" unsigned char* copyScratchPadToGlobal_ele_S_Int32_with_S_Time;

extern "C" unsigned char* copyScratchPadToGlobal_ele_S_Time_with_S_Float;
extern "C" unsigned char* copyScratchPadToGlobal_ele_S_Time_with_S_Int32;
////
extern "C" unsigned char* copyScratchPadToLocal_ele_S_Float_with_S_Int32;
extern "C" unsigned char* copyScratchPadToLocal_ele_S_Float_with_S_Time;

extern "C" unsigned char* copyScratchPadToLocal_ele_S_Int32_with_S_Float;
extern "C" unsigned char* copyScratchPadToLocal_ele_S_Int32_with_S_Time;

extern "C" unsigned char* copyScratchPadToLocal_ele_S_Time_with_S_Float;
extern "C" unsigned char* copyScratchPadToLocal_ele_S_Time_with_S_Int32;

extern "C" unsigned char* pushScPA_ele_S_Bool;
extern "C" unsigned char* pushScPA_ele_S_Color;
extern "C" unsigned char* pushScPA_ele_S_Float;
extern "C" unsigned char* pushScPA_ele_S_Image;
extern "C" unsigned char* pushScPA_ele_S_Int32;
extern "C" unsigned char* pushScPA_ele_NodePtr;
extern "C" unsigned char* pushScPA_ele_S_Rotation;
extern "C" unsigned char* pushScPA_ele_S_String;
extern "C" unsigned char* pushScPA_ele_S_Time;
extern "C" unsigned char* pushScPA_ele_S_Vec2f;
extern "C" unsigned char* pushScPA_ele_S_Vec3f;
extern "C" unsigned char* pushScPA_ele_M_Bool;
extern "C" unsigned char* pushScPA_ele_M_Color;
extern "C" unsigned char* pushScPA_ele_M_Float;
extern "C" unsigned char* pushScPA_ele_M_Image;
extern "C" unsigned char* pushScPA_ele_M_Int32;
extern "C" unsigned char* pushScPA_ele_M_Node;
extern "C" unsigned char* pushScPA_ele_M_Rotation;
extern "C" unsigned char* pushScPA_ele_M_String;
extern "C" unsigned char* pushScPA_ele_M_Time;
extern "C" unsigned char* pushScPA_ele_M_Vec2f;
extern "C" unsigned char* pushScPA_ele_M_Vec3f;

template <class T, class S> inline unsigned char** pushScPA_insert();
template <> inline unsigned char** pushScPA_insert<bool, Single>(){ return &pushScPA_ele_S_Bool;}
template <> inline unsigned char** pushScPA_insert<Color3f, Single>(){ return &pushScPA_ele_S_Color;}
template <> inline unsigned char** pushScPA_insert<float, Single>(){ return &pushScPA_ele_S_Float;}
template <> inline unsigned char** pushScPA_insert<Image, Single>(){ return &pushScPA_ele_S_Image;}
template <> inline unsigned char** pushScPA_insert<int, Single>(){ return &pushScPA_ele_S_Int32;}
template <> inline unsigned char** pushScPA_insert<NodePtr, Single>(){ return &pushScPA_ele_NodePtr;}
template <> inline unsigned char** pushScPA_insert<Rotation, Single>(){ return &pushScPA_ele_S_Rotation;}
template <> inline unsigned char** pushScPA_insert<String, Single>(){ return &pushScPA_ele_S_String;}
template <> inline unsigned char** pushScPA_insert<double, Single>(){ return &pushScPA_ele_S_Time;}
template <> inline unsigned char** pushScPA_insert<Vec2f, Single>(){ return &pushScPA_ele_S_Vec2f;}
template <> inline unsigned char** pushScPA_insert<Vec3f, Single>(){ return &pushScPA_ele_S_Vec3f;}
template <> inline unsigned char** pushScPA_insert<bool, Multiple>(){ return &pushScPA_ele_M_Bool;}
template <> inline unsigned char** pushScPA_insert<Color3f, Multiple>(){ return &pushScPA_ele_M_Color;}
template <> inline unsigned char** pushScPA_insert<float, Multiple>(){ return &pushScPA_ele_M_Float;}
template <> inline unsigned char** pushScPA_insert<Image, Multiple>(){ return &pushScPA_ele_M_Image;}
template <> inline unsigned char** pushScPA_insert<int, Multiple>(){ return &pushScPA_ele_M_Int32;}
template <> inline unsigned char** pushScPA_insert<NodePtr, Multiple>(){ return &pushScPA_ele_M_Node;}
template <> inline unsigned char** pushScPA_insert<Rotation, Multiple>(){ return &pushScPA_ele_M_Rotation;}
template <> inline unsigned char** pushScPA_insert<String, Multiple>(){ return &pushScPA_ele_M_String;}
template <> inline unsigned char** pushScPA_insert<double, Multiple>(){ return &pushScPA_ele_M_Time;}
template <> inline unsigned char** pushScPA_insert<Vec2f, Multiple>(){ return &pushScPA_ele_M_Vec2f;}
template <> inline unsigned char** pushScPA_insert<Vec3f, Multiple>(){ return &pushScPA_ele_M_Vec3f;}

extern "C" unsigned char* resizeContainer_ele_M_Bool;
extern "C" unsigned char* resizeContainer_ele_M_Color;
extern "C" unsigned char* resizeContainer_ele_M_Float;
extern "C" unsigned char* resizeContainer_ele_M_Image;
extern "C" unsigned char* resizeContainer_ele_M_Int32;
extern "C" unsigned char* resizeContainer_ele_M_Node;
extern "C" unsigned char* resizeContainer_ele_M_Rotation;
extern "C" unsigned char* resizeContainer_ele_M_String;
extern "C" unsigned char* resizeContainer_ele_M_Time;
extern "C" unsigned char* resizeContainer_ele_M_Vec2f;
extern "C" unsigned char* resizeContainer_ele_M_Vec3f;

template <class T> inline unsigned char** resizeContainer_insert();
template <> inline unsigned char** resizeContainer_insert<M_Bool>(){ return &resizeContainer_ele_M_Bool;}
template <> inline unsigned char** resizeContainer_insert<M_Color3f>(){ return &resizeContainer_ele_M_Vec3f;}
template <> inline unsigned char** resizeContainer_insert<M_Float>(){ return &resizeContainer_ele_M_Float;}
template <> inline unsigned char** resizeContainer_insert<M_Image>(){ return &resizeContainer_ele_M_Image;}
template <> inline unsigned char** resizeContainer_insert<M_Int>(){ return &resizeContainer_ele_M_Int32;}
template <> inline unsigned char** resizeContainer_insert<M_Node>(){ return &resizeContainer_ele_M_Node;}
template <> inline unsigned char** resizeContainer_insert<M_Rotation>(){ return &resizeContainer_ele_M_Rotation;}
template <> inline unsigned char** resizeContainer_insert<M_String>(){ return &resizeContainer_ele_M_String;}
template <> inline unsigned char** resizeContainer_insert<M_Double>(){ return &resizeContainer_ele_M_Time;}
template <> inline unsigned char** resizeContainer_insert<M_Vec2f>(){ return &resizeContainer_ele_M_Vec2f;}
template <> inline unsigned char** resizeContainer_insert<M_Vec3f>(){ return &resizeContainer_ele_M_Vec3f;}

extern "C" unsigned char* writeElement_ele_M_Bool;
extern "C" unsigned char* writeElement_ele_M_Color;
extern "C" unsigned char* writeElement_ele_M_Float;
extern "C" unsigned char* writeElement_ele_M_Image;
extern "C" unsigned char* writeElement_ele_M_Int32;
extern "C" unsigned char* writeElement_ele_M_Node;
extern "C" unsigned char* writeElement_ele_M_Rotation;
extern "C" unsigned char* writeElement_ele_M_String;
extern "C" unsigned char* writeElement_ele_M_Time;
extern "C" unsigned char* writeElement_ele_M_Vec2f;
extern "C" unsigned char* writeElement_ele_M_Vec3f;

template <class T> inline unsigned char** writeElement_insert();
template <> inline unsigned char** writeElement_insert<M_Bool>(){ return &writeElement_ele_M_Bool;}
template <> inline unsigned char** writeElement_insert<M_Color3f>(){ return &writeElement_ele_M_Vec3f;}
template <> inline unsigned char** writeElement_insert<M_Float>(){ return &writeElement_ele_M_Float;}
template <> inline unsigned char** writeElement_insert<M_Image>(){ return &writeElement_ele_M_Image;}
template <> inline unsigned char** writeElement_insert<M_Int>(){ return &writeElement_ele_M_Int32;}
template <> inline unsigned char** writeElement_insert<M_Node>(){ return &writeElement_ele_M_Node;}
template <> inline unsigned char** writeElement_insert<M_Rotation>(){ return &writeElement_ele_M_Rotation;}
template <> inline unsigned char** writeElement_insert<M_String>(){ return &writeElement_ele_M_String;}
template <> inline unsigned char** writeElement_insert<M_Double>(){ return &writeElement_ele_M_Time;}
template <> inline unsigned char** writeElement_insert<M_Vec2f>(){ return &writeElement_ele_M_Vec2f;}
template <> inline unsigned char** writeElement_insert<M_Vec3f>(){ return &writeElement_ele_M_Vec3f;}

extern "C" unsigned char* assignVal2FieldDirect_ele_S_Bool;
extern "C" unsigned char* assignVal2FieldDirect_ele_S_Color;
extern "C" unsigned char* assignVal2FieldDirect_ele_S_Float;
extern "C" unsigned char* assignVal2FieldDirect_ele_S_Image;
extern "C" unsigned char* assignVal2FieldDirect_ele_S_Int32;
extern "C" unsigned char* assignVal2FieldDirect_ele_NodePtr;
extern "C" unsigned char* assignVal2FieldDirect_ele_S_Rotation;
extern "C" unsigned char* assignVal2FieldDirect_ele_S_String;
extern "C" unsigned char* assignVal2FieldDirect_ele_S_Time;
extern "C" unsigned char* assignVal2FieldDirect_ele_S_Vec2f;
extern "C" unsigned char* assignVal2FieldDirect_ele_S_Vec3f;
extern "C" unsigned char* assignVal2FieldDirect_ele_M_Bool;
extern "C" unsigned char* assignVal2FieldDirect_ele_M_Color;
extern "C" unsigned char* assignVal2FieldDirect_ele_M_Float;
extern "C" unsigned char* assignVal2FieldDirect_ele_M_Image;
extern "C" unsigned char* assignVal2FieldDirect_ele_M_Int32;
extern "C" unsigned char* assignVal2FieldDirect_ele_M_Node;
extern "C" unsigned char* assignVal2FieldDirect_ele_M_Rotation;
extern "C" unsigned char* assignVal2FieldDirect_ele_M_String;
extern "C" unsigned char* assignVal2FieldDirect_ele_M_Time;
extern "C" unsigned char* assignVal2FieldDirect_ele_M_Vec2f;
extern "C" unsigned char* assignVal2FieldDirect_ele_M_Vec3f;

extern "C" unsigned char* assignVal2FieldDirect_ele_S_Bool;
extern "C" unsigned char* assignVal2FieldDirect_ele_S_Color;
extern "C" unsigned char* assignVal2FieldDirect_ele_S_Float;
extern "C" unsigned char* assignVal2FieldDirect_ele_S_Image;
extern "C" unsigned char* assignVal2FieldDirect_ele_S_Int32;
extern "C" unsigned char* assignVal2FieldDirect_ele_NodePtr;
extern "C" unsigned char* assignVal2FieldDirect_ele_S_Rotation;
extern "C" unsigned char* assignVal2FieldDirect_ele_S_String;
extern "C" unsigned char* assignVal2FieldDirect_ele_S_Time;
extern "C" unsigned char* assignVal2FieldDirect_ele_S_Vec2f;
extern "C" unsigned char* assignVal2FieldDirect_ele_S_Vec3f;
extern "C" unsigned char* assignVal2FieldDirect_ele_M_Bool;
extern "C" unsigned char* assignVal2FieldDirect_ele_M_Color;
extern "C" unsigned char* assignVal2FieldDirect_ele_M_Float;
extern "C" unsigned char* assignVal2FieldDirect_ele_M_Image;
extern "C" unsigned char* assignVal2FieldDirect_ele_M_Int32;
extern "C" unsigned char* assignVal2FieldDirect_ele_M_Node;
extern "C" unsigned char* assignVal2FieldDirect_ele_M_Rotation;
extern "C" unsigned char* assignVal2FieldDirect_ele_M_String;
extern "C" unsigned char* assignVal2FieldDirect_ele_M_Time;
extern "C" unsigned char* assignVal2FieldDirect_ele_M_Vec2f;
extern "C" unsigned char* assignVal2FieldDirect_ele_M_Vec3f;

template <class T> inline unsigned char** assignVal2FieldDirect_insert();
template <> inline unsigned char** assignVal2FieldDirect_insert<SFBool>(){ return &assignVal2FieldDirect_ele_S_Bool;}
template <> inline unsigned char** assignVal2FieldDirect_insert<SFColor>(){ return &assignVal2FieldDirect_ele_S_Color;}
template <> inline unsigned char** assignVal2FieldDirect_insert<SFFloat>(){ return &assignVal2FieldDirect_ele_S_Float;}
template <> inline unsigned char** assignVal2FieldDirect_insert<SFImage>(){ return &assignVal2FieldDirect_ele_S_Image;}
template <> inline unsigned char** assignVal2FieldDirect_insert<SFInt32>(){ return &assignVal2FieldDirect_ele_S_Int32;}
template <> inline unsigned char** assignVal2FieldDirect_insert<SFNode>(){ return &assignVal2FieldDirect_ele_NodePtr;}
template <> inline unsigned char** assignVal2FieldDirect_insert<SFRotation>(){ return &assignVal2FieldDirect_ele_S_Rotation;}
template <> inline unsigned char** assignVal2FieldDirect_insert<SFString>(){ return &assignVal2FieldDirect_ele_S_String;}
template <> inline unsigned char** assignVal2FieldDirect_insert<SFTime>(){ return &assignVal2FieldDirect_ele_S_Time;}
template <> inline unsigned char** assignVal2FieldDirect_insert<SFVec2f>(){ return &assignVal2FieldDirect_ele_S_Vec2f;}
template <> inline unsigned char** assignVal2FieldDirect_insert<SFVec3f>(){ return &assignVal2FieldDirect_ele_S_Vec3f;}
template <> inline unsigned char** assignVal2FieldDirect_insert<MFBool>(){ return &assignVal2FieldDirect_ele_M_Bool;}
template <> inline unsigned char** assignVal2FieldDirect_insert<MFColor>(){ return &assignVal2FieldDirect_ele_M_Color;}
template <> inline unsigned char** assignVal2FieldDirect_insert<MFFloat>(){ return &assignVal2FieldDirect_ele_M_Float;}
template <> inline unsigned char** assignVal2FieldDirect_insert<MFImage>(){ return &assignVal2FieldDirect_ele_M_Image;}
template <> inline unsigned char** assignVal2FieldDirect_insert<MFInt32>(){ return &assignVal2FieldDirect_ele_M_Int32;}
template <> inline unsigned char** assignVal2FieldDirect_insert<MFNode>(){ return &assignVal2FieldDirect_ele_M_Node;}
template <> inline unsigned char** assignVal2FieldDirect_insert<MFRotation>(){ return &assignVal2FieldDirect_ele_M_Rotation;}
template <> inline unsigned char** assignVal2FieldDirect_insert<MFString>(){ return &assignVal2FieldDirect_ele_M_String;}
template <> inline unsigned char** assignVal2FieldDirect_insert<MFTime>(){ return &assignVal2FieldDirect_ele_M_Time;}
template <> inline unsigned char** assignVal2FieldDirect_insert<MFVec2f>(){ return &assignVal2FieldDirect_ele_M_Vec2f;}
template <> inline unsigned char** assignVal2FieldDirect_insert<MFVec3f>(){ return &assignVal2FieldDirect_ele_M_Vec3f;}


extern "C" unsigned char* addAssignElement2FieldDirect_ele_S_Bool;
extern "C" unsigned char* addAssignElement2FieldDirect_ele_S_Color;
extern "C" unsigned char* addAssignElement2FieldDirect_ele_S_Float;
extern "C" unsigned char* addAssignElement2FieldDirect_ele_S_Image;
extern "C" unsigned char* addAssignElement2FieldDirect_ele_S_Int32;
extern "C" unsigned char* addAssignElement2FieldDirect_ele_NodePtr;
extern "C" unsigned char* addAssignElement2FieldDirect_ele_S_Rotation;
extern "C" unsigned char* addAssignElement2FieldDirect_ele_S_String;
extern "C" unsigned char* addAssignElement2FieldDirect_ele_S_Time;
extern "C" unsigned char* addAssignElement2FieldDirect_ele_S_Vec2f;
extern "C" unsigned char* addAssignElement2FieldDirect_ele_S_Vec3f;
extern "C" unsigned char* addAssignElement2FieldDirect_ele_M_Bool;
extern "C" unsigned char* addAssignElement2FieldDirect_ele_M_Color;
extern "C" unsigned char* addAssignElement2FieldDirect_ele_M_Float;
extern "C" unsigned char* addAssignElement2FieldDirect_ele_M_Image;
extern "C" unsigned char* addAssignElement2FieldDirect_ele_M_Int32;
extern "C" unsigned char* addAssignElement2FieldDirect_ele_M_Node;
extern "C" unsigned char* addAssignElement2FieldDirect_ele_M_Rotation;
extern "C" unsigned char* addAssignElement2FieldDirect_ele_M_String;
extern "C" unsigned char* addAssignElement2FieldDirect_ele_M_Time;
extern "C" unsigned char* addAssignElement2FieldDirect_ele_M_Vec2f;
extern "C" unsigned char* addAssignElement2FieldDirect_ele_M_Vec3f;

template <class T> inline unsigned char** addAssignElement2FieldDirect_insert();
template <> inline unsigned char** addAssignElement2FieldDirect_insert<SFBool>(){ return &addAssignElement2FieldDirect_ele_S_Bool;}
template <> inline unsigned char** addAssignElement2FieldDirect_insert<SFColor>(){ return &addAssignElement2FieldDirect_ele_S_Color;}
template <> inline unsigned char** addAssignElement2FieldDirect_insert<SFFloat>(){ return &addAssignElement2FieldDirect_ele_S_Float;}
template <> inline unsigned char** addAssignElement2FieldDirect_insert<SFImage>(){ return &addAssignElement2FieldDirect_ele_S_Image;}
template <> inline unsigned char** addAssignElement2FieldDirect_insert<SFInt32>(){ return &addAssignElement2FieldDirect_ele_S_Int32;}
template <> inline unsigned char** addAssignElement2FieldDirect_insert<SFNode>(){ return &addAssignElement2FieldDirect_ele_NodePtr;}
template <> inline unsigned char** addAssignElement2FieldDirect_insert<SFRotation>(){ return &addAssignElement2FieldDirect_ele_S_Rotation;}
template <> inline unsigned char** addAssignElement2FieldDirect_insert<SFString>(){ return &addAssignElement2FieldDirect_ele_S_String;}
template <> inline unsigned char** addAssignElement2FieldDirect_insert<SFTime>(){ return &addAssignElement2FieldDirect_ele_S_Time;}
template <> inline unsigned char** addAssignElement2FieldDirect_insert<SFVec2f>(){ return &addAssignElement2FieldDirect_ele_S_Vec2f;}
template <> inline unsigned char** addAssignElement2FieldDirect_insert<SFVec3f>(){ return &addAssignElement2FieldDirect_ele_S_Vec3f;}
template <> inline unsigned char** addAssignElement2FieldDirect_insert<MFBool>(){ return &addAssignElement2FieldDirect_ele_M_Bool;}
template <> inline unsigned char** addAssignElement2FieldDirect_insert<MFColor>(){ return &addAssignElement2FieldDirect_ele_M_Color;}
template <> inline unsigned char** addAssignElement2FieldDirect_insert<MFFloat>(){ return &addAssignElement2FieldDirect_ele_M_Float;}
template <> inline unsigned char** addAssignElement2FieldDirect_insert<MFImage>(){ return &addAssignElement2FieldDirect_ele_M_Image;}
template <> inline unsigned char** addAssignElement2FieldDirect_insert<MFInt32>(){ return &addAssignElement2FieldDirect_ele_M_Int32;}
template <> inline unsigned char** addAssignElement2FieldDirect_insert<MFNode>(){ return &addAssignElement2FieldDirect_ele_M_Node;}
template <> inline unsigned char** addAssignElement2FieldDirect_insert<MFRotation>(){ return &addAssignElement2FieldDirect_ele_M_Rotation;}
template <> inline unsigned char** addAssignElement2FieldDirect_insert<MFString>(){ return &addAssignElement2FieldDirect_ele_M_String;}
template <> inline unsigned char** addAssignElement2FieldDirect_insert<MFTime>(){ return &addAssignElement2FieldDirect_ele_M_Time;}
template <> inline unsigned char** addAssignElement2FieldDirect_insert<MFVec2f>(){ return &addAssignElement2FieldDirect_ele_M_Vec2f;}
template <> inline unsigned char** addAssignElement2FieldDirect_insert<MFVec3f>(){ return &addAssignElement2FieldDirect_ele_M_Vec3f;}

extern "C" unsigned char* incESI_ele_S_Bool;
extern "C" unsigned char* incESI_ele_S_Color;
extern "C" unsigned char* incESI_ele_S_Float;
extern "C" unsigned char* incESI_ele_S_Image;
extern "C" unsigned char* incESI_ele_S_Int32;
extern "C" unsigned char* incESI_ele_NodePtr;
extern "C" unsigned char* incESI_ele_S_Rotation;
extern "C" unsigned char* incESI_ele_S_String;
extern "C" unsigned char* incESI_ele_S_Time;
extern "C" unsigned char* incESI_ele_S_Vec2f;
extern "C" unsigned char* incESI_ele_S_Vec3f;
extern "C" unsigned char* incESI_ele_M_Bool;
extern "C" unsigned char* incESI_ele_M_Color;
extern "C" unsigned char* incESI_ele_M_Float;
extern "C" unsigned char* incESI_ele_M_Image;
extern "C" unsigned char* incESI_ele_M_Int32;
extern "C" unsigned char* incESI_ele_M_Node;
extern "C" unsigned char* incESI_ele_M_Rotation;
extern "C" unsigned char* incESI_ele_M_String;
extern "C" unsigned char* incESI_ele_M_Time;
extern "C" unsigned char* incESI_ele_M_Vec2f;
extern "C" unsigned char* incESI_ele_M_Vec3f;

template <class T> inline unsigned char** incESI_insert();
template <> inline unsigned char** incESI_insert<S_Bool>(){ return &incESI_ele_S_Bool;}
template <> inline unsigned char** incESI_insert<Color3f>(){ return &incESI_ele_S_Color;}
template <> inline unsigned char** incESI_insert<S_Float>(){ return &incESI_ele_S_Float;}
template <> inline unsigned char** incESI_insert<Image>(){ return &incESI_ele_S_Image;}
template <> inline unsigned char** incESI_insert<S_Int32>(){ return &incESI_ele_S_Int32;}
template <> inline unsigned char** incESI_insert<NodePtr>(){ return &incESI_ele_NodePtr;}
template <> inline unsigned char** incESI_insert<Rotation>(){ return &incESI_ele_S_Rotation;}
template <> inline unsigned char** incESI_insert<String>(){ return &incESI_ele_S_String;}
template <> inline unsigned char** incESI_insert<double>(){ return &incESI_ele_S_Time;}
template <> inline unsigned char** incESI_insert<Vec2f>(){ return &incESI_ele_S_Vec2f;}
template <> inline unsigned char** incESI_insert<Vec3f>(){ return &incESI_ele_S_Vec3f;}
template <> inline unsigned char** incESI_insert<M_Bool>(){ return &incESI_ele_M_Bool;}
template <> inline unsigned char** incESI_insert<M_Color3f>(){ return &incESI_ele_M_Color;}
template <> inline unsigned char** incESI_insert<M_Float>(){ return &incESI_ele_M_Float;}
template <> inline unsigned char** incESI_insert<M_Image>(){ return &incESI_ele_M_Image;}
template <> inline unsigned char** incESI_insert<M_Int>(){ return &incESI_ele_M_Int32;}
template <> inline unsigned char** incESI_insert<M_Node>(){ return &incESI_ele_M_Node;}
template <> inline unsigned char** incESI_insert<M_Rotation>(){ return &incESI_ele_M_Rotation;}
template <> inline unsigned char** incESI_insert<M_String>(){ return &incESI_ele_M_String;}
template <> inline unsigned char** incESI_insert<M_Double>(){ return &incESI_ele_M_Time;}
template <> inline unsigned char** incESI_insert<M_Vec2f>(){ return &incESI_ele_M_Vec2f;}
template <> inline unsigned char** incESI_insert<M_Vec3f>(){ return &incESI_ele_M_Vec3f;}

extern "C" unsigned char* negate_ele_S_Bool;
extern "C" unsigned char* negate_ele_S_Color;
extern "C" unsigned char* negate_ele_S_Float;
extern "C" unsigned char* negate_ele_S_Image;
extern "C" unsigned char* negate_ele_S_Int32;
extern "C" unsigned char* negate_ele_NodePtr;
extern "C" unsigned char* negate_ele_S_Rotation;
extern "C" unsigned char* negate_ele_S_String;
extern "C" unsigned char* negate_ele_S_Time;
extern "C" unsigned char* negate_ele_S_Vec2f;
extern "C" unsigned char* negate_ele_S_Vec3f;

template <class T> inline unsigned char** negate_insert();
template <> inline unsigned char** negate_insert<S_Bool>(){ return &negate_ele_S_Bool;}
template <> inline unsigned char** negate_insert<Color3f>(){ return &negate_ele_S_Color;}
template <> inline unsigned char** negate_insert<S_Float>(){ return &negate_ele_S_Float;}
template <> inline unsigned char** negate_insert<Image>(){ return &negate_ele_S_Image;}
template <> inline unsigned char** negate_insert<S_Int32>(){ return &negate_ele_S_Int32;}
template <> inline unsigned char** negate_insert<NodePtr>(){ return &negate_ele_NodePtr;}
template <> inline unsigned char** negate_insert<Rotation>(){ return &negate_ele_S_Rotation;}
template <> inline unsigned char** negate_insert<String>(){ return &negate_ele_S_String;}
template <> inline unsigned char** negate_insert<double>(){ return &negate_ele_S_Time;}
template <> inline unsigned char** negate_insert<Vec2f>(){ return &negate_ele_S_Vec2f;}
template <> inline unsigned char** negate_insert<Vec3f>(){ return &negate_ele_S_Vec3f;}
template <> inline unsigned char** negate_insert<M_Bool>(){ throw;}
template <> inline unsigned char** negate_insert<M_Color3f>(){ throw;}
template <> inline unsigned char** negate_insert<M_Float>(){ throw;}
template <> inline unsigned char** negate_insert<M_Image>(){ throw;}
template <> inline unsigned char** negate_insert<M_Int>(){ throw;}
template <> inline unsigned char** negate_insert<M_Node>(){ throw;}
template <> inline unsigned char** negate_insert<M_Rotation>(){ throw;}
template <> inline unsigned char** negate_insert<M_String>(){ throw;}
template <> inline unsigned char** negate_insert<M_Double>(){ throw;}
template <> inline unsigned char** negate_insert<M_Vec2f>(){ throw;}
template <> inline unsigned char** negate_insert<M_Vec3f>(){ throw;}

extern "C" unsigned char* not_ele_S_Bool;
template <class T> inline unsigned char** not_insert();
template <> inline unsigned char** not_insert<S_Bool>(){ return &not_ele_S_Bool;}

extern "C" unsigned char* sin_ele_S_Float;
extern "C" unsigned char* sin_ele_S_Time;

extern "C" unsigned char* cos_ele_S_Float;
extern "C" unsigned char* cos_ele_S_Time;

extern "C" unsigned char* asin_ele_S_Float;
extern "C" unsigned char* asin_ele_S_Time;

extern "C" unsigned char* sqrt_ele_S_Float;
extern "C" unsigned char* sqrt_ele_S_Time;

extern "C" unsigned char* fabs_ele_S_Float;
extern "C" unsigned char* fabs_ele_S_Time;

extern "C" unsigned char* pow_ele_S_Float;
extern "C" unsigned char* pow_ele_S_Time;

extern "C" unsigned char* getTime_ele_S_Time;

extern "C" unsigned char* end_ele_M_Vec3f;

extern "C" unsigned char* normalize_insert_S_Vec3f;

typedef M_ <unsigned char> M_Byte;

template <class T> inline T* virtAlloc(int bufLen){
	return reinterpret_cast<T*>(VirtualAlloc(NULL, sizeof(T)*bufLen, MEM_COMMIT, PAGE_READWRITE ));
};
template <class T> inline void virtFree(T* p){
	VirtualFree(p,0,MEM_RELEASE);
};

template <typename T>class M_DEP_ {
class Length{
	friend M_<T>;
	friend M_DEP_;
protected:
	unsigned bufLen;
	unsigned rawLen;
	T* pointer;
public:
	Length():
		rawLen(0), bufLen(0), pointer(0)
	{
	}
	Length(const T& f0):
		rawLen(1),
		bufLen(STANDARDBUFLEN),
		pointer(virtAlloc(STANDARDBUFLEN))
	{
		pointer[0]=f0;
	}
	Length(const T& f0,const T& f1):
		rawLen(2),
		bufLen(STANDARDBUFLEN),
		pointer(virtAlloc(STANDARDBUFLEN))
	{
		pointer[0]=f0;
		pointer[1]=f1;
	}
	Length(const T& f0,const T& f1,const T& f2):
		rawLen(3),
		bufLen(STANDARDBUFLEN),
		pointer(virtAlloc(STANDARDBUFLEN))
	{
		pointer[0]=f0;
		pointer[1]=f1;
		pointer[2]=f2;
	}
	Length(const T& f0,const T& f1,const T& f2, const T& f3):
		rawLen(4),
		bufLen(STANDARDBUFLEN),
		pointer(virtAlloc(STANDARDBUFLEN))
	{
		pointer[0]=f0;
		pointer[1]=f1;
		pointer[2]=f2;
		pointer[3]=f3;
	}
	Length(const T& f0,const T& f1,const T& f2, const T& f3, const T& f4):
		rawLen(5),
		bufLen(STANDARDBUFLEN),
		pointer(virtAlloc(STANDARDBUFLEN))
	{
		pointer[0]=f0;
		pointer[1]=f1;
		pointer[2]=f2;
		pointer[3]=f3;
		pointer[4]=f4;
	}
	Length(const T& f0,const T& f1,const T& f2, const T& f3, const T& f4, const T& f5):
		rawLen(6),
		bufLen(STANDARDBUFLEN),
		pointer(virtAlloc(STANDARDBUFLEN))
	{
		pointer[0]=f0;
		pointer[1]=f1;
		pointer[2]=f2;
		pointer[3]=f3;
		pointer[4]=f4;
		pointer[5]=f5;
	}
	Length(const T& f0,const T& f1,const T& f2, const T& f3, const T& f4, const T& f5, const T& f6):
		rawLen(7),
		bufLen(STANDARDBUFLEN),
		pointer(virtAlloc(STANDARDBUFLEN))
	{
		pointer[0]=f0;
		pointer[1]=f1;
		pointer[2]=f2;
		pointer[3]=f3;
		pointer[4]=f4;
		pointer[5]=f5;
		pointer[6]=f6;
	}
	Length(const T& f0,const T& f1,const T& f2, const T& f3, const T& f4, const T& f5, const T& f6, const T& f7):
		rawLen(8),
		bufLen(STANDARDBUFLEN),
		pointer(virtAlloc(STANDARDBUFLEN))
	{
		pointer[0]=f0;
		pointer[1]=f1;
		pointer[2]=f2;
		pointer[3]=f3;
		pointer[4]=f4;
		pointer[5]=f5;
		pointer[6]=f6;
		pointer[7]=f7;
	}
	Length(const T& f0,const T& f1,const T& f2, const T& f3, const T& f4, const T& f5, const T& f6, const T& f7, const T& f8):
		rawLen(9),
		bufLen(STANDARDBUFLEN),
		pointer(virtAlloc(STANDARDBUFLEN))
	{
		pointer[0]=f0;
		pointer[1]=f1;
		pointer[2]=f2;
		pointer[3]=f3;
		pointer[4]=f4;
		pointer[5]=f5;
		pointer[6]=f6;
		pointer[7]=f7;
		pointer[8]=f8;
	}
	Length(const T& f0,const T& f1,const T& f2, const T& f3, const T& f4, const T& f5, const T& f6, const T& f7, const T& f8, const T& f9):
		rawLen(10),
		bufLen(STANDARDBUFLEN),
		pointer(virtAlloc(STANDARDBUFLEN))
	{
		pointer[0]=f0;
		pointer[1]=f1;
		pointer[2]=f2;
		pointer[3]=f3;
		pointer[4]=f4;
		pointer[5]=f5;
		pointer[6]=f6;
		pointer[7]=f7;
		pointer[8]=f8;
		pointer[9]=f9;
	}
	Length(const T& f0,const T& f1,const T& f2, const T& f3, const T& f4, const T& f5, const T& f6, const T& f7, const T& f8, const T& f9, const T& f10):
		rawLen(11),
		bufLen(STANDARDBUFLEN),
		pointer(virtAlloc(STANDARDBUFLEN))
	{
		pointer[0]=f0;
		pointer[1]=f1;
		pointer[2]=f2;
		pointer[3]=f3;
		pointer[4]=f4;
		pointer[5]=f5;
		pointer[6]=f6;
		pointer[7]=f7;
		pointer[8]=f8;
		pointer[9]=f9;
		pointer[10]=f10;
	}
	Length(const T& f0,const T& f1,const T& f2, const T& f3, const T& f4, const T& f5, const T& f6, const T& f7, const T& f8, const T& f9, const T& f10, const T& f11):
		rawLen(12),
		bufLen(STANDARDBUFLEN),
		pointer(virtAlloc(STANDARDBUFLEN))
	{
		pointer[0]=f0;
		pointer[1]=f1;
		pointer[2]=f2;
		pointer[3]=f3;
		pointer[4]=f4;
		pointer[5]=f5;
		pointer[6]=f6;
		pointer[7]=f7;
		pointer[8]=f8;
		pointer[9]=f9;
		pointer[10]=f10;
		pointer[11]=f11;
	}
	Length(const T& f0,const T& f1,const T& f2, const T& f3, const T& f4, const T& f5, const T& f6, const T& f7, const T& f8, const T& f9, const T& f10, const T& f11, const T& f12):
		rawLen(13),
		bufLen(STANDARDBUFLEN),
		pointer(virtAlloc(STANDARDBUFLEN))
	{
		pointer[0]=f0;
		pointer[1]=f1;
		pointer[2]=f2;
		pointer[3]=f3;
		pointer[4]=f4;
		pointer[5]=f5;
		pointer[6]=f6;
		pointer[7]=f7;
		pointer[8]=f8;
		pointer[9]=f9;
		pointer[10]=f10;
		pointer[11]=f11;
		pointer[12]=f12;
	}
	Length(const T& f0,const T& f1,const T& f2, const T& f3, const T& f4, const T& f5, const T& f6, const T& f7, const T& f8, const T& f9, const T& f10, const T& f11, const T& f12, const T& f13):
		rawLen(14),
		bufLen(STANDARDBUFLEN),
		pointer(virtAlloc(STANDARDBUFLEN))
	{
		pointer[0]=f0;
		pointer[1]=f1;
		pointer[2]=f2;
		pointer[3]=f3;
		pointer[4]=f4;
		pointer[5]=f5;
		pointer[6]=f6;
		pointer[7]=f7;
		pointer[8]=f8;
		pointer[9]=f9;
		pointer[10]=f10;
		pointer[11]=f11;
		pointer[12]=f12;
		pointer[13]=f13;
	}
	Length(const T& f0,const T& f1,const T& f2, const T& f3, const T& f4, const T& f5, const T& f6, const T& f7, const T& f8, const T& f9, const T& f10, const T& f11, const T& f12, const T& f13, const T& f14, const T& f15):
		rawLen(16),
		bufLen(STANDARDBUFLEN),
		pointer(virtAlloc(STANDARDBUFLEN))
	{
		pointer[0]=f0;
		pointer[1]=f1;
		pointer[2]=f2;
		pointer[3]=f3;
		pointer[4]=f4;
		pointer[5]=f5;
		pointer[6]=f6;
		pointer[7]=f7;
		pointer[8]=f8;
		pointer[9]=f9;
		pointer[10]=f10;
		pointer[11]=f11;
		pointer[12]=f12;
		pointer[13]=f13;
		pointer[14]=f14;
		pointer[15]=f15;
	}
	Length(IntList2& par):
		rawLen(par.len*sizeof(int)/sizeof(T))
	{
		if(par.len % rawLen){
			throw this;
		}
		bufLen=rawLen;
		pointer=virtAlloc(bufLen);
		for(int i=0; i<par.len; i++){
			reinterpret_cast<int*>(pointer)[i]=par.p[i];
		}
	}
	Length(FloatList2& par):
		rawLen(par.len*sizeof(float)/sizeof(T))
	{
		if(par.len % rawLen){
			throw this;
		}
		bufLen=rawLen;
		pointer=virtAlloc(bufLen);
		for(int i=0; i<par.len; i++){
			reinterpret_cast<float*>(pointer)[i]=par.p[i];
		}
	}
	Length(EventList& par):
		rawLen(par.len*sizeof(InputEvent*)/sizeof(InputEvent*))
	{
		bufLen=rawLen;
		pointer=virtAlloc(bufLen);
		for(int i=0; i<par.len; i++){
			pointer[i]=par.p[i];
		}
	}
	Length(const M_DEP_& param):
		rawLen(param.length.rawLen),
		bufLen(param.length.bufLen),
		pointer(virtAlloc(bufLen))
	{
		for(unsigned i=0;i<rawLen;i++)pointer[i]=param.length.pointer[i];
	}
	~Length(){
		virtFree(pointer);
	}
	int operator =(const Length& par){
		unsigned newLength=par.rawLen;
		unsigned i;
		if(bufLen<newLength){
			bufLen=newLength*2;
			T* newPointer=virtAlloc(bufLen);
			for(i=0;i<rawLen;i++){
				newPointer[i]=pointer[i];
			}
			virtFree(pointer);
			pointer=newPointer;
		}
		for(i=rawLen;i<newLength;i++){
			pointer[i]=T();
		}
		rawLen=newLength;
		return rawLen;
	}
	inline int operator =(const int& newLength){
		unsigned i;
		if(bufLen<(unsigned)newLength){
			bufLen=newLength*2;
			T* newPointer=virtAlloc<T>(bufLen);
			for(i=0;i<rawLen;i++){
				newPointer[i]=pointer[i];
			}
			virtFree(pointer);
			pointer=newPointer;
		}
		for(i=rawLen;i<(unsigned)newLength;i++){
			pointer[i]=T();
		}
		rawLen=newLength;
		return rawLen;
	}
	int operator ++ (int){
		unsigned newLength=rawLen+1;

		if(bufLen<newLength){
			bufLen=newLength*2;
			T* newPointer=virtAlloc(bufLen);
			for(unsigned i=0;i<rawLen;i++){
				newPointer[i]=pointer[i];
			}
			virtFree(pointer);
			pointer=newPointer;
		}
		pointer[rawLen]=T();
		rawLen=newLength;
		return rawLen-1;
	}
	int operator -- (int){
		rawLen--;
		if((int)rawLen<0)throw this;
		return rawLen;
	}
	int operator -= (int par){
		rawLen-=par;
		if((int)rawLen<0)throw this;
		return rawLen;
	}
	int operator += (int par){
		unsigned newLength=par+rawLen;

		if(bufLen<newLength){
			bufLen=newLength*2;
			T* newPointer=virtAlloc<T>(bufLen);
			for(unsigned i=0;i<rawLen;i++){
				newPointer[i]=pointer[i];
			}
			virtFree(pointer);
			pointer=newPointer;
		}
		for(unsigned i=rawLen;i<newLength;i++){
			pointer[i]=T();
		}
		rawLen=newLength;
		return newLength;
	}
	operator int ()const{
		return rawLen;
	}
};

public:
	M_DEP_():
		length()
	{}
	M_DEP_(const T& f0):
		length(f0)
	{}
	M_DEP_(const T& f0,const T& f1):
		length(f0,f1)
	{}
	M_DEP_(const T& f0,const T& f1,const T& f2):
		length(f0,f1,f2)
	{}
	M_DEP_(const T& f0,const T& f1,const T& f2,const T& f3):
		length(f0,f1,f2,f3)
	{}
	M_DEP_(const T& f0,const T& f1,const T& f2, const T& f3, const T& f4):
		length(f0,f1,f2,f3,f4)
	{}
	M_DEP_(const T& f0,const T& f1,const T& f2, const T& f3, const T& f4, const T& f5):
		length(f0,f1,f2,f3,f4,f5)
	{}
	M_DEP_(const T& f0,const T& f1,const T& f2, const T& f3, const T& f4, const T& f5, const T& f6):
		length(f0,f1,f2,f3,f4,f5,f6)
	{}
	M_DEP_(const T& f0,const T& f1,const T& f2, const T& f3, const T& f4, const T& f5, const T& f6, const T& f7):
		length(f0,f1,f2,f3,f4,f5,f6,f7)
	{}
	M_DEP_(const T& f0,const T& f1,const T& f2, const T& f3, const T& f4, const T& f5, const T& f6, const T& f7, const T& f8):
		length(f0,f1,f2,f3,f4,f5,f6,f7,f8)
	{}
	M_DEP_(const T& f0,const T& f1,const T& f2, const T& f3, const T& f4, const T& f5, const T& f6, const T& f7, const T& f8, const T& f9):
		length(f0,f1,f2,f3,f4,f5,f6,f7,f8,f9)
	{}
	M_DEP_(const T& f0,const T& f1,const T& f2, const T& f3, const T& f4, const T& f5, const T& f6, const T& f7, const T& f8, const T& f9, const T& f10):
		length(f0,f1,f2,f3,f4,f5,f6,f7,f8,f9,f10)
	{}
	M_DEP_(const T& f0,const T& f1,const T& f2, const T& f3, const T& f4, const T& f5, const T& f6, const T& f7, const T& f8, const T& f9, const T& f10, const T& f11):
		length(f0,f1,f2,f3,f4,f5,f6,f7,f8,f9,f10,f11)
	{}
	M_DEP_(const T& f0,const T& f1,const T& f2, const T& f3, const T& f4, const T& f5, const T& f6, const T& f7, const T& f8, const T& f9, const T& f10, const T& f11, const T& f12):
		length(f0,f1,f2,f3,f4,f5,f6,f7,f8,f9,f10,f11,f12)
	{}
	M_DEP_(const T& f0,const T& f1,const T& f2, const T& f3, const T& f4, const T& f5, const T& f6, const T& f7, const T& f8, const T& f9, const T& f10, const T& f11, const T& f12, const T& f13):
		length(f0,f1,f2,f3,f4,f5,f6,f7,f8,f9,f10,f11,f12,f13)
	{}
	M_DEP_(const T& f0,const T& f1,const T& f2, const T& f3, const T& f4, const T& f5, const T& f6, const T& f7, const T& f8, const T& f9, const T& f10, const T& f11, const T& f12, const T& f13, const T& f14, const T& f15):
		length(f0,f1,f2,f3,f4,f5,f6,f7,f8,f9,f10,f11,f12,f13,f14,f15)
	{}
	M_DEP_(IntList2& par):
		length(par)
	{}
	M_DEP_(FloatList2& par):
		length(par)
	{}
	M_DEP_(EventList& par):
		length(par)
	{}
	M_DEP_(const M_DEP_& param):
		length(param)
	{}
	~M_DEP_(){
	}
	M_DEP_& operator= (const M_DEP_& par){
		if(par.length.pointer==0){
			length.rawLen=0;
		}
		else if(par.length.pointer!=length.pointer){
			if(length.bufLen<par.length.bufLen){
				virtFree(length.pointer);
				length.rawLen=par.length.rawLen;
				length.bufLen=length.rawLen*2;
				length.pointer=virtAlloc(length.bufLen);
			}
			else{
				length.rawLen=par.length.rawLen;
			}
			for(unsigned i=0;i<length.rawLen;i++)length.pointer[i]=par.length.pointer[i];
		}
		return *this;
	}
	//const M_DEP_& operator =(const SF<M_DEP_,	T>& par);

	M_DEP_& operator +=(T par){
		unsigned newLength=length.rawLen+1;

		if(length.bufLen<newLength){
			length.bufLen=newLength*2;
			T* newPointer=virtAlloc<T>(length.bufLen);
			for(unsigned i=0;i<length.rawLen;i++){
				newPointer[i]=length.pointer[i];
			}
			virtFree(length.pointer);
			length.pointer=newPointer;
		}
		length.pointer[length.rawLen]=par;
		length.rawLen=newLength;
		return *this;
	}
	M_DEP_& operator +=(const M_DEP_& par1){
		unsigned i;
		unsigned newLength=par1.length.rawLen+length.rawLen;
		if(length.bufLen<newLength){
			length.bufLen=newLength*2;
			T* newPointer=virtAlloc(length.bufLen);
			for(unsigned i=0;i<length.rawLen;i++){
				newPointer[i]=length.pointer[i];
			}
			virtFree(length.pointer);
			length.pointer=newPointer;
		}
		for(i=0;i<par1.length.rawLen;i++){
			length.pointer[i+length.rawLen]=par1.length.pointer[i];
		}
		length.rawLen=newLength;
		return *this;
	}
	M_DEP_& operator +=(const M_<T>& par1){
		unsigned i;
		unsigned newLength=par1.length.rawLen+length.rawLen;
		if(length.bufLen<newLength){
			length.bufLen=newLength*2;
			T* newPointer=virtAlloc<T>(length.bufLen);
			for(unsigned i=0;i<length.rawLen;i++){
				newPointer[i]=length.pointer[i];
			}
			virtFree(length.pointer);
			length.pointer=newPointer;
		}
		for(i=0;i<par1.length.rawLen;i++){
			length.pointer[i+length.rawLen]=par1.length.pointer[i];
		}
		length.rawLen=newLength;
		return *this;
	}
	M_DEP_ operator +(const T& par)const{
		M_DEP_ m=*this;
		m+=par;
		return m;
	}
	M_DEP_ operator +(const M_DEP_& par)const{
		M_DEP_ m=*this;
		m+=par;
		return m;
	}
    const char *showReason() const { 
		return "Exception in M_DEP_ class."; 
	}
	int size()const{
		return length.rawLen;
	}
	void setElement(const int index, const T& val){
		length.pointer[index]=val;
	}
	T& operator [](const int& index){
		if((unsigned)index>=length.rawLen){
			throw this;
			/*
			unsigned newLength=index+1;
			unsigned i;
			if(length.bufLen<newLength){
				length.bufLen=newLength*2;
				T* newPointer=virtAlloc(length.bufLen);
				for(i=0;i<length.rawLen;i++){
					newPointer[i]=length.pointer[i];
				}
				virtFree(length.pointer);
				length.pointer=newPointer;
			}
			for(i=length.rawLen;i<newLength;i++){
				length.pointer[i]=T();
			}
			length.rawLen=newLength;
			*/
		}
		return length.pointer[index];
	}
	const T& operator [](const int& index)const{
		if((unsigned)index>=length.rawLen){
			throw this;
		}
		return length.pointer[index];
	}
	const T* getArray()const{
		return length.pointer;
	}
	T* getArray(){
		return length.pointer;
	}
	bool operator ==(const M_DEP_& par)const{
		if(rawLen!=par->rawLen){
			return false;
		}
		for(int i=length; --i>=0;){
			if(length.pointer[i]!=par[i]){
				return false;
			}
		}
		return true;
	}
	bool operator !=(const M_DEP_& par)const{
		if(length!=par.length){
			return true;
		}
		for(int i=length; --i>=0;){
			if(length.pointer[i]!=par[i]){
				return true;
			}
		}
		return false;
	}
	void toString(String& str)const{
		str.cat(*this);
	}
	void freeMem(){
		length.rawLen=0;
		length.bufLen=0;
		virtFree(length.pointer);
		length.pointer=0;
	}
	void protect(){
		if(length.rawLen > 0){
			DWORD dwOldProtect = 0;
			if(!VirtualProtect(length.pointer, length.rawLen*sizeof(T), PAGE_EXECUTE_READ, &dwOldProtect)){
				dLog("protect() failed");
				throw this;
			}
		}
	}
	void enableWrite(){
		if(length.bufLen > 0){
			DWORD dwOldProtect = 0;
			if(!VirtualProtect(length.pointer, length.bufLen*sizeof(T), PAGE_READWRITE, &dwOldProtect)){
				dLog("enableWrite() failed");
				throw this;
			}
		}
	}

	void getField(const String& str,int& i);

	//'class M_DEP_<int> &(class SF<class M_DEP_<int>,int,69> &)'
//inline    M_DEP_<T, type> &	operator =(SF <M_DEP_<T,type>,T,type>& par);
//inline    M_DEP_<T, type> &	operator +=(const T par);
	void toMNodeString(String& str, bool newList=true);
	void removeNode(T par1);
	void removeNode(M_DEP_& par1);
	Art*& refPart(int i)const;
	int& refIndex(int i)const;
	int& refTexIndex(int i)const;
	int getTriangleLength()const{
		return length*3;
	};
	bool write(String fileName)const;

	void rotate(Rotation rot,int offset=0);
	void rotate( float x, float y, float z, float angle, int offset=0);
	void translate(T tran,int offset=0);
	void translate( float x, float y, float z, int offset=0);
	void scale(float xsc,float ysc,float zsc);
	void translateToZero(int offset,int numOf, Vec3f& center);
	void insertElement(int targetIndex, T value);
	void deleteElement(int targetIndex);
	void deleteElement(int from, int numOf);

	void toCppString(String& str);


	void translateToZero(int offset,int numOf,Rotation& center);
	void multMatrixLeft(Matrix& matrix);
	void setUpRoute( int& yPos);
	int getIndex(String str)const;
	Length length;
};

class Operation {
public:
	virtual void doOperation()=0;
};
typedef M_ <Operation*> M_Operation;
typedef M_DEP_ <unsigned char> M_DEP_Byte;
typedef M_ <Word*> M_Word;
typedef M_ <AReturn*> M_AReturn;
typedef M_ <ACall> M_ACall;
typedef M_ <Asm2CppAddy> M_Asm2CppAddy;


class Function : public Node, public MField1 {
	public:
	int getType()const{return FUNCTION;}
	int getFLen()const{return size();};
	Field* getField(const int i)const{return (*this)[i];}
	int getFieldGroupIndex(const int &i){throw this;}
	void setFieldGroupIndex(const int &i, const int &val){throw this;}
	M_Int* getIndexPtr(){return nullPtr;}

	MFNode codeLine;

	Function():
		MField1(
			&codeLine
		),

		codeLine(M_Node()),
/////////
		returnType(0),
		sizeOfLocalAndParam(0),
		localSize(0),
		paramLen(0),
		bodyTextBegin(0),
		startCodeIdx(0),
		startCodePtr(0)
	{
	}
	String functionName;
	static M_DEP_Byte compiled;
	static M_Byte constData;
	M_Word m_word;
	static M_Int breakStack;
	static M_Int continueStack;
	static bool reset_m_word;

	M_L_ m_local;
	M_Int stackOffset;
	int sizeOfLocalAndParam;
	int localSize;
	int returnType;
	int paramLen;
	int bodyTextBegin;
	M_ACall call2ndCompile;
	int startCodeIdx;
	unsigned char* startCodePtr;
	static Node* selectedLine;
	static bool readyForOutput;

	static void insertDWord(int mark, int data);
	void catConstQWord(unsigned char* data);

	void Function::decodeHeader(const String& str,int& i, int& bracketCntr);
	void Function::newDecodeHeader(const String& str, int& i, int& bracketCntr);
	void Function::decodeBody(const String& str);
	inline bool dealWithEndOfLine(const String& str, int& i, CodeLine*& codeLine, const int& remBracketCntr, int& bracketCntr);
	void Function::toString(String& str, bool newList=true);
	void processForOutput(M_Node& processed, int& n, Script* gs);
	void Function::addFunctionToList(const String& str, int& i);
	L_Variable* Function::isLocalVariable(const String& label);
	void Function::decodeCppLine(M_Node& clParent, int& codeLineIdx, CodeLine& cl, const String& str, int& i, int& bracketCntr);
	inline void Function::doBody(CodeLine& cl, const String& str, int& i, int& bracketCntr);
	void decodeTextCell(const String& str, int& i, int& bracketCntr);
	void Function::replaceNodeRef(NodePtr from, Node* to);
	void Function::replaceNodeFieldRef(NodePtr from, Node* to, Field* fieldTo);
	void compileScalePositionInterpolator(const Art* part, Matrix& matrix);
	void compileOrientationInterpolator(const Art* part, Rotation rotation);
	void deleteControlPointFromAllMorphers(Art* part, int idx);


template <class T> inline void CopyImmediateToScratchPad_insert(const String& str, int& i){throw;};
void inline CopyImmediateToScratchPad_insert(const int& data);
template <class T> inline void copyPiToScratchPad_insert(const String& str, int& i){throw;};

void copyMachineCode(unsigned char** address);
void copyMachineCode(unsigned char* address, int length);
template <typename C> inline void copyMachineCodeT(){
	copyMachineCode(C, sizeof(C));
}
template <class T> inline void copyCode_writeElement_ref_insert();

template <class T> inline void copyScratchPadToGlobal_insert2(void* valAddress);
template <class T> inline void copyScratchPadToLocal_insert2(const int& variableIndex);
template <class T> inline void copyScratchPadToLocal_insert2(const int& variableIndex, const int& offset);

template <class T> inline void addScratchPadToGlobal_insert2(void* valAddress);
template <class T> inline void subScratchPadToGlobal_insert2(void* valAddress);

template <class T> inline void addScratchPadToLocal_insert2(const int& variableIndex);
template <class T> inline void addScratchPadToLocal_insert2(const int& variableIndex, const int& offset);
template <class T> inline void subScratchPadToLocal_insert2(const int& variableIndex);

inline void Function::decodeFloatComponent(const int& varIndex, const char* componentName, const int& offset, const String& str, int& i, int& bracketCntr);


//template <class T, class S> inline void copyGlobalToScratchPad_insert2(G_Variable* gv, const String& str, int& i, int& bracketCntr);
template <class T, class S> inline void copyToScratchPad(G_Variable* v, const String& str, int& i, int& bracketCntr);
template <class T, class S> inline void copyToScratchPad(L_Variable* v, const String& str, int& i, int& bracketCntr);
template <class T, class S> inline void getOperandAfterGlobal(V_<T,S>* node, const String& str, int& i, int& bracketCntr );
template <class T, class S> inline void getOperandAfterGlobal(const T& value, const String& str, int& i, int& bracketCntr );
template <class T, class S> inline void getOperandAfterLocal(const int& varIndex, const String& str, int& i, int& bracketCntr );
//template <class T> inline void copyLocalToScratchPad_insert2(const int& varIndex, const String& str, int& i, int& bracketCntr);
template <class T, class S> inline void copyToScratchPad(Field* fieldPtr, const String& str, int& i, int& bracketCntr);



int getOperand(const String& str, int& i, int& bracketCntr, int optimizeFor=-1);
template <class T, class S> int doOperator(const String& str, int& i, int& bracketCntr);
template <class T> inline void wordDivide();

//int getExpression(const String& str, int& i, int& bracketCntr);
int getFirstOperand(const String& str, int& i, int& bracketCntr, int optimizeFor=-1);
template <class T, class S> int doHiddenBrackOperator(const String& str, int& i, int& bracketCntr);
inline int getNumber(const String& str, int& i, int& bracketCntr, int optimizeFor=-1);


//template <class T, class S> void getHiddenBrackOperator(const String& str, int& i, int& bracketCntr);
//template <class T, class S> void getOperand(const String& str, int& i, int& bracketCntr);
//template <class T, class S> inline void getFirstOperand(const String& str, int& i, int& bracketCntr );
//template <> inline void getFirstOperand<S_Bool,Single>(const String& str, int& i,int& bracketCntr);
//template <class T, class S> void getNextOperator(const String& str, int& i, int& bracketCntr);


template <class T, class S> void passToFunctionByVal(const String& str, int& i, int& bracketCntr, int& justThisFunctionCallOffset);
template <class T, class S> void passToFunctionByRef(const String& str, int& i, int& bracketCntr, int& justThisFunctionCallOffset);

template <class T, class S> inline int doDivisionOperator(int& opTypeB);
template <class T, class S> inline int doModuloOperator(int& opTypeB);
template <class T, class S> inline int doMultiplyOperator(int& opTypeB);

template <class T, class S> inline int doSubtractionOperator(int& opTypeB);
template <class T, class S> inline int doAdditionOperator(int& opTypeB);


void Function::getArg(Function& callFunction, const String& str, int& i, int& bracketCntr);
template <class T> inline bool builtInFunction(const String& label, const String& str, int& i, int& bracketCntr, int& opType){
	return false;
}

template <class T> inline void postIncLocal_insert2(const int& varIndex);
template <class T> inline void postDecLocal_insert2(const int& varIndex);


template <class T> inline void decodeAssign(V_<T,Multiple,Global>& v, const String& str, int& i);
template <class T> inline void decodeAssign(V_<T,Single,Global>& v, const String& str, int& i);
template <class T> inline void decodeAssign(V_<T,Single,Local>& v, const String& str, int& i);
template <class T> inline void decodeAddAssign(V_<T,Single,Global>& v, const String& str, int& i);
template <class T> inline void decodeAddAssign(V_<T,Multiple,Global>& v, const String& str, int& i);
template <class T> inline void decodeAddAssign(V_<T,Single,Local>& v, const String& str, int& i);
template <class T> inline void decodeSubAssign(V_<T,Single,Global>& v, const String& str, int& i);
template <class T> inline void decodeSubAssign(V_<T,Single,Local>& v, const String& str, int& i);

template <class T> inline void globalAssignElement(M_<T>& value, const String& str, int& i);
template <class T> inline void refAssignElement(V_<T,Multiple,Local>& ref, const String& str, int& i);
template <class T,class Single> inline void decodeSource(L_Variable* v, const String& str, int& i, int& bracketCntr);
template <class T,class Single> inline void decodeSource(G_Variable* v, const String& str, int& i, int& bracketCntr);

template <class T, class S> inline void decodeField(Field* fieldPtr, const String& str, int& i, int& bracketCntr);
template <class T, class S> inline void decodeMultipleGlobal(G_Variable* gv, const String& str, int& i, int& bracketCntr);
template <class S, class T> inline void decodePostSquareBrack(G_Variable* gv, const String& str, int& i, int& bracketCntr);
template <class T> void decodeGlobalElementProperty(M_<T>& value, const String& str, int& i);
template <class C> void decodeGlobalElementComponent(M_Vec3f& value, const String& str, int& i);

template <class S, class T> inline void addNewLocal(String& str);
template <> inline void addNewLocal<S_Bool,Single>(String& str){	m_local+=new L_S_Bool(str, m_local.length);}
template <> inline void addNewLocal<Color3f,Single>(String& str){	m_local+=new L_S_Color(str, m_local.length);}
template <> inline void addNewLocal<S_Float,Single>(String& str){	m_local+=new L_S_Float(str, m_local.length);}
template <> inline void addNewLocal<Image,Single>(String& str){	m_local+=new L_S_Image(str, m_local.length);}
template <> inline void addNewLocal<S_Int32,Single>(String& str){	m_local+=new L_S_Int32(str, m_local.length);}
template <> inline void addNewLocal<NodePtr,Single>(String& str){	m_local+=new L_NodePtr(str, m_local.length);}
template <> inline void addNewLocal<Rotation,Single>(String& str){	m_local+=new L_S_Rotation(str, m_local.length);}
template <> inline void addNewLocal<String,Single>(String& str){m_local+=new L_S_String(str, m_local.length);}
template <> inline void addNewLocal<double,Single>(String& str){	m_local+=new L_S_Time(str, m_local.length);}
template <> inline void addNewLocal<Vec2f,Single>(String& str){	m_local+=new L_S_Vec2f(str, m_local.length);}
template <> inline void addNewLocal<Vec3f,Single>(String& str){	m_local+=new L_S_Vec3f(str, m_local.length);}
template <> inline void addNewLocal<M_Bool,Multiple>(String& str){	m_local+=new L_M_Bool(str, m_local.length);}
template <> inline void addNewLocal<M_Color3f,Multiple>(String& str){	m_local+=new L_M_Color(str, m_local.length);}
template <> inline void addNewLocal<M_Float,Multiple>(String& str){	m_local+=new L_M_Float(str, m_local.length);}
template <> inline void addNewLocal<M_Image,Multiple>(String& str){	m_local+=new L_M_Image(str, m_local.length);}
template <> inline void addNewLocal<M_Int,Multiple>(String& str){	m_local+=new L_M_Int32(str, m_local.length);}
template <> inline void addNewLocal<M_Node,Multiple>(String& str){	m_local+=new L_M_Node(str, m_local.length);}
template <> inline void addNewLocal<M_Rotation,Multiple>(String& str){m_local+=new L_M_Rotation(str, m_local.length);}
template <> inline void addNewLocal<M_String,Multiple>(String& str){	m_local+=new L_M_String(str, m_local.length);}
template <> inline void addNewLocal<M_Double,Multiple>(String& str){	m_local+=new L_M_Time(str, m_local.length);}
template <> inline void addNewLocal<M_Vec2f,Multiple>(String& str){	m_local+=new L_M_Vec2f(str, m_local.length);}
template <> inline void addNewLocal<M_Vec3f,Multiple>(String& str){	m_local+=new L_M_Vec3f(str, m_local.length);}


template <class T,class S> inline void extractParam(const String& str, int& i){
	str.stbCpp(i);
	if(str[i]=='&'){
		str.stbCpp(++i);
		stackOffset+=REF_SIZE;
	}
	else{
		stackOffset+=sizeof(T);
	}
	addNewLocal<T,S>(str.extractCppLabel(i));
	str.stb(i);
}
void extractParameter(const String& str, int& i){
	String label=str.extractCppLabel(i);

	if		(label=="bool"){extractParam<bool, Single>(str, i);return;}
	else if	(label=="Color3f"){extractParam<Color3f, Single>(str, i);return;}
	else if	(label=="float"){extractParam<float, Single>(str, i);return;}
	else if	(label=="int"){extractParam<int, Single>(str, i);return;}
	else if	(label=="NodePtr"){extractParam<NodePtr, Single>(str, i);return;}
	else if	(label=="Rotation"){extractParam<Rotation, Single>(str, i);return;}
	else if	(label=="String"){extractParam<String, Single>(str, i);return;}
	else if	(label=="double"){extractParam<double, Single>(str, i);return;}
	else if	(label=="Vec2f"){extractParam<Vec2f, Single>(str, i);return;}
	else if	(label=="Vec3f"){extractParam<Vec3f, Single>(str, i);return;}
	else if	(label=="M_bool"){extractParam<M_Bool, Multiple>(str, i);return;}
	else if	(label=="M_Color3f"){extractParam<M_Color3f, Multiple>(str, i);return;}
	else if	(label=="M_float"){extractParam<M_Float, Multiple>(str, i);return;}
	else if	(label=="M_Int"){extractParam<M_Int, Multiple>(str, i);return;}
	else if	(label=="M_Node"){extractParam<M_Node, Multiple>(str, i);return;}
	else if	(label=="M_Rotation"){extractParam<M_Rotation, Multiple>(str, i);return;}
	else if	(label=="M_String"){extractParam<M_String, Multiple>(str, i);return;}
	else if	(label=="M_Double"){extractParam<M_Double, Multiple>(str, i);return;}
	else if	(label=="M_Vec2f"){extractParam<M_Vec2f, Multiple>(str, i);return;}
	else if	(label=="M_Vec3f"){extractParam<M_Vec3f, Multiple>(str, i);return;}
	else throw;
}

template <class T> inline void writeProperty(V_<T,Multiple,Global>& v, const String& str, int& i);
template <class T> inline void writeProperty(V_<T,Single,Global>& v, const String& str, int& i);
template <class T> inline void writeProperty(V_<T,Single,Local>& v, const String& str, int& i);
inline void writeGlobalFloatComponent(unsigned char*& valAddress, Word* componentName, const int& offset, const String& str, int& i, int& bracketCntr);


template <class T,class S> inline void convertVar(G_Variable* v, const String& str, int& i, int& bracketCntr){throw;};
template <class T,class S> inline void convertVar(L_Variable* v, const String& str, int& i, int& bracketCntr){throw;};
template <class T, class T2, class T3, class GV> inline void convertVar(GV*& v, const String& str, int& i, int& bracketCntr);
void getFloatOperand(const String& str, int& i, int& bracketCntr);


template<class T, class S> void copyScratchPadToLocal(const int& varIndex, const int& destOpType, const String& str, int& i);
template<class T, class S> void copyScratchPadToGlobal(V_<T,Single,Global>& v, const int& destOpType, const String& str, int& i);


	String Function::toCodeLineString();
	//void putInElse(CodeLine& cl, const String& str, int& i);
	inline void Function::setUpElse(int& bypassMark, int& exitMark);
	void Function::genCppLine(CodeLine& codeLine);
	void Function::doCall2ndCompile();
	void Function::headerToString(String& str);
	void Function::headerToColorFormat(String& str, M_<ColorFormat>& format, int x);
	Seamless* getSeamlessReference();
	void addInterpToListForExportBVH(Art* part, M_Bool& pos, M_Art& m_interpolatorPart);
};

class MemberFunction : public Function{
	int getType()const{return MEMBERFUNCTION;}
};


Function* isFunctionCall(const String& label);
NodePtr getNodeFromName(const String str);

inline ACall::ACall(Function* i_function, int i_byteIndex):
		function(i_function),
		byteIndex(i_byteIndex)
{
}
#define OPERATOR_COLOR 0x4080c0
class W_Operator : public Word{
public:
	unsigned int getColor(){
		return OPERATOR_COLOR;
	}
};
#define FUNCTION_COLOR 0xFF8000
class W_Function : public Word{
	unsigned int getColor(){
		return FUNCTION_COLOR;
	}
};
class W_MemberFunction : public Word{
	unsigned int getColor(){
		return 0xff8000;
	}
};
class W_Keyword : public Word{
	unsigned int getColor(){
		//return 0xff8080;
		return 0x0000b0;
		//return 0x008080;
	}
};
class W_Property : public Word{
	unsigned int getColor(){
		return PROPERTY_COLOR;
	}
};


class W_Type : public Word{
	unsigned int getColor(){
		return TYPE_COLOR;
	}
};
class W_Type2 : public Word{
	unsigned int getColor(){
		return TYPE_COLOR;
	}
};
class W_Constant : public Word{
	unsigned int getColor(){
//		return 0x005010;
		return CONSTANT_COLOR;
	}
};
class W_Global : public Word{
public:
G_Variable* variable;
	W_Global(G_Variable* i_node):
		variable(i_node)
	{}
	void toString(String& str){
		str+=variable->name;
	}
	unsigned int getColor(){
		return GLOBAL_COLOR;
	}
};
class W_Node : public Word{
public:
Node* node;
	W_Node(Node* i_node):
		node(i_node)
	{}
	void toString(String& str){
		str+=node->name;
	}
	unsigned int getColor(){
		return NODE_COLOR;
	}
};
class W_NoOutputNode : public Word{
public:
Node* node;
	W_NoOutputNode(Node* i_node):
		node(i_node)
	{}
	void toString(String& str){
		if(g_outputMode==false){
			str+=node->name;
		}
	}
	unsigned int getColor(){
		return NODE_COLOR;
	}
};

/*
class W_Node : public Word{
Node* node;
public:
	W_NodeDotField(Node* i_node,const int& i_fieldIdx):
		node(i_node),
		fieldIdx(i_fieldIdx)
	{}
	void toString(String& str){
		str+=node->name;
	}
	unsigned int getColor(){
		return NODE_COLOR;
	}
};
*/
class W_Field : public W_Property{
public:
int fieldIdx;
Node* node;
	W_Field(Node* i_node,const int& i_fieldIdx):
		node(i_node),
		fieldIdx(i_fieldIdx)
	{}
	void toString(String& str){
		if(g_outputMode){
			addNodeNameForScript(str, node, fieldIdx);
		}
		else{
			str+=node->getProperties()[fieldIdx].name;
		}
	}
};

class W_MemberCall : public W_Function{
public:
	W_MemberCall()
	{}
	void toString(String& str){
		str+="doBuild()";
	}
};

class W_Morph : public W_Function{
public:
	W_Morph()
	{}
	void toString(String& str){
		str+="morph";
	}
};

class W_Comp : public W_Function{
public:
	W_Comp()
	{}
	void toString(String& str){
		str+="comp";
	}
};

#define LOCAL_COLOR 0x900090
class W_Local : public Word{
Function* function;
int varIndex;
public:
	W_Local(Function* i_function, int i_varIndex):
		function(i_function),
		varIndex(i_varIndex)
	{}
	void toString(String& str){
		str+=function->m_local[varIndex]->name;
	}
	unsigned int getColor(){
		return LOCAL_COLOR;
	}
	L_Variable* getVar(){
		return function->m_local[varIndex];
	}
};
class W_Assignment : public W_Operator{
public:
	void toString(String& str){
		str+=" = ";
	}
};
class W_PlusAssignment : public W_Operator{
public:
	void toString(String& str){
		str+=" += ";
	}
};
class W_MinusAssignment : public W_Operator{
public:
	void toString(String& str){
		str+=" -= ";
	}
};
class W_Minus : public W_Operator{
public:
	void toString(String& str){
		str+="-";
	}
};
class W_Not : public W_Operator{
public:
	void toString(String& str){
		str+="!";
	}
};
class W_Ref : public W_Operator{
public:
	void toString(String& str){
		str+="& ";
	}
};
class W_Sub : public W_Operator{
public:
	void toString(String& str){
		str+=" - ";
	}
};
class W_Add : public W_Operator{
public:
	void toString(String& str){
		str+=" + ";
	}
};
class W_Divide : public W_Operator{
public:
	void toString(String& str){
		str+=" / ";
	}
};
class W_VecDivide : public W_Operator{
public:
	void toString(String& str){
		str+=" / ";
	}
};
class W_Modulo : public W_Operator{
public:
	void toString(String& str){
		str+=" % ";
	}
};
class W_Multiply : public W_Operator{
public:
	void toString(String& str){
		str+=" * ";
	}
};
class W_MinusMinus : public W_Operator{
public:
	void toString(String& str){
		str+="--";
	}
};
class W_PlusPlus : public W_Operator{
public:
	void toString(String& str){
		str+="++";
	}
};

class W_LessThan : public W_Operator{
public:
	void toString(String& str){
		str+=" < ";
	}
};
class W_LessEqual : public W_Operator{
public:
	void toString(String& str){
		str+=" <= ";
	}
};
class W_GreaterThan : public W_Operator{
public:
	void toString(String& str){
		str+=" > ";
	}
};
class W_GreaterEqual : public W_Operator{
public:
	void toString(String& str){
		str+=" >= ";
	}
};
class W_Equality : public W_Operator{
public:
	void toString(String& str){
		str+=" == ";
	}
};
class W_NonEquality : public W_Operator{
public:
	void toString(String& str){
		str+=" != ";
	}
};
class W_Logic{};
class W_And : public W_Operator, public W_Logic{
public:
	void toString(String& str){
		str+=" && ";
	}
};
class W_Or : public W_Operator, public W_Logic{
public:
	void toString(String& str){
		str+=" || ";
	}
};
class W_OpenBracket : public W_Operator{
public:
	void toString(String& str){
		str+='(';
	}
};
class W_CloseBracket : public W_Operator{
public:
	void toString(String& str){
		str+=')';
	}
};
class W_ForSemicolon : public W_Operator{
public:
	void toString(String& str){
		str+="; ";
	}
};
class W_Semicolon : public W_Operator{
public:
	void toString(String& str){
		str+=";";
	}
};
class W_ReturnVal : public W_Keyword{
public:
	void toString(String& str){
		str+="return ";
	}
};
class W_If : public W_Keyword{
public:
	void toString(String& str){
		str+="if";
	}
};
class W_Else : public W_Keyword{
public:
	void toString(String& str){
		str+="else ";
	}
};
class W_While : public W_Keyword{
public:
	void toString(String& str){
		str+="while";
	}
};
class W_Do : public W_Keyword{
public:
	void toString(String& str){
		str+="do";
	}
};
class W_For : public W_Keyword{
public:
	void toString(String& str){
		str+="for";
	}
};
class W_Break : public W_Keyword{
public:
	void toString(String& str){
		str+="break";
	}
};
class W_Continue : public W_Keyword{
public:
	void toString(String& str){
		str+="continue";
	}
};
class W_ReturnVoid : public W_Keyword{
public:
	void toString(String& str){
		str+="return";
	}
};
class W_NoOutputDot : public W_Operator{
public:
	void toString(String& str){
		if(g_outputMode==false){
			str+=".";
		}
	}
};
class W_Dot : public W_Operator{
public:
	void toString(String& str){
		str+=".";
	}
};
class W_Length : public W_Property{
public:
	void toString(String& str){
		str+="length";
	}
};
class W_X : public W_Property{
public:
	void toString(String& str){
		str+="x";
	}
};
class W_Y : public W_Property{
public:
	void toString(String& str){
		str+="y";
	}
};
class W_Z : public W_Property{
public:
	void toString(String& str){
		str+="z";
	}
};
class W_Angle : public W_Property{
public:
	void toString(String& str){
		str+="angle";
	}
};
class W_R : public W_Property{
public:
	void toString(String& str){
		str+="x";
	}
};
class W_G : public W_Property{
public:
	void toString(String& str){
		str+="x";
	}
};
class W_B : public W_Property{
public:
	void toString(String& str){
		str+="x";
	}
};
class W_S_Bool : public W_Type{
public:
	void toString(String& str){
		if(!g_outputMode){
			str+="bool ";
		}
	}
};
class W_S_Float : public W_Type{
public:
	void toString(String& str){
		if(!g_outputMode){
			str+="float ";
		}
	}
};
class W_S_Image : public W_Type{
public:
	void toString(String& str){
		if(!g_outputMode){
			str+="Image ";
		}
	}
};
class W_S_Int32 : public W_Type{
public:
	void toString(String& str){
		if(!g_outputMode){
			str+="int ";
		}
	}
};
class W_S_NodePtr : public W_Type{
public:
	void toString(String& str){
		if(!g_outputMode){
			str+="Node* ";
		}
	}
};
class W_S_String : public W_Type{
public:
	void toString(String& str){
		if(!g_outputMode){
			str+="String ";
		}
	}
};
class W_S_Time : public W_Type{
public:
	void toString(String& str){
		if(!g_outputMode){
			str+="double ";
		}
	}
};
class W_S_Color : public W_Type{
public:
	void toString(String& str){
		if(!g_outputMode){
			str+="Color3f ";
		}
	}
};
class W_S_Rotation : public W_Type{
public:
	void toString(String& str){
		if(!g_outputMode){
			str+="Rotation ";
		}
	}
};
class W_S_Vec2f : public W_Type{
public:
	void toString(String& str){
		if(!g_outputMode){
			str+="Vec2f ";
		}
	}
};
class W_S_Vec3f : public W_Type{
public:
	void toString(String& str){
		if(!g_outputMode){
			str+="Vec3f ";
		}
	}
};
///////////
class W_M_Bool : public W_Type{
public:
	void toString(String& str){
		if(!g_outputMode){
			str+="M_Bool ";
		}
	}
};
class W_M_Float : public W_Type{
public:
	void toString(String& str){
		if(!g_outputMode){
			str+="M_Float ";
		}
	}
};
class W_M_Image : public W_Type{
public:
	void toString(String& str){
		if(!g_outputMode){
			str+="M_Image ";
		}
	}
};
class W_M_Int32 : public W_Type{
public:
	void toString(String& str){
		if(!g_outputMode){
			str+="M_Int ";
		}
	}
};
class W_M_NodePtr : public W_Type{
public:
	void toString(String& str){
		if(!g_outputMode){
			str+="M_Node ";
		}
	}
};
class W_M_String : public W_Type{
public:
	void toString(String& str){
		if(!g_outputMode){
			str+="M_String ";
		}
	}
};
class W_M_Double : public W_Type{
public:
	void toString(String& str){
		if(!g_outputMode){
			str+="M_Double ";
		}
	}
};
class W_M_Color : public W_Type{
public:
	void toString(String& str){
		if(!g_outputMode){
			str+="M_Color3f ";
		}
	}
};
class W_M_Rotation : public W_Type{
public:
	void toString(String& str){
		if(!g_outputMode){
			str+="M_Rotation ";
		}
	}
};
class W_M_Vec2f : public W_Type{
public:
	void toString(String& str){
		if(!g_outputMode){
			str+="M_Vec2f ";
		}
	}
};
class W_M_Vec3f : public W_Type{
public:
	void toString(String& str){
		if(!g_outputMode){
			str+="M_Vec3f ";
		}
	}
};
///////
class W_S_Bool2 : public W_Type2{
public:
	void toString(String& str){
		if(g_outputMode){
			str+="new SFBool";
		}
		else{
			str+="bool";
		}
	}
};
class W_S_Float2 : public W_Type2{
public:
	void toString(String& str){
		if(g_outputMode){
			str+="new SFFloat";
		}
		else{
			str+="float";
		}
	}
};
class W_S_Image2 : public W_Type2{
public:
	void toString(String& str){
		if(g_outputMode){
			str+="new SFImage";
		}
		else{
			str+="Image";
		}
	}
};
class W_S_Int322 : public W_Type2{
public:
	void toString(String& str){
		if(g_outputMode){
			str+="new SFInt32";
		}
		else{
			str+="int";
		}
	}
};
class W_S_NodePtr2 : public W_Type2{
public:
	void toString(String& str){
		if(g_outputMode){
			str+="new SFNode";
		}
		else{
			str+="Node*";
		}
	}
};
class W_S_String2 : public W_Type2{
public:
	void toString(String& str){
		if(g_outputMode){
			str+="new SFString";
		}
		else{
			str+="String";
		}
	}
};
class W_S_Time2 : public W_Type2{
public:
	void toString(String& str){
		if(g_outputMode){
			str+="new SFTime";
		}
		else{
			str+="double";
		}
	}
};
class W_S_Color2 : public W_Type2{
public:
	void toString(String& str){
		if(g_outputMode){
			str+="new SFColor";
		}
		else{
			str+="Color3f";
		}
	}
};
class W_S_Rotation2 : public W_Type2{
public:
	void toString(String& str){
		if(g_outputMode){
			str+="new SFRotation";
		}
		else{
			str+="Rotation";
		}
	}
};
class W_S_Vec2f2 : public W_Type2{
public:
	void toString(String& str){
		if(g_outputMode){
			str+="new SFVec2f";
		}
		else{
			str+="Vec2f";
		}
	}
};
class W_S_Vec3f2 : public W_Type2{
public:
	void toString(String& str){
		if(g_outputMode){
			str+="new SFVec3f";
		}
		else{
			str+="Vec3f";
		}
	}
};
///////////
class W_M_Bool2 : public W_Type2{
public:
	void toString(String& str){
		if(g_outputMode){
			str+="new MFBool";
		}
		else{
			str+="M_Bool";
		}
	}
};
class W_M_Float2 : public W_Type2{
public:
	void toString(String& str){
		if(g_outputMode){
			str+="new MFFloat";
		}
		else{
			str+="M_Float";
		}
	}
};
class W_M_Image2 : public W_Type2{
public:
	void toString(String& str){
		if(g_outputMode){
			str+="new MFImage";
		}
		else{
			str+="M_Image";
		}
	}
};
class W_M_Int322 : public W_Type2{
public:
	void toString(String& str){
		if(g_outputMode){
			str+="new MFInt32";
		}
		else{
			str+="M_Int";
		}
	}
};
class W_M_NodePtr2 : public W_Type2{
public:
	void toString(String& str){
		if(g_outputMode){
			str+="new MFNode";
		}
		else{
			str+="M_Node";
		}
	}
};
class W_M_String2 : public W_Type2{
public:
	void toString(String& str){
		if(g_outputMode){
			str+="new MFString";
		}
		else{
			str+="M_String";
		}
	}
};
class W_M_Double2 : public W_Type2{
public:
	void toString(String& str){
		if(g_outputMode){
			str+="new MFTime";
		}
		else{
			str+="M_Double";
		}
	}
};
class W_M_Color2 : public W_Type2{
public:
	void toString(String& str){
		if(g_outputMode){
			str+="new MFColor";
		}
		else{
			str+="M_RGB";
		}
	}
};
class W_M_Rotation2 : public W_Type2{
public:
	void toString(String& str){
		if(g_outputMode){
			str+="new MFRotation";
		}
		else{
			str+="Rotation";
		}
	}
};
class W_M_Vec2f2 : public W_Type2{
public:
	void toString(String& str){
		if(g_outputMode){
			str+="new SFVec2f";
		}
		else{
			str+="M_Vec2f";
		}
	}
};
class W_M_Vec3f2 : public W_Type2{
public:
	void toString(String& str){
		if(g_outputMode){
			str+="new MFVec3f";
		}
		else{
			str+="M_Vec3f";
		}
	}
};
class W_M_Int2 : public W_Type2{
public:
	void toString(String& str){
		if(g_outputMode){
			str+="new MFInt32";
		}
		else{
			str+="M_Int";
		}
	}
};

template <class T> class W_Immediate : public W_Constant{
public:
T value;
	W_Immediate(T i_value):
		value(i_value)
	{}
	void toString(String& str);
	inline unsigned int getColor();
};

template <class T> inline unsigned int W_Immediate<T>::getColor(){
	return CONSTANT_COLOR;
}
template <> inline unsigned int W_Immediate<S_Int32>::getColor(){
	return 0x00c0c0;
}
template <> inline unsigned int W_Immediate<double>::getColor(){
	return 0xff00ff;
}

template <> inline void W_Immediate<S_Bool>::toString(String& str){
	if(value){
		str+="true";
	}
	else{
		str+="false";
	}
}
template <> inline void W_Immediate<S_Float>::toString(String& str){
	str+=value;
}
template <> inline void W_Immediate<S_Int32>::toString(String& str){
	str+=value;
}
template <> inline void W_Immediate<double>::toString(String& str){
	str+=value;
}
class W_FloatList : public W_Type2{
public:
	void toString(String& str){
		if(g_outputMode){
			str+="new MFInt32";
		}
		else{
			str+="FloatList";
		}
	}
};
class W_IntListData : public Word{
public:
	M_Int value;
	void toString(String& str){
		value.toCppString(str);
	}
	void read(M_Word& m_word, const String& str, int& i);
	unsigned int getColor(){
		return 0x00c0c0;
	}
};
class W_FloatListData : public Word{
public:
	M_Float value;
	void toString(String& str){
		value.toCppString(str);
	}
	void read(M_Word& m_word, const String& str, int& i);
	unsigned int getColor(){
		return CONSTANT_COLOR;
	}
};
class W_PI : public W_Constant{
public:
	void toString(String& str){
		if(g_outputMode){
			str+="Math.PI";
		}
		else{
			str+="PI";
		}
	}
};
template <class T,class S> class W_ConstructLocal : public Word{
public:
	W_ConstructLocal()
	{}
	void W_ConstructLocal::toString(String& str){
		if(g_outputMode==false){
			str+=typeNameLookUp[templateType<T,S>()];
			str+=' ';
		}
	}
	unsigned int getColor(){
		return 0x0000ff;
	}
};
class W_Call : public W_Function{
	Function* function;
public:
	W_Call(Function* i_function):
		function(i_function)
	{}
	void toString(String& str){
		str+=function->functionName;
	}
};
class W_Comma : public W_Operator{
public:
	W_Comma()
	{}
	void toString(String& str){
		str+=", ";
	}
};
class W_OpenSquareBracket : public W_Operator{
public:
	W_OpenSquareBracket()
	{}
	void toString(String& str){
		str+='[';
	}
};
class W_CloseSquareBracket : public W_Operator{
public:
	W_CloseSquareBracket()
	{}
	void toString(String& str){
		str+=']';
	}
};
class W_Sin : public W_Function{
public:
	W_Sin()
	{}
	void toString(String& str){
		if(g_outputMode){
			str+="Math.sin";
		}
		else{
			str+="sin";
		}
	}
};
class W_Cos : public W_Function{
public:
	W_Cos()
	{}
	void toString(String& str){
		if(g_outputMode){
			str+="Math.cos";
		}
		else{
			str+="cos";
		}
	}
};
class W_Asin : public W_Function{
public:
	W_Asin()
	{}
	void toString(String& str){
		if(g_outputMode){
			str+="Math.asin";
		}
		else{
			str+="asin";
		}
	}
};
class W_Acos : public W_Function{
public:
	W_Acos()
	{}
	void toString(String& str){
		if(g_outputMode){
			str+="Math.acos";
		}
		else{
			str+="acos";
		}
	}
};
class W_Sqrt : public W_Function{
public:
	W_Sqrt()
	{}
	void toString(String& str){
		if(g_outputMode){
			str+="Math.sqrt";
		}
		else{
			str+="sqrt";
		}
	}
};
class W_Fabs : public W_Function{
public:
	W_Fabs()
	{}
	void toString(String& str){
		if(g_outputMode){
			str+="Math.abs";
		}
		else{
			str+="fabs";
		}
	}
};
class W_Ceil : public W_Function{
public:
	W_Ceil()
	{}
	void toString(String& str){
		if(g_outputMode){
			str+="Math.ceil";
		}
		else{
			str+="ceil";
		}
	}
};
class W_Floor : public W_Function{
public:
	W_Floor()
	{}
	void toString(String& str){
		if(g_outputMode){
			str+="Math.floor";
		}
		else{
			str+="floor";
		}
	}
};
class W_Pow : public W_Function{
public:
	W_Pow()
	{}
	void toString(String& str){
		if(g_outputMode){
			str+="Math.pow";
		}
		else{
			str+="pow";
		}
	}
};
class W_Powf : public W_Function{
public:
	W_Powf()
	{}
	void toString(String& str){
		if(g_outputMode){
			str+="Math.pow";
		}
		else{
			str+="powf";
		}
	}
};
class W_GetTime : public W_Function{
public:
	W_GetTime()
	{}
	void toString(String& str){
		if(g_outputMode){
			str+="Browser.";
		}
		str+="getTime";
	}
};
class W_Normalize : public W_MemberFunction{
public:
	W_Normalize()
	{}
	void toString(String& str){
		str+="normalize()";
	}
};

class W_Name : public Word{
	const char* cstr;
public:
	W_Name(const char* i_str):
		cstr(i_str)
	{}
	void toString(String& str){
		str+=cstr;
	}
	unsigned int getColor(){
		return 0;
	}
};
//hello
class W_Comment : public Word{
	String comment;
public:
	W_Comment(const String& i_comment):
		comment(i_comment)
	{}
	void toString(String& str){
		str+=comment;
	}
	unsigned int getColor(){
		return 0x808080;
	}
};



extern const char typeNameLookUp[VARABLETYPE_LOOKUPLEN][11];
extern const char ecmaScriptTypeNameLookUp[VARABLETYPE_LOOKUPLEN][11];
void onlyFunctionsToString(String& str);
void decodeFunctions(const String& str);
void initCompileForBody();
Script* getOutputScriptPtr();
void pasteSprite( int dx, int dy, int sx, int sy, int dstWidth, int dstHeight, int srcWidth, int srcHeight, int iconWidth, int iconHeight, int* dstPix, unsigned* srcPix);
void outputUserScripts(String& str, bool& newList);
bool getEquiveRoute(M_Word& m_word, W_Field*& w_dstField, W_Field*& w_srcField, M_Node& m_codeLine, int i);
bool getNodeFieldEqualsSomething(M_Word& m_word, W_Field*& w_dstField, M_Node& m_codeLine, int i);

