 /* Seamless3d

Copyright (C) 2001-2011 Graham John Perrett (AKA thyme) thyme@seamless3d.com www.seamless3d.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this
software and associated documentation files (the "Software"), to deal in the Software
without restriction, including without limitation the rights to use, copy, modify,
merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to the following
conditions:

The above copyright notice and this permission notice shall be included in all copies or
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT
OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.
*/


#include <windows.h>
#include <direct.h>

#include <sys/stat.h>
#include <share.h>
#include <fcntl.h>
#include <io.h>

#include "seamless3d.h"
#include "techuelife_triangle_editor.h"
#include "seamless3d_gui_render.h"
#include "gui.h"
#include "scene_tree.h"
#include "robot_aid.h"

extern bool g_insertControlPoint;
extern const char* title_save_file_as;

S_Triangle triVertexRem=S_Triangle(0,0,0,0,0,0);

#define halfArrowlen .1
#define g_spaghettiRad .025

extern bool RotateMode;
extern bool MouseFlag;
extern bool tugFlag;
extern bool g_tugRotFlag;
extern bool orientationTugFlag;
extern float g_version;
extern bool outPutIndices;
extern String g_bindSmlsFile;

extern BOOL g_autoNormalsGen;
extern String g_seamlessDir;
extern bool	g_mirrorModeVertexFound;
extern int	g_mirrorVertex;

extern OpenNetFilePtr g_openNetFile;
Vec3f Seamless::selCoordPos;
void addNodesForNewFile();

bool g_paletteMode = false;
bool g_colorOwnership = false;
extern bool g_dragMorphFlag;

extern Ptr<PasteOptions> g_pasteOptions;
extern Ptr<DeleteOptions> g_deleteOptions;
extern Ptr<TimelineFilter> g_timelineFilter;

M_Node g_nailDown;

ArtPtr g_mirrorPart;
Node* g_visualClue = nullPtr;
SeamlessPtr g_workPolyMapSmlsColor = new Seamless;
SeamlessPtr g_workPolyMapSmlsTexture = new Seamless;
ImmediatePart* g_workPolyMapColor = new ImmediatePart;
ImmediatePart* g_workPolyMapTexture = new ImmediatePart;

SeamlessPtr g_vcSeamless = new Seamless;
SeamlessPtr g_vcOscSmls;
TransformPtr g_vcOscBase=new Transform;
TransformPtr g_vcRingBuildBase;
TransformPtr g_vcRingBuildPivotPoint;
TransformPtr g_vcRingBuildOrientationA;
TransformPtr g_vcRingBuildOrientationB;
TransformPtr g_vcRingBuildOrientationC;
TransformPtr g_vcRingBuildOrientationD;
SeamlessPtr g_vcTransferBCyl;
CylinderBuild* g_vcTransferBCylCylinderBuild = new CylinderBuild;
TransformPtr g_vcTransferBCylBase=new Transform;
TransformPtr g_vcECScaleBase=new Transform;
TransformPtr g_vcBezierBase=new Transform;
SeamlessPtr g_vcNurbsCtrlPoly = new Seamless;
ArtPtr g_vcNurbsCtrlPolyPart;
TransformPtr g_vcCCLatheBase;
Ptr<AnimClue> g_vcArtAnimBase;
Ptr<PartClue> g_vcArtTriBase;

IndexedLineSet* g_triClue = new IndexedLineSet;
Coordinate* g_triClueCoord = new Coordinate;
Color* g_triClueColor = new Color;
Shape* g_triClueShape = new Shape;

IndexedLineSet* g_nurbsIls = new IndexedLineSet;
Coordinate* g_nurbsIlsCoord = new Coordinate;
Color* g_nurbsIlsColor = new Color;
Shape* g_nurbsIlsShape = new Shape;
TransformPtr g_nurbsIlsBase = new Transform;
Ptr<PartCageClue> g_partCageClue = new PartCageClue;

Ptr<CenterJointClue> g_vcCenterJointBase = new CenterJointClue;
Ptr<TransferBCylClue> g_transferBCylClue = new TransferBCylClue;
Ptr<TransferOwnershipClue> g_transferOwnershipClue = new TransferOwnershipClue;
ShapePtr g_vcSkeletonLinesShape;
Coordinate* g_vcSkeletonLinesCoordinate = 0;



Transform* g_vcECScaleTangentArrow;
Shape* g_vcEllipticalPolyline2D = new Shape;
Shape* g_transformLeverClue = new Shape;
ShapePtr g_vcFuseSurfaceFlatStageAidShape = new Shape;
Material* g_vcEllipticalMaterial = new Material;
Material* g_stickManMaterial = new Material;

ShapePtr g_vcSelectRectShape = new Shape;
Material* g_vcSelectRectMaterial = new Material;
Group* g_vcSelectRectGroup = new Group;

AppearancePtr g_vcCommonAppearance = new Appearance;


ArtPtr g_vcPart;
StemPtr g_vcStem;
SurfaceGeneratorPtr g_vcSurface;
CCLathePtr g_vcCCLathe;
SurfaceTransformPtr g_vcSurfaceTransform;
CCBendPtr g_vcCCBend;
SurfaceTransformPtr g_vcBaseRad;
Matrix g_viewMatrix;

bool NurbsPatch::tugRingPivotPoint = false;
Node* SenseTugGlobals::node=0;
Art* SenseTugGlobals::actTriPart=0;
int SenseTugGlobals::ActTri=0;
float SenseTugGlobals::hitZ=0;
int SenseTugGlobals::corner=0;
int SenseTugGlobals::ActCtrlPnt=0;
//bool SenseTugGlobals::sameVertexHit=false;
Vec3f SenseTugGlobals::pointer;
bool MultiSelect::selectRectMode = false;
Vec3f MultiSelect::topLeft;
Vec3f MultiSelect::bottomRight;
int MultiSelect::uRingDim = 0;
int MultiSelect::vRingDim = 0;

Vec2f MultiSelect::topLeft2d;
Vec2f MultiSelect::bottomRight2d;
M_Vec3f MultiSelect::beginDragPos;
Vec3f MultiSelect::autoPivotPoint;
int MultiSelect::clickedIndex = 0;
bool g_allowPolyMorphCapture = false;

Stem* g_radialTugStem;
RadialTug* g_radialTug;

int BendyLathe::radIdx = 0;

//bool g_spaghetti=true;

//extern ScenePtr g_scene;
Pointer g_pointer;
extern Rotation g_beginViewRot;
extern Vec2i g_rightDragCounter;
int g_lockFile = 0;

extern int getFileSize(FILE* Stream);
extern bool g_createdINI;

void search(char* match ,char*& from,char* BufEnd);
void searchTillText(char*& p);
void searchTillNextWord(char*& p);
void compileScalePositionInterpolator(const Art* part, Matrix& matrix);
void compileOrientationInterpolator(const Art* part, Rotation rotation, M_<SceneTreeListItem>& sceneTreeList);
Vec2f convertSouthPoleToMercator(const Vec2f& coord);
Vec2f convertNorthPoleToMercator(const Vec2f& coord);

void fillPolyLineIndex(M_Vec3f& point, M_Int& coordIndex);
bool checkAllPivotPoints(bool perspective, StickMan* stickMan);
void insertBone(Art* part);

float g_viewDist = 3;
void setTitleName();
extern Ptr<Palette> g_palette;

unsigned colorToHex(Color3f color);
int BackUp::backedUpFileCntr=0;
int BackUp::backedUpFileCntrEnd=0;
bool SurfacePoint(
				  float u, float v, int uOrder, int vOrder, int uDimension, int vDimension,
				  const float *uKnot, const float * vKnot,
				  const Vec3f *P,const float *W,
				  Vec3f *dest // result
				  );



extern M_Int progIndices;
void changeFloatFormat(int numOfDigits);
bool decodeSmlsVrmlHeader(String& str, int& i);
extern bool forceSaveAs=false;

M_Matrix m_matrix;
M_Int m_smi;
M_Vec3f g_m_coord;

Matrix g_catchMatrix;
Matrix g_catchIMatrix;
String BackUp::ourBackUpName="0_";
Ptr<_3DWin> g_3d = new _3DWin;

/*
enum WinSizePosIndices{
	THREE_D_POS_X=0,
	THREE_D_POS_Y=1,
	THREE_D_CX=2,
	THREE_D_CY=3,

	TOOLBAR_POS_X=4,
	TOOLBAR_POS_Y=5,
	TOOLBAR_CX=6,
	TOOLBAR_CY=7,

	SCENETREE_POS_X=8,
	SCENETREE_POS_Y=9,
	SCENETREE_CX=10,
	SCENETREE_CY=11,

	NODE_POS_X=12,
	NODE_POS_Y=13
};
*/
bool checkKey(M_<ShortCutKey>& shortCut, unsigned key){
	for(int i=shortCut.length; --i>=0;){
		if(shortCut[i].keyCode==key){
			shortCut[i].function();
			return true;
		}
	}
	return false;
}
unsigned g_key=0;
bool keyDown(unsigned key){
	g_key=key;
	if(checkKey(globalShortCut, key))return true;
	return Window::activeWindow->keyDown(key);
}
bool keyDownAlt(unsigned key){
	if(checkKey(globalShortCutAlt, key))return true;
	return Window::activeWindow->keyDownAlt(key);
}
bool keyDownCtrl(unsigned key){
	if(checkKey(globalShortCutCtrl, key))return true;
	return Window::activeWindow->keyDownCtrl(key);
}
bool blockShortCutKeys(unsigned key){
	if(key>=0x21 && key <= 0x28){
		return false; //includes arrow keys
	}
	if(Window::chatMode())return true;
	return false;
}
bool keyDownShift(unsigned key){
	if(blockShortCutKeys(key))return false;
	if(checkKey(globalShortCutShift, key))return true;
	return Window::activeWindow->keyDownShift(key);
}
bool _3DWin::keyDown(unsigned key){
	if(Window::chatMode())return false;
	return checkKey(shortCut, key);
}
bool _3DWin::keyDownAlt(unsigned key){
	return checkKey(shortCutAlt, key);
}
bool _3DWin::keyDownCtrl(unsigned key){
	return checkKey(shortCutCtrl, key);
}
bool _3DWin::keyDownShift(unsigned key){
	if(Window::chatMode())return false;
	return checkKey(shortCutShift, key);
}
bool ToolBarWin::keyDown(unsigned key){
	if(Window::chatMode())return false;
	return checkKey(_3DWin::shortCut, key);
}
bool ToolBarWin::keyDownAlt(unsigned key){
	return checkKey(_3DWin::shortCutAlt, key);
}
bool ToolBarWin::keyDownCtrl(unsigned key){
	return checkKey(_3DWin::shortCutCtrl, key);
}
bool ToolBarWin::keyDownShift(unsigned key){
	if(Window::chatMode())return false;
	return checkKey(_3DWin::shortCutShift, key);
}
bool SceneTree::keyDown(unsigned key){
	return checkKey(shortCut, key);
}
bool SceneTree::keyDownAlt(unsigned key){
	return checkKey(shortCutAlt, key);
}
bool SceneTree::keyDownCtrl(unsigned key){
	return checkKey(shortCutCtrl, key);
}
bool SceneTree::keyDownShift(unsigned key){
	return checkKey(shortCutShift, key);
}

bool ControlPanel::keyDown(unsigned key){
	return checkKey(shortCut, key);
}
bool ControlPanel::keyDownAlt(unsigned key){
	return checkKey(shortCutAlt, key);
}
bool ControlPanel::keyDownCtrl(unsigned key){
	return checkKey(shortCutCtrl, key);
}
bool ControlPanel::keyDownShift(unsigned key){
	return checkKey(shortCutShift, key);
}
bool ChatWin::keyDown(unsigned key){
	return checkKey(shortCut, key);
}
bool ChatWin::keyDownCtrl(unsigned key){
	return checkKey(shortCutCtrl, key);
}
bool ChatWin::keyDownShift(unsigned key){
	return checkKey(shortCutShift, key);
}

/*
void Scene::setViewFromRaw(){
	position.refValue()=-translation+(-rotation*Vec3f(0,0,g_viewDist));
	orientation.refValue()=-rotation;
	if(dynamic_cast<Scene*>(g_pw->node.p)){
		g_pw->paint();
	}
	//BackUp::monostable.beginCycle();
}
*/
void Scene::rotateScene(long dimslX, long dimslY){
	g_rightDragCounter+=Vec2i(dimslX, dimslY);
//	trackball code
	Vec2f diff = Vec2f(g_rightDragCounter.x, g_rightDragCounter.y);
	float a = diff.getAngle2();
	float distance=(diff.x*diff.x + diff.y*diff.y);
	Vec3f axis=Vec3f(0,1,0);
	axis=Rotation(0,0,1,-a)*axis;

	rotation=g_beginViewRot*Rotation(axis, distance*.0025);

//	rotation=rotation*Rotation(0,1,0, PI*.005*(float)dimslX)*Rotation(1,0,0, PI*.005*(float)dimslY);
//	BackUp::monostable.beginCycle();

}
void BackUp::commonUndoRedo(){

	decodeFunctions(g_loadStr);
	g_stw->genNodeList();
	postLoadProcessNodes(g_sceneTreeList);

	g_scene->postLoadProcessCollapse();
	resetLooseEnds();
	Window::postLoadProcessWin(*g_scene);
}
void BackUp::deleteBackupFiles(){
	String filename;
	_close(g_lockFile);
	g_lockFile = 0;

	filename = g_undoDir;
	filename += ourBackUpName;
	filename += ".lock";
	DeleteFile(filename.c_str());

	BOOL noError = FALSE;
	int i = 0;
	do{
		filename = g_undoDir;
		filename += ourBackUpName;
		filename += i++;
		filename += ".smls";
		noError = DeleteFile(filename.c_str());
	}while (noError);
	backedUpFileCntr = 0;
}


void BackUp::createUniqueBackupName(bool copy){
	String filename;
	int i = 0;
	while(1){
		ourBackUpName = i++;
		ourBackUpName += '_';
		filename = g_undoDir;
		filename += ourBackUpName;
		filename += "0.smls";
		if(FILE* file = fopen(filename.c_str(), "r+b")){
			fclose(file);
		}
		else{
			String lockName = g_undoDir;
			lockName += ourBackUpName;
			lockName += ".lock";
			if(g_lockFile > 0){
				_close(g_lockFile);
			}
			g_lockFile = 0;
			SetFileAttributes(lockName.c_str(), FILE_ATTRIBUTE_NORMAL);
			if(_sopen_s(&g_lockFile, lockName.c_str(), _O_RDWR | _O_CREAT | _O_EXCL, _SH_DENYRW, _S_IREAD | _S_IWRITE) == 0){
				//const String& file0 = g_settings->file.refValue()[0];
				_write(g_lockFile, g_currentFilename.c_str(), g_currentFilename.length);
				backupToCache(copy);
				break;
			}
			else{
				g_lockFile = 0;
			}
		}
	}
}

void BackUp::backupToCache(bool copy){
	static String filename;
	filename = g_undoDir;
	filename += ourBackUpName;
	filename += backedUpFileCntr;
	filename += ".smls";
	SetFileAttributes(filename.c_str(), FILE_ATTRIBUTE_NORMAL);
	if(copy && g_settings->file.length){
		CopyFile(g_settings->file.refValue()[0].c_str(), filename.c_str(), FALSE);
	}
	else{
		writeSmlsFile(filename);
	}
	backedUpFileCntrEnd = backedUpFileCntr;

	if(copy){
		smlsSaved.set();
	}
	//setTitleName();
}

void splitPath(String wholePath, String& drive, String& dir, String& fname, String& ext){
	char _dir[_MAX_DIR];
	char _fname[_MAX_FNAME];
	char _ext[_MAX_EXT];
	char _drive[_MAX_DRIVE];
	_splitpath( wholePath.c_str(), _drive, _dir, _fname, _ext );
	drive=_drive;
	dir=_dir;
	fname=_fname;
	ext=_ext;
}
bool checkVertexBelongsToTriangle(Art* vertexPart, Art* triPart, int vertexIndex, int &corner, int triangleIndex){
	if(triPart==0)return 0;
	int i=triangleIndex*3+corner;
	M_Triangle& triangle=triPart->triangle;
	if((unsigned)i>=(unsigned)triangle.length*3)return false;
	if(vertexPart!=triangle.refPart(i))return false;
	if(vertexIndex!=triangle.refIndex(i))return false;
	return true;
}
inline Art* findFirstRelivantPartContainingTriangles(M_Art& relevantPart){
	for(int k=0; k<relevantPart.length;k++){
		if(relevantPart[k]->triangle.length){
			return relevantPart[k];
		}
	}
	throw;
}
inline int findRelivantPartIndex(M_Art& relevantPart, Art*& part){
	for(int k=0; k<relevantPart.length;k++){
		if(relevantPart[k]==part){
			return k;
		}
	}
	throw;
}
int findNextTriangleFromVertex(Art* vertexPart, Art*& triPart, const int vertexIndex, int &corner, int triangleIndex){
	if(vertexIndex>=vertexPart->coord.refValue().length)throw vertexPart;

	M_Art& relevantPart=vertexPart->relevantPart;

	if(checkVertexBelongsToTriangle(vertexPart, triPart, vertexIndex, corner, triangleIndex)==false){
		triPart=findFirstRelivantPartContainingTriangles( relevantPart);
		corner=0;
		triangleIndex=0;
	}
	M_Triangle* triangle=&triPart->triangle;
	int k=findRelivantPartIndex(relevantPart, triPart);

	while(1){
		if(++triangleIndex>=triangle->length){
			if(++k>=relevantPart.length){
				k=0;
			}
			triangle=&relevantPart[k]->triangle;
			triangleIndex=0;
		}
		if(triangle->length){
			S_Triangle tri=(*triangle)[triangleIndex];

			if(	vertexIndex==tri.ia && tri.pa == vertexPart){
				corner=0;
				triPart=relevantPart[k];
				return triangleIndex;
			}
			if(	vertexIndex==tri.ib && tri.pb == vertexPart){
				corner=1;
				triPart=relevantPart[k];
				return triangleIndex;
			}
			if(	vertexIndex==tri.ic && tri.pc == vertexPart){
				corner=2;
				triPart=relevantPart[k];
				return triangleIndex;
			}
		}
	}
}
int findPrevTriangleFromVertex(Art* vertexPart, Art*& triPart, const int vertexIndex, int &corner, int triangleIndex){
	if(vertexIndex>=vertexPart->coord.refValue().length)throw vertexPart;

	M_Art& relevantPart=vertexPart->relevantPart;

	if(checkVertexBelongsToTriangle(vertexPart, triPart, vertexIndex, corner, triangleIndex)==false){
		triPart=findFirstRelivantPartContainingTriangles( relevantPart);
		corner=0;
		triangleIndex=0;
	}
	M_Triangle* triangle=&triPart->triangle;
	int k=findRelivantPartIndex(relevantPart, triPart);

	while(1){
		if(--triangleIndex<0){
			if(--k<0){
				k=relevantPart.length-1;
			}
			triangle=&relevantPart[k]->triangle;
			triangleIndex=triangle->length-1;
		}
		S_Triangle tri=(*triangle)[triangleIndex];

		if(triangle->length){
			if(	vertexIndex==tri.ia && tri.pa == vertexPart){
				corner=0;
				triPart=relevantPart[k];
				return triangleIndex;
			}
			if(	vertexIndex==tri.ib && tri.pb == vertexPart){
				corner=1;
				triPart=relevantPart[k];
				return triangleIndex;
			}
			if(	vertexIndex==tri.ic && tri.pc == vertexPart){
				corner=2;
				triPart=relevantPart[k];
				return triangleIndex;
			}
		}
	}
}

inline bool findPolyThatDoesBelong(M_Triangle& triangle, Art*& target, int& triIndex, int vi){
	for(int i=0; i<triangle.length; i++ ){
		S_Triangle tri=triangle[i];
		if(tri.pa==target && tri.ia==vi){
			triIndex=i;
			return true;
		}
		if(tri.pb==target && tri.ib==vi){
			triIndex=i;
			return true;
		}
		if(tri.pc==target && tri.ic==vi){
			triIndex=i;
			return true;
		}
	}
	return false;
}
inline void refreshCorner(Art* part, const int& actCtrlPnt, S_Triangle& tri, int& corner){
	Vec3f* activeVertex = &part->coord.refValue()[actCtrlPnt];
	if(activeVertex==&tri.pa->coord.refValue()[tri.ia]){
		corner=0;
	}
	else if(activeVertex==&tri.pb->coord.refValue()[tri.ib]){
		corner=1;
	}
	else if(activeVertex==&tri.pc->coord.refValue()[tri.ic]){
		corner=2;
	}
	else{
		throw;
	}
}
inline void refreshActTriPart(Art*& part, Seamless*& smls){
	if(checkVertexBelongsToTriangle(part, smls->actTriPart, smls->ActCtrlPnt, smls->ActTriCorner, smls->ActTri)){
		return;
	}
/*
	if(	smls->actTriPart!=0 &&
		smls->actTriPart->triangle.length &&
		(unsigned)smls->ActTri<(unsigned)smls->actTriPart->triangle.length){

		S_Triangle& actTri=smls->actTriPart->triangle[smls->ActTri];
		if(
		   (actTri.ia==smls->ActCtrlPnt && actTri.pa==part ||
			actTri.ib==smls->ActCtrlPnt && actTri.pb==part ||
			actTri.ic==smls->ActCtrlPnt && actTri.pc==part )
		){
			return;
		}
	}
*/
	//first look for a poly owned by us that references us
	if(findPolyThatDoesBelong(part->triangle, part, smls->ActTri, smls->ActCtrlPnt)){
		smls->actTriPart=part;
		return;
	}
	//then look for a poly from a relevant part that references us
	M_Art& relevantPart=part->relevantPart;
	for(int k=relevantPart.length; --k>=0;){
		if(findPolyThatDoesBelong(relevantPart[k]->triangle, part, smls->ActTri, smls->ActCtrlPnt)){
			smls->actTriPart=relevantPart[k];
			return;
		}
	}
	smls->ActTri=0;
	smls->ActCtrlPnt=0;
	smls->actTriPart=0;
}
inline void setCoordIndexToPointerIndex(Art& part, Seamless& smls){
	if(part.index[0]!=smls.ActCtrlPnt||g_before!=g_selNode.getRaw()){
		part.index[0]=smls.ActCtrlPnt;
		if(part.workEffect!=0){
			int tri=smls.ActTri;
			int corner=smls.ActTriCorner;
			Art* triPart =smls.actTriPart;
			int ti=tri*3+corner;
			if(triPart!=0&&triPart->triangle.length>ti){
				part.workEffect->setColorIndex(triPart->triangle.refTexIndex(ti));
			}
		}
		if(Art* part=dynamic_cast<Art*>(g_stw->cp->node.p)){
			if(part->index[0]>=part->coord.refValue().length){
				part->index[0]=0;
			}
			g_stw->cp->paint();
		}
	}
}
void makePointerEqualIndexLocation(){
	Art* part=dynamic_cast<Art*>(g_selNode.getRaw());
	if(part==0)return;
	Seamless* smls=part->smls;
	const M_Vec3f& coord=part->coord;

	if(coord.length==0){
		smls->ActTriCorner=0;
		smls->ActTri=0;
		smls->ActCtrlPnt=0;
		smls->actTriPart=0;
		setCoordIndexToPointerIndex(*part,*smls);
		return;
	}
	if((unsigned)smls->ActCtrlPnt>=(unsigned)coord.length){
		smls->ActCtrlPnt=coord.length-1;
	}
	if(smls->ActCtrlPnt<0){
		smls->ActCtrlPnt=0;
		smls->ActTri=0;
		smls->ActTriCorner=0;
	}
	refreshActTriPart(part, smls);
	if(smls->actTriPart){
		refreshCorner(part, smls->ActCtrlPnt, smls->actTriPart->triangle[smls->ActTri], smls->ActTriCorner);
	}

	Matrix viewMatrix=g_scene->getViewMatrix();
	
	g_pointer.translation=part->matrix * viewMatrix * coord[smls->ActCtrlPnt];

	setCoordIndexToPointerIndex(*part,*smls);
	if(g_before!=g_selNode.getRaw()){
		setTitleName();
		g_stw->setSelNodeLine(part);
		g_stw->paint();
		g_stw->cp->paint();
	}
}

void Settings::writeINIFile(){

	String str="#VRML V2.0 utf8\n";
	str+="#generated by Seamless3d ";
	str+=g_version;
	str+='\n';

//	nodeWinPos=Vec2f(g_stw->cp->pos.x, g_stw->cp->pos.y);
//	sceneTreePos=Vec2f(g_stw->pos.x, g_stw->pos.y);
//	sceneTreeSize=Vec2f(g_stw->size.x, g_stw->size.y);
	toolBarPos=Vec2f(g_tb->pos.x, g_tb->pos.y);

	toString(str,true);
	g_openNetFile->toString(str, true);
	g_pasteOptions->toString(str, true);
	g_deleteOptions->toString(str, true);
	g_timelineFilter->toString(str, true);
	str.write(g_setupPath);
}

void buildSceneMatrices( ArtPtr part, Matrix catmat);

inline void genTransferBCylClue(){
	g_vcTransferBCyl = new Seamless;
	ColorEffect* colorEffect = new ColorEffect;
	colorEffect->transparency.refValue()=.5;
	g_vcTransferBCyl->effect=colorEffect;
	Part* part = new Part;
	g_vcTransferBCyl->skeleton.refValue()+=part;
	g_vcTransferBCylCylinderBuild->part.refValue()=part;
	g_vcTransferBCyl->build.refValue()+=g_vcTransferBCylCylinderBuild;
}
inline void genMultiSelClue(){
	Seamless* smls = new Seamless;
	ColorEffect* colorEffect = new ColorEffect;
	colorEffect->transparency.refValue()=.5;
	smls->effect=colorEffect;
	ImmediatePart* part = new ImmediatePart;
	smls->skeleton.refValue()+=part;
	smls->process();
	g_vcArtTriBase->transpSmls = smls;
}
inline void genOscClue(){
	g_vcOscSmls = new Seamless;
	ColorEffect* colorEffect = new ColorEffect;
	g_vcOscSmls->effect=colorEffect;
	Part* part = new Part;
	g_vcOscSmls->skeleton+=part;
	Stem* stem = new Stem;
	stem->part+=part;
	//stem->slide=Vec2f(.5,0);
	stem->xCentred=false;
	stem->xDistance = 1;
	stem->sync = false;
	stem->staves = 64;
	stem->xJoinEnds = false;
	stem->add = false;
	stem->store.refValue() = true;
	//stem->insideOut=true;
	NurbsPatch* nurbsPatch = new NurbsPatch;
	nurbsPatch->uJoinEnds = false;
	nurbsPatch->stem.refValue() += stem;
	g_vcOscSmls->build+=nurbsPatch;
	g_vcOscBase->children+=g_vcOscSmls.p;
	g_vcOscBase->children+=g_vcBezierBase.p;

/////////
}
IndexedTriangleSet* convertToTriangleSet(GeomSource& gs){
	gs.genNormals();
	IndexedTriangleSet* its=new IndexedTriangleSet;
	Coordinate* coord=new Coordinate;
	Color* color=new Color;
	Normal* normal=new Normal;
	its->coord=coord;
	its->color=color;
	its->normal=normal;
	coord->point=gs.coord;
	color->color=gs.color;
	normal->vector=gs.normal;
	its->index=gs.coordIndex;
	if(gs.texCoord.length>0){
		TextureCoordinate* textureCoordinate=new TextureCoordinate;
		textureCoordinate->point=gs.texCoord;
		its->texCoord=textureCoordinate;
	}
	return its;
}
IndexedFaceSet* convertToFaceSet(GeomSourcePtr gs){
	gs->genNormals();
	IndexedFaceSetPtr its=new IndexedFaceSet;
	Coordinate* coord=new Coordinate;
	Color* color=new Color;
	Normal* normal=new Normal;
	its->coord=coord;
	its->color=color;
	its->normal=normal;
	coord->point=gs->coord;
	color->color=gs->color;
	normal->vector=gs->normal;
	its->index=gs->coordIndex;
	if(gs->texCoord.length>0){
		TextureCoordinate* textureCoordinate=new TextureCoordinate;
		textureCoordinate->point=gs->texCoord;
		its->texCoord=textureCoordinate;
	}
	return its;
}

void initVisualClue(){
//do pointer stuff
	g_vcArtTriBase = new PartClue;

	genSphere(MouseSense*2., 8, 8, *g_redPoint, new ColorEffect);
	genSphere(MouseSense*2., 8, 8, *g_nextPoint, new ColorEffect);
	genSphere(MouseSense*2., 8, 8, *g_lastPoint, new ColorEffect);
	genSphere(MouseSense*2., 8, 8, *g_yellowPoint, new ColorEffect);
	genSphere(MouseSense*2., 8, 8, *g_cyanPoint, new ColorEffect);

	g_redPoint->genColorList(Color3f(1,0,0));
	g_lastPoint->genColorList(Color3f(0,0,1));
	g_nextPoint->genColorList(Color3f(0,1,0));
	g_yellowPoint->genColorList(Color3f(1,1,0));
	g_cyanPoint->genColorList(Color3f(0,1,1));

	
	g_vcCommonAppearance->material = new Material;

	g_redPoint->appearance = g_vcCommonAppearance;
	g_lastPoint->appearance = g_vcCommonAppearance;
	g_nextPoint->appearance = g_vcCommonAppearance;
	g_yellowPoint->appearance = g_vcCommonAppearance;
	g_cyanPoint->appearance = g_vcCommonAppearance;
	
	genTransferBCylClue();
	genMultiSelClue();
	genOscClue();


	{
		g_triClue->coord.refValue() = g_triClueCoord;
		g_triClue->color.refValue() = g_triClueColor;
		Appearance* appearance = new Appearance;
		appearance->material = new Material;
		g_triClueShape->appearance.refValue() = appearance;
		g_triClueShape->geometry.refValue() = g_triClue;
		g_triClueCoord->point.refValue().length = 6;
		g_triClueCoord->point.refValue() = M_Vec3f(Vec3f(-1, 0, 0), Vec3f(0, 1, 0), Vec3f(0, 1, 0), Vec3f(1, 0, 0), Vec3f(1, 0, 0), Vec3f(-1, 0, 0));
		g_triClueColor->color.refValue() = M_Color3f(Color3f(1, 0, 0), Color3f(1, 0, 0), Color3f(0, 1, 0), Color3f(0, 1, 0), Color3f(0, 0, 1), Color3f(0, 0, 1));

		g_triClue->coordIndex.refValue() = M_Int(0, 1, -1, 2, 3, -1, 4, 5);
		g_triClue->noZbuf = true;
	}

	g_nurbsIls->coord.refValue() = g_nurbsIlsCoord;
	g_nurbsIls->color.refValue() = g_nurbsIlsColor;
	g_nurbsIlsShape->appearance.refValue() = new Appearance;
	g_nurbsIlsShape->geometry.refValue() = g_nurbsIls;
	g_nurbsIlsBase->children.refValue() += g_nurbsIlsShape;


	ColorEffect* colorEffect = new ColorEffect;
	colorEffect->transparency = .5;
	g_vcNurbsCtrlPoly->effect = colorEffect;
	g_vcNurbsCtrlPolyPart = new Part;
	g_vcNurbsCtrlPoly->skeleton += g_vcNurbsCtrlPolyPart;

	//g_visualClue=g_vcSeamless;
	g_workPolyMapSmlsColor->effect = new ColorEffect;;
	g_workPolyMapSmlsTexture->effect = new ColorEffect;
	g_workPolyMapColor->mirror.refValue() = false;
	g_workPolyMapTexture->mirror.refValue() = false;
	g_workPolyMapSmlsColor->skeleton += g_workPolyMapColor;
	g_workPolyMapSmlsTexture->skeleton += g_workPolyMapTexture;
	g_workPolyMapSmlsColor->process();
	g_workPolyMapSmlsTexture->process();
	g_workPolyMapSmlsColor->resetRelevantParts();
	g_workPolyMapSmlsTexture->resetRelevantParts();


	g_vcSeamless->effect = new ColorEffect;
	g_vcPart = new ImmediatePart;
	g_vcSeamless->skeleton += g_vcPart;
	g_vcStem = new Stem;
	g_vcStem->sync = true;
	g_vcStem->sync.refValue()=true;
	g_vcSurface = new SurfaceGenerator;
	g_vcCCLathe = new CCLathe;
	g_vcSurfaceTransform = new SurfaceTransform;
	g_vcCCBend = new CCBend;
	g_vcCCBend->location=Vec3f(0, -.0000001, 0);
	g_vcBaseRad = new SurfaceTransform;

	M_Build build;
	build += g_vcStem.p;
	g_vcSeamless->build.refValue() += g_vcStem.p;

	g_vcStem->part = g_vcPart;
	M_Node& generator = g_vcSurface->generator.refValue();
	generator += g_vcCCLathe.p;
	generator += g_vcSurfaceTransform.p;
	generator += g_vcCCBend.p;
	generator += g_vcBaseRad.p;


	g_vcStem->surface = g_vcSurface;
	g_vcStem->staves.refValue() += 6;
	g_vcStem->bands.refValue() += 16;

	g_vcCCLathe->distance.refValue() += 1;
	g_vcCCLathe->curve.refValue() += -.00001;
	g_vcCCLathe->curve.refValue() += .00001;
	g_vcCCLathe->baseRad.refValue() = g_spaghettiRad;

	g_vcSeamless->process();

	IndexedLineSet* ils = new IndexedLineSet;
	ils->coord = new Coordinate;
	g_vcEllipticalPolyline2D->geometry=ils;
	IndexedLineSet* ils2 = new IndexedLineSet;
	ils2->coord = new Coordinate;
	g_vcFuseSurfaceFlatStageAidShape->geometry = ils2;
	IndexedLineSet* ils3 = new IndexedLineSet;
	ils3->coord = new Coordinate;
	g_transformLeverClue->geometry = ils3;
	

	g_vcCCLatheBase = new Transform;
	g_vcRingBuildBase = new Transform;
	g_vcRingBuildPivotPoint = new Transform;
	g_vcRingBuildOrientationA = new Transform;
	g_vcRingBuildOrientationB = new Transform;
	g_vcRingBuildOrientationC = new Transform;
	g_vcRingBuildOrientationD = new Transform;
	g_vcArtAnimBase = new AnimClue;

	
	g_vcRingBuildBase->children.refValue() += g_vcRingBuildPivotPoint.p;
	g_vcRingBuildPivotPoint->children.refValue() += g_yellowPoint.p;
	g_vcRingBuildBase->children.refValue() += g_vcRingBuildOrientationA.p;
	g_vcRingBuildBase->children.refValue() += g_vcRingBuildOrientationB.p;
	g_vcRingBuildBase->children.refValue() += g_vcRingBuildOrientationC.p;
	g_vcRingBuildBase->children.refValue() += g_vcRingBuildOrientationD.p;
	g_vcRingBuildOrientationA->children.refValue() += g_redPoint.p;
	g_vcRingBuildOrientationB->children.refValue() += g_redPoint.p;
	g_vcRingBuildOrientationC->children.refValue() += g_redPoint.p;
	g_vcRingBuildOrientationD->children.refValue() += g_redPoint.p;

	AppearancePtr g_vcCommonAppearance = new Appearance;
	g_vcCommonAppearance->material = new Material;

	g_vcSelectRectMaterial->emissiveColor = Color3f(1,1,0);
	g_vcSelectRectMaterial->diffuseColor = Color3f(1,1,0);
	Appearance* selectRectAppearance = new Appearance;
	selectRectAppearance->material = g_vcSelectRectMaterial;
	g_vcSelectRectShape->appearance = selectRectAppearance;
	g_transformLeverClue->appearance = selectRectAppearance;

	{
		IndexedLineSet* ils=new IndexedLineSet;
		ils->noZbuf = true;
		ils->coord=new Coordinate;
		g_vcSelectRectShape->geometry=ils;
	}


	g_stickManMaterial->emissiveColor=Color3f(0,1,0);
	g_stickManMaterial->diffuseColor=Color3f(0,1,0);
	g_vcEllipticalMaterial->emissiveColor=Color3f(0,1,1);
	g_vcEllipticalMaterial->diffuseColor=Color3f(0,1,1);
	Appearance* appearance =new Appearance;
	appearance->material=g_vcEllipticalMaterial;
	g_vcEllipticalPolyline2D->appearance = appearance;
	g_vcFuseSurfaceFlatStageAidShape->appearance = appearance;
	g_vcECScaleBase->children.refValue()+=g_vcEllipticalPolyline2D;
	g_vcBezierBase->children.refValue()+=g_vcEllipticalPolyline2D;
	g_vcCCLatheBase->children.refValue()+=g_vcEllipticalPolyline2D;
	g_vcTransferBCylBase->children.refValue()+=g_vcEllipticalPolyline2D;
	g_vcTransferBCylBase->children.refValue()+=g_vcTransferBCyl.p;

	g_vcECScaleTangentArrow=new Transform;
	Shape* shape=new Shape;
	shape->appearance=appearance;
	IndexedLineSet* polyline2D=new IndexedLineSet;
	Coordinate* polyline2DCoord=new Coordinate;
	polyline2D->coord=polyline2DCoord;
	polyline2DCoord->point=M_Vec3f(Vec3f(-halfArrowlen,0,0),Vec3f(halfArrowlen,0,0));
	shape->geometry=polyline2D;
	g_vcECScaleTangentArrow->children.refValue()+=shape;

	Transform* transform=new Transform;
	Appearance* sharedAppearance=new Appearance;
	shape->appearance=sharedAppearance;
	transform->translation=Vec3f(-halfArrowlen,0,0);
	transform->children.refValue()+=g_lastPoint.p;
	g_vcECScaleTangentArrow->children.refValue()+=transform;


	shape=new Shape;
	appearance=new Appearance;
	Material* material=new Material;
	material->diffuseColor=Color3f(1,0,0);
	appearance->material=material;
	shape->appearance=appearance;
	Cone* cone=new Cone;
	shape->geometry=cone;
	cone->bottomRadius=MouseSense*3;
	cone->height=MouseSense*5;
	transform=new Transform;
	transform->translation=Vec3f(halfArrowlen,0,0);
	transform->rotation=Rotation(0,0,1,-PI*.5);
	transform->children.refValue()+=shape;
	g_vcECScaleTangentArrow->children.refValue()+=transform;

	g_vcECScaleBase->children.refValue()+=g_vcECScaleTangentArrow;


	g_nailDown += g_visualClue;
	g_nailDown += g_vcEllipticalMaterial;
	g_nailDown += g_stickManMaterial;
	g_nailDown += g_vcSelectRectMaterial;
	g_nailDown += g_vcSelectRectGroup;
}
void initRadialTug(){
	g_radialTugStem=new Stem;
	g_radialTugStem->sync=true;
	g_radialTugStem->part.resize(1);
	SurfaceGenerator* surface=new SurfaceGenerator;
	g_radialTugStem->surface=surface;

	g_radialTug=new RadialTug;
	g_radialTug->radius=0;
	g_radialTug->location.refValue().length=1;
	g_radialTug->displacement.refValue().length=1;
	g_radialTug->slope.refValue().length=1;
	g_radialTug->name="immediate";
	surface->generator.refValue()+=g_radialTug;
}
/*
void pointToPartByDefault(){
	g_selNode=g_scene;
	if(!g_scene->children.length)return;
	if(g_scene->children[0]->getType()!=SEAMLESS)return;
	SeamlessPtr smls=g_scene->children[0];
	if(!smls->skeleton.length)return;
	g_selNode=smls->skeleton[0];
}
*/
void genCylinder(float height, float radius, bool side, bool top, bool bottom, int staves, int bands, GeomSource& geomSource , Effect* effect){
	Seamless smls;
	smls.effect=effect;
	Art* part = new Part;
	smls.skeleton+=part;
	M_Vec3f& coord=part->coord;
	if(side){
		CCLathe* lathe=new CCLathe;
		SurfaceGenerator* surface=new SurfaceGenerator;
		surface->generator.refValue()+=lathe;
		SurfaceTransform* transform = new SurfaceTransform;
		surface->generator.refValue()+=transform;
		transform->translation=Vec3f(0,-height*.5,0);

		Stem stem;
		stem.sync.refValue()=true;
		stem.surface=surface;
		stem.part.refValue()+=part;
		stem.staves.refValue()=staves;
		stem.bands=bands;
		effect->addTextureMapper(stem);

		lathe->curve.refValue()+=0;
		lathe->curve.refValue()+=-.0001;
		lathe->distance.refValue()+=height;
		lathe->baseRad=radius;

		stem.doBuild();
		JoinCloseVertex joinCloseVertex;
		joinCloseVertex.part+=part;
		joinCloseVertex.doBuild();
	}
	if(top){
		Art* topCap = new Part;
		smls.skeleton+=topCap;
		CCLathe* lathe=new CCLathe;
		SurfaceGenerator* surface=new SurfaceGenerator;
		surface->generator.refValue()+=lathe;
		SurfaceTransform* transform = new SurfaceTransform;
		surface->generator.refValue()+=transform;
		transform->translation=Vec3f(0,height*.5,0);

		Stem stem;
		stem.sync.refValue()=true;
		stem.surface=surface;
		stem.part.refValue()+=topCap;
		stem.staves.refValue()=staves;
		stem.bands=bands;
		stem.beginSP=true;
		stem.insideOut=true;

		lathe->curve.refValue()+=PI*.5;
		lathe->curve.refValue()+=-.0001;
		lathe->distance.refValue()+=radius;

		stem.doBuild();
		JoinCloseVertex joinCloseVertex;
		joinCloseVertex.part+=topCap;
		joinCloseVertex.doBuild();

		if(side){

			M_Vec3f& coordTopCap=topCap->coord;

			for(int i=staves; --i>=0;){
				coordTopCap[i+1]=coord[staves+i];
			}
		}
		effect->textureMapDisk(topCap, radius, -PI*.5);

		TransferVertex transferVertex;
		transferVertex.source=topCap;
		transferVertex.dest=part;
		transferVertex.doBuild();
	}
	if(bottom){
		Art* base = new Part;
		smls.skeleton+=base;
		CCLathe* lathe=new CCLathe;
		SurfaceGenerator* surface=new SurfaceGenerator;
		surface->generator.refValue()+=lathe;
		SurfaceTransform* transform = new SurfaceTransform;
		surface->generator.refValue()+=transform;
		transform->translation=Vec3f(0,-height*.5,0);

		Stem stem;
		stem.sync.refValue()=true;
		stem.surface=surface;
		stem.part.refValue()+=base;
		stem.staves.refValue()=staves;
		stem.bands=bands;
		stem.beginSP=true;
		effect->addTextureMapper(stem);

		lathe->curve.refValue()+=PI*.5;
		lathe->curve.refValue()+=-.0001;
		lathe->distance.refValue()+=radius;

		TransferVertex transferVertex;
		transferVertex.source=base;
		transferVertex.dest=part;

		stem.doBuild();
		JoinCloseVertex joinCloseVertex;
		joinCloseVertex.part+=base;
		joinCloseVertex.doBuild();

		if(side){

			M_Vec3f& coordBase=base->coord;

			for(int i=staves; --i>=0;){
				coordBase[i+1]=coord[i];
			}
		}
		effect->textureMapDisk(base, radius, PI*.5);

		transferVertex.doBuild();
	}
	part->extractGeometryGS(geomSource, g_tmpIndex);
}
void genSphere(float rad, int staves, int bands, GeomSource& geomSource, Effect* effect){
	Seamless smls;
	smls.effect=effect;
	Art* part = new Part;
	smls.skeleton+=part;
	CCLathe* lathe=new CCLathe;
	SurfaceGenerator* surface=new SurfaceGenerator;
	surface->generator.refValue()+=lathe;
	SurfaceTransform* transform = new SurfaceTransform;
	surface->generator.refValue()+=transform;

	transform->translation=Vec3f(0,-rad,0);

	Stem stem;
	stem.sync=true;
	stem.surface=surface;
	stem.beginSP=true;
	stem.endSP=true;
	stem.part+=part;
	stem.staves=staves;
	stem.bands=bands;
	effect->addTextureMapper(stem);

	lathe->curve.refValue()+=-PI*.5;
	lathe->curve.refValue()+=PI;
	lathe->distance.refValue()+=rad*PI;

	stem.doBuild();
	JoinCloseVertex joinCloseVertex;
	joinCloseVertex.part+=part;
	joinCloseVertex.doBuild();

	part->extractGeometryGS(geomSource, g_tmpIndex);
}
void genCone(float height, float radius, bool side, bool bottom, int staves, GeomSource& geomSource, Effect* effect){
	Seamless smls;
	smls.effect=effect;
	Art* part = new Part;
	smls.skeleton+=part;
	M_Vec3f& coord=part->coord;
	if(side){
		CCLathe* lathe=new CCLathe;
		SurfaceGenerator* surface=new SurfaceGenerator;
		surface->generator.refValue()+=lathe;
		SurfaceTransform* transform = new SurfaceTransform;
		surface->generator.refValue()+=transform;
		transform->translation=Vec3f(0,-height*.5,0);

		Stem stem;
		stem.sync.refValue()=true;
		stem.surface=surface;
		stem.part.refValue()+=part;
		stem.staves.refValue()=staves;
		stem.bands=1;
		stem.xJoinEnds=false;
		effect->addTextureMapper(stem);

		DelPoly delPoly;
		delPoly.part=part;
		M_Int& poly=delPoly.poly;
		poly.length=staves;
		for(int i=0; i<staves; i++){
			poly[i]=i*2+1;
		}

		float hyp=sqrt(height*height+radius*radius);
		float angle;
		if(height>radius){
			angle=PI*.5-asin(height/hyp);
		}
		else{
			angle=PI*.5-acos(radius/hyp);
		}
		lathe->curve.refValue()+=angle;
		lathe->curve.refValue()+=-.0001;
		lathe->distance.refValue()+=hyp;
		lathe->baseRad=radius;
		stem.doBuild();
		delPoly.doBuild();

		JoinVertex joinVertex;
		joinVertex.part=part;
		joinVertex.srcCoord.refValue()+=2;
		joinVertex.dstCoord.refValue()+=3*(staves-1);
		joinVertex.doBuild();

		for( int i=staves; --i>=0;){
			coord[staves+i]=Vec3f(0,height*.5,0);
		}
	}
	if(bottom){
		Art* base = new Part;
		smls.skeleton+=base;
		CCLathe* lathe=new CCLathe;
		SurfaceGenerator* surface=new SurfaceGenerator;
		surface->generator.refValue()+=lathe;
		SurfaceTransform* transform = new SurfaceTransform;
		surface->generator.refValue()+=transform;
		transform->translation=Vec3f(0,-height*.5,0);

		Stem stem;
		stem.sync.refValue()=true;
		stem.surface=surface;
		stem.part.refValue()+=base;
		stem.staves.refValue()=staves;
		stem.bands=1;
		stem.beginSP=true;
		effect->addTextureMapper(stem);

		lathe->curve.refValue()+=PI*.5;
		lathe->curve.refValue()+=-.0001;
		lathe->distance.refValue()+=radius;

		TransferVertex transferVertex;
		transferVertex.source=base;
		transferVertex.dest=part;

		stem.doBuild();
		JoinCloseVertex joinCloseVertex;
		joinCloseVertex.part+=base;
		joinCloseVertex.doBuild();

		if(side){

			M_Vec3f& coordBase=base->coord;

			for(int i=staves; --i>=0;){
				coordBase[i+1]=coord[i];
			}
		}
		effect->textureMapDisk(base, radius, PI*.5);

		transferVertex.doBuild();
	}
	part->extractGeometryGS(geomSource, g_tmpIndex);
}
void genBox(Vec3f size, int staves, int bands, GeomSource& geomSource, Effect* effect){
	Seamless smls;
	smls.effect=effect;
	Art* part = new Part;
	Art* part2 = new Part;
	smls.skeleton+=part;
	smls.skeleton+=part2;
	smls.process();
	smls.doBuild();

	Stem stem;
	M_Vec3f& coord=part->coord;
	effect->addTextureMapper(stem);
	TextureEffect* e=dynamic_cast<TextureEffect*>(effect);
	M_Vec2f& texCoord=e->texCoord;
	stem.part.refValue()+=part2;
	stem.staves.refValue()=staves;
	stem.bands.refValue()=bands;
	stem.xDistance=1;
	stem.yDistance=1;
	stem.xJoinEnds=false;
	stem.doBuild();
	Place place;
	place.part+=part2;
	place.translation=Vec3f(0,-.5,.5);
	place.doBuild();
	CopyPart copyPart;
	copyPart.dest+=part;
	copyPart.source+=part2;
	copyPart.doBuild();
	copyPart.mirrorX=true;
	copyPart.mirrorZ=true;
	copyPart.add=true;
	copyPart.doBuild();
	copyPart.mirrorX=false;
	copyPart.mirrorZ=false;
	place.translation=Vec3f(0,0,0);
	place.rotation=Rotation(0,1,0,PI*.5);
	place.doBuild();
	copyPart.doBuild();
	copyPart.mirrorX=true;
	copyPart.mirrorZ=true;
	copyPart.doBuild();
	place.rotation=Rotation(0,1,0,-PI*.5)*Rotation(1,0,0,-PI*.5);
	place.doBuild();
	copyPart.mirrorX=false;
	copyPart.mirrorZ=false;
	copyPart.doBuild();
	copyPart.mirrorY=true;
	copyPart.mirrorZ=true;
	copyPart.doBuild();
	place.part.refValue()[0]=part;
	place.scale=size;
	place.rotation=Rotation(0,0,1,0);
	place.doBuild();

	part->extractGeometryGS(geomSource, g_tmpIndex);
}
void Cylinder::convertToGeomSource(GeomSource& geomSource, const Appearance*& appearance, M_Int& maskIndex)const{
	geomSource.reset();
	genCylinder( height, radius, side, top, bottom, 12, 1, geomSource,workOutEffectType(appearance));
}
void Sphere::convertToGeomSource(GeomSource& geomSource, const Appearance*& appearance, M_Int& maskIndex)const{
	geomSource.reset();
	genSphere(radius, 12, 6, geomSource,workOutEffectType(appearance));
}
void Box::convertToGeomSource(GeomSource& geomSource, const Appearance*& appearance, M_Int& maskIndex)const{
	geomSource.reset();
	genBox( size, 1, 1, geomSource, workOutEffectType(appearance));
}
void Cone::convertToGeomSource(GeomSource& geomSource, const Appearance*& appearance, M_Int& maskIndex)const{
	geomSource.reset();
	genCone(height, bottomRadius, side, bottom, 16, geomSource,workOutEffectType(appearance));
}

/*
void workOutActTriFromActPoint(){
	ArtPtr pj=g_selNode;
	SeamlessPtr smls=pj->smls;
	M_Int& coordIndex=smls->refCoordIndex();

	for(int t=0;t<coordIndex.length;t+=4){
		if(coordIndex[t]==smls->ActCtrlPnt){
			smls->ActTri=t/4;
			smls->ActTriCorner=t=0;
			break;
		}
		if(coordIndex[t+1]==smls->ActCtrlPnt){
			smls->ActTri=t/4;
			smls->ActTriCorner=t=1;
			break;
		}
		if(coordIndex[t+2]==smls->ActCtrlPnt){
			smls->ActTri=t/4;
			smls->ActTriCorner=t=2;
			break;
		}
	}
}
*/
bool backFace(const Vec3f& v0, const Vec3f& v1, const Vec3f& v2){
Vec3f fNormal;
	calcNormal(
		v0,
		v1,
		v2,
		fNormal
	);
	return fNormal.z < 0;
}
bool backFace(const Vec2f& v0, const Vec2f& v1, const Vec2f& v2){
Vec3f fNormal;
	calcNormal(
		Vec3f(v0.x, v0.y, 0),
		Vec3f(v1.x, v1.y, 0),
		Vec3f(v2.x, v2.y, 0),
		fNormal
	);
	return fNormal.z < 0;
}
bool backFace(Vec2f* v){
Vec3f fNormal;
	calcNormal(
		Vec3f(v[0].x, v[0].y, 0),
		Vec3f(v[1].x, v[1].y, 0),
		Vec3f(v[2].x, v[2].y, 0),
		fNormal
	);
	return fNormal.z < 0;
}

int backFace3(Vec3f *p){
	Vec3f v0=p[0];
	Vec3f v1=p[1];
	Vec3f v2=p[2];
	float x1=v0.x;
	float x2=v1.x;
	float x3=v2.x;

	float y1=v0.y;
	float y2=v1.y;
	float y3=v2.y;

	float z1=v0.z;
	float z2=v1.z;
	float z3=v2.z;

	float c=	(x3*((z1*y2)-(y1*z2)))+
				(y3*((x1*z2)-(z1*x2)))+
				(z3*((y1*x2)-(x1*y2)));

	return (c<0);
}

int backFace2(Vec3f& v0, Vec3f& v1, Vec3f& v2){
float c=(v2.x*((v0.z*v1.y)-(v0.y*v1.z)))+
		(v2.y*((v0.x*v1.z)-(v0.z*v1.x)))+
		(v2.z*((v0.y*v1.x)-(v0.x*v1.y)));
	return (c<0);
}


bool checkOrientationPoint(const bool perspective, const Vec2f point, const float tangent, Vec3f& pt, bool& tugFlag, Matrix &partMatrix){
	Vec3f sr;
//	if(point.length==0)return false;
	Vec2f spoint=Vec2f(-halfArrowlen*g_vcECScaleTangentArrow->scale.x,0);
	rotateZ(spoint, tangent);
	spoint+=point;



	sr= partMatrix * Vec3f(spoint.x, spoint.y, 0);

	Vec3f v;
	Vec3f mv;

	if(perspective){
		v.x = sr.x / (-sr.z + g_viewDist);
		v.y = sr.y / (-sr.z + g_viewDist);

		mv.x=pt.x/(-pt.z + g_viewDist);
		mv.y=pt.y/(-pt.z + g_viewDist);
	}
	else{
		v.x=sr.x;
		v.y=sr.y;

		mv.x=pt.x;
		mv.y=pt.y;
	}
	
	if(
		v.x < mv.x+MouseSense&&
		v.x > mv.x-MouseSense&&
		v.y < mv.y+MouseSense&&
		v.y > mv.y-MouseSense){

				
		pt.x=sr.x;
		pt.y=sr.y;
		pt.z=sr.z;

		if(!(0x8000&GetKeyState(VK_LCONTROL))){
			tugFlag=true;
			showCursor(FALSE);
		}
		orientationTugFlag=true;

		return true;
	}
	else return false;
}
bool RingBuild::getGoodQuadrant(Matrix& matrix){
	int i = index[0];
	matrix = rotationMatrix(orientation.refValue()[i]) * translationMatrix(location.refValue()[i]) * 
		scaleMatrix(scale.refValue()) * rotationMatrix(rotation.refValue()) * translationMatrix(translation.refValue());

	g_vcRingBuildPivotPoint->translation=matrix * Vec3f();
	Vec3f a = matrix * g_scene->getViewMatrix() * Vec3f(scaleX.refValue()[i].y,0,0);
	Vec3f b = matrix * g_scene->getViewMatrix() * Vec3f(0,0,scaleZ.refValue()[i].y);
	Vec3f screenPivotPoint = matrix * g_scene->getViewMatrix() * Vec3f();
	Vec2f aDiff = Vec2f(a.x - screenPivotPoint.x, a.y - screenPivotPoint.y);
	Vec2f bDiff = Vec2f(b.x - screenPivotPoint.x, b.y - screenPivotPoint.y);
	float distanceA = sqrt(aDiff.x*aDiff.x + aDiff.y*aDiff.y);
	float distanceB = sqrt(bDiff.x*bDiff.x + bDiff.y*bDiff.y);
	return distanceA > distanceB;
}
inline bool BendyLathe::checkPoint(const bool& perspective, int& idx, const Matrix& matrix){
	M_Vec3f& dpoint = patch->dpoint;
	int numOfU = patch->dim.x;
	int numOfV = patch->dim.y;
	Vec3f selPoint = ringBase[idx];

	Vec3f sr = matrix * selPoint;
	Vec3f v;
	Vec3f mv;

	if(perspective){
		v.x = sr.x / (-sr.z + g_viewDist);
		v.y = sr.y / (-sr.z + g_viewDist);

		mv.x = g_pointer.translation.x / (-g_pointer.translation.z + g_viewDist);
		mv.y = g_pointer.translation.y / (-g_pointer.translation.z + g_viewDist);
	}
	else{
		v.x=sr.x;
		v.y=sr.y;

		mv.x=g_pointer.translation.x;
		mv.y=g_pointer.translation.y;
	}
	
	if(
		v.x < mv.x+MouseSense&&
		v.x > mv.x-MouseSense&&
		v.y < mv.y+MouseSense&&
		v.y > mv.y-MouseSense){

		g_pointer.translation=sr;

		if(!(0x8000&GetKeyState(VK_LCONTROL))){
			tugFlag=true;
			showCursor(FALSE);
		}
		radIdx = -1;

		return true;
	}
	double a = PI/ctrlStaves;
	for(int q = 0, c = 1; q < ctrlStaves; q++, c+=2){
		double b = a*c;
		sr = matrix * (ringRot[idx] * Vec3f(-sin(b)*radius[idx]*ratio, 0, -cos(b)*radius[idx]*ratio) + selPoint);
		
		Vec3f v;
		Vec3f mv;

		if(perspective){
			v.x = sr.x / (-sr.z + g_viewDist);
			v.y = sr.y / (-sr.z + g_viewDist);

			mv.x = g_pointer.translation.x / (-g_pointer.translation.z + g_viewDist);
			mv.y = g_pointer.translation.y / (-g_pointer.translation.z + g_viewDist);
		}
		else{
			v.x=sr.x;
			v.y=sr.y;

			mv.x=g_pointer.translation.x;
			mv.y=g_pointer.translation.y;
		}
		if(
			v.x < mv.x+MouseSense&&
			v.x > mv.x-MouseSense&&
			v.y < mv.y+MouseSense&&
			v.y > mv.y-MouseSense){

			g_pointer.translation=sr;
//			String str;
//			sr.toString(str);
//			dLog(str);

			if(!(0x8000&GetKeyState(VK_LCONTROL))){
				tugFlag=true;
				showCursor(FALSE);
			}
			radIdx=q;
			dLog(radIdx);

			return true;
		}
	}
	return false;
}
inline bool RingBuild::checkPoint(const bool& perspective, int& idx, const int& i, const Matrix& baseMatrix){
	Vec3f sr;
	if(scaleX.length==0)return false;

	Matrix matrix = rotationMatrix(orientation.refValue()[i]) * translationMatrix(location.refValue()[i]) * baseMatrix;

	for(int q = 5; --q>=0;){
		switch(q){
			case 0:
				sr= matrix * Vec3f(0, 0, -scaleZ[i].x);
				break;
			case 1:
				sr= matrix * Vec3f(-scaleX[i].x, 0, 0);
				break;
			case 2:
				sr= matrix * Vec3f(0, 0, scaleZ[i].y);
				break;
			case 3:
				sr= matrix * Vec3f(scaleX[i].y, 0, 0);
				break;
			case 4:
				sr= matrix * Vec3f(0, 0, 0);
				break;
		}

		Vec3f v;
		Vec3f mv;

		if(perspective){
			v.x = sr.x / (-sr.z + g_viewDist);
			v.y = sr.y / (-sr.z + g_viewDist);

			mv.x = g_pointer.translation.x / (-g_pointer.translation.z + g_viewDist);
			mv.y = g_pointer.translation.y / (-g_pointer.translation.z + g_viewDist);
		}
		else{
			v.x=sr.x;
			v.y=sr.y;

			mv.x=g_pointer.translation.x;
			mv.y=g_pointer.translation.y;
		}
		
		if(
			v.x < mv.x+MouseSense&&
			v.x > mv.x-MouseSense&&
			v.y < mv.y+MouseSense&&
			v.y > mv.y-MouseSense){


			{/*
				Matrix matrix;
				if(getGoodQuadrant(matrix)){
					if(q == 0 || q == 2){
						return false;
					}
				}
				else{
					if(q == 1 || q == 3){
						return false;
					}
				}*/
			}

			g_pointer.translation.x=sr.x;
			g_pointer.translation.y=sr.y;
			g_pointer.translation.z=sr.z;

			if(!(0x8000&GetKeyState(VK_LCONTROL))){
				tugFlag=true;
				showCursor(FALSE);
			}
			idx=i;
			quadrant=q;
			dynamic_cast<VisualAid*>(g_selNode.getRaw())->houseKeep();

			return true;
		}
	}
	return false;
}
inline bool checkPoint(const bool perspective, const Vec2f point, int& idx, int i, Matrix &partMatrix){//used by both bezier and EC2, and CCLathe
	Vec3f sr;

	sr= partMatrix * Vec3f(point.x, point.y, 0);

	Vec3f v;
	Vec3f mv;

	if(perspective){
		v.x = sr.x / (-sr.z + g_viewDist);
		v.y = sr.y / (-sr.z + g_viewDist);

		mv.x = g_pointer.translation.x / (-g_pointer.translation.z + g_viewDist);
		mv.y = g_pointer.translation.y / (-g_pointer.translation.z + g_viewDist);
	}
	else{
		v.x=sr.x;
		v.y=sr.y;

		mv.x=g_pointer.translation.x;
		mv.y=g_pointer.translation.y;
	}
	
	if(
		v.x < mv.x+MouseSense&&
		v.x > mv.x-MouseSense&&
		v.y < mv.y+MouseSense&&
		v.y > mv.y-MouseSense){

				
		g_pointer.translation.x=sr.x;
		g_pointer.translation.y=sr.y;
		g_pointer.translation.z=sr.z;

		if(!(0x8000&GetKeyState(VK_LCONTROL))){
			tugFlag=true;
			showCursor(FALSE);
		}
		idx=i;
		dynamic_cast<VisualAid*>(g_selNode.getRaw())->houseKeep();

		return true;
	}
	else return false;
}
Vec2f convertPointToScreen2d(Vec3f point){
	Vec3f sr = g_viewMatrix * point;

	Vec2f v;
	if(g_settings->perspective){
		v.x = sr.x / (-sr.z + g_viewDist);
		v.y = sr.y / (-sr.z + g_viewDist);
	}
	else{
		v.x=sr.x;
		v.y=sr.y;
	}
	return v;
}

Vec2f convertPointToScreen2dNoMatrix(Vec3f sr){

	Vec2f v;
	if(g_settings->perspective){
		v.x = sr.x / (-sr.z + g_viewDist);
		v.y = sr.y / (-sr.z + g_viewDist);
	}
	else{
		v.x=sr.x;
		v.y=sr.y;
	}
	return v;
}
void MultiSelect::processRectangle(){
	if(topLeft.x > bottomRight.x){
		float store = topLeft.x;
		topLeft.x = bottomRight.x;
		bottomRight.x = store;
	}
	if(topLeft.y < bottomRight.y){
		float store = topLeft.y;
		topLeft.y = bottomRight.y;
		bottomRight.y = store;
	}
	if(g_settings->perspective){

		Vec2f vec = convertPointToScreen2dNoMatrix(MultiSelect::topLeft);
		MultiSelect::topLeft.x = vec.x;
		MultiSelect::topLeft.y = vec.y;
		vec = convertPointToScreen2dNoMatrix(MultiSelect::bottomRight);
		MultiSelect::bottomRight.x = vec.x;
		MultiSelect::bottomRight.y = vec.y;
	}
}

bool MultiSelect::isInBox(Vec3f point){
	Vec2f v = convertPointToScreen2d(point);
	if(v.x > topLeft.x && v.x < bottomRight.x && v.y < topLeft.y && v.y > bottomRight.y){
		return true;
	}
	return false;
}
bool MultiSelect::isInBox(Vec2f v){
	if(v.x > topLeft.x && v.x < bottomRight.x && v.y < topLeft.y && v.y > bottomRight.y){
		return true;
	}
	return false;
}
Vec2f convertMouseToScreen2d(){
	Vec2f mv;
	if(g_settings->perspective){

		mv.x = g_pointer.translation.x / (-g_pointer.translation.z + g_viewDist);
		mv.y = g_pointer.translation.y / (-g_pointer.translation.z + g_viewDist);
	}
	else{
		mv.x=g_pointer.translation.x;
		mv.y=g_pointer.translation.y;
	}
	return mv;
}
inline bool checkPoint(const bool perspective, const Vec3f sr, int& idx, int i){//used by NurbsPatch, bezier3f and BezierPatch
	Vec3f v;
	Vec3f mv;

	if(perspective){
		v.x = sr.x / (-sr.z + g_viewDist);
		v.y = sr.y / (-sr.z + g_viewDist);

		mv.x = g_pointer.translation.x / (-g_pointer.translation.z + g_viewDist);
		mv.y = g_pointer.translation.y / (-g_pointer.translation.z + g_viewDist);
	}
	else{
		v.x=sr.x;
		v.y=sr.y;

		mv.x=g_pointer.translation.x;
		mv.y=g_pointer.translation.y;
	}
	
	if(
		v.x < mv.x+MouseSense&&
		v.x > mv.x-MouseSense&&
		v.y < mv.y+MouseSense&&
		v.y > mv.y-MouseSense){

				
		g_pointer.translation.x=sr.x;
		g_pointer.translation.y=sr.y;
		g_pointer.translation.z=sr.z;

		if(!(0x8000&GetKeyState(VK_LCONTROL))){
			tugFlag=true;
			showCursor(FALSE);
		}
		idx=i;
		dynamic_cast<VisualAid*>(g_selNode.getRaw())->houseKeep();

		return true;
	}
	else return false;
}
inline bool valid_TL(const bool& perspective, const Vec2f& sel, const int& u, const int& v, const int& uDim, const int& vDim, const Matrix& matrix, const M_Vec3f& dpoint, bool insideOut){
	if(u > 0 && v < vDim - 1){
		int i0 = (v+1)*uDim + u;
		Vec2f pnt0 = procPerspective(perspective, matrix*dpoint[i0]);
		int i1 = v*uDim + (u-1);
		Vec2f pnt1 = procPerspective(perspective, matrix*dpoint[i1]);
		
		if(insideOut){
			if(backFace(pnt0, pnt1, sel)){
				return false;
			}
			return true;
		}
		else{
			if(backFace(pnt0, pnt1, sel)){
				return true;
			}
			return false;
		}
	}
	return false;
}

inline bool valid_BL(const bool& perspective, const Vec2f& sel, const int& u, const int& v, const int& uDim, const int& vDim, const Matrix& matrix, const M_Vec3f& dpoint, bool insideOut){
	if(u > 0 && v > 0){
		int i0 = v*uDim + (u-1);
		Vec2f pnt0 = procPerspective(perspective, matrix*dpoint[i0]);
		int i1 = (v-1)*uDim + u;
		Vec2f pnt1 = procPerspective(perspective, matrix*dpoint[i1]);
		
		if(insideOut){
			if(backFace(pnt0, pnt1, sel)){
				return false;
			}
			return true;
		}
		else{
			if(backFace(pnt0, pnt1, sel)){
				return true;
			}
			return false;
		}
	}
	return false;
}

inline bool valid_TR(const bool& perspective, const Vec2f& sel, const int& u, const int& v, const int& uDim, const int& vDim, const Matrix& matrix, const M_Vec3f& dpoint, bool insideOut){
	if(u < uDim - 1 && v < vDim - 1){
		int i0 = v*uDim + (u+1);
		Vec2f pnt0 = procPerspective(perspective, matrix*dpoint[i0]);
		int i1 = (v+1)*uDim + u;
		Vec2f pnt1 = procPerspective(perspective, matrix*dpoint[i1]);
		
		if(insideOut){
			if(backFace(pnt0, pnt1, sel)){
				return false;
			}
			return true;
		}
		else{
			if(backFace(pnt0, pnt1, sel)){
				return true;
			}
			return false;
		}
	}
	return false;
}

inline bool valid_BR(const bool& perspective, const Vec2f& sel, const int& u, const int& v, const int& uDim, const int& vDim, const Matrix& matrix, const M_Vec3f& dpoint, bool insideOut){
	if(u < uDim - 1 && v > 0){
		int i0 = (v-1)*uDim + u;
		Vec2f pnt0 = procPerspective(perspective, matrix*dpoint[i0]);
		int i1 = v*uDim + (u+1);
		Vec2f pnt1 = procPerspective(perspective, matrix*dpoint[i1]);
		
		if(insideOut){
			if(backFace(pnt0, pnt1, sel)){
				return false;
			}
			return true;
		}
		else{
			if(backFace(pnt0, pnt1, sel)){
				return true;
			}
			return false;
		}
	}
	return false;
}

inline bool NurbsPatch::checkPointIgnoreBackface(const bool& perspective, const Vec3f& sr, int& idx, const int& u, const int& v, const int& uDim, const int& vDim, const int& i, const Matrix& matrix, const M_Vec3f& dpoint){//used by NurbsPatch
	Vec2f sel;
	Vec3f mv;

	if(perspective){
		sel.x = sr.x / (-sr.z + g_viewDist);
		sel.y = sr.y / (-sr.z + g_viewDist);

		mv.x = g_pointer.translation.x / (-g_pointer.translation.z + g_viewDist);
		mv.y = g_pointer.translation.y / (-g_pointer.translation.z + g_viewDist);
	}
	else{
		sel.x=sr.x;
		sel.y=sr.y;

		mv.x=g_pointer.translation.x;
		mv.y=g_pointer.translation.y;
	}
	
	if(
		sel.x < mv.x+MouseSense&&
		sel.x > mv.x-MouseSense&&
		sel.y < mv.y+MouseSense&&
		sel.y > mv.y-MouseSense){

		if(MultiSelect::backfaceIgnore && !(beginSP && v == 0) && !(endSP && v == vDim - 1)){
			bool insideOut = false;
			if(stem.length){
				if(Stem* stem = dynamic_cast<Stem*>(this->stem[0].p)){
					insideOut = stem->insideOut;
				}
			}
			if(
				!valid_TL(perspective, sel, u, v, uDim, vDim, matrix, dpoint, insideOut) &&
				!valid_BL(perspective, sel, u, v, uDim, vDim, matrix, dpoint, insideOut) &&
				!valid_TR(perspective, sel, u, v, uDim, vDim, matrix, dpoint, insideOut) &&
				!valid_BR(perspective, sel, u, v, uDim, vDim, matrix, dpoint, insideOut)){

				return false;
			}
		}

		g_pointer.remTranslation = g_pointer.translation = sr;
		g_transformCP->remRotCtrlPnt = g_transformCP->rotCtrlPnt;

		tugFlag = true;
		showCursor(FALSE);
		idx = i;

		return true;
	}
	return false;
}


bool checkPoint(const bool perspective, EC3& ec, int i, Matrix &partMatrix){
	M_Vec2f& point=ec.point;
	Vec3f sr;
	if(point.length==0)return false;


	sr= partMatrix * Vec3f(point[i].x, point[i].y, 0);

	Vec3f v;
	Vec3f mv;

	if(perspective){
		v.x = sr.x / (-sr.z + g_viewDist);
		v.y = sr.y / (-sr.z + g_viewDist);

		mv.x = g_pointer.translation.x / (-g_pointer.translation.z + g_viewDist);
		mv.y = g_pointer.translation.y / (-g_pointer.translation.z + g_viewDist);
	}
	else{
		v.x=sr.x;
		v.y=sr.y;

		mv.x=g_pointer.translation.x;
		mv.y=g_pointer.translation.y;
	}
	
	if(
		v.x < mv.x+MouseSense&&
		v.x > mv.x-MouseSense&&
		v.y < mv.y+MouseSense&&
		v.y > mv.y-MouseSense){

				
		g_pointer.translation=sr;

		if(!(0x8000&GetKeyState(VK_LCONTROL))){
			tugFlag=true;
			showCursor(FALSE);
		}
		ec.index[0]=i;
		dynamic_cast<VisualAid*>(g_selNode.getRaw())->houseKeep();

		return true;
	}
	else return false;
}

bool Art::checkPointInVertex(const bool perspective, int i, Matrix& partMatrix){
	Vec3f sr;
	const M_Vec3f& coord=this->coord;
	if(coord.length==0)return false;

	sr = partMatrix * coord[i];

	Vec3f v;
	Vec3f mv;

	if(perspective){
		v.x = sr.x / (-sr.z + g_viewDist);
		v.y = sr.y / (-sr.z + g_viewDist);

		mv.x = g_pointer.translation.x / (-g_pointer.translation.z + g_viewDist);
		mv.y = g_pointer.translation.y / (-g_pointer.translation.z + g_viewDist);
	}
	else{
		v.x=sr.x;
		v.y=sr.y;

		mv.x=g_pointer.translation.x;
		mv.y=g_pointer.translation.y;
	}
	
	if(
		v.x < mv.x+MouseSense&&
		v.x > mv.x-MouseSense&&
		v.y < mv.y+MouseSense&&
		v.y > mv.y-MouseSense){

				
		int corner;
		S_Triangle* triangle;
		S_Triangle* beginTri;
		int tri;

		Art* triPart=0;
		tri = findNextTriangleFromVertex( this, triPart, i, corner, smls->ActTri-1);
		triangle=&triPart->triangle[tri];
		beginTri=triangle;

		while(1){

			int vecIndexA=triangle->ia;
			int vecIndexB=triangle->ib;
			int vecIndexC=triangle->ic;

			Art* partA=triangle->pa;
			Art* partB=triangle->pb;
			Art* partC=triangle->pc;

			Vec3f vA = partA->matrix * g_viewMatrix * partA->coord.refValue()[vecIndexA];
			Vec3f vB = partB->matrix * g_viewMatrix * partB->coord.refValue()[vecIndexB];
			Vec3f vC = partC->matrix * g_viewMatrix * partC->coord.refValue()[vecIndexC];

			if(g_settings->backfaceIgnore){
				if(backFace(vA,vB,vC))	break; //if success exit loop 
			}
			else {
				break;
			}

			tri=findNextTriangleFromVertex(this, triPart, i, corner, tri);
			triangle=&triPart->triangle[tri];
			if(triangle!=beginTri) continue;
			else return false; //looped full circle
		}
		if(SenseTugGlobals::node == 0 || sr.z >= SenseTugGlobals::hitZ){

			SenseTugGlobals::pointer=sr;
			SenseTugGlobals::hitZ=sr.z;
			SenseTugGlobals::node=this;
			SenseTugGlobals::actTriPart=triPart;
			SenseTugGlobals::ActTri=tri;
			SenseTugGlobals::corner=corner;
			SenseTugGlobals::ActCtrlPnt=i;
			return true;
		}

	}
	return false;
}
//=IF(AND(((D$4-D6)*(C7-C6)-(C$4-C6)*(D7-D6))*((D$4-D7)*(C8-C7)-(C$4-C7)*(D8-D7))>-0.0001,((D$4-D7)*(C8-C7)-(C$4-C7)*(D8-D7))*((D$4-D8)*(C6-C8)-(C$4-C8)*(D6-D8))>-0.001),"Inside","Not Inside")
//=((D$4-D6)*(C7-C6)-(C$4-C6)*(D7-D6))*((D$4-D7)*(C8-C7)-(C$4-C7)*(D8-D7))
//=((D$4-D7)*(C8-C7)-(C$4-C7)*(D8-D7))*((D$4-D8)*(C6-C8)-(C$4-C8)*(D6-D8))
//4=p
//6=a
//7=b
//8=c
//C=x
//D=y
//(p.y-a.y)*(b.x-a.x)-(p.x-a.x)*(b.y-a.y))*((p.y-b.y)*(c.x-b.x)-(p.x-b.x)*(c.y-b.y) > -.00001 &&
//(p.y-a.y)*(b.x-a.x)-(p.x-a.x)*(b.y-a.y))*((p.y-b.y)*(c.x-b.x)-(p.x-b.x)*(c.y-b.y))

//(p.y-b.y)*(c.x-b.x)-(p.x-b.x)*(c.y-b.y))*((p.y-c.y)*(a.x-c.x)-(p.x-c.x)*(a.y-c.y) > -.00001){
//(p.y-b.y)*(c.x-b.x)-(p.x-b.x)*(c.y-b.y))*((p.y-c.y)*(a.x-c.x)-(p.x-c.x)*(a.y-c.y))


//thanks to:
//http://mcraefamily.com/MathHelp/GeometryPointAndTriangle.htm
//for showing how to find a point in a triangle

bool pointInTriangle(const Vec2f& p, const Vec2f& a, const Vec2f& b, const Vec2f& c){

	if(	((p.y-a.y)*(b.x-a.x)-(p.x-a.x)*(b.y-a.y))*((p.y-b.y)*(c.x-b.x)-(p.x-b.x)*(c.y-b.y)) > 0 &&
		((p.y-b.y)*(c.x-b.x)-(p.x-b.x)*(c.y-b.y))*((p.y-c.y)*(a.x-c.x)-(p.x-c.x)*(a.y-c.y)) > 0){
		return true;
	}
	return false;
}

float getDistanceFromVertex(const Vec2f& p0, const Vec2f& p1){
	float xDiff = fabs(p0.x - p1.x);
	float yDiff = fabs(p0.y - p1.y);
	return sqrt(xDiff * xDiff + yDiff * yDiff);
}
float getDistanceFromSide(Vec2f mv, const Vec2f* v, const int i){
	int nextCorner = (i + 1) % 3;

	Vec2f c = Vec2f(v[nextCorner].x - v[i].x, v[nextCorner].y - v[i].y);
	mv -= Vec2f(v[i].x, v[i].y);

	float a = -c.getAngle();
	c = c.rotate(a);
	mv = mv.rotate(a);
	return fabs(c.y - mv.y);
}
Art* g_testPart;
extern Vec2f g_storeClickPos;

inline bool findLineIntersectInTriangle2(int& c, const Vec2f& pt, const Vec3f* v, Vec3f& edgeHit){
	for(int i = 0; i < 2; i++, c++){
		c = c % 3;
		int n = (c + 1) % 3;
		if(pt.y >= v[n].y && pt.y <= v[c].y){
			Vec3f diff = v[c] - v[n];
			edgeHit.y = pt.y - v[n].y;
			float yRatio = edgeHit.y / diff.y;
			edgeHit.x = diff.x * yRatio;
			edgeHit.z = diff.z * yRatio;
			edgeHit += v[n];
			return true;
		}
	}
	return false;
}
inline bool findLineIntersectInTriangle2(int& c, const Vec2f& pt, const Vec2f* v, Vec2f& edgeHit){
	for(int i = 0; i < 2; i++, c++){
		c = c % 3;
		int n = (c + 1) % 3;
		if(pt.y >= v[n].y && pt.y <= v[c].y){
			Vec2f diff = v[c] - v[n];
			edgeHit.y = pt.y - v[n].y;
			float yRatio = edgeHit.y / diff.y;
			edgeHit.x = diff.x * yRatio;
			edgeHit += v[n];
			return true;
		}
	}
	return false;
}
inline bool findLineIntersectInTriangle(int& c, const Vec2f& pt, const Vec3f* v, Vec3f& edgeHit){
	for(int i = 0; i < 3; i++, c++){
		c = c % 3;
		int n = (c + 1) % 3;
		if(pt.y >= v[c].y && pt.y <= v[n].y){
			Vec3f diff = v[n] - v[c];
			edgeHit.y = pt.y - v[c].y;
			float yRatio = edgeHit.y / diff.y;
			edgeHit.x = diff.x * yRatio;
			edgeHit.z = diff.z * yRatio;
			edgeHit += v[c];
			return true;
		}
	}
	return false;
}
inline bool findLineIntersectInTriangle(int& c, const Vec2f& pt, const Vec2f* v, Vec2f& edgeHit){
	for(int i = 0; i < 3; i++, c++){
		c = c % 3;
		int n = (c + 1) % 3;
		if(pt.y >= v[c].y && pt.y <= v[n].y){
			Vec2f diff = v[n] - v[c];
			edgeHit.y = pt.y - v[c].y;
			float yRatio = edgeHit.y / diff.y;
			edgeHit.x = diff.x * yRatio;
			edgeHit += v[c];
			return true;
		}
	}
	return false;
}
bool getTriangleHitPos(Vec3f& clickPos, const Vec2f& pt, const Vec3f* v){
	Vec3f edgeHitA;
	Vec3f edgeHitB;
	int c = 0;
	if(	findLineIntersectInTriangle(c,	  pt, v, edgeHitA) &&
		findLineIntersectInTriangle2(++c, pt, v, edgeHitB)){
		//if findLineIntersectInTriangle2 fails we could ask if either end in the first line was a direct hit (== || ==) return;

		if(pt.x <= edgeHitA.x && pt.x >= edgeHitB.x){

			float xDiff = edgeHitA.x - edgeHitB.x;
			float zDiff = edgeHitA.z - edgeHitB.z;
			float xRatio = (pt.x - edgeHitB.x) / xDiff;

			clickPos = Vec3f(pt.x, pt.y, edgeHitB.z + zDiff * xRatio);

			return true;
		}
	}
	return false;
}
bool isTriangleHit(const Vec2f& pt, const Vec2f* v){
	Vec2f edgeHitA;
	Vec2f edgeHitB;
	int c = 0;
	if(	findLineIntersectInTriangle(c,	  pt, v, edgeHitA) &&
		findLineIntersectInTriangle2(++c, pt, v, edgeHitB)){

		if(pt.x <= edgeHitA.x && pt.x >= edgeHitB.x){
			return true;
		}
	}
	return false;
}
bool Art::checkPointInTriangle(const bool perspective, const int& triIdx){
	Art* part[3] = {
		triangle.refPart(triIdx),
		triangle.refPart(triIdx+1),
		triangle.refPart(triIdx+2)
	};

	int crdIdx[3] = {
		triangle.refIndex(triIdx),
		triangle.refIndex(triIdx+1),
		triangle.refIndex(triIdx+2)
	};

	Vec3f sr[3] = {
		part[0]->matrix * g_viewMatrix * part[0]->coord.refValue()[crdIdx[0]],
		part[1]->matrix * g_viewMatrix * part[1]->coord.refValue()[crdIdx[1]],
		part[2]->matrix * g_viewMatrix * part[2]->coord.refValue()[crdIdx[2]]
	};

	Vec2f v[3];
	Vec2f mv;

	if(perspective){
		v[0].x = sr[0].x / (-sr[0].z + g_viewDist);
		v[0].y = sr[0].y / (-sr[0].z + g_viewDist);

		v[1].x = sr[1].x / (-sr[1].z + g_viewDist);
		v[1].y = sr[1].y / (-sr[1].z + g_viewDist);

		v[2].x = sr[2].x / (-sr[2].z + g_viewDist);
		v[2].y = sr[2].y / (-sr[2].z + g_viewDist);

		mv.x = g_pointer.translation.x / (-g_pointer.translation.z + g_viewDist);
		mv.y = g_pointer.translation.y / (-g_pointer.translation.z + g_viewDist);
	}
	else{
		v[0].x=sr[0].x;
		v[0].y=sr[0].y;

		v[1].x=sr[1].x;
		v[1].y=sr[1].y;

		v[2].x=sr[2].x;
		v[2].y=sr[2].y;

		mv.x=g_pointer.translation.x;
		mv.y=g_pointer.translation.y;
	}
	
	//if(pointInTriangle(Vec2f(mv.x, mv.y), Vec2f(v[0].x, v[0].y), Vec2f(v[1].x, v[1].y), Vec2f(v[2].x, v[2].y))){

	if(isTriangleHit(Vec2f(mv.x, mv.y), v)){
		if(g_settings->backfaceIgnore){
			if(backFace(v) == false)	return false; //if success exit loop 
		}
		g_storeClickPos = mv;
		//dLog("pointer: " + String(mv.x)  + ", " + String(mv.y));

		Vec3f v3[3]={
			Vec3f(v[0].x, v[0].y, sr[0].z),
			Vec3f(v[1].x, v[1].y, sr[1].z),
			Vec3f(v[2].x, v[2].y, sr[2].z)
		};
		Vec3f clickPos;
		getTriangleHitPos(clickPos, Vec2f(g_pointer.translation.x, g_pointer.translation.y), v3);
		if(SenseTugGlobals::node == 0 || clickPos.z > SenseTugGlobals::hitZ){

			int corner = 0;
			if((0x8000&GetKeyState(VK_LMENU) || InputEvent::l_altDown)){
				corner = 0;
				float closest = getDistanceFromSide(Vec2f(mv.x, mv.y), v, 0);
				for(int i = 1; i < 3; i++){
					float distance = getDistanceFromSide(Vec2f(mv.x, mv.y), v, i);

					if(distance < closest){
						closest = distance;
						corner = i;
					}
				}
			}
			else{
				float closest = getDistanceFromVertex(mv, v[0]);
				for(int i = 1; i < 3; i++){
					float distance = getDistanceFromVertex(mv, v[i]);
					if(distance < closest){
						closest = distance;
						corner = i;
					}
				}	
			}
			g_testPart=part[corner];


			SenseTugGlobals::pointer=sr[corner];
			SenseTugGlobals::hitZ=clickPos.z;
			SenseTugGlobals::node=part[corner];
			SenseTugGlobals::actTriPart=this;
			SenseTugGlobals::ActTri=triIdx/3;
			SenseTugGlobals::corner=corner;
			SenseTugGlobals::ActCtrlPnt=crdIdx[corner];

			return true;
		}
	}
	return false;
}
bool Art::checkPointInTriangle(const bool perspective, const int& triIdx, TouchSensor* sensor){
	Art* part[3] = {
		triangle.refPart(triIdx),
		triangle.refPart(triIdx+1),
		triangle.refPart(triIdx+2)
	};

	int crdIdx[3] = {
		triangle.refIndex(triIdx),
		triangle.refIndex(triIdx+1),
		triangle.refIndex(triIdx+2)
	};

	Vec3f sr[3] = {
		part[0]->matrix * g_viewMatrix * part[0]->coord.refValue()[crdIdx[0]],
		part[1]->matrix * g_viewMatrix * part[1]->coord.refValue()[crdIdx[1]],
		part[2]->matrix * g_viewMatrix * part[2]->coord.refValue()[crdIdx[2]]
	};

	Vec2f v[3];
	Vec2f mv;

	if(perspective){
		v[0].x = sr[0].x / (-sr[0].z + g_viewDist);
		v[0].y = sr[0].y / (-sr[0].z + g_viewDist);

		v[1].x = sr[1].x / (-sr[1].z + g_viewDist);
		v[1].y = sr[1].y / (-sr[1].z + g_viewDist);

		v[2].x = sr[2].x / (-sr[2].z + g_viewDist);
		v[2].y = sr[2].y / (-sr[2].z + g_viewDist);

		mv.x = g_pointer.translation.x / (-g_pointer.translation.z + g_viewDist);
		mv.y = g_pointer.translation.y / (-g_pointer.translation.z + g_viewDist);
	}
	else{
		v[0].x=sr[0].x;
		v[0].y=sr[0].y;

		v[1].x=sr[1].x;
		v[1].y=sr[1].y;

		v[2].x=sr[2].x;
		v[2].y=sr[2].y;

		mv.x=g_pointer.translation.x;
		mv.y=g_pointer.translation.y;
	}
	
	//if(pointInTriangle(Vec2f(mv.x, mv.y), Vec2f(v[0].x, v[0].y), Vec2f(v[1].x, v[1].y), Vec2f(v[2].x, v[2].y))){

	if(isTriangleHit(Vec2f(mv.x, mv.y), v)){
		if(g_settings->backfaceIgnore){
			if(backFace(v) == false)	return false; //if success exit loop 
		}
		g_storeClickPos = mv;
		//dLog("pointer: " + String(mv.x)  + ", " + String(mv.y));

		Vec3f v3[3]={
			Vec3f(v[0].x, v[0].y, sr[0].z),
			Vec3f(v[1].x, v[1].y, sr[1].z),
			Vec3f(v[2].x, v[2].y, sr[2].z)
		};
		Vec3f clickPos;
		getTriangleHitPos(clickPos, Vec2f(g_pointer.translation.x, g_pointer.translation.y), v3);
		if(SenseTugGlobals::node == 0 || clickPos.z > SenseTugGlobals::hitZ){

			if(sensor->name != ""){
				sensor->sendClickToServer();
			}
			sensor->hit=true;
			SenseTugGlobals::hitZ=clickPos.z;
			SenseTugGlobals::node=sensor;
			SenseTugGlobals::actTriPart=this;

			return true;
		}
	}
	return false;
}
bool Art::checkPointInTriangleForProjectVertex(const bool perspective, const int& triIdx){
	Art* part[3] = {
		triangle.refPart(triIdx),
		triangle.refPart(triIdx+1),
		triangle.refPart(triIdx+2)
	};

	int crdIdx[3] = {
		triangle.refIndex(triIdx),
		triangle.refIndex(triIdx+1),
		triangle.refIndex(triIdx+2)
	};

	Vec3f sr[3] = {
		part[0]->matrix * g_viewMatrix * part[0]->coord.refValue()[crdIdx[0]],
		part[1]->matrix * g_viewMatrix * part[1]->coord.refValue()[crdIdx[1]],
		part[2]->matrix * g_viewMatrix * part[2]->coord.refValue()[crdIdx[2]]
	};

	Vec2f v[3];
	Vec2f mv;

	if(perspective){
		v[0].x = sr[0].x / (-sr[0].z + g_viewDist);
		v[0].y = sr[0].y / (-sr[0].z + g_viewDist);

		v[1].x = sr[1].x / (-sr[1].z + g_viewDist);
		v[1].y = sr[1].y / (-sr[1].z + g_viewDist);

		v[2].x = sr[2].x / (-sr[2].z + g_viewDist);
		v[2].y = sr[2].y / (-sr[2].z + g_viewDist);

		mv.x = g_pointer.translation.x / (-g_pointer.translation.z + g_viewDist);
		mv.y = g_pointer.translation.y / (-g_pointer.translation.z + g_viewDist);
	}
	else{
		v[0].x=sr[0].x;
		v[0].y=sr[0].y;

		v[1].x=sr[1].x;
		v[1].y=sr[1].y;

		v[2].x=sr[2].x;
		v[2].y=sr[2].y;

		mv.x=g_pointer.translation.x;
		mv.y=g_pointer.translation.y;
	}
	
	//if(pointInTriangle(Vec2f(mv.x, mv.y), Vec2f(v[0].x, v[0].y), Vec2f(v[1].x, v[1].y), Vec2f(v[2].x, v[2].y))){

	if(isTriangleHit(Vec2f(mv.x, mv.y), v)){
		if(g_settings->backfaceIgnore){
			if(backFace(v) == false)	return false; //if success exit loop 
		}
		g_storeClickPos = mv;
		//dLog("pointer: " + String(mv.x)  + ", " + String(mv.y));
		return true;
	}
	return false;
}

void SceneTree::setResultsForSelectingVertex(Art* part){
	setSelNodeLine(part);
	lastSelNodeLine = selLine;
	cp->setupNodeWindow(part);
	cp->paint();
	setNodeWinTitleBar(part);
	textCell.cellStr = g_selNode->name;
	removeTypeFromList<CaretWindow>(g_callBackList);
	textCell.cellActive = false;
	paint();
	setTitleName();
}

void SceneTree::setResultsForSelectingPoint(NurbsPatch* patch){
	setSelNodeLine(patch);
	lastSelNodeLine = selLine;
	cp->setupNodeWindow(patch);
	cp->paint();
	setNodeWinTitleBar(patch);
	textCell.cellStr = g_selNode->name;
	removeTypeFromList<CaretWindow>(g_callBackList);
	textCell.cellActive = false;
	paint();
	setTitleName();
	patch->houseKeep();
}

//anim control point for part
bool checkAnimCtrlPoint(const bool perspective, Vec3f sr){
	
	Vec3f v;
	Vec3f mv;

	if(perspective){
		v.x = sr.x / (-sr.z + g_viewDist);
		v.y = sr.y / (-sr.z + g_viewDist);

		mv.x = g_pointer.translation.x / (-g_pointer.translation.z + g_viewDist);
		mv.y = g_pointer.translation.y / (-g_pointer.translation.z + g_viewDist);
	}
	else{
		v.x=sr.x;
		v.y=sr.y;

		mv.x=g_pointer.translation.x;
		mv.y=g_pointer.translation.y;
	}
	
	if(
		v.x < mv.x+MouseSense&&
		v.x > mv.x-MouseSense&&
		v.y < mv.y+MouseSense&&
		v.y > mv.y-MouseSense){

				
		g_pointer.translation.x = sr.x;
		g_pointer.translation.y = sr.y;
		g_pointer.translation.z = sr.z;
		SenseTugGlobals::node = g_visualClue;

//		if(false == Window::animMode()){
		tugFlag = true;
		showCursor(FALSE);
//		}

		return true;
	}
	else return false;
}
bool checkRadialTugSize(const bool perspective, const ImmediatePart* part){
	if(part->coord.length == 0){
		return false;
	}
	const Seamless* smls=part->smls;
	Vec3f sr = part->matrix * g_viewMatrix * part->coord[smls->ActCtrlPnt];
	sr.x += g_scene->getZoom() * g_settings->tugRadius;

	Vec3f v;
	Vec3f mv;

	if(perspective){
		v.x = sr.x / (-sr.z + g_viewDist);
		v.y = sr.y / (-sr.z + g_viewDist);

		mv.x = g_pointer.translation.x / (-g_pointer.translation.z + g_viewDist);
		mv.y = g_pointer.translation.y / (-g_pointer.translation.z + g_viewDist);
	}
	else{
		v.x=sr.x;
		v.y=sr.y;

		mv.x=g_pointer.translation.x;
		mv.y=g_pointer.translation.y;
	}
	
	if(
		v.x < mv.x+MouseSense&&
		v.x > mv.x-MouseSense&&
		v.y < mv.y+MouseSense&&
		v.y > mv.y-MouseSense){

				
		g_pointer.translation.x=sr.x;
		g_pointer.translation.y=sr.y;
		g_pointer.translation.z=sr.z;
		SenseTugGlobals::node=g_visualClue;

		tugFlag=true;
		showCursor(FALSE);
		Art::tugRadialTugSize=true;

		return true;
	}
	Art::tugRadialTugSize=false;
	return false;
}
bool checkRadialTugSizeNurbs(const bool perspective, NurbsPatch* patch){
	if(patch->point.length == 0){
		return false;
	}
	Vec3f sr = g_viewMatrix * patch->dpoint[patch->index[0]];
	sr.x += g_scene->getZoom() * g_settings->tugRadius;

	Vec3f v;
	Vec3f mv;

	if(perspective){
		v.x = sr.x / (-sr.z + g_viewDist);
		v.y = sr.y / (-sr.z + g_viewDist);

		mv.x = g_pointer.translation.x / (-g_pointer.translation.z + g_viewDist);
		mv.y = g_pointer.translation.y / (-g_pointer.translation.z + g_viewDist);
	}
	else{
		v.x=sr.x;
		v.y=sr.y;

		mv.x=g_pointer.translation.x;
		mv.y=g_pointer.translation.y;
	}
	
	if(
		v.x < mv.x+MouseSense&&
		v.x > mv.x-MouseSense&&
		v.y < mv.y+MouseSense&&
		v.y > mv.y-MouseSense){

				
		g_pointer.translation.x=sr.x;
		g_pointer.translation.y=sr.y;
		g_pointer.translation.z=sr.z;
		SenseTugGlobals::node=g_visualClue;

		tugFlag=true;
		showCursor(FALSE);
		Art::tugRadialTugSize=true;

		return true;
	}
	Art::tugRadialTugSize=false;
	return false;
}
void Node::senseTug(Matrix matrix, Matrix imatrix, TouchSensor* sensor){
	return;
}
void ProtoDeclare::senseTug(Matrix matrix, Matrix imatrix, TouchSensor* sensor){
	//if(SenseTugGlobals::sameVertexHit)return;
	if(SenseTugGlobals::node)return;
	M_Node& children=protoBody->children.refValue();
	for(int k=0; k<children.length;k++){
		children[k]->senseTug(matrix,imatrix, 0);
	}
}
int getSeamlessParent(int line){//diminished better to use Seamless* getSeamless(int& selLine){
	do{
		line = g_sceneTreeList[line].parent;
	}while(dynamic_cast<Seamless*>(g_sceneTreeList[line].node.p)==0);
	return line;
}
void BezierPatch::setNumOfGroupFromSelPoint(){
	const M_Vec2f& dim=this->dim;
	int pointIndex=index[pointGroup];
	int numOfUGroupIdx=0;

	for(int k=0; k<dim.length; k++){
		int noU=dim[k].x;
		int noV=dim[k].y;
		numOfUGroupIdx+=noU*noV;
		if(numOfUGroupIdx>pointIndex){
			index[numOfUGroup]=k;
			break;
		}
	}
}
M_Vec3f g_backUpCoord;
void backUpSmlsCoord(const M_Art& partList){
	g_backUpCoord.length=0;
	for(int c=0, k=0; k<partList.length; k++){
		Art* part=partList[k];
		M_Vec3f& coord=part->coord.refValue();
		int length=part->coord.length;
		g_backUpCoord.length+=length;
		{
			for(int i=0; i<length; c++, i++){
				g_backUpCoord[c]=coord[i];
			}
		}
	}
}

void restoreCoord(const M_Art& partList){
	for(int c=0, k=0; k<partList.length; k++){
		Art* part=partList[k];
		M_Vec3f& coord=part->coord.refValue();
		int length=part->coord.length;
		{
			for(int i=0; i<length; c++, i++){
				coord[i]=g_backUpCoord[c];
			}
		}
	}
}
/*
inline void NurbsPatch::initTug(const bool& perspective){
	tugRingPivotPoint=false;
	int& idx = index[0];
	Matrix matrix = this->matrix * g_viewMatrix;
	int j=index[0];
	const int uDim = dim.x;
	const int vDim = dim.y;

	Vec2i idx2i = Vec2i(idx % uDim, idx / uDim);
	if(checkPoint(perspective, matrix*dpoint[idx], idx, idx)){
		g_selNode->processFields();
		g_stw->cp->paint();
		tugFlag=true;
		return;
	}
	if(showAll){
		for(int v=0; v < vDim; v++){
			for(int u=0; u < uDim; u++){
				int i = v*uDim + u;
				Vec3f pnt = matrix * dpoint[i];
				if(checkPointIgnoreBackface(perspective, pnt, idx, u, v, uDim, vDim, i, matrix, dpoint)){
					g_selNode->processFields();
					g_stw->cp->paint();
					tugFlag = true;
					return;
				}
			}
		}
	}
	else{
		if(uJoinEnds){
			for(int v=0; v < vDim; v++){
				for(int u=1; u < uDim; u+=2){
					int i = v*uDim + u;
					Vec3f pnt = matrix * dpoint[i];
					if(checkPointIgnoreBackface(perspective, pnt, idx, u, v, uDim, vDim, i, matrix, dpoint)){
						g_selNode->processFields();
						g_stw->cp->paint();
						tugFlag=true;
						return;
					}
				}
			}
		}
		else{
			if(NurbsPatch::MID_1 == checkNurbsType(vDim, vKnot)){
				for(int v=0; v < vDim; v += 2){
					for(int u=0; u<uDim; u += 2){
						int i = v * uDim + u;
						Vec3f pnt = matrix*dpoint[i];
						if(checkPointIgnoreBackface(perspective, pnt, idx, u, v, uDim, vDim, i, matrix, dpoint)){
							g_selNode->processFields();
							g_stw->cp->paint();
							tugFlag=true;
							return;
						}
						if(u == 0 || u == uDim - 2){
							u--;
						}
					}
					if(v == 0 || v == vDim - 2){
						v--;
					}
				}
			}
			else{
				for(int v = 0; v < vDim; v++){
					for(int u = 0; u < uDim; u += 2){
						int i = v * uDim + u;
						Vec3f pnt = matrix * dpoint[i];
						if(checkPointIgnoreBackface(perspective, pnt, idx, u, v, uDim, vDim, i, matrix, dpoint)){
							g_selNode->processFields();
							g_stw->cp->paint();
							tugFlag = true;
							return;
						}
						if(u == 0 || u == uDim - 2){
							u--;
						}
					}
				}
			}
		}
	}
	return;
}
*/

inline void NurbsPatch::initTugCP(const bool& perspective){
	tugRingPivotPoint = false;
	int& idx = index[0];
	Matrix matrix = this->matrix * g_viewMatrix;
	int j=index[0];
	const int uDim = dim.x;
	const int vDim = dim.y;

	Vec2i idx2i = Vec2i(idx % uDim, idx / uDim);
/*
	if(uJoinEnds){
		if(idx2i.x == uDim-1){
			idx2i.x = 0;
		}
		if(beginSP){
			if(idx2i.y == 0){
				idx2i.x = 0;
				if(showAll==false){
					idx2i.x = 1;
				}
			}
		}
		if(endSP){
			if(idx2i.y == vDim-1){
				idx2i.x = 0;
				if(showAll==false){
					idx2i.x = 1;
				}
			}
		}
	}
	else{
		if(beginSP){
			pointLen-=uDim-1;
			if(idx2i.y == 0){
				idx2i.x = 0;
			}
		}
		if(endSP){
			pointLen-=uDim-1;
			if(idx2i.y == vDim-1){
				idx2i.x = 0;
			}
		}
	}
	idx = idx2i.y * uDim + idx2i.x;
*/

	if(checkPoint(perspective, matrix*dpoint[idx], idx, idx)){
		g_selNode->processFields();
		g_stw->cp->paint();
		tugFlag=true;
		return;
	}
	if(showAll){
		for(int v=0; v < vDim; v++){
			for(int u=0; u < uDim; u++){
				int i = v*uDim + u;
				Vec3f pnt = matrix*dpoint[i];
				if(checkPointIgnoreBackface(perspective, pnt, idx, u, v, uDim, vDim, i, matrix, dpoint)){
					g_selNode->processFields();
					g_stw->cp->paint();
					tugFlag=true;
					return;
				}
			}
		}
	}
	else{
		if(uJoinEnds){
			for(int v=0; v < vDim; v++){
				for(int u=1; u < uDim; u+=2){
					int i = v*uDim + u;
					Vec3f pnt = matrix*dpoint[i];
					if(checkPointIgnoreBackface(perspective, pnt, idx, u, v, uDim, vDim, i, matrix, dpoint)){
						g_selNode->processFields();
						g_stw->cp->paint();
						tugFlag=true;
						return;
					}
				}
			}
		}
		else{
			if(NurbsPatch::MID_1 == checkNurbsType(vDim, vKnot)){
				for(int v=0; v<vDim; v+=2){
					for(int u=0; u<uDim; u+=2){
						int i = v*uDim + u;
						Vec3f pnt = matrix*dpoint[i];
						if(checkPointIgnoreBackface(perspective, pnt, idx, u, v, uDim, vDim, i, matrix, dpoint)){
							g_selNode->processFields();
							g_stw->cp->paint();
							tugFlag=true;
							return;
						}
						if(u == 0 || u == uDim - 2){
							u--;
						}
					}
					if(v == 0 || v == vDim - 2){
						v--;
					}
				}
			}
			else{
				if(NurbsPatch::MID_1 == checkNurbsType(uDim, uKnot)){
					for(int v = 0; v < vDim; v++){
						for(int u = 0; u < uDim; u += 2){
							int i = v * uDim + u;
							Vec3f pnt = matrix * dpoint[i];
							if(checkPointIgnoreBackface(perspective, pnt, idx, u, v, uDim, vDim, i, matrix, dpoint)){
								g_selNode->processFields();
								g_stw->cp->paint();
								tugFlag = true;
								return;
							}
							if(u == 0 || u == uDim - 2){
								u--;
							}
						}
					}
				}
				else{
					for(int v = 0; v < vDim; v++){
						for(int u = 0; u < uDim; u++){
							int i = v * uDim + u;
							Vec3f pnt = matrix * dpoint[i];
							if(checkPointIgnoreBackface(perspective, pnt, idx, u, v, uDim, vDim, i, matrix, dpoint)){
								g_selNode->processFields();
								g_stw->cp->paint();
								tugFlag = true;
								return;
							}
						}
					}
				}
			}
		}
	}
}

Vec3f calcRingPivotPoint2(int selUdim, int from){
	Vec3f pivotPoint;
	for(int u = 0; u < selUdim; u++){
		pivotPoint += MultiSelect::beginDragPos[from + u];
	}

	return pivotPoint / selUdim;
}

bool NurbsPatch::selectSelectedPoint(const bool& perspective){
	for(int i = MultiSelect::pointList.length; --i >= 0;){
		NurbsPatch* patch = MultiSelect::pointList[i].p;
		int idx = MultiSelect::pointList[i].i;
		Matrix matrix = patch->matrix * g_viewMatrix;
		M_Vec3f& dpoint = patch->dpoint;
		Vec3f pnt = matrix*dpoint[idx];

		const int uDim = patch->dim.x;
		const int vDim = patch->dim.y;
		int v = idx / uDim;
		int u = idx - v * uDim;

		if(patch->checkPointIgnoreBackface(perspective, pnt, idx, u, v, uDim, vDim, idx, matrix, dpoint)){
			MultiSelect::clickedIndex = i;
			if(patch != g_selNode.getRaw()){
				g_stw->setResultsForSelectingPoint(patch);
			}
//				if(g_transformCP->selectRings == false){
			return true;
//				}
		}
	}
	return false;
}


inline void NurbsPatch::initMultiCageTug(const bool& perspective, M_<NurbsPatch*>& cpNodeList){

	if(MultiSelect::accumulate || (g_transformCP->selectRing == false && g_transformCP->selectRings == false)){
		if(selectSelectedPoint(perspective)){
			return;
		}
	}

	//check non selected points
	if(g_transformCP->selectRings == false){
		MultiSelect::clickedIndex = 0;
	}
	for(int k = cpNodeList.length; --k >= 0;){
		NurbsPatch* patch = cpNodeList[k];
		if(patch->cage){
			patch->initTugCP(perspective);
			if(tugFlag){
				int& idx = patch->index[0];
//					MultiSelect::selectedPoint = i;
				if(g_transformCP->selectRing || g_transformCP->selectRings){
					if(MultiSelect::accumulate == false){
						MultiSelect::beginDragPos.length = 0;
						MultiSelect::pointList.length = 0;
					}
					int idx = patch->index[0];
					int uDim = patch->dim.x;
					int vDim = patch->dim.y;
					int v = idx / uDim;
					if(g_transformCP->selectRing){
						patch->selectRing(v, uDim, vDim);
					}
					else{
						patch->selectRings(v, uDim, vDim);
					}
					Vec3f pivotPoint = calcRingPivotPoint2(MultiSelect::uRingDim, 0);

					g_transformCP->scale.refValue() = Vec3f(1, 1, 1);
					g_transformCP->scaleOrientation.refValue() = Rotation();
					g_transformCP->pivotPoint.refValue() = pivotPoint;
					g_transformCP->rotation.refValue() = Rotation();
					g_transformCP->translation.refValue() = Vec3f();

					IndexedLineSet* polyLine = dynamic_cast<IndexedLineSet*>(g_transformLeverClue->geometry.p);
					M_Vec3f& coord = dynamic_cast<Coordinate*>(polyLine->coord.p)->point;
					g_transformCP->rotCtrlPnt = g_scene->getViewMatrix() * g_transformCP->pivotPoint + Vec3f(.5, 0, 0);
					g_transformCP->remRotCtrlPnt = g_transformCP->rotCtrlPnt;

					//g_transformCP->remRotCtrlPnt = g_transformCP->rotCtrlPnt;
					coord.length = 0;
					coord += g_transformCP->pivotPoint;
					coord += g_transformCP->rotCtrlPnt;
					fillPolyLineIndex(coord, polyLine->coordIndex.refValue());
					polyLine->convertToRenderFormat(g_transformLeverClue->render);

					selectSelectedPoint(perspective);
					return;
				}
				else{
					MultiSelect::beginDragPos = patch->dpoint[idx];
					MultiSelect::pointList = ControlPoint(patch, idx);
				}

				if(patch != g_selNode.getRaw()){
					g_stw->setResultsForSelectingPoint(patch);
				}
				return;
			}
		}
	}
}

void findPartsCPReference(NurbsPatch* patch, Seamless* smls){
	int& idx = patch->index[0];
	M_Art& partList = smls->partList;
	for(int k = partList.length; --k >= 0;){
		Art* part = partList[k];
		M_ControlPoint& m_controlPoint = part->controlPoint;

		for(int i = m_controlPoint.length; --i >= 0;){
			ControlPoint& cp = m_controlPoint[i];
			if(cp.p == patch && cp.i == idx){
				MultiSelect::selectedPoint = i;

				MultiSelect::beginDragPos += patch->dpoint[idx];
				MultiSelect::pointList += ControlPoint(patch, idx);

				if(part != g_selNode.getRaw()){
					g_stw->setResultsForSelectingVertex(part);
				}

				return;
			}
		}
	}
}

inline void Art::initTugCP(const bool& perspective){

	M_<NurbsPatch*>& cpNodeList = smls->cpNodeList;
	
	for(int i = MultiSelect::pointList.length; --i >= 0;){
		NurbsPatch* patch = MultiSelect::pointList[i].p;
		int idx = MultiSelect::pointList[i].i;
		Matrix matrix = patch->matrix * g_viewMatrix;
		M_Vec3f& dpoint = patch->dpoint;
		Vec3f pnt = matrix*dpoint[idx];

		const int uDim = patch->dim.x;
		const int vDim = patch->dim.y;
		int v = idx / uDim;
		int u = idx - v * uDim;

		if(patch->checkPointIgnoreBackface(perspective, pnt, idx, u, v, uDim, vDim, idx, matrix, dpoint)){
			MultiSelect::clickedIndex = i;
			MultiSelect::beginDragPos.length = MultiSelect::pointList.length;
			for(int i = MultiSelect::pointList.length; --i >= 0;){
				NurbsPatch* patch = MultiSelect::pointList[i].p;
				int idx = MultiSelect::pointList[i].i;
				MultiSelect::beginDragPos[i] = patch->dpoint[idx];
			}
			return;
		}
	}
	//check non selected points
	MultiSelect::clickedIndex = 0;
	MultiSelect::beginDragPos.length = MultiSelect::pointList.length = 0;

	for(int k = cpNodeList.length; --k >= 0;){
		NurbsPatch* patch = cpNodeList[k];
		if(patch->cage){
			patch->initTugCP(perspective);
			if(tugFlag){
				findPartsCPReference(patch, smls);
				return;
			}
		}
	}
}

void MultiSelect::calcPivotPoint(){
	autoPivotPoint = Vec3f();
	for(int i = beginDragPos.length; --i>=0;){
		autoPivotPoint += MultiSelect::beginDragPos[i];
	}
	autoPivotPoint /= beginDragPos.length;
}
int NurbsPatch::getSelRing()const{
	int dimU = this->dim.x;
	return index[0]/dimU;
}
Vec3f NurbsPatch::calcRingPivotPoint(int v)const{
	int dimU = dim.x;
	int from = v * dimU;

	Vec3f pivotPoint;
	for(int i = dimU; --i>=0;){
		pivotPoint += MultiSelect::beginDragPos[from + i];
	}
	return pivotPoint /= dimU;
}

bool twoLineSegmentsIntersect(Vec2f AS, Vec2f AE, Vec2f BS, Vec2f BE);

bool clickLine(Vec2f click, Vec2f ptA, Vec2f ptB, Vec2f& crossPoint){
	Vec2f normal = calcNormal(ptA, ptB);
	if(normal == Vec2f()){
		return false;
	}
	Vec2f clickA = click + normal * MouseSense;
	Vec2f clickB = click + normal * -MouseSense;

	if(twoLineSegmentsIntersect(clickA, clickB, ptA, ptB)){
		crossPoint = lineIntersect(clickA, clickB, ptA, ptB);
		return true;
	}
	return false;
}

float getMidPointRatio(Vec2f ptA, Vec2f ptB, Vec2f crossPoint){
	float xDiff = ptB.x - ptA.x;
	float yDiff = ptB.y - ptA.y;
	if(fabs(xDiff) > fabs(yDiff)){
		return (crossPoint.x - ptA.x) / xDiff;
	}
	else{
		return (crossPoint.y - ptA.y) / yDiff;
	}
}

void updateNurbsCurveKnotMid0(int dim, M_Float& knot);

void NurbsLathe::insertControlPoint(){
	DPoint& dpoint = this->dpoint;
	M_Vec2f& point = this->point;
	M_Float& weight = this->weight;
	M_Float& knot = this->knot;

	for(int k = 0; k < dpoint.length - 1; k++){

		//Vec2f click3f = g_pointer.translation;
		Vec2f ptA2f = dpoint[k];
		Vec2f ptB2f = dpoint[k + 1];

		Matrix matrix = this->matrix * g_viewMatrix;

		Vec3f ptA3f = matrix * Vec3f(ptA2f.x, ptA2f.y, 0);
		Vec3f ptB3f = matrix * Vec3f(ptB2f.x, ptB2f.y, 0);

		Vec2f click = Vec2f(g_pointer.translation.x, g_pointer.translation.y);
		Vec2f ptA = Vec2f(ptA3f.x, ptA3f.y);
		Vec2f ptB = Vec2f(ptB3f.x, ptB3f.y);
		Vec2f crossPoint;
		if(clickLine(click, ptA, ptB, crossPoint)){
			
			float ratio = getMidPointRatio(ptA, ptB, crossPoint);
			Vec2f newPoint = ptA2f + (ptB2f - ptA2f) * ratio;

			point.insertElement(k + 1, newPoint);
			weight.insertElement(k + 1, 1);

			updateNurbsCurveKnotMid0(point.length, knot);

			index[0] = k + 1;
			processFields();
			updateBuild3dWindow();
			g_insertControlPoint = true;
			break;
		}
	}
/*
		g_stw->textCell.cellActive = false;
		Art* newPart = createPartOfType(part);
		//newPart->pivotPoint.refValue() = g_pointer.translation - part->absolutePivotPoint;;
		Vec3f vz = g_scene->getViewMatrix() * part->pivotPointMatrix * Vec3f(0,0,0);
		g_pointer.translation.z = vz.z;
		Vec3f newAbsolutePivotPoint = g_scene->getIViewMatrix() * g_catchIMatrix * part->imatrix * g_pointer.translation;
		Vec3f newPivotPoint = newAbsolutePivotPoint - part->absolutePivotPoint;
		part->part.refValue() += newPart;
		if(part->mirror){
			if(part->part.length == 1 && fabs(part->absolutePivotPoint.x) < .000001){
				if(fabs(newAbsolutePivotPoint.x) < 1 / g_scene->getZoom() * .05){
					newPivotPoint.x = 0;
				}
				else{
					addMirrorPart(part, newPivotPoint);
				}
			}
			else if(Art* mirrorPart = part->mirrorPart){
				addMirrorPart(mirrorPart, newPivotPoint);
			}
			else{
				addMirrorPart(part, newPivotPoint);
			}
		}
		newPart->pivotPoint.refValue() = newPivotPoint;
		
		part->smls->process();
		g_stw->genNodeList();
		g_stw->selLine = getNodeLine(newPart);
		g_stw->commonPostInsertNodeStuff();
		return;
	}
*/
}

void newNurbsTrackSegment(){
	NurbsTrack* nurbsTrack = dynamic_cast<NurbsTrack*>(g_selNode.getRaw());
	int atIndex = nurbsTrack->index[1];
	M_Float& key = nurbsTrack->key;
	M_Vec3f& keyValue = nurbsTrack->keyValue;
	M_Float& weight = nurbsTrack->weight;
	M_Float& knot = nurbsTrack->knot;
	M_Int& dim = nurbsTrack->dim;
	int d = -1;
	int i = 0;
	while(i + dim[++d] - 1 <= atIndex){
		i += dim[d] - 1;
	}
	int g = atIndex - i + 1;
	int s = dim[d];
	int h = s - g + 1;
	dim[d] = g;
	dim.insertElement(d + 1, h);

	float t = key[d] + (key[d + 1] - key[d]) * .5;
	key.insertElement(d + 1, t);

	updateNurbsCurveKnotMid0(keyValue.length, knot);
	nurbsTrack->processFields();
	updateBuild3dWindow();
	nurbsTrack->houseKeep();
}

void NurbsTrack::removeControlPoint(int atIndex){
	M_Float& key = this->key;
	M_Vec3f& keyValue = this->keyValue;
	M_Float& weight = this->weight;
	M_Float& knot = this->knot;
	M_Int& dim = this->dim;
	int d = -1;
	int i = 0;
	while(++d < dim.length && i + dim[d] - 1 <= atIndex){
		i += dim[d] - 1;
	}

	if(atIndex > 0 && i == atIndex){
		if(d == dim.length){
			d--;
			goto here;
		}
		d--;
		keyValue.deleteElement(atIndex);
		weight.deleteElement(atIndex);
		dim[d] += dim[d + 1] - 2;
		dim.deleteElement(d + 1);
		key.deleteElement(d * 2 + 1);
	}
	else{
here:
		if(dim[d] <= 3){
			return;
		}
		keyValue.deleteElement(atIndex);
		weight.deleteElement(atIndex);
		dim[d]--;
	}

	if(index[0] >= key.length){
		index[0]--;
	}
	if(atIndex >= keyValue.length){
		index[1]--;
	}
	if(index[2] >= dim.length){
		index[2]--;
	}

	updateNurbsCurveKnotMid0(keyValue.length, knot);
	processFields();
	updateBuild3dWindow();
	houseKeep();
}
void NurbsTrack::insertControlPoint(int atIndex, float ratio){
	M_Vec3f& keyValue = this->keyValue;
	M_Float& weight = this->weight;
	M_Float& knot = this->knot;

	Vec3f ptA3f = keyValue[atIndex];
	Vec3f ptB3f = keyValue[atIndex + 1];
	Vec3f newPoint = ptA3f + (ptB3f - ptA3f) * ratio;

	keyValue.insertElement(atIndex + 1, newPoint);
	weight.insertElement(atIndex + 1, 1);

	int d = -1;
	for(int i = 0; i <= atIndex + 1; i++){
		i += dim[++d] - 1;
	}
	dim.refValue()[d]++;

	index[1] = atIndex + 1;

	updateNurbsCurveKnotMid0(keyValue.length, knot);
	processFields();
	updateBuild3dWindow();
	houseKeep();
}
void NurbsTrack::insertControlPoint(){
	M_Vec3f& keyValue = this->keyValue;
	M_Float& weight = this->weight;
	M_Float& knot = this->knot;

	for(int k = 0; k < keyValue.length - 1; k++){
		Matrix matrix = g_viewMatrix;

		Vec3f ptA3f = matrix * keyValue[k];
		Vec3f ptB3f = matrix * keyValue[k + 1];

		Vec2f click = Vec2f(g_pointer.translation.x, g_pointer.translation.y);
		Vec2f ptA = Vec2f(ptA3f.x, ptA3f.y);
		Vec2f ptB = Vec2f(ptB3f.x, ptB3f.y);
		Vec2f crossPoint;
		if(clickLine(click, ptA, ptB, crossPoint)){
			
			float ratio = getMidPointRatio(ptA, ptB, crossPoint);

			insertControlPoint(k, ratio);

			g_insertControlPoint = true;
			break;
		}
	}
}

bool NurbsPatch::insertControlRowColum(const Vec2f par, const int idxA, const int idxB, Matrix matrix, M_Vec3f& dpoint, M_Vec3f& point, M_Float& weight, M_Float& uKnot, int uDim, int vDim, int& idx, Vec2f& dim, Vec2f& order){
	if(this->cage == false){
		return false;
	}

	Vec3f dpA = dpoint[idxA];
	Vec3f dpB = dpoint[idxB];

	Vec3f ptA3f = matrix * dpA;
	Vec3f ptB3f = matrix * dpB;

	Vec2f click = Vec2f(g_pointer.translation.x, g_pointer.translation.y);
	Vec2f ptA = Vec2f(ptA3f.x, ptA3f.y);
	Vec2f ptB = Vec2f(ptB3f.x, ptB3f.y);
	Vec2f crossPoint;
	if(clickLine(click, ptA, ptB, crossPoint)){
		
		float ratio = getMidPointRatio(ptA, ptB, crossPoint);
//		Vec3f newPoint = (dpB - dpA) * ratio;
		idx = idxA;
		
		doDim(par, ratio, dpoint, weight, order, dim, idx, uDim, vDim);
		if(showAll==false){
			Vec2f par2 = dim;
			if(uDim < par.x){
				if(idx%uDim==0){
					idx=20;
				}
				else{
					idx+=2;
				}
				uDim++;
				par2.x=uDim+1;
			}
			else if(uDim > par.x){
				uDim--;
				par2.x=uDim-1;
			}
			else{
				goto here;
			}
			uDim = dim.x;
			vDim = dim.y;
			doDim(par2, ratio, dpoint, weight, order, dim, idx, uDim, vDim);
		}
here:

		updateKnot(dim.x, uKnot, order.x);
		updateKnot(dim.y, vKnot, order.y);
		
		dpointsToPoints();

		if(idx >= dpoint.length){
			idx=dpoint.length-1;
		}

		processFields();
		MultiSelect::pointList.length = 0;
		smls->setPointFrom();
		smls->doBuild();
		smls->ctrf = true;
		houseKeep();
		BackUp::monostable.beginCycle();
		return true;
	}
	return false;
}

bool NurbsPatch::clickedRowColum(const int idxA, const int idxB, Matrix matrix){

	Vec3f dpA = dpoint[idxA];
	Vec3f dpB = dpoint[idxB];

	Vec3f ptA3f = matrix * dpA;
	Vec3f ptB3f = matrix * dpB;

	Vec2f click = Vec2f(g_pointer.translation.x, g_pointer.translation.y);
	Vec2f ptA = Vec2f(ptA3f.x, ptA3f.y);
	Vec2f ptB = Vec2f(ptB3f.x, ptB3f.y);
	Vec2f crossPoint;
	if(clickLine(click, ptA, ptB, crossPoint)){
		return true;
	}
	return false;
}


bool NurbsPatch::deleteControlRowColum(const Vec2f par, const int idxA, const int idxB, Matrix matrix, M_Vec3f& dpoint, M_Vec3f& point, M_Float& weight, M_Float& uKnot, int uDim, int vDim, int& idx, Vec2f& dim, Vec2f& order){

	Vec3f dpA = dpoint[idxA];
	Vec3f dpB = dpoint[idxB];

	Vec3f ptA3f = matrix * dpA;
	Vec3f ptB3f = matrix * dpB;

	Vec2f click = Vec2f(g_pointer.translation.x, g_pointer.translation.y);
	Vec2f ptA = Vec2f(ptA3f.x, ptA3f.y);
	Vec2f ptB = Vec2f(ptB3f.x, ptB3f.y);
	Vec2f crossPoint;
	if(clickLine(click, ptA, ptB, crossPoint)){
		
		idx = idxA;
		
		doDim(par, .5, dpoint, weight, order, dim, idx, uDim, vDim);

		updateKnot(dim.x, uKnot, order.x);
		updateKnot(dim.y, vKnot, order.y);
		dpointsToPoints();
		
		if(idx >= dpoint.length){
			idx=dpoint.length-1;
		}

		processFields();
		MultiSelect::pointList.length = 0;
		smls->setPointFrom();
		houseKeep();
		return true;
	}
	return false;
}

bool NurbsPatch::insertControlRowColum(){
	M_Vec3f& dpoint = this->dpoint.refValue();
	M_Vec3f& point = this->point.refValue();
	M_Float& weight = this->weight.refValue();
	M_Float& uKnot = this->uKnot.refValue();
	Vec2f& dim = this->dim.refValue();
	Vec2f& order = this->order.refValue();
	int uDim = dim.x;
	int vDim = dim.y;
	int& idx = index[0];
	Matrix matrix = this->matrix * g_viewMatrix;

	for(int v = 0; v < vDim; v++){
		for(int u = 0; u < uDim - 1; u++){
			int idxA = v * uDim + u;
			int idxB = v * uDim + u + 1;
			Vec2f par = Vec2f(uDim + 1, vDim);
			if(insertControlRowColum(par, idxA, idxB, matrix, dpoint, point, weight, uKnot, uDim, vDim, idx, dim, order)){
				return true;
			}
		}
	}
	for(int u = 0; u < uDim; u++){
		for(int v = 0; v < vDim - 1; v++){
			int idxA = v * uDim + u;
			int idxB = (v + 1) * uDim + u;
			Vec2f par = Vec2f(uDim, vDim + 1);
			if(insertControlRowColum(par, idxA, idxB, matrix, dpoint, point, weight, uKnot, uDim, vDim, idx, dim, order)){
				return true;
			}
		}
	}
	return false;
}

bool NurbsPatch::deleteControlRowColum(){
	if(cage == false || showAll == false){
		return false;
	}
	M_Vec3f& dpoint = this->dpoint.refValue();
	M_Vec3f& point = this->point.refValue();
	M_Float& weight = this->weight.refValue();
	M_Float& uKnot = this->uKnot.refValue();
	Vec2f& dim = this->dim.refValue();
	Vec2f& order = this->order.refValue();
	int uDim = dim.x;
	int vDim = dim.y;
	int& idx = index[0];
	Matrix matrix = this->matrix * g_viewMatrix;

	for(int v = 0; v < vDim; v++){
		for(int u = 0; u < uDim - 1; u++){
			int idxA = v * uDim + u;
			int idxB = v * uDim + u + 1;
			Vec2f par = Vec2f(uDim, vDim - 1);
			if(deleteControlRowColum(par, idxA, idxB, matrix, dpoint, point, weight, uKnot, uDim, vDim, idx, dim, order)){
				return true;
			}
		}
	}
	for(int u = 0; u < uDim; u++){
		for(int v = 0; v < vDim - 1; v++){
			int idxA = v * uDim + u;
			int idxB = (v + 1) * uDim + u;
			Vec2f par = Vec2f(uDim - 1, vDim);
			if(deleteControlRowColum(par, idxA, idxB, matrix, dpoint, point, weight, uKnot, uDim, vDim, idx, dim, order)){
				return true;
			}
		}
	}
	return false;
}

void Seamless::deleteControlRowColum(){
	for(int k = 0; k < cpNodeList.length; k++){
		if(cpNodeList[k]->deleteControlRowColum()){
			updateBuild3dWindow();
			return;
		}
	}
}

void Seamless::insertControlRowColum(){
	for(int k = 0; k < cpNodeList.length; k++){
		if(cpNodeList[k]->insertControlRowColum()){
			return;
		}
	}
}

void SenseTugGlobals::initTug(){
//	if(!(0x8000&GetKeyState(VK_LSHIFT))){
//			int g = 0;
//	}

	node = 0;
	g_viewMatrix = g_scene->getViewMatrix();
	const bool perspective = g_settings->perspective;

	g_transformCP->dragPivotPointMode = false;
	g_transformCP->dragRotPointMode = false;
	g_transformCP->stretchHappened = false;
	if(NurbsPatch* selPatch = dynamic_cast<NurbsPatch*>(g_selNode.getRaw())){
		if(selPatch->smls){
			Seamless* smls = selPatch->smls;
			if(MultiSelect::pointList.length > 3 && g_transformCP->sameScreenPosition == false){
				Matrix pivotPointMatrix;
				g_transformCP->getPivotPointMatrix(pivotPointMatrix, smls);

				if(checkAnimCtrlPoint(perspective, g_transformCP->rotCtrlPnt)){
					g_transformCP->dragRotPointMode = true;
					g_transformCP->refreshRemPoints();

					//if((0x8000&GetKeyState(VK_LSHIFT) || InputEvent::l_shiftDown)){
					//	if((0x8000&GetKeyState(VK_LCONTROL) || InputEvent::l_ctrlDown)){
					//		if((0x8000&GetKeyState(VK_LMENU) || InputEvent::l_altDown)){
					//		}
					//		else{
					//			g_transformCP->refreshRemPoints();
					//		}
					//	}
					//	else{
					//		g_transformCP->refreshRemPoints();
					//	}
					//}
					//else if((0x8000&GetKeyState(VK_LMENU) || InputEvent::l_altDown)){
					//	if((0x8000&GetKeyState(VK_LCONTROL) || InputEvent::l_ctrlDown)){
					//		g_transformCP->refreshRemPoints();
					//	}
					//	else{
					//		g_transformCP->refreshRemPoints();
					//	}
					//}


	//				g_transformCP->remRotCtrlPnt = g_transformCP->rotCtrlPnt;
					//g_transformCP->remRotation = g_transformCP->rotation;
					return;
				}
				else if(checkAnimCtrlPoint(perspective, pivotPointMatrix * Vec3f())){
					g_transformCP->dragPivotPointMode = true;
					return;
				}
			}
			else if((((0x8000&GetKeyState(VK_RMENU)) || InputEvent::r_altDown) && (0x8000&GetKeyState(VK_RCONTROL)) || InputEvent::r_ctrlDown)){
				selPatch->splitPatch();
				return;
			}
			if((0x8000&GetKeyState(VK_DELETE))){
				smls->deleteControlRowColum();
				return;
			}
			if(((0x8000&GetKeyState(VK_RCONTROL)) || InputEvent::r_altDown) && false == ((0x8000&GetKeyState(VK_RSHIFT) ||  InputEvent::r_shiftDown))){
				smls->insertControlRowColum();
				return;
			}

			if((false == (0x8000&GetKeyState(VK_LMENU)) && ((0x8000&GetKeyState(VK_LSHIFT)) || (0x8000&GetKeyState(VK_RSHIFT)))) ||
				(false == InputEvent::l_altDown && (InputEvent::l_shiftDown || InputEvent::r_shiftDown))
			
				){
				if(false == (0x8000&GetKeyState(VK_LCONTROL)) && false == InputEvent::l_ctrlDown && false == (0x8000&GetKeyState(VK_RCONTROL)) && false == InputEvent::r_ctrlDown){
					MultiSelect::reset();
				}
				showCursor(FALSE);
				MultiSelect::selectRectMode = true;

				MultiSelect::bottomRight=MultiSelect::topLeft=g_pointer.translation;
				g_vcSelectRectGroup->children.refValue().length=0;
				g_vcSelectRectGroup->children.refValue()+=g_vcSelectRectShape.p;
				g_visualClue=g_vcSelectRectGroup;
				selPatch->houseKeep();
				return;
			}

			if(g_transformCP->sameScreenPosition){
				MultiSelect::reset();
				//showCursor(FALSE);
				//MultiSelect::selectRectMode = true;

				MultiSelect::bottomRight=MultiSelect::topLeft = g_pointer.translation + Vec3f(-.01, .01, 0);
				MultiSelect::bottomRight=MultiSelect::bottomRight = g_pointer.translation + Vec3f(.01, -.01, 0);
				smls->selectControlPoints();

				//g_vcSelectRectGroup->children.refValue().length=0;
				//g_vcSelectRectGroup->children.refValue()+=g_vcSelectRectShape.p;
				//g_visualClue=g_vcSelectRectGroup;
				if(MultiSelect::pointList.length){
					MultiSelect::pointList[0].p->index[0] = MultiSelect::pointList[0].i;
					NurbsPatch* patch = MultiSelect::pointList[0].p;
					g_pointer.translation = g_scene->getViewMatrix() * patch->dpoint[MultiSelect::pointList[0].i];

					if(patch != selPatch){
						g_stw->setResultsForSelectingPoint(patch);
					}
					tugFlag=true;
					MultiSelect::clickedIndex = 0;
					showCursor(FALSE);
					g_transformCP->refreshRemPoints();

					selPatch->houseKeep();
					fastFullBuildForAutoIn(false);
				}
				return;
			}

			selPatch->initMultiCageTug(perspective, smls->cpNodeList);

			if(MultiSelect::pointList.length && g_settings->tugRadius.refValue() > 0){
				selPatch = MultiSelect::pointList[MultiSelect::clickedIndex].p;
				int idx = MultiSelect::pointList[MultiSelect::clickedIndex].i;
				MultiSelect::pointList = ControlPoint(selPatch, idx);
				MultiSelect::beginDragPos = selPatch->dpoint[idx];
				MultiSelect::clickedIndex = 0;
				selPatch->index[0] = idx;
				if(checkRadialTugSizeNurbs(perspective, selPatch)){
					return;
				}
				M_<NurbsPatch*>& cpNodeList = smls->cpNodeList;
				g_backUpCoord.length = 0;
				for(int k = 0; k < cpNodeList.length; k++){
					g_backUpCoord += cpNodeList[k]->dpoint;
				}
			}
	/*
			selPatch->initTug(perspective);
			*/
	//		if(selPatch->ringTug){
	//			MultiSelect::beginDragPos = selPatch->dpoint;
	//		}
		

			if(tugFlag){
				selPatch->houseKeep();
			}

			fastFullBuildForAutoIn(false);
			return;
		}
	}
	if(Subdivision* subdivision=dynamic_cast<Subdivision*>(g_selNode.getRaw())){
		int& idx = subdivision->index[0];
		Matrix matrix = g_viewMatrix;
		int j=subdivision->index[0];
		M_Vec3f& point = subdivision->point;
		for(int i=point.length; --i>=0;){
			if (checkPoint(perspective, matrix * point[i], idx, i)){
				g_selNode->processFields();
				subdivision->makePointerEqualIndexLocation();
				subdivision->houseKeep();
				g_stw->cp->paint();
				fastFullBuildForAutoIn(false);
				return;
			}
		}
		return;
	}
	else if(EC2* ec2=dynamic_cast<EC2*>(g_selNode.getRaw())){
		Matrix matrix = rotationMatrix(ec2->orientation) * translationMatrix(ec2->location) * g_viewMatrix;
		int j=ec2->index[0];
		if(checkOrientationPoint(perspective, ec2->point[j], ec2->tangent[j], g_pointer.translation, tugFlag, matrix)){
			g_selNode->processFields();
			g_stw->cp->paint();
			if(g_controlPanelOpen==true){
//					g_stw->setSelNodeLine();
			}
			return;
		}
		for(int i=ec2->point.length; --i>=0;){
			if(checkPoint(perspective, ec2->point[i], ec2->index[0], i, matrix)){
				g_selNode->processFields();
				g_stw->cp->paint();
				return;
			}
		}
	}
	else if(EC3* ec3=dynamic_cast<EC3*>(g_selNode.getRaw())){
		Matrix matrix = rotationMatrix(ec3->orientation) * translationMatrix(ec3->location) * g_viewMatrix;
		int j=ec3->index[0];
		if (checkOrientationPoint(perspective, Vec2f(ec3->point[j].x, ec3->point[j].y), ec3->tangent[j], g_pointer.translation, tugFlag, matrix)){
			g_selNode->processFields();
			g_stw->cp->paint();
			if(g_controlPanelOpen==true){
//					g_stw->setSelNodeLine();
			}
			return;
		}
		for(int i=ec3->point.length; --i>=0;){
			if (checkPoint(perspective, *ec3, i, matrix)){
				g_selNode->processFields();
				g_stw->cp->paint();
				if(g_controlPanelOpen==true){
//						g_stw->setSelNodeLine();
				}
				return;
			}
		}
	}
	else if(BezierTape* bezier=dynamic_cast<BezierTape*>(g_selNode.getRaw())){
		Matrix matrix = rotationMatrix(Rotation(0,-1,0,PI*.5)) * g_viewMatrix;
		int j=bezier->index[0];
		for(int i=bezier->point.length; --i>=0;){
			if (checkPoint(perspective, bezier->point.refValue()[i], bezier->index[0], i, matrix)){
				g_selNode->processFields();
				g_stw->cp->paint();
				return;
			}
		}
		return;
	}
	else if(Bezier3f* bezier=dynamic_cast<Bezier3f*>(g_selNode.getRaw())){
		Matrix matrix = bezier->matrix * g_viewMatrix;
		int j=bezier->index[0];
		for(int i=bezier->point.length; --i>=0;){
			if (checkPoint(perspective, matrix * bezier->point[i], bezier->index[0], i)){
				g_selNode->processFields();
				g_stw->cp->paint();
				return;
			}
		}
		return;
	}
	else if(Bezier* bezier=dynamic_cast<Bezier*>(g_selNode.getRaw())){
		Matrix matrix = bezier->matrix * g_viewMatrix;
		int j=bezier->index[0];
		for(int i=bezier->point.length; --i>=0;){
			if (checkPoint(perspective, bezier->point[i], bezier->index[0], i, matrix)){
				g_selNode->processFields();
				g_stw->cp->paint();
				return;
			}
		}
		return;
	}
	else if(BezierPatch* bezierPatch=dynamic_cast<BezierPatch*>(g_selNode.getRaw())){
		if(MultiSelect::selectRectMode){
			MultiSelect::bottomRight=MultiSelect::topLeft=g_pointer.translation;
			//Patch::bottomRight+=Vec3f(.2, -.2, .0);
			g_vcSelectRectGroup->children.refValue().length=0;
			g_vcSelectRectGroup->children.refValue()+=g_vcSelectRectShape.p;
			g_vcSelectRectGroup->children.refValue()+=g_vcBezierBase.p;
			g_visualClue=g_vcSelectRectGroup;
			bezierPatch->houseKeep();
		}
		else{
			Matrix matrix = g_viewMatrix;
			int j=bezierPatch->index[0];
			int len=bezierPatch->point.length;
			const M_Vec3f& point = bezierPatch->point;
			for(int i=0; i<len; i++){
				if (checkPoint(perspective, matrix * point[i], bezierPatch->index[0], i)){
					bezierPatch->setNumOfGroupFromSelPoint();
					bezierPatch->houseKeep();
					g_selNode->processFields();
					g_stw->cp->paint();
					return;
				}
			}
		}
		return;
	}
	else if(RingBuild* ringBuild=dynamic_cast<RingBuild*>(g_selNode.getRaw())){
		Matrix matrix = scaleMatrix(ringBuild->scale.refValue()) * rotationMatrix(ringBuild->rotation.refValue()) * translationMatrix(ringBuild->translation.refValue()) * g_viewMatrix;
		int j=ringBuild->index[0];
		if (ringBuild->checkPoint(perspective, ringBuild->index[0], j, matrix)){
			g_selNode->processFields();
			g_stw->cp->paint();
			return;
		}
		for(int i=ringBuild->scaleX.length; --i>=0;){
			if (ringBuild->checkPoint(perspective, ringBuild->index[0], i, matrix)){
				g_selNode->processFields();
				g_stw->cp->paint();
				return;
			}
		}
		return;
	}
	else if(SineTug* sineTug=dynamic_cast<SineTug*>(g_selNode.getRaw())){
		const M_Vec3f& location = sineTug->location;
		const M_Rotation& orientation = sineTug->orientation;
		const M_Float& distance = sineTug->distance;
		int j=sineTug->index[0];
		//if (sineTug->checkPoint(perspective, ringBuild->index[0], j, matrix)){
		//	g_selNode->processFields();
		//	g_stw->cp->paint();
		//	return;
		//}
		for(int i=sineTug->radius.length; --i>=0;){
			Matrix matrix = rotationMatrix(orientation[i]) * translationMatrix(location[i]) * g_viewMatrix;
			if(checkPoint(perspective, matrix * Vec3f(0, distance[i]*8, 0), sineTug->index[0], i)){
				sineTug->houseKeep();
				g_selNode->processFields();
				g_stw->cp->paint();
				fastFullBuildForAutoIn(false);
			}
		}
		return;
	}
	else if(NurbsLathe* nurbsLathe = dynamic_cast<NurbsLathe*>(g_selNode.getRaw())){
		if((0x8000&GetKeyState(VK_RCONTROL)) || InputEvent::r_ctrlDown){
			nurbsLathe->insertControlPoint();
			return;
		}
		int& idx = nurbsLathe->index[0];
		Matrix matrix = nurbsLathe->matrix * g_viewMatrix;
		int j = nurbsLathe->index[0];
		M_Vec2f& dpoint = nurbsLathe->dpoint;
		for(int i = dpoint.length; --i >= 0;){
			if(checkPoint(perspective, dpoint[i], idx, i, matrix)){
				g_selNode->processFields();
				g_stw->cp->paint();
				fastFullBuildForAutoIn(false);
				return;
			}
		}
		return;
	}
	else if(BendyLathe* bendyLathe=dynamic_cast<BendyLathe*>(g_selNode.getRaw())){
		Matrix matrix = bendyLathe->matrix * g_viewMatrix;
		if(bendyLathe->showRingControlPoints){
			for(int i = bendyLathe->radius.length; --i>=0;){
				if(bendyLathe->checkPoint(perspective, i, matrix)){
					bendyLathe->index[2]=i;
					g_selNode->processFields();
					bendyLathe->houseKeep();
					g_stw->cp->paint();
					fastFullBuildForAutoIn(false);
					return;
				}
			}
		}
		else{
			int& idx = bendyLathe->index[0];
			M_Vec3f& dpoint = bendyLathe->dpoint;
			for(int i=dpoint.length; --i>=0;){
				if(checkPoint(perspective, matrix * dpoint[i], idx, i)){
					g_selNode->processFields();
					g_stw->cp->paint();
					fastFullBuildForAutoIn(false);
					return;
				}
			}
		}
		return;
	}
	else if(NurbsTrack* nurbsTrack = dynamic_cast<NurbsTrack*>(g_selNode.getRaw())){
		if((0x8000&GetKeyState(VK_RCONTROL)) || InputEvent::r_ctrlDown){
			nurbsTrack->insertControlPoint();
			return;
		}
		Matrix matrix = g_viewMatrix;
		int& idx = nurbsTrack->index[1];
		M_Vec3f& keyValue = nurbsTrack->keyValue.refValue();
		for(int i = keyValue.length; --i>=0;){
			if(checkPoint(perspective, matrix * keyValue[i], idx, i)){
				nurbsTrack->processFields();
				g_stw->cp->paint();
				fastFullBuildForAutoIn(false);
				return;
			}
		}
		return;
	}
	else if(Oscillator* oscillator=dynamic_cast<Oscillator*>(g_selNode.getRaw())){
		int& idx = oscillator->index[0];
		Matrix matrix = g_viewMatrix;
		int j=oscillator->index[0];
		M_Vec2f& fPoint = oscillator->note;
		M_Float& fWeight = oscillator->fWeight;
		for(int i=fPoint.length; --i>=0;){
			float w = 1;
			if(fWeight.length){
				w = fWeight[i];
			}
			if (checkPoint(perspective, Vec2f(fPoint[i].x/w, fPoint[i].y/w*.1), idx, i, matrix)){
				g_selNode->processFields();
				g_stw->cp->paint();
				return;
			}
		}
		return;
	}
	else if(Envelope* envelope=dynamic_cast<Envelope*>(g_selNode.getRaw())){
		int& idx = envelope->index[0];
		Matrix matrix = g_viewMatrix;
		int j=envelope->index[0];
		M_Vec2f& point = envelope->point;
		M_Float& weight = envelope->weight;
		for(int i = point.length; --i>=0;){
			float w = 1;
			if(weight.length){
				w = weight[i];
			}
			if (checkPoint(perspective, Vec2f(point[i].x/w, point[i].y/w*.1), idx, i, matrix)){
				g_selNode->processFields();
				g_stw->cp->paint();
				return;
			}
		}
		return;
	}
	else if(AmplitudeModulator* modulator=dynamic_cast<AmplitudeModulator*>(g_selNode.getRaw())){
		int& idx = modulator->index[0];
		Matrix matrix = g_viewMatrix;
		int j=modulator->index[0];
		M_Vec2f& point = modulator->wet;
		M_Float& weight = modulator->weight;
		for(int i = point.length; --i>=0;){
			float w = 1;
			if(weight.length){
				w = weight[i];
			}
			if (checkPoint(perspective, Vec2f(point[i].x/w, point[i].y/w*.1), idx, i, matrix)){
				g_selNode->processFields();
				g_stw->cp->paint();
				return;
			}
		}
		return;
	}
	else if(FrequencyModulator* modulator=dynamic_cast<FrequencyModulator*>(g_selNode.getRaw())){
		int& idx = modulator->index[0];
		Matrix matrix = g_viewMatrix;
		int j=modulator->index[0];
		M_Vec2f& point = modulator->wet;
		M_Float& weight = modulator->weight;
		for(int i = point.length; --i>=0;){
			float w = 1;
			if(weight.length){
				w = weight[i];
			}
			if (checkPoint(perspective, Vec2f(point[i].x/w, point[i].y/w*.1), idx, i, matrix)){
				g_selNode->processFields();
				g_stw->cp->paint();
				return;
			}
		}
		return;
	}
	else if(Chords* chords=dynamic_cast<Chords*>(g_selNode.getRaw())){
		int& idx = chords->index[0];
		Matrix matrix = g_viewMatrix;
		int j=chords->index[0];
		M_Vec2f& point = chords->spectrum;
		M_Float& weight = chords->weight;
		for(int i = point.length; --i>=0;){
			float w = 1;
			if(weight.length){
				w = weight[i];
			}
			if (checkPoint(perspective, Vec2f(point[i].x/w, point[i].y/w*.1), idx, i, matrix)){
				g_selNode->processFields();
				g_stw->cp->paint();
				return;
			}
		}
		return;
	}
	else if(HarmonicPatch* harmonicPatch=dynamic_cast<HarmonicPatch*>(g_selNode.getRaw())){
		int& idx = harmonicPatch->index[0];
		Matrix matrix = g_viewMatrix;
		int j=harmonicPatch->index[0];
		M_Vec3f& point = harmonicPatch->point;
		M_Float& weight = harmonicPatch->weight;
		for(int i=point.length; --i>=0;){
			if (checkPoint(perspective, matrix * point[i]/weight[i], idx, i)){
				g_selNode->processFields();
				g_stw->cp->paint();
				return;
			}
		}
		return;
	}
	else if(ColorSweep* colorSweep=dynamic_cast<ColorSweep*>(g_selNode.getRaw())){
		int& idx = colorSweep->index[0];
		Matrix matrix = colorSweep->matrix * g_viewMatrix;
		int j=colorSweep->index[0];
		M_Float& distance = colorSweep->distance;
		float posY = 0;

		for(int i = 0; ;i++){
			if(checkPoint(perspective, Vec2f(0, posY), idx, i, matrix)){
				g_selNode->processFields();
				g_stw->cp->paint();
				return;
			}
			if(i == distance.length){
				break;
			}
			posY+=distance[i];
		}
		fastFullBuildForAutoIn(false);
		return;
	}
	else if(CCLathe* ccLathe=dynamic_cast<CCLathe*>(g_selNode.getRaw())){
		fastFullBuildForAutoIn(false);
		Matrix matrix = g_viewMatrix;
		int idx = ccLathe->index[0];

		ccLathe->angleDrag = false;

		if ( idx < ccLathe->seg.length )
		{
			// check  current first
			if (checkPoint(perspective, ccLathe->seg[idx].Pf, ccLathe->index[0], idx, matrix))
			{
				g_selNode->processFields();
				g_stw->cp->paint();
				return;
			}

			// check angle control point
			if (checkPoint(perspective, ccLathe->anglePt, ccLathe->angleDrag, 1, matrix))
			{
				g_selNode->processFields();
				g_stw->cp->paint();
				return;
			}

		}
		for(int i=ccLathe->seg.length; --i>=0;){
			if (checkPoint(perspective, ccLathe->seg[i].Pf, ccLathe->index[0], i, matrix)){
				g_selNode->processFields();
				g_stw->cp->paint();
				return;
			}
		}
		// add point
		if ( 0x8000 & GetKeyState(VK_LCONTROL) ) {
			ccLathe->addControlPoint(g_pointer.translation);
		}

		return;
	}
	if(Art* part = dynamic_cast<Art*>(g_selNode.getRaw())){
		Seamless* smls = part->smls;
		if((false == (0x8000&GetKeyState(VK_LMENU)) && ((0x8000&GetKeyState(VK_LSHIFT)) || (0x8000&GetKeyState(VK_RSHIFT)))) ||
			(false == InputEvent::l_altDown && (InputEvent::l_shiftDown || InputEvent::r_shiftDown))
		
			){
			MultiSelect::reset();
			showCursor(FALSE);
			MultiSelect::selectRectMode = true;

			MultiSelect::bottomRight=MultiSelect::topLeft=g_pointer.translation;
			//Patch::bottomRight+=Vec3f(.2, -.2, .0);
			g_vcSelectRectGroup->children.refValue().length=0;
			g_vcSelectRectGroup->children.refValue()+=g_vcSelectRectShape.p;
			//g_vcSelectRectGroup->children.refValue()+=g_vcBezierBase.p;
			g_visualClue=g_vcSelectRectGroup;
			part->houseKeep();
			return;
		}
		//else if(smls->workAnimCP && (smls->xRay == false || g_scene->xRayEdit == false)){
		else if(dynamic_cast<AnimClue*>(g_visualClue)){
			g_partCageClue->cpTug = false;
			g_vcArtAnimBase->dragTranPointMode = false;
			if(false == (part->ik && part->pin == false) && checkAnimCtrlPoint(perspective, g_vcArtAnimBase->rotCtrlPnt)){
			}
			else if(checkAllPivotPoints(perspective, part->smls->stickMan)){
				part = dynamic_cast<Art*>(SenseTugGlobals::node);
				g_stw->setResultsForSelectingVertex(part);
				SenseTugGlobals::node = g_vcArtAnimBase;
//				else if(checkAnimCtrlPoint(perspective, part->pivotPointMatrix * Vec3f(0,0,0))){
				g_vcArtAnimBase->dragTranPointMode = true;
				g_vcArtAnimBase->remTranValue = part->translation;
				smls->ActTri=0;
				smls->ActTriCorner=0;
				smls->actTriPart=0;
				return;
			}
			else if(smls->workAnimCP){
				part->initTugCP(perspective);
				if(tugFlag){
					part->houseKeep();
					fastFullBuildForAutoIn(false);
					g_partCageClue->cpTug = true;
					return;
				}
			}
			//else if(Window::animMode() == 0 || g_allowPolyMorphCapture){
			goto here;
		}
		else if(part->smls->xRay && g_scene->xRayEdit && ((0x8000&GetKeyState(VK_RCONTROL)) || InputEvent::r_ctrlDown)){
			insertBone(part);
			return;
		}
		else if(part->smls->xRay && checkAllPivotPoints(perspective, part->smls->stickMan)){
			if(node != g_selNode.getRaw()){
				if(Art* part = dynamic_cast<Art*>(node)){
					g_stw->setResultsForSelectingVertex(part);
					smls->ActTri=0;
					smls->ActTriCorner=0;
					smls->actTriPart=0;
				}
			}
			return;
		}
		else if(g_settings->tugRadius > 0 && Window::animMode() == 0){
			if(ImmediatePart* part=dynamic_cast<ImmediatePart*>(g_selNode.getRaw())){
				if(checkRadialTugSize(perspective, part)){
					return;
				}
			}
		}
	}
here:
	if(Window::animMode()){
		MultiSelect::clickedIndex = 0;
	}
	g_scene->senseTug(Matrix(),Matrix(), 0);

	if(Art* part = dynamic_cast<Art*>(node)){
		Seamless* smls = part->smls;
		if(g_dragMorphFlag == false && (Window::animMode() || g_scene->xRayEdit) && smls->xRay){
			if(part != g_selNode.getRaw()){
				g_stw->setResultsForSelectingVertex(part);
				smls->ActTri=0;
				smls->ActTriCorner=0;
				smls->actTriPart=0;
			}
			return;
		}

		if(g_settings->tugRadius.refValue()>0){
			backUpSmlsCoord(smls->partList);
		}
		smls->ActCtrlPnt=ActCtrlPnt;
		tugFlag=true;
		if(!(0x8000&GetKeyState(VK_LCONTROL))){
			showCursor(FALSE);
			//coord[i] = getIViewMatrix() * triPart->imatrix * pt;
		}
		g_pointer.translation = pointer;
		smls->ActTri=ActTri;
		smls->ActTriCorner=corner;
		smls->actTriPart=actTriPart;
		part->index[0]=ActCtrlPnt;
		part->workEffect->setColorIndex(actTriPart->triangle.refTexIndex(ActTri*3+corner));
		if(part != g_selNode.getRaw()){
			g_stw->setResultsForSelectingVertex(part);
		}
		g_stw->cp->paint();
		g_tb->updateToolBarContext();
		fastFullBuildForAutoIn(true);


		if(MultiSelect::coordList.length != MultiSelect::beginDragPos.length){
			MultiSelect::checkToReset(part);
			MultiSelect::backUpCoords(part);
			MultiSelect::calcPivotPoint();
		}

	//	g_vcArtTriBase->circle=false;
	}
}
void ModelMorpher::senseTug(Matrix matrix, Matrix imatrix, TouchSensor* sensor){
	M_Node& sequence = this->sequence.refValue();
	for(int k = sequence.length; --k>=0;){
		ModelSequence* modelSequence = dynamic_cast<ModelSequence*>(sequence[k].p);
		M_Node& model = modelSequence->model.refValue();
		for(int k = model.length; --k>=0;){
			Seamless* smls = dynamic_cast<Seamless*>(model[k].p);
			smls->senseTug(matrix, imatrix, sensor);
		}
	}
}

bool checkAllPivotPoints(bool perspective, StickMan* stickMan){
	if(stickMan == 0) return false;
	M_Art& partList = stickMan->partList;

	if(Art* part = dynamic_cast<Art*>(g_selNode.getRaw())){
		if(part->ik){
			if(checkAnimCtrlPoint(perspective, g_scene->getViewMatrix() * part->ikPos)){
				SenseTugGlobals::node = part;
				return true;
			}
		}
		else{
			if(checkAnimCtrlPoint(perspective, part->pivotPointMatrix * Vec3f(0,0,0))){
				SenseTugGlobals::node = part;
				return true;
			}
		}
	}

	for(int k = partList.length; --k >= 0;){
		Art* part = partList[k];
		if(part->ik){
			if(checkAnimCtrlPoint(perspective, g_scene->getViewMatrix() * part->ikPos)){
				SenseTugGlobals::node = part;
				return true;
			}
		}
		else{
			if(checkAnimCtrlPoint(perspective, part->pivotPointMatrix * Vec3f(0,0,0))){
				SenseTugGlobals::node = part;
				return true;
			}
		}
	}
	return false;
}

bool isChildOf(Art* partA, Art* partB, int& k){

	M_SArt& m_part = partA->part.refValue();

	for(k = m_part.length; --k >=0;){
		Art* part = m_part[k].p;
		if(part == partB){
			return true;
		}
	}
	return false;
}

Art* createPartOfType(Art* part){
	if(dynamic_cast<Part*>(part)){
		return new Part;
	}
	return new ImmediatePart;
}

void mirrorInsertPart(ArtPtr mirrorPartA, ArtPtr mirrorPartB, int mirrorK, Vec3f pivotPoint){

	M_SArt& m_part = mirrorPartA->part.refValue();

	Art* mirrorNewPart = createPartOfType(mirrorPartA);
	m_part[mirrorK] = mirrorNewPart;
	mirrorNewPart->part.refValue() += mirrorPartB;

	Vec3f mirrorPivotPoint = pivotPoint;
	mirrorPivotPoint.x = -mirrorPivotPoint.x;

	mirrorNewPart->pivotPoint.refValue() = mirrorPivotPoint;
	mirrorPartB->pivotPoint.refValue() -= mirrorNewPart->pivotPoint;
}

void addMirrorPart(Art* part, Vec3f newPivotPoint){
	Art* newMirrorPart = createPartOfType(part);
	Vec3f newMirrorPivotPoint = newPivotPoint;
	newMirrorPivotPoint.x = -newMirrorPivotPoint.x;
	newMirrorPart->pivotPoint.refValue() = newMirrorPivotPoint;
	part->part.refValue() += newMirrorPart;
}

void insertBone(Art* part){
	if(StickMan* stickMan = part->smls->stickMan){
		M_Art& partList = stickMan->partList;
		for(int k = partList.length; --k >= 0;){
			Art* partA = partList[k];
			M_SArt& m_part = partA->part.refValue();
			for(int k = m_part.length; --k >= 0; ){
				ArtPtr partB = m_part[k];

				//Vec2f click3f = g_pointer.translation;
				Vec3f ptA3f = partA->pivotPointMatrix * Vec3f(0,0,0);
				Vec3f ptB3f = partB->pivotPointMatrix * Vec3f(0,0,0);

				Vec2f click = Vec2f(g_pointer.translation.x, g_pointer.translation.y);
				Vec2f ptA = Vec2f(ptA3f.x, ptA3f.y);
				Vec2f ptB = Vec2f(ptB3f.x, ptB3f.y);
				Vec2f crossPoint;
				if(clickLine(click, ptA, ptB, crossPoint)){
					g_stw->textCell.cellActive = false;
					Art* newPart = createPartOfType(partA);
					m_part[k] = newPart;
					newPart->part.refValue() += partB;
					
					float ratio = getMidPointRatio(ptA, ptB, crossPoint);
					Vec3f pivotPoint = (partB->absolutePivotPoint - partA->absolutePivotPoint) * ratio;

					newPart->pivotPoint.refValue() = pivotPoint;
					partB->pivotPoint.refValue() -= newPart->pivotPoint;

					int mirrorK = 0;
					if(partA->mirrorPart && partB->mirrorPart && isChildOf(partA->mirrorPart, partB->mirrorPart, mirrorK)){
						mirrorInsertPart(partA->mirrorPart, partB->mirrorPart, mirrorK, pivotPoint);
					}
					else if(partB->mirrorPart && isChildOf(partA, partB->mirrorPart, mirrorK)){
						mirrorInsertPart(partA, partB->mirrorPart, mirrorK, pivotPoint);
					}
					part->smls->process();
					g_stw->genNodeList();
					g_stw->selLine = getNodeLine(newPart);
					g_stw->commonPostInsertNodeStuff();
					return;
				}
			}
		}
		g_stw->textCell.cellActive = false;
		Art* newPart = createPartOfType(part);
		//newPart->pivotPoint.refValue() = g_pointer.translation - part->absolutePivotPoint;;
		Vec3f vz = g_scene->getViewMatrix() * part->pivotPointMatrix * Vec3f(0,0,0);
		g_pointer.translation.z = vz.z;
		Vec3f newAbsolutePivotPoint = g_scene->getIViewMatrix() * g_catchIMatrix * part->imatrix * g_pointer.translation;
		Vec3f newPivotPoint = newAbsolutePivotPoint - part->absolutePivotPoint;
		part->part.refValue() += newPart;
		if(part->mirror){
			if(part->part.length == 1 && fabs(part->absolutePivotPoint.x) < .000001){
				if(fabs(newAbsolutePivotPoint.x) < 1 / g_scene->getZoom() * .05){
					newPivotPoint.x = 0;
				}
				else{
					addMirrorPart(part, newPivotPoint);
				}
			}
			else if(Art* mirrorPart = part->mirrorPart){
				addMirrorPart(mirrorPart, newPivotPoint);
			}
			else{
				addMirrorPart(part, newPivotPoint);
			}
		}
		newPart->pivotPoint.refValue() = newPivotPoint;
		
		part->smls->process();
		g_stw->genNodeList();
		g_stw->selLine = getNodeLine(newPart);
		g_stw->commonPostInsertNodeStuff();
		return;
	}
}

void Seamless::senseTug(Matrix matrix, Matrix imatrix, TouchSensor* sensor){
	if(SenseTugGlobals::node) return;;

	const bool perspective=g_settings->perspective;

	g_catchMatrix=matrix;
	g_catchIMatrix=imatrix;
	catMatrix(matrix);
	catIMatrix(imatrix);

//	if(dynamic_cast<Build*>(g_selNode.getRaw()) || dynamic_cast<Generator*>(g_selNode.getRaw())){
//		if(g_stw->selLine > 0 && this==g_sceneTreeList[getSeamlessParent(g_stw->selLine)].node){
//		}
//	}
	if(sensor && sensor->enabled){
		g_stw->senseAllPartsTriangle(this, sensor);
	}
	if(SenseTugGlobals::node == 0){
		if(Window::animMode()){
			if(checkAllPivotPoints(perspective, stickMan)){
				Art* part = dynamic_cast<Art*>(SenseTugGlobals::node);
				g_stw->setResultsForSelectingVertex(part);
				SenseTugGlobals::node = g_vcArtAnimBase;
				g_vcArtAnimBase->dragTranPointMode = true;
				g_vcArtAnimBase->remTranValue = part->translation;
				return;
			}
		}
		if(0x8000&GetKeyState(VK_LMENU) || InputEvent::l_altDown){
			g_stw->senseAllPartsTriangle(this);
		}
		else if(Window::animMode() == 0 || g_allowPolyMorphCapture){
			senseAllPartsVertex();
		}
	}
}

void Transform::senseTug(Matrix matrix, Matrix imatrix, TouchSensor* sensor){
	//if(SenseTugGlobals::sameVertexHit)return;
	if(SenseTugGlobals::node)return;
	Vec3f scale=this->scale;
	Rotation rotation=this->rotation;
	Vec3f translation=this->translation;

	matrix= scaleMatrix(scale)*rotationMatrix(rotation)*translationMatrix(translation)*matrix;
	imatrix=imatrix*translationMatrix(Vec3f(-translation.x, -translation.y, -translation.z)) * rotationMatrix( -rotation) * scaleMatrix(Vec3f(1./scale.x, 1./scale.y, 1./scale.z));

	M_Node& children=this->children.refValue();

	for(int k=0; k<children.length;k++){
		children[k]->senseTug(matrix, imatrix, this->sensor);
	}
}
void AGroup::senseTug(Matrix matrix, Matrix imatrix, TouchSensor* sensor){
	//if(SenseTugGlobals::sameVertexHit)return;
	if(SenseTugGlobals::node)return;
	M_Node& children=this->children.refValue();

	for(int k=0; k<children.length;k++){
		children[k]->senseTug(matrix, imatrix, this->sensor);
	}
}
//void NoOutputGroup::senseTug(Matrix matrix, Matrix imatrix){
//}
void Node::resetRotations(){
	return;
}
void ProtoDeclare::resetRotations(){
	M_Node& children=protoBody->children.refValue();
	for(int k=0; k<children.length;k++){
		children[k]->resetRotations();
	}
}
void Seamless::resetRotations(){
	for(int k=partList.length; --k>=0;){
		partList[k]->rotation=Rotation();
	}
	catMatrix();
	catIMatrix();
	updateAllVertexInBuffers();//needed to stop dark normals but dont see y its needed?
}
void AGroup::resetRotations(){

	M_Node& children=this->children.refValue();

	for(int k=0; k<children.length;k++){
		children[k]->resetRotations();
	}
}


/*
Vec3f S_Triangle::calculateNormal(){

	Vec3f cornerA=pa->matrix * pa->coord[ia];
					
	Vec3f cornerB=pb->matrix * pb->coord[ib];

	Vec3f cornerC=pc->matrix * pc->coord[ic];

    float Qx, Qy, Qz, Px, Py, Pz;

    Qx = cornerC.x-cornerA.x;
    Qy = cornerC.y-cornerA.y;
    Qz = cornerC.z-cornerA.z;
    Px = cornerB.x-cornerA.x;
    Py = cornerB.y-cornerA.y;
    Pz = cornerB.z-cornerA.z;

	return Vec3f(
		Py*Qz - Pz*Qy,
		Pz*Qx - Px*Qz,
		Px*Qy - Py*Qx
		).normalize();

} 

inline Vec3f genSingleVertexNormal(Art* part, int index){
	M_Art& relevantPart=part->relevantPart;
	Vec3f norm;
	int hits=0;
	for(int k=relevantPart.length;  --k>-0; ){
		M_Triangle triangle=relevantPart[k]->triangle;
		for(int i=triangle.length*3; --i>=0; ){
			if(triangle.refIndex(i)==index && triangle.refPart(i)==part){
				S_Triangle tri=triangle[i/3];
				norm+=tri.calculateNormal();
				hits++;
			}
		}
	}
	return (norm/(float)hits).normalize();
}
*/
inline int convertToLinearVertexIndex(Art* target, M_Art& partList){
	int accumIndex=0;
	for(int k=0; k<partList.length; k++){
		Art* part=partList[k];
		if(part==target){
			return accumIndex;
		}
		accumIndex+=part->coord.refValue().length;
	}
	throw target;
}
int sensePartForRadialTug(M_Vec3f& coord, const Vec3f location, const Vec3f displacement, const float radius, const float slope){
	float pp=slope*PI;
	float ipp=PI-pp;
	float h=1/(cos(ipp)+1);
	for (int i=0;i<coord.length;i++){
		Vec3f vec=coord[i];
		float vecX=vec.x-location.x;
		float vecY=vec.y-location.y;
		float vecZ=vec.z-location.z;
		float HYP=(float)(sqrt(vecX*vecX + vecY*vecY + vecZ*vecZ));

		if(radius>HYP){
			return i;
		}
	}
	return -1;
}
void singleRadialTug(M_Vec3f& coord, const Vec3f location, const Vec3f displacement, const float radius, const float slope){
	float pp=slope*PI;
	float ipp=PI-pp;
	float h=1/(cos(ipp)+1);
	for (int i=0;i<coord.length;i++){
		Vec3f vec=coord[i];
		float vecX=vec.x-location.x;
		float vecY=vec.y-location.y;
		float vecZ=vec.z-location.z;
		float HYP=(float)(sqrt(vecX*vecX + vecY*vecY + vecZ*vecZ));

		if(radius>HYP){
			float scaleDecay=(cos((fabs(HYP)/radius)*pp+ipp)+1)*h;
			vec.x+=displacement.x*scaleDecay;
			vec.y+=displacement.y*scaleDecay;
			vec.z+=displacement.z*scaleDecay;
			coord[i]=vec;
		}
	}
}
void singleMirrorRadialTug(M_Vec3f& coord, const Vec3f location, const Vec3f mirrorLocation, const Vec3f displacement, const Vec3f mirrorDisplacement, const float radius, const float slope){
	float pp=slope*PI;
	float ipp=PI-pp;
	float h=1/(cos(ipp)+1);
	for(int i=0;i<coord.length;i++){
		Vec3f vec=coord[i];
		if(vec.x==0){
			float vecX=vec.x-location.x;
			float vecY=vec.y-location.y;
			float vecZ=vec.z-location.z;
			float HYP=(float)(sqrt(vecX*vecX + vecY*vecY + vecZ*vecZ));

			if(radius>HYP){
				float scaleDecay=(cos((fabs(HYP)/radius)*pp+ipp)+1)*h;
				vec.x=0;
				vec.y+=displacement.y*scaleDecay;
				vec.z+=displacement.z*scaleDecay;
				coord[i]=vec;
			}
		}
		else if(vec.x>0){
			float vecX=vec.x-location.x;
			float vecY=vec.y-location.y;
			float vecZ=vec.z-location.z;
			float HYP=(float)(sqrt(vecX*vecX + vecY*vecY + vecZ*vecZ));

			if(radius>HYP){
				float scaleDecay=(cos((fabs(HYP)/radius)*pp+ipp)+1)*h;
				vec.x+=displacement.x*scaleDecay;
				vec.y+=displacement.y*scaleDecay;
				vec.z+=displacement.z*scaleDecay;
				coord[i]=vec;
			}
		}
		else if(vec.x<0){
			float vecX=vec.x-mirrorLocation.x;
			float vecY=vec.y-mirrorLocation.y;
			float vecZ=vec.z-mirrorLocation.z;
			float HYP=(float)(sqrt(vecX*vecX + vecY*vecY + vecZ*vecZ));

			if(radius>HYP){
				float scaleDecay=(cos((fabs(HYP)/radius)*pp+ipp)+1)*h;
				vec.x+=mirrorDisplacement.x*scaleDecay;
				vec.y+=mirrorDisplacement.y*scaleDecay;
				vec.z+=mirrorDisplacement.z*scaleDecay;
				coord[i]=vec;
			}
		}
	}
}
/*
void singleSineTugNurbs(const M_Vec3f& dpoint, M_Vec3f& point, const M_Float& weight, const Vec3f location, const Rotation& orientation, const float height, const float rad){
	
	Matrix matrix=rotationMatrix(orientation)*translationMatrix(location);
	Matrix imatrix=translationMatrix(-location)*rotationMatrix(-orientation);
	const float sdecay=rad;

	float xsc=1;
	
	for(int i=point.length; --i>=0;){
		Vec3f ivec=imatrix*(point[i] / weight[i]);
		if(ivec.y<-sdecay)continue;
		float ADJ=(float)(fabs(ivec.x)*xsc);
		float OPP=(float)(fabs(ivec.z));
		float HYP=(float)(sqrt(ADJ*ADJ + OPP*OPP));

		if(rad>HYP && abs(ivec.y)<sdecay){
			float sineCurve=cos((fabs(ivec.y)/sdecay)*PI)*.5+.5;
			ivec.y+=((1+cos(HYP/rad*PI))/2.*height)*sineCurve;
			point[i]=matrix*(ivec * weight[i]);
		}
	}
}
*/

void NurbsPatch::mirrorSingleCP(int idx){
	if(mirror){
		if(mirrorPatch){
			Vec3f vec = point[idx];
			vec.x = -vec.x;
			mirrorPatch->point.refValue()[idx] = vec;
			vec = dpoint[idx];
			vec.x = -vec.x;
			mirrorPatch->dpoint.refValue()[idx] = vec;
		}
		else{
			int uDim = dim.x;
			int vDim = dim.y;
			int u = idx % uDim;
			int v = idx / uDim;
			int middleIdx = (uDim >> 1);

			if(u == middleIdx){
				point.refValue()[idx].x = 0;
				dpoint.refValue()[idx].x = 0;
			}
			else{
				int midx = v * uDim + uDim - 1 - u;
				Vec3f vec = point[idx];
				vec.x = -vec.x;
				point.refValue()[midx] = vec;
				vec = dpoint[idx];
				vec.x = -vec.x;
				dpoint.refValue()[midx] = vec;
			}
		}
	}
}

void NurbsPatch::singleRadialTugMirrorRight(M_Vec3f& point, const M_Float& weight, const Vec3f location, const Vec3f displacement, const float radius, const float slope){
	float pp=slope*PI;
	float ipp=PI-pp;
	float h=1/(cos(ipp)+1);
	for (int i = 0; i < point.length; i++){
		int idx = fromPoint + i;
		Vec3f vec = g_backUpCoord[idx];
		if(vec.x <= 0){
			float vecX=vec.x-location.x;
			float vecY=vec.y-location.y;
			float vecZ=vec.z-location.z;
			float HYP=(float)(sqrt(vecX*vecX + vecY*vecY + vecZ*vecZ));

			if(radius>HYP){
				float scaleDecay=(cos((fabs(HYP)/radius)*pp+ipp)+1)*h;
				vec+=displacement*scaleDecay;
				//point[i] = vec * weight[i];
				setFilteredDimension(point, i, vec * weight[i]);
				mirrorSingleCP(i);
			}
		}
	}
}

void NurbsPatch::singleRadialTugMirrorLeft(M_Vec3f& point, const M_Float& weight, const Vec3f location, const Vec3f displacement, const float radius, const float slope){
	float pp=slope*PI;
	float ipp=PI-pp;
	float h=1/(cos(ipp)+1);
	for (int i = 0; i < point.length; i++){
		int idx = fromPoint + i;
		Vec3f vec = g_backUpCoord[idx];
		if(vec.x >= 0){
			float vecX=vec.x-location.x;
			float vecY=vec.y-location.y;
			float vecZ=vec.z-location.z;
			float HYP=(float)(sqrt(vecX*vecX + vecY*vecY + vecZ*vecZ));

			if(radius>HYP){
				float scaleDecay=(cos((fabs(HYP)/radius)*pp+ipp)+1)*h;
				vec+=displacement*scaleDecay;
				//point[i] = vec * weight[i];
				setFilteredDimension(point, i, vec * weight[i]);
				mirrorSingleCP(i);
			}
		}
	}
}

void NurbsPatch::singleRadialTug(M_Vec3f& point, const M_Float& weight, const Vec3f location, const Vec3f displacement, const float radius, const float slope){
	float pp=slope*PI;
	float ipp=PI-pp;
	float h=1/(cos(ipp)+1);
	for (int i = 0; i < point.length; i++){
		int idx = fromPoint + i;
		Vec3f vec = g_backUpCoord[idx];
		float vecX=vec.x-location.x;
		float vecY=vec.y-location.y;
		float vecZ=vec.z-location.z;
		float HYP=(float)(sqrt(vecX*vecX + vecY*vecY + vecZ*vecZ));

		if(radius>HYP){
			float scaleDecay=(cos((fabs(HYP)/radius)*pp+ipp)+1)*h;
			vec+=displacement*scaleDecay;
			//point[i] = vec * weight[i];
			setFilteredDimension(point, i, vec * weight[i]);
			mirrorSingleCP(i);
		}
	}
}

void NurbsPatch::radialTug(Vec3f location, Vec3f pt){

	float radius = g_settings->tugRadius.refValue();


	const Vec3f newPt = g_scene->getIViewMatrix() * g_catchIMatrix * pt;
	const Vec3f displacement = Vec3f(newPt.x - location.x, newPt.y - location.y, newPt.z - location.z);

	const float slope = g_settings->tugSlope.refValue();

	M_<NurbsPatch*>& cpNodeList = smls->cpNodeList;

	for(int k = 0; k < cpNodeList.length; k++){
		NurbsPatch* patch = cpNodeList[k];
		if(patch->cage){
			if(patch->mirror){
				if(location.x < 0){
					patch->singleRadialTugMirrorRight(patch->point.refValue(), patch->weight.refValue(), location, displacement, radius, slope);
				}
				else{
					patch->singleRadialTugMirrorLeft(patch->point.refValue(), patch->weight.refValue(), location, displacement, radius, slope);
				}
			}
			else{
				patch->singleRadialTug(patch->point.refValue(), patch->weight.refValue(), location, displacement, radius, slope);
			}
		}
	}
}

int findMirrorVertex(const M_Vec3f& cPtr, Vec3f svec);

void radialTug3(const Vec3f coord, const Vec3f displacement, Vec3f disp, const Vec3f selCoord, const float radius, const float slope, Art* part, Art* prt){
	if(prt->show==false){
		return;
	}
	Vec3f location=part->matrix * (prt->imatrix * coord);

	int idx=sensePartForRadialTug(prt->coord, location, displacement, radius, slope);
	if(idx==-1){
		return;
	}
	Vec3f v = prt->matrix*prt->coord[idx];

	Art* mirrorPart = prt->mirrorPart;

	Vec3f mirrorDisp=displacement;

	Vec3f mirrorLocation=location;
	if(prt->mirror && mirrorPart == 0){
		if(location.x>0){
			mirrorLocation.x=-mirrorLocation.x;
			mirrorDisp.x=-mirrorDisp.x;
		}
		else{
			location.x=-location.x;
			disp.x=-displacement.x;
		}
		singleMirrorRadialTug(prt->coord, location, mirrorLocation, disp, mirrorDisp, radius, slope);
	}
	else if(mirrorPart){
		mirrorLocation.x=-mirrorLocation.x;
		mirrorDisp.x=-mirrorDisp.x;

		if(	(prt->plidx > mirrorPart->plidx && v.x<0 && selCoord.x<0)||
			(prt->plidx < mirrorPart->plidx && v.x>0 && selCoord.x>0)||
			(prt->plidx < mirrorPart->plidx && v.x<0 && selCoord.x<0)||
			(prt->plidx > mirrorPart->plidx && v.x>0 && selCoord.x>0)
		){

			singleRadialTug(prt->coord, location, disp, radius, slope);
			singleRadialTug(mirrorPart->coord, mirrorLocation, mirrorDisp, radius, slope);
		}
	}
	else{
		singleRadialTug(prt->coord, location, disp, radius, slope);
	}
}
void radialTug4(const Vec3f coord, const Vec3f displacement, Vec3f disp, const Vec3f selCoord, const float radius, const float slope, Art* part, Art* prt){
	if(prt->show==false){
		return;
	}
	Vec3f location=part->matrix * (prt->imatrix * coord);

	int idx=sensePartForRadialTug(prt->coord, location, displacement, radius, slope);
	if(idx==-1){
		return;
	}
	Vec3f v = prt->matrix*prt->coord[idx];

	Art* mirrorPart = prt->mirrorPart;


	Vec3f mirrorDisp=displacement;

	Vec3f mirrorLocation=location;
	if(prt->mirror && mirrorPart){
		if(location.x>0){
			mirrorLocation.x=-mirrorLocation.x;
			mirrorDisp.x=-mirrorDisp.x;
		}
		else{
			location.x=-location.x;
			disp.x=-displacement.x;
		}
		singleMirrorRadialTug(prt->coord, location, mirrorLocation, disp, mirrorDisp, radius, slope);
	}
	else if(mirrorPart){
		mirrorLocation.x=-mirrorLocation.x;
		mirrorDisp.x=-mirrorDisp.x;

		singleRadialTug(prt->coord, location, disp, radius, slope);
		singleRadialTug(mirrorPart->coord, mirrorLocation, mirrorDisp, radius, slope);
	}
	else{
		singleRadialTug(prt->coord, location, disp, radius, slope);
	}
}

void radialTug(Vec3f coord, Vec3f pt){

	float radius=g_settings->tugRadius.refValue();
	Art* part=dynamic_cast<Art*>(g_selNode.getRaw());
	Seamless* smls=part->smls;
	M_Art& partList=smls->partList;

	const Vec3f newPt=g_scene->getIViewMatrix() * g_catchIMatrix * part->imatrix * pt;
	const Vec3f displacement=Vec3f(newPt.x-coord.x,newPt.y-coord.y,newPt.z-coord.z);
	Vec3f disp=displacement;

	const Vec3f selCoord = part->matrix*coord;

	const float slope=g_settings->tugSlope.refValue();
	if(g_settings->multiPartTug.refValue()){
		for(int k=partList.length; --k>=0;){
			radialTug3(coord, displacement, disp, selCoord, radius, slope, part, partList[k]);
		}
	}
	else{
		radialTug4(coord, displacement, disp, selCoord, radius, slope, part, part);
	}
}
void radialTug2(Vec3f coord, Vec3f pt){
	const float radius=g_settings->tugRadius.refValue();
	Art* part=dynamic_cast<Art*>(g_selNode.getRaw());
	Seamless* smls=part->smls;
	M_Art& partList=smls->partList;

	Vec3f mtxPt=g_scene->getIViewMatrix() * part->imatrix * pt;
	Vec3f displacement=Vec3f(mtxPt.x-coord.x,mtxPt.y-coord.y,mtxPt.z-coord.z);

	const float slope=g_settings->tugSlope.refValue();
	if(g_settings->multiPartTug.refValue()){
		for(int k=partList.length; --k>=0;){
			Art* prt=partList[k];
			g_radialTugStem->part.refValue()[0]=prt;
			Vec3f location=part->matrix * (prt->imatrix * coord);
			//g_radialTugStem->doBuild();
			singleRadialTug(prt->coord, location, displacement, radius, slope);
		}
	}
	else{
		g_radialTugStem->part.refValue()[0]=part;
		g_radialTug->location.refValue()[0]=coord;
		g_radialTugStem->doBuild();
	}
}
bool findAndSetMirrorVertex(M_Vec3f& cPtr, Vec3f svec, Vec3f newPt){
	for(int i = cPtr.length; --i >= 0;){
		if(svec == cPtr[i]){
			cPtr[i] = Vec3f(-newPt.x, newPt.y, newPt.z);
			return true;
		}
	}
	return false;
}
int findMirrorVertex(const M_Vec3f& cPtr, Vec3f svec){
	for(int i = cPtr.length; --i >= 0;){
		if(svec == cPtr[i]){
			return i;
		}
	}
	return -1;
}

Art* Seamless::getMirrorVertex(Art* part, const int vertexIndex, int& returnVertex){
	if(part->mirror == false){
		return 0;
	}
	M_Vec3f& coord=part->coord.refValue();
	Vec3f svec = coord[vertexIndex];
	svec.x=-svec.x;

	if(Art* mirrorPart = part->mirrorPart){
		if((returnVertex = findMirrorVertex(mirrorPart->coord.refValue(), svec))!=-1){
			return mirrorPart;
		}
	}
	else{
		if((returnVertex = findMirrorVertex(coord, svec))!=-1){
			return part;
		}
	}
	return 0;
}
Art* Seamless::getMirrorTriangle(Art* part, int& returnCoordIdxIdx){
	M_Triangle triangle=actTriPart->triangle;

	const int ia=triangle.refIndex(ActTri*3+0);
	const int ib=triangle.refIndex(ActTri*3+1);
	const int ic=triangle.refIndex(ActTri*3+2);

	Art* pa=triangle.refPart(ActTri*3+0);
	Art* pb=triangle.refPart(ActTri*3+1);
	Art* pc=triangle.refPart(ActTri*3+2);

	int mia=0;
	int mib=0;
	int mic=0;

	Art* mpa=getMirrorVertex(pa, ia, mia);
	Art* mpb=getMirrorVertex(pb, ib, mib);
	Art* mpc=getMirrorVertex(pc, ic, mic);

	if(mpa && mpb && mpc){
		for(int k=partList.length; --k>=0;){
			Art* part=partList[k];
			M_Triangle& triangle=part->triangle;
			for(int i=triangle.getTriangleLength(); (i-=3)>=0;){
				if(
					triangle.refPart(i+0)==mpa && triangle.refIndex(i+0)==mia&&
					triangle.refPart(i+2)==mpb && triangle.refIndex(i+2)==mib&&
					triangle.refPart(i+1)==mpc && triangle.refIndex(i+1)==mic
					){
					int mirrorCorner=0;
					if(ActTriCorner==1){
						mirrorCorner=2;
					}
					else if(ActTriCorner==2){
						mirrorCorner=1;
					}
					returnCoordIdxIdx=i+mirrorCorner;
					return part;
				}
			}
		}
	}
	return 0;
}

void Seamless::senseMirrorVertex(ImmediatePart* part, int idx, Vec3f& newPt){
	if(part->mirror == false){
		return;
	}

	Seamless* smls = part->smls;
	M_Vec3f& coord = part->coord.refValue();

	Vec3f svec = coord[idx];
	svec.x=-svec.x;

	if(part->mirrorPart == 0){
		if(part->isLeftRightName() == false){
			if(fabs(svec.x) < .000001){
				newPt.x=0;
				return;
			}
			int i=0;
			if(-1 != (i = findMirrorVertex(coord, svec))){
				coord[i] = Vec3f(-newPt.x, newPt.y, newPt.z);
				return;
			}
		}
	}
	else{
		if(findAndSetMirrorVertex(part->mirrorPart->coord.refValue(), svec, newPt)){
			return;
		}
	}
}

bool getPolyMap(NurbsPatch*& patch, Stem*& stem, Polymap*& polymap, Art* part, int selTriIndex, ImmediatePart*& sourcePart){
	for(int k = g_sceneTreeList.length; --k>=0;){
		if(patch = dynamic_cast<NurbsPatch*>(g_sceneTreeList[k].node.p)){
			if(patch->stem.length){
				M_Node& m_stem = patch->stem.refValue();
				for(int j = m_stem.length; --j>=0;){
					if(stem = dynamic_cast<Stem*>(m_stem[j].p)){
						if(stem->store && selTriIndex >= stem->beginTriOffset){
							if(stem->part.length == 1 && stem->part.refValue()[0] == part){
								if(stem->source.length == 1){
									if(ImmediatePart* bluePart = dynamic_cast<ImmediatePart*>(stem->source.refValue()[0].p)){
										//if(polyMap->coord.length == part->coord.length && polyMap->triangle.length == part->triangle.length){
										sourcePart = bluePart;
										return true;
									}
									return false;
								}
								else{
									return true;
								}
							}
						}
					}
				}
			}
		}
		else if(polymap = dynamic_cast<Polymap*>(g_sceneTreeList[k].node.p)){
			if(selTriIndex >= polymap->beginTriOffset){
				if(polymap->part.refValue() == part){
					return true;
				}
			}
		}
	}
	return false;
}

Vec2f curvePoint2d(int dim, int degree, const float *knot, const Vec2f *pnt, const float *w, const float t);
void CurvePoint (int dim, int degree, const float *knot, const Vec3f *pnt, const float *w, const float t, Vec3f& outPoint);
inline float getDiff(Vec2f t, Vec3f& curvedEdgePoint, float maxU, float maxV, float minU, float minV, int uOrder, int vOrder, int uDegree, int vDegree, int uDim, int vDim, float* uKnotPtr, float* vKnotPtr, Vec3f* pointPtr, float* weightPtr, Vec3f* topPointPtr, float* topWeightPtr, Vec3f* r_pointPtr, float* r_weightPtr, Vec3f coord){
	//if(t.x < 0 || t.x > 1 || t.y < 0 || t.y > 1){
	//	return 0;
	//}

	float u = (maxU-minU)*t.x + minU;
	float v = (maxV-minV)*t.y + minV;

	if(u <= 0){
		u = 0;
	}
	if(v <= 0){
		v = 0;
	}

	if(t.x==1){
		CurvePoint(vDim-1, vDegree, vKnotPtr, r_pointPtr, r_weightPtr, v, curvedEdgePoint);
	}
	else if(t.y==1){
		CurvePoint(uDim-1, uDegree, uKnotPtr, topPointPtr, topWeightPtr, u, curvedEdgePoint);
	}
	else{
		SurfacePoint(u, v, uOrder, vOrder, uDim, vDim, uKnotPtr, vKnotPtr, pointPtr, weightPtr, &curvedEdgePoint);
	}

	Vec3f crd = coord;
	Vec3f crd2 = curvedEdgePoint;

	Vec3f diff = crd - crd2;
	return fabs(sqrt(diff.x*diff.x + diff.y*diff.y + diff.z*diff.z));
}
inline Vec3f getSurfaceVertex(const Vec2f& t, float maxU, float maxV, float minU, float minV, int uOrder, int vOrder, int uDegree, int vDegree, int uDim, int vDim, float* uKnotPtr, float* vKnotPtr, Vec3f* pointPtr, float* weightPtr, Vec3f* topPointPtr, float* topWeightPtr, Vec3f* r_pointPtr, float* r_weightPtr){
	Vec3f curvedEdgePoint;
	float u = (maxU-minU)*t.x + minU;
	float v = (maxV-minV)*t.y + minV;

	if(t.x==1){
		CurvePoint(vDim-1, vDegree, vKnotPtr, r_pointPtr, r_weightPtr, v, curvedEdgePoint);
	}
	else if(t.y==1){
		CurvePoint(uDim-1, uDegree, uKnotPtr, topPointPtr, topWeightPtr, u, curvedEdgePoint);
	}
	else{
		SurfacePoint(u, v, uOrder, vOrder, uDim, vDim, uKnotPtr, vKnotPtr, pointPtr, weightPtr, &curvedEdgePoint);
	}

	return curvedEdgePoint;
}
inline void guessPolyMapTug(NurbsPatch& patch, float& xCrd, float& yCrd, const Vec3f& pt, const bool& northPole, const bool& southPole){
	int uOrder = patch.order.x;
	int vOrder = patch.order.y;
	int uDegree = patch.uDegree;
	int vDegree = patch.vDegree;
	int uDim = patch.dim.x;
	int vDim = patch.dim.y;

	float* uKnotPtr = patch.uKnot.refValue().getArray();
	float* vKnotPtr = patch.vKnot.refValue().getArray();
	Vec3f* pointPtr = patch.point.refValue().getArray();
	float* weightPtr = patch.weight.refValue().getArray();

	Vec3f* topPointPtr = pointPtr+uDim*vDim-uDim;
	float* topWeightPtr = weightPtr+uDim*vDim-uDim;

	Vec3f* r_pointPtr = patch.r_point.getArray();
	float* r_weightPtr = patch.r_weight.getArray();


	float coordChange = .001;

	Vec2f diff;

	Vec3f curvedEdgePoint0;

	float maxU = patch.maxU();
	float maxV = patch.maxV();
	float minU = patch.minU();
	float minV = patch.minV();

	float diffArray[4];
	Vec3f destPoint[4];
	Vec2f base0 = Vec2f(xCrd, yCrd);

	if(northPole){
		for(int i = 200; --i>=0;){
			Vec2f basePointOffset[4] = {
				Vec2f(-coordChange, 0),
				Vec2f( coordChange, 0),

				Vec2f(0,-coordChange),
				Vec2f(0, coordChange),
			};
			float error = getDiff(convertNorthPoleToMercator(base0), destPoint[0], maxU, maxV, minU, minV, uOrder, vOrder, uDegree, vDegree, uDim, vDim, uKnotPtr, vKnotPtr, pointPtr, weightPtr, topPointPtr, topWeightPtr, r_pointPtr, r_weightPtr, pt);

			diffArray[0] = getDiff(convertNorthPoleToMercator(base0 + basePointOffset[0]), destPoint[0], maxU, maxV, minU, minV, uOrder, vOrder, uDegree, vDegree, uDim, vDim, uKnotPtr, vKnotPtr, pointPtr, weightPtr, topPointPtr, topWeightPtr, r_pointPtr, r_weightPtr, pt);
			diffArray[1] = getDiff(convertNorthPoleToMercator(base0 + basePointOffset[1]), destPoint[1], maxU, maxV, minU, minV, uOrder, vOrder, uDegree, vDegree, uDim, vDim, uKnotPtr, vKnotPtr, pointPtr, weightPtr, topPointPtr, topWeightPtr, r_pointPtr, r_weightPtr, pt);
			diffArray[2] = getDiff(convertNorthPoleToMercator(base0 + basePointOffset[2]), destPoint[2], maxU, maxV, minU, minV, uOrder, vOrder, uDegree, vDegree, uDim, vDim, uKnotPtr, vKnotPtr, pointPtr, weightPtr, topPointPtr, topWeightPtr, r_pointPtr, r_weightPtr, pt);
			diffArray[3] = getDiff(convertNorthPoleToMercator(base0 + basePointOffset[3]), destPoint[3], maxU, maxV, minU, minV, uOrder, vOrder, uDegree, vDegree, uDim, vDim, uKnotPtr, vKnotPtr, pointPtr, weightPtr, topPointPtr, topWeightPtr, r_pointPtr, r_weightPtr, pt);

			float leastDiff = diffArray[0];
			int k = 0;

			for(int j = 1; j < 4; j++){
				float curDiff = diffArray[j];
				if(curDiff < leastDiff){
					leastDiff = curDiff;
					k = j;
				}
			}
			if(diffArray[k] < error){
				base0+=basePointOffset[k];
				coordChange*=2;
			}
			else{
				coordChange*=.5;
				if(coordChange==0){
					break;
				}
			}
		}
	}
	else if(southPole){
		for(int i = 200; --i>=0;){
			Vec2f basePointOffset[4] = {
				Vec2f(-coordChange, 0),
				Vec2f( coordChange, 0),

				Vec2f(0,-coordChange),
				Vec2f(0, coordChange),
			};
			float error = getDiff(convertSouthPoleToMercator(base0), destPoint[0], maxU, maxV, minU, minV, uOrder, vOrder, uDegree, vDegree, uDim, vDim, uKnotPtr, vKnotPtr, pointPtr, weightPtr, topPointPtr, topWeightPtr, r_pointPtr, r_weightPtr, pt);

			diffArray[0] = getDiff(convertSouthPoleToMercator(base0 + basePointOffset[0]), destPoint[0], maxU, maxV, minU, minV, uOrder, vOrder, uDegree, vDegree, uDim, vDim, uKnotPtr, vKnotPtr, pointPtr, weightPtr, topPointPtr, topWeightPtr, r_pointPtr, r_weightPtr, pt);
			diffArray[1] = getDiff(convertSouthPoleToMercator(base0 + basePointOffset[1]), destPoint[1], maxU, maxV, minU, minV, uOrder, vOrder, uDegree, vDegree, uDim, vDim, uKnotPtr, vKnotPtr, pointPtr, weightPtr, topPointPtr, topWeightPtr, r_pointPtr, r_weightPtr, pt);
			diffArray[2] = getDiff(convertSouthPoleToMercator(base0 + basePointOffset[2]), destPoint[2], maxU, maxV, minU, minV, uOrder, vOrder, uDegree, vDegree, uDim, vDim, uKnotPtr, vKnotPtr, pointPtr, weightPtr, topPointPtr, topWeightPtr, r_pointPtr, r_weightPtr, pt);
			diffArray[3] = getDiff(convertSouthPoleToMercator(base0 + basePointOffset[3]), destPoint[3], maxU, maxV, minU, minV, uOrder, vOrder, uDegree, vDegree, uDim, vDim, uKnotPtr, vKnotPtr, pointPtr, weightPtr, topPointPtr, topWeightPtr, r_pointPtr, r_weightPtr, pt);

			float leastDiff = diffArray[0];
			int k = 0;

			for(int j = 1; j < 4; j++){
				float curDiff = diffArray[j];
				if(curDiff < leastDiff){
					leastDiff = curDiff;
					k = j;
				}
			}
			if(diffArray[k] < error){
				base0+=basePointOffset[k];
				coordChange*=2;
			}
			else{
				coordChange*=.5;
				if(coordChange==0){
					break;
				}
			}
		}
	}
	else{
		for(int i = 200; --i>=0;){
			Vec2f basePointOffset[4] = {
				Vec2f(-coordChange, 0),
				Vec2f( coordChange, 0),

				Vec2f(0,-coordChange),
				Vec2f(0, coordChange),
			};
			float error = getDiff(base0, destPoint[0], maxU, maxV, minU, minV, uOrder, vOrder, uDegree, vDegree, uDim, vDim, uKnotPtr, vKnotPtr, pointPtr, weightPtr, topPointPtr, topWeightPtr, r_pointPtr, r_weightPtr, pt);

			diffArray[0] = getDiff(base0 + basePointOffset[0], destPoint[0], maxU, maxV, minU, minV, uOrder, vOrder, uDegree, vDegree, uDim, vDim, uKnotPtr, vKnotPtr, pointPtr, weightPtr, topPointPtr, topWeightPtr, r_pointPtr, r_weightPtr, pt);
			diffArray[1] = getDiff(base0 + basePointOffset[1], destPoint[1], maxU, maxV, minU, minV, uOrder, vOrder, uDegree, vDegree, uDim, vDim, uKnotPtr, vKnotPtr, pointPtr, weightPtr, topPointPtr, topWeightPtr, r_pointPtr, r_weightPtr, pt);
			diffArray[2] = getDiff(base0 + basePointOffset[2], destPoint[2], maxU, maxV, minU, minV, uOrder, vOrder, uDegree, vDegree, uDim, vDim, uKnotPtr, vKnotPtr, pointPtr, weightPtr, topPointPtr, topWeightPtr, r_pointPtr, r_weightPtr, pt);
			diffArray[3] = getDiff(base0 + basePointOffset[3], destPoint[3], maxU, maxV, minU, minV, uOrder, vOrder, uDegree, vDegree, uDim, vDim, uKnotPtr, vKnotPtr, pointPtr, weightPtr, topPointPtr, topWeightPtr, r_pointPtr, r_weightPtr, pt);

			float leastDiff = diffArray[0];
			int k = 0;

			for(int j = 1; j < 4; j++){
				float curDiff = diffArray[j];
				if(curDiff < leastDiff){
					leastDiff = curDiff;
					k = j;
				}
			}
			if(diffArray[k] < error){
				base0+=basePointOffset[k];
				coordChange*=2;
			}
			else{
				coordChange*=.5;
				if(coordChange==0){
					break;
				}
			}
		}
		if(base0.x <= 0){
			base0.x = 0;
		}
		else if(base0.x >= 1){
			base0.x = 1;
		}
		if(base0.y <= 0){
			base0.y = 0;
		}
		else if(base0.y >= 1){
			base0.y = 1;
		}
	}
	xCrd = base0.x;
	yCrd = base0.y;
}

void addEdge0(NurbsPatch* patch, M_Vec3f& edgePoint, M_Float& edgeWeight){
	int uDim = patch->dim.x;
	edgePoint.length = uDim;
	edgeWeight.length = uDim;
	M_Vec3f& point = patch->point.refValue();
	M_Float& weight = patch->weight.refValue();

	for(int i = uDim; --i>=0;){
		edgePoint[i] = point[i];
		edgeWeight[i] = weight[i];
	}
}

void addEdge1(NurbsPatch* patch, M_Vec3f& edgePoint, M_Float& edgeWeight){
	int uDim = patch->dim.x;
	int vDim = patch->dim.y;
	edgePoint.length = vDim;
	edgeWeight.length = vDim;
	M_Vec3f& point = patch->point.refValue();
	M_Float& weight = patch->weight.refValue();

	for(int i = vDim; --i>=0;){
		int idx = i * uDim + uDim - 1;
		edgePoint[i] = point[idx];
		edgeWeight[i] = weight[idx];
	}
}

void addEdge2(NurbsPatch* patch, M_Vec3f& edgePoint, M_Float& edgeWeight){
	int uDim = patch->dim.x;
	int vDim = patch->dim.y;
	edgePoint.length = uDim;
	edgeWeight.length = uDim;
	M_Vec3f& point = patch->point.refValue();
	M_Float& weight = patch->weight.refValue();

	for(int i = uDim; --i>=0;){
		int idx = uDim * vDim - uDim + i;
		edgePoint[i] = point[idx];
		edgeWeight[i] = weight[idx];
	}
}

void addEdge3(NurbsPatch* patch, M_Vec3f& edgePoint, M_Float& edgeWeight){
	int uDim = patch->dim.x;
	int vDim = patch->dim.y;
	edgePoint.length = vDim;
	edgeWeight.length = vDim;
	M_Vec3f& point = patch->point.refValue();
	M_Float& weight = patch->weight.refValue();

	for(int i = vDim; --i>=0;){
		int idx = i * uDim;
		edgePoint[i] = point[idx];
		edgeWeight[i] = weight[idx];
	}
}

void guessEdge(const Vec3f& pt, float& closestDistance, int& closestPatch, int& closestEdge, int patch, int edge, int blockPatch, int blockEdge, float& closestT, float& closestOther, float other, const Vec3f* edgePointPtr, const float* edgeWeightPtr, const float* edgeKnotPtr, const float order, const float degree, const int dim, const float minT, const float maxT){
	if(patch == blockPatch && edge == blockEdge){
		return;
	}

	for(int i = 0; i < 1000; i++){
		float t = 1.0f / 999 * i;
		t = (maxT - minT) * t + minT;
		Vec3f testCrd;
		CurvePoint(dim-1, degree, edgeKnotPtr, edgePointPtr, edgeWeightPtr, t, testCrd);
		Vec3f diff = pt - testCrd;
		float distance = fabs(sqrt(diff.x*diff.x + diff.y*diff.y + diff.z*diff.z));
		if(distance < .1 && distance < closestDistance){
			closestDistance = distance;
			closestPatch = patch;
			closestEdge = edge;
			closestT = t;
			closestOther = other;
		}
	}
}

bool guessEdge(const Vec3f& pt, Vec2f& base0, int& patchOwner, const int blockEdge, M_Node& surface){
	static M_Vec3f edgePoint;
	static M_Float edgeWeight;
	int closestPatch = -1;
	int closestEdge = -1;
	float closestU;
	float closestV;
	float closestDistance = 1000000000000;

	for(int k = surface.length; --k>=0;){
		NurbsPatch* patch = dynamic_cast<NurbsPatch*>(surface[k].p);
		const float* uKnotPtr = patch->uKnot.refValue().getArray();
		const float* vKnotPtr = patch->vKnot.refValue().getArray();
		const Vec3f* pointPtr = patch->point.refValue().getArray();
		const float* weightPtr = patch->weight.refValue().getArray();
		const int uOrder = patch->order.x;
		const int vOrder = patch->order.y;
		const int uDim = patch->dim.x;
		const int vDim = patch->dim.y;

		const float maxU = patch->maxU();
		const float maxV = patch->maxV();
		const float minU = patch->minU();
		const float minV = patch->minV();

		float uDegree = patch->uDegree;
		float vDegree = patch->vDegree;


		addEdge0(patch, edgePoint, edgeWeight);
		guessEdge(pt, closestDistance, closestPatch, closestEdge, k, 0, patchOwner, blockEdge, closestU, closestV, 0, edgePoint.getArray(), edgeWeight.getArray(), uKnotPtr, uOrder, uDegree, uDim, minU, maxU);

		addEdge1(patch, edgePoint, edgeWeight);
		guessEdge(pt, closestDistance, closestPatch, closestEdge, k, 1, patchOwner, blockEdge, closestV, closestU, 1, edgePoint.getArray(), edgeWeight.getArray(), vKnotPtr, vOrder, vDegree, vDim, minV, maxV);

		addEdge2(patch, edgePoint, edgeWeight);
		guessEdge(pt, closestDistance, closestPatch, closestEdge, k, 2, patchOwner, blockEdge, closestU, closestV, 1, edgePoint.getArray(), edgeWeight.getArray(), uKnotPtr, uOrder, uDegree, uDim, minU, maxU);

		addEdge3(patch, edgePoint, edgeWeight);
		guessEdge(pt, closestDistance, closestPatch, closestEdge, k, 3, patchOwner, blockEdge, closestV, closestU, 0, edgePoint.getArray(), edgeWeight.getArray(), vKnotPtr, vOrder, vDegree, vDim, minV, maxV);
	}
	if(closestPatch != -1){
		base0.x = closestU;
		base0.y = closestV;
		patchOwner = closestPatch;
		return true;
	}
	return false;
}

inline void guessPolyMapTug(NurbsPatch& patch, M_Node& surface, float& xCrd, float& yCrd, int& patchOwner, const Vec3f& pt){
	int uOrder = patch.order.x;
	int vOrder = patch.order.y;
	int uDegree = patch.uDegree;
	int vDegree = patch.vDegree;
	int uDim = patch.dim.x;
	int vDim = patch.dim.y;

	float* uKnotPtr = patch.uKnot.refValue().getArray();
	float* vKnotPtr = patch.vKnot.refValue().getArray();
	Vec3f* pointPtr = patch.point.refValue().getArray();
	float* weightPtr = patch.weight.refValue().getArray();

	Vec3f* topPointPtr = pointPtr+uDim*vDim-uDim;
	float* topWeightPtr = weightPtr+uDim*vDim-uDim;

	Vec3f* r_pointPtr = patch.r_point.getArray();
	float* r_weightPtr = patch.r_weight.getArray();


	float coordChange = .001;

	Vec2f diff;

	Vec3f curvedEdgePoint0;

	float maxU = patch.maxU();
	float maxV = patch.maxV();
	float minU = patch.minU();
	float minV = patch.minV();

	float diffArray[4];
	Vec3f destPoint[4];
	Vec2f base0 = Vec2f(xCrd, yCrd);

	for(int i = 200; --i>=0;){
		Vec2f basePointOffset[4] = {
			Vec2f(-coordChange, 0),
			Vec2f( coordChange, 0),

			Vec2f(0,-coordChange),
			Vec2f(0, coordChange),
		};
		float error = getDiff(base0, destPoint[0], maxU, maxV, minU, minV, uOrder, vOrder, uDegree, vDegree, uDim, vDim, uKnotPtr, vKnotPtr, pointPtr, weightPtr, topPointPtr, topWeightPtr, r_pointPtr, r_weightPtr, pt);

		diffArray[0] = getDiff(base0 + basePointOffset[0], destPoint[0], maxU, maxV, minU, minV, uOrder, vOrder, uDegree, vDegree, uDim, vDim, uKnotPtr, vKnotPtr, pointPtr, weightPtr, topPointPtr, topWeightPtr, r_pointPtr, r_weightPtr, pt);
		diffArray[1] = getDiff(base0 + basePointOffset[1], destPoint[1], maxU, maxV, minU, minV, uOrder, vOrder, uDegree, vDegree, uDim, vDim, uKnotPtr, vKnotPtr, pointPtr, weightPtr, topPointPtr, topWeightPtr, r_pointPtr, r_weightPtr, pt);
		diffArray[2] = getDiff(base0 + basePointOffset[2], destPoint[2], maxU, maxV, minU, minV, uOrder, vOrder, uDegree, vDegree, uDim, vDim, uKnotPtr, vKnotPtr, pointPtr, weightPtr, topPointPtr, topWeightPtr, r_pointPtr, r_weightPtr, pt);
		diffArray[3] = getDiff(base0 + basePointOffset[3], destPoint[3], maxU, maxV, minU, minV, uOrder, vOrder, uDegree, vDegree, uDim, vDim, uKnotPtr, vKnotPtr, pointPtr, weightPtr, topPointPtr, topWeightPtr, r_pointPtr, r_weightPtr, pt);

		float leastDiff = diffArray[0];
		int k = 0;

		for(int j = 1; j < 4; j++){
			float curDiff = diffArray[j];
			if(curDiff < leastDiff){
				leastDiff = curDiff;
				k = j;
			}
		}
		if(diffArray[k] < error){
			base0+=basePointOffset[k];
			coordChange*=2;
		}
		else{
			coordChange*=.5;
			if(coordChange==0){
				break;
			}
		}
	}
	if(base0.x <= 0){
		if(false == guessEdge(pt, base0, patchOwner, 3, surface)){
			base0.x = 0;
		}
	}
	else if(base0.x >= 1){
		if(false == guessEdge(pt, base0, patchOwner, 1, surface)){
			base0.x = 1;
		}
	}
	if(base0.y <= 0){
		if(false == guessEdge(pt, base0, patchOwner, 0, surface)){
			base0.y = 0;
		}
	}
	else if(base0.y >= 1){
		if(false == guessEdge(pt, base0, patchOwner, 2, surface)){
			base0.y = 1;
		}
	}
	xCrd = base0.x;
	yCrd = base0.y;
}


void MultiSelect::reset(){
	clickedIndex = 0;
	selectRectMode=false;

	if(MultiSelect::accumulate == false){
		if(g_paletteMode == false){
			triList.length = 0;
		}
		coordList.length = 0;
		pointList.length = 0;
		ctrlPolySelList.length = 0;
	}
	//g_vcArtTriBase->transpSmls->skeleton.refValue()[0]->deleteGeometry();
	IndexedLineSet* polyLine=dynamic_cast<IndexedLineSet*>(g_vcSelectRectShape->geometry.p);
	dynamic_cast<Coordinate*>(polyLine->coord.p)->point.refValue().length=0;
	polyLine->coordIndex.refValue().length=0;
	beginDragPos.length = 0;
}
void MultiSelect::checkToReset(Art* part){
	Seamless* smls = part->smls;
	int len = coordList.length;
	for(int i = 0; i<len; i++){
		Art* prt = coordList[i].p;
		int idx = coordList[i].i;
		if(part == prt && smls->ActCtrlPnt == idx){
			return;
		}
	}
	reset();
}

void MultiSelect::backUpCoords(Art* part){
	Seamless* smls = part->smls;
	int len = coordList.length;
	beginDragPos.length = len;
	for(int i = 0; i<len; i++){
		Art* prt = coordList[i].p;
		int idx = coordList[i].i;
		if(part == prt && smls->ActCtrlPnt == idx){
			clickedIndex = i;
		}
		beginDragPos[i] = prt->coord[idx];
	}
}

inline void MultiSelect::rotateSelection(Art* part, Vec3f vec){

	Vec3f normal = -g_scene->rotation * Vec3f(0,0,1);

	Rotation normRot = Rotation(normal, Vec3f(0,0,1));

	Vec3f a = normRot * (vec-autoPivotPoint);
	Vec3f b = normRot * (beginDragPos[clickedIndex]-autoPivotPoint);
	//Rotation twistRot = Rotation(a, b);
	float angleA = Vec2f(a.y, a.x).getAngle();
	float angleB = Vec2f(b.y, b.x).getAngle();

	float rotateAngle = angleB - angleA;

	Rotation rot = Rotation(normal.x, normal.y, normal.z, rotateAngle);
	int len = beginDragPos.length;
	M_Vec3f& coord = part->coord.refValue();
	for(int i = triList.length; --i>=0;){
		coord[triList[i].i] = rot * (beginDragPos[i] - autoPivotPoint) + autoPivotPoint;
	}
}

int findNonRefPart(int seamlessLine, Art* part){
	int line=seamlessLine+part->plidx;
	while(g_sceneTreeList[line].node.p!= part){
		line++;
	};
	return line;
}

void processAllPatches(M_<NurbsPatch*>& cpNodeList);
void commonPostTug();


void updateCage(Seamless* smls){
	if(Art* part = dynamic_cast<Art*>(g_selNode.getRaw())){
		if(part->smls == smls){
			g_partCageClue->genCage();
		}
	}
}

int getEdgeInnerPoint(NurbsPatch* patch, int idx);

void dragPivotPoint(Seamless* smls, Vec3f pt){
	smls->catIMatrix();
	int line = g_scene->selLine;
	Art* part = dynamic_cast<Art*>(g_sceneTreeList[line].node.p);
	if(part == 0){
		return;
	}
	if(g_sceneTreeList[line].ref){
		line = findNonRefPart(getSeamlessParent(line), part);
	}
	//smls->catIMatrix();//is this needed again?

	int i = g_sceneTreeList[line].parent;

	Vec3f newPt;
	if(Art* parent=dynamic_cast<Art*>(g_sceneTreeList[i].node.p)){
		newPt = g_scene->getIViewMatrix() * g_catchIMatrix * parent->imatrix * pt - part->translation;
		newPt = newPt - parent->absolutePivotPoint;
	}
	else if(Seamless* parent=dynamic_cast<Seamless*>(g_sceneTreeList[i].node.p)){
		newPt = g_scene->getIViewMatrix() * g_catchIMatrix * parent->imatrix * pt - part->translation;
	}
	part->pivotPoint = newPt;
}

void PartCageClue::dragPart(Vec3f pt){
	Art* art = dynamic_cast<Art*>(g_selNode.getRaw());
	Seamless* smls = art->smls;
	if(g_scene->xRayEdit && smls->xRay){
		dragPivotPoint(smls, pt);
		return;
	}
	if(MultiSelect::pointList.length){
		g_dragActuallyHappned = true;
		Vec3f newPt = g_scene->getIViewMatrix() * g_catchIMatrix * pt;
		Vec3f diff = newPt - MultiSelect::beginDragPos[MultiSelect::clickedIndex];
		for(int i = MultiSelect::pointList.length; --i>=0;){
			int idx = MultiSelect::pointList[i].i;
			NurbsPatch* patch = MultiSelect::pointList[i].p;
			Vec3f svec = MultiSelect::beginDragPos[i] + diff;
			patch->setFilteredDimension(patch->point.refValue(), idx, svec * patch->weight[idx]);

			patch->smls->dragCtrlPoly(patch, idx);
		}
		processAllPatches(smls->cpNodeList);
		commonPostTug();
		if(Art* part = dynamic_cast<Art*>(g_selNode.getRaw())){
			updateCage(part->smls);
		}
	}
}

void PartClue::dragPart(Vec3f pt){
	if(0x8000&GetKeyState(VK_LMENU) || InputEvent::l_altDown){
		return;
	}
	Art* art = dynamic_cast<Art*>(g_selNode.getRaw());
	Seamless* smls = art->smls;
	if(g_scene->xRayEdit && smls->xRay){
		dragPivotPoint(smls, pt);
		return;
	}
	g_dragActuallyHappned = true;

	if(ImmediatePart* part = dynamic_cast<ImmediatePart*>(art)){
		smls->catIMatrix();

		if(g_settings->tugRadius.refValue()>0){
			if(Art::tugRadialTugSize){
				float inverseSize=(1/g_scene->getZoom());
				Vec3f sr = part->matrix * g_viewMatrix * part->coord[smls->ActCtrlPnt];
				g_pointer.translation.y = sr.y;
				g_settings->tugRadius = inverseSize * abs(sr.x - pt.x);
				part->houseKeep();
			}
			else{
				restoreCoord(smls->partList);
				M_Vec3f& coord = part->coord.refValue();
				radialTug(coord[smls->ActCtrlPnt], pt);
			}
		}
		else if(MultiSelect::coordList.length){
			//if(g_tugRotFlag){
			//	Vec3f vec = g_scene->getIViewMatrix() * part->imatrix * pt;
			//	MultiSelect::rotateSelection(part, vec);
			//}
			//else
			{
				Vec3f newPt = g_scene->getIViewMatrix() * g_catchIMatrix * part->imatrix * pt;
				Vec3f diff = newPt - MultiSelect::beginDragPos[MultiSelect::clickedIndex];
				for(int i = MultiSelect::coordList.length; --i >= 0;){
					if(ImmediatePart* part = dynamic_cast<ImmediatePart*>(MultiSelect::coordList[i].p)){
						int idx = MultiSelect::coordList[i].i;
						M_Vec3f& coord = part->coord.refValue();
						Vec3f newPt = MultiSelect::beginDragPos[i] + diff;
						smls->senseMirrorVertex(part, idx, newPt);
						coord[idx] = newPt;
					}
				}
			}
		}
		else{
			M_Vec3f& coord = part->coord.refValue();
			Vec3f newPt = g_scene->getIViewMatrix() * g_catchIMatrix * part->imatrix * pt;
			smls->senseMirrorVertex(part, smls->ActCtrlPnt, newPt);
			coord[smls->ActCtrlPnt] = newPt;
		}
	}
	if(Part* whitePart = dynamic_cast<Part*>(art)){
		smls->catIMatrix();
		M_Vec3f& coord = whitePart->coord.refValue();

		NurbsPatch* patch = 0;
		Stem* stem = 0;
		Polymap* polymap = 0;
		ImmediatePart* sourcePart = 0;
		if(getPolyMap(patch, stem, polymap, whitePart, smls->ActTri, sourcePart)){
			Vec3f newPt = g_scene->getIViewMatrix() * g_catchIMatrix * whitePart->imatrix * pt;
			if(sourcePart){
				Vec3f& crd = sourcePart->coord.refValue()[smls->ActCtrlPnt-stem->beginCoordOffset];
				guessPolyMapTug(*patch, crd.x, crd.y, newPt, stem->northPole, stem->southPole);
			}
			else if(polymap){
				int ActTri = (smls->ActTri - polymap->beginTriOffset) * 3;
				int ActTriCorner = smls->ActTriCorner;
				int crdIdx = polymap->rcCoord(ActTri + ActTriCorner);
				Vec2f& crd = polymap->coord.refValue()[crdIdx];
				int& patchOwner = polymap->patchOwner.refValue()[crdIdx];
				patch = dynamic_cast<NurbsPatch*>(polymap->surface.refValue()[patchOwner].p);
				guessPolyMapTug(*patch, polymap->surface.refValue(), crd.x, crd.y, patchOwner, newPt);
			}
			else{
				int ActTri = (smls->ActTri - stem->beginTriOffset) * 3;
				int ActTriCorner = smls->ActTriCorner;
				if(stem->mould == false){
					if(stem->mirror && stem->mirrorStem == 0){
						if(ActTri >= stem->coordIndex.length){
							ActTri -= stem->coordIndex.length;
							ActTriCorner = abs(3 - ActTriCorner) % 3;
							newPt.x = -(newPt.x - .5) + .5;
						}
					}
					Vec2f& crd = stem->coord.refValue()[stem->rcCoord(ActTri + ActTriCorner)];
					if(false == (stem->mirror && stem->mirrorStem == 0 && crd.x == .5)){
						crd.x = newPt.x;
					}
					crd.y = newPt.y;
					if(crd.x <= 0){
						crd.x = 0;
					}
					else if(crd.x >= 1){
						crd.x = 1;
					}
					if(crd.y <= 0){
						crd.y = 0;
					}
					else if(crd.y >= 1){
						crd.y = 1;
					}
				}
				else{
					if(stem->mirror && stem->mirrorStem == 0){
						if(ActTri >= stem->coordIndex.length){
							ActTri -= stem->coordIndex.length;
							ActTriCorner = abs(3 - ActTriCorner) % 3;
							newPt.x = -newPt.x;
						}
						Vec2f& crd = stem->coord.refValue()[stem->rcCoord(ActTri + ActTriCorner)];
						if(crd.x == .5){
							guessPolyMapTug(*patch, crd.x, crd.y, newPt, stem->northPole, stem->southPole);
							crd.x = .5;
						}
						else{
							guessPolyMapTug(*patch, crd.x, crd.y, newPt, stem->northPole, stem->southPole);
						}
					}
					else{
						if(stem->mirrorStem && stem->coordIndex.length == 0){
							stem = stem->mirrorStem;
							ActTriCorner = abs(3 - ActTriCorner) % 3;
						}
						Vec2f& crd = stem->coord.refValue()[stem->rcCoord(ActTri + ActTriCorner)];
						guessPolyMapTug(*patch, crd.x, crd.y, newPt, stem->northPole, stem->southPole);
					}
				}
			}
			patch->smls->doBuild();
			g_pointer.translation = g_scene->getViewMatrix() * coord[whitePart->triangle.refIndex(smls->ActTri * 3 + smls->ActTriCorner)];
		}
	}
}

void Node::catInverseRotation(Rotation& r){
}
void ProtoDeclare::catInverseRotation(Rotation& r){
}
void Seamless::catInverseRotation(Rotation& r){
	r = -rotation.refValue() * r;
}
void Art::catInverseRotation(Rotation& r){
	r = -rotation.refValue() * r;
}
void Transform::catInverseRotation(Rotation& r){
	r = -rotation.refValue() * r;
}
void AGroup::catInverseRotation(Rotation& r){
}

void AnimClue::dragPart(Vec3f pt){
	if(g_dragMorphFlag){
		g_dragActuallyHappned = true;
		dragPartForMorph(pt);
		return;
	}
	


	if(g_partCageClue->cpTug){
		g_partCageClue->dragPart(pt);
		return;
	}

	int line = g_scene->selLine;
	Art* part=dynamic_cast<Art*>(g_sceneTreeList[line].node.p);
	if(part==0){
		return;
	}
	Seamless* smls=part->smls;
	if(g_sceneTreeList[line].ref){
		line = findNonRefPart(getSeamlessParent(line), part);
	}

	if(dragTranPointMode){
		if(part->ik || part == part->smls->partList[0]){
			g_dragActuallyHappned = true;
			if(part->ik){
				Vec3f newPt = g_scene->getIViewMatrix() * g_catchIMatrix * pt;
				part->ikPos = newPt;
			}
			else{
				smls->catIMatrix();
				int i = g_sceneTreeList[line].parent;
				if(Art* parent=dynamic_cast<Art*>(g_sceneTreeList[i].node.p)){
					Vec3f newPt = g_scene->getIViewMatrix() * g_catchIMatrix * parent->imatrix * pt;
					part->translation = newPt - part->absolutePivotPoint;
				}
				else if(Seamless* parent=dynamic_cast<Seamless*>(g_sceneTreeList[i].node.p)){
					Vec3f newPt = g_scene->getIViewMatrix() * g_catchIMatrix * parent->imatrix * pt;
					part->translation = newPt - part->absolutePivotPoint;
				}
			}
		}
	}
	else{
		g_dragActuallyHappned = true;
		Vec3f pivotPoint  = part->pivotPointMatrix * Vec3f(0,0,0);

		//Vec3f pivotPoint=Rotation(1, 0, 0, .698132) * g_scene->rotation * Vec3f(0, .5, 0);
		Vec2f toDiff = Vec2f(pt.x-pivotPoint.x, pt.y-pivotPoint.y);
		Vec2f fromDiff = Vec2f(rotCtrlPnt.x-pivotPoint.x, rotCtrlPnt.y-pivotPoint.y);
		float toAngle = toDiff.getAngle();
		float fromAngle = fromDiff.getAngle();

		float angle = toAngle - fromAngle;

		if(part->ik){
			Vec3f newPt = g_scene->getIViewMatrix() * pt;
			Vec2f toDiff = Vec2f(pt.x - pivotPoint.x, pt.y - pivotPoint.y);
			Vec3f axis;
			axis = (-g_scene->rotation * Vec3f(0, 0, 1)).normalize();
			part->ikRot = part->ikRot.refValue() * Rotation(axis, angle);
		}
		else{
			Rotation unwind;
			int i=line;
			do{
				i = g_sceneTreeList[i].parent;
				g_sceneTreeList[i].node->catInverseRotation(unwind);
			}while(i>0);

			Vec3f axis;
			if(g_transformCP->rotAxisZ){
				axis = (-g_scene->rotation * unwind * Vec3f(0, 0, 1)).normalize();
			}
			else if(g_transformCP->rotAxisY){
				axis = (-g_scene->rotation * unwind * Vec3f(0, 1, 0)).normalize();
			}
			else if(g_transformCP->rotAxisX){
				axis = (-g_scene->rotation * unwind * Vec3f(-1, 0, 0)).normalize();
			}

			part->rotation = part->rotation.refValue() * Rotation(axis, angle);
		}
		rotCtrlPnt=pt;
	}

	part->smls->catIMatrix();
	part->smls->catMatrix();

	anim->setMorphersForTransformDrag(smls);
}

inline Matrix dxWindMatrix(Matrix& parentMatrix, Vec3f& scale, Rotation& rotation, Vec3f& translation);

void TransformCP::getPivotPointMatrix(Matrix& pivotPointMatrix, Seamless* smls)const{
	Matrix parentMatrix = dxWindMatrix(g_scene->getViewMatrix(), smls->scale, smls->rotation, smls->translation);
	pivotPointMatrix = translationMatrix(translation + pivotPoint) * parentMatrix;
	pivotPointMatrix = rotationMatrix(scaleOrientation) * scaleMatrix(scale) * rotationMatrix(-scaleOrientation) * rotationMatrix(rotation) * pivotPointMatrix;
}


Vec3f calcRingNormal(int v, Vec3f pivotPoint, int selUdim, int from){
	Vec3f ringNormal;
	for(int u = 0; u < selUdim - 1; u++){
		Vec3f pa = pivotPoint;
		Vec3f pb = MultiSelect::beginDragPos[from + u + 1];
		Vec3f pc = MultiSelect::beginDragPos[from + u];
		Vec3f test = pb - pa;
		if(fabs(test.x) < .00001 && fabs(test.y) < .00001 && fabs(test.z) < .00001){
			continue;
		}
		test = pc - pa;
		if(fabs(test.x) < .00001 && fabs(test.y) < .00001 && fabs(test.z) < .00001){
			continue;
		}
		test = pb - pc;
		if(fabs(test.x) < .00001 && fabs(test.y) < .00001 && fabs(test.z) < .00001){
			continue;
		}

		ringNormal += calcNormal(pa, pb, pc);
	}
	return ringNormal.normalize();
}

inline void twistRing(const int& i, const int& selUdim, const int from, M_Vec3f& point, M_Vec3f& dpoint, M_Float& weight, float& twistAngle){

	Vec3f pivotPoint = calcRingPivotPoint2(selUdim, from);
	Vec3f normal = calcRingNormal(i, pivotPoint, selUdim, from);

	Rotation normRot = -Rotation(Vec3f(0, 1, 0), normal);

	Rotation rot = Rotation(normal.x, normal.y, normal.z, twistAngle);
	for(int u = 0; u < selUdim; u++){
		int srcIdx = from + u;
		int dstIdx = MultiSelect::pointList[srcIdx].i;
		dpoint[dstIdx] = rot * (MultiSelect::beginDragPos[srcIdx] - pivotPoint) + pivotPoint;
		point[dstIdx] = dpoint[dstIdx] * weight[dstIdx];
	}
}

void stretch(bool lock, bool x, bool z, M_Vec3f& dpoint){
	Vec3f pp = g_scene->getViewMatrix() * g_transformCP->pivotPoint;
	Vec3f d = g_scene->getViewMatrix() * MultiSelect::beginDragPos[MultiSelect::clickedIndex];
	pp.z = d.z;
	g_pointer.translation.z = pp.z;
	pp = g_scene->getIViewMatrix() * pp;
	//g_transformCP->pivotPoint.refValue() = pp;
	//pp = g_scene->getIViewMatrix() * pp;

	Vec3f newPt = g_scene->getIViewMatrix() * g_catchIMatrix * g_pointer.translation;
	Vec3f ev = newPt - (g_transformCP->translation + pp);
	Vec3f bv = MultiSelect::beginDragPos[MultiSelect::clickedIndex] - pp;
	float bdis = bv.length();
	float edis = ev.length();
	Vec3f baxis = bv.normalize();
	Vec3f eaxis = ev.normalize();

	if(fabs(bdis) > .0001){
		float s = edis / bdis;
		if(x){
			g_transformCP->scale.refValue().x = s;
		}
		g_transformCP->scale.refValue().y = s;
		if(z){
			g_transformCP->scale.refValue().z = s;
		}
		if(lock == false){
			g_transformCP->rotation.refValue() = Rotation(baxis, eaxis);
		}
		g_transformCP->scaleOrientation = Rotation(baxis, Vec3f(0, 1, 0));
		g_pointer.translation = g_scene->getViewMatrix() * dpoint[MultiSelect::pointList[MultiSelect::clickedIndex].i];
//		g_pointer.translation.z = 0;
	}
}

void scalePoints(bool lock, bool x, bool z, M_Vec3f& dpoint){
	Vec3f pp = g_scene->getViewMatrix() * g_transformCP->pivotPoint;
	Vec3f d = g_transformCP->remRotCtrlPnt;
	pp.z = d.z;
	g_pointer.translation.z = pp.z;
	pp = g_scene->getIViewMatrix() * pp;

	Vec3f newPt = g_scene->getIViewMatrix() * g_catchIMatrix * g_pointer.translation;
	Vec3f ev = newPt - (g_transformCP->translation + pp);
	Vec3f bv = g_scene->getIViewMatrix() * g_transformCP->remRotCtrlPnt - pp;
	float bdis = bv.length();
	float edis = ev.length();
	Vec3f baxis = bv.normalize();
	Vec3f eaxis = ev.normalize();
	Rotation so = Rotation(baxis, Vec3f(0, 1, 0));

	if(fabs(bdis) > .0001){
		float s = edis / bdis;
		if(x){
			g_transformCP->scale.refValue().x = s;
		}
		g_transformCP->scale.refValue().y = s;
		if(z){
			g_transformCP->scale.refValue().z = s;
		}
		if(lock == false){
			g_transformCP->rotation.refValue() = Rotation(baxis, eaxis);
			g_transformCP->rotCtrlPnt = g_pointer.translation;
		}
		else{
			g_pointer.translation = g_transformCP->rotCtrlPnt = g_scene->getViewMatrix() * (g_transformCP->rotation * -so * Vec3f(0, edis, 0) + pp);
		}
		g_transformCP->scaleOrientation = so;
		//g_transformCP->rotCtrlPnt = g_pointer.translation;
//		g_pointer.translation.z = 0;
	}
}

void TransformCP::quantizeRotationLever(){
	if(false == ((0x8000&GetKeyState(VK_LCONTROL) || InputEvent::l_ctrlDown)) || stretchHappened){
		return;
	}

	Vec3f rp = g_transformCP->rotCtrlPnt;
	Vec3f pp = g_scene->getViewMatrix() * (g_transformCP->pivotPoint + g_transformCP->translation);
	rp.z = pp.z;
	rp = (rp - pp);
	Vec3f nv = rp.normalize();
	float closest = 100000;
	for(int i = 0; i < 8; i++){
		Vec3f v = Rotation(0, 0, 1, PI * 2 / 8 * i) * Vec3f(1, 0, 0);
		float diff = abs((nv - v).length());
		if(diff < closest){
			closest = diff;
			leverAxis = v;
		}
	}
	g_transformCP->rotCtrlPnt = leverAxis * rp.length() + pp;
}

void NurbsPatch::drag(Vec3f pt){
	g_viewMatrix = g_scene->getViewMatrix();
	Matrix iViewMatrix = g_scene->getIViewMatrix();
	M_Vec3f& point = this->point;
	M_Vec3f& dpoint = this->dpoint;
	int idx = index[0];

	//if((0x8000&GetKeyState(VK_RCONTROL) || InputEvent::r_ctrlDown) && (0x8000 & GetKeyState(VK_RSHIFT))){
	//	


	//		Matrix parentMatrix = g_scene->getViewMatrix();
	//		Matrix pivotPointMatrix = translationMatrix(g_transformCP->translation + g_transformCP->pivotPoint) * parentMatrix;
	//		pivotPointMatrix = rotationMatrix(g_transformCP->rotation) * pivotPointMatrix;
	//		Vec3f screenPivotPoint = pivotPointMatrix * Vec3f();

	//		Vec3f selCoord  = pivotPointMatrix * (MultiSelect::beginDragPos[MultiSelect::clickedIndex] - g_transformCP->pivotPoint);

	//		Vec3f bv = selCoord - screenPivotPoint;
	//		Vec3f ev = pt - screenPivotPoint;

	//		float bdis = bv.length();
	//		float edis = ev.length();

	//		Vec3f baxis = bv.normalize();
	//		Vec3f eaxis = ev.normalize();
	//		if(fabs(bdis) > .0001){
	//			float s = edis / bdis;
	//			Vec3f scale = Vec3f(s, s, s);
	//			g_transformCP->scale = scale;
	//			g_pointer.translation = g_scene->getViewMatrix() * dpoint[MultiSelect::pointList[MultiSelect::clickedIndex].i];
	//		}
	//	}
	if(g_transformCP->dragRotPointMode){
		if((0x8000&GetKeyState(VK_LSHIFT) || InputEvent::l_shiftDown)){
			if((0x8000&GetKeyState(VK_LCONTROL) || InputEvent::l_ctrlDown)){
				if((0x8000&GetKeyState(VK_LMENU) || InputEvent::l_altDown)){
				}
				else{
					scalePoints(false, false, false, dpoint);
					g_transformCP->stretchHappened = true;
				}
			}
			else{
				
			}
		}
		else if((0x8000&GetKeyState(VK_LMENU) || InputEvent::l_altDown)){
			if((0x8000&GetKeyState(VK_LCONTROL) || InputEvent::l_ctrlDown)){
				scalePoints(true, false, true, dpoint);
			}
			else{
				//scalePoints(true, true, true, dpoint);
				scalePoints(true, false, false, dpoint);
			}
		}
		else if((0x8000&GetKeyState(VK_LCONTROL) || InputEvent::l_ctrlDown)){
			g_transformCP->rotCtrlPnt = g_pointer.translation;
			Vec3f pivotPoint = g_transformCP->pivotPoint + g_transformCP->translation;
			g_transformCP->refreshRemPoints();
			g_transformCP->pivotPoint.refValue() = pivotPoint; 
		}
		else{
			g_transformCP->dragRot(this, pt);
		}
	}
	else if((0x8000&GetKeyState(VK_LSHIFT) || InputEvent::l_shiftDown)){
		if((0x8000&GetKeyState(VK_LCONTROL) || InputEvent::l_ctrlDown)){
			if((0x8000&GetKeyState(VK_LMENU) || InputEvent::l_altDown)){
				g_transformCP->scaleOrientation.refValue() = Rotation();
				g_transformCP->remRotCtrlPnt = g_transformCP->rotCtrlPnt;
				Vec3f newPt = g_scene->getIViewMatrix() * g_catchIMatrix * pt;
				Vec3f a = MultiSelect::beginDragPos[MultiSelect::clickedIndex] - g_transformCP->pivotPoint;
				Vec3f b = -g_transformCP->rotation.refValue() * (newPt - g_transformCP->pivotPoint - g_transformCP->translation);
				if(fabs(a.x) < .00001){
					a.x = 1;
					b.x = 1;
				}
				if(fabs(a.y) < .00001){
					a.y = 1;
					b.y = 1;
				}
				if(fabs(a.z) < .00001){
					a.z = 1;
					b.z = 1;
				}
				Vec3f scale = b / a;
				g_transformCP->scale = scale;
			}
			else{
			}
		}
		else{
			stretch(true, false, true, dpoint);
		}
	}
	else if((0x8000&GetKeyState(VK_LMENU) || InputEvent::l_altDown)){
		if((0x8000&GetKeyState(VK_LCONTROL) || InputEvent::l_ctrlDown)){
			stretch(true, false, false, dpoint);
		}
		else{
			stretch(true, true, true, dpoint);
			//stretch(false, false, false, dpoint);
		}
	}
	else if(0x8000 & GetKeyState(VK_LCONTROL)){
	}
	else if(g_transformCP->dragPivotPointMode){
		g_transformCP->pivotPoint.refValue() = g_scene->getIViewMatrix() * g_catchIMatrix * translationMatrix(-g_transformCP->translation) * pt;
		g_transformCP->rotCtrlPnt = g_scene->getViewMatrix() * g_transformCP->pivotPoint + g_transformCP->leverAxis * .5;
	}
	else if(g_settings->tugRadius > 0){
		if(Art::tugRadialTugSize){
			float inverseSize = (1/g_scene->getZoom());
			Vec3f sr = matrix * g_viewMatrix * dpoint[index[0]];
			g_pointer.translation.y = sr.y;
			g_settings->tugRadius = inverseSize * abs(sr.x - pt.x);
			houseKeep();
		}
		else{
			//restorePatchPoint(this);
			radialTug(g_backUpCoord[fromPoint + idx], pt);
			smls->calcAllPolyMidPoints();
			commonPostTug();
			updateCage(smls);
			houseKeep();
		}
	}
	else{
		Matrix pivotPointMatrix;
		g_transformCP->getPivotPointMatrix(pivotPointMatrix, smls);

		Vec3f pt2 = pt - (pivotPointMatrix * MultiSelect::beginDragPos[MultiSelect::clickedIndex] - pivotPointMatrix * g_transformCP->pivotPoint);

		Vec3f newPt = g_scene->getIViewMatrix() * g_catchIMatrix * pt2;

		g_transformCP->rotCtrlPnt = g_transformCP->remRotCtrlPnt + g_pointer.translation - g_pointer.remTranslation;

		g_transformCP->translation = newPt - g_transformCP->pivotPoint;
	}
	g_dragActuallyHappned = true;
}

void TransformCP::dragRot(NurbsPatch* selPatch, Vec3f pt){

	Seamless* smls = selPatch->smls;

	Vec3f& pivotPoint = this->pivotPoint.refValue();
	//Rotation& rotation = this->rotation.refValue();
	Vec3f& translation = this->translation.refValue();

	Matrix pivotPointMatrix;
	getPivotPointMatrix(pivotPointMatrix, smls);

	int line = g_scene->selLine;

	Vec3f pivotPointVM  = pivotPointMatrix * Vec3f(0,0,0);

	//Vec3f pivotPoint=Rotation(1, 0, 0, .698132) * g_scene->rotation * Vec3f(0, .5, 0);
	Vec2f toDiff = Vec2f(pt.x - pivotPointVM.x, pt.y - pivotPointVM.y);
	Vec2f fromDiff = Vec2f(remRotCtrlPnt.x - pivotPointVM.x, remRotCtrlPnt.y - pivotPointVM.y);
	float toAngle = toDiff.getAngle();
	float fromAngle = fromDiff.getAngle();

	float angle = toAngle - fromAngle;

	if(selectRings && 0x8000 & GetKeyState(VK_RMENU)){
		M_Vec3f& point = selPatch->point.refValue();
		M_Vec3f& dpoint = selPatch->dpoint.refValue();
		M_Float& weight = selPatch->weight.refValue();

		for(int i = 0; i < MultiSelect::vRingDim; i++){
			int from = i * MultiSelect::uRingDim;
			twistRing(i, MultiSelect::uRingDim, from, point, dpoint, weight, angle);
		}
		rotCtrlPnt = pt;
		rotation.refValue() = Rotation();
		selPatch->processFields();
		smls->doBuild();
		smls->ctrf = true;

		selPatch->houseKeep();
		return;
	}

	Rotation unwind;
	int i = line;
	do{
		i = g_sceneTreeList[i].parent;
		g_sceneTreeList[i].node->catInverseRotation(unwind);
	}while(i>0);

	Vec3f axis;
	if(rotAxisZ){
		axis = (-g_scene->rotation * unwind * Vec3f(0, 0, 1)).normalize();
	}
	else if(rotAxisY){
		axis = (-g_scene->rotation * unwind * Vec3f(0, 1, 0)).normalize();
	}
	else if(rotAxisX){
		axis = (-g_scene->rotation * unwind * Vec3f(-1, 0, 0)).normalize();
	}
	
	rotCtrlPnt = pt;
	rotation = Rotation(axis, angle);
/*	
	Matrix matrix = translationMatrix(-pivotPoint) * scaleMatrix(scale) * rotationMatrix(rotation) * translationMatrix(translation) * translationMatrix(pivotPoint);


	if(MultiSelect::pointList.length){
		for(int i = MultiSelect::pointList.length; --i>=0;){
			int idx = MultiSelect::pointList[i].i;
			NurbsPatch* patch = MultiSelect::pointList[i].p;
			Vec3f svec = matrix * MultiSelect::beginDragPos[i];
			patch->setFilteredDimension(patch->point.refValue(), idx, svec * patch->weight[idx]);
		}
		smls->calcAllPolyMidPoints();
		processAllPatches(smls->cpNodeList);
		commonPostTug();
	}
	selPatch->houseKeep();
*/

}

const char* commonSpecStr ="\
DEF pelvis CommonSpec{\n\
color 1 1 1\n\
alternative \"seg0\"\n\
}\n\
DEF tummy0 CommonSpec{\n\
color 0 0 1\n\
alternative [\"seg1\" \"tummy\"]\n\
}\n\
DEF tummy1 CommonSpec{\n\
color 0 1 1\n\
alternative \"seg2\"\n\
}\n\
DEF tummy2 CommonSpec{\n\
color 1 .5 0\n\
alternative \"seg3\"\n\
}\n\
DEF thorax CommonSpec{\n\
color .7 0 .7\n\
alternative \"seg4\"\n\
}\n\
DEF neck0 CommonSpec{\n\
color 1 1 0\n\
alternative [\"seg5\" \"neck\"]\n\
}\n\
DEF neck1 CommonSpec{\n\
color 1 0 1\n\
alternative \"seg6\"\n\
}\n\
DEF head CommonSpec{\n\
color 0 0 .8\n\
alternative \"seg7\"\n\
}\n\
DEF l_ear CommonSpec{\n\
color 1 1 1\n\
}\n\
DEF r_ear CommonSpec{\n\
color 1 1 1\n\
}\n\
DEF hair CommonSpec{\n\
color 0 0 1\n\
}\n\
DEF l_eyeCrater CommonSpec{\n\
color 0 1 0\n\
}\n\
DEF r_eyeCrater CommonSpec{\n\
color 0 0 1\n\
}\n\
DEF l_eyeball CommonSpec{\n\
color 1 1 0\n\
}\n\
DEF r_eyeball CommonSpec{\n\
color 1 1 0\n\
}\n\
DEF l_collar CommonSpec{\n\
color 1 1 1\n\
}\n\
DEF l_shoulder CommonSpec{\n\
color [1 0 0 0 1 0 0 0 1 1 1 0 1 0 1 0 1 1 1 .5 0]\n\
}\n\
DEF l_elbow CommonSpec{\n\
color [0 1 1 .7 .8 0 1 .5 1]\n\
}\n\
DEF l_wrist CommonSpec{\n\
color [.5 0 1 0 1 0]\n\
}\n\
DEF l_thumb0 CommonSpec{\n\
color 0 .8 0\n\
alternative \"thumb\"\n\
}\n\
DEF l_index CommonSpec{\n\
color 1 1 1\n\
}\n\
DEF l_middle CommonSpec{\n\
color 1 1 1\n\
}\n\
DEF l_ring CommonSpec{\n\
color 1 1 1\n\
}\n\
DEF l_pinky CommonSpec{\n\
color 1 1 1\n\
}\n\
DEF r_collar CommonSpec{\n\
color 1 1 1\n\
}\n\
DEF r_shoulder CommonSpec{\n\
color [1 0 0 0 1 0 0 0 1 1 1 0 1 0 1 0 1 1 1 .5 0]\n\
}\n\
DEF r_elbow CommonSpec{\n\
color [0 1 1 .7 .8 0 1 .5 1]\n\
}\n\
DEF r_wrist CommonSpec{\n\
color [.5 0 1 0 1 0]\n\
}\n\
DEF r_thumb CommonSpec{\n\
color 1 1 1\n\
}\n\
DEF r_index CommonSpec{\n\
color 1 1 1\n\
}\n\
DEF r_middle CommonSpec{\n\
color 1 1 1\n\
}\n\
DEF r_ring CommonSpec{\n\
color 1 1 1\n\
}\n\
DEF r_pinky CommonSpec{\n\
color 1 1 1\n\
}\n\
DEF l_breast CommonSpec{\n\
color 1 1 1\n\
}\n\
DEF r_breast CommonSpec{\n\
color 1 1 1\n\
}\n\
DEF navel CommonSpec{\n\
color 1 1 1\n\
}\n\
DEF l_hip CommonSpec{\n\
color [1 0 0 0 1 0 0 0 1 1 1 0 1 0 1 0 1 1 1 .5 0]\n\
}\n\
DEF l_knee CommonSpec{\n\
color [0 1 1 .7 .8 0 1 .5 1]\n\
}\n\
DEF l_ankle CommonSpec{\n\
color [.5 0 1 0 1 0]\n\
}\n\
DEF r_hip CommonSpec{\n\
color [.5 0 1 0 1 0 0 0 1 1 1 0 1 0 1 0 1 1 1 .5 0]\n\
}\n\
DEF r_knee CommonSpec{\n\
color [0 1 1 .7 .8 0 1 .5 1]\n\
}\n\
DEF r_ankle CommonSpec{\n\
color [.5 0 1 0 1 0]\n\
}\n\
DEF jaw CommonSpec{\n\
color [0 1 0 1 0 0]\n\
}\n\
DEF l_eyebrow CommonSpec{\n\
color [1 1 1]\n\
}\n\
DEF r_eyebrow CommonSpec{\n\
color [1 1 1]\n\
}\n\
DEF top_teeth CommonSpec{\n\
color [1 1 1]\n\
}\n\
DEF bottom_teeth CommonSpec{\n\
color [1 1 1]\n\
}\n\
";

void Settings::loadINI(){
	#define maxLen 0x200
	//char tmpStr[maxLen+1];

	M_Node mnode;
	if(Browser::createVrmlFromFile(g_setupPath, mnode)){
		for(int k=0; k<mnode.length; k++){
			switch(mnode[k]->getType()){
				case SETTINGS:	 g_settings = dynamic_cast<Settings*>(mnode[k].p); break;
				case OPENNETFILE:	g_openNetFile = dynamic_cast<OpenNetFile*>(mnode[k].p); break;
				case PASTEOPTIONS:	g_pasteOptions = dynamic_cast<PasteOptions*>(mnode[k].p); break;
				case DELETEOPTIONS:	g_deleteOptions = dynamic_cast<DeleteOptions*>(mnode[k].p); break;
				case TIMELINEFILTER: g_timelineFilter = dynamic_cast<TimelineFilter*>(mnode[k].p); break;
				case FREEZETIME: break;
			}
		}
	}
	else{
		g_settings = new Settings;
		//strcpy(tmpStr,"seamless3dSetup.wrl");
		//_fullpath( tmpStr, "seamless3dSetup.wrl", maxLen );
		g_settings->writeINIFile();
		g_createdINI = true;
	}
	if(g_settings->palette == 0){
		g_settings->palette = new Palette;
	}
	g_palette = dynamic_cast<Palette*>(g_settings->palette.p);
	if(g_settings->commonSpec.length == 0){
		createSmlsFromString(g_settings->commonSpec.refValue(), g_loadStr = commonSpecStr);
	}

	if(g_settings->file.length){
		String drive;
		String dir;
		splitPath(g_settings->file[0], drive, dir, String(), String());
		if(drive.length){
			g_seamlessDir = drive;
			g_seamlessDir += dir;
		}
		else{
			g_settings->file.refValue().length = 0;
		}
		_chdir(g_seamlessDir.c_str());
	}

	g_writeGzipped=g_settings->saveGZipped.refValue();

	g_settings->version.refValue()=g_version;
	g_settings->freeCPU = g_settings->freeCPU.refValue();

/*	const M_Int& win=g_settings->win;
	if(g_settings->win.length){

		g_3dwinSize.x=win[THREE_D_POS_X];
		g_3dwinSize.y=win[THREE_D_POS_Y];
		g_3dwinPos.x=win[THREE_D_CX];
		g_3dwinPos.y=win[THREE_D_CY];

		g_toolbarPos.x=win[TOOLBAR_POS_X];
		g_toolbarPos.y=win[TOOLBAR_POS_Y];
		//g_toolbarSize.x=win[TOOLBAR_CX];
		//g_toolbarSize.y=win[TOOLBAR_CY];

		g_sceneTreePos.x=win[SCENETREE_POS_X];
		g_sceneTreePos.y=win[SCENETREE_POS_Y];
		g_sceneTreeSize.x=win[SCENETREE_CX];
		g_sceneTreeSize.y=win[SCENETREE_CY];

		g_nodeWinPos.x=win[NODE_POS_X];
		g_nodeWinPos.y=win[NODE_POS_Y];

	}
*/
}

/*
void localPermanentTransform(){
	Art* part=dynamic_cast<Art*>(g_selNode.getRawPtr());
	Seamless* smls=part->smls;
	BackUp::monostable.beginCycle();
	part->geometry->scale(part->scale.refValue());
	part->geometry->rotate(part->rotation.refValue());
	part->geometry->translate(part->translation.refValue());
	part->scale=Vec3f(1,1,1);
	part->rotation=Rotation(0,1,0,0);
	part->translation=Vec3f(0,0,0);
	smls->convertToRenderFormat();
}
void permanentTransform(){
	Art* part=dynamic_cast<Art*>(g_selNode.getRawPtr());
	Seamless* smls=part->smls;
	BackUp::monostable.beginCycle();

	IndexedFaceSet* cpy=new IndexedFaceSet;
	Matrix matrix;
	matrix=scaleMatrix(g_scene->scale);
	matrix=matrix*translationMatrix(g_scene->translation);
	matrix=matrix*rotationMatrix(g_scene->rotation);
	CoordinatePtr(part->geometry->coord)->point.refValue().multMatrixLeft(matrix);
	NormalPtr(part->geometry->normal)->vector.refValue().multMatrixLeft(rotationMatrix(g_scene->rotation));

	g_scene->scale=Vec3f(1,1,1);
	g_scene->translation=Vec3f(0,0,0);
	g_scene->rotation=Rotation();
	smls->convertToRenderFormat();
}
/*
void Art::genUndoFilename(){
	char buf [0x200];
	sprintf(buf,"%s%i_",name.getArray(),undoUniqueNameCntr++);
	undoName=buf;
	backupToCache();
	saved=true;
}
void Art::genCacheUrl(String& str ,int cacheEntry){
	str=g_undoDir;
	str+=undoName;
	str+=cacheEntry;
	str+=".wrl";
}

void Art::backupToCache(){
	String backUpFileName;
	if(geometry==NULL)return; //never NULL the way it is so this line could be removed
	saved=false;
	setTitleName();
	highestEntry=cacheEntry;
	genCacheUrl(backUpFileName, cacheEntry++);
	writePart(backUpFileName);
}
void unDo(){
	ArtPtr pj=g_selNode;
	String backUpFileName;
	if(pj->cacheEntry<2)return;

	pj->genCacheUrl(backUpFileName ,--pj->cacheEntry-1);
	pj->loadGZippedPart(backUpFileName);

	if(pj->geometry->coord!=NULL){
		if(g_seamless->ActCtrlPnt>=pj->geometry->coord->point.length||g_seamless->ActTri>=pj->geometry->coordIndex.length/3){
			g_seamless->ActCtrlPnt=0;
			g_seamless->ActTri=0;
		}
	}
	g_seamless->convertToRenderFormat();
}
void reDo(){
	ArtPtr pj=g_selNode;
	String backUpFileName;
	if(pj->cacheEntry>pj->highestEntry)
		return;

	pj->genCacheUrl(backUpFileName, pj->cacheEntry++);
	pj->loadGZippedPart(backUpFileName);

	if(g_seamless->ActCtrlPnt>=pj->geometry->coord->point.length||g_seamless->ActTri>=pj->geometry->coordIndex.length/3){
		g_seamless->ActCtrlPnt=0;
		g_seamless->ActTri=0;
	}
	g_seamless->convertToRenderFormat();
}
*//*
void transformActiveVector(){
	Art* part=dynamic_cast<Art*>(g_selNode.getRawPtr());
	Matrix matrix = part->matrix;

	CoordinatePtr(pj->geometry->coord)->point.refValue().multMatrixLeft(matrix);
	NormalPtr(pj->geometry->normal)->vector.refValue().multMatrixLeft(rotationMatrix(g_scene->rotation));
}
void unTransformActiveVector(){
	ArtPtr pj=g_selNode;
	Matrix matrix;
	NormalPtr(pj->geometry->normal)->vector.refValue().multMatrixLeft(rotationMatrix(-g_scene->rotation));
	matrix=rotationMatrix(-g_scene->rotation);
	matrix=matrix*translationMatrix(-g_scene->translation);
	matrix=matrix*scaleMatrix(Vec3f(1,1,1)/g_scene->scale);
	CoordinatePtr(pj->geometry->coord)->point.refValue().multMatrixLeft(matrix);
}
*/

void modifyMirrorVertexX(ArtPtr selPart){
	/*
	if(!g_mirrorVertexModeX||!g_mirrorModeVertexFound)return;
	SeamlessPtr smls=selPart->smls;
	M_Vec3f& point=smls->refPoint();
	M_Vec3f& renderPoint=smls->renderIfs->refPoint();

	Vec3f svec = selPart->matrix*point[smls->ActCtrlPnt];
	svec.x=-svec.x;
	point[g_mirrorVertex]=svec-g_mirrorPart->matrix*Vec3f();

	renderPoint[g_mirrorVertex]=g_mirrorPart->matrix*point[g_mirrorVertex];
	*/
}
/*
void snapSinglePointToRem(){
	ArtPtr part=g_selNode;
	Seamless* smls=part->smls;
	M_Vec3f& coord=part->coord.refValue();

	Vec3f svec = remPJ->matrix*remPJ->coord.refValue()[smls->remActCtrlPnt];
	coord[smls->ActCtrlPnt]=svec-part->matrix*Vec3f();

	smls->convertToRenderFormat();
	makePointerEqualIndexLocation();
}
/*
void Seamless::joinVertex(){
	ArtPtr selPart=selectPartFromVertex(this, ActCtrlPnt);
	ArtPtr remPart=selectPartFromVertex(this, remActCtrlPnt);
	M_Int &coordIndex = refCoordIndex();
	
	int i,j;
	//check to see if a point is being joined to another point within the same triangle
	for( i=0; i<coordIndex.length; i++){
		if(coordIndex[i]==ActCtrlPnt){
			j=i/4*4;
			if( coordIndex[j+0]==remActCtrlPnt||
				coordIndex[j+1]==remActCtrlPnt||
				coordIndex[j+2]==remActCtrlPnt
				){
				msG(String("you have been prevented from joining 2 points together that are both from the same triangle"));
				return;
			}
		}
	}

	int from=ActCtrlPnt;
	int to=remActCtrlPnt;

	for(i=0;i<coordIndex.length;i+=4){
		if(coordIndex[i]==ActCtrlPnt){
			coordIndex[i]=remActCtrlPnt;
		}
		if(coordIndex[i+1]==ActCtrlPnt){
			coordIndex[i+1]=remActCtrlPnt;
		}
		if(coordIndex[i+2]==ActCtrlPnt){
			coordIndex[i+2]=remActCtrlPnt;
		}
	}
	deleteVertex(ActCtrlPnt);

	ActCtrlPnt=remActCtrlPnt;
	makePointerEqualIndexLocation(this);
	convertToRenderFormat();
}
*/
ArtPtr selectPartFromVertex(SeamlessPtr smls, int targetVertex){
	for(int i=0;i<smls->partList.length;i++){
		ArtPtr part=smls->partList[i];
		if( targetVertex>=part->fromCoord&&targetVertex<part->toCoord){
			return part;
		}
	}
	throw;
}
void Art::compileTransformLoop(Matrix matrix, Rotation accumRot){

	M_Vec3f& coord = this->coord.refValue();

	pivotPoint.refValue() += translation;

	if(accumRot.angle != 0 || rotation.refValue().angle != 0){
		compileOrientationInterpolator(this, accumRot, g_sceneTreeList);
	}
	accumRot = rotation.refValue() * accumRot;
	
	matrix = translationMatrix(-absolutePivotPoint) * scaleMatrix(scale) * rotationMatrix(rotation) * translationMatrix(absolutePivotPoint+translation) * matrix;

	for(int i=coord.length; --i>=0; ){
		coord[i] *= matrix;
	}

	if(translation != Vec3f()){
		compileScalePositionInterpolator(this, translationMatrix(-translation));
	}

	scale.refValue() = Vec3f(1,1,1);
	rotation.refValue() = Rotation();
	translation.refValue() = Vec3f();

	for(int k=0;k<part.length;k++){
		part.refValue()[k]->compileTransformLoop(matrix, accumRot);
	}
}

void Art::compileTransformLoopAnimCP(Matrix matrix, Rotation accumRot){

	pivotPoint.refValue() += translation;

	if(accumRot.angle != 0 || rotation.refValue().angle != 0){
		compileOrientationInterpolator(this, accumRot, g_sceneTreeList);
	}
	accumRot = rotation.refValue() * accumRot;
	
	matrix = translationMatrix(-absolutePivotPoint) * scaleMatrix(scale) * rotationMatrix(rotation) * translationMatrix(absolutePivotPoint+translation) * matrix;

	for(int i = controlPoint.length; --i>=0; ){
		NurbsPatch* patch = controlPoint[i].p;
		int idx = controlPoint[i].i;
		float w = patch->weight[idx];
		Vec3f vec = patch->tpoint[idx] / w;
		vec *= matrix;
		patch->tpoint[idx] = vec * w;
	}

	if(translation != Vec3f()){
		compileScalePositionInterpolator(this, translationMatrix(-translation));
	}

	scale.refValue() = Vec3f(1,1,1);
	rotation.refValue() = Rotation();
	translation.refValue() = Vec3f();

	for(int k=0;k<part.length;k++){
		part.refValue()[k]->compileTransformLoopAnimCP(matrix, accumRot);
	}
}

void compileTransform(){
	Seamless* smls=dynamic_cast<Seamless*>(g_selNode.getRaw());
	Art* part=dynamic_cast<Art*>(g_selNode.getRaw());
	if(smls){
		Matrix matrix= scaleMatrix(smls->scale.refValue())*rotationMatrix(smls->rotation.refValue())*translationMatrix(smls->translation.refValue());
		M_SArt& skeleton=smls->skeleton.refValue();

		if(smls->workAnimCP){
			for(int k=skeleton.length; --k>=0;){
				skeleton[k]->compileTransformLoopAnimCP(matrix, smls->rotation);
			}
		}
		else{
			for(int k=skeleton.length; --k>=0;){
				skeleton[k]->compileTransformLoop(matrix, smls->rotation);
			}
		}
		smls->scale=Vec3f(1,1,1);
		smls->rotation=Rotation();
		smls->translation=Vec3f();
	}
	else if(part){
		smls = part->smls;
		if(smls->workAnimCP){
			part->compileTransformLoopAnimCP(Matrix(), Rotation());
		}
		else{
			part->compileTransformLoop(Matrix(), Rotation());
		}
	}
	else return;

	smls->calcAbsolutePivotPoint();
	smls->catTranslationMatrix();
	smls->ctrf = true;
	smls->catMatrix();
	smls->catIMatrix();
	Window::paintAllControlPanels();
	BackUp::monostable.beginCycle();
}

void Art::compileScaleLoop(Matrix matrix, Matrix scaleRotMatrix, Rotation accumRot, M_Rotation& m_rot){
	m_rot[plidx] = rotation;

	M_Vec3f& coord = this->coord.refValue();

	translation.refValue() = scaleRotMatrix * translation;
	pivotPoint.refValue() = scaleRotMatrix * pivotPoint;
	scaleRotMatrix = scaleMatrix(scale) * rotationMatrix(rotation) * scaleRotMatrix;

	Rotation return_rot;
	if(accumRot.angle != 0 || rotation.refValue().angle != 0){

		if(rotation.angle != 0){
			Rotation accumRot2 = rotation * accumRot;
			return_rot = -accumRot2 * accumRot;
		}
	}
	accumRot = rotation.refValue() * accumRot;
	
	matrix = translationMatrix(-absolutePivotPoint) * scaleMatrix(scale) * rotationMatrix(rotation) * translationMatrix(absolutePivotPoint) * matrix;

	for(int i=coord.length; --i>=0; ){
		coord[i] *= matrix;
	}

	scale.refValue() = Vec3f(1,1,1);
	rotation.refValue() = return_rot;

	for(int k=0;k<part.length;k++){
		part.refValue()[k]->compileScaleLoop(matrix, scaleRotMatrix, accumRot, m_rot);
	}
}

void Art::compileScaleLoopAnimCP(Matrix matrix, Matrix scaleRotMatrix, Rotation accumRot, M_Rotation& m_rot){
	m_rot[plidx] = rotation;

	translation.refValue() = scaleRotMatrix * translation;
	pivotPoint.refValue() = scaleRotMatrix * pivotPoint;
	scaleRotMatrix = scaleMatrix(scale) * rotationMatrix(rotation) * scaleRotMatrix;

	Rotation return_rot;
	if(accumRot.angle != 0 || rotation.refValue().angle != 0){

		if(rotation.angle != 0){
			Rotation accumRot2 = rotation * accumRot;
			return_rot = -accumRot2 * accumRot;
		}
	}
	accumRot = rotation.refValue() * accumRot;
	
	matrix = translationMatrix(-absolutePivotPoint) * scaleMatrix(scale) * rotationMatrix(rotation) * translationMatrix(absolutePivotPoint) * matrix;

	for(int i = controlPoint.length; --i>=0; ){
		NurbsPatch* patch = controlPoint[i].p;
		int idx = controlPoint[i].i;
		float w = patch->weight[idx];
		Vec3f vec = patch->tpoint[idx] / w;
		vec *= matrix;
		patch->tpoint[idx] = vec * w;
	}

	scale.refValue() = Vec3f(1,1,1);
	rotation.refValue() = return_rot;

	for(int k=0;k<part.length;k++){
		part.refValue()[k]->compileScaleLoopAnimCP(matrix, scaleRotMatrix, accumRot, m_rot);
	}
}

void Art::compileScaleRotBackLoop(Matrix matrix, Matrix rotMatrix, Rotation accumRot, M_Rotation& m_rot){
	M_Vec3f& coord = this->coord.refValue();

	translation.refValue() = rotMatrix * translation;
	pivotPoint.refValue() = rotMatrix * pivotPoint;
	rotMatrix = scaleMatrix(scale) * rotationMatrix(rotation) * rotMatrix;

	accumRot = rotation.refValue() * accumRot;
	
	matrix = translationMatrix(-absolutePivotPoint) * scaleMatrix(scale) * rotationMatrix(rotation) * translationMatrix(absolutePivotPoint) * matrix;

	for(int i=coord.length; --i>=0; ){
		coord[i] *= matrix;
	}

	scale.refValue() = Vec3f(1,1,1);
	rotation.refValue() = m_rot[plidx];

	for(int k=0;k<part.length;k++){
		part.refValue()[k]->compileScaleRotBackLoop(matrix, rotMatrix, accumRot, m_rot);
	}
}

void Art::compileScaleRotBackLoopAnimCP(Matrix matrix, Matrix rotMatrix, Rotation accumRot, M_Rotation& m_rot){

	translation.refValue() = rotMatrix * translation;
	pivotPoint.refValue() = rotMatrix * pivotPoint;
	rotMatrix = scaleMatrix(scale) * rotationMatrix(rotation) * rotMatrix;

	accumRot = rotation.refValue() * accumRot;
	
	matrix = translationMatrix(-absolutePivotPoint) * scaleMatrix(scale) * rotationMatrix(rotation) * translationMatrix(absolutePivotPoint) * matrix;

	for(int i = controlPoint.length; --i>=0; ){
		NurbsPatch* patch = controlPoint[i].p;
		int idx = controlPoint[i].i;
		float w = patch->weight[idx];
		Vec3f vec = patch->tpoint[idx] / w;
		vec *= matrix;
		patch->tpoint[idx] = vec * w;
	}

	scale.refValue() = Vec3f(1,1,1);
	rotation.refValue() = m_rot[plidx];

	for(int k=0;k<part.length;k++){
		part.refValue()[k]->compileScaleRotBackLoopAnimCP(matrix, rotMatrix, accumRot, m_rot);
	}
}

/*
void Art::compileScaleLoop(Matrix matrix, Matrix scaleRotMatrix, Matrix unRotMatrix){
	Matrix tmpMtx=matrix*unRotMatrix;
	translation.refValue()*=tmpMtx;

	M_Vec3f& coord=this->coord.refValue();
	matrix= scaleMatrix(scale)*rotationMatrix(rotation)*matrix;
	unRotMatrix=unRotMatrix*rotationMatrix(-rotation);
	tmpMtx=matrix*unRotMatrix;

	for(int i=coord.length; --i>=0; ){
		coord[i]*=tmpMtx;
	}
	scale=Vec3f(1,1,1);
//	compileScalePositionInterpolator(this, tmpMtx);

	M_SArt& children=part.refValue();
	for(int k=0; k<children.length; k++){
		children[k]->compileScaleLoop(matrix, scaleRotMatrix, unRotMatrix);
	}
}
*/

void compileScale(){
	Seamless* smls=dynamic_cast<Seamless*>(g_selNode.getRaw());
	Art* part=dynamic_cast<Art*>(g_selNode.getRaw());
	M_Rotation m_rot;

	if(smls){
		m_rot.length = smls->partList.length;

		Vec3f sc=smls->scale;
		Matrix matrix= scaleMatrix(sc)*rotationMatrix(smls->rotation);
		Matrix unRotMatrix= rotationMatrix(-smls->rotation);

		smls->scale=Vec3f(1,1,1);

		M_SArt& skeleton=smls->skeleton.refValue();
		if(smls->workAnimCP){
			for(int k=skeleton.length; --k>=0;){
				skeleton[k]->compileScaleLoopAnimCP(matrix, matrix, smls->rotation, m_rot);
				smls->calcAbsolutePivotPoint();
				skeleton[k]->compileScaleRotBackLoopAnimCP(unRotMatrix, unRotMatrix, -smls->rotation, m_rot);
			}
		}
		else{
			for(int k=skeleton.length; --k>=0;){
				skeleton[k]->compileScaleLoop(matrix, matrix, smls->rotation, m_rot);
				smls->calcAbsolutePivotPoint();
				skeleton[k]->compileScaleRotBackLoop(unRotMatrix, unRotMatrix, -smls->rotation, m_rot);
			}
		}
	}
	else if(part){
		smls = part->smls;
		m_rot.length = smls->partList.length;
		if(smls->workAnimCP){
			part->compileScaleLoopAnimCP(Matrix(), Matrix(), Rotation(), m_rot);
			part->smls->calcAbsolutePivotPoint();
			part->compileScaleRotBackLoopAnimCP(Matrix(), Matrix(), Rotation(), m_rot);
		}
		else{
			part->compileScaleLoop(Matrix(), Matrix(), Rotation(), m_rot);
			part->smls->calcAbsolutePivotPoint();
			part->compileScaleRotBackLoop(Matrix(), Matrix(), Rotation(), m_rot);
		}
	}
	else return;
	smls->calcAbsolutePivotPoint();
	smls->catTranslationMatrix();
	smls->ctrf = true;
	smls->catMatrix();
	smls->catIMatrix();
	Window::paintAllControlPanels();
	BackUp::monostable.beginCycle();
}
/*
void compileScaleRotationLoop(Art* part, Matrix matrix, Rotation accumRot){
	part->translation.refValue()*=matrix;

	M_Vec3f &coord=part->coord.refValue();
	matrix= scaleMatrix(part->scale)*rotationMatrix(part->rotation)*matrix;
	accumRot=part->rotation.refValue()*accumRot;
	//matrix= scaleMatrix(part->scale)*rotationMatrix(accumRot);
	

	for(int i=coord.length; --i>=0; ){
		coord[i]*=matrix;
	}
//	Vec3f v = accumRot*Vec3f(1,0,0);
	//v = Rotation(0,0,1,PI);
//	Rotation r = Rotation (Vec3f(0,0,0), v);
	compileOrientationInterpolator(part, accumRot);

	part->scale=Vec3f(1,1,1);
	part->rotation=Rotation();

	M_SArt& children=part->part.refValue();
	for(int k=0; k<children.length; k++){
		compileScaleRotationLoop(children[k], matrix, accumRot);
	}
}
*/
void Art::compileScaleRotationLoop(Matrix matrix, Matrix scaleRotMatrix, Rotation accumRot, M_<SceneTreeListItem>& sceneTreeList){

	M_Vec3f& coord = this->coord.refValue();

	translation.refValue() = scaleRotMatrix * translation;
	pivotPoint.refValue() = scaleRotMatrix * pivotPoint;
	scaleRotMatrix = scaleMatrix(scale) * rotationMatrix(rotation) * scaleRotMatrix;

	if(accumRot.angle != 0 || rotation.refValue().angle != 0){
		compileOrientationInterpolator(this, accumRot, sceneTreeList);
	}
	accumRot = rotation.refValue() * accumRot;
	
	matrix = translationMatrix(-absolutePivotPoint) * scaleMatrix(scale) * rotationMatrix(rotation) * translationMatrix(absolutePivotPoint) * matrix;

	for(int i=coord.length; --i>=0; ){
		coord[i] *= matrix;
	}

	scale.refValue() = Vec3f(1,1,1);
	rotation.refValue() = Rotation();

	for(int k=0;k<part.length;k++){
		part.refValue()[k]->compileScaleRotationLoop(matrix, scaleRotMatrix, accumRot, sceneTreeList);
	}
}

void Art::compileScaleRotationLoopAnimCP(Matrix matrix, Matrix scaleRotMatrix, Rotation accumRot, M_<SceneTreeListItem>& sceneTreeList){

	translation.refValue() = scaleRotMatrix * translation;
	pivotPoint.refValue() = scaleRotMatrix * pivotPoint;
	scaleRotMatrix = scaleMatrix(scale) * rotationMatrix(rotation) * scaleRotMatrix;

	if(accumRot.angle != 0 || rotation.refValue().angle != 0){
		compileOrientationInterpolator(this, accumRot, sceneTreeList);
	}
	accumRot = rotation.refValue() * accumRot;
	
	matrix = translationMatrix(-absolutePivotPoint) * scaleMatrix(scale) * rotationMatrix(rotation) * translationMatrix(absolutePivotPoint) * matrix;

	for(int i = controlPoint.length; --i>=0; ){
		NurbsPatch* patch = controlPoint[i].p;
		int idx = controlPoint[i].i;
		float w = patch->weight[idx];
		Vec3f vec = patch->tpoint[idx] / w;
		vec *= matrix;
		patch->tpoint[idx] = vec * w;
	}

	scale.refValue() = Vec3f(1,1,1);
	rotation.refValue() = Rotation();

	for(int k=0;k<part.length;k++){
		part.refValue()[k]->compileScaleRotationLoopAnimCP(matrix, scaleRotMatrix, accumRot, sceneTreeList);
	}
}

void compileScaleRotation(Seamless* smls, M_<SceneTreeListItem>& sceneTreeList){
	Vec3f sc=smls->scale;
	Matrix matrix= scaleMatrix(sc)*rotationMatrix(smls->rotation);

	smls->scale=Vec3f(1,1,1);
	smls->rotation=Rotation();

	M_SArt& skeleton=smls->skeleton.refValue();
	if(smls->workAnimCP){
		for(int k = skeleton.length; --k>=0;){
			skeleton[k]->compileScaleRotationLoopAnimCP(matrix, matrix, Rotation(), sceneTreeList);
		}
	}
	else{
		for(int k = skeleton.length; --k>=0;){
			skeleton[k]->compileScaleRotationLoop(matrix, matrix, Rotation(), sceneTreeList);
		}
	}
}


void compileScaleRotation(){
	Seamless* smls=dynamic_cast<Seamless*>(g_selNode.getRaw());
	Art* part=dynamic_cast<Art*>(g_selNode.getRaw());
	if(smls){
		compileScaleRotation(smls, g_sceneTreeList);
	}
	else if(part){
		smls = part->smls;
		if(smls->workAnimCP){
			part->compileScaleRotationLoopAnimCP(Matrix(), Matrix(), Rotation(), g_sceneTreeList);
		}
		else{
			part->compileScaleRotationLoop(Matrix(), Matrix(), Rotation(), g_sceneTreeList);
		}
	}
	else return;

	smls->calcAbsolutePivotPoint();
	smls->catTranslationMatrix();
	smls->ctrf = true;
	smls->catMatrix();
	smls->catIMatrix();
	Window::paintAllControlPanels();
	BackUp::monostable.beginCycle();
}


void listRoutes(String&);
bool writeSmlsFile( String filename){
	g_scene->preSaveProcessCollapse();
	Window::preSaveProcessWin(*g_scene);

	changeFloatFormat(g_formatCoords=g_settings->roundCoordsSave.getValue());
	g_scene->version=g_version;

	static String str;
	str="#VRML V2.0 utf8\n";
	g_scene->toString(str);

	listRoutes(str);


	if(str.write(filename)==false){
		msG("\
*ERROR*\nFailed to save file!!!!\n\
This can happen if the file has been opened from a computer on the network and the computer\n\
has been turned off or if there is a file write protected.\n\
Try saving the file with a new name in a folder that exists\n\
or try deleting all files in your Recover folder which is located in:\n\
My Documents\\Seamless3d\\Application Data For Seamless3d");
		return false;
	}
	return true;
}
void commonEndCCVisualClue(){
	g_vcSeamless->process();
	g_vcSeamless->doBuild();
	g_vcSeamless->convertToRenderFormat();
	g_visualClue=g_vcSeamless;
}

void fillPolyLineIndex(M_Vec3f& point, M_Int& coordIndex){
	coordIndex.length=point.length;
	for(int i=0; i<point.length; i++){
		coordIndex[i]=i;
	}
}
void fillPolyLineIndex(M_Vec3f& point, int markPoint, M_Int& coordIndex){
	//coordIndex+=-1;
	int mark = coordIndex.length;
	int pointsToAdd = point.length - markPoint;
	coordIndex.length+=pointsToAdd;
	for(int i = 0; i < pointsToAdd; i++){
		coordIndex[mark + i] = markPoint + i;
	}
}

// draw a circle starting at 'startAng' angle and rotate through 'deltaAng' radians 
void circle(M_Vec3f &coord, Vec2f centerPos, double radius, double startAng, double deltaAng)
{
   int numPoints = max(2, 100 * (fabs(deltaAng) / (2 * PI)));
   double delta = deltaAng / numPoints;
   double angle = startAng;
   for ( int i = 0; i <= numPoints; i++ )
   {     
      coord += Vec3f(radius * cos(angle) + centerPos.x, radius * sin(angle) + centerPos.y, 0);
      angle += delta;
   }
}

// length of angle control point stick
const float CCLathe::controlLen = 0.2;

void CCLathe::calcSegments()
{
   seg.length = 0;

   if ( curve.length == 0 )
   {
      return;
   }

   Vec2f P = Vec2f(baseRad, 0);
   double a = curve[0];

   seg += CCSegment(Vec2f(0, 0), P, P, a);

   for ( int i = 1; i < curve.length; i++ ){
      double r = (curve[i] == 0) ? distance[i - 1] : distance[i - 1] / curve[i];
      Vec2f Ro = Vec2f(r, 0).rotate(a);
      Vec2f C  = P - Ro;

      a += curve[i];
      Vec2f Rf = Vec2f(r, 0).rotate(a);        
      P += Rf - Ro;

      seg += CCSegment(C, Ro, P, curve[i]);
   }
}

void CCLathe::setPoint(int idx, Vec2f point)
{

   setMathHiPresision();

   if ( idx == 0 ) {
      baseRad = max(0, point.x);
   }
   else if ( idx < curve.length ) {
      Vec2f Po = Vec2f(baseRad, 0);  // point at start of arc
      double ao = curve[0];

      for ( int i = 1; i < idx; i++ ){
         double r = (curve[i] == 0) ? distance[i - 1] : distance[i - 1] / curve[i];

         Vec2f Ro = Vec2f(r, 0).rotate(ao);
         ao += curve[i];
         Vec2f Rf = Vec2f(r, 0).rotate(ao);  
         Po += Rf - Ro;
      }

      float zoom = g_scene->getZoom();
      bool shift = (0x8000 & GetKeyState(VK_LSHIFT)) != 0;
      double ed = shift ? 0.03 / zoom : 1e-5;
      double ea = shift ? 0.03 : 1e-5;

      // normalize so P1 is at the origin and the start tangent is along the x-axis
      Vec2f Pf = (point - Po).rotate(-(ao + PI / 2));

      // calc angle and radius of arc
      double a = curve[idx];
      double d = distance[idx - 1];

      if ( abs(Pf.y) < ed && Pf.x > 1e-2 && abs(a) < ea) {
         a = 0.0001;
         d = fabs(Pf.x); 
      }
      else if ( abs(Pf.y) > 1e-6 ) {
         double r = (Pf.y + Pf.x * Pf.x / Pf.y) / 2;
         Vec2f C  = Vec2f(0, r);
         Vec2f T2 = (Pf - C).rotate(PI / 2);
         a = T2.angle();

         if ( C.y < 0 ){
            a -= PI;
         }
         else if ( Pf.x < 0 ){
            a += 2 * PI;
         }

         d = abs(a * r); 
      }
      else {
         restoreMathPresision();
         return;  // limit of precision - no work around
      }

      curve.refValue()[idx] = a;
      distance.refValue()[idx - 1] = d; 
   }

   restoreMathPresision();
}

void CCLathe::houseKeep()
{

   setMathHiPresision();
   calcSegments();

   float inverseSize = (1/g_scene->getZoom());
   int idx = index[0]; 
   IndexedLineSet* polyline = dynamic_cast<IndexedLineSet*>(g_vcEllipticalPolyline2D->geometry.p);
   M_Vec3f& coord=dynamic_cast<Coordinate*>(polyline->coord.p)->point;
   M_Node& children = g_vcCCLatheBase->children;
   children.length = 1; // IndexedLineSet
   coord.length = 0;

   // draw the cc control curve visual clue 
   float crossLen = controlLen * inverseSize * 0.2;
   float senseLen = MouseSense * inverseSize * 1.75;

   coord += Vec3f(0, 0, 0);
   coord += Vec3f(0, -crossLen, 0);
   coord += Vec3f(0, +crossLen, 0);
   coord += Vec3f(0, 0, 0);
   coord += Vec3f(-crossLen, 0, 0);
   coord += Vec3f(+crossLen, 0, 0);
   circle(coord, Vec2f(0, 0), senseLen, PI, 2 * PI);

   coord += Vec3f(baseRad, 0, 0);
   coord += Vec3f(baseRad, -crossLen / 2, 0);
   coord += Vec3f(baseRad, +crossLen / 2, 0);
   coord += Vec3f(baseRad, 0, 0);

   for ( int i = 0; i < seg.length; i++ )
   {
      if ( i == 0 ) {
         if ( i == idx ) {
            anglePt = Vec2f(controlLen * inverseSize, 0).rotate(seg[i].a );
            anglePt += seg[i].Pf;
            coord += Vec3f( seg[i].Pf.x, seg[i].Pf.y, 0);
            coord += Vec3f( anglePt.x, anglePt.y, 0); 
            coord += Vec3f( seg[i].Pf.x, seg[i].Pf.y, 0);
         }
      }
      else {
         circle(coord, seg[i].C, seg[i].Ro.length(), seg[i].Ro.angle(), seg[i].a);
         if ( i == idx ){
            anglePt = seg[i].Pf - seg[i].C;
            double a = anglePt.angle() / PI * 4;
            anglePt *= (controlLen * inverseSize / max(1e-9, anglePt.length()));
            anglePt += seg[i].Pf;

            coord += Vec3f( seg[i].C.x, seg[i].C.y, 0);
            if ( fabs(a - floor(a + 0.5)) < 0.0001 ) {
               Vec2f T = (seg[i].Pf - seg[i].C).rotate(PI/2);
               T *= senseLen * 2 / max(1e-5, T.length());
               coord += Vec3f( seg[i].C.x + T.x, seg[i].C.y + T.y, 0);
               coord += Vec3f( seg[i].C.x - T.x, seg[i].C.y - T.y, 0);
               coord += Vec3f( seg[i].C.x, seg[i].C.y, 0);
            }
            circle(coord, seg[i].C, senseLen, seg[i].Ro.angle() + seg[i].a, 2 * PI);

            coord += Vec3f( anglePt.x, anglePt.y, 0);
            coord += Vec3f( seg[i].Pf.x, seg[i].Pf.y, 0);
         }
      }
   }
   fillPolyLineIndex(coord, polyline->coordIndex.refValue());

   // setup the control points
   if ( seg.length > 0 ){
      int childDiff = (seg.length + 1) - children.length + 1;
      if(childDiff<0){
         children.length+=childDiff;
      }
      else{
         for(int i=childDiff; --i>=0;){
            Transform* transform=new Transform;
            transform->children.refValue()+=g_nextPoint.p; //could just add null?
            children+=transform;
         }
      }

      for(int i=seg.length; --i>=0;){
         Transform* transform=dynamic_cast<Transform*>(children[i+1].p);
         transform->translation = Vec3f(seg[i].Pf.x, seg[i].Pf.y, 0);

         if(i==idx){
            transform->children.refValue()[0]=g_redPoint;
         }
         else{
            if( i == 0 || (i == seg.length - 1 && autoClose) ){
               transform->children.refValue()[0]=g_cyanPoint;
            }
            else{
               transform->children.refValue()[0]=g_nextPoint;
            }
         }
         transform->scale=Vec3f( inverseSize, inverseSize, inverseSize);
      }

      // angle-drag control point
      Transform* transform=dynamic_cast<Transform*>(children[seg.length + 1].p);
      transform->translation = Vec3f(anglePt.x, anglePt.y, 0);
      if ( angleDrag ) {
         transform->children.refValue()[0]=g_redPoint;
      }
      else{
         transform->children.refValue()[0]=g_nextPoint;
      }
      transform->scale=Vec3f( inverseSize, inverseSize, inverseSize);
   }

   g_visualClue = g_vcCCLatheBase;

   restoreMathPresision();
//   g_vcEllipticalMaterial->diffuseColor=g_settings->backgroundColor.refValue().invert();
   polyline->convertToRenderFormat(g_vcEllipticalPolyline2D->render);
}

void CCScale::houseKeep(){

	if(!distance.length){
		return;
	}
	int index=this->index[0];
	if(index==0){
		index=1;
	}

	g_vcCCLathe->distance.refValue()[0]=distance[index-1];
	if(index<=1){
		g_vcSurfaceTransform->translation=Vec3f();
	}
	else {
		g_vcSurfaceTransform->translation=Vec3f(0,getDistance(index-2),0);
	}

	g_vcCCBend->distance.resize(distance.length);
	g_vcCCBend->yAngle.resize(curve.length);
	g_vcCCBend->curve.resize(curve.length);
	g_vcCCBend->distance.refValue()=distance;
	g_vcCCBend->curve.refValue()=curve;

	g_vcCCLathe->baseRad.refValue()=g_spaghettiRad*(1/g_scene->getZoom());

	g_vcBaseRad->translation=Vec3f(baseRad+location.x,location.y,location.z);
	g_vcBaseRad->rotation=orientation;
	commonEndCCVisualClue();
}
void rotateZ(Vec2f *p,int numberOf,float angle);
void genSegment(	long double  pt1x,long double  pt1y,
		long double  ang1z,
		long double  pt2x,long double  pt2y,
		long double  ang2z,
		long double a, long double b,
		Vec3f* v,int step){

	long double x,y,xyt,t,j;

	x=pt2x-pt1x;
	y=pt2y-pt1y;

	rotateZ(x, y, ang1z);

	j=y/b;
	if(j>1)j=1;
	xyt=asinl(j);

	Matrix3 matrix=rotationMatrix(ang1z-PI*.5)*translationMatrix(Vec2f(pt1x,pt1y));
	for(int i=0;i<step;i++){
		t=xyt/step*i;
		Vec2f vec2f=matrix*Vec2f(a*cosl(t)-a, b*sinl(t));
		v[i]=Vec3f(vec2f.x,vec2f.y,0);
	}
}


void EC2::genCurve(M_Vec3f& line){

	int step=32;
	int pointLen=point.length;
	line.length=(pointLen-1)*step+1;
	int offset=0;
	int i=0;
	for(; i<pointLen-1; i++){
		genSegment(
			point[i].x,point[i].y,tangent[i],
			point[i+1].x,point[i+1].y,tangent[i+1],
			a[i], b[i],
			&line[offset],step
		);
		offset+=step;
	}
	line[offset]=Vec3f(point[i].x,point[i].y,0);
}

void ECBend::genCurve(M_Vec3f& line){

	int step=32;
	int pointLen=point.length;
	line.length=(pointLen-1)*step+1;
	int offset=0;
	int i=0;
	for(; i<pointLen-1; i++){
		genSegment(
			point[i].x,point[i].y,tangent[i],
			point[i+1].x,point[i+1].y,tangent[i+1],
			a[i], b[i],
			&line[offset],step
		);
		offset+=step;
	}
	line[offset]=Vec3f(point[i].x,point[i].y,0);
}
void BezierPatch::houseKeep(){
	if(!point.length){
		g_visualClue=0;
		return;
	}
	setMathHiPresision();

	if(g_visualClue==g_vcSelectRectGroup){
		IndexedLineSet* polyLine=dynamic_cast<IndexedLineSet*>(g_vcSelectRectShape->geometry.p);
		M_Vec3f& coord=dynamic_cast<Coordinate*>(polyLine->coord.p)->point;

		coord.length=0;
		coord+=g_scene->getIViewMatrix() * Vec3f(MultiSelect::topLeft.x, MultiSelect::bottomRight.y, MultiSelect::topLeft.z);
		coord+=g_scene->getIViewMatrix() * MultiSelect::bottomRight;
		coord+=g_scene->getIViewMatrix() * Vec3f(MultiSelect::bottomRight.x, MultiSelect::topLeft.y, MultiSelect::topLeft.z);
		coord+=g_scene->getIViewMatrix() * MultiSelect::topLeft;
		fillPolyLineIndex(coord, polyLine->coordIndex.refValue());
		polyLine->coordIndex.refValue()+=0;
		polyLine->convertToRenderFormat(g_vcSelectRectShape->render);
	}
	else{
		float inverseSize=(1/g_scene->getZoom());
		int idx=index[0];

		IndexedLineSet* polyLine=dynamic_cast<IndexedLineSet*>(g_vcEllipticalPolyline2D->geometry.p);

		M_Vec3f& coord=dynamic_cast<Coordinate*>(polyLine->coord.p)->point;

		coord = point;

		M_Int& coordIndex = polyLine->coordIndex.refValue();
		coordIndex.length=0;

		int numOfUGroupIdx=index[numOfUGroup];
		int bp=0;

		for(int k=0; k < dim.refValue().length; k++){
			int numOfU=this->dim.refValue()[k].x;
			int numOfV=this->dim.refValue()[k].y;

			for(int v=0; v<numOfV; v++){
				for(int u=0; u<numOfU; u++){
					coordIndex+=bp+v*numOfU+u;
				}
				coordIndex+=-1;
			}
			for(int u=0; u<numOfU; u++){
				for(int v=0; v<numOfV; v++){
					coordIndex+=bp+v*numOfU+u;
				}
				coordIndex+=-1;
			}
			bp+=numOfU*numOfV;
		}
	/////////////
		bp=getBasePoint(numOfUGroupIdx);
		{
			int numOfU=this->dim.refValue()[numOfUGroupIdx].x;
			int numOfV=this->dim.refValue()[numOfUGroupIdx].y;
			int pointLen=numOfU*numOfV;

			M_Node& children=g_vcBezierBase->children;
			children.length=1;
			int childDiff=pointLen-children.length+1;

			if(childDiff<0){
				children.length+=childDiff;
			}
			else{
				for(int i=childDiff; --i>=0;){
					Transform* transform=new Transform;
					transform->children.refValue()+=g_nextPoint.p; //could just add null?
					children+=transform;
				}
			}
			for(int i=pointLen; --i>=0;){
				Transform* transform=dynamic_cast<Transform*>(children[i+1].p);
				Vec3f pt=point[bp+i];
				transform->translation=pt;

				if(bp+i==idx){
					transform->children.refValue()[0]=g_redPoint;
				}
				else{
					transform->children.refValue()[0]=g_nextPoint;
				}

				transform->scale=Vec3f( inverseSize, inverseSize, inverseSize);
			}
		}

	///////////
		g_vcBezierBase->rotation=Rotation();

		g_visualClue=g_vcBezierBase;
	//	g_scene->children.refValue()+=g_vcBezierBase.p;
	//	g_stw->genNodeList();


		//g_vcEllipticalMaterial->diffuseColor=g_settings->backgroundColor.refValue().invert();
		polyLine->convertToRenderFormat(g_vcEllipticalPolyline2D->render);
	}
	restoreMathPresision();
}


void houseKeep2dNurbs(const M_Vec2f& dpoint, const M_Vec2f& point, const M_Float& weight, const M_Float& knot, const int idx, float minT, float maxT, const int degree, float scale, Matrix matrix=Matrix()){
	static M_Vec2f line;
	if(!point.length){
		g_visualClue=0;
		return;
	}
	setMathHiPresision();
	const float inverseSize=(1/g_scene->getZoom());

	const int dim=point.length;

	const int steps = (dim == 2)? 0: 1000;

	line.length=steps;


	
	const Vec2f* p = point.getArray();
	const float* w=weight.getArray();


	for(int i=0; i<steps; i++){
		float t = static_cast<float>(i)/(steps+1);
		t = (maxT-minT)*t + minT;
		line[i]=curvePoint2d(dim-1, degree, knot.getArray(), p, w, t);
	}

	IndexedLineSet* polyLine=dynamic_cast<IndexedLineSet*>(g_vcEllipticalPolyline2D->geometry.p);

	M_Vec3f& coord=dynamic_cast<Coordinate*>(polyLine->coord.p)->point;
	coord.length=line.length+point.length;//normA.length*5;
	for(int i=0; i<line.length; i++){
		coord[i]=matrix*Vec3f(line[i].x, line[i].y*scale, 0);
	}
	for(int j=line.length, i=point.length; --i>=0; j++){
		coord[j]=matrix*Vec3f(dpoint[i].x, dpoint[i].y*scale, 0);
	}
	fillPolyLineIndex(coord, polyLine->coordIndex.refValue());
/////////////
	M_Node& children=g_vcBezierBase->children;
	children.length=1;
	int childDiff=point.length-children.length+1;

	if(childDiff<0){
		children.length+=childDiff;
	}
	else{
		for(int i=childDiff; --i>=0;){
			Transform* transform=new Transform;
			transform->children.refValue()+=g_nextPoint.p; //could just add null?
			children+=transform;
		}
	}
	for(int i=point.length; --i>=0;){
		Transform* transform=dynamic_cast<Transform*>(children[i+1].p);
		Vec3f pt = matrix * Vec3f(dpoint[i].x, dpoint[i].y*scale, 0);
		transform->translation = pt;

		if(i==idx){
			transform->children.refValue()[0]=g_redPoint;
		}
		else{
			transform->children.refValue()[0]=g_nextPoint;
		}

		transform->scale=Vec3f( inverseSize, inverseSize, inverseSize);
	}

///////////
	g_vcBezierBase->rotation=Rotation();

	g_visualClue=g_vcBezierBase;

	restoreMathPresision();
	//g_vcEllipticalMaterial->diffuseColor=g_settings->backgroundColor.refValue().invert();
	polyLine->convertToRenderFormat(g_vcEllipticalPolyline2D->render);
}
void BendyLathe::dragRadius(Vec3f pt, int idx){
	Vec3f loc = ringBase[idx];
	Rotation orien = ringRot[idx];

	double a = PI/ctrlStaves;
	double b = a*(radIdx*2+1);
	Rotation polar = Rotation(0,1,0,-b);

	Matrix imatrix=translationMatrix(Vec3f(-translation.x, -translation.y, -translation.z)) * rotationMatrix( -rotation) * scaleMatrix(Vec3f(1./scale.x, 1./scale.y, 1./scale.z)) *
		translationMatrix(-loc) * rotationMatrix(-orien) * rotationMatrix(polar);

	Vec3f newPt = g_scene->getIViewMatrix() * imatrix * pt;

//	double distance = abs(sqrt(newPt.x*newPt.x + newPt.y*newPt.y + newPt.z*newPt.z));

//	radius.refValue()[idx] = distance/ratio;
	radius.refValue()[idx] = -newPt.z/ratio;


	Vec3f sr = matrix * g_scene->getViewMatrix() * (ringRot[idx] * Vec3f(-sin(b)*radius[idx]*ratio, 0, -cos(b)*radius[idx]*ratio) + loc);
	g_pointer.translation=sr;
}
void BendyLathe::dragRingBase(Vec3f pt, int idx, M_Vec3f& point, const M_Float& weight, const M_Float& knot, float minT, float maxT, Vec3f& atVec){
	M_Float& at = this->at.refValue();

	const int dim=point.length;
	Matrix windMatrix = matrix * g_viewMatrix;
	const Vec3f* p = point.getArray();
	const float* w=weight.getArray();
	const float* knotPtr = knot.getArray();

	float t = (maxT-minT)*at[idx] + minT;
	double tp = (maxT-minT)*(at[idx]+.0001) + minT;

	Vec3f vb;
	Vec3f vc;

	CurvePoint(dim-1, degree, knotPtr, p, w, t, vb);
	CurvePoint(dim-1, degree, knotPtr, p, w, tp, vc);
	vb*=windMatrix;
	vc*=windMatrix;
	double ang = Vec2f(vc.x-vb.x, vc.y-vb.y).getAngle();

	Vec2f dragDiff = Vec2f(pt.x - vb.x, pt.y - vb.y);

	Vec3f dragDiff3f = Rotation(0,0,1,-ang)*Vec3f(dragDiff.x, dragDiff.y, 0);

	at[idx] += dragDiff3f.x*.1;
	if(at[idx] < 0){
		at[idx] = 0;
	}
	if(at[idx] > 1){
		at[idx] = 1;
	}

	CurvePoint(dim-1, degree, knotPtr, p, w, t, atVec);

	g_pointer.translation = atVec = windMatrix * atVec;
}
void BendyLathe::houseKeep3dNurbs(const M_Vec3f& dpoint, const M_Vec3f& point, const M_Float& weight, const M_Float& knot, const int idx, float minT, float maxT, Matrix matrix, NurbsPatch* patch, int selRadius){
	float scale = 1;
	static M_Vec3f line;
	if(!point.length){
		g_visualClue=0;
		return;
	}
	setMathHiPresision();
	const float inverseSize=(1/g_scene->getZoom());

	const int dim=point.length;

	const int steps = (dim == 2)? 0: 1000;

	line.length=steps;

	const Vec3f* p = point.getArray();
	const float* w=weight.getArray();
	Vec3f* linePtr = line.getArray();
	const float* knotPtr = knot.getArray();

	for(int i=0; i<steps; i++){
		float t = static_cast<float>(i)/(steps+1);
		t = (maxT-minT)*t + minT;
		CurvePoint(dim-1, degree, knotPtr, p, w, t, *(linePtr+i));
	}

	IndexedLineSet* polyLine=dynamic_cast<IndexedLineSet*>(g_vcEllipticalPolyline2D->geometry.p);

	M_Vec3f& coord=dynamic_cast<Coordinate*>(polyLine->coord.p)->point;
	coord.length=line.length+point.length;//normA.length*5;
	for(int i=0; i<line.length; i++){
		coord[i]=matrix*Vec3f(line[i].x, line[i].y*scale, line[i].z);
	}
	for(int j=line.length, i=point.length; --i>=0; j++){
		coord[j]=matrix*Vec3f(dpoint[i].x, dpoint[i].y*scale, dpoint[i].z);
	}
	M_Int& coordIndex = polyLine->coordIndex.refValue();
	fillPolyLineIndex(coord, coordIndex);

/////////////
	M_Node& children=g_vcBezierBase->children;
	children.length=1;
	if(showRingControlPoints==false){
		int childDiff=point.length-children.length+1;
		if(childDiff<0){
			children.length+=childDiff;
		}
		else{
			for(int i=childDiff; --i>=0;){
				Transform* transform=new Transform;
				transform->children.refValue()+=g_nextPoint.p; //could just add null?
				children+=transform;
			}
		}
		for(int i=point.length; --i>=0;){
			Transform* transform=dynamic_cast<Transform*>(children[i+1].p);
			Vec3f pt=matrix*Vec3f(dpoint[i].x, dpoint[i].y*scale, dpoint[i].z);
			transform->translation=pt;

			if(i==idx){
				transform->children.refValue()[0]=g_redPoint;
			}
			else{
				transform->children.refValue()[0]=g_nextPoint;
			}

			transform->scale=Vec3f( inverseSize, inverseSize, inverseSize);
		}
	}
	M_Vec3f& dpoint = patch->dpoint;
	int numOfU = patch->dim.x;
	int numOfV = patch->dim.y;
	int offset = coord.length;
	Vec3f selPoint = ringBase[selRadius];
	if(showRingControlPoints){
		int childDiff=((numOfU-1)>>1)-children.length+2;
		if(childDiff<0){
			children.length+=childDiff;
		}
		else{
			for(int i=childDiff; --i>=0;){
				Transform* transform=new Transform;
				transform->children.refValue()+=g_nextPoint.p; //could just add null?
				children+=transform;
			}
		}
		Transform* transform=dynamic_cast<Transform*>(children[1].p);
		transform->children.refValue()=M_Node(g_yellowPoint.p);
		Vec3f pt=matrix*Vec3f(selPoint.x, selPoint.y*scale, selPoint.z);
		transform->translation=pt;
		transform->scale=Vec3f(inverseSize, inverseSize, inverseSize);
	}

	coordIndex+=-1;

	for(int v=0; v<numOfV; v++){
		for(int u=0; u<numOfU; u++){
			Vec3f pt=dpoint[v*numOfU+u];
			if(v == selRadius && (u & 1) && showRingControlPoints){
				Transform* transform=dynamic_cast<Transform*>(children[2+(u>>1)].p);
				transform->children.refValue()=M_Node(g_redPoint.p);
				transform->translation=pt;
				transform->scale=Vec3f(inverseSize, inverseSize, inverseSize);
			}
			coord+=pt;
			coordIndex+=offset+v*numOfU+u;
		}
		coordIndex+=-1;
	}

///////////
	g_vcBezierBase->rotation=Rotation();

	g_visualClue=g_vcBezierBase;

	restoreMathPresision();
	//g_vcEllipticalMaterial->diffuseColor=g_settings->backgroundColor.refValue().invert();
	polyLine->convertToRenderFormat(g_vcEllipticalPolyline2D->render);
}

void Oscillator::houseKeep(){
	houseKeep2dNurbs(fdpoint, note, fWeight, fKnot, index[0], minT(), maxT(), fDegree, .1);
}

void Envelope::houseKeep(){
	houseKeep2dNurbs(point, point, weight, knot, index[0], minT(), maxT(), degree, .1);

	IndexedLineSet* polyLine=dynamic_cast<IndexedLineSet*>(g_vcEllipticalPolyline2D->geometry.p);

	M_Vec3f& ilsCoord = dynamic_cast<Coordinate*>(polyLine->coord.p)->point;
	M_Int& coordIndex = polyLine->coordIndex.refValue();

	int ci = ilsCoord.length;
	for(int i = 0; i < coord.length; i++){
		ilsCoord+=Vec3f(coord[i].x, 0, 0);
		ilsCoord+=Vec3f(coord[i].x, coord[i].y * .1, 0);
		coordIndex+=ci;
		ci++;
		coordIndex+=ci;
		ci++;
		coordIndex+=-1;
	}
	polyLine->convertToRenderFormat(g_vcEllipticalPolyline2D->render);
}

void AmplitudeModulator::houseKeep(){
	houseKeep2dNurbs(wet, wet, weight, knot, index[0], minT(), maxT(), degree, .1);

	IndexedLineSet* polyLine=dynamic_cast<IndexedLineSet*>(g_vcEllipticalPolyline2D->geometry.p);

	M_Vec3f& ilsCoord = dynamic_cast<Coordinate*>(polyLine->coord.p)->point;
	M_Int& coordIndex = polyLine->coordIndex.refValue();

	int ci = ilsCoord.length;
	for(int i = 0; i < coord.length; i++){
		ilsCoord+=Vec3f(coord[i].x, 0, 0);
		ilsCoord+=Vec3f(coord[i].x, coord[i].y * .1, 0);
		coordIndex+=ci;
		ci++;
		coordIndex+=ci;
		ci++;
		coordIndex+=-1;
	}
	polyLine->convertToRenderFormat(g_vcEllipticalPolyline2D->render);
}

void FrequencyModulator::houseKeep(){
	houseKeep2dNurbs(wet, wet, weight, knot, index[0], minT(), maxT(), degree, .1);
}

void Chords::houseKeep(){
	houseKeep2dNurbs(spectrum, spectrum, weight, knot, index[0], minT(), maxT(), degree, .1);

	IndexedLineSet* polyLine=dynamic_cast<IndexedLineSet*>(g_vcEllipticalPolyline2D->geometry.p);

	M_Vec3f& ilsCoord = dynamic_cast<Coordinate*>(polyLine->coord.p)->point;
	M_Int& coordIndex = polyLine->coordIndex.refValue();

	int ci = ilsCoord.length;
	for(int i = 0; i < coord.length; i++){
		//coord[i].y = amplitude[i] * (i + 1.0);
		ilsCoord+=Vec3f(coord[i].x, 0, 0);
		ilsCoord+=Vec3f(coord[i].x, coord[i].y * .1, 0);
		coordIndex+=ci;
		ci++;
		coordIndex+=ci;
		ci++;
		coordIndex+=-1;
	}
	polyLine->convertToRenderFormat(g_vcEllipticalPolyline2D->render);
}

void NurbsLathe::houseKeep(){
	houseKeep2dNurbs(dpoint, point, weight, knot, index[0], minT(), maxT(), degree, 1, matrix);
}
void BendyLathe::houseKeep(){
	houseKeep3dNurbs(dpoint, point, weight, knot, index[0], minT(), maxT(), matrix, patch, index[2]);
}

void NurbsTrack::houseKeep(){
	int idx = index[1];

	static M_Vec3f line;
	if(!keyValue.length){
		g_visualClue=0;
		return;
	}

	setMathHiPresision();
	line.length = 0;

	genCurve(line, 1000);

	const float inverseSize=(1 / g_scene->getZoom());
	
	IndexedLineSet* polyLine = dynamic_cast<IndexedLineSet*>(g_vcEllipticalPolyline2D->geometry.p);

	M_Vec3f& coord = dynamic_cast<Coordinate*>(polyLine->coord.p)->point;
	coord.length = line.length + keyValue.length;
	for(int i = 0; i < line.length; i++){
		coord[i] = line[i];
	}
	for(int j = line.length, i = keyValue.length; --i >= 0; j++){
		coord[j] = keyValue[i];
	}
	M_Int& coordIndex = polyLine->coordIndex.refValue();
	fillPolyLineIndex(coord, coordIndex);

/////////////
	coordIndex += -1;
	M_Node& children = g_vcBezierBase->children;
	children.length = 1;
	int childDiff = keyValue.length - children.length + 1;

	if(childDiff<0){
		children.length+=childDiff;
	}
	else{
		for(int i=childDiff; --i>=0;){
			Transform* transform=new Transform;
			transform->children.refValue()+=g_nextPoint.p; //could just add null?
			children+=transform;
		}
	}
	for(int i = keyValue.length; --i>=0;){
		Transform* transform=dynamic_cast<Transform*>(children[i+1].p);
		Vec3f pt = keyValue[i];
		transform->translation = pt;

		if(i == idx){
			transform->children.refValue()[0] = g_redPoint;
		}
		else{
			transform->children.refValue()[0] = g_nextPoint;
		}

		transform->scale=Vec3f(inverseSize, inverseSize, inverseSize);
	}

///////////
	g_vcBezierBase->rotation = Rotation();

	g_visualClue = g_vcBezierBase;

	restoreMathPresision();
	//g_vcEllipticalMaterial->diffuseColor=g_settings->backgroundColor.refValue().invert();
	polyLine->convertToRenderFormat(g_vcEllipticalPolyline2D->render);
}

inline void createCurvedPolyMapVisualAid(NurbsPatch& patch, M_Vec3f& coord, M_Int& coordIndex, M_Vec2f& baseCoord, const int qty, Vec3f& average, const bool& northPole, const bool& southPole){
	int uOrder = patch.order.x;
	int vOrder = patch.order.y;
	int uDegree = patch.uDegree;
	int vDegree = patch.vDegree;
	int uDim = patch.dim.x;
	int vDim = patch.dim.y;

	float* uKnotPtr = patch.uKnot.refValue().getArray();
	float* vKnotPtr = patch.vKnot.refValue().getArray();
	Vec3f* pointPtr = patch.point.refValue().getArray();
	float* weightPtr = patch.weight.refValue().getArray();

	Vec3f* topPointPtr = pointPtr+uDim*vDim-uDim;
	float* topWeightPtr = weightPtr+uDim*vDim-uDim;

	Vec3f* r_pointPtr = patch.r_point.getArray();
	float* r_weightPtr = patch.r_weight.getArray();


	float coordChange = .001;

	Vec2f diff;

	Vec3f curvedEdgePoint0;

	float maxU = patch.maxU();
	float maxV = patch.maxV();
	float minU = patch.minU();
	float minV = patch.minV();


	if(northPole){
		for(int i = 0; i < qty; i++){
			coordIndex+=coord.length;
			Vec3f v = getSurfaceVertex(convertNorthPoleToMercator(baseCoord[i]), maxU, maxV, minU, minV, uOrder, vOrder, uDegree, vDegree, uDim, vDim, uKnotPtr, vKnotPtr, pointPtr, weightPtr, topPointPtr, topWeightPtr, r_pointPtr, r_weightPtr);
			average+=v;
			coord+=v;
		}
	}
	else if(southPole){
		for(int i = 0; i < qty; i++){
			coordIndex+=coord.length;
			Vec3f v = getSurfaceVertex(convertSouthPoleToMercator(baseCoord[i]), maxU, maxV, minU, minV, uOrder, vOrder, uDegree, vDegree, uDim, vDim, uKnotPtr, vKnotPtr, pointPtr, weightPtr, topPointPtr, topWeightPtr, r_pointPtr, r_weightPtr);
			average+=v;
			coord+=v;
		}
	}
	else{
		for(int i = 0; i < qty; i++){
			coordIndex+=coord.length;
			Vec3f v = getSurfaceVertex(baseCoord[i], maxU, maxV, minU, minV, uOrder, vOrder, uDegree, vDegree, uDim, vDim, uKnotPtr, vKnotPtr, pointPtr, weightPtr, topPointPtr, topWeightPtr, r_pointPtr, r_weightPtr);
			average+=v;
			coord+=v;
		}
	}
}
void FuseSurface::houseKeep(){
	if(polyMapAid==false || parentPatch == 0 || parentStem == 0){
		g_visualClue=0;
		return;
	}

	setMathHiPresision();

	M_Node& children=g_vcBezierBase->children;
	children.length=1;

	IndexedLineSet* polyLine=dynamic_cast<IndexedLineSet*>(g_vcFuseSurfaceFlatStageAidShape->geometry.p);

	M_Vec3f& coord=dynamic_cast<Coordinate*>(polyLine->coord.p)->point;
	coord.length=2;


	M_Int& coordIndex = polyLine->coordIndex.refValue();
	coordIndex.length=0;
/*
	const Rotation orientation = this->orientation;
	Vec3f endPoint = Vec3f(0,1,0);
	if(baseCoord.length && curvedEdgePoint.length){

		M_Vec2f& baseCoord = this->baseCoord.refValue();
		const float vSpan=this->vSpan;
		NurbsPatch* stemPatch = parentPatch;
		if(stemPatch == 0){
			stemPatch = dynamic_cast<NurbsPatch*>(this->patch.p);
			if(stemPatch == 0){
				if(Stem* stem = dynamic_cast<Stem*>(this->patch.p)){
					stemPatch = dynamic_cast<NurbsPatch*>(stem->surface.p);
				}
			}
		}
		if(stemPatch){
			const int i=index[0];

			int uOrder = stemPatch->order.x;
			int vOrder = stemPatch->order.y;
			int uDim = stemPatch->dim.x;
			int vDim = stemPatch->dim.y;
			int uDegree = stemPatch->uDegree;
			int vDegree = stemPatch->vDegree;

			float* uKnotPtr = stemPatch->uKnot.refValue().getArray();
			float* vKnotPtr = stemPatch->vKnot.refValue().getArray();
			Vec3f* pointPtr = stemPatch->point.refValue().getArray();
			float* weightPtr = stemPatch->weight.refValue().getArray();
			Vec3f* topPointPtr = pointPtr+uDim*vDim-uDim;
			float* topWeightPtr = weightPtr+uDim*vDim-uDim;
			Vec3f* r_pointPtr = stemPatch->r_point.getArray();
			float* r_weightPtr = stemPatch->r_weight.getArray();


			Vec2f t = baseCoord[i];
			if(parentStem->northPole){
				t = convertNorthPoleToMercator(t);
			}
			else if(parentStem->southPole){
				t = convertSouthPoleToMercator(t);
			}


			float u = (stemPatch->maxU()-stemPatch->minU())*t.x + stemPatch->minU();
			float v = (stemPatch->maxV()-stemPatch->minV())*t.y + stemPatch->minV();

			if(t.x==1){
				CurvePoint(vDim-1, vDegree, vKnotPtr, r_pointPtr, r_weightPtr, v, curvedEdgePoint[i]);
			}
			else if(t.y==1){
				CurvePoint(uDim-1, uDegree, uKnotPtr, topPointPtr, topWeightPtr, u, curvedEdgePoint[i]);
			}
			else{
				SurfacePoint(u, v, uOrder, vOrder, uDim, vDim, uKnotPtr, vKnotPtr, pointPtr, weightPtr, &curvedEdgePoint[i]);
			}

			Vec3f ce = curvedEdgePoint[i];

			coord[0]=ce+orientation*endPoint;
			coord[1]=ce;
			fillPolyLineIndex(coord, coordIndex);
			coordIndex+=-1;
		}
	}
	else{
		coord[0]=orientation*endPoint;
		coord[1]=Vec3f();
		fillPolyLineIndex(coord, coordIndex);
		coord+=orientation*Vec3f(.03, .1, 0);
		coord+=orientation*Vec3f(-.03, .1, 0);
		coord+=orientation*Vec3f(0, .1, .03);
		coord+=orientation*Vec3f(0, .1, -.03);
		coordIndex+=-1;
		coordIndex+=2;
		coordIndex+=1;
		coordIndex+=4;
		coordIndex+=-1;
		coordIndex+=3;
		coordIndex+=1;
		coordIndex+=5;
		coordIndex+=-1;
	}
	*/
	if((polyMapAidMirror | polyMapAid) && baseCoord.length){
		Vec3f average;
		int remCoordLen=coord.length;
		if(dynamic_cast<ImmediatePart*>(g_selNode.getRaw())==0){
			createCurvedPolyMapVisualAid(*parentPatch, coord, coordIndex, baseCoord, qty, average, parentStem->northPole, parentStem->southPole);
		}
		else{
			if(polyMapAidMirror){
				for(int i = 0; i < qty; i++){
					coordIndex+=coord.length;
					Vec2f vec2f = baseCoord[i];
					vec2f.x-=.5;
					vec2f.x=-vec2f.x;
					vec2f.x+=.5;
					average.x+=vec2f.x;
					average.y+=vec2f.y;
					coord+=Vec3f(vec2f.x, vec2f.y, 0);
				}
			}
			else{
				for(int i = 0; i < qty; i++){
					coordIndex+=coord.length;
					Vec2f vec2f = baseCoord[i];
					average.x+=vec2f.x;
					average.y+=vec2f.y;
					coord+=Vec3f(vec2f.x, vec2f.y, 0);
				}
			}
		}
		coordIndex+=remCoordLen;
		coordIndex+=-1;
		average/=qty;

		for(int i = 0; i < baseCoord.length; i++){
			coordIndex+=remCoordLen+i;
			coordIndex+=coord.length;
			coordIndex+=-1;
			//coord+=(average+coord[remCoordLen+i]*7)*.125;
			coord+=average;
		}
	}

	g_vcBezierBase->rotation=Rotation();

	if(dynamic_cast<Art*>(g_selNode.getRaw())){
		g_vcArtTriBase->fuseSurface=this;
	}
	else{
		g_visualClue=g_vcFuseSurfaceFlatStageAidShape;
	}
	restoreMathPresision();
	//g_vcEllipticalMaterial->diffuseColor=g_settings->backgroundColor.refValue().invert();
	//polyLine->convertToRenderFormat(g_vcFuseSurfaceFlatStageAidShape->render);
	polyLine->renderProcessed=false;
}
void PolyToNurbs::houseKeep(){
	if(NurbsPatch* mimic = dynamic_cast<NurbsPatch*>(this->mimic.p)){
		if(mimic->stem.length){
			Stem* stem = dynamic_cast<Stem*>(mimic->stem[0].p);
			if(Art* originalPart = dynamic_cast<Art*>(this->original.p)){
				M_SArt& part = stem->part.refValue();
				if(part.length){
					M_SArt& source = stem->source.refValue();
					Art* guessPart = part[0];
					const M_Vec3f& mimicCoord = guessPart->coord;
					const M_Vec3f& originalCoord = originalPart->coord;

					if(source.length){
						Art* flatPart = source[0];

						setMathHiPresision();

						M_Node& children=g_vcBezierBase->children;
						children.length=1;

						IndexedLineSet* polyLine=dynamic_cast<IndexedLineSet*>(g_vcFuseSurfaceFlatStageAidShape->geometry.p);

						M_Vec3f& coord=dynamic_cast<Coordinate*>(polyLine->coord.p)->point;
						coord.length = 8;
						M_Int& coordIndex = polyLine->coordIndex.refValue();
						coordIndex.length=0;

						if(selMimicVertex >= mimicCoord.length){
							selMimicVertex.refValue() = mimicCoord.length-1;
						}
						Vec3f cb = mimicCoord[selMimicVertex];
						coord[0] = cb;
						Vec3f ce = cb;
						ce.x += 1;
						coord[1] = ce;

						coord[2] = cb + Vec3f(.1, .03, 0);
						coord[3] = cb + Vec3f(.1, -.03, 0);

						coordIndex+=0;
						coordIndex+=1;
						coordIndex+=-1;

						coordIndex+=2;
						coordIndex+=0;
						coordIndex+=3;
						coordIndex+=-1;

						int i = 0;
						float nearest = 10000000;
						int nearestIdx = 0;
						for(i = originalCoord.length; --i>=0;){
							Vec3f diff = originalCoord[i] - cb;
							float n = fabs(sqrt(diff.x*diff.x + diff.y*diff.y + diff.z*diff.z));
							if(n < nearest){
								nearest = n;
								nearestIdx=i;
							}
						}

						cb = originalCoord[nearestIdx];
						coord[4] = cb;
						ce = cb;
						ce.x -= 1;
						coord[5] = ce;

						coord[6] = cb + Vec3f(-.1, .03, 0);
						coord[7] = cb + Vec3f(-.1, -.03, 0);

						coordIndex+=4;
						coordIndex+=5;
						coordIndex+=-1;

						coordIndex+=6;
						coordIndex+=4;
						coordIndex+=7;

						selOriginalVertex.refValue() = nearestIdx;
						mimicOriginalIndex.refValue() = nearestIdx == selMimicVertex;

						g_visualClue=g_vcFuseSurfaceFlatStageAidShape;
						restoreMathPresision();
						polyLine->renderProcessed=false;
					}
				}
			}
		}
	}
}

void Harmonic::houseKeep(){
	houseKeep2dNurbs(dpoint, point, weight, knot, index[0], minT(), maxT(), degree, 1);
}

void HarmonicPatch::houseKeep(){
	if(!point.length){
		g_visualClue=0;
		return;
	}
	setMathHiPresision();
	int numOfU=this->dim.x;
	int numOfV=this->dim.y;

	{
		float inverseSize=(1/g_scene->getZoom());
		int idx=index[0];

		IndexedLineSet* polyLine=dynamic_cast<IndexedLineSet*>(g_vcEllipticalPolyline2D->geometry.p);

		M_Vec3f& coord=dynamic_cast<Coordinate*>(polyLine->coord.p)->point;

		coord.length=0;

		for(int i=point.length; --i>=0;){
			coord += point[i] / weight[i];
		}

		M_Int& coordIndex = polyLine->coordIndex.refValue();
		coordIndex.length=0;


		for(int v=0; v<numOfV; v++){
			for(int u=0; u<numOfU; u++){
				coordIndex+=v*numOfU+u;
			}
			coordIndex+=-1;
		}
		for(int u=0; u<numOfU; u++){
			for(int v=0; v<numOfV; v++){
				coordIndex+=v*numOfU+u;
			}
			coordIndex+=-1;
		}
	/////////////
		{
			int pointLen=numOfU*numOfV;

			M_Node& children=g_vcBezierBase->children;
			children.length=1;
			int childDiff=pointLen-children.length+1;

			if(childDiff<0){
				children.length+=childDiff;
			}
			else{
				for(int i=childDiff; --i>=0;){
					Transform* transform=new Transform;
					transform->children.refValue()+=g_nextPoint.p; //could just add null?
					children+=transform;
				}
			}
			for(int i=pointLen; --i>=0;){
				Transform* transform=dynamic_cast<Transform*>(children[i+1].p);
				Vec3f pt=point[i] / weight[i];
				transform->translation=pt;

				if(i==idx){
					transform->children.refValue()[0]=g_redPoint;
				}
				else{
					transform->children.refValue()[0]=g_nextPoint;
				}

				transform->scale=Vec3f( inverseSize, inverseSize, inverseSize);
			}
		}

	///////////
		g_vcBezierBase->rotation=Rotation();

		//g_visualClue=g_vcBezierBase;
		NurbsPatch* nurbsPatch = dynamic_cast<NurbsPatch*>(g_vcOscSmls->build[0].p);
		Stem* stem = dynamic_cast<Stem*>(nurbsPatch->stem[0].p);
		M_Int& stemCoordIndex = stem->coordIndex.refValue();

		stem->staves.refValue() = staves;
		stem->coord.refValue().length = stem->color.refValue().length = vCoord.length;
		stemCoordIndex.length = (harmonics - 1) * staves * 3 * 2;
		int staves = this->staves;

		int tri = 0;
		for(int h = 0; h < harmonics - 1; h++){
			for(int s = 0; s < staves; s++){
				stemCoordIndex[tri++] = (staves + 1) * h + s + 1;
				stemCoordIndex[tri++] = (staves + 1) * (h + 1) + s;
				stemCoordIndex[tri++] = (staves + 1) * h + s;

				stemCoordIndex[tri++] = (staves + 1) * h + s + 1;
				stemCoordIndex[tri++] = (staves + 1) * (h + 1) + s + 1;
				stemCoordIndex[tri++] = (staves + 1) * (h + 1) + s;
			}
		}
		//String str;
		//for(int i = 0; i < stemCoordIndex.length; i++){
		//	str += stemCoordIndex[i];
		//	str += " ";
		//	if(stemCoordIndex[i] >= stem->coord.length){
		//		break;
		//	}
		//}
		//dLog(str);

		if(vCoord.length){
			float f = 1.0 / staves;
			for(int h = 0; h < harmonics; h++){
				for(int i = 0; i < staves + 1; i++){
					int i2 = h * (staves + 1) + i;
					//vCoord.refValue()[i2] = .01 * h;
					stem->coord.refValue()[i2] = Vec2f(f * i, vCoord[i2]);
				}
			}
		}
		nurbsPatch->point.refValue() = point;
		nurbsPatch->weight.refValue() = weight;
		nurbsPatch->uKnot.refValue() = uKnot;
		nurbsPatch->vKnot.refValue() = vKnot;
		nurbsPatch->dim.refValue() = dim;
		nurbsPatch->order.refValue() = order;
		nurbsPatch->showAll.refValue() = true;
		nurbsPatch->processFields();

		g_vcOscSmls->doBuild();
		Art* part = dynamic_cast<Art*>(stem->part[0].p);
		this->coord.refValue() = part->coord;
		g_visualClue = g_vcOscBase;

		g_vcOscSmls->convertToRenderFormat();

	//	g_scene->children.refValue()+=g_vcBezierBase.p;
	//	g_stw->genNodeList();


		//g_vcEllipticalMaterial->diffuseColor=g_settings->backgroundColor.refValue().invert();
		polyLine->convertToRenderFormat(g_vcEllipticalPolyline2D->render);
	}
	restoreMathPresision();
}
/*
void Subdivision::houseKeep(){
	if(!point.length){
		g_visualClue=0;
		return;
	}
	setMathHiPresision();
	float inverseSize=(1/g_scene->getZoom());
	int idx=index[0];

	IndexedLineSet* polyLine=dynamic_cast<IndexedLineSet*>(g_vcEllipticalPolyline2D->geometry.p);

	M_Vec3f& coord=dynamic_cast<Coordinate*>(polyLine->coord.p)->point;

	coord.length=0;

	coord = point;

	M_Int& polyCoordIndex = polyLine->coordIndex.refValue();
	polyCoordIndex.length=0;

	M_Int& coordIndex = this->coordIndex.refValue();

	for(int i=0; i<coordIndex.length; ){
		polyCoordIndex += coordIndex[i++];
		polyCoordIndex += coordIndex[i++];
		polyCoordIndex += coordIndex[i++];
		polyCoordIndex += coordIndex[i++];
		polyCoordIndex += coordIndex[i-4];
		polyCoordIndex += -1;
	}


/////////////
	{
		int pointLen=point.length;

		M_Node& children=g_vcBezierBase->children;
		children.length=1;
		int childDiff=pointLen-children.length+1;

		if(childDiff<0){
			children.length+=childDiff;
		}
		else{
			for(int i=childDiff; --i>=0;){
				Transform* transform=new Transform;
				transform->children.refValue()+=g_nextPoint.p; //could just add null?
				children+=transform;
			}
		}
		for(int i=pointLen; --i>=0;){
			Transform* transform=dynamic_cast<Transform*>(children[i+1].p);
			Vec3f pt=point[i];
			transform->translation=pt;

			if(i==idx){
				transform->children.refValue()[0]=g_redPoint;
			}
			else{
				transform->children.refValue()[0]=g_nextPoint;
			}

			transform->scale=Vec3f( inverseSize, inverseSize, inverseSize);
		}
	}

///////////
	g_vcBezierBase->rotation=Rotation();

	g_visualClue=g_vcBezierBase;

	polyLine->convertToRenderFormat(g_vcEllipticalPolyline2D->render);
	restoreMathPresision();
}
*/

void Subdivision::houseKeep(){
	if(!point.length){
		g_visualClue=0;
		return;
	}
	setMathHiPresision();
	float inverseSize=(1/g_scene->getZoom());
	int idx=index[0];

	IndexedLineSet* polyLine=dynamic_cast<IndexedLineSet*>(g_vcEllipticalPolyline2D->geometry.p);

	M_Vec3f& coord=dynamic_cast<Coordinate*>(polyLine->coord.p)->point;

	coord.length=0;

	coord = point;

	M_Int& polyCoordIndex = polyLine->coordIndex.refValue();
	polyCoordIndex.length=0;

	M_Int& coordIndex = this->coordIndex.refValue();

	for(int i=0; i<coordIndex.length; ){
		polyCoordIndex += coordIndex[i++];
		polyCoordIndex += coordIndex[i++];
		polyCoordIndex += coordIndex[i++];
		polyCoordIndex += coordIndex[i++];
		polyCoordIndex += coordIndex[i-4];
		polyCoordIndex += -1;
	}


/////////////
	{
		int pointLen=point.length;

		M_Node& children=g_vcBezierBase->children;
		children.length=1;
		int childDiff=4-children.length+1;

		if(childDiff<0){
			children.length+=childDiff;
		}

		{
			for(int i=childDiff; --i>=0;){
				Transform* transform=new Transform;
				transform->children.refValue()+=g_nextPoint.p; //could just add null?
				children+=transform;
			}

			M_Int& coordIndex = this->coordIndex.refValue();

			for(int i=4; --i>=0;){
				int j = coordIndex[actPoly*4+i];
				Transform* transform=dynamic_cast<Transform*>(children[i+1].p);
				Vec3f pt = point[j];
				transform->translation = pt;
				if(i==actCorner){
					transform->children.refValue()[0]=g_redPoint;
				}
				else if(i==(actCorner+1)%4){
					transform->children.refValue()[0]=g_nextPoint;
				}
				else{
					transform->children.refValue()[0]=g_lastPoint;
				}
				transform->scale=Vec3f( inverseSize, inverseSize, inverseSize);
			}
		}
	}

///////////
	g_vcBezierBase->rotation=Rotation();

	g_visualClue=g_vcBezierBase;

	polyLine->convertToRenderFormat(g_vcEllipticalPolyline2D->render);
	restoreMathPresision();
}
Vec3f g_immediateRadialTugResizeLocation;

void NurbsPatch::addCageShowMid(IndexedLineSet* polyLine){
	int uDim = this->dim.x;
	int vDim = this->dim.y;

	M_Vec3f& coord = dynamic_cast<Coordinate*>(polyLine->coord.p)->point;
	M_Color3f& color = dynamic_cast<Color*>(polyLine->color.p)->color;
	M_Int& coordIndex = polyLine->coordIndex.refValue();

	Color3f midColor = Color3f(1, 0, 1);
	Color3f visColor = Color3f(0, 1, 1);

	for(int i = dpoint.length; --i>=0;){
		coord[fromPoint + i] = matrix * dpoint[i];
		color[fromPoint + i] = midColor;
	}
	if(cage == false){
		return;
	}

	for(int v = 0; v < vDim; v++){
		for(int u=0; u<uDim; u++){
			coordIndex += fromPoint + v * uDim+u;
		}
		coordIndex += -1;
	}
	for(int u = 0; u < uDim; u++){
		for(int v=0; v < vDim; v++){
			coordIndex += fromPoint + v * uDim + u;
		}
		coordIndex += -1;
	}

	if(NurbsPatch::MID_1 == checkNurbsType(vDim, vKnot)){
		if(NurbsPatch::MID_1 == checkNurbsType(uDim, uKnot)){
			for(int v = 1; v < vDim - 1; v += 2){
				for(int u = 1; u < uDim; u += 2){
					color[fromPoint + v * uDim + u] = visColor;
				}
				if(v == 0 || v == vDim - 2){
					v--;
				}
			}
		}
		else{
			for(int v = 1; v < vDim - 1; v += 2){
				for(int u = 1; u < uDim - 1; u += 1){
					color[fromPoint + v * uDim + u] = visColor;
				}
				if(v == 0 || v == vDim - 2){
					v--;
				}
			}
		}
	}
	else{
		if(NurbsPatch::MID_1 == checkNurbsType(uDim, uKnot)){
			for(int v = 1; v < vDim - 1; v++){
				for(int u = 1; u < uDim; u += 2){
					color[fromPoint + v * uDim + u] = visColor;
				}
			}
		}
		else{
			for(int v = 1; v < vDim - 1; v++){
				for(int u = 1; u < uDim -  1; u++){
					color[fromPoint + v * uDim + u] = visColor;
				}
			}
		}
	}
}

void NurbsPatch::addCage(IndexedLineSet* polyLine){
	int uDim = this->dim.x;
	int vDim = this->dim.y;

	M_Vec3f& coord = dynamic_cast<Coordinate*>(polyLine->coord.p)->point;
	M_Int& coordIndex = polyLine->coordIndex.refValue();

	for(int i = dpoint.length; --i>=0;){
		coord[fromPoint + i] = matrix * dpoint[i];
	}
	if(cage == false){
		return;
	}

	if(uJoinEnds){
		if(NurbsPatch::MID_1 == checkNurbsType(vDim, vKnot)){
			if(isCtrlPolyRef){
				for(int v = 1; v < vDim - 1; v += 2){
					for(int u = 1; u < uDim - 1; u += 2){
						coordIndex += fromPoint + v * uDim+u;
					}
					coordIndex += -1;
				}
				for(int u = 1; u < uDim - 1; u += 2){
					for(int v = 1; v < vDim - 1; v++){
						coordIndex += fromPoint + v * uDim + u;
					}
					coordIndex += -1;
				}
			}
			else{
				for(int v=0; v<vDim; v+=2){
					coordIndex += fromPoint + v * uDim;
					for(int u=1; u<uDim; u+=2){
						coordIndex += fromPoint + v * uDim + u;
					}
					coordIndex += fromPoint + v * uDim + uDim - 1;
					coordIndex+=-1;

					if(v == 0 || v == vDim - 2){
						v--;
					}
				}
				//for(int v=0; v<vDim; v += 2){
				//	coordIndex += fromPoint + v * uDim + uDim - 2;
				//	for(int u=1; u<uDim; u+=2){
				//		coordIndex += fromPoint + v * uDim + u;
				//	}
				//	coordIndex += -1;
				//}
				for(int u=1; u<uDim; u += 2){
					for(int v=0; v<vDim; v++){
						coordIndex += fromPoint + v * uDim + u;
					}
					coordIndex += -1;
				}
			}
		}
		else{
			if(isCtrlPolyRef){
				if(vDim > 3){
					for(int v = 1; v < vDim - 1; v++){
						for(int u = 1; u < uDim - 1; u += 2){
							coordIndex += fromPoint + v * uDim+u;
						}
						coordIndex += -1;
					}
					for(int u = 1; u < uDim - 1; u += 2){
						for(int v = 1; v < vDim - 1; v++){
							coordIndex += fromPoint + v * uDim + u;
						}
						coordIndex += -1;
					}
				}
			}
			else{
				for(int v=0; v < vDim; v++){
					coordIndex += fromPoint + v * uDim + uDim - 2;
					for(int u=1; u<uDim; u+=2){
						coordIndex += fromPoint + v * uDim + u;
					}
					coordIndex += -1;
				}
				for(int u=1; u<uDim; u+=2){
					for(int v=0; v<vDim; v++){
						coordIndex += fromPoint + v * uDim + u;
					}
					coordIndex += -1;
				}
			}
		}
	}
	else {
		if(NurbsPatch::MID_1 == checkNurbsType(vDim, vKnot)){
			if(isCtrlPolyRef){
				if(vDim > 5){
					for(int v = 1; v < vDim - 1; v += 2){
						for(int u = 1; u < uDim - 1; u += 2){
							coordIndex += fromPoint + v * uDim+u;
						}
						coordIndex += -1;
					}
					for(int u = 1; u < uDim - 1; u += 2){
						for(int v = 1; v < vDim - 1; v += 2){
							coordIndex += fromPoint + v * uDim + u;
						}
						coordIndex += -1;
					}
				}
			}
			else{
				for(int v=0; v<vDim; v+=2){
					coordIndex += fromPoint + v * uDim;
					for(int u=1; u<uDim; u+=2){
						coordIndex += fromPoint + v * uDim + u;
					}
					coordIndex += fromPoint + v * uDim + uDim - 1;
					coordIndex+=-1;

					if(v == 0 || v == vDim - 2){
						v--;
					}
				}
				for(int u=0; u<uDim; u += 2){
					coordIndex += fromPoint + u;
					for(int v=1; v<vDim; v+=2){
						coordIndex += fromPoint + v * uDim + u;
					}
					coordIndex += fromPoint + uDim * vDim - uDim + u;
					coordIndex += -1;
					if(u == 0 || u == uDim - 2){
						u--;
					}
				}
			}
		}
		else{
			if(NurbsPatch::MID_1 == checkNurbsType(uDim, uKnot)){
				if(isCtrlPolyRef){
					if(vDim > 3){
						for(int v = 1; v < vDim - 1; v++){
							for(int u = 1; u < uDim - 1; u += 2){
								coordIndex += fromPoint + v * uDim+u;
							}
							coordIndex += -1;
						}
						for(int u = 1; u < uDim - 1; u += 2){
							for(int v = 1; v < vDim - 1; v++){
								coordIndex += fromPoint + v * uDim + u;
							}
							coordIndex += -1;
						}
					}
				}
				else{
					for(int v = 0; v < vDim; v++){
						coordIndex += fromPoint + v * uDim;
						for(int u = 1; u < uDim; u +=2){
							coordIndex += fromPoint + v * uDim + u;
						}
						coordIndex += fromPoint + v * uDim + uDim - 1;
						coordIndex += -1;
					}
					for(int u = 0; u < uDim; u += 2){
						coordIndex += fromPoint + u;
						for(int v = 0; v < vDim; v++){
							coordIndex += fromPoint + v * uDim + u;
						}
						coordIndex += fromPoint + uDim * vDim - uDim + u;
						coordIndex += -1;
						if(u == 0 || u == uDim - 2){
							u--;
						}
					}
				}
			}
			else{
				if(isCtrlPolyRef){
					for(int v = 1; v < vDim - 1; v++){
						for(int u = 1; u < uDim - 1; u++){
							coordIndex += fromPoint + v * uDim + u;
						}
						coordIndex += -1;
					}
					for(int u = 1; u < uDim - 1; u++){
						for(int v = 1; v < vDim - 1; v++){
							coordIndex += fromPoint + v * uDim + u;
						}
						coordIndex += -1;
					}
				}
				else {
					for(int v = 0; v < vDim; v++){
						for(int u = 0; u < uDim; u++){
							coordIndex += fromPoint + v * uDim + u;
						}
						coordIndex += -1;
					}
					for(int u = 0; u < uDim; u++){
						for(int v = 0; v < vDim; v++){
							coordIndex += fromPoint + v * uDim + u;
						}
						coordIndex += -1;
					}
				}
			}
		}
	}
}


void NurbsPatch::selectRing(int v, int uDim, int vDim){
	int idx = 0;
	if(uJoinEnds){
		if(NurbsPatch::MID_1 == checkNurbsType(vDim, vKnot)){
			if(isCtrlPolyRef){
				for(int u = 1; u < uDim - 1; u += 2){
					idx = v * uDim + u;
					MultiSelect::beginDragPos += dpoint[idx];
					MultiSelect::pointList += ControlPoint(this, idx);
				}
			}
			else{
				idx = v * uDim;
				MultiSelect::beginDragPos += dpoint[idx];
				MultiSelect::pointList += ControlPoint(this, idx);
				for(int u = 1; u<uDim; u+=2){
					idx = v * uDim + u;
					MultiSelect::beginDragPos += dpoint[idx];
					MultiSelect::pointList += ControlPoint(this, idx);
				}
				idx = v * uDim + uDim - 1;
				MultiSelect::beginDragPos += dpoint[idx];
				MultiSelect::pointList += ControlPoint(this, idx);
			}
		}
		else if(NurbsPatch::MID_0 == checkNurbsType(vDim, vKnot)){
			if(isCtrlPolyRef){
				MultiSelect::uRingDim = uDim / 2;
				MultiSelect::vRingDim = vDim - v - 1;
				for(int u = 1; u < uDim - 1; u += 2){
					idx = v * uDim+u;
					MultiSelect::beginDragPos += dpoint[idx];
					MultiSelect::pointList += ControlPoint(this, idx);
				}
			}
			else{
				MultiSelect::uRingDim = uDim / 2;
				MultiSelect::vRingDim = vDim - v;
				idx = v * uDim + uDim - 2;
				for(int u = 1; u < uDim - 1; u+=2){
					idx = v * uDim + u;
					MultiSelect::beginDragPos += dpoint[idx];
					MultiSelect::pointList += ControlPoint(this, idx);
				}
			}
		}
	}
	else {
		if(NurbsPatch::MID_1 == checkNurbsType(vDim, vKnot)){
			if(isCtrlPolyRef){
				MultiSelect::uRingDim = uDim / 2;
				MultiSelect::vRingDim = (vDim / 2 - 1) - v / 2;
				for(int u = 1; u < uDim - 1; u += 2){
					idx = v * uDim + u;
					MultiSelect::beginDragPos += dpoint[idx];
					MultiSelect::pointList += ControlPoint(this, idx);
				}
			}
			else{
				MultiSelect::uRingDim = uDim / 2;
				MultiSelect::vRingDim = (vDim / 2) - v / 2;
				int idx = v * uDim;
				MultiSelect::beginDragPos += dpoint[idx];
				MultiSelect::pointList += ControlPoint(this, idx);
				for(int u = 1; u < uDim; u += 2){
					idx = v * uDim + u;
					MultiSelect::beginDragPos += dpoint[idx];
					MultiSelect::pointList += ControlPoint(this, idx);
				}
				idx = v * uDim + uDim - 1;
				MultiSelect::beginDragPos += dpoint[idx];
				MultiSelect::pointList += ControlPoint(this, idx);
			}
		}
		else{
			if(NurbsPatch::MID_1 == checkNurbsType(uDim, uKnot)){
				if(isCtrlPolyRef){
					MultiSelect::uRingDim = uDim / 2;
					MultiSelect::vRingDim = (vDim / 2 - 1) - v / 2;
					for(int u = 1; u < uDim - 1; u += 2){
						idx = v * uDim + u;
						MultiSelect::beginDragPos += dpoint[idx];
						MultiSelect::pointList += ControlPoint(this, idx);
					}
				}
				else{
					idx = v * uDim;
					MultiSelect::beginDragPos += dpoint[idx];
					MultiSelect::pointList += ControlPoint(this, idx);
					for(int u = 1; u < uDim; u +=2){
						idx = v * uDim + u;
						MultiSelect::beginDragPos += dpoint[idx];
						MultiSelect::pointList += ControlPoint(this, idx);
					}
					idx = v * uDim + uDim - 1;
					MultiSelect::beginDragPos += dpoint[idx];
					MultiSelect::pointList += ControlPoint(this, idx);
				}
			}
			else{
				if(isCtrlPolyRef){
					for(int u = 1; u < uDim - 1; u++){
						idx = v * uDim + u;
						MultiSelect::beginDragPos += dpoint[idx];
						MultiSelect::pointList += ControlPoint(this, idx);
					}
				}
				else {
					for(int u = 0; u < uDim; u++){
						idx = v * uDim + u;
						MultiSelect::beginDragPos += dpoint[idx];
						MultiSelect::pointList += ControlPoint(this, idx);
					}
				}
			}
		}
	}
}

void NurbsPatch::selectRings(int v, int uDim, int vDim){
	int idx = 0;
	if(uJoinEnds){
		if(NurbsPatch::MID_1 == checkNurbsType(vDim, vKnot)){
			if(isCtrlPolyRef){
				for(; v < vDim - 1; v += 2){
					for(int u = 1; u < uDim - 1; u += 2){
						idx = v * uDim + u;
						MultiSelect::beginDragPos += dpoint[idx];
						MultiSelect::pointList += ControlPoint(this, idx);
					}
				}
			}
			else{
				for(; v < vDim; v += 2){
					idx = v * uDim;
					MultiSelect::beginDragPos += dpoint[idx];
					MultiSelect::pointList += ControlPoint(this, idx);
					for(int u = 1; u<uDim; u+=2){
						idx = v * uDim + u;
						MultiSelect::beginDragPos += dpoint[idx];
						MultiSelect::pointList += ControlPoint(this, idx);
					}
					idx = v * uDim + uDim - 1;
					MultiSelect::beginDragPos += dpoint[idx];
					MultiSelect::pointList += ControlPoint(this, idx);

					if(v == 0 || v == vDim - 2){
						v--;
					}
				}
			}
		}
		else if(NurbsPatch::MID_0 == checkNurbsType(vDim, vKnot)){
			if(isCtrlPolyRef){
				if(vDim > 3){
					MultiSelect::uRingDim = uDim / 2;
					MultiSelect::vRingDim = vDim - v - 1;
					for(; v < vDim - 1; v++){
						for(int u = 1; u < uDim - 1; u += 2){
							idx = v * uDim+u;
							MultiSelect::beginDragPos += dpoint[idx];
							MultiSelect::pointList += ControlPoint(this, idx);
						}
					}
				}
			}
			else{
				MultiSelect::uRingDim = uDim / 2;
				MultiSelect::vRingDim = vDim - v;
				for(; v < vDim; v++){
					idx = v * uDim + uDim - 2;
					for(int u = 1; u < uDim - 1; u+=2){
						idx = v * uDim + u;
						MultiSelect::beginDragPos += dpoint[idx];
						MultiSelect::pointList += ControlPoint(this, idx);
					}
				}
			}
		}
	}
	else {
		if(NurbsPatch::MID_1 == checkNurbsType(vDim, vKnot)){
			if(isCtrlPolyRef){
				if(vDim > 5){
					MultiSelect::uRingDim = uDim / 2;
					MultiSelect::vRingDim = (vDim / 2 - 1) - v / 2;
					for(; v < vDim - 1; v += 2){
						for(int u = 1; u < uDim - 1; u += 2){
							idx = v * uDim + u;
							MultiSelect::beginDragPos += dpoint[idx];
							MultiSelect::pointList += ControlPoint(this, idx);
						}
					}
				}
			}
			else{
				MultiSelect::uRingDim = uDim / 2;
				MultiSelect::vRingDim = (vDim / 2) - v / 2;
				for(; v < vDim; v += 2){
					int idx = v * uDim;
					MultiSelect::beginDragPos += dpoint[idx];
					MultiSelect::pointList += ControlPoint(this, idx);
					for(int u = 1; u < uDim; u += 2){
						idx = v * uDim + u;
						MultiSelect::beginDragPos += dpoint[idx];
						MultiSelect::pointList += ControlPoint(this, idx);
					}
					idx = v * uDim + uDim - 1;
					MultiSelect::beginDragPos += dpoint[idx];
					MultiSelect::pointList += ControlPoint(this, idx);

					if(v == 0 || v == vDim - 2){
						v--;
					}
				}
			}
		}
		else{
			if(NurbsPatch::MID_1 == checkNurbsType(uDim, uKnot)){
				if(isCtrlPolyRef){
					if(vDim > 3){
						for(; v < vDim - 1; v++){
							for(int u = 1; u < uDim - 1; u += 2){
								idx = v * uDim+u;
								MultiSelect::beginDragPos += dpoint[idx];
								MultiSelect::pointList += ControlPoint(this, idx);
							}
						}
					}
				}
				else{
					for(; v < vDim; v++){
						idx = v * uDim;
						MultiSelect::beginDragPos += dpoint[idx];
						MultiSelect::pointList += ControlPoint(this, idx);
						for(int u = 1; u < uDim; u +=2){
							idx = v * uDim + u;
							MultiSelect::beginDragPos += dpoint[idx];
							MultiSelect::pointList += ControlPoint(this, idx);
						}
						idx = v * uDim + uDim - 1;
						MultiSelect::beginDragPos += dpoint[idx];
						MultiSelect::pointList += ControlPoint(this, idx);
					}
				}
			}
			else{
				if(isCtrlPolyRef){
					for(; v < vDim - 1; v++){
						for(int u = 1; u < uDim - 1; u++){
							idx = v * uDim + u;
							MultiSelect::beginDragPos += dpoint[idx];
							MultiSelect::pointList += ControlPoint(this, idx);
						}
					}
				}
				else {
					for(; v < vDim; v++){
						for(int u = 0; u < uDim; u++){
							idx = v * uDim + u;
							MultiSelect::beginDragPos += dpoint[idx];
							MultiSelect::pointList += ControlPoint(this, idx);
						}
					}
				}
			}
		}
	}
}

int Seamless::getCtrlPolyLengthInTri(int& coordLen){
	int triLen = 0;
	for(int i = 0; i < ctrlPolyPoint.length;){
		int len = ctrlPolyPoint[i].len;
		coordLen += len;
		triLen += len - 2;
		i = getNextCtrlPoly(i);
	}
	return triLen;
}

void Seamless::genTransparentCtrlPolys(Art* part, const M_Vec3f& point){
	if(part->workEffect==0)return;
	part->deleteGeometry();
	Seamless* smls = part->smls;
	Effect* effect=part->workEffect;
	int effectLength=effect->getEffectLength();
	int coordLen = 0;
	int triLen = getCtrlPolyLengthInTri(coordLen);
	M_Color3f& color = dynamic_cast<ColorEffect*>(effect)->color.refValue();
	color.length = coordLen;
	for(int i = coordLen; --i >= 0;){
		color[i] = Color3f(0,1,1);
	}

	M_Vec3f& coord = part->coord.refValue();
	coord.length = coordLen;
	M_Triangle& m_triangle = part->triangle;
	m_triangle.length = triLen;

	int ctrlPolyPointLen = ctrlPolyPoint.length;

	for(int i = 0, t = 0, c = 0, s = 0; i < ctrlPolyPointLen;){
		int len = ctrlPolyPoint[i].len;
		if(s < MultiSelect::ctrlPolySelList.length){
			if(i == MultiSelect::ctrlPolySelList[s]){
				for(int j = 0; j < len; j++){
					color[c + j] = Color3f(1,0,0);
				}
				s++;
			}
		}
		int c0 = c;
		coord[c0] = point[ctrlPolyPoint[i + 0].p->fromPoint + ctrlPolyPoint[i + 0].i];
		coord[c + 1] = point[ctrlPolyPoint[i + 1].p->fromPoint + ctrlPolyPoint[i + 1].i];
		coord[c + 2] = point[ctrlPolyPoint[i + 2].p->fromPoint + ctrlPolyPoint[i + 2].i];
		m_triangle[t++] = S_Triangle(part, c + 0, c + 0, part, c + 1, c + 1, part, c + 2, c + 2);

		for(int j = 2; j < len - 1; j++){
			coord[c + j + 1] = point[ctrlPolyPoint[i + j + 1].p->fromPoint + ctrlPolyPoint[i + j + 1].i];
			m_triangle[t++] = S_Triangle(part, c0, c0, part, c + j, c + j, part, c + j + 1, c + j + 1);
		}
		c += len;
		i = getNextCtrlPoly(i);
	}

	part->relevantPart=M_Art(part);
	smls->process();
}

bool g_showControlPolys = false;

void Seamless::colorLinkPoints(){
	M_Color3f& color = dynamic_cast<Color*>(g_nurbsIlsColor)->color;

	for(int i = link.length; --i>=0;){
		color[link[i].patch->fromPoint + link[i].point] = Color3f(0,0,1);
	}
}

void Seamless::addCageControlPolys(){
	M_Vec3f& point=dynamic_cast<Coordinate*>(g_nurbsIlsCoord)->point;
	M_Int& coordIndex = g_nurbsIls->coordIndex.refValue();

	int ctrlPolyPointLen = ctrlPolyPoint.length;
	for(int i = 0; i < ctrlPolyPointLen;){
		int ctrlPoly = i;
		int len = ctrlPolyPoint[ctrlPoly].len;
		int mark = coordIndex.length;
		bool flag = false;
		int m = i + len;
		for(int k = len; --k >= 0;){
			NurbsPatch* patch = ctrlPolyPoint[i].p;
			flag |= patch->cage;
			coordIndex += patch->fromPoint + ctrlPolyPoint[i].i;
			i++;
		}
		if(flag == false){
			coordIndex.length = mark;
		}
		else{
			coordIndex += ctrlPolyPoint[ctrlPoly].p->fromPoint + ctrlPolyPoint[ctrlPoly].i;
			coordIndex += -1;
		}
		i = getNextCtrlPoly(ctrlPoly);
	}
	if(g_showControlPolys && ctrlPolyPoint.length){
		genTransparentCtrlPolys(g_vcNurbsCtrlPolyPart, point);
	}
}

void Seamless::addCageControlPolysColorCode(){
	M_Vec3f& point=dynamic_cast<Coordinate*>(g_nurbsIlsCoord)->point;
	M_Color3f& color = dynamic_cast<Color*>(g_nurbsIlsColor)->color;
	M_Int& coordIndex = g_nurbsIls->coordIndex.refValue();

	int ctrlPolyPointLen = ctrlPolyPoint.length;
	for(int i = 0; i < ctrlPolyPointLen;){
		int ctrlPoly = i;
		int len = ctrlPolyPoint[ctrlPoly].len;
		int mark = coordIndex.length;
		bool flag = false;
		int m = i + len;
		for(int k = len; --k >= 0;){
			NurbsPatch* patch = ctrlPolyPoint[i].p;
			flag |= patch->cage;
			coordIndex += patch->fromPoint + ctrlPolyPoint[i].i;
			i++;
		}
		if(flag == false){
			coordIndex.length = mark;
			i = getNextCtrlPoly(ctrlPoly);
			continue;
		}
		for(int o = len + 1; --o>=0;){
			int midLen = ctrlPolyPoint[m].len;
			if(midLen == 0){
				m++;
				continue;
			}
			for(int j = midLen; --j>=0;){
				color[ctrlPolyPoint[m].p->fromPoint + ctrlPolyPoint[m].i] = Color3f(0,1,0);
				m++;
			}
		}
		i = m;

		coordIndex += ctrlPolyPoint[ctrlPoly].p->fromPoint + ctrlPolyPoint[ctrlPoly].i;
		coordIndex += -1;
//		i = getNextCtrlPoly(ctrlPoly);
	}
	if(g_showControlPolys && ctrlPolyPoint.length){
		genTransparentCtrlPolys(g_vcNurbsCtrlPolyPart, point);
	}
}

void genCage(Seamless* smls){
	M_Vec3f& point = dynamic_cast<Coordinate*>(g_nurbsIlsCoord)->point;

	point.length = 0;
	M_Int& coordIndex = g_nurbsIls->coordIndex.refValue();
	coordIndex.length = 0;

	M_<NurbsPatch*>& cpNodeList = smls->cpNodeList;

	g_nurbsIlsCoord->point.refValue().length = smls->pointTotal;
	M_Color3f& color = g_nurbsIlsColor->color.refValue();
	color.length = smls->pointTotal;

	for(int k = 0; k < cpNodeList.length; k++){
		NurbsPatch* patch = cpNodeList[k];
		patch->matrix = scaleMatrix(patch->scale) * rotationMatrix(patch->rotation) * translationMatrix(patch->translation);
		if(patch->showAll){
			patch->addCageShowMid(g_nurbsIls);
		}
		else{
			int fromPoint = patch->fromPoint;
			for(int i = patch->dpoint.length; --i>=0;){
				color[fromPoint + i] = Color3f(0, 1, 1);
			}
			patch->addCage(g_nurbsIls);
		}
	}

	smls->colorLinkPoints();
	smls->addCageControlPolysColorCode();
	//int d = 0;
	//int pointCount = 0;
	//for(int s = 0 ; s < coordIndex.length; ){
	//	int idx = coordIndex[s++];
	//	if(idx == -1){
	//		if(pointCount == 1){
	//			d--;
	//		}
	//		pointCount = 0;
	//		coordIndex[d++] = idx;
	//		continue;
	//	}
	//	Color3f clr = color[idx];
	//	if(clr == Color3f(0,1,1)){
	//		coordIndex[d++] = idx;
	//		pointCount++;
	//	}
	//}
	//coordIndex.length = d;


	g_vcBezierBase->rotation=Rotation();

	g_nurbsIls->convertToRenderFormat(g_nurbsIlsShape->render);
}

bool isSelected(NurbsPatch* patch, int idx){
	for(int k = MultiSelect::pointList.length; --k>= 0;){
		if(patch == MultiSelect::pointList[k].p && idx == MultiSelect::pointList[k].i){
			return true;
		}
	}
	return false;
}

void NurbsPatch::houseKeep(){
	if(smls == 0){
		return;
	}

	if(smls->workAnimCP == true){
		smls->workAnimCP = false;
		M_<NurbsPatch*>& cpNodeList = smls->cpNodeList;
		for(int k = 0; k < cpNodeList.length; k++){
			NurbsPatch* patch = cpNodeList[k];
			patch->point = patch->tpoint;
			patch->tpoint.length = 0;
		}
		smls->calcAllPolyMidPoints();
		smls->doBuild();
		smls->ctrf = true;
	}

	if(!point.length){
		g_visualClue=0;
		return;
	}
	setMathHiPresision();
	float inverseSize=(1/g_scene->getZoom());

	if(MultiSelect::pointList.length){
		if(MultiSelect::pointList[0].p->smls != smls){
			MultiSelect::pointList.length = 0;
			MultiSelect::beginDragPos.length = 0;
		}
	}

	if(g_visualClue == g_vcSelectRectGroup){
		IndexedLineSet* polyLine=dynamic_cast<IndexedLineSet*>(g_vcSelectRectShape->geometry.p);
		M_Vec3f& coord=dynamic_cast<Coordinate*>(polyLine->coord.p)->point;

		coord.length = 0;
		coord += g_scene->getIViewMatrix() * Vec3f(MultiSelect::topLeft.x, MultiSelect::bottomRight.y, MultiSelect::topLeft.z);
		coord += g_scene->getIViewMatrix() * MultiSelect::bottomRight;
		coord += g_scene->getIViewMatrix() * Vec3f(MultiSelect::bottomRight.x, MultiSelect::topLeft.y, MultiSelect::topLeft.z);
		coord += g_scene->getIViewMatrix() * MultiSelect::topLeft;
		fillPolyLineIndex(coord, polyLine->coordIndex.refValue());
		polyLine->coordIndex.refValue() += 0;
		polyLine->convertToRenderFormat(g_vcSelectRectShape->render);
		g_vcSelectRectGroup->children.refValue().length = 0;
		//g_vcBezierBase->children.refValue() += g_vcSelectRectShape.p;
	}
	genCage(smls);
	NurbsPatch* selPatch = dynamic_cast<NurbsPatch*>(g_selNode.getRaw()); //selected patch may have changed from this
	int uDim = selPatch->dim.x;
	int vDim = selPatch->dim.y;
	int idx = selPatch->index[0];

	M_Node& children = g_nurbsIlsBase->children;
	children.length = 1;

	for(int k = 0; k < MultiSelect::pointList.length; k++){
		Transform* transform=new Transform;
		transform->children.refValue() += g_redPoint.p;
		NurbsPatch* patch = MultiSelect::pointList[k].p;
		int i = MultiSelect::pointList[k].i;
		transform->scale=Vec3f(inverseSize, inverseSize, inverseSize);
		transform->translation.refValue() = patch->matrix * patch->dpoint[i];

		children += transform;
	}
	//if(smls->selCtrlPoly < smls->ctrlPolyPoint.length){
	//	int len = smls->ctrlPolyPoint[smls->selCtrlPoly].len;
	//	for(int q = len; --q >= 0;){
	//		NurbsPatch* patch =  smls->ctrlPolyPoint[smls->selCtrlPoly + q].p;
	//		int i =  smls->ctrlPolyPoint[smls->selCtrlPoly + q].i;
	//		if(isSelected(patch, i)){
	//			continue;
	//		}
	//		Transform* transform = new Transform;
	//		transform->children.refValue() += g_nextPoint.p;
	//		transform->scale=Vec3f(inverseSize, inverseSize, inverseSize);
	//		transform->translation.refValue() = patch->dpoint[i];
	//		children += transform;
	//	}
	//}
	if(g_visualClue == g_vcSelectRectGroup){
		g_vcSelectRectGroup->children.refValue() += g_nurbsIlsBase.p;
		g_vcSelectRectGroup->children.refValue() += g_vcSelectRectShape.p;
		if(g_showControlPolys && smls->ctrlPolyPoint.length){
			g_vcSelectRectGroup->children.refValue() += g_vcNurbsCtrlPoly.p;
		}
		if(MultiSelect::pointList.length > 3 && g_transformCP->sameScreenPosition == false){
			children += g_transformCP.p;
		}
	}
	else{
		g_visualClue = g_nurbsIlsBase;
		if(g_showControlPolys && smls->ctrlPolyPoint.length){
			children += g_vcNurbsCtrlPoly.p;
		}
		if(MultiSelect::pointList.length > 3 && g_transformCP->sameScreenPosition == false){
			children += g_transformCP.p;
		}
	}
	float tugRadius=g_settings->tugRadius;
	if(tugRadius > 0){
		IndexedLineSet* polyLine = dynamic_cast<IndexedLineSet*>(g_nurbsIlsShape->geometry.p);
		M_Vec3f& coord = dynamic_cast<Coordinate*>(polyLine->coord.p)->point.refValue();
		M_Color3f& color = dynamic_cast<Color*>(polyLine->color.p)->color.refValue();
		int mark = coord.length;
		int numOfPoints=129;
		coord.length+=numOfPoints;
		color.length+=numOfPoints;

		Vec3f location = dpoint[index[0]];
		Matrix matrix = rotationMatrix( -g_scene->rotation) * scaleMatrix(Vec3f(tugRadius,tugRadius,tugRadius)) * translationMatrix(location);

		for(int i=0; i<numOfPoints; i++){
			coord[mark+i]= matrix * Vec3f(cos(i*(PI*2/(numOfPoints-1))), sin(i*(PI*2/(numOfPoints-1))),0);
			color[mark+i]= Color3f(0,1,0);
		}
		fillPolyLineIndex(coord, mark, polyLine->coordIndex.refValue());

		Transform* transform=new Transform;
		transform->children.refValue() += g_redPoint.p;
		children+=transform;

		transform->translation = matrix * Vec3f(1, 0, 0);
		transform->scale=Vec3f(inverseSize, inverseSize, inverseSize);
		polyLine->convertToRenderFormat(g_nurbsIlsShape->render);
	}
	return;
	/*
	}
	else{
		int uDim=this->dim.x;
		int vDim=this->dim.y;
		int idx = index[0];
		IndexedLineSet* polyLine = dynamic_cast<IndexedLineSet*>(g_vcEllipticalPolyline2D->geometry.p);
		M_Vec3f& coord = dynamic_cast<Coordinate*>(polyLine->coord.p)->point.refValue();
		coord.length = 0;
		M_Int& coordIndex = polyLine->coordIndex.refValue();
		coordIndex.length = 0;
		addCage(polyLine);

		M_Node& children=g_vcBezierBase->children;
		children.length=1;
		Transform* transform=new Transform;
		transform->children.refValue() += g_redPoint.p;
		children+=transform;

		Vec2i idx2i = Vec2i(idx % uDim, idx / uDim);

		Vec3f pt = matrix * dpoint[idx2i.y*uDim+idx2i.x];
		transform->translation = pt;
		transform->scale=Vec3f(inverseSize, inverseSize, inverseSize);

		if(ringTug){
			if(MultiSelect::beginDragPos.length==0){
				MultiSelect::beginDragPos = dpoint;
			}
			transform=new Transform;
			transform->children.refValue() += g_yellowPoint.p;
			children+=transform;

			if(tugRingPivotPoint){
				transform->translation = g_scene->getIViewMatrix() * g_pointer.translation;
			}
			else{
				transform->translation = matrix * calcRingPivotPoint(getSelRing());
			}
			
			transform->scale=Vec3f(inverseSize, inverseSize, inverseSize);
		}

		
	///////////
		g_vcBezierBase->rotation=Rotation();

		g_visualClue=g_vcBezierBase;
	//	g_scene->children.refValue()+=g_vcBezierBase.p;
	//	g_stw->genNodeList();


		//g_vcEllipticalMaterial->diffuseColor=g_settings->backgroundColor.refValue().invert();
		polyLine->convertToRenderFormat(g_vcEllipticalPolyline2D->render);
	}
	*/
	restoreMathPresision();
}
//#include <crtdbg.h>
//static _CrtMemState s1, s2, s3;



void Bezier::bezHouseKeep(){
	if(!point.length){
		g_visualClue=0;
		return;
	}
	setMathHiPresision();
	float inverseSize=(1/g_scene->getZoom());
	int idx=index[0];

	IndexedLineSet* polyLine=dynamic_cast<IndexedLineSet*>(g_vcEllipticalPolyline2D->geometry.p);

	M_Vec3f& coord=dynamic_cast<Coordinate*>(polyLine->coord.p)->point;
	coord.length=line.length+point.length;//normA.length*5;
	for(int i=0; i<line.length; i++){
		coord[i]=matrix*Vec3f(line[i].x, line[i].y, 0);
	}
	for(int j=line.length, i=point.length; --i>=0; j++){
		coord[j]=matrix*Vec3f(point[i].x, point[i].y, 0);
	}
	/*
	for(int j=point.length+line.length, i=normA.length; --i>=0; ){
		coord[j++]=Vec3f(normA[i].x, normA[i].y, 0);
		coord[j++]=Vec3f(normB[i].x, normB[i].y, 0);
		coord[j++]=Vec3f(normC[i].x, normC[i].y, 0);
		coord[j++]=Vec3f(normD[i].x, normD[i].y, 0);
		coord[j++]=Vec3f(normA[i].x, normA[i].y, 0);
	}
	*/
	fillPolyLineIndex(coord, polyLine->coordIndex.refValue());
/////////////
//	_CrtMemCheckpoint( &s1 );

	M_Node& children=g_vcBezierBase->children;
	children.length=1;
	int childDiff=point.length-children.length+1;

	if(childDiff<0){
		children.length+=childDiff;
	}
	else{
		for(int i=childDiff; --i>=0;){
			Transform* transform=new Transform;
			transform->children.refValue()+=g_nextPoint.p; //could just add null?
			children+=transform;
		}
	}
	for(int i=point.length; --i>=0;){
		Transform* transform=dynamic_cast<Transform*>(children[i+1].p);
		Vec3f pt=matrix*Vec3f(point[i].x, point[i].y, 0);
		transform->translation=pt;

		if(i==idx){
			transform->children.refValue()[0]=g_redPoint;
		}
		else{
			if(fix[i]){
				transform->children.refValue()[0]=g_cyanPoint;
			}
			else{
				transform->children.refValue()[0]=g_nextPoint;
			}
		}

		transform->scale=Vec3f( inverseSize, inverseSize, inverseSize);
	}
//	_CrtMemCheckpoint( &s2 );
//	int flag = _CrtMemDifference(&s3, &s1, &s2);

///////////
	if(dynamic_cast<BezierTape*>(this)){
		g_vcBezierBase->rotation=Rotation(0,1,0,-PI*.5);
	}
	else{
		g_vcBezierBase->rotation=Rotation();
	}

	g_visualClue=g_vcBezierBase;

	restoreMathPresision();
	//g_vcEllipticalMaterial->diffuseColor=g_settings->backgroundColor.refValue().invert();
	polyLine->convertToRenderFormat(g_vcEllipticalPolyline2D->render);

//	dLog(flag);
}
void Bezier3f::bezHouseKeep(){
	if(!point.length){
		g_visualClue=0;
		return;
	}
	setMathHiPresision();
	float inverseSize=(1/g_scene->getZoom());
	int idx=index[0];

	IndexedLineSet* polyLine=dynamic_cast<IndexedLineSet*>(g_vcEllipticalPolyline2D->geometry.p);

	M_Vec3f& coord=dynamic_cast<Coordinate*>(polyLine->coord.p)->point;
	coord.length=line.length+point.length;//normA.length*5;
	for(int i=0; i<line.length; i++){
		coord[i]=matrix*Vec3f(line[i].x, line[i].y, line[i].z);
	}
	for(int j=line.length, i=point.length; --i>=0; j++){
		coord[j]=matrix*Vec3f(point[i].x, point[i].y, point[i].z);
	}
	fillPolyLineIndex(coord, polyLine->coordIndex.refValue());
/////////////
	M_Node& children=g_vcBezierBase->children;
	children.length=1;
	int childDiff=point.length-children.length+1;

	if(childDiff<0){
		children.length+=childDiff;
	}
	else{
		for(int i=childDiff; --i>=0;){
			Transform* transform=new Transform;
			transform->children.refValue()+=g_nextPoint.p; //could just add null?
			children+=transform;
		}
	}
	for(int i=point.length; --i>=0;){
		Transform* transform=dynamic_cast<Transform*>(children[i+1].p);
		Vec3f pt=matrix*point[i];
		transform->translation=pt;

		if(i==idx){
			transform->children.refValue()[0]=g_redPoint;
		}
		else{
			if(fix[i]){
				transform->children.refValue()[0]=g_cyanPoint;
			}
			else{
				transform->children.refValue()[0]=g_nextPoint;
			}
		}

		transform->scale=Vec3f( inverseSize, inverseSize, inverseSize);
	}

///////////
	g_vcBezierBase->rotation=Rotation();

	g_visualClue=g_vcBezierBase;

	restoreMathPresision();
	//g_vcEllipticalMaterial->diffuseColor=g_settings->backgroundColor.refValue().invert();
	polyLine->convertToRenderFormat(g_vcEllipticalPolyline2D->render);
}
void BezierLathe::houseKeep(){
	bezHouseKeep();
}
void BezierLatheBuild::houseKeep(){
	bezHouseKeep();
}
void BezierScale::houseKeep(){
	bezHouseKeep();
}
void BezierCircumScale::houseKeep(){
	bezHouseKeep();
}
void BezierShift::houseKeep(){
	bezHouseKeep();
}
void BezierTwist::houseKeep(){
	bezHouseKeep();
}
void BezierTape::houseKeep(){
	bezHouseKeep();
}
void BezierBend::houseKeep(){
	bezHouseKeep();
}
Seamless* getSeamless(int& selLine);
ModelMorpher* getModelMorpher(){
	int selLine = g_stw->selLine;
	if(Seamless* selSmls = getSeamless(selLine)){
		selLine = getParent(selLine);
		if(ModelSequence* modelSequence = dynamic_cast<ModelSequence*>(g_sceneTreeList[selLine].node.p)){
			return dynamic_cast<ModelMorpher*>(g_sceneTreeList[getParent(selLine)].node.p);
		}
	}
	return 0;
}


CommonSpec* getCommonSpec(const Art* part);

void PartCageClue::genCage(){
	if(smls == 0){
		return;
	}
	M_Vec3f& point = dynamic_cast<Coordinate*>(g_nurbsIlsCoord)->point;

	point.length = 0;
	M_Int& coordIndex = g_nurbsIls->coordIndex.refValue();
	coordIndex.length = 0;

	M_<NurbsPatch*>& cpNodeList = smls->cpNodeList;

	g_nurbsIlsCoord->point.refValue().length = smls->pointTotal;
	M_Color3f& m_color = g_nurbsIlsColor->color.refValue();
	m_color.length = smls->pointTotal;


	for(int k = 0; k < cpNodeList.length; k++){
		NurbsPatch* patch = cpNodeList[k];
		patch->addCage(g_nurbsIls);
	}

	for(int i = m_color.length; --i >= 0;){
		m_color[i] = Color3f(.5, .5, .5);
	}
	M_Art& partList = smls->partList;

	for(int k = partList.length; --k >= 0;){

		Art* part = partList[k];
		const M_Int& coordGroup=part->coordGroup;
		const M_ControlPoint& controlPoint = part->controlPoint;

		if(const CommonSpec* commonSpec = getCommonSpec(part)){
			int k = 0;
			Color3f groupColor;
			if(commonSpec->color.length){

				for(int cg = 0; cg < coordGroup.length; cg++){
					if(cg < commonSpec->color.length){
						groupColor = commonSpec->color[cg];
					}
					for(int c = coordGroup[cg]; --c >= 0 && k < controlPoint.length; k++){
						const ControlPoint& cp = controlPoint[k];
						int idx = cp.i + cp.p->fromPoint;
						m_color[idx] = groupColor;
					}
				}
			}
			if(commonSpec->color.length){
				groupColor = commonSpec->color[commonSpec->color.length-1];
			}
			for(; k < part->controlPoint.length; k++){
				const ControlPoint& cp = controlPoint[k];
				int idx = cp.i + cp.p->fromPoint;
				m_color[idx] = groupColor;
			}
		}
	}
	smls->addCageControlPolys();
	g_nurbsIls->convertToRenderFormat(g_nurbsIlsShape->render);
}

void houseKeepAnim(Art* part, Anim* anim){
	g_vcArtAnimBase->dragTranPointMode = false;
	g_vcArtAnimBase->anim = anim;

	IndexedLineSet* polyLine = dynamic_cast<IndexedLineSet*>(g_vcEllipticalPolyline2D->geometry.p);
	M_Vec3f& coord = dynamic_cast<Coordinate*>(polyLine->coord.p)->point;
	coord.length = 0;
	coord += Vec3f();
	coord += Vec3f(1, 0, 0);
	fillPolyLineIndex(coord, polyLine->coordIndex.refValue());
	g_vcArtAnimBase->rotCtrlPnt = part->pivotPointMatrix * Vec3f(0, 0, 0) + Vec3f(.5, 0, 0);
	polyLine->convertToRenderFormat(g_vcEllipticalPolyline2D->render);
}

void setAllIns(M_Build& m_build, const bool inState);
void storePoints(M_<NurbsPatch*>& cpNodeList);

void Art::houseKeep(){

	if(smls->animCP && smls->workAnimCP == false){
		M_<NurbsPatch*>& cpNodeList = smls->cpNodeList;
		if(smls->workAnimCP == false){
			setAllIns(smls->build.refValue(), false);
			smls->workAnimCP = true;
			storePoints(cpNodeList);
			smls->doBuild();
			smls->ctrf = true;
		}
	}

	float inverseSize=(1/g_scene->getZoom());
	g_vcArtTriBase->smls = smls;
	Anim* anim=Window::animMode();
	g_nurbsIlsBase->children.refValue().length = 1;
	g_partCageClue->cpTug = false;

	if(MultiSelect::selectRectMode){
		IndexedLineSet* polyLine=dynamic_cast<IndexedLineSet*>(g_vcSelectRectShape->geometry.p);
		M_Vec3f& coord = dynamic_cast<Coordinate*>(polyLine->coord.p)->point;

		coord.length = 0;
		Matrix iViewMatrix;
		if(anim == 0 || smls->animCP){
			iViewMatrix = g_scene->getIViewMatrix();
		}
		coord+=iViewMatrix * Vec3f(MultiSelect::topLeft.x, MultiSelect::bottomRight.y, MultiSelect::topLeft.z);
		coord+=iViewMatrix * MultiSelect::bottomRight;
		coord+=iViewMatrix * Vec3f(MultiSelect::bottomRight.x, MultiSelect::topLeft.y, MultiSelect::topLeft.z);
		coord+=iViewMatrix * MultiSelect::topLeft;
		fillPolyLineIndex(coord, polyLine->coordIndex.refValue());
		polyLine->coordIndex.refValue() += 0;
		polyLine->convertToRenderFormat(g_vcSelectRectShape->render);
		if(smls->workAnimCP){
			g_nurbsIlsBase->children.refValue() += g_vcSelectRectShape.p;
		}
	}
	if(smls->workAnimCP){
		g_partCageClue->smls = smls;
		g_visualClue = g_partCageClue;
		g_partCageClue->genCage();
		if(anim == 0){
			g_visualClue = g_partCageClue;
			return;
		}
	}
	if(anim && getModelMorpher() == 0){
		houseKeepAnim(this, anim);
		g_visualClue = g_vcArtAnimBase;
	}
	else{
		if(MultiSelect::coordList.length){
			Seamless* smls = g_vcArtTriBase->transpSmls;
			Art* part = smls->skeleton.refValue()[0];
		}
		float tugRadius = g_settings->tugRadius * g_scene->scale.x;
		if(dynamic_cast<ImmediatePart*>(this) && tugRadius>0){
			IndexedLineSet* polyLine=dynamic_cast<IndexedLineSet*>(g_vcEllipticalPolyline2D->geometry.p);
			M_Vec3f& coord=dynamic_cast<Coordinate*>(polyLine->coord.p)->point;
			int numOfPoints=129;
			coord.length=numOfPoints;
			for(int i=0; i<numOfPoints; i++){
				coord[i]=Vec3f(tugRadius*cos(i*(PI*2/(numOfPoints-1))), tugRadius*sin(i*(PI*2/(numOfPoints-1))),0);
			}
			g_immediateRadialTugResizeLocation = Vec3f(tugRadius,0,0);

			fillPolyLineIndex(coord, polyLine->coordIndex.refValue());
			//g_vcEllipticalMaterial->diffuseColor=g_settings->backgroundColor.refValue().invert();
			polyLine->convertToRenderFormat(g_vcEllipticalPolyline2D->render);
			g_vcArtTriBase->circle=true;
		}
		else{
			g_vcArtTriBase->circle=false;
		}
		

		g_vcArtTriBase->smls = smls;
		g_visualClue = g_vcArtTriBase;
	}
}

void CenterJoint::houseKeep(){
	g_visualClue = g_vcCenterJointBase;
}
void TransferOwnership::houseKeep(){
	g_visualClue = g_transferOwnershipClue;
}

void RingBuild::houseKeep(){
	int i=index[0];
	if(i< scaleX.length){
		Matrix matrix;
		getGoodQuadrant(matrix);
		g_vcRingBuildOrientationA->translation=matrix * Vec3f(-scaleX.refValue()[i].x,0,0);
		g_vcRingBuildOrientationB->translation=matrix * Vec3f(scaleX.refValue()[i].y,0,0);
		g_vcRingBuildOrientationC->translation=matrix * Vec3f(0,0,-scaleZ.refValue()[i].x);
		g_vcRingBuildOrientationD->translation=matrix * Vec3f(0,0,scaleZ.refValue()[i].y);
//		}
//		else{
//			g_vcRingBuildOrientationA->translation=matrix * Vec3f(0,0,-scaleZ.refValue()[i].x);
//			g_vcRingBuildOrientationB->translation=matrix * Vec3f(0,0,scaleZ.refValue()[i].y);
//		}
		float inverseSize=(1/g_scene->getZoom());
		g_vcRingBuildPivotPoint->scale=Vec3f(inverseSize, inverseSize, inverseSize);
		g_vcRingBuildOrientationA->scale=Vec3f(inverseSize, inverseSize, inverseSize);
		g_vcRingBuildOrientationB->scale=Vec3f(inverseSize, inverseSize, inverseSize);
		g_vcRingBuildOrientationC->scale=Vec3f(inverseSize, inverseSize, inverseSize);
		g_vcRingBuildOrientationD->scale=Vec3f(inverseSize, inverseSize, inverseSize);
		g_visualClue = g_vcRingBuildBase;
	}
}
void ColorSweep::houseKeep(){
	if(!distance.length){
		g_visualClue=0;
		return;
	}
	setMathHiPresision();
	int idx=index[0];
	IndexedLineSet* polyLine=dynamic_cast<IndexedLineSet*>(g_vcEllipticalPolyline2D->geometry.p);
	M_Vec3f& coord=dynamic_cast<Coordinate*>(polyLine->coord.p)->point;
	M_Int& coordIndex = polyLine->coordIndex.refValue();
	coordIndex.length=0;

	M_Float& distance = this->distance.refValue();

	coord.length=0;
	coord.length=2;

	M_Node& children=g_vcBezierBase->children;
	children.length=0;
	children+=g_vcEllipticalPolyline2D;
	int childDiff=color.length-children.length+1;

	if(childDiff<0){
		children.length+=childDiff;
	}
	else{
		for(int i=childDiff; --i>=0;){
			Transform* transform=new Transform;
			transform->children.refValue()+=g_nextPoint.p; //could just add null?
			children+=transform;
		}
	}
	float inverseSize=(1/g_scene->getZoom());
	float sum=0;
	for(int i = 0; i<color.length; i++){
		Transform* transform=dynamic_cast<Transform*>(children[i+1].p);
		Vec3f pt=matrix*Vec3f(0, sum, 0);
		transform->translation=pt;
		if(i==0 ){
			coord[0] = pt;
		}

		if(i==idx){
			transform->children.refValue()[0]=g_redPoint;
		}
		else{
			transform->children.refValue()[0]=g_nextPoint;
		}

		transform->scale=Vec3f(inverseSize, inverseSize, inverseSize);

		if(i==distance.length){
			coord[1] = pt;
		}
		else{
			sum+=distance[i];
		}
	}
	fillPolyLineIndex(coord, polyLine->coordIndex.refValue());

///////////
	g_vcBezierBase->rotation=Rotation();

	g_visualClue=g_vcBezierBase;

	restoreMathPresision();
	//g_vcEllipticalMaterial->diffuseColor=g_settings->backgroundColor.refValue().invert();
	polyLine->convertToRenderFormat(g_vcEllipticalPolyline2D->render);
}
void EC2::ecHouseKeep(){
	if(!point.length)return;
	setMathHiPresision();
	float inverseSize=(1/g_scene->getZoom());
	g_vcECScaleTangentArrow->scale=Vec3f( inverseSize, inverseSize, inverseSize);

	int idx=index[0];
	processFields();

	IndexedLineSet* polyLine=dynamic_cast<IndexedLineSet*>(g_vcEllipticalPolyline2D->geometry.p);

	M_Vec3f& line=dynamic_cast<Coordinate*>(polyLine->coord.p)->point;

	genCurve( line);
	fillPolyLineIndex(line, polyLine->coordIndex.refValue());

	M_Node& children=g_vcECScaleBase->children;

	int childDiff=point.length-children.length+2;

	if(childDiff<0){
		children.length+=childDiff;
	}
	else{
		for(int i=childDiff; --i>=0;){
			Transform* transform=new Transform;
			transform->children.refValue()+=g_nextPoint.p; //could just add null?
			children+=transform;
		}
	}
	for(int i=point.length; --i>=0;){
		Transform* transform=dynamic_cast<Transform*>(children[i+2].p);
		Vec3f pt=Vec3f(point[i].x, point[i].y, 0);
		transform->translation=pt;

		if(i==idx){
			transform->children.refValue()[0]=g_redPoint;
			g_vcECScaleTangentArrow->translation=pt;
			g_vcECScaleTangentArrow->rotation=Rotation(0,0,1,tangent[i]);
		}
		else{
			transform->children.refValue()[0]=g_nextPoint;
		}
		
		transform->scale=Vec3f( inverseSize, inverseSize, inverseSize);
	}

	g_vcECScaleBase->translation=location;
	g_vcECScaleBase->rotation=orientation;

	g_visualClue=g_vcECScaleBase;

	restoreMathPresision();

	//g_vcEllipticalMaterial->diffuseColor=g_settings->backgroundColor.refValue().invert();
	polyLine->convertToRenderFormat(g_vcEllipticalPolyline2D->render);
}
void EC3::ecHouseKeep(){
	if(!point.length)return;
	setMathHiPresision();
	float inverseSize=(1/g_scene->getZoom());
	g_vcECScaleTangentArrow->scale=Vec3f( inverseSize, inverseSize, inverseSize);

	int idx=index[0];
	processFields();

	IndexedLineSet* polyLine=dynamic_cast<IndexedLineSet*>(g_vcEllipticalPolyline2D->geometry.p);

	M_Vec3f& line=dynamic_cast<Coordinate*>(polyLine->coord.p)->point;

	genCurve( line);
	fillPolyLineIndex(line, polyLine->coordIndex.refValue());

	M_Node& children=g_vcECScaleBase->children;

	int childDiff=point.length-children.length+2;

	if(childDiff<0){
		children.length+=childDiff;
	}
	else{
		for(int i=childDiff; --i>=0;){
			Transform* transform=new Transform;
			transform->children.refValue()+=g_nextPoint.p; //could just add null?
			children+=transform;
		}
	}
	for(int i=point.length; --i>=0;){
		Transform* transform=dynamic_cast<Transform*>(children[i+2].p);
		Vec3f pt=Vec3f(point[i].x, point[i].y, 0);
		transform->translation=pt;

		if(i==idx){
			transform->children.refValue()[0]=g_redPoint;
			g_vcECScaleTangentArrow->translation=pt;
			g_vcECScaleTangentArrow->rotation=Rotation(0,0,1,tangent[i]);
		}
		else{
			transform->children.refValue()[0]=g_nextPoint;
		}
		
		transform->scale=Vec3f( inverseSize, inverseSize, inverseSize);
	}


/*
	{
		Rotation tangRot=Rotation(0,0,1,PI*.5);
		for( int i=1; i<pointX.length; i++){
			tangRot=Rotation(0,1,0,-yAngle[i-1])*Rotation(0,0,1,relTangent[i])*tangRot;

			if(i==idx){
				g_vcECScaleTangentArrow->rotation=tangRot;
				break;
			}
		}
	}
*/
	g_vcECScaleBase->translation=location;
	g_vcECScaleBase->rotation=orientation;

	g_visualClue=g_vcECScaleBase;

	restoreMathPresision();

	//g_vcEllipticalMaterial->diffuseColor=g_settings->backgroundColor.refValue().invert();
	polyLine->convertToRenderFormat(g_vcEllipticalPolyline2D->render);
}

void ECLathe::houseKeep(){
	ecHouseKeep();
}
void ECScale::houseKeep(){
	ecHouseKeep();
}
void ECShift::houseKeep(){
	ecHouseKeep();
}
void ECSyncBend::houseKeep(){
	ecHouseKeep();
}
void EvolveCoord::houseKeep(){
	ecHouseKeep();
}
/*
void Progressor::houseKeep(){
	ecHouseKeep();
}
*/
void ECBend::houseKeep(){
	ecHouseKeep();
}
void ECTwist::houseKeep(){
	ecHouseKeep();
}

void CCBend::houseKeep(){

	if(!distance.length){
		return;
	}
	int index=this->index[0];
	if(index==0){
		index=1;
	}

	g_vcCCLathe->distance.refValue()[0]=distance[index-1];
	if(index<=1){
		g_vcSurfaceTransform->translation=Vec3f();
	}
	else {
		g_vcSurfaceTransform->translation=Vec3f(0,getDistance(index-2),0);
	}
	g_vcCCBend->distance.resize(distance.length);
	g_vcCCBend->yAngle.resize(curve.length);
	g_vcCCBend->curve.resize(curve.length);

	g_vcCCBend->distance.refValue()=distance;
	g_vcCCBend->curve.refValue()=curve;
	g_vcCCBend->yAngle.refValue()=yAngle;

	g_vcCCLathe->baseRad.refValue()=g_spaghettiRad*(1/g_scene->getZoom());

	g_vcBaseRad->translation=location+Vec3f(baseRad,0,0);
	g_vcBaseRad->rotation=orientation;

	commonEndCCVisualClue();
}
void CCTranslate::houseKeep(){
	if(!distance.length){
		return;
	}
	int index=this->index[0];
	if(index==0){
		index=1;
	}

	g_vcCCLathe->distance.refValue()[0]=distance[index-1];
	if(index<=1){
		g_vcSurfaceTransform->translation=Vec3f();
	}
	else {
		g_vcSurfaceTransform->translation=Vec3f(0,getDistance(index-2),0);
	}

	g_vcCCBend->distance.resize(distance.length);
	g_vcCCBend->yAngle.resize(curve.length);
	g_vcCCBend->curve.resize(curve.length);
	g_vcCCBend->distance.refValue()=distance;
	g_vcCCBend->curve.refValue()=curve;

	g_vcCCLathe->baseRad.refValue()=g_spaghettiRad*(1/g_scene->getZoom());

	g_vcBaseRad->translation=Vec3f(baseRad+location.x,location.y,location.z);
	g_vcBaseRad->rotation=orientation;
	commonEndCCVisualClue();
}
float getDistance(int index, const M_Float& distance){
	float distanceSum=0;
	int i;
	for(i=0;i<=index;i++){
		distanceSum+=distance[i];
	}
	return distanceSum;
}
void SineTug::houseKeep(){
	if(!distance.refValue().length){
		return;
	}
	setMathHiPresision();

	int index=this->index[0];
	int numOf = this->radius.refValue().length;

	float radius=this->radius[index];
	const M_Float& distance=this->distance;
	float xScale=this->xScale[index];

	IndexedLineSet* polyLine=dynamic_cast<IndexedLineSet*>(g_vcEllipticalPolyline2D->geometry.p);
	M_Vec3f& coord=dynamic_cast<Coordinate*>(polyLine->coord.p)->point;
	coord.length=2002;
	Matrix matrix = scaleMatrix(Vec3f(radius*xScale,radius,radius)) * rotationMatrix(orientation.refValue()[index])*translationMatrix(location.refValue()[index]);
	for(int i=0; i<1000; i++){
		coord[i] = matrix * Vec3f(cos(i/999.0*(PI*2)), 0, sin(i/999.0*(PI*2)));
	}
	
	coord[1000]=matrix*Vec3f(0,0,0);
	
	matrix = scaleMatrix(Vec3f(radius*xScale,radius,radius)) * translationMatrix(Vec3f(0,distance[index]*8,0)) * rotationMatrix(orientation.refValue()[index])*translationMatrix(location.refValue()[index]);

	coord[1001]=matrix*Vec3f(0,0,0);

	M_Int& coordIndex = polyLine->coordIndex;
	coordIndex.length = 2004;

	for(int i = 0; i<1000; i++){
		coordIndex[i] = i;
	}
	coordIndex[1000]=-1;
	coordIndex[1001]=1000;
	coordIndex[1002]=1001;
	coordIndex[1003]=-1;

	matrix = scaleMatrix(Vec3f(radius*xScale,radius,radius)) * translationMatrix(Vec3f(0,-decay[index],0)) * rotationMatrix(orientation.refValue()[index])*translationMatrix(location.refValue()[index]);
	for(int i=0; i<1000; i++){
		coord[1002+i] = matrix * Vec3f(cos(i/999.0*(PI*2)), 0, sin(i/999.0*(PI*2)));
	}
	for(int i = 0; i<1000; i++){
		coordIndex[1004+i] = 1002+i;
	}
	
	float inverseSize = (1/g_scene->getZoom());
	M_Node& children=g_vcBezierBase->children;
	children.length=1;
	int childDiff=numOf-children.length+1;

	if(childDiff<0){
		children.length+=childDiff;
	}
	else{
		for(int i=childDiff; --i>=0;){
			Transform* transform=new Transform;
			transform->children.refValue()+=g_nextPoint.p; //could just add null?
			children+=transform;
		}
	}
	for(int i=numOf; --i>=0;){
		Transform* transform=dynamic_cast<Transform*>(children[i+1].p);
		Vec3f pt=orientation[i]*Vec3f(0,distance[i]*8,0)+location[i];
		transform->translation=pt;

		if(i==index){
			transform->children.refValue()[0]=g_redPoint;
		}
		else{
			transform->children.refValue()[0]=g_nextPoint;
		}

		transform->scale=Vec3f( inverseSize, inverseSize, inverseSize);
	}

	g_vcBezierBase->rotation=Rotation();
	g_visualClue=g_vcBezierBase;

	polyLine->convertToRenderFormat(g_vcEllipticalPolyline2D->render);

	restoreMathPresision();
}
void SimulSineTug::houseKeep(){
}
void RadialTug::houseKeep(){
	if(!location.refValue().length){
		return;
	}
	setMathHiPresision();

	int index=this->index[0];

	float radius=this->radius[index];
	Vec3f displacement=this->displacement[index];

	IndexedLineSet* polyLine=dynamic_cast<IndexedLineSet*>(g_vcEllipticalPolyline2D->geometry.p);
	M_Vec3f& coord=dynamic_cast<Coordinate*>(polyLine->coord.p)->point;
	coord.length=1002;
	
	Matrix matrix = rotationMatrix( -g_scene->rotation) * scaleMatrix(Vec3f(radius,radius,radius)) * translationMatrix(location.refValue()[index]);
	for(int i=0; i<1000; i++){
		coord[i] = matrix * Vec3f(cos(i/999.0*(PI*2)), sin(i/999.0*(PI*2)),0);
	}
	
	coord[1000]=matrix*Vec3f(0,0,0);
	
	matrix = translationMatrix(location.refValue()[index]+displacement);

	coord[1001]=matrix*Vec3f(0,0,0);

	M_Int& coordIndex = polyLine->coordIndex;
	coordIndex.length = 1003;

	for(int i = 0; i<1000; i++){
		coordIndex[i] = i;
	}
	coordIndex[1000]=-1;
	coordIndex[1001]=1000;
	coordIndex[1002]=1001;
	
	g_vcBezierBase->rotation=Rotation();
	g_visualClue=g_vcBezierBase;

	polyLine->convertToRenderFormat(g_vcEllipticalPolyline2D->render);

	restoreMathPresision();
}
void TransferBCyl::houseKeep(){
	if(in){
		g_vcTransferBCylCylinderBuild->height.refValue()=height;
		g_vcTransferBCylCylinderBuild->radius.refValue()=radius;

		M_Node& children = g_vcTransferBCylBase->children.refValue();

		IndexedLineSet* polyLine=dynamic_cast<IndexedLineSet*>(g_vcEllipticalPolyline2D->geometry.p);
		M_Vec3f& coord=dynamic_cast<Coordinate*>(polyLine->coord.p)->point;
		coord.length=0;
		M_Int& coordIndex = polyLine->coordIndex.refValue();
		coordIndex.length=0;

		if(groupingHeight>0){
			coord+=Vec3f(-radius, groupingHeight, 0);
			coord+=Vec3f(radius, groupingHeight, 0);
			coord+=Vec3f(0, groupingHeight, -radius);
			coord+=Vec3f(0, groupingHeight, radius);
			coordIndex+=0;
			coordIndex+=1;
			coordIndex+=-1;
			coordIndex+=2;
			coordIndex+=3;
			coordIndex+=-1;
			polyLine->renderProcessed=false;
		}
		g_vcTransferBCylBase->rotation=orientation;
		g_vcTransferBCylBase->translation=location;

		g_vcTransferBCylCylinderBuild->processFields();
		g_vcTransferBCyl->process();
		g_vcTransferBCyl->doBuild();
		g_vcTransferBCyl->convertToRenderFormat();

		g_visualClue=g_transferBCylClue;
	}
	else{
		g_visualClue=0;
	}
}
void houseKeepVisualClue(){
	if(VisualAid* visualAid=dynamic_cast<VisualAid*>(g_selNode.getRaw())){
		visualAid->houseKeep();
/*
		if(g_visualClue){
			String str;
			g_visualClue->toString(str);
			dLog(str);
		}
*/
	}
	else{
		g_visualClue=nullPtr;
	}
}
void houseKeepVisualClueViewChange(){
	if(RingBuild* ringBuild=dynamic_cast<RingBuild*>(g_selNode.getRaw())){
		ringBuild->houseKeep();
	}
	else if(NurbsPatch* patch = dynamic_cast<NurbsPatch*>(g_selNode.getRaw())){
		if(g_settings->tugRadius > 0){
			patch->houseKeep();
		}
	}
}
void ColorEffect::breakColorJoin(Seamless& smls){
	Art* triPart=smls.actTriPart;
	M_Color3f& color=this->color.refValue();
	color+=color[triPart->triangle.refTexIndex(smls.ActTri*3+smls.ActTriCorner)];
	triPart->triangle.refTexIndex(smls.ActTri*3+smls.ActTriCorner)=color.length-1;
	smls.convertToRenderFormat();
	BackUp::monostable.beginCycle();
}
void MonoColorEffect::breakColorJoin(Seamless& smls){
	Art* triPart=smls.actTriPart;
	M_Color3f& color=this->color.refValue();
	color+=color[triPart->triangle.refTexIndex(smls.ActTri*3+smls.ActTriCorner)];
	triPart->triangle.refTexIndex(smls.ActTri*3+smls.ActTriCorner)=color.length-1;
	smls.convertToRenderFormat();
	BackUp::monostable.beginCycle();
}
void TextureEffect::breakColorJoin(Seamless& smls){
	Art* triPart=smls.actTriPart;
	M_Vec2f& texCoord=this->texCoord.refValue();
	texCoord+=texCoord[triPart->triangle.refTexIndex(smls.ActTri*3+smls.ActTriCorner)];
	triPart->triangle.refTexIndex(smls.ActTri*3+smls.ActTriCorner)=texCoord.length-1;
	smls.convertToRenderFormat();
	BackUp::monostable.beginCycle();
}
void findThinTriangle(Seamless* smls){
	float triThinAmount=.0005;
	M_Art& partList=smls->partList;
	for(int k=partList.length; --k>=0;){
		ArtPtr part=partList[k];
		M_Triangle& triangle=part->triangle;
		for(int i=triangle.length*3; --i>=0;){
			Art* p;
			p=triangle.refPart(i);
			Vec3f a=p->matrix*p->coord.refValue()[triangle.refIndex(i)];

			p=triangle.refPart(--i);
			Vec3f b=p->matrix*p->coord.refValue()[triangle.refIndex(i)];

			p=triangle.refPart(--i);
			Vec3f c=p->matrix*p->coord.refValue()[triangle.refIndex(i)];

			if(
				fabs(a.x-b.x)<triThinAmount&&
				fabs(a.y-b.y)<triThinAmount&&
				fabs(a.z-b.z)<triThinAmount||

				fabs(a.x-c.x)<triThinAmount&&
				fabs(a.y-c.y)<triThinAmount&&
				fabs(a.z-c.z)<triThinAmount||

				fabs(b.x-c.x)<triThinAmount&&
				fabs(b.y-c.y)<triThinAmount&&
				fabs(b.z-c.z)<triThinAmount
				){
				smls->ActTri=i/3;
				smls->ActCtrlPnt=triangle.refIndex(smls->ActTri*3);// corner a
				g_stw->setSelNodeLine(triangle.refPart(smls->ActTri*3));// corner a
				smls->ActTriCorner=0;// corner a
				smls->actTriPart=part;
				checkVertexBelongsToTriangle(dynamic_cast<Art*>(g_selNode.getRaw()), smls->actTriPart, smls->ActCtrlPnt, smls->ActTriCorner, smls->ActTri);
				makePointerEqualIndexLocation();
				return;
			}
		}
	}
}
void searchDoBuild(M_Node& children){
	for(int i=0;i<children.length;i++){
		Node* node=children[i];
		if(node->getType()==SEAMLESS){
			Seamless* smls=dynamic_cast<Seamless*>(node);
			smls->doBuild();
			smls->convertToRenderFormat();
		}
		else if(dynamic_cast<AGroup*>(node)!=0){
			searchDoBuild(dynamic_cast<AGroup*>(node)->children.refValue());
		}
		else if(node->getType()==PROTODECLARE){
			searchDoBuild(dynamic_cast<ProtoDeclare*>(node)->protoBody->children.refValue());
		}
	}
}
void CaretWindow::copyToClipboard(bool cut){
	if(highlightPos<0)return;
	if(	caretPos!=highlightPos){
		if (OpenClipboard(hWndMain)){
			EmptyClipboard(); 
	
			// Allocate a global memory object for the text. 
		
			int highlightLen=abs(caretPos-highlightPos);
			int index=highlightPos;
			if(caretPos<highlightPos){
				index=caretPos;
			}

			HGLOBAL hglbCopy = GlobalAlloc(GMEM_MOVEABLE, abs(caretPos-highlightPos) + 1);
			if (hglbCopy == NULL) 
			{ 
				CloseClipboard(); 
			}
			else{
		
				// Lock the handle and copy the text to the buffer. 
			
				LPVOID lptstrCopy = GlobalLock(hglbCopy); 
				memcpy(lptstrCopy, &cellStr[index], highlightLen + 1); 

				((char*)lptstrCopy)[highlightLen] = 0;    // null character 
				GlobalUnlock(hglbCopy); 
			
				// Place the handle on the clipboard. 
			
				SetClipboardData(CF_TEXT, hglbCopy);
				CloseClipboard();

				if(cut){
					String str = cellStr;
					cellStr = str.substr(0, index);
					cellStr += str.substr(index + highlightLen);
					caretPos = highlightPos =index;
					onChange();
					win.paint();
				}
			}
		} 
	}
}
void CaretWindow::pasteFromClipboard(){
	if (!IsClipboardFormatAvailable(CF_TEXT)) 
		return; 
	if (!OpenClipboard(hWndMain)) 
		return; 

	HGLOBAL hglb = GetClipboardData(CF_TEXT); 
	if (hglb != NULL) 
	{ 
		LPTSTR lptstr = (LPTSTR)GlobalLock(hglb); 
		if (lptstr != NULL) 
		{ 
			// Call the application-defined ReplaceSelection 
			// function to insert the text and repaint the 
			// window. 
			deleteHighlight();

			int remLen=cellStr.length;
			cellStr=cellStr.replace(caretPos, 0, lptstr);
			
			GlobalUnlock(hglb); 
			caretPos+=cellStr.length-remLen;
			highlightPos=caretPos;

		} 
	} 
	CloseClipboard();

	updateCaret();
	onChange();
	win.paint();
}
void SceneTree::pasteFromClipboard(){
	g_stw->textCell.pasteFromClipboard();
}
void ChatWin::pasteFromClipboard(){
	g_cw->textCell.pasteFromClipboard();
}
void ControlPanel::pasteFromClipboard(){
	if(ControlPanel* controlPanel=dynamic_cast<ControlPanel*>(Window::activeWindow.getActiveWindow())){
		controlPanel->textCell.pasteFromClipboard();
	}
}
CaretWindow* activeTextCell(){
	CaretWindow* textCell=0;
	for(int i=g_callBackList.length; --i>=0;){
		if(textCell=dynamic_cast<CaretWindow*>(g_callBackList[i])){
			break;
		}
	}
	return textCell;
}

void incCaret(){
	if(CaretWindow* tc=activeTextCell()){
		tc->incCaretPosNH();
	}
	else if(Anim* anim = Window::animMode()){
		anim->slider.refValue() += 1.0 / (anim->stepsPerSecond * anim->period);
		if(anim->slider > 1){
			anim->slider.refValue() = 1;
		}
		anim->quantizeSlider();
		anim->updateDisplayFrom();
	}
}

void decCaret(){
	if(CaretWindow* tc=activeTextCell()){
		tc->decCaretPosNH();
	}
	else if(Anim* anim = Window::animMode()){
		anim->slider.refValue() -= 1.0 / (anim->stepsPerSecond * anim->period);
		if(anim->slider < 0){
			anim->slider.refValue() = 0;
		}
		anim->quantizeSlider();
		anim->updateDisplayFrom();
	}
}

void homeCaret(){
	if(CaretWindow* tc = activeTextCell()){
		tc->doHome();
	}
	else if(Anim* anim = Window::animMode()){
		anim->slider = 0;
		anim->updateDisplayFrom();
	}
}
void endCaret(){
	if(CaretWindow* tc=activeTextCell()){
		tc->doEnd();
	}
	else if(Anim* anim = Window::animMode()){
		anim->slider = 1;
		anim->updateDisplayFrom();
	}
}
void deleteChar(){
	if(CaretWindow* tc=activeTextCell()){
		tc->doDelete();
	}
}

void incCaretShift(){
	if(CaretWindow* tc=activeTextCell()){
		tc->incCaretPos();
	}
}
void decCaretShift(){
	if(CaretWindow* tc=activeTextCell()){
		tc->decCaretPos();
	}
}
void homeCaretShift(){
	if(CaretWindow* tc=activeTextCell()){
		tc->doHomeShift();
	}
}
void endCaretShift(){
	if(CaretWindow* tc=activeTextCell()){
		tc->doEndShift();
	}
}
void deleteCharShift(){
	if(CaretWindow* tc=activeTextCell()){
		tc->doDeleteShift();
	}
}
void BezierPatch::mirrorPatch(){
	BackUp::monostable.beginCycle();

	int idx=index[numOfUGroup];
	int pointIdx=index[pointGroup];

	M_Vec3f& point=this->point.refValue();
	M_Vec2f& dim=this->dim.refValue();
	M_Vec2f& zone=this->zone.refValue();
//	M_Int& fixIdx=this->fixIdx.refValue();
//	M_Int& fixIdxIdx=this->fixIdxIdx.refValue();
	M_Int& link=this->link.refValue();

	int noU=dim[idx].x;
	int noV=dim[idx].y;

	int bp=getBasePoint(idx);
	int nBp=point.length;

	int pointLen=noU*noV;

	if(pointIdx-bp<noU-1){//bottom edge
		for(int v=0; v<noV; v++){
			for(int u=0; u<noU; u++){
				int sw=bp+(noV-1-v)*noU+u;
				int h=bp+u;
				point+=point[h]+(point[h]-point[sw]);
			}
		}
/*
		for(int u=0; u<noU; u++){
			int hingeA=u+bp;
			int swingA=hingeA+noU;
			int hingeB=pointLen-noU+u+nBp;
			int swingB=hingeB-noU;

			link+=swingA;
			link+=hingeA;
			link+=hingeB;
			link+=swingB;
		}
		*/
	}
	if(pointIdx-bp>pointLen-noU){//top edge
		for(int v=0; v<noV; v++){
			for(int u=0; u<noU; u++){
				int sw=bp+(noV-1-v)*noU+u;
				int h=bp+pointLen-noU+u;
				point+=point[h]+(point[h]-point[sw]);
			}
		}
		/*
		for(int u=0; u<noU; u++){
			int hingeA=bp+pointLen-noU+u;
			int swingA=hingeA-noU;
			int hingeB=nBp+u;
			int swingB=hingeB+noU;

			link+=swingA;
			link+=hingeA;
			link+=hingeB;
			link+=swingB;
		}
		*/
	}
	else{
		if(pointIdx%noU>(noU>>1)){//right edge
			for(int v=0; v<noV; v++){
				for(int u=noU; --u>=0;){
					int h=bp+v*noU+noU-1;
					int sw=bp+v*noU+u;
					point+=point[h]+(point[h]-point[sw]);
				}
			}
			/*
			for(int v=0; v<noV; v++){
				int hingeA=bp+v*noU+noU-1;;
				int swingA=hingeA-1;
				int hingeB=nBp+v*noU;
				int swingB=hingeB+1;

				link+=swingA;
				link+=hingeA;
				link+=hingeB;
				link+=swingB;
			}
			*/
		}
		else{
			for(int v=0; v<noV; v++){
				for(int u=noU; --u>=0;){
					int h=bp+v*noU;
					int sw=bp+v*noU+u;
					point+=point[h]+(point[h]-point[sw]);
				}
			}
			/*
			for(int v=0; v<noV; v++){
				int hingeA=bp+v*noU;;
				int swingA=hingeA+1;
				int hingeB=nBp+v*noU+noU-1;
				int swingB=hingeB-1;

				link+=swingA;
				link+=hingeA;
				link+=hingeB;
				link+=swingB;
			}
			*/
		}
	}
	dim+=Vec2f(noU, noV);
	zone+=zone[idx];
//	fixIdxIdx+=fixIdx.length;

	processFields();

	updateBuild3dWindow();
}
bool BezierPatch::findMatchingPoint(int exclude){
	M_Vec3f& point=this->point.refValue();
	M_Vec2f& dim=this->dim.refValue();
	M_Vec2f& zone=this->zone.refValue();
//	M_Int& fixIdx=this->fixIdx.refValue();
//	M_Int& fixIdxIdx=this->fixIdxIdx.refValue();
	M_Int& link=this->link.refValue();

	for(int k=0, bp=0; k<dim.length; k++){
		int noU=dim[k].x;
		int noV=dim[k].y;
		int pointLen=noU*noV;

		for(int v=0; v<noV; v++){//check for left edge
			int i=v*noU;
			if(MultiSelect::isInBox(point[bp+i])){
				if(MultiSelect::isInBox(point[bp+i+1])){
					if(bp+i!=exclude){
						int j=link.length;
						link.length+=4;
						link[j+swingB]=bp+i+1;
						link[j+hingeB]=bp+i;
						return true;
					}
				}
			}
			i+=noU-1;//check for right edge
			if(MultiSelect::isInBox(point[bp+i])){
				if(MultiSelect::isInBox(point[bp+i-1])){
					if(bp+i!=exclude){
						int j=link.length;
						link.length+=4;
						link[j+swingB]=bp+i-1;
						link[j+hingeB]=bp+i;
						return true;
					}
				}
			}
		}
		for(int u=0; u<noU; u++){//check for bottom edge
			int i=u;
			if(MultiSelect::isInBox(point[bp+i])){
				if(MultiSelect::isInBox(point[bp+i+noU])){
					if(bp+i!=exclude){
						int j=link.length;
						link.length+=4;
						link[j+swingB]=bp+i+noU;
						link[j+hingeB]=bp+i;
						return true;
					}
				}
			}
			i+=pointLen-noU;//check for top edge
			if(MultiSelect::isInBox(point[bp+i])){
				if(MultiSelect::isInBox(point[bp+i-noU])){
					if(bp+i!=exclude){
						int j=link.length;
						link.length+=4;
						link[j+swingB]=bp+i-noU;
						link[j+hingeB]=bp+i;
						return true;
					}
				}
			}
		}
		bp+=pointLen;
	}
	return false;
}
void BezierPatch::joinEdge(){
	MultiSelect::topLeft2d=convertPointToScreen2d(MultiSelect::topLeft);
	MultiSelect::bottomRight2d=convertPointToScreen2d(MultiSelect::bottomRight);

	M_Vec3f& point=this->point.refValue();
	M_Vec2f& dim=this->dim.refValue();
	M_Vec2f& zone=this->zone.refValue();
//	M_Int& fixIdx=this->fixIdx.refValue();
//	M_Int& fixIdxIdx=this->fixIdxIdx.refValue();
	M_Int& link=this->link.refValue();


	for(int k=0, bp=0; k<dim.length; k++){
		int noU=dim[k].x;
		int noV=dim[k].y;
		int pointLen=noU*noV;

		for(int v=0; v<noV; v++){//check for left edge
			int i=v*noU;
			if(MultiSelect::isInBox(point[bp+i])){
				if(MultiSelect::isInBox(point[bp+i+1])){
					if(findMatchingPoint(bp+i)){
						int j=link.length-4;
						link[j+hingeA]=bp+i;
						link[j+swingA]=bp+i+1;
						doLink(point, link, bp+i, Vec3f());
					}
				}
			}
			i+=noU-1;//check for right edge
			if(MultiSelect::isInBox(point[bp+i])){
				if(MultiSelect::isInBox(point[bp+i-1])){
					if(findMatchingPoint(bp+i)){
						int j=link.length-4;
						link[j+hingeA]=bp+i;
						link[j+swingA]=bp+i-1;
						doLink(point, link, bp+i, Vec3f());
					}
				}
			}
		}
		for(int u=0; u<noU; u++){//check for bottom edge
			int i=u;
			if(MultiSelect::isInBox(point[bp+i])){
				if(MultiSelect::isInBox(point[bp+i+noU])){
					if(findMatchingPoint(bp+i)){
						int j=link.length-4;
						link[j+hingeA]=bp+i;
						link[j+swingA]=bp+i+noU;
						doLink(point, link, bp+i, Vec3f());
					}
				}
			}
			i+=pointLen-noU;//check for top edge
			if(MultiSelect::isInBox(point[bp+i])){
				if(MultiSelect::isInBox(point[bp+i-noU])){
					if(findMatchingPoint(bp+i)){
						int j=link.length-4;
						link[j+hingeA]=bp+i;
						link[j+swingA]=bp+i-noU;
						doLink(point, link, bp+i, Vec3f());
					}
				}
			}
		}
		bp+=pointLen;
	}
	refreshLink(link);
	houseKeep();
}
void BezierPatch::deletePatch(){
	M_Vec3f& point=this->point.refValue();
	M_Vec2f& dim=this->dim.refValue();
	M_Vec2f& zone=this->zone.refValue();
//	M_Int& fixIdx=this->fixIdx.refValue();
//	M_Int& fixIdxIdx=this->fixIdxIdx.refValue();
	M_Int& link=this->link.refValue();

	int idx = index[numOfUGroup];
	int pointIdx = index[pointGroup];
	int pointLen = dim[idx].x*dim[idx].y;
	int bp = m_bp[idx];

	point.deleteElement(bp, pointLen);
	dim.deleteElement(idx);
	zone.deleteElement(idx);
//	fixIdxIdx.deleteElement(idx);

	if(idx>=dim.length){
		idx=dim.length-1;
	}
	if(idx<0){
		idx=0;
	}
	index[numOfUGroup]=idx;

	if(pointIdx>=point.length){
		pointIdx=point.length-1;
	}
	if(pointIdx<0){
		pointIdx=0;
	}
	index[pointGroup]=pointIdx;

	processFields();

	updateBuild3dWindow();
}
/*
inline void FuseSurface::doIn(Stem* stem, Art* source, M_Vec2f& baseCoord, M_Int& srcMatch, M_Int& dstMatch, const float maxU, const float maxV, const float minU, const float minV, float* stemUKnotPtr, float* stemVKnotPtr, Vec3f* stemPointPtr, float* stemWeightPtr, int stemUOrder, int stemVOrder, int stemUDim, int stemVDim){
	if(stem->part.length){
		if(Art* part = dynamic_cast<Art*>(stem->part.refValue()[0].p)){
			const M_Vec3f& coord = part->coord;

			srcMatch.length=0;
			dstMatch.length=0;
			for(int i=0;i<baseCoord.length;i++){

				const Vec2f base0 = baseCoord[i];

				float u = (maxU-minU)*base0.x + minU;
				float v = (maxV-minV)*base0.y + minV;
				Vec3f curvedEdgePoint;
				SurfacePoint(u, v, stemUOrder, stemVOrder, stemUDim, stemVDim, stemUKnotPtr, stemVKnotPtr, stemPointPtr, stemWeightPtr, &curvedEdgePoint);

				for(int j=0;j<coord.length;j++){
					Vec3f crd=coord[j];
					if(
						fabs(crd.x-curvedEdgePoint.x)<range&&
						fabs(crd.y-curvedEdgePoint.y)<range&&
						fabs(crd.z-curvedEdgePoint.z)<range
						){
						srcMatch+=j;
						dstMatch+=i;
					}
				}
			}
		}
	}
}
*/
inline void FuseSurface::doIn(Stem* stem, M_Vec2f& baseCoord, M_Int& srcMatch, M_Int& dstMatch){
	if(stem->part.length){
		if(ImmediatePart* source = dynamic_cast<ImmediatePart*>(stem->source.refValue()[0].p)){
			const M_Vec3f& coord = source->coord;

			srcMatch.length=0;
			dstMatch.length=0;
			for(int i=0;i<baseCoord.length;i++){

				const Vec2f base0 = baseCoord[i];

				for(int j=0;j<coord.length;j++){
					Vec3f crd=coord[j];
					if(
						fabs(crd.x-base0.x)<range&&
						fabs(crd.y-base0.y)<range
						){
						srcMatch+=j;
						dstMatch+=i;
					}
				}
			}
		}
	}
}
/*
inline void FuseSurface::guessPoints(NurbsPatch* patch, Vec3f amount, const M_Vec3f& parentFlatCoord, const M_Vec3f& baseCurvedPoint, float& guessSumDiff, float& currentSumDiff, const float maxU, const float maxV, const float minU, const float minV, float* uKnotPtr, float* vKnotPtr, Vec3f* pointPtr, float* weightPtr, Vec3f* topPointPtr, float* topWeightPtr, Vec3f* r_pointPtr, float* r_weightPtr, int uOrder, int vOrder, int uDegree, int vDegree, int uDim, int vDim){
	Vec3f dummy;
	for(int j = 1; j < uDim; j+=2){
		guessSumDiff = 0;
		Vec3f storePoint = pointPtr[j];
		pointPtr[j] += amount;
		patch->processFields();
		for(int i = srcMatch.length; --i>=0;){
			int d = dstMatch[i];
			guessSumDiff += getDiff(Vec2f(parentFlatCoord[d].x, parentFlatCoord[d].y), dummy, maxU, maxV, minU, minV, uOrder, vOrder, uDegree, vDegree, uDim, vDim, uKnotPtr, vKnotPtr, pointPtr, weightPtr, topPointPtr, topWeightPtr, r_pointPtr, r_weightPtr, baseCurvedPoint[srcMatch[i]]);
		}
		if(guessSumDiff < currentSumDiff){
			currentSumDiff = guessSumDiff;
		}
		else{
			pointPtr[j] = storePoint;
		}
	}
}
*/
inline void FuseSurface::guessPoints(NurbsPatch* patch, Vec3f amount, const M_Vec3f& parentFlatCoord, const M_Vec3f& baseCurvedPoint, float& guessSumDiff, float& currentSumDiff, const float maxU, const float maxV, const float minU, const float minV, float* uKnotPtr, float* vKnotPtr, Vec3f* pointPtr, float* weightPtr, Vec3f* topPointPtr, float* topWeightPtr, Vec3f* r_pointPtr, float* r_weightPtr, int uOrder, int vOrder, int uDegree, int vDegree, int uDim, int vDim){
	Vec3f dummy;
	for(int j = 1; j < uDim; j+=2){
		guessSumDiff = 0;
		Vec3f storePoint = pointPtr[j];
		pointPtr[j] += amount;
		patch->processFields();
		for(int i = qty; --i>=0;){
			guessSumDiff += getDiff(Vec2f(parentFlatCoord[i].x, parentFlatCoord[i].y), dummy, maxU, maxV, minU, minV, uOrder, vOrder, uDegree, vDegree, uDim, vDim, uKnotPtr, vKnotPtr, pointPtr, weightPtr, topPointPtr, topWeightPtr, r_pointPtr, r_weightPtr, baseCurvedPoint[i]);
		}
		if(guessSumDiff < currentSumDiff){
			currentSumDiff = guessSumDiff;
		}
		else{
			pointPtr[j] = storePoint;
		}
	}
}
void FuseSurface::doFuse(NurbsPatch& parentPatch, Stem& parentStem){
	this->parentPatch = &parentPatch;
	this->parentStem = &parentStem;
	NurbsPatch* childPatch = dynamic_cast<NurbsPatch*>(this->patch.p);
	M_Vec3f* childCoordPtr = 0;
	Seamless* smls = 0;
	const bool southPole = parentStem.southPole;
	const bool northPole = parentStem.northPole;
	if(childPatch){
		for(int s = 0; s < childPatch->stem.length; s++){
			if(Stem* childStem = dynamic_cast<Stem*>(childPatch->stem[s].p)){
				if(s == 0 && childStem->mould){
					childStem->doBuild(this, childPatch);
					if(childStem->part.length){
						if(Art* part = childStem->part[0].p){
							childCoordPtr = &part->coord.refValue();
							smls = part->smls;
						}
					}
				}
				else{
					childStem->doBuild(0, childPatch);
				}
			}
		}
	}
	else{
		return;
	}
	if(childCoordPtr == 0){
		return;
	}
	M_Vec3f& childCoord = *childCoordPtr;

	M_Vec2f& baseCoord = this->baseCoord.refValue();
	if(baseCoord.length == 0){
		return;
	}
	if(spread){
		spread.refValue()=false;
		doSpread();
	}

	const float vSpan=this->vSpan;

	int uOrder = parentPatch.order.x;
	int vOrder = parentPatch.order.y;
	int uDegree = parentPatch.uDegree;
	int vDegree = parentPatch.vDegree;
	int uDim = parentPatch.dim.x;
	int vDim = parentPatch.dim.y;

	float* uKnotPtr = parentPatch.uKnot.refValue().getArray();
	float* vKnotPtr = parentPatch.vKnot.refValue().getArray();
	Vec3f* pointPtr = parentPatch.point.refValue().getArray();
	float* weightPtr = parentPatch.weight.refValue().getArray();

	Vec3f* topPointPtr = pointPtr+uDim*vDim-uDim;
	float* topWeightPtr = weightPtr+uDim*vDim-uDim;

	Vec3f* r_pointPtr = parentPatch.r_point.getArray();
	float* r_weightPtr = parentPatch.r_weight.getArray();


	const float coordChange = this->coordChange;

	int count = vPan.length;

	curvedEdgePoint.length = count;

	Vec2f diff;
	const int idx=index[0];

	Vec3f curvedEdgePoint0;

	float maxU = parentPatch.maxU();
	float maxV = parentPatch.maxV();
	float minU = parentPatch.minU();
	float minV = parentPatch.minV();

	if(bond){
		if(northPole){
			for(int i=0; i<count; i++){
				const Vec2f base0 = baseCoord[i];

				float diffArray[4];
				Vec3f destPoint[4];

				float error = getDiff(convertNorthPoleToMercator(base0), curvedEdgePoint0, maxU, maxV, minU, minV, uOrder, vOrder, uDegree, vDegree, uDim, vDim, uKnotPtr, vKnotPtr, pointPtr, weightPtr, topPointPtr, topWeightPtr, r_pointPtr, r_weightPtr, childCoord[i]);

				Vec2f basePointOffset[4] = {
					Vec2f(-coordChange, 0),
					Vec2f( coordChange, 0),

					Vec2f(0,-coordChange),
					Vec2f(0, coordChange),
				};

				diffArray[0] = getDiff(convertNorthPoleToMercator(base0 + basePointOffset[0]), destPoint[0], maxU, maxV, minU, minV, uOrder, vOrder, uDegree, vDegree, uDim, vDim, uKnotPtr, vKnotPtr, pointPtr, weightPtr, topPointPtr, topWeightPtr, r_pointPtr, r_weightPtr, childCoord[i]);
				diffArray[1] = getDiff(convertNorthPoleToMercator(base0 + basePointOffset[1]), destPoint[1], maxU, maxV, minU, minV, uOrder, vOrder, uDegree, vDegree, uDim, vDim, uKnotPtr, vKnotPtr, pointPtr, weightPtr, topPointPtr, topWeightPtr, r_pointPtr, r_weightPtr, childCoord[i]);
				diffArray[2] = getDiff(convertNorthPoleToMercator(base0 + basePointOffset[2]), destPoint[2], maxU, maxV, minU, minV, uOrder, vOrder, uDegree, vDegree, uDim, vDim, uKnotPtr, vKnotPtr, pointPtr, weightPtr, topPointPtr, topWeightPtr, r_pointPtr, r_weightPtr, childCoord[i]);
				diffArray[3] = getDiff(convertNorthPoleToMercator(base0 + basePointOffset[3]), destPoint[3], maxU, maxV, minU, minV, uOrder, vOrder, uDegree, vDegree, uDim, vDim, uKnotPtr, vKnotPtr, pointPtr, weightPtr, topPointPtr, topWeightPtr, r_pointPtr, r_weightPtr, childCoord[i]);


				float leastDiff = diffArray[0];
				int k = 0;

				for(int j = 1; j < 4; j++){
					float curDiff = diffArray[j];
					if(curDiff < leastDiff){
						leastDiff = curDiff;
						k = j;
					}
				}
				if(i == idx){
					this->diff.refValue().x = diffArray[k];
				}

				float a = 1.0 - vPan[i]/vSpan;
				//float pan = sqrt(1.0 - a * a);
				float pan = cos(a * PI) * .5 + .5;
				float ipan = 1.0 - pan;

				if(error >= diffArray[k]){
					baseCoord[i] += basePointOffset[k];
					if(bond){
						childCoord[i]=destPoint[k]*ipan+childCoord[i]*pan;
					}
				}
				else{
					if(bond){
						childCoord[i]=curvedEdgePoint0*ipan+childCoord[i]*pan;
					}
				}
			}
		}
		else if(southPole){
			for(int i=0; i<count; i++){
				const Vec2f base0 = baseCoord[i];

				float diffArray[4];
				Vec3f destPoint[4];

				float error = getDiff(convertSouthPoleToMercator(base0), curvedEdgePoint0, maxU, maxV, minU, minV, uOrder, vOrder, uDegree, vDegree, uDim, vDim, uKnotPtr, vKnotPtr, pointPtr, weightPtr, topPointPtr, topWeightPtr, r_pointPtr, r_weightPtr, childCoord[i]);

				Vec2f basePointOffset[4] = {
					Vec2f(-coordChange, 0),
					Vec2f( coordChange, 0),

					Vec2f(0,-coordChange),
					Vec2f(0, coordChange),
				};

				diffArray[0] = getDiff(convertSouthPoleToMercator(base0 + basePointOffset[0]), destPoint[0], maxU, maxV, minU, minV, uOrder, vOrder, uDegree, vDegree, uDim, vDim, uKnotPtr, vKnotPtr, pointPtr, weightPtr, topPointPtr, topWeightPtr, r_pointPtr, r_weightPtr, childCoord[i]);
				diffArray[1] = getDiff(convertSouthPoleToMercator(base0 + basePointOffset[1]), destPoint[1], maxU, maxV, minU, minV, uOrder, vOrder, uDegree, vDegree, uDim, vDim, uKnotPtr, vKnotPtr, pointPtr, weightPtr, topPointPtr, topWeightPtr, r_pointPtr, r_weightPtr, childCoord[i]);
				diffArray[2] = getDiff(convertSouthPoleToMercator(base0 + basePointOffset[2]), destPoint[2], maxU, maxV, minU, minV, uOrder, vOrder, uDegree, vDegree, uDim, vDim, uKnotPtr, vKnotPtr, pointPtr, weightPtr, topPointPtr, topWeightPtr, r_pointPtr, r_weightPtr, childCoord[i]);
				diffArray[3] = getDiff(convertSouthPoleToMercator(base0 + basePointOffset[3]), destPoint[3], maxU, maxV, minU, minV, uOrder, vOrder, uDegree, vDegree, uDim, vDim, uKnotPtr, vKnotPtr, pointPtr, weightPtr, topPointPtr, topWeightPtr, r_pointPtr, r_weightPtr, childCoord[i]);


				float leastDiff = diffArray[0];
				int k = 0;

				for(int j = 1; j < 4; j++){
					float curDiff = diffArray[j];
					if(curDiff < leastDiff){
						leastDiff = curDiff;
						k = j;
					}
				}
				if(i == idx){
					this->diff.refValue().x = diffArray[k];
				}

				float a = 1.0 - vPan[i]/vSpan;
				//float pan = sqrt(1.0 - a * a);
				float pan = cos(a * PI) * .5 + .5;
				float ipan = 1.0 - pan;

				if(error >= diffArray[k]){
					baseCoord[i] += basePointOffset[k];
					if(bond){
						childCoord[i]=destPoint[k]*ipan+childCoord[i]*pan;
					}
				}
				else{
					if(bond){
						childCoord[i]=curvedEdgePoint0*ipan+childCoord[i]*pan;
					}
				}
			}
		}
		else{
			for(int i=0; i<count; i++){
				const Vec2f base0 = baseCoord[i];

				float diffArray[4];
				Vec3f destPoint[4];

				float error = getDiff(base0, curvedEdgePoint0, maxU, maxV, minU, minV, uOrder, vOrder, uDegree, vDegree, uDim, vDim, uKnotPtr, vKnotPtr, pointPtr, weightPtr, topPointPtr, topWeightPtr, r_pointPtr, r_weightPtr, childCoord[i]);

				Vec2f basePointOffset[4] = {
					Vec2f(-coordChange, 0),
					Vec2f( coordChange, 0),

					Vec2f(0,-coordChange),
					Vec2f(0, coordChange),
				};

				diffArray[0] = getDiff(base0 + basePointOffset[0], destPoint[0], maxU, maxV, minU, minV, uOrder, vOrder, uDegree, vDegree, uDim, vDim, uKnotPtr, vKnotPtr, pointPtr, weightPtr, topPointPtr, topWeightPtr, r_pointPtr, r_weightPtr, childCoord[i]);
				diffArray[1] = getDiff(base0 + basePointOffset[1], destPoint[1], maxU, maxV, minU, minV, uOrder, vOrder, uDegree, vDegree, uDim, vDim, uKnotPtr, vKnotPtr, pointPtr, weightPtr, topPointPtr, topWeightPtr, r_pointPtr, r_weightPtr, childCoord[i]);
				diffArray[2] = getDiff(base0 + basePointOffset[2], destPoint[2], maxU, maxV, minU, minV, uOrder, vOrder, uDegree, vDegree, uDim, vDim, uKnotPtr, vKnotPtr, pointPtr, weightPtr, topPointPtr, topWeightPtr, r_pointPtr, r_weightPtr, childCoord[i]);
				diffArray[3] = getDiff(base0 + basePointOffset[3], destPoint[3], maxU, maxV, minU, minV, uOrder, vOrder, uDegree, vDegree, uDim, vDim, uKnotPtr, vKnotPtr, pointPtr, weightPtr, topPointPtr, topWeightPtr, r_pointPtr, r_weightPtr, childCoord[i]);


				float leastDiff = diffArray[0];
				int k = 0;

				for(int j = 1; j < 4; j++){
					float curDiff = diffArray[j];
					if(curDiff < leastDiff){
						leastDiff = curDiff;
						k = j;
					}
				}
				if(i == idx){
					this->diff.refValue().x = diffArray[k];
				}

				float a = 1.0 - vPan[i]/vSpan;
				//float pan = sqrt(1.0 - a * a);
				float pan = cos(a * PI) * .5 + .5;
				float ipan = 1.0 - pan;

				if(error >= diffArray[k]){
					baseCoord[i] += basePointOffset[k];
					if(bond){
						childCoord[i]=destPoint[k]*ipan+childCoord[i]*pan;
					}
				}
				else{
					if(bond){
						childCoord[i]=curvedEdgePoint0*ipan+childCoord[i]*pan;
					}
				}
			}
		}
	}
	if(parentStem.source.length){
		if(Art* source = dynamic_cast<Art*>(parentStem.source.refValue()[0].p)){
			M_Vec2f& baseCoord = this->baseCoord.refValue();
			M_Int& srcMatch = this->srcMatch.refValue();
			M_Int& dstMatch = this->dstMatch.refValue();
			if(in){
				//doIn(&parentStem, source, baseCoord, srcMatch, dstMatch, maxU, maxV, minU, minV, parentUKnotPtr, parentVKnotPtr, parentPointPtr, parentWeightPtr, parentUOrder, parentVOrder, parentUDim, parentVDim);
				doIn(&parentStem, baseCoord, srcMatch, dstMatch);
			}
			M_Vec3f& coord = source->coord.refValue();
			if(snap){
				if(in == false && smls->autoIn){
					in.refValue()=true;
					doIn(&parentStem, baseCoord, srcMatch, dstMatch);
				}
				/*
				if(in == false && g_potentialTrouble){
					int len = coord.length;
					for(int i = srcMatch.length; --i>=0;){
						int si = srcMatch[i];
						if(si<len){
							Vec3f srcCrd = coord[si];
							Vec2f dstCrd = baseCoord[dstMatch[i]];
							if(	fabs(srcCrd.x-dstCrd.x)<range&&
								fabs(srcCrd.y-dstCrd.y)<range){

								continue;
							}
						}
						in.refValue()=true;
						doIn(&parentStem, baseCoord, srcMatch, dstMatch);
						break;
					}
				}
				*/
				for(int i = srcMatch.length; --i>=0;){
					int d = dstMatch[i];
					coord[srcMatch[i]] = Vec3f(baseCoord[d].x, baseCoord[d].y, 0);
				}
			}
/*
			if(evolve && pointChange > 0){
				if(parentStem.part.length){
					if(Art* part = dynamic_cast<Art*>(parentStem.part.refValue()[0].p)){

						int uOrder = childPatch->order.x;
						int vOrder = childPatch->order.y;
						int uDegree = childPatch->uDegree;
						int vDegree = childPatch->vDegree;
						int uDim = childPatch->dim.x;
						int vDim = childPatch->dim.y;

						float* uKnotPtr = childPatch->uKnot.refValue().getArray();
						float* vKnotPtr = childPatch->vKnot.refValue().getArray();
						Vec3f* pointPtr = childPatch->point.refValue().getArray();
						float* weightPtr = childPatch->weight.refValue().getArray();

						Vec3f* topPointPtr = pointPtr+uDim*vDim-uDim;
						float* topWeightPtr = weightPtr+uDim*vDim-uDim;

						Vec3f* r_pointPtr = childPatch->r_point.getArray();
						float* r_weightPtr = childPatch->r_weight.getArray();


						const M_Vec3f& baseCurvedPoint = part->coord;
						float currentSumDiff=0;
						float guessSumDiff=0;
						Vec3f dummy;
						for(int i = srcMatch.length; --i>=0;){
							int d = dstMatch[i];
							currentSumDiff += getDiff(Vec2f(parentFlatCoord[d].x, parentFlatCoord[d].y), dummy, maxU, maxV, minU, minV, uOrder, vOrder, uDegree, vDegree, uDim, vDim, uKnotPtr, vKnotPtr, pointPtr, weightPtr, topPointPtr, topWeightPtr, r_pointPtr, r_weightPtr, childCoord[i]);
						}
						guessPoints(childPatch, Vec3f(-pointChange, 0, 0), parentFlatCoord, baseCurvedPoint, guessSumDiff, currentSumDiff, maxU, maxV, minU, minV, uKnotPtr, vKnotPtr, pointPtr, weightPtr, topPointPtr, topWeightPtr, r_pointPtr, r_weightPtr, uOrder, vOrder, uDegree, vDegree, uDim, vDim);
						guessPoints(childPatch, Vec3f(pointChange, 0, 0), parentFlatCoord, baseCurvedPoint, guessSumDiff, currentSumDiff, maxU, maxV, minU, minV, uKnotPtr, vKnotPtr, pointPtr, weightPtr, topPointPtr, topWeightPtr, r_pointPtr, r_weightPtr, uOrder, vOrder, uDegree, vDegree, uDim, vDim);

						guessPoints(childPatch, Vec3f(0, -pointChange, 0), parentFlatCoord, baseCurvedPoint, guessSumDiff, currentSumDiff, maxU, maxV, minU, minV, uKnotPtr, vKnotPtr, pointPtr, weightPtr, topPointPtr, topWeightPtr, r_pointPtr, r_weightPtr, uOrder, vOrder, uDegree, vDegree, uDim, vDim);
						guessPoints(childPatch, Vec3f(0, pointChange, 0), parentFlatCoord, baseCurvedPoint, guessSumDiff, currentSumDiff, maxU, maxV, minU, minV, uKnotPtr, vKnotPtr, pointPtr, weightPtr, topPointPtr, topWeightPtr, r_pointPtr, r_weightPtr, uOrder, vOrder, uDegree, vDegree, uDim, vDim);

						guessPoints(childPatch, Vec3f(0, 0, -pointChange), parentFlatCoord, baseCurvedPoint, guessSumDiff, currentSumDiff, maxU, maxV, minU, minV, uKnotPtr, vKnotPtr, pointPtr, weightPtr, topPointPtr, topWeightPtr, r_pointPtr, r_weightPtr, uOrder, vOrder, uDegree, vDegree, uDim, vDim);
						guessPoints(childPatch, Vec3f(0, 0, pointChange), parentFlatCoord, baseCurvedPoint, guessSumDiff, currentSumDiff, maxU, maxV, minU, minV, uKnotPtr, vKnotPtr, pointPtr, weightPtr, topPointPtr, topWeightPtr, r_pointPtr, r_weightPtr, uOrder, vOrder, uDegree, vDegree, uDim, vDim);

						sumDiff.refValue()=currentSumDiff;
						averageDiff.refValue()=currentSumDiff/qty;
						childPatch->processFields();
					}
				}
			}
			*/
		}
	}
	if(evolve && pointChange > 0 && childCoord.length >= qty){

		if(bond == false){
			if(northPole){
				for(int i=0; i<qty; i++){
					childCoord[i]=getSurfaceVertex(convertNorthPoleToMercator(baseCoord[i]), maxU, maxV, minU, minV, uOrder, vOrder, uDegree, vDegree, uDim, vDim, uKnotPtr, vKnotPtr, pointPtr, weightPtr, topPointPtr, topWeightPtr, r_pointPtr, r_weightPtr);
				}
			}
			else if(southPole){
				for(int i=0; i<qty; i++){
					childCoord[i]=getSurfaceVertex(convertSouthPoleToMercator(baseCoord[i]), maxU, maxV, minU, minV, uOrder, vOrder, uDegree, vDegree, uDim, vDim, uKnotPtr, vKnotPtr, pointPtr, weightPtr, topPointPtr, topWeightPtr, r_pointPtr, r_weightPtr);
				}
			}
			else{
				for(int i=0; i<qty; i++){
					childCoord[i]=getSurfaceVertex(baseCoord[i], maxU, maxV, minU, minV, uOrder, vOrder, uDegree, vDegree, uDim, vDim, uKnotPtr, vKnotPtr, pointPtr, weightPtr, topPointPtr, topWeightPtr, r_pointPtr, r_weightPtr);
				}
			}			
		}

		int uOrder = childPatch->order.x;
		int vOrder = childPatch->order.y;
		int uDegree = childPatch->uDegree;
		int vDegree = childPatch->vDegree;
		int uDim = childPatch->dim.x;
		int vDim = childPatch->dim.y;

		float* uKnotPtr = childPatch->uKnot.refValue().getArray();
		float* vKnotPtr = childPatch->vKnot.refValue().getArray();
		Vec3f* pointPtr = childPatch->point.refValue().getArray();
		float* weightPtr = childPatch->weight.refValue().getArray();

		Vec3f* topPointPtr = pointPtr+uDim*vDim-uDim;
		float* topWeightPtr = weightPtr+uDim*vDim-uDim;

		Vec3f* r_pointPtr = childPatch->r_point.getArray();
		float* r_weightPtr = childPatch->r_weight.getArray();

		float currentSumDiff=0;
		float guessSumDiff=0;
		Vec3f dummy;
		for(int i = qty; --i>=0;){
			currentSumDiff += getDiff(Vec2f(parentFlatCoord[i].x, parentFlatCoord[i].y), dummy, maxU, maxV, minU, minV, uOrder, vOrder, uDegree, vDegree, uDim, vDim, uKnotPtr, vKnotPtr, pointPtr, weightPtr, topPointPtr, topWeightPtr, r_pointPtr, r_weightPtr, childCoord[i]);
		}
		guessPoints(childPatch, Vec3f(-pointChange, 0, 0), parentFlatCoord, childCoord, guessSumDiff, currentSumDiff, maxU, maxV, minU, minV, uKnotPtr, vKnotPtr, pointPtr, weightPtr, topPointPtr, topWeightPtr, r_pointPtr, r_weightPtr, uOrder, vOrder, uDegree, vDegree, uDim, vDim);
		guessPoints(childPatch, Vec3f(pointChange, 0, 0), parentFlatCoord, childCoord, guessSumDiff, currentSumDiff, maxU, maxV, minU, minV, uKnotPtr, vKnotPtr, pointPtr, weightPtr, topPointPtr, topWeightPtr, r_pointPtr, r_weightPtr, uOrder, vOrder, uDegree, vDegree, uDim, vDim);

		guessPoints(childPatch, Vec3f(0, -pointChange, 0), parentFlatCoord, childCoord, guessSumDiff, currentSumDiff, maxU, maxV, minU, minV, uKnotPtr, vKnotPtr, pointPtr, weightPtr, topPointPtr, topWeightPtr, r_pointPtr, r_weightPtr, uOrder, vOrder, uDegree, vDegree, uDim, vDim);
		guessPoints(childPatch, Vec3f(0, pointChange, 0), parentFlatCoord, childCoord, guessSumDiff, currentSumDiff, maxU, maxV, minU, minV, uKnotPtr, vKnotPtr, pointPtr, weightPtr, topPointPtr, topWeightPtr, r_pointPtr, r_weightPtr, uOrder, vOrder, uDegree, vDegree, uDim, vDim);

		guessPoints(childPatch, Vec3f(0, 0, -pointChange), parentFlatCoord, childCoord, guessSumDiff, currentSumDiff, maxU, maxV, minU, minV, uKnotPtr, vKnotPtr, pointPtr, weightPtr, topPointPtr, topWeightPtr, r_pointPtr, r_weightPtr, uOrder, vOrder, uDegree, vDegree, uDim, vDim);
		guessPoints(childPatch, Vec3f(0, 0, pointChange), parentFlatCoord, childCoord, guessSumDiff, currentSumDiff, maxU, maxV, minU, minV, uKnotPtr, vKnotPtr, pointPtr, weightPtr, topPointPtr, topWeightPtr, r_pointPtr, r_weightPtr, uOrder, vOrder, uDegree, vDegree, uDim, vDim);

		sumDiff.refValue()=currentSumDiff;
		averageDiff.refValue()=currentSumDiff/qty;
		childPatch->processFields();
	}

	if((polyMapAid || polyMapAidMirror) && g_visualClue == g_vcArtTriBase && g_vcArtTriBase->fuseSurface==this){
		this->houseKeep();
	}
	else if(g_selNode.getRaw()==this){
		this->houseKeep();
	}
}
void FuseSurface::doFuse(M_Vec3f& coord, NurbsPatch& branchPatch){//old
	this->parentPatch = 0;
	M_Vec2f& baseCoord = this->baseCoord.refValue();
	//Matrix imatrix = rotationMatrix(-orientation);
	const float vSpan=this->vSpan;
	NurbsPatch* stemPatch = dynamic_cast<NurbsPatch*>(this->patch.p);
	if(stemPatch==0){
		if(Stem* stem = dynamic_cast<Stem*>(this->patch.p)){
			stemPatch = dynamic_cast<NurbsPatch*>(stem->surface.p);
		}
	}
	if(stemPatch){
		int uOrder = stemPatch->order.x;
		int vOrder = stemPatch->order.y;
		int uDegree = stemPatch->uDegree;
		int vDegree = stemPatch->vDegree;
		int uDim = stemPatch->dim.x;
		int vDim = stemPatch->dim.y;

		float* uKnotPtr = stemPatch->uKnot.refValue().getArray();
		float* vKnotPtr = stemPatch->vKnot.refValue().getArray();
		Vec3f* pointPtr = stemPatch->point.refValue().getArray();
		float* weightPtr = stemPatch->weight.refValue().getArray();

		Vec3f* topPointPtr = pointPtr+uDim*vDim-uDim;
		float* topWeightPtr = weightPtr+uDim*vDim-uDim;

		Vec3f* r_pointPtr = stemPatch->r_point.getArray();
		float* r_weightPtr = stemPatch->r_weight.getArray();

		const float coordChange = this->coordChange;

		int count = vPan.length;

		curvedEdgePoint.length = count;

		Vec2f diff;
		const int idx=index[0];

		Vec3f curvedEdgePoint0;

		float maxU = stemPatch->maxU();
		float maxV = stemPatch->maxV();
		float minU = stemPatch->minU();
		float minV = stemPatch->minV();

		for(int i=0; i<count; i++){
			Vec2f base0 = baseCoord[i];


			float diffArray[4];
			Vec3f destPoint[4];

			float error = getDiff(baseCoord[i], curvedEdgePoint0, maxU, maxV, minU, minV, uOrder, vOrder, uDegree, vDegree, uDim, vDim, uKnotPtr, vKnotPtr, pointPtr, weightPtr, topPointPtr, topWeightPtr, r_pointPtr, r_weightPtr, coord[i]);

			Vec2f basePointOffset[4] = {
				Vec2f(-coordChange, 0),
				Vec2f( coordChange, 0),

				Vec2f(0,-coordChange),
				Vec2f(0, coordChange),

				//Vec2f(-coordChange, -coordChange),
				//Vec2f(-coordChange,  coordChange),

				//Vec2f( coordChange, -coordChange),
				//Vec2f( coordChange,  coordChange),
			};

			diffArray[0] = getDiff(baseCoord[i] + basePointOffset[0], destPoint[0], maxU, maxV, minU, minV, uOrder, vOrder, uDegree, vDegree, uDim, vDim, uKnotPtr, vKnotPtr, pointPtr, weightPtr, topPointPtr, topWeightPtr, r_pointPtr, r_weightPtr, coord[i]);
			diffArray[1] = getDiff(baseCoord[i] + basePointOffset[1], destPoint[1], maxU, maxV, minU, minV, uOrder, vOrder, uDegree, vDegree, uDim, vDim, uKnotPtr, vKnotPtr, pointPtr, weightPtr, topPointPtr, topWeightPtr, r_pointPtr, r_weightPtr, coord[i]);
			diffArray[2] = getDiff(baseCoord[i] + basePointOffset[2], destPoint[2], maxU, maxV, minU, minV, uOrder, vOrder, uDegree, vDegree, uDim, vDim, uKnotPtr, vKnotPtr, pointPtr, weightPtr, topPointPtr, topWeightPtr, r_pointPtr, r_weightPtr, coord[i]);
			diffArray[3] = getDiff(baseCoord[i] + basePointOffset[3], destPoint[3], maxU, maxV, minU, minV, uOrder, vOrder, uDegree, vDegree, uDim, vDim, uKnotPtr, vKnotPtr, pointPtr, weightPtr, topPointPtr, topWeightPtr, r_pointPtr, r_weightPtr, coord[i]);

//			diffArray[4] = getDiff(baseCoord[i] + basePointOffset[4], destPoint[4], maxU, maxV, minU, minV, stemUOrder, stemVOrder, stemUDim, stemVDim, stemUKnotPtr, stemVKnotPtr, stemPointPtr, stemWeightPtr, coord[i]);
//			diffArray[5] = getDiff(baseCoord[i] + basePointOffset[5], destPoint[5], maxU, maxV, minU, minV, stemUOrder, stemVOrder, stemUDim, stemVDim, stemUKnotPtr, stemVKnotPtr, stemPointPtr, stemWeightPtr, coord[i]);
//			diffArray[6] = getDiff(baseCoord[i] + basePointOffset[6], destPoint[6], maxU, maxV, minU, minV, stemUOrder, stemVOrder, stemUDim, stemVDim, stemUKnotPtr, stemVKnotPtr, stemPointPtr, stemWeightPtr, coord[i]);
//			diffArray[7] = getDiff(baseCoord[i] + basePointOffset[7], destPoint[7], maxU, maxV, minU, minV, stemUOrder, stemVOrder, stemUDim, stemVDim, stemUKnotPtr, stemVKnotPtr, stemPointPtr, stemWeightPtr, coord[i]);

			float leastDiff = diffArray[0];
			int k = 0;

			for(int j = 1; j < 4; j++){
				float curDiff = diffArray[j];
				if(curDiff < leastDiff){
					leastDiff = curDiff;
					k = j;
				}
			}
			if(i == idx){
				this->diff.refValue().x = diffArray[k];
			}

			float a = 1.0 - vPan[i]/vSpan;
			//float pan = sqrt(1.0 - a * a);
			float pan = cos(a * PI) * .5 + .5;
			float ipan = 1.0 - pan;

			if(error >= diffArray[k]){
				baseCoord[i] += basePointOffset[k];
				if(bond){
					coord[i]=destPoint[k]*ipan+coord[i]*pan;
				}
			}
			else{
				if(bond){
					coord[i]=curvedEdgePoint0*ipan+coord[i]*pan;
				}
			}
		}
		if(Stem* stem = dynamic_cast<Stem*>(this->patch.refValue().p)){
			if(stem->source.length){
				if(Art* source = dynamic_cast<Art*>(stem->source.refValue()[0].p)){
					M_Vec2f& baseCoord = this->baseCoord.refValue();
					M_Int& srcMatch = this->srcMatch.refValue();
					M_Int& dstMatch = this->dstMatch.refValue();
					if(in){
						//old way never had an in?
						//doIn(stem, source, baseCoord, srcMatch, dstMatch, maxU, maxV, minU, minV, stemUKnotPtr, stemVKnotPtr, stemPointPtr, stemWeightPtr, stemUOrder, stemVOrder, stemUDim, stemVDim);
					}
					M_Vec3f& coord = source->coord.refValue();
					if(snap){
						for(int i = srcMatch.length; --i>=0;){
							int d = dstMatch[i];
							coord[srcMatch[i]] = Vec3f(baseCoord[d].x, baseCoord[d].y, 0);
						}
					}
					if(evolve && pointChange > 0){
						if(stem->part.length){
							if(Art* part = dynamic_cast<Art*>(stem->part.refValue()[0].p)){

								int uOrder = branchPatch.order.x;
								int vOrder = branchPatch.order.y;
								int uDegree = branchPatch.uDegree;
								int vDegree = branchPatch.vDegree;
								int uDim = branchPatch.dim.x;
								int vDim = branchPatch.dim.y;

								float* uKnotPtr = branchPatch.uKnot.refValue().getArray();
								float* vKnotPtr = branchPatch.vKnot.refValue().getArray();
								Vec3f* pointPtr = branchPatch.point.refValue().getArray();
								float* weightPtr = branchPatch.weight.refValue().getArray();

								Vec3f* topPointPtr = pointPtr+uDim*vDim-uDim;
								float* topWeightPtr = weightPtr+uDim*vDim-uDim;

								Vec3f* r_pointPtr = branchPatch.r_point.getArray();
								float* r_weightPtr = branchPatch.r_weight.getArray();

								float maxU = branchPatch.maxU();
								float maxV = branchPatch.maxV();
								float minU = branchPatch.minU();
								float minV = branchPatch.minV();


								const M_Vec3f& baseCurvedPoint = part->coord;
								float currentSumDiff=0;
								float guessSumDiff=0;
								Vec3f dummy;
								for(int i = srcMatch.length; --i>=0;){
									int d = dstMatch[i];
									currentSumDiff += getDiff(Vec2f(parentFlatCoord[d].x, parentFlatCoord[d].y), dummy, maxU, maxV, minU, minV, uOrder, vOrder, uDegree, vDegree, uDim, vDim, uKnotPtr, vKnotPtr, pointPtr, weightPtr, topPointPtr, topWeightPtr, r_pointPtr, r_weightPtr, baseCurvedPoint[srcMatch[i]]);
								}
								guessPoints(&branchPatch, Vec3f(-pointChange, 0, 0), parentFlatCoord, baseCurvedPoint, guessSumDiff, currentSumDiff, maxU, maxV, minU, minV, uKnotPtr, vKnotPtr, pointPtr, weightPtr, topPointPtr, topWeightPtr, r_pointPtr, r_weightPtr, uOrder, vOrder, uDegree, vDegree, uDim, vDim);
								guessPoints(&branchPatch, Vec3f(pointChange, 0, 0), parentFlatCoord, baseCurvedPoint, guessSumDiff, currentSumDiff, maxU, maxV, minU, minV, uKnotPtr, vKnotPtr, pointPtr, weightPtr, topPointPtr, topWeightPtr, r_pointPtr, r_weightPtr, uOrder, vOrder, uDegree, vDegree, uDim, vDim);

								guessPoints(&branchPatch, Vec3f(0, -pointChange, 0), parentFlatCoord, baseCurvedPoint, guessSumDiff, currentSumDiff, maxU, maxV, minU, minV, uKnotPtr, vKnotPtr, pointPtr, weightPtr, topPointPtr, topWeightPtr, r_pointPtr, r_weightPtr, uOrder, vOrder, uDegree, vDegree, uDim, vDim);
								guessPoints(&branchPatch, Vec3f(0, pointChange, 0), parentFlatCoord, baseCurvedPoint, guessSumDiff, currentSumDiff, maxU, maxV, minU, minV, uKnotPtr, vKnotPtr, pointPtr, weightPtr, topPointPtr, topWeightPtr, r_pointPtr, r_weightPtr, uOrder, vOrder, uDegree, vDegree, uDim, vDim);

								guessPoints(&branchPatch, Vec3f(0, 0, -pointChange), parentFlatCoord, baseCurvedPoint, guessSumDiff, currentSumDiff, maxU, maxV, minU, minV, uKnotPtr, vKnotPtr, pointPtr, weightPtr, topPointPtr, topWeightPtr, r_pointPtr, r_weightPtr, uOrder, vOrder, uDegree, vDegree, uDim, vDim);
								guessPoints(&branchPatch, Vec3f(0, 0, pointChange), parentFlatCoord, baseCurvedPoint, guessSumDiff, currentSumDiff, maxU, maxV, minU, minV, uKnotPtr, vKnotPtr, pointPtr, weightPtr, topPointPtr, topWeightPtr, r_pointPtr, r_weightPtr, uOrder, vOrder, uDegree, vDegree, uDim, vDim);

								sumDiff.refValue()=currentSumDiff;
								averageDiff.refValue()=currentSumDiff/qty;
								branchPatch.processFields();
							}
						}
					}
				}
			}
		}
		if((polyMapAid || polyMapAidMirror) && g_visualClue == g_vcArtTriBase && g_vcArtTriBase->fuseSurface==this){
			this->houseKeep();
		}
		else if(g_selNode.getRaw()==this){
			this->houseKeep();
		}
	}
}
void updateFuseSurfaceClue(){
	if(dynamic_cast<Art*>(g_selNode.getRaw())){
		g_vcArtTriBase->fuseSurface = 0;
		for(int k = g_sceneTreeList.length; --k>=0;){
			if(FuseSurface* fuseSurface = dynamic_cast<FuseSurface*>(g_sceneTreeList[k].node.p)){
				if(fuseSurface->polyMapAid || fuseSurface->polyMapAidMirror){
					g_vcArtTriBase->fuseSurface=fuseSurface;
					fuseSurface->houseKeep();
					break;
				}
			}
		}
	}
}

void clearAllFuseSurfaceClues(FuseSurface* block){
	g_vcArtTriBase->fuseSurface = 0;
	for(int k = g_sceneTreeList.length; --k>=0;){
		if(FuseSurface* fuseSurface = dynamic_cast<FuseSurface*>(g_sceneTreeList[k].node.p)){
			if(block!=fuseSurface){
				fuseSurface->polyMapAid.refValue()=false;
				fuseSurface->polyMapAidMirror.refValue()=false;
			}
		}
	}
}
void FuseSurface::PolyMapAid::operator=(const bool& par){
	refValue()=par;
	home.polyMapAidMirror.refValue()=false;
	if(par){
		g_vcArtTriBase->fuseSurface=&home;
	}
	clearAllFuseSurfaceClues(&home);
}
void FuseSurface::PolyMapAidMirror::operator=(const bool& par){
	refValue()=par;
	home.polyMapAid.refValue()=false;
	if(par){
		g_vcArtTriBase->fuseSurface=&home;
	}
	clearAllFuseSurfaceClues(&home);
}
//#include <time.h>
template <class T> void inline updateFlatVertices(NurbsPatch& nurbsPatch, M_<T>& flatCoord, const M_Vec3f& guessCoord, const M_Vec3f& originalCoord, const int& bits, const float& coordChange){
	float* stemUKnotPtr = nurbsPatch.uKnot.refValue().getArray();
	float* stemVKnotPtr = nurbsPatch.vKnot.refValue().getArray();
	Vec3f* stemPointPtr = nurbsPatch.point.refValue().getArray();
	float* stemWeightPtr = nurbsPatch.weight.refValue().getArray();
	int uOrder = nurbsPatch.order.x;
	int vOrder = nurbsPatch.order.y;
	int uDim = nurbsPatch.dim.x;
	int vDim = nurbsPatch.dim.y;

	float maxU = nurbsPatch.maxU();
	float maxV = nurbsPatch.maxV();
	float minU = nurbsPatch.minU();
	float minV = nurbsPatch.minV();

	for(int i = flatCoord.length; --i>=0;){
		T pt = flatCoord[i];

		if(1 & bits){
			pt.x-=coordChange;
		}
		if(2 & bits){
			pt.x+=coordChange;
		}
		if(pt.y > 0 && pt.y < 1){
			if(4 & bits){
				pt.y-=coordChange;
			}
			if(8 & bits){
				pt.y+=coordChange;
			}
		}
		if(pt.x < 0){
			pt.x += 1;
		}
		else if(pt.x > 1){
			pt.x -= 1;
		}
		if(pt.y < 0){
			pt.y = 0;
		}
		else if(pt.y > 1){
			pt.y = 1;
		}

		float u = (maxU-minU)*pt.x + minU;
		float v = (maxV-minV)*pt.y + minV;

		Vec3f curvedEdgePoint;

		SurfacePoint(u, v, uOrder, vOrder, uDim, vDim, stemUKnotPtr, stemVKnotPtr, stemPointPtr, stemWeightPtr, &curvedEdgePoint);

		Vec3f diffA = guessCoord[i] - originalCoord[i];
		Vec3f diffB = curvedEdgePoint - originalCoord[i];
		float fdA = fabs(sqrt(diffA.x*diffA.x + diffA.y*diffA.y + diffA.z*diffA.z));
		float fdB = fabs(sqrt(diffB.x*diffB.x + diffB.y*diffB.y + diffB.z*diffB.z));
//		float fdA = fabs(sqrt(diffA.x*diffA.x + diffA.z*diffA.z));
//		float fdB = fabs(sqrt(diffB.x*diffB.x + diffB.z*diffB.z));
		if(fdB < fdA){
			flatCoord[i] = pt;
		}
	}
}



void inline updateFlatVertices(M_Node& surface, M_Int& patchOwner, M_Vec2f& flatCoord, const M_Vec3f& guessCoord, const M_Vec3f& originalCoord, const int& bits, const float& coordChange){

	for(int i = flatCoord.length; --i>=0;){
		Vec2f pt = flatCoord[i];

		if(1 & bits){
			pt.x-=coordChange;
		}
		if(2 & bits){
			pt.x+=coordChange;
		}
		if(pt.y > 0 && pt.y < 1){
			if(4 & bits){
				pt.y-=coordChange;
			}
			if(8 & bits){
				pt.y+=coordChange;
			}
		}
		if(pt.x < 0){
			pt.x += 1;
		}
		else if(pt.x > 1){
			pt.x -= 1;
		}
		if(pt.y < 0){
			pt.y = 0;
		}
		else if(pt.y > 1){
			pt.y = 1;
		}


		if(NurbsPatch* nurbsPatch = dynamic_cast<NurbsPatch*>(surface[patchOwner[i]].p)){
			float* stemUKnotPtr = nurbsPatch->uKnot.refValue().getArray();
			float* stemVKnotPtr = nurbsPatch->vKnot.refValue().getArray();
			Vec3f* stemPointPtr = nurbsPatch->point.refValue().getArray();
			float* stemWeightPtr = nurbsPatch->weight.refValue().getArray();
			int uOrder = nurbsPatch->order.x;
			int vOrder = nurbsPatch->order.y;
			int uDim = nurbsPatch->dim.x;
			int vDim = nurbsPatch->dim.y;

			float maxU = nurbsPatch->maxU();
			float maxV = nurbsPatch->maxV();
			float minU = nurbsPatch->minU();
			float minV = nurbsPatch->minV();

			Vec3f curvedEdgePoint;

			float u = (maxU-minU)*pt.x + minU;
			float v = (maxV-minV)*pt.y + minV;
			SurfacePoint(u, v, uOrder, vOrder, uDim, vDim, stemUKnotPtr, stemVKnotPtr, stemPointPtr, stemWeightPtr, &curvedEdgePoint);

			Vec3f diffA = guessCoord[i] - originalCoord[i];
			Vec3f diffB = curvedEdgePoint - originalCoord[i];
			float fdA = fabs(sqrt(diffA.x*diffA.x + diffA.y*diffA.y + diffA.z*diffA.z));
			float fdB = fabs(sqrt(diffB.x*diffB.x + diffB.y*diffB.y + diffB.z*diffB.z));
			if(fdB < fdA){
				flatCoord[i] = pt;
			}
		}
	}
}

void tugVertex(int i, Vec3f pt, M_Vec3f& dpoint, M_Vec3f& point, const M_Float& weight, const M_Vec3f& originalCoord, const M_Vec3f& guessCoord, NurbsPatch* patch, float& sumDiff, float& averageDiff, bool mirrorTug){
	Vec3f remPoint = point[i];
	Vec3f remMirrorPoint;
	int mirrorIdx = -1;

	if(mirrorTug){
		Vec3f svec = dpoint[i];
		if(svec.x==0){
			pt.x=0;
		}
		svec.x=-svec.x;
		mirrorIdx = findMirrorVertex(dpoint, svec);
		if(mirrorIdx>=0){
			remMirrorPoint = point[mirrorIdx];
			svec=pt;
			svec.x=-svec.x;
			point[mirrorIdx]=svec*weight[mirrorIdx];
		}
	}

	point[i]=pt*weight[i];

	patch->processFields();
	
	patch->doBuild();

	float testSumDiff=0;
	for(int j = guessCoord.length; --j>=0;){
		Vec3f diff = guessCoord[j] - originalCoord[j];
		testSumDiff += fabs(sqrt(diff.x*diff.x + diff.y*diff.y + diff.z*diff.z));
	}
	//dLog(testSumDiff);
	if(testSumDiff<sumDiff){
		sumDiff = testSumDiff;
		averageDiff = testSumDiff/guessCoord.length;
		//polyToNurbs->processFields();
	}
	else{//undo
		point[i] = remPoint;
		if(mirrorIdx>=0){
			point[mirrorIdx] = remMirrorPoint;
		}
		patch->processFields();
	}
}

void tugVertex(int i, Vec3f pt, M_Vec3f& dpoint, M_Vec3f& point, const M_Float& weight, const M_Vec3f& originalCoord, const M_Vec3f& guessCoord, Polymap* polymap, NurbsPatch* patch, float& sumDiff, float& averageDiff){
	Seamless* smls = patch->smls;
	Vec3f remPoint = point[i];
	Vec3f remMirrorPoint;
	int mirrorIdx = -1;

	point[i]=pt*weight[i];

	smls->calcAllPolyMidPoints();
	patch->processFields();
	
	polymap->doBuild();

	float testSumDiff=0;
	for(int j = guessCoord.length; --j>=0;){
		Vec3f diff = guessCoord[j] - originalCoord[j];
		testSumDiff += fabs(sqrt(diff.x*diff.x + diff.y*diff.y + diff.z*diff.z));
	}
	//dLog(testSumDiff);
	if(testSumDiff<sumDiff){
		sumDiff = testSumDiff;
		averageDiff = testSumDiff/guessCoord.length;
		//polyToNurbs->processFields();
	}
	else{//undo
		point[i] = remPoint;
		if(mirrorIdx>=0){
			point[mirrorIdx] = remMirrorPoint;
		}
		smls->calcAllPolyMidPoints();
		patch->processFields();
	}
}

void updateDiff(const M_Vec3f& originalCoord, const M_Vec3f& guessCoord, float& sumDiff, float& averageDiff){
	float testSumDiff=0;
	for(int i = guessCoord.length; --i>=0;){
		Vec3f diff = guessCoord[i] - originalCoord[i];
		testSumDiff += fabs(sqrt(diff.x*diff.x + diff.y*diff.y + diff.z*diff.z));
	}
	sumDiff = testSumDiff;
	averageDiff = testSumDiff/guessCoord.length;
}

void PolyToNurbs::doBuild(){
	const int randomLen = 127;
	if(NurbsPatch* patch = dynamic_cast<NurbsPatch*>(this->mimic.p)){
		bool xTug = patch->xTug;
		bool yTug = patch->yTug;
		bool zTug = patch->zTug;
		const bool mirrorTug = patch->mirror;
		if(patch->stem.length){
			Stem* stem = dynamic_cast<Stem*>(patch->stem[0].p);
			stem->add.refValue() = false;
			patch->doBuild();
			if(Art* originalPart = dynamic_cast<Art*>(this->original.p)){
				M_SArt& part = stem->part.refValue();
				if(part.length){
					M_SArt& source = stem->source.refValue();
					Art* guessPart = part[0];
					const M_Vec3f& guessCoord = guessPart->coord;
					const M_Vec3f& originalCoord = originalPart->coord;
					const int uDim = patch->dim.x;
					const int vDim = patch->dim.y;


					if(source.length){
						Art* flatPart = source[0];
						M_Vec3f& flatCoord = flatPart->coord;

						M_Vec3f& dpoint = patch->dpoint;
						M_Vec3f& point = patch->point;
						M_Float& weight = patch->weight;

						int r = offset;
						r++;
						if(r>=randomLen){
							r = 0;
						}
						offset.refValue()=r;

						int bits=initPolyToNurbs_random[r];

						updateDiff(originalCoord, guessCoord, sumDiff.refValue(), averageDiff.refValue());

						if(bond && coordChange != 0){
							updateFlatVertices(*patch, flatCoord, guessCoord, originalCoord, bits, coordChange);
						}
							//srand( (unsigned)time( NULL ) );
						if(evolve && pointChange != 0){
							for(int v = 0; v < vDim; v++){
								for(int u = 0; u < uDim; u+=1){
									int i = v * uDim + u;

									int r = offset;
									r++;
									if(r>=randomLen){
										r = 0;
									}
									offset.refValue()=r;

									int bits=initPolyToNurbs_random[r];


									Vec3f pt = dpoint[i];
									//int bits = rand();
									//bits&=0x7;

									//dLog(bits);

									if(mirrorTug && pt.x < 0){
										continue;
									}
									if(1 & bits){
										pt.x-=pointChange;
									}
									if(2 & bits){
										pt.x+=pointChange;
									}
									if(4 & bits){
										pt.y-=pointChange;
									}
									if(8 & bits){
										pt.y+=pointChange;
									}
									if(16 & bits){
										pt.z-=pointChange;
									}
									if(32 & bits){
										pt.z+=pointChange;
									}
									tugVertex(i, pt, dpoint, point, weight, originalCoord, guessCoord, patch, sumDiff.refValue(), averageDiff.refValue(), mirrorTug);
								}
							}
						}
					}
					else{
						M_Vec2f& flatCoord = stem->coord.refValue();

						M_Vec3f& dpoint = patch->dpoint;
						M_Vec3f& point = patch->point;
						M_Float& weight = patch->weight;

						int r = offset;
						r++;
						if(r>=randomLen){
							r = 0;
						}
						offset.refValue()=r;

						int bits=initPolyToNurbs_random[r];

						updateDiff(originalCoord, guessCoord, sumDiff.refValue(), averageDiff.refValue());

						if(bond && coordChange != 0){
							updateFlatVertices(*patch, flatCoord, guessCoord, originalCoord, bits, coordChange);
						}
							//srand( (unsigned)time( NULL ) );
						if(evolve && pointChange != 0){
							for(int v = 0; v < vDim; v++){
								for(int u = 0; u < uDim; u+=1){
									int i = v * uDim + u;

									int r = offset;
									r++;
									if(r>=randomLen){
										r = 0;
									}
									offset.refValue()=r;

									int bits=initPolyToNurbs_random[r];


									Vec3f pt = dpoint[i];
									//int bits = rand();
									//bits&=0x7;

									//dLog(bits);

									if(mirrorTug && pt.x < 0){
										continue;
									}
									if(xTug){
										if(1 & bits){
											pt.x-=pointChange;
										}
										if(2 & bits){
											pt.x+=pointChange;
										}
									}
									if(yTug){
										if(4 & bits){
											pt.y-=pointChange;
										}
										if(8 & bits){
											pt.y+=pointChange;
										}
										if(pt.y < 0){
											pt.y = 0;
										}
									}
									if(zTug){
										if(16 & bits){
											pt.z-=pointChange;
										}
										if(32 & bits){
											pt.z+=pointChange;
										}
									}
									tugVertex(i, pt, dpoint, point, weight, originalCoord, guessCoord, patch, sumDiff.refValue(), averageDiff.refValue(), mirrorTug);
								}
							}
						}
					}
				}
			}
		}
	}
	if(Polymap* polymap = dynamic_cast<Polymap*>(this->mimic.p)){
		M_Node& surface = polymap->surface.refValue();
		M_Int& patchOwner = polymap->patchOwner.refValue();
		if(surface.length){
			polymap->doBuild();
			polymap->add.refValue() = false;
			if(Art* originalPart = dynamic_cast<Art*>(this->original.p)){
				Art* guessPart = polymap->part.refValue().p;
				const M_Vec3f& guessCoord = guessPart->coord;
				const M_Vec3f& originalCoord = originalPart->coord;

				M_Vec2f& flatCoord = polymap->coord.refValue();

				int r = offset;
				r++;
				if(r>=randomLen){
					r = 0;
				}
				offset.refValue()=r;

				int bits=initPolyToNurbs_random[r];

				updateDiff(originalCoord, guessCoord, sumDiff.refValue(), averageDiff.refValue());

				if(bond && coordChange != 0){
					updateFlatVertices(surface, patchOwner, flatCoord, guessCoord, originalCoord, bits, coordChange);
				}
					//srand( (unsigned)time( NULL ) );
				if(evolve && pointChange != 0){
					for(int k = surface.length; --k>=0;){
						NurbsPatch* patch = dynamic_cast<NurbsPatch*>(surface[k].p);
						const int uDim = patch->dim.x;
						const int vDim = patch->dim.y;
						M_Vec3f& dpoint = patch->dpoint;
						M_Vec3f& point = patch->point;
						M_Float& weight = patch->weight;

						for(int v = 0; v < vDim; v++){
							for(int u = 0; u < uDim; u+=1){
								int i = v * uDim + u;

								int r = offset;
								r++;
								if(r>=randomLen){
									r = 0;
								}
								offset.refValue()=r;

								int bits=initPolyToNurbs_random[r];


								Vec3f pt = dpoint[i];
								//int bits = rand();
								//bits&=0x7;

								//dLog(bits);

								//if(mirrorTug && pt.x < 0){
								//	continue;
								//}
								if(1 & bits){
									pt.x-=pointChange;
								}
								if(2 & bits){
									pt.x+=pointChange;
								}
								if(4 & bits){
									pt.y-=pointChange;
								}
								if(8 & bits){
									pt.y+=pointChange;
								}
								if(16 & bits){
									pt.z-=pointChange;
								}
								if(32 & bits){
									pt.z+=pointChange;
								}
								tugVertex(i, pt, dpoint, point, weight, originalCoord, guessCoord, polymap, patch, sumDiff.refValue(), averageDiff.refValue());
							}
						}
					}
				}
			}
		}
	}
}
/*
void inline updateFlatVertices(NurbsPatch& nurbsPatch, M_Vec3f& flatCoord, const M_Vec3f& guessCoord, const M_Vec3f& originalCoord, const M_Int& random, const int& bits, const float& coordChange){
	float* stemUKnotPtr = nurbsPatch.uKnot.refValue().getArray();
	float* stemVKnotPtr = nurbsPatch.vKnot.refValue().getArray();
	Vec3f* stemPointPtr = nurbsPatch.point.refValue().getArray();
	float* stemWeightPtr = nurbsPatch.weight.refValue().getArray();
	int uOrder = nurbsPatch.order.x;
	int vOrder = nurbsPatch.order.y;
	int uDim = nurbsPatch.dim.x;
	int vDim = nurbsPatch.dim.y;

	float maxU = nurbsPatch.maxU();
	float maxV = nurbsPatch.maxV();
	float minU = nurbsPatch.minU();
	float minV = nurbsPatch.minV();

	for(int i = flatCoord.length; --i>=0;){
		Vec3f pt = flatCoord[i];

		if(1 & bits){
			pt.x-=coordChange;
		}
		if(2 & bits){
			pt.x+=coordChange;
		}
		if(pt.y > 0 && pt.y < 1){
			if(4 & bits){
				pt.y-=coordChange;
			}
			if(8 & bits){
				pt.y+=coordChange;
			}
		}
		if(pt.x < 0){
			pt.x += 1;
		}
		else if(pt.x > 1){
			pt.x -= 1;
		}
		if(pt.y < 0){
			pt.y = 0;
		}
		else if(pt.y > 1){
			pt.y = 1;
		}

		float u = (maxU-minU)*pt.x + minU;
		float v = (maxV-minV)*pt.y + minV;

		Vec3f curvedEdgePoint;

		SurfacePoint(u, v, uOrder, vOrder, uDim, vDim, stemUKnotPtr, stemVKnotPtr, stemPointPtr, stemWeightPtr, &curvedEdgePoint);

		Vec3f diffA = guessCoord[i] - originalCoord[i];
		Vec3f diffB = curvedEdgePoint - originalCoord[i];
		float fdA = fabs(sqrt(diffA.x*diffA.x + diffA.y*diffA.y + diffA.z*diffA.z));
		float fdB = fabs(sqrt(diffB.x*diffB.x + diffB.y*diffB.y + diffB.z*diffB.z));
		if(fdB < fdA){
			flatCoord[i] = pt;
		}
	}
}
void PolyToNurbs::doBuild(){
	if(NurbsPatch* patch = dynamic_cast<NurbsPatch*>(this->mimic.p)){
		const bool mirrorTug = patch->mirrorTug;
		if(patch->stem.length){
			Stem* stem = dynamic_cast<Stem*>(patch->stem[0].p);
			patch->doBuild();
			if(Art* originalPart = dynamic_cast<Art*>(this->original.p)){
				M_SArt& part = stem->part.refValue();
				if(part.length){
					M_SArt& source = stem->source.refValue();
					Art* guessPart = part[0];
					const M_Vec3f& guessCoord = guessPart->coord;
					const M_Vec3f& originalCoord = originalPart->coord;
					const int uDim = patch->dim.x;
					const int vDim = patch->dim.y;


					if(source.length){
						Art* flatPart = source[0];
						M_Vec3f& flatCoord = flatPart->coord;

						M_Vec3f& dpoint = patch->dpoint;
						M_Vec3f& point = patch->point;
						M_Float& weight = patch->weight;

						if(update){

							int r = offset;
							r++;

							offset.refValue()=r;

							int random2[] = {1, 2, 4, 8};
							int bits=random2[r & 0x3];

							updateFlatVertices(*patch, flatCoord, guessCoord, originalCoord, random, bits, coordChange);
							//srand( (unsigned)time( NULL ) );
							if(pointChange){
								int random2[] = {1, 2, 4, 8, 16, 32};
								for(int k = cycle; --k>=0;){
									for(int v = 0; v < vDim; v++){
										for(int u = 1; u < uDim; u+=2){
											int i = v * uDim + u;
											if(i == 63){
												int a = 0;
											}

											if(r>=sizeof(random2)/sizeof(int)){
												r = 0;
											}
											//offset.refValue()=r;

											int bits=random2[r];
											//int i = 48;{
											//if(i!=46)continue;
			//								if(i==46){
			//									i=i;
			//								}


											Vec3f pt = dpoint[i];
											//int bits = rand();
											//bits&=0x7;

											//dLog(bits);


											if(mirrorTug && pt.x < 0){
												continue;
											}
											if(1 & bits){
												pt.x-=pointChange;
											}
											if(2 & bits){
												pt.x+=pointChange;
											}
											if(4 & bits){
												pt.y-=pointChange;
											}
											if(8 & bits){
												pt.y+=pointChange;
											}
											if(16 & bits){
												pt.z-=pointChange;
											}
											if(32 & bits){
												pt.z+=pointChange;
											}
											tugVertex(i, pt, dpoint, point, weight, originalCoord, guessCoord, patch, sumDiff.refValue(), averageDiff.refValue(), mirrorTug);
										}
									}
								}
								patch->doBuild();
								houseKeepVisualClue();
							}
							else{
								goto here;
							}
						}
						else{
here:
							float testSumDiff=0;
							for(int i = guessCoord.length; --i>=0;){
								Vec3f diff = guessPart->coord[i] - originalPart->coord[i];
								testSumDiff += fabs(sqrt(diff.x*diff.x + diff.y*diff.y + diff.z*diff.z));
							}
							sumDiff.refValue() = testSumDiff;
							averageDiff.refValue() = testSumDiff/guessCoord.length;
						}
					}
				}
			}
		}
	}
}
*/

template <class T> void inline spreadVertices(NurbsPatch& nurbsPatch, M_<T>& flatCoord, M_Vec3f& guessCoord, const M_Vec3f& originalCoord, const int& spreadRes, const float& spreadResFloat){
	float* stemUKnotPtr = nurbsPatch.uKnot.refValue().getArray();
	float* stemVKnotPtr = nurbsPatch.vKnot.refValue().getArray();
	Vec3f* stemPointPtr = nurbsPatch.point.refValue().getArray();
	float* stemWeightPtr = nurbsPatch.weight.refValue().getArray();
	int uOrder = nurbsPatch.order.x;
	int vOrder = nurbsPatch.order.y;
	int uDim = nurbsPatch.dim.x;
	int vDim = nurbsPatch.dim.y;

	float maxU = nurbsPatch.maxU();
	float maxV = nurbsPatch.maxV();
	float minU = nurbsPatch.minU();
	float minV = nurbsPatch.minV();

	for(int i = flatCoord.length; --i>=0;){
		for(int uPt = 0; uPt < spreadRes; uPt++){
			for(int vPt = 0; vPt < spreadRes; vPt++){
				T pt;
				pt.x = uPt / spreadResFloat;
				pt.y = vPt / spreadResFloat;

				float u = (maxU-minU)*pt.x + minU;
				float v = (maxV-minV)*pt.y + minV;

				Vec3f curvedEdgePoint;

				SurfacePoint(u, v, uOrder, vOrder, uDim, vDim, stemUKnotPtr, stemVKnotPtr, stemPointPtr, stemWeightPtr, &curvedEdgePoint);

				Vec3f diffA = guessCoord[i] - originalCoord[i];
				Vec3f diffB = curvedEdgePoint - originalCoord[i];
				float fdA = fabs(sqrt(diffA.x*diffA.x + diffA.y*diffA.y + diffA.z*diffA.z));
				float fdB = fabs(sqrt(diffB.x*diffB.x + diffB.y*diffB.y + diffB.z*diffB.z));
				if(fdB < fdA){
					flatCoord[i] = pt;
					guessCoord[i] = curvedEdgePoint;
				}
			}
		}
	}
}

void inline spreadVertices(NurbsPatch& nurbsPatch, M_Vec2f& flatCoord, M_Int& patchOwner, const int& patchOwnerIdx, M_Vec3f& guessCoord, const M_Vec3f& originalCoord, const int& spreadRes, const float& spreadResFloat){
	float* stemUKnotPtr = nurbsPatch.uKnot.refValue().getArray();
	float* stemVKnotPtr = nurbsPatch.vKnot.refValue().getArray();
	Vec3f* stemPointPtr = nurbsPatch.point.refValue().getArray();
	float* stemWeightPtr = nurbsPatch.weight.refValue().getArray();
	int uOrder = nurbsPatch.order.x;
	int vOrder = nurbsPatch.order.y;
	int uDim = nurbsPatch.dim.x;
	int vDim = nurbsPatch.dim.y;

	float maxU = nurbsPatch.maxU();
	float maxV = nurbsPatch.maxV();
	float minU = nurbsPatch.minU();
	float minV = nurbsPatch.minV();

	for(int i = flatCoord.length; --i>=0;){
		for(int uPt = 0; uPt < spreadRes; uPt++){
			for(int vPt = 0; vPt < spreadRes; vPt++){
				Vec2f pt = Vec2f(uPt / spreadResFloat, vPt / spreadResFloat);

				float u = (maxU-minU)*pt.x + minU;
				float v = (maxV-minV)*pt.y + minV;

				Vec3f curvedEdgePoint;

				SurfacePoint(u, v, uOrder, vOrder, uDim, vDim, stemUKnotPtr, stemVKnotPtr, stemPointPtr, stemWeightPtr, &curvedEdgePoint);

				Vec3f diffA = guessCoord[i] - originalCoord[i];
				Vec3f diffB = curvedEdgePoint - originalCoord[i];
				float fdA = fabs(sqrt(diffA.x*diffA.x + diffA.y*diffA.y + diffA.z*diffA.z));
				float fdB = fabs(sqrt(diffB.x*diffB.x + diffB.y*diffB.y + diffB.z*diffB.z));
				if(fdB < fdA){
					flatCoord[i] = pt;
					guessCoord[i] = curvedEdgePoint;
					patchOwner[i] = patchOwnerIdx;
				}
			}
		}
	}
}

void PolyToNurbs::Spread::operator=(const bool& par){
	if(NurbsPatch* patch = dynamic_cast<NurbsPatch*>(home.mimic.p)){
		const bool mirrorTug = patch->mirror;
		if(patch->stem.length){
			Stem* stem = dynamic_cast<Stem*>(patch->stem[0].p);
			patch->doBuild();
			if(Art* originalPart = dynamic_cast<Art*>(home.original.p)){
				M_SArt& part = stem->part.refValue();
				if(part.length){
					M_SArt& source = stem->source.refValue();
					Art* guessPart = part[0];
					M_Vec3f& guessCoord = guessPart->coord.refValue();
					const M_Vec3f& originalCoord = originalPart->coord;
					const int uDim = patch->dim.x;
					const int vDim = patch->dim.y;


					if(source.length){
						Art* flatPart = source[0];
						M_Vec3f& flatCoord = flatPart->coord;

						M_Vec3f& dpoint = patch->dpoint;
						M_Vec3f& point = patch->point;
						M_Float& weight = patch->weight;

						spreadVertices(*patch, flatCoord, guessCoord, originalCoord, home.spreadRes, home.spreadRes);
					}
					else{
						M_Vec2f& flatCoord = stem->coord.refValue();

						M_Vec3f& dpoint = patch->dpoint;
						M_Vec3f& point = patch->point;
						M_Float& weight = patch->weight;

						spreadVertices(*patch, flatCoord, guessCoord, originalCoord, home.spreadRes, home.spreadRes);
					}
				}
			}
		}
	}
	else if(Polymap* polymap = dynamic_cast<Polymap*>(home.mimic.p)){
		if(Art* originalPart = dynamic_cast<Art*>(home.original.p)){
			if(Art* guessPart = polymap->part.refValue()){
				M_Vec3f& guessCoord = guessPart->coord.refValue();
				M_Vec2f& flatCoord = polymap->coord.refValue();
				M_Int& patchOwner = polymap->patchOwner.refValue();
				patchOwner.length = flatCoord.length;
				const M_Vec3f& originalCoord = originalPart->coord;
				M_Node& surface = polymap->surface.refValue();

				for(int k = 0; k < surface.length; k++){
					NurbsPatch* patch = dynamic_cast<NurbsPatch*>(surface[k].p);

					spreadVertices(*patch, flatCoord, patchOwner, k, guessCoord, originalCoord, home.spreadRes, home.spreadRes);

				}
			}
		}
	}
}
void convertMercatorToNorthPole(M_Vec3f& coord);
void convertMercatorToSouthPole(M_Vec3f& coord);
void convertNorthPoleToMercator(M_Vec3f& coord);
void convertSouthPoleToMercator(M_Vec3f& coord);

void FuseSurface::doSpread(){
	M_Vec2f& baseCoord = this->baseCoord.refValue();
	if(NurbsPatch* parentPatch = this->parentPatch){
		if(NurbsPatch* patch = dynamic_cast<NurbsPatch*>(this->patch.p)){
			if(patch->stem.length){
				if(Stem* stem = dynamic_cast<Stem*>(patch->stem.refValue()[0].p)){
					M_SArt& part = stem->part.refValue();
					if(part.length){
						M_SArt& source = stem->source.refValue();
						Art* originalPart = part[0];
						const M_Vec3f& originalCoord = originalPart->coord;
						const int uDim = parentPatch->dim.x;
						const int vDim = parentPatch->dim.y;

						M_Vec3f& dpoint = parentPatch->dpoint;
						M_Vec3f& point = parentPatch->point;
						M_Float& weight = parentPatch->weight;

						M_Vec3f guessCoord;
						guessCoord.length = baseCoord.length;
						M_Vec3f baseCoord3f;
						baseCoord3f.length = baseCoord.length;

						spreadVertices(*parentPatch, baseCoord3f, guessCoord, originalCoord, spreadRes, spreadRes);

						if(parentStem->northPole){
							convertMercatorToNorthPole(baseCoord3f);
						}
						else if(parentStem->southPole){
							convertMercatorToSouthPole(baseCoord3f);
						}
						for(int i = baseCoord.length; --i>=0;){
							baseCoord[i] = Vec2f(baseCoord3f[i].x, baseCoord3f[i].y);
						}
					}
				}
			}
		}
	}
}
void FuseSurface::Spread::operator=(const bool& par){
	refValue()=true;
}
bool catchSelCoord(Seamless* smls){
	if(Part* part = dynamic_cast<Part*>(g_selNode.getRaw())){
		M_Vec3f& coord = part->coord.refValue();
		if(smls->ActCtrlPnt < coord.length){
			if(smls->selCoordPos == coord[smls->ActCtrlPnt]){
				return true;
			}
		}
	}
	return false;
}

void insertSection(M_Vec3f& keyValue, int channels, int sectionIndex){
	int from = keyValue.length;
	keyValue.length += channels;
	int to = sectionIndex * channels;
	for(int i = from; --i >= to;){
		keyValue[i + channels] = keyValue[i];
	}
}

void CPMorpher::insertChannel(int& channel, const int k, M_Int& cpIndex, M_Vec3f& keyValue, int& numOfChannels, const int numOfSections){
	int from = keyValue.length;
	keyValue.length += numOfSections;
	int dist = numOfSections;
	int cycle = numOfChannels - channel;

	for(int i = from; --i >= 0;){
		if(cycle == 0){
			keyValue[i + dist] = Vec3f();
			cycle = numOfChannels;
			dist--;
		}
		keyValue[i + dist] = keyValue[i];
		--cycle;
	}
	numOfChannels++;
	cpIndex.insertElement(channel, k);
	channel++;
}

void CPMorpher::insertChannel(int& channel, const int k, M_Int& cpIndex, const float slider, const Vec3f& diff, M_Vec3f& keyValue, int& numOfChannels, const int numOfSections){
	if(key.length == 0){
		keyValue.length = key.refValue().length = 2;
		key.refValue()[1] = 1;
	}

	int remChannel = channel;
	insertChannel(channel, k, cpIndex, keyValue, numOfChannels, numOfSections);
	take(slider, diff, remChannel, numOfChannels);
}

inline void CPMorpher::deleteChannel(const int& channel, M_Int& cpIndex, M_Float& key, M_Vec3f& keyValue, const int numOfChannels){
	index[0] = 0;
	index[1] = 0;
	cpIndex.deleteElement(channel);
	if(cpIndex.length == 0){
		key.length = 0;
		keyValue.length = 0;
		return;
	}
	int cycle = numOfChannels - 1;
	int dst = channel;
	int src = dst + 1;
	int less = 0;
	while(src < keyValue.length){
		keyValue[dst++] = keyValue[src++];

		if(--cycle == 0){
			src++;
			less++;
			cycle = numOfChannels - 1;
		}
	}
	keyValue.length -= key.length;
}

void CPMorpher::deleteChannelCoords(const Art* part){
	for(int k = MultiSelect::coordList.length; --k>=0;){
		if(part == MultiSelect::coordList[k].p){
			int idx = MultiSelect::coordList[k].i;
			for(int i = cpIndex.length; --i>=0;){
				if(cpIndex[i] == idx){
					deleteChannel(i, cpIndex.refValue(), key.refValue(), keyValue.refValue(), cpIndex.length);
					break;
				}
			}
		}
	}
}

int Art::getControlPointIndexFromPointListItem(const NurbsPatch* patch, const int idx)const{
	for(int i = controlPoint.length; --i>=0;){
		if(controlPoint[i].p == patch && controlPoint[i].i == idx){
			return i;
		}
	}
	return -1;
}

void CPMorpher::deleteChannelPoints(const Art* part){
	for(int k = MultiSelect::pointList.length; --k>=0;){
		if(int idx = part->getControlPointIndexFromPointListItem(MultiSelect::pointList[k].p, MultiSelect::pointList[k].i)){
			for(int i = cpIndex.length; --i>=0;){
				if(cpIndex[i] == idx){
					deleteChannel(i, cpIndex.refValue(), key.refValue(), keyValue.refValue(), keyValue.length / key.length);
					break;
				}
			}
		}
	}
}

void CPMorpher::take(float slider, const Vec3f& diff, const int channel, const int numOfChannels){
	M_Float& key = this->key;
	M_Vec3f& keyValue = this->keyValue;

	if(key.length == 0){
		key += slider;
		keyValue.length = key.length * numOfChannels;
		keyValue[channel] = diff;
	}
	else{
		
		for(int i = 0; i < key.length; i++){
			if(roughEQ(slider, key[i])){
				keyValue[numOfChannels * i + channel] = diff;
				index[0] = i;
				break;
			}
			else if(roughGT(key[i], slider)){
				key.insertElement(i, slider);
				insertSection(keyValue, numOfChannels, i);
				keyValue[numOfChannels * i + channel] = diff;
				index[0] = i;
				break;
			}
			else if(i == key.length - 1){
				i++;
				key.insertElement(i, slider);
				insertSection(keyValue, numOfChannels, i);
				keyValue[numOfChannels * i + channel] = diff;
				index[0] = i;
				break;
			}
		}
	}
}

void NurbsLathe::Scale::operator=(const Vec3f& par){
	refValue()=par;
	home.processFields();
	if(NurbsLathe* mirrorBuild = home.getMirrorLathe()){
		mirrorBuild->scale.refValue() = par;
		mirrorBuild->processFields();
	}
}

void NurbsLathe::HRotation::operator=(const Rotation& par){
	refValue()=par;
	home.processFields();
	if(NurbsLathe* mirrorBuild = home.getMirrorLathe()){
		mirrorBuild->rotation.refValue() = Rotation(par.x, -par.y, -par.z, par.angle);
		mirrorBuild->processFields();
	}
}

void NurbsLathe::Translation::operator=(const Vec3f& par){
	refValue()=par;
	home.processFields();
	if(NurbsLathe* mirrorBuild = home.getMirrorLathe()){
		mirrorBuild->translation.refValue() = Vec3f(-par.x, par.y, par.z);
		mirrorBuild->processFields();
	}
}

void SphereBuild::Scale::operator=(const Vec3f& par){
	refValue()=par;
	home.processFields();
	if(SphereBuild* mirrorBuild = home.getMirrorSphere()){
		mirrorBuild->scale.refValue() = par;
		mirrorBuild->processFields();
	}
}

void SphereBuild::HRotation::operator=(const Rotation& par){
	refValue()=par;
	home.processFields();
	if(SphereBuild* mirrorBuild = home.getMirrorSphere()){
		mirrorBuild->rotation.refValue() = Rotation(par.x, -par.y, -par.z, par.angle);
		mirrorBuild->processFields();
	}
}

void SphereBuild::Translation::operator=(const Vec3f& par){
	refValue()=par;
	home.processFields();
	if(SphereBuild* mirrorBuild = home.getMirrorSphere()){
		mirrorBuild->translation.refValue() = Vec3f(-par.x, par.y, par.z);
		mirrorBuild->processFields();
	}
}

void Art::genDynamicTransforms(M_Matrix& m_matrix, M_Int& m_smi, Matrix parentMatrix, int& smi){
	
	Matrix pivotPointMatrix = translationMatrix(translation + pivotPoint) * parentMatrix;
	int coordLen = coord.length;
	int groupLen = coordGroup.length;
	m_smi.length += coordLen;
	int k = 0;
	int coordsInGroups = 0;

	for(int cg = 0; cg < groupLen; cg++){
		int mark = k;
		for(int c = coordGroup[cg]; --c >= 0 && k < coordLen; k++){
			m_smi[fromCoord + k] = smi;
		}
		if(k > mark){
			m_matrix += scaleMatrix(scale) * rotationMatrix(Rotation(rotation.x, rotation.y, rotation.z, rotation.angle * weight[cg])) * pivotPointMatrix;
			smi++;
		}
	}

	pivotPointMatrix = scaleMatrix(scale) * rotationMatrix(rotation) * pivotPointMatrix;

	int mark = k;
	for(int len = coord.length; k < len; k++){
		m_smi[fromCoord + k] = smi;
	}
	if(k > mark){
		m_matrix += pivotPointMatrix;
		smi++;
	}

	for(int k = 0; k < part.length; k++){
		part.refValue()[k]->genDynamicTransforms(m_matrix, m_smi, pivotPointMatrix, smi);
	}
}

inline bool checkPointInVertex(Art*& part, const bool& perspective, const int& i){
	if(part->coord.length==0)return false;

	Vec3f sr = m_matrix[m_smi[part->fromCoord + i]] * g_viewMatrix * (g_m_coord[part->fromCoord + i] - part->absolutePivotPoint);

	Vec3f v;
	Vec3f mv;

	if(perspective){
		v.x = sr.x / (-sr.z + g_viewDist);
		v.y = sr.y / (-sr.z + g_viewDist);

		mv.x = g_pointer.translation.x / (-g_pointer.translation.z + g_viewDist);
		mv.y = g_pointer.translation.y / (-g_pointer.translation.z + g_viewDist);
	}
	else{
		v.x=sr.x;
		v.y=sr.y;

		mv.x=g_pointer.translation.x;
		mv.y=g_pointer.translation.y;
	}
	
	if(
		v.x < mv.x+MouseSense&&
		v.x > mv.x-MouseSense&&
		v.y < mv.y+MouseSense&&
		v.y > mv.y-MouseSense){

				
		int corner;
		S_Triangle* triangle;
		S_Triangle* beginTri;
		int tri;

		Art* triPart=0;
		tri = findNextTriangleFromVertex(part, triPart, i, corner, part->smls->ActTri-1);
		triangle=&triPart->triangle[tri];
		beginTri=triangle;

		while(1){

			int vecIndexA=triangle->ia;
			int vecIndexB=triangle->ib;
			int vecIndexC=triangle->ic;

			Art* partA=triangle->pa;
			Art* partB=triangle->pb;
			Art* partC=triangle->pc;

			int idxA = partA->fromCoord + vecIndexA;
			int idxB = partB->fromCoord + vecIndexB;
			int idxC = partC->fromCoord + vecIndexC;
			Vec3f vA = m_matrix[m_smi[idxA]] * g_viewMatrix * (g_m_coord[idxA] - partA->absolutePivotPoint);
			Vec3f vB = m_matrix[m_smi[idxB]] * g_viewMatrix * (g_m_coord[idxB] - partB->absolutePivotPoint);
			Vec3f vC = m_matrix[m_smi[idxC]] * g_viewMatrix * (g_m_coord[idxC] - partC->absolutePivotPoint);


			if(g_settings->backfaceIgnore){
				if(backFace(vA,vB,vC))	break; //if success exit loop 
			}
			else {
				break;
			}

			tri=findNextTriangleFromVertex(part, triPart, i, corner, tri);
			triangle=&triPart->triangle[tri];
			if(triangle!=beginTri) continue;
			else return false; //looped full circle
		}
		if(SenseTugGlobals::node == 0 || sr.z >= SenseTugGlobals::hitZ){

			SenseTugGlobals::pointer=sr;
			SenseTugGlobals::hitZ=sr.z;
			SenseTugGlobals::node=part;
			SenseTugGlobals::actTriPart=triPart;
			SenseTugGlobals::ActTri=tri;
			SenseTugGlobals::corner=corner;
			SenseTugGlobals::ActCtrlPnt=i;
			return true;
		}

	}
	return false;
}

void MultiSelect::selectFrontAndBackFaceTransformedVertices(Seamless* smls){
	smls->genDynamicMatricesAndPopulate();

	M_Art& partList = smls->partList;

	DXRender* render = smls->render;
	if(render==0)return;


	for(int k = 0; k < partList.length; k++){
		Art* part = partList[k];
		if(part->show == false){
			continue;
		}
		int coordLen = part->coord.length;
		for(int i = 0; i <coordLen ; i++){
			if(isInBox(m_matrix[m_smi[part->fromCoord + i]] * (g_m_coord[part->fromCoord + i] - part->absolutePivotPoint))){
				coordList += S_VertexIndex(part, i);
			}
		}
	}
}

Vec2f convertPointToScreen2d(Vec3f point);

void selectFrontFaceRenderedVertices2(M_Art& partList, const bool& backfaceIgnore, M_Bool& selectCoord){
	for(int k = 0; k < partList.length; k++){
		Art* part = partList[k];
		if(part->show == false){
			continue;
		}
		M_Triangle& triangle = part->triangle;
		for(int i = triangle.length; --i>=0;){
			S_Triangle& tri = triangle[i];

			Art* partA = tri.pa;
			Art* partB = tri.pb;
			Art* partC = tri.pc;

			if(backfaceIgnore){
				Vec3f coordA = m_matrix[m_smi[partA->fromCoord + tri.ia]] * g_viewMatrix * (g_m_coord[partA->fromCoord + tri.ia] - partA->absolutePivotPoint);
				Vec3f coordB = m_matrix[m_smi[partB->fromCoord + tri.ib]] * g_viewMatrix * (g_m_coord[partB->fromCoord + tri.ib] - partB->absolutePivotPoint);
				Vec3f coordC = m_matrix[m_smi[partC->fromCoord + tri.ic]] * g_viewMatrix * (g_m_coord[partC->fromCoord + tri.ic] - partC->absolutePivotPoint);

				if(!backFace(coordA, coordB, coordC)){
					continue;
				}
			}
			Vec2f va = convertPointToScreen2d(m_matrix[m_smi[partA->fromCoord + tri.ia]] * (g_m_coord[partA->fromCoord + tri.ia] - partA->absolutePivotPoint));
			Vec2f vb = convertPointToScreen2d(m_matrix[m_smi[partB->fromCoord + tri.ib]] * (g_m_coord[partB->fromCoord + tri.ib] - partB->absolutePivotPoint));
			Vec2f vc = convertPointToScreen2d(m_matrix[m_smi[partC->fromCoord + tri.ic]] * (g_m_coord[partC->fromCoord + tri.ic] - partC->absolutePivotPoint));

			Vec2f tri2d[3] = {va, vb, vc};

			int contigIdx = partA->fromCoord + tri.ia;
			if(selectCoord[contigIdx] == false && MultiSelect::isInBox(va)){
				selectCoord[contigIdx] = true;
				MultiSelect::coordList += S_VertexIndex(partA, tri.ia);
			}

			contigIdx = partB->fromCoord + tri.ib;
			if(selectCoord[contigIdx] == false && MultiSelect::isInBox(vb)){
				selectCoord[contigIdx] = true;
				MultiSelect::coordList += S_VertexIndex(partB, tri.ib);
			}

			contigIdx = partC->fromCoord + tri.ic;
			if(selectCoord[contigIdx] == false && MultiSelect::isInBox(vc)){
				selectCoord[contigIdx] = true;
				MultiSelect::coordList += S_VertexIndex(partC, tri.ic);
			}
		}
	}
}


void MultiSelect::selectFrontFaceRenderedVertices(Seamless* smls){
	M_Art& partList = smls->partList;
	M_Bool selectCoord;
	int selectCoordLen = 0;

	for(int k = 0; k < partList.length; k++){
		Art* part = partList[k];
		selectCoordLen += part->coord.length;
	}
	selectCoord.length = selectCoordLen;

	for(int k = coordList.length; --k >= 0;){
		selectCoord[coordList[k].p->fromCoord + coordList[k].i] = true;
	}

	DXRender* render = smls->render;
	if(render==0)return;

	smls->genDynamicMatricesAndPopulate();

	selectFrontFaceRenderedVertices2(partList, backfaceIgnore, selectCoord);
}


void CPMorpher::do_displayPolyFrame(const Art* part){
	M_Float& key = this->key.refValue();
	M_Vec3f& keyValue = this->keyValue.refValue();
	if(!key.length)return;

//	for(int k = controlPoint.length; --k >= 0;){
//	}
	int portionLen = cpIndex.length;

	float frac = frame;
	int i = 0;
	if(key.length < 2)goto here;
	for(; i<key.length-1; i++){
		if(key[i] > frac){
			i--;
			break;
		}
	}
	if(i < 0){
here:
		if(key.length > 0){
			//value_changed.sendOutRoute(keyValue[0]);
		}
		return;
	}
	if(i == key.length - 1){
		i = key.length - 2;
	}
	float diff = key[i + 1] - key[i];

	float interp = 1.0 / diff * (frac - key[i]);
	if(interp < 0){
		interp =- interp;
	}

	int coordLen = part->coord.length;
	int fromCoord = part->fromCoord;

	for(int j = 0; j < portionLen; j++){
		int h = i * portionLen + j;
		Vec3f result = keyValue[h] + (keyValue[(i + 1) * portionLen + j] - keyValue[h]) * interp;
		int idx = cpIndex[j];
		if(idx < coordLen){
			g_m_coord[fromCoord + idx] += result;
		}
	}
}
void Seamless::genDynamicMatricesAndPopulate(){
	g_m_coord.length = setCoordFroms();
	genDynamicMatricesForSelectedPolyCoords();
	populate_g_m_coord();
}

void Seamless::populate_g_m_coord(){
	int q = 0;
	for(int k = 0; k < partList.length; k++){
		const M_Vec3f& coord = partList[k]->coord;
		for(int i = 0, len = coord.length; i < len;){
			g_m_coord[q++] = coord[i++];
		}
	}
	for(int k = cpMorpherPart.length; --k >= 0;){
		CPMorpherPart& cpmp = cpMorpherPart[k];
		cpmp.morpher->do_displayPolyFrame(cpmp.part);
	}
}

void Seamless::genDynamicMatricesForSelectedPolyCoords(){
	m_matrix.length = 0;
	m_smi.length = 0;
	Matrix parentMatrix = scaleMatrix(scale) * rotationMatrix(rotation) * translationMatrix(translation);

	int smi = 0;
	M_SArt& skeleton = this->skeleton.refValue();
	for(int i = 0, smi = 0; i < skeleton.length; i++){
		skeleton[i]->genDynamicTransforms(m_matrix, m_smi, parentMatrix, smi);
	}
}
//bool checkInList(Art* part, int idx){
//	for(int i = MultiSelect::coordList.length; --i>=0;){
//		if(MultiSelect::coordList[i].p == part && MultiSelect::coordList[i].i == idx){
//			return true;
//		}
//	}
//	return false;
//}
void Seamless::senseAllPartsVertex(){
	genDynamicMatricesAndPopulate();

	const bool perspective=g_settings->perspective;
	NodePtr before = g_selNode.getRaw();
	Anim* anim = Window::animMode();



	//check selected vertices first
	if(MultiSelect::coordList.length){
		if(MultiSelect::coordList[0].p->smls == this){
			for(int i = MultiSelect::coordList.length; --i >= 0;){
				Art* part = MultiSelect::coordList[i].p;
				int idx = MultiSelect::coordList[i].i;
				bool remBackfaceIgnore = g_settings->backfaceIgnore;
				g_settings->backfaceIgnore = false;
				if(checkPointInVertex(part, perspective, idx)){
					MultiSelect::clickedIndex = i;
					part->index[0] = idx;
					ActCtrlPnt = idx;
					g_settings->backfaceIgnore = remBackfaceIgnore;

					MultiSelect::beginDragPos.length = MultiSelect::coordList.length;
					for(int i = MultiSelect::coordList.length; --i >= 0;){
						Art* part = MultiSelect::coordList[i].p;
						int idx = MultiSelect::coordList[i].i;
						MultiSelect::beginDragPos[i] = g_m_coord[part->fromCoord + idx];
					}
					if(anim && g_allowPolyMorphCapture){
						g_dragMorphFlag =  true;
					}
					return;
				}
				g_settings->backfaceIgnore = remBackfaceIgnore;
			}
		}
	}
	//before anything else sense the 3 vertices in selected triangle
	//bool clickedSelectedTriangleCoord = false;
	if(anim == 0){
		if(Art* part = dynamic_cast<Art*>(before.p)){
			if(part->smls == this && part->show && anim == 0){
				Matrix partMatrix = part->matrix * g_viewMatrix;
				if(part->show && part->coord.refValue().length){
					if(actTriPart){
						M_Triangle& triangle = actTriPart->triangle;
						for(int i=ActTri*3+ActTriCorner, j=3; --j>=0;){//check other 2 corners in tri first
							Art* prt = triangle.refPart(i);
							int idx = triangle.refIndex(i);
							if(checkPointInVertex(prt, perspective, idx)){
								SenseTugGlobals::ActTri=ActTri;
								SenseTugGlobals::actTriPart=actTriPart;
								SenseTugGlobals::corner=i%3;

								MultiSelect::coordList.length = 0;
								MultiSelect::beginDragPos.length = 0;
								MultiSelect::coordList += S_VertexIndex(prt, idx);
								MultiSelect::beginDragPos+=prt->coord[idx];
								ActCtrlPnt = idx;
								MultiSelect::clickedIndex = 0;
								prt->index[0] = idx;
								return;
							}
							i=ActTri*3+(i+1)%3;
						}
					}
				}
			}
		}
	}
	//if(clickedSelectedTriangleCoord){
	//	if(MultiSelect::coordList.length == 1){
	//		MultiSelect::coordList.length = 0;
	//		MultiSelect::beginDragPos.length = 0;
	//	}
	//	return;
	//}

	//check NON selected vertices next
	bool clear = true; //flag to indicate that we must clear all the selected vertices if we hit a vertex from here on


	for(int k = 0; k < partList.length; k++){

		Art* part=partList[k];
		M_Triangle& triangle = part->triangle;


		if(part->show){
			for(int i=triangle.getTriangleLength(); --i>=0; ){
				Art* part = triangle.refPart(i);
				if (checkPointInVertex(part, perspective, triangle.refIndex(i))){
					if(clear){
						MultiSelect::beginDragPos.length = MultiSelect::coordList.length = 0;
						clear = false;
					}
					part->index[0] = triangle.refIndex(i);
					MultiSelect::coordList += S_VertexIndex(part, part->index[0]);
					MultiSelect::clickedIndex = 0;
					ActCtrlPnt = part->index[0];
					MultiSelect::checkToReset(part);

					//if(anim == 0 && MultiSelect::coordList.length == 1){
					//	MultiSelect::coordList.length = 0;
					//	MultiSelect::beginDragPos.length = 0;
					//}

//						MultiSelect::backUpCoords(part);

					MultiSelect::beginDragPos.length = MultiSelect::coordList.length;
					for(int i = MultiSelect::coordList.length; --i >= 0;){
						Art* part = MultiSelect::coordList[i].p;
						int idx = MultiSelect::coordList[i].i;
						MultiSelect::beginDragPos[i] = g_m_coord[part->fromCoord + idx];
					}

					if(anim && g_allowPolyMorphCapture){
						g_dragMorphFlag =  true;
					}

					MultiSelect::calcPivotPoint();
					return;

				}
			}
		}
	}
}

//void fillWithDraggedMorphedVertices(){//beginDragPos was use for the destination not the sourse
//	for(int i = MultiSelect::coordList.length; --i >= 0;){
//		Art* part = MultiSelect::coordList[i].p;
//		int idx = MultiSelect::coordList[i].i;
//		g_m_coord[part->fromCoord + idx] = MultiSelect::beginDragPos[i];
//	}
//}

void NurbsTrack::KeyValue::operator=(const M_Vec3f& par){
	refValue() = par;
	home.houseKeep();
	if(Anim* anim = Window::animMode()){
		anim->update = true;
		anim->renderNodeFirst();
		anim->update = false;
	}
}

